var l9=Object.defineProperty;var u9=(e,t,n)=>t in e?l9(e,t,{enumerable:!0,configurable:!0,writable:!0,value:n}):e[t]=n;var Xr=(e,t,n)=>u9(e,typeof t!="symbol"?t+"":t,n);function c9(e,t){for(var n=0;n<t.length;n++){const r=t[n];if(typeof r!="string"&&!Array.isArray(r)){for(const a in r)if(a!=="default"&&!(a in e)){const o=Object.getOwnPropertyDescriptor(r,a);o&&Object.defineProperty(e,a,o.get?o:{enumerable:!0,get:()=>r[a]})}}}return Object.freeze(Object.defineProperty(e,Symbol.toStringTag,{value:"Module"}))}(function(){const t=document.createElement("link").relList;if(t&&t.supports&&t.supports("modulepreload"))return;for(const a of document.querySelectorAll('link[rel="modulepreload"]'))r(a);new MutationObserver(a=>{for(const o of a)if(o.type==="childList")for(const s of o.addedNodes)s.tagName==="LINK"&&s.rel==="modulepreload"&&r(s)}).observe(document,{childList:!0,subtree:!0});function n(a){const o={};return a.integrity&&(o.integrity=a.integrity),a.referrerPolicy&&(o.referrerPolicy=a.referrerPolicy),a.crossOrigin==="use-credentials"?o.credentials="include":a.crossOrigin==="anonymous"?o.credentials="omit":o.credentials="same-origin",o}function r(a){if(a.ep)return;a.ep=!0;const o=n(a);fetch(a.href,o)}})();var df=typeof globalThis<"u"?globalThis:typeof window<"u"?window:typeof global<"u"?global:typeof self<"u"?self:{};function un(e){return e&&e.__esModule&&Object.prototype.hasOwnProperty.call(e,"default")?e.default:e}function d9(e){if(e.__esModule)return e;var t=e.default;if(typeof t=="function"){var n=function r(){return this instanceof r?Reflect.construct(t,arguments,this.constructor):t.apply(this,arguments)};n.prototype=t.prototype}else n={};return Object.defineProperty(n,"__esModule",{value:!0}),Object.keys(e).forEach(function(r){var a=Object.getOwnPropertyDescriptor(e,r);Object.defineProperty(n,r,a.get?a:{enumerable:!0,get:function(){return e[r]}})}),n}var $h={exports:{}},Xl={};/**
 * @license React
 * react-jsx-runtime.production.js
 *
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */var q_;function f9(){if(q_)return Xl;q_=1;var e=Symbol.for("react.transitional.element"),t=Symbol.for("react.fragment");function n(r,a,o){var s=null;if(o!==void 0&&(s=""+o),a.key!==void 0&&(s=""+a.key),"key"in a){o={};for(var u in a)u!=="key"&&(o[u]=a[u])}else o=a;return a=o.ref,{$$typeof:e,type:r,key:s,ref:a!==void 0?a:null,props:o}}return Xl.Fragment=t,Xl.jsx=n,Xl.jsxs=n,Xl}var V_;function p9(){return V_||(V_=1,$h.exports=f9()),$h.exports}var w=p9(),qh={exports:{}},it={};/**
 * @license React
 * react.production.js
 *
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */var W_;function g9(){if(W_)return it;W_=1;var e=Symbol.for("react.transitional.element"),t=Symbol.for("react.portal"),n=Symbol.for("react.fragment"),r=Symbol.for("react.strict_mode"),a=Symbol.for("react.profiler"),o=Symbol.for("react.consumer"),s=Symbol.for("react.context"),u=Symbol.for("react.forward_ref"),c=Symbol.for("react.suspense"),d=Symbol.for("react.memo"),p=Symbol.for("react.lazy"),g=Symbol.iterator;function m(D){return D===null||typeof D!="object"?null:(D=g&&D[g]||D["@@iterator"],typeof D=="function"?D:null)}var b={isMounted:function(){return!1},enqueueForceUpdate:function(){},enqueueReplaceState:function(){},enqueueSetState:function(){}},y=Object.assign,S={};function x(D,V,B){this.props=D,this.context=V,this.refs=S,this.updater=B||b}x.prototype.isReactComponent={},x.prototype.setState=function(D,V){if(typeof D!="object"&&typeof D!="function"&&D!=null)throw Error("takes an object of state variables to update or a function which returns an object of state variables.");this.updater.enqueueSetState(this,D,V,"setState")},x.prototype.forceUpdate=function(D){this.updater.enqueueForceUpdate(this,D,"forceUpdate")};function R(){}R.prototype=x.prototype;function k(D,V,B){this.props=D,this.context=V,this.refs=S,this.updater=B||b}var A=k.prototype=new R;A.constructor=k,y(A,x.prototype),A.isPureReactComponent=!0;var C=Array.isArray,N={H:null,A:null,T:null,S:null},_=Object.prototype.hasOwnProperty;function O(D,V,B,M,W,Q){return B=Q.ref,{$$typeof:e,type:D,key:V,ref:B!==void 0?B:null,props:Q}}function F(D,V){return O(D.type,V,void 0,void 0,void 0,D.props)}function L(D){return typeof D=="object"&&D!==null&&D.$$typeof===e}function I(D){var V={"=":"=0",":":"=2"};return"$"+D.replace(/[=:]/g,function(B){return V[B]})}var H=/\/+/g;function $(D,V){return typeof D=="object"&&D!==null&&D.key!=null?I(""+D.key):V.toString(36)}function U(){}function Y(D){switch(D.status){case"fulfilled":return D.value;case"rejected":throw D.reason;default:switch(typeof D.status=="string"?D.then(U,U):(D.status="pending",D.then(function(V){D.status==="pending"&&(D.status="fulfilled",D.value=V)},function(V){D.status==="pending"&&(D.status="rejected",D.reason=V)})),D.status){case"fulfilled":return D.value;case"rejected":throw D.reason}}throw D}function Z(D,V,B,M,W){var Q=typeof D;(Q==="undefined"||Q==="boolean")&&(D=null);var oe=!1;if(D===null)oe=!0;else switch(Q){case"bigint":case"string":case"number":oe=!0;break;case"object":switch(D.$$typeof){case e:case t:oe=!0;break;case p:return oe=D._init,Z(oe(D._payload),V,B,M,W)}}if(oe)return W=W(D),oe=M===""?"."+$(D,0):M,C(W)?(B="",oe!=null&&(B=oe.replace(H,"$&/")+"/"),Z(W,V,B,"",function(Se){return Se})):W!=null&&(L(W)&&(W=F(W,B+(W.key==null||D&&D.key===W.key?"":(""+W.key).replace(H,"$&/")+"/")+oe)),V.push(W)),1;oe=0;var re=M===""?".":M+":";if(C(D))for(var ie=0;ie<D.length;ie++)M=D[ie],Q=re+$(M,ie),oe+=Z(M,V,B,Q,W);else if(ie=m(D),typeof ie=="function")for(D=ie.call(D),ie=0;!(M=D.next()).done;)M=M.value,Q=re+$(M,ie++),oe+=Z(M,V,B,Q,W);else if(Q==="object"){if(typeof D.then=="function")return Z(Y(D),V,B,M,W);throw V=String(D),Error("Objects are not valid as a React child (found: "+(V==="[object Object]"?"object with keys {"+Object.keys(D).join(", ")+"}":V)+"). If you meant to render a collection of children, use an array instead.")}return oe}function j(D,V,B){if(D==null)return D;var M=[],W=0;return Z(D,M,"","",function(Q){return V.call(B,Q,W++)}),M}function G(D){if(D._status===-1){var V=D._result;V=V(),V.then(function(B){(D._status===0||D._status===-1)&&(D._status=1,D._result=B)},function(B){(D._status===0||D._status===-1)&&(D._status=2,D._result=B)}),D._status===-1&&(D._status=0,D._result=V)}if(D._status===1)return D._result.default;throw D._result}var z=typeof reportError=="function"?reportError:function(D){if(typeof window=="object"&&typeof window.ErrorEvent=="function"){var V=new window.ErrorEvent("error",{bubbles:!0,cancelable:!0,message:typeof D=="object"&&D!==null&&typeof D.message=="string"?String(D.message):String(D),error:D});if(!window.dispatchEvent(V))return}else if(typeof process=="object"&&typeof process.emit=="function"){process.emit("uncaughtException",D);return}console.error(D)};function K(){}return it.Children={map:j,forEach:function(D,V,B){j(D,function(){V.apply(this,arguments)},B)},count:function(D){var V=0;return j(D,function(){V++}),V},toArray:function(D){return j(D,function(V){return V})||[]},only:function(D){if(!L(D))throw Error("React.Children.only expected to receive a single React element child.");return D}},it.Component=x,it.Fragment=n,it.Profiler=a,it.PureComponent=k,it.StrictMode=r,it.Suspense=c,it.__CLIENT_INTERNALS_DO_NOT_USE_OR_WARN_USERS_THEY_CANNOT_UPGRADE=N,it.act=function(){throw Error("act(...) is not supported in production builds of React.")},it.cache=function(D){return function(){return D.apply(null,arguments)}},it.cloneElement=function(D,V,B){if(D==null)throw Error("The argument must be a React element, but you passed "+D+".");var M=y({},D.props),W=D.key,Q=void 0;if(V!=null)for(oe in V.ref!==void 0&&(Q=void 0),V.key!==void 0&&(W=""+V.key),V)!_.call(V,oe)||oe==="key"||oe==="__self"||oe==="__source"||oe==="ref"&&V.ref===void 0||(M[oe]=V[oe]);var oe=arguments.length-2;if(oe===1)M.children=B;else if(1<oe){for(var re=Array(oe),ie=0;ie<oe;ie++)re[ie]=arguments[ie+2];M.children=re}return O(D.type,W,void 0,void 0,Q,M)},it.createContext=function(D){return D={$$typeof:s,_currentValue:D,_currentValue2:D,_threadCount:0,Provider:null,Consumer:null},D.Provider=D,D.Consumer={$$typeof:o,_context:D},D},it.createElement=function(D,V,B){var M,W={},Q=null;if(V!=null)for(M in V.key!==void 0&&(Q=""+V.key),V)_.call(V,M)&&M!=="key"&&M!=="__self"&&M!=="__source"&&(W[M]=V[M]);var oe=arguments.length-2;if(oe===1)W.children=B;else if(1<oe){for(var re=Array(oe),ie=0;ie<oe;ie++)re[ie]=arguments[ie+2];W.children=re}if(D&&D.defaultProps)for(M in oe=D.defaultProps,oe)W[M]===void 0&&(W[M]=oe[M]);return O(D,Q,void 0,void 0,null,W)},it.createRef=function(){return{current:null}},it.forwardRef=function(D){return{$$typeof:u,render:D}},it.isValidElement=L,it.lazy=function(D){return{$$typeof:p,_payload:{_status:-1,_result:D},_init:G}},it.memo=function(D,V){return{$$typeof:d,type:D,compare:V===void 0?null:V}},it.startTransition=function(D){var V=N.T,B={};N.T=B;try{var M=D(),W=N.S;W!==null&&W(B,M),typeof M=="object"&&M!==null&&typeof M.then=="function"&&M.then(K,z)}catch(Q){z(Q)}finally{N.T=V}},it.unstable_useCacheRefresh=function(){return N.H.useCacheRefresh()},it.use=function(D){return N.H.use(D)},it.useActionState=function(D,V,B){return N.H.useActionState(D,V,B)},it.useCallback=function(D,V){return N.H.useCallback(D,V)},it.useContext=function(D){return N.H.useContext(D)},it.useDebugValue=function(){},it.useDeferredValue=function(D,V){return N.H.useDeferredValue(D,V)},it.useEffect=function(D,V){return N.H.useEffect(D,V)},it.useId=function(){return N.H.useId()},it.useImperativeHandle=function(D,V,B){return N.H.useImperativeHandle(D,V,B)},it.useInsertionEffect=function(D,V){return N.H.useInsertionEffect(D,V)},it.useLayoutEffect=function(D,V){return N.H.useLayoutEffect(D,V)},it.useMemo=function(D,V){return N.H.useMemo(D,V)},it.useOptimistic=function(D,V){return N.H.useOptimistic(D,V)},it.useReducer=function(D,V,B){return N.H.useReducer(D,V,B)},it.useRef=function(D){return N.H.useRef(D)},it.useState=function(D){return N.H.useState(D)},it.useSyncExternalStore=function(D,V,B){return N.H.useSyncExternalStore(D,V,B)},it.useTransition=function(){return N.H.useTransition()},it.version="19.0.0",it}var Y_;function $f(){return Y_||(Y_=1,qh.exports=g9()),qh.exports}var E=$f();const ye=un(E),h9=c9({__proto__:null,default:ye},[E]);var Vh={exports:{}},Zl={},Wh={exports:{}},Yh={};/**
 * @license React
 * scheduler.production.js
 *
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */var K_;function m9(){return K_||(K_=1,function(e){function t(j,G){var z=j.length;j.push(G);e:for(;0<z;){var K=z-1>>>1,D=j[K];if(0<a(D,G))j[K]=G,j[z]=D,z=K;else break e}}function n(j){return j.length===0?null:j[0]}function r(j){if(j.length===0)return null;var G=j[0],z=j.pop();if(z!==G){j[0]=z;e:for(var K=0,D=j.length,V=D>>>1;K<V;){var B=2*(K+1)-1,M=j[B],W=B+1,Q=j[W];if(0>a(M,z))W<D&&0>a(Q,M)?(j[K]=Q,j[W]=z,K=W):(j[K]=M,j[B]=z,K=B);else if(W<D&&0>a(Q,z))j[K]=Q,j[W]=z,K=W;else break e}}return G}function a(j,G){var z=j.sortIndex-G.sortIndex;return z!==0?z:j.id-G.id}if(e.unstable_now=void 0,typeof performance=="object"&&typeof performance.now=="function"){var o=performance;e.unstable_now=function(){return o.now()}}else{var s=Date,u=s.now();e.unstable_now=function(){return s.now()-u}}var c=[],d=[],p=1,g=null,m=3,b=!1,y=!1,S=!1,x=typeof setTimeout=="function"?setTimeout:null,R=typeof clearTimeout=="function"?clearTimeout:null,k=typeof setImmediate<"u"?setImmediate:null;function A(j){for(var G=n(d);G!==null;){if(G.callback===null)r(d);else if(G.startTime<=j)r(d),G.sortIndex=G.expirationTime,t(c,G);else break;G=n(d)}}function C(j){if(S=!1,A(j),!y)if(n(c)!==null)y=!0,Y();else{var G=n(d);G!==null&&Z(C,G.startTime-j)}}var N=!1,_=-1,O=5,F=-1;function L(){return!(e.unstable_now()-F<O)}function I(){if(N){var j=e.unstable_now();F=j;var G=!0;try{e:{y=!1,S&&(S=!1,R(_),_=-1),b=!0;var z=m;try{t:{for(A(j),g=n(c);g!==null&&!(g.expirationTime>j&&L());){var K=g.callback;if(typeof K=="function"){g.callback=null,m=g.priorityLevel;var D=K(g.expirationTime<=j);if(j=e.unstable_now(),typeof D=="function"){g.callback=D,A(j),G=!0;break t}g===n(c)&&r(c),A(j)}else r(c);g=n(c)}if(g!==null)G=!0;else{var V=n(d);V!==null&&Z(C,V.startTime-j),G=!1}}break e}finally{g=null,m=z,b=!1}G=void 0}}finally{G?H():N=!1}}}var H;if(typeof k=="function")H=function(){k(I)};else if(typeof MessageChannel<"u"){var $=new MessageChannel,U=$.port2;$.port1.onmessage=I,H=function(){U.postMessage(null)}}else H=function(){x(I,0)};function Y(){N||(N=!0,H())}function Z(j,G){_=x(function(){j(e.unstable_now())},G)}e.unstable_IdlePriority=5,e.unstable_ImmediatePriority=1,e.unstable_LowPriority=4,e.unstable_NormalPriority=3,e.unstable_Profiling=null,e.unstable_UserBlockingPriority=2,e.unstable_cancelCallback=function(j){j.callback=null},e.unstable_continueExecution=function(){y||b||(y=!0,Y())},e.unstable_forceFrameRate=function(j){0>j||125<j?console.error("forceFrameRate takes a positive int between 0 and 125, forcing frame rates higher than 125 fps is not supported"):O=0<j?Math.floor(1e3/j):5},e.unstable_getCurrentPriorityLevel=function(){return m},e.unstable_getFirstCallbackNode=function(){return n(c)},e.unstable_next=function(j){switch(m){case 1:case 2:case 3:var G=3;break;default:G=m}var z=m;m=G;try{return j()}finally{m=z}},e.unstable_pauseExecution=function(){},e.unstable_requestPaint=function(){},e.unstable_runWithPriority=function(j,G){switch(j){case 1:case 2:case 3:case 4:case 5:break;default:j=3}var z=m;m=j;try{return G()}finally{m=z}},e.unstable_scheduleCallback=function(j,G,z){var K=e.unstable_now();switch(typeof z=="object"&&z!==null?(z=z.delay,z=typeof z=="number"&&0<z?K+z:K):z=K,j){case 1:var D=-1;break;case 2:D=250;break;case 5:D=1073741823;break;case 4:D=1e4;break;default:D=5e3}return D=z+D,j={id:p++,callback:G,priorityLevel:j,startTime:z,expirationTime:D,sortIndex:-1},z>K?(j.sortIndex=z,t(d,j),n(c)===null&&j===n(d)&&(S?(R(_),_=-1):S=!0,Z(C,z-K))):(j.sortIndex=D,t(c,j),y||b||(y=!0,Y())),j},e.unstable_shouldYield=L,e.unstable_wrapCallback=function(j){var G=m;return function(){var z=m;m=G;try{return j.apply(this,arguments)}finally{m=z}}}}(Yh)),Yh}var X_;function b9(){return X_||(X_=1,Wh.exports=m9()),Wh.exports}var Kh={exports:{}},wn={};/**
 * @license React
 * react-dom.production.js
 *
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */var Z_;function y9(){if(Z_)return wn;Z_=1;var e=$f();function t(c){var d="https://react.dev/errors/"+c;if(1<arguments.length){d+="?args[]="+encodeURIComponent(arguments[1]);for(var p=2;p<arguments.length;p++)d+="&args[]="+encodeURIComponent(arguments[p])}return"Minified React error #"+c+"; visit "+d+" for the full message or use the non-minified dev environment for full errors and additional helpful warnings."}function n(){}var r={d:{f:n,r:function(){throw Error(t(522))},D:n,C:n,L:n,m:n,X:n,S:n,M:n},p:0,findDOMNode:null},a=Symbol.for("react.portal");function o(c,d,p){var g=3<arguments.length&&arguments[3]!==void 0?arguments[3]:null;return{$$typeof:a,key:g==null?null:""+g,children:c,containerInfo:d,implementation:p}}var s=e.__CLIENT_INTERNALS_DO_NOT_USE_OR_WARN_USERS_THEY_CANNOT_UPGRADE;function u(c,d){if(c==="font")return"";if(typeof d=="string")return d==="use-credentials"?d:""}return wn.__DOM_INTERNALS_DO_NOT_USE_OR_WARN_USERS_THEY_CANNOT_UPGRADE=r,wn.createPortal=function(c,d){var p=2<arguments.length&&arguments[2]!==void 0?arguments[2]:null;if(!d||d.nodeType!==1&&d.nodeType!==9&&d.nodeType!==11)throw Error(t(299));return o(c,d,null,p)},wn.flushSync=function(c){var d=s.T,p=r.p;try{if(s.T=null,r.p=2,c)return c()}finally{s.T=d,r.p=p,r.d.f()}},wn.preconnect=function(c,d){typeof c=="string"&&(d?(d=d.crossOrigin,d=typeof d=="string"?d==="use-credentials"?d:"":void 0):d=null,r.d.C(c,d))},wn.prefetchDNS=function(c){typeof c=="string"&&r.d.D(c)},wn.preinit=function(c,d){if(typeof c=="string"&&d&&typeof d.as=="string"){var p=d.as,g=u(p,d.crossOrigin),m=typeof d.integrity=="string"?d.integrity:void 0,b=typeof d.fetchPriority=="string"?d.fetchPriority:void 0;p==="style"?r.d.S(c,typeof d.precedence=="string"?d.precedence:void 0,{crossOrigin:g,integrity:m,fetchPriority:b}):p==="script"&&r.d.X(c,{crossOrigin:g,integrity:m,fetchPriority:b,nonce:typeof d.nonce=="string"?d.nonce:void 0})}},wn.preinitModule=function(c,d){if(typeof c=="string")if(typeof d=="object"&&d!==null){if(d.as==null||d.as==="script"){var p=u(d.as,d.crossOrigin);r.d.M(c,{crossOrigin:p,integrity:typeof d.integrity=="string"?d.integrity:void 0,nonce:typeof d.nonce=="string"?d.nonce:void 0})}}else d==null&&r.d.M(c)},wn.preload=function(c,d){if(typeof c=="string"&&typeof d=="object"&&d!==null&&typeof d.as=="string"){var p=d.as,g=u(p,d.crossOrigin);r.d.L(c,p,{crossOrigin:g,integrity:typeof d.integrity=="string"?d.integrity:void 0,nonce:typeof d.nonce=="string"?d.nonce:void 0,type:typeof d.type=="string"?d.type:void 0,fetchPriority:typeof d.fetchPriority=="string"?d.fetchPriority:void 0,referrerPolicy:typeof d.referrerPolicy=="string"?d.referrerPolicy:void 0,imageSrcSet:typeof d.imageSrcSet=="string"?d.imageSrcSet:void 0,imageSizes:typeof d.imageSizes=="string"?d.imageSizes:void 0,media:typeof d.media=="string"?d.media:void 0})}},wn.preloadModule=function(c,d){if(typeof c=="string")if(d){var p=u(d.as,d.crossOrigin);r.d.m(c,{as:typeof d.as=="string"&&d.as!=="script"?d.as:void 0,crossOrigin:p,integrity:typeof d.integrity=="string"?d.integrity:void 0})}else r.d.m(c)},wn.requestFormReset=function(c){r.d.r(c)},wn.unstable_batchedUpdates=function(c,d){return c(d)},wn.useFormState=function(c,d,p){return s.H.useFormState(c,d,p)},wn.useFormStatus=function(){return s.H.useHostTransitionStatus()},wn.version="19.0.0",wn}var Q_;function Pz(){if(Q_)return Kh.exports;Q_=1;function e(){if(!(typeof __REACT_DEVTOOLS_GLOBAL_HOOK__>"u"||typeof __REACT_DEVTOOLS_GLOBAL_HOOK__.checkDCE!="function"))try{__REACT_DEVTOOLS_GLOBAL_HOOK__.checkDCE(e)}catch(t){console.error(t)}}return e(),Kh.exports=y9(),Kh.exports}/**
 * @license React
 * react-dom-client.production.js
 *
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */var J_;function v9(){if(J_)return Zl;J_=1;var e=b9(),t=$f(),n=Pz();function r(i){var l="https://react.dev/errors/"+i;if(1<arguments.length){l+="?args[]="+encodeURIComponent(arguments[1]);for(var f=2;f<arguments.length;f++)l+="&args[]="+encodeURIComponent(arguments[f])}return"Minified React error #"+i+"; visit "+l+" for the full message or use the non-minified dev environment for full errors and additional helpful warnings."}function a(i){return!(!i||i.nodeType!==1&&i.nodeType!==9&&i.nodeType!==11)}var o=Symbol.for("react.element"),s=Symbol.for("react.transitional.element"),u=Symbol.for("react.portal"),c=Symbol.for("react.fragment"),d=Symbol.for("react.strict_mode"),p=Symbol.for("react.profiler"),g=Symbol.for("react.provider"),m=Symbol.for("react.consumer"),b=Symbol.for("react.context"),y=Symbol.for("react.forward_ref"),S=Symbol.for("react.suspense"),x=Symbol.for("react.suspense_list"),R=Symbol.for("react.memo"),k=Symbol.for("react.lazy"),A=Symbol.for("react.offscreen"),C=Symbol.for("react.memo_cache_sentinel"),N=Symbol.iterator;function _(i){return i===null||typeof i!="object"?null:(i=N&&i[N]||i["@@iterator"],typeof i=="function"?i:null)}var O=Symbol.for("react.client.reference");function F(i){if(i==null)return null;if(typeof i=="function")return i.$$typeof===O?null:i.displayName||i.name||null;if(typeof i=="string")return i;switch(i){case c:return"Fragment";case u:return"Portal";case p:return"Profiler";case d:return"StrictMode";case S:return"Suspense";case x:return"SuspenseList"}if(typeof i=="object")switch(i.$$typeof){case b:return(i.displayName||"Context")+".Provider";case m:return(i._context.displayName||"Context")+".Consumer";case y:var l=i.render;return i=i.displayName,i||(i=l.displayName||l.name||"",i=i!==""?"ForwardRef("+i+")":"ForwardRef"),i;case R:return l=i.displayName||null,l!==null?l:F(i.type)||"Memo";case k:l=i._payload,i=i._init;try{return F(i(l))}catch{}}return null}var L=t.__CLIENT_INTERNALS_DO_NOT_USE_OR_WARN_USERS_THEY_CANNOT_UPGRADE,I=Object.assign,H,$;function U(i){if(H===void 0)try{throw Error()}catch(f){var l=f.stack.trim().match(/\n( *(at )?)/);H=l&&l[1]||"",$=-1<f.stack.indexOf(`
    at`)?" (<anonymous>)":-1<f.stack.indexOf("@")?"@unknown:0:0":""}return`
`+H+i+$}var Y=!1;function Z(i,l){if(!i||Y)return"";Y=!0;var f=Error.prepareStackTrace;Error.prepareStackTrace=void 0;try{var h={DetermineComponentFrameRoot:function(){try{if(l){var Ae=function(){throw Error()};if(Object.defineProperty(Ae.prototype,"props",{set:function(){throw Error()}}),typeof Reflect=="object"&&Reflect.construct){try{Reflect.construct(Ae,[])}catch(ge){var de=ge}Reflect.construct(i,[],Ae)}else{try{Ae.call()}catch(ge){de=ge}i.call(Ae.prototype)}}else{try{throw Error()}catch(ge){de=ge}(Ae=i())&&typeof Ae.catch=="function"&&Ae.catch(function(){})}}catch(ge){if(ge&&de&&typeof ge.stack=="string")return[ge.stack,de.stack]}return[null,null]}};h.DetermineComponentFrameRoot.displayName="DetermineComponentFrameRoot";var v=Object.getOwnPropertyDescriptor(h.DetermineComponentFrameRoot,"name");v&&v.configurable&&Object.defineProperty(h.DetermineComponentFrameRoot,"name",{value:"DetermineComponentFrameRoot"});var T=h.DetermineComponentFrameRoot(),P=T[0],q=T[1];if(P&&q){var X=P.split(`
`),ne=q.split(`
`);for(v=h=0;h<X.length&&!X[h].includes("DetermineComponentFrameRoot");)h++;for(;v<ne.length&&!ne[v].includes("DetermineComponentFrameRoot");)v++;if(h===X.length||v===ne.length)for(h=X.length-1,v=ne.length-1;1<=h&&0<=v&&X[h]!==ne[v];)v--;for(;1<=h&&0<=v;h--,v--)if(X[h]!==ne[v]){if(h!==1||v!==1)do if(h--,v--,0>v||X[h]!==ne[v]){var be=`
`+X[h].replace(" at new "," at ");return i.displayName&&be.includes("<anonymous>")&&(be=be.replace("<anonymous>",i.displayName)),be}while(1<=h&&0<=v);break}}}finally{Y=!1,Error.prepareStackTrace=f}return(f=i?i.displayName||i.name:"")?U(f):""}function j(i){switch(i.tag){case 26:case 27:case 5:return U(i.type);case 16:return U("Lazy");case 13:return U("Suspense");case 19:return U("SuspenseList");case 0:case 15:return i=Z(i.type,!1),i;case 11:return i=Z(i.type.render,!1),i;case 1:return i=Z(i.type,!0),i;default:return""}}function G(i){try{var l="";do l+=j(i),i=i.return;while(i);return l}catch(f){return`
Error generating stack: `+f.message+`
`+f.stack}}function z(i){var l=i,f=i;if(i.alternate)for(;l.return;)l=l.return;else{i=l;do l=i,l.flags&4098&&(f=l.return),i=l.return;while(i)}return l.tag===3?f:null}function K(i){if(i.tag===13){var l=i.memoizedState;if(l===null&&(i=i.alternate,i!==null&&(l=i.memoizedState)),l!==null)return l.dehydrated}return null}function D(i){if(z(i)!==i)throw Error(r(188))}function V(i){var l=i.alternate;if(!l){if(l=z(i),l===null)throw Error(r(188));return l!==i?null:i}for(var f=i,h=l;;){var v=f.return;if(v===null)break;var T=v.alternate;if(T===null){if(h=v.return,h!==null){f=h;continue}break}if(v.child===T.child){for(T=v.child;T;){if(T===f)return D(v),i;if(T===h)return D(v),l;T=T.sibling}throw Error(r(188))}if(f.return!==h.return)f=v,h=T;else{for(var P=!1,q=v.child;q;){if(q===f){P=!0,f=v,h=T;break}if(q===h){P=!0,h=v,f=T;break}q=q.sibling}if(!P){for(q=T.child;q;){if(q===f){P=!0,f=T,h=v;break}if(q===h){P=!0,h=T,f=v;break}q=q.sibling}if(!P)throw Error(r(189))}}if(f.alternate!==h)throw Error(r(190))}if(f.tag!==3)throw Error(r(188));return f.stateNode.current===f?i:l}function B(i){var l=i.tag;if(l===5||l===26||l===27||l===6)return i;for(i=i.child;i!==null;){if(l=B(i),l!==null)return l;i=i.sibling}return null}var M=Array.isArray,W=n.__DOM_INTERNALS_DO_NOT_USE_OR_WARN_USERS_THEY_CANNOT_UPGRADE,Q={pending:!1,data:null,method:null,action:null},oe=[],re=-1;function ie(i){return{current:i}}function Se(i){0>re||(i.current=oe[re],oe[re]=null,re--)}function ae(i,l){re++,oe[re]=i.current,i.current=l}var ve=ie(null),xe=ie(null),Ie=ie(null),Ce=ie(null);function ke(i,l){switch(ae(Ie,l),ae(xe,i),ae(ve,null),i=l.nodeType,i){case 9:case 11:l=(l=l.documentElement)&&(l=l.namespaceURI)?S_(l):0;break;default:if(i=i===8?l.parentNode:l,l=i.tagName,i=i.namespaceURI)i=S_(i),l=E_(i,l);else switch(l){case"svg":l=1;break;case"math":l=2;break;default:l=0}}Se(ve),ae(ve,l)}function J(){Se(ve),Se(xe),Se(Ie)}function fe(i){i.memoizedState!==null&&ae(Ce,i);var l=ve.current,f=E_(l,i.type);l!==f&&(ae(xe,i),ae(ve,f))}function Te(i){xe.current===i&&(Se(ve),Se(xe)),Ce.current===i&&(Se(Ce),ql._currentValue=Q)}var me=Object.prototype.hasOwnProperty,Ee=e.unstable_scheduleCallback,le=e.unstable_cancelCallback,Be=e.unstable_shouldYield,Ue=e.unstable_requestPaint,he=e.unstable_now,Ne=e.unstable_getCurrentPriorityLevel,ee=e.unstable_ImmediatePriority,ce=e.unstable_UserBlockingPriority,_e=e.unstable_NormalPriority,Pe=e.unstable_LowPriority,We=e.unstable_IdlePriority,yt=e.log,kt=e.unstable_setDisableYieldValue,ht=null,et=null;function Tt(i){if(et&&typeof et.onCommitFiberRoot=="function")try{et.onCommitFiberRoot(ht,i,void 0,(i.current.flags&128)===128)}catch{}}function ot(i){if(typeof yt=="function"&&kt(i),et&&typeof et.setStrictMode=="function")try{et.setStrictMode(ht,i)}catch{}}var St=Math.clz32?Math.clz32:zt,Ht=Math.log,fn=Math.LN2;function zt(i){return i>>>=0,i===0?32:31-(Ht(i)/fn|0)|0}var or=128,$r=4194304;function Jt(i){var l=i&42;if(l!==0)return l;switch(i&-i){case 1:return 1;case 2:return 2;case 4:return 4;case 8:return 8;case 16:return 16;case 32:return 32;case 64:return 64;case 128:case 256:case 512:case 1024:case 2048:case 4096:case 8192:case 16384:case 32768:case 65536:case 131072:case 262144:case 524288:case 1048576:case 2097152:return i&4194176;case 4194304:case 8388608:case 16777216:case 33554432:return i&62914560;case 67108864:return 67108864;case 134217728:return 134217728;case 268435456:return 268435456;case 536870912:return 536870912;case 1073741824:return 0;default:return i}}function fa(i,l){var f=i.pendingLanes;if(f===0)return 0;var h=0,v=i.suspendedLanes,T=i.pingedLanes,P=i.warmLanes;i=i.finishedLanes!==0;var q=f&134217727;return q!==0?(f=q&~v,f!==0?h=Jt(f):(T&=q,T!==0?h=Jt(T):i||(P=q&~P,P!==0&&(h=Jt(P))))):(q=f&~v,q!==0?h=Jt(q):T!==0?h=Jt(T):i||(P=f&~P,P!==0&&(h=Jt(P)))),h===0?0:l!==0&&l!==h&&!(l&v)&&(v=h&-h,P=l&-l,v>=P||v===32&&(P&4194176)!==0)?l:h}function Xe(i,l){return(i.pendingLanes&~(i.suspendedLanes&~i.pingedLanes)&l)===0}function mt(i,l){switch(i){case 1:case 2:case 4:case 8:return l+250;case 16:case 32:case 64:case 128:case 256:case 512:case 1024:case 2048:case 4096:case 8192:case 16384:case 32768:case 65536:case 131072:case 262144:case 524288:case 1048576:case 2097152:return l+5e3;case 4194304:case 8388608:case 16777216:case 33554432:return-1;case 67108864:case 134217728:case 268435456:case 536870912:case 1073741824:return-1;default:return-1}}function Ct(){var i=or;return or<<=1,!(or&4194176)&&(or=128),i}function Dn(){var i=$r;return $r<<=1,!($r&62914560)&&($r=4194304),i}function _n(i){for(var l=[],f=0;31>f;f++)l.push(i);return l}function In(i,l){i.pendingLanes|=l,l!==268435456&&(i.suspendedLanes=0,i.pingedLanes=0,i.warmLanes=0)}function pa(i,l,f,h,v,T){var P=i.pendingLanes;i.pendingLanes=f,i.suspendedLanes=0,i.pingedLanes=0,i.warmLanes=0,i.expiredLanes&=f,i.entangledLanes&=f,i.errorRecoveryDisabledLanes&=f,i.shellSuspendCounter=0;var q=i.entanglements,X=i.expirationTimes,ne=i.hiddenUpdates;for(f=P&~f;0<f;){var be=31-St(f),Ae=1<<be;q[be]=0,X[be]=-1;var de=ne[be];if(de!==null)for(ne[be]=null,be=0;be<de.length;be++){var ge=de[be];ge!==null&&(ge.lane&=-536870913)}f&=~Ae}h!==0&&Ln(i,h,0),T!==0&&v===0&&i.tag!==0&&(i.suspendedLanes|=T&~(P&~l))}function Ln(i,l,f){i.pendingLanes|=l,i.suspendedLanes&=~l;var h=31-St(l);i.entangledLanes|=l,i.entanglements[h]=i.entanglements[h]|1073741824|f&4194218}function Rr(i,l){var f=i.entangledLanes|=l;for(i=i.entanglements;f;){var h=31-St(f),v=1<<h;v&l|i[h]&l&&(i[h]|=l),f&=~v}}function mA(i){return i&=-i,2<i?8<i?i&134217727?32:268435456:8:2}function bA(){var i=W.p;return i!==0?i:(i=window.event,i===void 0?32:B_(i.type))}function QH(i,l){var f=W.p;try{return W.p=i,l()}finally{W.p=f}}var Ka=Math.random().toString(36).slice(2),Sn="__reactFiber$"+Ka,Mn="__reactProps$"+Ka,Ri="__reactContainer$"+Ka,Up="__reactEvents$"+Ka,JH="__reactListeners$"+Ka,e$="__reactHandles$"+Ka,yA="__reactResources$"+Ka,rl="__reactMarker$"+Ka;function jp(i){delete i[Sn],delete i[Mn],delete i[Up],delete i[JH],delete i[e$]}function Fo(i){var l=i[Sn];if(l)return l;for(var f=i.parentNode;f;){if(l=f[Ri]||f[Sn]){if(f=l.alternate,l.child!==null||f!==null&&f.child!==null)for(i=k_(i);i!==null;){if(f=i[Sn])return f;i=k_(i)}return l}i=f,f=i.parentNode}return null}function _i(i){if(i=i[Sn]||i[Ri]){var l=i.tag;if(l===5||l===6||l===13||l===26||l===27||l===3)return i}return null}function al(i){var l=i.tag;if(l===5||l===26||l===27||l===6)return i.stateNode;throw Error(r(33))}function Ci(i){var l=i[yA];return l||(l=i[yA]={hoistableStyles:new Map,hoistableScripts:new Map}),l}function rn(i){i[rl]=!0}var vA=new Set,SA={};function Po(i,l){Ni(i,l),Ni(i+"Capture",l)}function Ni(i,l){for(SA[i]=l,i=0;i<l.length;i++)vA.add(l[i])}var ga=!(typeof window>"u"||typeof window.document>"u"||typeof window.document.createElement>"u"),t$=RegExp("^[:A-Z_a-z\\u00C0-\\u00D6\\u00D8-\\u00F6\\u00F8-\\u02FF\\u0370-\\u037D\\u037F-\\u1FFF\\u200C-\\u200D\\u2070-\\u218F\\u2C00-\\u2FEF\\u3001-\\uD7FF\\uF900-\\uFDCF\\uFDF0-\\uFFFD][:A-Z_a-z\\u00C0-\\u00D6\\u00D8-\\u00F6\\u00F8-\\u02FF\\u0370-\\u037D\\u037F-\\u1FFF\\u200C-\\u200D\\u2070-\\u218F\\u2C00-\\u2FEF\\u3001-\\uD7FF\\uF900-\\uFDCF\\uFDF0-\\uFFFD\\-.0-9\\u00B7\\u0300-\\u036F\\u203F-\\u2040]*$"),EA={},wA={};function n$(i){return me.call(wA,i)?!0:me.call(EA,i)?!1:t$.test(i)?wA[i]=!0:(EA[i]=!0,!1)}function sc(i,l,f){if(n$(l))if(f===null)i.removeAttribute(l);else{switch(typeof f){case"undefined":case"function":case"symbol":i.removeAttribute(l);return;case"boolean":var h=l.toLowerCase().slice(0,5);if(h!=="data-"&&h!=="aria-"){i.removeAttribute(l);return}}i.setAttribute(l,""+f)}}function lc(i,l,f){if(f===null)i.removeAttribute(l);else{switch(typeof f){case"undefined":case"function":case"symbol":case"boolean":i.removeAttribute(l);return}i.setAttribute(l,""+f)}}function ha(i,l,f,h){if(h===null)i.removeAttribute(f);else{switch(typeof h){case"undefined":case"function":case"symbol":case"boolean":i.removeAttribute(f);return}i.setAttributeNS(l,f,""+h)}}function ir(i){switch(typeof i){case"bigint":case"boolean":case"number":case"string":case"undefined":return i;case"object":return i;default:return""}}function xA(i){var l=i.type;return(i=i.nodeName)&&i.toLowerCase()==="input"&&(l==="checkbox"||l==="radio")}function r$(i){var l=xA(i)?"checked":"value",f=Object.getOwnPropertyDescriptor(i.constructor.prototype,l),h=""+i[l];if(!i.hasOwnProperty(l)&&typeof f<"u"&&typeof f.get=="function"&&typeof f.set=="function"){var v=f.get,T=f.set;return Object.defineProperty(i,l,{configurable:!0,get:function(){return v.call(this)},set:function(P){h=""+P,T.call(this,P)}}),Object.defineProperty(i,l,{enumerable:f.enumerable}),{getValue:function(){return h},setValue:function(P){h=""+P},stopTracking:function(){i._valueTracker=null,delete i[l]}}}}function uc(i){i._valueTracker||(i._valueTracker=r$(i))}function kA(i){if(!i)return!1;var l=i._valueTracker;if(!l)return!0;var f=l.getValue(),h="";return i&&(h=xA(i)?i.checked?"true":"false":i.value),i=h,i!==f?(l.setValue(i),!0):!1}function cc(i){if(i=i||(typeof document<"u"?document:void 0),typeof i>"u")return null;try{return i.activeElement||i.body}catch{return i.body}}var a$=/[\n"\\]/g;function sr(i){return i.replace(a$,function(l){return"\\"+l.charCodeAt(0).toString(16)+" "})}function Gp(i,l,f,h,v,T,P,q){i.name="",P!=null&&typeof P!="function"&&typeof P!="symbol"&&typeof P!="boolean"?i.type=P:i.removeAttribute("type"),l!=null?P==="number"?(l===0&&i.value===""||i.value!=l)&&(i.value=""+ir(l)):i.value!==""+ir(l)&&(i.value=""+ir(l)):P!=="submit"&&P!=="reset"||i.removeAttribute("value"),l!=null?Hp(i,P,ir(l)):f!=null?Hp(i,P,ir(f)):h!=null&&i.removeAttribute("value"),v==null&&T!=null&&(i.defaultChecked=!!T),v!=null&&(i.checked=v&&typeof v!="function"&&typeof v!="symbol"),q!=null&&typeof q!="function"&&typeof q!="symbol"&&typeof q!="boolean"?i.name=""+ir(q):i.removeAttribute("name")}function TA(i,l,f,h,v,T,P,q){if(T!=null&&typeof T!="function"&&typeof T!="symbol"&&typeof T!="boolean"&&(i.type=T),l!=null||f!=null){if(!(T!=="submit"&&T!=="reset"||l!=null))return;f=f!=null?""+ir(f):"",l=l!=null?""+ir(l):f,q||l===i.value||(i.value=l),i.defaultValue=l}h=h??v,h=typeof h!="function"&&typeof h!="symbol"&&!!h,i.checked=q?i.checked:!!h,i.defaultChecked=!!h,P!=null&&typeof P!="function"&&typeof P!="symbol"&&typeof P!="boolean"&&(i.name=P)}function Hp(i,l,f){l==="number"&&cc(i.ownerDocument)===i||i.defaultValue===""+f||(i.defaultValue=""+f)}function Oi(i,l,f,h){if(i=i.options,l){l={};for(var v=0;v<f.length;v++)l["$"+f[v]]=!0;for(f=0;f<i.length;f++)v=l.hasOwnProperty("$"+i[f].value),i[f].selected!==v&&(i[f].selected=v),v&&h&&(i[f].defaultSelected=!0)}else{for(f=""+ir(f),l=null,v=0;v<i.length;v++){if(i[v].value===f){i[v].selected=!0,h&&(i[v].defaultSelected=!0);return}l!==null||i[v].disabled||(l=i[v])}l!==null&&(l.selected=!0)}}function AA(i,l,f){if(l!=null&&(l=""+ir(l),l!==i.value&&(i.value=l),f==null)){i.defaultValue!==l&&(i.defaultValue=l);return}i.defaultValue=f!=null?""+ir(f):""}function RA(i,l,f,h){if(l==null){if(h!=null){if(f!=null)throw Error(r(92));if(M(h)){if(1<h.length)throw Error(r(93));h=h[0]}f=h}f==null&&(f=""),l=f}f=ir(l),i.defaultValue=f,h=i.textContent,h===f&&h!==""&&h!==null&&(i.value=h)}function Di(i,l){if(l){var f=i.firstChild;if(f&&f===i.lastChild&&f.nodeType===3){f.nodeValue=l;return}}i.textContent=l}var o$=new Set("animationIterationCount aspectRatio borderImageOutset borderImageSlice borderImageWidth boxFlex boxFlexGroup boxOrdinalGroup columnCount columns flex flexGrow flexPositive flexShrink flexNegative flexOrder gridArea gridRow gridRowEnd gridRowSpan gridRowStart gridColumn gridColumnEnd gridColumnSpan gridColumnStart fontWeight lineClamp lineHeight opacity order orphans scale tabSize widows zIndex zoom fillOpacity floodOpacity stopOpacity strokeDasharray strokeDashoffset strokeMiterlimit strokeOpacity strokeWidth MozAnimationIterationCount MozBoxFlex MozBoxFlexGroup MozLineClamp msAnimationIterationCount msFlex msZoom msFlexGrow msFlexNegative msFlexOrder msFlexPositive msFlexShrink msGridColumn msGridColumnSpan msGridRow msGridRowSpan WebkitAnimationIterationCount WebkitBoxFlex WebKitBoxFlexGroup WebkitBoxOrdinalGroup WebkitColumnCount WebkitColumns WebkitFlex WebkitFlexGrow WebkitFlexPositive WebkitFlexShrink WebkitLineClamp".split(" "));function _A(i,l,f){var h=l.indexOf("--")===0;f==null||typeof f=="boolean"||f===""?h?i.setProperty(l,""):l==="float"?i.cssFloat="":i[l]="":h?i.setProperty(l,f):typeof f!="number"||f===0||o$.has(l)?l==="float"?i.cssFloat=f:i[l]=(""+f).trim():i[l]=f+"px"}function CA(i,l,f){if(l!=null&&typeof l!="object")throw Error(r(62));if(i=i.style,f!=null){for(var h in f)!f.hasOwnProperty(h)||l!=null&&l.hasOwnProperty(h)||(h.indexOf("--")===0?i.setProperty(h,""):h==="float"?i.cssFloat="":i[h]="");for(var v in l)h=l[v],l.hasOwnProperty(v)&&f[v]!==h&&_A(i,v,h)}else for(var T in l)l.hasOwnProperty(T)&&_A(i,T,l[T])}function $p(i){if(i.indexOf("-")===-1)return!1;switch(i){case"annotation-xml":case"color-profile":case"font-face":case"font-face-src":case"font-face-uri":case"font-face-format":case"font-face-name":case"missing-glyph":return!1;default:return!0}}var i$=new Map([["acceptCharset","accept-charset"],["htmlFor","for"],["httpEquiv","http-equiv"],["crossOrigin","crossorigin"],["accentHeight","accent-height"],["alignmentBaseline","alignment-baseline"],["arabicForm","arabic-form"],["baselineShift","baseline-shift"],["capHeight","cap-height"],["clipPath","clip-path"],["clipRule","clip-rule"],["colorInterpolation","color-interpolation"],["colorInterpolationFilters","color-interpolation-filters"],["colorProfile","color-profile"],["colorRendering","color-rendering"],["dominantBaseline","dominant-baseline"],["enableBackground","enable-background"],["fillOpacity","fill-opacity"],["fillRule","fill-rule"],["floodColor","flood-color"],["floodOpacity","flood-opacity"],["fontFamily","font-family"],["fontSize","font-size"],["fontSizeAdjust","font-size-adjust"],["fontStretch","font-stretch"],["fontStyle","font-style"],["fontVariant","font-variant"],["fontWeight","font-weight"],["glyphName","glyph-name"],["glyphOrientationHorizontal","glyph-orientation-horizontal"],["glyphOrientationVertical","glyph-orientation-vertical"],["horizAdvX","horiz-adv-x"],["horizOriginX","horiz-origin-x"],["imageRendering","image-rendering"],["letterSpacing","letter-spacing"],["lightingColor","lighting-color"],["markerEnd","marker-end"],["markerMid","marker-mid"],["markerStart","marker-start"],["overlinePosition","overline-position"],["overlineThickness","overline-thickness"],["paintOrder","paint-order"],["panose-1","panose-1"],["pointerEvents","pointer-events"],["renderingIntent","rendering-intent"],["shapeRendering","shape-rendering"],["stopColor","stop-color"],["stopOpacity","stop-opacity"],["strikethroughPosition","strikethrough-position"],["strikethroughThickness","strikethrough-thickness"],["strokeDasharray","stroke-dasharray"],["strokeDashoffset","stroke-dashoffset"],["strokeLinecap","stroke-linecap"],["strokeLinejoin","stroke-linejoin"],["strokeMiterlimit","stroke-miterlimit"],["strokeOpacity","stroke-opacity"],["strokeWidth","stroke-width"],["textAnchor","text-anchor"],["textDecoration","text-decoration"],["textRendering","text-rendering"],["transformOrigin","transform-origin"],["underlinePosition","underline-position"],["underlineThickness","underline-thickness"],["unicodeBidi","unicode-bidi"],["unicodeRange","unicode-range"],["unitsPerEm","units-per-em"],["vAlphabetic","v-alphabetic"],["vHanging","v-hanging"],["vIdeographic","v-ideographic"],["vMathematical","v-mathematical"],["vectorEffect","vector-effect"],["vertAdvY","vert-adv-y"],["vertOriginX","vert-origin-x"],["vertOriginY","vert-origin-y"],["wordSpacing","word-spacing"],["writingMode","writing-mode"],["xmlnsXlink","xmlns:xlink"],["xHeight","x-height"]]),s$=/^[\u0000-\u001F ]*j[\r\n\t]*a[\r\n\t]*v[\r\n\t]*a[\r\n\t]*s[\r\n\t]*c[\r\n\t]*r[\r\n\t]*i[\r\n\t]*p[\r\n\t]*t[\r\n\t]*:/i;function dc(i){return s$.test(""+i)?"javascript:throw new Error('React has blocked a javascript: URL as a security precaution.')":i}var qp=null;function Vp(i){return i=i.target||i.srcElement||window,i.correspondingUseElement&&(i=i.correspondingUseElement),i.nodeType===3?i.parentNode:i}var Ii=null,Li=null;function NA(i){var l=_i(i);if(l&&(i=l.stateNode)){var f=i[Mn]||null;e:switch(i=l.stateNode,l.type){case"input":if(Gp(i,f.value,f.defaultValue,f.defaultValue,f.checked,f.defaultChecked,f.type,f.name),l=f.name,f.type==="radio"&&l!=null){for(f=i;f.parentNode;)f=f.parentNode;for(f=f.querySelectorAll('input[name="'+sr(""+l)+'"][type="radio"]'),l=0;l<f.length;l++){var h=f[l];if(h!==i&&h.form===i.form){var v=h[Mn]||null;if(!v)throw Error(r(90));Gp(h,v.value,v.defaultValue,v.defaultValue,v.checked,v.defaultChecked,v.type,v.name)}}for(l=0;l<f.length;l++)h=f[l],h.form===i.form&&kA(h)}break e;case"textarea":AA(i,f.value,f.defaultValue);break e;case"select":l=f.value,l!=null&&Oi(i,!!f.multiple,l,!1)}}}var Wp=!1;function OA(i,l,f){if(Wp)return i(l,f);Wp=!0;try{var h=i(l);return h}finally{if(Wp=!1,(Ii!==null||Li!==null)&&(Yc(),Ii&&(l=Ii,i=Li,Li=Ii=null,NA(l),i)))for(l=0;l<i.length;l++)NA(i[l])}}function ol(i,l){var f=i.stateNode;if(f===null)return null;var h=f[Mn]||null;if(h===null)return null;f=h[l];e:switch(l){case"onClick":case"onClickCapture":case"onDoubleClick":case"onDoubleClickCapture":case"onMouseDown":case"onMouseDownCapture":case"onMouseMove":case"onMouseMoveCapture":case"onMouseUp":case"onMouseUpCapture":case"onMouseEnter":(h=!h.disabled)||(i=i.type,h=!(i==="button"||i==="input"||i==="select"||i==="textarea")),i=!h;break e;default:i=!1}if(i)return null;if(f&&typeof f!="function")throw Error(r(231,l,typeof f));return f}var Yp=!1;if(ga)try{var il={};Object.defineProperty(il,"passive",{get:function(){Yp=!0}}),window.addEventListener("test",il,il),window.removeEventListener("test",il,il)}catch{Yp=!1}var Xa=null,Kp=null,fc=null;function DA(){if(fc)return fc;var i,l=Kp,f=l.length,h,v="value"in Xa?Xa.value:Xa.textContent,T=v.length;for(i=0;i<f&&l[i]===v[i];i++);var P=f-i;for(h=1;h<=P&&l[f-h]===v[T-h];h++);return fc=v.slice(i,1<h?1-h:void 0)}function pc(i){var l=i.keyCode;return"charCode"in i?(i=i.charCode,i===0&&l===13&&(i=13)):i=l,i===10&&(i=13),32<=i||i===13?i:0}function gc(){return!0}function IA(){return!1}function Fn(i){function l(f,h,v,T,P){this._reactName=f,this._targetInst=v,this.type=h,this.nativeEvent=T,this.target=P,this.currentTarget=null;for(var q in i)i.hasOwnProperty(q)&&(f=i[q],this[q]=f?f(T):T[q]);return this.isDefaultPrevented=(T.defaultPrevented!=null?T.defaultPrevented:T.returnValue===!1)?gc:IA,this.isPropagationStopped=IA,this}return I(l.prototype,{preventDefault:function(){this.defaultPrevented=!0;var f=this.nativeEvent;f&&(f.preventDefault?f.preventDefault():typeof f.returnValue!="unknown"&&(f.returnValue=!1),this.isDefaultPrevented=gc)},stopPropagation:function(){var f=this.nativeEvent;f&&(f.stopPropagation?f.stopPropagation():typeof f.cancelBubble!="unknown"&&(f.cancelBubble=!0),this.isPropagationStopped=gc)},persist:function(){},isPersistent:gc}),l}var zo={eventPhase:0,bubbles:0,cancelable:0,timeStamp:function(i){return i.timeStamp||Date.now()},defaultPrevented:0,isTrusted:0},hc=Fn(zo),sl=I({},zo,{view:0,detail:0}),l$=Fn(sl),Xp,Zp,ll,mc=I({},sl,{screenX:0,screenY:0,clientX:0,clientY:0,pageX:0,pageY:0,ctrlKey:0,shiftKey:0,altKey:0,metaKey:0,getModifierState:Jp,button:0,buttons:0,relatedTarget:function(i){return i.relatedTarget===void 0?i.fromElement===i.srcElement?i.toElement:i.fromElement:i.relatedTarget},movementX:function(i){return"movementX"in i?i.movementX:(i!==ll&&(ll&&i.type==="mousemove"?(Xp=i.screenX-ll.screenX,Zp=i.screenY-ll.screenY):Zp=Xp=0,ll=i),Xp)},movementY:function(i){return"movementY"in i?i.movementY:Zp}}),LA=Fn(mc),u$=I({},mc,{dataTransfer:0}),c$=Fn(u$),d$=I({},sl,{relatedTarget:0}),Qp=Fn(d$),f$=I({},zo,{animationName:0,elapsedTime:0,pseudoElement:0}),p$=Fn(f$),g$=I({},zo,{clipboardData:function(i){return"clipboardData"in i?i.clipboardData:window.clipboardData}}),h$=Fn(g$),m$=I({},zo,{data:0}),MA=Fn(m$),b$={Esc:"Escape",Spacebar:" ",Left:"ArrowLeft",Up:"ArrowUp",Right:"ArrowRight",Down:"ArrowDown",Del:"Delete",Win:"OS",Menu:"ContextMenu",Apps:"ContextMenu",Scroll:"ScrollLock",MozPrintableKey:"Unidentified"},y$={8:"Backspace",9:"Tab",12:"Clear",13:"Enter",16:"Shift",17:"Control",18:"Alt",19:"Pause",20:"CapsLock",27:"Escape",32:" ",33:"PageUp",34:"PageDown",35:"End",36:"Home",37:"ArrowLeft",38:"ArrowUp",39:"ArrowRight",40:"ArrowDown",45:"Insert",46:"Delete",112:"F1",113:"F2",114:"F3",115:"F4",116:"F5",117:"F6",118:"F7",119:"F8",120:"F9",121:"F10",122:"F11",123:"F12",144:"NumLock",145:"ScrollLock",224:"Meta"},v$={Alt:"altKey",Control:"ctrlKey",Meta:"metaKey",Shift:"shiftKey"};function S$(i){var l=this.nativeEvent;return l.getModifierState?l.getModifierState(i):(i=v$[i])?!!l[i]:!1}function Jp(){return S$}var E$=I({},sl,{key:function(i){if(i.key){var l=b$[i.key]||i.key;if(l!=="Unidentified")return l}return i.type==="keypress"?(i=pc(i),i===13?"Enter":String.fromCharCode(i)):i.type==="keydown"||i.type==="keyup"?y$[i.keyCode]||"Unidentified":""},code:0,location:0,ctrlKey:0,shiftKey:0,altKey:0,metaKey:0,repeat:0,locale:0,getModifierState:Jp,charCode:function(i){return i.type==="keypress"?pc(i):0},keyCode:function(i){return i.type==="keydown"||i.type==="keyup"?i.keyCode:0},which:function(i){return i.type==="keypress"?pc(i):i.type==="keydown"||i.type==="keyup"?i.keyCode:0}}),w$=Fn(E$),x$=I({},mc,{pointerId:0,width:0,height:0,pressure:0,tangentialPressure:0,tiltX:0,tiltY:0,twist:0,pointerType:0,isPrimary:0}),FA=Fn(x$),k$=I({},sl,{touches:0,targetTouches:0,changedTouches:0,altKey:0,metaKey:0,ctrlKey:0,shiftKey:0,getModifierState:Jp}),T$=Fn(k$),A$=I({},zo,{propertyName:0,elapsedTime:0,pseudoElement:0}),R$=Fn(A$),_$=I({},mc,{deltaX:function(i){return"deltaX"in i?i.deltaX:"wheelDeltaX"in i?-i.wheelDeltaX:0},deltaY:function(i){return"deltaY"in i?i.deltaY:"wheelDeltaY"in i?-i.wheelDeltaY:"wheelDelta"in i?-i.wheelDelta:0},deltaZ:0,deltaMode:0}),C$=Fn(_$),N$=I({},zo,{newState:0,oldState:0}),O$=Fn(N$),D$=[9,13,27,32],eg=ga&&"CompositionEvent"in window,ul=null;ga&&"documentMode"in document&&(ul=document.documentMode);var I$=ga&&"TextEvent"in window&&!ul,PA=ga&&(!eg||ul&&8<ul&&11>=ul),zA=" ",BA=!1;function UA(i,l){switch(i){case"keyup":return D$.indexOf(l.keyCode)!==-1;case"keydown":return l.keyCode!==229;case"keypress":case"mousedown":case"focusout":return!0;default:return!1}}function jA(i){return i=i.detail,typeof i=="object"&&"data"in i?i.data:null}var Mi=!1;function L$(i,l){switch(i){case"compositionend":return jA(l);case"keypress":return l.which!==32?null:(BA=!0,zA);case"textInput":return i=l.data,i===zA&&BA?null:i;default:return null}}function M$(i,l){if(Mi)return i==="compositionend"||!eg&&UA(i,l)?(i=DA(),fc=Kp=Xa=null,Mi=!1,i):null;switch(i){case"paste":return null;case"keypress":if(!(l.ctrlKey||l.altKey||l.metaKey)||l.ctrlKey&&l.altKey){if(l.char&&1<l.char.length)return l.char;if(l.which)return String.fromCharCode(l.which)}return null;case"compositionend":return PA&&l.locale!=="ko"?null:l.data;default:return null}}var F$={color:!0,date:!0,datetime:!0,"datetime-local":!0,email:!0,month:!0,number:!0,password:!0,range:!0,search:!0,tel:!0,text:!0,time:!0,url:!0,week:!0};function GA(i){var l=i&&i.nodeName&&i.nodeName.toLowerCase();return l==="input"?!!F$[i.type]:l==="textarea"}function HA(i,l,f,h){Ii?Li?Li.push(h):Li=[h]:Ii=h,l=Jc(l,"onChange"),0<l.length&&(f=new hc("onChange","change",null,f,h),i.push({event:f,listeners:l}))}var cl=null,dl=null;function P$(i){h_(i,0)}function bc(i){var l=al(i);if(kA(l))return i}function $A(i,l){if(i==="change")return l}var qA=!1;if(ga){var tg;if(ga){var ng="oninput"in document;if(!ng){var VA=document.createElement("div");VA.setAttribute("oninput","return;"),ng=typeof VA.oninput=="function"}tg=ng}else tg=!1;qA=tg&&(!document.documentMode||9<document.documentMode)}function WA(){cl&&(cl.detachEvent("onpropertychange",YA),dl=cl=null)}function YA(i){if(i.propertyName==="value"&&bc(dl)){var l=[];HA(l,dl,i,Vp(i)),OA(P$,l)}}function z$(i,l,f){i==="focusin"?(WA(),cl=l,dl=f,cl.attachEvent("onpropertychange",YA)):i==="focusout"&&WA()}function B$(i){if(i==="selectionchange"||i==="keyup"||i==="keydown")return bc(dl)}function U$(i,l){if(i==="click")return bc(l)}function j$(i,l){if(i==="input"||i==="change")return bc(l)}function G$(i,l){return i===l&&(i!==0||1/i===1/l)||i!==i&&l!==l}var Yn=typeof Object.is=="function"?Object.is:G$;function fl(i,l){if(Yn(i,l))return!0;if(typeof i!="object"||i===null||typeof l!="object"||l===null)return!1;var f=Object.keys(i),h=Object.keys(l);if(f.length!==h.length)return!1;for(h=0;h<f.length;h++){var v=f[h];if(!me.call(l,v)||!Yn(i[v],l[v]))return!1}return!0}function KA(i){for(;i&&i.firstChild;)i=i.firstChild;return i}function XA(i,l){var f=KA(i);i=0;for(var h;f;){if(f.nodeType===3){if(h=i+f.textContent.length,i<=l&&h>=l)return{node:f,offset:l-i};i=h}e:{for(;f;){if(f.nextSibling){f=f.nextSibling;break e}f=f.parentNode}f=void 0}f=KA(f)}}function ZA(i,l){return i&&l?i===l?!0:i&&i.nodeType===3?!1:l&&l.nodeType===3?ZA(i,l.parentNode):"contains"in i?i.contains(l):i.compareDocumentPosition?!!(i.compareDocumentPosition(l)&16):!1:!1}function QA(i){i=i!=null&&i.ownerDocument!=null&&i.ownerDocument.defaultView!=null?i.ownerDocument.defaultView:window;for(var l=cc(i.document);l instanceof i.HTMLIFrameElement;){try{var f=typeof l.contentWindow.location.href=="string"}catch{f=!1}if(f)i=l.contentWindow;else break;l=cc(i.document)}return l}function rg(i){var l=i&&i.nodeName&&i.nodeName.toLowerCase();return l&&(l==="input"&&(i.type==="text"||i.type==="search"||i.type==="tel"||i.type==="url"||i.type==="password")||l==="textarea"||i.contentEditable==="true")}function H$(i,l){var f=QA(l);l=i.focusedElem;var h=i.selectionRange;if(f!==l&&l&&l.ownerDocument&&ZA(l.ownerDocument.documentElement,l)){if(h!==null&&rg(l)){if(i=h.start,f=h.end,f===void 0&&(f=i),"selectionStart"in l)l.selectionStart=i,l.selectionEnd=Math.min(f,l.value.length);else if(f=(i=l.ownerDocument||document)&&i.defaultView||window,f.getSelection){f=f.getSelection();var v=l.textContent.length,T=Math.min(h.start,v);h=h.end===void 0?T:Math.min(h.end,v),!f.extend&&T>h&&(v=h,h=T,T=v),v=XA(l,T);var P=XA(l,h);v&&P&&(f.rangeCount!==1||f.anchorNode!==v.node||f.anchorOffset!==v.offset||f.focusNode!==P.node||f.focusOffset!==P.offset)&&(i=i.createRange(),i.setStart(v.node,v.offset),f.removeAllRanges(),T>h?(f.addRange(i),f.extend(P.node,P.offset)):(i.setEnd(P.node,P.offset),f.addRange(i)))}}for(i=[],f=l;f=f.parentNode;)f.nodeType===1&&i.push({element:f,left:f.scrollLeft,top:f.scrollTop});for(typeof l.focus=="function"&&l.focus(),l=0;l<i.length;l++)f=i[l],f.element.scrollLeft=f.left,f.element.scrollTop=f.top}}var $$=ga&&"documentMode"in document&&11>=document.documentMode,Fi=null,ag=null,pl=null,og=!1;function JA(i,l,f){var h=f.window===f?f.document:f.nodeType===9?f:f.ownerDocument;og||Fi==null||Fi!==cc(h)||(h=Fi,"selectionStart"in h&&rg(h)?h={start:h.selectionStart,end:h.selectionEnd}:(h=(h.ownerDocument&&h.ownerDocument.defaultView||window).getSelection(),h={anchorNode:h.anchorNode,anchorOffset:h.anchorOffset,focusNode:h.focusNode,focusOffset:h.focusOffset}),pl&&fl(pl,h)||(pl=h,h=Jc(ag,"onSelect"),0<h.length&&(l=new hc("onSelect","select",null,l,f),i.push({event:l,listeners:h}),l.target=Fi)))}function Bo(i,l){var f={};return f[i.toLowerCase()]=l.toLowerCase(),f["Webkit"+i]="webkit"+l,f["Moz"+i]="moz"+l,f}var Pi={animationend:Bo("Animation","AnimationEnd"),animationiteration:Bo("Animation","AnimationIteration"),animationstart:Bo("Animation","AnimationStart"),transitionrun:Bo("Transition","TransitionRun"),transitionstart:Bo("Transition","TransitionStart"),transitioncancel:Bo("Transition","TransitionCancel"),transitionend:Bo("Transition","TransitionEnd")},ig={},e1={};ga&&(e1=document.createElement("div").style,"AnimationEvent"in window||(delete Pi.animationend.animation,delete Pi.animationiteration.animation,delete Pi.animationstart.animation),"TransitionEvent"in window||delete Pi.transitionend.transition);function Uo(i){if(ig[i])return ig[i];if(!Pi[i])return i;var l=Pi[i],f;for(f in l)if(l.hasOwnProperty(f)&&f in e1)return ig[i]=l[f];return i}var t1=Uo("animationend"),n1=Uo("animationiteration"),r1=Uo("animationstart"),q$=Uo("transitionrun"),V$=Uo("transitionstart"),W$=Uo("transitioncancel"),a1=Uo("transitionend"),o1=new Map,i1="abort auxClick beforeToggle cancel canPlay canPlayThrough click close contextMenu copy cut drag dragEnd dragEnter dragExit dragLeave dragOver dragStart drop durationChange emptied encrypted ended error gotPointerCapture input invalid keyDown keyPress keyUp load loadedData loadedMetadata loadStart lostPointerCapture mouseDown mouseMove mouseOut mouseOver mouseUp paste pause play playing pointerCancel pointerDown pointerMove pointerOut pointerOver pointerUp progress rateChange reset resize seeked seeking stalled submit suspend timeUpdate touchCancel touchEnd touchStart volumeChange scroll scrollEnd toggle touchMove waiting wheel".split(" ");function _r(i,l){o1.set(i,l),Po(l,[i])}var lr=[],zi=0,sg=0;function yc(){for(var i=zi,l=sg=zi=0;l<i;){var f=lr[l];lr[l++]=null;var h=lr[l];lr[l++]=null;var v=lr[l];lr[l++]=null;var T=lr[l];if(lr[l++]=null,h!==null&&v!==null){var P=h.pending;P===null?v.next=v:(v.next=P.next,P.next=v),h.pending=v}T!==0&&s1(f,v,T)}}function vc(i,l,f,h){lr[zi++]=i,lr[zi++]=l,lr[zi++]=f,lr[zi++]=h,sg|=h,i.lanes|=h,i=i.alternate,i!==null&&(i.lanes|=h)}function lg(i,l,f,h){return vc(i,l,f,h),Sc(i)}function Za(i,l){return vc(i,null,null,l),Sc(i)}function s1(i,l,f){i.lanes|=f;var h=i.alternate;h!==null&&(h.lanes|=f);for(var v=!1,T=i.return;T!==null;)T.childLanes|=f,h=T.alternate,h!==null&&(h.childLanes|=f),T.tag===22&&(i=T.stateNode,i===null||i._visibility&1||(v=!0)),i=T,T=T.return;v&&l!==null&&i.tag===3&&(T=i.stateNode,v=31-St(f),T=T.hiddenUpdates,i=T[v],i===null?T[v]=[l]:i.push(l),l.lane=f|536870912)}function Sc(i){if(50<zl)throw zl=0,gh=null,Error(r(185));for(var l=i.return;l!==null;)i=l,l=i.return;return i.tag===3?i.stateNode:null}var Bi={},l1=new WeakMap;function ur(i,l){if(typeof i=="object"&&i!==null){var f=l1.get(i);return f!==void 0?f:(l={value:i,source:l,stack:G(l)},l1.set(i,l),l)}return{value:i,source:l,stack:G(l)}}var Ui=[],ji=0,Ec=null,wc=0,cr=[],dr=0,jo=null,ma=1,ba="";function Go(i,l){Ui[ji++]=wc,Ui[ji++]=Ec,Ec=i,wc=l}function u1(i,l,f){cr[dr++]=ma,cr[dr++]=ba,cr[dr++]=jo,jo=i;var h=ma;i=ba;var v=32-St(h)-1;h&=~(1<<v),f+=1;var T=32-St(l)+v;if(30<T){var P=v-v%5;T=(h&(1<<P)-1).toString(32),h>>=P,v-=P,ma=1<<32-St(l)+v|f<<v|h,ba=T+i}else ma=1<<T|f<<v|h,ba=i}function ug(i){i.return!==null&&(Go(i,1),u1(i,1,0))}function cg(i){for(;i===Ec;)Ec=Ui[--ji],Ui[ji]=null,wc=Ui[--ji],Ui[ji]=null;for(;i===jo;)jo=cr[--dr],cr[dr]=null,ba=cr[--dr],cr[dr]=null,ma=cr[--dr],cr[dr]=null}var Cn=null,pn=null,Et=!1,Cr=null,qr=!1,dg=Error(r(519));function Ho(i){var l=Error(r(418,""));throw ml(ur(l,i)),dg}function c1(i){var l=i.stateNode,f=i.type,h=i.memoizedProps;switch(l[Sn]=i,l[Mn]=h,f){case"dialog":pt("cancel",l),pt("close",l);break;case"iframe":case"object":case"embed":pt("load",l);break;case"video":case"audio":for(f=0;f<Ul.length;f++)pt(Ul[f],l);break;case"source":pt("error",l);break;case"img":case"image":case"link":pt("error",l),pt("load",l);break;case"details":pt("toggle",l);break;case"input":pt("invalid",l),TA(l,h.value,h.defaultValue,h.checked,h.defaultChecked,h.type,h.name,!0),uc(l);break;case"select":pt("invalid",l);break;case"textarea":pt("invalid",l),RA(l,h.value,h.defaultValue,h.children),uc(l)}f=h.children,typeof f!="string"&&typeof f!="number"&&typeof f!="bigint"||l.textContent===""+f||h.suppressHydrationWarning===!0||v_(l.textContent,f)?(h.popover!=null&&(pt("beforetoggle",l),pt("toggle",l)),h.onScroll!=null&&pt("scroll",l),h.onScrollEnd!=null&&pt("scrollend",l),h.onClick!=null&&(l.onclick=ed),l=!0):l=!1,l||Ho(i)}function d1(i){for(Cn=i.return;Cn;)switch(Cn.tag){case 3:case 27:qr=!0;return;case 5:case 13:qr=!1;return;default:Cn=Cn.return}}function gl(i){if(i!==Cn)return!1;if(!Et)return d1(i),Et=!0,!1;var l=!1,f;if((f=i.tag!==3&&i.tag!==27)&&((f=i.tag===5)&&(f=i.type,f=!(f!=="form"&&f!=="button")||Oh(i.type,i.memoizedProps)),f=!f),f&&(l=!0),l&&pn&&Ho(i),d1(i),i.tag===13){if(i=i.memoizedState,i=i!==null?i.dehydrated:null,!i)throw Error(r(317));e:{for(i=i.nextSibling,l=0;i;){if(i.nodeType===8)if(f=i.data,f==="/$"){if(l===0){pn=Or(i.nextSibling);break e}l--}else f!=="$"&&f!=="$!"&&f!=="$?"||l++;i=i.nextSibling}pn=null}}else pn=Cn?Or(i.stateNode.nextSibling):null;return!0}function hl(){pn=Cn=null,Et=!1}function ml(i){Cr===null?Cr=[i]:Cr.push(i)}var bl=Error(r(460)),f1=Error(r(474)),fg={then:function(){}};function p1(i){return i=i.status,i==="fulfilled"||i==="rejected"}function xc(){}function g1(i,l,f){switch(f=i[f],f===void 0?i.push(l):f!==l&&(l.then(xc,xc),l=f),l.status){case"fulfilled":return l.value;case"rejected":throw i=l.reason,i===bl?Error(r(483)):i;default:if(typeof l.status=="string")l.then(xc,xc);else{if(i=It,i!==null&&100<i.shellSuspendCounter)throw Error(r(482));i=l,i.status="pending",i.then(function(h){if(l.status==="pending"){var v=l;v.status="fulfilled",v.value=h}},function(h){if(l.status==="pending"){var v=l;v.status="rejected",v.reason=h}})}switch(l.status){case"fulfilled":return l.value;case"rejected":throw i=l.reason,i===bl?Error(r(483)):i}throw yl=l,bl}}var yl=null;function h1(){if(yl===null)throw Error(r(459));var i=yl;return yl=null,i}var Gi=null,vl=0;function kc(i){var l=vl;return vl+=1,Gi===null&&(Gi=[]),g1(Gi,i,l)}function Sl(i,l){l=l.props.ref,i.ref=l!==void 0?l:null}function Tc(i,l){throw l.$$typeof===o?Error(r(525)):(i=Object.prototype.toString.call(l),Error(r(31,i==="[object Object]"?"object with keys {"+Object.keys(l).join(", ")+"}":i)))}function m1(i){var l=i._init;return l(i._payload)}function b1(i){function l(se,te){if(i){var ue=se.deletions;ue===null?(se.deletions=[te],se.flags|=16):ue.push(te)}}function f(se,te){if(!i)return null;for(;te!==null;)l(se,te),te=te.sibling;return null}function h(se){for(var te=new Map;se!==null;)se.key!==null?te.set(se.key,se):te.set(se.index,se),se=se.sibling;return te}function v(se,te){return se=uo(se,te),se.index=0,se.sibling=null,se}function T(se,te,ue){return se.index=ue,i?(ue=se.alternate,ue!==null?(ue=ue.index,ue<te?(se.flags|=33554434,te):ue):(se.flags|=33554434,te)):(se.flags|=1048576,te)}function P(se){return i&&se.alternate===null&&(se.flags|=33554434),se}function q(se,te,ue,we){return te===null||te.tag!==6?(te=ih(ue,se.mode,we),te.return=se,te):(te=v(te,ue),te.return=se,te)}function X(se,te,ue,we){var Ge=ue.type;return Ge===c?be(se,te,ue.props.children,we,ue.key):te!==null&&(te.elementType===Ge||typeof Ge=="object"&&Ge!==null&&Ge.$$typeof===k&&m1(Ge)===te.type)?(te=v(te,ue.props),Sl(te,ue),te.return=se,te):(te=Hc(ue.type,ue.key,ue.props,null,se.mode,we),Sl(te,ue),te.return=se,te)}function ne(se,te,ue,we){return te===null||te.tag!==4||te.stateNode.containerInfo!==ue.containerInfo||te.stateNode.implementation!==ue.implementation?(te=sh(ue,se.mode,we),te.return=se,te):(te=v(te,ue.children||[]),te.return=se,te)}function be(se,te,ue,we,Ge){return te===null||te.tag!==7?(te=Jo(ue,se.mode,we,Ge),te.return=se,te):(te=v(te,ue),te.return=se,te)}function Ae(se,te,ue){if(typeof te=="string"&&te!==""||typeof te=="number"||typeof te=="bigint")return te=ih(""+te,se.mode,ue),te.return=se,te;if(typeof te=="object"&&te!==null){switch(te.$$typeof){case s:return ue=Hc(te.type,te.key,te.props,null,se.mode,ue),Sl(ue,te),ue.return=se,ue;case u:return te=sh(te,se.mode,ue),te.return=se,te;case k:var we=te._init;return te=we(te._payload),Ae(se,te,ue)}if(M(te)||_(te))return te=Jo(te,se.mode,ue,null),te.return=se,te;if(typeof te.then=="function")return Ae(se,kc(te),ue);if(te.$$typeof===b)return Ae(se,Uc(se,te),ue);Tc(se,te)}return null}function de(se,te,ue,we){var Ge=te!==null?te.key:null;if(typeof ue=="string"&&ue!==""||typeof ue=="number"||typeof ue=="bigint")return Ge!==null?null:q(se,te,""+ue,we);if(typeof ue=="object"&&ue!==null){switch(ue.$$typeof){case s:return ue.key===Ge?X(se,te,ue,we):null;case u:return ue.key===Ge?ne(se,te,ue,we):null;case k:return Ge=ue._init,ue=Ge(ue._payload),de(se,te,ue,we)}if(M(ue)||_(ue))return Ge!==null?null:be(se,te,ue,we,null);if(typeof ue.then=="function")return de(se,te,kc(ue),we);if(ue.$$typeof===b)return de(se,te,Uc(se,ue),we);Tc(se,ue)}return null}function ge(se,te,ue,we,Ge){if(typeof we=="string"&&we!==""||typeof we=="number"||typeof we=="bigint")return se=se.get(ue)||null,q(te,se,""+we,Ge);if(typeof we=="object"&&we!==null){switch(we.$$typeof){case s:return se=se.get(we.key===null?ue:we.key)||null,X(te,se,we,Ge);case u:return se=se.get(we.key===null?ue:we.key)||null,ne(te,se,we,Ge);case k:var lt=we._init;return we=lt(we._payload),ge(se,te,ue,we,Ge)}if(M(we)||_(we))return se=se.get(ue)||null,be(te,se,we,Ge,null);if(typeof we.then=="function")return ge(se,te,ue,kc(we),Ge);if(we.$$typeof===b)return ge(se,te,ue,Uc(te,we),Ge);Tc(te,we)}return null}function qe(se,te,ue,we){for(var Ge=null,lt=null,Ye=te,Je=te=0,sn=null;Ye!==null&&Je<ue.length;Je++){Ye.index>Je?(sn=Ye,Ye=null):sn=Ye.sibling;var wt=de(se,Ye,ue[Je],we);if(wt===null){Ye===null&&(Ye=sn);break}i&&Ye&&wt.alternate===null&&l(se,Ye),te=T(wt,te,Je),lt===null?Ge=wt:lt.sibling=wt,lt=wt,Ye=sn}if(Je===ue.length)return f(se,Ye),Et&&Go(se,Je),Ge;if(Ye===null){for(;Je<ue.length;Je++)Ye=Ae(se,ue[Je],we),Ye!==null&&(te=T(Ye,te,Je),lt===null?Ge=Ye:lt.sibling=Ye,lt=Ye);return Et&&Go(se,Je),Ge}for(Ye=h(Ye);Je<ue.length;Je++)sn=ge(Ye,se,Je,ue[Je],we),sn!==null&&(i&&sn.alternate!==null&&Ye.delete(sn.key===null?Je:sn.key),te=T(sn,te,Je),lt===null?Ge=sn:lt.sibling=sn,lt=sn);return i&&Ye.forEach(function(bo){return l(se,bo)}),Et&&Go(se,Je),Ge}function nt(se,te,ue,we){if(ue==null)throw Error(r(151));for(var Ge=null,lt=null,Ye=te,Je=te=0,sn=null,wt=ue.next();Ye!==null&&!wt.done;Je++,wt=ue.next()){Ye.index>Je?(sn=Ye,Ye=null):sn=Ye.sibling;var bo=de(se,Ye,wt.value,we);if(bo===null){Ye===null&&(Ye=sn);break}i&&Ye&&bo.alternate===null&&l(se,Ye),te=T(bo,te,Je),lt===null?Ge=bo:lt.sibling=bo,lt=bo,Ye=sn}if(wt.done)return f(se,Ye),Et&&Go(se,Je),Ge;if(Ye===null){for(;!wt.done;Je++,wt=ue.next())wt=Ae(se,wt.value,we),wt!==null&&(te=T(wt,te,Je),lt===null?Ge=wt:lt.sibling=wt,lt=wt);return Et&&Go(se,Je),Ge}for(Ye=h(Ye);!wt.done;Je++,wt=ue.next())wt=ge(Ye,se,Je,wt.value,we),wt!==null&&(i&&wt.alternate!==null&&Ye.delete(wt.key===null?Je:wt.key),te=T(wt,te,Je),lt===null?Ge=wt:lt.sibling=wt,lt=wt);return i&&Ye.forEach(function(s9){return l(se,s9)}),Et&&Go(se,Je),Ge}function Vt(se,te,ue,we){if(typeof ue=="object"&&ue!==null&&ue.type===c&&ue.key===null&&(ue=ue.props.children),typeof ue=="object"&&ue!==null){switch(ue.$$typeof){case s:e:{for(var Ge=ue.key;te!==null;){if(te.key===Ge){if(Ge=ue.type,Ge===c){if(te.tag===7){f(se,te.sibling),we=v(te,ue.props.children),we.return=se,se=we;break e}}else if(te.elementType===Ge||typeof Ge=="object"&&Ge!==null&&Ge.$$typeof===k&&m1(Ge)===te.type){f(se,te.sibling),we=v(te,ue.props),Sl(we,ue),we.return=se,se=we;break e}f(se,te);break}else l(se,te);te=te.sibling}ue.type===c?(we=Jo(ue.props.children,se.mode,we,ue.key),we.return=se,se=we):(we=Hc(ue.type,ue.key,ue.props,null,se.mode,we),Sl(we,ue),we.return=se,se=we)}return P(se);case u:e:{for(Ge=ue.key;te!==null;){if(te.key===Ge)if(te.tag===4&&te.stateNode.containerInfo===ue.containerInfo&&te.stateNode.implementation===ue.implementation){f(se,te.sibling),we=v(te,ue.children||[]),we.return=se,se=we;break e}else{f(se,te);break}else l(se,te);te=te.sibling}we=sh(ue,se.mode,we),we.return=se,se=we}return P(se);case k:return Ge=ue._init,ue=Ge(ue._payload),Vt(se,te,ue,we)}if(M(ue))return qe(se,te,ue,we);if(_(ue)){if(Ge=_(ue),typeof Ge!="function")throw Error(r(150));return ue=Ge.call(ue),nt(se,te,ue,we)}if(typeof ue.then=="function")return Vt(se,te,kc(ue),we);if(ue.$$typeof===b)return Vt(se,te,Uc(se,ue),we);Tc(se,ue)}return typeof ue=="string"&&ue!==""||typeof ue=="number"||typeof ue=="bigint"?(ue=""+ue,te!==null&&te.tag===6?(f(se,te.sibling),we=v(te,ue),we.return=se,se=we):(f(se,te),we=ih(ue,se.mode,we),we.return=se,se=we),P(se)):f(se,te)}return function(se,te,ue,we){try{vl=0;var Ge=Vt(se,te,ue,we);return Gi=null,Ge}catch(Ye){if(Ye===bl)throw Ye;var lt=hr(29,Ye,null,se.mode);return lt.lanes=we,lt.return=se,lt}finally{}}}var $o=b1(!0),y1=b1(!1),Hi=ie(null),Ac=ie(0);function v1(i,l){i=_a,ae(Ac,i),ae(Hi,l),_a=i|l.baseLanes}function pg(){ae(Ac,_a),ae(Hi,Hi.current)}function gg(){_a=Ac.current,Se(Hi),Se(Ac)}var fr=ie(null),Vr=null;function Qa(i){var l=i.alternate;ae(en,en.current&1),ae(fr,i),Vr===null&&(l===null||Hi.current!==null||l.memoizedState!==null)&&(Vr=i)}function S1(i){if(i.tag===22){if(ae(en,en.current),ae(fr,i),Vr===null){var l=i.alternate;l!==null&&l.memoizedState!==null&&(Vr=i)}}else Ja()}function Ja(){ae(en,en.current),ae(fr,fr.current)}function ya(i){Se(fr),Vr===i&&(Vr=null),Se(en)}var en=ie(0);function Rc(i){for(var l=i;l!==null;){if(l.tag===13){var f=l.memoizedState;if(f!==null&&(f=f.dehydrated,f===null||f.data==="$?"||f.data==="$!"))return l}else if(l.tag===19&&l.memoizedProps.revealOrder!==void 0){if(l.flags&128)return l}else if(l.child!==null){l.child.return=l,l=l.child;continue}if(l===i)break;for(;l.sibling===null;){if(l.return===null||l.return===i)return null;l=l.return}l.sibling.return=l.return,l=l.sibling}return null}var Y$=typeof AbortController<"u"?AbortController:function(){var i=[],l=this.signal={aborted:!1,addEventListener:function(f,h){i.push(h)}};this.abort=function(){l.aborted=!0,i.forEach(function(f){return f()})}},K$=e.unstable_scheduleCallback,X$=e.unstable_NormalPriority,tn={$$typeof:b,Consumer:null,Provider:null,_currentValue:null,_currentValue2:null,_threadCount:0};function hg(){return{controller:new Y$,data:new Map,refCount:0}}function El(i){i.refCount--,i.refCount===0&&K$(X$,function(){i.controller.abort()})}var wl=null,mg=0,$i=0,qi=null;function Z$(i,l){if(wl===null){var f=wl=[];mg=0,$i=wh(),qi={status:"pending",value:void 0,then:function(h){f.push(h)}}}return mg++,l.then(E1,E1),l}function E1(){if(--mg===0&&wl!==null){qi!==null&&(qi.status="fulfilled");var i=wl;wl=null,$i=0,qi=null;for(var l=0;l<i.length;l++)(0,i[l])()}}function Q$(i,l){var f=[],h={status:"pending",value:null,reason:null,then:function(v){f.push(v)}};return i.then(function(){h.status="fulfilled",h.value=l;for(var v=0;v<f.length;v++)(0,f[v])(l)},function(v){for(h.status="rejected",h.reason=v,v=0;v<f.length;v++)(0,f[v])(void 0)}),h}var w1=L.S;L.S=function(i,l){typeof l=="object"&&l!==null&&typeof l.then=="function"&&Z$(i,l),w1!==null&&w1(i,l)};var qo=ie(null);function bg(){var i=qo.current;return i!==null?i:It.pooledCache}function _c(i,l){l===null?ae(qo,qo.current):ae(qo,l.pool)}function x1(){var i=bg();return i===null?null:{parent:tn._currentValue,pool:i}}var eo=0,st=null,At=null,Kt=null,Cc=!1,Vi=!1,Vo=!1,Nc=0,xl=0,Wi=null,J$=0;function Wt(){throw Error(r(321))}function yg(i,l){if(l===null)return!1;for(var f=0;f<l.length&&f<i.length;f++)if(!Yn(i[f],l[f]))return!1;return!0}function vg(i,l,f,h,v,T){return eo=T,st=l,l.memoizedState=null,l.updateQueue=null,l.lanes=0,L.H=i===null||i.memoizedState===null?Wo:to,Vo=!1,T=f(h,v),Vo=!1,Vi&&(T=T1(l,f,h,v)),k1(i),T}function k1(i){L.H=Wr;var l=At!==null&&At.next!==null;if(eo=0,Kt=At=st=null,Cc=!1,xl=0,Wi=null,l)throw Error(r(300));i===null||an||(i=i.dependencies,i!==null&&Bc(i)&&(an=!0))}function T1(i,l,f,h){st=i;var v=0;do{if(Vi&&(Wi=null),xl=0,Vi=!1,25<=v)throw Error(r(301));if(v+=1,Kt=At=null,i.updateQueue!=null){var T=i.updateQueue;T.lastEffect=null,T.events=null,T.stores=null,T.memoCache!=null&&(T.memoCache.index=0)}L.H=Yo,T=l(f,h)}while(Vi);return T}function e6(){var i=L.H,l=i.useState()[0];return l=typeof l.then=="function"?kl(l):l,i=i.useState()[0],(At!==null?At.memoizedState:null)!==i&&(st.flags|=1024),l}function Sg(){var i=Nc!==0;return Nc=0,i}function Eg(i,l,f){l.updateQueue=i.updateQueue,l.flags&=-2053,i.lanes&=~f}function wg(i){if(Cc){for(i=i.memoizedState;i!==null;){var l=i.queue;l!==null&&(l.pending=null),i=i.next}Cc=!1}eo=0,Kt=At=st=null,Vi=!1,xl=Nc=0,Wi=null}function Pn(){var i={memoizedState:null,baseState:null,baseQueue:null,queue:null,next:null};return Kt===null?st.memoizedState=Kt=i:Kt=Kt.next=i,Kt}function Xt(){if(At===null){var i=st.alternate;i=i!==null?i.memoizedState:null}else i=At.next;var l=Kt===null?st.memoizedState:Kt.next;if(l!==null)Kt=l,At=i;else{if(i===null)throw st.alternate===null?Error(r(467)):Error(r(310));At=i,i={memoizedState:At.memoizedState,baseState:At.baseState,baseQueue:At.baseQueue,queue:At.queue,next:null},Kt===null?st.memoizedState=Kt=i:Kt=Kt.next=i}return Kt}var Oc;Oc=function(){return{lastEffect:null,events:null,stores:null,memoCache:null}};function kl(i){var l=xl;return xl+=1,Wi===null&&(Wi=[]),i=g1(Wi,i,l),l=st,(Kt===null?l.memoizedState:Kt.next)===null&&(l=l.alternate,L.H=l===null||l.memoizedState===null?Wo:to),i}function Dc(i){if(i!==null&&typeof i=="object"){if(typeof i.then=="function")return kl(i);if(i.$$typeof===b)return En(i)}throw Error(r(438,String(i)))}function xg(i){var l=null,f=st.updateQueue;if(f!==null&&(l=f.memoCache),l==null){var h=st.alternate;h!==null&&(h=h.updateQueue,h!==null&&(h=h.memoCache,h!=null&&(l={data:h.data.map(function(v){return v.slice()}),index:0})))}if(l==null&&(l={data:[],index:0}),f===null&&(f=Oc(),st.updateQueue=f),f.memoCache=l,f=l.data[l.index],f===void 0)for(f=l.data[l.index]=Array(i),h=0;h<i;h++)f[h]=C;return l.index++,f}function va(i,l){return typeof l=="function"?l(i):l}function Ic(i){var l=Xt();return kg(l,At,i)}function kg(i,l,f){var h=i.queue;if(h===null)throw Error(r(311));h.lastRenderedReducer=f;var v=i.baseQueue,T=h.pending;if(T!==null){if(v!==null){var P=v.next;v.next=T.next,T.next=P}l.baseQueue=v=T,h.pending=null}if(T=i.baseState,v===null)i.memoizedState=T;else{l=v.next;var q=P=null,X=null,ne=l,be=!1;do{var Ae=ne.lane&-536870913;if(Ae!==ne.lane?(bt&Ae)===Ae:(eo&Ae)===Ae){var de=ne.revertLane;if(de===0)X!==null&&(X=X.next={lane:0,revertLane:0,action:ne.action,hasEagerState:ne.hasEagerState,eagerState:ne.eagerState,next:null}),Ae===$i&&(be=!0);else if((eo&de)===de){ne=ne.next,de===$i&&(be=!0);continue}else Ae={lane:0,revertLane:ne.revertLane,action:ne.action,hasEagerState:ne.hasEagerState,eagerState:ne.eagerState,next:null},X===null?(q=X=Ae,P=T):X=X.next=Ae,st.lanes|=de,co|=de;Ae=ne.action,Vo&&f(T,Ae),T=ne.hasEagerState?ne.eagerState:f(T,Ae)}else de={lane:Ae,revertLane:ne.revertLane,action:ne.action,hasEagerState:ne.hasEagerState,eagerState:ne.eagerState,next:null},X===null?(q=X=de,P=T):X=X.next=de,st.lanes|=Ae,co|=Ae;ne=ne.next}while(ne!==null&&ne!==l);if(X===null?P=T:X.next=q,!Yn(T,i.memoizedState)&&(an=!0,be&&(f=qi,f!==null)))throw f;i.memoizedState=T,i.baseState=P,i.baseQueue=X,h.lastRenderedState=T}return v===null&&(h.lanes=0),[i.memoizedState,h.dispatch]}function Tg(i){var l=Xt(),f=l.queue;if(f===null)throw Error(r(311));f.lastRenderedReducer=i;var h=f.dispatch,v=f.pending,T=l.memoizedState;if(v!==null){f.pending=null;var P=v=v.next;do T=i(T,P.action),P=P.next;while(P!==v);Yn(T,l.memoizedState)||(an=!0),l.memoizedState=T,l.baseQueue===null&&(l.baseState=T),f.lastRenderedState=T}return[T,h]}function A1(i,l,f){var h=st,v=Xt(),T=Et;if(T){if(f===void 0)throw Error(r(407));f=f()}else f=l();var P=!Yn((At||v).memoizedState,f);if(P&&(v.memoizedState=f,an=!0),v=v.queue,_g(C1.bind(null,h,v,i),[i]),v.getSnapshot!==l||P||Kt!==null&&Kt.memoizedState.tag&1){if(h.flags|=2048,Yi(9,_1.bind(null,h,v,f,l),{destroy:void 0},null),It===null)throw Error(r(349));T||eo&60||R1(h,l,f)}return f}function R1(i,l,f){i.flags|=16384,i={getSnapshot:l,value:f},l=st.updateQueue,l===null?(l=Oc(),st.updateQueue=l,l.stores=[i]):(f=l.stores,f===null?l.stores=[i]:f.push(i))}function _1(i,l,f,h){l.value=f,l.getSnapshot=h,N1(l)&&O1(i)}function C1(i,l,f){return f(function(){N1(l)&&O1(i)})}function N1(i){var l=i.getSnapshot;i=i.value;try{var f=l();return!Yn(i,f)}catch{return!0}}function O1(i){var l=Za(i,2);l!==null&&Nn(l,i,2)}function Ag(i){var l=Pn();if(typeof i=="function"){var f=i;if(i=f(),Vo){ot(!0);try{f()}finally{ot(!1)}}}return l.memoizedState=l.baseState=i,l.queue={pending:null,lanes:0,dispatch:null,lastRenderedReducer:va,lastRenderedState:i},l}function D1(i,l,f,h){return i.baseState=f,kg(i,At,typeof h=="function"?h:va)}function t6(i,l,f,h,v){if(Fc(i))throw Error(r(485));if(i=l.action,i!==null){var T={payload:v,action:i,next:null,isTransition:!0,status:"pending",value:null,reason:null,listeners:[],then:function(P){T.listeners.push(P)}};L.T!==null?f(!0):T.isTransition=!1,h(T),f=l.pending,f===null?(T.next=l.pending=T,I1(l,T)):(T.next=f.next,l.pending=f.next=T)}}function I1(i,l){var f=l.action,h=l.payload,v=i.state;if(l.isTransition){var T=L.T,P={};L.T=P;try{var q=f(v,h),X=L.S;X!==null&&X(P,q),L1(i,l,q)}catch(ne){Rg(i,l,ne)}finally{L.T=T}}else try{T=f(v,h),L1(i,l,T)}catch(ne){Rg(i,l,ne)}}function L1(i,l,f){f!==null&&typeof f=="object"&&typeof f.then=="function"?f.then(function(h){M1(i,l,h)},function(h){return Rg(i,l,h)}):M1(i,l,f)}function M1(i,l,f){l.status="fulfilled",l.value=f,F1(l),i.state=f,l=i.pending,l!==null&&(f=l.next,f===l?i.pending=null:(f=f.next,l.next=f,I1(i,f)))}function Rg(i,l,f){var h=i.pending;if(i.pending=null,h!==null){h=h.next;do l.status="rejected",l.reason=f,F1(l),l=l.next;while(l!==h)}i.action=null}function F1(i){i=i.listeners;for(var l=0;l<i.length;l++)(0,i[l])()}function P1(i,l){return l}function z1(i,l){if(Et){var f=It.formState;if(f!==null){e:{var h=st;if(Et){if(pn){t:{for(var v=pn,T=qr;v.nodeType!==8;){if(!T){v=null;break t}if(v=Or(v.nextSibling),v===null){v=null;break t}}T=v.data,v=T==="F!"||T==="F"?v:null}if(v){pn=Or(v.nextSibling),h=v.data==="F!";break e}}Ho(h)}h=!1}h&&(l=f[0])}}return f=Pn(),f.memoizedState=f.baseState=l,h={pending:null,lanes:0,dispatch:null,lastRenderedReducer:P1,lastRenderedState:l},f.queue=h,f=nR.bind(null,st,h),h.dispatch=f,h=Ag(!1),T=Ig.bind(null,st,!1,h.queue),h=Pn(),v={state:l,dispatch:null,action:i,pending:null},h.queue=v,f=t6.bind(null,st,v,T,f),v.dispatch=f,h.memoizedState=i,[l,f,!1]}function B1(i){var l=Xt();return U1(l,At,i)}function U1(i,l,f){l=kg(i,l,P1)[0],i=Ic(va)[0],l=typeof l=="object"&&l!==null&&typeof l.then=="function"?kl(l):l;var h=Xt(),v=h.queue,T=v.dispatch;return f!==h.memoizedState&&(st.flags|=2048,Yi(9,n6.bind(null,v,f),{destroy:void 0},null)),[l,T,i]}function n6(i,l){i.action=l}function j1(i){var l=Xt(),f=At;if(f!==null)return U1(l,f,i);Xt(),l=l.memoizedState,f=Xt();var h=f.queue.dispatch;return f.memoizedState=i,[l,h,!1]}function Yi(i,l,f,h){return i={tag:i,create:l,inst:f,deps:h,next:null},l=st.updateQueue,l===null&&(l=Oc(),st.updateQueue=l),f=l.lastEffect,f===null?l.lastEffect=i.next=i:(h=f.next,f.next=i,i.next=h,l.lastEffect=i),i}function G1(){return Xt().memoizedState}function Lc(i,l,f,h){var v=Pn();st.flags|=i,v.memoizedState=Yi(1|l,f,{destroy:void 0},h===void 0?null:h)}function Mc(i,l,f,h){var v=Xt();h=h===void 0?null:h;var T=v.memoizedState.inst;At!==null&&h!==null&&yg(h,At.memoizedState.deps)?v.memoizedState=Yi(l,f,T,h):(st.flags|=i,v.memoizedState=Yi(1|l,f,T,h))}function H1(i,l){Lc(8390656,8,i,l)}function _g(i,l){Mc(2048,8,i,l)}function $1(i,l){return Mc(4,2,i,l)}function q1(i,l){return Mc(4,4,i,l)}function V1(i,l){if(typeof l=="function"){i=i();var f=l(i);return function(){typeof f=="function"?f():l(null)}}if(l!=null)return i=i(),l.current=i,function(){l.current=null}}function W1(i,l,f){f=f!=null?f.concat([i]):null,Mc(4,4,V1.bind(null,l,i),f)}function Cg(){}function Y1(i,l){var f=Xt();l=l===void 0?null:l;var h=f.memoizedState;return l!==null&&yg(l,h[1])?h[0]:(f.memoizedState=[i,l],i)}function K1(i,l){var f=Xt();l=l===void 0?null:l;var h=f.memoizedState;if(l!==null&&yg(l,h[1]))return h[0];if(h=i(),Vo){ot(!0);try{i()}finally{ot(!1)}}return f.memoizedState=[h,l],h}function Ng(i,l,f){return f===void 0||eo&1073741824?i.memoizedState=l:(i.memoizedState=f,i=ZR(),st.lanes|=i,co|=i,f)}function X1(i,l,f,h){return Yn(f,l)?f:Hi.current!==null?(i=Ng(i,f,h),Yn(i,l)||(an=!0),i):eo&42?(i=ZR(),st.lanes|=i,co|=i,l):(an=!0,i.memoizedState=f)}function Z1(i,l,f,h,v){var T=W.p;W.p=T!==0&&8>T?T:8;var P=L.T,q={};L.T=q,Ig(i,!1,l,f);try{var X=v(),ne=L.S;if(ne!==null&&ne(q,X),X!==null&&typeof X=="object"&&typeof X.then=="function"){var be=Q$(X,h);Tl(i,l,be,Qn(i))}else Tl(i,l,h,Qn(i))}catch(Ae){Tl(i,l,{then:function(){},status:"rejected",reason:Ae},Qn())}finally{W.p=T,L.T=P}}function r6(){}function Og(i,l,f,h){if(i.tag!==5)throw Error(r(476));var v=Q1(i).queue;Z1(i,v,l,Q,f===null?r6:function(){return J1(i),f(h)})}function Q1(i){var l=i.memoizedState;if(l!==null)return l;l={memoizedState:Q,baseState:Q,baseQueue:null,queue:{pending:null,lanes:0,dispatch:null,lastRenderedReducer:va,lastRenderedState:Q},next:null};var f={};return l.next={memoizedState:f,baseState:f,baseQueue:null,queue:{pending:null,lanes:0,dispatch:null,lastRenderedReducer:va,lastRenderedState:f},next:null},i.memoizedState=l,i=i.alternate,i!==null&&(i.memoizedState=l),l}function J1(i){var l=Q1(i).next.queue;Tl(i,l,{},Qn())}function Dg(){return En(ql)}function eR(){return Xt().memoizedState}function tR(){return Xt().memoizedState}function a6(i){for(var l=i.return;l!==null;){switch(l.tag){case 24:case 3:var f=Qn();i=ao(f);var h=oo(l,i,f);h!==null&&(Nn(h,l,f),_l(h,l,f)),l={cache:hg()},i.payload=l;return}l=l.return}}function o6(i,l,f){var h=Qn();f={lane:h,revertLane:0,action:f,hasEagerState:!1,eagerState:null,next:null},Fc(i)?rR(l,f):(f=lg(i,l,f,h),f!==null&&(Nn(f,i,h),aR(f,l,h)))}function nR(i,l,f){var h=Qn();Tl(i,l,f,h)}function Tl(i,l,f,h){var v={lane:h,revertLane:0,action:f,hasEagerState:!1,eagerState:null,next:null};if(Fc(i))rR(l,v);else{var T=i.alternate;if(i.lanes===0&&(T===null||T.lanes===0)&&(T=l.lastRenderedReducer,T!==null))try{var P=l.lastRenderedState,q=T(P,f);if(v.hasEagerState=!0,v.eagerState=q,Yn(q,P))return vc(i,l,v,0),It===null&&yc(),!1}catch{}finally{}if(f=lg(i,l,v,h),f!==null)return Nn(f,i,h),aR(f,l,h),!0}return!1}function Ig(i,l,f,h){if(h={lane:2,revertLane:wh(),action:h,hasEagerState:!1,eagerState:null,next:null},Fc(i)){if(l)throw Error(r(479))}else l=lg(i,f,h,2),l!==null&&Nn(l,i,2)}function Fc(i){var l=i.alternate;return i===st||l!==null&&l===st}function rR(i,l){Vi=Cc=!0;var f=i.pending;f===null?l.next=l:(l.next=f.next,f.next=l),i.pending=l}function aR(i,l,f){if(f&4194176){var h=l.lanes;h&=i.pendingLanes,f|=h,l.lanes=f,Rr(i,f)}}var Wr={readContext:En,use:Dc,useCallback:Wt,useContext:Wt,useEffect:Wt,useImperativeHandle:Wt,useLayoutEffect:Wt,useInsertionEffect:Wt,useMemo:Wt,useReducer:Wt,useRef:Wt,useState:Wt,useDebugValue:Wt,useDeferredValue:Wt,useTransition:Wt,useSyncExternalStore:Wt,useId:Wt};Wr.useCacheRefresh=Wt,Wr.useMemoCache=Wt,Wr.useHostTransitionStatus=Wt,Wr.useFormState=Wt,Wr.useActionState=Wt,Wr.useOptimistic=Wt;var Wo={readContext:En,use:Dc,useCallback:function(i,l){return Pn().memoizedState=[i,l===void 0?null:l],i},useContext:En,useEffect:H1,useImperativeHandle:function(i,l,f){f=f!=null?f.concat([i]):null,Lc(4194308,4,V1.bind(null,l,i),f)},useLayoutEffect:function(i,l){return Lc(4194308,4,i,l)},useInsertionEffect:function(i,l){Lc(4,2,i,l)},useMemo:function(i,l){var f=Pn();l=l===void 0?null:l;var h=i();if(Vo){ot(!0);try{i()}finally{ot(!1)}}return f.memoizedState=[h,l],h},useReducer:function(i,l,f){var h=Pn();if(f!==void 0){var v=f(l);if(Vo){ot(!0);try{f(l)}finally{ot(!1)}}}else v=l;return h.memoizedState=h.baseState=v,i={pending:null,lanes:0,dispatch:null,lastRenderedReducer:i,lastRenderedState:v},h.queue=i,i=i.dispatch=o6.bind(null,st,i),[h.memoizedState,i]},useRef:function(i){var l=Pn();return i={current:i},l.memoizedState=i},useState:function(i){i=Ag(i);var l=i.queue,f=nR.bind(null,st,l);return l.dispatch=f,[i.memoizedState,f]},useDebugValue:Cg,useDeferredValue:function(i,l){var f=Pn();return Ng(f,i,l)},useTransition:function(){var i=Ag(!1);return i=Z1.bind(null,st,i.queue,!0,!1),Pn().memoizedState=i,[!1,i]},useSyncExternalStore:function(i,l,f){var h=st,v=Pn();if(Et){if(f===void 0)throw Error(r(407));f=f()}else{if(f=l(),It===null)throw Error(r(349));bt&60||R1(h,l,f)}v.memoizedState=f;var T={value:f,getSnapshot:l};return v.queue=T,H1(C1.bind(null,h,T,i),[i]),h.flags|=2048,Yi(9,_1.bind(null,h,T,f,l),{destroy:void 0},null),f},useId:function(){var i=Pn(),l=It.identifierPrefix;if(Et){var f=ba,h=ma;f=(h&~(1<<32-St(h)-1)).toString(32)+f,l=":"+l+"R"+f,f=Nc++,0<f&&(l+="H"+f.toString(32)),l+=":"}else f=J$++,l=":"+l+"r"+f.toString(32)+":";return i.memoizedState=l},useCacheRefresh:function(){return Pn().memoizedState=a6.bind(null,st)}};Wo.useMemoCache=xg,Wo.useHostTransitionStatus=Dg,Wo.useFormState=z1,Wo.useActionState=z1,Wo.useOptimistic=function(i){var l=Pn();l.memoizedState=l.baseState=i;var f={pending:null,lanes:0,dispatch:null,lastRenderedReducer:null,lastRenderedState:null};return l.queue=f,l=Ig.bind(null,st,!0,f),f.dispatch=l,[i,l]};var to={readContext:En,use:Dc,useCallback:Y1,useContext:En,useEffect:_g,useImperativeHandle:W1,useInsertionEffect:$1,useLayoutEffect:q1,useMemo:K1,useReducer:Ic,useRef:G1,useState:function(){return Ic(va)},useDebugValue:Cg,useDeferredValue:function(i,l){var f=Xt();return X1(f,At.memoizedState,i,l)},useTransition:function(){var i=Ic(va)[0],l=Xt().memoizedState;return[typeof i=="boolean"?i:kl(i),l]},useSyncExternalStore:A1,useId:eR};to.useCacheRefresh=tR,to.useMemoCache=xg,to.useHostTransitionStatus=Dg,to.useFormState=B1,to.useActionState=B1,to.useOptimistic=function(i,l){var f=Xt();return D1(f,At,i,l)};var Yo={readContext:En,use:Dc,useCallback:Y1,useContext:En,useEffect:_g,useImperativeHandle:W1,useInsertionEffect:$1,useLayoutEffect:q1,useMemo:K1,useReducer:Tg,useRef:G1,useState:function(){return Tg(va)},useDebugValue:Cg,useDeferredValue:function(i,l){var f=Xt();return At===null?Ng(f,i,l):X1(f,At.memoizedState,i,l)},useTransition:function(){var i=Tg(va)[0],l=Xt().memoizedState;return[typeof i=="boolean"?i:kl(i),l]},useSyncExternalStore:A1,useId:eR};Yo.useCacheRefresh=tR,Yo.useMemoCache=xg,Yo.useHostTransitionStatus=Dg,Yo.useFormState=j1,Yo.useActionState=j1,Yo.useOptimistic=function(i,l){var f=Xt();return At!==null?D1(f,At,i,l):(f.baseState=i,[i,f.queue.dispatch])};function Lg(i,l,f,h){l=i.memoizedState,f=f(h,l),f=f==null?l:I({},l,f),i.memoizedState=f,i.lanes===0&&(i.updateQueue.baseState=f)}var Mg={isMounted:function(i){return(i=i._reactInternals)?z(i)===i:!1},enqueueSetState:function(i,l,f){i=i._reactInternals;var h=Qn(),v=ao(h);v.payload=l,f!=null&&(v.callback=f),l=oo(i,v,h),l!==null&&(Nn(l,i,h),_l(l,i,h))},enqueueReplaceState:function(i,l,f){i=i._reactInternals;var h=Qn(),v=ao(h);v.tag=1,v.payload=l,f!=null&&(v.callback=f),l=oo(i,v,h),l!==null&&(Nn(l,i,h),_l(l,i,h))},enqueueForceUpdate:function(i,l){i=i._reactInternals;var f=Qn(),h=ao(f);h.tag=2,l!=null&&(h.callback=l),l=oo(i,h,f),l!==null&&(Nn(l,i,f),_l(l,i,f))}};function oR(i,l,f,h,v,T,P){return i=i.stateNode,typeof i.shouldComponentUpdate=="function"?i.shouldComponentUpdate(h,T,P):l.prototype&&l.prototype.isPureReactComponent?!fl(f,h)||!fl(v,T):!0}function iR(i,l,f,h){i=l.state,typeof l.componentWillReceiveProps=="function"&&l.componentWillReceiveProps(f,h),typeof l.UNSAFE_componentWillReceiveProps=="function"&&l.UNSAFE_componentWillReceiveProps(f,h),l.state!==i&&Mg.enqueueReplaceState(l,l.state,null)}function Ko(i,l){var f=l;if("ref"in l){f={};for(var h in l)h!=="ref"&&(f[h]=l[h])}if(i=i.defaultProps){f===l&&(f=I({},f));for(var v in i)f[v]===void 0&&(f[v]=i[v])}return f}var Pc=typeof reportError=="function"?reportError:function(i){if(typeof window=="object"&&typeof window.ErrorEvent=="function"){var l=new window.ErrorEvent("error",{bubbles:!0,cancelable:!0,message:typeof i=="object"&&i!==null&&typeof i.message=="string"?String(i.message):String(i),error:i});if(!window.dispatchEvent(l))return}else if(typeof process=="object"&&typeof process.emit=="function"){process.emit("uncaughtException",i);return}console.error(i)};function sR(i){Pc(i)}function lR(i){console.error(i)}function uR(i){Pc(i)}function zc(i,l){try{var f=i.onUncaughtError;f(l.value,{componentStack:l.stack})}catch(h){setTimeout(function(){throw h})}}function cR(i,l,f){try{var h=i.onCaughtError;h(f.value,{componentStack:f.stack,errorBoundary:l.tag===1?l.stateNode:null})}catch(v){setTimeout(function(){throw v})}}function Fg(i,l,f){return f=ao(f),f.tag=3,f.payload={element:null},f.callback=function(){zc(i,l)},f}function dR(i){return i=ao(i),i.tag=3,i}function fR(i,l,f,h){var v=f.type.getDerivedStateFromError;if(typeof v=="function"){var T=h.value;i.payload=function(){return v(T)},i.callback=function(){cR(l,f,h)}}var P=f.stateNode;P!==null&&typeof P.componentDidCatch=="function"&&(i.callback=function(){cR(l,f,h),typeof v!="function"&&(fo===null?fo=new Set([this]):fo.add(this));var q=h.stack;this.componentDidCatch(h.value,{componentStack:q!==null?q:""})})}function i6(i,l,f,h,v){if(f.flags|=32768,h!==null&&typeof h=="object"&&typeof h.then=="function"){if(l=f.alternate,l!==null&&Rl(l,f,v,!0),f=fr.current,f!==null){switch(f.tag){case 13:return Vr===null?bh():f.alternate===null&&qt===0&&(qt=3),f.flags&=-257,f.flags|=65536,f.lanes=v,h===fg?f.flags|=16384:(l=f.updateQueue,l===null?f.updateQueue=new Set([h]):l.add(h),vh(i,h,v)),!1;case 22:return f.flags|=65536,h===fg?f.flags|=16384:(l=f.updateQueue,l===null?(l={transitions:null,markerInstances:null,retryQueue:new Set([h])},f.updateQueue=l):(f=l.retryQueue,f===null?l.retryQueue=new Set([h]):f.add(h)),vh(i,h,v)),!1}throw Error(r(435,f.tag))}return vh(i,h,v),bh(),!1}if(Et)return l=fr.current,l!==null?(!(l.flags&65536)&&(l.flags|=256),l.flags|=65536,l.lanes=v,h!==dg&&(i=Error(r(422),{cause:h}),ml(ur(i,f)))):(h!==dg&&(l=Error(r(423),{cause:h}),ml(ur(l,f))),i=i.current.alternate,i.flags|=65536,v&=-v,i.lanes|=v,h=ur(h,f),v=Fg(i.stateNode,h,v),Zg(i,v),qt!==4&&(qt=2)),!1;var T=Error(r(520),{cause:h});if(T=ur(T,f),Fl===null?Fl=[T]:Fl.push(T),qt!==4&&(qt=2),l===null)return!0;h=ur(h,f),f=l;do{switch(f.tag){case 3:return f.flags|=65536,i=v&-v,f.lanes|=i,i=Fg(f.stateNode,h,i),Zg(f,i),!1;case 1:if(l=f.type,T=f.stateNode,(f.flags&128)===0&&(typeof l.getDerivedStateFromError=="function"||T!==null&&typeof T.componentDidCatch=="function"&&(fo===null||!fo.has(T))))return f.flags|=65536,v&=-v,f.lanes|=v,v=dR(v),fR(v,i,f,h),Zg(f,v),!1}f=f.return}while(f!==null);return!1}var pR=Error(r(461)),an=!1;function gn(i,l,f,h){l.child=i===null?y1(l,null,f,h):$o(l,i.child,f,h)}function gR(i,l,f,h,v){f=f.render;var T=l.ref;if("ref"in h){var P={};for(var q in h)q!=="ref"&&(P[q]=h[q])}else P=h;return Zo(l),h=vg(i,l,f,P,T,v),q=Sg(),i!==null&&!an?(Eg(i,l,v),Sa(i,l,v)):(Et&&q&&ug(l),l.flags|=1,gn(i,l,h,v),l.child)}function hR(i,l,f,h,v){if(i===null){var T=f.type;return typeof T=="function"&&!oh(T)&&T.defaultProps===void 0&&f.compare===null?(l.tag=15,l.type=T,mR(i,l,T,h,v)):(i=Hc(f.type,null,h,l,l.mode,v),i.ref=l.ref,i.return=l,l.child=i)}if(T=i.child,!qg(i,v)){var P=T.memoizedProps;if(f=f.compare,f=f!==null?f:fl,f(P,h)&&i.ref===l.ref)return Sa(i,l,v)}return l.flags|=1,i=uo(T,h),i.ref=l.ref,i.return=l,l.child=i}function mR(i,l,f,h,v){if(i!==null){var T=i.memoizedProps;if(fl(T,h)&&i.ref===l.ref)if(an=!1,l.pendingProps=h=T,qg(i,v))i.flags&131072&&(an=!0);else return l.lanes=i.lanes,Sa(i,l,v)}return Pg(i,l,f,h,v)}function bR(i,l,f){var h=l.pendingProps,v=h.children,T=(l.stateNode._pendingVisibility&2)!==0,P=i!==null?i.memoizedState:null;if(Al(i,l),h.mode==="hidden"||T){if(l.flags&128){if(h=P!==null?P.baseLanes|f:f,i!==null){for(v=l.child=i.child,T=0;v!==null;)T=T|v.lanes|v.childLanes,v=v.sibling;l.childLanes=T&~h}else l.childLanes=0,l.child=null;return yR(i,l,h,f)}if(f&536870912)l.memoizedState={baseLanes:0,cachePool:null},i!==null&&_c(l,P!==null?P.cachePool:null),P!==null?v1(l,P):pg(),S1(l);else return l.lanes=l.childLanes=536870912,yR(i,l,P!==null?P.baseLanes|f:f,f)}else P!==null?(_c(l,P.cachePool),v1(l,P),Ja(),l.memoizedState=null):(i!==null&&_c(l,null),pg(),Ja());return gn(i,l,v,f),l.child}function yR(i,l,f,h){var v=bg();return v=v===null?null:{parent:tn._currentValue,pool:v},l.memoizedState={baseLanes:f,cachePool:v},i!==null&&_c(l,null),pg(),S1(l),i!==null&&Rl(i,l,h,!0),null}function Al(i,l){var f=l.ref;if(f===null)i!==null&&i.ref!==null&&(l.flags|=2097664);else{if(typeof f!="function"&&typeof f!="object")throw Error(r(284));(i===null||i.ref!==f)&&(l.flags|=2097664)}}function Pg(i,l,f,h,v){return Zo(l),f=vg(i,l,f,h,void 0,v),h=Sg(),i!==null&&!an?(Eg(i,l,v),Sa(i,l,v)):(Et&&h&&ug(l),l.flags|=1,gn(i,l,f,v),l.child)}function vR(i,l,f,h,v,T){return Zo(l),l.updateQueue=null,f=T1(l,h,f,v),k1(i),h=Sg(),i!==null&&!an?(Eg(i,l,T),Sa(i,l,T)):(Et&&h&&ug(l),l.flags|=1,gn(i,l,f,T),l.child)}function SR(i,l,f,h,v){if(Zo(l),l.stateNode===null){var T=Bi,P=f.contextType;typeof P=="object"&&P!==null&&(T=En(P)),T=new f(h,T),l.memoizedState=T.state!==null&&T.state!==void 0?T.state:null,T.updater=Mg,l.stateNode=T,T._reactInternals=l,T=l.stateNode,T.props=h,T.state=l.memoizedState,T.refs={},Kg(l),P=f.contextType,T.context=typeof P=="object"&&P!==null?En(P):Bi,T.state=l.memoizedState,P=f.getDerivedStateFromProps,typeof P=="function"&&(Lg(l,f,P,h),T.state=l.memoizedState),typeof f.getDerivedStateFromProps=="function"||typeof T.getSnapshotBeforeUpdate=="function"||typeof T.UNSAFE_componentWillMount!="function"&&typeof T.componentWillMount!="function"||(P=T.state,typeof T.componentWillMount=="function"&&T.componentWillMount(),typeof T.UNSAFE_componentWillMount=="function"&&T.UNSAFE_componentWillMount(),P!==T.state&&Mg.enqueueReplaceState(T,T.state,null),Nl(l,h,T,v),Cl(),T.state=l.memoizedState),typeof T.componentDidMount=="function"&&(l.flags|=4194308),h=!0}else if(i===null){T=l.stateNode;var q=l.memoizedProps,X=Ko(f,q);T.props=X;var ne=T.context,be=f.contextType;P=Bi,typeof be=="object"&&be!==null&&(P=En(be));var Ae=f.getDerivedStateFromProps;be=typeof Ae=="function"||typeof T.getSnapshotBeforeUpdate=="function",q=l.pendingProps!==q,be||typeof T.UNSAFE_componentWillReceiveProps!="function"&&typeof T.componentWillReceiveProps!="function"||(q||ne!==P)&&iR(l,T,h,P),ro=!1;var de=l.memoizedState;T.state=de,Nl(l,h,T,v),Cl(),ne=l.memoizedState,q||de!==ne||ro?(typeof Ae=="function"&&(Lg(l,f,Ae,h),ne=l.memoizedState),(X=ro||oR(l,f,X,h,de,ne,P))?(be||typeof T.UNSAFE_componentWillMount!="function"&&typeof T.componentWillMount!="function"||(typeof T.componentWillMount=="function"&&T.componentWillMount(),typeof T.UNSAFE_componentWillMount=="function"&&T.UNSAFE_componentWillMount()),typeof T.componentDidMount=="function"&&(l.flags|=4194308)):(typeof T.componentDidMount=="function"&&(l.flags|=4194308),l.memoizedProps=h,l.memoizedState=ne),T.props=h,T.state=ne,T.context=P,h=X):(typeof T.componentDidMount=="function"&&(l.flags|=4194308),h=!1)}else{T=l.stateNode,Xg(i,l),P=l.memoizedProps,be=Ko(f,P),T.props=be,Ae=l.pendingProps,de=T.context,ne=f.contextType,X=Bi,typeof ne=="object"&&ne!==null&&(X=En(ne)),q=f.getDerivedStateFromProps,(ne=typeof q=="function"||typeof T.getSnapshotBeforeUpdate=="function")||typeof T.UNSAFE_componentWillReceiveProps!="function"&&typeof T.componentWillReceiveProps!="function"||(P!==Ae||de!==X)&&iR(l,T,h,X),ro=!1,de=l.memoizedState,T.state=de,Nl(l,h,T,v),Cl();var ge=l.memoizedState;P!==Ae||de!==ge||ro||i!==null&&i.dependencies!==null&&Bc(i.dependencies)?(typeof q=="function"&&(Lg(l,f,q,h),ge=l.memoizedState),(be=ro||oR(l,f,be,h,de,ge,X)||i!==null&&i.dependencies!==null&&Bc(i.dependencies))?(ne||typeof T.UNSAFE_componentWillUpdate!="function"&&typeof T.componentWillUpdate!="function"||(typeof T.componentWillUpdate=="function"&&T.componentWillUpdate(h,ge,X),typeof T.UNSAFE_componentWillUpdate=="function"&&T.UNSAFE_componentWillUpdate(h,ge,X)),typeof T.componentDidUpdate=="function"&&(l.flags|=4),typeof T.getSnapshotBeforeUpdate=="function"&&(l.flags|=1024)):(typeof T.componentDidUpdate!="function"||P===i.memoizedProps&&de===i.memoizedState||(l.flags|=4),typeof T.getSnapshotBeforeUpdate!="function"||P===i.memoizedProps&&de===i.memoizedState||(l.flags|=1024),l.memoizedProps=h,l.memoizedState=ge),T.props=h,T.state=ge,T.context=X,h=be):(typeof T.componentDidUpdate!="function"||P===i.memoizedProps&&de===i.memoizedState||(l.flags|=4),typeof T.getSnapshotBeforeUpdate!="function"||P===i.memoizedProps&&de===i.memoizedState||(l.flags|=1024),h=!1)}return T=h,Al(i,l),h=(l.flags&128)!==0,T||h?(T=l.stateNode,f=h&&typeof f.getDerivedStateFromError!="function"?null:T.render(),l.flags|=1,i!==null&&h?(l.child=$o(l,i.child,null,v),l.child=$o(l,null,f,v)):gn(i,l,f,v),l.memoizedState=T.state,i=l.child):i=Sa(i,l,v),i}function ER(i,l,f,h){return hl(),l.flags|=256,gn(i,l,f,h),l.child}var zg={dehydrated:null,treeContext:null,retryLane:0};function Bg(i){return{baseLanes:i,cachePool:x1()}}function Ug(i,l,f){return i=i!==null?i.childLanes&~f:0,l&&(i|=mr),i}function wR(i,l,f){var h=l.pendingProps,v=!1,T=(l.flags&128)!==0,P;if((P=T)||(P=i!==null&&i.memoizedState===null?!1:(en.current&2)!==0),P&&(v=!0,l.flags&=-129),P=(l.flags&32)!==0,l.flags&=-33,i===null){if(Et){if(v?Qa(l):Ja(),Et){var q=pn,X;if(X=q){e:{for(X=q,q=qr;X.nodeType!==8;){if(!q){q=null;break e}if(X=Or(X.nextSibling),X===null){q=null;break e}}q=X}q!==null?(l.memoizedState={dehydrated:q,treeContext:jo!==null?{id:ma,overflow:ba}:null,retryLane:536870912},X=hr(18,null,null,0),X.stateNode=q,X.return=l,l.child=X,Cn=l,pn=null,X=!0):X=!1}X||Ho(l)}if(q=l.memoizedState,q!==null&&(q=q.dehydrated,q!==null))return q.data==="$!"?l.lanes=16:l.lanes=536870912,null;ya(l)}return q=h.children,h=h.fallback,v?(Ja(),v=l.mode,q=Gg({mode:"hidden",children:q},v),h=Jo(h,v,f,null),q.return=l,h.return=l,q.sibling=h,l.child=q,v=l.child,v.memoizedState=Bg(f),v.childLanes=Ug(i,P,f),l.memoizedState=zg,h):(Qa(l),jg(l,q))}if(X=i.memoizedState,X!==null&&(q=X.dehydrated,q!==null)){if(T)l.flags&256?(Qa(l),l.flags&=-257,l=Hg(i,l,f)):l.memoizedState!==null?(Ja(),l.child=i.child,l.flags|=128,l=null):(Ja(),v=h.fallback,q=l.mode,h=Gg({mode:"visible",children:h.children},q),v=Jo(v,q,f,null),v.flags|=2,h.return=l,v.return=l,h.sibling=v,l.child=h,$o(l,i.child,null,f),h=l.child,h.memoizedState=Bg(f),h.childLanes=Ug(i,P,f),l.memoizedState=zg,l=v);else if(Qa(l),q.data==="$!"){if(P=q.nextSibling&&q.nextSibling.dataset,P)var ne=P.dgst;P=ne,h=Error(r(419)),h.stack="",h.digest=P,ml({value:h,source:null,stack:null}),l=Hg(i,l,f)}else if(an||Rl(i,l,f,!1),P=(f&i.childLanes)!==0,an||P){if(P=It,P!==null){if(h=f&-f,h&42)h=1;else switch(h){case 2:h=1;break;case 8:h=4;break;case 32:h=16;break;case 128:case 256:case 512:case 1024:case 2048:case 4096:case 8192:case 16384:case 32768:case 65536:case 131072:case 262144:case 524288:case 1048576:case 2097152:case 4194304:case 8388608:case 16777216:case 33554432:h=64;break;case 268435456:h=134217728;break;default:h=0}if(h=h&(P.suspendedLanes|f)?0:h,h!==0&&h!==X.retryLane)throw X.retryLane=h,Za(i,h),Nn(P,i,h),pR}q.data==="$?"||bh(),l=Hg(i,l,f)}else q.data==="$?"?(l.flags|=128,l.child=i.child,l=E6.bind(null,i),q._reactRetry=l,l=null):(i=X.treeContext,pn=Or(q.nextSibling),Cn=l,Et=!0,Cr=null,qr=!1,i!==null&&(cr[dr++]=ma,cr[dr++]=ba,cr[dr++]=jo,ma=i.id,ba=i.overflow,jo=l),l=jg(l,h.children),l.flags|=4096);return l}return v?(Ja(),v=h.fallback,q=l.mode,X=i.child,ne=X.sibling,h=uo(X,{mode:"hidden",children:h.children}),h.subtreeFlags=X.subtreeFlags&31457280,ne!==null?v=uo(ne,v):(v=Jo(v,q,f,null),v.flags|=2),v.return=l,h.return=l,h.sibling=v,l.child=h,h=v,v=l.child,q=i.child.memoizedState,q===null?q=Bg(f):(X=q.cachePool,X!==null?(ne=tn._currentValue,X=X.parent!==ne?{parent:ne,pool:ne}:X):X=x1(),q={baseLanes:q.baseLanes|f,cachePool:X}),v.memoizedState=q,v.childLanes=Ug(i,P,f),l.memoizedState=zg,h):(Qa(l),f=i.child,i=f.sibling,f=uo(f,{mode:"visible",children:h.children}),f.return=l,f.sibling=null,i!==null&&(P=l.deletions,P===null?(l.deletions=[i],l.flags|=16):P.push(i)),l.child=f,l.memoizedState=null,f)}function jg(i,l){return l=Gg({mode:"visible",children:l},i.mode),l.return=i,i.child=l}function Gg(i,l){return YR(i,l,0,null)}function Hg(i,l,f){return $o(l,i.child,null,f),i=jg(l,l.pendingProps.children),i.flags|=2,l.memoizedState=null,i}function xR(i,l,f){i.lanes|=l;var h=i.alternate;h!==null&&(h.lanes|=l),Wg(i.return,l,f)}function $g(i,l,f,h,v){var T=i.memoizedState;T===null?i.memoizedState={isBackwards:l,rendering:null,renderingStartTime:0,last:h,tail:f,tailMode:v}:(T.isBackwards=l,T.rendering=null,T.renderingStartTime=0,T.last=h,T.tail=f,T.tailMode=v)}function kR(i,l,f){var h=l.pendingProps,v=h.revealOrder,T=h.tail;if(gn(i,l,h.children,f),h=en.current,h&2)h=h&1|2,l.flags|=128;else{if(i!==null&&i.flags&128)e:for(i=l.child;i!==null;){if(i.tag===13)i.memoizedState!==null&&xR(i,f,l);else if(i.tag===19)xR(i,f,l);else if(i.child!==null){i.child.return=i,i=i.child;continue}if(i===l)break e;for(;i.sibling===null;){if(i.return===null||i.return===l)break e;i=i.return}i.sibling.return=i.return,i=i.sibling}h&=1}switch(ae(en,h),v){case"forwards":for(f=l.child,v=null;f!==null;)i=f.alternate,i!==null&&Rc(i)===null&&(v=f),f=f.sibling;f=v,f===null?(v=l.child,l.child=null):(v=f.sibling,f.sibling=null),$g(l,!1,v,f,T);break;case"backwards":for(f=null,v=l.child,l.child=null;v!==null;){if(i=v.alternate,i!==null&&Rc(i)===null){l.child=v;break}i=v.sibling,v.sibling=f,f=v,v=i}$g(l,!0,f,null,T);break;case"together":$g(l,!1,null,null,void 0);break;default:l.memoizedState=null}return l.child}function Sa(i,l,f){if(i!==null&&(l.dependencies=i.dependencies),co|=l.lanes,!(f&l.childLanes))if(i!==null){if(Rl(i,l,f,!1),(f&l.childLanes)===0)return null}else return null;if(i!==null&&l.child!==i.child)throw Error(r(153));if(l.child!==null){for(i=l.child,f=uo(i,i.pendingProps),l.child=f,f.return=l;i.sibling!==null;)i=i.sibling,f=f.sibling=uo(i,i.pendingProps),f.return=l;f.sibling=null}return l.child}function qg(i,l){return i.lanes&l?!0:(i=i.dependencies,!!(i!==null&&Bc(i)))}function s6(i,l,f){switch(l.tag){case 3:ke(l,l.stateNode.containerInfo),no(l,tn,i.memoizedState.cache),hl();break;case 27:case 5:fe(l);break;case 4:ke(l,l.stateNode.containerInfo);break;case 10:no(l,l.type,l.memoizedProps.value);break;case 13:var h=l.memoizedState;if(h!==null)return h.dehydrated!==null?(Qa(l),l.flags|=128,null):f&l.child.childLanes?wR(i,l,f):(Qa(l),i=Sa(i,l,f),i!==null?i.sibling:null);Qa(l);break;case 19:var v=(i.flags&128)!==0;if(h=(f&l.childLanes)!==0,h||(Rl(i,l,f,!1),h=(f&l.childLanes)!==0),v){if(h)return kR(i,l,f);l.flags|=128}if(v=l.memoizedState,v!==null&&(v.rendering=null,v.tail=null,v.lastEffect=null),ae(en,en.current),h)break;return null;case 22:case 23:return l.lanes=0,bR(i,l,f);case 24:no(l,tn,i.memoizedState.cache)}return Sa(i,l,f)}function TR(i,l,f){if(i!==null)if(i.memoizedProps!==l.pendingProps)an=!0;else{if(!qg(i,f)&&!(l.flags&128))return an=!1,s6(i,l,f);an=!!(i.flags&131072)}else an=!1,Et&&l.flags&1048576&&u1(l,wc,l.index);switch(l.lanes=0,l.tag){case 16:e:{i=l.pendingProps;var h=l.elementType,v=h._init;if(h=v(h._payload),l.type=h,typeof h=="function")oh(h)?(i=Ko(h,i),l.tag=1,l=SR(null,l,h,i,f)):(l.tag=0,l=Pg(null,l,h,i,f));else{if(h!=null){if(v=h.$$typeof,v===y){l.tag=11,l=gR(null,l,h,i,f);break e}else if(v===R){l.tag=14,l=hR(null,l,h,i,f);break e}}throw l=F(h)||h,Error(r(306,l,""))}}return l;case 0:return Pg(i,l,l.type,l.pendingProps,f);case 1:return h=l.type,v=Ko(h,l.pendingProps),SR(i,l,h,v,f);case 3:e:{if(ke(l,l.stateNode.containerInfo),i===null)throw Error(r(387));var T=l.pendingProps;v=l.memoizedState,h=v.element,Xg(i,l),Nl(l,T,null,f);var P=l.memoizedState;if(T=P.cache,no(l,tn,T),T!==v.cache&&Yg(l,[tn],f,!0),Cl(),T=P.element,v.isDehydrated)if(v={element:T,isDehydrated:!1,cache:P.cache},l.updateQueue.baseState=v,l.memoizedState=v,l.flags&256){l=ER(i,l,T,f);break e}else if(T!==h){h=ur(Error(r(424)),l),ml(h),l=ER(i,l,T,f);break e}else for(pn=Or(l.stateNode.containerInfo.firstChild),Cn=l,Et=!0,Cr=null,qr=!0,f=y1(l,null,T,f),l.child=f;f;)f.flags=f.flags&-3|4096,f=f.sibling;else{if(hl(),T===h){l=Sa(i,l,f);break e}gn(i,l,T,f)}l=l.child}return l;case 26:return Al(i,l),i===null?(f=__(l.type,null,l.pendingProps,null))?l.memoizedState=f:Et||(f=l.type,i=l.pendingProps,h=td(Ie.current).createElement(f),h[Sn]=l,h[Mn]=i,hn(h,f,i),rn(h),l.stateNode=h):l.memoizedState=__(l.type,i.memoizedProps,l.pendingProps,i.memoizedState),null;case 27:return fe(l),i===null&&Et&&(h=l.stateNode=T_(l.type,l.pendingProps,Ie.current),Cn=l,qr=!0,pn=Or(h.firstChild)),h=l.pendingProps.children,i!==null||Et?gn(i,l,h,f):l.child=$o(l,null,h,f),Al(i,l),l.child;case 5:return i===null&&Et&&((v=h=pn)&&(h=P6(h,l.type,l.pendingProps,qr),h!==null?(l.stateNode=h,Cn=l,pn=Or(h.firstChild),qr=!1,v=!0):v=!1),v||Ho(l)),fe(l),v=l.type,T=l.pendingProps,P=i!==null?i.memoizedProps:null,h=T.children,Oh(v,T)?h=null:P!==null&&Oh(v,P)&&(l.flags|=32),l.memoizedState!==null&&(v=vg(i,l,e6,null,null,f),ql._currentValue=v),Al(i,l),gn(i,l,h,f),l.child;case 6:return i===null&&Et&&((i=f=pn)&&(f=z6(f,l.pendingProps,qr),f!==null?(l.stateNode=f,Cn=l,pn=null,i=!0):i=!1),i||Ho(l)),null;case 13:return wR(i,l,f);case 4:return ke(l,l.stateNode.containerInfo),h=l.pendingProps,i===null?l.child=$o(l,null,h,f):gn(i,l,h,f),l.child;case 11:return gR(i,l,l.type,l.pendingProps,f);case 7:return gn(i,l,l.pendingProps,f),l.child;case 8:return gn(i,l,l.pendingProps.children,f),l.child;case 12:return gn(i,l,l.pendingProps.children,f),l.child;case 10:return h=l.pendingProps,no(l,l.type,h.value),gn(i,l,h.children,f),l.child;case 9:return v=l.type._context,h=l.pendingProps.children,Zo(l),v=En(v),h=h(v),l.flags|=1,gn(i,l,h,f),l.child;case 14:return hR(i,l,l.type,l.pendingProps,f);case 15:return mR(i,l,l.type,l.pendingProps,f);case 19:return kR(i,l,f);case 22:return bR(i,l,f);case 24:return Zo(l),h=En(tn),i===null?(v=bg(),v===null&&(v=It,T=hg(),v.pooledCache=T,T.refCount++,T!==null&&(v.pooledCacheLanes|=f),v=T),l.memoizedState={parent:h,cache:v},Kg(l),no(l,tn,v)):(i.lanes&f&&(Xg(i,l),Nl(l,null,null,f),Cl()),v=i.memoizedState,T=l.memoizedState,v.parent!==h?(v={parent:h,cache:h},l.memoizedState=v,l.lanes===0&&(l.memoizedState=l.updateQueue.baseState=v),no(l,tn,h)):(h=T.cache,no(l,tn,h),h!==v.cache&&Yg(l,[tn],f,!0))),gn(i,l,l.pendingProps.children,f),l.child;case 29:throw l.pendingProps}throw Error(r(156,l.tag))}var Vg=ie(null),Xo=null,Ea=null;function no(i,l,f){ae(Vg,l._currentValue),l._currentValue=f}function wa(i){i._currentValue=Vg.current,Se(Vg)}function Wg(i,l,f){for(;i!==null;){var h=i.alternate;if((i.childLanes&l)!==l?(i.childLanes|=l,h!==null&&(h.childLanes|=l)):h!==null&&(h.childLanes&l)!==l&&(h.childLanes|=l),i===f)break;i=i.return}}function Yg(i,l,f,h){var v=i.child;for(v!==null&&(v.return=i);v!==null;){var T=v.dependencies;if(T!==null){var P=v.child;T=T.firstContext;e:for(;T!==null;){var q=T;T=v;for(var X=0;X<l.length;X++)if(q.context===l[X]){T.lanes|=f,q=T.alternate,q!==null&&(q.lanes|=f),Wg(T.return,f,i),h||(P=null);break e}T=q.next}}else if(v.tag===18){if(P=v.return,P===null)throw Error(r(341));P.lanes|=f,T=P.alternate,T!==null&&(T.lanes|=f),Wg(P,f,i),P=null}else P=v.child;if(P!==null)P.return=v;else for(P=v;P!==null;){if(P===i){P=null;break}if(v=P.sibling,v!==null){v.return=P.return,P=v;break}P=P.return}v=P}}function Rl(i,l,f,h){i=null;for(var v=l,T=!1;v!==null;){if(!T){if(v.flags&524288)T=!0;else if(v.flags&262144)break}if(v.tag===10){var P=v.alternate;if(P===null)throw Error(r(387));if(P=P.memoizedProps,P!==null){var q=v.type;Yn(v.pendingProps.value,P.value)||(i!==null?i.push(q):i=[q])}}else if(v===Ce.current){if(P=v.alternate,P===null)throw Error(r(387));P.memoizedState.memoizedState!==v.memoizedState.memoizedState&&(i!==null?i.push(ql):i=[ql])}v=v.return}i!==null&&Yg(l,i,f,h),l.flags|=262144}function Bc(i){for(i=i.firstContext;i!==null;){if(!Yn(i.context._currentValue,i.memoizedValue))return!0;i=i.next}return!1}function Zo(i){Xo=i,Ea=null,i=i.dependencies,i!==null&&(i.firstContext=null)}function En(i){return AR(Xo,i)}function Uc(i,l){return Xo===null&&Zo(i),AR(i,l)}function AR(i,l){var f=l._currentValue;if(l={context:l,memoizedValue:f,next:null},Ea===null){if(i===null)throw Error(r(308));Ea=l,i.dependencies={lanes:0,firstContext:l},i.flags|=524288}else Ea=Ea.next=l;return f}var ro=!1;function Kg(i){i.updateQueue={baseState:i.memoizedState,firstBaseUpdate:null,lastBaseUpdate:null,shared:{pending:null,lanes:0,hiddenCallbacks:null},callbacks:null}}function Xg(i,l){i=i.updateQueue,l.updateQueue===i&&(l.updateQueue={baseState:i.baseState,firstBaseUpdate:i.firstBaseUpdate,lastBaseUpdate:i.lastBaseUpdate,shared:i.shared,callbacks:null})}function ao(i){return{lane:i,tag:0,payload:null,callback:null,next:null}}function oo(i,l,f){var h=i.updateQueue;if(h===null)return null;if(h=h.shared,Ut&2){var v=h.pending;return v===null?l.next=l:(l.next=v.next,v.next=l),h.pending=l,l=Sc(i),s1(i,null,f),l}return vc(i,h,l,f),Sc(i)}function _l(i,l,f){if(l=l.updateQueue,l!==null&&(l=l.shared,(f&4194176)!==0)){var h=l.lanes;h&=i.pendingLanes,f|=h,l.lanes=f,Rr(i,f)}}function Zg(i,l){var f=i.updateQueue,h=i.alternate;if(h!==null&&(h=h.updateQueue,f===h)){var v=null,T=null;if(f=f.firstBaseUpdate,f!==null){do{var P={lane:f.lane,tag:f.tag,payload:f.payload,callback:null,next:null};T===null?v=T=P:T=T.next=P,f=f.next}while(f!==null);T===null?v=T=l:T=T.next=l}else v=T=l;f={baseState:h.baseState,firstBaseUpdate:v,lastBaseUpdate:T,shared:h.shared,callbacks:h.callbacks},i.updateQueue=f;return}i=f.lastBaseUpdate,i===null?f.firstBaseUpdate=l:i.next=l,f.lastBaseUpdate=l}var Qg=!1;function Cl(){if(Qg){var i=qi;if(i!==null)throw i}}function Nl(i,l,f,h){Qg=!1;var v=i.updateQueue;ro=!1;var T=v.firstBaseUpdate,P=v.lastBaseUpdate,q=v.shared.pending;if(q!==null){v.shared.pending=null;var X=q,ne=X.next;X.next=null,P===null?T=ne:P.next=ne,P=X;var be=i.alternate;be!==null&&(be=be.updateQueue,q=be.lastBaseUpdate,q!==P&&(q===null?be.firstBaseUpdate=ne:q.next=ne,be.lastBaseUpdate=X))}if(T!==null){var Ae=v.baseState;P=0,be=ne=X=null,q=T;do{var de=q.lane&-536870913,ge=de!==q.lane;if(ge?(bt&de)===de:(h&de)===de){de!==0&&de===$i&&(Qg=!0),be!==null&&(be=be.next={lane:0,tag:q.tag,payload:q.payload,callback:null,next:null});e:{var qe=i,nt=q;de=l;var Vt=f;switch(nt.tag){case 1:if(qe=nt.payload,typeof qe=="function"){Ae=qe.call(Vt,Ae,de);break e}Ae=qe;break e;case 3:qe.flags=qe.flags&-65537|128;case 0:if(qe=nt.payload,de=typeof qe=="function"?qe.call(Vt,Ae,de):qe,de==null)break e;Ae=I({},Ae,de);break e;case 2:ro=!0}}de=q.callback,de!==null&&(i.flags|=64,ge&&(i.flags|=8192),ge=v.callbacks,ge===null?v.callbacks=[de]:ge.push(de))}else ge={lane:de,tag:q.tag,payload:q.payload,callback:q.callback,next:null},be===null?(ne=be=ge,X=Ae):be=be.next=ge,P|=de;if(q=q.next,q===null){if(q=v.shared.pending,q===null)break;ge=q,q=ge.next,ge.next=null,v.lastBaseUpdate=ge,v.shared.pending=null}}while(!0);be===null&&(X=Ae),v.baseState=X,v.firstBaseUpdate=ne,v.lastBaseUpdate=be,T===null&&(v.shared.lanes=0),co|=P,i.lanes=P,i.memoizedState=Ae}}function RR(i,l){if(typeof i!="function")throw Error(r(191,i));i.call(l)}function _R(i,l){var f=i.callbacks;if(f!==null)for(i.callbacks=null,i=0;i<f.length;i++)RR(f[i],l)}function Ol(i,l){try{var f=l.updateQueue,h=f!==null?f.lastEffect:null;if(h!==null){var v=h.next;f=v;do{if((f.tag&i)===i){h=void 0;var T=f.create,P=f.inst;h=T(),P.destroy=h}f=f.next}while(f!==v)}}catch(q){Nt(l,l.return,q)}}function io(i,l,f){try{var h=l.updateQueue,v=h!==null?h.lastEffect:null;if(v!==null){var T=v.next;h=T;do{if((h.tag&i)===i){var P=h.inst,q=P.destroy;if(q!==void 0){P.destroy=void 0,v=l;var X=f;try{q()}catch(ne){Nt(v,X,ne)}}}h=h.next}while(h!==T)}}catch(ne){Nt(l,l.return,ne)}}function CR(i){var l=i.updateQueue;if(l!==null){var f=i.stateNode;try{_R(l,f)}catch(h){Nt(i,i.return,h)}}}function NR(i,l,f){f.props=Ko(i.type,i.memoizedProps),f.state=i.memoizedState;try{f.componentWillUnmount()}catch(h){Nt(i,l,h)}}function Qo(i,l){try{var f=i.ref;if(f!==null){var h=i.stateNode;switch(i.tag){case 26:case 27:case 5:var v=h;break;default:v=h}typeof f=="function"?i.refCleanup=f(v):f.current=v}}catch(T){Nt(i,l,T)}}function Kn(i,l){var f=i.ref,h=i.refCleanup;if(f!==null)if(typeof h=="function")try{h()}catch(v){Nt(i,l,v)}finally{i.refCleanup=null,i=i.alternate,i!=null&&(i.refCleanup=null)}else if(typeof f=="function")try{f(null)}catch(v){Nt(i,l,v)}else f.current=null}function OR(i){var l=i.type,f=i.memoizedProps,h=i.stateNode;try{e:switch(l){case"button":case"input":case"select":case"textarea":f.autoFocus&&h.focus();break e;case"img":f.src?h.src=f.src:f.srcSet&&(h.srcset=f.srcSet)}}catch(v){Nt(i,i.return,v)}}function DR(i,l,f){try{var h=i.stateNode;D6(h,i.type,f,l),h[Mn]=l}catch(v){Nt(i,i.return,v)}}function IR(i){return i.tag===5||i.tag===3||i.tag===26||i.tag===27||i.tag===4}function Jg(i){e:for(;;){for(;i.sibling===null;){if(i.return===null||IR(i.return))return null;i=i.return}for(i.sibling.return=i.return,i=i.sibling;i.tag!==5&&i.tag!==6&&i.tag!==27&&i.tag!==18;){if(i.flags&2||i.child===null||i.tag===4)continue e;i.child.return=i,i=i.child}if(!(i.flags&2))return i.stateNode}}function eh(i,l,f){var h=i.tag;if(h===5||h===6)i=i.stateNode,l?f.nodeType===8?f.parentNode.insertBefore(i,l):f.insertBefore(i,l):(f.nodeType===8?(l=f.parentNode,l.insertBefore(i,f)):(l=f,l.appendChild(i)),f=f._reactRootContainer,f!=null||l.onclick!==null||(l.onclick=ed));else if(h!==4&&h!==27&&(i=i.child,i!==null))for(eh(i,l,f),i=i.sibling;i!==null;)eh(i,l,f),i=i.sibling}function jc(i,l,f){var h=i.tag;if(h===5||h===6)i=i.stateNode,l?f.insertBefore(i,l):f.appendChild(i);else if(h!==4&&h!==27&&(i=i.child,i!==null))for(jc(i,l,f),i=i.sibling;i!==null;)jc(i,l,f),i=i.sibling}var xa=!1,$t=!1,th=!1,LR=typeof WeakSet=="function"?WeakSet:Set,on=null,MR=!1;function l6(i,l){if(i=i.containerInfo,Ch=sd,i=QA(i),rg(i)){if("selectionStart"in i)var f={start:i.selectionStart,end:i.selectionEnd};else e:{f=(f=i.ownerDocument)&&f.defaultView||window;var h=f.getSelection&&f.getSelection();if(h&&h.rangeCount!==0){f=h.anchorNode;var v=h.anchorOffset,T=h.focusNode;h=h.focusOffset;try{f.nodeType,T.nodeType}catch{f=null;break e}var P=0,q=-1,X=-1,ne=0,be=0,Ae=i,de=null;t:for(;;){for(var ge;Ae!==f||v!==0&&Ae.nodeType!==3||(q=P+v),Ae!==T||h!==0&&Ae.nodeType!==3||(X=P+h),Ae.nodeType===3&&(P+=Ae.nodeValue.length),(ge=Ae.firstChild)!==null;)de=Ae,Ae=ge;for(;;){if(Ae===i)break t;if(de===f&&++ne===v&&(q=P),de===T&&++be===h&&(X=P),(ge=Ae.nextSibling)!==null)break;Ae=de,de=Ae.parentNode}Ae=ge}f=q===-1||X===-1?null:{start:q,end:X}}else f=null}f=f||{start:0,end:0}}else f=null;for(Nh={focusedElem:i,selectionRange:f},sd=!1,on=l;on!==null;)if(l=on,i=l.child,(l.subtreeFlags&1028)!==0&&i!==null)i.return=l,on=i;else for(;on!==null;){switch(l=on,T=l.alternate,i=l.flags,l.tag){case 0:break;case 11:case 15:break;case 1:if(i&1024&&T!==null){i=void 0,f=l,v=T.memoizedProps,T=T.memoizedState,h=f.stateNode;try{var qe=Ko(f.type,v,f.elementType===f.type);i=h.getSnapshotBeforeUpdate(qe,T),h.__reactInternalSnapshotBeforeUpdate=i}catch(nt){Nt(f,f.return,nt)}}break;case 3:if(i&1024){if(i=l.stateNode.containerInfo,f=i.nodeType,f===9)Lh(i);else if(f===1)switch(i.nodeName){case"HEAD":case"HTML":case"BODY":Lh(i);break;default:i.textContent=""}}break;case 5:case 26:case 27:case 6:case 4:case 17:break;default:if(i&1024)throw Error(r(163))}if(i=l.sibling,i!==null){i.return=l.return,on=i;break}on=l.return}return qe=MR,MR=!1,qe}function FR(i,l,f){var h=f.flags;switch(f.tag){case 0:case 11:case 15:Ta(i,f),h&4&&Ol(5,f);break;case 1:if(Ta(i,f),h&4)if(i=f.stateNode,l===null)try{i.componentDidMount()}catch(q){Nt(f,f.return,q)}else{var v=Ko(f.type,l.memoizedProps);l=l.memoizedState;try{i.componentDidUpdate(v,l,i.__reactInternalSnapshotBeforeUpdate)}catch(q){Nt(f,f.return,q)}}h&64&&CR(f),h&512&&Qo(f,f.return);break;case 3:if(Ta(i,f),h&64&&(h=f.updateQueue,h!==null)){if(i=null,f.child!==null)switch(f.child.tag){case 27:case 5:i=f.child.stateNode;break;case 1:i=f.child.stateNode}try{_R(h,i)}catch(q){Nt(f,f.return,q)}}break;case 26:Ta(i,f),h&512&&Qo(f,f.return);break;case 27:case 5:Ta(i,f),l===null&&h&4&&OR(f),h&512&&Qo(f,f.return);break;case 12:Ta(i,f);break;case 13:Ta(i,f),h&4&&BR(i,f);break;case 22:if(v=f.memoizedState!==null||xa,!v){l=l!==null&&l.memoizedState!==null||$t;var T=xa,P=$t;xa=v,($t=l)&&!P?so(i,f,(f.subtreeFlags&8772)!==0):Ta(i,f),xa=T,$t=P}h&512&&(f.memoizedProps.mode==="manual"?Qo(f,f.return):Kn(f,f.return));break;default:Ta(i,f)}}function PR(i){var l=i.alternate;l!==null&&(i.alternate=null,PR(l)),i.child=null,i.deletions=null,i.sibling=null,i.tag===5&&(l=i.stateNode,l!==null&&jp(l)),i.stateNode=null,i.return=null,i.dependencies=null,i.memoizedProps=null,i.memoizedState=null,i.pendingProps=null,i.stateNode=null,i.updateQueue=null}var Zt=null,Xn=!1;function ka(i,l,f){for(f=f.child;f!==null;)zR(i,l,f),f=f.sibling}function zR(i,l,f){if(et&&typeof et.onCommitFiberUnmount=="function")try{et.onCommitFiberUnmount(ht,f)}catch{}switch(f.tag){case 26:$t||Kn(f,l),ka(i,l,f),f.memoizedState?f.memoizedState.count--:f.stateNode&&(f=f.stateNode,f.parentNode.removeChild(f));break;case 27:$t||Kn(f,l);var h=Zt,v=Xn;for(Zt=f.stateNode,ka(i,l,f),f=f.stateNode,l=f.attributes;l.length;)f.removeAttributeNode(l[0]);jp(f),Zt=h,Xn=v;break;case 5:$t||Kn(f,l);case 6:v=Zt;var T=Xn;if(Zt=null,ka(i,l,f),Zt=v,Xn=T,Zt!==null)if(Xn)try{i=Zt,h=f.stateNode,i.nodeType===8?i.parentNode.removeChild(h):i.removeChild(h)}catch(P){Nt(f,l,P)}else try{Zt.removeChild(f.stateNode)}catch(P){Nt(f,l,P)}break;case 18:Zt!==null&&(Xn?(l=Zt,f=f.stateNode,l.nodeType===8?Ih(l.parentNode,f):l.nodeType===1&&Ih(l,f),Kl(l)):Ih(Zt,f.stateNode));break;case 4:h=Zt,v=Xn,Zt=f.stateNode.containerInfo,Xn=!0,ka(i,l,f),Zt=h,Xn=v;break;case 0:case 11:case 14:case 15:$t||io(2,f,l),$t||io(4,f,l),ka(i,l,f);break;case 1:$t||(Kn(f,l),h=f.stateNode,typeof h.componentWillUnmount=="function"&&NR(f,l,h)),ka(i,l,f);break;case 21:ka(i,l,f);break;case 22:$t||Kn(f,l),$t=(h=$t)||f.memoizedState!==null,ka(i,l,f),$t=h;break;default:ka(i,l,f)}}function BR(i,l){if(l.memoizedState===null&&(i=l.alternate,i!==null&&(i=i.memoizedState,i!==null&&(i=i.dehydrated,i!==null))))try{Kl(i)}catch(f){Nt(l,l.return,f)}}function u6(i){switch(i.tag){case 13:case 19:var l=i.stateNode;return l===null&&(l=i.stateNode=new LR),l;case 22:return i=i.stateNode,l=i._retryCache,l===null&&(l=i._retryCache=new LR),l;default:throw Error(r(435,i.tag))}}function nh(i,l){var f=u6(i);l.forEach(function(h){var v=w6.bind(null,i,h);f.has(h)||(f.add(h),h.then(v,v))})}function pr(i,l){var f=l.deletions;if(f!==null)for(var h=0;h<f.length;h++){var v=f[h],T=i,P=l,q=P;e:for(;q!==null;){switch(q.tag){case 27:case 5:Zt=q.stateNode,Xn=!1;break e;case 3:Zt=q.stateNode.containerInfo,Xn=!0;break e;case 4:Zt=q.stateNode.containerInfo,Xn=!0;break e}q=q.return}if(Zt===null)throw Error(r(160));zR(T,P,v),Zt=null,Xn=!1,T=v.alternate,T!==null&&(T.return=null),v.return=null}if(l.subtreeFlags&13878)for(l=l.child;l!==null;)UR(l,i),l=l.sibling}var Nr=null;function UR(i,l){var f=i.alternate,h=i.flags;switch(i.tag){case 0:case 11:case 14:case 15:pr(l,i),gr(i),h&4&&(io(3,i,i.return),Ol(3,i),io(5,i,i.return));break;case 1:pr(l,i),gr(i),h&512&&($t||f===null||Kn(f,f.return)),h&64&&xa&&(i=i.updateQueue,i!==null&&(h=i.callbacks,h!==null&&(f=i.shared.hiddenCallbacks,i.shared.hiddenCallbacks=f===null?h:f.concat(h))));break;case 26:var v=Nr;if(pr(l,i),gr(i),h&512&&($t||f===null||Kn(f,f.return)),h&4){var T=f!==null?f.memoizedState:null;if(h=i.memoizedState,f===null)if(h===null)if(i.stateNode===null){e:{h=i.type,f=i.memoizedProps,v=v.ownerDocument||v;t:switch(h){case"title":T=v.getElementsByTagName("title")[0],(!T||T[rl]||T[Sn]||T.namespaceURI==="http://www.w3.org/2000/svg"||T.hasAttribute("itemprop"))&&(T=v.createElement(h),v.head.insertBefore(T,v.querySelector("head > title"))),hn(T,h,f),T[Sn]=i,rn(T),h=T;break e;case"link":var P=O_("link","href",v).get(h+(f.href||""));if(P){for(var q=0;q<P.length;q++)if(T=P[q],T.getAttribute("href")===(f.href==null?null:f.href)&&T.getAttribute("rel")===(f.rel==null?null:f.rel)&&T.getAttribute("title")===(f.title==null?null:f.title)&&T.getAttribute("crossorigin")===(f.crossOrigin==null?null:f.crossOrigin)){P.splice(q,1);break t}}T=v.createElement(h),hn(T,h,f),v.head.appendChild(T);break;case"meta":if(P=O_("meta","content",v).get(h+(f.content||""))){for(q=0;q<P.length;q++)if(T=P[q],T.getAttribute("content")===(f.content==null?null:""+f.content)&&T.getAttribute("name")===(f.name==null?null:f.name)&&T.getAttribute("property")===(f.property==null?null:f.property)&&T.getAttribute("http-equiv")===(f.httpEquiv==null?null:f.httpEquiv)&&T.getAttribute("charset")===(f.charSet==null?null:f.charSet)){P.splice(q,1);break t}}T=v.createElement(h),hn(T,h,f),v.head.appendChild(T);break;default:throw Error(r(468,h))}T[Sn]=i,rn(T),h=T}i.stateNode=h}else D_(v,i.type,i.stateNode);else i.stateNode=N_(v,h,i.memoizedProps);else T!==h?(T===null?f.stateNode!==null&&(f=f.stateNode,f.parentNode.removeChild(f)):T.count--,h===null?D_(v,i.type,i.stateNode):N_(v,h,i.memoizedProps)):h===null&&i.stateNode!==null&&DR(i,i.memoizedProps,f.memoizedProps)}break;case 27:if(h&4&&i.alternate===null){v=i.stateNode,T=i.memoizedProps;try{for(var X=v.firstChild;X;){var ne=X.nextSibling,be=X.nodeName;X[rl]||be==="HEAD"||be==="BODY"||be==="SCRIPT"||be==="STYLE"||be==="LINK"&&X.rel.toLowerCase()==="stylesheet"||v.removeChild(X),X=ne}for(var Ae=i.type,de=v.attributes;de.length;)v.removeAttributeNode(de[0]);hn(v,Ae,T),v[Sn]=i,v[Mn]=T}catch(qe){Nt(i,i.return,qe)}}case 5:if(pr(l,i),gr(i),h&512&&($t||f===null||Kn(f,f.return)),i.flags&32){v=i.stateNode;try{Di(v,"")}catch(qe){Nt(i,i.return,qe)}}h&4&&i.stateNode!=null&&(v=i.memoizedProps,DR(i,v,f!==null?f.memoizedProps:v)),h&1024&&(th=!0);break;case 6:if(pr(l,i),gr(i),h&4){if(i.stateNode===null)throw Error(r(162));h=i.memoizedProps,f=i.stateNode;try{f.nodeValue=h}catch(qe){Nt(i,i.return,qe)}}break;case 3:if(ad=null,v=Nr,Nr=nd(l.containerInfo),pr(l,i),Nr=v,gr(i),h&4&&f!==null&&f.memoizedState.isDehydrated)try{Kl(l.containerInfo)}catch(qe){Nt(i,i.return,qe)}th&&(th=!1,jR(i));break;case 4:h=Nr,Nr=nd(i.stateNode.containerInfo),pr(l,i),gr(i),Nr=h;break;case 12:pr(l,i),gr(i);break;case 13:pr(l,i),gr(i),i.child.flags&8192&&i.memoizedState!==null!=(f!==null&&f.memoizedState!==null)&&(dh=he()),h&4&&(h=i.updateQueue,h!==null&&(i.updateQueue=null,nh(i,h)));break;case 22:if(h&512&&($t||f===null||Kn(f,f.return)),X=i.memoizedState!==null,ne=f!==null&&f.memoizedState!==null,be=xa,Ae=$t,xa=be||X,$t=Ae||ne,pr(l,i),$t=Ae,xa=be,gr(i),l=i.stateNode,l._current=i,l._visibility&=-3,l._visibility|=l._pendingVisibility&2,h&8192&&(l._visibility=X?l._visibility&-2:l._visibility|1,X&&(l=xa||$t,f===null||ne||l||Ki(i)),i.memoizedProps===null||i.memoizedProps.mode!=="manual"))e:for(f=null,l=i;;){if(l.tag===5||l.tag===26||l.tag===27){if(f===null){ne=f=l;try{if(v=ne.stateNode,X)T=v.style,typeof T.setProperty=="function"?T.setProperty("display","none","important"):T.display="none";else{P=ne.stateNode,q=ne.memoizedProps.style;var ge=q!=null&&q.hasOwnProperty("display")?q.display:null;P.style.display=ge==null||typeof ge=="boolean"?"":(""+ge).trim()}}catch(qe){Nt(ne,ne.return,qe)}}}else if(l.tag===6){if(f===null){ne=l;try{ne.stateNode.nodeValue=X?"":ne.memoizedProps}catch(qe){Nt(ne,ne.return,qe)}}}else if((l.tag!==22&&l.tag!==23||l.memoizedState===null||l===i)&&l.child!==null){l.child.return=l,l=l.child;continue}if(l===i)break e;for(;l.sibling===null;){if(l.return===null||l.return===i)break e;f===l&&(f=null),l=l.return}f===l&&(f=null),l.sibling.return=l.return,l=l.sibling}h&4&&(h=i.updateQueue,h!==null&&(f=h.retryQueue,f!==null&&(h.retryQueue=null,nh(i,f))));break;case 19:pr(l,i),gr(i),h&4&&(h=i.updateQueue,h!==null&&(i.updateQueue=null,nh(i,h)));break;case 21:break;default:pr(l,i),gr(i)}}function gr(i){var l=i.flags;if(l&2){try{if(i.tag!==27){e:{for(var f=i.return;f!==null;){if(IR(f)){var h=f;break e}f=f.return}throw Error(r(160))}switch(h.tag){case 27:var v=h.stateNode,T=Jg(i);jc(i,T,v);break;case 5:var P=h.stateNode;h.flags&32&&(Di(P,""),h.flags&=-33);var q=Jg(i);jc(i,q,P);break;case 3:case 4:var X=h.stateNode.containerInfo,ne=Jg(i);eh(i,ne,X);break;default:throw Error(r(161))}}}catch(be){Nt(i,i.return,be)}i.flags&=-3}l&4096&&(i.flags&=-4097)}function jR(i){if(i.subtreeFlags&1024)for(i=i.child;i!==null;){var l=i;jR(l),l.tag===5&&l.flags&1024&&l.stateNode.reset(),i=i.sibling}}function Ta(i,l){if(l.subtreeFlags&8772)for(l=l.child;l!==null;)FR(i,l.alternate,l),l=l.sibling}function Ki(i){for(i=i.child;i!==null;){var l=i;switch(l.tag){case 0:case 11:case 14:case 15:io(4,l,l.return),Ki(l);break;case 1:Kn(l,l.return);var f=l.stateNode;typeof f.componentWillUnmount=="function"&&NR(l,l.return,f),Ki(l);break;case 26:case 27:case 5:Kn(l,l.return),Ki(l);break;case 22:Kn(l,l.return),l.memoizedState===null&&Ki(l);break;default:Ki(l)}i=i.sibling}}function so(i,l,f){for(f=f&&(l.subtreeFlags&8772)!==0,l=l.child;l!==null;){var h=l.alternate,v=i,T=l,P=T.flags;switch(T.tag){case 0:case 11:case 15:so(v,T,f),Ol(4,T);break;case 1:if(so(v,T,f),h=T,v=h.stateNode,typeof v.componentDidMount=="function")try{v.componentDidMount()}catch(ne){Nt(h,h.return,ne)}if(h=T,v=h.updateQueue,v!==null){var q=h.stateNode;try{var X=v.shared.hiddenCallbacks;if(X!==null)for(v.shared.hiddenCallbacks=null,v=0;v<X.length;v++)RR(X[v],q)}catch(ne){Nt(h,h.return,ne)}}f&&P&64&&CR(T),Qo(T,T.return);break;case 26:case 27:case 5:so(v,T,f),f&&h===null&&P&4&&OR(T),Qo(T,T.return);break;case 12:so(v,T,f);break;case 13:so(v,T,f),f&&P&4&&BR(v,T);break;case 22:T.memoizedState===null&&so(v,T,f),Qo(T,T.return);break;default:so(v,T,f)}l=l.sibling}}function rh(i,l){var f=null;i!==null&&i.memoizedState!==null&&i.memoizedState.cachePool!==null&&(f=i.memoizedState.cachePool.pool),i=null,l.memoizedState!==null&&l.memoizedState.cachePool!==null&&(i=l.memoizedState.cachePool.pool),i!==f&&(i!=null&&i.refCount++,f!=null&&El(f))}function ah(i,l){i=null,l.alternate!==null&&(i=l.alternate.memoizedState.cache),l=l.memoizedState.cache,l!==i&&(l.refCount++,i!=null&&El(i))}function lo(i,l,f,h){if(l.subtreeFlags&10256)for(l=l.child;l!==null;)GR(i,l,f,h),l=l.sibling}function GR(i,l,f,h){var v=l.flags;switch(l.tag){case 0:case 11:case 15:lo(i,l,f,h),v&2048&&Ol(9,l);break;case 3:lo(i,l,f,h),v&2048&&(i=null,l.alternate!==null&&(i=l.alternate.memoizedState.cache),l=l.memoizedState.cache,l!==i&&(l.refCount++,i!=null&&El(i)));break;case 12:if(v&2048){lo(i,l,f,h),i=l.stateNode;try{var T=l.memoizedProps,P=T.id,q=T.onPostCommit;typeof q=="function"&&q(P,l.alternate===null?"mount":"update",i.passiveEffectDuration,-0)}catch(X){Nt(l,l.return,X)}}else lo(i,l,f,h);break;case 23:break;case 22:T=l.stateNode,l.memoizedState!==null?T._visibility&4?lo(i,l,f,h):Dl(i,l):T._visibility&4?lo(i,l,f,h):(T._visibility|=4,Xi(i,l,f,h,(l.subtreeFlags&10256)!==0)),v&2048&&rh(l.alternate,l);break;case 24:lo(i,l,f,h),v&2048&&ah(l.alternate,l);break;default:lo(i,l,f,h)}}function Xi(i,l,f,h,v){for(v=v&&(l.subtreeFlags&10256)!==0,l=l.child;l!==null;){var T=i,P=l,q=f,X=h,ne=P.flags;switch(P.tag){case 0:case 11:case 15:Xi(T,P,q,X,v),Ol(8,P);break;case 23:break;case 22:var be=P.stateNode;P.memoizedState!==null?be._visibility&4?Xi(T,P,q,X,v):Dl(T,P):(be._visibility|=4,Xi(T,P,q,X,v)),v&&ne&2048&&rh(P.alternate,P);break;case 24:Xi(T,P,q,X,v),v&&ne&2048&&ah(P.alternate,P);break;default:Xi(T,P,q,X,v)}l=l.sibling}}function Dl(i,l){if(l.subtreeFlags&10256)for(l=l.child;l!==null;){var f=i,h=l,v=h.flags;switch(h.tag){case 22:Dl(f,h),v&2048&&rh(h.alternate,h);break;case 24:Dl(f,h),v&2048&&ah(h.alternate,h);break;default:Dl(f,h)}l=l.sibling}}var Il=8192;function Zi(i){if(i.subtreeFlags&Il)for(i=i.child;i!==null;)HR(i),i=i.sibling}function HR(i){switch(i.tag){case 26:Zi(i),i.flags&Il&&i.memoizedState!==null&&Z6(Nr,i.memoizedState,i.memoizedProps);break;case 5:Zi(i);break;case 3:case 4:var l=Nr;Nr=nd(i.stateNode.containerInfo),Zi(i),Nr=l;break;case 22:i.memoizedState===null&&(l=i.alternate,l!==null&&l.memoizedState!==null?(l=Il,Il=16777216,Zi(i),Il=l):Zi(i));break;default:Zi(i)}}function $R(i){var l=i.alternate;if(l!==null&&(i=l.child,i!==null)){l.child=null;do l=i.sibling,i.sibling=null,i=l;while(i!==null)}}function Ll(i){var l=i.deletions;if(i.flags&16){if(l!==null)for(var f=0;f<l.length;f++){var h=l[f];on=h,VR(h,i)}$R(i)}if(i.subtreeFlags&10256)for(i=i.child;i!==null;)qR(i),i=i.sibling}function qR(i){switch(i.tag){case 0:case 11:case 15:Ll(i),i.flags&2048&&io(9,i,i.return);break;case 3:Ll(i);break;case 12:Ll(i);break;case 22:var l=i.stateNode;i.memoizedState!==null&&l._visibility&4&&(i.return===null||i.return.tag!==13)?(l._visibility&=-5,Gc(i)):Ll(i);break;default:Ll(i)}}function Gc(i){var l=i.deletions;if(i.flags&16){if(l!==null)for(var f=0;f<l.length;f++){var h=l[f];on=h,VR(h,i)}$R(i)}for(i=i.child;i!==null;){switch(l=i,l.tag){case 0:case 11:case 15:io(8,l,l.return),Gc(l);break;case 22:f=l.stateNode,f._visibility&4&&(f._visibility&=-5,Gc(l));break;default:Gc(l)}i=i.sibling}}function VR(i,l){for(;on!==null;){var f=on;switch(f.tag){case 0:case 11:case 15:io(8,f,l);break;case 23:case 22:if(f.memoizedState!==null&&f.memoizedState.cachePool!==null){var h=f.memoizedState.cachePool.pool;h!=null&&h.refCount++}break;case 24:El(f.memoizedState.cache)}if(h=f.child,h!==null)h.return=f,on=h;else e:for(f=i;on!==null;){h=on;var v=h.sibling,T=h.return;if(PR(h),h===f){on=null;break e}if(v!==null){v.return=T,on=v;break e}on=T}}}function c6(i,l,f,h){this.tag=i,this.key=f,this.sibling=this.child=this.return=this.stateNode=this.type=this.elementType=null,this.index=0,this.refCleanup=this.ref=null,this.pendingProps=l,this.dependencies=this.memoizedState=this.updateQueue=this.memoizedProps=null,this.mode=h,this.subtreeFlags=this.flags=0,this.deletions=null,this.childLanes=this.lanes=0,this.alternate=null}function hr(i,l,f,h){return new c6(i,l,f,h)}function oh(i){return i=i.prototype,!(!i||!i.isReactComponent)}function uo(i,l){var f=i.alternate;return f===null?(f=hr(i.tag,l,i.key,i.mode),f.elementType=i.elementType,f.type=i.type,f.stateNode=i.stateNode,f.alternate=i,i.alternate=f):(f.pendingProps=l,f.type=i.type,f.flags=0,f.subtreeFlags=0,f.deletions=null),f.flags=i.flags&31457280,f.childLanes=i.childLanes,f.lanes=i.lanes,f.child=i.child,f.memoizedProps=i.memoizedProps,f.memoizedState=i.memoizedState,f.updateQueue=i.updateQueue,l=i.dependencies,f.dependencies=l===null?null:{lanes:l.lanes,firstContext:l.firstContext},f.sibling=i.sibling,f.index=i.index,f.ref=i.ref,f.refCleanup=i.refCleanup,f}function WR(i,l){i.flags&=31457282;var f=i.alternate;return f===null?(i.childLanes=0,i.lanes=l,i.child=null,i.subtreeFlags=0,i.memoizedProps=null,i.memoizedState=null,i.updateQueue=null,i.dependencies=null,i.stateNode=null):(i.childLanes=f.childLanes,i.lanes=f.lanes,i.child=f.child,i.subtreeFlags=0,i.deletions=null,i.memoizedProps=f.memoizedProps,i.memoizedState=f.memoizedState,i.updateQueue=f.updateQueue,i.type=f.type,l=f.dependencies,i.dependencies=l===null?null:{lanes:l.lanes,firstContext:l.firstContext}),i}function Hc(i,l,f,h,v,T){var P=0;if(h=i,typeof i=="function")oh(i)&&(P=1);else if(typeof i=="string")P=K6(i,f,ve.current)?26:i==="html"||i==="head"||i==="body"?27:5;else e:switch(i){case c:return Jo(f.children,v,T,l);case d:P=8,v|=24;break;case p:return i=hr(12,f,l,v|2),i.elementType=p,i.lanes=T,i;case S:return i=hr(13,f,l,v),i.elementType=S,i.lanes=T,i;case x:return i=hr(19,f,l,v),i.elementType=x,i.lanes=T,i;case A:return YR(f,v,T,l);default:if(typeof i=="object"&&i!==null)switch(i.$$typeof){case g:case b:P=10;break e;case m:P=9;break e;case y:P=11;break e;case R:P=14;break e;case k:P=16,h=null;break e}P=29,f=Error(r(130,i===null?"null":typeof i,"")),h=null}return l=hr(P,f,l,v),l.elementType=i,l.type=h,l.lanes=T,l}function Jo(i,l,f,h){return i=hr(7,i,h,l),i.lanes=f,i}function YR(i,l,f,h){i=hr(22,i,h,l),i.elementType=A,i.lanes=f;var v={_visibility:1,_pendingVisibility:1,_pendingMarkers:null,_retryCache:null,_transitions:null,_current:null,detach:function(){var T=v._current;if(T===null)throw Error(r(456));if(!(v._pendingVisibility&2)){var P=Za(T,2);P!==null&&(v._pendingVisibility|=2,Nn(P,T,2))}},attach:function(){var T=v._current;if(T===null)throw Error(r(456));if(v._pendingVisibility&2){var P=Za(T,2);P!==null&&(v._pendingVisibility&=-3,Nn(P,T,2))}}};return i.stateNode=v,i}function ih(i,l,f){return i=hr(6,i,null,l),i.lanes=f,i}function sh(i,l,f){return l=hr(4,i.children!==null?i.children:[],i.key,l),l.lanes=f,l.stateNode={containerInfo:i.containerInfo,pendingChildren:null,implementation:i.implementation},l}function Aa(i){i.flags|=4}function KR(i,l){if(l.type!=="stylesheet"||l.state.loading&4)i.flags&=-16777217;else if(i.flags|=16777216,!I_(l)){if(l=fr.current,l!==null&&((bt&4194176)===bt?Vr!==null:(bt&62914560)!==bt&&!(bt&536870912)||l!==Vr))throw yl=fg,f1;i.flags|=8192}}function $c(i,l){l!==null&&(i.flags|=4),i.flags&16384&&(l=i.tag!==22?Dn():536870912,i.lanes|=l,Ji|=l)}function Ml(i,l){if(!Et)switch(i.tailMode){case"hidden":l=i.tail;for(var f=null;l!==null;)l.alternate!==null&&(f=l),l=l.sibling;f===null?i.tail=null:f.sibling=null;break;case"collapsed":f=i.tail;for(var h=null;f!==null;)f.alternate!==null&&(h=f),f=f.sibling;h===null?l||i.tail===null?i.tail=null:i.tail.sibling=null:h.sibling=null}}function Bt(i){var l=i.alternate!==null&&i.alternate.child===i.child,f=0,h=0;if(l)for(var v=i.child;v!==null;)f|=v.lanes|v.childLanes,h|=v.subtreeFlags&31457280,h|=v.flags&31457280,v.return=i,v=v.sibling;else for(v=i.child;v!==null;)f|=v.lanes|v.childLanes,h|=v.subtreeFlags,h|=v.flags,v.return=i,v=v.sibling;return i.subtreeFlags|=h,i.childLanes=f,l}function d6(i,l,f){var h=l.pendingProps;switch(cg(l),l.tag){case 16:case 15:case 0:case 11:case 7:case 8:case 12:case 9:case 14:return Bt(l),null;case 1:return Bt(l),null;case 3:return f=l.stateNode,h=null,i!==null&&(h=i.memoizedState.cache),l.memoizedState.cache!==h&&(l.flags|=2048),wa(tn),J(),f.pendingContext&&(f.context=f.pendingContext,f.pendingContext=null),(i===null||i.child===null)&&(gl(l)?Aa(l):i===null||i.memoizedState.isDehydrated&&!(l.flags&256)||(l.flags|=1024,Cr!==null&&(hh(Cr),Cr=null))),Bt(l),null;case 26:return f=l.memoizedState,i===null?(Aa(l),f!==null?(Bt(l),KR(l,f)):(Bt(l),l.flags&=-16777217)):f?f!==i.memoizedState?(Aa(l),Bt(l),KR(l,f)):(Bt(l),l.flags&=-16777217):(i.memoizedProps!==h&&Aa(l),Bt(l),l.flags&=-16777217),null;case 27:Te(l),f=Ie.current;var v=l.type;if(i!==null&&l.stateNode!=null)i.memoizedProps!==h&&Aa(l);else{if(!h){if(l.stateNode===null)throw Error(r(166));return Bt(l),null}i=ve.current,gl(l)?c1(l):(i=T_(v,h,f),l.stateNode=i,Aa(l))}return Bt(l),null;case 5:if(Te(l),f=l.type,i!==null&&l.stateNode!=null)i.memoizedProps!==h&&Aa(l);else{if(!h){if(l.stateNode===null)throw Error(r(166));return Bt(l),null}if(i=ve.current,gl(l))c1(l);else{switch(v=td(Ie.current),i){case 1:i=v.createElementNS("http://www.w3.org/2000/svg",f);break;case 2:i=v.createElementNS("http://www.w3.org/1998/Math/MathML",f);break;default:switch(f){case"svg":i=v.createElementNS("http://www.w3.org/2000/svg",f);break;case"math":i=v.createElementNS("http://www.w3.org/1998/Math/MathML",f);break;case"script":i=v.createElement("div"),i.innerHTML="<script><\/script>",i=i.removeChild(i.firstChild);break;case"select":i=typeof h.is=="string"?v.createElement("select",{is:h.is}):v.createElement("select"),h.multiple?i.multiple=!0:h.size&&(i.size=h.size);break;default:i=typeof h.is=="string"?v.createElement(f,{is:h.is}):v.createElement(f)}}i[Sn]=l,i[Mn]=h;e:for(v=l.child;v!==null;){if(v.tag===5||v.tag===6)i.appendChild(v.stateNode);else if(v.tag!==4&&v.tag!==27&&v.child!==null){v.child.return=v,v=v.child;continue}if(v===l)break e;for(;v.sibling===null;){if(v.return===null||v.return===l)break e;v=v.return}v.sibling.return=v.return,v=v.sibling}l.stateNode=i;e:switch(hn(i,f,h),f){case"button":case"input":case"select":case"textarea":i=!!h.autoFocus;break e;case"img":i=!0;break e;default:i=!1}i&&Aa(l)}}return Bt(l),l.flags&=-16777217,null;case 6:if(i&&l.stateNode!=null)i.memoizedProps!==h&&Aa(l);else{if(typeof h!="string"&&l.stateNode===null)throw Error(r(166));if(i=Ie.current,gl(l)){if(i=l.stateNode,f=l.memoizedProps,h=null,v=Cn,v!==null)switch(v.tag){case 27:case 5:h=v.memoizedProps}i[Sn]=l,i=!!(i.nodeValue===f||h!==null&&h.suppressHydrationWarning===!0||v_(i.nodeValue,f)),i||Ho(l)}else i=td(i).createTextNode(h),i[Sn]=l,l.stateNode=i}return Bt(l),null;case 13:if(h=l.memoizedState,i===null||i.memoizedState!==null&&i.memoizedState.dehydrated!==null){if(v=gl(l),h!==null&&h.dehydrated!==null){if(i===null){if(!v)throw Error(r(318));if(v=l.memoizedState,v=v!==null?v.dehydrated:null,!v)throw Error(r(317));v[Sn]=l}else hl(),!(l.flags&128)&&(l.memoizedState=null),l.flags|=4;Bt(l),v=!1}else Cr!==null&&(hh(Cr),Cr=null),v=!0;if(!v)return l.flags&256?(ya(l),l):(ya(l),null)}if(ya(l),l.flags&128)return l.lanes=f,l;if(f=h!==null,i=i!==null&&i.memoizedState!==null,f){h=l.child,v=null,h.alternate!==null&&h.alternate.memoizedState!==null&&h.alternate.memoizedState.cachePool!==null&&(v=h.alternate.memoizedState.cachePool.pool);var T=null;h.memoizedState!==null&&h.memoizedState.cachePool!==null&&(T=h.memoizedState.cachePool.pool),T!==v&&(h.flags|=2048)}return f!==i&&f&&(l.child.flags|=8192),$c(l,l.updateQueue),Bt(l),null;case 4:return J(),i===null&&Ah(l.stateNode.containerInfo),Bt(l),null;case 10:return wa(l.type),Bt(l),null;case 19:if(Se(en),v=l.memoizedState,v===null)return Bt(l),null;if(h=(l.flags&128)!==0,T=v.rendering,T===null)if(h)Ml(v,!1);else{if(qt!==0||i!==null&&i.flags&128)for(i=l.child;i!==null;){if(T=Rc(i),T!==null){for(l.flags|=128,Ml(v,!1),i=T.updateQueue,l.updateQueue=i,$c(l,i),l.subtreeFlags=0,i=f,f=l.child;f!==null;)WR(f,i),f=f.sibling;return ae(en,en.current&1|2),l.child}i=i.sibling}v.tail!==null&&he()>qc&&(l.flags|=128,h=!0,Ml(v,!1),l.lanes=4194304)}else{if(!h)if(i=Rc(T),i!==null){if(l.flags|=128,h=!0,i=i.updateQueue,l.updateQueue=i,$c(l,i),Ml(v,!0),v.tail===null&&v.tailMode==="hidden"&&!T.alternate&&!Et)return Bt(l),null}else 2*he()-v.renderingStartTime>qc&&f!==536870912&&(l.flags|=128,h=!0,Ml(v,!1),l.lanes=4194304);v.isBackwards?(T.sibling=l.child,l.child=T):(i=v.last,i!==null?i.sibling=T:l.child=T,v.last=T)}return v.tail!==null?(l=v.tail,v.rendering=l,v.tail=l.sibling,v.renderingStartTime=he(),l.sibling=null,i=en.current,ae(en,h?i&1|2:i&1),l):(Bt(l),null);case 22:case 23:return ya(l),gg(),h=l.memoizedState!==null,i!==null?i.memoizedState!==null!==h&&(l.flags|=8192):h&&(l.flags|=8192),h?f&536870912&&!(l.flags&128)&&(Bt(l),l.subtreeFlags&6&&(l.flags|=8192)):Bt(l),f=l.updateQueue,f!==null&&$c(l,f.retryQueue),f=null,i!==null&&i.memoizedState!==null&&i.memoizedState.cachePool!==null&&(f=i.memoizedState.cachePool.pool),h=null,l.memoizedState!==null&&l.memoizedState.cachePool!==null&&(h=l.memoizedState.cachePool.pool),h!==f&&(l.flags|=2048),i!==null&&Se(qo),null;case 24:return f=null,i!==null&&(f=i.memoizedState.cache),l.memoizedState.cache!==f&&(l.flags|=2048),wa(tn),Bt(l),null;case 25:return null}throw Error(r(156,l.tag))}function f6(i,l){switch(cg(l),l.tag){case 1:return i=l.flags,i&65536?(l.flags=i&-65537|128,l):null;case 3:return wa(tn),J(),i=l.flags,i&65536&&!(i&128)?(l.flags=i&-65537|128,l):null;case 26:case 27:case 5:return Te(l),null;case 13:if(ya(l),i=l.memoizedState,i!==null&&i.dehydrated!==null){if(l.alternate===null)throw Error(r(340));hl()}return i=l.flags,i&65536?(l.flags=i&-65537|128,l):null;case 19:return Se(en),null;case 4:return J(),null;case 10:return wa(l.type),null;case 22:case 23:return ya(l),gg(),i!==null&&Se(qo),i=l.flags,i&65536?(l.flags=i&-65537|128,l):null;case 24:return wa(tn),null;case 25:return null;default:return null}}function XR(i,l){switch(cg(l),l.tag){case 3:wa(tn),J();break;case 26:case 27:case 5:Te(l);break;case 4:J();break;case 13:ya(l);break;case 19:Se(en);break;case 10:wa(l.type);break;case 22:case 23:ya(l),gg(),i!==null&&Se(qo);break;case 24:wa(tn)}}var p6={getCacheForType:function(i){var l=En(tn),f=l.data.get(i);return f===void 0&&(f=i(),l.data.set(i,f)),f}},g6=typeof WeakMap=="function"?WeakMap:Map,Ut=0,It=null,ct=null,bt=0,Lt=0,Zn=null,Ra=!1,Qi=!1,lh=!1,_a=0,qt=0,co=0,ei=0,uh=0,mr=0,Ji=0,Fl=null,Yr=null,ch=!1,dh=0,qc=1/0,Vc=null,fo=null,Wc=!1,ti=null,Pl=0,fh=0,ph=null,zl=0,gh=null;function Qn(){if(Ut&2&&bt!==0)return bt&-bt;if(L.T!==null){var i=$i;return i!==0?i:wh()}return bA()}function ZR(){mr===0&&(mr=!(bt&536870912)||Et?Ct():536870912);var i=fr.current;return i!==null&&(i.flags|=32),mr}function Nn(i,l,f){(i===It&&Lt===2||i.cancelPendingCommit!==null)&&(es(i,0),Ca(i,bt,mr,!1)),In(i,f),(!(Ut&2)||i!==It)&&(i===It&&(!(Ut&2)&&(ei|=f),qt===4&&Ca(i,bt,mr,!1)),Kr(i))}function QR(i,l,f){if(Ut&6)throw Error(r(327));var h=!f&&(l&60)===0&&(l&i.expiredLanes)===0||Xe(i,l),v=h?b6(i,l):yh(i,l,!0),T=h;do{if(v===0){Qi&&!h&&Ca(i,l,0,!1);break}else if(v===6)Ca(i,l,0,!Ra);else{if(f=i.current.alternate,T&&!h6(f)){v=yh(i,l,!1),T=!1;continue}if(v===2){if(T=l,i.errorRecoveryDisabledLanes&T)var P=0;else P=i.pendingLanes&-536870913,P=P!==0?P:P&536870912?536870912:0;if(P!==0){l=P;e:{var q=i;v=Fl;var X=q.current.memoizedState.isDehydrated;if(X&&(es(q,P).flags|=256),P=yh(q,P,!1),P!==2){if(lh&&!X){q.errorRecoveryDisabledLanes|=T,ei|=T,v=4;break e}T=Yr,Yr=v,T!==null&&hh(T)}v=P}if(T=!1,v!==2)continue}}if(v===1){es(i,0),Ca(i,l,0,!0);break}e:{switch(h=i,v){case 0:case 1:throw Error(r(345));case 4:if((l&4194176)===l){Ca(h,l,mr,!Ra);break e}break;case 2:Yr=null;break;case 3:case 5:break;default:throw Error(r(329))}if(h.finishedWork=f,h.finishedLanes=l,(l&62914560)===l&&(T=dh+300-he(),10<T)){if(Ca(h,l,mr,!Ra),fa(h,0)!==0)break e;h.timeoutHandle=w_(JR.bind(null,h,f,Yr,Vc,ch,l,mr,ei,Ji,Ra,2,-0,0),T);break e}JR(h,f,Yr,Vc,ch,l,mr,ei,Ji,Ra,0,-0,0)}}break}while(!0);Kr(i)}function hh(i){Yr===null?Yr=i:Yr.push.apply(Yr,i)}function JR(i,l,f,h,v,T,P,q,X,ne,be,Ae,de){var ge=l.subtreeFlags;if((ge&8192||(ge&16785408)===16785408)&&($l={stylesheets:null,count:0,unsuspend:X6},HR(l),l=Q6(),l!==null)){i.cancelPendingCommit=l(i_.bind(null,i,f,h,v,P,q,X,1,Ae,de)),Ca(i,T,P,!ne);return}i_(i,f,h,v,P,q,X,be,Ae,de)}function h6(i){for(var l=i;;){var f=l.tag;if((f===0||f===11||f===15)&&l.flags&16384&&(f=l.updateQueue,f!==null&&(f=f.stores,f!==null)))for(var h=0;h<f.length;h++){var v=f[h],T=v.getSnapshot;v=v.value;try{if(!Yn(T(),v))return!1}catch{return!1}}if(f=l.child,l.subtreeFlags&16384&&f!==null)f.return=l,l=f;else{if(l===i)break;for(;l.sibling===null;){if(l.return===null||l.return===i)return!0;l=l.return}l.sibling.return=l.return,l=l.sibling}}return!0}function Ca(i,l,f,h){l&=~uh,l&=~ei,i.suspendedLanes|=l,i.pingedLanes&=~l,h&&(i.warmLanes|=l),h=i.expirationTimes;for(var v=l;0<v;){var T=31-St(v),P=1<<T;h[T]=-1,v&=~P}f!==0&&Ln(i,f,l)}function Yc(){return Ut&6?!0:(Bl(0),!1)}function mh(){if(ct!==null){if(Lt===0)var i=ct.return;else i=ct,Ea=Xo=null,wg(i),Gi=null,vl=0,i=ct;for(;i!==null;)XR(i.alternate,i),i=i.return;ct=null}}function es(i,l){i.finishedWork=null,i.finishedLanes=0;var f=i.timeoutHandle;f!==-1&&(i.timeoutHandle=-1,L6(f)),f=i.cancelPendingCommit,f!==null&&(i.cancelPendingCommit=null,f()),mh(),It=i,ct=f=uo(i.current,null),bt=l,Lt=0,Zn=null,Ra=!1,Qi=Xe(i,l),lh=!1,Ji=mr=uh=ei=co=qt=0,Yr=Fl=null,ch=!1,l&8&&(l|=l&32);var h=i.entangledLanes;if(h!==0)for(i=i.entanglements,h&=l;0<h;){var v=31-St(h),T=1<<v;l|=i[v],h&=~T}return _a=l,yc(),f}function e_(i,l){st=null,L.H=Wr,l===bl?(l=h1(),Lt=3):l===f1?(l=h1(),Lt=4):Lt=l===pR?8:l!==null&&typeof l=="object"&&typeof l.then=="function"?6:1,Zn=l,ct===null&&(qt=1,zc(i,ur(l,i.current)))}function t_(){var i=L.H;return L.H=Wr,i===null?Wr:i}function n_(){var i=L.A;return L.A=p6,i}function bh(){qt=4,Ra||(bt&4194176)!==bt&&fr.current!==null||(Qi=!0),!(co&134217727)&&!(ei&134217727)||It===null||Ca(It,bt,mr,!1)}function yh(i,l,f){var h=Ut;Ut|=2;var v=t_(),T=n_();(It!==i||bt!==l)&&(Vc=null,es(i,l)),l=!1;var P=qt;e:do try{if(Lt!==0&&ct!==null){var q=ct,X=Zn;switch(Lt){case 8:mh(),P=6;break e;case 3:case 2:case 6:fr.current===null&&(l=!0);var ne=Lt;if(Lt=0,Zn=null,ts(i,q,X,ne),f&&Qi){P=0;break e}break;default:ne=Lt,Lt=0,Zn=null,ts(i,q,X,ne)}}m6(),P=qt;break}catch(be){e_(i,be)}while(!0);return l&&i.shellSuspendCounter++,Ea=Xo=null,Ut=h,L.H=v,L.A=T,ct===null&&(It=null,bt=0,yc()),P}function m6(){for(;ct!==null;)r_(ct)}function b6(i,l){var f=Ut;Ut|=2;var h=t_(),v=n_();It!==i||bt!==l?(Vc=null,qc=he()+500,es(i,l)):Qi=Xe(i,l);e:do try{if(Lt!==0&&ct!==null){l=ct;var T=Zn;t:switch(Lt){case 1:Lt=0,Zn=null,ts(i,l,T,1);break;case 2:if(p1(T)){Lt=0,Zn=null,a_(l);break}l=function(){Lt===2&&It===i&&(Lt=7),Kr(i)},T.then(l,l);break e;case 3:Lt=7;break e;case 4:Lt=5;break e;case 7:p1(T)?(Lt=0,Zn=null,a_(l)):(Lt=0,Zn=null,ts(i,l,T,7));break;case 5:var P=null;switch(ct.tag){case 26:P=ct.memoizedState;case 5:case 27:var q=ct;if(!P||I_(P)){Lt=0,Zn=null;var X=q.sibling;if(X!==null)ct=X;else{var ne=q.return;ne!==null?(ct=ne,Kc(ne)):ct=null}break t}}Lt=0,Zn=null,ts(i,l,T,5);break;case 6:Lt=0,Zn=null,ts(i,l,T,6);break;case 8:mh(),qt=6;break e;default:throw Error(r(462))}}y6();break}catch(be){e_(i,be)}while(!0);return Ea=Xo=null,L.H=h,L.A=v,Ut=f,ct!==null?0:(It=null,bt=0,yc(),qt)}function y6(){for(;ct!==null&&!Be();)r_(ct)}function r_(i){var l=TR(i.alternate,i,_a);i.memoizedProps=i.pendingProps,l===null?Kc(i):ct=l}function a_(i){var l=i,f=l.alternate;switch(l.tag){case 15:case 0:l=vR(f,l,l.pendingProps,l.type,void 0,bt);break;case 11:l=vR(f,l,l.pendingProps,l.type.render,l.ref,bt);break;case 5:wg(l);default:XR(f,l),l=ct=WR(l,_a),l=TR(f,l,_a)}i.memoizedProps=i.pendingProps,l===null?Kc(i):ct=l}function ts(i,l,f,h){Ea=Xo=null,wg(l),Gi=null,vl=0;var v=l.return;try{if(i6(i,v,l,f,bt)){qt=1,zc(i,ur(f,i.current)),ct=null;return}}catch(T){if(v!==null)throw ct=v,T;qt=1,zc(i,ur(f,i.current)),ct=null;return}l.flags&32768?(Et||h===1?i=!0:Qi||bt&536870912?i=!1:(Ra=i=!0,(h===2||h===3||h===6)&&(h=fr.current,h!==null&&h.tag===13&&(h.flags|=16384))),o_(l,i)):Kc(l)}function Kc(i){var l=i;do{if(l.flags&32768){o_(l,Ra);return}i=l.return;var f=d6(l.alternate,l,_a);if(f!==null){ct=f;return}if(l=l.sibling,l!==null){ct=l;return}ct=l=i}while(l!==null);qt===0&&(qt=5)}function o_(i,l){do{var f=f6(i.alternate,i);if(f!==null){f.flags&=32767,ct=f;return}if(f=i.return,f!==null&&(f.flags|=32768,f.subtreeFlags=0,f.deletions=null),!l&&(i=i.sibling,i!==null)){ct=i;return}ct=i=f}while(i!==null);qt=6,ct=null}function i_(i,l,f,h,v,T,P,q,X,ne){var be=L.T,Ae=W.p;try{W.p=2,L.T=null,v6(i,l,f,h,Ae,v,T,P,q,X,ne)}finally{L.T=be,W.p=Ae}}function v6(i,l,f,h,v,T,P,q){do ns();while(ti!==null);if(Ut&6)throw Error(r(327));var X=i.finishedWork;if(h=i.finishedLanes,X===null)return null;if(i.finishedWork=null,i.finishedLanes=0,X===i.current)throw Error(r(177));i.callbackNode=null,i.callbackPriority=0,i.cancelPendingCommit=null;var ne=X.lanes|X.childLanes;if(ne|=sg,pa(i,h,ne,T,P,q),i===It&&(ct=It=null,bt=0),!(X.subtreeFlags&10256)&&!(X.flags&10256)||Wc||(Wc=!0,fh=ne,ph=f,x6(_e,function(){return ns(),null})),f=(X.flags&15990)!==0,X.subtreeFlags&15990||f?(f=L.T,L.T=null,T=W.p,W.p=2,P=Ut,Ut|=4,l6(i,X),UR(X,i),H$(Nh,i.containerInfo),sd=!!Ch,Nh=Ch=null,i.current=X,FR(i,X.alternate,X),Ue(),Ut=P,W.p=T,L.T=f):i.current=X,Wc?(Wc=!1,ti=i,Pl=h):s_(i,ne),ne=i.pendingLanes,ne===0&&(fo=null),Tt(X.stateNode),Kr(i),l!==null)for(v=i.onRecoverableError,X=0;X<l.length;X++)ne=l[X],v(ne.value,{componentStack:ne.stack});return Pl&3&&ns(),ne=i.pendingLanes,h&4194218&&ne&42?i===gh?zl++:(zl=0,gh=i):zl=0,Bl(0),null}function s_(i,l){(i.pooledCacheLanes&=l)===0&&(l=i.pooledCache,l!=null&&(i.pooledCache=null,El(l)))}function ns(){if(ti!==null){var i=ti,l=fh;fh=0;var f=mA(Pl),h=L.T,v=W.p;try{if(W.p=32>f?32:f,L.T=null,ti===null)var T=!1;else{f=ph,ph=null;var P=ti,q=Pl;if(ti=null,Pl=0,Ut&6)throw Error(r(331));var X=Ut;if(Ut|=4,qR(P.current),GR(P,P.current,q,f),Ut=X,Bl(0,!1),et&&typeof et.onPostCommitFiberRoot=="function")try{et.onPostCommitFiberRoot(ht,P)}catch{}T=!0}return T}finally{W.p=v,L.T=h,s_(i,l)}}return!1}function l_(i,l,f){l=ur(f,l),l=Fg(i.stateNode,l,2),i=oo(i,l,2),i!==null&&(In(i,2),Kr(i))}function Nt(i,l,f){if(i.tag===3)l_(i,i,f);else for(;l!==null;){if(l.tag===3){l_(l,i,f);break}else if(l.tag===1){var h=l.stateNode;if(typeof l.type.getDerivedStateFromError=="function"||typeof h.componentDidCatch=="function"&&(fo===null||!fo.has(h))){i=ur(f,i),f=dR(2),h=oo(l,f,2),h!==null&&(fR(f,h,l,i),In(h,2),Kr(h));break}}l=l.return}}function vh(i,l,f){var h=i.pingCache;if(h===null){h=i.pingCache=new g6;var v=new Set;h.set(l,v)}else v=h.get(l),v===void 0&&(v=new Set,h.set(l,v));v.has(f)||(lh=!0,v.add(f),i=S6.bind(null,i,l,f),l.then(i,i))}function S6(i,l,f){var h=i.pingCache;h!==null&&h.delete(l),i.pingedLanes|=i.suspendedLanes&f,i.warmLanes&=~f,It===i&&(bt&f)===f&&(qt===4||qt===3&&(bt&62914560)===bt&&300>he()-dh?!(Ut&2)&&es(i,0):uh|=f,Ji===bt&&(Ji=0)),Kr(i)}function u_(i,l){l===0&&(l=Dn()),i=Za(i,l),i!==null&&(In(i,l),Kr(i))}function E6(i){var l=i.memoizedState,f=0;l!==null&&(f=l.retryLane),u_(i,f)}function w6(i,l){var f=0;switch(i.tag){case 13:var h=i.stateNode,v=i.memoizedState;v!==null&&(f=v.retryLane);break;case 19:h=i.stateNode;break;case 22:h=i.stateNode._retryCache;break;default:throw Error(r(314))}h!==null&&h.delete(l),u_(i,f)}function x6(i,l){return Ee(i,l)}var Xc=null,rs=null,Sh=!1,Zc=!1,Eh=!1,ni=0;function Kr(i){i!==rs&&i.next===null&&(rs===null?Xc=rs=i:rs=rs.next=i),Zc=!0,Sh||(Sh=!0,T6(k6))}function Bl(i,l){if(!Eh&&Zc){Eh=!0;do for(var f=!1,h=Xc;h!==null;){if(i!==0){var v=h.pendingLanes;if(v===0)var T=0;else{var P=h.suspendedLanes,q=h.pingedLanes;T=(1<<31-St(42|i)+1)-1,T&=v&~(P&~q),T=T&201326677?T&201326677|1:T?T|2:0}T!==0&&(f=!0,f_(h,T))}else T=bt,T=fa(h,h===It?T:0),!(T&3)||Xe(h,T)||(f=!0,f_(h,T));h=h.next}while(f);Eh=!1}}function k6(){Zc=Sh=!1;var i=0;ni!==0&&(I6()&&(i=ni),ni=0);for(var l=he(),f=null,h=Xc;h!==null;){var v=h.next,T=c_(h,l);T===0?(h.next=null,f===null?Xc=v:f.next=v,v===null&&(rs=f)):(f=h,(i!==0||T&3)&&(Zc=!0)),h=v}Bl(i)}function c_(i,l){for(var f=i.suspendedLanes,h=i.pingedLanes,v=i.expirationTimes,T=i.pendingLanes&-62914561;0<T;){var P=31-St(T),q=1<<P,X=v[P];X===-1?(!(q&f)||q&h)&&(v[P]=mt(q,l)):X<=l&&(i.expiredLanes|=q),T&=~q}if(l=It,f=bt,f=fa(i,i===l?f:0),h=i.callbackNode,f===0||i===l&&Lt===2||i.cancelPendingCommit!==null)return h!==null&&h!==null&&le(h),i.callbackNode=null,i.callbackPriority=0;if(!(f&3)||Xe(i,f)){if(l=f&-f,l===i.callbackPriority)return l;switch(h!==null&&le(h),mA(f)){case 2:case 8:f=ce;break;case 32:f=_e;break;case 268435456:f=We;break;default:f=_e}return h=d_.bind(null,i),f=Ee(f,h),i.callbackPriority=l,i.callbackNode=f,l}return h!==null&&h!==null&&le(h),i.callbackPriority=2,i.callbackNode=null,2}function d_(i,l){var f=i.callbackNode;if(ns()&&i.callbackNode!==f)return null;var h=bt;return h=fa(i,i===It?h:0),h===0?null:(QR(i,h,l),c_(i,he()),i.callbackNode!=null&&i.callbackNode===f?d_.bind(null,i):null)}function f_(i,l){if(ns())return null;QR(i,l,!0)}function T6(i){M6(function(){Ut&6?Ee(ee,i):i()})}function wh(){return ni===0&&(ni=Ct()),ni}function p_(i){return i==null||typeof i=="symbol"||typeof i=="boolean"?null:typeof i=="function"?i:dc(""+i)}function g_(i,l){var f=l.ownerDocument.createElement("input");return f.name=l.name,f.value=l.value,i.id&&f.setAttribute("form",i.id),l.parentNode.insertBefore(f,l),i=new FormData(i),f.parentNode.removeChild(f),i}function A6(i,l,f,h,v){if(l==="submit"&&f&&f.stateNode===v){var T=p_((v[Mn]||null).action),P=h.submitter;P&&(l=(l=P[Mn]||null)?p_(l.formAction):P.getAttribute("formAction"),l!==null&&(T=l,P=null));var q=new hc("action","action",null,h,v);i.push({event:q,listeners:[{instance:null,listener:function(){if(h.defaultPrevented){if(ni!==0){var X=P?g_(v,P):new FormData(v);Og(f,{pending:!0,data:X,method:v.method,action:T},null,X)}}else typeof T=="function"&&(q.preventDefault(),X=P?g_(v,P):new FormData(v),Og(f,{pending:!0,data:X,method:v.method,action:T},T,X))},currentTarget:v}]})}}for(var xh=0;xh<i1.length;xh++){var kh=i1[xh],R6=kh.toLowerCase(),_6=kh[0].toUpperCase()+kh.slice(1);_r(R6,"on"+_6)}_r(t1,"onAnimationEnd"),_r(n1,"onAnimationIteration"),_r(r1,"onAnimationStart"),_r("dblclick","onDoubleClick"),_r("focusin","onFocus"),_r("focusout","onBlur"),_r(q$,"onTransitionRun"),_r(V$,"onTransitionStart"),_r(W$,"onTransitionCancel"),_r(a1,"onTransitionEnd"),Ni("onMouseEnter",["mouseout","mouseover"]),Ni("onMouseLeave",["mouseout","mouseover"]),Ni("onPointerEnter",["pointerout","pointerover"]),Ni("onPointerLeave",["pointerout","pointerover"]),Po("onChange","change click focusin focusout input keydown keyup selectionchange".split(" ")),Po("onSelect","focusout contextmenu dragend focusin keydown keyup mousedown mouseup selectionchange".split(" ")),Po("onBeforeInput",["compositionend","keypress","textInput","paste"]),Po("onCompositionEnd","compositionend focusout keydown keypress keyup mousedown".split(" ")),Po("onCompositionStart","compositionstart focusout keydown keypress keyup mousedown".split(" ")),Po("onCompositionUpdate","compositionupdate focusout keydown keypress keyup mousedown".split(" "));var Ul="abort canplay canplaythrough durationchange emptied encrypted ended error loadeddata loadedmetadata loadstart pause play playing progress ratechange resize seeked seeking stalled suspend timeupdate volumechange waiting".split(" "),C6=new Set("beforetoggle cancel close invalid load scroll scrollend toggle".split(" ").concat(Ul));function h_(i,l){l=(l&4)!==0;for(var f=0;f<i.length;f++){var h=i[f],v=h.event;h=h.listeners;e:{var T=void 0;if(l)for(var P=h.length-1;0<=P;P--){var q=h[P],X=q.instance,ne=q.currentTarget;if(q=q.listener,X!==T&&v.isPropagationStopped())break e;T=q,v.currentTarget=ne;try{T(v)}catch(be){Pc(be)}v.currentTarget=null,T=X}else for(P=0;P<h.length;P++){if(q=h[P],X=q.instance,ne=q.currentTarget,q=q.listener,X!==T&&v.isPropagationStopped())break e;T=q,v.currentTarget=ne;try{T(v)}catch(be){Pc(be)}v.currentTarget=null,T=X}}}}function pt(i,l){var f=l[Up];f===void 0&&(f=l[Up]=new Set);var h=i+"__bubble";f.has(h)||(m_(l,i,2,!1),f.add(h))}function Th(i,l,f){var h=0;l&&(h|=4),m_(f,i,h,l)}var Qc="_reactListening"+Math.random().toString(36).slice(2);function Ah(i){if(!i[Qc]){i[Qc]=!0,vA.forEach(function(f){f!=="selectionchange"&&(C6.has(f)||Th(f,!1,i),Th(f,!0,i))});var l=i.nodeType===9?i:i.ownerDocument;l===null||l[Qc]||(l[Qc]=!0,Th("selectionchange",!1,l))}}function m_(i,l,f,h){switch(B_(l)){case 2:var v=t9;break;case 8:v=n9;break;default:v=Bh}f=v.bind(null,l,f,i),v=void 0,!Yp||l!=="touchstart"&&l!=="touchmove"&&l!=="wheel"||(v=!0),h?v!==void 0?i.addEventListener(l,f,{capture:!0,passive:v}):i.addEventListener(l,f,!0):v!==void 0?i.addEventListener(l,f,{passive:v}):i.addEventListener(l,f,!1)}function Rh(i,l,f,h,v){var T=h;if(!(l&1)&&!(l&2)&&h!==null)e:for(;;){if(h===null)return;var P=h.tag;if(P===3||P===4){var q=h.stateNode.containerInfo;if(q===v||q.nodeType===8&&q.parentNode===v)break;if(P===4)for(P=h.return;P!==null;){var X=P.tag;if((X===3||X===4)&&(X=P.stateNode.containerInfo,X===v||X.nodeType===8&&X.parentNode===v))return;P=P.return}for(;q!==null;){if(P=Fo(q),P===null)return;if(X=P.tag,X===5||X===6||X===26||X===27){h=T=P;continue e}q=q.parentNode}}h=h.return}OA(function(){var ne=T,be=Vp(f),Ae=[];e:{var de=o1.get(i);if(de!==void 0){var ge=hc,qe=i;switch(i){case"keypress":if(pc(f)===0)break e;case"keydown":case"keyup":ge=w$;break;case"focusin":qe="focus",ge=Qp;break;case"focusout":qe="blur",ge=Qp;break;case"beforeblur":case"afterblur":ge=Qp;break;case"click":if(f.button===2)break e;case"auxclick":case"dblclick":case"mousedown":case"mousemove":case"mouseup":case"mouseout":case"mouseover":case"contextmenu":ge=LA;break;case"drag":case"dragend":case"dragenter":case"dragexit":case"dragleave":case"dragover":case"dragstart":case"drop":ge=c$;break;case"touchcancel":case"touchend":case"touchmove":case"touchstart":ge=T$;break;case t1:case n1:case r1:ge=p$;break;case a1:ge=R$;break;case"scroll":case"scrollend":ge=l$;break;case"wheel":ge=C$;break;case"copy":case"cut":case"paste":ge=h$;break;case"gotpointercapture":case"lostpointercapture":case"pointercancel":case"pointerdown":case"pointermove":case"pointerout":case"pointerover":case"pointerup":ge=FA;break;case"toggle":case"beforetoggle":ge=O$}var nt=(l&4)!==0,Vt=!nt&&(i==="scroll"||i==="scrollend"),se=nt?de!==null?de+"Capture":null:de;nt=[];for(var te=ne,ue;te!==null;){var we=te;if(ue=we.stateNode,we=we.tag,we!==5&&we!==26&&we!==27||ue===null||se===null||(we=ol(te,se),we!=null&&nt.push(jl(te,we,ue))),Vt)break;te=te.return}0<nt.length&&(de=new ge(de,qe,null,f,be),Ae.push({event:de,listeners:nt}))}}if(!(l&7)){e:{if(de=i==="mouseover"||i==="pointerover",ge=i==="mouseout"||i==="pointerout",de&&f!==qp&&(qe=f.relatedTarget||f.fromElement)&&(Fo(qe)||qe[Ri]))break e;if((ge||de)&&(de=be.window===be?be:(de=be.ownerDocument)?de.defaultView||de.parentWindow:window,ge?(qe=f.relatedTarget||f.toElement,ge=ne,qe=qe?Fo(qe):null,qe!==null&&(Vt=z(qe),nt=qe.tag,qe!==Vt||nt!==5&&nt!==27&&nt!==6)&&(qe=null)):(ge=null,qe=ne),ge!==qe)){if(nt=LA,we="onMouseLeave",se="onMouseEnter",te="mouse",(i==="pointerout"||i==="pointerover")&&(nt=FA,we="onPointerLeave",se="onPointerEnter",te="pointer"),Vt=ge==null?de:al(ge),ue=qe==null?de:al(qe),de=new nt(we,te+"leave",ge,f,be),de.target=Vt,de.relatedTarget=ue,we=null,Fo(be)===ne&&(nt=new nt(se,te+"enter",qe,f,be),nt.target=ue,nt.relatedTarget=Vt,we=nt),Vt=we,ge&&qe)t:{for(nt=ge,se=qe,te=0,ue=nt;ue;ue=as(ue))te++;for(ue=0,we=se;we;we=as(we))ue++;for(;0<te-ue;)nt=as(nt),te--;for(;0<ue-te;)se=as(se),ue--;for(;te--;){if(nt===se||se!==null&&nt===se.alternate)break t;nt=as(nt),se=as(se)}nt=null}else nt=null;ge!==null&&b_(Ae,de,ge,nt,!1),qe!==null&&Vt!==null&&b_(Ae,Vt,qe,nt,!0)}}e:{if(de=ne?al(ne):window,ge=de.nodeName&&de.nodeName.toLowerCase(),ge==="select"||ge==="input"&&de.type==="file")var Ge=$A;else if(GA(de))if(qA)Ge=j$;else{Ge=B$;var lt=z$}else ge=de.nodeName,!ge||ge.toLowerCase()!=="input"||de.type!=="checkbox"&&de.type!=="radio"?ne&&$p(ne.elementType)&&(Ge=$A):Ge=U$;if(Ge&&(Ge=Ge(i,ne))){HA(Ae,Ge,f,be);break e}lt&&lt(i,de,ne),i==="focusout"&&ne&&de.type==="number"&&ne.memoizedProps.value!=null&&Hp(de,"number",de.value)}switch(lt=ne?al(ne):window,i){case"focusin":(GA(lt)||lt.contentEditable==="true")&&(Fi=lt,ag=ne,pl=null);break;case"focusout":pl=ag=Fi=null;break;case"mousedown":og=!0;break;case"contextmenu":case"mouseup":case"dragend":og=!1,JA(Ae,f,be);break;case"selectionchange":if($$)break;case"keydown":case"keyup":JA(Ae,f,be)}var Ye;if(eg)e:{switch(i){case"compositionstart":var Je="onCompositionStart";break e;case"compositionend":Je="onCompositionEnd";break e;case"compositionupdate":Je="onCompositionUpdate";break e}Je=void 0}else Mi?UA(i,f)&&(Je="onCompositionEnd"):i==="keydown"&&f.keyCode===229&&(Je="onCompositionStart");Je&&(PA&&f.locale!=="ko"&&(Mi||Je!=="onCompositionStart"?Je==="onCompositionEnd"&&Mi&&(Ye=DA()):(Xa=be,Kp="value"in Xa?Xa.value:Xa.textContent,Mi=!0)),lt=Jc(ne,Je),0<lt.length&&(Je=new MA(Je,i,null,f,be),Ae.push({event:Je,listeners:lt}),Ye?Je.data=Ye:(Ye=jA(f),Ye!==null&&(Je.data=Ye)))),(Ye=I$?L$(i,f):M$(i,f))&&(Je=Jc(ne,"onBeforeInput"),0<Je.length&&(lt=new MA("onBeforeInput","beforeinput",null,f,be),Ae.push({event:lt,listeners:Je}),lt.data=Ye)),A6(Ae,i,ne,f,be)}h_(Ae,l)})}function jl(i,l,f){return{instance:i,listener:l,currentTarget:f}}function Jc(i,l){for(var f=l+"Capture",h=[];i!==null;){var v=i,T=v.stateNode;v=v.tag,v!==5&&v!==26&&v!==27||T===null||(v=ol(i,f),v!=null&&h.unshift(jl(i,v,T)),v=ol(i,l),v!=null&&h.push(jl(i,v,T))),i=i.return}return h}function as(i){if(i===null)return null;do i=i.return;while(i&&i.tag!==5&&i.tag!==27);return i||null}function b_(i,l,f,h,v){for(var T=l._reactName,P=[];f!==null&&f!==h;){var q=f,X=q.alternate,ne=q.stateNode;if(q=q.tag,X!==null&&X===h)break;q!==5&&q!==26&&q!==27||ne===null||(X=ne,v?(ne=ol(f,T),ne!=null&&P.unshift(jl(f,ne,X))):v||(ne=ol(f,T),ne!=null&&P.push(jl(f,ne,X)))),f=f.return}P.length!==0&&i.push({event:l,listeners:P})}var N6=/\r\n?/g,O6=/\u0000|\uFFFD/g;function y_(i){return(typeof i=="string"?i:""+i).replace(N6,`
`).replace(O6,"")}function v_(i,l){return l=y_(l),y_(i)===l}function ed(){}function Rt(i,l,f,h,v,T){switch(f){case"children":typeof h=="string"?l==="body"||l==="textarea"&&h===""||Di(i,h):(typeof h=="number"||typeof h=="bigint")&&l!=="body"&&Di(i,""+h);break;case"className":lc(i,"class",h);break;case"tabIndex":lc(i,"tabindex",h);break;case"dir":case"role":case"viewBox":case"width":case"height":lc(i,f,h);break;case"style":CA(i,h,T);break;case"data":if(l!=="object"){lc(i,"data",h);break}case"src":case"href":if(h===""&&(l!=="a"||f!=="href")){i.removeAttribute(f);break}if(h==null||typeof h=="function"||typeof h=="symbol"||typeof h=="boolean"){i.removeAttribute(f);break}h=dc(""+h),i.setAttribute(f,h);break;case"action":case"formAction":if(typeof h=="function"){i.setAttribute(f,"javascript:throw new Error('A React form was unexpectedly submitted. If you called form.submit() manually, consider using form.requestSubmit() instead. If you\\'re trying to use event.stopPropagation() in a submit event handler, consider also calling event.preventDefault().')");break}else typeof T=="function"&&(f==="formAction"?(l!=="input"&&Rt(i,l,"name",v.name,v,null),Rt(i,l,"formEncType",v.formEncType,v,null),Rt(i,l,"formMethod",v.formMethod,v,null),Rt(i,l,"formTarget",v.formTarget,v,null)):(Rt(i,l,"encType",v.encType,v,null),Rt(i,l,"method",v.method,v,null),Rt(i,l,"target",v.target,v,null)));if(h==null||typeof h=="symbol"||typeof h=="boolean"){i.removeAttribute(f);break}h=dc(""+h),i.setAttribute(f,h);break;case"onClick":h!=null&&(i.onclick=ed);break;case"onScroll":h!=null&&pt("scroll",i);break;case"onScrollEnd":h!=null&&pt("scrollend",i);break;case"dangerouslySetInnerHTML":if(h!=null){if(typeof h!="object"||!("__html"in h))throw Error(r(61));if(f=h.__html,f!=null){if(v.children!=null)throw Error(r(60));i.innerHTML=f}}break;case"multiple":i.multiple=h&&typeof h!="function"&&typeof h!="symbol";break;case"muted":i.muted=h&&typeof h!="function"&&typeof h!="symbol";break;case"suppressContentEditableWarning":case"suppressHydrationWarning":case"defaultValue":case"defaultChecked":case"innerHTML":case"ref":break;case"autoFocus":break;case"xlinkHref":if(h==null||typeof h=="function"||typeof h=="boolean"||typeof h=="symbol"){i.removeAttribute("xlink:href");break}f=dc(""+h),i.setAttributeNS("http://www.w3.org/1999/xlink","xlink:href",f);break;case"contentEditable":case"spellCheck":case"draggable":case"value":case"autoReverse":case"externalResourcesRequired":case"focusable":case"preserveAlpha":h!=null&&typeof h!="function"&&typeof h!="symbol"?i.setAttribute(f,""+h):i.removeAttribute(f);break;case"inert":case"allowFullScreen":case"async":case"autoPlay":case"controls":case"default":case"defer":case"disabled":case"disablePictureInPicture":case"disableRemotePlayback":case"formNoValidate":case"hidden":case"loop":case"noModule":case"noValidate":case"open":case"playsInline":case"readOnly":case"required":case"reversed":case"scoped":case"seamless":case"itemScope":h&&typeof h!="function"&&typeof h!="symbol"?i.setAttribute(f,""):i.removeAttribute(f);break;case"capture":case"download":h===!0?i.setAttribute(f,""):h!==!1&&h!=null&&typeof h!="function"&&typeof h!="symbol"?i.setAttribute(f,h):i.removeAttribute(f);break;case"cols":case"rows":case"size":case"span":h!=null&&typeof h!="function"&&typeof h!="symbol"&&!isNaN(h)&&1<=h?i.setAttribute(f,h):i.removeAttribute(f);break;case"rowSpan":case"start":h==null||typeof h=="function"||typeof h=="symbol"||isNaN(h)?i.removeAttribute(f):i.setAttribute(f,h);break;case"popover":pt("beforetoggle",i),pt("toggle",i),sc(i,"popover",h);break;case"xlinkActuate":ha(i,"http://www.w3.org/1999/xlink","xlink:actuate",h);break;case"xlinkArcrole":ha(i,"http://www.w3.org/1999/xlink","xlink:arcrole",h);break;case"xlinkRole":ha(i,"http://www.w3.org/1999/xlink","xlink:role",h);break;case"xlinkShow":ha(i,"http://www.w3.org/1999/xlink","xlink:show",h);break;case"xlinkTitle":ha(i,"http://www.w3.org/1999/xlink","xlink:title",h);break;case"xlinkType":ha(i,"http://www.w3.org/1999/xlink","xlink:type",h);break;case"xmlBase":ha(i,"http://www.w3.org/XML/1998/namespace","xml:base",h);break;case"xmlLang":ha(i,"http://www.w3.org/XML/1998/namespace","xml:lang",h);break;case"xmlSpace":ha(i,"http://www.w3.org/XML/1998/namespace","xml:space",h);break;case"is":sc(i,"is",h);break;case"innerText":case"textContent":break;default:(!(2<f.length)||f[0]!=="o"&&f[0]!=="O"||f[1]!=="n"&&f[1]!=="N")&&(f=i$.get(f)||f,sc(i,f,h))}}function _h(i,l,f,h,v,T){switch(f){case"style":CA(i,h,T);break;case"dangerouslySetInnerHTML":if(h!=null){if(typeof h!="object"||!("__html"in h))throw Error(r(61));if(f=h.__html,f!=null){if(v.children!=null)throw Error(r(60));i.innerHTML=f}}break;case"children":typeof h=="string"?Di(i,h):(typeof h=="number"||typeof h=="bigint")&&Di(i,""+h);break;case"onScroll":h!=null&&pt("scroll",i);break;case"onScrollEnd":h!=null&&pt("scrollend",i);break;case"onClick":h!=null&&(i.onclick=ed);break;case"suppressContentEditableWarning":case"suppressHydrationWarning":case"innerHTML":case"ref":break;case"innerText":case"textContent":break;default:if(!SA.hasOwnProperty(f))e:{if(f[0]==="o"&&f[1]==="n"&&(v=f.endsWith("Capture"),l=f.slice(2,v?f.length-7:void 0),T=i[Mn]||null,T=T!=null?T[f]:null,typeof T=="function"&&i.removeEventListener(l,T,v),typeof h=="function")){typeof T!="function"&&T!==null&&(f in i?i[f]=null:i.hasAttribute(f)&&i.removeAttribute(f)),i.addEventListener(l,h,v);break e}f in i?i[f]=h:h===!0?i.setAttribute(f,""):sc(i,f,h)}}}function hn(i,l,f){switch(l){case"div":case"span":case"svg":case"path":case"a":case"g":case"p":case"li":break;case"img":pt("error",i),pt("load",i);var h=!1,v=!1,T;for(T in f)if(f.hasOwnProperty(T)){var P=f[T];if(P!=null)switch(T){case"src":h=!0;break;case"srcSet":v=!0;break;case"children":case"dangerouslySetInnerHTML":throw Error(r(137,l));default:Rt(i,l,T,P,f,null)}}v&&Rt(i,l,"srcSet",f.srcSet,f,null),h&&Rt(i,l,"src",f.src,f,null);return;case"input":pt("invalid",i);var q=T=P=v=null,X=null,ne=null;for(h in f)if(f.hasOwnProperty(h)){var be=f[h];if(be!=null)switch(h){case"name":v=be;break;case"type":P=be;break;case"checked":X=be;break;case"defaultChecked":ne=be;break;case"value":T=be;break;case"defaultValue":q=be;break;case"children":case"dangerouslySetInnerHTML":if(be!=null)throw Error(r(137,l));break;default:Rt(i,l,h,be,f,null)}}TA(i,T,q,X,ne,P,v,!1),uc(i);return;case"select":pt("invalid",i),h=P=T=null;for(v in f)if(f.hasOwnProperty(v)&&(q=f[v],q!=null))switch(v){case"value":T=q;break;case"defaultValue":P=q;break;case"multiple":h=q;default:Rt(i,l,v,q,f,null)}l=T,f=P,i.multiple=!!h,l!=null?Oi(i,!!h,l,!1):f!=null&&Oi(i,!!h,f,!0);return;case"textarea":pt("invalid",i),T=v=h=null;for(P in f)if(f.hasOwnProperty(P)&&(q=f[P],q!=null))switch(P){case"value":h=q;break;case"defaultValue":v=q;break;case"children":T=q;break;case"dangerouslySetInnerHTML":if(q!=null)throw Error(r(91));break;default:Rt(i,l,P,q,f,null)}RA(i,h,v,T),uc(i);return;case"option":for(X in f)if(f.hasOwnProperty(X)&&(h=f[X],h!=null))switch(X){case"selected":i.selected=h&&typeof h!="function"&&typeof h!="symbol";break;default:Rt(i,l,X,h,f,null)}return;case"dialog":pt("cancel",i),pt("close",i);break;case"iframe":case"object":pt("load",i);break;case"video":case"audio":for(h=0;h<Ul.length;h++)pt(Ul[h],i);break;case"image":pt("error",i),pt("load",i);break;case"details":pt("toggle",i);break;case"embed":case"source":case"link":pt("error",i),pt("load",i);case"area":case"base":case"br":case"col":case"hr":case"keygen":case"meta":case"param":case"track":case"wbr":case"menuitem":for(ne in f)if(f.hasOwnProperty(ne)&&(h=f[ne],h!=null))switch(ne){case"children":case"dangerouslySetInnerHTML":throw Error(r(137,l));default:Rt(i,l,ne,h,f,null)}return;default:if($p(l)){for(be in f)f.hasOwnProperty(be)&&(h=f[be],h!==void 0&&_h(i,l,be,h,f,void 0));return}}for(q in f)f.hasOwnProperty(q)&&(h=f[q],h!=null&&Rt(i,l,q,h,f,null))}function D6(i,l,f,h){switch(l){case"div":case"span":case"svg":case"path":case"a":case"g":case"p":case"li":break;case"input":var v=null,T=null,P=null,q=null,X=null,ne=null,be=null;for(ge in f){var Ae=f[ge];if(f.hasOwnProperty(ge)&&Ae!=null)switch(ge){case"checked":break;case"value":break;case"defaultValue":X=Ae;default:h.hasOwnProperty(ge)||Rt(i,l,ge,null,h,Ae)}}for(var de in h){var ge=h[de];if(Ae=f[de],h.hasOwnProperty(de)&&(ge!=null||Ae!=null))switch(de){case"type":T=ge;break;case"name":v=ge;break;case"checked":ne=ge;break;case"defaultChecked":be=ge;break;case"value":P=ge;break;case"defaultValue":q=ge;break;case"children":case"dangerouslySetInnerHTML":if(ge!=null)throw Error(r(137,l));break;default:ge!==Ae&&Rt(i,l,de,ge,h,Ae)}}Gp(i,P,q,X,ne,be,T,v);return;case"select":ge=P=q=de=null;for(T in f)if(X=f[T],f.hasOwnProperty(T)&&X!=null)switch(T){case"value":break;case"multiple":ge=X;default:h.hasOwnProperty(T)||Rt(i,l,T,null,h,X)}for(v in h)if(T=h[v],X=f[v],h.hasOwnProperty(v)&&(T!=null||X!=null))switch(v){case"value":de=T;break;case"defaultValue":q=T;break;case"multiple":P=T;default:T!==X&&Rt(i,l,v,T,h,X)}l=q,f=P,h=ge,de!=null?Oi(i,!!f,de,!1):!!h!=!!f&&(l!=null?Oi(i,!!f,l,!0):Oi(i,!!f,f?[]:"",!1));return;case"textarea":ge=de=null;for(q in f)if(v=f[q],f.hasOwnProperty(q)&&v!=null&&!h.hasOwnProperty(q))switch(q){case"value":break;case"children":break;default:Rt(i,l,q,null,h,v)}for(P in h)if(v=h[P],T=f[P],h.hasOwnProperty(P)&&(v!=null||T!=null))switch(P){case"value":de=v;break;case"defaultValue":ge=v;break;case"children":break;case"dangerouslySetInnerHTML":if(v!=null)throw Error(r(91));break;default:v!==T&&Rt(i,l,P,v,h,T)}AA(i,de,ge);return;case"option":for(var qe in f)if(de=f[qe],f.hasOwnProperty(qe)&&de!=null&&!h.hasOwnProperty(qe))switch(qe){case"selected":i.selected=!1;break;default:Rt(i,l,qe,null,h,de)}for(X in h)if(de=h[X],ge=f[X],h.hasOwnProperty(X)&&de!==ge&&(de!=null||ge!=null))switch(X){case"selected":i.selected=de&&typeof de!="function"&&typeof de!="symbol";break;default:Rt(i,l,X,de,h,ge)}return;case"img":case"link":case"area":case"base":case"br":case"col":case"embed":case"hr":case"keygen":case"meta":case"param":case"source":case"track":case"wbr":case"menuitem":for(var nt in f)de=f[nt],f.hasOwnProperty(nt)&&de!=null&&!h.hasOwnProperty(nt)&&Rt(i,l,nt,null,h,de);for(ne in h)if(de=h[ne],ge=f[ne],h.hasOwnProperty(ne)&&de!==ge&&(de!=null||ge!=null))switch(ne){case"children":case"dangerouslySetInnerHTML":if(de!=null)throw Error(r(137,l));break;default:Rt(i,l,ne,de,h,ge)}return;default:if($p(l)){for(var Vt in f)de=f[Vt],f.hasOwnProperty(Vt)&&de!==void 0&&!h.hasOwnProperty(Vt)&&_h(i,l,Vt,void 0,h,de);for(be in h)de=h[be],ge=f[be],!h.hasOwnProperty(be)||de===ge||de===void 0&&ge===void 0||_h(i,l,be,de,h,ge);return}}for(var se in f)de=f[se],f.hasOwnProperty(se)&&de!=null&&!h.hasOwnProperty(se)&&Rt(i,l,se,null,h,de);for(Ae in h)de=h[Ae],ge=f[Ae],!h.hasOwnProperty(Ae)||de===ge||de==null&&ge==null||Rt(i,l,Ae,de,h,ge)}var Ch=null,Nh=null;function td(i){return i.nodeType===9?i:i.ownerDocument}function S_(i){switch(i){case"http://www.w3.org/2000/svg":return 1;case"http://www.w3.org/1998/Math/MathML":return 2;default:return 0}}function E_(i,l){if(i===0)switch(l){case"svg":return 1;case"math":return 2;default:return 0}return i===1&&l==="foreignObject"?0:i}function Oh(i,l){return i==="textarea"||i==="noscript"||typeof l.children=="string"||typeof l.children=="number"||typeof l.children=="bigint"||typeof l.dangerouslySetInnerHTML=="object"&&l.dangerouslySetInnerHTML!==null&&l.dangerouslySetInnerHTML.__html!=null}var Dh=null;function I6(){var i=window.event;return i&&i.type==="popstate"?i===Dh?!1:(Dh=i,!0):(Dh=null,!1)}var w_=typeof setTimeout=="function"?setTimeout:void 0,L6=typeof clearTimeout=="function"?clearTimeout:void 0,x_=typeof Promise=="function"?Promise:void 0,M6=typeof queueMicrotask=="function"?queueMicrotask:typeof x_<"u"?function(i){return x_.resolve(null).then(i).catch(F6)}:w_;function F6(i){setTimeout(function(){throw i})}function Ih(i,l){var f=l,h=0;do{var v=f.nextSibling;if(i.removeChild(f),v&&v.nodeType===8)if(f=v.data,f==="/$"){if(h===0){i.removeChild(v),Kl(l);return}h--}else f!=="$"&&f!=="$?"&&f!=="$!"||h++;f=v}while(f);Kl(l)}function Lh(i){var l=i.firstChild;for(l&&l.nodeType===10&&(l=l.nextSibling);l;){var f=l;switch(l=l.nextSibling,f.nodeName){case"HTML":case"HEAD":case"BODY":Lh(f),jp(f);continue;case"SCRIPT":case"STYLE":continue;case"LINK":if(f.rel.toLowerCase()==="stylesheet")continue}i.removeChild(f)}}function P6(i,l,f,h){for(;i.nodeType===1;){var v=f;if(i.nodeName.toLowerCase()!==l.toLowerCase()){if(!h&&(i.nodeName!=="INPUT"||i.type!=="hidden"))break}else if(h){if(!i[rl])switch(l){case"meta":if(!i.hasAttribute("itemprop"))break;return i;case"link":if(T=i.getAttribute("rel"),T==="stylesheet"&&i.hasAttribute("data-precedence"))break;if(T!==v.rel||i.getAttribute("href")!==(v.href==null?null:v.href)||i.getAttribute("crossorigin")!==(v.crossOrigin==null?null:v.crossOrigin)||i.getAttribute("title")!==(v.title==null?null:v.title))break;return i;case"style":if(i.hasAttribute("data-precedence"))break;return i;case"script":if(T=i.getAttribute("src"),(T!==(v.src==null?null:v.src)||i.getAttribute("type")!==(v.type==null?null:v.type)||i.getAttribute("crossorigin")!==(v.crossOrigin==null?null:v.crossOrigin))&&T&&i.hasAttribute("async")&&!i.hasAttribute("itemprop"))break;return i;default:return i}}else if(l==="input"&&i.type==="hidden"){var T=v.name==null?null:""+v.name;if(v.type==="hidden"&&i.getAttribute("name")===T)return i}else return i;if(i=Or(i.nextSibling),i===null)break}return null}function z6(i,l,f){if(l==="")return null;for(;i.nodeType!==3;)if((i.nodeType!==1||i.nodeName!=="INPUT"||i.type!=="hidden")&&!f||(i=Or(i.nextSibling),i===null))return null;return i}function Or(i){for(;i!=null;i=i.nextSibling){var l=i.nodeType;if(l===1||l===3)break;if(l===8){if(l=i.data,l==="$"||l==="$!"||l==="$?"||l==="F!"||l==="F")break;if(l==="/$")return null}}return i}function k_(i){i=i.previousSibling;for(var l=0;i;){if(i.nodeType===8){var f=i.data;if(f==="$"||f==="$!"||f==="$?"){if(l===0)return i;l--}else f==="/$"&&l++}i=i.previousSibling}return null}function T_(i,l,f){switch(l=td(f),i){case"html":if(i=l.documentElement,!i)throw Error(r(452));return i;case"head":if(i=l.head,!i)throw Error(r(453));return i;case"body":if(i=l.body,!i)throw Error(r(454));return i;default:throw Error(r(451))}}var br=new Map,A_=new Set;function nd(i){return typeof i.getRootNode=="function"?i.getRootNode():i.ownerDocument}var Na=W.d;W.d={f:B6,r:U6,D:j6,C:G6,L:H6,m:$6,X:V6,S:q6,M:W6};function B6(){var i=Na.f(),l=Yc();return i||l}function U6(i){var l=_i(i);l!==null&&l.tag===5&&l.type==="form"?J1(l):Na.r(i)}var os=typeof document>"u"?null:document;function R_(i,l,f){var h=os;if(h&&typeof l=="string"&&l){var v=sr(l);v='link[rel="'+i+'"][href="'+v+'"]',typeof f=="string"&&(v+='[crossorigin="'+f+'"]'),A_.has(v)||(A_.add(v),i={rel:i,crossOrigin:f,href:l},h.querySelector(v)===null&&(l=h.createElement("link"),hn(l,"link",i),rn(l),h.head.appendChild(l)))}}function j6(i){Na.D(i),R_("dns-prefetch",i,null)}function G6(i,l){Na.C(i,l),R_("preconnect",i,l)}function H6(i,l,f){Na.L(i,l,f);var h=os;if(h&&i&&l){var v='link[rel="preload"][as="'+sr(l)+'"]';l==="image"&&f&&f.imageSrcSet?(v+='[imagesrcset="'+sr(f.imageSrcSet)+'"]',typeof f.imageSizes=="string"&&(v+='[imagesizes="'+sr(f.imageSizes)+'"]')):v+='[href="'+sr(i)+'"]';var T=v;switch(l){case"style":T=is(i);break;case"script":T=ss(i)}br.has(T)||(i=I({rel:"preload",href:l==="image"&&f&&f.imageSrcSet?void 0:i,as:l},f),br.set(T,i),h.querySelector(v)!==null||l==="style"&&h.querySelector(Gl(T))||l==="script"&&h.querySelector(Hl(T))||(l=h.createElement("link"),hn(l,"link",i),rn(l),h.head.appendChild(l)))}}function $6(i,l){Na.m(i,l);var f=os;if(f&&i){var h=l&&typeof l.as=="string"?l.as:"script",v='link[rel="modulepreload"][as="'+sr(h)+'"][href="'+sr(i)+'"]',T=v;switch(h){case"audioworklet":case"paintworklet":case"serviceworker":case"sharedworker":case"worker":case"script":T=ss(i)}if(!br.has(T)&&(i=I({rel:"modulepreload",href:i},l),br.set(T,i),f.querySelector(v)===null)){switch(h){case"audioworklet":case"paintworklet":case"serviceworker":case"sharedworker":case"worker":case"script":if(f.querySelector(Hl(T)))return}h=f.createElement("link"),hn(h,"link",i),rn(h),f.head.appendChild(h)}}}function q6(i,l,f){Na.S(i,l,f);var h=os;if(h&&i){var v=Ci(h).hoistableStyles,T=is(i);l=l||"default";var P=v.get(T);if(!P){var q={loading:0,preload:null};if(P=h.querySelector(Gl(T)))q.loading=5;else{i=I({rel:"stylesheet",href:i,"data-precedence":l},f),(f=br.get(T))&&Mh(i,f);var X=P=h.createElement("link");rn(X),hn(X,"link",i),X._p=new Promise(function(ne,be){X.onload=ne,X.onerror=be}),X.addEventListener("load",function(){q.loading|=1}),X.addEventListener("error",function(){q.loading|=2}),q.loading|=4,rd(P,l,h)}P={type:"stylesheet",instance:P,count:1,state:q},v.set(T,P)}}}function V6(i,l){Na.X(i,l);var f=os;if(f&&i){var h=Ci(f).hoistableScripts,v=ss(i),T=h.get(v);T||(T=f.querySelector(Hl(v)),T||(i=I({src:i,async:!0},l),(l=br.get(v))&&Fh(i,l),T=f.createElement("script"),rn(T),hn(T,"link",i),f.head.appendChild(T)),T={type:"script",instance:T,count:1,state:null},h.set(v,T))}}function W6(i,l){Na.M(i,l);var f=os;if(f&&i){var h=Ci(f).hoistableScripts,v=ss(i),T=h.get(v);T||(T=f.querySelector(Hl(v)),T||(i=I({src:i,async:!0,type:"module"},l),(l=br.get(v))&&Fh(i,l),T=f.createElement("script"),rn(T),hn(T,"link",i),f.head.appendChild(T)),T={type:"script",instance:T,count:1,state:null},h.set(v,T))}}function __(i,l,f,h){var v=(v=Ie.current)?nd(v):null;if(!v)throw Error(r(446));switch(i){case"meta":case"title":return null;case"style":return typeof f.precedence=="string"&&typeof f.href=="string"?(l=is(f.href),f=Ci(v).hoistableStyles,h=f.get(l),h||(h={type:"style",instance:null,count:0,state:null},f.set(l,h)),h):{type:"void",instance:null,count:0,state:null};case"link":if(f.rel==="stylesheet"&&typeof f.href=="string"&&typeof f.precedence=="string"){i=is(f.href);var T=Ci(v).hoistableStyles,P=T.get(i);if(P||(v=v.ownerDocument||v,P={type:"stylesheet",instance:null,count:0,state:{loading:0,preload:null}},T.set(i,P),(T=v.querySelector(Gl(i)))&&!T._p&&(P.instance=T,P.state.loading=5),br.has(i)||(f={rel:"preload",as:"style",href:f.href,crossOrigin:f.crossOrigin,integrity:f.integrity,media:f.media,hrefLang:f.hrefLang,referrerPolicy:f.referrerPolicy},br.set(i,f),T||Y6(v,i,f,P.state))),l&&h===null)throw Error(r(528,""));return P}if(l&&h!==null)throw Error(r(529,""));return null;case"script":return l=f.async,f=f.src,typeof f=="string"&&l&&typeof l!="function"&&typeof l!="symbol"?(l=ss(f),f=Ci(v).hoistableScripts,h=f.get(l),h||(h={type:"script",instance:null,count:0,state:null},f.set(l,h)),h):{type:"void",instance:null,count:0,state:null};default:throw Error(r(444,i))}}function is(i){return'href="'+sr(i)+'"'}function Gl(i){return'link[rel="stylesheet"]['+i+"]"}function C_(i){return I({},i,{"data-precedence":i.precedence,precedence:null})}function Y6(i,l,f,h){i.querySelector('link[rel="preload"][as="style"]['+l+"]")?h.loading=1:(l=i.createElement("link"),h.preload=l,l.addEventListener("load",function(){return h.loading|=1}),l.addEventListener("error",function(){return h.loading|=2}),hn(l,"link",f),rn(l),i.head.appendChild(l))}function ss(i){return'[src="'+sr(i)+'"]'}function Hl(i){return"script[async]"+i}function N_(i,l,f){if(l.count++,l.instance===null)switch(l.type){case"style":var h=i.querySelector('style[data-href~="'+sr(f.href)+'"]');if(h)return l.instance=h,rn(h),h;var v=I({},f,{"data-href":f.href,"data-precedence":f.precedence,href:null,precedence:null});return h=(i.ownerDocument||i).createElement("style"),rn(h),hn(h,"style",v),rd(h,f.precedence,i),l.instance=h;case"stylesheet":v=is(f.href);var T=i.querySelector(Gl(v));if(T)return l.state.loading|=4,l.instance=T,rn(T),T;h=C_(f),(v=br.get(v))&&Mh(h,v),T=(i.ownerDocument||i).createElement("link"),rn(T);var P=T;return P._p=new Promise(function(q,X){P.onload=q,P.onerror=X}),hn(T,"link",h),l.state.loading|=4,rd(T,f.precedence,i),l.instance=T;case"script":return T=ss(f.src),(v=i.querySelector(Hl(T)))?(l.instance=v,rn(v),v):(h=f,(v=br.get(T))&&(h=I({},f),Fh(h,v)),i=i.ownerDocument||i,v=i.createElement("script"),rn(v),hn(v,"link",h),i.head.appendChild(v),l.instance=v);case"void":return null;default:throw Error(r(443,l.type))}else l.type==="stylesheet"&&!(l.state.loading&4)&&(h=l.instance,l.state.loading|=4,rd(h,f.precedence,i));return l.instance}function rd(i,l,f){for(var h=f.querySelectorAll('link[rel="stylesheet"][data-precedence],style[data-precedence]'),v=h.length?h[h.length-1]:null,T=v,P=0;P<h.length;P++){var q=h[P];if(q.dataset.precedence===l)T=q;else if(T!==v)break}T?T.parentNode.insertBefore(i,T.nextSibling):(l=f.nodeType===9?f.head:f,l.insertBefore(i,l.firstChild))}function Mh(i,l){i.crossOrigin==null&&(i.crossOrigin=l.crossOrigin),i.referrerPolicy==null&&(i.referrerPolicy=l.referrerPolicy),i.title==null&&(i.title=l.title)}function Fh(i,l){i.crossOrigin==null&&(i.crossOrigin=l.crossOrigin),i.referrerPolicy==null&&(i.referrerPolicy=l.referrerPolicy),i.integrity==null&&(i.integrity=l.integrity)}var ad=null;function O_(i,l,f){if(ad===null){var h=new Map,v=ad=new Map;v.set(f,h)}else v=ad,h=v.get(f),h||(h=new Map,v.set(f,h));if(h.has(i))return h;for(h.set(i,null),f=f.getElementsByTagName(i),v=0;v<f.length;v++){var T=f[v];if(!(T[rl]||T[Sn]||i==="link"&&T.getAttribute("rel")==="stylesheet")&&T.namespaceURI!=="http://www.w3.org/2000/svg"){var P=T.getAttribute(l)||"";P=i+P;var q=h.get(P);q?q.push(T):h.set(P,[T])}}return h}function D_(i,l,f){i=i.ownerDocument||i,i.head.insertBefore(f,l==="title"?i.querySelector("head > title"):null)}function K6(i,l,f){if(f===1||l.itemProp!=null)return!1;switch(i){case"meta":case"title":return!0;case"style":if(typeof l.precedence!="string"||typeof l.href!="string"||l.href==="")break;return!0;case"link":if(typeof l.rel!="string"||typeof l.href!="string"||l.href===""||l.onLoad||l.onError)break;switch(l.rel){case"stylesheet":return i=l.disabled,typeof l.precedence=="string"&&i==null;default:return!0}case"script":if(l.async&&typeof l.async!="function"&&typeof l.async!="symbol"&&!l.onLoad&&!l.onError&&l.src&&typeof l.src=="string")return!0}return!1}function I_(i){return!(i.type==="stylesheet"&&!(i.state.loading&3))}var $l=null;function X6(){}function Z6(i,l,f){if($l===null)throw Error(r(475));var h=$l;if(l.type==="stylesheet"&&(typeof f.media!="string"||matchMedia(f.media).matches!==!1)&&!(l.state.loading&4)){if(l.instance===null){var v=is(f.href),T=i.querySelector(Gl(v));if(T){i=T._p,i!==null&&typeof i=="object"&&typeof i.then=="function"&&(h.count++,h=od.bind(h),i.then(h,h)),l.state.loading|=4,l.instance=T,rn(T);return}T=i.ownerDocument||i,f=C_(f),(v=br.get(v))&&Mh(f,v),T=T.createElement("link"),rn(T);var P=T;P._p=new Promise(function(q,X){P.onload=q,P.onerror=X}),hn(T,"link",f),l.instance=T}h.stylesheets===null&&(h.stylesheets=new Map),h.stylesheets.set(l,i),(i=l.state.preload)&&!(l.state.loading&3)&&(h.count++,l=od.bind(h),i.addEventListener("load",l),i.addEventListener("error",l))}}function Q6(){if($l===null)throw Error(r(475));var i=$l;return i.stylesheets&&i.count===0&&Ph(i,i.stylesheets),0<i.count?function(l){var f=setTimeout(function(){if(i.stylesheets&&Ph(i,i.stylesheets),i.unsuspend){var h=i.unsuspend;i.unsuspend=null,h()}},6e4);return i.unsuspend=l,function(){i.unsuspend=null,clearTimeout(f)}}:null}function od(){if(this.count--,this.count===0){if(this.stylesheets)Ph(this,this.stylesheets);else if(this.unsuspend){var i=this.unsuspend;this.unsuspend=null,i()}}}var id=null;function Ph(i,l){i.stylesheets=null,i.unsuspend!==null&&(i.count++,id=new Map,l.forEach(J6,i),id=null,od.call(i))}function J6(i,l){if(!(l.state.loading&4)){var f=id.get(i);if(f)var h=f.get(null);else{f=new Map,id.set(i,f);for(var v=i.querySelectorAll("link[data-precedence],style[data-precedence]"),T=0;T<v.length;T++){var P=v[T];(P.nodeName==="LINK"||P.getAttribute("media")!=="not all")&&(f.set(P.dataset.precedence,P),h=P)}h&&f.set(null,h)}v=l.instance,P=v.getAttribute("data-precedence"),T=f.get(P)||h,T===h&&f.set(null,v),f.set(P,v),this.count++,h=od.bind(this),v.addEventListener("load",h),v.addEventListener("error",h),T?T.parentNode.insertBefore(v,T.nextSibling):(i=i.nodeType===9?i.head:i,i.insertBefore(v,i.firstChild)),l.state.loading|=4}}var ql={$$typeof:b,Provider:null,Consumer:null,_currentValue:Q,_currentValue2:Q,_threadCount:0};function e9(i,l,f,h,v,T,P,q){this.tag=1,this.containerInfo=i,this.finishedWork=this.pingCache=this.current=this.pendingChildren=null,this.timeoutHandle=-1,this.callbackNode=this.next=this.pendingContext=this.context=this.cancelPendingCommit=null,this.callbackPriority=0,this.expirationTimes=_n(-1),this.entangledLanes=this.shellSuspendCounter=this.errorRecoveryDisabledLanes=this.finishedLanes=this.expiredLanes=this.warmLanes=this.pingedLanes=this.suspendedLanes=this.pendingLanes=0,this.entanglements=_n(0),this.hiddenUpdates=_n(null),this.identifierPrefix=h,this.onUncaughtError=v,this.onCaughtError=T,this.onRecoverableError=P,this.pooledCache=null,this.pooledCacheLanes=0,this.formState=q,this.incompleteTransitions=new Map}function L_(i,l,f,h,v,T,P,q,X,ne,be,Ae){return i=new e9(i,l,f,P,q,X,ne,Ae),l=1,T===!0&&(l|=24),T=hr(3,null,null,l),i.current=T,T.stateNode=i,l=hg(),l.refCount++,i.pooledCache=l,l.refCount++,T.memoizedState={element:h,isDehydrated:f,cache:l},Kg(T),i}function M_(i){return i?(i=Bi,i):Bi}function F_(i,l,f,h,v,T){v=M_(v),h.context===null?h.context=v:h.pendingContext=v,h=ao(l),h.payload={element:f},T=T===void 0?null:T,T!==null&&(h.callback=T),f=oo(i,h,l),f!==null&&(Nn(f,i,l),_l(f,i,l))}function P_(i,l){if(i=i.memoizedState,i!==null&&i.dehydrated!==null){var f=i.retryLane;i.retryLane=f!==0&&f<l?f:l}}function zh(i,l){P_(i,l),(i=i.alternate)&&P_(i,l)}function z_(i){if(i.tag===13){var l=Za(i,67108864);l!==null&&Nn(l,i,67108864),zh(i,67108864)}}var sd=!0;function t9(i,l,f,h){var v=L.T;L.T=null;var T=W.p;try{W.p=2,Bh(i,l,f,h)}finally{W.p=T,L.T=v}}function n9(i,l,f,h){var v=L.T;L.T=null;var T=W.p;try{W.p=8,Bh(i,l,f,h)}finally{W.p=T,L.T=v}}function Bh(i,l,f,h){if(sd){var v=Uh(h);if(v===null)Rh(i,l,h,ld,f),U_(i,h);else if(a9(v,i,l,f,h))h.stopPropagation();else if(U_(i,h),l&4&&-1<r9.indexOf(i)){for(;v!==null;){var T=_i(v);if(T!==null)switch(T.tag){case 3:if(T=T.stateNode,T.current.memoizedState.isDehydrated){var P=Jt(T.pendingLanes);if(P!==0){var q=T;for(q.pendingLanes|=2,q.entangledLanes|=2;P;){var X=1<<31-St(P);q.entanglements[1]|=X,P&=~X}Kr(T),!(Ut&6)&&(qc=he()+500,Bl(0))}}break;case 13:q=Za(T,2),q!==null&&Nn(q,T,2),Yc(),zh(T,2)}if(T=Uh(h),T===null&&Rh(i,l,h,ld,f),T===v)break;v=T}v!==null&&h.stopPropagation()}else Rh(i,l,h,null,f)}}function Uh(i){return i=Vp(i),jh(i)}var ld=null;function jh(i){if(ld=null,i=Fo(i),i!==null){var l=z(i);if(l===null)i=null;else{var f=l.tag;if(f===13){if(i=K(l),i!==null)return i;i=null}else if(f===3){if(l.stateNode.current.memoizedState.isDehydrated)return l.tag===3?l.stateNode.containerInfo:null;i=null}else l!==i&&(i=null)}}return ld=i,null}function B_(i){switch(i){case"beforetoggle":case"cancel":case"click":case"close":case"contextmenu":case"copy":case"cut":case"auxclick":case"dblclick":case"dragend":case"dragstart":case"drop":case"focusin":case"focusout":case"input":case"invalid":case"keydown":case"keypress":case"keyup":case"mousedown":case"mouseup":case"paste":case"pause":case"play":case"pointercancel":case"pointerdown":case"pointerup":case"ratechange":case"reset":case"resize":case"seeked":case"submit":case"toggle":case"touchcancel":case"touchend":case"touchstart":case"volumechange":case"change":case"selectionchange":case"textInput":case"compositionstart":case"compositionend":case"compositionupdate":case"beforeblur":case"afterblur":case"beforeinput":case"blur":case"fullscreenchange":case"focus":case"hashchange":case"popstate":case"select":case"selectstart":return 2;case"drag":case"dragenter":case"dragexit":case"dragleave":case"dragover":case"mousemove":case"mouseout":case"mouseover":case"pointermove":case"pointerout":case"pointerover":case"scroll":case"touchmove":case"wheel":case"mouseenter":case"mouseleave":case"pointerenter":case"pointerleave":return 8;case"message":switch(Ne()){case ee:return 2;case ce:return 8;case _e:case Pe:return 32;case We:return 268435456;default:return 32}default:return 32}}var Gh=!1,po=null,go=null,ho=null,Vl=new Map,Wl=new Map,mo=[],r9="mousedown mouseup touchcancel touchend touchstart auxclick dblclick pointercancel pointerdown pointerup dragend dragstart drop compositionend compositionstart keydown keypress keyup input textInput copy cut paste click change contextmenu reset".split(" ");function U_(i,l){switch(i){case"focusin":case"focusout":po=null;break;case"dragenter":case"dragleave":go=null;break;case"mouseover":case"mouseout":ho=null;break;case"pointerover":case"pointerout":Vl.delete(l.pointerId);break;case"gotpointercapture":case"lostpointercapture":Wl.delete(l.pointerId)}}function Yl(i,l,f,h,v,T){return i===null||i.nativeEvent!==T?(i={blockedOn:l,domEventName:f,eventSystemFlags:h,nativeEvent:T,targetContainers:[v]},l!==null&&(l=_i(l),l!==null&&z_(l)),i):(i.eventSystemFlags|=h,l=i.targetContainers,v!==null&&l.indexOf(v)===-1&&l.push(v),i)}function a9(i,l,f,h,v){switch(l){case"focusin":return po=Yl(po,i,l,f,h,v),!0;case"dragenter":return go=Yl(go,i,l,f,h,v),!0;case"mouseover":return ho=Yl(ho,i,l,f,h,v),!0;case"pointerover":var T=v.pointerId;return Vl.set(T,Yl(Vl.get(T)||null,i,l,f,h,v)),!0;case"gotpointercapture":return T=v.pointerId,Wl.set(T,Yl(Wl.get(T)||null,i,l,f,h,v)),!0}return!1}function j_(i){var l=Fo(i.target);if(l!==null){var f=z(l);if(f!==null){if(l=f.tag,l===13){if(l=K(f),l!==null){i.blockedOn=l,QH(i.priority,function(){if(f.tag===13){var h=Qn(),v=Za(f,h);v!==null&&Nn(v,f,h),zh(f,h)}});return}}else if(l===3&&f.stateNode.current.memoizedState.isDehydrated){i.blockedOn=f.tag===3?f.stateNode.containerInfo:null;return}}}i.blockedOn=null}function ud(i){if(i.blockedOn!==null)return!1;for(var l=i.targetContainers;0<l.length;){var f=Uh(i.nativeEvent);if(f===null){f=i.nativeEvent;var h=new f.constructor(f.type,f);qp=h,f.target.dispatchEvent(h),qp=null}else return l=_i(f),l!==null&&z_(l),i.blockedOn=f,!1;l.shift()}return!0}function G_(i,l,f){ud(i)&&f.delete(l)}function o9(){Gh=!1,po!==null&&ud(po)&&(po=null),go!==null&&ud(go)&&(go=null),ho!==null&&ud(ho)&&(ho=null),Vl.forEach(G_),Wl.forEach(G_)}function cd(i,l){i.blockedOn===l&&(i.blockedOn=null,Gh||(Gh=!0,e.unstable_scheduleCallback(e.unstable_NormalPriority,o9)))}var dd=null;function H_(i){dd!==i&&(dd=i,e.unstable_scheduleCallback(e.unstable_NormalPriority,function(){dd===i&&(dd=null);for(var l=0;l<i.length;l+=3){var f=i[l],h=i[l+1],v=i[l+2];if(typeof h!="function"){if(jh(h||f)===null)continue;break}var T=_i(f);T!==null&&(i.splice(l,3),l-=3,Og(T,{pending:!0,data:v,method:f.method,action:h},h,v))}}))}function Kl(i){function l(X){return cd(X,i)}po!==null&&cd(po,i),go!==null&&cd(go,i),ho!==null&&cd(ho,i),Vl.forEach(l),Wl.forEach(l);for(var f=0;f<mo.length;f++){var h=mo[f];h.blockedOn===i&&(h.blockedOn=null)}for(;0<mo.length&&(f=mo[0],f.blockedOn===null);)j_(f),f.blockedOn===null&&mo.shift();if(f=(i.ownerDocument||i).$$reactFormReplay,f!=null)for(h=0;h<f.length;h+=3){var v=f[h],T=f[h+1],P=v[Mn]||null;if(typeof T=="function")P||H_(f);else if(P){var q=null;if(T&&T.hasAttribute("formAction")){if(v=T,P=T[Mn]||null)q=P.formAction;else if(jh(v)!==null)continue}else q=P.action;typeof q=="function"?f[h+1]=q:(f.splice(h,3),h-=3),H_(f)}}}function Hh(i){this._internalRoot=i}fd.prototype.render=Hh.prototype.render=function(i){var l=this._internalRoot;if(l===null)throw Error(r(409));var f=l.current,h=Qn();F_(f,h,i,l,null,null)},fd.prototype.unmount=Hh.prototype.unmount=function(){var i=this._internalRoot;if(i!==null){this._internalRoot=null;var l=i.containerInfo;i.tag===0&&ns(),F_(i.current,2,null,i,null,null),Yc(),l[Ri]=null}};function fd(i){this._internalRoot=i}fd.prototype.unstable_scheduleHydration=function(i){if(i){var l=bA();i={blockedOn:null,target:i,priority:l};for(var f=0;f<mo.length&&l!==0&&l<mo[f].priority;f++);mo.splice(f,0,i),f===0&&j_(i)}};var $_=t.version;if($_!=="19.0.0")throw Error(r(527,$_,"19.0.0"));W.findDOMNode=function(i){var l=i._reactInternals;if(l===void 0)throw typeof i.render=="function"?Error(r(188)):(i=Object.keys(i).join(","),Error(r(268,i)));return i=V(l),i=i!==null?B(i):null,i=i===null?null:i.stateNode,i};var i9={bundleType:0,version:"19.0.0",rendererPackageName:"react-dom",currentDispatcherRef:L,findFiberByHostInstance:Fo,reconcilerVersion:"19.0.0"};if(typeof __REACT_DEVTOOLS_GLOBAL_HOOK__<"u"){var pd=__REACT_DEVTOOLS_GLOBAL_HOOK__;if(!pd.isDisabled&&pd.supportsFiber)try{ht=pd.inject(i9),et=pd}catch{}}return Zl.createRoot=function(i,l){if(!a(i))throw Error(r(299));var f=!1,h="",v=sR,T=lR,P=uR,q=null;return l!=null&&(l.unstable_strictMode===!0&&(f=!0),l.identifierPrefix!==void 0&&(h=l.identifierPrefix),l.onUncaughtError!==void 0&&(v=l.onUncaughtError),l.onCaughtError!==void 0&&(T=l.onCaughtError),l.onRecoverableError!==void 0&&(P=l.onRecoverableError),l.unstable_transitionCallbacks!==void 0&&(q=l.unstable_transitionCallbacks)),l=L_(i,1,!1,null,null,f,h,v,T,P,q,null),i[Ri]=l.current,Ah(i.nodeType===8?i.parentNode:i),new Hh(l)},Zl.hydrateRoot=function(i,l,f){if(!a(i))throw Error(r(299));var h=!1,v="",T=sR,P=lR,q=uR,X=null,ne=null;return f!=null&&(f.unstable_strictMode===!0&&(h=!0),f.identifierPrefix!==void 0&&(v=f.identifierPrefix),f.onUncaughtError!==void 0&&(T=f.onUncaughtError),f.onCaughtError!==void 0&&(P=f.onCaughtError),f.onRecoverableError!==void 0&&(q=f.onRecoverableError),f.unstable_transitionCallbacks!==void 0&&(X=f.unstable_transitionCallbacks),f.formState!==void 0&&(ne=f.formState)),l=L_(i,1,!0,l,f??null,h,v,T,P,q,X,ne),l.context=M_(null),f=l.current,h=Qn(),v=ao(h),v.callback=null,oo(f,v,h),l.current.lanes=h,In(l,h),Kr(l),i[Ri]=l.current,Ah(i),new fd(l)},Zl.version="19.0.0",Zl}var eC;function S9(){if(eC)return Vh.exports;eC=1;function e(){if(!(typeof __REACT_DEVTOOLS_GLOBAL_HOOK__>"u"||typeof __REACT_DEVTOOLS_GLOBAL_HOOK__.checkDCE!="function"))try{__REACT_DEVTOOLS_GLOBAL_HOOK__.checkDCE(e)}catch(t){console.error(t)}}return e(),Vh.exports=v9(),Vh.exports}var E9=S9(),Ql={},tC;function w9(){if(tC)return Ql;tC=1,Object.defineProperty(Ql,"__esModule",{value:!0}),Ql.parse=s,Ql.serialize=d;const e=/^[\u0021-\u003A\u003C\u003E-\u007E]+$/,t=/^[\u0021-\u003A\u003C-\u007E]*$/,n=/^([.]?[a-z0-9]([a-z0-9-]{0,61}[a-z0-9])?)([.][a-z0-9]([a-z0-9-]{0,61}[a-z0-9])?)*$/i,r=/^[\u0020-\u003A\u003D-\u007E]*$/,a=Object.prototype.toString,o=(()=>{const m=function(){};return m.prototype=Object.create(null),m})();function s(m,b){const y=new o,S=m.length;if(S<2)return y;const x=(b==null?void 0:b.decode)||p;let R=0;do{const k=m.indexOf("=",R);if(k===-1)break;const A=m.indexOf(";",R),C=A===-1?S:A;if(k>C){R=m.lastIndexOf(";",k-1)+1;continue}const N=u(m,R,k),_=c(m,k,N),O=m.slice(N,_);if(y[O]===void 0){let F=u(m,k+1,C),L=c(m,C,F);const I=x(m.slice(F,L));y[O]=I}R=C+1}while(R<S);return y}function u(m,b,y){do{const S=m.charCodeAt(b);if(S!==32&&S!==9)return b}while(++b<y);return y}function c(m,b,y){for(;b>y;){const S=m.charCodeAt(--b);if(S!==32&&S!==9)return b+1}return y}function d(m,b,y){const S=(y==null?void 0:y.encode)||encodeURIComponent;if(!e.test(m))throw new TypeError(`argument name is invalid: ${m}`);const x=S(b);if(!t.test(x))throw new TypeError(`argument val is invalid: ${b}`);let R=m+"="+x;if(!y)return R;if(y.maxAge!==void 0){if(!Number.isInteger(y.maxAge))throw new TypeError(`option maxAge is invalid: ${y.maxAge}`);R+="; Max-Age="+y.maxAge}if(y.domain){if(!n.test(y.domain))throw new TypeError(`option domain is invalid: ${y.domain}`);R+="; Domain="+y.domain}if(y.path){if(!r.test(y.path))throw new TypeError(`option path is invalid: ${y.path}`);R+="; Path="+y.path}if(y.expires){if(!g(y.expires)||!Number.isFinite(y.expires.valueOf()))throw new TypeError(`option expires is invalid: ${y.expires}`);R+="; Expires="+y.expires.toUTCString()}if(y.httpOnly&&(R+="; HttpOnly"),y.secure&&(R+="; Secure"),y.partitioned&&(R+="; Partitioned"),y.priority)switch(typeof y.priority=="string"?y.priority.toLowerCase():void 0){case"low":R+="; Priority=Low";break;case"medium":R+="; Priority=Medium";break;case"high":R+="; Priority=High";break;default:throw new TypeError(`option priority is invalid: ${y.priority}`)}if(y.sameSite)switch(typeof y.sameSite=="string"?y.sameSite.toLowerCase():y.sameSite){case!0:case"strict":R+="; SameSite=Strict";break;case"lax":R+="; SameSite=Lax";break;case"none":R+="; SameSite=None";break;default:throw new TypeError(`option sameSite is invalid: ${y.sameSite}`)}return R}function p(m){if(m.indexOf("%")===-1)return m;try{return decodeURIComponent(m)}catch{return m}}function g(m){return a.call(m)==="[object Date]"}return Ql}w9();/**
 * react-router v7.3.0
 *
 * Copyright (c) Remix Software Inc.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE.md file in the root directory of this source tree.
 *
 * @license MIT
 */var nC="popstate";function x9(e={}){function t(a,o){let{pathname:s="/",search:u="",hash:c=""}=Ei(a.location.hash.substring(1));return!s.startsWith("/")&&!s.startsWith(".")&&(s="/"+s),rk("",{pathname:s,search:u,hash:c},o.state&&o.state.usr||null,o.state&&o.state.key||"default")}function n(a,o){let s=a.document.querySelector("base"),u="";if(s&&s.getAttribute("href")){let c=a.location.href,d=c.indexOf("#");u=d===-1?c:c.slice(0,d)}return u+"#"+(typeof o=="string"?o:Eu(o))}function r(a,o){Br(a.pathname.charAt(0)==="/",`relative pathnames are not supported in hash history.push(${JSON.stringify(o)})`)}return T9(t,n,r,e)}function Gt(e,t){if(e===!1||e===null||typeof e>"u")throw new Error(t)}function Br(e,t){if(!e){typeof console<"u"&&console.warn(t);try{throw new Error(t)}catch{}}}function k9(){return Math.random().toString(36).substring(2,10)}function rC(e,t){return{usr:e.state,key:e.key,idx:t}}function rk(e,t,n=null,r){return{pathname:typeof e=="string"?e:e.pathname,search:"",hash:"",...typeof t=="string"?Ei(t):t,state:n,key:t&&t.key||r||k9()}}function Eu({pathname:e="/",search:t="",hash:n=""}){return t&&t!=="?"&&(e+=t.charAt(0)==="?"?t:"?"+t),n&&n!=="#"&&(e+=n.charAt(0)==="#"?n:"#"+n),e}function Ei(e){let t={};if(e){let n=e.indexOf("#");n>=0&&(t.hash=e.substring(n),e=e.substring(0,n));let r=e.indexOf("?");r>=0&&(t.search=e.substring(r),e=e.substring(0,r)),e&&(t.pathname=e)}return t}function T9(e,t,n,r={}){let{window:a=document.defaultView,v5Compat:o=!1}=r,s=a.history,u="POP",c=null,d=p();d==null&&(d=0,s.replaceState({...s.state,idx:d},""));function p(){return(s.state||{idx:null}).idx}function g(){u="POP";let x=p(),R=x==null?null:x-d;d=x,c&&c({action:u,location:S.location,delta:R})}function m(x,R){u="PUSH";let k=rk(S.location,x,R);n&&n(k,x),d=p()+1;let A=rC(k,d),C=S.createHref(k);try{s.pushState(A,"",C)}catch(N){if(N instanceof DOMException&&N.name==="DataCloneError")throw N;a.location.assign(C)}o&&c&&c({action:u,location:S.location,delta:1})}function b(x,R){u="REPLACE";let k=rk(S.location,x,R);n&&n(k,x),d=p();let A=rC(k,d),C=S.createHref(k);s.replaceState(A,"",C),o&&c&&c({action:u,location:S.location,delta:0})}function y(x){let R=a.location.origin!=="null"?a.location.origin:a.location.href,k=typeof x=="string"?x:Eu(x);return k=k.replace(/ $/,"%20"),Gt(R,`No window.location.(origin|href) available to create URL for href: ${k}`),new URL(k,R)}let S={get action(){return u},get location(){return e(a,s)},listen(x){if(c)throw new Error("A history only accepts one active listener");return a.addEventListener(nC,g),c=x,()=>{a.removeEventListener(nC,g),c=null}},createHref(x){return t(a,x)},createURL:y,encodeLocation(x){let R=y(x);return{pathname:R.pathname,search:R.search,hash:R.hash}},push:m,replace:b,go(x){return s.go(x)}};return S}function zz(e,t,n="/"){return A9(e,t,n,!1)}function A9(e,t,n,r){let a=typeof t=="string"?Ei(t):t,o=Ua(a.pathname||"/",n);if(o==null)return null;let s=Bz(e);R9(s);let u=null;for(let c=0;u==null&&c<s.length;++c){let d=z9(o);u=F9(s[c],d,r)}return u}function Bz(e,t=[],n=[],r=""){let a=(o,s,u)=>{let c={relativePath:u===void 0?o.path||"":u,caseSensitive:o.caseSensitive===!0,childrenIndex:s,route:o};c.relativePath.startsWith("/")&&(Gt(c.relativePath.startsWith(r),`Absolute route path "${c.relativePath}" nested under path "${r}" is not valid. An absolute child route path must start with the combined path of all its parent routes.`),c.relativePath=c.relativePath.slice(r.length));let d=za([r,c.relativePath]),p=n.concat(c);o.children&&o.children.length>0&&(Gt(o.index!==!0,`Index routes must not have child routes. Please remove all child routes from route path "${d}".`),Bz(o.children,t,p,d)),!(o.path==null&&!o.index)&&t.push({path:d,score:L9(d,o.index),routesMeta:p})};return e.forEach((o,s)=>{var u;if(o.path===""||!((u=o.path)!=null&&u.includes("?")))a(o,s);else for(let c of Uz(o.path))a(o,s,c)}),t}function Uz(e){let t=e.split("/");if(t.length===0)return[];let[n,...r]=t,a=n.endsWith("?"),o=n.replace(/\?$/,"");if(r.length===0)return a?[o,""]:[o];let s=Uz(r.join("/")),u=[];return u.push(...s.map(c=>c===""?o:[o,c].join("/"))),a&&u.push(...s),u.map(c=>e.startsWith("/")&&c===""?"/":c)}function R9(e){e.sort((t,n)=>t.score!==n.score?n.score-t.score:M9(t.routesMeta.map(r=>r.childrenIndex),n.routesMeta.map(r=>r.childrenIndex)))}var _9=/^:[\w-]+$/,C9=3,N9=2,O9=1,D9=10,I9=-2,aC=e=>e==="*";function L9(e,t){let n=e.split("/"),r=n.length;return n.some(aC)&&(r+=I9),t&&(r+=N9),n.filter(a=>!aC(a)).reduce((a,o)=>a+(_9.test(o)?C9:o===""?O9:D9),r)}function M9(e,t){return e.length===t.length&&e.slice(0,-1).every((r,a)=>r===t[a])?e[e.length-1]-t[t.length-1]:0}function F9(e,t,n=!1){let{routesMeta:r}=e,a={},o="/",s=[];for(let u=0;u<r.length;++u){let c=r[u],d=u===r.length-1,p=o==="/"?t:t.slice(o.length)||"/",g=ff({path:c.relativePath,caseSensitive:c.caseSensitive,end:d},p),m=c.route;if(!g&&d&&n&&!r[r.length-1].route.index&&(g=ff({path:c.relativePath,caseSensitive:c.caseSensitive,end:!1},p)),!g)return null;Object.assign(a,g.params),s.push({params:a,pathname:za([o,g.pathname]),pathnameBase:G9(za([o,g.pathnameBase])),route:m}),g.pathnameBase!=="/"&&(o=za([o,g.pathnameBase]))}return s}function ff(e,t){typeof e=="string"&&(e={path:e,caseSensitive:!1,end:!0});let[n,r]=P9(e.path,e.caseSensitive,e.end),a=t.match(n);if(!a)return null;let o=a[0],s=o.replace(/(.)\/+$/,"$1"),u=a.slice(1);return{params:r.reduce((d,{paramName:p,isOptional:g},m)=>{if(p==="*"){let y=u[m]||"";s=o.slice(0,o.length-y.length).replace(/(.)\/+$/,"$1")}const b=u[m];return g&&!b?d[p]=void 0:d[p]=(b||"").replace(/%2F/g,"/"),d},{}),pathname:o,pathnameBase:s,pattern:e}}function P9(e,t=!1,n=!0){Br(e==="*"||!e.endsWith("*")||e.endsWith("/*"),`Route path "${e}" will be treated as if it were "${e.replace(/\*$/,"/*")}" because the \`*\` character must always follow a \`/\` in the pattern. To get rid of this warning, please change the route path to "${e.replace(/\*$/,"/*")}".`);let r=[],a="^"+e.replace(/\/*\*?$/,"").replace(/^\/*/,"/").replace(/[\\.*+^${}|()[\]]/g,"\\$&").replace(/\/:([\w-]+)(\?)?/g,(s,u,c)=>(r.push({paramName:u,isOptional:c!=null}),c?"/?([^\\/]+)?":"/([^\\/]+)"));return e.endsWith("*")?(r.push({paramName:"*"}),a+=e==="*"||e==="/*"?"(.*)$":"(?:\\/(.+)|\\/*)$"):n?a+="\\/*$":e!==""&&e!=="/"&&(a+="(?:(?=\\/|$))"),[new RegExp(a,t?void 0:"i"),r]}function z9(e){try{return e.split("/").map(t=>decodeURIComponent(t).replace(/\//g,"%2F")).join("/")}catch(t){return Br(!1,`The URL path "${e}" could not be decoded because it is a malformed URL segment. This is probably due to a bad percent encoding (${t}).`),e}}function Ua(e,t){if(t==="/")return e;if(!e.toLowerCase().startsWith(t.toLowerCase()))return null;let n=t.endsWith("/")?t.length-1:t.length,r=e.charAt(n);return r&&r!=="/"?null:e.slice(n)||"/"}function B9(e,t="/"){let{pathname:n,search:r="",hash:a=""}=typeof e=="string"?Ei(e):e;return{pathname:n?n.startsWith("/")?n:U9(n,t):t,search:H9(r),hash:$9(a)}}function U9(e,t){let n=t.replace(/\/+$/,"").split("/");return e.split("/").forEach(a=>{a===".."?n.length>1&&n.pop():a!=="."&&n.push(a)}),n.length>1?n.join("/"):"/"}function Xh(e,t,n,r){return`Cannot include a '${e}' character in a manually specified \`to.${t}\` field [${JSON.stringify(r)}].  Please separate it out to the \`to.${n}\` field. Alternatively you may provide the full path as a string in <Link to="..."> and the router will parse it for you.`}function j9(e){return e.filter((t,n)=>n===0||t.route.path&&t.route.path.length>0)}function jz(e){let t=j9(e);return t.map((n,r)=>r===t.length-1?n.pathname:n.pathnameBase)}function Gz(e,t,n,r=!1){let a;typeof e=="string"?a=Ei(e):(a={...e},Gt(!a.pathname||!a.pathname.includes("?"),Xh("?","pathname","search",a)),Gt(!a.pathname||!a.pathname.includes("#"),Xh("#","pathname","hash",a)),Gt(!a.search||!a.search.includes("#"),Xh("#","search","hash",a)));let o=e===""||a.pathname==="",s=o?"/":a.pathname,u;if(s==null)u=n;else{let g=t.length-1;if(!r&&s.startsWith("..")){let m=s.split("/");for(;m[0]==="..";)m.shift(),g-=1;a.pathname=m.join("/")}u=g>=0?t[g]:"/"}let c=B9(a,u),d=s&&s!=="/"&&s.endsWith("/"),p=(o||s===".")&&n.endsWith("/");return!c.pathname.endsWith("/")&&(d||p)&&(c.pathname+="/"),c}var za=e=>e.join("/").replace(/\/\/+/g,"/"),G9=e=>e.replace(/\/+$/,"").replace(/^\/*/,"/"),H9=e=>!e||e==="?"?"":e.startsWith("?")?e:"?"+e,$9=e=>!e||e==="#"?"":e.startsWith("#")?e:"#"+e;function q9(e){return e!=null&&typeof e.status=="number"&&typeof e.statusText=="string"&&typeof e.internal=="boolean"&&"data"in e}var Hz=["POST","PUT","PATCH","DELETE"];new Set(Hz);var V9=["GET",...Hz];new Set(V9);var Bs=E.createContext(null);Bs.displayName="DataRouter";var qf=E.createContext(null);qf.displayName="DataRouterState";var $z=E.createContext({isTransitioning:!1});$z.displayName="ViewTransition";var W9=E.createContext(new Map);W9.displayName="Fetchers";var Y9=E.createContext(null);Y9.displayName="Await";var ia=E.createContext(null);ia.displayName="Navigation";var Bu=E.createContext(null);Bu.displayName="Location";var qa=E.createContext({outlet:null,matches:[],isDataRoute:!1});qa.displayName="Route";var v0=E.createContext(null);v0.displayName="RouteError";function K9(e,{relative:t}={}){Gt(Uu(),"useHref() may be used only in the context of a <Router> component.");let{basename:n,navigator:r}=E.useContext(ia),{hash:a,pathname:o,search:s}=ju(e,{relative:t}),u=o;return n!=="/"&&(u=o==="/"?n:za([n,o])),r.createHref({pathname:u,search:s,hash:a})}function Uu(){return E.useContext(Bu)!=null}function wi(){return Gt(Uu(),"useLocation() may be used only in the context of a <Router> component."),E.useContext(Bu).location}var qz="You should call navigate() in a React.useEffect(), not when your component is first rendered.";function Vz(e){E.useContext(ia).static||E.useLayoutEffect(e)}function Vf(){let{isDataRoute:e}=E.useContext(qa);return e?lq():X9()}function X9(){Gt(Uu(),"useNavigate() may be used only in the context of a <Router> component.");let e=E.useContext(Bs),{basename:t,navigator:n}=E.useContext(ia),{matches:r}=E.useContext(qa),{pathname:a}=wi(),o=JSON.stringify(jz(r)),s=E.useRef(!1);return Vz(()=>{s.current=!0}),E.useCallback((c,d={})=>{if(Br(s.current,qz),!s.current)return;if(typeof c=="number"){n.go(c);return}let p=Gz(c,JSON.parse(o),a,d.relative==="path");e==null&&t!=="/"&&(p.pathname=p.pathname==="/"?t:za([t,p.pathname])),(d.replace?n.replace:n.push)(p,d.state,d)},[t,n,o,a,e])}E.createContext(null);function ju(e,{relative:t}={}){let{matches:n}=E.useContext(qa),{pathname:r}=wi(),a=JSON.stringify(jz(n));return E.useMemo(()=>Gz(e,JSON.parse(a),r,t==="path"),[e,a,r,t])}function Z9(e,t){return Wz(e,t)}function Wz(e,t,n,r){var k;Gt(Uu(),"useRoutes() may be used only in the context of a <Router> component.");let{navigator:a,static:o}=E.useContext(ia),{matches:s}=E.useContext(qa),u=s[s.length-1],c=u?u.params:{},d=u?u.pathname:"/",p=u?u.pathnameBase:"/",g=u&&u.route;{let A=g&&g.path||"";Yz(d,!g||A.endsWith("*")||A.endsWith("*?"),`You rendered descendant <Routes> (or called \`useRoutes()\`) at "${d}" (under <Route path="${A}">) but the parent route path has no trailing "*". This means if you navigate deeper, the parent won't match anymore and therefore the child routes will never render.

Please change the parent <Route path="${A}"> to <Route path="${A==="/"?"*":`${A}/*`}">.`)}let m=wi(),b;if(t){let A=typeof t=="string"?Ei(t):t;Gt(p==="/"||((k=A.pathname)==null?void 0:k.startsWith(p)),`When overriding the location using \`<Routes location>\` or \`useRoutes(routes, location)\`, the location pathname must begin with the portion of the URL pathname that was matched by all parent routes. The current pathname base is "${p}" but pathname "${A.pathname}" was given in the \`location\` prop.`),b=A}else b=m;let y=b.pathname||"/",S=y;if(p!=="/"){let A=p.replace(/^\//,"").split("/");S="/"+y.replace(/^\//,"").split("/").slice(A.length).join("/")}let x=!o&&n&&n.matches&&n.matches.length>0?n.matches:zz(e,{pathname:S});Br(g||x!=null,`No routes matched location "${b.pathname}${b.search}${b.hash}" `),Br(x==null||x[x.length-1].route.element!==void 0||x[x.length-1].route.Component!==void 0||x[x.length-1].route.lazy!==void 0,`Matched leaf route at location "${b.pathname}${b.search}${b.hash}" does not have an element or Component. This means it will render an <Outlet /> with a null value by default resulting in an "empty" page.`);let R=nq(x&&x.map(A=>Object.assign({},A,{params:Object.assign({},c,A.params),pathname:za([p,a.encodeLocation?a.encodeLocation(A.pathname).pathname:A.pathname]),pathnameBase:A.pathnameBase==="/"?p:za([p,a.encodeLocation?a.encodeLocation(A.pathnameBase).pathname:A.pathnameBase])})),s,n,r);return t&&R?E.createElement(Bu.Provider,{value:{location:{pathname:"/",search:"",hash:"",state:null,key:"default",...b},navigationType:"POP"}},R):R}function Q9(){let e=sq(),t=q9(e)?`${e.status} ${e.statusText}`:e instanceof Error?e.message:JSON.stringify(e),n=e instanceof Error?e.stack:null,r="rgba(200,200,200, 0.5)",a={padding:"0.5rem",backgroundColor:r},o={padding:"2px 4px",backgroundColor:r},s=null;return console.error("Error handled by React Router default ErrorBoundary:",e),s=E.createElement(E.Fragment,null,E.createElement("p",null,"💿 Hey developer 👋"),E.createElement("p",null,"You can provide a way better UX than this when your app throws errors by providing your own ",E.createElement("code",{style:o},"ErrorBoundary")," or"," ",E.createElement("code",{style:o},"errorElement")," prop on your route.")),E.createElement(E.Fragment,null,E.createElement("h2",null,"Unexpected Application Error!"),E.createElement("h3",{style:{fontStyle:"italic"}},t),n?E.createElement("pre",{style:a},n):null,s)}var J9=E.createElement(Q9,null),eq=class extends E.Component{constructor(e){super(e),this.state={location:e.location,revalidation:e.revalidation,error:e.error}}static getDerivedStateFromError(e){return{error:e}}static getDerivedStateFromProps(e,t){return t.location!==e.location||t.revalidation!=="idle"&&e.revalidation==="idle"?{error:e.error,location:e.location,revalidation:e.revalidation}:{error:e.error!==void 0?e.error:t.error,location:t.location,revalidation:e.revalidation||t.revalidation}}componentDidCatch(e,t){console.error("React Router caught the following error during render",e,t)}render(){return this.state.error!==void 0?E.createElement(qa.Provider,{value:this.props.routeContext},E.createElement(v0.Provider,{value:this.state.error,children:this.props.component})):this.props.children}};function tq({routeContext:e,match:t,children:n}){let r=E.useContext(Bs);return r&&r.static&&r.staticContext&&(t.route.errorElement||t.route.ErrorBoundary)&&(r.staticContext._deepestRenderedBoundaryId=t.route.id),E.createElement(qa.Provider,{value:e},n)}function nq(e,t=[],n=null,r=null){if(e==null){if(!n)return null;if(n.errors)e=n.matches;else if(t.length===0&&!n.initialized&&n.matches.length>0)e=n.matches;else return null}let a=e,o=n==null?void 0:n.errors;if(o!=null){let c=a.findIndex(d=>d.route.id&&(o==null?void 0:o[d.route.id])!==void 0);Gt(c>=0,`Could not find a matching route for errors on route IDs: ${Object.keys(o).join(",")}`),a=a.slice(0,Math.min(a.length,c+1))}let s=!1,u=-1;if(n)for(let c=0;c<a.length;c++){let d=a[c];if((d.route.HydrateFallback||d.route.hydrateFallbackElement)&&(u=c),d.route.id){let{loaderData:p,errors:g}=n,m=d.route.loader&&!p.hasOwnProperty(d.route.id)&&(!g||g[d.route.id]===void 0);if(d.route.lazy||m){s=!0,u>=0?a=a.slice(0,u+1):a=[a[0]];break}}}return a.reduceRight((c,d,p)=>{let g,m=!1,b=null,y=null;n&&(g=o&&d.route.id?o[d.route.id]:void 0,b=d.route.errorElement||J9,s&&(u<0&&p===0?(Yz("route-fallback",!1,"No `HydrateFallback` element provided to render during initial hydration"),m=!0,y=null):u===p&&(m=!0,y=d.route.hydrateFallbackElement||null)));let S=t.concat(a.slice(0,p+1)),x=()=>{let R;return g?R=b:m?R=y:d.route.Component?R=E.createElement(d.route.Component,null):d.route.element?R=d.route.element:R=c,E.createElement(tq,{match:d,routeContext:{outlet:c,matches:S,isDataRoute:n!=null},children:R})};return n&&(d.route.ErrorBoundary||d.route.errorElement||p===0)?E.createElement(eq,{location:n.location,revalidation:n.revalidation,component:b,error:g,children:x(),routeContext:{outlet:null,matches:S,isDataRoute:!0}}):x()},null)}function S0(e){return`${e} must be used within a data router.  See https://reactrouter.com/en/main/routers/picking-a-router.`}function rq(e){let t=E.useContext(Bs);return Gt(t,S0(e)),t}function aq(e){let t=E.useContext(qf);return Gt(t,S0(e)),t}function oq(e){let t=E.useContext(qa);return Gt(t,S0(e)),t}function E0(e){let t=oq(e),n=t.matches[t.matches.length-1];return Gt(n.route.id,`${e} can only be used on routes that contain a unique "id"`),n.route.id}function iq(){return E0("useRouteId")}function sq(){var r;let e=E.useContext(v0),t=aq("useRouteError"),n=E0("useRouteError");return e!==void 0?e:(r=t.errors)==null?void 0:r[n]}function lq(){let{router:e}=rq("useNavigate"),t=E0("useNavigate"),n=E.useRef(!1);return Vz(()=>{n.current=!0}),E.useCallback(async(a,o={})=>{Br(n.current,qz),n.current&&(typeof a=="number"?e.navigate(a):await e.navigate(a,{fromRouteId:t,...o}))},[e,t])}var oC={};function Yz(e,t,n){!t&&!oC[e]&&(oC[e]=!0,Br(!1,n))}E.memo(uq);function uq({routes:e,future:t,state:n}){return Wz(e,void 0,n,t)}function ak(e){Gt(!1,"A <Route> is only ever to be used as the child of <Routes> element, never rendered directly. Please wrap your <Route> in a <Routes>.")}function cq({basename:e="/",children:t=null,location:n,navigationType:r="POP",navigator:a,static:o=!1}){Gt(!Uu(),"You cannot render a <Router> inside another <Router>. You should never have more than one in your app.");let s=e.replace(/^\/*/,"/"),u=E.useMemo(()=>({basename:s,navigator:a,static:o,future:{}}),[s,a,o]);typeof n=="string"&&(n=Ei(n));let{pathname:c="/",search:d="",hash:p="",state:g=null,key:m="default"}=n,b=E.useMemo(()=>{let y=Ua(c,s);return y==null?null:{location:{pathname:y,search:d,hash:p,state:g,key:m},navigationType:r}},[s,c,d,p,g,m,r]);return Br(b!=null,`<Router basename="${s}"> is not able to match the URL "${c}${d}${p}" because it does not start with the basename, so the <Router> won't render anything.`),b==null?null:E.createElement(ia.Provider,{value:u},E.createElement(Bu.Provider,{children:t,value:b}))}function dq({children:e,location:t}){return Z9(ok(e),t)}function ok(e,t=[]){let n=[];return E.Children.forEach(e,(r,a)=>{if(!E.isValidElement(r))return;let o=[...t,a];if(r.type===E.Fragment){n.push.apply(n,ok(r.props.children,o));return}Gt(r.type===ak,`[${typeof r.type=="string"?r.type:r.type.name}] is not a <Route> component. All component children of <Routes> must be a <Route> or <React.Fragment>`),Gt(!r.props.index||!r.props.children,"An index route cannot have child routes.");let s={id:r.props.id||o.join("-"),caseSensitive:r.props.caseSensitive,element:r.props.element,Component:r.props.Component,index:r.props.index,path:r.props.path,loader:r.props.loader,action:r.props.action,hydrateFallbackElement:r.props.hydrateFallbackElement,HydrateFallback:r.props.HydrateFallback,errorElement:r.props.errorElement,ErrorBoundary:r.props.ErrorBoundary,hasErrorBoundary:r.props.hasErrorBoundary===!0||r.props.ErrorBoundary!=null||r.props.errorElement!=null,shouldRevalidate:r.props.shouldRevalidate,handle:r.props.handle,lazy:r.props.lazy};r.props.children&&(s.children=ok(r.props.children,o)),n.push(s)}),n}var qd="get",Vd="application/x-www-form-urlencoded";function Wf(e){return e!=null&&typeof e.tagName=="string"}function fq(e){return Wf(e)&&e.tagName.toLowerCase()==="button"}function pq(e){return Wf(e)&&e.tagName.toLowerCase()==="form"}function gq(e){return Wf(e)&&e.tagName.toLowerCase()==="input"}function hq(e){return!!(e.metaKey||e.altKey||e.ctrlKey||e.shiftKey)}function mq(e,t){return e.button===0&&(!t||t==="_self")&&!hq(e)}var gd=null;function bq(){if(gd===null)try{new FormData(document.createElement("form"),0),gd=!1}catch{gd=!0}return gd}var yq=new Set(["application/x-www-form-urlencoded","multipart/form-data","text/plain"]);function Zh(e){return e!=null&&!yq.has(e)?(Br(!1,`"${e}" is not a valid \`encType\` for \`<Form>\`/\`<fetcher.Form>\` and will default to "${Vd}"`),null):e}function vq(e,t){let n,r,a,o,s;if(pq(e)){let u=e.getAttribute("action");r=u?Ua(u,t):null,n=e.getAttribute("method")||qd,a=Zh(e.getAttribute("enctype"))||Vd,o=new FormData(e)}else if(fq(e)||gq(e)&&(e.type==="submit"||e.type==="image")){let u=e.form;if(u==null)throw new Error('Cannot submit a <button> or <input type="submit"> without a <form>');let c=e.getAttribute("formaction")||u.getAttribute("action");if(r=c?Ua(c,t):null,n=e.getAttribute("formmethod")||u.getAttribute("method")||qd,a=Zh(e.getAttribute("formenctype"))||Zh(u.getAttribute("enctype"))||Vd,o=new FormData(u,e),!bq()){let{name:d,type:p,value:g}=e;if(p==="image"){let m=d?`${d}.`:"";o.append(`${m}x`,"0"),o.append(`${m}y`,"0")}else d&&o.append(d,g)}}else{if(Wf(e))throw new Error('Cannot submit element that is not <form>, <button>, or <input type="submit|image">');n=qd,r=null,a=Vd,s=e}return o&&a==="text/plain"&&(s=o,o=void 0),{action:r,method:n.toLowerCase(),encType:a,formData:o,body:s}}function w0(e,t){if(e===!1||e===null||typeof e>"u")throw new Error(t)}async function Sq(e,t){if(e.id in t)return t[e.id];try{let n=await import(e.module);return t[e.id]=n,n}catch(n){return console.error(`Error loading route module \`${e.module}\`, reloading page...`),console.error(n),window.__reactRouterContext&&window.__reactRouterContext.isSpaMode,window.location.reload(),new Promise(()=>{})}}function Eq(e){return e==null?!1:e.href==null?e.rel==="preload"&&typeof e.imageSrcSet=="string"&&typeof e.imageSizes=="string":typeof e.rel=="string"&&typeof e.href=="string"}async function wq(e,t,n){let r=await Promise.all(e.map(async a=>{let o=t.routes[a.route.id];if(o){let s=await Sq(o,n);return s.links?s.links():[]}return[]}));return Aq(r.flat(1).filter(Eq).filter(a=>a.rel==="stylesheet"||a.rel==="preload").map(a=>a.rel==="stylesheet"?{...a,rel:"prefetch",as:"style"}:{...a,rel:"prefetch"}))}function iC(e,t,n,r,a,o){let s=(c,d)=>n[d]?c.route.id!==n[d].route.id:!0,u=(c,d)=>{var p;return n[d].pathname!==c.pathname||((p=n[d].route.path)==null?void 0:p.endsWith("*"))&&n[d].params["*"]!==c.params["*"]};return o==="assets"?t.filter((c,d)=>s(c,d)||u(c,d)):o==="data"?t.filter((c,d)=>{var g;let p=r.routes[c.route.id];if(!p||!p.hasLoader)return!1;if(s(c,d)||u(c,d))return!0;if(c.route.shouldRevalidate){let m=c.route.shouldRevalidate({currentUrl:new URL(a.pathname+a.search+a.hash,window.origin),currentParams:((g=n[0])==null?void 0:g.params)||{},nextUrl:new URL(e,window.origin),nextParams:c.params,defaultShouldRevalidate:!0});if(typeof m=="boolean")return m}return!0}):[]}function xq(e,t,{includeHydrateFallback:n}={}){return kq(e.map(r=>{let a=t.routes[r.route.id];if(!a)return[];let o=[a.module];return a.clientActionModule&&(o=o.concat(a.clientActionModule)),a.clientLoaderModule&&(o=o.concat(a.clientLoaderModule)),n&&a.hydrateFallbackModule&&(o=o.concat(a.hydrateFallbackModule)),a.imports&&(o=o.concat(a.imports)),o}).flat(1))}function kq(e){return[...new Set(e)]}function Tq(e){let t={},n=Object.keys(e).sort();for(let r of n)t[r]=e[r];return t}function Aq(e,t){let n=new Set;return new Set(t),e.reduce((r,a)=>{let o=JSON.stringify(Tq(a));return n.has(o)||(n.add(o),r.push({key:o,link:a})),r},[])}function Rq(e,t){let n=typeof e=="string"?new URL(e,typeof window>"u"?"server://singlefetch/":window.location.origin):e;return n.pathname==="/"?n.pathname="_root.data":t&&Ua(n.pathname,t)==="/"?n.pathname=`${t.replace(/\/$/,"")}/_root.data`:n.pathname=`${n.pathname.replace(/\/$/,"")}.data`,n}function Kz(){let e=E.useContext(Bs);return w0(e,"You must render this element inside a <DataRouterContext.Provider> element"),e}function _q(){let e=E.useContext(qf);return w0(e,"You must render this element inside a <DataRouterStateContext.Provider> element"),e}var x0=E.createContext(void 0);x0.displayName="FrameworkContext";function Xz(){let e=E.useContext(x0);return w0(e,"You must render this element inside a <HydratedRouter> element"),e}function Cq(e,t){let n=E.useContext(x0),[r,a]=E.useState(!1),[o,s]=E.useState(!1),{onFocus:u,onBlur:c,onMouseEnter:d,onMouseLeave:p,onTouchStart:g}=t,m=E.useRef(null);E.useEffect(()=>{if(e==="render"&&s(!0),e==="viewport"){let S=R=>{R.forEach(k=>{s(k.isIntersecting)})},x=new IntersectionObserver(S,{threshold:.5});return m.current&&x.observe(m.current),()=>{x.disconnect()}}},[e]),E.useEffect(()=>{if(r){let S=setTimeout(()=>{s(!0)},100);return()=>{clearTimeout(S)}}},[r]);let b=()=>{a(!0)},y=()=>{a(!1),s(!1)};return n?e!=="intent"?[o,m,{}]:[o,m,{onFocus:Jl(u,b),onBlur:Jl(c,y),onMouseEnter:Jl(d,b),onMouseLeave:Jl(p,y),onTouchStart:Jl(g,b)}]:[!1,m,{}]}function Jl(e,t){return n=>{e&&e(n),n.defaultPrevented||t(n)}}function Nq({page:e,...t}){let{router:n}=Kz(),r=E.useMemo(()=>zz(n.routes,e,n.basename),[n.routes,e,n.basename]);return r?E.createElement(Dq,{page:e,matches:r,...t}):null}function Oq(e){let{manifest:t,routeModules:n}=Xz(),[r,a]=E.useState([]);return E.useEffect(()=>{let o=!1;return wq(e,t,n).then(s=>{o||a(s)}),()=>{o=!0}},[e,t,n]),r}function Dq({page:e,matches:t,...n}){let r=wi(),{manifest:a,routeModules:o}=Xz(),{basename:s}=Kz(),{loaderData:u,matches:c}=_q(),d=E.useMemo(()=>iC(e,t,c,a,r,"data"),[e,t,c,a,r]),p=E.useMemo(()=>iC(e,t,c,a,r,"assets"),[e,t,c,a,r]),g=E.useMemo(()=>{if(e===r.pathname+r.search+r.hash)return[];let y=new Set,S=!1;if(t.forEach(R=>{var A;let k=a.routes[R.route.id];!k||!k.hasLoader||(!d.some(C=>C.route.id===R.route.id)&&R.route.id in u&&((A=o[R.route.id])!=null&&A.shouldRevalidate)||k.hasClientLoader?S=!0:y.add(R.route.id))}),y.size===0)return[];let x=Rq(e,s);return S&&y.size>0&&x.searchParams.set("_routes",t.filter(R=>y.has(R.route.id)).map(R=>R.route.id).join(",")),[x.pathname+x.search]},[s,u,r,a,d,t,e,o]),m=E.useMemo(()=>xq(p,a),[p,a]),b=Oq(p);return E.createElement(E.Fragment,null,g.map(y=>E.createElement("link",{key:y,rel:"prefetch",as:"fetch",href:y,...n})),m.map(y=>E.createElement("link",{key:y,rel:"modulepreload",href:y,...n})),b.map(({key:y,link:S})=>E.createElement("link",{key:y,...S})))}function Iq(...e){return t=>{e.forEach(n=>{typeof n=="function"?n(t):n!=null&&(n.current=t)})}}var Zz=typeof window<"u"&&typeof window.document<"u"&&typeof window.document.createElement<"u";try{Zz&&(window.__reactRouterVersion="7.3.0")}catch{}function Lq({basename:e,children:t,window:n}){let r=E.useRef();r.current==null&&(r.current=x9({window:n,v5Compat:!0}));let a=r.current,[o,s]=E.useState({action:a.action,location:a.location}),u=E.useCallback(c=>{E.startTransition(()=>s(c))},[s]);return E.useLayoutEffect(()=>a.listen(u),[a,u]),E.createElement(cq,{basename:e,children:t,location:o.location,navigationType:o.action,navigator:a})}var Qz=/^(?:[a-z][a-z0-9+.-]*:|\/\/)/i,Jz=E.forwardRef(function({onClick:t,discover:n="render",prefetch:r="none",relative:a,reloadDocument:o,replace:s,state:u,target:c,to:d,preventScrollReset:p,viewTransition:g,...m},b){let{basename:y}=E.useContext(ia),S=typeof d=="string"&&Qz.test(d),x,R=!1;if(typeof d=="string"&&S&&(x=d,Zz))try{let L=new URL(window.location.href),I=d.startsWith("//")?new URL(L.protocol+d):new URL(d),H=Ua(I.pathname,y);I.origin===L.origin&&H!=null?d=H+I.search+I.hash:R=!0}catch{Br(!1,`<Link to="${d}"> contains an invalid URL which will probably break when clicked - please update to a valid URL path.`)}let k=K9(d,{relative:a}),[A,C,N]=Cq(r,m),_=zq(d,{replace:s,state:u,target:c,preventScrollReset:p,relative:a,viewTransition:g});function O(L){t&&t(L),L.defaultPrevented||_(L)}let F=E.createElement("a",{...m,...N,href:x||k,onClick:R||o?t:O,ref:Iq(b,C),target:c,"data-discover":!S&&n==="render"?"true":void 0});return A&&!S?E.createElement(E.Fragment,null,F,E.createElement(Nq,{page:k})):F});Jz.displayName="Link";var Mq=E.forwardRef(function({"aria-current":t="page",caseSensitive:n=!1,className:r="",end:a=!1,style:o,to:s,viewTransition:u,children:c,...d},p){let g=ju(s,{relative:d.relative}),m=wi(),b=E.useContext(qf),{navigator:y,basename:S}=E.useContext(ia),x=b!=null&&Hq(g)&&u===!0,R=y.encodeLocation?y.encodeLocation(g).pathname:g.pathname,k=m.pathname,A=b&&b.navigation&&b.navigation.location?b.navigation.location.pathname:null;n||(k=k.toLowerCase(),A=A?A.toLowerCase():null,R=R.toLowerCase()),A&&S&&(A=Ua(A,S)||A);const C=R!=="/"&&R.endsWith("/")?R.length-1:R.length;let N=k===R||!a&&k.startsWith(R)&&k.charAt(C)==="/",_=A!=null&&(A===R||!a&&A.startsWith(R)&&A.charAt(R.length)==="/"),O={isActive:N,isPending:_,isTransitioning:x},F=N?t:void 0,L;typeof r=="function"?L=r(O):L=[r,N?"active":null,_?"pending":null,x?"transitioning":null].filter(Boolean).join(" ");let I=typeof o=="function"?o(O):o;return E.createElement(Jz,{...d,"aria-current":F,className:L,ref:p,style:I,to:s,viewTransition:u},typeof c=="function"?c(O):c)});Mq.displayName="NavLink";var Fq=E.forwardRef(({discover:e="render",fetcherKey:t,navigate:n,reloadDocument:r,replace:a,state:o,method:s=qd,action:u,onSubmit:c,relative:d,preventScrollReset:p,viewTransition:g,...m},b)=>{let y=jq(),S=Gq(u,{relative:d}),x=s.toLowerCase()==="get"?"get":"post",R=typeof u=="string"&&Qz.test(u),k=A=>{if(c&&c(A),A.defaultPrevented)return;A.preventDefault();let C=A.nativeEvent.submitter,N=(C==null?void 0:C.getAttribute("formmethod"))||s;y(C||A.currentTarget,{fetcherKey:t,method:N,navigate:n,replace:a,state:o,relative:d,preventScrollReset:p,viewTransition:g})};return E.createElement("form",{ref:b,method:x,action:S,onSubmit:r?c:k,...m,"data-discover":!R&&e==="render"?"true":void 0})});Fq.displayName="Form";function Pq(e){return`${e} must be used within a data router.  See https://reactrouter.com/en/main/routers/picking-a-router.`}function eB(e){let t=E.useContext(Bs);return Gt(t,Pq(e)),t}function zq(e,{target:t,replace:n,state:r,preventScrollReset:a,relative:o,viewTransition:s}={}){let u=Vf(),c=wi(),d=ju(e,{relative:o});return E.useCallback(p=>{if(mq(p,t)){p.preventDefault();let g=n!==void 0?n:Eu(c)===Eu(d);u(e,{replace:g,state:r,preventScrollReset:a,relative:o,viewTransition:s})}},[c,u,d,n,r,t,e,a,o,s])}var Bq=0,Uq=()=>`__${String(++Bq)}__`;function jq(){let{router:e}=eB("useSubmit"),{basename:t}=E.useContext(ia),n=iq();return E.useCallback(async(r,a={})=>{let{action:o,method:s,encType:u,formData:c,body:d}=vq(r,t);if(a.navigate===!1){let p=a.fetcherKey||Uq();await e.fetch(p,n,a.action||o,{preventScrollReset:a.preventScrollReset,formData:c,body:d,formMethod:a.method||s,formEncType:a.encType||u,flushSync:a.flushSync})}else await e.navigate(a.action||o,{preventScrollReset:a.preventScrollReset,formData:c,body:d,formMethod:a.method||s,formEncType:a.encType||u,replace:a.replace,state:a.state,fromRouteId:n,flushSync:a.flushSync,viewTransition:a.viewTransition})},[e,t,n])}function Gq(e,{relative:t}={}){let{basename:n}=E.useContext(ia),r=E.useContext(qa);Gt(r,"useFormAction must be used inside a RouteContext");let[a]=r.matches.slice(-1),o={...ju(e||".",{relative:t})},s=wi();if(e==null){o.search=s.search;let u=new URLSearchParams(o.search),c=u.getAll("index");if(c.some(p=>p==="")){u.delete("index"),c.filter(g=>g).forEach(g=>u.append("index",g));let p=u.toString();o.search=p?`?${p}`:""}}return(!e||e===".")&&a.route.index&&(o.search=o.search?o.search.replace(/^\?/,"?index&"):"?index"),n!=="/"&&(o.pathname=o.pathname==="/"?n:za([n,o.pathname])),Eu(o)}function Hq(e,t={}){let n=E.useContext($z);Gt(n!=null,"`useViewTransitionState` must be used within `react-router-dom`'s `RouterProvider`.  Did you accidentally import `RouterProvider` from `react-router`?");let{basename:r}=eB("useViewTransitionState"),a=ju(e,{relative:t.relative});if(!n.isTransitioning)return!1;let o=Ua(n.currentLocation.pathname,r)||n.currentLocation.pathname,s=Ua(n.nextLocation.pathname,r)||n.nextLocation.pathname;return ff(a.pathname,s)!=null||ff(a.pathname,o)!=null}new TextEncoder;var Gu=Pz();const tB=un(Gu),sC=e=>{let t;const n=new Set,r=(d,p)=>{const g=typeof d=="function"?d(t):d;if(!Object.is(g,t)){const m=t;t=p??(typeof g!="object"||g===null)?g:Object.assign({},t,g),n.forEach(b=>b(t,m))}},a=()=>t,u={setState:r,getState:a,getInitialState:()=>c,subscribe:d=>(n.add(d),()=>n.delete(d))},c=t=e(r,a,u);return u},$q=e=>e?sC(e):sC,qq=e=>e;function Vq(e,t=qq){const n=ye.useSyncExternalStore(e.subscribe,()=>t(e.getState()),()=>t(e.getInitialState()));return ye.useDebugValue(n),n}const lC=e=>{const t=$q(e),n=r=>Vq(t,r);return Object.assign(n,t),n},Yf=e=>e?lC(e):lC;function nB(e){var t,n,r="";if(typeof e=="string"||typeof e=="number")r+=e;else if(typeof e=="object")if(Array.isArray(e)){var a=e.length;for(t=0;t<a;t++)e[t]&&(n=nB(e[t]))&&(r&&(r+=" "),r+=n)}else for(n in e)e[n]&&(r&&(r+=" "),r+=n);return r}function rB(){for(var e,t,n=0,r="",a=arguments.length;n<a;n++)(e=arguments[n])&&(t=nB(e))&&(r&&(r+=" "),r+=t);return r}const k0="-",Wq=e=>{const t=Kq(e),{conflictingClassGroups:n,conflictingClassGroupModifiers:r}=e;return{getClassGroupId:s=>{const u=s.split(k0);return u[0]===""&&u.length!==1&&u.shift(),aB(u,t)||Yq(s)},getConflictingClassGroupIds:(s,u)=>{const c=n[s]||[];return u&&r[s]?[...c,...r[s]]:c}}},aB=(e,t)=>{var s;if(e.length===0)return t.classGroupId;const n=e[0],r=t.nextPart.get(n),a=r?aB(e.slice(1),r):void 0;if(a)return a;if(t.validators.length===0)return;const o=e.join(k0);return(s=t.validators.find(({validator:u})=>u(o)))==null?void 0:s.classGroupId},uC=/^\[(.+)\]$/,Yq=e=>{if(uC.test(e)){const t=uC.exec(e)[1],n=t==null?void 0:t.substring(0,t.indexOf(":"));if(n)return"arbitrary.."+n}},Kq=e=>{const{theme:t,classGroups:n}=e,r={nextPart:new Map,validators:[]};for(const a in n)ik(n[a],r,a,t);return r},ik=(e,t,n,r)=>{e.forEach(a=>{if(typeof a=="string"){const o=a===""?t:cC(t,a);o.classGroupId=n;return}if(typeof a=="function"){if(Xq(a)){ik(a(r),t,n,r);return}t.validators.push({validator:a,classGroupId:n});return}Object.entries(a).forEach(([o,s])=>{ik(s,cC(t,o),n,r)})})},cC=(e,t)=>{let n=e;return t.split(k0).forEach(r=>{n.nextPart.has(r)||n.nextPart.set(r,{nextPart:new Map,validators:[]}),n=n.nextPart.get(r)}),n},Xq=e=>e.isThemeGetter,Zq=e=>{if(e<1)return{get:()=>{},set:()=>{}};let t=0,n=new Map,r=new Map;const a=(o,s)=>{n.set(o,s),t++,t>e&&(t=0,r=n,n=new Map)};return{get(o){let s=n.get(o);if(s!==void 0)return s;if((s=r.get(o))!==void 0)return a(o,s),s},set(o,s){n.has(o)?n.set(o,s):a(o,s)}}},sk="!",lk=":",Qq=lk.length,Jq=e=>{const{prefix:t,experimentalParseClassName:n}=e;let r=a=>{const o=[];let s=0,u=0,c=0,d;for(let y=0;y<a.length;y++){let S=a[y];if(s===0&&u===0){if(S===lk){o.push(a.slice(c,y)),c=y+Qq;continue}if(S==="/"){d=y;continue}}S==="["?s++:S==="]"?s--:S==="("?u++:S===")"&&u--}const p=o.length===0?a:a.substring(c),g=e8(p),m=g!==p,b=d&&d>c?d-c:void 0;return{modifiers:o,hasImportantModifier:m,baseClassName:g,maybePostfixModifierPosition:b}};if(t){const a=t+lk,o=r;r=s=>s.startsWith(a)?o(s.substring(a.length)):{isExternal:!0,modifiers:[],hasImportantModifier:!1,baseClassName:s,maybePostfixModifierPosition:void 0}}if(n){const a=r;r=o=>n({className:o,parseClassName:a})}return r},e8=e=>e.endsWith(sk)?e.substring(0,e.length-1):e.startsWith(sk)?e.substring(1):e,t8=e=>{const t=Object.fromEntries(e.orderSensitiveModifiers.map(r=>[r,!0]));return r=>{if(r.length<=1)return r;const a=[];let o=[];return r.forEach(s=>{s[0]==="["||t[s]?(a.push(...o.sort(),s),o=[]):o.push(s)}),a.push(...o.sort()),a}},n8=e=>({cache:Zq(e.cacheSize),parseClassName:Jq(e),sortModifiers:t8(e),...Wq(e)}),r8=/\s+/,a8=(e,t)=>{const{parseClassName:n,getClassGroupId:r,getConflictingClassGroupIds:a,sortModifiers:o}=t,s=[],u=e.trim().split(r8);let c="";for(let d=u.length-1;d>=0;d-=1){const p=u[d],{isExternal:g,modifiers:m,hasImportantModifier:b,baseClassName:y,maybePostfixModifierPosition:S}=n(p);if(g){c=p+(c.length>0?" "+c:c);continue}let x=!!S,R=r(x?y.substring(0,S):y);if(!R){if(!x){c=p+(c.length>0?" "+c:c);continue}if(R=r(y),!R){c=p+(c.length>0?" "+c:c);continue}x=!1}const k=o(m).join(":"),A=b?k+sk:k,C=A+R;if(s.includes(C))continue;s.push(C);const N=a(R,x);for(let _=0;_<N.length;++_){const O=N[_];s.push(A+O)}c=p+(c.length>0?" "+c:c)}return c};function o8(){let e=0,t,n,r="";for(;e<arguments.length;)(t=arguments[e++])&&(n=oB(t))&&(r&&(r+=" "),r+=n);return r}const oB=e=>{if(typeof e=="string")return e;let t,n="";for(let r=0;r<e.length;r++)e[r]&&(t=oB(e[r]))&&(n&&(n+=" "),n+=t);return n};function i8(e,...t){let n,r,a,o=s;function s(c){const d=t.reduce((p,g)=>g(p),e());return n=n8(d),r=n.cache.get,a=n.cache.set,o=u,u(c)}function u(c){const d=r(c);if(d)return d;const p=a8(c,n);return a(c,p),p}return function(){return o(o8.apply(null,arguments))}}const ln=e=>{const t=n=>n[e]||[];return t.isThemeGetter=!0,t},iB=/^\[(?:(\w[\w-]*):)?(.+)\]$/i,sB=/^\((?:(\w[\w-]*):)?(.+)\)$/i,s8=/^\d+\/\d+$/,l8=/^(\d+(\.\d+)?)?(xs|sm|md|lg|xl)$/,u8=/\d+(%|px|r?em|[sdl]?v([hwib]|min|max)|pt|pc|in|cm|mm|cap|ch|ex|r?lh|cq(w|h|i|b|min|max))|\b(calc|min|max|clamp)\(.+\)|^0$/,c8=/^(rgba?|hsla?|hwb|(ok)?(lab|lch))\(.+\)$/,d8=/^(inset_)?-?((\d+)?\.?(\d+)[a-z]+|0)_-?((\d+)?\.?(\d+)[a-z]+|0)/,f8=/^(url|image|image-set|cross-fade|element|(repeating-)?(linear|radial|conic)-gradient)\(.+\)$/,ls=e=>s8.test(e),ft=e=>!!e&&!Number.isNaN(Number(e)),ri=e=>!!e&&Number.isInteger(Number(e)),dC=e=>e.endsWith("%")&&ft(e.slice(0,-1)),yo=e=>l8.test(e),p8=()=>!0,g8=e=>u8.test(e)&&!c8.test(e),T0=()=>!1,h8=e=>d8.test(e),m8=e=>f8.test(e),b8=e=>!He(e)&&!$e(e),y8=e=>Us(e,cB,T0),He=e=>iB.test(e),ai=e=>Us(e,dB,g8),Qh=e=>Us(e,C8,ft),v8=e=>Us(e,lB,T0),S8=e=>Us(e,uB,m8),E8=e=>Us(e,T0,h8),$e=e=>sB.test(e),hd=e=>js(e,dB),w8=e=>js(e,N8),x8=e=>js(e,lB),k8=e=>js(e,cB),T8=e=>js(e,uB),A8=e=>js(e,O8,!0),Us=(e,t,n)=>{const r=iB.exec(e);return r?r[1]?t(r[1]):n(r[2]):!1},js=(e,t,n=!1)=>{const r=sB.exec(e);return r?r[1]?t(r[1]):n:!1},lB=e=>e==="position",R8=new Set(["image","url"]),uB=e=>R8.has(e),_8=new Set(["length","size","percentage"]),cB=e=>_8.has(e),dB=e=>e==="length",C8=e=>e==="number",N8=e=>e==="family-name",O8=e=>e==="shadow",D8=()=>{const e=ln("color"),t=ln("font"),n=ln("text"),r=ln("font-weight"),a=ln("tracking"),o=ln("leading"),s=ln("breakpoint"),u=ln("container"),c=ln("spacing"),d=ln("radius"),p=ln("shadow"),g=ln("inset-shadow"),m=ln("drop-shadow"),b=ln("blur"),y=ln("perspective"),S=ln("aspect"),x=ln("ease"),R=ln("animate"),k=()=>["auto","avoid","all","avoid-page","page","left","right","column"],A=()=>["bottom","center","left","left-bottom","left-top","right","right-bottom","right-top","top"],C=()=>["auto","hidden","clip","visible","scroll"],N=()=>["auto","contain","none"],_=()=>[$e,He,c],O=()=>[ls,"full","auto",..._()],F=()=>[ri,"none","subgrid",$e,He],L=()=>["auto",{span:["full",ri,$e,He]},$e,He],I=()=>[ri,"auto",$e,He],H=()=>["auto","min","max","fr",$e,He],$=()=>["start","end","center","between","around","evenly","stretch","baseline"],U=()=>["start","end","center","stretch"],Y=()=>["auto",..._()],Z=()=>[ls,"auto","full","dvw","dvh","lvw","lvh","svw","svh","min","max","fit",..._()],j=()=>[e,$e,He],G=()=>[dC,ai],z=()=>["","none","full",d,$e,He],K=()=>["",ft,hd,ai],D=()=>["solid","dashed","dotted","double"],V=()=>["normal","multiply","screen","overlay","darken","lighten","color-dodge","color-burn","hard-light","soft-light","difference","exclusion","hue","saturation","color","luminosity"],B=()=>["","none",b,$e,He],M=()=>["center","top","top-right","right","bottom-right","bottom","bottom-left","left","top-left",$e,He],W=()=>["none",ft,$e,He],Q=()=>["none",ft,$e,He],oe=()=>[ft,$e,He],re=()=>[ls,"full",..._()];return{cacheSize:500,theme:{animate:["spin","ping","pulse","bounce"],aspect:["video"],blur:[yo],breakpoint:[yo],color:[p8],container:[yo],"drop-shadow":[yo],ease:["in","out","in-out"],font:[b8],"font-weight":["thin","extralight","light","normal","medium","semibold","bold","extrabold","black"],"inset-shadow":[yo],leading:["none","tight","snug","normal","relaxed","loose"],perspective:["dramatic","near","normal","midrange","distant","none"],radius:[yo],shadow:[yo],spacing:["px",ft],text:[yo],tracking:["tighter","tight","normal","wide","wider","widest"]},classGroups:{aspect:[{aspect:["auto","square",ls,He,$e,S]}],container:["container"],columns:[{columns:[ft,He,$e,u]}],"break-after":[{"break-after":k()}],"break-before":[{"break-before":k()}],"break-inside":[{"break-inside":["auto","avoid","avoid-page","avoid-column"]}],"box-decoration":[{"box-decoration":["slice","clone"]}],box:[{box:["border","content"]}],display:["block","inline-block","inline","flex","inline-flex","table","inline-table","table-caption","table-cell","table-column","table-column-group","table-footer-group","table-header-group","table-row-group","table-row","flow-root","grid","inline-grid","contents","list-item","hidden"],sr:["sr-only","not-sr-only"],float:[{float:["right","left","none","start","end"]}],clear:[{clear:["left","right","both","none","start","end"]}],isolation:["isolate","isolation-auto"],"object-fit":[{object:["contain","cover","fill","none","scale-down"]}],"object-position":[{object:[...A(),He,$e]}],overflow:[{overflow:C()}],"overflow-x":[{"overflow-x":C()}],"overflow-y":[{"overflow-y":C()}],overscroll:[{overscroll:N()}],"overscroll-x":[{"overscroll-x":N()}],"overscroll-y":[{"overscroll-y":N()}],position:["static","fixed","absolute","relative","sticky"],inset:[{inset:O()}],"inset-x":[{"inset-x":O()}],"inset-y":[{"inset-y":O()}],start:[{start:O()}],end:[{end:O()}],top:[{top:O()}],right:[{right:O()}],bottom:[{bottom:O()}],left:[{left:O()}],visibility:["visible","invisible","collapse"],z:[{z:[ri,"auto",$e,He]}],basis:[{basis:[ls,"full","auto",u,..._()]}],"flex-direction":[{flex:["row","row-reverse","col","col-reverse"]}],"flex-wrap":[{flex:["nowrap","wrap","wrap-reverse"]}],flex:[{flex:[ft,ls,"auto","initial","none",He]}],grow:[{grow:["",ft,$e,He]}],shrink:[{shrink:["",ft,$e,He]}],order:[{order:[ri,"first","last","none",$e,He]}],"grid-cols":[{"grid-cols":F()}],"col-start-end":[{col:L()}],"col-start":[{"col-start":I()}],"col-end":[{"col-end":I()}],"grid-rows":[{"grid-rows":F()}],"row-start-end":[{row:L()}],"row-start":[{"row-start":I()}],"row-end":[{"row-end":I()}],"grid-flow":[{"grid-flow":["row","col","dense","row-dense","col-dense"]}],"auto-cols":[{"auto-cols":H()}],"auto-rows":[{"auto-rows":H()}],gap:[{gap:_()}],"gap-x":[{"gap-x":_()}],"gap-y":[{"gap-y":_()}],"justify-content":[{justify:[...$(),"normal"]}],"justify-items":[{"justify-items":[...U(),"normal"]}],"justify-self":[{"justify-self":["auto",...U()]}],"align-content":[{content:["normal",...$()]}],"align-items":[{items:[...U(),"baseline"]}],"align-self":[{self:["auto",...U(),"baseline"]}],"place-content":[{"place-content":$()}],"place-items":[{"place-items":[...U(),"baseline"]}],"place-self":[{"place-self":["auto",...U()]}],p:[{p:_()}],px:[{px:_()}],py:[{py:_()}],ps:[{ps:_()}],pe:[{pe:_()}],pt:[{pt:_()}],pr:[{pr:_()}],pb:[{pb:_()}],pl:[{pl:_()}],m:[{m:Y()}],mx:[{mx:Y()}],my:[{my:Y()}],ms:[{ms:Y()}],me:[{me:Y()}],mt:[{mt:Y()}],mr:[{mr:Y()}],mb:[{mb:Y()}],ml:[{ml:Y()}],"space-x":[{"space-x":_()}],"space-x-reverse":["space-x-reverse"],"space-y":[{"space-y":_()}],"space-y-reverse":["space-y-reverse"],size:[{size:Z()}],w:[{w:[u,"screen",...Z()]}],"min-w":[{"min-w":[u,"screen","none",...Z()]}],"max-w":[{"max-w":[u,"screen","none","prose",{screen:[s]},...Z()]}],h:[{h:["screen",...Z()]}],"min-h":[{"min-h":["screen","none",...Z()]}],"max-h":[{"max-h":["screen",...Z()]}],"font-size":[{text:["base",n,hd,ai]}],"font-smoothing":["antialiased","subpixel-antialiased"],"font-style":["italic","not-italic"],"font-weight":[{font:[r,$e,Qh]}],"font-stretch":[{"font-stretch":["ultra-condensed","extra-condensed","condensed","semi-condensed","normal","semi-expanded","expanded","extra-expanded","ultra-expanded",dC,He]}],"font-family":[{font:[w8,He,t]}],"fvn-normal":["normal-nums"],"fvn-ordinal":["ordinal"],"fvn-slashed-zero":["slashed-zero"],"fvn-figure":["lining-nums","oldstyle-nums"],"fvn-spacing":["proportional-nums","tabular-nums"],"fvn-fraction":["diagonal-fractions","stacked-fractions"],tracking:[{tracking:[a,$e,He]}],"line-clamp":[{"line-clamp":[ft,"none",$e,Qh]}],leading:[{leading:[o,..._()]}],"list-image":[{"list-image":["none",$e,He]}],"list-style-position":[{list:["inside","outside"]}],"list-style-type":[{list:["disc","decimal","none",$e,He]}],"text-alignment":[{text:["left","center","right","justify","start","end"]}],"placeholder-color":[{placeholder:j()}],"text-color":[{text:j()}],"text-decoration":["underline","overline","line-through","no-underline"],"text-decoration-style":[{decoration:[...D(),"wavy"]}],"text-decoration-thickness":[{decoration:[ft,"from-font","auto",$e,ai]}],"text-decoration-color":[{decoration:j()}],"underline-offset":[{"underline-offset":[ft,"auto",$e,He]}],"text-transform":["uppercase","lowercase","capitalize","normal-case"],"text-overflow":["truncate","text-ellipsis","text-clip"],"text-wrap":[{text:["wrap","nowrap","balance","pretty"]}],indent:[{indent:_()}],"vertical-align":[{align:["baseline","top","middle","bottom","text-top","text-bottom","sub","super",$e,He]}],whitespace:[{whitespace:["normal","nowrap","pre","pre-line","pre-wrap","break-spaces"]}],break:[{break:["normal","words","all","keep"]}],hyphens:[{hyphens:["none","manual","auto"]}],content:[{content:["none",$e,He]}],"bg-attachment":[{bg:["fixed","local","scroll"]}],"bg-clip":[{"bg-clip":["border","padding","content","text"]}],"bg-origin":[{"bg-origin":["border","padding","content"]}],"bg-position":[{bg:[...A(),x8,v8]}],"bg-repeat":[{bg:["no-repeat",{repeat:["","x","y","space","round"]}]}],"bg-size":[{bg:["auto","cover","contain",k8,y8]}],"bg-image":[{bg:["none",{linear:[{to:["t","tr","r","br","b","bl","l","tl"]},ri,$e,He],radial:["",$e,He],conic:[ri,$e,He]},T8,S8]}],"bg-color":[{bg:j()}],"gradient-from-pos":[{from:G()}],"gradient-via-pos":[{via:G()}],"gradient-to-pos":[{to:G()}],"gradient-from":[{from:j()}],"gradient-via":[{via:j()}],"gradient-to":[{to:j()}],rounded:[{rounded:z()}],"rounded-s":[{"rounded-s":z()}],"rounded-e":[{"rounded-e":z()}],"rounded-t":[{"rounded-t":z()}],"rounded-r":[{"rounded-r":z()}],"rounded-b":[{"rounded-b":z()}],"rounded-l":[{"rounded-l":z()}],"rounded-ss":[{"rounded-ss":z()}],"rounded-se":[{"rounded-se":z()}],"rounded-ee":[{"rounded-ee":z()}],"rounded-es":[{"rounded-es":z()}],"rounded-tl":[{"rounded-tl":z()}],"rounded-tr":[{"rounded-tr":z()}],"rounded-br":[{"rounded-br":z()}],"rounded-bl":[{"rounded-bl":z()}],"border-w":[{border:K()}],"border-w-x":[{"border-x":K()}],"border-w-y":[{"border-y":K()}],"border-w-s":[{"border-s":K()}],"border-w-e":[{"border-e":K()}],"border-w-t":[{"border-t":K()}],"border-w-r":[{"border-r":K()}],"border-w-b":[{"border-b":K()}],"border-w-l":[{"border-l":K()}],"divide-x":[{"divide-x":K()}],"divide-x-reverse":["divide-x-reverse"],"divide-y":[{"divide-y":K()}],"divide-y-reverse":["divide-y-reverse"],"border-style":[{border:[...D(),"hidden","none"]}],"divide-style":[{divide:[...D(),"hidden","none"]}],"border-color":[{border:j()}],"border-color-x":[{"border-x":j()}],"border-color-y":[{"border-y":j()}],"border-color-s":[{"border-s":j()}],"border-color-e":[{"border-e":j()}],"border-color-t":[{"border-t":j()}],"border-color-r":[{"border-r":j()}],"border-color-b":[{"border-b":j()}],"border-color-l":[{"border-l":j()}],"divide-color":[{divide:j()}],"outline-style":[{outline:[...D(),"none","hidden"]}],"outline-offset":[{"outline-offset":[ft,$e,He]}],"outline-w":[{outline:["",ft,hd,ai]}],"outline-color":[{outline:[e]}],shadow:[{shadow:["","none",p,A8,E8]}],"shadow-color":[{shadow:j()}],"inset-shadow":[{"inset-shadow":["none",$e,He,g]}],"inset-shadow-color":[{"inset-shadow":j()}],"ring-w":[{ring:K()}],"ring-w-inset":["ring-inset"],"ring-color":[{ring:j()}],"ring-offset-w":[{"ring-offset":[ft,ai]}],"ring-offset-color":[{"ring-offset":j()}],"inset-ring-w":[{"inset-ring":K()}],"inset-ring-color":[{"inset-ring":j()}],opacity:[{opacity:[ft,$e,He]}],"mix-blend":[{"mix-blend":[...V(),"plus-darker","plus-lighter"]}],"bg-blend":[{"bg-blend":V()}],filter:[{filter:["","none",$e,He]}],blur:[{blur:B()}],brightness:[{brightness:[ft,$e,He]}],contrast:[{contrast:[ft,$e,He]}],"drop-shadow":[{"drop-shadow":["","none",m,$e,He]}],grayscale:[{grayscale:["",ft,$e,He]}],"hue-rotate":[{"hue-rotate":[ft,$e,He]}],invert:[{invert:["",ft,$e,He]}],saturate:[{saturate:[ft,$e,He]}],sepia:[{sepia:["",ft,$e,He]}],"backdrop-filter":[{"backdrop-filter":["","none",$e,He]}],"backdrop-blur":[{"backdrop-blur":B()}],"backdrop-brightness":[{"backdrop-brightness":[ft,$e,He]}],"backdrop-contrast":[{"backdrop-contrast":[ft,$e,He]}],"backdrop-grayscale":[{"backdrop-grayscale":["",ft,$e,He]}],"backdrop-hue-rotate":[{"backdrop-hue-rotate":[ft,$e,He]}],"backdrop-invert":[{"backdrop-invert":["",ft,$e,He]}],"backdrop-opacity":[{"backdrop-opacity":[ft,$e,He]}],"backdrop-saturate":[{"backdrop-saturate":[ft,$e,He]}],"backdrop-sepia":[{"backdrop-sepia":["",ft,$e,He]}],"border-collapse":[{border:["collapse","separate"]}],"border-spacing":[{"border-spacing":_()}],"border-spacing-x":[{"border-spacing-x":_()}],"border-spacing-y":[{"border-spacing-y":_()}],"table-layout":[{table:["auto","fixed"]}],caption:[{caption:["top","bottom"]}],transition:[{transition:["","all","colors","opacity","shadow","transform","none",$e,He]}],"transition-behavior":[{transition:["normal","discrete"]}],duration:[{duration:[ft,"initial",$e,He]}],ease:[{ease:["linear","initial",x,$e,He]}],delay:[{delay:[ft,$e,He]}],animate:[{animate:["none",R,$e,He]}],backface:[{backface:["hidden","visible"]}],perspective:[{perspective:[y,$e,He]}],"perspective-origin":[{"perspective-origin":M()}],rotate:[{rotate:W()}],"rotate-x":[{"rotate-x":W()}],"rotate-y":[{"rotate-y":W()}],"rotate-z":[{"rotate-z":W()}],scale:[{scale:Q()}],"scale-x":[{"scale-x":Q()}],"scale-y":[{"scale-y":Q()}],"scale-z":[{"scale-z":Q()}],"scale-3d":["scale-3d"],skew:[{skew:oe()}],"skew-x":[{"skew-x":oe()}],"skew-y":[{"skew-y":oe()}],transform:[{transform:[$e,He,"","none","gpu","cpu"]}],"transform-origin":[{origin:M()}],"transform-style":[{transform:["3d","flat"]}],translate:[{translate:re()}],"translate-x":[{"translate-x":re()}],"translate-y":[{"translate-y":re()}],"translate-z":[{"translate-z":re()}],"translate-none":["translate-none"],accent:[{accent:j()}],appearance:[{appearance:["none","auto"]}],"caret-color":[{caret:j()}],"color-scheme":[{scheme:["normal","dark","light","light-dark","only-dark","only-light"]}],cursor:[{cursor:["auto","default","pointer","wait","text","move","help","not-allowed","none","context-menu","progress","cell","crosshair","vertical-text","alias","copy","no-drop","grab","grabbing","all-scroll","col-resize","row-resize","n-resize","e-resize","s-resize","w-resize","ne-resize","nw-resize","se-resize","sw-resize","ew-resize","ns-resize","nesw-resize","nwse-resize","zoom-in","zoom-out",$e,He]}],"field-sizing":[{"field-sizing":["fixed","content"]}],"pointer-events":[{"pointer-events":["auto","none"]}],resize:[{resize:["none","","y","x"]}],"scroll-behavior":[{scroll:["auto","smooth"]}],"scroll-m":[{"scroll-m":_()}],"scroll-mx":[{"scroll-mx":_()}],"scroll-my":[{"scroll-my":_()}],"scroll-ms":[{"scroll-ms":_()}],"scroll-me":[{"scroll-me":_()}],"scroll-mt":[{"scroll-mt":_()}],"scroll-mr":[{"scroll-mr":_()}],"scroll-mb":[{"scroll-mb":_()}],"scroll-ml":[{"scroll-ml":_()}],"scroll-p":[{"scroll-p":_()}],"scroll-px":[{"scroll-px":_()}],"scroll-py":[{"scroll-py":_()}],"scroll-ps":[{"scroll-ps":_()}],"scroll-pe":[{"scroll-pe":_()}],"scroll-pt":[{"scroll-pt":_()}],"scroll-pr":[{"scroll-pr":_()}],"scroll-pb":[{"scroll-pb":_()}],"scroll-pl":[{"scroll-pl":_()}],"snap-align":[{snap:["start","end","center","align-none"]}],"snap-stop":[{snap:["normal","always"]}],"snap-type":[{snap:["none","x","y","both"]}],"snap-strictness":[{snap:["mandatory","proximity"]}],touch:[{touch:["auto","none","manipulation"]}],"touch-x":[{"touch-pan":["x","left","right"]}],"touch-y":[{"touch-pan":["y","up","down"]}],"touch-pz":["touch-pinch-zoom"],select:[{select:["none","text","all","auto"]}],"will-change":[{"will-change":["auto","scroll","contents","transform",$e,He]}],fill:[{fill:["none",...j()]}],"stroke-w":[{stroke:[ft,hd,ai,Qh]}],stroke:[{stroke:["none",...j()]}],"forced-color-adjust":[{"forced-color-adjust":["auto","none"]}]},conflictingClassGroups:{overflow:["overflow-x","overflow-y"],overscroll:["overscroll-x","overscroll-y"],inset:["inset-x","inset-y","start","end","top","right","bottom","left"],"inset-x":["right","left"],"inset-y":["top","bottom"],flex:["basis","grow","shrink"],gap:["gap-x","gap-y"],p:["px","py","ps","pe","pt","pr","pb","pl"],px:["pr","pl"],py:["pt","pb"],m:["mx","my","ms","me","mt","mr","mb","ml"],mx:["mr","ml"],my:["mt","mb"],size:["w","h"],"font-size":["leading"],"fvn-normal":["fvn-ordinal","fvn-slashed-zero","fvn-figure","fvn-spacing","fvn-fraction"],"fvn-ordinal":["fvn-normal"],"fvn-slashed-zero":["fvn-normal"],"fvn-figure":["fvn-normal"],"fvn-spacing":["fvn-normal"],"fvn-fraction":["fvn-normal"],"line-clamp":["display","overflow"],rounded:["rounded-s","rounded-e","rounded-t","rounded-r","rounded-b","rounded-l","rounded-ss","rounded-se","rounded-ee","rounded-es","rounded-tl","rounded-tr","rounded-br","rounded-bl"],"rounded-s":["rounded-ss","rounded-es"],"rounded-e":["rounded-se","rounded-ee"],"rounded-t":["rounded-tl","rounded-tr"],"rounded-r":["rounded-tr","rounded-br"],"rounded-b":["rounded-br","rounded-bl"],"rounded-l":["rounded-tl","rounded-bl"],"border-spacing":["border-spacing-x","border-spacing-y"],"border-w":["border-w-s","border-w-e","border-w-t","border-w-r","border-w-b","border-w-l"],"border-w-x":["border-w-r","border-w-l"],"border-w-y":["border-w-t","border-w-b"],"border-color":["border-color-s","border-color-e","border-color-t","border-color-r","border-color-b","border-color-l"],"border-color-x":["border-color-r","border-color-l"],"border-color-y":["border-color-t","border-color-b"],translate:["translate-x","translate-y","translate-none"],"translate-none":["translate","translate-x","translate-y","translate-z"],"scroll-m":["scroll-mx","scroll-my","scroll-ms","scroll-me","scroll-mt","scroll-mr","scroll-mb","scroll-ml"],"scroll-mx":["scroll-mr","scroll-ml"],"scroll-my":["scroll-mt","scroll-mb"],"scroll-p":["scroll-px","scroll-py","scroll-ps","scroll-pe","scroll-pt","scroll-pr","scroll-pb","scroll-pl"],"scroll-px":["scroll-pr","scroll-pl"],"scroll-py":["scroll-pt","scroll-pb"],touch:["touch-x","touch-y","touch-pz"],"touch-x":["touch"],"touch-y":["touch"],"touch-pz":["touch"]},conflictingClassGroupModifiers:{"font-size":["leading"]},orderSensitiveModifiers:["before","after","placeholder","file","marker","selection","first-line","first-letter","backdrop","*","**"]}},I8=i8(D8);function Me(...e){return I8(rB(e))}function fB(){const e="0123456789abcdef";let t="#";for(let n=0;n<6;n++)t+=e.charAt(Math.floor(Math.random()*16));return t}function Sr(e){return e instanceof Error?e.message:`${e}`}const A0=e=>{const t=e;t.use={};for(const n of Object.keys(t.getState()))t.use[n]=()=>t(r=>r[n]);return t};function pB(e,t){return function(){return e.apply(t,arguments)}}const{toString:L8}=Object.prototype,{getPrototypeOf:R0}=Object,Kf=(e=>t=>{const n=L8.call(t);return e[n]||(e[n]=n.slice(8,-1).toLowerCase())})(Object.create(null)),Gr=e=>(e=e.toLowerCase(),t=>Kf(t)===e),Xf=e=>t=>typeof t===e,{isArray:Gs}=Array,wu=Xf("undefined");function M8(e){return e!==null&&!wu(e)&&e.constructor!==null&&!wu(e.constructor)&&tr(e.constructor.isBuffer)&&e.constructor.isBuffer(e)}const gB=Gr("ArrayBuffer");function F8(e){let t;return typeof ArrayBuffer<"u"&&ArrayBuffer.isView?t=ArrayBuffer.isView(e):t=e&&e.buffer&&gB(e.buffer),t}const P8=Xf("string"),tr=Xf("function"),hB=Xf("number"),Zf=e=>e!==null&&typeof e=="object",z8=e=>e===!0||e===!1,Wd=e=>{if(Kf(e)!=="object")return!1;const t=R0(e);return(t===null||t===Object.prototype||Object.getPrototypeOf(t)===null)&&!(Symbol.toStringTag in e)&&!(Symbol.iterator in e)},B8=Gr("Date"),U8=Gr("File"),j8=Gr("Blob"),G8=Gr("FileList"),H8=e=>Zf(e)&&tr(e.pipe),$8=e=>{let t;return e&&(typeof FormData=="function"&&e instanceof FormData||tr(e.append)&&((t=Kf(e))==="formdata"||t==="object"&&tr(e.toString)&&e.toString()==="[object FormData]"))},q8=Gr("URLSearchParams"),[V8,W8,Y8,K8]=["ReadableStream","Request","Response","Headers"].map(Gr),X8=e=>e.trim?e.trim():e.replace(/^[\s\uFEFF\xA0]+|[\s\uFEFF\xA0]+$/g,"");function Hu(e,t,{allOwnKeys:n=!1}={}){if(e===null||typeof e>"u")return;let r,a;if(typeof e!="object"&&(e=[e]),Gs(e))for(r=0,a=e.length;r<a;r++)t.call(null,e[r],r,e);else{const o=n?Object.getOwnPropertyNames(e):Object.keys(e),s=o.length;let u;for(r=0;r<s;r++)u=o[r],t.call(null,e[u],u,e)}}function mB(e,t){t=t.toLowerCase();const n=Object.keys(e);let r=n.length,a;for(;r-- >0;)if(a=n[r],t===a.toLowerCase())return a;return null}const di=typeof globalThis<"u"?globalThis:typeof self<"u"?self:typeof window<"u"?window:global,bB=e=>!wu(e)&&e!==di;function uk(){const{caseless:e}=bB(this)&&this||{},t={},n=(r,a)=>{const o=e&&mB(t,a)||a;Wd(t[o])&&Wd(r)?t[o]=uk(t[o],r):Wd(r)?t[o]=uk({},r):Gs(r)?t[o]=r.slice():t[o]=r};for(let r=0,a=arguments.length;r<a;r++)arguments[r]&&Hu(arguments[r],n);return t}const Z8=(e,t,n,{allOwnKeys:r}={})=>(Hu(t,(a,o)=>{n&&tr(a)?e[o]=pB(a,n):e[o]=a},{allOwnKeys:r}),e),Q8=e=>(e.charCodeAt(0)===65279&&(e=e.slice(1)),e),J8=(e,t,n,r)=>{e.prototype=Object.create(t.prototype,r),e.prototype.constructor=e,Object.defineProperty(e,"super",{value:t.prototype}),n&&Object.assign(e.prototype,n)},e7=(e,t,n,r)=>{let a,o,s;const u={};if(t=t||{},e==null)return t;do{for(a=Object.getOwnPropertyNames(e),o=a.length;o-- >0;)s=a[o],(!r||r(s,e,t))&&!u[s]&&(t[s]=e[s],u[s]=!0);e=n!==!1&&R0(e)}while(e&&(!n||n(e,t))&&e!==Object.prototype);return t},t7=(e,t,n)=>{e=String(e),(n===void 0||n>e.length)&&(n=e.length),n-=t.length;const r=e.indexOf(t,n);return r!==-1&&r===n},n7=e=>{if(!e)return null;if(Gs(e))return e;let t=e.length;if(!hB(t))return null;const n=new Array(t);for(;t-- >0;)n[t]=e[t];return n},r7=(e=>t=>e&&t instanceof e)(typeof Uint8Array<"u"&&R0(Uint8Array)),a7=(e,t)=>{const r=(e&&e[Symbol.iterator]).call(e);let a;for(;(a=r.next())&&!a.done;){const o=a.value;t.call(e,o[0],o[1])}},o7=(e,t)=>{let n;const r=[];for(;(n=e.exec(t))!==null;)r.push(n);return r},i7=Gr("HTMLFormElement"),s7=e=>e.toLowerCase().replace(/[-_\s]([a-z\d])(\w*)/g,function(n,r,a){return r.toUpperCase()+a}),fC=(({hasOwnProperty:e})=>(t,n)=>e.call(t,n))(Object.prototype),l7=Gr("RegExp"),yB=(e,t)=>{const n=Object.getOwnPropertyDescriptors(e),r={};Hu(n,(a,o)=>{let s;(s=t(a,o,e))!==!1&&(r[o]=s||a)}),Object.defineProperties(e,r)},u7=e=>{yB(e,(t,n)=>{if(tr(e)&&["arguments","caller","callee"].indexOf(n)!==-1)return!1;const r=e[n];if(tr(r)){if(t.enumerable=!1,"writable"in t){t.writable=!1;return}t.set||(t.set=()=>{throw Error("Can not rewrite read-only method '"+n+"'")})}})},c7=(e,t)=>{const n={},r=a=>{a.forEach(o=>{n[o]=!0})};return Gs(e)?r(e):r(String(e).split(t)),n},d7=()=>{},f7=(e,t)=>e!=null&&Number.isFinite(e=+e)?e:t,Jh="abcdefghijklmnopqrstuvwxyz",pC="0123456789",vB={DIGIT:pC,ALPHA:Jh,ALPHA_DIGIT:Jh+Jh.toUpperCase()+pC},p7=(e=16,t=vB.ALPHA_DIGIT)=>{let n="";const{length:r}=t;for(;e--;)n+=t[Math.random()*r|0];return n};function g7(e){return!!(e&&tr(e.append)&&e[Symbol.toStringTag]==="FormData"&&e[Symbol.iterator])}const h7=e=>{const t=new Array(10),n=(r,a)=>{if(Zf(r)){if(t.indexOf(r)>=0)return;if(!("toJSON"in r)){t[a]=r;const o=Gs(r)?[]:{};return Hu(r,(s,u)=>{const c=n(s,a+1);!wu(c)&&(o[u]=c)}),t[a]=void 0,o}}return r};return n(e,0)},m7=Gr("AsyncFunction"),b7=e=>e&&(Zf(e)||tr(e))&&tr(e.then)&&tr(e.catch),SB=((e,t)=>e?setImmediate:t?((n,r)=>(di.addEventListener("message",({source:a,data:o})=>{a===di&&o===n&&r.length&&r.shift()()},!1),a=>{r.push(a),di.postMessage(n,"*")}))(`axios@${Math.random()}`,[]):n=>setTimeout(n))(typeof setImmediate=="function",tr(di.postMessage)),y7=typeof queueMicrotask<"u"?queueMicrotask.bind(di):typeof process<"u"&&process.nextTick||SB,pe={isArray:Gs,isArrayBuffer:gB,isBuffer:M8,isFormData:$8,isArrayBufferView:F8,isString:P8,isNumber:hB,isBoolean:z8,isObject:Zf,isPlainObject:Wd,isReadableStream:V8,isRequest:W8,isResponse:Y8,isHeaders:K8,isUndefined:wu,isDate:B8,isFile:U8,isBlob:j8,isRegExp:l7,isFunction:tr,isStream:H8,isURLSearchParams:q8,isTypedArray:r7,isFileList:G8,forEach:Hu,merge:uk,extend:Z8,trim:X8,stripBOM:Q8,inherits:J8,toFlatObject:e7,kindOf:Kf,kindOfTest:Gr,endsWith:t7,toArray:n7,forEachEntry:a7,matchAll:o7,isHTMLForm:i7,hasOwnProperty:fC,hasOwnProp:fC,reduceDescriptors:yB,freezeMethods:u7,toObjectSet:c7,toCamelCase:s7,noop:d7,toFiniteNumber:f7,findKey:mB,global:di,isContextDefined:bB,ALPHABET:vB,generateString:p7,isSpecCompliantForm:g7,toJSONObject:h7,isAsyncFn:m7,isThenable:b7,setImmediate:SB,asap:y7};function at(e,t,n,r,a){Error.call(this),Error.captureStackTrace?Error.captureStackTrace(this,this.constructor):this.stack=new Error().stack,this.message=e,this.name="AxiosError",t&&(this.code=t),n&&(this.config=n),r&&(this.request=r),a&&(this.response=a,this.status=a.status?a.status:null)}pe.inherits(at,Error,{toJSON:function(){return{message:this.message,name:this.name,description:this.description,number:this.number,fileName:this.fileName,lineNumber:this.lineNumber,columnNumber:this.columnNumber,stack:this.stack,config:pe.toJSONObject(this.config),code:this.code,status:this.status}}});const EB=at.prototype,wB={};["ERR_BAD_OPTION_VALUE","ERR_BAD_OPTION","ECONNABORTED","ETIMEDOUT","ERR_NETWORK","ERR_FR_TOO_MANY_REDIRECTS","ERR_DEPRECATED","ERR_BAD_RESPONSE","ERR_BAD_REQUEST","ERR_CANCELED","ERR_NOT_SUPPORT","ERR_INVALID_URL"].forEach(e=>{wB[e]={value:e}});Object.defineProperties(at,wB);Object.defineProperty(EB,"isAxiosError",{value:!0});at.from=(e,t,n,r,a,o)=>{const s=Object.create(EB);return pe.toFlatObject(e,s,function(c){return c!==Error.prototype},u=>u!=="isAxiosError"),at.call(s,e.message,t,n,r,a),s.cause=e,s.name=e.name,o&&Object.assign(s,o),s};const v7=null;function ck(e){return pe.isPlainObject(e)||pe.isArray(e)}function xB(e){return pe.endsWith(e,"[]")?e.slice(0,-2):e}function gC(e,t,n){return e?e.concat(t).map(function(a,o){return a=xB(a),!n&&o?"["+a+"]":a}).join(n?".":""):t}function S7(e){return pe.isArray(e)&&!e.some(ck)}const E7=pe.toFlatObject(pe,{},null,function(t){return/^is[A-Z]/.test(t)});function Qf(e,t,n){if(!pe.isObject(e))throw new TypeError("target must be an object");t=t||new FormData,n=pe.toFlatObject(n,{metaTokens:!0,dots:!1,indexes:!1},!1,function(S,x){return!pe.isUndefined(x[S])});const r=n.metaTokens,a=n.visitor||p,o=n.dots,s=n.indexes,c=(n.Blob||typeof Blob<"u"&&Blob)&&pe.isSpecCompliantForm(t);if(!pe.isFunction(a))throw new TypeError("visitor must be a function");function d(y){if(y===null)return"";if(pe.isDate(y))return y.toISOString();if(!c&&pe.isBlob(y))throw new at("Blob is not supported. Use a Buffer instead.");return pe.isArrayBuffer(y)||pe.isTypedArray(y)?c&&typeof Blob=="function"?new Blob([y]):Buffer.from(y):y}function p(y,S,x){let R=y;if(y&&!x&&typeof y=="object"){if(pe.endsWith(S,"{}"))S=r?S:S.slice(0,-2),y=JSON.stringify(y);else if(pe.isArray(y)&&S7(y)||(pe.isFileList(y)||pe.endsWith(S,"[]"))&&(R=pe.toArray(y)))return S=xB(S),R.forEach(function(A,C){!(pe.isUndefined(A)||A===null)&&t.append(s===!0?gC([S],C,o):s===null?S:S+"[]",d(A))}),!1}return ck(y)?!0:(t.append(gC(x,S,o),d(y)),!1)}const g=[],m=Object.assign(E7,{defaultVisitor:p,convertValue:d,isVisitable:ck});function b(y,S){if(!pe.isUndefined(y)){if(g.indexOf(y)!==-1)throw Error("Circular reference detected in "+S.join("."));g.push(y),pe.forEach(y,function(R,k){(!(pe.isUndefined(R)||R===null)&&a.call(t,R,pe.isString(k)?k.trim():k,S,m))===!0&&b(R,S?S.concat(k):[k])}),g.pop()}}if(!pe.isObject(e))throw new TypeError("data must be an object");return b(e),t}function hC(e){const t={"!":"%21","'":"%27","(":"%28",")":"%29","~":"%7E","%20":"+","%00":"\0"};return encodeURIComponent(e).replace(/[!'()~]|%20|%00/g,function(r){return t[r]})}function _0(e,t){this._pairs=[],e&&Qf(e,this,t)}const kB=_0.prototype;kB.append=function(t,n){this._pairs.push([t,n])};kB.toString=function(t){const n=t?function(r){return t.call(this,r,hC)}:hC;return this._pairs.map(function(a){return n(a[0])+"="+n(a[1])},"").join("&")};function w7(e){return encodeURIComponent(e).replace(/%3A/gi,":").replace(/%24/g,"$").replace(/%2C/gi,",").replace(/%20/g,"+").replace(/%5B/gi,"[").replace(/%5D/gi,"]")}function TB(e,t,n){if(!t)return e;const r=n&&n.encode||w7;pe.isFunction(n)&&(n={serialize:n});const a=n&&n.serialize;let o;if(a?o=a(t,n):o=pe.isURLSearchParams(t)?t.toString():new _0(t,n).toString(r),o){const s=e.indexOf("#");s!==-1&&(e=e.slice(0,s)),e+=(e.indexOf("?")===-1?"?":"&")+o}return e}class mC{constructor(){this.handlers=[]}use(t,n,r){return this.handlers.push({fulfilled:t,rejected:n,synchronous:r?r.synchronous:!1,runWhen:r?r.runWhen:null}),this.handlers.length-1}eject(t){this.handlers[t]&&(this.handlers[t]=null)}clear(){this.handlers&&(this.handlers=[])}forEach(t){pe.forEach(this.handlers,function(r){r!==null&&t(r)})}}const AB={silentJSONParsing:!0,forcedJSONParsing:!0,clarifyTimeoutError:!1},x7=typeof URLSearchParams<"u"?URLSearchParams:_0,k7=typeof FormData<"u"?FormData:null,T7=typeof Blob<"u"?Blob:null,A7={isBrowser:!0,classes:{URLSearchParams:x7,FormData:k7,Blob:T7},protocols:["http","https","file","blob","url","data"]},C0=typeof window<"u"&&typeof document<"u",dk=typeof navigator=="object"&&navigator||void 0,R7=C0&&(!dk||["ReactNative","NativeScript","NS"].indexOf(dk.product)<0),_7=typeof WorkerGlobalScope<"u"&&self instanceof WorkerGlobalScope&&typeof self.importScripts=="function",C7=C0&&window.location.href||"http://localhost",N7=Object.freeze(Object.defineProperty({__proto__:null,hasBrowserEnv:C0,hasStandardBrowserEnv:R7,hasStandardBrowserWebWorkerEnv:_7,navigator:dk,origin:C7},Symbol.toStringTag,{value:"Module"})),xn={...N7,...A7};function O7(e,t){return Qf(e,new xn.classes.URLSearchParams,Object.assign({visitor:function(n,r,a,o){return xn.isNode&&pe.isBuffer(n)?(this.append(r,n.toString("base64")),!1):o.defaultVisitor.apply(this,arguments)}},t))}function D7(e){return pe.matchAll(/\w+|\[(\w*)]/g,e).map(t=>t[0]==="[]"?"":t[1]||t[0])}function I7(e){const t={},n=Object.keys(e);let r;const a=n.length;let o;for(r=0;r<a;r++)o=n[r],t[o]=e[o];return t}function RB(e){function t(n,r,a,o){let s=n[o++];if(s==="__proto__")return!0;const u=Number.isFinite(+s),c=o>=n.length;return s=!s&&pe.isArray(a)?a.length:s,c?(pe.hasOwnProp(a,s)?a[s]=[a[s],r]:a[s]=r,!u):((!a[s]||!pe.isObject(a[s]))&&(a[s]=[]),t(n,r,a[s],o)&&pe.isArray(a[s])&&(a[s]=I7(a[s])),!u)}if(pe.isFormData(e)&&pe.isFunction(e.entries)){const n={};return pe.forEachEntry(e,(r,a)=>{t(D7(r),a,n,0)}),n}return null}function L7(e,t,n){if(pe.isString(e))try{return(t||JSON.parse)(e),pe.trim(e)}catch(r){if(r.name!=="SyntaxError")throw r}return(n||JSON.stringify)(e)}const $u={transitional:AB,adapter:["xhr","http","fetch"],transformRequest:[function(t,n){const r=n.getContentType()||"",a=r.indexOf("application/json")>-1,o=pe.isObject(t);if(o&&pe.isHTMLForm(t)&&(t=new FormData(t)),pe.isFormData(t))return a?JSON.stringify(RB(t)):t;if(pe.isArrayBuffer(t)||pe.isBuffer(t)||pe.isStream(t)||pe.isFile(t)||pe.isBlob(t)||pe.isReadableStream(t))return t;if(pe.isArrayBufferView(t))return t.buffer;if(pe.isURLSearchParams(t))return n.setContentType("application/x-www-form-urlencoded;charset=utf-8",!1),t.toString();let u;if(o){if(r.indexOf("application/x-www-form-urlencoded")>-1)return O7(t,this.formSerializer).toString();if((u=pe.isFileList(t))||r.indexOf("multipart/form-data")>-1){const c=this.env&&this.env.FormData;return Qf(u?{"files[]":t}:t,c&&new c,this.formSerializer)}}return o||a?(n.setContentType("application/json",!1),L7(t)):t}],transformResponse:[function(t){const n=this.transitional||$u.transitional,r=n&&n.forcedJSONParsing,a=this.responseType==="json";if(pe.isResponse(t)||pe.isReadableStream(t))return t;if(t&&pe.isString(t)&&(r&&!this.responseType||a)){const s=!(n&&n.silentJSONParsing)&&a;try{return JSON.parse(t)}catch(u){if(s)throw u.name==="SyntaxError"?at.from(u,at.ERR_BAD_RESPONSE,this,null,this.response):u}}return t}],timeout:0,xsrfCookieName:"XSRF-TOKEN",xsrfHeaderName:"X-XSRF-TOKEN",maxContentLength:-1,maxBodyLength:-1,env:{FormData:xn.classes.FormData,Blob:xn.classes.Blob},validateStatus:function(t){return t>=200&&t<300},headers:{common:{Accept:"application/json, text/plain, */*","Content-Type":void 0}}};pe.forEach(["delete","get","head","post","put","patch"],e=>{$u.headers[e]={}});const M7=pe.toObjectSet(["age","authorization","content-length","content-type","etag","expires","from","host","if-modified-since","if-unmodified-since","last-modified","location","max-forwards","proxy-authorization","referer","retry-after","user-agent"]),F7=e=>{const t={};let n,r,a;return e&&e.split(`
`).forEach(function(s){a=s.indexOf(":"),n=s.substring(0,a).trim().toLowerCase(),r=s.substring(a+1).trim(),!(!n||t[n]&&M7[n])&&(n==="set-cookie"?t[n]?t[n].push(r):t[n]=[r]:t[n]=t[n]?t[n]+", "+r:r)}),t},bC=Symbol("internals");function eu(e){return e&&String(e).trim().toLowerCase()}function Yd(e){return e===!1||e==null?e:pe.isArray(e)?e.map(Yd):String(e)}function P7(e){const t=Object.create(null),n=/([^\s,;=]+)\s*(?:=\s*([^,;]+))?/g;let r;for(;r=n.exec(e);)t[r[1]]=r[2];return t}const z7=e=>/^[-_a-zA-Z0-9^`|~,!#$%&'*+.]+$/.test(e.trim());function em(e,t,n,r,a){if(pe.isFunction(r))return r.call(this,t,n);if(a&&(t=n),!!pe.isString(t)){if(pe.isString(r))return t.indexOf(r)!==-1;if(pe.isRegExp(r))return r.test(t)}}function B7(e){return e.trim().toLowerCase().replace(/([a-z\d])(\w*)/g,(t,n,r)=>n.toUpperCase()+r)}function U7(e,t){const n=pe.toCamelCase(" "+t);["get","set","has"].forEach(r=>{Object.defineProperty(e,r+n,{value:function(a,o,s){return this[r].call(this,t,a,o,s)},configurable:!0})})}let Gn=class{constructor(t){t&&this.set(t)}set(t,n,r){const a=this;function o(u,c,d){const p=eu(c);if(!p)throw new Error("header name must be a non-empty string");const g=pe.findKey(a,p);(!g||a[g]===void 0||d===!0||d===void 0&&a[g]!==!1)&&(a[g||c]=Yd(u))}const s=(u,c)=>pe.forEach(u,(d,p)=>o(d,p,c));if(pe.isPlainObject(t)||t instanceof this.constructor)s(t,n);else if(pe.isString(t)&&(t=t.trim())&&!z7(t))s(F7(t),n);else if(pe.isHeaders(t))for(const[u,c]of t.entries())o(c,u,r);else t!=null&&o(n,t,r);return this}get(t,n){if(t=eu(t),t){const r=pe.findKey(this,t);if(r){const a=this[r];if(!n)return a;if(n===!0)return P7(a);if(pe.isFunction(n))return n.call(this,a,r);if(pe.isRegExp(n))return n.exec(a);throw new TypeError("parser must be boolean|regexp|function")}}}has(t,n){if(t=eu(t),t){const r=pe.findKey(this,t);return!!(r&&this[r]!==void 0&&(!n||em(this,this[r],r,n)))}return!1}delete(t,n){const r=this;let a=!1;function o(s){if(s=eu(s),s){const u=pe.findKey(r,s);u&&(!n||em(r,r[u],u,n))&&(delete r[u],a=!0)}}return pe.isArray(t)?t.forEach(o):o(t),a}clear(t){const n=Object.keys(this);let r=n.length,a=!1;for(;r--;){const o=n[r];(!t||em(this,this[o],o,t,!0))&&(delete this[o],a=!0)}return a}normalize(t){const n=this,r={};return pe.forEach(this,(a,o)=>{const s=pe.findKey(r,o);if(s){n[s]=Yd(a),delete n[o];return}const u=t?B7(o):String(o).trim();u!==o&&delete n[o],n[u]=Yd(a),r[u]=!0}),this}concat(...t){return this.constructor.concat(this,...t)}toJSON(t){const n=Object.create(null);return pe.forEach(this,(r,a)=>{r!=null&&r!==!1&&(n[a]=t&&pe.isArray(r)?r.join(", "):r)}),n}[Symbol.iterator](){return Object.entries(this.toJSON())[Symbol.iterator]()}toString(){return Object.entries(this.toJSON()).map(([t,n])=>t+": "+n).join(`
`)}get[Symbol.toStringTag](){return"AxiosHeaders"}static from(t){return t instanceof this?t:new this(t)}static concat(t,...n){const r=new this(t);return n.forEach(a=>r.set(a)),r}static accessor(t){const r=(this[bC]=this[bC]={accessors:{}}).accessors,a=this.prototype;function o(s){const u=eu(s);r[u]||(U7(a,s),r[u]=!0)}return pe.isArray(t)?t.forEach(o):o(t),this}};Gn.accessor(["Content-Type","Content-Length","Accept","Accept-Encoding","User-Agent","Authorization"]);pe.reduceDescriptors(Gn.prototype,({value:e},t)=>{let n=t[0].toUpperCase()+t.slice(1);return{get:()=>e,set(r){this[n]=r}}});pe.freezeMethods(Gn);function tm(e,t){const n=this||$u,r=t||n,a=Gn.from(r.headers);let o=r.data;return pe.forEach(e,function(u){o=u.call(n,o,a.normalize(),t?t.status:void 0)}),a.normalize(),o}function _B(e){return!!(e&&e.__CANCEL__)}function Hs(e,t,n){at.call(this,e??"canceled",at.ERR_CANCELED,t,n),this.name="CanceledError"}pe.inherits(Hs,at,{__CANCEL__:!0});function CB(e,t,n){const r=n.config.validateStatus;!n.status||!r||r(n.status)?e(n):t(new at("Request failed with status code "+n.status,[at.ERR_BAD_REQUEST,at.ERR_BAD_RESPONSE][Math.floor(n.status/100)-4],n.config,n.request,n))}function j7(e){const t=/^([-+\w]{1,25})(:?\/\/|:)/.exec(e);return t&&t[1]||""}function G7(e,t){e=e||10;const n=new Array(e),r=new Array(e);let a=0,o=0,s;return t=t!==void 0?t:1e3,function(c){const d=Date.now(),p=r[o];s||(s=d),n[a]=c,r[a]=d;let g=o,m=0;for(;g!==a;)m+=n[g++],g=g%e;if(a=(a+1)%e,a===o&&(o=(o+1)%e),d-s<t)return;const b=p&&d-p;return b?Math.round(m*1e3/b):void 0}}function H7(e,t){let n=0,r=1e3/t,a,o;const s=(d,p=Date.now())=>{n=p,a=null,o&&(clearTimeout(o),o=null),e.apply(null,d)};return[(...d)=>{const p=Date.now(),g=p-n;g>=r?s(d,p):(a=d,o||(o=setTimeout(()=>{o=null,s(a)},r-g)))},()=>a&&s(a)]}const pf=(e,t,n=3)=>{let r=0;const a=G7(50,250);return H7(o=>{const s=o.loaded,u=o.lengthComputable?o.total:void 0,c=s-r,d=a(c),p=s<=u;r=s;const g={loaded:s,total:u,progress:u?s/u:void 0,bytes:c,rate:d||void 0,estimated:d&&u&&p?(u-s)/d:void 0,event:o,lengthComputable:u!=null,[t?"download":"upload"]:!0};e(g)},n)},yC=(e,t)=>{const n=e!=null;return[r=>t[0]({lengthComputable:n,total:e,loaded:r}),t[1]]},vC=e=>(...t)=>pe.asap(()=>e(...t)),$7=xn.hasStandardBrowserEnv?((e,t)=>n=>(n=new URL(n,xn.origin),e.protocol===n.protocol&&e.host===n.host&&(t||e.port===n.port)))(new URL(xn.origin),xn.navigator&&/(msie|trident)/i.test(xn.navigator.userAgent)):()=>!0,q7=xn.hasStandardBrowserEnv?{write(e,t,n,r,a,o){const s=[e+"="+encodeURIComponent(t)];pe.isNumber(n)&&s.push("expires="+new Date(n).toGMTString()),pe.isString(r)&&s.push("path="+r),pe.isString(a)&&s.push("domain="+a),o===!0&&s.push("secure"),document.cookie=s.join("; ")},read(e){const t=document.cookie.match(new RegExp("(^|;\\s*)("+e+")=([^;]*)"));return t?decodeURIComponent(t[3]):null},remove(e){this.write(e,"",Date.now()-864e5)}}:{write(){},read(){return null},remove(){}};function V7(e){return/^([a-z][a-z\d+\-.]*:)?\/\//i.test(e)}function W7(e,t){return t?e.replace(/\/?\/$/,"")+"/"+t.replace(/^\/+/,""):e}function NB(e,t){return e&&!V7(t)?W7(e,t):t}const SC=e=>e instanceof Gn?{...e}:e;function gi(e,t){t=t||{};const n={};function r(d,p,g,m){return pe.isPlainObject(d)&&pe.isPlainObject(p)?pe.merge.call({caseless:m},d,p):pe.isPlainObject(p)?pe.merge({},p):pe.isArray(p)?p.slice():p}function a(d,p,g,m){if(pe.isUndefined(p)){if(!pe.isUndefined(d))return r(void 0,d,g,m)}else return r(d,p,g,m)}function o(d,p){if(!pe.isUndefined(p))return r(void 0,p)}function s(d,p){if(pe.isUndefined(p)){if(!pe.isUndefined(d))return r(void 0,d)}else return r(void 0,p)}function u(d,p,g){if(g in t)return r(d,p);if(g in e)return r(void 0,d)}const c={url:o,method:o,data:o,baseURL:s,transformRequest:s,transformResponse:s,paramsSerializer:s,timeout:s,timeoutMessage:s,withCredentials:s,withXSRFToken:s,adapter:s,responseType:s,xsrfCookieName:s,xsrfHeaderName:s,onUploadProgress:s,onDownloadProgress:s,decompress:s,maxContentLength:s,maxBodyLength:s,beforeRedirect:s,transport:s,httpAgent:s,httpsAgent:s,cancelToken:s,socketPath:s,responseEncoding:s,validateStatus:u,headers:(d,p,g)=>a(SC(d),SC(p),g,!0)};return pe.forEach(Object.keys(Object.assign({},e,t)),function(p){const g=c[p]||a,m=g(e[p],t[p],p);pe.isUndefined(m)&&g!==u||(n[p]=m)}),n}const OB=e=>{const t=gi({},e);let{data:n,withXSRFToken:r,xsrfHeaderName:a,xsrfCookieName:o,headers:s,auth:u}=t;t.headers=s=Gn.from(s),t.url=TB(NB(t.baseURL,t.url),e.params,e.paramsSerializer),u&&s.set("Authorization","Basic "+btoa((u.username||"")+":"+(u.password?unescape(encodeURIComponent(u.password)):"")));let c;if(pe.isFormData(n)){if(xn.hasStandardBrowserEnv||xn.hasStandardBrowserWebWorkerEnv)s.setContentType(void 0);else if((c=s.getContentType())!==!1){const[d,...p]=c?c.split(";").map(g=>g.trim()).filter(Boolean):[];s.setContentType([d||"multipart/form-data",...p].join("; "))}}if(xn.hasStandardBrowserEnv&&(r&&pe.isFunction(r)&&(r=r(t)),r||r!==!1&&$7(t.url))){const d=a&&o&&q7.read(o);d&&s.set(a,d)}return t},Y7=typeof XMLHttpRequest<"u",K7=Y7&&function(e){return new Promise(function(n,r){const a=OB(e);let o=a.data;const s=Gn.from(a.headers).normalize();let{responseType:u,onUploadProgress:c,onDownloadProgress:d}=a,p,g,m,b,y;function S(){b&&b(),y&&y(),a.cancelToken&&a.cancelToken.unsubscribe(p),a.signal&&a.signal.removeEventListener("abort",p)}let x=new XMLHttpRequest;x.open(a.method.toUpperCase(),a.url,!0),x.timeout=a.timeout;function R(){if(!x)return;const A=Gn.from("getAllResponseHeaders"in x&&x.getAllResponseHeaders()),N={data:!u||u==="text"||u==="json"?x.responseText:x.response,status:x.status,statusText:x.statusText,headers:A,config:e,request:x};CB(function(O){n(O),S()},function(O){r(O),S()},N),x=null}"onloadend"in x?x.onloadend=R:x.onreadystatechange=function(){!x||x.readyState!==4||x.status===0&&!(x.responseURL&&x.responseURL.indexOf("file:")===0)||setTimeout(R)},x.onabort=function(){x&&(r(new at("Request aborted",at.ECONNABORTED,e,x)),x=null)},x.onerror=function(){r(new at("Network Error",at.ERR_NETWORK,e,x)),x=null},x.ontimeout=function(){let C=a.timeout?"timeout of "+a.timeout+"ms exceeded":"timeout exceeded";const N=a.transitional||AB;a.timeoutErrorMessage&&(C=a.timeoutErrorMessage),r(new at(C,N.clarifyTimeoutError?at.ETIMEDOUT:at.ECONNABORTED,e,x)),x=null},o===void 0&&s.setContentType(null),"setRequestHeader"in x&&pe.forEach(s.toJSON(),function(C,N){x.setRequestHeader(N,C)}),pe.isUndefined(a.withCredentials)||(x.withCredentials=!!a.withCredentials),u&&u!=="json"&&(x.responseType=a.responseType),d&&([m,y]=pf(d,!0),x.addEventListener("progress",m)),c&&x.upload&&([g,b]=pf(c),x.upload.addEventListener("progress",g),x.upload.addEventListener("loadend",b)),(a.cancelToken||a.signal)&&(p=A=>{x&&(r(!A||A.type?new Hs(null,e,x):A),x.abort(),x=null)},a.cancelToken&&a.cancelToken.subscribe(p),a.signal&&(a.signal.aborted?p():a.signal.addEventListener("abort",p)));const k=j7(a.url);if(k&&xn.protocols.indexOf(k)===-1){r(new at("Unsupported protocol "+k+":",at.ERR_BAD_REQUEST,e));return}x.send(o||null)})},X7=(e,t)=>{const{length:n}=e=e?e.filter(Boolean):[];if(t||n){let r=new AbortController,a;const o=function(d){if(!a){a=!0,u();const p=d instanceof Error?d:this.reason;r.abort(p instanceof at?p:new Hs(p instanceof Error?p.message:p))}};let s=t&&setTimeout(()=>{s=null,o(new at(`timeout ${t} of ms exceeded`,at.ETIMEDOUT))},t);const u=()=>{e&&(s&&clearTimeout(s),s=null,e.forEach(d=>{d.unsubscribe?d.unsubscribe(o):d.removeEventListener("abort",o)}),e=null)};e.forEach(d=>d.addEventListener("abort",o));const{signal:c}=r;return c.unsubscribe=()=>pe.asap(u),c}},Z7=function*(e,t){let n=e.byteLength;if(n<t){yield e;return}let r=0,a;for(;r<n;)a=r+t,yield e.slice(r,a),r=a},Q7=async function*(e,t){for await(const n of J7(e))yield*Z7(n,t)},J7=async function*(e){if(e[Symbol.asyncIterator]){yield*e;return}const t=e.getReader();try{for(;;){const{done:n,value:r}=await t.read();if(n)break;yield r}}finally{await t.cancel()}},EC=(e,t,n,r)=>{const a=Q7(e,t);let o=0,s,u=c=>{s||(s=!0,r&&r(c))};return new ReadableStream({async pull(c){try{const{done:d,value:p}=await a.next();if(d){u(),c.close();return}let g=p.byteLength;if(n){let m=o+=g;n(m)}c.enqueue(new Uint8Array(p))}catch(d){throw u(d),d}},cancel(c){return u(c),a.return()}},{highWaterMark:2})},Jf=typeof fetch=="function"&&typeof Request=="function"&&typeof Response=="function",DB=Jf&&typeof ReadableStream=="function",eV=Jf&&(typeof TextEncoder=="function"?(e=>t=>e.encode(t))(new TextEncoder):async e=>new Uint8Array(await new Response(e).arrayBuffer())),IB=(e,...t)=>{try{return!!e(...t)}catch{return!1}},tV=DB&&IB(()=>{let e=!1;const t=new Request(xn.origin,{body:new ReadableStream,method:"POST",get duplex(){return e=!0,"half"}}).headers.has("Content-Type");return e&&!t}),wC=64*1024,fk=DB&&IB(()=>pe.isReadableStream(new Response("").body)),gf={stream:fk&&(e=>e.body)};Jf&&(e=>{["text","arrayBuffer","blob","formData","stream"].forEach(t=>{!gf[t]&&(gf[t]=pe.isFunction(e[t])?n=>n[t]():(n,r)=>{throw new at(`Response type '${t}' is not supported`,at.ERR_NOT_SUPPORT,r)})})})(new Response);const nV=async e=>{if(e==null)return 0;if(pe.isBlob(e))return e.size;if(pe.isSpecCompliantForm(e))return(await new Request(xn.origin,{method:"POST",body:e}).arrayBuffer()).byteLength;if(pe.isArrayBufferView(e)||pe.isArrayBuffer(e))return e.byteLength;if(pe.isURLSearchParams(e)&&(e=e+""),pe.isString(e))return(await eV(e)).byteLength},rV=async(e,t)=>{const n=pe.toFiniteNumber(e.getContentLength());return n??nV(t)},aV=Jf&&(async e=>{let{url:t,method:n,data:r,signal:a,cancelToken:o,timeout:s,onDownloadProgress:u,onUploadProgress:c,responseType:d,headers:p,withCredentials:g="same-origin",fetchOptions:m}=OB(e);d=d?(d+"").toLowerCase():"text";let b=X7([a,o&&o.toAbortSignal()],s),y;const S=b&&b.unsubscribe&&(()=>{b.unsubscribe()});let x;try{if(c&&tV&&n!=="get"&&n!=="head"&&(x=await rV(p,r))!==0){let N=new Request(t,{method:"POST",body:r,duplex:"half"}),_;if(pe.isFormData(r)&&(_=N.headers.get("content-type"))&&p.setContentType(_),N.body){const[O,F]=yC(x,pf(vC(c)));r=EC(N.body,wC,O,F)}}pe.isString(g)||(g=g?"include":"omit");const R="credentials"in Request.prototype;y=new Request(t,{...m,signal:b,method:n.toUpperCase(),headers:p.normalize().toJSON(),body:r,duplex:"half",credentials:R?g:void 0});let k=await fetch(y);const A=fk&&(d==="stream"||d==="response");if(fk&&(u||A&&S)){const N={};["status","statusText","headers"].forEach(L=>{N[L]=k[L]});const _=pe.toFiniteNumber(k.headers.get("content-length")),[O,F]=u&&yC(_,pf(vC(u),!0))||[];k=new Response(EC(k.body,wC,O,()=>{F&&F(),S&&S()}),N)}d=d||"text";let C=await gf[pe.findKey(gf,d)||"text"](k,e);return!A&&S&&S(),await new Promise((N,_)=>{CB(N,_,{data:C,headers:Gn.from(k.headers),status:k.status,statusText:k.statusText,config:e,request:y})})}catch(R){throw S&&S(),R&&R.name==="TypeError"&&/fetch/i.test(R.message)?Object.assign(new at("Network Error",at.ERR_NETWORK,e,y),{cause:R.cause||R}):at.from(R,R&&R.code,e,y)}}),pk={http:v7,xhr:K7,fetch:aV};pe.forEach(pk,(e,t)=>{if(e){try{Object.defineProperty(e,"name",{value:t})}catch{}Object.defineProperty(e,"adapterName",{value:t})}});const xC=e=>`- ${e}`,oV=e=>pe.isFunction(e)||e===null||e===!1,LB={getAdapter:e=>{e=pe.isArray(e)?e:[e];const{length:t}=e;let n,r;const a={};for(let o=0;o<t;o++){n=e[o];let s;if(r=n,!oV(n)&&(r=pk[(s=String(n)).toLowerCase()],r===void 0))throw new at(`Unknown adapter '${s}'`);if(r)break;a[s||"#"+o]=r}if(!r){const o=Object.entries(a).map(([u,c])=>`adapter ${u} `+(c===!1?"is not supported by the environment":"is not available in the build"));let s=t?o.length>1?`since :
`+o.map(xC).join(`
`):" "+xC(o[0]):"as no adapter specified";throw new at("There is no suitable adapter to dispatch the request "+s,"ERR_NOT_SUPPORT")}return r},adapters:pk};function nm(e){if(e.cancelToken&&e.cancelToken.throwIfRequested(),e.signal&&e.signal.aborted)throw new Hs(null,e)}function kC(e){return nm(e),e.headers=Gn.from(e.headers),e.data=tm.call(e,e.transformRequest),["post","put","patch"].indexOf(e.method)!==-1&&e.headers.setContentType("application/x-www-form-urlencoded",!1),LB.getAdapter(e.adapter||$u.adapter)(e).then(function(r){return nm(e),r.data=tm.call(e,e.transformResponse,r),r.headers=Gn.from(r.headers),r},function(r){return _B(r)||(nm(e),r&&r.response&&(r.response.data=tm.call(e,e.transformResponse,r.response),r.response.headers=Gn.from(r.response.headers))),Promise.reject(r)})}const MB="1.7.9",ep={};["object","boolean","number","function","string","symbol"].forEach((e,t)=>{ep[e]=function(r){return typeof r===e||"a"+(t<1?"n ":" ")+e}});const TC={};ep.transitional=function(t,n,r){function a(o,s){return"[Axios v"+MB+"] Transitional option '"+o+"'"+s+(r?". "+r:"")}return(o,s,u)=>{if(t===!1)throw new at(a(s," has been removed"+(n?" in "+n:"")),at.ERR_DEPRECATED);return n&&!TC[s]&&(TC[s]=!0,console.warn(a(s," has been deprecated since v"+n+" and will be removed in the near future"))),t?t(o,s,u):!0}};ep.spelling=function(t){return(n,r)=>(console.warn(`${r} is likely a misspelling of ${t}`),!0)};function iV(e,t,n){if(typeof e!="object")throw new at("options must be an object",at.ERR_BAD_OPTION_VALUE);const r=Object.keys(e);let a=r.length;for(;a-- >0;){const o=r[a],s=t[o];if(s){const u=e[o],c=u===void 0||s(u,o,e);if(c!==!0)throw new at("option "+o+" must be "+c,at.ERR_BAD_OPTION_VALUE);continue}if(n!==!0)throw new at("Unknown option "+o,at.ERR_BAD_OPTION)}}const Kd={assertOptions:iV,validators:ep},Zr=Kd.validators;let fi=class{constructor(t){this.defaults=t,this.interceptors={request:new mC,response:new mC}}async request(t,n){try{return await this._request(t,n)}catch(r){if(r instanceof Error){let a={};Error.captureStackTrace?Error.captureStackTrace(a):a=new Error;const o=a.stack?a.stack.replace(/^.+\n/,""):"";try{r.stack?o&&!String(r.stack).endsWith(o.replace(/^.+\n.+\n/,""))&&(r.stack+=`
`+o):r.stack=o}catch{}}throw r}}_request(t,n){typeof t=="string"?(n=n||{},n.url=t):n=t||{},n=gi(this.defaults,n);const{transitional:r,paramsSerializer:a,headers:o}=n;r!==void 0&&Kd.assertOptions(r,{silentJSONParsing:Zr.transitional(Zr.boolean),forcedJSONParsing:Zr.transitional(Zr.boolean),clarifyTimeoutError:Zr.transitional(Zr.boolean)},!1),a!=null&&(pe.isFunction(a)?n.paramsSerializer={serialize:a}:Kd.assertOptions(a,{encode:Zr.function,serialize:Zr.function},!0)),Kd.assertOptions(n,{baseUrl:Zr.spelling("baseURL"),withXsrfToken:Zr.spelling("withXSRFToken")},!0),n.method=(n.method||this.defaults.method||"get").toLowerCase();let s=o&&pe.merge(o.common,o[n.method]);o&&pe.forEach(["delete","get","head","post","put","patch","common"],y=>{delete o[y]}),n.headers=Gn.concat(s,o);const u=[];let c=!0;this.interceptors.request.forEach(function(S){typeof S.runWhen=="function"&&S.runWhen(n)===!1||(c=c&&S.synchronous,u.unshift(S.fulfilled,S.rejected))});const d=[];this.interceptors.response.forEach(function(S){d.push(S.fulfilled,S.rejected)});let p,g=0,m;if(!c){const y=[kC.bind(this),void 0];for(y.unshift.apply(y,u),y.push.apply(y,d),m=y.length,p=Promise.resolve(n);g<m;)p=p.then(y[g++],y[g++]);return p}m=u.length;let b=n;for(g=0;g<m;){const y=u[g++],S=u[g++];try{b=y(b)}catch(x){S.call(this,x);break}}try{p=kC.call(this,b)}catch(y){return Promise.reject(y)}for(g=0,m=d.length;g<m;)p=p.then(d[g++],d[g++]);return p}getUri(t){t=gi(this.defaults,t);const n=NB(t.baseURL,t.url);return TB(n,t.params,t.paramsSerializer)}};pe.forEach(["delete","get","head","options"],function(t){fi.prototype[t]=function(n,r){return this.request(gi(r||{},{method:t,url:n,data:(r||{}).data}))}});pe.forEach(["post","put","patch"],function(t){function n(r){return function(o,s,u){return this.request(gi(u||{},{method:t,headers:r?{"Content-Type":"multipart/form-data"}:{},url:o,data:s}))}}fi.prototype[t]=n(),fi.prototype[t+"Form"]=n(!0)});let sV=class FB{constructor(t){if(typeof t!="function")throw new TypeError("executor must be a function.");let n;this.promise=new Promise(function(o){n=o});const r=this;this.promise.then(a=>{if(!r._listeners)return;let o=r._listeners.length;for(;o-- >0;)r._listeners[o](a);r._listeners=null}),this.promise.then=a=>{let o;const s=new Promise(u=>{r.subscribe(u),o=u}).then(a);return s.cancel=function(){r.unsubscribe(o)},s},t(function(o,s,u){r.reason||(r.reason=new Hs(o,s,u),n(r.reason))})}throwIfRequested(){if(this.reason)throw this.reason}subscribe(t){if(this.reason){t(this.reason);return}this._listeners?this._listeners.push(t):this._listeners=[t]}unsubscribe(t){if(!this._listeners)return;const n=this._listeners.indexOf(t);n!==-1&&this._listeners.splice(n,1)}toAbortSignal(){const t=new AbortController,n=r=>{t.abort(r)};return this.subscribe(n),t.signal.unsubscribe=()=>this.unsubscribe(n),t.signal}static source(){let t;return{token:new FB(function(a){t=a}),cancel:t}}};function lV(e){return function(n){return e.apply(null,n)}}function uV(e){return pe.isObject(e)&&e.isAxiosError===!0}const gk={Continue:100,SwitchingProtocols:101,Processing:102,EarlyHints:103,Ok:200,Created:201,Accepted:202,NonAuthoritativeInformation:203,NoContent:204,ResetContent:205,PartialContent:206,MultiStatus:207,AlreadyReported:208,ImUsed:226,MultipleChoices:300,MovedPermanently:301,Found:302,SeeOther:303,NotModified:304,UseProxy:305,Unused:306,TemporaryRedirect:307,PermanentRedirect:308,BadRequest:400,Unauthorized:401,PaymentRequired:402,Forbidden:403,NotFound:404,MethodNotAllowed:405,NotAcceptable:406,ProxyAuthenticationRequired:407,RequestTimeout:408,Conflict:409,Gone:410,LengthRequired:411,PreconditionFailed:412,PayloadTooLarge:413,UriTooLong:414,UnsupportedMediaType:415,RangeNotSatisfiable:416,ExpectationFailed:417,ImATeapot:418,MisdirectedRequest:421,UnprocessableEntity:422,Locked:423,FailedDependency:424,TooEarly:425,UpgradeRequired:426,PreconditionRequired:428,TooManyRequests:429,RequestHeaderFieldsTooLarge:431,UnavailableForLegalReasons:451,InternalServerError:500,NotImplemented:501,BadGateway:502,ServiceUnavailable:503,GatewayTimeout:504,HttpVersionNotSupported:505,VariantAlsoNegotiates:506,InsufficientStorage:507,LoopDetected:508,NotExtended:510,NetworkAuthenticationRequired:511};Object.entries(gk).forEach(([e,t])=>{gk[t]=e});function PB(e){const t=new fi(e),n=pB(fi.prototype.request,t);return pe.extend(n,fi.prototype,t,{allOwnKeys:!0}),pe.extend(n,t,null,{allOwnKeys:!0}),n.create=function(a){return PB(gi(e,a))},n}const Yt=PB($u);Yt.Axios=fi;Yt.CanceledError=Hs;Yt.CancelToken=sV;Yt.isCancel=_B;Yt.VERSION=MB;Yt.toFormData=Qf;Yt.AxiosError=at;Yt.Cancel=Yt.CanceledError;Yt.all=function(t){return Promise.all(t)};Yt.spread=lV;Yt.isAxiosError=uV;Yt.mergeConfig=gi;Yt.AxiosHeaders=Gn;Yt.formToJSON=e=>RB(pe.isHTMLForm(e)?new FormData(e):e);Yt.getAdapter=LB.getAdapter;Yt.HttpStatusCode=gk;Yt.default=Yt;const{Axios:pxe,AxiosError:gxe,CanceledError:hxe,isCancel:mxe,CancelToken:bxe,VERSION:yxe,all:vxe,Cancel:Sxe,isAxiosError:Exe,spread:wxe,toFormData:xxe,AxiosHeaders:kxe,HttpStatusCode:Txe,formToJSON:Axe,getAdapter:Rxe,mergeConfig:_xe}=Yt,zB="",cV="/webui/",Fr="ghost",dV="#B2EBF2",fV="#000",pV="#E2E2E2",BB="#EEEEEE",gV="#F57F17",hV="#969696",mV="#F57F17",AC="#B2EBF2",md=50,RC=100,gu=4,UB=20,bV=15,_C="*",yV={"text/plain":[".txt",".md",".html",".htm",".tex",".json",".xml",".yaml",".yml",".rtf",".odt",".epub",".csv",".log",".conf",".ini",".properties",".sql",".bat",".sh",".c",".cpp",".py",".java",".js",".ts",".swift",".go",".rb",".php",".css",".scss",".less"],"application/pdf":[".pdf"],"application/msword":[".doc"],"application/vnd.openxmlformats-officedocument.wordprocessingml.document":[".docx"],"application/vnd.openxmlformats-officedocument.presentationml.presentation":[".pptx"]},CC={name:"LightRAG",github:"https://github.com/HKUDS/LightRAG"},vV="modulepreload",SV=function(e){return"/webui/"+e},NC={},EV=function(t,n,r){let a=Promise.resolve();if(n&&n.length>0){document.getElementsByTagName("link");const s=document.querySelector("meta[property=csp-nonce]"),u=(s==null?void 0:s.nonce)||(s==null?void 0:s.getAttribute("nonce"));a=Promise.allSettled(n.map(c=>{if(c=SV(c),c in NC)return;NC[c]=!0;const d=c.endsWith(".css"),p=d?'[rel="stylesheet"]':"";if(document.querySelector(`link[href="${c}"]${p}`))return;const g=document.createElement("link");if(g.rel=d?"stylesheet":vV,d||(g.as="script"),g.crossOrigin="",g.href=c,u&&g.setAttribute("nonce",u),document.head.appendChild(g),d)return new Promise((m,b)=>{g.addEventListener("load",m),g.addEventListener("error",()=>b(new Error(`Unable to preload CSS for ${c}`)))})}))}function o(s){const u=new Event("vite:preloadError",{cancelable:!0});if(u.payload=s,window.dispatchEvent(u),!u.defaultPrevented)throw s}return a.then(s=>{for(const u of s||[])u.status==="rejected"&&o(u.reason);return t().catch(o)})};function jB(e,t){let n;try{n=e()}catch{return}return{getItem:a=>{var o;const s=c=>c===null?null:JSON.parse(c,void 0),u=(o=n.getItem(a))!=null?o:null;return u instanceof Promise?u.then(s):s(u)},setItem:(a,o)=>n.setItem(a,JSON.stringify(o,void 0)),removeItem:a=>n.removeItem(a)}}const hk=e=>t=>{try{const n=e(t);return n instanceof Promise?n:{then(r){return hk(r)(n)},catch(r){return this}}}catch(n){return{then(r){return this},catch(r){return hk(r)(n)}}}},wV=(e,t)=>(n,r,a)=>{let o={storage:jB(()=>localStorage),partialize:S=>S,version:0,merge:(S,x)=>({...x,...S}),...t},s=!1;const u=new Set,c=new Set;let d=o.storage;if(!d)return e((...S)=>{console.warn(`[zustand persist middleware] Unable to update item '${o.name}', the given storage is currently unavailable.`),n(...S)},r,a);const p=()=>{const S=o.partialize({...r()});return d.setItem(o.name,{state:S,version:o.version})},g=a.setState;a.setState=(S,x)=>{g(S,x),p()};const m=e((...S)=>{n(...S),p()},r,a);a.getInitialState=()=>m;let b;const y=()=>{var S,x;if(!d)return;s=!1,u.forEach(k=>{var A;return k((A=r())!=null?A:m)});const R=((x=o.onRehydrateStorage)==null?void 0:x.call(o,(S=r())!=null?S:m))||void 0;return hk(d.getItem.bind(d))(o.name).then(k=>{if(k)if(typeof k.version=="number"&&k.version!==o.version){if(o.migrate){const A=o.migrate(k.state,k.version);return A instanceof Promise?A.then(C=>[!0,C]):[!0,A]}console.error("State loaded from storage couldn't be migrated since no migrate function was provided")}else return[!1,k.state];return[!1,void 0]}).then(k=>{var A;const[C,N]=k;if(b=o.merge(N,(A=r())!=null?A:m),n(b,!0),C)return p()}).then(()=>{R==null||R(b,void 0),b=r(),s=!0,c.forEach(k=>k(b))}).catch(k=>{R==null||R(void 0,k)})};return a.persist={setOptions:S=>{o={...o,...S},S.storage&&(d=S.storage)},clearStorage:()=>{d==null||d.removeItem(o.name)},getOptions:()=>o,rehydrate:()=>y(),hasHydrated:()=>s,onHydrate:S=>(u.add(S),()=>{u.delete(S)}),onFinishHydration:S=>(c.add(S),()=>{c.delete(S)})},o.skipHydration||y(),b||m},xV=wV,kV=Yf()(xV(e=>({theme:"system",language:"en",showPropertyPanel:!0,showNodeSearchBar:!0,showNodeLabel:!0,enableNodeDrag:!0,showEdgeLabel:!1,enableHideUnselectedEdges:!0,enableEdgeEvents:!1,graphQueryMaxDepth:3,graphMinDegree:0,graphLayoutMaxIterations:15,queryLabel:_C,enableHealthCheck:!0,apiKey:null,currentTab:"documents",retrievalHistory:[],querySettings:{mode:"global",response_type:"Multiple Paragraphs",top_k:10,max_token_for_text_unit:4e3,max_token_for_global_context:4e3,max_token_for_local_context:4e3,only_need_context:!1,only_need_prompt:!1,stream:!0,history_turns:3,hl_keywords:[],ll_keywords:[]},setTheme:t=>e({theme:t}),setLanguage:t=>{e({language:t}),EV(async()=>{const{default:n}=await Promise.resolve().then(()=>Vwe);return{default:n}},[]).then(({default:n})=>{n.language!==t&&n.changeLanguage(t)})},setGraphLayoutMaxIterations:t=>e({graphLayoutMaxIterations:t}),setQueryLabel:t=>e({queryLabel:t}),setGraphQueryMaxDepth:t=>e({graphQueryMaxDepth:t}),setGraphMinDegree:t=>e({graphMinDegree:t}),setEnableHealthCheck:t=>e({enableHealthCheck:t}),setApiKey:t=>e({apiKey:t}),setCurrentTab:t=>e({currentTab:t}),setRetrievalHistory:t=>e({retrievalHistory:t}),updateQuerySettings:t=>e(n=>({querySettings:{...n.querySettings,...t}}))}),{name:"settings-storage",storage:jB(()=>localStorage),version:8,migrate:(e,t)=>(t<2&&(e.showEdgeLabel=!1),t<3&&(e.queryLabel=_C),t<4&&(e.showPropertyPanel=!0,e.showNodeSearchBar=!0,e.showNodeLabel=!0,e.enableHealthCheck=!0,e.apiKey=null),t<5&&(e.currentTab="documents"),t<6&&(e.querySettings={mode:"global",response_type:"Multiple Paragraphs",top_k:10,max_token_for_text_unit:4e3,max_token_for_global_context:4e3,max_token_for_local_context:4e3,only_need_context:!1,only_need_prompt:!1,stream:!0,history_turns:3,hl_keywords:[],ll_keywords:[]},e.retrievalHistory=[]),t<7&&(e.graphQueryMaxDepth=3,e.graphLayoutMaxIterations=15),t<8&&(e.graphMinDegree=0,e.language="en"),e)})),Fe=A0(kV);class TV{constructor(){Xr(this,"nodes",[]);Xr(this,"edges",[]);Xr(this,"nodeIdMap",{});Xr(this,"edgeIdMap",{});Xr(this,"edgeDynamicIdMap",{});Xr(this,"getNode",t=>{const n=this.nodeIdMap[t];if(n!==void 0)return this.nodes[n]});Xr(this,"getEdge",(t,n=!0)=>{const r=n?this.edgeDynamicIdMap[t]:this.edgeIdMap[t];if(r!==void 0)return this.edges[r]});Xr(this,"buildDynamicMap",()=>{this.edgeDynamicIdMap={};for(let t=0;t<this.edges.length;t++){const n=this.edges[t];this.edgeDynamicIdMap[n.dynamicId]=t}})}}const AV=Yf()(e=>({selectedNode:null,focusedNode:null,selectedEdge:null,focusedEdge:null,moveToSelectedNode:!1,isFetching:!1,graphDataFetchAttempted:!1,labelsFetchAttempted:!1,rawGraph:null,sigmaGraph:null,sigmaInstance:null,allDatabaseLabels:["*"],searchEngine:null,setIsFetching:t=>e({isFetching:t}),setSelectedNode:(t,n)=>e({selectedNode:t,moveToSelectedNode:n}),setFocusedNode:t=>e({focusedNode:t}),setSelectedEdge:t=>e({selectedEdge:t}),setFocusedEdge:t=>e({focusedEdge:t}),clearSelection:()=>e({selectedNode:null,focusedNode:null,selectedEdge:null,focusedEdge:null}),reset:()=>{e({selectedNode:null,focusedNode:null,selectedEdge:null,focusedEdge:null,rawGraph:null,sigmaGraph:null,searchEngine:null,moveToSelectedNode:!1})},setRawGraph:t=>e({rawGraph:t}),setSigmaGraph:t=>{e({sigmaGraph:t})},setAllDatabaseLabels:t=>e({allDatabaseLabels:t}),fetchAllDatabaseLabels:async()=>{try{console.log("Fetching all database labels...");const t=await _V();e({allDatabaseLabels:["*",...t]});return}catch(t){throw console.error("Failed to fetch all database labels:",t),e({allDatabaseLabels:["*"]}),t}},setMoveToSelectedNode:t=>e({moveToSelectedNode:t}),setSigmaInstance:t=>e({sigmaInstance:t}),setSearchEngine:t=>e({searchEngine:t}),resetSearchEngine:()=>e({searchEngine:null}),setGraphDataFetchAttempted:t=>e({graphDataFetchAttempted:t}),setLabelsFetchAttempted:t=>e({labelsFetchAttempted:t}),nodeToExpand:null,nodeToPrune:null,triggerNodeExpand:t=>e({nodeToExpand:t}),triggerNodePrune:t=>e({nodeToPrune:t})})),ze=A0(AV);class RV{constructor(){Xr(this,"navigate",null)}setNavigate(t){this.navigate=t}resetAllApplicationState(){console.log("Resetting all application state...");const t=ze.getState(),n=t.sigmaInstance;t.reset(),t.setGraphDataFetchAttempted(!1),t.setLabelsFetchAttempted(!1),t.setSigmaInstance(null),t.setIsFetching(!1),Hn.getState().clear(),Fe.getState().setRetrievalHistory([]),sessionStorage.clear(),n&&(n.getGraph().clear(),n.kill(),ze.getState().setSigmaInstance(null))}handleDirectLoginAccess(){const t=!document.referrer;return t&&this.resetAllApplicationState(),t}navigateToLogin(){if(!this.navigate){console.error("Navigation function not set");return}this.navigate("/login"),setTimeout(()=>{this.resetAllApplicationState(),na.getState().logout()},0)}navigateToHome(){if(!this.navigate){console.error("Navigation function not set");return}this.navigate("/")}}const xu=new RV,GB="Invalid API Key",HB="API Key required",ar=Yt.create({baseURL:zB,headers:{"Content-Type":"application/json"}});ar.interceptors.request.use(e=>{const t=Fe.getState().apiKey,n=localStorage.getItem("LIGHTRAG-API-TOKEN");return n&&(e.headers.Authorization=`Bearer ${n}`),t&&(e.headers["X-API-Key"]=t),e});ar.interceptors.response.use(e=>e,e=>{var t,n,r,a;if(e.response){if(((t=e.response)==null?void 0:t.status)===401){if((r=(n=e.config)==null?void 0:n.url)!=null&&r.includes("/login"))throw e;return xu.navigateToLogin(),new Promise(()=>{})}throw new Error(`${e.response.status} ${e.response.statusText}
${JSON.stringify(e.response.data)}
${(a=e.config)==null?void 0:a.url}`)}throw e});const $B=async(e,t,n)=>(await ar.get(`/graphs?label=${encodeURIComponent(e)}&max_depth=${t}&min_degree=${n}`)).data,_V=async()=>(await ar.get("/graph/label/list")).data,CV=async()=>{try{return(await ar.get("/health")).data}catch(e){return{status:"error",message:Sr(e)}}},NV=async()=>(await ar.get("/documents")).data,OV=async()=>(await ar.post("/documents/scan")).data,DV=async e=>(await ar.post("/query",e)).data,IV=async(e,t,n)=>{try{let r="";if(await ar.post("/query/stream",e,{responseType:"text",headers:{Accept:"application/x-ndjson"},transformResponse:[a=>{r+=a;const o=r.split(`
`);r=o.pop()||"";for(const s of o)if(s.trim())try{const u=JSON.parse(s);u.response?t(u.response):u.error&&n&&n(u.error)}catch(u){console.error("Error parsing stream chunk:",u),n&&n("Error parsing server response")}return a}]}).catch(a=>{n&&n(Sr(a))}),r.trim())try{const a=JSON.parse(r);a.response?t(a.response):a.error&&n&&n(a.error)}catch(a){console.error("Error parsing final chunk:",a),n&&n("Error parsing server response")}}catch(r){const a=Sr(r);console.error("Stream request failed:",a),n&&n(a)}},LV=async(e,t)=>{const n=new FormData;return n.append("file",e),(await ar.post("/documents/upload",n,{headers:{"Content-Type":"multipart/form-data"},onUploadProgress:t!==void 0?a=>{const o=Math.round(a.loaded*100/a.total);t(o)}:void 0})).data},MV=async()=>(await ar.delete("/documents")).data,N0=async()=>{try{const e=await ar.get("/auth-status",{timeout:5e3,headers:{Accept:"application/json"}});if((e.headers["content-type"]||"").includes("text/html"))return console.warn("Received HTML response instead of JSON for auth-status endpoint"),{auth_configured:!0,auth_mode:"enabled"};if(e.data&&typeof e.data=="object"&&"auth_configured"in e.data&&typeof e.data.auth_configured=="boolean"){if(e.data.auth_configured)return e.data;if(e.data.access_token&&typeof e.data.access_token=="string")return e.data;console.warn("Auth not configured but no valid access token provided")}return console.warn("Received invalid auth status response:",e.data),{auth_configured:!0,auth_mode:"enabled"}}catch(e){return console.error("Failed to get auth status:",Sr(e)),{auth_configured:!0,auth_mode:"enabled"}}},FV=async(e,t)=>{const n=new FormData;return n.append("username",e),n.append("password",t),(await ar.post("/login",n,{headers:{"Content-Type":"multipart/form-data"}})).data},PV=Yf()(e=>({health:!0,message:null,messageTitle:null,lastCheckTime:Date.now(),status:null,check:async()=>{const t=await CV();return t.status==="healthy"?(e({health:!0,message:null,messageTitle:null,lastCheckTime:Date.now(),status:t}),!0):(e({health:!1,message:t.message,messageTitle:"Backend Health Check Error!",lastCheckTime:Date.now(),status:null}),!1)},clear:()=>{e({health:!0,message:null,messageTitle:null})},setErrorMessage:(t,n)=>{e({health:!1,message:t,messageTitle:n})}})),Hn=A0(PV),zV=e=>{try{const t=e.split(".");return t.length!==3?!1:JSON.parse(atob(t[1])).role==="guest"}catch(t){return console.error("Error parsing token:",t),!1}},BV=()=>{const e=localStorage.getItem("LIGHTRAG-API-TOKEN");return e?{isAuthenticated:!0,isGuestMode:zV(e)}:{isAuthenticated:!1,isGuestMode:!1}},na=Yf(e=>{const t=BV();return{isAuthenticated:t.isAuthenticated,isGuestMode:t.isGuestMode,login:(n,r=!1)=>{localStorage.setItem("LIGHTRAG-API-TOKEN",n),e({isAuthenticated:!0,isGuestMode:r})},logout:()=>{localStorage.removeItem("LIGHTRAG-API-TOKEN"),e({isAuthenticated:!1,isGuestMode:!1})}}});var UV=e=>{switch(e){case"success":return HV;case"info":return qV;case"warning":return $V;case"error":return VV;default:return null}},jV=Array(12).fill(0),GV=({visible:e,className:t})=>ye.createElement("div",{className:["sonner-loading-wrapper",t].filter(Boolean).join(" "),"data-visible":e},ye.createElement("div",{className:"sonner-spinner"},jV.map((n,r)=>ye.createElement("div",{className:"sonner-loading-bar",key:`spinner-bar-${r}`})))),HV=ye.createElement("svg",{xmlns:"http://www.w3.org/2000/svg",viewBox:"0 0 20 20",fill:"currentColor",height:"20",width:"20"},ye.createElement("path",{fillRule:"evenodd",d:"M10 18a8 8 0 100-16 8 8 0 000 16zm3.857-9.809a.75.75 0 00-1.214-.882l-3.483 4.79-1.88-1.88a.75.75 0 10-1.06 1.061l2.5 2.5a.75.75 0 001.137-.089l4-5.5z",clipRule:"evenodd"})),$V=ye.createElement("svg",{xmlns:"http://www.w3.org/2000/svg",viewBox:"0 0 24 24",fill:"currentColor",height:"20",width:"20"},ye.createElement("path",{fillRule:"evenodd",d:"M9.401 3.003c1.155-2 4.043-2 5.197 0l7.355 12.748c1.154 2-.29 4.5-2.599 4.5H4.645c-2.309 0-3.752-2.5-2.598-4.5L9.4 3.003zM12 8.25a.75.75 0 01.75.75v3.75a.75.75 0 01-1.5 0V9a.75.75 0 01.75-.75zm0 8.25a.75.75 0 100-1.5.75.75 0 000 1.5z",clipRule:"evenodd"})),qV=ye.createElement("svg",{xmlns:"http://www.w3.org/2000/svg",viewBox:"0 0 20 20",fill:"currentColor",height:"20",width:"20"},ye.createElement("path",{fillRule:"evenodd",d:"M18 10a8 8 0 11-16 0 8 8 0 0116 0zm-7-4a1 1 0 11-2 0 1 1 0 012 0zM9 9a.75.75 0 000 1.5h.253a.25.25 0 01.244.304l-.459 2.066A1.75 1.75 0 0010.747 15H11a.75.75 0 000-1.5h-.253a.25.25 0 01-.244-.304l.459-2.066A1.75 1.75 0 009.253 9H9z",clipRule:"evenodd"})),VV=ye.createElement("svg",{xmlns:"http://www.w3.org/2000/svg",viewBox:"0 0 20 20",fill:"currentColor",height:"20",width:"20"},ye.createElement("path",{fillRule:"evenodd",d:"M18 10a8 8 0 11-16 0 8 8 0 0116 0zm-8-5a.75.75 0 01.75.75v4.5a.75.75 0 01-1.5 0v-4.5A.75.75 0 0110 5zm0 10a1 1 0 100-2 1 1 0 000 2z",clipRule:"evenodd"})),WV=ye.createElement("svg",{xmlns:"http://www.w3.org/2000/svg",width:"12",height:"12",viewBox:"0 0 24 24",fill:"none",stroke:"currentColor",strokeWidth:"1.5",strokeLinecap:"round",strokeLinejoin:"round"},ye.createElement("line",{x1:"18",y1:"6",x2:"6",y2:"18"}),ye.createElement("line",{x1:"6",y1:"6",x2:"18",y2:"18"})),YV=()=>{let[e,t]=ye.useState(document.hidden);return ye.useEffect(()=>{let n=()=>{t(document.hidden)};return document.addEventListener("visibilitychange",n),()=>window.removeEventListener("visibilitychange",n)},[]),e},mk=1,KV=class{constructor(){this.subscribe=e=>(this.subscribers.push(e),()=>{let t=this.subscribers.indexOf(e);this.subscribers.splice(t,1)}),this.publish=e=>{this.subscribers.forEach(t=>t(e))},this.addToast=e=>{this.publish(e),this.toasts=[...this.toasts,e]},this.create=e=>{var t;let{message:n,...r}=e,a=typeof(e==null?void 0:e.id)=="number"||((t=e.id)==null?void 0:t.length)>0?e.id:mk++,o=this.toasts.find(u=>u.id===a),s=e.dismissible===void 0?!0:e.dismissible;return this.dismissedToasts.has(a)&&this.dismissedToasts.delete(a),o?this.toasts=this.toasts.map(u=>u.id===a?(this.publish({...u,...e,id:a,title:n}),{...u,...e,id:a,dismissible:s,title:n}):u):this.addToast({title:n,...r,dismissible:s,id:a}),a},this.dismiss=e=>(this.dismissedToasts.add(e),e||this.toasts.forEach(t=>{this.subscribers.forEach(n=>n({id:t.id,dismiss:!0}))}),this.subscribers.forEach(t=>t({id:e,dismiss:!0})),e),this.message=(e,t)=>this.create({...t,message:e}),this.error=(e,t)=>this.create({...t,message:e,type:"error"}),this.success=(e,t)=>this.create({...t,type:"success",message:e}),this.info=(e,t)=>this.create({...t,type:"info",message:e}),this.warning=(e,t)=>this.create({...t,type:"warning",message:e}),this.loading=(e,t)=>this.create({...t,type:"loading",message:e}),this.promise=(e,t)=>{if(!t)return;let n;t.loading!==void 0&&(n=this.create({...t,promise:e,type:"loading",message:t.loading,description:typeof t.description!="function"?t.description:void 0}));let r=e instanceof Promise?e:e(),a=n!==void 0,o,s=r.then(async c=>{if(o=["resolve",c],ye.isValidElement(c))a=!1,this.create({id:n,type:"default",message:c});else if(ZV(c)&&!c.ok){a=!1;let d=typeof t.error=="function"?await t.error(`HTTP error! status: ${c.status}`):t.error,p=typeof t.description=="function"?await t.description(`HTTP error! status: ${c.status}`):t.description;this.create({id:n,type:"error",message:d,description:p})}else if(t.success!==void 0){a=!1;let d=typeof t.success=="function"?await t.success(c):t.success,p=typeof t.description=="function"?await t.description(c):t.description;this.create({id:n,type:"success",message:d,description:p})}}).catch(async c=>{if(o=["reject",c],t.error!==void 0){a=!1;let d=typeof t.error=="function"?await t.error(c):t.error,p=typeof t.description=="function"?await t.description(c):t.description;this.create({id:n,type:"error",message:d,description:p})}}).finally(()=>{var c;a&&(this.dismiss(n),n=void 0),(c=t.finally)==null||c.call(t)}),u=()=>new Promise((c,d)=>s.then(()=>o[0]==="reject"?d(o[1]):c(o[1])).catch(d));return typeof n!="string"&&typeof n!="number"?{unwrap:u}:Object.assign(n,{unwrap:u})},this.custom=(e,t)=>{let n=(t==null?void 0:t.id)||mk++;return this.create({jsx:e(n),id:n,...t}),n},this.getActiveToasts=()=>this.toasts.filter(e=>!this.dismissedToasts.has(e.id)),this.subscribers=[],this.toasts=[],this.dismissedToasts=new Set}},Un=new KV,XV=(e,t)=>{let n=(t==null?void 0:t.id)||mk++;return Un.addToast({title:e,...t,id:n}),n},ZV=e=>e&&typeof e=="object"&&"ok"in e&&typeof e.ok=="boolean"&&"status"in e&&typeof e.status=="number",QV=XV,JV=()=>Un.toasts,eW=()=>Un.getActiveToasts(),Ft=Object.assign(QV,{success:Un.success,info:Un.info,warning:Un.warning,error:Un.error,custom:Un.custom,message:Un.message,promise:Un.promise,dismiss:Un.dismiss,loading:Un.loading},{getHistory:JV,getToasts:eW});function tW(e,{insertAt:t}={}){if(typeof document>"u")return;let n=document.head||document.getElementsByTagName("head")[0],r=document.createElement("style");r.type="text/css",t==="top"&&n.firstChild?n.insertBefore(r,n.firstChild):n.appendChild(r),r.styleSheet?r.styleSheet.cssText=e:r.appendChild(document.createTextNode(e))}tW(`:where(html[dir="ltr"]),:where([data-sonner-toaster][dir="ltr"]){--toast-icon-margin-start: -3px;--toast-icon-margin-end: 4px;--toast-svg-margin-start: -1px;--toast-svg-margin-end: 0px;--toast-button-margin-start: auto;--toast-button-margin-end: 0;--toast-close-button-start: 0;--toast-close-button-end: unset;--toast-close-button-transform: translate(-35%, -35%)}:where(html[dir="rtl"]),:where([data-sonner-toaster][dir="rtl"]){--toast-icon-margin-start: 4px;--toast-icon-margin-end: -3px;--toast-svg-margin-start: 0px;--toast-svg-margin-end: -1px;--toast-button-margin-start: 0;--toast-button-margin-end: auto;--toast-close-button-start: unset;--toast-close-button-end: 0;--toast-close-button-transform: translate(35%, -35%)}:where([data-sonner-toaster]){position:fixed;width:var(--width);font-family:ui-sans-serif,system-ui,-apple-system,BlinkMacSystemFont,Segoe UI,Roboto,Helvetica Neue,Arial,Noto Sans,sans-serif,Apple Color Emoji,Segoe UI Emoji,Segoe UI Symbol,Noto Color Emoji;--gray1: hsl(0, 0%, 99%);--gray2: hsl(0, 0%, 97.3%);--gray3: hsl(0, 0%, 95.1%);--gray4: hsl(0, 0%, 93%);--gray5: hsl(0, 0%, 90.9%);--gray6: hsl(0, 0%, 88.7%);--gray7: hsl(0, 0%, 85.8%);--gray8: hsl(0, 0%, 78%);--gray9: hsl(0, 0%, 56.1%);--gray10: hsl(0, 0%, 52.3%);--gray11: hsl(0, 0%, 43.5%);--gray12: hsl(0, 0%, 9%);--border-radius: 8px;box-sizing:border-box;padding:0;margin:0;list-style:none;outline:none;z-index:999999999;transition:transform .4s ease}:where([data-sonner-toaster][data-lifted="true"]){transform:translateY(-10px)}@media (hover: none) and (pointer: coarse){:where([data-sonner-toaster][data-lifted="true"]){transform:none}}:where([data-sonner-toaster][data-x-position="right"]){right:var(--offset-right)}:where([data-sonner-toaster][data-x-position="left"]){left:var(--offset-left)}:where([data-sonner-toaster][data-x-position="center"]){left:50%;transform:translate(-50%)}:where([data-sonner-toaster][data-y-position="top"]){top:var(--offset-top)}:where([data-sonner-toaster][data-y-position="bottom"]){bottom:var(--offset-bottom)}:where([data-sonner-toast]){--y: translateY(100%);--lift-amount: calc(var(--lift) * var(--gap));z-index:var(--z-index);position:absolute;opacity:0;transform:var(--y);filter:blur(0);touch-action:none;transition:transform .4s,opacity .4s,height .4s,box-shadow .2s;box-sizing:border-box;outline:none;overflow-wrap:anywhere}:where([data-sonner-toast][data-styled="true"]){padding:16px;background:var(--normal-bg);border:1px solid var(--normal-border);color:var(--normal-text);border-radius:var(--border-radius);box-shadow:0 4px 12px #0000001a;width:var(--width);font-size:13px;display:flex;align-items:center;gap:6px}:where([data-sonner-toast]:focus-visible){box-shadow:0 4px 12px #0000001a,0 0 0 2px #0003}:where([data-sonner-toast][data-y-position="top"]){top:0;--y: translateY(-100%);--lift: 1;--lift-amount: calc(1 * var(--gap))}:where([data-sonner-toast][data-y-position="bottom"]){bottom:0;--y: translateY(100%);--lift: -1;--lift-amount: calc(var(--lift) * var(--gap))}:where([data-sonner-toast]) :where([data-description]){font-weight:400;line-height:1.4;color:inherit}:where([data-sonner-toast]) :where([data-title]){font-weight:500;line-height:1.5;color:inherit}:where([data-sonner-toast]) :where([data-icon]){display:flex;height:16px;width:16px;position:relative;justify-content:flex-start;align-items:center;flex-shrink:0;margin-left:var(--toast-icon-margin-start);margin-right:var(--toast-icon-margin-end)}:where([data-sonner-toast][data-promise="true"]) :where([data-icon])>svg{opacity:0;transform:scale(.8);transform-origin:center;animation:sonner-fade-in .3s ease forwards}:where([data-sonner-toast]) :where([data-icon])>*{flex-shrink:0}:where([data-sonner-toast]) :where([data-icon]) svg{margin-left:var(--toast-svg-margin-start);margin-right:var(--toast-svg-margin-end)}:where([data-sonner-toast]) :where([data-content]){display:flex;flex-direction:column;gap:2px}[data-sonner-toast][data-styled=true] [data-button]{border-radius:4px;padding-left:8px;padding-right:8px;height:24px;font-size:12px;color:var(--normal-bg);background:var(--normal-text);margin-left:var(--toast-button-margin-start);margin-right:var(--toast-button-margin-end);border:none;cursor:pointer;outline:none;display:flex;align-items:center;flex-shrink:0;transition:opacity .4s,box-shadow .2s}:where([data-sonner-toast]) :where([data-button]):focus-visible{box-shadow:0 0 0 2px #0006}:where([data-sonner-toast]) :where([data-button]):first-of-type{margin-left:var(--toast-button-margin-start);margin-right:var(--toast-button-margin-end)}:where([data-sonner-toast]) :where([data-cancel]){color:var(--normal-text);background:rgba(0,0,0,.08)}:where([data-sonner-toast][data-theme="dark"]) :where([data-cancel]){background:rgba(255,255,255,.3)}:where([data-sonner-toast]) :where([data-close-button]){position:absolute;left:var(--toast-close-button-start);right:var(--toast-close-button-end);top:0;height:20px;width:20px;display:flex;justify-content:center;align-items:center;padding:0;color:var(--gray12);border:1px solid var(--gray4);transform:var(--toast-close-button-transform);border-radius:50%;cursor:pointer;z-index:1;transition:opacity .1s,background .2s,border-color .2s}[data-sonner-toast] [data-close-button]{background:var(--gray1)}:where([data-sonner-toast]) :where([data-close-button]):focus-visible{box-shadow:0 4px 12px #0000001a,0 0 0 2px #0003}:where([data-sonner-toast]) :where([data-disabled="true"]){cursor:not-allowed}:where([data-sonner-toast]):hover :where([data-close-button]):hover{background:var(--gray2);border-color:var(--gray5)}:where([data-sonner-toast][data-swiping="true"]):before{content:"";position:absolute;left:-50%;right:-50%;height:100%;z-index:-1}:where([data-sonner-toast][data-y-position="top"][data-swiping="true"]):before{bottom:50%;transform:scaleY(3) translateY(50%)}:where([data-sonner-toast][data-y-position="bottom"][data-swiping="true"]):before{top:50%;transform:scaleY(3) translateY(-50%)}:where([data-sonner-toast][data-swiping="false"][data-removed="true"]):before{content:"";position:absolute;inset:0;transform:scaleY(2)}:where([data-sonner-toast]):after{content:"";position:absolute;left:0;height:calc(var(--gap) + 1px);bottom:100%;width:100%}:where([data-sonner-toast][data-mounted="true"]){--y: translateY(0);opacity:1}:where([data-sonner-toast][data-expanded="false"][data-front="false"]){--scale: var(--toasts-before) * .05 + 1;--y: translateY(calc(var(--lift-amount) * var(--toasts-before))) scale(calc(-1 * var(--scale)));height:var(--front-toast-height)}:where([data-sonner-toast])>*{transition:opacity .4s}:where([data-sonner-toast][data-expanded="false"][data-front="false"][data-styled="true"])>*{opacity:0}:where([data-sonner-toast][data-visible="false"]){opacity:0;pointer-events:none}:where([data-sonner-toast][data-mounted="true"][data-expanded="true"]){--y: translateY(calc(var(--lift) * var(--offset)));height:var(--initial-height)}:where([data-sonner-toast][data-removed="true"][data-front="true"][data-swipe-out="false"]){--y: translateY(calc(var(--lift) * -100%));opacity:0}:where([data-sonner-toast][data-removed="true"][data-front="false"][data-swipe-out="false"][data-expanded="true"]){--y: translateY(calc(var(--lift) * var(--offset) + var(--lift) * -100%));opacity:0}:where([data-sonner-toast][data-removed="true"][data-front="false"][data-swipe-out="false"][data-expanded="false"]){--y: translateY(40%);opacity:0;transition:transform .5s,opacity .2s}:where([data-sonner-toast][data-removed="true"][data-front="false"]):before{height:calc(var(--initial-height) + 20%)}[data-sonner-toast][data-swiping=true]{transform:var(--y) translateY(var(--swipe-amount-y, 0px)) translate(var(--swipe-amount-x, 0px));transition:none}[data-sonner-toast][data-swiped=true]{user-select:none}[data-sonner-toast][data-swipe-out=true][data-y-position=bottom],[data-sonner-toast][data-swipe-out=true][data-y-position=top]{animation-duration:.2s;animation-timing-function:ease-out;animation-fill-mode:forwards}[data-sonner-toast][data-swipe-out=true][data-swipe-direction=left]{animation-name:swipe-out-left}[data-sonner-toast][data-swipe-out=true][data-swipe-direction=right]{animation-name:swipe-out-right}[data-sonner-toast][data-swipe-out=true][data-swipe-direction=up]{animation-name:swipe-out-up}[data-sonner-toast][data-swipe-out=true][data-swipe-direction=down]{animation-name:swipe-out-down}@keyframes swipe-out-left{0%{transform:var(--y) translate(var(--swipe-amount-x));opacity:1}to{transform:var(--y) translate(calc(var(--swipe-amount-x) - 100%));opacity:0}}@keyframes swipe-out-right{0%{transform:var(--y) translate(var(--swipe-amount-x));opacity:1}to{transform:var(--y) translate(calc(var(--swipe-amount-x) + 100%));opacity:0}}@keyframes swipe-out-up{0%{transform:var(--y) translateY(var(--swipe-amount-y));opacity:1}to{transform:var(--y) translateY(calc(var(--swipe-amount-y) - 100%));opacity:0}}@keyframes swipe-out-down{0%{transform:var(--y) translateY(var(--swipe-amount-y));opacity:1}to{transform:var(--y) translateY(calc(var(--swipe-amount-y) + 100%));opacity:0}}@media (max-width: 600px){[data-sonner-toaster]{position:fixed;right:var(--mobile-offset-right);left:var(--mobile-offset-left);width:100%}[data-sonner-toaster][dir=rtl]{left:calc(var(--mobile-offset-left) * -1)}[data-sonner-toaster] [data-sonner-toast]{left:0;right:0;width:calc(100% - var(--mobile-offset-left) * 2)}[data-sonner-toaster][data-x-position=left]{left:var(--mobile-offset-left)}[data-sonner-toaster][data-y-position=bottom]{bottom:var(--mobile-offset-bottom)}[data-sonner-toaster][data-y-position=top]{top:var(--mobile-offset-top)}[data-sonner-toaster][data-x-position=center]{left:var(--mobile-offset-left);right:var(--mobile-offset-right);transform:none}}[data-sonner-toaster][data-theme=light]{--normal-bg: #fff;--normal-border: var(--gray4);--normal-text: var(--gray12);--success-bg: hsl(143, 85%, 96%);--success-border: hsl(145, 92%, 91%);--success-text: hsl(140, 100%, 27%);--info-bg: hsl(208, 100%, 97%);--info-border: hsl(221, 91%, 91%);--info-text: hsl(210, 92%, 45%);--warning-bg: hsl(49, 100%, 97%);--warning-border: hsl(49, 91%, 91%);--warning-text: hsl(31, 92%, 45%);--error-bg: hsl(359, 100%, 97%);--error-border: hsl(359, 100%, 94%);--error-text: hsl(360, 100%, 45%)}[data-sonner-toaster][data-theme=light] [data-sonner-toast][data-invert=true]{--normal-bg: #000;--normal-border: hsl(0, 0%, 20%);--normal-text: var(--gray1)}[data-sonner-toaster][data-theme=dark] [data-sonner-toast][data-invert=true]{--normal-bg: #fff;--normal-border: var(--gray3);--normal-text: var(--gray12)}[data-sonner-toaster][data-theme=dark]{--normal-bg: #000;--normal-bg-hover: hsl(0, 0%, 12%);--normal-border: hsl(0, 0%, 20%);--normal-border-hover: hsl(0, 0%, 25%);--normal-text: var(--gray1);--success-bg: hsl(150, 100%, 6%);--success-border: hsl(147, 100%, 12%);--success-text: hsl(150, 86%, 65%);--info-bg: hsl(215, 100%, 6%);--info-border: hsl(223, 100%, 12%);--info-text: hsl(216, 87%, 65%);--warning-bg: hsl(64, 100%, 6%);--warning-border: hsl(60, 100%, 12%);--warning-text: hsl(46, 87%, 65%);--error-bg: hsl(358, 76%, 10%);--error-border: hsl(357, 89%, 16%);--error-text: hsl(358, 100%, 81%)}[data-sonner-toaster][data-theme=dark] [data-sonner-toast] [data-close-button]{background:var(--normal-bg);border-color:var(--normal-border);color:var(--normal-text)}[data-sonner-toaster][data-theme=dark] [data-sonner-toast] [data-close-button]:hover{background:var(--normal-bg-hover);border-color:var(--normal-border-hover)}[data-rich-colors=true][data-sonner-toast][data-type=success],[data-rich-colors=true][data-sonner-toast][data-type=success] [data-close-button]{background:var(--success-bg);border-color:var(--success-border);color:var(--success-text)}[data-rich-colors=true][data-sonner-toast][data-type=info],[data-rich-colors=true][data-sonner-toast][data-type=info] [data-close-button]{background:var(--info-bg);border-color:var(--info-border);color:var(--info-text)}[data-rich-colors=true][data-sonner-toast][data-type=warning],[data-rich-colors=true][data-sonner-toast][data-type=warning] [data-close-button]{background:var(--warning-bg);border-color:var(--warning-border);color:var(--warning-text)}[data-rich-colors=true][data-sonner-toast][data-type=error],[data-rich-colors=true][data-sonner-toast][data-type=error] [data-close-button]{background:var(--error-bg);border-color:var(--error-border);color:var(--error-text)}.sonner-loading-wrapper{--size: 16px;height:var(--size);width:var(--size);position:absolute;inset:0;z-index:10}.sonner-loading-wrapper[data-visible=false]{transform-origin:center;animation:sonner-fade-out .2s ease forwards}.sonner-spinner{position:relative;top:50%;left:50%;height:var(--size);width:var(--size)}.sonner-loading-bar{animation:sonner-spin 1.2s linear infinite;background:var(--gray11);border-radius:6px;height:8%;left:-10%;position:absolute;top:-3.9%;width:24%}.sonner-loading-bar:nth-child(1){animation-delay:-1.2s;transform:rotate(.0001deg) translate(146%)}.sonner-loading-bar:nth-child(2){animation-delay:-1.1s;transform:rotate(30deg) translate(146%)}.sonner-loading-bar:nth-child(3){animation-delay:-1s;transform:rotate(60deg) translate(146%)}.sonner-loading-bar:nth-child(4){animation-delay:-.9s;transform:rotate(90deg) translate(146%)}.sonner-loading-bar:nth-child(5){animation-delay:-.8s;transform:rotate(120deg) translate(146%)}.sonner-loading-bar:nth-child(6){animation-delay:-.7s;transform:rotate(150deg) translate(146%)}.sonner-loading-bar:nth-child(7){animation-delay:-.6s;transform:rotate(180deg) translate(146%)}.sonner-loading-bar:nth-child(8){animation-delay:-.5s;transform:rotate(210deg) translate(146%)}.sonner-loading-bar:nth-child(9){animation-delay:-.4s;transform:rotate(240deg) translate(146%)}.sonner-loading-bar:nth-child(10){animation-delay:-.3s;transform:rotate(270deg) translate(146%)}.sonner-loading-bar:nth-child(11){animation-delay:-.2s;transform:rotate(300deg) translate(146%)}.sonner-loading-bar:nth-child(12){animation-delay:-.1s;transform:rotate(330deg) translate(146%)}@keyframes sonner-fade-in{0%{opacity:0;transform:scale(.8)}to{opacity:1;transform:scale(1)}}@keyframes sonner-fade-out{0%{opacity:1;transform:scale(1)}to{opacity:0;transform:scale(.8)}}@keyframes sonner-spin{0%{opacity:1}to{opacity:.15}}@media (prefers-reduced-motion){[data-sonner-toast],[data-sonner-toast]>*,.sonner-loading-bar{transition:none!important;animation:none!important}}.sonner-loader{position:absolute;top:50%;left:50%;transform:translate(-50%,-50%);transform-origin:center;transition:opacity .2s,transform .2s}.sonner-loader[data-visible=false]{opacity:0;transform:scale(.8) translate(-50%,-50%)}
`);function bd(e){return e.label!==void 0}var nW=3,rW="32px",aW="16px",OC=4e3,oW=356,iW=14,sW=20,lW=200;function Dr(...e){return e.filter(Boolean).join(" ")}function uW(e){let[t,n]=e.split("-"),r=[];return t&&r.push(t),n&&r.push(n),r}var cW=e=>{var t,n,r,a,o,s,u,c,d,p,g;let{invert:m,toast:b,unstyled:y,interacting:S,setHeights:x,visibleToasts:R,heights:k,index:A,toasts:C,expanded:N,removeToast:_,defaultRichColors:O,closeButton:F,style:L,cancelButtonStyle:I,actionButtonStyle:H,className:$="",descriptionClassName:U="",duration:Y,position:Z,gap:j,loadingIcon:G,expandByDefault:z,classNames:K,icons:D,closeButtonAriaLabel:V="Close toast",pauseWhenPageIsHidden:B}=e,[M,W]=ye.useState(null),[Q,oe]=ye.useState(null),[re,ie]=ye.useState(!1),[Se,ae]=ye.useState(!1),[ve,xe]=ye.useState(!1),[Ie,Ce]=ye.useState(!1),[ke,J]=ye.useState(!1),[fe,Te]=ye.useState(0),[me,Ee]=ye.useState(0),le=ye.useRef(b.duration||Y||OC),Be=ye.useRef(null),Ue=ye.useRef(null),he=A===0,Ne=A+1<=R,ee=b.type,ce=b.dismissible!==!1,_e=b.className||"",Pe=b.descriptionClassName||"",We=ye.useMemo(()=>k.findIndex(Xe=>Xe.toastId===b.id)||0,[k,b.id]),yt=ye.useMemo(()=>{var Xe;return(Xe=b.closeButton)!=null?Xe:F},[b.closeButton,F]),kt=ye.useMemo(()=>b.duration||Y||OC,[b.duration,Y]),ht=ye.useRef(0),et=ye.useRef(0),Tt=ye.useRef(0),ot=ye.useRef(null),[St,Ht]=Z.split("-"),fn=ye.useMemo(()=>k.reduce((Xe,mt,Ct)=>Ct>=We?Xe:Xe+mt.height,0),[k,We]),zt=YV(),or=b.invert||m,$r=ee==="loading";et.current=ye.useMemo(()=>We*j+fn,[We,fn]),ye.useEffect(()=>{le.current=kt},[kt]),ye.useEffect(()=>{ie(!0)},[]),ye.useEffect(()=>{let Xe=Ue.current;if(Xe){let mt=Xe.getBoundingClientRect().height;return Ee(mt),x(Ct=>[{toastId:b.id,height:mt,position:b.position},...Ct]),()=>x(Ct=>Ct.filter(Dn=>Dn.toastId!==b.id))}},[x,b.id]),ye.useLayoutEffect(()=>{if(!re)return;let Xe=Ue.current,mt=Xe.style.height;Xe.style.height="auto";let Ct=Xe.getBoundingClientRect().height;Xe.style.height=mt,Ee(Ct),x(Dn=>Dn.find(_n=>_n.toastId===b.id)?Dn.map(_n=>_n.toastId===b.id?{..._n,height:Ct}:_n):[{toastId:b.id,height:Ct,position:b.position},...Dn])},[re,b.title,b.description,x,b.id]);let Jt=ye.useCallback(()=>{ae(!0),Te(et.current),x(Xe=>Xe.filter(mt=>mt.toastId!==b.id)),setTimeout(()=>{_(b)},lW)},[b,_,x,et]);ye.useEffect(()=>{if(b.promise&&ee==="loading"||b.duration===1/0||b.type==="loading")return;let Xe;return N||S||B&&zt?(()=>{if(Tt.current<ht.current){let mt=new Date().getTime()-ht.current;le.current=le.current-mt}Tt.current=new Date().getTime()})():le.current!==1/0&&(ht.current=new Date().getTime(),Xe=setTimeout(()=>{var mt;(mt=b.onAutoClose)==null||mt.call(b,b),Jt()},le.current)),()=>clearTimeout(Xe)},[N,S,b,ee,B,zt,Jt]),ye.useEffect(()=>{b.delete&&Jt()},[Jt,b.delete]);function fa(){var Xe,mt,Ct;return D!=null&&D.loading?ye.createElement("div",{className:Dr(K==null?void 0:K.loader,(Xe=b==null?void 0:b.classNames)==null?void 0:Xe.loader,"sonner-loader"),"data-visible":ee==="loading"},D.loading):G?ye.createElement("div",{className:Dr(K==null?void 0:K.loader,(mt=b==null?void 0:b.classNames)==null?void 0:mt.loader,"sonner-loader"),"data-visible":ee==="loading"},G):ye.createElement(GV,{className:Dr(K==null?void 0:K.loader,(Ct=b==null?void 0:b.classNames)==null?void 0:Ct.loader),visible:ee==="loading"})}return ye.createElement("li",{tabIndex:0,ref:Ue,className:Dr($,_e,K==null?void 0:K.toast,(t=b==null?void 0:b.classNames)==null?void 0:t.toast,K==null?void 0:K.default,K==null?void 0:K[ee],(n=b==null?void 0:b.classNames)==null?void 0:n[ee]),"data-sonner-toast":"","data-rich-colors":(r=b.richColors)!=null?r:O,"data-styled":!(b.jsx||b.unstyled||y),"data-mounted":re,"data-promise":!!b.promise,"data-swiped":ke,"data-removed":Se,"data-visible":Ne,"data-y-position":St,"data-x-position":Ht,"data-index":A,"data-front":he,"data-swiping":ve,"data-dismissible":ce,"data-type":ee,"data-invert":or,"data-swipe-out":Ie,"data-swipe-direction":Q,"data-expanded":!!(N||z&&re),style:{"--index":A,"--toasts-before":A,"--z-index":C.length-A,"--offset":`${Se?fe:et.current}px`,"--initial-height":z?"auto":`${me}px`,...L,...b.style},onDragEnd:()=>{xe(!1),W(null),ot.current=null},onPointerDown:Xe=>{$r||!ce||(Be.current=new Date,Te(et.current),Xe.target.setPointerCapture(Xe.pointerId),Xe.target.tagName!=="BUTTON"&&(xe(!0),ot.current={x:Xe.clientX,y:Xe.clientY}))},onPointerUp:()=>{var Xe,mt,Ct,Dn;if(Ie||!ce)return;ot.current=null;let _n=Number(((Xe=Ue.current)==null?void 0:Xe.style.getPropertyValue("--swipe-amount-x").replace("px",""))||0),In=Number(((mt=Ue.current)==null?void 0:mt.style.getPropertyValue("--swipe-amount-y").replace("px",""))||0),pa=new Date().getTime()-((Ct=Be.current)==null?void 0:Ct.getTime()),Ln=M==="x"?_n:In,Rr=Math.abs(Ln)/pa;if(Math.abs(Ln)>=sW||Rr>.11){Te(et.current),(Dn=b.onDismiss)==null||Dn.call(b,b),oe(M==="x"?_n>0?"right":"left":In>0?"down":"up"),Jt(),Ce(!0),J(!1);return}xe(!1),W(null)},onPointerMove:Xe=>{var mt,Ct,Dn,_n;if(!ot.current||!ce||((mt=window.getSelection())==null?void 0:mt.toString().length)>0)return;let In=Xe.clientY-ot.current.y,pa=Xe.clientX-ot.current.x,Ln=(Ct=e.swipeDirections)!=null?Ct:uW(Z);!M&&(Math.abs(pa)>1||Math.abs(In)>1)&&W(Math.abs(pa)>Math.abs(In)?"x":"y");let Rr={x:0,y:0};M==="y"?(Ln.includes("top")||Ln.includes("bottom"))&&(Ln.includes("top")&&In<0||Ln.includes("bottom")&&In>0)&&(Rr.y=In):M==="x"&&(Ln.includes("left")||Ln.includes("right"))&&(Ln.includes("left")&&pa<0||Ln.includes("right")&&pa>0)&&(Rr.x=pa),(Math.abs(Rr.x)>0||Math.abs(Rr.y)>0)&&J(!0),(Dn=Ue.current)==null||Dn.style.setProperty("--swipe-amount-x",`${Rr.x}px`),(_n=Ue.current)==null||_n.style.setProperty("--swipe-amount-y",`${Rr.y}px`)}},yt&&!b.jsx?ye.createElement("button",{"aria-label":V,"data-disabled":$r,"data-close-button":!0,onClick:$r||!ce?()=>{}:()=>{var Xe;Jt(),(Xe=b.onDismiss)==null||Xe.call(b,b)},className:Dr(K==null?void 0:K.closeButton,(a=b==null?void 0:b.classNames)==null?void 0:a.closeButton)},(o=D==null?void 0:D.close)!=null?o:WV):null,b.jsx||E.isValidElement(b.title)?b.jsx?b.jsx:typeof b.title=="function"?b.title():b.title:ye.createElement(ye.Fragment,null,ee||b.icon||b.promise?ye.createElement("div",{"data-icon":"",className:Dr(K==null?void 0:K.icon,(s=b==null?void 0:b.classNames)==null?void 0:s.icon)},b.promise||b.type==="loading"&&!b.icon?b.icon||fa():null,b.type!=="loading"?b.icon||(D==null?void 0:D[ee])||UV(ee):null):null,ye.createElement("div",{"data-content":"",className:Dr(K==null?void 0:K.content,(u=b==null?void 0:b.classNames)==null?void 0:u.content)},ye.createElement("div",{"data-title":"",className:Dr(K==null?void 0:K.title,(c=b==null?void 0:b.classNames)==null?void 0:c.title)},typeof b.title=="function"?b.title():b.title),b.description?ye.createElement("div",{"data-description":"",className:Dr(U,Pe,K==null?void 0:K.description,(d=b==null?void 0:b.classNames)==null?void 0:d.description)},typeof b.description=="function"?b.description():b.description):null),E.isValidElement(b.cancel)?b.cancel:b.cancel&&bd(b.cancel)?ye.createElement("button",{"data-button":!0,"data-cancel":!0,style:b.cancelButtonStyle||I,onClick:Xe=>{var mt,Ct;bd(b.cancel)&&ce&&((Ct=(mt=b.cancel).onClick)==null||Ct.call(mt,Xe),Jt())},className:Dr(K==null?void 0:K.cancelButton,(p=b==null?void 0:b.classNames)==null?void 0:p.cancelButton)},b.cancel.label):null,E.isValidElement(b.action)?b.action:b.action&&bd(b.action)?ye.createElement("button",{"data-button":!0,"data-action":!0,style:b.actionButtonStyle||H,onClick:Xe=>{var mt,Ct;bd(b.action)&&((Ct=(mt=b.action).onClick)==null||Ct.call(mt,Xe),!Xe.defaultPrevented&&Jt())},className:Dr(K==null?void 0:K.actionButton,(g=b==null?void 0:b.classNames)==null?void 0:g.actionButton)},b.action.label):null))};function DC(){if(typeof window>"u"||typeof document>"u")return"ltr";let e=document.documentElement.getAttribute("dir");return e==="auto"||!e?window.getComputedStyle(document.documentElement).direction:e}function dW(e,t){let n={};return[e,t].forEach((r,a)=>{let o=a===1,s=o?"--mobile-offset":"--offset",u=o?aW:rW;function c(d){["top","right","bottom","left"].forEach(p=>{n[`${s}-${p}`]=typeof d=="number"?`${d}px`:d})}typeof r=="number"||typeof r=="string"?c(r):typeof r=="object"?["top","right","bottom","left"].forEach(d=>{r[d]===void 0?n[`${s}-${d}`]=u:n[`${s}-${d}`]=typeof r[d]=="number"?`${r[d]}px`:r[d]}):c(u)}),n}var fW=E.forwardRef(function(e,t){let{invert:n,position:r="bottom-right",hotkey:a=["altKey","KeyT"],expand:o,closeButton:s,className:u,offset:c,mobileOffset:d,theme:p="light",richColors:g,duration:m,style:b,visibleToasts:y=nW,toastOptions:S,dir:x=DC(),gap:R=iW,loadingIcon:k,icons:A,containerAriaLabel:C="Notifications",pauseWhenPageIsHidden:N}=e,[_,O]=ye.useState([]),F=ye.useMemo(()=>Array.from(new Set([r].concat(_.filter(B=>B.position).map(B=>B.position)))),[_,r]),[L,I]=ye.useState([]),[H,$]=ye.useState(!1),[U,Y]=ye.useState(!1),[Z,j]=ye.useState(p!=="system"?p:typeof window<"u"&&window.matchMedia&&window.matchMedia("(prefers-color-scheme: dark)").matches?"dark":"light"),G=ye.useRef(null),z=a.join("+").replace(/Key/g,"").replace(/Digit/g,""),K=ye.useRef(null),D=ye.useRef(!1),V=ye.useCallback(B=>{O(M=>{var W;return(W=M.find(Q=>Q.id===B.id))!=null&&W.delete||Un.dismiss(B.id),M.filter(({id:Q})=>Q!==B.id)})},[]);return ye.useEffect(()=>Un.subscribe(B=>{if(B.dismiss){O(M=>M.map(W=>W.id===B.id?{...W,delete:!0}:W));return}setTimeout(()=>{tB.flushSync(()=>{O(M=>{let W=M.findIndex(Q=>Q.id===B.id);return W!==-1?[...M.slice(0,W),{...M[W],...B},...M.slice(W+1)]:[B,...M]})})})}),[]),ye.useEffect(()=>{if(p!=="system"){j(p);return}if(p==="system"&&(window.matchMedia&&window.matchMedia("(prefers-color-scheme: dark)").matches?j("dark"):j("light")),typeof window>"u")return;let B=window.matchMedia("(prefers-color-scheme: dark)");try{B.addEventListener("change",({matches:M})=>{j(M?"dark":"light")})}catch{B.addListener(({matches:W})=>{try{j(W?"dark":"light")}catch(Q){console.error(Q)}})}},[p]),ye.useEffect(()=>{_.length<=1&&$(!1)},[_]),ye.useEffect(()=>{let B=M=>{var W,Q;a.every(oe=>M[oe]||M.code===oe)&&($(!0),(W=G.current)==null||W.focus()),M.code==="Escape"&&(document.activeElement===G.current||(Q=G.current)!=null&&Q.contains(document.activeElement))&&$(!1)};return document.addEventListener("keydown",B),()=>document.removeEventListener("keydown",B)},[a]),ye.useEffect(()=>{if(G.current)return()=>{K.current&&(K.current.focus({preventScroll:!0}),K.current=null,D.current=!1)}},[G.current]),ye.createElement("section",{ref:t,"aria-label":`${C} ${z}`,tabIndex:-1,"aria-live":"polite","aria-relevant":"additions text","aria-atomic":"false",suppressHydrationWarning:!0},F.map((B,M)=>{var W;let[Q,oe]=B.split("-");return _.length?ye.createElement("ol",{key:B,dir:x==="auto"?DC():x,tabIndex:-1,ref:G,className:u,"data-sonner-toaster":!0,"data-theme":Z,"data-y-position":Q,"data-lifted":H&&_.length>1&&!o,"data-x-position":oe,style:{"--front-toast-height":`${((W=L[0])==null?void 0:W.height)||0}px`,"--width":`${oW}px`,"--gap":`${R}px`,...b,...dW(c,d)},onBlur:re=>{D.current&&!re.currentTarget.contains(re.relatedTarget)&&(D.current=!1,K.current&&(K.current.focus({preventScroll:!0}),K.current=null))},onFocus:re=>{re.target instanceof HTMLElement&&re.target.dataset.dismissible==="false"||D.current||(D.current=!0,K.current=re.relatedTarget)},onMouseEnter:()=>$(!0),onMouseMove:()=>$(!0),onMouseLeave:()=>{U||$(!1)},onDragEnd:()=>$(!1),onPointerDown:re=>{re.target instanceof HTMLElement&&re.target.dataset.dismissible==="false"||Y(!0)},onPointerUp:()=>Y(!1)},_.filter(re=>!re.position&&M===0||re.position===B).map((re,ie)=>{var Se,ae;return ye.createElement(cW,{key:re.id,icons:A,index:ie,toast:re,defaultRichColors:g,duration:(Se=S==null?void 0:S.duration)!=null?Se:m,className:S==null?void 0:S.className,descriptionClassName:S==null?void 0:S.descriptionClassName,invert:n,visibleToasts:y,closeButton:(ae=S==null?void 0:S.closeButton)!=null?ae:s,interacting:U,position:B,style:S==null?void 0:S.style,unstyled:S==null?void 0:S.unstyled,classNames:S==null?void 0:S.classNames,cancelButtonStyle:S==null?void 0:S.cancelButtonStyle,actionButtonStyle:S==null?void 0:S.actionButtonStyle,removeToast:V,toasts:_.filter(ve=>ve.position==re.position),heights:L.filter(ve=>ve.position==re.position),setHeights:I,expandByDefault:o,gap:R,loadingIcon:k,expanded:H,pauseWhenPageIsHidden:N,swipeDirections:e.swipeDirections})})):null}))});const pW={theme:"system",setTheme:()=>null},qB=E.createContext(pW);function VB({children:e,...t}){const n=Fe.use.theme(),r=Fe.use.setTheme();E.useEffect(()=>{const o=window.document.documentElement;if(o.classList.remove("light","dark"),n==="system"){const s=window.matchMedia("(prefers-color-scheme: dark)"),u=c=>{o.classList.remove("light","dark"),o.classList.add(c.matches?"dark":"light")};return o.classList.add(s.matches?"dark":"light"),s.addEventListener("change",u),()=>s.removeEventListener("change",u)}else o.classList.add(n)},[n]);const a={theme:n,setTheme:r};return w.jsx(qB.Provider,{...t,value:a,children:e})}const gW={visibleTabs:{},setTabVisibility:()=>{},isTabVisible:()=>!1},WB=E.createContext(gW),hW=({children:e})=>{const t=Fe.use.currentTab(),[n,r]=E.useState(()=>({documents:!0,"knowledge-graph":!0,retrieval:!0,api:!0}));E.useEffect(()=>{r(o=>({...o,documents:!0,"knowledge-graph":!0,retrieval:!0,api:!0}))},[t]);const a=E.useMemo(()=>({visibleTabs:n,setTabVisibility:(o,s)=>{r(u=>({...u,[o]:s}))},isTabVisible:o=>!!n[o]}),[n]);return w.jsx(WB.Provider,{value:a,children:e})},IC=e=>typeof e=="boolean"?`${e}`:e===0?"0":e,LC=rB,YB=(e,t)=>n=>{var r;if((t==null?void 0:t.variants)==null)return LC(e,n==null?void 0:n.class,n==null?void 0:n.className);const{variants:a,defaultVariants:o}=t,s=Object.keys(a).map(d=>{const p=n==null?void 0:n[d],g=o==null?void 0:o[d];if(p===null)return null;const m=IC(p)||IC(g);return a[d][m]}),u=n&&Object.entries(n).reduce((d,p)=>{let[g,m]=p;return m===void 0||(d[g]=m),d},{}),c=t==null||(r=t.compoundVariants)===null||r===void 0?void 0:r.reduce((d,p)=>{let{class:g,className:m,...b}=p;return Object.entries(b).every(y=>{let[S,x]=y;return Array.isArray(x)?x.includes({...o,...u}[S]):{...o,...u}[S]===x})?[...d,g,m]:d},[]);return LC(e,s,c,n==null?void 0:n.class,n==null?void 0:n.className)},mW=YB("relative w-full rounded-lg border px-4 py-3 text-sm [&>svg+div]:translate-y-[-3px] [&>svg]:absolute [&>svg]:left-4 [&>svg]:top-4 [&>svg]:text-foreground [&>svg~*]:pl-7",{variants:{variant:{default:"bg-background text-foreground",destructive:"border-destructive/50 text-destructive dark:border-destructive [&>svg]:text-destructive"}},defaultVariants:{variant:"default"}}),KB=E.forwardRef(({className:e,variant:t,...n},r)=>w.jsx("div",{ref:r,role:"alert",className:Me(mW({variant:t}),e),...n}));KB.displayName="Alert";const XB=E.forwardRef(({className:e,...t},n)=>w.jsx("h5",{ref:n,className:Me("mb-1 leading-none font-medium tracking-tight",e),...t}));XB.displayName="AlertTitle";const ZB=E.forwardRef(({className:e,...t},n)=>w.jsx("div",{ref:n,className:Me("text-sm [&_p]:leading-relaxed",e),...t}));ZB.displayName="AlertDescription";/**
 * @license lucide-react v0.475.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */const bW=e=>e.replace(/([a-z0-9])([A-Z])/g,"$1-$2").toLowerCase(),QB=(...e)=>e.filter((t,n,r)=>!!t&&t.trim()!==""&&r.indexOf(t)===n).join(" ").trim();/**
 * @license lucide-react v0.475.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */var yW={xmlns:"http://www.w3.org/2000/svg",width:24,height:24,viewBox:"0 0 24 24",fill:"none",stroke:"currentColor",strokeWidth:2,strokeLinecap:"round",strokeLinejoin:"round"};/**
 * @license lucide-react v0.475.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */const vW=E.forwardRef(({color:e="currentColor",size:t=24,strokeWidth:n=2,absoluteStrokeWidth:r,className:a="",children:o,iconNode:s,...u},c)=>E.createElement("svg",{ref:c,...yW,width:t,height:t,stroke:e,strokeWidth:r?Number(n)*24/Number(t):n,className:QB("lucide",a),...u},[...s.map(([d,p])=>E.createElement(d,p)),...Array.isArray(o)?o:[o]]));/**
 * @license lucide-react v0.475.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */const vt=(e,t)=>{const n=E.forwardRef(({className:r,...a},o)=>E.createElement(vW,{ref:o,iconNode:t,className:QB(`lucide-${bW(e)}`,r),...a}));return n.displayName=`${e}`,n};/**
 * @license lucide-react v0.475.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */const SW=[["path",{d:"M20 6 9 17l-5-5",key:"1gmf2c"}]],O0=vt("Check",SW);/**
 * @license lucide-react v0.475.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */const EW=[["path",{d:"m6 9 6 6 6-6",key:"qrunsl"}]],D0=vt("ChevronDown",EW);/**
 * @license lucide-react v0.475.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */const wW=[["path",{d:"m18 15-6-6-6 6",key:"153udz"}]],JB=vt("ChevronUp",wW);/**
 * @license lucide-react v0.475.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */const xW=[["path",{d:"m7 15 5 5 5-5",key:"1hf1tw"}],["path",{d:"m7 9 5-5 5 5",key:"sgt6xg"}]],kW=vt("ChevronsUpDown",xW);/**
 * @license lucide-react v0.475.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */const TW=[["circle",{cx:"12",cy:"12",r:"10",key:"1mglay"}],["line",{x1:"12",x2:"12",y1:"8",y2:"12",key:"1pkeuh"}],["line",{x1:"12",x2:"12.01",y1:"16",y2:"16",key:"4dfq90"}]],AW=vt("CircleAlert",TW);/**
 * @license lucide-react v0.475.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */const RW=[["rect",{width:"14",height:"14",x:"8",y:"8",rx:"2",ry:"2",key:"17jyea"}],["path",{d:"M4 16c-1.1 0-2-.9-2-2V4c0-1.1.9-2 2-2h10c1.1 0 2 .9 2 2",key:"zix9uf"}]],_W=vt("Copy",RW);/**
 * @license lucide-react v0.475.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */const CW=[["path",{d:"m7 21-4.3-4.3c-1-1-1-2.5 0-3.4l9.6-9.6c1-1 2.5-1 3.4 0l5.6 5.6c1 1 1 2.5 0 3.4L13 21",key:"182aya"}],["path",{d:"M22 21H7",key:"t4ddhn"}],["path",{d:"m5 11 9 9",key:"1mo9qw"}]],eU=vt("Eraser",CW);/**
 * @license lucide-react v0.475.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */const NW=[["path",{d:"M15 2H6a2 2 0 0 0-2 2v16a2 2 0 0 0 2 2h12a2 2 0 0 0 2-2V7Z",key:"1rqfz7"}],["path",{d:"M14 2v4a2 2 0 0 0 2 2h4",key:"tnqrlb"}],["path",{d:"M10 9H8",key:"b1mrlr"}],["path",{d:"M16 13H8",key:"t4e002"}],["path",{d:"M16 17H8",key:"z1uh3a"}]],OW=vt("FileText",NW);/**
 * @license lucide-react v0.475.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */const DW=[["path",{d:"M20 7h-3a2 2 0 0 1-2-2V2",key:"x099mo"}],["path",{d:"M9 18a2 2 0 0 1-2-2V4a2 2 0 0 1 2-2h7l4 4v10a2 2 0 0 1-2 2Z",key:"18t6ie"}],["path",{d:"M3 7.6v12.8A1.6 1.6 0 0 0 4.6 22h9.8",key:"1nja0z"}]],IW=vt("Files",DW);/**
 * @license lucide-react v0.475.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */const LW=[["path",{d:"M3 7V5a2 2 0 0 1 2-2h2",key:"aa7l1z"}],["path",{d:"M17 3h2a2 2 0 0 1 2 2v2",key:"4qcy5o"}],["path",{d:"M21 17v2a2 2 0 0 1-2 2h-2",key:"6vwrx8"}],["path",{d:"M7 21H5a2 2 0 0 1-2-2v-2",key:"ioqczr"}],["rect",{width:"10",height:"8",x:"7",y:"8",rx:"1",key:"vys8me"}]],MW=vt("Fullscreen",LW);/**
 * @license lucide-react v0.475.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */const FW=[["path",{d:"M6 3v12",key:"qpgusn"}],["path",{d:"M18 9a3 3 0 1 0 0-6 3 3 0 0 0 0 6z",key:"1d02ji"}],["path",{d:"M6 21a3 3 0 1 0 0-6 3 3 0 0 0 0 6z",key:"chk6ph"}],["path",{d:"M15 6a9 9 0 0 0-9 9",key:"or332x"}],["path",{d:"M18 15v6",key:"9wciyi"}],["path",{d:"M21 18h-6",key:"139f0c"}]],PW=vt("GitBranchPlus",FW);/**
 * @license lucide-react v0.475.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */const zW=[["path",{d:"M15 22v-4a4.8 4.8 0 0 0-1-3.5c3 0 6-2 6-5.5.08-1.25-.27-2.48-1-3.5.28-1.15.28-2.35 0-3.5 0 0-1 0-3 1.5-2.64-.5-5.36-.5-8 0C6 2 5 2 5 2c-.3 1.15-.3 2.35 0 3.5A5.403 5.403 0 0 0 4 9c0 3.5 3 5.5 6 5.5-.39.49-.68 1.05-.85 1.65-.17.6-.22 1.23-.15 1.85v4",key:"tonef"}],["path",{d:"M9 18c-4.51 2-5-2-7-2",key:"9comsn"}]],BW=vt("Github",zW);/**
 * @license lucide-react v0.475.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */const UW=[["circle",{cx:"12",cy:"5",r:"1",key:"gxeob9"}],["circle",{cx:"19",cy:"5",r:"1",key:"w8mnmm"}],["circle",{cx:"5",cy:"5",r:"1",key:"lttvr7"}],["circle",{cx:"12",cy:"12",r:"1",key:"41hilf"}],["circle",{cx:"19",cy:"12",r:"1",key:"1wjl8i"}],["circle",{cx:"5",cy:"12",r:"1",key:"1pcz8c"}],["circle",{cx:"12",cy:"19",r:"1",key:"lyex9k"}],["circle",{cx:"19",cy:"19",r:"1",key:"shf9b7"}],["circle",{cx:"5",cy:"19",r:"1",key:"bfqh0e"}]],jW=vt("Grip",UW);/**
 * @license lucide-react v0.475.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */const GW=[["path",{d:"M21 12a9 9 0 1 1-6.219-8.56",key:"13zald"}]],tU=vt("LoaderCircle",GW);/**
 * @license lucide-react v0.475.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */const HW=[["path",{d:"M12 2v4",key:"3427ic"}],["path",{d:"m16.2 7.8 2.9-2.9",key:"r700ao"}],["path",{d:"M18 12h4",key:"wj9ykh"}],["path",{d:"m16.2 16.2 2.9 2.9",key:"1bxg5t"}],["path",{d:"M12 18v4",key:"jadmvz"}],["path",{d:"m4.9 19.1 2.9-2.9",key:"bwix9q"}],["path",{d:"M2 12h4",key:"j09sii"}],["path",{d:"m4.9 4.9 2.9 2.9",key:"giyufr"}]],$W=vt("Loader",HW);/**
 * @license lucide-react v0.475.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */const qW=[["path",{d:"M9 21H5a2 2 0 0 1-2-2V5a2 2 0 0 1 2-2h4",key:"1uf3rs"}],["polyline",{points:"16 17 21 12 16 7",key:"1gabdz"}],["line",{x1:"21",x2:"9",y1:"12",y2:"12",key:"1uyos4"}]],VW=vt("LogOut",qW);/**
 * @license lucide-react v0.475.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */const WW=[["path",{d:"M8 3H5a2 2 0 0 0-2 2v3",key:"1dcmit"}],["path",{d:"M21 8V5a2 2 0 0 0-2-2h-3",key:"1e4gt3"}],["path",{d:"M3 16v3a2 2 0 0 0 2 2h3",key:"wsl5sc"}],["path",{d:"M16 21h3a2 2 0 0 0 2-2v-3",key:"18trek"}]],YW=vt("Maximize",WW);/**
 * @license lucide-react v0.475.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */const KW=[["path",{d:"M8 3v3a2 2 0 0 1-2 2H3",key:"hohbtr"}],["path",{d:"M21 8h-3a2 2 0 0 1-2-2V3",key:"5jw1f3"}],["path",{d:"M3 16h3a2 2 0 0 1 2 2v3",key:"198tvr"}],["path",{d:"M16 21v-3a2 2 0 0 1 2-2h3",key:"ph8mxp"}]],XW=vt("Minimize",KW);/**
 * @license lucide-react v0.475.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */const ZW=[["circle",{cx:"13.5",cy:"6.5",r:".5",fill:"currentColor",key:"1okk4w"}],["circle",{cx:"17.5",cy:"10.5",r:".5",fill:"currentColor",key:"f64h9f"}],["circle",{cx:"8.5",cy:"7.5",r:".5",fill:"currentColor",key:"fotxhn"}],["circle",{cx:"6.5",cy:"12.5",r:".5",fill:"currentColor",key:"qy21gx"}],["path",{d:"M12 2C6.5 2 2 6.5 2 12s4.5 10 10 10c.926 0 1.648-.746 1.648-1.688 0-.437-.18-.835-.437-1.125-.29-.289-.438-.652-.438-1.125a1.64 1.64 0 0 1 1.668-1.668h1.996c3.051 0 5.555-2.503 5.555-5.554C21.965 6.012 17.461 2 12 2z",key:"12rzf8"}]],QW=vt("Palette",ZW);/**
 * @license lucide-react v0.475.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */const JW=[["rect",{x:"14",y:"4",width:"4",height:"16",rx:"1",key:"zuxfzm"}],["rect",{x:"6",y:"4",width:"4",height:"16",rx:"1",key:"1okwgv"}]],eY=vt("Pause",JW);/**
 * @license lucide-react v0.475.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */const tY=[["polygon",{points:"6 3 20 12 6 21 6 3",key:"1oa8hb"}]],nY=vt("Play",tY);/**
 * @license lucide-react v0.475.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */const rY=[["path",{d:"M3 12a9 9 0 0 1 9-9 9.75 9.75 0 0 1 6.74 2.74L21 8",key:"v9h5vc"}],["path",{d:"M21 3v5h-5",key:"1q7to0"}],["path",{d:"M21 12a9 9 0 0 1-9 9 9.75 9.75 0 0 1-6.74-2.74L3 16",key:"3uifl3"}],["path",{d:"M8 16H3v5",key:"1cv678"}]],nU=vt("RefreshCw",rY);/**
 * @license lucide-react v0.475.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */const aY=[["path",{d:"M3 12a9 9 0 1 0 9-9 9.75 9.75 0 0 0-6.74 2.74L3 8",key:"1357e3"}],["path",{d:"M3 3v5h5",key:"1xhq8a"}]],oY=vt("RotateCcw",aY);/**
 * @license lucide-react v0.475.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */const iY=[["path",{d:"M21 12a9 9 0 1 1-9-9c2.52 0 4.93 1 6.74 2.74L21 8",key:"1p45f6"}],["path",{d:"M21 3v5h-5",key:"1q7to0"}]],sY=vt("RotateCw",iY);/**
 * @license lucide-react v0.475.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */const lY=[["circle",{cx:"6",cy:"6",r:"3",key:"1lh9wr"}],["path",{d:"M8.12 8.12 12 12",key:"1alkpv"}],["path",{d:"M20 4 8.12 15.88",key:"xgtan2"}],["circle",{cx:"6",cy:"18",r:"3",key:"fqmcym"}],["path",{d:"M14.8 14.8 20 20",key:"ptml3r"}]],uY=vt("Scissors",lY);/**
 * @license lucide-react v0.475.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */const cY=[["circle",{cx:"11",cy:"11",r:"8",key:"4ej97u"}],["path",{d:"m21 21-4.3-4.3",key:"1qie3q"}]],dY=vt("Search",cY);/**
 * @license lucide-react v0.475.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */const fY=[["path",{d:"M14.536 21.686a.5.5 0 0 0 .937-.024l6.5-19a.496.496 0 0 0-.635-.635l-19 6.5a.5.5 0 0 0-.024.937l7.93 3.18a2 2 0 0 1 1.112 1.11z",key:"1ffxy3"}],["path",{d:"m21.854 2.147-10.94 10.939",key:"12cjpa"}]],pY=vt("Send",fY);/**
 * @license lucide-react v0.475.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */const gY=[["path",{d:"M12.22 2h-.44a2 2 0 0 0-2 2v.18a2 2 0 0 1-1 1.73l-.43.25a2 2 0 0 1-2 0l-.15-.08a2 2 0 0 0-2.73.73l-.22.38a2 2 0 0 0 .73 2.73l.15.1a2 2 0 0 1 1 1.72v.51a2 2 0 0 1-1 1.74l-.15.09a2 2 0 0 0-.73 2.73l.22.38a2 2 0 0 0 2.73.73l.15-.08a2 2 0 0 1 2 0l.43.25a2 2 0 0 1 1 1.73V20a2 2 0 0 0 2 2h.44a2 2 0 0 0 2-2v-.18a2 2 0 0 1 1-1.73l.43-.25a2 2 0 0 1 2 0l.15.08a2 2 0 0 0 2.73-.73l.22-.39a2 2 0 0 0-.73-2.73l-.15-.08a2 2 0 0 1-1-1.74v-.5a2 2 0 0 1 1-1.74l.15-.09a2 2 0 0 0 .73-2.73l-.22-.38a2 2 0 0 0-2.73-.73l-.15.08a2 2 0 0 1-2 0l-.43-.25a2 2 0 0 1-1-1.73V4a2 2 0 0 0-2-2z",key:"1qme2f"}],["circle",{cx:"12",cy:"12",r:"3",key:"1v7zrd"}]],hY=vt("Settings",gY);/**
 * @license lucide-react v0.475.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */const mY=[["path",{d:"M21 15v4a2 2 0 0 1-2 2H5a2 2 0 0 1-2-2v-4",key:"ih7n3h"}],["polyline",{points:"17 8 12 3 7 8",key:"t8dd8p"}],["line",{x1:"12",x2:"12",y1:"3",y2:"15",key:"widbto"}]],bk=vt("Upload",mY);/**
 * @license lucide-react v0.475.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */const bY=[["path",{d:"M18 6 6 18",key:"1bl5f8"}],["path",{d:"m6 6 12 12",key:"d8bk6v"}]],rU=vt("X",bY);/**
 * @license lucide-react v0.475.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */const yY=[["path",{d:"M4 14a1 1 0 0 1-.78-1.63l9.9-10.2a.5.5 0 0 1 .86.46l-1.92 6.02A1 1 0 0 0 13 10h7a1 1 0 0 1 .78 1.63l-9.9 10.2a.5.5 0 0 1-.86-.46l1.92-6.02A1 1 0 0 0 11 14z",key:"1xq2db"}]],aU=vt("Zap",yY);/**
 * @license lucide-react v0.475.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */const vY=[["circle",{cx:"11",cy:"11",r:"8",key:"4ej97u"}],["line",{x1:"21",x2:"16.65",y1:"21",y2:"16.65",key:"13gj7c"}],["line",{x1:"11",x2:"11",y1:"8",y2:"14",key:"1vmskp"}],["line",{x1:"8",x2:"14",y1:"11",y2:"11",key:"durymu"}]],SY=vt("ZoomIn",vY);/**
 * @license lucide-react v0.475.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */const EY=[["circle",{cx:"11",cy:"11",r:"8",key:"4ej97u"}],["line",{x1:"21",x2:"16.65",y1:"21",y2:"16.65",key:"13gj7c"}],["line",{x1:"8",x2:"14",y1:"11",y2:"11",key:"durymu"}]],wY=vt("ZoomOut",EY),xY=()=>{const e=Hn.use.health(),t=Hn.use.message(),n=Hn.use.messageTitle(),[r,a]=E.useState(!1);return E.useEffect(()=>{setTimeout(()=>{a(!0)},50)},[]),w.jsxs(KB,{className:Me("bg-background/90 absolute top-12 left-1/2 flex w-auto max-w-lg -translate-x-1/2 transform items-center gap-4 shadow-md backdrop-blur-lg transition-all duration-500 ease-in-out",r?"translate-y-0 opacity-100":"-translate-y-20 opacity-0",!e&&"bg-red-700 text-white"),children:[!e&&w.jsx("div",{children:w.jsx(AW,{className:"size-4"})}),w.jsxs("div",{children:[w.jsx(XB,{className:"font-bold",children:n}),w.jsx(ZB,{children:t})]})]})};function kY(e,t){const n=E.createContext(t),r=o=>{const{children:s,...u}=o,c=E.useMemo(()=>u,Object.values(u));return w.jsx(n.Provider,{value:c,children:s})};r.displayName=e+"Provider";function a(o){const s=E.useContext(n);if(s)return s;if(t!==void 0)return t;throw new Error(`\`${o}\` must be used within \`${e}\``)}return[r,a]}function wr(e,t=[]){let n=[];function r(o,s){const u=E.createContext(s),c=n.length;n=[...n,s];const d=g=>{var R;const{scope:m,children:b,...y}=g,S=((R=m==null?void 0:m[e])==null?void 0:R[c])||u,x=E.useMemo(()=>y,Object.values(y));return w.jsx(S.Provider,{value:x,children:b})};d.displayName=o+"Provider";function p(g,m){var S;const b=((S=m==null?void 0:m[e])==null?void 0:S[c])||u,y=E.useContext(b);if(y)return y;if(s!==void 0)return s;throw new Error(`\`${g}\` must be used within \`${o}\``)}return[d,p]}const a=()=>{const o=n.map(s=>E.createContext(s));return function(u){const c=(u==null?void 0:u[e])||o;return E.useMemo(()=>({[`__scope${e}`]:{...u,[e]:c}}),[u,c])}};return a.scopeName=e,[r,TY(a,...t)]}function TY(...e){const t=e[0];if(e.length===1)return t;const n=()=>{const r=e.map(a=>({useScope:a(),scopeName:a.scopeName}));return function(o){const s=r.reduce((u,{useScope:c,scopeName:d})=>{const g=c(o)[`__scope${d}`];return{...u,...g}},{});return E.useMemo(()=>({[`__scope${t.scopeName}`]:s}),[s])}};return n.scopeName=t.scopeName,n}function MC(e,t){if(typeof e=="function")return e(t);e!=null&&(e.current=t)}function oU(...e){return t=>{let n=!1;const r=e.map(a=>{const o=MC(a,t);return!n&&typeof o=="function"&&(n=!0),o});if(n)return()=>{for(let a=0;a<r.length;a++){const o=r[a];typeof o=="function"?o():MC(e[a],null)}}}}function gt(...e){return E.useCallback(oU(...e),e)}function Ke(e,t,{checkForDefaultPrevented:n=!0}={}){return function(a){if(e==null||e(a),n===!1||!a.defaultPrevented)return t==null?void 0:t(a)}}var An=globalThis!=null&&globalThis.document?E.useLayoutEffect:()=>{},AY=h9.useId||(()=>{}),RY=0;function Tn(e){const[t,n]=E.useState(AY());return An(()=>{n(r=>r??String(RY++))},[e]),t?`radix-${t}`:""}function bn(e){const t=E.useRef(e);return E.useEffect(()=>{t.current=e}),E.useMemo(()=>(...n)=>{var r;return(r=t.current)==null?void 0:r.call(t,...n)},[])}function ja({prop:e,defaultProp:t,onChange:n=()=>{}}){const[r,a]=_Y({defaultProp:t,onChange:n}),o=e!==void 0,s=o?e:r,u=bn(n),c=E.useCallback(d=>{if(o){const g=typeof d=="function"?d(e):d;g!==e&&u(g)}else a(d)},[o,e,a,u]);return[s,c]}function _Y({defaultProp:e,onChange:t}){const n=E.useState(e),[r]=n,a=E.useRef(r),o=bn(t);return E.useEffect(()=>{a.current!==r&&(o(r),a.current=r)},[r,a,o]),n}var Co=E.forwardRef((e,t)=>{const{children:n,...r}=e,a=E.Children.toArray(n),o=a.find(CY);if(o){const s=o.props.children,u=a.map(c=>c===o?E.Children.count(s)>1?E.Children.only(null):E.isValidElement(s)?s.props.children:null:c);return w.jsx(yk,{...r,ref:t,children:E.isValidElement(s)?E.cloneElement(s,void 0,u):null})}return w.jsx(yk,{...r,ref:t,children:n})});Co.displayName="Slot";var yk=E.forwardRef((e,t)=>{const{children:n,...r}=e;if(E.isValidElement(n)){const a=OY(n),o=NY(r,n.props);return n.type!==E.Fragment&&(o.ref=t?oU(t,a):a),E.cloneElement(n,o)}return E.Children.count(n)>1?E.Children.only(null):null});yk.displayName="SlotClone";var I0=({children:e})=>w.jsx(w.Fragment,{children:e});function CY(e){return E.isValidElement(e)&&e.type===I0}function NY(e,t){const n={...t};for(const r in t){const a=e[r],o=t[r];/^on[A-Z]/.test(r)?a&&o?n[r]=(...u)=>{o(...u),a(...u)}:a&&(n[r]=a):r==="style"?n[r]={...a,...o}:r==="className"&&(n[r]=[a,o].filter(Boolean).join(" "))}return{...e,...n}}function OY(e){var r,a;let t=(r=Object.getOwnPropertyDescriptor(e.props,"ref"))==null?void 0:r.get,n=t&&"isReactWarning"in t&&t.isReactWarning;return n?e.ref:(t=(a=Object.getOwnPropertyDescriptor(e,"ref"))==null?void 0:a.get,n=t&&"isReactWarning"in t&&t.isReactWarning,n?e.props.ref:e.props.ref||e.ref)}var DY=["a","button","div","form","h2","h3","img","input","label","li","nav","ol","p","span","svg","ul"],Ze=DY.reduce((e,t)=>{const n=E.forwardRef((r,a)=>{const{asChild:o,...s}=r,u=o?Co:t;return typeof window<"u"&&(window[Symbol.for("radix-ui")]=!0),w.jsx(u,{...s,ref:a})});return n.displayName=`Primitive.${t}`,{...e,[t]:n}},{});function IY(e,t){e&&Gu.flushSync(()=>e.dispatchEvent(t))}function LY(e,t=globalThis==null?void 0:globalThis.document){const n=bn(e);E.useEffect(()=>{const r=a=>{a.key==="Escape"&&n(a)};return t.addEventListener("keydown",r,{capture:!0}),()=>t.removeEventListener("keydown",r,{capture:!0})},[n,t])}var MY="DismissableLayer",vk="dismissableLayer.update",FY="dismissableLayer.pointerDownOutside",PY="dismissableLayer.focusOutside",FC,iU=E.createContext({layers:new Set,layersWithOutsidePointerEventsDisabled:new Set,branches:new Set}),qu=E.forwardRef((e,t)=>{const{disableOutsidePointerEvents:n=!1,onEscapeKeyDown:r,onPointerDownOutside:a,onFocusOutside:o,onInteractOutside:s,onDismiss:u,...c}=e,d=E.useContext(iU),[p,g]=E.useState(null),m=(p==null?void 0:p.ownerDocument)??(globalThis==null?void 0:globalThis.document),[,b]=E.useState({}),y=gt(t,O=>g(O)),S=Array.from(d.layers),[x]=[...d.layersWithOutsidePointerEventsDisabled].slice(-1),R=S.indexOf(x),k=p?S.indexOf(p):-1,A=d.layersWithOutsidePointerEventsDisabled.size>0,C=k>=R,N=UY(O=>{const F=O.target,L=[...d.branches].some(I=>I.contains(F));!C||L||(a==null||a(O),s==null||s(O),O.defaultPrevented||u==null||u())},m),_=jY(O=>{const F=O.target;[...d.branches].some(I=>I.contains(F))||(o==null||o(O),s==null||s(O),O.defaultPrevented||u==null||u())},m);return LY(O=>{k===d.layers.size-1&&(r==null||r(O),!O.defaultPrevented&&u&&(O.preventDefault(),u()))},m),E.useEffect(()=>{if(p)return n&&(d.layersWithOutsidePointerEventsDisabled.size===0&&(FC=m.body.style.pointerEvents,m.body.style.pointerEvents="none"),d.layersWithOutsidePointerEventsDisabled.add(p)),d.layers.add(p),PC(),()=>{n&&d.layersWithOutsidePointerEventsDisabled.size===1&&(m.body.style.pointerEvents=FC)}},[p,m,n,d]),E.useEffect(()=>()=>{p&&(d.layers.delete(p),d.layersWithOutsidePointerEventsDisabled.delete(p),PC())},[p,d]),E.useEffect(()=>{const O=()=>b({});return document.addEventListener(vk,O),()=>document.removeEventListener(vk,O)},[]),w.jsx(Ze.div,{...c,ref:y,style:{pointerEvents:A?C?"auto":"none":void 0,...e.style},onFocusCapture:Ke(e.onFocusCapture,_.onFocusCapture),onBlurCapture:Ke(e.onBlurCapture,_.onBlurCapture),onPointerDownCapture:Ke(e.onPointerDownCapture,N.onPointerDownCapture)})});qu.displayName=MY;var zY="DismissableLayerBranch",BY=E.forwardRef((e,t)=>{const n=E.useContext(iU),r=E.useRef(null),a=gt(t,r);return E.useEffect(()=>{const o=r.current;if(o)return n.branches.add(o),()=>{n.branches.delete(o)}},[n.branches]),w.jsx(Ze.div,{...e,ref:a})});BY.displayName=zY;function UY(e,t=globalThis==null?void 0:globalThis.document){const n=bn(e),r=E.useRef(!1),a=E.useRef(()=>{});return E.useEffect(()=>{const o=u=>{if(u.target&&!r.current){let c=function(){sU(FY,n,d,{discrete:!0})};const d={originalEvent:u};u.pointerType==="touch"?(t.removeEventListener("click",a.current),a.current=c,t.addEventListener("click",a.current,{once:!0})):c()}else t.removeEventListener("click",a.current);r.current=!1},s=window.setTimeout(()=>{t.addEventListener("pointerdown",o)},0);return()=>{window.clearTimeout(s),t.removeEventListener("pointerdown",o),t.removeEventListener("click",a.current)}},[t,n]),{onPointerDownCapture:()=>r.current=!0}}function jY(e,t=globalThis==null?void 0:globalThis.document){const n=bn(e),r=E.useRef(!1);return E.useEffect(()=>{const a=o=>{o.target&&!r.current&&sU(PY,n,{originalEvent:o},{discrete:!1})};return t.addEventListener("focusin",a),()=>t.removeEventListener("focusin",a)},[t,n]),{onFocusCapture:()=>r.current=!0,onBlurCapture:()=>r.current=!1}}function PC(){const e=new CustomEvent(vk);document.dispatchEvent(e)}function sU(e,t,n,{discrete:r}){const a=n.originalEvent.target,o=new CustomEvent(e,{bubbles:!1,cancelable:!0,detail:n});t&&a.addEventListener(e,t,{once:!0}),r?IY(a,o):a.dispatchEvent(o)}var rm="focusScope.autoFocusOnMount",am="focusScope.autoFocusOnUnmount",zC={bubbles:!1,cancelable:!0},GY="FocusScope",tp=E.forwardRef((e,t)=>{const{loop:n=!1,trapped:r=!1,onMountAutoFocus:a,onUnmountAutoFocus:o,...s}=e,[u,c]=E.useState(null),d=bn(a),p=bn(o),g=E.useRef(null),m=gt(t,S=>c(S)),b=E.useRef({paused:!1,pause(){this.paused=!0},resume(){this.paused=!1}}).current;E.useEffect(()=>{if(r){let S=function(A){if(b.paused||!u)return;const C=A.target;u.contains(C)?g.current=C:xo(g.current,{select:!0})},x=function(A){if(b.paused||!u)return;const C=A.relatedTarget;C!==null&&(u.contains(C)||xo(g.current,{select:!0}))},R=function(A){if(document.activeElement===document.body)for(const N of A)N.removedNodes.length>0&&xo(u)};document.addEventListener("focusin",S),document.addEventListener("focusout",x);const k=new MutationObserver(R);return u&&k.observe(u,{childList:!0,subtree:!0}),()=>{document.removeEventListener("focusin",S),document.removeEventListener("focusout",x),k.disconnect()}}},[r,u,b.paused]),E.useEffect(()=>{if(u){UC.add(b);const S=document.activeElement;if(!u.contains(S)){const R=new CustomEvent(rm,zC);u.addEventListener(rm,d),u.dispatchEvent(R),R.defaultPrevented||(HY(YY(lU(u)),{select:!0}),document.activeElement===S&&xo(u))}return()=>{u.removeEventListener(rm,d),setTimeout(()=>{const R=new CustomEvent(am,zC);u.addEventListener(am,p),u.dispatchEvent(R),R.defaultPrevented||xo(S??document.body,{select:!0}),u.removeEventListener(am,p),UC.remove(b)},0)}}},[u,d,p,b]);const y=E.useCallback(S=>{if(!n&&!r||b.paused)return;const x=S.key==="Tab"&&!S.altKey&&!S.ctrlKey&&!S.metaKey,R=document.activeElement;if(x&&R){const k=S.currentTarget,[A,C]=$Y(k);A&&C?!S.shiftKey&&R===C?(S.preventDefault(),n&&xo(A,{select:!0})):S.shiftKey&&R===A&&(S.preventDefault(),n&&xo(C,{select:!0})):R===k&&S.preventDefault()}},[n,r,b.paused]);return w.jsx(Ze.div,{tabIndex:-1,...s,ref:m,onKeyDown:y})});tp.displayName=GY;function HY(e,{select:t=!1}={}){const n=document.activeElement;for(const r of e)if(xo(r,{select:t}),document.activeElement!==n)return}function $Y(e){const t=lU(e),n=BC(t,e),r=BC(t.reverse(),e);return[n,r]}function lU(e){const t=[],n=document.createTreeWalker(e,NodeFilter.SHOW_ELEMENT,{acceptNode:r=>{const a=r.tagName==="INPUT"&&r.type==="hidden";return r.disabled||r.hidden||a?NodeFilter.FILTER_SKIP:r.tabIndex>=0?NodeFilter.FILTER_ACCEPT:NodeFilter.FILTER_SKIP}});for(;n.nextNode();)t.push(n.currentNode);return t}function BC(e,t){for(const n of e)if(!qY(n,{upTo:t}))return n}function qY(e,{upTo:t}){if(getComputedStyle(e).visibility==="hidden")return!0;for(;e;){if(t!==void 0&&e===t)return!1;if(getComputedStyle(e).display==="none")return!0;e=e.parentElement}return!1}function VY(e){return e instanceof HTMLInputElement&&"select"in e}function xo(e,{select:t=!1}={}){if(e&&e.focus){const n=document.activeElement;e.focus({preventScroll:!0}),e!==n&&VY(e)&&t&&e.select()}}var UC=WY();function WY(){let e=[];return{add(t){const n=e[0];t!==n&&(n==null||n.pause()),e=jC(e,t),e.unshift(t)},remove(t){var n;e=jC(e,t),(n=e[0])==null||n.resume()}}}function jC(e,t){const n=[...e],r=n.indexOf(t);return r!==-1&&n.splice(r,1),n}function YY(e){return e.filter(t=>t.tagName!=="A")}var KY="Portal",L0=E.forwardRef((e,t)=>{var u;const{container:n,...r}=e,[a,o]=E.useState(!1);An(()=>o(!0),[]);const s=n||a&&((u=globalThis==null?void 0:globalThis.document)==null?void 0:u.body);return s?tB.createPortal(w.jsx(Ze.div,{...r,ref:t}),s):null});L0.displayName=KY;function XY(e,t){return E.useReducer((n,r)=>t[n][r]??n,e)}var xr=e=>{const{present:t,children:n}=e,r=ZY(t),a=typeof n=="function"?n({present:r.isPresent}):E.Children.only(n),o=gt(r.ref,QY(a));return typeof n=="function"||r.isPresent?E.cloneElement(a,{ref:o}):null};xr.displayName="Presence";function ZY(e){const[t,n]=E.useState(),r=E.useRef({}),a=E.useRef(e),o=E.useRef("none"),s=e?"mounted":"unmounted",[u,c]=XY(s,{mounted:{UNMOUNT:"unmounted",ANIMATION_OUT:"unmountSuspended"},unmountSuspended:{MOUNT:"mounted",ANIMATION_END:"unmounted"},unmounted:{MOUNT:"mounted"}});return E.useEffect(()=>{const d=yd(r.current);o.current=u==="mounted"?d:"none"},[u]),An(()=>{const d=r.current,p=a.current;if(p!==e){const m=o.current,b=yd(d);e?c("MOUNT"):b==="none"||(d==null?void 0:d.display)==="none"?c("UNMOUNT"):c(p&&m!==b?"ANIMATION_OUT":"UNMOUNT"),a.current=e}},[e,c]),An(()=>{if(t){let d;const p=t.ownerDocument.defaultView??window,g=b=>{const S=yd(r.current).includes(b.animationName);if(b.target===t&&S&&(c("ANIMATION_END"),!a.current)){const x=t.style.animationFillMode;t.style.animationFillMode="forwards",d=p.setTimeout(()=>{t.style.animationFillMode==="forwards"&&(t.style.animationFillMode=x)})}},m=b=>{b.target===t&&(o.current=yd(r.current))};return t.addEventListener("animationstart",m),t.addEventListener("animationcancel",g),t.addEventListener("animationend",g),()=>{p.clearTimeout(d),t.removeEventListener("animationstart",m),t.removeEventListener("animationcancel",g),t.removeEventListener("animationend",g)}}else c("ANIMATION_END")},[t,c]),{isPresent:["mounted","unmountSuspended"].includes(u),ref:E.useCallback(d=>{d&&(r.current=getComputedStyle(d)),n(d)},[])}}function yd(e){return(e==null?void 0:e.animationName)||"none"}function QY(e){var r,a;let t=(r=Object.getOwnPropertyDescriptor(e.props,"ref"))==null?void 0:r.get,n=t&&"isReactWarning"in t&&t.isReactWarning;return n?e.ref:(t=(a=Object.getOwnPropertyDescriptor(e,"ref"))==null?void 0:a.get,n=t&&"isReactWarning"in t&&t.isReactWarning,n?e.props.ref:e.props.ref||e.ref)}var om=0;function M0(){E.useEffect(()=>{const e=document.querySelectorAll("[data-radix-focus-guard]");return document.body.insertAdjacentElement("afterbegin",e[0]??GC()),document.body.insertAdjacentElement("beforeend",e[1]??GC()),om++,()=>{om===1&&document.querySelectorAll("[data-radix-focus-guard]").forEach(t=>t.remove()),om--}},[])}function GC(){const e=document.createElement("span");return e.setAttribute("data-radix-focus-guard",""),e.tabIndex=0,e.style.outline="none",e.style.opacity="0",e.style.position="fixed",e.style.pointerEvents="none",e}var ea=function(){return ea=Object.assign||function(t){for(var n,r=1,a=arguments.length;r<a;r++){n=arguments[r];for(var o in n)Object.prototype.hasOwnProperty.call(n,o)&&(t[o]=n[o])}return t},ea.apply(this,arguments)};function uU(e,t){var n={};for(var r in e)Object.prototype.hasOwnProperty.call(e,r)&&t.indexOf(r)<0&&(n[r]=e[r]);if(e!=null&&typeof Object.getOwnPropertySymbols=="function")for(var a=0,r=Object.getOwnPropertySymbols(e);a<r.length;a++)t.indexOf(r[a])<0&&Object.prototype.propertyIsEnumerable.call(e,r[a])&&(n[r[a]]=e[r[a]]);return n}function xi(e,t,n,r){function a(o){return o instanceof n?o:new n(function(s){s(o)})}return new(n||(n=Promise))(function(o,s){function u(p){try{d(r.next(p))}catch(g){s(g)}}function c(p){try{d(r.throw(p))}catch(g){s(g)}}function d(p){p.done?o(p.value):a(p.value).then(u,c)}d((r=r.apply(e,t||[])).next())})}function JY(e,t,n){if(n||arguments.length===2)for(var r=0,a=t.length,o;r<a;r++)(o||!(r in t))&&(o||(o=Array.prototype.slice.call(t,0,r)),o[r]=t[r]);return e.concat(o||Array.prototype.slice.call(t))}var Xd="right-scroll-bar-position",Zd="width-before-scroll-bar",eK="with-scroll-bars-hidden",tK="--removed-body-scroll-bar-size";function im(e,t){return typeof e=="function"?e(t):e&&(e.current=t),e}function nK(e,t){var n=E.useState(function(){return{value:e,callback:t,facade:{get current(){return n.value},set current(r){var a=n.value;a!==r&&(n.value=r,n.callback(r,a))}}}})[0];return n.callback=t,n.facade}var rK=typeof window<"u"?E.useLayoutEffect:E.useEffect,HC=new WeakMap;function aK(e,t){var n=nK(null,function(r){return e.forEach(function(a){return im(a,r)})});return rK(function(){var r=HC.get(n);if(r){var a=new Set(r),o=new Set(e),s=n.current;a.forEach(function(u){o.has(u)||im(u,null)}),o.forEach(function(u){a.has(u)||im(u,s)})}HC.set(n,e)},[e]),n}function oK(e){return e}function iK(e,t){t===void 0&&(t=oK);var n=[],r=!1,a={read:function(){if(r)throw new Error("Sidecar: could not `read` from an `assigned` medium. `read` could be used only with `useMedium`.");return n.length?n[n.length-1]:e},useMedium:function(o){var s=t(o,r);return n.push(s),function(){n=n.filter(function(u){return u!==s})}},assignSyncMedium:function(o){for(r=!0;n.length;){var s=n;n=[],s.forEach(o)}n={push:function(u){return o(u)},filter:function(){return n}}},assignMedium:function(o){r=!0;var s=[];if(n.length){var u=n;n=[],u.forEach(o),s=n}var c=function(){var p=s;s=[],p.forEach(o)},d=function(){return Promise.resolve().then(c)};d(),n={push:function(p){s.push(p),d()},filter:function(p){return s=s.filter(p),n}}}};return a}function sK(e){e===void 0&&(e={});var t=iK(null);return t.options=ea({async:!0,ssr:!1},e),t}var cU=function(e){var t=e.sideCar,n=uU(e,["sideCar"]);if(!t)throw new Error("Sidecar: please provide `sideCar` property to import the right car");var r=t.read();if(!r)throw new Error("Sidecar medium not found");return E.createElement(r,ea({},n))};cU.isSideCarExport=!0;function lK(e,t){return e.useMedium(t),cU}var dU=sK(),sm=function(){},np=E.forwardRef(function(e,t){var n=E.useRef(null),r=E.useState({onScrollCapture:sm,onWheelCapture:sm,onTouchMoveCapture:sm}),a=r[0],o=r[1],s=e.forwardProps,u=e.children,c=e.className,d=e.removeScrollBar,p=e.enabled,g=e.shards,m=e.sideCar,b=e.noIsolation,y=e.inert,S=e.allowPinchZoom,x=e.as,R=x===void 0?"div":x,k=e.gapMode,A=uU(e,["forwardProps","children","className","removeScrollBar","enabled","shards","sideCar","noIsolation","inert","allowPinchZoom","as","gapMode"]),C=m,N=aK([n,t]),_=ea(ea({},A),a);return E.createElement(E.Fragment,null,p&&E.createElement(C,{sideCar:dU,removeScrollBar:d,shards:g,noIsolation:b,inert:y,setCallbacks:o,allowPinchZoom:!!S,lockRef:n,gapMode:k}),s?E.cloneElement(E.Children.only(u),ea(ea({},_),{ref:N})):E.createElement(R,ea({},_,{className:c,ref:N}),u))});np.defaultProps={enabled:!0,removeScrollBar:!0,inert:!1};np.classNames={fullWidth:Zd,zeroRight:Xd};var uK=function(){if(typeof __webpack_nonce__<"u")return __webpack_nonce__};function cK(){if(!document)return null;var e=document.createElement("style");e.type="text/css";var t=uK();return t&&e.setAttribute("nonce",t),e}function dK(e,t){e.styleSheet?e.styleSheet.cssText=t:e.appendChild(document.createTextNode(t))}function fK(e){var t=document.head||document.getElementsByTagName("head")[0];t.appendChild(e)}var pK=function(){var e=0,t=null;return{add:function(n){e==0&&(t=cK())&&(dK(t,n),fK(t)),e++},remove:function(){e--,!e&&t&&(t.parentNode&&t.parentNode.removeChild(t),t=null)}}},gK=function(){var e=pK();return function(t,n){E.useEffect(function(){return e.add(t),function(){e.remove()}},[t&&n])}},fU=function(){var e=gK(),t=function(n){var r=n.styles,a=n.dynamic;return e(r,a),null};return t},hK={left:0,top:0,right:0,gap:0},lm=function(e){return parseInt(e||"",10)||0},mK=function(e){var t=window.getComputedStyle(document.body),n=t[e==="padding"?"paddingLeft":"marginLeft"],r=t[e==="padding"?"paddingTop":"marginTop"],a=t[e==="padding"?"paddingRight":"marginRight"];return[lm(n),lm(r),lm(a)]},bK=function(e){if(e===void 0&&(e="margin"),typeof window>"u")return hK;var t=mK(e),n=document.documentElement.clientWidth,r=window.innerWidth;return{left:t[0],top:t[1],right:t[2],gap:Math.max(0,r-n+t[2]-t[0])}},yK=fU(),ks="data-scroll-locked",vK=function(e,t,n,r){var a=e.left,o=e.top,s=e.right,u=e.gap;return n===void 0&&(n="margin"),`
  .`.concat(eK,` {
   overflow: hidden `).concat(r,`;
   padding-right: `).concat(u,"px ").concat(r,`;
  }
  body[`).concat(ks,`] {
    overflow: hidden `).concat(r,`;
    overscroll-behavior: contain;
    `).concat([t&&"position: relative ".concat(r,";"),n==="margin"&&`
    padding-left: `.concat(a,`px;
    padding-top: `).concat(o,`px;
    padding-right: `).concat(s,`px;
    margin-left:0;
    margin-top:0;
    margin-right: `).concat(u,"px ").concat(r,`;
    `),n==="padding"&&"padding-right: ".concat(u,"px ").concat(r,";")].filter(Boolean).join(""),`
  }
  
  .`).concat(Xd,` {
    right: `).concat(u,"px ").concat(r,`;
  }
  
  .`).concat(Zd,` {
    margin-right: `).concat(u,"px ").concat(r,`;
  }
  
  .`).concat(Xd," .").concat(Xd,` {
    right: 0 `).concat(r,`;
  }
  
  .`).concat(Zd," .").concat(Zd,` {
    margin-right: 0 `).concat(r,`;
  }
  
  body[`).concat(ks,`] {
    `).concat(tK,": ").concat(u,`px;
  }
`)},$C=function(){var e=parseInt(document.body.getAttribute(ks)||"0",10);return isFinite(e)?e:0},SK=function(){E.useEffect(function(){return document.body.setAttribute(ks,($C()+1).toString()),function(){var e=$C()-1;e<=0?document.body.removeAttribute(ks):document.body.setAttribute(ks,e.toString())}},[])},EK=function(e){var t=e.noRelative,n=e.noImportant,r=e.gapMode,a=r===void 0?"margin":r;SK();var o=E.useMemo(function(){return bK(a)},[a]);return E.createElement(yK,{styles:vK(o,!t,a,n?"":"!important")})},Sk=!1;if(typeof window<"u")try{var vd=Object.defineProperty({},"passive",{get:function(){return Sk=!0,!0}});window.addEventListener("test",vd,vd),window.removeEventListener("test",vd,vd)}catch{Sk=!1}var us=Sk?{passive:!1}:!1,wK=function(e){return e.tagName==="TEXTAREA"},pU=function(e,t){if(!(e instanceof Element))return!1;var n=window.getComputedStyle(e);return n[t]!=="hidden"&&!(n.overflowY===n.overflowX&&!wK(e)&&n[t]==="visible")},xK=function(e){return pU(e,"overflowY")},kK=function(e){return pU(e,"overflowX")},qC=function(e,t){var n=t.ownerDocument,r=t;do{typeof ShadowRoot<"u"&&r instanceof ShadowRoot&&(r=r.host);var a=gU(e,r);if(a){var o=hU(e,r),s=o[1],u=o[2];if(s>u)return!0}r=r.parentNode}while(r&&r!==n.body);return!1},TK=function(e){var t=e.scrollTop,n=e.scrollHeight,r=e.clientHeight;return[t,n,r]},AK=function(e){var t=e.scrollLeft,n=e.scrollWidth,r=e.clientWidth;return[t,n,r]},gU=function(e,t){return e==="v"?xK(t):kK(t)},hU=function(e,t){return e==="v"?TK(t):AK(t)},RK=function(e,t){return e==="h"&&t==="rtl"?-1:1},_K=function(e,t,n,r,a){var o=RK(e,window.getComputedStyle(t).direction),s=o*r,u=n.target,c=t.contains(u),d=!1,p=s>0,g=0,m=0;do{var b=hU(e,u),y=b[0],S=b[1],x=b[2],R=S-x-o*y;(y||R)&&gU(e,u)&&(g+=R,m+=y),u instanceof ShadowRoot?u=u.host:u=u.parentNode}while(!c&&u!==document.body||c&&(t.contains(u)||t===u));return(p&&Math.abs(g)<1||!p&&Math.abs(m)<1)&&(d=!0),d},Sd=function(e){return"changedTouches"in e?[e.changedTouches[0].clientX,e.changedTouches[0].clientY]:[0,0]},VC=function(e){return[e.deltaX,e.deltaY]},WC=function(e){return e&&"current"in e?e.current:e},CK=function(e,t){return e[0]===t[0]&&e[1]===t[1]},NK=function(e){return`
  .block-interactivity-`.concat(e,` {pointer-events: none;}
  .allow-interactivity-`).concat(e,` {pointer-events: all;}
`)},OK=0,cs=[];function DK(e){var t=E.useRef([]),n=E.useRef([0,0]),r=E.useRef(),a=E.useState(OK++)[0],o=E.useState(fU)[0],s=E.useRef(e);E.useEffect(function(){s.current=e},[e]),E.useEffect(function(){if(e.inert){document.body.classList.add("block-interactivity-".concat(a));var S=JY([e.lockRef.current],(e.shards||[]).map(WC),!0).filter(Boolean);return S.forEach(function(x){return x.classList.add("allow-interactivity-".concat(a))}),function(){document.body.classList.remove("block-interactivity-".concat(a)),S.forEach(function(x){return x.classList.remove("allow-interactivity-".concat(a))})}}},[e.inert,e.lockRef.current,e.shards]);var u=E.useCallback(function(S,x){if("touches"in S&&S.touches.length===2||S.type==="wheel"&&S.ctrlKey)return!s.current.allowPinchZoom;var R=Sd(S),k=n.current,A="deltaX"in S?S.deltaX:k[0]-R[0],C="deltaY"in S?S.deltaY:k[1]-R[1],N,_=S.target,O=Math.abs(A)>Math.abs(C)?"h":"v";if("touches"in S&&O==="h"&&_.type==="range")return!1;var F=qC(O,_);if(!F)return!0;if(F?N=O:(N=O==="v"?"h":"v",F=qC(O,_)),!F)return!1;if(!r.current&&"changedTouches"in S&&(A||C)&&(r.current=N),!N)return!0;var L=r.current||N;return _K(L,x,S,L==="h"?A:C)},[]),c=E.useCallback(function(S){var x=S;if(!(!cs.length||cs[cs.length-1]!==o)){var R="deltaY"in x?VC(x):Sd(x),k=t.current.filter(function(N){return N.name===x.type&&(N.target===x.target||x.target===N.shadowParent)&&CK(N.delta,R)})[0];if(k&&k.should){x.cancelable&&x.preventDefault();return}if(!k){var A=(s.current.shards||[]).map(WC).filter(Boolean).filter(function(N){return N.contains(x.target)}),C=A.length>0?u(x,A[0]):!s.current.noIsolation;C&&x.cancelable&&x.preventDefault()}}},[]),d=E.useCallback(function(S,x,R,k){var A={name:S,delta:x,target:R,should:k,shadowParent:IK(R)};t.current.push(A),setTimeout(function(){t.current=t.current.filter(function(C){return C!==A})},1)},[]),p=E.useCallback(function(S){n.current=Sd(S),r.current=void 0},[]),g=E.useCallback(function(S){d(S.type,VC(S),S.target,u(S,e.lockRef.current))},[]),m=E.useCallback(function(S){d(S.type,Sd(S),S.target,u(S,e.lockRef.current))},[]);E.useEffect(function(){return cs.push(o),e.setCallbacks({onScrollCapture:g,onWheelCapture:g,onTouchMoveCapture:m}),document.addEventListener("wheel",c,us),document.addEventListener("touchmove",c,us),document.addEventListener("touchstart",p,us),function(){cs=cs.filter(function(S){return S!==o}),document.removeEventListener("wheel",c,us),document.removeEventListener("touchmove",c,us),document.removeEventListener("touchstart",p,us)}},[]);var b=e.removeScrollBar,y=e.inert;return E.createElement(E.Fragment,null,y?E.createElement(o,{styles:NK(a)}):null,b?E.createElement(EK,{gapMode:e.gapMode}):null)}function IK(e){for(var t=null;e!==null;)e instanceof ShadowRoot&&(t=e.host,e=e.host),e=e.parentNode;return t}const LK=lK(dU,DK);var rp=E.forwardRef(function(e,t){return E.createElement(np,ea({},e,{ref:t,sideCar:LK}))});rp.classNames=np.classNames;var MK=function(e){if(typeof document>"u")return null;var t=Array.isArray(e)?e[0]:e;return t.ownerDocument.body},ds=new WeakMap,Ed=new WeakMap,wd={},um=0,mU=function(e){return e&&(e.host||mU(e.parentNode))},FK=function(e,t){return t.map(function(n){if(e.contains(n))return n;var r=mU(n);return r&&e.contains(r)?r:(console.error("aria-hidden",n,"in not contained inside",e,". Doing nothing"),null)}).filter(function(n){return!!n})},PK=function(e,t,n,r){var a=FK(t,Array.isArray(e)?e:[e]);wd[n]||(wd[n]=new WeakMap);var o=wd[n],s=[],u=new Set,c=new Set(a),d=function(g){!g||u.has(g)||(u.add(g),d(g.parentNode))};a.forEach(d);var p=function(g){!g||c.has(g)||Array.prototype.forEach.call(g.children,function(m){if(u.has(m))p(m);else try{var b=m.getAttribute(r),y=b!==null&&b!=="false",S=(ds.get(m)||0)+1,x=(o.get(m)||0)+1;ds.set(m,S),o.set(m,x),s.push(m),S===1&&y&&Ed.set(m,!0),x===1&&m.setAttribute(n,"true"),y||m.setAttribute(r,"true")}catch(R){console.error("aria-hidden: cannot operate on ",m,R)}})};return p(t),u.clear(),um++,function(){s.forEach(function(g){var m=ds.get(g)-1,b=o.get(g)-1;ds.set(g,m),o.set(g,b),m||(Ed.has(g)||g.removeAttribute(r),Ed.delete(g)),b||g.removeAttribute(n)}),um--,um||(ds=new WeakMap,ds=new WeakMap,Ed=new WeakMap,wd={})}},F0=function(e,t,n){n===void 0&&(n="data-aria-hidden");var r=Array.from(Array.isArray(e)?e:[e]),a=MK(e);return a?(r.push.apply(r,Array.from(a.querySelectorAll("[aria-live]"))),PK(r,a,n,"aria-hidden")):function(){return null}},P0="Dialog",[bU,yU]=wr(P0),[zK,Hr]=bU(P0),vU=e=>{const{__scopeDialog:t,children:n,open:r,defaultOpen:a,onOpenChange:o,modal:s=!0}=e,u=E.useRef(null),c=E.useRef(null),[d=!1,p]=ja({prop:r,defaultProp:a,onChange:o});return w.jsx(zK,{scope:t,triggerRef:u,contentRef:c,contentId:Tn(),titleId:Tn(),descriptionId:Tn(),open:d,onOpenChange:p,onOpenToggle:E.useCallback(()=>p(g=>!g),[p]),modal:s,children:n})};vU.displayName=P0;var SU="DialogTrigger",EU=E.forwardRef((e,t)=>{const{__scopeDialog:n,...r}=e,a=Hr(SU,n),o=gt(t,a.triggerRef);return w.jsx(Ze.button,{type:"button","aria-haspopup":"dialog","aria-expanded":a.open,"aria-controls":a.contentId,"data-state":U0(a.open),...r,ref:o,onClick:Ke(e.onClick,a.onOpenToggle)})});EU.displayName=SU;var z0="DialogPortal",[BK,wU]=bU(z0,{forceMount:void 0}),xU=e=>{const{__scopeDialog:t,forceMount:n,children:r,container:a}=e,o=Hr(z0,t);return w.jsx(BK,{scope:t,forceMount:n,children:E.Children.map(r,s=>w.jsx(xr,{present:n||o.open,children:w.jsx(L0,{asChild:!0,container:a,children:s})}))})};xU.displayName=z0;var hf="DialogOverlay",kU=E.forwardRef((e,t)=>{const n=wU(hf,e.__scopeDialog),{forceMount:r=n.forceMount,...a}=e,o=Hr(hf,e.__scopeDialog);return o.modal?w.jsx(xr,{present:r||o.open,children:w.jsx(UK,{...a,ref:t})}):null});kU.displayName=hf;var UK=E.forwardRef((e,t)=>{const{__scopeDialog:n,...r}=e,a=Hr(hf,n);return w.jsx(rp,{as:Co,allowPinchZoom:!0,shards:[a.contentRef],children:w.jsx(Ze.div,{"data-state":U0(a.open),...r,ref:t,style:{pointerEvents:"auto",...r.style}})})}),hi="DialogContent",TU=E.forwardRef((e,t)=>{const n=wU(hi,e.__scopeDialog),{forceMount:r=n.forceMount,...a}=e,o=Hr(hi,e.__scopeDialog);return w.jsx(xr,{present:r||o.open,children:o.modal?w.jsx(jK,{...a,ref:t}):w.jsx(GK,{...a,ref:t})})});TU.displayName=hi;var jK=E.forwardRef((e,t)=>{const n=Hr(hi,e.__scopeDialog),r=E.useRef(null),a=gt(t,n.contentRef,r);return E.useEffect(()=>{const o=r.current;if(o)return F0(o)},[]),w.jsx(AU,{...e,ref:a,trapFocus:n.open,disableOutsidePointerEvents:!0,onCloseAutoFocus:Ke(e.onCloseAutoFocus,o=>{var s;o.preventDefault(),(s=n.triggerRef.current)==null||s.focus()}),onPointerDownOutside:Ke(e.onPointerDownOutside,o=>{const s=o.detail.originalEvent,u=s.button===0&&s.ctrlKey===!0;(s.button===2||u)&&o.preventDefault()}),onFocusOutside:Ke(e.onFocusOutside,o=>o.preventDefault())})}),GK=E.forwardRef((e,t)=>{const n=Hr(hi,e.__scopeDialog),r=E.useRef(!1),a=E.useRef(!1);return w.jsx(AU,{...e,ref:t,trapFocus:!1,disableOutsidePointerEvents:!1,onCloseAutoFocus:o=>{var s,u;(s=e.onCloseAutoFocus)==null||s.call(e,o),o.defaultPrevented||(r.current||(u=n.triggerRef.current)==null||u.focus(),o.preventDefault()),r.current=!1,a.current=!1},onInteractOutside:o=>{var c,d;(c=e.onInteractOutside)==null||c.call(e,o),o.defaultPrevented||(r.current=!0,o.detail.originalEvent.type==="pointerdown"&&(a.current=!0));const s=o.target;((d=n.triggerRef.current)==null?void 0:d.contains(s))&&o.preventDefault(),o.detail.originalEvent.type==="focusin"&&a.current&&o.preventDefault()}})}),AU=E.forwardRef((e,t)=>{const{__scopeDialog:n,trapFocus:r,onOpenAutoFocus:a,onCloseAutoFocus:o,...s}=e,u=Hr(hi,n),c=E.useRef(null),d=gt(t,c);return M0(),w.jsxs(w.Fragment,{children:[w.jsx(tp,{asChild:!0,loop:!0,trapped:r,onMountAutoFocus:a,onUnmountAutoFocus:o,children:w.jsx(qu,{role:"dialog",id:u.contentId,"aria-describedby":u.descriptionId,"aria-labelledby":u.titleId,"data-state":U0(u.open),...s,ref:d,onDismiss:()=>u.onOpenChange(!1)})}),w.jsxs(w.Fragment,{children:[w.jsx($K,{titleId:u.titleId}),w.jsx(VK,{contentRef:c,descriptionId:u.descriptionId})]})]})}),B0="DialogTitle",RU=E.forwardRef((e,t)=>{const{__scopeDialog:n,...r}=e,a=Hr(B0,n);return w.jsx(Ze.h2,{id:a.titleId,...r,ref:t})});RU.displayName=B0;var _U="DialogDescription",CU=E.forwardRef((e,t)=>{const{__scopeDialog:n,...r}=e,a=Hr(_U,n);return w.jsx(Ze.p,{id:a.descriptionId,...r,ref:t})});CU.displayName=_U;var NU="DialogClose",OU=E.forwardRef((e,t)=>{const{__scopeDialog:n,...r}=e,a=Hr(NU,n);return w.jsx(Ze.button,{type:"button",...r,ref:t,onClick:Ke(e.onClick,()=>a.onOpenChange(!1))})});OU.displayName=NU;function U0(e){return e?"open":"closed"}var DU="DialogTitleWarning",[HK,IU]=kY(DU,{contentName:hi,titleName:B0,docsSlug:"dialog"}),$K=({titleId:e})=>{const t=IU(DU),n=`\`${t.contentName}\` requires a \`${t.titleName}\` for the component to be accessible for screen reader users.

If you want to hide the \`${t.titleName}\`, you can wrap it with our VisuallyHidden component.

For more information, see https://radix-ui.com/primitives/docs/components/${t.docsSlug}`;return E.useEffect(()=>{e&&(document.getElementById(e)||console.error(n))},[n,e]),null},qK="DialogDescriptionWarning",VK=({contentRef:e,descriptionId:t})=>{const r=`Warning: Missing \`Description\` or \`aria-describedby={undefined}\` for {${IU(qK).contentName}}.`;return E.useEffect(()=>{var o;const a=(o=e.current)==null?void 0:o.getAttribute("aria-describedby");t&&a&&(document.getElementById(t)||console.warn(r))},[r,e,t]),null},j0=vU,LU=EU,G0=xU,ap=kU,op=TU,H0=RU,$0=CU,q0=OU,MU="AlertDialog",[WK,Cxe]=wr(MU,[yU]),Va=yU(),FU=e=>{const{__scopeAlertDialog:t,...n}=e,r=Va(t);return w.jsx(j0,{...r,...n,modal:!0})};FU.displayName=MU;var YK="AlertDialogTrigger",KK=E.forwardRef((e,t)=>{const{__scopeAlertDialog:n,...r}=e,a=Va(n);return w.jsx(LU,{...a,...r,ref:t})});KK.displayName=YK;var XK="AlertDialogPortal",PU=e=>{const{__scopeAlertDialog:t,...n}=e,r=Va(t);return w.jsx(G0,{...r,...n})};PU.displayName=XK;var ZK="AlertDialogOverlay",zU=E.forwardRef((e,t)=>{const{__scopeAlertDialog:n,...r}=e,a=Va(n);return w.jsx(ap,{...a,...r,ref:t})});zU.displayName=ZK;var Ts="AlertDialogContent",[QK,JK]=WK(Ts),BU=E.forwardRef((e,t)=>{const{__scopeAlertDialog:n,children:r,...a}=e,o=Va(n),s=E.useRef(null),u=gt(t,s),c=E.useRef(null);return w.jsx(HK,{contentName:Ts,titleName:UU,docsSlug:"alert-dialog",children:w.jsx(QK,{scope:n,cancelRef:c,children:w.jsxs(op,{role:"alertdialog",...o,...a,ref:u,onOpenAutoFocus:Ke(a.onOpenAutoFocus,d=>{var p;d.preventDefault(),(p=c.current)==null||p.focus({preventScroll:!0})}),onPointerDownOutside:d=>d.preventDefault(),onInteractOutside:d=>d.preventDefault(),children:[w.jsx(I0,{children:r}),w.jsx(tX,{contentRef:s})]})})})});BU.displayName=Ts;var UU="AlertDialogTitle",jU=E.forwardRef((e,t)=>{const{__scopeAlertDialog:n,...r}=e,a=Va(n);return w.jsx(H0,{...a,...r,ref:t})});jU.displayName=UU;var GU="AlertDialogDescription",HU=E.forwardRef((e,t)=>{const{__scopeAlertDialog:n,...r}=e,a=Va(n);return w.jsx($0,{...a,...r,ref:t})});HU.displayName=GU;var eX="AlertDialogAction",$U=E.forwardRef((e,t)=>{const{__scopeAlertDialog:n,...r}=e,a=Va(n);return w.jsx(q0,{...a,...r,ref:t})});$U.displayName=eX;var qU="AlertDialogCancel",VU=E.forwardRef((e,t)=>{const{__scopeAlertDialog:n,...r}=e,{cancelRef:a}=JK(qU,n),o=Va(n),s=gt(t,a);return w.jsx(q0,{...o,...r,ref:s})});VU.displayName=qU;var tX=({contentRef:e})=>{const t=`\`${Ts}\` requires a description for the component to be accessible for screen reader users.

You can add a description to the \`${Ts}\` by passing a \`${GU}\` component as a child, which also benefits sighted users by adding visible context to the dialog.

Alternatively, you can use your own component as a description by assigning it an \`id\` and passing the same value to the \`aria-describedby\` prop in \`${Ts}\`. If the description is confusing or duplicative for sighted users, you can use the \`@radix-ui/react-visually-hidden\` primitive as a wrapper around your description component.

For more information, see https://radix-ui.com/primitives/docs/components/alert-dialog`;return E.useEffect(()=>{var r;document.getElementById((r=e.current)==null?void 0:r.getAttribute("aria-describedby"))||console.warn(t)},[t,e]),null},nX=FU,rX=PU,WU=zU,YU=BU,KU=$U,XU=VU,ZU=jU,QU=HU;const aX=["top","right","bottom","left"],No=Math.min,er=Math.max,mf=Math.round,xd=Math.floor,aa=e=>({x:e,y:e}),oX={left:"right",right:"left",bottom:"top",top:"bottom"},iX={start:"end",end:"start"};function Ek(e,t,n){return er(e,No(t,n))}function Ga(e,t){return typeof e=="function"?e(t):e}function Ha(e){return e.split("-")[0]}function $s(e){return e.split("-")[1]}function V0(e){return e==="x"?"y":"x"}function W0(e){return e==="y"?"height":"width"}function Oo(e){return["top","bottom"].includes(Ha(e))?"y":"x"}function Y0(e){return V0(Oo(e))}function sX(e,t,n){n===void 0&&(n=!1);const r=$s(e),a=Y0(e),o=W0(a);let s=a==="x"?r===(n?"end":"start")?"right":"left":r==="start"?"bottom":"top";return t.reference[o]>t.floating[o]&&(s=bf(s)),[s,bf(s)]}function lX(e){const t=bf(e);return[wk(e),t,wk(t)]}function wk(e){return e.replace(/start|end/g,t=>iX[t])}function uX(e,t,n){const r=["left","right"],a=["right","left"],o=["top","bottom"],s=["bottom","top"];switch(e){case"top":case"bottom":return n?t?a:r:t?r:a;case"left":case"right":return t?o:s;default:return[]}}function cX(e,t,n,r){const a=$s(e);let o=uX(Ha(e),n==="start",r);return a&&(o=o.map(s=>s+"-"+a),t&&(o=o.concat(o.map(wk)))),o}function bf(e){return e.replace(/left|right|bottom|top/g,t=>oX[t])}function dX(e){return{top:0,right:0,bottom:0,left:0,...e}}function JU(e){return typeof e!="number"?dX(e):{top:e,right:e,bottom:e,left:e}}function yf(e){const{x:t,y:n,width:r,height:a}=e;return{width:r,height:a,top:n,left:t,right:t+r,bottom:n+a,x:t,y:n}}function YC(e,t,n){let{reference:r,floating:a}=e;const o=Oo(t),s=Y0(t),u=W0(s),c=Ha(t),d=o==="y",p=r.x+r.width/2-a.width/2,g=r.y+r.height/2-a.height/2,m=r[u]/2-a[u]/2;let b;switch(c){case"top":b={x:p,y:r.y-a.height};break;case"bottom":b={x:p,y:r.y+r.height};break;case"right":b={x:r.x+r.width,y:g};break;case"left":b={x:r.x-a.width,y:g};break;default:b={x:r.x,y:r.y}}switch($s(t)){case"start":b[s]-=m*(n&&d?-1:1);break;case"end":b[s]+=m*(n&&d?-1:1);break}return b}const fX=async(e,t,n)=>{const{placement:r="bottom",strategy:a="absolute",middleware:o=[],platform:s}=n,u=o.filter(Boolean),c=await(s.isRTL==null?void 0:s.isRTL(t));let d=await s.getElementRects({reference:e,floating:t,strategy:a}),{x:p,y:g}=YC(d,r,c),m=r,b={},y=0;for(let S=0;S<u.length;S++){const{name:x,fn:R}=u[S],{x:k,y:A,data:C,reset:N}=await R({x:p,y:g,initialPlacement:r,placement:m,strategy:a,middlewareData:b,rects:d,platform:s,elements:{reference:e,floating:t}});p=k??p,g=A??g,b={...b,[x]:{...b[x],...C}},N&&y<=50&&(y++,typeof N=="object"&&(N.placement&&(m=N.placement),N.rects&&(d=N.rects===!0?await s.getElementRects({reference:e,floating:t,strategy:a}):N.rects),{x:p,y:g}=YC(d,m,c)),S=-1)}return{x:p,y:g,placement:m,strategy:a,middlewareData:b}};async function ku(e,t){var n;t===void 0&&(t={});const{x:r,y:a,platform:o,rects:s,elements:u,strategy:c}=e,{boundary:d="clippingAncestors",rootBoundary:p="viewport",elementContext:g="floating",altBoundary:m=!1,padding:b=0}=Ga(t,e),y=JU(b),x=u[m?g==="floating"?"reference":"floating":g],R=yf(await o.getClippingRect({element:(n=await(o.isElement==null?void 0:o.isElement(x)))==null||n?x:x.contextElement||await(o.getDocumentElement==null?void 0:o.getDocumentElement(u.floating)),boundary:d,rootBoundary:p,strategy:c})),k=g==="floating"?{x:r,y:a,width:s.floating.width,height:s.floating.height}:s.reference,A=await(o.getOffsetParent==null?void 0:o.getOffsetParent(u.floating)),C=await(o.isElement==null?void 0:o.isElement(A))?await(o.getScale==null?void 0:o.getScale(A))||{x:1,y:1}:{x:1,y:1},N=yf(o.convertOffsetParentRelativeRectToViewportRelativeRect?await o.convertOffsetParentRelativeRectToViewportRelativeRect({elements:u,rect:k,offsetParent:A,strategy:c}):k);return{top:(R.top-N.top+y.top)/C.y,bottom:(N.bottom-R.bottom+y.bottom)/C.y,left:(R.left-N.left+y.left)/C.x,right:(N.right-R.right+y.right)/C.x}}const pX=e=>({name:"arrow",options:e,async fn(t){const{x:n,y:r,placement:a,rects:o,platform:s,elements:u,middlewareData:c}=t,{element:d,padding:p=0}=Ga(e,t)||{};if(d==null)return{};const g=JU(p),m={x:n,y:r},b=Y0(a),y=W0(b),S=await s.getDimensions(d),x=b==="y",R=x?"top":"left",k=x?"bottom":"right",A=x?"clientHeight":"clientWidth",C=o.reference[y]+o.reference[b]-m[b]-o.floating[y],N=m[b]-o.reference[b],_=await(s.getOffsetParent==null?void 0:s.getOffsetParent(d));let O=_?_[A]:0;(!O||!await(s.isElement==null?void 0:s.isElement(_)))&&(O=u.floating[A]||o.floating[y]);const F=C/2-N/2,L=O/2-S[y]/2-1,I=No(g[R],L),H=No(g[k],L),$=I,U=O-S[y]-H,Y=O/2-S[y]/2+F,Z=Ek($,Y,U),j=!c.arrow&&$s(a)!=null&&Y!==Z&&o.reference[y]/2-(Y<$?I:H)-S[y]/2<0,G=j?Y<$?Y-$:Y-U:0;return{[b]:m[b]+G,data:{[b]:Z,centerOffset:Y-Z-G,...j&&{alignmentOffset:G}},reset:j}}}),gX=function(e){return e===void 0&&(e={}),{name:"flip",options:e,async fn(t){var n,r;const{placement:a,middlewareData:o,rects:s,initialPlacement:u,platform:c,elements:d}=t,{mainAxis:p=!0,crossAxis:g=!0,fallbackPlacements:m,fallbackStrategy:b="bestFit",fallbackAxisSideDirection:y="none",flipAlignment:S=!0,...x}=Ga(e,t);if((n=o.arrow)!=null&&n.alignmentOffset)return{};const R=Ha(a),k=Oo(u),A=Ha(u)===u,C=await(c.isRTL==null?void 0:c.isRTL(d.floating)),N=m||(A||!S?[bf(u)]:lX(u)),_=y!=="none";!m&&_&&N.push(...cX(u,S,y,C));const O=[u,...N],F=await ku(t,x),L=[];let I=((r=o.flip)==null?void 0:r.overflows)||[];if(p&&L.push(F[R]),g){const Y=sX(a,s,C);L.push(F[Y[0]],F[Y[1]])}if(I=[...I,{placement:a,overflows:L}],!L.every(Y=>Y<=0)){var H,$;const Y=(((H=o.flip)==null?void 0:H.index)||0)+1,Z=O[Y];if(Z)return{data:{index:Y,overflows:I},reset:{placement:Z}};let j=($=I.filter(G=>G.overflows[0]<=0).sort((G,z)=>G.overflows[1]-z.overflows[1])[0])==null?void 0:$.placement;if(!j)switch(b){case"bestFit":{var U;const G=(U=I.filter(z=>{if(_){const K=Oo(z.placement);return K===k||K==="y"}return!0}).map(z=>[z.placement,z.overflows.filter(K=>K>0).reduce((K,D)=>K+D,0)]).sort((z,K)=>z[1]-K[1])[0])==null?void 0:U[0];G&&(j=G);break}case"initialPlacement":j=u;break}if(a!==j)return{reset:{placement:j}}}return{}}}};function KC(e,t){return{top:e.top-t.height,right:e.right-t.width,bottom:e.bottom-t.height,left:e.left-t.width}}function XC(e){return aX.some(t=>e[t]>=0)}const hX=function(e){return e===void 0&&(e={}),{name:"hide",options:e,async fn(t){const{rects:n}=t,{strategy:r="referenceHidden",...a}=Ga(e,t);switch(r){case"referenceHidden":{const o=await ku(t,{...a,elementContext:"reference"}),s=KC(o,n.reference);return{data:{referenceHiddenOffsets:s,referenceHidden:XC(s)}}}case"escaped":{const o=await ku(t,{...a,altBoundary:!0}),s=KC(o,n.floating);return{data:{escapedOffsets:s,escaped:XC(s)}}}default:return{}}}}};async function mX(e,t){const{placement:n,platform:r,elements:a}=e,o=await(r.isRTL==null?void 0:r.isRTL(a.floating)),s=Ha(n),u=$s(n),c=Oo(n)==="y",d=["left","top"].includes(s)?-1:1,p=o&&c?-1:1,g=Ga(t,e);let{mainAxis:m,crossAxis:b,alignmentAxis:y}=typeof g=="number"?{mainAxis:g,crossAxis:0,alignmentAxis:null}:{mainAxis:g.mainAxis||0,crossAxis:g.crossAxis||0,alignmentAxis:g.alignmentAxis};return u&&typeof y=="number"&&(b=u==="end"?y*-1:y),c?{x:b*p,y:m*d}:{x:m*d,y:b*p}}const bX=function(e){return e===void 0&&(e=0),{name:"offset",options:e,async fn(t){var n,r;const{x:a,y:o,placement:s,middlewareData:u}=t,c=await mX(t,e);return s===((n=u.offset)==null?void 0:n.placement)&&(r=u.arrow)!=null&&r.alignmentOffset?{}:{x:a+c.x,y:o+c.y,data:{...c,placement:s}}}}},yX=function(e){return e===void 0&&(e={}),{name:"shift",options:e,async fn(t){const{x:n,y:r,placement:a}=t,{mainAxis:o=!0,crossAxis:s=!1,limiter:u={fn:x=>{let{x:R,y:k}=x;return{x:R,y:k}}},...c}=Ga(e,t),d={x:n,y:r},p=await ku(t,c),g=Oo(Ha(a)),m=V0(g);let b=d[m],y=d[g];if(o){const x=m==="y"?"top":"left",R=m==="y"?"bottom":"right",k=b+p[x],A=b-p[R];b=Ek(k,b,A)}if(s){const x=g==="y"?"top":"left",R=g==="y"?"bottom":"right",k=y+p[x],A=y-p[R];y=Ek(k,y,A)}const S=u.fn({...t,[m]:b,[g]:y});return{...S,data:{x:S.x-n,y:S.y-r,enabled:{[m]:o,[g]:s}}}}}},vX=function(e){return e===void 0&&(e={}),{options:e,fn(t){const{x:n,y:r,placement:a,rects:o,middlewareData:s}=t,{offset:u=0,mainAxis:c=!0,crossAxis:d=!0}=Ga(e,t),p={x:n,y:r},g=Oo(a),m=V0(g);let b=p[m],y=p[g];const S=Ga(u,t),x=typeof S=="number"?{mainAxis:S,crossAxis:0}:{mainAxis:0,crossAxis:0,...S};if(c){const A=m==="y"?"height":"width",C=o.reference[m]-o.floating[A]+x.mainAxis,N=o.reference[m]+o.reference[A]-x.mainAxis;b<C?b=C:b>N&&(b=N)}if(d){var R,k;const A=m==="y"?"width":"height",C=["top","left"].includes(Ha(a)),N=o.reference[g]-o.floating[A]+(C&&((R=s.offset)==null?void 0:R[g])||0)+(C?0:x.crossAxis),_=o.reference[g]+o.reference[A]+(C?0:((k=s.offset)==null?void 0:k[g])||0)-(C?x.crossAxis:0);y<N?y=N:y>_&&(y=_)}return{[m]:b,[g]:y}}}},SX=function(e){return e===void 0&&(e={}),{name:"size",options:e,async fn(t){var n,r;const{placement:a,rects:o,platform:s,elements:u}=t,{apply:c=()=>{},...d}=Ga(e,t),p=await ku(t,d),g=Ha(a),m=$s(a),b=Oo(a)==="y",{width:y,height:S}=o.floating;let x,R;g==="top"||g==="bottom"?(x=g,R=m===(await(s.isRTL==null?void 0:s.isRTL(u.floating))?"start":"end")?"left":"right"):(R=g,x=m==="end"?"top":"bottom");const k=S-p.top-p.bottom,A=y-p.left-p.right,C=No(S-p[x],k),N=No(y-p[R],A),_=!t.middlewareData.shift;let O=C,F=N;if((n=t.middlewareData.shift)!=null&&n.enabled.x&&(F=A),(r=t.middlewareData.shift)!=null&&r.enabled.y&&(O=k),_&&!m){const I=er(p.left,0),H=er(p.right,0),$=er(p.top,0),U=er(p.bottom,0);b?F=y-2*(I!==0||H!==0?I+H:er(p.left,p.right)):O=S-2*($!==0||U!==0?$+U:er(p.top,p.bottom))}await c({...t,availableWidth:F,availableHeight:O});const L=await s.getDimensions(u.floating);return y!==L.width||S!==L.height?{reset:{rects:!0}}:{}}}};function ip(){return typeof window<"u"}function qs(e){return e3(e)?(e.nodeName||"").toLowerCase():"#document"}function nr(e){var t;return(e==null||(t=e.ownerDocument)==null?void 0:t.defaultView)||window}function sa(e){var t;return(t=(e3(e)?e.ownerDocument:e.document)||window.document)==null?void 0:t.documentElement}function e3(e){return ip()?e instanceof Node||e instanceof nr(e).Node:!1}function Ur(e){return ip()?e instanceof Element||e instanceof nr(e).Element:!1}function oa(e){return ip()?e instanceof HTMLElement||e instanceof nr(e).HTMLElement:!1}function ZC(e){return!ip()||typeof ShadowRoot>"u"?!1:e instanceof ShadowRoot||e instanceof nr(e).ShadowRoot}function Vu(e){const{overflow:t,overflowX:n,overflowY:r,display:a}=jr(e);return/auto|scroll|overlay|hidden|clip/.test(t+r+n)&&!["inline","contents"].includes(a)}function EX(e){return["table","td","th"].includes(qs(e))}function sp(e){return[":popover-open",":modal"].some(t=>{try{return e.matches(t)}catch{return!1}})}function K0(e){const t=X0(),n=Ur(e)?jr(e):e;return["transform","translate","scale","rotate","perspective"].some(r=>n[r]?n[r]!=="none":!1)||(n.containerType?n.containerType!=="normal":!1)||!t&&(n.backdropFilter?n.backdropFilter!=="none":!1)||!t&&(n.filter?n.filter!=="none":!1)||["transform","translate","scale","rotate","perspective","filter"].some(r=>(n.willChange||"").includes(r))||["paint","layout","strict","content"].some(r=>(n.contain||"").includes(r))}function wX(e){let t=Do(e);for(;oa(t)&&!Ns(t);){if(K0(t))return t;if(sp(t))return null;t=Do(t)}return null}function X0(){return typeof CSS>"u"||!CSS.supports?!1:CSS.supports("-webkit-backdrop-filter","none")}function Ns(e){return["html","body","#document"].includes(qs(e))}function jr(e){return nr(e).getComputedStyle(e)}function lp(e){return Ur(e)?{scrollLeft:e.scrollLeft,scrollTop:e.scrollTop}:{scrollLeft:e.scrollX,scrollTop:e.scrollY}}function Do(e){if(qs(e)==="html")return e;const t=e.assignedSlot||e.parentNode||ZC(e)&&e.host||sa(e);return ZC(t)?t.host:t}function t3(e){const t=Do(e);return Ns(t)?e.ownerDocument?e.ownerDocument.body:e.body:oa(t)&&Vu(t)?t:t3(t)}function Tu(e,t,n){var r;t===void 0&&(t=[]),n===void 0&&(n=!0);const a=t3(e),o=a===((r=e.ownerDocument)==null?void 0:r.body),s=nr(a);if(o){const u=xk(s);return t.concat(s,s.visualViewport||[],Vu(a)?a:[],u&&n?Tu(u):[])}return t.concat(a,Tu(a,[],n))}function xk(e){return e.parent&&Object.getPrototypeOf(e.parent)?e.frameElement:null}function n3(e){const t=jr(e);let n=parseFloat(t.width)||0,r=parseFloat(t.height)||0;const a=oa(e),o=a?e.offsetWidth:n,s=a?e.offsetHeight:r,u=mf(n)!==o||mf(r)!==s;return u&&(n=o,r=s),{width:n,height:r,$:u}}function Z0(e){return Ur(e)?e:e.contextElement}function As(e){const t=Z0(e);if(!oa(t))return aa(1);const n=t.getBoundingClientRect(),{width:r,height:a,$:o}=n3(t);let s=(o?mf(n.width):n.width)/r,u=(o?mf(n.height):n.height)/a;return(!s||!Number.isFinite(s))&&(s=1),(!u||!Number.isFinite(u))&&(u=1),{x:s,y:u}}const xX=aa(0);function r3(e){const t=nr(e);return!X0()||!t.visualViewport?xX:{x:t.visualViewport.offsetLeft,y:t.visualViewport.offsetTop}}function kX(e,t,n){return t===void 0&&(t=!1),!n||t&&n!==nr(e)?!1:t}function mi(e,t,n,r){t===void 0&&(t=!1),n===void 0&&(n=!1);const a=e.getBoundingClientRect(),o=Z0(e);let s=aa(1);t&&(r?Ur(r)&&(s=As(r)):s=As(e));const u=kX(o,n,r)?r3(o):aa(0);let c=(a.left+u.x)/s.x,d=(a.top+u.y)/s.y,p=a.width/s.x,g=a.height/s.y;if(o){const m=nr(o),b=r&&Ur(r)?nr(r):r;let y=m,S=xk(y);for(;S&&r&&b!==y;){const x=As(S),R=S.getBoundingClientRect(),k=jr(S),A=R.left+(S.clientLeft+parseFloat(k.paddingLeft))*x.x,C=R.top+(S.clientTop+parseFloat(k.paddingTop))*x.y;c*=x.x,d*=x.y,p*=x.x,g*=x.y,c+=A,d+=C,y=nr(S),S=xk(y)}}return yf({width:p,height:g,x:c,y:d})}function Q0(e,t){const n=lp(e).scrollLeft;return t?t.left+n:mi(sa(e)).left+n}function a3(e,t,n){n===void 0&&(n=!1);const r=e.getBoundingClientRect(),a=r.left+t.scrollLeft-(n?0:Q0(e,r)),o=r.top+t.scrollTop;return{x:a,y:o}}function TX(e){let{elements:t,rect:n,offsetParent:r,strategy:a}=e;const o=a==="fixed",s=sa(r),u=t?sp(t.floating):!1;if(r===s||u&&o)return n;let c={scrollLeft:0,scrollTop:0},d=aa(1);const p=aa(0),g=oa(r);if((g||!g&&!o)&&((qs(r)!=="body"||Vu(s))&&(c=lp(r)),oa(r))){const b=mi(r);d=As(r),p.x=b.x+r.clientLeft,p.y=b.y+r.clientTop}const m=s&&!g&&!o?a3(s,c,!0):aa(0);return{width:n.width*d.x,height:n.height*d.y,x:n.x*d.x-c.scrollLeft*d.x+p.x+m.x,y:n.y*d.y-c.scrollTop*d.y+p.y+m.y}}function AX(e){return Array.from(e.getClientRects())}function RX(e){const t=sa(e),n=lp(e),r=e.ownerDocument.body,a=er(t.scrollWidth,t.clientWidth,r.scrollWidth,r.clientWidth),o=er(t.scrollHeight,t.clientHeight,r.scrollHeight,r.clientHeight);let s=-n.scrollLeft+Q0(e);const u=-n.scrollTop;return jr(r).direction==="rtl"&&(s+=er(t.clientWidth,r.clientWidth)-a),{width:a,height:o,x:s,y:u}}function _X(e,t){const n=nr(e),r=sa(e),a=n.visualViewport;let o=r.clientWidth,s=r.clientHeight,u=0,c=0;if(a){o=a.width,s=a.height;const d=X0();(!d||d&&t==="fixed")&&(u=a.offsetLeft,c=a.offsetTop)}return{width:o,height:s,x:u,y:c}}function CX(e,t){const n=mi(e,!0,t==="fixed"),r=n.top+e.clientTop,a=n.left+e.clientLeft,o=oa(e)?As(e):aa(1),s=e.clientWidth*o.x,u=e.clientHeight*o.y,c=a*o.x,d=r*o.y;return{width:s,height:u,x:c,y:d}}function QC(e,t,n){let r;if(t==="viewport")r=_X(e,n);else if(t==="document")r=RX(sa(e));else if(Ur(t))r=CX(t,n);else{const a=r3(e);r={x:t.x-a.x,y:t.y-a.y,width:t.width,height:t.height}}return yf(r)}function o3(e,t){const n=Do(e);return n===t||!Ur(n)||Ns(n)?!1:jr(n).position==="fixed"||o3(n,t)}function NX(e,t){const n=t.get(e);if(n)return n;let r=Tu(e,[],!1).filter(u=>Ur(u)&&qs(u)!=="body"),a=null;const o=jr(e).position==="fixed";let s=o?Do(e):e;for(;Ur(s)&&!Ns(s);){const u=jr(s),c=K0(s);!c&&u.position==="fixed"&&(a=null),(o?!c&&!a:!c&&u.position==="static"&&!!a&&["absolute","fixed"].includes(a.position)||Vu(s)&&!c&&o3(e,s))?r=r.filter(p=>p!==s):a=u,s=Do(s)}return t.set(e,r),r}function OX(e){let{element:t,boundary:n,rootBoundary:r,strategy:a}=e;const s=[...n==="clippingAncestors"?sp(t)?[]:NX(t,this._c):[].concat(n),r],u=s[0],c=s.reduce((d,p)=>{const g=QC(t,p,a);return d.top=er(g.top,d.top),d.right=No(g.right,d.right),d.bottom=No(g.bottom,d.bottom),d.left=er(g.left,d.left),d},QC(t,u,a));return{width:c.right-c.left,height:c.bottom-c.top,x:c.left,y:c.top}}function DX(e){const{width:t,height:n}=n3(e);return{width:t,height:n}}function IX(e,t,n){const r=oa(t),a=sa(t),o=n==="fixed",s=mi(e,!0,o,t);let u={scrollLeft:0,scrollTop:0};const c=aa(0);if(r||!r&&!o)if((qs(t)!=="body"||Vu(a))&&(u=lp(t)),r){const m=mi(t,!0,o,t);c.x=m.x+t.clientLeft,c.y=m.y+t.clientTop}else a&&(c.x=Q0(a));const d=a&&!r&&!o?a3(a,u):aa(0),p=s.left+u.scrollLeft-c.x-d.x,g=s.top+u.scrollTop-c.y-d.y;return{x:p,y:g,width:s.width,height:s.height}}function cm(e){return jr(e).position==="static"}function JC(e,t){if(!oa(e)||jr(e).position==="fixed")return null;if(t)return t(e);let n=e.offsetParent;return sa(e)===n&&(n=n.ownerDocument.body),n}function i3(e,t){const n=nr(e);if(sp(e))return n;if(!oa(e)){let a=Do(e);for(;a&&!Ns(a);){if(Ur(a)&&!cm(a))return a;a=Do(a)}return n}let r=JC(e,t);for(;r&&EX(r)&&cm(r);)r=JC(r,t);return r&&Ns(r)&&cm(r)&&!K0(r)?n:r||wX(e)||n}const LX=async function(e){const t=this.getOffsetParent||i3,n=this.getDimensions,r=await n(e.floating);return{reference:IX(e.reference,await t(e.floating),e.strategy),floating:{x:0,y:0,width:r.width,height:r.height}}};function MX(e){return jr(e).direction==="rtl"}const FX={convertOffsetParentRelativeRectToViewportRelativeRect:TX,getDocumentElement:sa,getClippingRect:OX,getOffsetParent:i3,getElementRects:LX,getClientRects:AX,getDimensions:DX,getScale:As,isElement:Ur,isRTL:MX};function s3(e,t){return e.x===t.x&&e.y===t.y&&e.width===t.width&&e.height===t.height}function PX(e,t){let n=null,r;const a=sa(e);function o(){var u;clearTimeout(r),(u=n)==null||u.disconnect(),n=null}function s(u,c){u===void 0&&(u=!1),c===void 0&&(c=1),o();const d=e.getBoundingClientRect(),{left:p,top:g,width:m,height:b}=d;if(u||t(),!m||!b)return;const y=xd(g),S=xd(a.clientWidth-(p+m)),x=xd(a.clientHeight-(g+b)),R=xd(p),A={rootMargin:-y+"px "+-S+"px "+-x+"px "+-R+"px",threshold:er(0,No(1,c))||1};let C=!0;function N(_){const O=_[0].intersectionRatio;if(O!==c){if(!C)return s();O?s(!1,O):r=setTimeout(()=>{s(!1,1e-7)},1e3)}O===1&&!s3(d,e.getBoundingClientRect())&&s(),C=!1}try{n=new IntersectionObserver(N,{...A,root:a.ownerDocument})}catch{n=new IntersectionObserver(N,A)}n.observe(e)}return s(!0),o}function zX(e,t,n,r){r===void 0&&(r={});const{ancestorScroll:a=!0,ancestorResize:o=!0,elementResize:s=typeof ResizeObserver=="function",layoutShift:u=typeof IntersectionObserver=="function",animationFrame:c=!1}=r,d=Z0(e),p=a||o?[...d?Tu(d):[],...Tu(t)]:[];p.forEach(R=>{a&&R.addEventListener("scroll",n,{passive:!0}),o&&R.addEventListener("resize",n)});const g=d&&u?PX(d,n):null;let m=-1,b=null;s&&(b=new ResizeObserver(R=>{let[k]=R;k&&k.target===d&&b&&(b.unobserve(t),cancelAnimationFrame(m),m=requestAnimationFrame(()=>{var A;(A=b)==null||A.observe(t)})),n()}),d&&!c&&b.observe(d),b.observe(t));let y,S=c?mi(e):null;c&&x();function x(){const R=mi(e);S&&!s3(S,R)&&n(),S=R,y=requestAnimationFrame(x)}return n(),()=>{var R;p.forEach(k=>{a&&k.removeEventListener("scroll",n),o&&k.removeEventListener("resize",n)}),g==null||g(),(R=b)==null||R.disconnect(),b=null,c&&cancelAnimationFrame(y)}}const BX=bX,UX=yX,jX=gX,GX=SX,HX=hX,eN=pX,$X=vX,qX=(e,t,n)=>{const r=new Map,a={platform:FX,...n},o={...a.platform,_c:r};return fX(e,t,{...a,platform:o})};var Qd=typeof document<"u"?E.useLayoutEffect:E.useEffect;function vf(e,t){if(e===t)return!0;if(typeof e!=typeof t)return!1;if(typeof e=="function"&&e.toString()===t.toString())return!0;let n,r,a;if(e&&t&&typeof e=="object"){if(Array.isArray(e)){if(n=e.length,n!==t.length)return!1;for(r=n;r--!==0;)if(!vf(e[r],t[r]))return!1;return!0}if(a=Object.keys(e),n=a.length,n!==Object.keys(t).length)return!1;for(r=n;r--!==0;)if(!{}.hasOwnProperty.call(t,a[r]))return!1;for(r=n;r--!==0;){const o=a[r];if(!(o==="_owner"&&e.$$typeof)&&!vf(e[o],t[o]))return!1}return!0}return e!==e&&t!==t}function l3(e){return typeof window>"u"?1:(e.ownerDocument.defaultView||window).devicePixelRatio||1}function tN(e,t){const n=l3(e);return Math.round(t*n)/n}function dm(e){const t=E.useRef(e);return Qd(()=>{t.current=e}),t}function VX(e){e===void 0&&(e={});const{placement:t="bottom",strategy:n="absolute",middleware:r=[],platform:a,elements:{reference:o,floating:s}={},transform:u=!0,whileElementsMounted:c,open:d}=e,[p,g]=E.useState({x:0,y:0,strategy:n,placement:t,middlewareData:{},isPositioned:!1}),[m,b]=E.useState(r);vf(m,r)||b(r);const[y,S]=E.useState(null),[x,R]=E.useState(null),k=E.useCallback(z=>{z!==_.current&&(_.current=z,S(z))},[]),A=E.useCallback(z=>{z!==O.current&&(O.current=z,R(z))},[]),C=o||y,N=s||x,_=E.useRef(null),O=E.useRef(null),F=E.useRef(p),L=c!=null,I=dm(c),H=dm(a),$=dm(d),U=E.useCallback(()=>{if(!_.current||!O.current)return;const z={placement:t,strategy:n,middleware:m};H.current&&(z.platform=H.current),qX(_.current,O.current,z).then(K=>{const D={...K,isPositioned:$.current!==!1};Y.current&&!vf(F.current,D)&&(F.current=D,Gu.flushSync(()=>{g(D)}))})},[m,t,n,H,$]);Qd(()=>{d===!1&&F.current.isPositioned&&(F.current.isPositioned=!1,g(z=>({...z,isPositioned:!1})))},[d]);const Y=E.useRef(!1);Qd(()=>(Y.current=!0,()=>{Y.current=!1}),[]),Qd(()=>{if(C&&(_.current=C),N&&(O.current=N),C&&N){if(I.current)return I.current(C,N,U);U()}},[C,N,U,I,L]);const Z=E.useMemo(()=>({reference:_,floating:O,setReference:k,setFloating:A}),[k,A]),j=E.useMemo(()=>({reference:C,floating:N}),[C,N]),G=E.useMemo(()=>{const z={position:n,left:0,top:0};if(!j.floating)return z;const K=tN(j.floating,p.x),D=tN(j.floating,p.y);return u?{...z,transform:"translate("+K+"px, "+D+"px)",...l3(j.floating)>=1.5&&{willChange:"transform"}}:{position:n,left:K,top:D}},[n,u,j.floating,p.x,p.y]);return E.useMemo(()=>({...p,update:U,refs:Z,elements:j,floatingStyles:G}),[p,U,Z,j,G])}const WX=e=>{function t(n){return{}.hasOwnProperty.call(n,"current")}return{name:"arrow",options:e,fn(n){const{element:r,padding:a}=typeof e=="function"?e(n):e;return r&&t(r)?r.current!=null?eN({element:r.current,padding:a}).fn(n):{}:r?eN({element:r,padding:a}).fn(n):{}}}},YX=(e,t)=>({...BX(e),options:[e,t]}),KX=(e,t)=>({...UX(e),options:[e,t]}),XX=(e,t)=>({...$X(e),options:[e,t]}),ZX=(e,t)=>({...jX(e),options:[e,t]}),QX=(e,t)=>({...GX(e),options:[e,t]}),JX=(e,t)=>({...HX(e),options:[e,t]}),eZ=(e,t)=>({...WX(e),options:[e,t]});var tZ="Arrow",u3=E.forwardRef((e,t)=>{const{children:n,width:r=10,height:a=5,...o}=e;return w.jsx(Ze.svg,{...o,ref:t,width:r,height:a,viewBox:"0 0 30 10",preserveAspectRatio:"none",children:e.asChild?n:w.jsx("polygon",{points:"0,0 30,0 15,10"})})});u3.displayName=tZ;var nZ=u3;function c3(e){const[t,n]=E.useState(void 0);return An(()=>{if(e){n({width:e.offsetWidth,height:e.offsetHeight});const r=new ResizeObserver(a=>{if(!Array.isArray(a)||!a.length)return;const o=a[0];let s,u;if("borderBoxSize"in o){const c=o.borderBoxSize,d=Array.isArray(c)?c[0]:c;s=d.inlineSize,u=d.blockSize}else s=e.offsetWidth,u=e.offsetHeight;n({width:s,height:u})});return r.observe(e,{box:"border-box"}),()=>r.unobserve(e)}else n(void 0)},[e]),t}var J0="Popper",[d3,Vs]=wr(J0),[rZ,f3]=d3(J0),p3=e=>{const{__scopePopper:t,children:n}=e,[r,a]=E.useState(null);return w.jsx(rZ,{scope:t,anchor:r,onAnchorChange:a,children:n})};p3.displayName=J0;var g3="PopperAnchor",h3=E.forwardRef((e,t)=>{const{__scopePopper:n,virtualRef:r,...a}=e,o=f3(g3,n),s=E.useRef(null),u=gt(t,s);return E.useEffect(()=>{o.onAnchorChange((r==null?void 0:r.current)||s.current)}),r?null:w.jsx(Ze.div,{...a,ref:u})});h3.displayName=g3;var eT="PopperContent",[aZ,oZ]=d3(eT),m3=E.forwardRef((e,t)=>{var re,ie,Se,ae,ve,xe;const{__scopePopper:n,side:r="bottom",sideOffset:a=0,align:o="center",alignOffset:s=0,arrowPadding:u=0,avoidCollisions:c=!0,collisionBoundary:d=[],collisionPadding:p=0,sticky:g="partial",hideWhenDetached:m=!1,updatePositionStrategy:b="optimized",onPlaced:y,...S}=e,x=f3(eT,n),[R,k]=E.useState(null),A=gt(t,Ie=>k(Ie)),[C,N]=E.useState(null),_=c3(C),O=(_==null?void 0:_.width)??0,F=(_==null?void 0:_.height)??0,L=r+(o!=="center"?"-"+o:""),I=typeof p=="number"?p:{top:0,right:0,bottom:0,left:0,...p},H=Array.isArray(d)?d:[d],$=H.length>0,U={padding:I,boundary:H.filter(sZ),altBoundary:$},{refs:Y,floatingStyles:Z,placement:j,isPositioned:G,middlewareData:z}=VX({strategy:"fixed",placement:L,whileElementsMounted:(...Ie)=>zX(...Ie,{animationFrame:b==="always"}),elements:{reference:x.anchor},middleware:[YX({mainAxis:a+F,alignmentAxis:s}),c&&KX({mainAxis:!0,crossAxis:!1,limiter:g==="partial"?XX():void 0,...U}),c&&ZX({...U}),QX({...U,apply:({elements:Ie,rects:Ce,availableWidth:ke,availableHeight:J})=>{const{width:fe,height:Te}=Ce.reference,me=Ie.floating.style;me.setProperty("--radix-popper-available-width",`${ke}px`),me.setProperty("--radix-popper-available-height",`${J}px`),me.setProperty("--radix-popper-anchor-width",`${fe}px`),me.setProperty("--radix-popper-anchor-height",`${Te}px`)}}),C&&eZ({element:C,padding:u}),lZ({arrowWidth:O,arrowHeight:F}),m&&JX({strategy:"referenceHidden",...U})]}),[K,D]=v3(j),V=bn(y);An(()=>{G&&(V==null||V())},[G,V]);const B=(re=z.arrow)==null?void 0:re.x,M=(ie=z.arrow)==null?void 0:ie.y,W=((Se=z.arrow)==null?void 0:Se.centerOffset)!==0,[Q,oe]=E.useState();return An(()=>{R&&oe(window.getComputedStyle(R).zIndex)},[R]),w.jsx("div",{ref:Y.setFloating,"data-radix-popper-content-wrapper":"",style:{...Z,transform:G?Z.transform:"translate(0, -200%)",minWidth:"max-content",zIndex:Q,"--radix-popper-transform-origin":[(ae=z.transformOrigin)==null?void 0:ae.x,(ve=z.transformOrigin)==null?void 0:ve.y].join(" "),...((xe=z.hide)==null?void 0:xe.referenceHidden)&&{visibility:"hidden",pointerEvents:"none"}},dir:e.dir,children:w.jsx(aZ,{scope:n,placedSide:K,onArrowChange:N,arrowX:B,arrowY:M,shouldHideArrow:W,children:w.jsx(Ze.div,{"data-side":K,"data-align":D,...S,ref:A,style:{...S.style,animation:G?void 0:"none"}})})})});m3.displayName=eT;var b3="PopperArrow",iZ={top:"bottom",right:"left",bottom:"top",left:"right"},y3=E.forwardRef(function(t,n){const{__scopePopper:r,...a}=t,o=oZ(b3,r),s=iZ[o.placedSide];return w.jsx("span",{ref:o.onArrowChange,style:{position:"absolute",left:o.arrowX,top:o.arrowY,[s]:0,transformOrigin:{top:"",right:"0 0",bottom:"center 0",left:"100% 0"}[o.placedSide],transform:{top:"translateY(100%)",right:"translateY(50%) rotate(90deg) translateX(-50%)",bottom:"rotate(180deg)",left:"translateY(50%) rotate(-90deg) translateX(50%)"}[o.placedSide],visibility:o.shouldHideArrow?"hidden":void 0},children:w.jsx(nZ,{...a,ref:n,style:{...a.style,display:"block"}})})});y3.displayName=b3;function sZ(e){return e!==null}var lZ=e=>({name:"transformOrigin",options:e,fn(t){var x,R,k;const{placement:n,rects:r,middlewareData:a}=t,s=((x=a.arrow)==null?void 0:x.centerOffset)!==0,u=s?0:e.arrowWidth,c=s?0:e.arrowHeight,[d,p]=v3(n),g={start:"0%",center:"50%",end:"100%"}[p],m=(((R=a.arrow)==null?void 0:R.x)??0)+u/2,b=(((k=a.arrow)==null?void 0:k.y)??0)+c/2;let y="",S="";return d==="bottom"?(y=s?g:`${m}px`,S=`${-c}px`):d==="top"?(y=s?g:`${m}px`,S=`${r.floating.height+c}px`):d==="right"?(y=`${-c}px`,S=s?g:`${b}px`):d==="left"&&(y=`${r.floating.width+c}px`,S=s?g:`${b}px`),{data:{x:y,y:S}}}});function v3(e){const[t,n="center"]=e.split("-");return[t,n]}var tT=p3,up=h3,nT=m3,rT=y3,uZ="VisuallyHidden",aT=E.forwardRef((e,t)=>w.jsx(Ze.span,{...e,ref:t,style:{position:"absolute",border:0,width:1,height:1,padding:0,margin:-1,overflow:"hidden",clip:"rect(0, 0, 0, 0)",whiteSpace:"nowrap",wordWrap:"normal",...e.style}}));aT.displayName=uZ;var cZ=aT,[cp,Nxe]=wr("Tooltip",[Vs]),dp=Vs(),S3="TooltipProvider",dZ=700,kk="tooltip.open",[fZ,oT]=cp(S3),E3=e=>{const{__scopeTooltip:t,delayDuration:n=dZ,skipDelayDuration:r=300,disableHoverableContent:a=!1,children:o}=e,[s,u]=E.useState(!0),c=E.useRef(!1),d=E.useRef(0);return E.useEffect(()=>{const p=d.current;return()=>window.clearTimeout(p)},[]),w.jsx(fZ,{scope:t,isOpenDelayed:s,delayDuration:n,onOpen:E.useCallback(()=>{window.clearTimeout(d.current),u(!1)},[]),onClose:E.useCallback(()=>{window.clearTimeout(d.current),d.current=window.setTimeout(()=>u(!0),r)},[r]),isPointerInTransitRef:c,onPointerInTransitChange:E.useCallback(p=>{c.current=p},[]),disableHoverableContent:a,children:o})};E3.displayName=S3;var fp="Tooltip",[pZ,pp]=cp(fp),w3=e=>{const{__scopeTooltip:t,children:n,open:r,defaultOpen:a=!1,onOpenChange:o,disableHoverableContent:s,delayDuration:u}=e,c=oT(fp,e.__scopeTooltip),d=dp(t),[p,g]=E.useState(null),m=Tn(),b=E.useRef(0),y=s??c.disableHoverableContent,S=u??c.delayDuration,x=E.useRef(!1),[R=!1,k]=ja({prop:r,defaultProp:a,onChange:O=>{O?(c.onOpen(),document.dispatchEvent(new CustomEvent(kk))):c.onClose(),o==null||o(O)}}),A=E.useMemo(()=>R?x.current?"delayed-open":"instant-open":"closed",[R]),C=E.useCallback(()=>{window.clearTimeout(b.current),b.current=0,x.current=!1,k(!0)},[k]),N=E.useCallback(()=>{window.clearTimeout(b.current),b.current=0,k(!1)},[k]),_=E.useCallback(()=>{window.clearTimeout(b.current),b.current=window.setTimeout(()=>{x.current=!0,k(!0),b.current=0},S)},[S,k]);return E.useEffect(()=>()=>{b.current&&(window.clearTimeout(b.current),b.current=0)},[]),w.jsx(tT,{...d,children:w.jsx(pZ,{scope:t,contentId:m,open:R,stateAttribute:A,trigger:p,onTriggerChange:g,onTriggerEnter:E.useCallback(()=>{c.isOpenDelayed?_():C()},[c.isOpenDelayed,_,C]),onTriggerLeave:E.useCallback(()=>{y?N():(window.clearTimeout(b.current),b.current=0)},[N,y]),onOpen:C,onClose:N,disableHoverableContent:y,children:n})})};w3.displayName=fp;var Tk="TooltipTrigger",x3=E.forwardRef((e,t)=>{const{__scopeTooltip:n,...r}=e,a=pp(Tk,n),o=oT(Tk,n),s=dp(n),u=E.useRef(null),c=gt(t,u,a.onTriggerChange),d=E.useRef(!1),p=E.useRef(!1),g=E.useCallback(()=>d.current=!1,[]);return E.useEffect(()=>()=>document.removeEventListener("pointerup",g),[g]),w.jsx(up,{asChild:!0,...s,children:w.jsx(Ze.button,{"aria-describedby":a.open?a.contentId:void 0,"data-state":a.stateAttribute,...r,ref:c,onPointerMove:Ke(e.onPointerMove,m=>{m.pointerType!=="touch"&&!p.current&&!o.isPointerInTransitRef.current&&(a.onTriggerEnter(),p.current=!0)}),onPointerLeave:Ke(e.onPointerLeave,()=>{a.onTriggerLeave(),p.current=!1}),onPointerDown:Ke(e.onPointerDown,()=>{d.current=!0,document.addEventListener("pointerup",g,{once:!0})}),onFocus:Ke(e.onFocus,()=>{d.current||a.onOpen()}),onBlur:Ke(e.onBlur,a.onClose),onClick:Ke(e.onClick,a.onClose)})})});x3.displayName=Tk;var gZ="TooltipPortal",[Oxe,hZ]=cp(gZ,{forceMount:void 0}),Os="TooltipContent",k3=E.forwardRef((e,t)=>{const n=hZ(Os,e.__scopeTooltip),{forceMount:r=n.forceMount,side:a="top",...o}=e,s=pp(Os,e.__scopeTooltip);return w.jsx(xr,{present:r||s.open,children:s.disableHoverableContent?w.jsx(T3,{side:a,...o,ref:t}):w.jsx(mZ,{side:a,...o,ref:t})})}),mZ=E.forwardRef((e,t)=>{const n=pp(Os,e.__scopeTooltip),r=oT(Os,e.__scopeTooltip),a=E.useRef(null),o=gt(t,a),[s,u]=E.useState(null),{trigger:c,onClose:d}=n,p=a.current,{onPointerInTransitChange:g}=r,m=E.useCallback(()=>{u(null),g(!1)},[g]),b=E.useCallback((y,S)=>{const x=y.currentTarget,R={x:y.clientX,y:y.clientY},k=SZ(R,x.getBoundingClientRect()),A=EZ(R,k),C=wZ(S.getBoundingClientRect()),N=kZ([...A,...C]);u(N),g(!0)},[g]);return E.useEffect(()=>()=>m(),[m]),E.useEffect(()=>{if(c&&p){const y=x=>b(x,p),S=x=>b(x,c);return c.addEventListener("pointerleave",y),p.addEventListener("pointerleave",S),()=>{c.removeEventListener("pointerleave",y),p.removeEventListener("pointerleave",S)}}},[c,p,b,m]),E.useEffect(()=>{if(s){const y=S=>{const x=S.target,R={x:S.clientX,y:S.clientY},k=(c==null?void 0:c.contains(x))||(p==null?void 0:p.contains(x)),A=!xZ(R,s);k?m():A&&(m(),d())};return document.addEventListener("pointermove",y),()=>document.removeEventListener("pointermove",y)}},[c,p,s,d,m]),w.jsx(T3,{...e,ref:o})}),[bZ,yZ]=cp(fp,{isInside:!1}),T3=E.forwardRef((e,t)=>{const{__scopeTooltip:n,children:r,"aria-label":a,onEscapeKeyDown:o,onPointerDownOutside:s,...u}=e,c=pp(Os,n),d=dp(n),{onClose:p}=c;return E.useEffect(()=>(document.addEventListener(kk,p),()=>document.removeEventListener(kk,p)),[p]),E.useEffect(()=>{if(c.trigger){const g=m=>{const b=m.target;b!=null&&b.contains(c.trigger)&&p()};return window.addEventListener("scroll",g,{capture:!0}),()=>window.removeEventListener("scroll",g,{capture:!0})}},[c.trigger,p]),w.jsx(qu,{asChild:!0,disableOutsidePointerEvents:!1,onEscapeKeyDown:o,onPointerDownOutside:s,onFocusOutside:g=>g.preventDefault(),onDismiss:p,children:w.jsxs(nT,{"data-state":c.stateAttribute,...d,...u,ref:t,style:{...u.style,"--radix-tooltip-content-transform-origin":"var(--radix-popper-transform-origin)","--radix-tooltip-content-available-width":"var(--radix-popper-available-width)","--radix-tooltip-content-available-height":"var(--radix-popper-available-height)","--radix-tooltip-trigger-width":"var(--radix-popper-anchor-width)","--radix-tooltip-trigger-height":"var(--radix-popper-anchor-height)"},children:[w.jsx(I0,{children:r}),w.jsx(bZ,{scope:n,isInside:!0,children:w.jsx(cZ,{id:c.contentId,role:"tooltip",children:a||r})})]})})});k3.displayName=Os;var A3="TooltipArrow",vZ=E.forwardRef((e,t)=>{const{__scopeTooltip:n,...r}=e,a=dp(n);return yZ(A3,n).isInside?null:w.jsx(rT,{...a,...r,ref:t})});vZ.displayName=A3;function SZ(e,t){const n=Math.abs(t.top-e.y),r=Math.abs(t.bottom-e.y),a=Math.abs(t.right-e.x),o=Math.abs(t.left-e.x);switch(Math.min(n,r,a,o)){case o:return"left";case a:return"right";case n:return"top";case r:return"bottom";default:throw new Error("unreachable")}}function EZ(e,t,n=5){const r=[];switch(t){case"top":r.push({x:e.x-n,y:e.y+n},{x:e.x+n,y:e.y+n});break;case"bottom":r.push({x:e.x-n,y:e.y-n},{x:e.x+n,y:e.y-n});break;case"left":r.push({x:e.x+n,y:e.y-n},{x:e.x+n,y:e.y+n});break;case"right":r.push({x:e.x-n,y:e.y-n},{x:e.x-n,y:e.y+n});break}return r}function wZ(e){const{top:t,right:n,bottom:r,left:a}=e;return[{x:a,y:t},{x:n,y:t},{x:n,y:r},{x:a,y:r}]}function xZ(e,t){const{x:n,y:r}=e;let a=!1;for(let o=0,s=t.length-1;o<t.length;s=o++){const u=t[o].x,c=t[o].y,d=t[s].x,p=t[s].y;c>r!=p>r&&n<(d-u)*(r-c)/(p-c)+u&&(a=!a)}return a}function kZ(e){const t=e.slice();return t.sort((n,r)=>n.x<r.x?-1:n.x>r.x?1:n.y<r.y?-1:n.y>r.y?1:0),TZ(t)}function TZ(e){if(e.length<=1)return e.slice();const t=[];for(let r=0;r<e.length;r++){const a=e[r];for(;t.length>=2;){const o=t[t.length-1],s=t[t.length-2];if((o.x-s.x)*(a.y-s.y)>=(o.y-s.y)*(a.x-s.x))t.pop();else break}t.push(a)}t.pop();const n=[];for(let r=e.length-1;r>=0;r--){const a=e[r];for(;n.length>=2;){const o=n[n.length-1],s=n[n.length-2];if((o.x-s.x)*(a.y-s.y)>=(o.y-s.y)*(a.x-s.x))n.pop();else break}n.push(a)}return n.pop(),t.length===1&&n.length===1&&t[0].x===n[0].x&&t[0].y===n[0].y?t:t.concat(n)}var AZ=E3,RZ=w3,_Z=x3,R3=k3;const _3=AZ,C3=RZ,N3=_Z,CZ=e=>typeof e!="string"?e:w.jsx("div",{className:"relative top-0 pt-1 whitespace-pre-wrap break-words",children:e}),iT=E.forwardRef(({className:e,side:t="left",align:n="start",children:r,...a},o)=>{const s=E.useRef(null);return E.useEffect(()=>{s.current&&(s.current.scrollTop=0)},[r]),w.jsx(R3,{ref:o,side:t,align:n,className:Me("bg-popover text-popover-foreground animate-in fade-in-0 zoom-in-95 data-[state=closed]:animate-out data-[state=closed]:fade-out-0 data-[state=closed]:zoom-out-95 data-[side=bottom]:slide-in-from-top-2 data-[side=left]:slide-in-from-right-2 data-[side=right]:slide-in-from-left-2 data-[side=top]:slide-in-from-bottom-2 max-h-[60vh] overflow-y-auto whitespace-pre-wrap break-words rounded-md border px-3 py-2 text-sm shadow-md z-60",e),...a,children:typeof r=="string"?CZ(r):r})});iT.displayName=R3.displayName;const Sf=YB("inline-flex items-center justify-center gap-2 whitespace-nowrap rounded-md text-sm font-medium ring-offset-background transition-colors focus-visible:outline-none focus-visible:ring-2 focus-visible:ring-ring focus-visible:ring-offset-2 disabled:pointer-events-none disabled:opacity-50 [&_svg]:pointer-events-none [&_svg]:size-4 [&_svg]:shrink-0",{variants:{variant:{default:"bg-primary text-primary-foreground hover:bg-primary/90",destructive:"bg-destructive text-destructive-foreground hover:bg-destructive/90",outline:"border border-input bg-background hover:bg-accent hover:text-accent-foreground",secondary:"bg-secondary text-secondary-foreground hover:bg-secondary/80",ghost:"hover:bg-accent hover:text-accent-foreground",link:"text-primary underline-offset-4 hover:underline"},size:{default:"h-10 px-4 py-2",sm:"h-9 rounded-md px-3",lg:"h-11 rounded-md px-8",icon:"size-8"}},defaultVariants:{variant:"default",size:"default"}}),xt=E.forwardRef(({className:e,variant:t,tooltip:n,size:r,side:a="right",asChild:o=!1,...s},u)=>{const c=o?Co:"button";return n?w.jsx(_3,{children:w.jsxs(C3,{children:[w.jsx(N3,{asChild:!0,children:w.jsx(c,{className:Me(Sf({variant:t,size:r,className:e}),"cursor-pointer"),ref:u,...s})}),w.jsx(iT,{side:a,children:n})]})}):w.jsx(c,{className:Me(Sf({variant:t,size:r,className:e}),"cursor-pointer"),ref:u,...s})});xt.displayName="Button";const NZ=nX,OZ=rX,O3=E.forwardRef(({className:e,...t},n)=>w.jsx(WU,{className:Me("data-[state=open]:animate-in data-[state=closed]:animate-out data-[state=closed]:fade-out-0 data-[state=open]:fade-in-0 fixed inset-0 z-50 bg-black/80",e),...t,ref:n}));O3.displayName=WU.displayName;const D3=E.forwardRef(({className:e,...t},n)=>w.jsxs(OZ,{children:[w.jsx(O3,{}),w.jsx(YU,{ref:n,className:Me("bg-background data-[state=open]:animate-in data-[state=closed]:animate-out data-[state=closed]:fade-out-0 data-[state=open]:fade-in-0 data-[state=closed]:zoom-out-95 data-[state=open]:zoom-in-95 data-[state=closed]:slide-out-to-top-[48%] data-[state=open]:slide-in-from-top-[48%] fixed top-[50%] left-[50%] z-50 grid w-full max-w-lg translate-x-[-50%] translate-y-[-50%] gap-4 border p-6 shadow-lg duration-200 sm:rounded-lg",e),...t})]}));D3.displayName=YU.displayName;const I3=({className:e,...t})=>w.jsx("div",{className:Me("flex flex-col space-y-2 text-center sm:text-left",e),...t});I3.displayName="AlertDialogHeader";const L3=E.forwardRef(({className:e,...t},n)=>w.jsx(ZU,{ref:n,className:Me("text-lg font-semibold",e),...t}));L3.displayName=ZU.displayName;const M3=E.forwardRef(({className:e,...t},n)=>w.jsx(QU,{ref:n,className:Me("text-muted-foreground text-sm",e),...t}));M3.displayName=QU.displayName;const DZ=E.forwardRef(({className:e,...t},n)=>w.jsx(KU,{ref:n,className:Me(Sf(),e),...t}));DZ.displayName=KU.displayName;const IZ=E.forwardRef(({className:e,...t},n)=>w.jsx(XU,{ref:n,className:Me(Sf({variant:"outline"}),"mt-2 sm:mt-0",e),...t}));IZ.displayName=XU.displayName;const $a=E.forwardRef(({className:e,type:t,...n},r)=>w.jsx("input",{type:t,className:Me("border-input file:text-foreground placeholder:text-muted-foreground focus-visible:ring-ring flex h-9 rounded-md border bg-transparent px-3 py-1 text-base shadow-sm transition-colors file:border-0 file:bg-transparent file:text-sm file:font-medium focus-visible:ring-1 focus-visible:outline-none disabled:cursor-not-allowed disabled:opacity-50 md:text-sm [&::-webkit-inner-spin-button]:opacity-100 [&::-webkit-outer-spin-button]:opacity-100",e),ref:r,...n}));$a.displayName="Input";const LZ=()=>{const[e,t]=E.useState(!0),n=Fe.use.apiKey(),[r,a]=E.useState(""),o=Hn.use.message();E.useEffect(()=>{a(n||"")},[n,e]),E.useEffect(()=>{o&&(o.includes(GB)||o.includes(HB))&&t(!0)},[o,t]);const s=E.useCallback(async()=>{if(Fe.setState({apiKey:r||null}),await Hn.getState().check()){t(!1);return}Ft.error("API Key is invalid")},[r]),u=E.useCallback(c=>{a(c.target.value)},[a]);return w.jsx(NZ,{open:e,onOpenChange:t,children:w.jsxs(D3,{children:[w.jsxs(I3,{children:[w.jsx(L3,{children:"API Key is required"}),w.jsx(M3,{children:"Please enter your API key"})]}),w.jsxs("form",{className:"flex gap-2",onSubmit:c=>c.preventDefault(),children:[w.jsx($a,{type:"password",value:r,onChange:u,placeholder:"Enter your API key",className:"max-h-full w-full min-w-0",autoComplete:"off"}),w.jsx(xt,{onClick:s,variant:"outline",size:"sm",children:"Save"})]})]})})};var sT="Popover",[F3,Dxe]=wr(sT,[Vs]),Wu=Vs(),[MZ,ki]=F3(sT),P3=e=>{const{__scopePopover:t,children:n,open:r,defaultOpen:a,onOpenChange:o,modal:s=!1}=e,u=Wu(t),c=E.useRef(null),[d,p]=E.useState(!1),[g=!1,m]=ja({prop:r,defaultProp:a,onChange:o});return w.jsx(tT,{...u,children:w.jsx(MZ,{scope:t,contentId:Tn(),triggerRef:c,open:g,onOpenChange:m,onOpenToggle:E.useCallback(()=>m(b=>!b),[m]),hasCustomAnchor:d,onCustomAnchorAdd:E.useCallback(()=>p(!0),[]),onCustomAnchorRemove:E.useCallback(()=>p(!1),[]),modal:s,children:n})})};P3.displayName=sT;var z3="PopoverAnchor",FZ=E.forwardRef((e,t)=>{const{__scopePopover:n,...r}=e,a=ki(z3,n),o=Wu(n),{onCustomAnchorAdd:s,onCustomAnchorRemove:u}=a;return E.useEffect(()=>(s(),()=>u()),[s,u]),w.jsx(up,{...o,...r,ref:t})});FZ.displayName=z3;var B3="PopoverTrigger",U3=E.forwardRef((e,t)=>{const{__scopePopover:n,...r}=e,a=ki(B3,n),o=Wu(n),s=gt(t,a.triggerRef),u=w.jsx(Ze.button,{type:"button","aria-haspopup":"dialog","aria-expanded":a.open,"aria-controls":a.contentId,"data-state":$3(a.open),...r,ref:s,onClick:Ke(e.onClick,a.onOpenToggle)});return a.hasCustomAnchor?u:w.jsx(up,{asChild:!0,...o,children:u})});U3.displayName=B3;var PZ="PopoverPortal",[Ixe,zZ]=F3(PZ,{forceMount:void 0}),Ds="PopoverContent",j3=E.forwardRef((e,t)=>{const n=zZ(Ds,e.__scopePopover),{forceMount:r=n.forceMount,...a}=e,o=ki(Ds,e.__scopePopover);return w.jsx(xr,{present:r||o.open,children:o.modal?w.jsx(BZ,{...a,ref:t}):w.jsx(UZ,{...a,ref:t})})});j3.displayName=Ds;var BZ=E.forwardRef((e,t)=>{const n=ki(Ds,e.__scopePopover),r=E.useRef(null),a=gt(t,r),o=E.useRef(!1);return E.useEffect(()=>{const s=r.current;if(s)return F0(s)},[]),w.jsx(rp,{as:Co,allowPinchZoom:!0,children:w.jsx(G3,{...e,ref:a,trapFocus:n.open,disableOutsidePointerEvents:!0,onCloseAutoFocus:Ke(e.onCloseAutoFocus,s=>{var u;s.preventDefault(),o.current||(u=n.triggerRef.current)==null||u.focus()}),onPointerDownOutside:Ke(e.onPointerDownOutside,s=>{const u=s.detail.originalEvent,c=u.button===0&&u.ctrlKey===!0,d=u.button===2||c;o.current=d},{checkForDefaultPrevented:!1}),onFocusOutside:Ke(e.onFocusOutside,s=>s.preventDefault(),{checkForDefaultPrevented:!1})})})}),UZ=E.forwardRef((e,t)=>{const n=ki(Ds,e.__scopePopover),r=E.useRef(!1),a=E.useRef(!1);return w.jsx(G3,{...e,ref:t,trapFocus:!1,disableOutsidePointerEvents:!1,onCloseAutoFocus:o=>{var s,u;(s=e.onCloseAutoFocus)==null||s.call(e,o),o.defaultPrevented||(r.current||(u=n.triggerRef.current)==null||u.focus(),o.preventDefault()),r.current=!1,a.current=!1},onInteractOutside:o=>{var c,d;(c=e.onInteractOutside)==null||c.call(e,o),o.defaultPrevented||(r.current=!0,o.detail.originalEvent.type==="pointerdown"&&(a.current=!0));const s=o.target;((d=n.triggerRef.current)==null?void 0:d.contains(s))&&o.preventDefault(),o.detail.originalEvent.type==="focusin"&&a.current&&o.preventDefault()}})}),G3=E.forwardRef((e,t)=>{const{__scopePopover:n,trapFocus:r,onOpenAutoFocus:a,onCloseAutoFocus:o,disableOutsidePointerEvents:s,onEscapeKeyDown:u,onPointerDownOutside:c,onFocusOutside:d,onInteractOutside:p,...g}=e,m=ki(Ds,n),b=Wu(n);return M0(),w.jsx(tp,{asChild:!0,loop:!0,trapped:r,onMountAutoFocus:a,onUnmountAutoFocus:o,children:w.jsx(qu,{asChild:!0,disableOutsidePointerEvents:s,onInteractOutside:p,onEscapeKeyDown:u,onPointerDownOutside:c,onFocusOutside:d,onDismiss:()=>m.onOpenChange(!1),children:w.jsx(nT,{"data-state":$3(m.open),role:"dialog",id:m.contentId,...b,...g,ref:t,style:{...g.style,"--radix-popover-content-transform-origin":"var(--radix-popper-transform-origin)","--radix-popover-content-available-width":"var(--radix-popper-available-width)","--radix-popover-content-available-height":"var(--radix-popper-available-height)","--radix-popover-trigger-width":"var(--radix-popper-anchor-width)","--radix-popover-trigger-height":"var(--radix-popper-anchor-height)"}})})})}),H3="PopoverClose",jZ=E.forwardRef((e,t)=>{const{__scopePopover:n,...r}=e,a=ki(H3,n);return w.jsx(Ze.button,{type:"button",...r,ref:t,onClick:Ke(e.onClick,()=>a.onOpenChange(!1))})});jZ.displayName=H3;var GZ="PopoverArrow",HZ=E.forwardRef((e,t)=>{const{__scopePopover:n,...r}=e,a=Wu(n);return w.jsx(rT,{...a,...r,ref:t})});HZ.displayName=GZ;function $3(e){return e?"open":"closed"}var $Z=P3,qZ=U3,q3=j3;const Yu=$Z,Ku=qZ,Ws=E.forwardRef(({className:e,align:t="center",sideOffset:n=4,...r},a)=>w.jsx(q3,{ref:a,align:t,sideOffset:n,className:Me("bg-popover text-popover-foreground data-[state=open]:animate-in data-[state=closed]:animate-out data-[state=closed]:fade-out-0 data-[state=open]:fade-in-0 data-[state=closed]:zoom-out-95 data-[state=open]:zoom-in-95 data-[side=bottom]:slide-in-from-top-2 data-[side=left]:slide-in-from-right-2 data-[side=right]:slide-in-from-left-2 data-[side=top]:slide-in-from-bottom-2 z-50 rounded-md border p-4 shadow-md outline-none",e),...r}));Ws.displayName=q3.displayName;const VZ=(e,t,n,r)=>{var o,s,u,c;const a=[n,{code:t,...r||{}}];if((s=(o=e==null?void 0:e.services)==null?void 0:o.logger)!=null&&s.forward)return e.services.logger.forward(a,"warn","react-i18next::",!0);pi(a[0])&&(a[0]=`react-i18next:: ${a[0]}`),(c=(u=e==null?void 0:e.services)==null?void 0:u.logger)!=null&&c.warn?e.services.logger.warn(...a):console!=null&&console.warn&&console.warn(...a)},nN={},Ak=(e,t,n,r)=>{pi(n)&&nN[n]||(pi(n)&&(nN[n]=new Date),VZ(e,t,n,r))},V3=(e,t)=>()=>{if(e.isInitialized)t();else{const n=()=>{setTimeout(()=>{e.off("initialized",n)},0),t()};e.on("initialized",n)}},Rk=(e,t,n)=>{e.loadNamespaces(t,V3(e,n))},rN=(e,t,n,r)=>{if(pi(n)&&(n=[n]),e.options.preload&&e.options.preload.indexOf(t)>-1)return Rk(e,n,r);n.forEach(a=>{e.options.ns.indexOf(a)<0&&e.options.ns.push(a)}),e.loadLanguages(t,V3(e,r))},WZ=(e,t,n={})=>!t.languages||!t.languages.length?(Ak(t,"NO_LANGUAGES","i18n.languages were undefined or empty",{languages:t.languages}),!0):t.hasLoadedNamespace(e,{lng:n.lng,precheck:(r,a)=>{var o;if(((o=n.bindI18n)==null?void 0:o.indexOf("languageChanging"))>-1&&r.services.backendConnector.backend&&r.isLanguageChangingTo&&!a(r.isLanguageChangingTo,e))return!1}}),pi=e=>typeof e=="string",YZ=e=>typeof e=="object"&&e!==null,KZ=/&(?:amp|#38|lt|#60|gt|#62|apos|#39|quot|#34|nbsp|#160|copy|#169|reg|#174|hellip|#8230|#x2F|#47);/g,XZ={"&amp;":"&","&#38;":"&","&lt;":"<","&#60;":"<","&gt;":">","&#62;":">","&apos;":"'","&#39;":"'","&quot;":'"',"&#34;":'"',"&nbsp;":" ","&#160;":" ","&copy;":"©","&#169;":"©","&reg;":"®","&#174;":"®","&hellip;":"…","&#8230;":"…","&#x2F;":"/","&#47;":"/"},ZZ=e=>XZ[e],QZ=e=>e.replace(KZ,ZZ);let _k={bindI18n:"languageChanged",bindI18nStore:"",transEmptyNodeValue:"",transSupportBasicHtmlNodes:!0,transWrapTextNodes:"",transKeepBasicHtmlNodesFor:["br","strong","i","p"],useSuspense:!0,unescape:QZ};const JZ=(e={})=>{_k={..._k,...e}},eQ=()=>_k;let W3;const tQ=e=>{W3=e},nQ=()=>W3,rQ={type:"3rdParty",init(e){JZ(e.options.react),tQ(e)}},aQ=E.createContext();class oQ{constructor(){this.usedNamespaces={}}addUsedNamespaces(t){t.forEach(n=>{this.usedNamespaces[n]||(this.usedNamespaces[n]=!0)})}getUsedNamespaces(){return Object.keys(this.usedNamespaces)}}const iQ=(e,t)=>{const n=E.useRef();return E.useEffect(()=>{n.current=e},[e,t]),n.current},Y3=(e,t,n,r)=>e.getFixedT(t,n,r),sQ=(e,t,n,r)=>E.useCallback(Y3(e,t,n,r),[e,t,n,r]),Pt=(e,t={})=>{var C,N,_,O;const{i18n:n}=t,{i18n:r,defaultNS:a}=E.useContext(aQ)||{},o=n||r||nQ();if(o&&!o.reportNamespaces&&(o.reportNamespaces=new oQ),!o){Ak(o,"NO_I18NEXT_INSTANCE","useTranslation: You will need to pass in an i18next instance by using initReactI18next");const F=(I,H)=>pi(H)?H:YZ(H)&&pi(H.defaultValue)?H.defaultValue:Array.isArray(I)?I[I.length-1]:I,L=[F,{},!1];return L.t=F,L.i18n={},L.ready=!1,L}(C=o.options.react)!=null&&C.wait&&Ak(o,"DEPRECATED_OPTION","useTranslation: It seems you are still using the old wait option, you may migrate to the new useSuspense behaviour.");const s={...eQ(),...o.options.react,...t},{useSuspense:u,keyPrefix:c}=s;let d=a||((N=o.options)==null?void 0:N.defaultNS);d=pi(d)?[d]:d||["translation"],(O=(_=o.reportNamespaces).addUsedNamespaces)==null||O.call(_,d);const p=(o.isInitialized||o.initializedStoreOnce)&&d.every(F=>WZ(F,o,s)),g=sQ(o,t.lng||null,s.nsMode==="fallback"?d:d[0],c),m=()=>g,b=()=>Y3(o,t.lng||null,s.nsMode==="fallback"?d:d[0],c),[y,S]=E.useState(m);let x=d.join();t.lng&&(x=`${t.lng}${x}`);const R=iQ(x),k=E.useRef(!0);E.useEffect(()=>{const{bindI18n:F,bindI18nStore:L}=s;k.current=!0,!p&&!u&&(t.lng?rN(o,t.lng,d,()=>{k.current&&S(b)}):Rk(o,d,()=>{k.current&&S(b)})),p&&R&&R!==x&&k.current&&S(b);const I=()=>{k.current&&S(b)};return F&&(o==null||o.on(F,I)),L&&(o==null||o.store.on(L,I)),()=>{k.current=!1,o&&(F==null||F.split(" ").forEach(H=>o.off(H,I))),L&&o&&L.split(" ").forEach(H=>o.store.off(H,I))}},[o,x]),E.useEffect(()=>{k.current&&p&&S(m)},[o,c,p]);const A=[y,o,p];if(A.t=y,A.i18n=o,A.ready=p,p||!p&&!u)return A;throw new Promise(F=>{t.lng?rN(o,t.lng,d,()=>F()):Rk(o,d,()=>F())})},lQ=({status:e})=>{const{t}=Pt();return e?w.jsxs("div",{className:"min-w-[300px] space-y-3 text-sm",children:[w.jsxs("div",{className:"space-y-1",children:[w.jsx("h4",{className:"font-medium",children:t("graphPanel.statusCard.storageInfo")}),w.jsxs("div",{className:"text-muted-foreground grid grid-cols-2 gap-1",children:[w.jsxs("span",{children:[t("graphPanel.statusCard.workingDirectory"),":"]}),w.jsx("span",{className:"truncate",children:e.working_directory}),w.jsxs("span",{children:[t("graphPanel.statusCard.inputDirectory"),":"]}),w.jsx("span",{className:"truncate",children:e.input_directory})]})]}),w.jsxs("div",{className:"space-y-1",children:[w.jsx("h4",{className:"font-medium",children:t("graphPanel.statusCard.llmConfig")}),w.jsxs("div",{className:"text-muted-foreground grid grid-cols-2 gap-1",children:[w.jsxs("span",{children:[t("graphPanel.statusCard.llmBinding"),":"]}),w.jsx("span",{children:e.configuration.llm_binding}),w.jsxs("span",{children:[t("graphPanel.statusCard.llmBindingHost"),":"]}),w.jsx("span",{children:e.configuration.llm_binding_host}),w.jsxs("span",{children:[t("graphPanel.statusCard.llmModel"),":"]}),w.jsx("span",{children:e.configuration.llm_model}),w.jsxs("span",{children:[t("graphPanel.statusCard.maxTokens"),":"]}),w.jsx("span",{children:e.configuration.max_tokens})]})]}),w.jsxs("div",{className:"space-y-1",children:[w.jsx("h4",{className:"font-medium",children:t("graphPanel.statusCard.embeddingConfig")}),w.jsxs("div",{className:"text-muted-foreground grid grid-cols-2 gap-1",children:[w.jsxs("span",{children:[t("graphPanel.statusCard.embeddingBinding"),":"]}),w.jsx("span",{children:e.configuration.embedding_binding}),w.jsxs("span",{children:[t("graphPanel.statusCard.embeddingBindingHost"),":"]}),w.jsx("span",{children:e.configuration.embedding_binding_host}),w.jsxs("span",{children:[t("graphPanel.statusCard.embeddingModel"),":"]}),w.jsx("span",{children:e.configuration.embedding_model})]})]}),w.jsxs("div",{className:"space-y-1",children:[w.jsx("h4",{className:"font-medium",children:t("graphPanel.statusCard.storageConfig")}),w.jsxs("div",{className:"text-muted-foreground grid grid-cols-2 gap-1",children:[w.jsxs("span",{children:[t("graphPanel.statusCard.kvStorage"),":"]}),w.jsx("span",{children:e.configuration.kv_storage}),w.jsxs("span",{children:[t("graphPanel.statusCard.docStatusStorage"),":"]}),w.jsx("span",{children:e.configuration.doc_status_storage}),w.jsxs("span",{children:[t("graphPanel.statusCard.graphStorage"),":"]}),w.jsx("span",{children:e.configuration.graph_storage}),w.jsxs("span",{children:[t("graphPanel.statusCard.vectorStorage"),":"]}),w.jsx("span",{children:e.configuration.vector_storage})]})]})]}):w.jsx("div",{className:"text-muted-foreground text-sm",children:t("graphPanel.statusCard.unavailable")})},uQ=()=>{const{t:e}=Pt(),t=Hn.use.health(),n=Hn.use.lastCheckTime(),r=Hn.use.status(),[a,o]=E.useState(!1);return E.useEffect(()=>{o(!0);const s=setTimeout(()=>o(!1),300);return()=>clearTimeout(s)},[n]),w.jsx("div",{className:"fixed right-4 bottom-4 flex items-center gap-2 opacity-80 select-none",children:w.jsxs(Yu,{children:[w.jsx(Ku,{asChild:!0,children:w.jsxs("div",{className:"flex cursor-help items-center gap-2",children:[w.jsx("div",{className:Me("h-3 w-3 rounded-full transition-all duration-300","shadow-[0_0_8px_rgba(0,0,0,0.2)]",t?"bg-green-500":"bg-red-500",a&&"scale-125",a&&t&&"shadow-[0_0_12px_rgba(34,197,94,0.4)]",a&&!t&&"shadow-[0_0_12px_rgba(239,68,68,0.4)]")}),w.jsx("span",{className:"text-muted-foreground text-xs",children:e(t?"graphPanel.statusIndicator.connected":"graphPanel.statusIndicator.disconnected")})]})}),w.jsx(Ws,{className:"w-auto",side:"top",align:"end",children:w.jsx(lQ,{status:r})})]})})};function Ck(e,[t,n]){return Math.min(n,Math.max(t,e))}function K3(e){const t=e+"CollectionProvider",[n,r]=wr(t),[a,o]=n(t,{collectionRef:{current:null},itemMap:new Map}),s=b=>{const{scope:y,children:S}=b,x=ye.useRef(null),R=ye.useRef(new Map).current;return w.jsx(a,{scope:y,itemMap:R,collectionRef:x,children:S})};s.displayName=t;const u=e+"CollectionSlot",c=ye.forwardRef((b,y)=>{const{scope:S,children:x}=b,R=o(u,S),k=gt(y,R.collectionRef);return w.jsx(Co,{ref:k,children:x})});c.displayName=u;const d=e+"CollectionItemSlot",p="data-radix-collection-item",g=ye.forwardRef((b,y)=>{const{scope:S,children:x,...R}=b,k=ye.useRef(null),A=gt(y,k),C=o(d,S);return ye.useEffect(()=>(C.itemMap.set(k,{ref:k,...R}),()=>void C.itemMap.delete(k))),w.jsx(Co,{[p]:"",ref:A,children:x})});g.displayName=d;function m(b){const y=o(e+"CollectionConsumer",b);return ye.useCallback(()=>{const x=y.collectionRef.current;if(!x)return[];const R=Array.from(x.querySelectorAll(`[${p}]`));return Array.from(y.itemMap.values()).sort((C,N)=>R.indexOf(C.ref.current)-R.indexOf(N.ref.current))},[y.collectionRef,y.itemMap])}return[{Provider:s,Slot:c,ItemSlot:g},m,r]}var cQ=E.createContext(void 0);function gp(e){const t=E.useContext(cQ);return e||t||"ltr"}function X3(e){const t=E.useRef({value:e,previous:e});return E.useMemo(()=>(t.current.value!==e&&(t.current.previous=t.current.value,t.current.value=e),t.current.previous),[e])}var dQ=[" ","Enter","ArrowUp","ArrowDown"],fQ=[" ","Enter"],Xu="Select",[hp,mp,pQ]=K3(Xu),[Ys,Lxe]=wr(Xu,[pQ,Vs]),bp=Vs(),[gQ,Io]=Ys(Xu),[hQ,mQ]=Ys(Xu),Z3=e=>{const{__scopeSelect:t,children:n,open:r,defaultOpen:a,onOpenChange:o,value:s,defaultValue:u,onValueChange:c,dir:d,name:p,autoComplete:g,disabled:m,required:b,form:y}=e,S=bp(t),[x,R]=E.useState(null),[k,A]=E.useState(null),[C,N]=E.useState(!1),_=gp(d),[O=!1,F]=ja({prop:r,defaultProp:a,onChange:o}),[L,I]=ja({prop:s,defaultProp:u,onChange:c}),H=E.useRef(null),$=x?y||!!x.closest("form"):!0,[U,Y]=E.useState(new Set),Z=Array.from(U).map(j=>j.props.value).join(";");return w.jsx(tT,{...S,children:w.jsxs(gQ,{required:b,scope:t,trigger:x,onTriggerChange:R,valueNode:k,onValueNodeChange:A,valueNodeHasChildren:C,onValueNodeHasChildrenChange:N,contentId:Tn(),value:L,onValueChange:I,open:O,onOpenChange:F,dir:_,triggerPointerDownPosRef:H,disabled:m,children:[w.jsx(hp.Provider,{scope:t,children:w.jsx(hQ,{scope:e.__scopeSelect,onNativeOptionAdd:E.useCallback(j=>{Y(G=>new Set(G).add(j))},[]),onNativeOptionRemove:E.useCallback(j=>{Y(G=>{const z=new Set(G);return z.delete(j),z})},[]),children:n})}),$?w.jsxs(xj,{"aria-hidden":!0,required:b,tabIndex:-1,name:p,autoComplete:g,value:L,onChange:j=>I(j.target.value),disabled:m,form:y,children:[L===void 0?w.jsx("option",{value:""}):null,Array.from(U)]},Z):null]})})};Z3.displayName=Xu;var Q3="SelectTrigger",J3=E.forwardRef((e,t)=>{const{__scopeSelect:n,disabled:r=!1,...a}=e,o=bp(n),s=Io(Q3,n),u=s.disabled||r,c=gt(t,s.onTriggerChange),d=mp(n),p=E.useRef("touch"),[g,m,b]=kj(S=>{const x=d().filter(A=>!A.disabled),R=x.find(A=>A.value===s.value),k=Tj(x,S,R);k!==void 0&&s.onValueChange(k.value)}),y=S=>{u||(s.onOpenChange(!0),b()),S&&(s.triggerPointerDownPosRef.current={x:Math.round(S.pageX),y:Math.round(S.pageY)})};return w.jsx(up,{asChild:!0,...o,children:w.jsx(Ze.button,{type:"button",role:"combobox","aria-controls":s.contentId,"aria-expanded":s.open,"aria-required":s.required,"aria-autocomplete":"none",dir:s.dir,"data-state":s.open?"open":"closed",disabled:u,"data-disabled":u?"":void 0,"data-placeholder":wj(s.value)?"":void 0,...a,ref:c,onClick:Ke(a.onClick,S=>{S.currentTarget.focus(),p.current!=="mouse"&&y(S)}),onPointerDown:Ke(a.onPointerDown,S=>{p.current=S.pointerType;const x=S.target;x.hasPointerCapture(S.pointerId)&&x.releasePointerCapture(S.pointerId),S.button===0&&S.ctrlKey===!1&&S.pointerType==="mouse"&&(y(S),S.preventDefault())}),onKeyDown:Ke(a.onKeyDown,S=>{const x=g.current!=="";!(S.ctrlKey||S.altKey||S.metaKey)&&S.key.length===1&&m(S.key),!(x&&S.key===" ")&&dQ.includes(S.key)&&(y(),S.preventDefault())})})})});J3.displayName=Q3;var ej="SelectValue",tj=E.forwardRef((e,t)=>{const{__scopeSelect:n,className:r,style:a,children:o,placeholder:s="",...u}=e,c=Io(ej,n),{onValueNodeHasChildrenChange:d}=c,p=o!==void 0,g=gt(t,c.onValueNodeChange);return An(()=>{d(p)},[d,p]),w.jsx(Ze.span,{...u,ref:g,style:{pointerEvents:"none"},children:wj(c.value)?w.jsx(w.Fragment,{children:s}):o})});tj.displayName=ej;var bQ="SelectIcon",nj=E.forwardRef((e,t)=>{const{__scopeSelect:n,children:r,...a}=e;return w.jsx(Ze.span,{"aria-hidden":!0,...a,ref:t,children:r||"▼"})});nj.displayName=bQ;var yQ="SelectPortal",rj=e=>w.jsx(L0,{asChild:!0,...e});rj.displayName=yQ;var bi="SelectContent",aj=E.forwardRef((e,t)=>{const n=Io(bi,e.__scopeSelect),[r,a]=E.useState();if(An(()=>{a(new DocumentFragment)},[]),!n.open){const o=r;return o?Gu.createPortal(w.jsx(oj,{scope:e.__scopeSelect,children:w.jsx(hp.Slot,{scope:e.__scopeSelect,children:w.jsx("div",{children:e.children})})}),o):null}return w.jsx(ij,{...e,ref:t})});aj.displayName=bi;var Lr=10,[oj,Lo]=Ys(bi),vQ="SelectContentImpl",ij=E.forwardRef((e,t)=>{const{__scopeSelect:n,position:r="item-aligned",onCloseAutoFocus:a,onEscapeKeyDown:o,onPointerDownOutside:s,side:u,sideOffset:c,align:d,alignOffset:p,arrowPadding:g,collisionBoundary:m,collisionPadding:b,sticky:y,hideWhenDetached:S,avoidCollisions:x,...R}=e,k=Io(bi,n),[A,C]=E.useState(null),[N,_]=E.useState(null),O=gt(t,re=>C(re)),[F,L]=E.useState(null),[I,H]=E.useState(null),$=mp(n),[U,Y]=E.useState(!1),Z=E.useRef(!1);E.useEffect(()=>{if(A)return F0(A)},[A]),M0();const j=E.useCallback(re=>{const[ie,...Se]=$().map(xe=>xe.ref.current),[ae]=Se.slice(-1),ve=document.activeElement;for(const xe of re)if(xe===ve||(xe==null||xe.scrollIntoView({block:"nearest"}),xe===ie&&N&&(N.scrollTop=0),xe===ae&&N&&(N.scrollTop=N.scrollHeight),xe==null||xe.focus(),document.activeElement!==ve))return},[$,N]),G=E.useCallback(()=>j([F,A]),[j,F,A]);E.useEffect(()=>{U&&G()},[U,G]);const{onOpenChange:z,triggerPointerDownPosRef:K}=k;E.useEffect(()=>{if(A){let re={x:0,y:0};const ie=ae=>{var ve,xe;re={x:Math.abs(Math.round(ae.pageX)-(((ve=K.current)==null?void 0:ve.x)??0)),y:Math.abs(Math.round(ae.pageY)-(((xe=K.current)==null?void 0:xe.y)??0))}},Se=ae=>{re.x<=10&&re.y<=10?ae.preventDefault():A.contains(ae.target)||z(!1),document.removeEventListener("pointermove",ie),K.current=null};return K.current!==null&&(document.addEventListener("pointermove",ie),document.addEventListener("pointerup",Se,{capture:!0,once:!0})),()=>{document.removeEventListener("pointermove",ie),document.removeEventListener("pointerup",Se,{capture:!0})}}},[A,z,K]),E.useEffect(()=>{const re=()=>z(!1);return window.addEventListener("blur",re),window.addEventListener("resize",re),()=>{window.removeEventListener("blur",re),window.removeEventListener("resize",re)}},[z]);const[D,V]=kj(re=>{const ie=$().filter(ve=>!ve.disabled),Se=ie.find(ve=>ve.ref.current===document.activeElement),ae=Tj(ie,re,Se);ae&&setTimeout(()=>ae.ref.current.focus())}),B=E.useCallback((re,ie,Se)=>{const ae=!Z.current&&!Se;(k.value!==void 0&&k.value===ie||ae)&&(L(re),ae&&(Z.current=!0))},[k.value]),M=E.useCallback(()=>A==null?void 0:A.focus(),[A]),W=E.useCallback((re,ie,Se)=>{const ae=!Z.current&&!Se;(k.value!==void 0&&k.value===ie||ae)&&H(re)},[k.value]),Q=r==="popper"?Nk:sj,oe=Q===Nk?{side:u,sideOffset:c,align:d,alignOffset:p,arrowPadding:g,collisionBoundary:m,collisionPadding:b,sticky:y,hideWhenDetached:S,avoidCollisions:x}:{};return w.jsx(oj,{scope:n,content:A,viewport:N,onViewportChange:_,itemRefCallback:B,selectedItem:F,onItemLeave:M,itemTextRefCallback:W,focusSelectedItem:G,selectedItemText:I,position:r,isPositioned:U,searchRef:D,children:w.jsx(rp,{as:Co,allowPinchZoom:!0,children:w.jsx(tp,{asChild:!0,trapped:k.open,onMountAutoFocus:re=>{re.preventDefault()},onUnmountAutoFocus:Ke(a,re=>{var ie;(ie=k.trigger)==null||ie.focus({preventScroll:!0}),re.preventDefault()}),children:w.jsx(qu,{asChild:!0,disableOutsidePointerEvents:!0,onEscapeKeyDown:o,onPointerDownOutside:s,onFocusOutside:re=>re.preventDefault(),onDismiss:()=>k.onOpenChange(!1),children:w.jsx(Q,{role:"listbox",id:k.contentId,"data-state":k.open?"open":"closed",dir:k.dir,onContextMenu:re=>re.preventDefault(),...R,...oe,onPlaced:()=>Y(!0),ref:O,style:{display:"flex",flexDirection:"column",outline:"none",...R.style},onKeyDown:Ke(R.onKeyDown,re=>{const ie=re.ctrlKey||re.altKey||re.metaKey;if(re.key==="Tab"&&re.preventDefault(),!ie&&re.key.length===1&&V(re.key),["ArrowUp","ArrowDown","Home","End"].includes(re.key)){let ae=$().filter(ve=>!ve.disabled).map(ve=>ve.ref.current);if(["ArrowUp","End"].includes(re.key)&&(ae=ae.slice().reverse()),["ArrowUp","ArrowDown"].includes(re.key)){const ve=re.target,xe=ae.indexOf(ve);ae=ae.slice(xe+1)}setTimeout(()=>j(ae)),re.preventDefault()}})})})})})})});ij.displayName=vQ;var SQ="SelectItemAlignedPosition",sj=E.forwardRef((e,t)=>{const{__scopeSelect:n,onPlaced:r,...a}=e,o=Io(bi,n),s=Lo(bi,n),[u,c]=E.useState(null),[d,p]=E.useState(null),g=gt(t,O=>p(O)),m=mp(n),b=E.useRef(!1),y=E.useRef(!0),{viewport:S,selectedItem:x,selectedItemText:R,focusSelectedItem:k}=s,A=E.useCallback(()=>{if(o.trigger&&o.valueNode&&u&&d&&S&&x&&R){const O=o.trigger.getBoundingClientRect(),F=d.getBoundingClientRect(),L=o.valueNode.getBoundingClientRect(),I=R.getBoundingClientRect();if(o.dir!=="rtl"){const ve=I.left-F.left,xe=L.left-ve,Ie=O.left-xe,Ce=O.width+Ie,ke=Math.max(Ce,F.width),J=window.innerWidth-Lr,fe=Ck(xe,[Lr,Math.max(Lr,J-ke)]);u.style.minWidth=Ce+"px",u.style.left=fe+"px"}else{const ve=F.right-I.right,xe=window.innerWidth-L.right-ve,Ie=window.innerWidth-O.right-xe,Ce=O.width+Ie,ke=Math.max(Ce,F.width),J=window.innerWidth-Lr,fe=Ck(xe,[Lr,Math.max(Lr,J-ke)]);u.style.minWidth=Ce+"px",u.style.right=fe+"px"}const H=m(),$=window.innerHeight-Lr*2,U=S.scrollHeight,Y=window.getComputedStyle(d),Z=parseInt(Y.borderTopWidth,10),j=parseInt(Y.paddingTop,10),G=parseInt(Y.borderBottomWidth,10),z=parseInt(Y.paddingBottom,10),K=Z+j+U+z+G,D=Math.min(x.offsetHeight*5,K),V=window.getComputedStyle(S),B=parseInt(V.paddingTop,10),M=parseInt(V.paddingBottom,10),W=O.top+O.height/2-Lr,Q=$-W,oe=x.offsetHeight/2,re=x.offsetTop+oe,ie=Z+j+re,Se=K-ie;if(ie<=W){const ve=H.length>0&&x===H[H.length-1].ref.current;u.style.bottom="0px";const xe=d.clientHeight-S.offsetTop-S.offsetHeight,Ie=Math.max(Q,oe+(ve?M:0)+xe+G),Ce=ie+Ie;u.style.height=Ce+"px"}else{const ve=H.length>0&&x===H[0].ref.current;u.style.top="0px";const Ie=Math.max(W,Z+S.offsetTop+(ve?B:0)+oe)+Se;u.style.height=Ie+"px",S.scrollTop=ie-W+S.offsetTop}u.style.margin=`${Lr}px 0`,u.style.minHeight=D+"px",u.style.maxHeight=$+"px",r==null||r(),requestAnimationFrame(()=>b.current=!0)}},[m,o.trigger,o.valueNode,u,d,S,x,R,o.dir,r]);An(()=>A(),[A]);const[C,N]=E.useState();An(()=>{d&&N(window.getComputedStyle(d).zIndex)},[d]);const _=E.useCallback(O=>{O&&y.current===!0&&(A(),k==null||k(),y.current=!1)},[A,k]);return w.jsx(wQ,{scope:n,contentWrapper:u,shouldExpandOnScrollRef:b,onScrollButtonChange:_,children:w.jsx("div",{ref:c,style:{display:"flex",flexDirection:"column",position:"fixed",zIndex:C},children:w.jsx(Ze.div,{...a,ref:g,style:{boxSizing:"border-box",maxHeight:"100%",...a.style}})})})});sj.displayName=SQ;var EQ="SelectPopperPosition",Nk=E.forwardRef((e,t)=>{const{__scopeSelect:n,align:r="start",collisionPadding:a=Lr,...o}=e,s=bp(n);return w.jsx(nT,{...s,...o,ref:t,align:r,collisionPadding:a,style:{boxSizing:"border-box",...o.style,"--radix-select-content-transform-origin":"var(--radix-popper-transform-origin)","--radix-select-content-available-width":"var(--radix-popper-available-width)","--radix-select-content-available-height":"var(--radix-popper-available-height)","--radix-select-trigger-width":"var(--radix-popper-anchor-width)","--radix-select-trigger-height":"var(--radix-popper-anchor-height)"}})});Nk.displayName=EQ;var[wQ,lT]=Ys(bi,{}),Ok="SelectViewport",lj=E.forwardRef((e,t)=>{const{__scopeSelect:n,nonce:r,...a}=e,o=Lo(Ok,n),s=lT(Ok,n),u=gt(t,o.onViewportChange),c=E.useRef(0);return w.jsxs(w.Fragment,{children:[w.jsx("style",{dangerouslySetInnerHTML:{__html:"[data-radix-select-viewport]{scrollbar-width:none;-ms-overflow-style:none;-webkit-overflow-scrolling:touch;}[data-radix-select-viewport]::-webkit-scrollbar{display:none}"},nonce:r}),w.jsx(hp.Slot,{scope:n,children:w.jsx(Ze.div,{"data-radix-select-viewport":"",role:"presentation",...a,ref:u,style:{position:"relative",flex:1,overflow:"hidden auto",...a.style},onScroll:Ke(a.onScroll,d=>{const p=d.currentTarget,{contentWrapper:g,shouldExpandOnScrollRef:m}=s;if(m!=null&&m.current&&g){const b=Math.abs(c.current-p.scrollTop);if(b>0){const y=window.innerHeight-Lr*2,S=parseFloat(g.style.minHeight),x=parseFloat(g.style.height),R=Math.max(S,x);if(R<y){const k=R+b,A=Math.min(y,k),C=k-A;g.style.height=A+"px",g.style.bottom==="0px"&&(p.scrollTop=C>0?C:0,g.style.justifyContent="flex-end")}}}c.current=p.scrollTop})})})]})});lj.displayName=Ok;var uj="SelectGroup",[xQ,kQ]=Ys(uj),cj=E.forwardRef((e,t)=>{const{__scopeSelect:n,...r}=e,a=Tn();return w.jsx(xQ,{scope:n,id:a,children:w.jsx(Ze.div,{role:"group","aria-labelledby":a,...r,ref:t})})});cj.displayName=uj;var dj="SelectLabel",fj=E.forwardRef((e,t)=>{const{__scopeSelect:n,...r}=e,a=kQ(dj,n);return w.jsx(Ze.div,{id:a.id,...r,ref:t})});fj.displayName=dj;var Ef="SelectItem",[TQ,pj]=Ys(Ef),gj=E.forwardRef((e,t)=>{const{__scopeSelect:n,value:r,disabled:a=!1,textValue:o,...s}=e,u=Io(Ef,n),c=Lo(Ef,n),d=u.value===r,[p,g]=E.useState(o??""),[m,b]=E.useState(!1),y=gt(t,k=>{var A;return(A=c.itemRefCallback)==null?void 0:A.call(c,k,r,a)}),S=Tn(),x=E.useRef("touch"),R=()=>{a||(u.onValueChange(r),u.onOpenChange(!1))};if(r==="")throw new Error("A <Select.Item /> must have a value prop that is not an empty string. This is because the Select value can be set to an empty string to clear the selection and show the placeholder.");return w.jsx(TQ,{scope:n,value:r,disabled:a,textId:S,isSelected:d,onItemTextChange:E.useCallback(k=>{g(A=>A||((k==null?void 0:k.textContent)??"").trim())},[]),children:w.jsx(hp.ItemSlot,{scope:n,value:r,disabled:a,textValue:p,children:w.jsx(Ze.div,{role:"option","aria-labelledby":S,"data-highlighted":m?"":void 0,"aria-selected":d&&m,"data-state":d?"checked":"unchecked","aria-disabled":a||void 0,"data-disabled":a?"":void 0,tabIndex:a?void 0:-1,...s,ref:y,onFocus:Ke(s.onFocus,()=>b(!0)),onBlur:Ke(s.onBlur,()=>b(!1)),onClick:Ke(s.onClick,()=>{x.current!=="mouse"&&R()}),onPointerUp:Ke(s.onPointerUp,()=>{x.current==="mouse"&&R()}),onPointerDown:Ke(s.onPointerDown,k=>{x.current=k.pointerType}),onPointerMove:Ke(s.onPointerMove,k=>{var A;x.current=k.pointerType,a?(A=c.onItemLeave)==null||A.call(c):x.current==="mouse"&&k.currentTarget.focus({preventScroll:!0})}),onPointerLeave:Ke(s.onPointerLeave,k=>{var A;k.currentTarget===document.activeElement&&((A=c.onItemLeave)==null||A.call(c))}),onKeyDown:Ke(s.onKeyDown,k=>{var C;((C=c.searchRef)==null?void 0:C.current)!==""&&k.key===" "||(fQ.includes(k.key)&&R(),k.key===" "&&k.preventDefault())})})})})});gj.displayName=Ef;var fu="SelectItemText",hj=E.forwardRef((e,t)=>{const{__scopeSelect:n,className:r,style:a,...o}=e,s=Io(fu,n),u=Lo(fu,n),c=pj(fu,n),d=mQ(fu,n),[p,g]=E.useState(null),m=gt(t,R=>g(R),c.onItemTextChange,R=>{var k;return(k=u.itemTextRefCallback)==null?void 0:k.call(u,R,c.value,c.disabled)}),b=p==null?void 0:p.textContent,y=E.useMemo(()=>w.jsx("option",{value:c.value,disabled:c.disabled,children:b},c.value),[c.disabled,c.value,b]),{onNativeOptionAdd:S,onNativeOptionRemove:x}=d;return An(()=>(S(y),()=>x(y)),[S,x,y]),w.jsxs(w.Fragment,{children:[w.jsx(Ze.span,{id:c.textId,...o,ref:m}),c.isSelected&&s.valueNode&&!s.valueNodeHasChildren?Gu.createPortal(o.children,s.valueNode):null]})});hj.displayName=fu;var mj="SelectItemIndicator",bj=E.forwardRef((e,t)=>{const{__scopeSelect:n,...r}=e;return pj(mj,n).isSelected?w.jsx(Ze.span,{"aria-hidden":!0,...r,ref:t}):null});bj.displayName=mj;var Dk="SelectScrollUpButton",yj=E.forwardRef((e,t)=>{const n=Lo(Dk,e.__scopeSelect),r=lT(Dk,e.__scopeSelect),[a,o]=E.useState(!1),s=gt(t,r.onScrollButtonChange);return An(()=>{if(n.viewport&&n.isPositioned){let u=function(){const d=c.scrollTop>0;o(d)};const c=n.viewport;return u(),c.addEventListener("scroll",u),()=>c.removeEventListener("scroll",u)}},[n.viewport,n.isPositioned]),a?w.jsx(Sj,{...e,ref:s,onAutoScroll:()=>{const{viewport:u,selectedItem:c}=n;u&&c&&(u.scrollTop=u.scrollTop-c.offsetHeight)}}):null});yj.displayName=Dk;var Ik="SelectScrollDownButton",vj=E.forwardRef((e,t)=>{const n=Lo(Ik,e.__scopeSelect),r=lT(Ik,e.__scopeSelect),[a,o]=E.useState(!1),s=gt(t,r.onScrollButtonChange);return An(()=>{if(n.viewport&&n.isPositioned){let u=function(){const d=c.scrollHeight-c.clientHeight,p=Math.ceil(c.scrollTop)<d;o(p)};const c=n.viewport;return u(),c.addEventListener("scroll",u),()=>c.removeEventListener("scroll",u)}},[n.viewport,n.isPositioned]),a?w.jsx(Sj,{...e,ref:s,onAutoScroll:()=>{const{viewport:u,selectedItem:c}=n;u&&c&&(u.scrollTop=u.scrollTop+c.offsetHeight)}}):null});vj.displayName=Ik;var Sj=E.forwardRef((e,t)=>{const{__scopeSelect:n,onAutoScroll:r,...a}=e,o=Lo("SelectScrollButton",n),s=E.useRef(null),u=mp(n),c=E.useCallback(()=>{s.current!==null&&(window.clearInterval(s.current),s.current=null)},[]);return E.useEffect(()=>()=>c(),[c]),An(()=>{var p;const d=u().find(g=>g.ref.current===document.activeElement);(p=d==null?void 0:d.ref.current)==null||p.scrollIntoView({block:"nearest"})},[u]),w.jsx(Ze.div,{"aria-hidden":!0,...a,ref:t,style:{flexShrink:0,...a.style},onPointerDown:Ke(a.onPointerDown,()=>{s.current===null&&(s.current=window.setInterval(r,50))}),onPointerMove:Ke(a.onPointerMove,()=>{var d;(d=o.onItemLeave)==null||d.call(o),s.current===null&&(s.current=window.setInterval(r,50))}),onPointerLeave:Ke(a.onPointerLeave,()=>{c()})})}),AQ="SelectSeparator",Ej=E.forwardRef((e,t)=>{const{__scopeSelect:n,...r}=e;return w.jsx(Ze.div,{"aria-hidden":!0,...r,ref:t})});Ej.displayName=AQ;var Lk="SelectArrow",RQ=E.forwardRef((e,t)=>{const{__scopeSelect:n,...r}=e,a=bp(n),o=Io(Lk,n),s=Lo(Lk,n);return o.open&&s.position==="popper"?w.jsx(rT,{...a,...r,ref:t}):null});RQ.displayName=Lk;function wj(e){return e===""||e===void 0}var xj=E.forwardRef((e,t)=>{const{value:n,...r}=e,a=E.useRef(null),o=gt(t,a),s=X3(n);return E.useEffect(()=>{const u=a.current,c=window.HTMLSelectElement.prototype,p=Object.getOwnPropertyDescriptor(c,"value").set;if(s!==n&&p){const g=new Event("change",{bubbles:!0});p.call(u,n),u.dispatchEvent(g)}},[s,n]),w.jsx(aT,{asChild:!0,children:w.jsx("select",{...r,ref:o,defaultValue:n})})});xj.displayName="BubbleSelect";function kj(e){const t=bn(e),n=E.useRef(""),r=E.useRef(0),a=E.useCallback(s=>{const u=n.current+s;t(u),function c(d){n.current=d,window.clearTimeout(r.current),d!==""&&(r.current=window.setTimeout(()=>c(""),1e3))}(u)},[t]),o=E.useCallback(()=>{n.current="",window.clearTimeout(r.current)},[]);return E.useEffect(()=>()=>window.clearTimeout(r.current),[]),[n,a,o]}function Tj(e,t,n){const a=t.length>1&&Array.from(t).every(d=>d===t[0])?t[0]:t,o=n?e.indexOf(n):-1;let s=_Q(e,Math.max(o,0));a.length===1&&(s=s.filter(d=>d!==n));const c=s.find(d=>d.textValue.toLowerCase().startsWith(a.toLowerCase()));return c!==n?c:void 0}function _Q(e,t){return e.map((n,r)=>e[(t+r)%e.length])}var CQ=Z3,Aj=J3,NQ=tj,OQ=nj,DQ=rj,Rj=aj,IQ=lj,LQ=cj,_j=fj,Cj=gj,MQ=hj,FQ=bj,Nj=yj,Oj=vj,Dj=Ej;const wf=CQ,aN=LQ,xf=NQ,Au=E.forwardRef(({className:e,children:t,...n},r)=>w.jsxs(Aj,{ref:r,className:Me("border-input bg-background ring-offset-background placeholder:text-muted-foreground focus:ring-ring flex h-10 w-full items-center justify-between rounded-md border px-3 py-2 text-sm focus:ring-2 focus:ring-offset-2 focus:outline-none disabled:cursor-not-allowed disabled:opacity-50 [&>span]:line-clamp-1",e),...n,children:[t,w.jsx(OQ,{asChild:!0,children:w.jsx(D0,{className:"h-4 w-4 opacity-50"})})]}));Au.displayName=Aj.displayName;const Ij=E.forwardRef(({className:e,...t},n)=>w.jsx(Nj,{ref:n,className:Me("flex cursor-default items-center justify-center py-1",e),...t,children:w.jsx(JB,{className:"h-4 w-4"})}));Ij.displayName=Nj.displayName;const Lj=E.forwardRef(({className:e,...t},n)=>w.jsx(Oj,{ref:n,className:Me("flex cursor-default items-center justify-center py-1",e),...t,children:w.jsx(D0,{className:"h-4 w-4"})}));Lj.displayName=Oj.displayName;const Ru=E.forwardRef(({className:e,children:t,position:n="popper",...r},a)=>w.jsx(DQ,{children:w.jsxs(Rj,{ref:a,className:Me("bg-popover text-popover-foreground data-[state=open]:animate-in data-[state=closed]:animate-out data-[state=closed]:fade-out-0 data-[state=open]:fade-in-0 data-[state=closed]:zoom-out-95 data-[state=open]:zoom-in-95 data-[side=bottom]:slide-in-from-top-2 data-[side=left]:slide-in-from-right-2 data-[side=right]:slide-in-from-left-2 data-[side=top]:slide-in-from-bottom-2 relative z-50 max-h-96 min-w-[8rem] overflow-hidden rounded-md border shadow-md",n==="popper"&&"data-[side=bottom]:translate-y-1 data-[side=left]:-translate-x-1 data-[side=right]:translate-x-1 data-[side=top]:-translate-y-1",e),position:n,...r,children:[w.jsx(Ij,{}),w.jsx(IQ,{className:Me("p-1",n==="popper"&&"h-[var(--radix-select-trigger-height)] w-full min-w-[var(--radix-select-trigger-width)]"),children:t}),w.jsx(Lj,{})]})}));Ru.displayName=Rj.displayName;const PQ=E.forwardRef(({className:e,...t},n)=>w.jsx(_j,{ref:n,className:Me("py-1.5 pr-2 pl-8 text-sm font-semibold",e),...t}));PQ.displayName=_j.displayName;const jn=E.forwardRef(({className:e,children:t,...n},r)=>w.jsxs(Cj,{ref:r,className:Me("focus:bg-accent focus:text-accent-foreground relative flex w-full cursor-default items-center rounded-sm py-1.5 pr-2 pl-8 text-sm outline-none select-none data-[disabled]:pointer-events-none data-[disabled]:opacity-50",e),...n,children:[w.jsx("span",{className:"absolute left-2 flex h-3.5 w-3.5 items-center justify-center",children:w.jsx(FQ,{children:w.jsx(O0,{className:"h-4 w-4"})})}),w.jsx(MQ,{children:t})]}));jn.displayName=Cj.displayName;const zQ=E.forwardRef(({className:e,...t},n)=>w.jsx(Dj,{ref:n,className:Me("bg-muted -mx-1 my-1 h-px",e),...t}));zQ.displayName=Dj.displayName;function Mj({className:e}){const[t,n]=E.useState(!1),{t:r}=Pt(),a=Fe.use.language(),o=Fe.use.setLanguage(),s=Fe.use.theme(),u=Fe.use.setTheme(),c=E.useCallback(p=>{o(p)},[o]),d=E.useCallback(p=>{u(p)},[u]);return w.jsxs(Yu,{open:t,onOpenChange:n,children:[w.jsx(Ku,{asChild:!0,children:w.jsx(xt,{variant:"ghost",size:"icon",className:Me("h-9 w-9",e),children:w.jsx(QW,{className:"h-5 w-5"})})}),w.jsx(Ws,{side:"bottom",align:"end",className:"w-56",children:w.jsxs("div",{className:"flex flex-col gap-4",children:[w.jsxs("div",{className:"flex flex-col gap-2",children:[w.jsx("label",{className:"text-sm font-medium",children:r("settings.language")}),w.jsxs(wf,{value:a,onValueChange:c,children:[w.jsx(Au,{children:w.jsx(xf,{})}),w.jsxs(Ru,{children:[w.jsx(jn,{value:"en",children:"English"}),w.jsx(jn,{value:"zh",children:"中文"})]})]})]}),w.jsxs("div",{className:"flex flex-col gap-2",children:[w.jsx("label",{className:"text-sm font-medium",children:r("settings.theme")}),w.jsxs(wf,{value:s,onValueChange:d,children:[w.jsx(Au,{children:w.jsx(xf,{})}),w.jsxs(Ru,{children:[w.jsx(jn,{value:"light",children:r("settings.light")}),w.jsx(jn,{value:"dark",children:r("settings.dark")}),w.jsx(jn,{value:"system",children:r("settings.system")})]})]})]})]})})]})}var fm="rovingFocusGroup.onEntryFocus",BQ={bubbles:!1,cancelable:!0},yp="RovingFocusGroup",[Mk,Fj,UQ]=K3(yp),[jQ,Pj]=wr(yp,[UQ]),[GQ,HQ]=jQ(yp),zj=E.forwardRef((e,t)=>w.jsx(Mk.Provider,{scope:e.__scopeRovingFocusGroup,children:w.jsx(Mk.Slot,{scope:e.__scopeRovingFocusGroup,children:w.jsx($Q,{...e,ref:t})})}));zj.displayName=yp;var $Q=E.forwardRef((e,t)=>{const{__scopeRovingFocusGroup:n,orientation:r,loop:a=!1,dir:o,currentTabStopId:s,defaultCurrentTabStopId:u,onCurrentTabStopIdChange:c,onEntryFocus:d,preventScrollOnEntryFocus:p=!1,...g}=e,m=E.useRef(null),b=gt(t,m),y=gp(o),[S=null,x]=ja({prop:s,defaultProp:u,onChange:c}),[R,k]=E.useState(!1),A=bn(d),C=Fj(n),N=E.useRef(!1),[_,O]=E.useState(0);return E.useEffect(()=>{const F=m.current;if(F)return F.addEventListener(fm,A),()=>F.removeEventListener(fm,A)},[A]),w.jsx(GQ,{scope:n,orientation:r,dir:y,loop:a,currentTabStopId:S,onItemFocus:E.useCallback(F=>x(F),[x]),onItemShiftTab:E.useCallback(()=>k(!0),[]),onFocusableItemAdd:E.useCallback(()=>O(F=>F+1),[]),onFocusableItemRemove:E.useCallback(()=>O(F=>F-1),[]),children:w.jsx(Ze.div,{tabIndex:R||_===0?-1:0,"data-orientation":r,...g,ref:b,style:{outline:"none",...e.style},onMouseDown:Ke(e.onMouseDown,()=>{N.current=!0}),onFocus:Ke(e.onFocus,F=>{const L=!N.current;if(F.target===F.currentTarget&&L&&!R){const I=new CustomEvent(fm,BQ);if(F.currentTarget.dispatchEvent(I),!I.defaultPrevented){const H=C().filter(j=>j.focusable),$=H.find(j=>j.active),U=H.find(j=>j.id===S),Z=[$,U,...H].filter(Boolean).map(j=>j.ref.current);jj(Z,p)}}N.current=!1}),onBlur:Ke(e.onBlur,()=>k(!1))})})}),Bj="RovingFocusGroupItem",Uj=E.forwardRef((e,t)=>{const{__scopeRovingFocusGroup:n,focusable:r=!0,active:a=!1,tabStopId:o,...s}=e,u=Tn(),c=o||u,d=HQ(Bj,n),p=d.currentTabStopId===c,g=Fj(n),{onFocusableItemAdd:m,onFocusableItemRemove:b}=d;return E.useEffect(()=>{if(r)return m(),()=>b()},[r,m,b]),w.jsx(Mk.ItemSlot,{scope:n,id:c,focusable:r,active:a,children:w.jsx(Ze.span,{tabIndex:p?0:-1,"data-orientation":d.orientation,...s,ref:t,onMouseDown:Ke(e.onMouseDown,y=>{r?d.onItemFocus(c):y.preventDefault()}),onFocus:Ke(e.onFocus,()=>d.onItemFocus(c)),onKeyDown:Ke(e.onKeyDown,y=>{if(y.key==="Tab"&&y.shiftKey){d.onItemShiftTab();return}if(y.target!==y.currentTarget)return;const S=WQ(y,d.orientation,d.dir);if(S!==void 0){if(y.metaKey||y.ctrlKey||y.altKey||y.shiftKey)return;y.preventDefault();let R=g().filter(k=>k.focusable).map(k=>k.ref.current);if(S==="last")R.reverse();else if(S==="prev"||S==="next"){S==="prev"&&R.reverse();const k=R.indexOf(y.currentTarget);R=d.loop?YQ(R,k+1):R.slice(k+1)}setTimeout(()=>jj(R))}})})})});Uj.displayName=Bj;var qQ={ArrowLeft:"prev",ArrowUp:"prev",ArrowRight:"next",ArrowDown:"next",PageUp:"first",Home:"first",PageDown:"last",End:"last"};function VQ(e,t){return t!=="rtl"?e:e==="ArrowLeft"?"ArrowRight":e==="ArrowRight"?"ArrowLeft":e}function WQ(e,t,n){const r=VQ(e.key,n);if(!(t==="vertical"&&["ArrowLeft","ArrowRight"].includes(r))&&!(t==="horizontal"&&["ArrowUp","ArrowDown"].includes(r)))return qQ[r]}function jj(e,t=!1){const n=document.activeElement;for(const r of e)if(r===n||(r.focus({preventScroll:t}),document.activeElement!==n))return}function YQ(e,t){return e.map((n,r)=>e[(t+r)%e.length])}var KQ=zj,XQ=Uj,uT="Tabs",[ZQ,Mxe]=wr(uT,[Pj]),Gj=Pj(),[QQ,cT]=ZQ(uT),Hj=E.forwardRef((e,t)=>{const{__scopeTabs:n,value:r,onValueChange:a,defaultValue:o,orientation:s="horizontal",dir:u,activationMode:c="automatic",...d}=e,p=gp(u),[g,m]=ja({prop:r,onChange:a,defaultProp:o});return w.jsx(QQ,{scope:n,baseId:Tn(),value:g,onValueChange:m,orientation:s,dir:p,activationMode:c,children:w.jsx(Ze.div,{dir:p,"data-orientation":s,...d,ref:t})})});Hj.displayName=uT;var $j="TabsList",qj=E.forwardRef((e,t)=>{const{__scopeTabs:n,loop:r=!0,...a}=e,o=cT($j,n),s=Gj(n);return w.jsx(KQ,{asChild:!0,...s,orientation:o.orientation,dir:o.dir,loop:r,children:w.jsx(Ze.div,{role:"tablist","aria-orientation":o.orientation,...a,ref:t})})});qj.displayName=$j;var Vj="TabsTrigger",Wj=E.forwardRef((e,t)=>{const{__scopeTabs:n,value:r,disabled:a=!1,...o}=e,s=cT(Vj,n),u=Gj(n),c=Xj(s.baseId,r),d=Zj(s.baseId,r),p=r===s.value;return w.jsx(XQ,{asChild:!0,...u,focusable:!a,active:p,children:w.jsx(Ze.button,{type:"button",role:"tab","aria-selected":p,"aria-controls":d,"data-state":p?"active":"inactive","data-disabled":a?"":void 0,disabled:a,id:c,...o,ref:t,onMouseDown:Ke(e.onMouseDown,g=>{!a&&g.button===0&&g.ctrlKey===!1?s.onValueChange(r):g.preventDefault()}),onKeyDown:Ke(e.onKeyDown,g=>{[" ","Enter"].includes(g.key)&&s.onValueChange(r)}),onFocus:Ke(e.onFocus,()=>{const g=s.activationMode!=="manual";!p&&!a&&g&&s.onValueChange(r)})})})});Wj.displayName=Vj;var Yj="TabsContent",Kj=E.forwardRef((e,t)=>{const{__scopeTabs:n,value:r,forceMount:a,children:o,...s}=e,u=cT(Yj,n),c=Xj(u.baseId,r),d=Zj(u.baseId,r),p=r===u.value,g=E.useRef(p);return E.useEffect(()=>{const m=requestAnimationFrame(()=>g.current=!1);return()=>cancelAnimationFrame(m)},[]),w.jsx(xr,{present:a||p,children:({present:m})=>w.jsx(Ze.div,{"data-state":p?"active":"inactive","data-orientation":u.orientation,role:"tabpanel","aria-labelledby":c,hidden:!m,id:d,tabIndex:0,...s,ref:t,style:{...e.style,animationDuration:g.current?"0s":void 0},children:m&&o})})});Kj.displayName=Yj;function Xj(e,t){return`${e}-trigger-${t}`}function Zj(e,t){return`${e}-content-${t}`}var JQ=Hj,Qj=qj,Jj=Wj,e4=Kj;const eJ=JQ,t4=E.forwardRef(({className:e,...t},n)=>w.jsx(Qj,{ref:n,className:Me("bg-muted text-muted-foreground inline-flex h-10 items-center justify-center rounded-md p-1",e),...t}));t4.displayName=Qj.displayName;const n4=E.forwardRef(({className:e,...t},n)=>w.jsx(Jj,{ref:n,className:Me("ring-offset-background focus-visible:ring-ring data-[state=active]:bg-background data-[state=active]:text-foreground inline-flex items-center justify-center rounded-sm px-3 py-1.5 text-sm font-medium whitespace-nowrap transition-all focus-visible:ring-2 focus-visible:ring-offset-2 focus-visible:outline-none disabled:pointer-events-none disabled:opacity-50 data-[state=active]:shadow-sm",e),...t}));n4.displayName=Jj.displayName;const pu=E.forwardRef(({className:e,...t},n)=>w.jsx(e4,{ref:n,className:Me("ring-offset-background focus-visible:ring-ring focus-visible:ring-2 focus-visible:ring-offset-2 focus-visible:outline-none","data-[state=inactive]:invisible data-[state=active]:visible","h-full w-full",e),forceMount:!0,...t}));pu.displayName=e4.displayName;function kd({value:e,currentTab:t,children:n}){return w.jsx(n4,{value:e,className:Me("cursor-pointer px-2 py-1 transition-all",t===e?"!bg-emerald-400 !text-zinc-50":"hover:bg-background/60"),children:n})}function tJ(){const e=Fe.use.currentTab(),{t}=Pt();return w.jsx("div",{className:"flex h-8 self-center",children:w.jsxs(t4,{className:"h-full gap-2",children:[w.jsx(kd,{value:"documents",currentTab:e,children:t("header.documents")}),w.jsx(kd,{value:"knowledge-graph",currentTab:e,children:t("header.knowledgeGraph")}),w.jsx(kd,{value:"retrieval",currentTab:e,children:t("header.retrieval")}),w.jsx(kd,{value:"api",currentTab:e,children:t("header.api")})]})})}function nJ(){const{t:e}=Pt(),{isGuestMode:t}=na(),n=()=>{xu.navigateToLogin()};return w.jsxs("header",{className:"border-border/40 bg-background/95 supports-[backdrop-filter]:bg-background/60 sticky top-0 z-50 flex h-10 w-full border-b px-4 backdrop-blur",children:[w.jsxs("a",{href:cV,className:"mr-6 flex items-center gap-2",children:[w.jsx(aU,{className:"size-4 text-emerald-400","aria-hidden":"true"}),w.jsx("span",{className:"font-bold md:inline-block",children:CC.name})]}),w.jsxs("div",{className:"flex h-10 flex-1 justify-center",children:[w.jsx(tJ,{}),t&&w.jsx("div",{className:"ml-2 self-center px-2 py-1 text-xs bg-amber-100 text-amber-800 dark:bg-amber-900 dark:text-amber-200 rounded-md",children:e("login.guestMode","Guest Mode")})]}),w.jsx("nav",{className:"flex items-center",children:w.jsxs("div",{className:"flex items-center gap-2",children:[w.jsx(xt,{variant:"ghost",size:"icon",side:"bottom",tooltip:e("header.projectRepository"),children:w.jsx("a",{href:CC.github,target:"_blank",rel:"noopener noreferrer",children:w.jsx(BW,{className:"size-4","aria-hidden":"true"})})}),w.jsx(Mj,{}),w.jsx(xt,{variant:"ghost",size:"icon",side:"bottom",tooltip:e("header.logout"),onClick:n,children:w.jsx(VW,{className:"size-4","aria-hidden":"true"})})]})})]})}var Td={exports:{}},oN;function rJ(){if(oN)return Td.exports;oN=1;var e=typeof Reflect=="object"?Reflect:null,t=e&&typeof e.apply=="function"?e.apply:function(N,_,O){return Function.prototype.apply.call(N,_,O)},n;e&&typeof e.ownKeys=="function"?n=e.ownKeys:Object.getOwnPropertySymbols?n=function(N){return Object.getOwnPropertyNames(N).concat(Object.getOwnPropertySymbols(N))}:n=function(N){return Object.getOwnPropertyNames(N)};function r(C){console&&console.warn&&console.warn(C)}var a=Number.isNaN||function(N){return N!==N};function o(){o.init.call(this)}Td.exports=o,Td.exports.once=R,o.EventEmitter=o,o.prototype._events=void 0,o.prototype._eventsCount=0,o.prototype._maxListeners=void 0;var s=10;function u(C){if(typeof C!="function")throw new TypeError('The "listener" argument must be of type Function. Received type '+typeof C)}Object.defineProperty(o,"defaultMaxListeners",{enumerable:!0,get:function(){return s},set:function(C){if(typeof C!="number"||C<0||a(C))throw new RangeError('The value of "defaultMaxListeners" is out of range. It must be a non-negative number. Received '+C+".");s=C}}),o.init=function(){(this._events===void 0||this._events===Object.getPrototypeOf(this)._events)&&(this._events=Object.create(null),this._eventsCount=0),this._maxListeners=this._maxListeners||void 0},o.prototype.setMaxListeners=function(N){if(typeof N!="number"||N<0||a(N))throw new RangeError('The value of "n" is out of range. It must be a non-negative number. Received '+N+".");return this._maxListeners=N,this};function c(C){return C._maxListeners===void 0?o.defaultMaxListeners:C._maxListeners}o.prototype.getMaxListeners=function(){return c(this)},o.prototype.emit=function(N){for(var _=[],O=1;O<arguments.length;O++)_.push(arguments[O]);var F=N==="error",L=this._events;if(L!==void 0)F=F&&L.error===void 0;else if(!F)return!1;if(F){var I;if(_.length>0&&(I=_[0]),I instanceof Error)throw I;var H=new Error("Unhandled error."+(I?" ("+I.message+")":""));throw H.context=I,H}var $=L[N];if($===void 0)return!1;if(typeof $=="function")t($,this,_);else for(var U=$.length,Y=y($,U),O=0;O<U;++O)t(Y[O],this,_);return!0};function d(C,N,_,O){var F,L,I;if(u(_),L=C._events,L===void 0?(L=C._events=Object.create(null),C._eventsCount=0):(L.newListener!==void 0&&(C.emit("newListener",N,_.listener?_.listener:_),L=C._events),I=L[N]),I===void 0)I=L[N]=_,++C._eventsCount;else if(typeof I=="function"?I=L[N]=O?[_,I]:[I,_]:O?I.unshift(_):I.push(_),F=c(C),F>0&&I.length>F&&!I.warned){I.warned=!0;var H=new Error("Possible EventEmitter memory leak detected. "+I.length+" "+String(N)+" listeners added. Use emitter.setMaxListeners() to increase limit");H.name="MaxListenersExceededWarning",H.emitter=C,H.type=N,H.count=I.length,r(H)}return C}o.prototype.addListener=function(N,_){return d(this,N,_,!1)},o.prototype.on=o.prototype.addListener,o.prototype.prependListener=function(N,_){return d(this,N,_,!0)};function p(){if(!this.fired)return this.target.removeListener(this.type,this.wrapFn),this.fired=!0,arguments.length===0?this.listener.call(this.target):this.listener.apply(this.target,arguments)}function g(C,N,_){var O={fired:!1,wrapFn:void 0,target:C,type:N,listener:_},F=p.bind(O);return F.listener=_,O.wrapFn=F,F}o.prototype.once=function(N,_){return u(_),this.on(N,g(this,N,_)),this},o.prototype.prependOnceListener=function(N,_){return u(_),this.prependListener(N,g(this,N,_)),this},o.prototype.removeListener=function(N,_){var O,F,L,I,H;if(u(_),F=this._events,F===void 0)return this;if(O=F[N],O===void 0)return this;if(O===_||O.listener===_)--this._eventsCount===0?this._events=Object.create(null):(delete F[N],F.removeListener&&this.emit("removeListener",N,O.listener||_));else if(typeof O!="function"){for(L=-1,I=O.length-1;I>=0;I--)if(O[I]===_||O[I].listener===_){H=O[I].listener,L=I;break}if(L<0)return this;L===0?O.shift():S(O,L),O.length===1&&(F[N]=O[0]),F.removeListener!==void 0&&this.emit("removeListener",N,H||_)}return this},o.prototype.off=o.prototype.removeListener,o.prototype.removeAllListeners=function(N){var _,O,F;if(O=this._events,O===void 0)return this;if(O.removeListener===void 0)return arguments.length===0?(this._events=Object.create(null),this._eventsCount=0):O[N]!==void 0&&(--this._eventsCount===0?this._events=Object.create(null):delete O[N]),this;if(arguments.length===0){var L=Object.keys(O),I;for(F=0;F<L.length;++F)I=L[F],I!=="removeListener"&&this.removeAllListeners(I);return this.removeAllListeners("removeListener"),this._events=Object.create(null),this._eventsCount=0,this}if(_=O[N],typeof _=="function")this.removeListener(N,_);else if(_!==void 0)for(F=_.length-1;F>=0;F--)this.removeListener(N,_[F]);return this};function m(C,N,_){var O=C._events;if(O===void 0)return[];var F=O[N];return F===void 0?[]:typeof F=="function"?_?[F.listener||F]:[F]:_?x(F):y(F,F.length)}o.prototype.listeners=function(N){return m(this,N,!0)},o.prototype.rawListeners=function(N){return m(this,N,!1)},o.listenerCount=function(C,N){return typeof C.listenerCount=="function"?C.listenerCount(N):b.call(C,N)},o.prototype.listenerCount=b;function b(C){var N=this._events;if(N!==void 0){var _=N[C];if(typeof _=="function")return 1;if(_!==void 0)return _.length}return 0}o.prototype.eventNames=function(){return this._eventsCount>0?n(this._events):[]};function y(C,N){for(var _=new Array(N),O=0;O<N;++O)_[O]=C[O];return _}function S(C,N){for(;N+1<C.length;N++)C[N]=C[N+1];C.pop()}function x(C){for(var N=new Array(C.length),_=0;_<N.length;++_)N[_]=C[_].listener||C[_];return N}function R(C,N){return new Promise(function(_,O){function F(I){C.removeListener(N,L),O(I)}function L(){typeof C.removeListener=="function"&&C.removeListener("error",F),_([].slice.call(arguments))}A(C,N,L,{once:!0}),N!=="error"&&k(C,F,{once:!0})})}function k(C,N,_){typeof C.on=="function"&&A(C,"error",N,_)}function A(C,N,_,O){if(typeof C.on=="function")O.once?C.once(N,_):C.on(N,_);else if(typeof C.addEventListener=="function")C.addEventListener(N,function F(L){O.once&&C.removeEventListener(N,F),_(L)});else throw new TypeError('The "emitter" argument must be of type EventEmitter. Received type '+typeof C)}return Td.exports}var r4=rJ();function aJ(){const e=arguments[0];for(let t=1,n=arguments.length;t<n;t++)if(arguments[t])for(const r in arguments[t])e[r]=arguments[t][r];return e}let nn=aJ;typeof Object.assign=="function"&&(nn=Object.assign);function Er(e,t,n,r){const a=e._nodes.get(t);let o=null;return a&&(r==="mixed"?o=a.out&&a.out[n]||a.undirected&&a.undirected[n]:r==="directed"?o=a.out&&a.out[n]:o=a.undirected&&a.undirected[n]),o}function yn(e){return typeof e=="object"&&e!==null}function a4(e){let t;for(t in e)return!1;return!0}function yr(e,t,n){Object.defineProperty(e,t,{enumerable:!1,configurable:!1,writable:!0,value:n})}function Ir(e,t,n){const r={enumerable:!0,configurable:!0};typeof n=="function"?r.get=n:(r.value=n,r.writable=!1),Object.defineProperty(e,t,r)}function iN(e){return!(!yn(e)||e.attributes&&!Array.isArray(e.attributes))}function oJ(){let e=Math.floor(Math.random()*256)&255;return()=>e++}function Ba(){const e=arguments;let t=null,n=-1;return{[Symbol.iterator](){return this},next(){let r=null;do{if(t===null){if(n++,n>=e.length)return{done:!0};t=e[n][Symbol.iterator]()}if(r=t.next(),r.done){t=null;continue}break}while(!0);return r}}}function Ks(){return{[Symbol.iterator](){return this},next(){return{done:!0}}}}class dT extends Error{constructor(t){super(),this.name="GraphError",this.message=t}}class Le extends dT{constructor(t){super(t),this.name="InvalidArgumentsGraphError",typeof Error.captureStackTrace=="function"&&Error.captureStackTrace(this,Le.prototype.constructor)}}class De extends dT{constructor(t){super(t),this.name="NotFoundGraphError",typeof Error.captureStackTrace=="function"&&Error.captureStackTrace(this,De.prototype.constructor)}}class Qe extends dT{constructor(t){super(t),this.name="UsageGraphError",typeof Error.captureStackTrace=="function"&&Error.captureStackTrace(this,Qe.prototype.constructor)}}function o4(e,t){this.key=e,this.attributes=t,this.clear()}o4.prototype.clear=function(){this.inDegree=0,this.outDegree=0,this.undirectedDegree=0,this.undirectedLoops=0,this.directedLoops=0,this.in={},this.out={},this.undirected={}};function i4(e,t){this.key=e,this.attributes=t,this.clear()}i4.prototype.clear=function(){this.inDegree=0,this.outDegree=0,this.directedLoops=0,this.in={},this.out={}};function s4(e,t){this.key=e,this.attributes=t,this.clear()}s4.prototype.clear=function(){this.undirectedDegree=0,this.undirectedLoops=0,this.undirected={}};function Xs(e,t,n,r,a){this.key=t,this.attributes=a,this.undirected=e,this.source=n,this.target=r}Xs.prototype.attach=function(){let e="out",t="in";this.undirected&&(e=t="undirected");const n=this.source.key,r=this.target.key;this.source[e][r]=this,!(this.undirected&&n===r)&&(this.target[t][n]=this)};Xs.prototype.attachMulti=function(){let e="out",t="in";const n=this.source.key,r=this.target.key;this.undirected&&(e=t="undirected");const a=this.source[e],o=a[r];if(typeof o>"u"){a[r]=this,this.undirected&&n===r||(this.target[t][n]=this);return}o.previous=this,this.next=o,a[r]=this,this.target[t][n]=this};Xs.prototype.detach=function(){const e=this.source.key,t=this.target.key;let n="out",r="in";this.undirected&&(n=r="undirected"),delete this.source[n][t],delete this.target[r][e]};Xs.prototype.detachMulti=function(){const e=this.source.key,t=this.target.key;let n="out",r="in";this.undirected&&(n=r="undirected"),this.previous===void 0?this.next===void 0?(delete this.source[n][t],delete this.target[r][e]):(this.next.previous=void 0,this.source[n][t]=this.next,this.target[r][e]=this.next):(this.previous.next=this.next,this.next!==void 0&&(this.next.previous=this.previous))};const l4=0,u4=1,iJ=2,c4=3;function Wa(e,t,n,r,a,o,s){let u,c,d,p;if(r=""+r,n===l4){if(u=e._nodes.get(r),!u)throw new De(`Graph.${t}: could not find the "${r}" node in the graph.`);d=a,p=o}else if(n===c4){if(a=""+a,c=e._edges.get(a),!c)throw new De(`Graph.${t}: could not find the "${a}" edge in the graph.`);const g=c.source.key,m=c.target.key;if(r===g)u=c.target;else if(r===m)u=c.source;else throw new De(`Graph.${t}: the "${r}" node is not attached to the "${a}" edge (${g}, ${m}).`);d=o,p=s}else{if(c=e._edges.get(r),!c)throw new De(`Graph.${t}: could not find the "${r}" edge in the graph.`);n===u4?u=c.source:u=c.target,d=a,p=o}return[u,d,p]}function sJ(e,t,n){e.prototype[t]=function(r,a,o){const[s,u]=Wa(this,t,n,r,a,o);return s.attributes[u]}}function lJ(e,t,n){e.prototype[t]=function(r,a){const[o]=Wa(this,t,n,r,a);return o.attributes}}function uJ(e,t,n){e.prototype[t]=function(r,a,o){const[s,u]=Wa(this,t,n,r,a,o);return s.attributes.hasOwnProperty(u)}}function cJ(e,t,n){e.prototype[t]=function(r,a,o,s){const[u,c,d]=Wa(this,t,n,r,a,o,s);return u.attributes[c]=d,this.emit("nodeAttributesUpdated",{key:u.key,type:"set",attributes:u.attributes,name:c}),this}}function dJ(e,t,n){e.prototype[t]=function(r,a,o,s){const[u,c,d]=Wa(this,t,n,r,a,o,s);if(typeof d!="function")throw new Le(`Graph.${t}: updater should be a function.`);const p=u.attributes,g=d(p[c]);return p[c]=g,this.emit("nodeAttributesUpdated",{key:u.key,type:"set",attributes:u.attributes,name:c}),this}}function fJ(e,t,n){e.prototype[t]=function(r,a,o){const[s,u]=Wa(this,t,n,r,a,o);return delete s.attributes[u],this.emit("nodeAttributesUpdated",{key:s.key,type:"remove",attributes:s.attributes,name:u}),this}}function pJ(e,t,n){e.prototype[t]=function(r,a,o){const[s,u]=Wa(this,t,n,r,a,o);if(!yn(u))throw new Le(`Graph.${t}: provided attributes are not a plain object.`);return s.attributes=u,this.emit("nodeAttributesUpdated",{key:s.key,type:"replace",attributes:s.attributes}),this}}function gJ(e,t,n){e.prototype[t]=function(r,a,o){const[s,u]=Wa(this,t,n,r,a,o);if(!yn(u))throw new Le(`Graph.${t}: provided attributes are not a plain object.`);return nn(s.attributes,u),this.emit("nodeAttributesUpdated",{key:s.key,type:"merge",attributes:s.attributes,data:u}),this}}function hJ(e,t,n){e.prototype[t]=function(r,a,o){const[s,u]=Wa(this,t,n,r,a,o);if(typeof u!="function")throw new Le(`Graph.${t}: provided updater is not a function.`);return s.attributes=u(s.attributes),this.emit("nodeAttributesUpdated",{key:s.key,type:"update",attributes:s.attributes}),this}}const mJ=[{name:e=>`get${e}Attribute`,attacher:sJ},{name:e=>`get${e}Attributes`,attacher:lJ},{name:e=>`has${e}Attribute`,attacher:uJ},{name:e=>`set${e}Attribute`,attacher:cJ},{name:e=>`update${e}Attribute`,attacher:dJ},{name:e=>`remove${e}Attribute`,attacher:fJ},{name:e=>`replace${e}Attributes`,attacher:pJ},{name:e=>`merge${e}Attributes`,attacher:gJ},{name:e=>`update${e}Attributes`,attacher:hJ}];function bJ(e){mJ.forEach(function({name:t,attacher:n}){n(e,t("Node"),l4),n(e,t("Source"),u4),n(e,t("Target"),iJ),n(e,t("Opposite"),c4)})}function yJ(e,t,n){e.prototype[t]=function(r,a){let o;if(this.type!=="mixed"&&n!=="mixed"&&n!==this.type)throw new Qe(`Graph.${t}: cannot find this type of edges in your ${this.type} graph.`);if(arguments.length>2){if(this.multi)throw new Qe(`Graph.${t}: cannot use a {source,target} combo when asking about an edge's attributes in a MultiGraph since we cannot infer the one you want information about.`);const s=""+r,u=""+a;if(a=arguments[2],o=Er(this,s,u,n),!o)throw new De(`Graph.${t}: could not find an edge for the given path ("${s}" - "${u}").`)}else{if(n!=="mixed")throw new Qe(`Graph.${t}: calling this method with only a key (vs. a source and target) does not make sense since an edge with this key could have the other type.`);if(r=""+r,o=this._edges.get(r),!o)throw new De(`Graph.${t}: could not find the "${r}" edge in the graph.`)}return o.attributes[a]}}function vJ(e,t,n){e.prototype[t]=function(r){let a;if(this.type!=="mixed"&&n!=="mixed"&&n!==this.type)throw new Qe(`Graph.${t}: cannot find this type of edges in your ${this.type} graph.`);if(arguments.length>1){if(this.multi)throw new Qe(`Graph.${t}: cannot use a {source,target} combo when asking about an edge's attributes in a MultiGraph since we cannot infer the one you want information about.`);const o=""+r,s=""+arguments[1];if(a=Er(this,o,s,n),!a)throw new De(`Graph.${t}: could not find an edge for the given path ("${o}" - "${s}").`)}else{if(n!=="mixed")throw new Qe(`Graph.${t}: calling this method with only a key (vs. a source and target) does not make sense since an edge with this key could have the other type.`);if(r=""+r,a=this._edges.get(r),!a)throw new De(`Graph.${t}: could not find the "${r}" edge in the graph.`)}return a.attributes}}function SJ(e,t,n){e.prototype[t]=function(r,a){let o;if(this.type!=="mixed"&&n!=="mixed"&&n!==this.type)throw new Qe(`Graph.${t}: cannot find this type of edges in your ${this.type} graph.`);if(arguments.length>2){if(this.multi)throw new Qe(`Graph.${t}: cannot use a {source,target} combo when asking about an edge's attributes in a MultiGraph since we cannot infer the one you want information about.`);const s=""+r,u=""+a;if(a=arguments[2],o=Er(this,s,u,n),!o)throw new De(`Graph.${t}: could not find an edge for the given path ("${s}" - "${u}").`)}else{if(n!=="mixed")throw new Qe(`Graph.${t}: calling this method with only a key (vs. a source and target) does not make sense since an edge with this key could have the other type.`);if(r=""+r,o=this._edges.get(r),!o)throw new De(`Graph.${t}: could not find the "${r}" edge in the graph.`)}return o.attributes.hasOwnProperty(a)}}function EJ(e,t,n){e.prototype[t]=function(r,a,o){let s;if(this.type!=="mixed"&&n!=="mixed"&&n!==this.type)throw new Qe(`Graph.${t}: cannot find this type of edges in your ${this.type} graph.`);if(arguments.length>3){if(this.multi)throw new Qe(`Graph.${t}: cannot use a {source,target} combo when asking about an edge's attributes in a MultiGraph since we cannot infer the one you want information about.`);const u=""+r,c=""+a;if(a=arguments[2],o=arguments[3],s=Er(this,u,c,n),!s)throw new De(`Graph.${t}: could not find an edge for the given path ("${u}" - "${c}").`)}else{if(n!=="mixed")throw new Qe(`Graph.${t}: calling this method with only a key (vs. a source and target) does not make sense since an edge with this key could have the other type.`);if(r=""+r,s=this._edges.get(r),!s)throw new De(`Graph.${t}: could not find the "${r}" edge in the graph.`)}return s.attributes[a]=o,this.emit("edgeAttributesUpdated",{key:s.key,type:"set",attributes:s.attributes,name:a}),this}}function wJ(e,t,n){e.prototype[t]=function(r,a,o){let s;if(this.type!=="mixed"&&n!=="mixed"&&n!==this.type)throw new Qe(`Graph.${t}: cannot find this type of edges in your ${this.type} graph.`);if(arguments.length>3){if(this.multi)throw new Qe(`Graph.${t}: cannot use a {source,target} combo when asking about an edge's attributes in a MultiGraph since we cannot infer the one you want information about.`);const u=""+r,c=""+a;if(a=arguments[2],o=arguments[3],s=Er(this,u,c,n),!s)throw new De(`Graph.${t}: could not find an edge for the given path ("${u}" - "${c}").`)}else{if(n!=="mixed")throw new Qe(`Graph.${t}: calling this method with only a key (vs. a source and target) does not make sense since an edge with this key could have the other type.`);if(r=""+r,s=this._edges.get(r),!s)throw new De(`Graph.${t}: could not find the "${r}" edge in the graph.`)}if(typeof o!="function")throw new Le(`Graph.${t}: updater should be a function.`);return s.attributes[a]=o(s.attributes[a]),this.emit("edgeAttributesUpdated",{key:s.key,type:"set",attributes:s.attributes,name:a}),this}}function xJ(e,t,n){e.prototype[t]=function(r,a){let o;if(this.type!=="mixed"&&n!=="mixed"&&n!==this.type)throw new Qe(`Graph.${t}: cannot find this type of edges in your ${this.type} graph.`);if(arguments.length>2){if(this.multi)throw new Qe(`Graph.${t}: cannot use a {source,target} combo when asking about an edge's attributes in a MultiGraph since we cannot infer the one you want information about.`);const s=""+r,u=""+a;if(a=arguments[2],o=Er(this,s,u,n),!o)throw new De(`Graph.${t}: could not find an edge for the given path ("${s}" - "${u}").`)}else{if(n!=="mixed")throw new Qe(`Graph.${t}: calling this method with only a key (vs. a source and target) does not make sense since an edge with this key could have the other type.`);if(r=""+r,o=this._edges.get(r),!o)throw new De(`Graph.${t}: could not find the "${r}" edge in the graph.`)}return delete o.attributes[a],this.emit("edgeAttributesUpdated",{key:o.key,type:"remove",attributes:o.attributes,name:a}),this}}function kJ(e,t,n){e.prototype[t]=function(r,a){let o;if(this.type!=="mixed"&&n!=="mixed"&&n!==this.type)throw new Qe(`Graph.${t}: cannot find this type of edges in your ${this.type} graph.`);if(arguments.length>2){if(this.multi)throw new Qe(`Graph.${t}: cannot use a {source,target} combo when asking about an edge's attributes in a MultiGraph since we cannot infer the one you want information about.`);const s=""+r,u=""+a;if(a=arguments[2],o=Er(this,s,u,n),!o)throw new De(`Graph.${t}: could not find an edge for the given path ("${s}" - "${u}").`)}else{if(n!=="mixed")throw new Qe(`Graph.${t}: calling this method with only a key (vs. a source and target) does not make sense since an edge with this key could have the other type.`);if(r=""+r,o=this._edges.get(r),!o)throw new De(`Graph.${t}: could not find the "${r}" edge in the graph.`)}if(!yn(a))throw new Le(`Graph.${t}: provided attributes are not a plain object.`);return o.attributes=a,this.emit("edgeAttributesUpdated",{key:o.key,type:"replace",attributes:o.attributes}),this}}function TJ(e,t,n){e.prototype[t]=function(r,a){let o;if(this.type!=="mixed"&&n!=="mixed"&&n!==this.type)throw new Qe(`Graph.${t}: cannot find this type of edges in your ${this.type} graph.`);if(arguments.length>2){if(this.multi)throw new Qe(`Graph.${t}: cannot use a {source,target} combo when asking about an edge's attributes in a MultiGraph since we cannot infer the one you want information about.`);const s=""+r,u=""+a;if(a=arguments[2],o=Er(this,s,u,n),!o)throw new De(`Graph.${t}: could not find an edge for the given path ("${s}" - "${u}").`)}else{if(n!=="mixed")throw new Qe(`Graph.${t}: calling this method with only a key (vs. a source and target) does not make sense since an edge with this key could have the other type.`);if(r=""+r,o=this._edges.get(r),!o)throw new De(`Graph.${t}: could not find the "${r}" edge in the graph.`)}if(!yn(a))throw new Le(`Graph.${t}: provided attributes are not a plain object.`);return nn(o.attributes,a),this.emit("edgeAttributesUpdated",{key:o.key,type:"merge",attributes:o.attributes,data:a}),this}}function AJ(e,t,n){e.prototype[t]=function(r,a){let o;if(this.type!=="mixed"&&n!=="mixed"&&n!==this.type)throw new Qe(`Graph.${t}: cannot find this type of edges in your ${this.type} graph.`);if(arguments.length>2){if(this.multi)throw new Qe(`Graph.${t}: cannot use a {source,target} combo when asking about an edge's attributes in a MultiGraph since we cannot infer the one you want information about.`);const s=""+r,u=""+a;if(a=arguments[2],o=Er(this,s,u,n),!o)throw new De(`Graph.${t}: could not find an edge for the given path ("${s}" - "${u}").`)}else{if(n!=="mixed")throw new Qe(`Graph.${t}: calling this method with only a key (vs. a source and target) does not make sense since an edge with this key could have the other type.`);if(r=""+r,o=this._edges.get(r),!o)throw new De(`Graph.${t}: could not find the "${r}" edge in the graph.`)}if(typeof a!="function")throw new Le(`Graph.${t}: provided updater is not a function.`);return o.attributes=a(o.attributes),this.emit("edgeAttributesUpdated",{key:o.key,type:"update",attributes:o.attributes}),this}}const RJ=[{name:e=>`get${e}Attribute`,attacher:yJ},{name:e=>`get${e}Attributes`,attacher:vJ},{name:e=>`has${e}Attribute`,attacher:SJ},{name:e=>`set${e}Attribute`,attacher:EJ},{name:e=>`update${e}Attribute`,attacher:wJ},{name:e=>`remove${e}Attribute`,attacher:xJ},{name:e=>`replace${e}Attributes`,attacher:kJ},{name:e=>`merge${e}Attributes`,attacher:TJ},{name:e=>`update${e}Attributes`,attacher:AJ}];function _J(e){RJ.forEach(function({name:t,attacher:n}){n(e,t("Edge"),"mixed"),n(e,t("DirectedEdge"),"directed"),n(e,t("UndirectedEdge"),"undirected")})}const CJ=[{name:"edges",type:"mixed"},{name:"inEdges",type:"directed",direction:"in"},{name:"outEdges",type:"directed",direction:"out"},{name:"inboundEdges",type:"mixed",direction:"in"},{name:"outboundEdges",type:"mixed",direction:"out"},{name:"directedEdges",type:"directed"},{name:"undirectedEdges",type:"undirected"}];function NJ(e,t,n,r){let a=!1;for(const o in t){if(o===r)continue;const s=t[o];if(a=n(s.key,s.attributes,s.source.key,s.target.key,s.source.attributes,s.target.attributes,s.undirected),e&&a)return s.key}}function OJ(e,t,n,r){let a,o,s,u=!1;for(const c in t)if(c!==r){a=t[c];do{if(o=a.source,s=a.target,u=n(a.key,a.attributes,o.key,s.key,o.attributes,s.attributes,a.undirected),e&&u)return a.key;a=a.next}while(a!==void 0)}}function pm(e,t){const n=Object.keys(e),r=n.length;let a,o=0;return{[Symbol.iterator](){return this},next(){do if(a)a=a.next;else{if(o>=r)return{done:!0};const s=n[o++];if(s===t){a=void 0;continue}a=e[s]}while(!a);return{done:!1,value:{edge:a.key,attributes:a.attributes,source:a.source.key,target:a.target.key,sourceAttributes:a.source.attributes,targetAttributes:a.target.attributes,undirected:a.undirected}}}}}function DJ(e,t,n,r){const a=t[n];if(!a)return;const o=a.source,s=a.target;if(r(a.key,a.attributes,o.key,s.key,o.attributes,s.attributes,a.undirected)&&e)return a.key}function IJ(e,t,n,r){let a=t[n];if(!a)return;let o=!1;do{if(o=r(a.key,a.attributes,a.source.key,a.target.key,a.source.attributes,a.target.attributes,a.undirected),e&&o)return a.key;a=a.next}while(a!==void 0)}function gm(e,t){let n=e[t];if(n.next!==void 0)return{[Symbol.iterator](){return this},next(){if(!n)return{done:!0};const a={edge:n.key,attributes:n.attributes,source:n.source.key,target:n.target.key,sourceAttributes:n.source.attributes,targetAttributes:n.target.attributes,undirected:n.undirected};return n=n.next,{done:!1,value:a}}};let r=!1;return{[Symbol.iterator](){return this},next(){return r===!0?{done:!0}:(r=!0,{done:!1,value:{edge:n.key,attributes:n.attributes,source:n.source.key,target:n.target.key,sourceAttributes:n.source.attributes,targetAttributes:n.target.attributes,undirected:n.undirected}})}}}function LJ(e,t){if(e.size===0)return[];if(t==="mixed"||t===e.type)return Array.from(e._edges.keys());const n=t==="undirected"?e.undirectedSize:e.directedSize,r=new Array(n),a=t==="undirected",o=e._edges.values();let s=0,u,c;for(;u=o.next(),u.done!==!0;)c=u.value,c.undirected===a&&(r[s++]=c.key);return r}function d4(e,t,n,r){if(t.size===0)return;const a=n!=="mixed"&&n!==t.type,o=n==="undirected";let s,u,c=!1;const d=t._edges.values();for(;s=d.next(),s.done!==!0;){if(u=s.value,a&&u.undirected!==o)continue;const{key:p,attributes:g,source:m,target:b}=u;if(c=r(p,g,m.key,b.key,m.attributes,b.attributes,u.undirected),e&&c)return p}}function MJ(e,t){if(e.size===0)return Ks();const n=t!=="mixed"&&t!==e.type,r=t==="undirected",a=e._edges.values();return{[Symbol.iterator](){return this},next(){let o,s;for(;;){if(o=a.next(),o.done)return o;if(s=o.value,!(n&&s.undirected!==r))break}return{value:{edge:s.key,attributes:s.attributes,source:s.source.key,target:s.target.key,sourceAttributes:s.source.attributes,targetAttributes:s.target.attributes,undirected:s.undirected},done:!1}}}}function fT(e,t,n,r,a,o){const s=t?OJ:NJ;let u;if(n!=="undirected"&&(r!=="out"&&(u=s(e,a.in,o),e&&u)||r!=="in"&&(u=s(e,a.out,o,r?void 0:a.key),e&&u))||n!=="directed"&&(u=s(e,a.undirected,o),e&&u))return u}function FJ(e,t,n,r){const a=[];return fT(!1,e,t,n,r,function(o){a.push(o)}),a}function PJ(e,t,n){let r=Ks();return e!=="undirected"&&(t!=="out"&&typeof n.in<"u"&&(r=Ba(r,pm(n.in))),t!=="in"&&typeof n.out<"u"&&(r=Ba(r,pm(n.out,t?void 0:n.key)))),e!=="directed"&&typeof n.undirected<"u"&&(r=Ba(r,pm(n.undirected))),r}function pT(e,t,n,r,a,o,s){const u=n?IJ:DJ;let c;if(t!=="undirected"&&(typeof a.in<"u"&&r!=="out"&&(c=u(e,a.in,o,s),e&&c)||typeof a.out<"u"&&r!=="in"&&(r||a.key!==o)&&(c=u(e,a.out,o,s),e&&c))||t!=="directed"&&typeof a.undirected<"u"&&(c=u(e,a.undirected,o,s),e&&c))return c}function zJ(e,t,n,r,a){const o=[];return pT(!1,e,t,n,r,a,function(s){o.push(s)}),o}function BJ(e,t,n,r){let a=Ks();return e!=="undirected"&&(typeof n.in<"u"&&t!=="out"&&r in n.in&&(a=Ba(a,gm(n.in,r))),typeof n.out<"u"&&t!=="in"&&r in n.out&&(t||n.key!==r)&&(a=Ba(a,gm(n.out,r)))),e!=="directed"&&typeof n.undirected<"u"&&r in n.undirected&&(a=Ba(a,gm(n.undirected,r))),a}function UJ(e,t){const{name:n,type:r,direction:a}=t;e.prototype[n]=function(o,s){if(r!=="mixed"&&this.type!=="mixed"&&r!==this.type)return[];if(!arguments.length)return LJ(this,r);if(arguments.length===1){o=""+o;const u=this._nodes.get(o);if(typeof u>"u")throw new De(`Graph.${n}: could not find the "${o}" node in the graph.`);return FJ(this.multi,r==="mixed"?this.type:r,a,u)}if(arguments.length===2){o=""+o,s=""+s;const u=this._nodes.get(o);if(!u)throw new De(`Graph.${n}:  could not find the "${o}" source node in the graph.`);if(!this._nodes.has(s))throw new De(`Graph.${n}:  could not find the "${s}" target node in the graph.`);return zJ(r,this.multi,a,u,s)}throw new Le(`Graph.${n}: too many arguments (expecting 0, 1 or 2 and got ${arguments.length}).`)}}function jJ(e,t){const{name:n,type:r,direction:a}=t,o="forEach"+n[0].toUpperCase()+n.slice(1,-1);e.prototype[o]=function(d,p,g){if(!(r!=="mixed"&&this.type!=="mixed"&&r!==this.type)){if(arguments.length===1)return g=d,d4(!1,this,r,g);if(arguments.length===2){d=""+d,g=p;const m=this._nodes.get(d);if(typeof m>"u")throw new De(`Graph.${o}: could not find the "${d}" node in the graph.`);return fT(!1,this.multi,r==="mixed"?this.type:r,a,m,g)}if(arguments.length===3){d=""+d,p=""+p;const m=this._nodes.get(d);if(!m)throw new De(`Graph.${o}:  could not find the "${d}" source node in the graph.`);if(!this._nodes.has(p))throw new De(`Graph.${o}:  could not find the "${p}" target node in the graph.`);return pT(!1,r,this.multi,a,m,p,g)}throw new Le(`Graph.${o}: too many arguments (expecting 1, 2 or 3 and got ${arguments.length}).`)}};const s="map"+n[0].toUpperCase()+n.slice(1);e.prototype[s]=function(){const d=Array.prototype.slice.call(arguments),p=d.pop();let g;if(d.length===0){let m=0;r!=="directed"&&(m+=this.undirectedSize),r!=="undirected"&&(m+=this.directedSize),g=new Array(m);let b=0;d.push((y,S,x,R,k,A,C)=>{g[b++]=p(y,S,x,R,k,A,C)})}else g=[],d.push((m,b,y,S,x,R,k)=>{g.push(p(m,b,y,S,x,R,k))});return this[o].apply(this,d),g};const u="filter"+n[0].toUpperCase()+n.slice(1);e.prototype[u]=function(){const d=Array.prototype.slice.call(arguments),p=d.pop(),g=[];return d.push((m,b,y,S,x,R,k)=>{p(m,b,y,S,x,R,k)&&g.push(m)}),this[o].apply(this,d),g};const c="reduce"+n[0].toUpperCase()+n.slice(1);e.prototype[c]=function(){let d=Array.prototype.slice.call(arguments);if(d.length<2||d.length>4)throw new Le(`Graph.${c}: invalid number of arguments (expecting 2, 3 or 4 and got ${d.length}).`);if(typeof d[d.length-1]=="function"&&typeof d[d.length-2]!="function")throw new Le(`Graph.${c}: missing initial value. You must provide it because the callback takes more than one argument and we cannot infer the initial value from the first iteration, as you could with a simple array.`);let p,g;d.length===2?(p=d[0],g=d[1],d=[]):d.length===3?(p=d[1],g=d[2],d=[d[0]]):d.length===4&&(p=d[2],g=d[3],d=[d[0],d[1]]);let m=g;return d.push((b,y,S,x,R,k,A)=>{m=p(m,b,y,S,x,R,k,A)}),this[o].apply(this,d),m}}function GJ(e,t){const{name:n,type:r,direction:a}=t,o="find"+n[0].toUpperCase()+n.slice(1,-1);e.prototype[o]=function(c,d,p){if(r!=="mixed"&&this.type!=="mixed"&&r!==this.type)return!1;if(arguments.length===1)return p=c,d4(!0,this,r,p);if(arguments.length===2){c=""+c,p=d;const g=this._nodes.get(c);if(typeof g>"u")throw new De(`Graph.${o}: could not find the "${c}" node in the graph.`);return fT(!0,this.multi,r==="mixed"?this.type:r,a,g,p)}if(arguments.length===3){c=""+c,d=""+d;const g=this._nodes.get(c);if(!g)throw new De(`Graph.${o}:  could not find the "${c}" source node in the graph.`);if(!this._nodes.has(d))throw new De(`Graph.${o}:  could not find the "${d}" target node in the graph.`);return pT(!0,r,this.multi,a,g,d,p)}throw new Le(`Graph.${o}: too many arguments (expecting 1, 2 or 3 and got ${arguments.length}).`)};const s="some"+n[0].toUpperCase()+n.slice(1,-1);e.prototype[s]=function(){const c=Array.prototype.slice.call(arguments),d=c.pop();return c.push((g,m,b,y,S,x,R)=>d(g,m,b,y,S,x,R)),!!this[o].apply(this,c)};const u="every"+n[0].toUpperCase()+n.slice(1,-1);e.prototype[u]=function(){const c=Array.prototype.slice.call(arguments),d=c.pop();return c.push((g,m,b,y,S,x,R)=>!d(g,m,b,y,S,x,R)),!this[o].apply(this,c)}}function HJ(e,t){const{name:n,type:r,direction:a}=t,o=n.slice(0,-1)+"Entries";e.prototype[o]=function(s,u){if(r!=="mixed"&&this.type!=="mixed"&&r!==this.type)return Ks();if(!arguments.length)return MJ(this,r);if(arguments.length===1){s=""+s;const c=this._nodes.get(s);if(!c)throw new De(`Graph.${o}: could not find the "${s}" node in the graph.`);return PJ(r,a,c)}if(arguments.length===2){s=""+s,u=""+u;const c=this._nodes.get(s);if(!c)throw new De(`Graph.${o}:  could not find the "${s}" source node in the graph.`);if(!this._nodes.has(u))throw new De(`Graph.${o}:  could not find the "${u}" target node in the graph.`);return BJ(r,a,c,u)}throw new Le(`Graph.${o}: too many arguments (expecting 0, 1 or 2 and got ${arguments.length}).`)}}function $J(e){CJ.forEach(t=>{UJ(e,t),jJ(e,t),GJ(e,t),HJ(e,t)})}const qJ=[{name:"neighbors",type:"mixed"},{name:"inNeighbors",type:"directed",direction:"in"},{name:"outNeighbors",type:"directed",direction:"out"},{name:"inboundNeighbors",type:"mixed",direction:"in"},{name:"outboundNeighbors",type:"mixed",direction:"out"},{name:"directedNeighbors",type:"directed"},{name:"undirectedNeighbors",type:"undirected"}];function vp(){this.A=null,this.B=null}vp.prototype.wrap=function(e){this.A===null?this.A=e:this.B===null&&(this.B=e)};vp.prototype.has=function(e){return this.A!==null&&e in this.A||this.B!==null&&e in this.B};function tu(e,t,n,r,a){for(const o in r){const s=r[o],u=s.source,c=s.target,d=u===n?c:u;if(t&&t.has(d.key))continue;const p=a(d.key,d.attributes);if(e&&p)return d.key}}function gT(e,t,n,r,a){if(t!=="mixed"){if(t==="undirected")return tu(e,null,r,r.undirected,a);if(typeof n=="string")return tu(e,null,r,r[n],a)}const o=new vp;let s;if(t!=="undirected"){if(n!=="out"){if(s=tu(e,null,r,r.in,a),e&&s)return s;o.wrap(r.in)}if(n!=="in"){if(s=tu(e,o,r,r.out,a),e&&s)return s;o.wrap(r.out)}}if(t!=="directed"&&(s=tu(e,o,r,r.undirected,a),e&&s))return s}function VJ(e,t,n){if(e!=="mixed"){if(e==="undirected")return Object.keys(n.undirected);if(typeof t=="string")return Object.keys(n[t])}const r=[];return gT(!1,e,t,n,function(a){r.push(a)}),r}function nu(e,t,n){const r=Object.keys(n),a=r.length;let o=0;return{[Symbol.iterator](){return this},next(){let s=null;do{if(o>=a)return e&&e.wrap(n),{done:!0};const u=n[r[o++]],c=u.source,d=u.target;if(s=c===t?d:c,e&&e.has(s.key)){s=null;continue}}while(s===null);return{done:!1,value:{neighbor:s.key,attributes:s.attributes}}}}}function WJ(e,t,n){if(e!=="mixed"){if(e==="undirected")return nu(null,n,n.undirected);if(typeof t=="string")return nu(null,n,n[t])}let r=Ks();const a=new vp;return e!=="undirected"&&(t!=="out"&&(r=Ba(r,nu(a,n,n.in))),t!=="in"&&(r=Ba(r,nu(a,n,n.out)))),e!=="directed"&&(r=Ba(r,nu(a,n,n.undirected))),r}function YJ(e,t){const{name:n,type:r,direction:a}=t;e.prototype[n]=function(o){if(r!=="mixed"&&this.type!=="mixed"&&r!==this.type)return[];o=""+o;const s=this._nodes.get(o);if(typeof s>"u")throw new De(`Graph.${n}: could not find the "${o}" node in the graph.`);return VJ(r==="mixed"?this.type:r,a,s)}}function KJ(e,t){const{name:n,type:r,direction:a}=t,o="forEach"+n[0].toUpperCase()+n.slice(1,-1);e.prototype[o]=function(d,p){if(r!=="mixed"&&this.type!=="mixed"&&r!==this.type)return;d=""+d;const g=this._nodes.get(d);if(typeof g>"u")throw new De(`Graph.${o}: could not find the "${d}" node in the graph.`);gT(!1,r==="mixed"?this.type:r,a,g,p)};const s="map"+n[0].toUpperCase()+n.slice(1);e.prototype[s]=function(d,p){const g=[];return this[o](d,(m,b)=>{g.push(p(m,b))}),g};const u="filter"+n[0].toUpperCase()+n.slice(1);e.prototype[u]=function(d,p){const g=[];return this[o](d,(m,b)=>{p(m,b)&&g.push(m)}),g};const c="reduce"+n[0].toUpperCase()+n.slice(1);e.prototype[c]=function(d,p,g){if(arguments.length<3)throw new Le(`Graph.${c}: missing initial value. You must provide it because the callback takes more than one argument and we cannot infer the initial value from the first iteration, as you could with a simple array.`);let m=g;return this[o](d,(b,y)=>{m=p(m,b,y)}),m}}function XJ(e,t){const{name:n,type:r,direction:a}=t,o=n[0].toUpperCase()+n.slice(1,-1),s="find"+o;e.prototype[s]=function(d,p){if(r!=="mixed"&&this.type!=="mixed"&&r!==this.type)return;d=""+d;const g=this._nodes.get(d);if(typeof g>"u")throw new De(`Graph.${s}: could not find the "${d}" node in the graph.`);return gT(!0,r==="mixed"?this.type:r,a,g,p)};const u="some"+o;e.prototype[u]=function(d,p){return!!this[s](d,p)};const c="every"+o;e.prototype[c]=function(d,p){return!this[s](d,(m,b)=>!p(m,b))}}function ZJ(e,t){const{name:n,type:r,direction:a}=t,o=n.slice(0,-1)+"Entries";e.prototype[o]=function(s){if(r!=="mixed"&&this.type!=="mixed"&&r!==this.type)return Ks();s=""+s;const u=this._nodes.get(s);if(typeof u>"u")throw new De(`Graph.${o}: could not find the "${s}" node in the graph.`);return WJ(r==="mixed"?this.type:r,a,u)}}function QJ(e){qJ.forEach(t=>{YJ(e,t),KJ(e,t),XJ(e,t),ZJ(e,t)})}function Ad(e,t,n,r,a){const o=r._nodes.values(),s=r.type;let u,c,d,p,g,m;for(;u=o.next(),u.done!==!0;){let b=!1;if(c=u.value,s!=="undirected"){p=c.out;for(d in p){g=p[d];do m=g.target,b=!0,a(c.key,m.key,c.attributes,m.attributes,g.key,g.attributes,g.undirected),g=g.next;while(g)}}if(s!=="directed"){p=c.undirected;for(d in p)if(!(t&&c.key>d)){g=p[d];do m=g.target,m.key!==d&&(m=g.source),b=!0,a(c.key,m.key,c.attributes,m.attributes,g.key,g.attributes,g.undirected),g=g.next;while(g)}}n&&!b&&a(c.key,null,c.attributes,null,null,null,null)}}function JJ(e,t){const n={key:e};return a4(t.attributes)||(n.attributes=nn({},t.attributes)),n}function eee(e,t,n){const r={key:t,source:n.source.key,target:n.target.key};return a4(n.attributes)||(r.attributes=nn({},n.attributes)),e==="mixed"&&n.undirected&&(r.undirected=!0),r}function tee(e){if(!yn(e))throw new Le('Graph.import: invalid serialized node. A serialized node should be a plain object with at least a "key" property.');if(!("key"in e))throw new Le("Graph.import: serialized node is missing its key.");if("attributes"in e&&(!yn(e.attributes)||e.attributes===null))throw new Le("Graph.import: invalid attributes. Attributes should be a plain object, null or omitted.")}function nee(e){if(!yn(e))throw new Le('Graph.import: invalid serialized edge. A serialized edge should be a plain object with at least a "source" & "target" property.');if(!("source"in e))throw new Le("Graph.import: serialized edge is missing its source.");if(!("target"in e))throw new Le("Graph.import: serialized edge is missing its target.");if("attributes"in e&&(!yn(e.attributes)||e.attributes===null))throw new Le("Graph.import: invalid attributes. Attributes should be a plain object, null or omitted.");if("undirected"in e&&typeof e.undirected!="boolean")throw new Le("Graph.import: invalid undirectedness information. Undirected should be boolean or omitted.")}const ree=oJ(),aee=new Set(["directed","undirected","mixed"]),sN=new Set(["domain","_events","_eventsCount","_maxListeners"]),oee=[{name:e=>`${e}Edge`,generateKey:!0},{name:e=>`${e}DirectedEdge`,generateKey:!0,type:"directed"},{name:e=>`${e}UndirectedEdge`,generateKey:!0,type:"undirected"},{name:e=>`${e}EdgeWithKey`},{name:e=>`${e}DirectedEdgeWithKey`,type:"directed"},{name:e=>`${e}UndirectedEdgeWithKey`,type:"undirected"}],iee={allowSelfLoops:!0,multi:!1,type:"mixed"};function see(e,t,n){if(n&&!yn(n))throw new Le(`Graph.addNode: invalid attributes. Expecting an object but got "${n}"`);if(t=""+t,n=n||{},e._nodes.has(t))throw new Qe(`Graph.addNode: the "${t}" node already exist in the graph.`);const r=new e.NodeDataClass(t,n);return e._nodes.set(t,r),e.emit("nodeAdded",{key:t,attributes:n}),r}function lN(e,t,n){const r=new e.NodeDataClass(t,n);return e._nodes.set(t,r),e.emit("nodeAdded",{key:t,attributes:n}),r}function f4(e,t,n,r,a,o,s,u){if(!r&&e.type==="undirected")throw new Qe(`Graph.${t}: you cannot add a directed edge to an undirected graph. Use the #.addEdge or #.addUndirectedEdge instead.`);if(r&&e.type==="directed")throw new Qe(`Graph.${t}: you cannot add an undirected edge to a directed graph. Use the #.addEdge or #.addDirectedEdge instead.`);if(u&&!yn(u))throw new Le(`Graph.${t}: invalid attributes. Expecting an object but got "${u}"`);if(o=""+o,s=""+s,u=u||{},!e.allowSelfLoops&&o===s)throw new Qe(`Graph.${t}: source & target are the same ("${o}"), thus creating a loop explicitly forbidden by this graph 'allowSelfLoops' option set to false.`);const c=e._nodes.get(o),d=e._nodes.get(s);if(!c)throw new De(`Graph.${t}: source node "${o}" not found.`);if(!d)throw new De(`Graph.${t}: target node "${s}" not found.`);const p={key:null,undirected:r,source:o,target:s,attributes:u};if(n)a=e._edgeKeyGenerator();else if(a=""+a,e._edges.has(a))throw new Qe(`Graph.${t}: the "${a}" edge already exists in the graph.`);if(!e.multi&&(r?typeof c.undirected[s]<"u":typeof c.out[s]<"u"))throw new Qe(`Graph.${t}: an edge linking "${o}" to "${s}" already exists. If you really want to add multiple edges linking those nodes, you should create a multi graph by using the 'multi' option.`);const g=new Xs(r,a,c,d,u);e._edges.set(a,g);const m=o===s;return r?(c.undirectedDegree++,d.undirectedDegree++,m&&(c.undirectedLoops++,e._undirectedSelfLoopCount++)):(c.outDegree++,d.inDegree++,m&&(c.directedLoops++,e._directedSelfLoopCount++)),e.multi?g.attachMulti():g.attach(),r?e._undirectedSize++:e._directedSize++,p.key=a,e.emit("edgeAdded",p),a}function lee(e,t,n,r,a,o,s,u,c){if(!r&&e.type==="undirected")throw new Qe(`Graph.${t}: you cannot merge/update a directed edge to an undirected graph. Use the #.mergeEdge/#.updateEdge or #.addUndirectedEdge instead.`);if(r&&e.type==="directed")throw new Qe(`Graph.${t}: you cannot merge/update an undirected edge to a directed graph. Use the #.mergeEdge/#.updateEdge or #.addDirectedEdge instead.`);if(u){if(c){if(typeof u!="function")throw new Le(`Graph.${t}: invalid updater function. Expecting a function but got "${u}"`)}else if(!yn(u))throw new Le(`Graph.${t}: invalid attributes. Expecting an object but got "${u}"`)}o=""+o,s=""+s;let d;if(c&&(d=u,u=void 0),!e.allowSelfLoops&&o===s)throw new Qe(`Graph.${t}: source & target are the same ("${o}"), thus creating a loop explicitly forbidden by this graph 'allowSelfLoops' option set to false.`);let p=e._nodes.get(o),g=e._nodes.get(s),m,b;if(!n&&(m=e._edges.get(a),m)){if((m.source.key!==o||m.target.key!==s)&&(!r||m.source.key!==s||m.target.key!==o))throw new Qe(`Graph.${t}: inconsistency detected when attempting to merge the "${a}" edge with "${o}" source & "${s}" target vs. ("${m.source.key}", "${m.target.key}").`);b=m}if(!b&&!e.multi&&p&&(b=r?p.undirected[s]:p.out[s]),b){const k=[b.key,!1,!1,!1];if(c?!d:!u)return k;if(c){const A=b.attributes;b.attributes=d(A),e.emit("edgeAttributesUpdated",{type:"replace",key:b.key,attributes:b.attributes})}else nn(b.attributes,u),e.emit("edgeAttributesUpdated",{type:"merge",key:b.key,attributes:b.attributes,data:u});return k}u=u||{},c&&d&&(u=d(u));const y={key:null,undirected:r,source:o,target:s,attributes:u};if(n)a=e._edgeKeyGenerator();else if(a=""+a,e._edges.has(a))throw new Qe(`Graph.${t}: the "${a}" edge already exists in the graph.`);let S=!1,x=!1;p||(p=lN(e,o,{}),S=!0,o===s&&(g=p,x=!0)),g||(g=lN(e,s,{}),x=!0),m=new Xs(r,a,p,g,u),e._edges.set(a,m);const R=o===s;return r?(p.undirectedDegree++,g.undirectedDegree++,R&&(p.undirectedLoops++,e._undirectedSelfLoopCount++)):(p.outDegree++,g.inDegree++,R&&(p.directedLoops++,e._directedSelfLoopCount++)),e.multi?m.attachMulti():m.attach(),r?e._undirectedSize++:e._directedSize++,y.key=a,e.emit("edgeAdded",y),[a,!0,S,x]}function fs(e,t){e._edges.delete(t.key);const{source:n,target:r,attributes:a}=t,o=t.undirected,s=n===r;o?(n.undirectedDegree--,r.undirectedDegree--,s&&(n.undirectedLoops--,e._undirectedSelfLoopCount--)):(n.outDegree--,r.inDegree--,s&&(n.directedLoops--,e._directedSelfLoopCount--)),e.multi?t.detachMulti():t.detach(),o?e._undirectedSize--:e._directedSize--,e.emit("edgeDropped",{key:t.key,attributes:a,source:n.key,target:r.key,undirected:o})}class _t extends r4.EventEmitter{constructor(t){if(super(),t=nn({},iee,t),typeof t.multi!="boolean")throw new Le(`Graph.constructor: invalid 'multi' option. Expecting a boolean but got "${t.multi}".`);if(!aee.has(t.type))throw new Le(`Graph.constructor: invalid 'type' option. Should be one of "mixed", "directed" or "undirected" but got "${t.type}".`);if(typeof t.allowSelfLoops!="boolean")throw new Le(`Graph.constructor: invalid 'allowSelfLoops' option. Expecting a boolean but got "${t.allowSelfLoops}".`);const n=t.type==="mixed"?o4:t.type==="directed"?i4:s4;yr(this,"NodeDataClass",n);const r="geid_"+ree()+"_";let a=0;const o=()=>{let s;do s=r+a++;while(this._edges.has(s));return s};yr(this,"_attributes",{}),yr(this,"_nodes",new Map),yr(this,"_edges",new Map),yr(this,"_directedSize",0),yr(this,"_undirectedSize",0),yr(this,"_directedSelfLoopCount",0),yr(this,"_undirectedSelfLoopCount",0),yr(this,"_edgeKeyGenerator",o),yr(this,"_options",t),sN.forEach(s=>yr(this,s,this[s])),Ir(this,"order",()=>this._nodes.size),Ir(this,"size",()=>this._edges.size),Ir(this,"directedSize",()=>this._directedSize),Ir(this,"undirectedSize",()=>this._undirectedSize),Ir(this,"selfLoopCount",()=>this._directedSelfLoopCount+this._undirectedSelfLoopCount),Ir(this,"directedSelfLoopCount",()=>this._directedSelfLoopCount),Ir(this,"undirectedSelfLoopCount",()=>this._undirectedSelfLoopCount),Ir(this,"multi",this._options.multi),Ir(this,"type",this._options.type),Ir(this,"allowSelfLoops",this._options.allowSelfLoops),Ir(this,"implementation",()=>"graphology")}_resetInstanceCounters(){this._directedSize=0,this._undirectedSize=0,this._directedSelfLoopCount=0,this._undirectedSelfLoopCount=0}hasNode(t){return this._nodes.has(""+t)}hasDirectedEdge(t,n){if(this.type==="undirected")return!1;if(arguments.length===1){const r=""+t,a=this._edges.get(r);return!!a&&!a.undirected}else if(arguments.length===2){t=""+t,n=""+n;const r=this._nodes.get(t);return r?r.out.hasOwnProperty(n):!1}throw new Le(`Graph.hasDirectedEdge: invalid arity (${arguments.length}, instead of 1 or 2). You can either ask for an edge id or for the existence of an edge between a source & a target.`)}hasUndirectedEdge(t,n){if(this.type==="directed")return!1;if(arguments.length===1){const r=""+t,a=this._edges.get(r);return!!a&&a.undirected}else if(arguments.length===2){t=""+t,n=""+n;const r=this._nodes.get(t);return r?r.undirected.hasOwnProperty(n):!1}throw new Le(`Graph.hasDirectedEdge: invalid arity (${arguments.length}, instead of 1 or 2). You can either ask for an edge id or for the existence of an edge between a source & a target.`)}hasEdge(t,n){if(arguments.length===1){const r=""+t;return this._edges.has(r)}else if(arguments.length===2){t=""+t,n=""+n;const r=this._nodes.get(t);return r?typeof r.out<"u"&&r.out.hasOwnProperty(n)||typeof r.undirected<"u"&&r.undirected.hasOwnProperty(n):!1}throw new Le(`Graph.hasEdge: invalid arity (${arguments.length}, instead of 1 or 2). You can either ask for an edge id or for the existence of an edge between a source & a target.`)}directedEdge(t,n){if(this.type==="undirected")return;if(t=""+t,n=""+n,this.multi)throw new Qe("Graph.directedEdge: this method is irrelevant with multigraphs since there might be multiple edges between source & target. See #.directedEdges instead.");const r=this._nodes.get(t);if(!r)throw new De(`Graph.directedEdge: could not find the "${t}" source node in the graph.`);if(!this._nodes.has(n))throw new De(`Graph.directedEdge: could not find the "${n}" target node in the graph.`);const a=r.out&&r.out[n]||void 0;if(a)return a.key}undirectedEdge(t,n){if(this.type==="directed")return;if(t=""+t,n=""+n,this.multi)throw new Qe("Graph.undirectedEdge: this method is irrelevant with multigraphs since there might be multiple edges between source & target. See #.undirectedEdges instead.");const r=this._nodes.get(t);if(!r)throw new De(`Graph.undirectedEdge: could not find the "${t}" source node in the graph.`);if(!this._nodes.has(n))throw new De(`Graph.undirectedEdge: could not find the "${n}" target node in the graph.`);const a=r.undirected&&r.undirected[n]||void 0;if(a)return a.key}edge(t,n){if(this.multi)throw new Qe("Graph.edge: this method is irrelevant with multigraphs since there might be multiple edges between source & target. See #.edges instead.");t=""+t,n=""+n;const r=this._nodes.get(t);if(!r)throw new De(`Graph.edge: could not find the "${t}" source node in the graph.`);if(!this._nodes.has(n))throw new De(`Graph.edge: could not find the "${n}" target node in the graph.`);const a=r.out&&r.out[n]||r.undirected&&r.undirected[n]||void 0;if(a)return a.key}areDirectedNeighbors(t,n){t=""+t,n=""+n;const r=this._nodes.get(t);if(!r)throw new De(`Graph.areDirectedNeighbors: could not find the "${t}" node in the graph.`);return this.type==="undirected"?!1:n in r.in||n in r.out}areOutNeighbors(t,n){t=""+t,n=""+n;const r=this._nodes.get(t);if(!r)throw new De(`Graph.areOutNeighbors: could not find the "${t}" node in the graph.`);return this.type==="undirected"?!1:n in r.out}areInNeighbors(t,n){t=""+t,n=""+n;const r=this._nodes.get(t);if(!r)throw new De(`Graph.areInNeighbors: could not find the "${t}" node in the graph.`);return this.type==="undirected"?!1:n in r.in}areUndirectedNeighbors(t,n){t=""+t,n=""+n;const r=this._nodes.get(t);if(!r)throw new De(`Graph.areUndirectedNeighbors: could not find the "${t}" node in the graph.`);return this.type==="directed"?!1:n in r.undirected}areNeighbors(t,n){t=""+t,n=""+n;const r=this._nodes.get(t);if(!r)throw new De(`Graph.areNeighbors: could not find the "${t}" node in the graph.`);return this.type!=="undirected"&&(n in r.in||n in r.out)||this.type!=="directed"&&n in r.undirected}areInboundNeighbors(t,n){t=""+t,n=""+n;const r=this._nodes.get(t);if(!r)throw new De(`Graph.areInboundNeighbors: could not find the "${t}" node in the graph.`);return this.type!=="undirected"&&n in r.in||this.type!=="directed"&&n in r.undirected}areOutboundNeighbors(t,n){t=""+t,n=""+n;const r=this._nodes.get(t);if(!r)throw new De(`Graph.areOutboundNeighbors: could not find the "${t}" node in the graph.`);return this.type!=="undirected"&&n in r.out||this.type!=="directed"&&n in r.undirected}inDegree(t){t=""+t;const n=this._nodes.get(t);if(!n)throw new De(`Graph.inDegree: could not find the "${t}" node in the graph.`);return this.type==="undirected"?0:n.inDegree}outDegree(t){t=""+t;const n=this._nodes.get(t);if(!n)throw new De(`Graph.outDegree: could not find the "${t}" node in the graph.`);return this.type==="undirected"?0:n.outDegree}directedDegree(t){t=""+t;const n=this._nodes.get(t);if(!n)throw new De(`Graph.directedDegree: could not find the "${t}" node in the graph.`);return this.type==="undirected"?0:n.inDegree+n.outDegree}undirectedDegree(t){t=""+t;const n=this._nodes.get(t);if(!n)throw new De(`Graph.undirectedDegree: could not find the "${t}" node in the graph.`);return this.type==="directed"?0:n.undirectedDegree}inboundDegree(t){t=""+t;const n=this._nodes.get(t);if(!n)throw new De(`Graph.inboundDegree: could not find the "${t}" node in the graph.`);let r=0;return this.type!=="directed"&&(r+=n.undirectedDegree),this.type!=="undirected"&&(r+=n.inDegree),r}outboundDegree(t){t=""+t;const n=this._nodes.get(t);if(!n)throw new De(`Graph.outboundDegree: could not find the "${t}" node in the graph.`);let r=0;return this.type!=="directed"&&(r+=n.undirectedDegree),this.type!=="undirected"&&(r+=n.outDegree),r}degree(t){t=""+t;const n=this._nodes.get(t);if(!n)throw new De(`Graph.degree: could not find the "${t}" node in the graph.`);let r=0;return this.type!=="directed"&&(r+=n.undirectedDegree),this.type!=="undirected"&&(r+=n.inDegree+n.outDegree),r}inDegreeWithoutSelfLoops(t){t=""+t;const n=this._nodes.get(t);if(!n)throw new De(`Graph.inDegreeWithoutSelfLoops: could not find the "${t}" node in the graph.`);return this.type==="undirected"?0:n.inDegree-n.directedLoops}outDegreeWithoutSelfLoops(t){t=""+t;const n=this._nodes.get(t);if(!n)throw new De(`Graph.outDegreeWithoutSelfLoops: could not find the "${t}" node in the graph.`);return this.type==="undirected"?0:n.outDegree-n.directedLoops}directedDegreeWithoutSelfLoops(t){t=""+t;const n=this._nodes.get(t);if(!n)throw new De(`Graph.directedDegreeWithoutSelfLoops: could not find the "${t}" node in the graph.`);return this.type==="undirected"?0:n.inDegree+n.outDegree-n.directedLoops*2}undirectedDegreeWithoutSelfLoops(t){t=""+t;const n=this._nodes.get(t);if(!n)throw new De(`Graph.undirectedDegreeWithoutSelfLoops: could not find the "${t}" node in the graph.`);return this.type==="directed"?0:n.undirectedDegree-n.undirectedLoops*2}inboundDegreeWithoutSelfLoops(t){t=""+t;const n=this._nodes.get(t);if(!n)throw new De(`Graph.inboundDegreeWithoutSelfLoops: could not find the "${t}" node in the graph.`);let r=0,a=0;return this.type!=="directed"&&(r+=n.undirectedDegree,a+=n.undirectedLoops*2),this.type!=="undirected"&&(r+=n.inDegree,a+=n.directedLoops),r-a}outboundDegreeWithoutSelfLoops(t){t=""+t;const n=this._nodes.get(t);if(!n)throw new De(`Graph.outboundDegreeWithoutSelfLoops: could not find the "${t}" node in the graph.`);let r=0,a=0;return this.type!=="directed"&&(r+=n.undirectedDegree,a+=n.undirectedLoops*2),this.type!=="undirected"&&(r+=n.outDegree,a+=n.directedLoops),r-a}degreeWithoutSelfLoops(t){t=""+t;const n=this._nodes.get(t);if(!n)throw new De(`Graph.degreeWithoutSelfLoops: could not find the "${t}" node in the graph.`);let r=0,a=0;return this.type!=="directed"&&(r+=n.undirectedDegree,a+=n.undirectedLoops*2),this.type!=="undirected"&&(r+=n.inDegree+n.outDegree,a+=n.directedLoops*2),r-a}source(t){t=""+t;const n=this._edges.get(t);if(!n)throw new De(`Graph.source: could not find the "${t}" edge in the graph.`);return n.source.key}target(t){t=""+t;const n=this._edges.get(t);if(!n)throw new De(`Graph.target: could not find the "${t}" edge in the graph.`);return n.target.key}extremities(t){t=""+t;const n=this._edges.get(t);if(!n)throw new De(`Graph.extremities: could not find the "${t}" edge in the graph.`);return[n.source.key,n.target.key]}opposite(t,n){t=""+t,n=""+n;const r=this._edges.get(n);if(!r)throw new De(`Graph.opposite: could not find the "${n}" edge in the graph.`);const a=r.source.key,o=r.target.key;if(t===a)return o;if(t===o)return a;throw new De(`Graph.opposite: the "${t}" node is not attached to the "${n}" edge (${a}, ${o}).`)}hasExtremity(t,n){t=""+t,n=""+n;const r=this._edges.get(t);if(!r)throw new De(`Graph.hasExtremity: could not find the "${t}" edge in the graph.`);return r.source.key===n||r.target.key===n}isUndirected(t){t=""+t;const n=this._edges.get(t);if(!n)throw new De(`Graph.isUndirected: could not find the "${t}" edge in the graph.`);return n.undirected}isDirected(t){t=""+t;const n=this._edges.get(t);if(!n)throw new De(`Graph.isDirected: could not find the "${t}" edge in the graph.`);return!n.undirected}isSelfLoop(t){t=""+t;const n=this._edges.get(t);if(!n)throw new De(`Graph.isSelfLoop: could not find the "${t}" edge in the graph.`);return n.source===n.target}addNode(t,n){return see(this,t,n).key}mergeNode(t,n){if(n&&!yn(n))throw new Le(`Graph.mergeNode: invalid attributes. Expecting an object but got "${n}"`);t=""+t,n=n||{};let r=this._nodes.get(t);return r?(n&&(nn(r.attributes,n),this.emit("nodeAttributesUpdated",{type:"merge",key:t,attributes:r.attributes,data:n})),[t,!1]):(r=new this.NodeDataClass(t,n),this._nodes.set(t,r),this.emit("nodeAdded",{key:t,attributes:n}),[t,!0])}updateNode(t,n){if(n&&typeof n!="function")throw new Le(`Graph.updateNode: invalid updater function. Expecting a function but got "${n}"`);t=""+t;let r=this._nodes.get(t);if(r){if(n){const o=r.attributes;r.attributes=n(o),this.emit("nodeAttributesUpdated",{type:"replace",key:t,attributes:r.attributes})}return[t,!1]}const a=n?n({}):{};return r=new this.NodeDataClass(t,a),this._nodes.set(t,r),this.emit("nodeAdded",{key:t,attributes:a}),[t,!0]}dropNode(t){t=""+t;const n=this._nodes.get(t);if(!n)throw new De(`Graph.dropNode: could not find the "${t}" node in the graph.`);let r;if(this.type!=="undirected"){for(const a in n.out){r=n.out[a];do fs(this,r),r=r.next;while(r)}for(const a in n.in){r=n.in[a];do fs(this,r),r=r.next;while(r)}}if(this.type!=="directed")for(const a in n.undirected){r=n.undirected[a];do fs(this,r),r=r.next;while(r)}this._nodes.delete(t),this.emit("nodeDropped",{key:t,attributes:n.attributes})}dropEdge(t){let n;if(arguments.length>1){const r=""+arguments[0],a=""+arguments[1];if(n=Er(this,r,a,this.type),!n)throw new De(`Graph.dropEdge: could not find the "${r}" -> "${a}" edge in the graph.`)}else if(t=""+t,n=this._edges.get(t),!n)throw new De(`Graph.dropEdge: could not find the "${t}" edge in the graph.`);return fs(this,n),this}dropDirectedEdge(t,n){if(arguments.length<2)throw new Qe("Graph.dropDirectedEdge: it does not make sense to try and drop a directed edge by key. What if the edge with this key is undirected? Use #.dropEdge for this purpose instead.");if(this.multi)throw new Qe("Graph.dropDirectedEdge: cannot use a {source,target} combo when dropping an edge in a MultiGraph since we cannot infer the one you want to delete as there could be multiple ones.");t=""+t,n=""+n;const r=Er(this,t,n,"directed");if(!r)throw new De(`Graph.dropDirectedEdge: could not find a "${t}" -> "${n}" edge in the graph.`);return fs(this,r),this}dropUndirectedEdge(t,n){if(arguments.length<2)throw new Qe("Graph.dropUndirectedEdge: it does not make sense to drop a directed edge by key. What if the edge with this key is undirected? Use #.dropEdge for this purpose instead.");if(this.multi)throw new Qe("Graph.dropUndirectedEdge: cannot use a {source,target} combo when dropping an edge in a MultiGraph since we cannot infer the one you want to delete as there could be multiple ones.");const r=Er(this,t,n,"undirected");if(!r)throw new De(`Graph.dropUndirectedEdge: could not find a "${t}" -> "${n}" edge in the graph.`);return fs(this,r),this}clear(){this._edges.clear(),this._nodes.clear(),this._resetInstanceCounters(),this.emit("cleared")}clearEdges(){const t=this._nodes.values();let n;for(;n=t.next(),n.done!==!0;)n.value.clear();this._edges.clear(),this._resetInstanceCounters(),this.emit("edgesCleared")}getAttribute(t){return this._attributes[t]}getAttributes(){return this._attributes}hasAttribute(t){return this._attributes.hasOwnProperty(t)}setAttribute(t,n){return this._attributes[t]=n,this.emit("attributesUpdated",{type:"set",attributes:this._attributes,name:t}),this}updateAttribute(t,n){if(typeof n!="function")throw new Le("Graph.updateAttribute: updater should be a function.");const r=this._attributes[t];return this._attributes[t]=n(r),this.emit("attributesUpdated",{type:"set",attributes:this._attributes,name:t}),this}removeAttribute(t){return delete this._attributes[t],this.emit("attributesUpdated",{type:"remove",attributes:this._attributes,name:t}),this}replaceAttributes(t){if(!yn(t))throw new Le("Graph.replaceAttributes: provided attributes are not a plain object.");return this._attributes=t,this.emit("attributesUpdated",{type:"replace",attributes:this._attributes}),this}mergeAttributes(t){if(!yn(t))throw new Le("Graph.mergeAttributes: provided attributes are not a plain object.");return nn(this._attributes,t),this.emit("attributesUpdated",{type:"merge",attributes:this._attributes,data:t}),this}updateAttributes(t){if(typeof t!="function")throw new Le("Graph.updateAttributes: provided updater is not a function.");return this._attributes=t(this._attributes),this.emit("attributesUpdated",{type:"update",attributes:this._attributes}),this}updateEachNodeAttributes(t,n){if(typeof t!="function")throw new Le("Graph.updateEachNodeAttributes: expecting an updater function.");if(n&&!iN(n))throw new Le("Graph.updateEachNodeAttributes: invalid hints. Expecting an object having the following shape: {attributes?: [string]}");const r=this._nodes.values();let a,o;for(;a=r.next(),a.done!==!0;)o=a.value,o.attributes=t(o.key,o.attributes);this.emit("eachNodeAttributesUpdated",{hints:n||null})}updateEachEdgeAttributes(t,n){if(typeof t!="function")throw new Le("Graph.updateEachEdgeAttributes: expecting an updater function.");if(n&&!iN(n))throw new Le("Graph.updateEachEdgeAttributes: invalid hints. Expecting an object having the following shape: {attributes?: [string]}");const r=this._edges.values();let a,o,s,u;for(;a=r.next(),a.done!==!0;)o=a.value,s=o.source,u=o.target,o.attributes=t(o.key,o.attributes,s.key,u.key,s.attributes,u.attributes,o.undirected);this.emit("eachEdgeAttributesUpdated",{hints:n||null})}forEachAdjacencyEntry(t){if(typeof t!="function")throw new Le("Graph.forEachAdjacencyEntry: expecting a callback.");Ad(!1,!1,!1,this,t)}forEachAdjacencyEntryWithOrphans(t){if(typeof t!="function")throw new Le("Graph.forEachAdjacencyEntryWithOrphans: expecting a callback.");Ad(!1,!1,!0,this,t)}forEachAssymetricAdjacencyEntry(t){if(typeof t!="function")throw new Le("Graph.forEachAssymetricAdjacencyEntry: expecting a callback.");Ad(!1,!0,!1,this,t)}forEachAssymetricAdjacencyEntryWithOrphans(t){if(typeof t!="function")throw new Le("Graph.forEachAssymetricAdjacencyEntryWithOrphans: expecting a callback.");Ad(!1,!0,!0,this,t)}nodes(){return Array.from(this._nodes.keys())}forEachNode(t){if(typeof t!="function")throw new Le("Graph.forEachNode: expecting a callback.");const n=this._nodes.values();let r,a;for(;r=n.next(),r.done!==!0;)a=r.value,t(a.key,a.attributes)}findNode(t){if(typeof t!="function")throw new Le("Graph.findNode: expecting a callback.");const n=this._nodes.values();let r,a;for(;r=n.next(),r.done!==!0;)if(a=r.value,t(a.key,a.attributes))return a.key}mapNodes(t){if(typeof t!="function")throw new Le("Graph.mapNode: expecting a callback.");const n=this._nodes.values();let r,a;const o=new Array(this.order);let s=0;for(;r=n.next(),r.done!==!0;)a=r.value,o[s++]=t(a.key,a.attributes);return o}someNode(t){if(typeof t!="function")throw new Le("Graph.someNode: expecting a callback.");const n=this._nodes.values();let r,a;for(;r=n.next(),r.done!==!0;)if(a=r.value,t(a.key,a.attributes))return!0;return!1}everyNode(t){if(typeof t!="function")throw new Le("Graph.everyNode: expecting a callback.");const n=this._nodes.values();let r,a;for(;r=n.next(),r.done!==!0;)if(a=r.value,!t(a.key,a.attributes))return!1;return!0}filterNodes(t){if(typeof t!="function")throw new Le("Graph.filterNodes: expecting a callback.");const n=this._nodes.values();let r,a;const o=[];for(;r=n.next(),r.done!==!0;)a=r.value,t(a.key,a.attributes)&&o.push(a.key);return o}reduceNodes(t,n){if(typeof t!="function")throw new Le("Graph.reduceNodes: expecting a callback.");if(arguments.length<2)throw new Le("Graph.reduceNodes: missing initial value. You must provide it because the callback takes more than one argument and we cannot infer the initial value from the first iteration, as you could with a simple array.");let r=n;const a=this._nodes.values();let o,s;for(;o=a.next(),o.done!==!0;)s=o.value,r=t(r,s.key,s.attributes);return r}nodeEntries(){const t=this._nodes.values();return{[Symbol.iterator](){return this},next(){const n=t.next();if(n.done)return n;const r=n.value;return{value:{node:r.key,attributes:r.attributes},done:!1}}}}export(){const t=new Array(this._nodes.size);let n=0;this._nodes.forEach((a,o)=>{t[n++]=JJ(o,a)});const r=new Array(this._edges.size);return n=0,this._edges.forEach((a,o)=>{r[n++]=eee(this.type,o,a)}),{options:{type:this.type,multi:this.multi,allowSelfLoops:this.allowSelfLoops},attributes:this.getAttributes(),nodes:t,edges:r}}import(t,n=!1){if(t instanceof _t)return t.forEachNode((c,d)=>{n?this.mergeNode(c,d):this.addNode(c,d)}),t.forEachEdge((c,d,p,g,m,b,y)=>{n?y?this.mergeUndirectedEdgeWithKey(c,p,g,d):this.mergeDirectedEdgeWithKey(c,p,g,d):y?this.addUndirectedEdgeWithKey(c,p,g,d):this.addDirectedEdgeWithKey(c,p,g,d)}),this;if(!yn(t))throw new Le("Graph.import: invalid argument. Expecting a serialized graph or, alternatively, a Graph instance.");if(t.attributes){if(!yn(t.attributes))throw new Le("Graph.import: invalid attributes. Expecting a plain object.");n?this.mergeAttributes(t.attributes):this.replaceAttributes(t.attributes)}let r,a,o,s,u;if(t.nodes){if(o=t.nodes,!Array.isArray(o))throw new Le("Graph.import: invalid nodes. Expecting an array.");for(r=0,a=o.length;r<a;r++){s=o[r],tee(s);const{key:c,attributes:d}=s;n?this.mergeNode(c,d):this.addNode(c,d)}}if(t.edges){let c=!1;if(this.type==="undirected"&&(c=!0),o=t.edges,!Array.isArray(o))throw new Le("Graph.import: invalid edges. Expecting an array.");for(r=0,a=o.length;r<a;r++){u=o[r],nee(u);const{source:d,target:p,attributes:g,undirected:m=c}=u;let b;"key"in u?(b=n?m?this.mergeUndirectedEdgeWithKey:this.mergeDirectedEdgeWithKey:m?this.addUndirectedEdgeWithKey:this.addDirectedEdgeWithKey,b.call(this,u.key,d,p,g)):(b=n?m?this.mergeUndirectedEdge:this.mergeDirectedEdge:m?this.addUndirectedEdge:this.addDirectedEdge,b.call(this,d,p,g))}}return this}nullCopy(t){const n=new _t(nn({},this._options,t));return n.replaceAttributes(nn({},this.getAttributes())),n}emptyCopy(t){const n=this.nullCopy(t);return this._nodes.forEach((r,a)=>{const o=nn({},r.attributes);r=new n.NodeDataClass(a,o),n._nodes.set(a,r)}),n}copy(t){if(t=t||{},typeof t.type=="string"&&t.type!==this.type&&t.type!=="mixed")throw new Qe(`Graph.copy: cannot create an incompatible copy from "${this.type}" type to "${t.type}" because this would mean losing information about the current graph.`);if(typeof t.multi=="boolean"&&t.multi!==this.multi&&t.multi!==!0)throw new Qe("Graph.copy: cannot create an incompatible copy by downgrading a multi graph to a simple one because this would mean losing information about the current graph.");if(typeof t.allowSelfLoops=="boolean"&&t.allowSelfLoops!==this.allowSelfLoops&&t.allowSelfLoops!==!0)throw new Qe("Graph.copy: cannot create an incompatible copy from a graph allowing self loops to one that does not because this would mean losing information about the current graph.");const n=this.emptyCopy(t),r=this._edges.values();let a,o;for(;a=r.next(),a.done!==!0;)o=a.value,f4(n,"copy",!1,o.undirected,o.key,o.source.key,o.target.key,nn({},o.attributes));return n}toJSON(){return this.export()}toString(){return"[object Graph]"}inspect(){const t={};this._nodes.forEach((o,s)=>{t[s]=o.attributes});const n={},r={};this._edges.forEach((o,s)=>{const u=o.undirected?"--":"->";let c="",d=o.source.key,p=o.target.key,g;o.undirected&&d>p&&(g=d,d=p,p=g);const m=`(${d})${u}(${p})`;s.startsWith("geid_")?this.multi&&(typeof r[m]>"u"?r[m]=0:r[m]++,c+=`${r[m]}. `):c+=`[${s}]: `,c+=m,n[c]=o.attributes});const a={};for(const o in this)this.hasOwnProperty(o)&&!sN.has(o)&&typeof this[o]!="function"&&typeof o!="symbol"&&(a[o]=this[o]);return a.attributes=this._attributes,a.nodes=t,a.edges=n,yr(a,"constructor",this.constructor),a}}typeof Symbol<"u"&&(_t.prototype[Symbol.for("nodejs.util.inspect.custom")]=_t.prototype.inspect);oee.forEach(e=>{["add","merge","update"].forEach(t=>{const n=e.name(t),r=t==="add"?f4:lee;e.generateKey?_t.prototype[n]=function(a,o,s){return r(this,n,!0,(e.type||this.type)==="undirected",null,a,o,s,t==="update")}:_t.prototype[n]=function(a,o,s,u){return r(this,n,!1,(e.type||this.type)==="undirected",a,o,s,u,t==="update")}})});bJ(_t);_J(_t);$J(_t);QJ(_t);class Sp extends _t{constructor(t){const n=nn({type:"directed"},t);if("multi"in n&&n.multi!==!1)throw new Le("DirectedGraph.from: inconsistent indication that the graph should be multi in given options!");if(n.type!=="directed")throw new Le('DirectedGraph.from: inconsistent "'+n.type+'" type in given options!');super(n)}}class p4 extends _t{constructor(t){const n=nn({type:"undirected"},t);if("multi"in n&&n.multi!==!1)throw new Le("UndirectedGraph.from: inconsistent indication that the graph should be multi in given options!");if(n.type!=="undirected")throw new Le('UndirectedGraph.from: inconsistent "'+n.type+'" type in given options!');super(n)}}class g4 extends _t{constructor(t){const n=nn({multi:!0},t);if("multi"in n&&n.multi!==!0)throw new Le("MultiGraph.from: inconsistent indication that the graph should be simple in given options!");super(n)}}class h4 extends _t{constructor(t){const n=nn({type:"directed",multi:!0},t);if("multi"in n&&n.multi!==!0)throw new Le("MultiDirectedGraph.from: inconsistent indication that the graph should be simple in given options!");if(n.type!=="directed")throw new Le('MultiDirectedGraph.from: inconsistent "'+n.type+'" type in given options!');super(n)}}class m4 extends _t{constructor(t){const n=nn({type:"undirected",multi:!0},t);if("multi"in n&&n.multi!==!0)throw new Le("MultiUndirectedGraph.from: inconsistent indication that the graph should be simple in given options!");if(n.type!=="undirected")throw new Le('MultiUndirectedGraph.from: inconsistent "'+n.type+'" type in given options!');super(n)}}function Zs(e){e.from=function(t,n){const r=nn({},t.options,n),a=new e(r);return a.import(t),a}}Zs(_t);Zs(Sp);Zs(p4);Zs(g4);Zs(h4);Zs(m4);_t.Graph=_t;_t.DirectedGraph=Sp;_t.UndirectedGraph=p4;_t.MultiGraph=g4;_t.MultiDirectedGraph=h4;_t.MultiUndirectedGraph=m4;_t.InvalidArgumentsGraphError=Le;_t.NotFoundGraphError=De;_t.UsageGraphError=Qe;function uee(e,t){if(typeof e!="object"||!e)return e;var n=e[Symbol.toPrimitive];if(n!==void 0){var r=n.call(e,t);if(typeof r!="object")return r;throw new TypeError("@@toPrimitive must return a primitive value.")}return String(e)}function hu(e){var t=uee(e,"string");return typeof t=="symbol"?t:t+""}function cn(e,t){if(!(e instanceof t))throw new TypeError("Cannot call a class as a function")}function uN(e,t){for(var n=0;n<t.length;n++){var r=t[n];r.enumerable=r.enumerable||!1,r.configurable=!0,"value"in r&&(r.writable=!0),Object.defineProperty(e,hu(r.key),r)}}function dn(e,t,n){return t&&uN(e.prototype,t),n&&uN(e,n),Object.defineProperty(e,"prototype",{writable:!1}),e}function Is(e){return Is=Object.setPrototypeOf?Object.getPrototypeOf.bind():function(t){return t.__proto__||Object.getPrototypeOf(t)},Is(e)}function b4(){try{var e=!Boolean.prototype.valueOf.call(Reflect.construct(Boolean,[],function(){}))}catch{}return(b4=function(){return!!e})()}function cee(e){if(e===void 0)throw new ReferenceError("this hasn't been initialised - super() hasn't been called");return e}function dee(e,t){if(t&&(typeof t=="object"||typeof t=="function"))return t;if(t!==void 0)throw new TypeError("Derived constructors may only return object or undefined");return cee(e)}function $n(e,t,n){return t=Is(t),dee(e,b4()?Reflect.construct(t,n||[],Is(e).constructor):t.apply(e,n))}function Fk(e,t){return Fk=Object.setPrototypeOf?Object.setPrototypeOf.bind():function(n,r){return n.__proto__=r,n},Fk(e,t)}function qn(e,t){if(typeof t!="function"&&t!==null)throw new TypeError("Super expression must either be null or a function");e.prototype=Object.create(t&&t.prototype,{constructor:{value:e,writable:!0,configurable:!0}}),Object.defineProperty(e,"prototype",{writable:!1}),t&&Fk(e,t)}function fee(e){if(Array.isArray(e))return e}function pee(e,t){var n=e==null?null:typeof Symbol<"u"&&e[Symbol.iterator]||e["@@iterator"];if(n!=null){var r,a,o,s,u=[],c=!0,d=!1;try{if(o=(n=n.call(e)).next,t===0){if(Object(n)!==n)return;c=!1}else for(;!(c=(r=o.call(n)).done)&&(u.push(r.value),u.length!==t);c=!0);}catch(p){d=!0,a=p}finally{try{if(!c&&n.return!=null&&(s=n.return(),Object(s)!==s))return}finally{if(d)throw a}}return u}}function Pk(e,t){(t==null||t>e.length)&&(t=e.length);for(var n=0,r=Array(t);n<t;n++)r[n]=e[n];return r}function y4(e,t){if(e){if(typeof e=="string")return Pk(e,t);var n={}.toString.call(e).slice(8,-1);return n==="Object"&&e.constructor&&(n=e.constructor.name),n==="Map"||n==="Set"?Array.from(e):n==="Arguments"||/^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n)?Pk(e,t):void 0}}function gee(){throw new TypeError(`Invalid attempt to destructure non-iterable instance.
In order to be iterable, non-array objects must have a [Symbol.iterator]() method.`)}function Ls(e,t){return fee(e)||pee(e,t)||y4(e,t)||gee()}var hm={black:"#000000",silver:"#C0C0C0",gray:"#808080",grey:"#808080",white:"#FFFFFF",maroon:"#800000",red:"#FF0000",purple:"#800080",fuchsia:"#FF00FF",green:"#008000",lime:"#00FF00",olive:"#808000",yellow:"#FFFF00",navy:"#000080",blue:"#0000FF",teal:"#008080",aqua:"#00FFFF",darkblue:"#00008B",mediumblue:"#0000CD",darkgreen:"#006400",darkcyan:"#008B8B",deepskyblue:"#00BFFF",darkturquoise:"#00CED1",mediumspringgreen:"#00FA9A",springgreen:"#00FF7F",cyan:"#00FFFF",midnightblue:"#191970",dodgerblue:"#1E90FF",lightseagreen:"#20B2AA",forestgreen:"#228B22",seagreen:"#2E8B57",darkslategray:"#2F4F4F",darkslategrey:"#2F4F4F",limegreen:"#32CD32",mediumseagreen:"#3CB371",turquoise:"#40E0D0",royalblue:"#4169E1",steelblue:"#4682B4",darkslateblue:"#483D8B",mediumturquoise:"#48D1CC",indigo:"#4B0082",darkolivegreen:"#556B2F",cadetblue:"#5F9EA0",cornflowerblue:"#6495ED",rebeccapurple:"#663399",mediumaquamarine:"#66CDAA",dimgray:"#696969",dimgrey:"#696969",slateblue:"#6A5ACD",olivedrab:"#6B8E23",slategray:"#708090",slategrey:"#708090",lightslategray:"#778899",lightslategrey:"#778899",mediumslateblue:"#7B68EE",lawngreen:"#7CFC00",chartreuse:"#7FFF00",aquamarine:"#7FFFD4",skyblue:"#87CEEB",lightskyblue:"#87CEFA",blueviolet:"#8A2BE2",darkred:"#8B0000",darkmagenta:"#8B008B",saddlebrown:"#8B4513",darkseagreen:"#8FBC8F",lightgreen:"#90EE90",mediumpurple:"#9370DB",darkviolet:"#9400D3",palegreen:"#98FB98",darkorchid:"#9932CC",yellowgreen:"#9ACD32",sienna:"#A0522D",brown:"#A52A2A",darkgray:"#A9A9A9",darkgrey:"#A9A9A9",lightblue:"#ADD8E6",greenyellow:"#ADFF2F",paleturquoise:"#AFEEEE",lightsteelblue:"#B0C4DE",powderblue:"#B0E0E6",firebrick:"#B22222",darkgoldenrod:"#B8860B",mediumorchid:"#BA55D3",rosybrown:"#BC8F8F",darkkhaki:"#BDB76B",mediumvioletred:"#C71585",indianred:"#CD5C5C",peru:"#CD853F",chocolate:"#D2691E",tan:"#D2B48C",lightgray:"#D3D3D3",lightgrey:"#D3D3D3",thistle:"#D8BFD8",orchid:"#DA70D6",goldenrod:"#DAA520",palevioletred:"#DB7093",crimson:"#DC143C",gainsboro:"#DCDCDC",plum:"#DDA0DD",burlywood:"#DEB887",lightcyan:"#E0FFFF",lavender:"#E6E6FA",darksalmon:"#E9967A",violet:"#EE82EE",palegoldenrod:"#EEE8AA",lightcoral:"#F08080",khaki:"#F0E68C",aliceblue:"#F0F8FF",honeydew:"#F0FFF0",azure:"#F0FFFF",sandybrown:"#F4A460",wheat:"#F5DEB3",beige:"#F5F5DC",whitesmoke:"#F5F5F5",mintcream:"#F5FFFA",ghostwhite:"#F8F8FF",salmon:"#FA8072",antiquewhite:"#FAEBD7",linen:"#FAF0E6",lightgoldenrodyellow:"#FAFAD2",oldlace:"#FDF5E6",magenta:"#FF00FF",deeppink:"#FF1493",orangered:"#FF4500",tomato:"#FF6347",hotpink:"#FF69B4",coral:"#FF7F50",darkorange:"#FF8C00",lightsalmon:"#FFA07A",orange:"#FFA500",lightpink:"#FFB6C1",pink:"#FFC0CB",gold:"#FFD700",peachpuff:"#FFDAB9",navajowhite:"#FFDEAD",moccasin:"#FFE4B5",bisque:"#FFE4C4",mistyrose:"#FFE4E1",blanchedalmond:"#FFEBCD",papayawhip:"#FFEFD5",lavenderblush:"#FFF0F5",seashell:"#FFF5EE",cornsilk:"#FFF8DC",lemonchiffon:"#FFFACD",floralwhite:"#FFFAF0",snow:"#FFFAFA",lightyellow:"#FFFFE0",ivory:"#FFFFF0"},v4=new Int8Array(4),Jd=new Int32Array(v4.buffer,0,1),S4=new Float32Array(v4.buffer,0,1),hee=/^\s*rgba?\s*\(/,mee=/^\s*rgba?\s*\(\s*([0-9]*)\s*,\s*([0-9]*)\s*,\s*([0-9]*)(?:\s*,\s*(.*)?)?\)\s*$/;function bee(e){var t=0,n=0,r=0,a=1;if(e[0]==="#")e.length===4?(t=parseInt(e.charAt(1)+e.charAt(1),16),n=parseInt(e.charAt(2)+e.charAt(2),16),r=parseInt(e.charAt(3)+e.charAt(3),16)):(t=parseInt(e.charAt(1)+e.charAt(2),16),n=parseInt(e.charAt(3)+e.charAt(4),16),r=parseInt(e.charAt(5)+e.charAt(6),16)),e.length===9&&(a=parseInt(e.charAt(7)+e.charAt(8),16)/255);else if(hee.test(e)){var o=e.match(mee);o&&(t=+o[1],n=+o[2],r=+o[3],o[4]&&(a=+o[4]))}return{r:t,g:n,b:r,a}}var Rs={};for(var Rd in hm)Rs[Rd]=la(hm[Rd]),Rs[hm[Rd]]=Rs[Rd];function E4(e,t,n,r,a){return Jd[0]=r<<24|n<<16|t<<8|e,Jd[0]=Jd[0]&4278190079,S4[0]}function la(e){if(e=e.toLowerCase(),typeof Rs[e]<"u")return Rs[e];var t=bee(e),n=t.r,r=t.g,a=t.b,o=t.a;o=o*255|0;var s=E4(n,r,a,o);return Rs[e]=s,s}function yee(e,t){S4[0]=la(e);var n=Jd[0],r=n&255,a=n>>8&255,o=n>>16&255,s=n>>24&255;return[r,a,o,s]}var mm={};function w4(e){if(typeof mm[e]<"u")return mm[e];var t=(e&16711680)>>>16,n=(e&65280)>>>8,r=e&255,a=255,o=E4(t,n,r,a);return mm[e]=o,o}function cN(e,t,n,r){return n+(t<<8)+(e<<16)}function dN(e,t,n,r,a,o){var s=Math.floor(n/o*a),u=Math.floor(e.drawingBufferHeight/o-r/o*a),c=new Uint8Array(4);e.bindFramebuffer(e.FRAMEBUFFER,t),e.readPixels(s,u,1,1,e.RGBA,e.UNSIGNED_BYTE,c);var d=Ls(c,4),p=d[0],g=d[1],m=d[2],b=d[3];return[p,g,m,b]}function Re(e,t,n){return(t=hu(t))in e?Object.defineProperty(e,t,{value:n,enumerable:!0,configurable:!0,writable:!0}):e[t]=n,e}function fN(e,t){var n=Object.keys(e);if(Object.getOwnPropertySymbols){var r=Object.getOwnPropertySymbols(e);t&&(r=r.filter(function(a){return Object.getOwnPropertyDescriptor(e,a).enumerable})),n.push.apply(n,r)}return n}function je(e){for(var t=1;t<arguments.length;t++){var n=arguments[t]!=null?arguments[t]:{};t%2?fN(Object(n),!0).forEach(function(r){Re(e,r,n[r])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(n)):fN(Object(n)).forEach(function(r){Object.defineProperty(e,r,Object.getOwnPropertyDescriptor(n,r))})}return e}function vee(e,t){for(;!{}.hasOwnProperty.call(e,t)&&(e=Is(e))!==null;);return e}function zk(){return zk=typeof Reflect<"u"&&Reflect.get?Reflect.get.bind():function(e,t,n){var r=vee(e,t);if(r){var a=Object.getOwnPropertyDescriptor(r,t);return a.get?a.get.call(arguments.length<3?e:n):a.value}},zk.apply(null,arguments)}function x4(e,t,n,r){var a=zk(Is(e.prototype),t,n);return typeof a=="function"?function(o){return a.apply(n,o)}:a}function See(e){return e.normalized?1:e.size}function bm(e){var t=0;return e.forEach(function(n){return t+=See(n)}),t}function k4(e,t,n){var r=e==="VERTEX"?t.VERTEX_SHADER:t.FRAGMENT_SHADER,a=t.createShader(r);if(a===null)throw new Error("loadShader: error while creating the shader");t.shaderSource(a,n),t.compileShader(a);var o=t.getShaderParameter(a,t.COMPILE_STATUS);if(!o){var s=t.getShaderInfoLog(a);throw t.deleteShader(a),new Error(`loadShader: error while compiling the shader:
`.concat(s,`
`).concat(n))}return a}function Eee(e,t){return k4("VERTEX",e,t)}function wee(e,t){return k4("FRAGMENT",e,t)}function xee(e,t){var n=e.createProgram();if(n===null)throw new Error("loadProgram: error while creating the program.");var r,a;for(r=0,a=t.length;r<a;r++)e.attachShader(n,t[r]);e.linkProgram(n);var o=e.getProgramParameter(n,e.LINK_STATUS);if(!o)throw e.deleteProgram(n),new Error("loadProgram: error while linking the program.");return n}function pN(e){var t=e.gl,n=e.buffer,r=e.program,a=e.vertexShader,o=e.fragmentShader;t.deleteShader(a),t.deleteShader(o),t.deleteProgram(r),t.deleteBuffer(n)}function gN(e){return e%1===0?e.toFixed(1):e.toString()}var hN=`#define PICKING_MODE
`,kee=Re(Re(Re(Re(Re(Re(Re(Re({},WebGL2RenderingContext.BOOL,1),WebGL2RenderingContext.BYTE,1),WebGL2RenderingContext.UNSIGNED_BYTE,1),WebGL2RenderingContext.SHORT,2),WebGL2RenderingContext.UNSIGNED_SHORT,2),WebGL2RenderingContext.INT,4),WebGL2RenderingContext.UNSIGNED_INT,4),WebGL2RenderingContext.FLOAT,4),T4=function(){function e(t,n,r){cn(this,e),Re(this,"array",new Float32Array),Re(this,"constantArray",new Float32Array),Re(this,"capacity",0),Re(this,"verticesCount",0);var a=this.getDefinition();if(this.VERTICES=a.VERTICES,this.VERTEX_SHADER_SOURCE=a.VERTEX_SHADER_SOURCE,this.FRAGMENT_SHADER_SOURCE=a.FRAGMENT_SHADER_SOURCE,this.UNIFORMS=a.UNIFORMS,this.ATTRIBUTES=a.ATTRIBUTES,this.METHOD=a.METHOD,this.CONSTANT_ATTRIBUTES="CONSTANT_ATTRIBUTES"in a?a.CONSTANT_ATTRIBUTES:[],this.CONSTANT_DATA="CONSTANT_DATA"in a?a.CONSTANT_DATA:[],this.isInstanced="CONSTANT_ATTRIBUTES"in a,this.ATTRIBUTES_ITEMS_COUNT=bm(this.ATTRIBUTES),this.STRIDE=this.VERTICES*this.ATTRIBUTES_ITEMS_COUNT,this.renderer=r,this.normalProgram=this.getProgramInfo("normal",t,a.VERTEX_SHADER_SOURCE,a.FRAGMENT_SHADER_SOURCE,null),this.pickProgram=n?this.getProgramInfo("pick",t,hN+a.VERTEX_SHADER_SOURCE,hN+a.FRAGMENT_SHADER_SOURCE,n):null,this.isInstanced){var o=bm(this.CONSTANT_ATTRIBUTES);if(this.CONSTANT_DATA.length!==this.VERTICES)throw new Error("Program: error while getting constant data (expected ".concat(this.VERTICES," items, received ").concat(this.CONSTANT_DATA.length," instead)"));this.constantArray=new Float32Array(this.CONSTANT_DATA.length*o);for(var s=0;s<this.CONSTANT_DATA.length;s++){var u=this.CONSTANT_DATA[s];if(u.length!==o)throw new Error("Program: error while getting constant data (one vector has ".concat(u.length," items instead of ").concat(o,")"));for(var c=0;c<u.length;c++)this.constantArray[s*o+c]=u[c]}this.STRIDE=this.ATTRIBUTES_ITEMS_COUNT}}return dn(e,[{key:"kill",value:function(){pN(this.normalProgram),this.pickProgram&&(pN(this.pickProgram),this.pickProgram=null)}},{key:"getProgramInfo",value:function(n,r,a,o,s){var u=this.getDefinition(),c=r.createBuffer();if(c===null)throw new Error("Program: error while creating the WebGL buffer.");var d=Eee(r,a),p=wee(r,o),g=xee(r,[d,p]),m={};u.UNIFORMS.forEach(function(S){var x=r.getUniformLocation(g,S);x&&(m[S]=x)});var b={};u.ATTRIBUTES.forEach(function(S){b[S.name]=r.getAttribLocation(g,S.name)});var y;if("CONSTANT_ATTRIBUTES"in u&&(u.CONSTANT_ATTRIBUTES.forEach(function(S){b[S.name]=r.getAttribLocation(g,S.name)}),y=r.createBuffer(),y===null))throw new Error("Program: error while creating the WebGL constant buffer.");return{name:n,program:g,gl:r,frameBuffer:s,buffer:c,constantBuffer:y||{},uniformLocations:m,attributeLocations:b,isPicking:n==="pick",vertexShader:d,fragmentShader:p}}},{key:"bindProgram",value:function(n){var r=this,a=0,o=n.gl,s=n.buffer;this.isInstanced?(o.bindBuffer(o.ARRAY_BUFFER,n.constantBuffer),a=0,this.CONSTANT_ATTRIBUTES.forEach(function(u){return a+=r.bindAttribute(u,n,a,!1)}),o.bufferData(o.ARRAY_BUFFER,this.constantArray,o.STATIC_DRAW),o.bindBuffer(o.ARRAY_BUFFER,n.buffer),a=0,this.ATTRIBUTES.forEach(function(u){return a+=r.bindAttribute(u,n,a,!0)}),o.bufferData(o.ARRAY_BUFFER,this.array,o.DYNAMIC_DRAW)):(o.bindBuffer(o.ARRAY_BUFFER,s),a=0,this.ATTRIBUTES.forEach(function(u){return a+=r.bindAttribute(u,n,a)}),o.bufferData(o.ARRAY_BUFFER,this.array,o.DYNAMIC_DRAW)),o.bindBuffer(o.ARRAY_BUFFER,null)}},{key:"unbindProgram",value:function(n){var r=this;this.isInstanced?(this.CONSTANT_ATTRIBUTES.forEach(function(a){return r.unbindAttribute(a,n,!1)}),this.ATTRIBUTES.forEach(function(a){return r.unbindAttribute(a,n,!0)})):this.ATTRIBUTES.forEach(function(a){return r.unbindAttribute(a,n)})}},{key:"bindAttribute",value:function(n,r,a,o){var s=kee[n.type];if(typeof s!="number")throw new Error('Program.bind: yet unsupported attribute type "'.concat(n.type,'"'));var u=r.attributeLocations[n.name],c=r.gl;if(u!==-1){c.enableVertexAttribArray(u);var d=this.isInstanced?(o?this.ATTRIBUTES_ITEMS_COUNT:bm(this.CONSTANT_ATTRIBUTES))*Float32Array.BYTES_PER_ELEMENT:this.ATTRIBUTES_ITEMS_COUNT*Float32Array.BYTES_PER_ELEMENT;if(c.vertexAttribPointer(u,n.size,n.type,n.normalized||!1,d,a),this.isInstanced&&o)if(c instanceof WebGL2RenderingContext)c.vertexAttribDivisor(u,1);else{var p=c.getExtension("ANGLE_instanced_arrays");p&&p.vertexAttribDivisorANGLE(u,1)}}return n.size*s}},{key:"unbindAttribute",value:function(n,r,a){var o=r.attributeLocations[n.name],s=r.gl;if(o!==-1&&(s.disableVertexAttribArray(o),this.isInstanced&&a))if(s instanceof WebGL2RenderingContext)s.vertexAttribDivisor(o,0);else{var u=s.getExtension("ANGLE_instanced_arrays");u&&u.vertexAttribDivisorANGLE(o,0)}}},{key:"reallocate",value:function(n){n!==this.capacity&&(this.capacity=n,this.verticesCount=this.VERTICES*n,this.array=new Float32Array(this.isInstanced?this.capacity*this.ATTRIBUTES_ITEMS_COUNT:this.verticesCount*this.ATTRIBUTES_ITEMS_COUNT))}},{key:"hasNothingToRender",value:function(){return this.verticesCount===0}},{key:"renderProgram",value:function(n,r){var a=r.gl,o=r.program;a.enable(a.BLEND),a.useProgram(o),this.setUniforms(n,r),this.drawWebGL(this.METHOD,r)}},{key:"render",value:function(n){this.hasNothingToRender()||(this.pickProgram&&(this.pickProgram.gl.viewport(0,0,n.width*n.pixelRatio/n.downSizingRatio,n.height*n.pixelRatio/n.downSizingRatio),this.bindProgram(this.pickProgram),this.renderProgram(je(je({},n),{},{pixelRatio:n.pixelRatio/n.downSizingRatio}),this.pickProgram),this.unbindProgram(this.pickProgram)),this.normalProgram.gl.viewport(0,0,n.width*n.pixelRatio,n.height*n.pixelRatio),this.bindProgram(this.normalProgram),this.renderProgram(n,this.normalProgram),this.unbindProgram(this.normalProgram))}},{key:"drawWebGL",value:function(n,r){var a=r.gl,o=r.frameBuffer;if(a.bindFramebuffer(a.FRAMEBUFFER,o),!this.isInstanced)a.drawArrays(n,0,this.verticesCount);else if(a instanceof WebGL2RenderingContext)a.drawArraysInstanced(n,0,this.VERTICES,this.capacity);else{var s=a.getExtension("ANGLE_instanced_arrays");s&&s.drawArraysInstancedANGLE(n,0,this.VERTICES,this.capacity)}}}])}(),hT=function(e){function t(){return cn(this,t),$n(this,t,arguments)}return qn(t,e),dn(t,[{key:"kill",value:function(){x4(t,"kill",this)([])}},{key:"process",value:function(r,a,o){var s=a*this.STRIDE;if(o.hidden){for(var u=s+this.STRIDE;s<u;s++)this.array[s]=0;return}return this.processVisibleItem(w4(r),s,o)}}])}(T4),Zu=function(e){function t(){var n;cn(this,t);for(var r=arguments.length,a=new Array(r),o=0;o<r;o++)a[o]=arguments[o];return n=$n(this,t,[].concat(a)),Re(n,"drawLabel",void 0),n}return qn(t,e),dn(t,[{key:"kill",value:function(){x4(t,"kill",this)([])}},{key:"process",value:function(r,a,o,s,u){var c=a*this.STRIDE;if(u.hidden||o.hidden||s.hidden){for(var d=c+this.STRIDE;c<d;c++)this.array[c]=0;return}return this.processVisibleItem(w4(r),c,o,s,u)}}])}(T4);function A4(e,t){return function(){function n(r,a,o){cn(this,n),Re(this,"drawLabel",t),this.programs=e.map(function(s){return new s(r,a,o)})}return dn(n,[{key:"reallocate",value:function(a){this.programs.forEach(function(o){return o.reallocate(a)})}},{key:"process",value:function(a,o,s,u,c){this.programs.forEach(function(d){return d.process(a,o,s,u,c)})}},{key:"render",value:function(a){this.programs.forEach(function(o){return o.render(a)})}},{key:"kill",value:function(){this.programs.forEach(function(a){return a.kill()})}}])}()}function Tee(e,t,n,r,a){var o=a.edgeLabelSize,s=a.edgeLabelFont,u=a.edgeLabelWeight,c=a.edgeLabelColor.attribute?t[a.edgeLabelColor.attribute]||a.edgeLabelColor.color||"#000":a.edgeLabelColor.color,d=t.label;if(d){e.fillStyle=c,e.font="".concat(u," ").concat(o,"px ").concat(s);var p=n.size,g=r.size,m=n.x,b=n.y,y=r.x,S=r.y,x=(m+y)/2,R=(b+S)/2,k=y-m,A=S-b,C=Math.sqrt(k*k+A*A);if(!(C<p+g)){m+=k*p/C,b+=A*p/C,y-=k*g/C,S-=A*g/C,x=(m+y)/2,R=(b+S)/2,k=y-m,A=S-b,C=Math.sqrt(k*k+A*A);var N=e.measureText(d).width;if(N>C){var _="…";for(d=d+_,N=e.measureText(d).width;N>C&&d.length>1;)d=d.slice(0,-2)+_,N=e.measureText(d).width;if(d.length<4)return}var O;k>0?A>0?O=Math.acos(k/C):O=Math.asin(A/C):A>0?O=Math.acos(k/C)+Math.PI:O=Math.asin(k/C)+Math.PI/2,e.save(),e.translate(x,R),e.rotate(O),e.fillText(d,-N/2,t.size/2+o),e.restore()}}}function R4(e,t,n){if(t.label){var r=n.labelSize,a=n.labelFont,o=n.labelWeight,s=n.labelColor.attribute?t[n.labelColor.attribute]||n.labelColor.color||"#000":n.labelColor.color;e.fillStyle=s,e.font="".concat(o," ").concat(r,"px ").concat(a),e.fillText(t.label,t.x+t.size+3,t.y+r/3)}}function Aee(e,t,n){var r=n.labelSize,a=n.labelFont,o=n.labelWeight;e.font="".concat(o," ").concat(r,"px ").concat(a),e.fillStyle="#FFF",e.shadowOffsetX=0,e.shadowOffsetY=0,e.shadowBlur=8,e.shadowColor="#000";var s=2;if(typeof t.label=="string"){var u=e.measureText(t.label).width,c=Math.round(u+5),d=Math.round(r+2*s),p=Math.max(t.size,r/2)+s,g=Math.asin(d/2/p),m=Math.sqrt(Math.abs(Math.pow(p,2)-Math.pow(d/2,2)));e.beginPath(),e.moveTo(t.x+m,t.y+d/2),e.lineTo(t.x+p+c,t.y+d/2),e.lineTo(t.x+p+c,t.y-d/2),e.lineTo(t.x+m,t.y-d/2),e.arc(t.x,t.y,p,g,-g),e.closePath(),e.fill()}else e.beginPath(),e.arc(t.x,t.y,t.size+s,0,Math.PI*2),e.closePath(),e.fill();e.shadowOffsetX=0,e.shadowOffsetY=0,e.shadowBlur=0,R4(e,t,n)}var Ree=`
precision highp float;

varying vec4 v_color;
varying vec2 v_diffVector;
varying float v_radius;

uniform float u_correctionRatio;

const vec4 transparent = vec4(0.0, 0.0, 0.0, 0.0);

void main(void) {
  float border = u_correctionRatio * 2.0;
  float dist = length(v_diffVector) - v_radius + border;

  // No antialiasing for picking mode:
  #ifdef PICKING_MODE
  if (dist > border)
    gl_FragColor = transparent;
  else
    gl_FragColor = v_color;

  #else
  float t = 0.0;
  if (dist > border)
    t = 1.0;
  else if (dist > 0.0)
    t = dist / border;

  gl_FragColor = mix(v_color, transparent, t);
  #endif
}
`,_ee=Ree,Cee=`
attribute vec4 a_id;
attribute vec4 a_color;
attribute vec2 a_position;
attribute float a_size;
attribute float a_angle;

uniform mat3 u_matrix;
uniform float u_sizeRatio;
uniform float u_correctionRatio;

varying vec4 v_color;
varying vec2 v_diffVector;
varying float v_radius;
varying float v_border;

const float bias = 255.0 / 254.0;

void main() {
  float size = a_size * u_correctionRatio / u_sizeRatio * 4.0;
  vec2 diffVector = size * vec2(cos(a_angle), sin(a_angle));
  vec2 position = a_position + diffVector;
  gl_Position = vec4(
    (u_matrix * vec3(position, 1)).xy,
    0,
    1
  );

  v_diffVector = diffVector;
  v_radius = size / 2.0;

  #ifdef PICKING_MODE
  // For picking mode, we use the ID as the color:
  v_color = a_id;
  #else
  // For normal mode, we use the color:
  v_color = a_color;
  #endif

  v_color.a *= bias;
}
`,Nee=Cee,_4=WebGLRenderingContext,mN=_4.UNSIGNED_BYTE,ym=_4.FLOAT,Oee=["u_sizeRatio","u_correctionRatio","u_matrix"],Qu=function(e){function t(){return cn(this,t),$n(this,t,arguments)}return qn(t,e),dn(t,[{key:"getDefinition",value:function(){return{VERTICES:3,VERTEX_SHADER_SOURCE:Nee,FRAGMENT_SHADER_SOURCE:_ee,METHOD:WebGLRenderingContext.TRIANGLES,UNIFORMS:Oee,ATTRIBUTES:[{name:"a_position",size:2,type:ym},{name:"a_size",size:1,type:ym},{name:"a_color",size:4,type:mN,normalized:!0},{name:"a_id",size:4,type:mN,normalized:!0}],CONSTANT_ATTRIBUTES:[{name:"a_angle",size:1,type:ym}],CONSTANT_DATA:[[t.ANGLE_1],[t.ANGLE_2],[t.ANGLE_3]]}}},{key:"processVisibleItem",value:function(r,a,o){var s=this.array,u=la(o.color);s[a++]=o.x,s[a++]=o.y,s[a++]=o.size,s[a++]=u,s[a++]=r}},{key:"setUniforms",value:function(r,a){var o=a.gl,s=a.uniformLocations,u=s.u_sizeRatio,c=s.u_correctionRatio,d=s.u_matrix;o.uniform1f(c,r.correctionRatio),o.uniform1f(u,r.sizeRatio),o.uniformMatrix3fv(d,!1,r.matrix)}}])}(hT);Re(Qu,"ANGLE_1",0);Re(Qu,"ANGLE_2",2*Math.PI/3);Re(Qu,"ANGLE_3",4*Math.PI/3);var Dee=`
precision mediump float;

varying vec4 v_color;

void main(void) {
  gl_FragColor = v_color;
}
`,Iee=Dee,Lee=`
attribute vec2 a_position;
attribute vec2 a_normal;
attribute float a_radius;
attribute vec3 a_barycentric;

#ifdef PICKING_MODE
attribute vec4 a_id;
#else
attribute vec4 a_color;
#endif

uniform mat3 u_matrix;
uniform float u_sizeRatio;
uniform float u_correctionRatio;
uniform float u_minEdgeThickness;
uniform float u_lengthToThicknessRatio;
uniform float u_widenessToThicknessRatio;

varying vec4 v_color;

const float bias = 255.0 / 254.0;

void main() {
  float minThickness = u_minEdgeThickness;

  float normalLength = length(a_normal);
  vec2 unitNormal = a_normal / normalLength;

  // These first computations are taken from edge.vert.glsl and
  // edge.clamped.vert.glsl. Please read it to get better comments on what's
  // happening:
  float pixelsThickness = max(normalLength / u_sizeRatio, minThickness);
  float webGLThickness = pixelsThickness * u_correctionRatio;
  float webGLNodeRadius = a_radius * 2.0 * u_correctionRatio / u_sizeRatio;
  float webGLArrowHeadLength = webGLThickness * u_lengthToThicknessRatio * 2.0;
  float webGLArrowHeadThickness = webGLThickness * u_widenessToThicknessRatio;

  float da = a_barycentric.x;
  float db = a_barycentric.y;
  float dc = a_barycentric.z;

  vec2 delta = vec2(
      da * (webGLNodeRadius * unitNormal.y)
    + db * ((webGLNodeRadius + webGLArrowHeadLength) * unitNormal.y + webGLArrowHeadThickness * unitNormal.x)
    + dc * ((webGLNodeRadius + webGLArrowHeadLength) * unitNormal.y - webGLArrowHeadThickness * unitNormal.x),

      da * (-webGLNodeRadius * unitNormal.x)
    + db * (-(webGLNodeRadius + webGLArrowHeadLength) * unitNormal.x + webGLArrowHeadThickness * unitNormal.y)
    + dc * (-(webGLNodeRadius + webGLArrowHeadLength) * unitNormal.x - webGLArrowHeadThickness * unitNormal.y)
  );

  vec2 position = (u_matrix * vec3(a_position + delta, 1)).xy;

  gl_Position = vec4(position, 0, 1);

  #ifdef PICKING_MODE
  // For picking mode, we use the ID as the color:
  v_color = a_id;
  #else
  // For normal mode, we use the color:
  v_color = a_color;
  #endif

  v_color.a *= bias;
}
`,Mee=Lee,C4=WebGLRenderingContext,bN=C4.UNSIGNED_BYTE,_d=C4.FLOAT,Fee=["u_matrix","u_sizeRatio","u_correctionRatio","u_minEdgeThickness","u_lengthToThicknessRatio","u_widenessToThicknessRatio"],Ju={extremity:"target",lengthToThicknessRatio:2.5,widenessToThicknessRatio:2};function kf(e){var t=je(je({},Ju),e||{});return function(n){function r(){return cn(this,r),$n(this,r,arguments)}return qn(r,n),dn(r,[{key:"getDefinition",value:function(){return{VERTICES:3,VERTEX_SHADER_SOURCE:Mee,FRAGMENT_SHADER_SOURCE:Iee,METHOD:WebGLRenderingContext.TRIANGLES,UNIFORMS:Fee,ATTRIBUTES:[{name:"a_position",size:2,type:_d},{name:"a_normal",size:2,type:_d},{name:"a_radius",size:1,type:_d},{name:"a_color",size:4,type:bN,normalized:!0},{name:"a_id",size:4,type:bN,normalized:!0}],CONSTANT_ATTRIBUTES:[{name:"a_barycentric",size:3,type:_d}],CONSTANT_DATA:[[1,0,0],[0,1,0],[0,0,1]]}}},{key:"processVisibleItem",value:function(o,s,u,c,d){if(t.extremity==="source"){var p=[c,u];u=p[0],c=p[1]}var g=d.size||1,m=c.size||1,b=u.x,y=u.y,S=c.x,x=c.y,R=la(d.color),k=S-b,A=x-y,C=k*k+A*A,N=0,_=0;C&&(C=1/Math.sqrt(C),N=-A*C*g,_=k*C*g);var O=this.array;O[s++]=S,O[s++]=x,O[s++]=-N,O[s++]=-_,O[s++]=m,O[s++]=R,O[s++]=o}},{key:"setUniforms",value:function(o,s){var u=s.gl,c=s.uniformLocations,d=c.u_matrix,p=c.u_sizeRatio,g=c.u_correctionRatio,m=c.u_minEdgeThickness,b=c.u_lengthToThicknessRatio,y=c.u_widenessToThicknessRatio;u.uniformMatrix3fv(d,!1,o.matrix),u.uniform1f(p,o.sizeRatio),u.uniform1f(g,o.correctionRatio),u.uniform1f(m,o.minEdgeThickness),u.uniform1f(b,t.lengthToThicknessRatio),u.uniform1f(y,t.widenessToThicknessRatio)}}])}(Zu)}kf();var Pee=`
precision mediump float;

varying vec4 v_color;
varying vec2 v_normal;
varying float v_thickness;
varying float v_feather;

const vec4 transparent = vec4(0.0, 0.0, 0.0, 0.0);

void main(void) {
  // We only handle antialiasing for normal mode:
  #ifdef PICKING_MODE
  gl_FragColor = v_color;
  #else
  float dist = length(v_normal) * v_thickness;

  float t = smoothstep(
    v_thickness - v_feather,
    v_thickness,
    dist
  );

  gl_FragColor = mix(v_color, transparent, t);
  #endif
}
`,mT=Pee,zee=`
attribute vec4 a_id;
attribute vec4 a_color;
attribute vec2 a_normal;
attribute float a_normalCoef;
attribute vec2 a_positionStart;
attribute vec2 a_positionEnd;
attribute float a_positionCoef;
attribute float a_radius;
attribute float a_radiusCoef;

uniform mat3 u_matrix;
uniform float u_zoomRatio;
uniform float u_sizeRatio;
uniform float u_pixelRatio;
uniform float u_correctionRatio;
uniform float u_minEdgeThickness;
uniform float u_lengthToThicknessRatio;
uniform float u_feather;

varying vec4 v_color;
varying vec2 v_normal;
varying float v_thickness;
varying float v_feather;

const float bias = 255.0 / 254.0;

void main() {
  float minThickness = u_minEdgeThickness;

  float radius = a_radius * a_radiusCoef;
  vec2 normal = a_normal * a_normalCoef;
  vec2 position = a_positionStart * (1.0 - a_positionCoef) + a_positionEnd * a_positionCoef;

  float normalLength = length(normal);
  vec2 unitNormal = normal / normalLength;

  // These first computations are taken from edge.vert.glsl. Please read it to
  // get better comments on what's happening:
  float pixelsThickness = max(normalLength, minThickness * u_sizeRatio);
  float webGLThickness = pixelsThickness * u_correctionRatio / u_sizeRatio;

  // Here, we move the point to leave space for the arrow head:
  float direction = sign(radius);
  float webGLNodeRadius = direction * radius * 2.0 * u_correctionRatio / u_sizeRatio;
  float webGLArrowHeadLength = webGLThickness * u_lengthToThicknessRatio * 2.0;

  vec2 compensationVector = vec2(-direction * unitNormal.y, direction * unitNormal.x) * (webGLNodeRadius + webGLArrowHeadLength);

  // Here is the proper position of the vertex
  gl_Position = vec4((u_matrix * vec3(position + unitNormal * webGLThickness + compensationVector, 1)).xy, 0, 1);

  v_thickness = webGLThickness / u_zoomRatio;

  v_normal = unitNormal;

  v_feather = u_feather * u_correctionRatio / u_zoomRatio / u_pixelRatio * 2.0;

  #ifdef PICKING_MODE
  // For picking mode, we use the ID as the color:
  v_color = a_id;
  #else
  // For normal mode, we use the color:
  v_color = a_color;
  #endif

  v_color.a *= bias;
}
`,Bee=zee,N4=WebGLRenderingContext,yN=N4.UNSIGNED_BYTE,oi=N4.FLOAT,Uee=["u_matrix","u_zoomRatio","u_sizeRatio","u_correctionRatio","u_pixelRatio","u_feather","u_minEdgeThickness","u_lengthToThicknessRatio"],jee={lengthToThicknessRatio:Ju.lengthToThicknessRatio};function O4(e){var t=je(je({},jee),{});return function(n){function r(){return cn(this,r),$n(this,r,arguments)}return qn(r,n),dn(r,[{key:"getDefinition",value:function(){return{VERTICES:6,VERTEX_SHADER_SOURCE:Bee,FRAGMENT_SHADER_SOURCE:mT,METHOD:WebGLRenderingContext.TRIANGLES,UNIFORMS:Uee,ATTRIBUTES:[{name:"a_positionStart",size:2,type:oi},{name:"a_positionEnd",size:2,type:oi},{name:"a_normal",size:2,type:oi},{name:"a_color",size:4,type:yN,normalized:!0},{name:"a_id",size:4,type:yN,normalized:!0},{name:"a_radius",size:1,type:oi}],CONSTANT_ATTRIBUTES:[{name:"a_positionCoef",size:1,type:oi},{name:"a_normalCoef",size:1,type:oi},{name:"a_radiusCoef",size:1,type:oi}],CONSTANT_DATA:[[0,1,0],[0,-1,0],[1,1,1],[1,1,1],[0,-1,0],[1,-1,-1]]}}},{key:"processVisibleItem",value:function(o,s,u,c,d){var p=d.size||1,g=u.x,m=u.y,b=c.x,y=c.y,S=la(d.color),x=b-g,R=y-m,k=c.size||1,A=x*x+R*R,C=0,N=0;A&&(A=1/Math.sqrt(A),C=-R*A*p,N=x*A*p);var _=this.array;_[s++]=g,_[s++]=m,_[s++]=b,_[s++]=y,_[s++]=C,_[s++]=N,_[s++]=S,_[s++]=o,_[s++]=k}},{key:"setUniforms",value:function(o,s){var u=s.gl,c=s.uniformLocations,d=c.u_matrix,p=c.u_zoomRatio,g=c.u_feather,m=c.u_pixelRatio,b=c.u_correctionRatio,y=c.u_sizeRatio,S=c.u_minEdgeThickness,x=c.u_lengthToThicknessRatio;u.uniformMatrix3fv(d,!1,o.matrix),u.uniform1f(p,o.zoomRatio),u.uniform1f(y,o.sizeRatio),u.uniform1f(b,o.correctionRatio),u.uniform1f(m,o.pixelRatio),u.uniform1f(g,o.antiAliasingFeather),u.uniform1f(S,o.minEdgeThickness),u.uniform1f(x,t.lengthToThicknessRatio)}}])}(Zu)}O4();function Gee(e){return A4([O4(),kf(e)])}var Hee=Gee(),D4=Hee,$ee=`
attribute vec4 a_id;
attribute vec4 a_color;
attribute vec2 a_normal;
attribute float a_normalCoef;
attribute vec2 a_positionStart;
attribute vec2 a_positionEnd;
attribute float a_positionCoef;

uniform mat3 u_matrix;
uniform float u_sizeRatio;
uniform float u_zoomRatio;
uniform float u_pixelRatio;
uniform float u_correctionRatio;
uniform float u_minEdgeThickness;
uniform float u_feather;

varying vec4 v_color;
varying vec2 v_normal;
varying float v_thickness;
varying float v_feather;

const float bias = 255.0 / 254.0;

void main() {
  float minThickness = u_minEdgeThickness;

  vec2 normal = a_normal * a_normalCoef;
  vec2 position = a_positionStart * (1.0 - a_positionCoef) + a_positionEnd * a_positionCoef;

  float normalLength = length(normal);
  vec2 unitNormal = normal / normalLength;

  // We require edges to be at least "minThickness" pixels thick *on screen*
  // (so we need to compensate the size ratio):
  float pixelsThickness = max(normalLength, minThickness * u_sizeRatio);

  // Then, we need to retrieve the normalized thickness of the edge in the WebGL
  // referential (in a ([0, 1], [0, 1]) space), using our "magic" correction
  // ratio:
  float webGLThickness = pixelsThickness * u_correctionRatio / u_sizeRatio;

  // Here is the proper position of the vertex
  gl_Position = vec4((u_matrix * vec3(position + unitNormal * webGLThickness, 1)).xy, 0, 1);

  // For the fragment shader though, we need a thickness that takes the "magic"
  // correction ratio into account (as in webGLThickness), but so that the
  // antialiasing effect does not depend on the zoom level. So here's yet
  // another thickness version:
  v_thickness = webGLThickness / u_zoomRatio;

  v_normal = unitNormal;

  v_feather = u_feather * u_correctionRatio / u_zoomRatio / u_pixelRatio * 2.0;

  #ifdef PICKING_MODE
  // For picking mode, we use the ID as the color:
  v_color = a_id;
  #else
  // For normal mode, we use the color:
  v_color = a_color;
  #endif

  v_color.a *= bias;
}
`,qee=$ee,I4=WebGLRenderingContext,vN=I4.UNSIGNED_BYTE,ru=I4.FLOAT,Vee=["u_matrix","u_zoomRatio","u_sizeRatio","u_correctionRatio","u_pixelRatio","u_feather","u_minEdgeThickness"],Wee=function(e){function t(){return cn(this,t),$n(this,t,arguments)}return qn(t,e),dn(t,[{key:"getDefinition",value:function(){return{VERTICES:6,VERTEX_SHADER_SOURCE:qee,FRAGMENT_SHADER_SOURCE:mT,METHOD:WebGLRenderingContext.TRIANGLES,UNIFORMS:Vee,ATTRIBUTES:[{name:"a_positionStart",size:2,type:ru},{name:"a_positionEnd",size:2,type:ru},{name:"a_normal",size:2,type:ru},{name:"a_color",size:4,type:vN,normalized:!0},{name:"a_id",size:4,type:vN,normalized:!0}],CONSTANT_ATTRIBUTES:[{name:"a_positionCoef",size:1,type:ru},{name:"a_normalCoef",size:1,type:ru}],CONSTANT_DATA:[[0,1],[0,-1],[1,1],[1,1],[0,-1],[1,-1]]}}},{key:"processVisibleItem",value:function(r,a,o,s,u){var c=u.size||1,d=o.x,p=o.y,g=s.x,m=s.y,b=la(u.color),y=g-d,S=m-p,x=y*y+S*S,R=0,k=0;x&&(x=1/Math.sqrt(x),R=-S*x*c,k=y*x*c);var A=this.array;A[a++]=d,A[a++]=p,A[a++]=g,A[a++]=m,A[a++]=R,A[a++]=k,A[a++]=b,A[a++]=r}},{key:"setUniforms",value:function(r,a){var o=a.gl,s=a.uniformLocations,u=s.u_matrix,c=s.u_zoomRatio,d=s.u_feather,p=s.u_pixelRatio,g=s.u_correctionRatio,m=s.u_sizeRatio,b=s.u_minEdgeThickness;o.uniformMatrix3fv(u,!1,r.matrix),o.uniform1f(c,r.zoomRatio),o.uniform1f(m,r.sizeRatio),o.uniform1f(g,r.correctionRatio),o.uniform1f(p,r.pixelRatio),o.uniform1f(d,r.antiAliasingFeather),o.uniform1f(b,r.minEdgeThickness)}}])}(Zu),bT=function(e){function t(){var n;return cn(this,t),n=$n(this,t),n.rawEmitter=n,n}return qn(t,e),dn(t)}(r4.EventEmitter),vm,SN;function ua(){return SN||(SN=1,vm=function(t){return t!==null&&typeof t=="object"&&typeof t.addUndirectedEdgeWithKey=="function"&&typeof t.dropNode=="function"&&typeof t.multi=="boolean"}),vm}var Yee=ua();const Kee=un(Yee);var Xee=function(t){return t},Zee=function(t){return t*t},Qee=function(t){return t*(2-t)},Jee=function(t){return(t*=2)<1?.5*t*t:-.5*(--t*(t-2)-1)},ete=function(t){return t*t*t},tte=function(t){return--t*t*t+1},nte=function(t){return(t*=2)<1?.5*t*t*t:.5*((t-=2)*t*t+2)},L4={linear:Xee,quadraticIn:Zee,quadraticOut:Qee,quadraticInOut:Jee,cubicIn:ete,cubicOut:tte,cubicInOut:nte},M4={easing:"quadraticInOut",duration:150};function F4(e,t,n,r){var a=Object.assign({},M4,n),o=typeof a.easing=="function"?a.easing:L4[a.easing],s=Date.now(),u={};for(var c in t){var d=t[c];u[c]={};for(var p in d)u[c][p]=e.getNodeAttribute(c,p)}var g=null,m=function(){g=null;var y=(Date.now()-s)/a.duration;if(y>=1){for(var S in t){var x=t[S];for(var R in x)e.setNodeAttribute(S,R,x[R])}return}y=o(y);for(var k in t){var A=t[k],C=u[k];for(var N in A)e.setNodeAttribute(k,N,A[N]*y+C[N]*(1-y))}g=requestAnimationFrame(m)};return m(),function(){g&&cancelAnimationFrame(g)}}function Mr(){return Float32Array.of(1,0,0,0,1,0,0,0,1)}function Cd(e,t,n){return e[0]=t,e[4]=typeof n=="number"?n:t,e}function EN(e,t){var n=Math.sin(t),r=Math.cos(t);return e[0]=r,e[1]=n,e[3]=-n,e[4]=r,e}function wN(e,t,n){return e[6]=t,e[7]=n,e}function vo(e,t){var n=e[0],r=e[1],a=e[2],o=e[3],s=e[4],u=e[5],c=e[6],d=e[7],p=e[8],g=t[0],m=t[1],b=t[2],y=t[3],S=t[4],x=t[5],R=t[6],k=t[7],A=t[8];return e[0]=g*n+m*o+b*c,e[1]=g*r+m*s+b*d,e[2]=g*a+m*u+b*p,e[3]=y*n+S*o+x*c,e[4]=y*r+S*s+x*d,e[5]=y*a+S*u+x*p,e[6]=R*n+k*o+A*c,e[7]=R*r+k*s+A*d,e[8]=R*a+k*u+A*p,e}function Bk(e,t){var n=arguments.length>2&&arguments[2]!==void 0?arguments[2]:1,r=e[0],a=e[1],o=e[3],s=e[4],u=e[6],c=e[7],d=t.x,p=t.y;return{x:d*r+p*o+u*n,y:d*a+p*s+c*n}}function rte(e,t){var n=e.height/e.width,r=t.height/t.width;return n<1&&r>1||n>1&&r<1?1:Math.min(Math.max(r,1/r),Math.max(1/n,n))}function au(e,t,n,r,a){var o=e.angle,s=e.ratio,u=e.x,c=e.y,d=t.width,p=t.height,g=Mr(),m=Math.min(d,p)-2*r,b=rte(t,n);return a?(vo(g,wN(Mr(),u,c)),vo(g,Cd(Mr(),s)),vo(g,EN(Mr(),o)),vo(g,Cd(Mr(),d/m/2/b,p/m/2/b))):(vo(g,Cd(Mr(),2*(m/d)*b,2*(m/p)*b)),vo(g,EN(Mr(),-o)),vo(g,Cd(Mr(),1/s)),vo(g,wN(Mr(),-u,-c))),g}function ate(e,t,n){var r=Bk(e,{x:Math.cos(t.angle),y:Math.sin(t.angle)},0),a=r.x,o=r.y;return 1/Math.sqrt(Math.pow(a,2)+Math.pow(o,2))/n.width}function ote(e){if(!e.order)return{x:[0,1],y:[0,1]};var t=1/0,n=-1/0,r=1/0,a=-1/0;return e.forEachNode(function(o,s){var u=s.x,c=s.y;u<t&&(t=u),u>n&&(n=u),c<r&&(r=c),c>a&&(a=c)}),{x:[t,n],y:[r,a]}}function ite(e){if(!Kee(e))throw new Error("Sigma: invalid graph instance.");e.forEachNode(function(t,n){if(!Number.isFinite(n.x)||!Number.isFinite(n.y))throw new Error("Sigma: Coordinates of node ".concat(t," are invalid. A node must have a numeric 'x' and 'y' attribute."))})}function ste(e,t,n){var r=document.createElement(e);if(t)for(var a in t)r.style[a]=t[a];if(n)for(var o in n)r.setAttribute(o,n[o]);return r}function xN(){return typeof window.devicePixelRatio<"u"?window.devicePixelRatio:1}function kN(e,t,n){return n.sort(function(r,a){var o=t(r)||0,s=t(a)||0;return o<s?-1:o>s?1:0})}function TN(e){var t=Ls(e.x,2),n=t[0],r=t[1],a=Ls(e.y,2),o=a[0],s=a[1],u=Math.max(r-n,s-o),c=(r+n)/2,d=(s+o)/2;(u===0||Math.abs(u)===1/0||isNaN(u))&&(u=1),isNaN(c)&&(c=0),isNaN(d)&&(d=0);var p=function(m){return{x:.5+(m.x-c)/u,y:.5+(m.y-d)/u}};return p.applyTo=function(g){g.x=.5+(g.x-c)/u,g.y=.5+(g.y-d)/u},p.inverse=function(g){return{x:c+u*(g.x-.5),y:d+u*(g.y-.5)}},p.ratio=u,p}function Uk(e){"@babel/helpers - typeof";return Uk=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(t){return typeof t}:function(t){return t&&typeof Symbol=="function"&&t.constructor===Symbol&&t!==Symbol.prototype?"symbol":typeof t},Uk(e)}function AN(e,t){var n=t.size;if(n!==0){var r=e.length;e.length+=n;var a=0;t.forEach(function(o){e[r+a]=o,a++})}}function Sm(e){e=e||{};for(var t=0,n=arguments.length<=1?0:arguments.length-1;t<n;t++){var r=t+1<1||arguments.length<=t+1?void 0:arguments[t+1];r&&Object.assign(e,r)}return e}var yT={hideEdgesOnMove:!1,hideLabelsOnMove:!1,renderLabels:!0,renderEdgeLabels:!1,enableEdgeEvents:!1,defaultNodeColor:"#999",defaultNodeType:"circle",defaultEdgeColor:"#ccc",defaultEdgeType:"line",labelFont:"Arial",labelSize:14,labelWeight:"normal",labelColor:{color:"#000"},edgeLabelFont:"Arial",edgeLabelSize:14,edgeLabelWeight:"normal",edgeLabelColor:{attribute:"color"},stagePadding:30,defaultDrawEdgeLabel:Tee,defaultDrawNodeLabel:R4,defaultDrawNodeHover:Aee,minEdgeThickness:1.7,antiAliasingFeather:1,dragTimeout:100,draggedEventsTolerance:3,inertiaDuration:200,inertiaRatio:3,zoomDuration:250,zoomingRatio:1.7,doubleClickTimeout:300,doubleClickZoomingRatio:2.2,doubleClickZoomingDuration:200,tapMoveTolerance:10,zoomToSizeRatioFunction:Math.sqrt,itemSizesReference:"screen",autoRescale:!0,autoCenter:!0,labelDensity:1,labelGridCellSize:100,labelRenderedSizeThreshold:6,nodeReducer:null,edgeReducer:null,zIndex:!1,minCameraRatio:null,maxCameraRatio:null,enableCameraZooming:!0,enableCameraPanning:!0,enableCameraRotation:!0,cameraPanBoundaries:null,allowInvalidContainer:!1,nodeProgramClasses:{},nodeHoverProgramClasses:{},edgeProgramClasses:{}},lte={circle:Qu},ute={arrow:D4,line:Wee};function Em(e){if(typeof e.labelDensity!="number"||e.labelDensity<0)throw new Error("Settings: invalid `labelDensity`. Expecting a positive number.");var t=e.minCameraRatio,n=e.maxCameraRatio;if(typeof t=="number"&&typeof n=="number"&&n<t)throw new Error("Settings: invalid camera ratio boundaries. Expecting `maxCameraRatio` to be greater than `minCameraRatio`.")}function cte(e){var t=Sm({},yT,e);return t.nodeProgramClasses=Sm({},lte,t.nodeProgramClasses),t.edgeProgramClasses=Sm({},ute,t.edgeProgramClasses),t}var Nd=1.5,RN=function(e){function t(){var n;return cn(this,t),n=$n(this,t),Re(n,"x",.5),Re(n,"y",.5),Re(n,"angle",0),Re(n,"ratio",1),Re(n,"minRatio",null),Re(n,"maxRatio",null),Re(n,"enabledZooming",!0),Re(n,"enabledPanning",!0),Re(n,"enabledRotation",!0),Re(n,"clean",null),Re(n,"nextFrame",null),Re(n,"previousState",null),Re(n,"enabled",!0),n.previousState=n.getState(),n}return qn(t,e),dn(t,[{key:"enable",value:function(){return this.enabled=!0,this}},{key:"disable",value:function(){return this.enabled=!1,this}},{key:"getState",value:function(){return{x:this.x,y:this.y,angle:this.angle,ratio:this.ratio}}},{key:"hasState",value:function(r){return this.x===r.x&&this.y===r.y&&this.ratio===r.ratio&&this.angle===r.angle}},{key:"getPreviousState",value:function(){var r=this.previousState;return r?{x:r.x,y:r.y,angle:r.angle,ratio:r.ratio}:null}},{key:"getBoundedRatio",value:function(r){var a=r;return typeof this.minRatio=="number"&&(a=Math.max(a,this.minRatio)),typeof this.maxRatio=="number"&&(a=Math.min(a,this.maxRatio)),a}},{key:"validateState",value:function(r){var a={};return this.enabledPanning&&typeof r.x=="number"&&(a.x=r.x),this.enabledPanning&&typeof r.y=="number"&&(a.y=r.y),this.enabledZooming&&typeof r.ratio=="number"&&(a.ratio=this.getBoundedRatio(r.ratio)),this.enabledRotation&&typeof r.angle=="number"&&(a.angle=r.angle),this.clean?this.clean(je(je({},this.getState()),a)):a}},{key:"isAnimated",value:function(){return!!this.nextFrame}},{key:"setState",value:function(r){if(!this.enabled)return this;this.previousState=this.getState();var a=this.validateState(r);return typeof a.x=="number"&&(this.x=a.x),typeof a.y=="number"&&(this.y=a.y),typeof a.ratio=="number"&&(this.ratio=a.ratio),typeof a.angle=="number"&&(this.angle=a.angle),this.hasState(this.previousState)||this.emit("updated",this.getState()),this}},{key:"updateState",value:function(r){return this.setState(r(this.getState())),this}},{key:"animate",value:function(r){var a=this,o=arguments.length>1&&arguments[1]!==void 0?arguments[1]:{},s=arguments.length>2?arguments[2]:void 0;if(!s)return new Promise(function(b){return a.animate(r,o,b)});if(this.enabled){var u=je(je({},M4),o),c=this.validateState(r),d=typeof u.easing=="function"?u.easing:L4[u.easing],p=Date.now(),g=this.getState(),m=function(){var y=(Date.now()-p)/u.duration;if(y>=1){a.nextFrame=null,a.setState(c),a.animationCallback&&(a.animationCallback.call(null),a.animationCallback=void 0);return}var S=d(y),x={};typeof c.x=="number"&&(x.x=g.x+(c.x-g.x)*S),typeof c.y=="number"&&(x.y=g.y+(c.y-g.y)*S),a.enabledRotation&&typeof c.angle=="number"&&(x.angle=g.angle+(c.angle-g.angle)*S),typeof c.ratio=="number"&&(x.ratio=g.ratio+(c.ratio-g.ratio)*S),a.setState(x),a.nextFrame=requestAnimationFrame(m)};this.nextFrame?(cancelAnimationFrame(this.nextFrame),this.animationCallback&&this.animationCallback.call(null),this.nextFrame=requestAnimationFrame(m)):m(),this.animationCallback=s}}},{key:"animatedZoom",value:function(r){return r?typeof r=="number"?this.animate({ratio:this.ratio/r}):this.animate({ratio:this.ratio/(r.factor||Nd)},r):this.animate({ratio:this.ratio/Nd})}},{key:"animatedUnzoom",value:function(r){return r?typeof r=="number"?this.animate({ratio:this.ratio*r}):this.animate({ratio:this.ratio*(r.factor||Nd)},r):this.animate({ratio:this.ratio*Nd})}},{key:"animatedReset",value:function(r){return this.animate({x:.5,y:.5,ratio:1,angle:0},r)}},{key:"copy",value:function(){return t.from(this.getState())}}],[{key:"from",value:function(r){var a=new t;return a.setState(r)}}])}(bT);function Pr(e,t){var n=t.getBoundingClientRect();return{x:e.clientX-n.left,y:e.clientY-n.top}}function La(e,t){var n=je(je({},Pr(e,t)),{},{sigmaDefaultPrevented:!1,preventSigmaDefault:function(){n.sigmaDefaultPrevented=!0},original:e});return n}function ou(e){var t="x"in e?e:je(je({},e.touches[0]||e.previousTouches[0]),{},{original:e.original,sigmaDefaultPrevented:e.sigmaDefaultPrevented,preventSigmaDefault:function(){e.sigmaDefaultPrevented=!0,t.sigmaDefaultPrevented=!0}});return t}function dte(e,t){return je(je({},La(e,t)),{},{delta:P4(e)})}var fte=2;function ef(e){for(var t=[],n=0,r=Math.min(e.length,fte);n<r;n++)t.push(e[n]);return t}function iu(e,t,n){var r={touches:ef(e.touches).map(function(a){return Pr(a,n)}),previousTouches:t.map(function(a){return Pr(a,n)}),sigmaDefaultPrevented:!1,preventSigmaDefault:function(){r.sigmaDefaultPrevented=!0},original:e};return r}function P4(e){if(typeof e.deltaY<"u")return e.deltaY*-3/360;if(typeof e.detail<"u")return e.detail/-9;throw new Error("Captor: could not extract delta from event.")}var z4=function(e){function t(n,r){var a;return cn(this,t),a=$n(this,t),a.container=n,a.renderer=r,a}return qn(t,e),dn(t)}(bT),pte=["doubleClickTimeout","doubleClickZoomingDuration","doubleClickZoomingRatio","dragTimeout","draggedEventsTolerance","inertiaDuration","inertiaRatio","zoomDuration","zoomingRatio"],gte=pte.reduce(function(e,t){return je(je({},e),{},Re({},t,yT[t]))},{}),hte=function(e){function t(n,r){var a;return cn(this,t),a=$n(this,t,[n,r]),Re(a,"enabled",!0),Re(a,"draggedEvents",0),Re(a,"downStartTime",null),Re(a,"lastMouseX",null),Re(a,"lastMouseY",null),Re(a,"isMouseDown",!1),Re(a,"isMoving",!1),Re(a,"movingTimeout",null),Re(a,"startCameraState",null),Re(a,"clicks",0),Re(a,"doubleClickTimeout",null),Re(a,"currentWheelDirection",0),Re(a,"settings",gte),a.handleClick=a.handleClick.bind(a),a.handleRightClick=a.handleRightClick.bind(a),a.handleDown=a.handleDown.bind(a),a.handleUp=a.handleUp.bind(a),a.handleMove=a.handleMove.bind(a),a.handleWheel=a.handleWheel.bind(a),a.handleLeave=a.handleLeave.bind(a),a.handleEnter=a.handleEnter.bind(a),n.addEventListener("click",a.handleClick,{capture:!1}),n.addEventListener("contextmenu",a.handleRightClick,{capture:!1}),n.addEventListener("mousedown",a.handleDown,{capture:!1}),n.addEventListener("wheel",a.handleWheel,{capture:!1}),n.addEventListener("mouseleave",a.handleLeave,{capture:!1}),n.addEventListener("mouseenter",a.handleEnter,{capture:!1}),document.addEventListener("mousemove",a.handleMove,{capture:!1}),document.addEventListener("mouseup",a.handleUp,{capture:!1}),a}return qn(t,e),dn(t,[{key:"kill",value:function(){var r=this.container;r.removeEventListener("click",this.handleClick),r.removeEventListener("contextmenu",this.handleRightClick),r.removeEventListener("mousedown",this.handleDown),r.removeEventListener("wheel",this.handleWheel),r.removeEventListener("mouseleave",this.handleLeave),r.removeEventListener("mouseenter",this.handleEnter),document.removeEventListener("mousemove",this.handleMove),document.removeEventListener("mouseup",this.handleUp)}},{key:"handleClick",value:function(r){var a=this;if(this.enabled){if(this.clicks++,this.clicks===2)return this.clicks=0,typeof this.doubleClickTimeout=="number"&&(clearTimeout(this.doubleClickTimeout),this.doubleClickTimeout=null),this.handleDoubleClick(r);setTimeout(function(){a.clicks=0,a.doubleClickTimeout=null},this.settings.doubleClickTimeout),this.draggedEvents<this.settings.draggedEventsTolerance&&this.emit("click",La(r,this.container))}}},{key:"handleRightClick",value:function(r){this.enabled&&this.emit("rightClick",La(r,this.container))}},{key:"handleDoubleClick",value:function(r){if(this.enabled){r.preventDefault(),r.stopPropagation();var a=La(r,this.container);if(this.emit("doubleClick",a),!a.sigmaDefaultPrevented){var o=this.renderer.getCamera(),s=o.getBoundedRatio(o.getState().ratio/this.settings.doubleClickZoomingRatio);o.animate(this.renderer.getViewportZoomedState(Pr(r,this.container),s),{easing:"quadraticInOut",duration:this.settings.doubleClickZoomingDuration})}}}},{key:"handleDown",value:function(r){if(this.enabled){if(r.button===0){this.startCameraState=this.renderer.getCamera().getState();var a=Pr(r,this.container),o=a.x,s=a.y;this.lastMouseX=o,this.lastMouseY=s,this.draggedEvents=0,this.downStartTime=Date.now(),this.isMouseDown=!0}this.emit("mousedown",La(r,this.container))}}},{key:"handleUp",value:function(r){var a=this;if(!(!this.enabled||!this.isMouseDown)){var o=this.renderer.getCamera();this.isMouseDown=!1,typeof this.movingTimeout=="number"&&(clearTimeout(this.movingTimeout),this.movingTimeout=null);var s=Pr(r,this.container),u=s.x,c=s.y,d=o.getState(),p=o.getPreviousState()||{x:0,y:0};this.isMoving?o.animate({x:d.x+this.settings.inertiaRatio*(d.x-p.x),y:d.y+this.settings.inertiaRatio*(d.y-p.y)},{duration:this.settings.inertiaDuration,easing:"quadraticOut"}):(this.lastMouseX!==u||this.lastMouseY!==c)&&o.setState({x:d.x,y:d.y}),this.isMoving=!1,setTimeout(function(){var g=a.draggedEvents>0;a.draggedEvents=0,g&&a.renderer.getSetting("hideEdgesOnMove")&&a.renderer.refresh()},0),this.emit("mouseup",La(r,this.container))}}},{key:"handleMove",value:function(r){var a=this;if(this.enabled){var o=La(r,this.container);if(this.emit("mousemovebody",o),(r.target===this.container||r.composedPath()[0]===this.container)&&this.emit("mousemove",o),!o.sigmaDefaultPrevented&&this.isMouseDown){this.isMoving=!0,this.draggedEvents++,typeof this.movingTimeout=="number"&&clearTimeout(this.movingTimeout),this.movingTimeout=window.setTimeout(function(){a.movingTimeout=null,a.isMoving=!1},this.settings.dragTimeout);var s=this.renderer.getCamera(),u=Pr(r,this.container),c=u.x,d=u.y,p=this.renderer.viewportToFramedGraph({x:this.lastMouseX,y:this.lastMouseY}),g=this.renderer.viewportToFramedGraph({x:c,y:d}),m=p.x-g.x,b=p.y-g.y,y=s.getState(),S=y.x+m,x=y.y+b;s.setState({x:S,y:x}),this.lastMouseX=c,this.lastMouseY=d,r.preventDefault(),r.stopPropagation()}}}},{key:"handleLeave",value:function(r){this.emit("mouseleave",La(r,this.container))}},{key:"handleEnter",value:function(r){this.emit("mouseenter",La(r,this.container))}},{key:"handleWheel",value:function(r){var a=this,o=this.renderer.getCamera();if(!(!this.enabled||!o.enabledZooming)){var s=P4(r);if(s){var u=dte(r,this.container);if(this.emit("wheel",u),u.sigmaDefaultPrevented){r.preventDefault(),r.stopPropagation();return}var c=o.getState().ratio,d=s>0?1/this.settings.zoomingRatio:this.settings.zoomingRatio,p=o.getBoundedRatio(c*d),g=s>0?1:-1,m=Date.now();c!==p&&(r.preventDefault(),r.stopPropagation(),!(this.currentWheelDirection===g&&this.lastWheelTriggerTime&&m-this.lastWheelTriggerTime<this.settings.zoomDuration/5)&&(o.animate(this.renderer.getViewportZoomedState(Pr(r,this.container),p),{easing:"quadraticOut",duration:this.settings.zoomDuration},function(){a.currentWheelDirection=0}),this.currentWheelDirection=g,this.lastWheelTriggerTime=m))}}}},{key:"setSettings",value:function(r){this.settings=r}}])}(z4),mte=["dragTimeout","inertiaDuration","inertiaRatio","doubleClickTimeout","doubleClickZoomingRatio","doubleClickZoomingDuration","tapMoveTolerance"],bte=mte.reduce(function(e,t){return je(je({},e),{},Re({},t,yT[t]))},{}),yte=function(e){function t(n,r){var a;return cn(this,t),a=$n(this,t,[n,r]),Re(a,"enabled",!0),Re(a,"isMoving",!1),Re(a,"hasMoved",!1),Re(a,"touchMode",0),Re(a,"startTouchesPositions",[]),Re(a,"lastTouches",[]),Re(a,"lastTap",null),Re(a,"settings",bte),a.handleStart=a.handleStart.bind(a),a.handleLeave=a.handleLeave.bind(a),a.handleMove=a.handleMove.bind(a),n.addEventListener("touchstart",a.handleStart,{capture:!1}),n.addEventListener("touchcancel",a.handleLeave,{capture:!1}),document.addEventListener("touchend",a.handleLeave,{capture:!1,passive:!1}),document.addEventListener("touchmove",a.handleMove,{capture:!1,passive:!1}),a}return qn(t,e),dn(t,[{key:"kill",value:function(){var r=this.container;r.removeEventListener("touchstart",this.handleStart),r.removeEventListener("touchcancel",this.handleLeave),document.removeEventListener("touchend",this.handleLeave),document.removeEventListener("touchmove",this.handleMove)}},{key:"getDimensions",value:function(){return{width:this.container.offsetWidth,height:this.container.offsetHeight}}},{key:"handleStart",value:function(r){var a=this;if(this.enabled){r.preventDefault();var o=ef(r.touches);if(this.touchMode=o.length,this.startCameraState=this.renderer.getCamera().getState(),this.startTouchesPositions=o.map(function(b){return Pr(b,a.container)}),this.touchMode===2){var s=Ls(this.startTouchesPositions,2),u=s[0],c=u.x,d=u.y,p=s[1],g=p.x,m=p.y;this.startTouchesAngle=Math.atan2(m-d,g-c),this.startTouchesDistance=Math.sqrt(Math.pow(g-c,2)+Math.pow(m-d,2))}this.emit("touchdown",iu(r,this.lastTouches,this.container)),this.lastTouches=o,this.lastTouchesPositions=this.startTouchesPositions}}},{key:"handleLeave",value:function(r){if(!(!this.enabled||!this.startTouchesPositions.length)){switch(r.cancelable&&r.preventDefault(),this.movingTimeout&&(this.isMoving=!1,clearTimeout(this.movingTimeout)),this.touchMode){case 2:if(r.touches.length===1){this.handleStart(r),r.preventDefault();break}case 1:if(this.isMoving){var a=this.renderer.getCamera(),o=a.getState(),s=a.getPreviousState()||{x:0,y:0};a.animate({x:o.x+this.settings.inertiaRatio*(o.x-s.x),y:o.y+this.settings.inertiaRatio*(o.y-s.y)},{duration:this.settings.inertiaDuration,easing:"quadraticOut"})}this.hasMoved=!1,this.isMoving=!1,this.touchMode=0;break}if(this.emit("touchup",iu(r,this.lastTouches,this.container)),!r.touches.length){var u=Pr(this.lastTouches[0],this.container),c=this.startTouchesPositions[0],d=Math.pow(u.x-c.x,2)+Math.pow(u.y-c.y,2);if(!r.touches.length&&d<Math.pow(this.settings.tapMoveTolerance,2))if(this.lastTap&&Date.now()-this.lastTap.time<this.settings.doubleClickTimeout){var p=iu(r,this.lastTouches,this.container);if(this.emit("doubletap",p),this.lastTap=null,!p.sigmaDefaultPrevented){var g=this.renderer.getCamera(),m=g.getBoundedRatio(g.getState().ratio/this.settings.doubleClickZoomingRatio);g.animate(this.renderer.getViewportZoomedState(u,m),{easing:"quadraticInOut",duration:this.settings.doubleClickZoomingDuration})}}else{var b=iu(r,this.lastTouches,this.container);this.emit("tap",b),this.lastTap={time:Date.now(),position:b.touches[0]||b.previousTouches[0]}}}this.lastTouches=ef(r.touches),this.startTouchesPositions=[]}}},{key:"handleMove",value:function(r){var a=this;if(!(!this.enabled||!this.startTouchesPositions.length)){r.preventDefault();var o=ef(r.touches),s=o.map(function(V){return Pr(V,a.container)}),u=this.lastTouches;this.lastTouches=o,this.lastTouchesPositions=s;var c=iu(r,u,this.container);if(this.emit("touchmove",c),!c.sigmaDefaultPrevented&&(this.hasMoved||(this.hasMoved=s.some(function(V,B){var M=a.startTouchesPositions[B];return M&&(V.x!==M.x||V.y!==M.y)})),!!this.hasMoved)){this.isMoving=!0,this.movingTimeout&&clearTimeout(this.movingTimeout),this.movingTimeout=window.setTimeout(function(){a.isMoving=!1},this.settings.dragTimeout);var d=this.renderer.getCamera(),p=this.startCameraState,g=this.renderer.getSetting("stagePadding");switch(this.touchMode){case 1:{var m=this.renderer.viewportToFramedGraph((this.startTouchesPositions||[])[0]),b=m.x,y=m.y,S=this.renderer.viewportToFramedGraph(s[0]),x=S.x,R=S.y;d.setState({x:p.x+b-x,y:p.y+y-R});break}case 2:{var k={x:.5,y:.5,angle:0,ratio:1},A=s[0],C=A.x,N=A.y,_=s[1],O=_.x,F=_.y,L=Math.atan2(F-N,O-C)-this.startTouchesAngle,I=Math.hypot(F-N,O-C)/this.startTouchesDistance,H=d.getBoundedRatio(p.ratio/I);k.ratio=H,k.angle=p.angle+L;var $=this.getDimensions(),U=this.renderer.viewportToFramedGraph((this.startTouchesPositions||[])[0],{cameraState:p}),Y=Math.min($.width,$.height)-2*g,Z=Y/$.width,j=Y/$.height,G=H/Y,z=C-Y/2/Z,K=N-Y/2/j,D=[z*Math.cos(-k.angle)-K*Math.sin(-k.angle),K*Math.cos(-k.angle)+z*Math.sin(-k.angle)];z=D[0],K=D[1],k.x=U.x-z*G,k.y=U.y+K*G,d.setState(k);break}}}}}},{key:"setSettings",value:function(r){this.settings=r}}])}(z4);function vte(e){if(Array.isArray(e))return Pk(e)}function Ste(e){if(typeof Symbol<"u"&&e[Symbol.iterator]!=null||e["@@iterator"]!=null)return Array.from(e)}function Ete(){throw new TypeError(`Invalid attempt to spread non-iterable instance.
In order to be iterable, non-array objects must have a [Symbol.iterator]() method.`)}function _N(e){return vte(e)||Ste(e)||y4(e)||Ete()}function wte(e,t){if(e==null)return{};var n={};for(var r in e)if({}.hasOwnProperty.call(e,r)){if(t.includes(r))continue;n[r]=e[r]}return n}function wm(e,t){if(e==null)return{};var n,r,a=wte(e,t);if(Object.getOwnPropertySymbols){var o=Object.getOwnPropertySymbols(e);for(r=0;r<o.length;r++)n=o[r],t.includes(n)||{}.propertyIsEnumerable.call(e,n)&&(a[n]=e[n])}return a}var CN=function(){function e(t,n){cn(this,e),this.key=t,this.size=n}return dn(e,null,[{key:"compare",value:function(n,r){return n.size>r.size?-1:n.size<r.size||n.key>r.key?1:-1}}])}(),NN=function(){function e(){cn(this,e),Re(this,"width",0),Re(this,"height",0),Re(this,"cellSize",0),Re(this,"columns",0),Re(this,"rows",0),Re(this,"cells",{})}return dn(e,[{key:"resizeAndClear",value:function(n,r){this.width=n.width,this.height=n.height,this.cellSize=r,this.columns=Math.ceil(n.width/r),this.rows=Math.ceil(n.height/r),this.cells={}}},{key:"getIndex",value:function(n){var r=Math.floor(n.x/this.cellSize),a=Math.floor(n.y/this.cellSize);return a*this.columns+r}},{key:"add",value:function(n,r,a){var o=new CN(n,r),s=this.getIndex(a),u=this.cells[s];u||(u=[],this.cells[s]=u),u.push(o)}},{key:"organize",value:function(){for(var n in this.cells){var r=this.cells[n];r.sort(CN.compare)}}},{key:"getLabelsToDisplay",value:function(n,r){var a=this.cellSize*this.cellSize,o=a/n/n,s=o*r/a,u=Math.ceil(s),c=[];for(var d in this.cells)for(var p=this.cells[d],g=0;g<Math.min(u,p.length);g++)c.push(p[g].key);return c}}])}();function xte(e){var t=e.graph,n=e.hoveredNode,r=e.highlightedNodes,a=e.displayedNodeLabels,o=[];return t.forEachEdge(function(s,u,c,d){(c===n||d===n||r.has(c)||r.has(d)||a.has(c)&&a.has(d))&&o.push(s)}),o}var kte=150,Tte=50,Pa=Object.prototype.hasOwnProperty;function Ate(e,t,n){if(!Pa.call(n,"x")||!Pa.call(n,"y"))throw new Error('Sigma: could not find a valid position (x, y) for node "'.concat(t,'". All your nodes must have a number "x" and "y". Maybe your forgot to apply a layout or your "nodeReducer" is not returning the correct data?'));return n.color||(n.color=e.defaultNodeColor),!n.label&&n.label!==""&&(n.label=null),n.label!==void 0&&n.label!==null?n.label=""+n.label:n.label=null,n.size||(n.size=2),Pa.call(n,"hidden")||(n.hidden=!1),Pa.call(n,"highlighted")||(n.highlighted=!1),Pa.call(n,"forceLabel")||(n.forceLabel=!1),(!n.type||n.type==="")&&(n.type=e.defaultNodeType),n.zIndex||(n.zIndex=0),n}function Rte(e,t,n){return n.color||(n.color=e.defaultEdgeColor),n.label||(n.label=""),n.size||(n.size=.5),Pa.call(n,"hidden")||(n.hidden=!1),Pa.call(n,"forceLabel")||(n.forceLabel=!1),(!n.type||n.type==="")&&(n.type=e.defaultEdgeType),n.zIndex||(n.zIndex=0),n}var _te=function(e){function t(n,r){var a,o=arguments.length>2&&arguments[2]!==void 0?arguments[2]:{};if(cn(this,t),a=$n(this,t),Re(a,"elements",{}),Re(a,"canvasContexts",{}),Re(a,"webGLContexts",{}),Re(a,"pickingLayers",new Set),Re(a,"textures",{}),Re(a,"frameBuffers",{}),Re(a,"activeListeners",{}),Re(a,"labelGrid",new NN),Re(a,"nodeDataCache",{}),Re(a,"edgeDataCache",{}),Re(a,"nodeProgramIndex",{}),Re(a,"edgeProgramIndex",{}),Re(a,"nodesWithForcedLabels",new Set),Re(a,"edgesWithForcedLabels",new Set),Re(a,"nodeExtent",{x:[0,1],y:[0,1]}),Re(a,"nodeZExtent",[1/0,-1/0]),Re(a,"edgeZExtent",[1/0,-1/0]),Re(a,"matrix",Mr()),Re(a,"invMatrix",Mr()),Re(a,"correctionRatio",1),Re(a,"customBBox",null),Re(a,"normalizationFunction",TN({x:[0,1],y:[0,1]})),Re(a,"graphToViewportRatio",1),Re(a,"itemIDsIndex",{}),Re(a,"nodeIndices",{}),Re(a,"edgeIndices",{}),Re(a,"width",0),Re(a,"height",0),Re(a,"pixelRatio",xN()),Re(a,"pickingDownSizingRatio",2*a.pixelRatio),Re(a,"displayedNodeLabels",new Set),Re(a,"displayedEdgeLabels",new Set),Re(a,"highlightedNodes",new Set),Re(a,"hoveredNode",null),Re(a,"hoveredEdge",null),Re(a,"renderFrame",null),Re(a,"renderHighlightedNodesFrame",null),Re(a,"needToProcess",!1),Re(a,"checkEdgesEventsFrame",null),Re(a,"nodePrograms",{}),Re(a,"nodeHoverPrograms",{}),Re(a,"edgePrograms",{}),a.settings=cte(o),Em(a.settings),ite(n),!(r instanceof HTMLElement))throw new Error("Sigma: container should be an html element.");a.graph=n,a.container=r,a.createWebGLContext("edges",{picking:o.enableEdgeEvents}),a.createCanvasContext("edgeLabels"),a.createWebGLContext("nodes",{picking:!0}),a.createCanvasContext("labels"),a.createCanvasContext("hovers"),a.createWebGLContext("hoverNodes"),a.createCanvasContext("mouse",{style:{touchAction:"none",userSelect:"none"}}),a.resize();for(var s in a.settings.nodeProgramClasses)a.registerNodeProgram(s,a.settings.nodeProgramClasses[s],a.settings.nodeHoverProgramClasses[s]);for(var u in a.settings.edgeProgramClasses)a.registerEdgeProgram(u,a.settings.edgeProgramClasses[u]);return a.camera=new RN,a.bindCameraHandlers(),a.mouseCaptor=new hte(a.elements.mouse,a),a.mouseCaptor.setSettings(a.settings),a.touchCaptor=new yte(a.elements.mouse,a),a.touchCaptor.setSettings(a.settings),a.bindEventHandlers(),a.bindGraphHandlers(),a.handleSettingsUpdate(),a.refresh(),a}return qn(t,e),dn(t,[{key:"registerNodeProgram",value:function(r,a,o){return this.nodePrograms[r]&&this.nodePrograms[r].kill(),this.nodeHoverPrograms[r]&&this.nodeHoverPrograms[r].kill(),this.nodePrograms[r]=new a(this.webGLContexts.nodes,this.frameBuffers.nodes,this),this.nodeHoverPrograms[r]=new(o||a)(this.webGLContexts.hoverNodes,null,this),this}},{key:"registerEdgeProgram",value:function(r,a){return this.edgePrograms[r]&&this.edgePrograms[r].kill(),this.edgePrograms[r]=new a(this.webGLContexts.edges,this.frameBuffers.edges,this),this}},{key:"unregisterNodeProgram",value:function(r){if(this.nodePrograms[r]){var a=this.nodePrograms,o=a[r],s=wm(a,[r].map(hu));o.kill(),this.nodePrograms=s}if(this.nodeHoverPrograms[r]){var u=this.nodeHoverPrograms,c=u[r],d=wm(u,[r].map(hu));c.kill(),this.nodePrograms=d}return this}},{key:"unregisterEdgeProgram",value:function(r){if(this.edgePrograms[r]){var a=this.edgePrograms,o=a[r],s=wm(a,[r].map(hu));o.kill(),this.edgePrograms=s}return this}},{key:"resetWebGLTexture",value:function(r){var a=this.webGLContexts[r],o=this.frameBuffers[r],s=this.textures[r];s&&a.deleteTexture(s);var u=a.createTexture();return a.bindFramebuffer(a.FRAMEBUFFER,o),a.bindTexture(a.TEXTURE_2D,u),a.texImage2D(a.TEXTURE_2D,0,a.RGBA,this.width,this.height,0,a.RGBA,a.UNSIGNED_BYTE,null),a.framebufferTexture2D(a.FRAMEBUFFER,a.COLOR_ATTACHMENT0,a.TEXTURE_2D,u,0),this.textures[r]=u,this}},{key:"bindCameraHandlers",value:function(){var r=this;return this.activeListeners.camera=function(){r.scheduleRender()},this.camera.on("updated",this.activeListeners.camera),this}},{key:"unbindCameraHandlers",value:function(){return this.camera.removeListener("updated",this.activeListeners.camera),this}},{key:"getNodeAtPosition",value:function(r){var a=r.x,o=r.y,s=dN(this.webGLContexts.nodes,this.frameBuffers.nodes,a,o,this.pixelRatio,this.pickingDownSizingRatio),u=cN.apply(void 0,_N(s)),c=this.itemIDsIndex[u];return c&&c.type==="node"?c.id:null}},{key:"bindEventHandlers",value:function(){var r=this;this.activeListeners.handleResize=function(){r.scheduleRefresh()},window.addEventListener("resize",this.activeListeners.handleResize),this.activeListeners.handleMove=function(o){var s=ou(o),u={event:s,preventSigmaDefault:function(){s.preventSigmaDefault()}},c=r.getNodeAtPosition(s);if(c&&r.hoveredNode!==c&&!r.nodeDataCache[c].hidden){r.hoveredNode&&r.emit("leaveNode",je(je({},u),{},{node:r.hoveredNode})),r.hoveredNode=c,r.emit("enterNode",je(je({},u),{},{node:c})),r.scheduleHighlightedNodesRender();return}if(r.hoveredNode&&r.getNodeAtPosition(s)!==r.hoveredNode){var d=r.hoveredNode;r.hoveredNode=null,r.emit("leaveNode",je(je({},u),{},{node:d})),r.scheduleHighlightedNodesRender();return}if(r.settings.enableEdgeEvents){var p=r.hoveredNode?null:r.getEdgeAtPoint(u.event.x,u.event.y);p!==r.hoveredEdge&&(r.hoveredEdge&&r.emit("leaveEdge",je(je({},u),{},{edge:r.hoveredEdge})),p&&r.emit("enterEdge",je(je({},u),{},{edge:p})),r.hoveredEdge=p)}},this.activeListeners.handleMoveBody=function(o){var s=ou(o);r.emit("moveBody",{event:s,preventSigmaDefault:function(){s.preventSigmaDefault()}})},this.activeListeners.handleLeave=function(o){var s=ou(o),u={event:s,preventSigmaDefault:function(){s.preventSigmaDefault()}};r.hoveredNode&&(r.emit("leaveNode",je(je({},u),{},{node:r.hoveredNode})),r.scheduleHighlightedNodesRender()),r.settings.enableEdgeEvents&&r.hoveredEdge&&(r.emit("leaveEdge",je(je({},u),{},{edge:r.hoveredEdge})),r.scheduleHighlightedNodesRender()),r.emit("leaveStage",je({},u))},this.activeListeners.handleEnter=function(o){var s=ou(o),u={event:s,preventSigmaDefault:function(){s.preventSigmaDefault()}};r.emit("enterStage",je({},u))};var a=function(s){return function(u){var c=ou(u),d={event:c,preventSigmaDefault:function(){c.preventSigmaDefault()}},p=r.getNodeAtPosition(c);if(p)return r.emit("".concat(s,"Node"),je(je({},d),{},{node:p}));if(r.settings.enableEdgeEvents){var g=r.getEdgeAtPoint(c.x,c.y);if(g)return r.emit("".concat(s,"Edge"),je(je({},d),{},{edge:g}))}return r.emit("".concat(s,"Stage"),d)}};return this.activeListeners.handleClick=a("click"),this.activeListeners.handleRightClick=a("rightClick"),this.activeListeners.handleDoubleClick=a("doubleClick"),this.activeListeners.handleWheel=a("wheel"),this.activeListeners.handleDown=a("down"),this.activeListeners.handleUp=a("up"),this.mouseCaptor.on("mousemove",this.activeListeners.handleMove),this.mouseCaptor.on("mousemovebody",this.activeListeners.handleMoveBody),this.mouseCaptor.on("click",this.activeListeners.handleClick),this.mouseCaptor.on("rightClick",this.activeListeners.handleRightClick),this.mouseCaptor.on("doubleClick",this.activeListeners.handleDoubleClick),this.mouseCaptor.on("wheel",this.activeListeners.handleWheel),this.mouseCaptor.on("mousedown",this.activeListeners.handleDown),this.mouseCaptor.on("mouseup",this.activeListeners.handleUp),this.mouseCaptor.on("mouseleave",this.activeListeners.handleLeave),this.mouseCaptor.on("mouseenter",this.activeListeners.handleEnter),this.touchCaptor.on("touchdown",this.activeListeners.handleDown),this.touchCaptor.on("touchdown",this.activeListeners.handleMove),this.touchCaptor.on("touchup",this.activeListeners.handleUp),this.touchCaptor.on("touchmove",this.activeListeners.handleMove),this.touchCaptor.on("tap",this.activeListeners.handleClick),this.touchCaptor.on("doubletap",this.activeListeners.handleDoubleClick),this.touchCaptor.on("touchmove",this.activeListeners.handleMoveBody),this}},{key:"bindGraphHandlers",value:function(){var r=this,a=this.graph,o=new Set(["x","y","zIndex","type"]);return this.activeListeners.eachNodeAttributesUpdatedGraphUpdate=function(s){var u,c=(u=s.hints)===null||u===void 0?void 0:u.attributes;r.graph.forEachNode(function(p){return r.updateNode(p)});var d=!c||c.some(function(p){return o.has(p)});r.refresh({partialGraph:{nodes:a.nodes()},skipIndexation:!d,schedule:!0})},this.activeListeners.eachEdgeAttributesUpdatedGraphUpdate=function(s){var u,c=(u=s.hints)===null||u===void 0?void 0:u.attributes;r.graph.forEachEdge(function(p){return r.updateEdge(p)});var d=c&&["zIndex","type"].some(function(p){return c==null?void 0:c.includes(p)});r.refresh({partialGraph:{edges:a.edges()},skipIndexation:!d,schedule:!0})},this.activeListeners.addNodeGraphUpdate=function(s){var u=s.key;r.addNode(u),r.refresh({partialGraph:{nodes:[u]},skipIndexation:!1,schedule:!0})},this.activeListeners.updateNodeGraphUpdate=function(s){var u=s.key;r.refresh({partialGraph:{nodes:[u]},skipIndexation:!1,schedule:!0})},this.activeListeners.dropNodeGraphUpdate=function(s){var u=s.key;r.removeNode(u),r.refresh({schedule:!0})},this.activeListeners.addEdgeGraphUpdate=function(s){var u=s.key;r.addEdge(u),r.refresh({partialGraph:{edges:[u]},schedule:!0})},this.activeListeners.updateEdgeGraphUpdate=function(s){var u=s.key;r.refresh({partialGraph:{edges:[u]},skipIndexation:!1,schedule:!0})},this.activeListeners.dropEdgeGraphUpdate=function(s){var u=s.key;r.removeEdge(u),r.refresh({schedule:!0})},this.activeListeners.clearEdgesGraphUpdate=function(){r.clearEdgeState(),r.clearEdgeIndices(),r.refresh({schedule:!0})},this.activeListeners.clearGraphUpdate=function(){r.clearEdgeState(),r.clearNodeState(),r.clearEdgeIndices(),r.clearNodeIndices(),r.refresh({schedule:!0})},a.on("nodeAdded",this.activeListeners.addNodeGraphUpdate),a.on("nodeDropped",this.activeListeners.dropNodeGraphUpdate),a.on("nodeAttributesUpdated",this.activeListeners.updateNodeGraphUpdate),a.on("eachNodeAttributesUpdated",this.activeListeners.eachNodeAttributesUpdatedGraphUpdate),a.on("edgeAdded",this.activeListeners.addEdgeGraphUpdate),a.on("edgeDropped",this.activeListeners.dropEdgeGraphUpdate),a.on("edgeAttributesUpdated",this.activeListeners.updateEdgeGraphUpdate),a.on("eachEdgeAttributesUpdated",this.activeListeners.eachEdgeAttributesUpdatedGraphUpdate),a.on("edgesCleared",this.activeListeners.clearEdgesGraphUpdate),a.on("cleared",this.activeListeners.clearGraphUpdate),this}},{key:"unbindGraphHandlers",value:function(){var r=this.graph;r.removeListener("nodeAdded",this.activeListeners.addNodeGraphUpdate),r.removeListener("nodeDropped",this.activeListeners.dropNodeGraphUpdate),r.removeListener("nodeAttributesUpdated",this.activeListeners.updateNodeGraphUpdate),r.removeListener("eachNodeAttributesUpdated",this.activeListeners.eachNodeAttributesUpdatedGraphUpdate),r.removeListener("edgeAdded",this.activeListeners.addEdgeGraphUpdate),r.removeListener("edgeDropped",this.activeListeners.dropEdgeGraphUpdate),r.removeListener("edgeAttributesUpdated",this.activeListeners.updateEdgeGraphUpdate),r.removeListener("eachEdgeAttributesUpdated",this.activeListeners.eachEdgeAttributesUpdatedGraphUpdate),r.removeListener("edgesCleared",this.activeListeners.clearEdgesGraphUpdate),r.removeListener("cleared",this.activeListeners.clearGraphUpdate)}},{key:"getEdgeAtPoint",value:function(r,a){var o=dN(this.webGLContexts.edges,this.frameBuffers.edges,r,a,this.pixelRatio,this.pickingDownSizingRatio),s=cN.apply(void 0,_N(o)),u=this.itemIDsIndex[s];return u&&u.type==="edge"?u.id:null}},{key:"process",value:function(){var r=this;this.emit("beforeProcess");var a=this.graph,o=this.settings,s=this.getDimensions();if(this.nodeExtent=ote(this.graph),!this.settings.autoRescale){var u=s.width,c=s.height,d=this.nodeExtent,p=d.x,g=d.y;this.nodeExtent={x:[(p[0]+p[1])/2-u/2,(p[0]+p[1])/2+u/2],y:[(g[0]+g[1])/2-c/2,(g[0]+g[1])/2+c/2]}}this.normalizationFunction=TN(this.customBBox||this.nodeExtent);var m=new RN,b=au(m.getState(),s,this.getGraphDimensions(),this.getStagePadding());this.labelGrid.resizeAndClear(s,o.labelGridCellSize);for(var y={},S={},x={},R={},k=1,A=a.nodes(),C=0,N=A.length;C<N;C++){var _=A[C],O=this.nodeDataCache[_],F=a.getNodeAttributes(_);O.x=F.x,O.y=F.y,this.normalizationFunction.applyTo(O),typeof O.label=="string"&&!O.hidden&&this.labelGrid.add(_,O.size,this.framedGraphToViewport(O,{matrix:b})),y[O.type]=(y[O.type]||0)+1}this.labelGrid.organize();for(var L in this.nodePrograms){if(!Pa.call(this.nodePrograms,L))throw new Error('Sigma: could not find a suitable program for node type "'.concat(L,'"!'));this.nodePrograms[L].reallocate(y[L]||0),y[L]=0}this.settings.zIndex&&this.nodeZExtent[0]!==this.nodeZExtent[1]&&(A=kN(this.nodeZExtent,function(Q){return r.nodeDataCache[Q].zIndex},A));for(var I=0,H=A.length;I<H;I++){var $=A[I];S[$]=k,R[S[$]]={type:"node",id:$},k++;var U=this.nodeDataCache[$];this.addNodeToProgram($,S[$],y[U.type]++)}for(var Y={},Z=a.edges(),j=0,G=Z.length;j<G;j++){var z=Z[j],K=this.edgeDataCache[z];Y[K.type]=(Y[K.type]||0)+1}this.settings.zIndex&&this.edgeZExtent[0]!==this.edgeZExtent[1]&&(Z=kN(this.edgeZExtent,function(Q){return r.edgeDataCache[Q].zIndex},Z));for(var D in this.edgePrograms){if(!Pa.call(this.edgePrograms,D))throw new Error('Sigma: could not find a suitable program for edge type "'.concat(D,'"!'));this.edgePrograms[D].reallocate(Y[D]||0),Y[D]=0}for(var V=0,B=Z.length;V<B;V++){var M=Z[V];x[M]=k,R[x[M]]={type:"edge",id:M},k++;var W=this.edgeDataCache[M];this.addEdgeToProgram(M,x[M],Y[W.type]++)}return this.itemIDsIndex=R,this.nodeIndices=S,this.edgeIndices=x,this.emit("afterProcess"),this}},{key:"handleSettingsUpdate",value:function(r){var a=this,o=this.settings;if(this.camera.minRatio=o.minCameraRatio,this.camera.maxRatio=o.maxCameraRatio,this.camera.enabledZooming=o.enableCameraZooming,this.camera.enabledPanning=o.enableCameraPanning,this.camera.enabledRotation=o.enableCameraRotation,o.cameraPanBoundaries?this.camera.clean=function(p){return a.cleanCameraState(p,o.cameraPanBoundaries&&Uk(o.cameraPanBoundaries)==="object"?o.cameraPanBoundaries:{})}:this.camera.clean=null,this.camera.setState(this.camera.validateState(this.camera.getState())),r){if(r.edgeProgramClasses!==o.edgeProgramClasses){for(var s in o.edgeProgramClasses)o.edgeProgramClasses[s]!==r.edgeProgramClasses[s]&&this.registerEdgeProgram(s,o.edgeProgramClasses[s]);for(var u in r.edgeProgramClasses)o.edgeProgramClasses[u]||this.unregisterEdgeProgram(u)}if(r.nodeProgramClasses!==o.nodeProgramClasses||r.nodeHoverProgramClasses!==o.nodeHoverProgramClasses){for(var c in o.nodeProgramClasses)(o.nodeProgramClasses[c]!==r.nodeProgramClasses[c]||o.nodeHoverProgramClasses[c]!==r.nodeHoverProgramClasses[c])&&this.registerNodeProgram(c,o.nodeProgramClasses[c],o.nodeHoverProgramClasses[c]);for(var d in r.nodeProgramClasses)o.nodeProgramClasses[d]||this.unregisterNodeProgram(d)}}return this.mouseCaptor.setSettings(this.settings),this.touchCaptor.setSettings(this.settings),this}},{key:"cleanCameraState",value:function(r){var a=arguments.length>1&&arguments[1]!==void 0?arguments[1]:{},o=a.tolerance,s=o===void 0?0:o,u=a.boundaries,c=je({},r),d=u||this.nodeExtent,p=Ls(d.x,2),g=p[0],m=p[1],b=Ls(d.y,2),y=b[0],S=b[1],x=[this.graphToViewport({x:g,y},{cameraState:r}),this.graphToViewport({x:m,y},{cameraState:r}),this.graphToViewport({x:g,y:S},{cameraState:r}),this.graphToViewport({x:m,y:S},{cameraState:r})],R=1/0,k=-1/0,A=1/0,C=-1/0;x.forEach(function(Y){var Z=Y.x,j=Y.y;R=Math.min(R,Z),k=Math.max(k,Z),A=Math.min(A,j),C=Math.max(C,j)});var N=k-R,_=C-A,O=this.getDimensions(),F=O.width,L=O.height,I=0,H=0;if(N>=F?k<F-s?I=k-(F-s):R>s&&(I=R-s):k>F+s?I=k-(F+s):R<-s&&(I=R+s),_>=L?C<L-s?H=C-(L-s):A>s&&(H=A-s):C>L+s?H=C-(L+s):A<-s&&(H=A+s),I||H){var $=this.viewportToFramedGraph({x:0,y:0},{cameraState:r}),U=this.viewportToFramedGraph({x:I,y:H},{cameraState:r});I=U.x-$.x,H=U.y-$.y,c.x+=I,c.y+=H}return c}},{key:"renderLabels",value:function(){if(!this.settings.renderLabels)return this;var r=this.camera.getState(),a=this.labelGrid.getLabelsToDisplay(r.ratio,this.settings.labelDensity);AN(a,this.nodesWithForcedLabels),this.displayedNodeLabels=new Set;for(var o=this.canvasContexts.labels,s=0,u=a.length;s<u;s++){var c=a[s],d=this.nodeDataCache[c];if(!this.displayedNodeLabels.has(c)&&!d.hidden){var p=this.framedGraphToViewport(d),g=p.x,m=p.y,b=this.scaleSize(d.size);if(!(!d.forceLabel&&b<this.settings.labelRenderedSizeThreshold)&&!(g<-150||g>this.width+kte||m<-50||m>this.height+Tte)){this.displayedNodeLabels.add(c);var y=this.settings.defaultDrawNodeLabel,S=this.nodePrograms[d.type],x=(S==null?void 0:S.drawLabel)||y;x(o,je(je({key:c},d),{},{size:b,x:g,y:m}),this.settings)}}}return this}},{key:"renderEdgeLabels",value:function(){if(!this.settings.renderEdgeLabels)return this;var r=this.canvasContexts.edgeLabels;r.clearRect(0,0,this.width,this.height);var a=xte({graph:this.graph,hoveredNode:this.hoveredNode,displayedNodeLabels:this.displayedNodeLabels,highlightedNodes:this.highlightedNodes});AN(a,this.edgesWithForcedLabels);for(var o=new Set,s=0,u=a.length;s<u;s++){var c=a[s],d=this.graph.extremities(c),p=this.nodeDataCache[d[0]],g=this.nodeDataCache[d[1]],m=this.edgeDataCache[c];if(!o.has(c)&&!(m.hidden||p.hidden||g.hidden)){var b=this.settings.defaultDrawEdgeLabel,y=this.edgePrograms[m.type],S=(y==null?void 0:y.drawLabel)||b;S(r,je(je({key:c},m),{},{size:this.scaleSize(m.size)}),je(je(je({key:d[0]},p),this.framedGraphToViewport(p)),{},{size:this.scaleSize(p.size)}),je(je(je({key:d[1]},g),this.framedGraphToViewport(g)),{},{size:this.scaleSize(g.size)}),this.settings),o.add(c)}}return this.displayedEdgeLabels=o,this}},{key:"renderHighlightedNodes",value:function(){var r=this,a=this.canvasContexts.hovers;a.clearRect(0,0,this.width,this.height);var o=function(b){var y=r.nodeDataCache[b],S=r.framedGraphToViewport(y),x=S.x,R=S.y,k=r.scaleSize(y.size),A=r.settings.defaultDrawNodeHover,C=r.nodePrograms[y.type],N=(C==null?void 0:C.drawHover)||A;N(a,je(je({key:b},y),{},{size:k,x,y:R}),r.settings)},s=[];this.hoveredNode&&!this.nodeDataCache[this.hoveredNode].hidden&&s.push(this.hoveredNode),this.highlightedNodes.forEach(function(m){m!==r.hoveredNode&&s.push(m)}),s.forEach(function(m){return o(m)});var u={};s.forEach(function(m){var b=r.nodeDataCache[m].type;u[b]=(u[b]||0)+1});for(var c in this.nodeHoverPrograms)this.nodeHoverPrograms[c].reallocate(u[c]||0),u[c]=0;s.forEach(function(m){var b=r.nodeDataCache[m];r.nodeHoverPrograms[b.type].process(0,u[b.type]++,b)}),this.webGLContexts.hoverNodes.clear(this.webGLContexts.hoverNodes.COLOR_BUFFER_BIT);var d=this.getRenderParams();for(var p in this.nodeHoverPrograms){var g=this.nodeHoverPrograms[p];g.render(d)}}},{key:"scheduleHighlightedNodesRender",value:function(){var r=this;this.renderHighlightedNodesFrame||this.renderFrame||(this.renderHighlightedNodesFrame=requestAnimationFrame(function(){r.renderHighlightedNodesFrame=null,r.renderHighlightedNodes(),r.renderEdgeLabels()}))}},{key:"render",value:function(){var r=this;this.emit("beforeRender");var a=function(){return r.emit("afterRender"),r};if(this.renderFrame&&(cancelAnimationFrame(this.renderFrame),this.renderFrame=null),this.resize(),this.needToProcess&&this.process(),this.needToProcess=!1,this.clear(),this.pickingLayers.forEach(function(x){return r.resetWebGLTexture(x)}),!this.graph.order)return a();var o=this.mouseCaptor,s=this.camera.isAnimated()||o.isMoving||o.draggedEvents||o.currentWheelDirection,u=this.camera.getState(),c=this.getDimensions(),d=this.getGraphDimensions(),p=this.getStagePadding();this.matrix=au(u,c,d,p),this.invMatrix=au(u,c,d,p,!0),this.correctionRatio=ate(this.matrix,u,c),this.graphToViewportRatio=this.getGraphToViewportRatio();var g=this.getRenderParams();for(var m in this.nodePrograms){var b=this.nodePrograms[m];b.render(g)}if(!this.settings.hideEdgesOnMove||!s)for(var y in this.edgePrograms){var S=this.edgePrograms[y];S.render(g)}return this.settings.hideLabelsOnMove&&s||(this.renderLabels(),this.renderEdgeLabels(),this.renderHighlightedNodes()),a()}},{key:"addNode",value:function(r){var a=Object.assign({},this.graph.getNodeAttributes(r));this.settings.nodeReducer&&(a=this.settings.nodeReducer(r,a));var o=Ate(this.settings,r,a);this.nodeDataCache[r]=o,this.nodesWithForcedLabels.delete(r),o.forceLabel&&!o.hidden&&this.nodesWithForcedLabels.add(r),this.highlightedNodes.delete(r),o.highlighted&&!o.hidden&&this.highlightedNodes.add(r),this.settings.zIndex&&(o.zIndex<this.nodeZExtent[0]&&(this.nodeZExtent[0]=o.zIndex),o.zIndex>this.nodeZExtent[1]&&(this.nodeZExtent[1]=o.zIndex))}},{key:"updateNode",value:function(r){this.addNode(r);var a=this.nodeDataCache[r];this.normalizationFunction.applyTo(a)}},{key:"removeNode",value:function(r){delete this.nodeDataCache[r],delete this.nodeProgramIndex[r],this.highlightedNodes.delete(r),this.hoveredNode===r&&(this.hoveredNode=null),this.nodesWithForcedLabels.delete(r)}},{key:"addEdge",value:function(r){var a=Object.assign({},this.graph.getEdgeAttributes(r));this.settings.edgeReducer&&(a=this.settings.edgeReducer(r,a));var o=Rte(this.settings,r,a);this.edgeDataCache[r]=o,this.edgesWithForcedLabels.delete(r),o.forceLabel&&!o.hidden&&this.edgesWithForcedLabels.add(r),this.settings.zIndex&&(o.zIndex<this.edgeZExtent[0]&&(this.edgeZExtent[0]=o.zIndex),o.zIndex>this.edgeZExtent[1]&&(this.edgeZExtent[1]=o.zIndex))}},{key:"updateEdge",value:function(r){this.addEdge(r)}},{key:"removeEdge",value:function(r){delete this.edgeDataCache[r],delete this.edgeProgramIndex[r],this.hoveredEdge===r&&(this.hoveredEdge=null),this.edgesWithForcedLabels.delete(r)}},{key:"clearNodeIndices",value:function(){this.labelGrid=new NN,this.nodeExtent={x:[0,1],y:[0,1]},this.nodeDataCache={},this.edgeProgramIndex={},this.nodesWithForcedLabels=new Set,this.nodeZExtent=[1/0,-1/0]}},{key:"clearEdgeIndices",value:function(){this.edgeDataCache={},this.edgeProgramIndex={},this.edgesWithForcedLabels=new Set,this.edgeZExtent=[1/0,-1/0]}},{key:"clearIndices",value:function(){this.clearEdgeIndices(),this.clearNodeIndices()}},{key:"clearNodeState",value:function(){this.displayedNodeLabels=new Set,this.highlightedNodes=new Set,this.hoveredNode=null}},{key:"clearEdgeState",value:function(){this.displayedEdgeLabels=new Set,this.highlightedNodes=new Set,this.hoveredEdge=null}},{key:"clearState",value:function(){this.clearEdgeState(),this.clearNodeState()}},{key:"addNodeToProgram",value:function(r,a,o){var s=this.nodeDataCache[r],u=this.nodePrograms[s.type];if(!u)throw new Error('Sigma: could not find a suitable program for node type "'.concat(s.type,'"!'));u.process(a,o,s),this.nodeProgramIndex[r]=o}},{key:"addEdgeToProgram",value:function(r,a,o){var s=this.edgeDataCache[r],u=this.edgePrograms[s.type];if(!u)throw new Error('Sigma: could not find a suitable program for edge type "'.concat(s.type,'"!'));var c=this.graph.extremities(r),d=this.nodeDataCache[c[0]],p=this.nodeDataCache[c[1]];u.process(a,o,d,p,s),this.edgeProgramIndex[r]=o}},{key:"getRenderParams",value:function(){return{matrix:this.matrix,invMatrix:this.invMatrix,width:this.width,height:this.height,pixelRatio:this.pixelRatio,zoomRatio:this.camera.ratio,cameraAngle:this.camera.angle,sizeRatio:1/this.scaleSize(),correctionRatio:this.correctionRatio,downSizingRatio:this.pickingDownSizingRatio,minEdgeThickness:this.settings.minEdgeThickness,antiAliasingFeather:this.settings.antiAliasingFeather}}},{key:"getStagePadding",value:function(){var r=this.settings,a=r.stagePadding,o=r.autoRescale;return o&&a||0}},{key:"createLayer",value:function(r,a){var o=arguments.length>2&&arguments[2]!==void 0?arguments[2]:{};if(this.elements[r])throw new Error('Sigma: a layer named "'.concat(r,'" already exists'));var s=ste(a,{position:"absolute"},{class:"sigma-".concat(r)});return o.style&&Object.assign(s.style,o.style),this.elements[r]=s,"beforeLayer"in o&&o.beforeLayer?this.elements[o.beforeLayer].before(s):"afterLayer"in o&&o.afterLayer?this.elements[o.afterLayer].after(s):this.container.appendChild(s),s}},{key:"createCanvas",value:function(r){var a=arguments.length>1&&arguments[1]!==void 0?arguments[1]:{};return this.createLayer(r,"canvas",a)}},{key:"createCanvasContext",value:function(r){var a=arguments.length>1&&arguments[1]!==void 0?arguments[1]:{},o=this.createCanvas(r,a),s={preserveDrawingBuffer:!1,antialias:!1};return this.canvasContexts[r]=o.getContext("2d",s),this}},{key:"createWebGLContext",value:function(r){var a=arguments.length>1&&arguments[1]!==void 0?arguments[1]:{},o=(a==null?void 0:a.canvas)||this.createCanvas(r,a);a.hidden&&o.remove();var s=je({preserveDrawingBuffer:!1,antialias:!1},a),u;u=o.getContext("webgl2",s),u||(u=o.getContext("webgl",s)),u||(u=o.getContext("experimental-webgl",s));var c=u;if(this.webGLContexts[r]=c,c.blendFunc(c.ONE,c.ONE_MINUS_SRC_ALPHA),a.picking){this.pickingLayers.add(r);var d=c.createFramebuffer();if(!d)throw new Error("Sigma: cannot create a new frame buffer for layer ".concat(r));this.frameBuffers[r]=d}return c}},{key:"killLayer",value:function(r){var a=this.elements[r];if(!a)throw new Error("Sigma: cannot kill layer ".concat(r,", which does not exist"));if(this.webGLContexts[r]){var o,s=this.webGLContexts[r];(o=s.getExtension("WEBGL_lose_context"))===null||o===void 0||o.loseContext(),delete this.webGLContexts[r]}else this.canvasContexts[r]&&delete this.canvasContexts[r];return a.remove(),delete this.elements[r],this}},{key:"getCamera",value:function(){return this.camera}},{key:"setCamera",value:function(r){this.unbindCameraHandlers(),this.camera=r,this.bindCameraHandlers()}},{key:"getContainer",value:function(){return this.container}},{key:"getGraph",value:function(){return this.graph}},{key:"setGraph",value:function(r){r!==this.graph&&(this.hoveredNode&&!r.hasNode(this.hoveredNode)&&(this.hoveredNode=null),this.hoveredEdge&&!r.hasEdge(this.hoveredEdge)&&(this.hoveredEdge=null),this.unbindGraphHandlers(),this.checkEdgesEventsFrame!==null&&(cancelAnimationFrame(this.checkEdgesEventsFrame),this.checkEdgesEventsFrame=null),this.graph=r,this.bindGraphHandlers(),this.refresh())}},{key:"getMouseCaptor",value:function(){return this.mouseCaptor}},{key:"getTouchCaptor",value:function(){return this.touchCaptor}},{key:"getDimensions",value:function(){return{width:this.width,height:this.height}}},{key:"getGraphDimensions",value:function(){var r=this.customBBox||this.nodeExtent;return{width:r.x[1]-r.x[0]||1,height:r.y[1]-r.y[0]||1}}},{key:"getNodeDisplayData",value:function(r){var a=this.nodeDataCache[r];return a?Object.assign({},a):void 0}},{key:"getEdgeDisplayData",value:function(r){var a=this.edgeDataCache[r];return a?Object.assign({},a):void 0}},{key:"getNodeDisplayedLabels",value:function(){return new Set(this.displayedNodeLabels)}},{key:"getEdgeDisplayedLabels",value:function(){return new Set(this.displayedEdgeLabels)}},{key:"getSettings",value:function(){return je({},this.settings)}},{key:"getSetting",value:function(r){return this.settings[r]}},{key:"setSetting",value:function(r,a){var o=je({},this.settings);return this.settings[r]=a,Em(this.settings),this.handleSettingsUpdate(o),this.scheduleRefresh(),this}},{key:"updateSetting",value:function(r,a){return this.setSetting(r,a(this.settings[r])),this}},{key:"setSettings",value:function(r){var a=je({},this.settings);return this.settings=je(je({},this.settings),r),Em(this.settings),this.handleSettingsUpdate(a),this.scheduleRefresh(),this}},{key:"resize",value:function(r){var a=this.width,o=this.height;if(this.width=this.container.offsetWidth,this.height=this.container.offsetHeight,this.pixelRatio=xN(),this.width===0)if(this.settings.allowInvalidContainer)this.width=1;else throw new Error("Sigma: Container has no width. You can set the allowInvalidContainer setting to true to stop seeing this error.");if(this.height===0)if(this.settings.allowInvalidContainer)this.height=1;else throw new Error("Sigma: Container has no height. You can set the allowInvalidContainer setting to true to stop seeing this error.");if(!r&&a===this.width&&o===this.height)return this;for(var s in this.elements){var u=this.elements[s];u.style.width=this.width+"px",u.style.height=this.height+"px"}for(var c in this.canvasContexts)this.elements[c].setAttribute("width",this.width*this.pixelRatio+"px"),this.elements[c].setAttribute("height",this.height*this.pixelRatio+"px"),this.pixelRatio!==1&&this.canvasContexts[c].scale(this.pixelRatio,this.pixelRatio);for(var d in this.webGLContexts){this.elements[d].setAttribute("width",this.width*this.pixelRatio+"px"),this.elements[d].setAttribute("height",this.height*this.pixelRatio+"px");var p=this.webGLContexts[d];if(p.viewport(0,0,this.width*this.pixelRatio,this.height*this.pixelRatio),this.pickingLayers.has(d)){var g=this.textures[d];g&&p.deleteTexture(g)}}return this.emit("resize"),this}},{key:"clear",value:function(){return this.emit("beforeClear"),this.webGLContexts.nodes.bindFramebuffer(WebGLRenderingContext.FRAMEBUFFER,null),this.webGLContexts.nodes.clear(WebGLRenderingContext.COLOR_BUFFER_BIT),this.webGLContexts.edges.bindFramebuffer(WebGLRenderingContext.FRAMEBUFFER,null),this.webGLContexts.edges.clear(WebGLRenderingContext.COLOR_BUFFER_BIT),this.webGLContexts.hoverNodes.clear(WebGLRenderingContext.COLOR_BUFFER_BIT),this.canvasContexts.labels.clearRect(0,0,this.width,this.height),this.canvasContexts.hovers.clearRect(0,0,this.width,this.height),this.canvasContexts.edgeLabels.clearRect(0,0,this.width,this.height),this.emit("afterClear"),this}},{key:"refresh",value:function(r){var a=this,o=(r==null?void 0:r.skipIndexation)!==void 0?r==null?void 0:r.skipIndexation:!1,s=(r==null?void 0:r.schedule)!==void 0?r.schedule:!1,u=!r||!r.partialGraph;if(u)this.clearEdgeIndices(),this.clearNodeIndices(),this.graph.forEachNode(function(C){return a.addNode(C)}),this.graph.forEachEdge(function(C){return a.addEdge(C)});else{for(var c,d,p=((c=r.partialGraph)===null||c===void 0?void 0:c.nodes)||[],g=0,m=(p==null?void 0:p.length)||0;g<m;g++){var b=p[g];if(this.updateNode(b),o){var y=this.nodeProgramIndex[b];if(y===void 0)throw new Error('Sigma: node "'.concat(b,`" can't be repaint`));this.addNodeToProgram(b,this.nodeIndices[b],y)}}for(var S=(r==null||(d=r.partialGraph)===null||d===void 0?void 0:d.edges)||[],x=0,R=S.length;x<R;x++){var k=S[x];if(this.updateEdge(k),o){var A=this.edgeProgramIndex[k];if(A===void 0)throw new Error('Sigma: edge "'.concat(k,`" can't be repaint`));this.addEdgeToProgram(k,this.edgeIndices[k],A)}}}return(u||!o)&&(this.needToProcess=!0),s?this.scheduleRender():this.render(),this}},{key:"scheduleRender",value:function(){var r=this;return this.renderFrame||(this.renderFrame=requestAnimationFrame(function(){r.render()})),this}},{key:"scheduleRefresh",value:function(r){return this.refresh(je(je({},r),{},{schedule:!0}))}},{key:"getViewportZoomedState",value:function(r,a){var o=this.camera.getState(),s=o.ratio,u=o.angle,c=o.x,d=o.y,p=this.settings,g=p.minCameraRatio,m=p.maxCameraRatio;typeof m=="number"&&(a=Math.min(a,m)),typeof g=="number"&&(a=Math.max(a,g));var b=a/s,y={x:this.width/2,y:this.height/2},S=this.viewportToFramedGraph(r),x=this.viewportToFramedGraph(y);return{angle:u,x:(S.x-x.x)*(1-b)+c,y:(S.y-x.y)*(1-b)+d,ratio:a}}},{key:"viewRectangle",value:function(){var r=this.viewportToFramedGraph({x:0,y:0}),a=this.viewportToFramedGraph({x:this.width,y:0}),o=this.viewportToFramedGraph({x:0,y:this.height});return{x1:r.x,y1:r.y,x2:a.x,y2:a.y,height:a.y-o.y}}},{key:"framedGraphToViewport",value:function(r){var a=arguments.length>1&&arguments[1]!==void 0?arguments[1]:{},o=!!a.cameraState||!!a.viewportDimensions||!!a.graphDimensions,s=a.matrix?a.matrix:o?au(a.cameraState||this.camera.getState(),a.viewportDimensions||this.getDimensions(),a.graphDimensions||this.getGraphDimensions(),a.padding||this.getStagePadding()):this.matrix,u=Bk(s,r);return{x:(1+u.x)*this.width/2,y:(1-u.y)*this.height/2}}},{key:"viewportToFramedGraph",value:function(r){var a=arguments.length>1&&arguments[1]!==void 0?arguments[1]:{},o=!!a.cameraState||!!a.viewportDimensions||!a.graphDimensions,s=a.matrix?a.matrix:o?au(a.cameraState||this.camera.getState(),a.viewportDimensions||this.getDimensions(),a.graphDimensions||this.getGraphDimensions(),a.padding||this.getStagePadding(),!0):this.invMatrix,u=Bk(s,{x:r.x/this.width*2-1,y:1-r.y/this.height*2});return isNaN(u.x)&&(u.x=0),isNaN(u.y)&&(u.y=0),u}},{key:"viewportToGraph",value:function(r){var a=arguments.length>1&&arguments[1]!==void 0?arguments[1]:{};return this.normalizationFunction.inverse(this.viewportToFramedGraph(r,a))}},{key:"graphToViewport",value:function(r){var a=arguments.length>1&&arguments[1]!==void 0?arguments[1]:{};return this.framedGraphToViewport(this.normalizationFunction(r),a)}},{key:"getGraphToViewportRatio",value:function(){var r={x:0,y:0},a={x:1,y:1},o=Math.sqrt(Math.pow(r.x-a.x,2)+Math.pow(r.y-a.y,2)),s=this.graphToViewport(r),u=this.graphToViewport(a),c=Math.sqrt(Math.pow(s.x-u.x,2)+Math.pow(s.y-u.y,2));return c/o}},{key:"getBBox",value:function(){return this.nodeExtent}},{key:"getCustomBBox",value:function(){return this.customBBox}},{key:"setCustomBBox",value:function(r){return this.customBBox=r,this.scheduleRender(),this}},{key:"kill",value:function(){this.emit("kill"),this.removeAllListeners(),this.unbindCameraHandlers(),window.removeEventListener("resize",this.activeListeners.handleResize),this.mouseCaptor.kill(),this.touchCaptor.kill(),this.unbindGraphHandlers(),this.clearIndices(),this.clearState(),this.nodeDataCache={},this.edgeDataCache={},this.highlightedNodes.clear(),this.renderFrame&&(cancelAnimationFrame(this.renderFrame),this.renderFrame=null),this.renderHighlightedNodesFrame&&(cancelAnimationFrame(this.renderHighlightedNodesFrame),this.renderHighlightedNodesFrame=null);for(var r=this.container;r.firstChild;)r.removeChild(r.firstChild);this.canvasContexts={},this.webGLContexts={},this.elements={};for(var a in this.nodePrograms)this.nodePrograms[a].kill();for(var o in this.nodeHoverPrograms)this.nodeHoverPrograms[o].kill();for(var s in this.edgePrograms)this.edgePrograms[s].kill();this.nodePrograms={},this.nodeHoverPrograms={},this.edgePrograms={};for(var u in this.elements)this.killLayer(u)}},{key:"scaleSize",value:function(){var r=arguments.length>0&&arguments[0]!==void 0?arguments[0]:1,a=arguments.length>1&&arguments[1]!==void 0?arguments[1]:this.camera.ratio;return r/this.settings.zoomToSizeRatioFunction(a)*(this.getSetting("itemSizesReference")==="positions"?a*this.graphToViewportRatio:1)}},{key:"getCanvases",value:function(){var r={};for(var a in this.elements)this.elements[a]instanceof HTMLCanvasElement&&(r[a]=this.elements[a]);return r}}])}(bT);const B4=E.createContext(null),Cte=B4.Provider;function vT(){const e=E.useContext(B4);if(e==null)throw new Error("No context provided: useSigmaContext() can only be used in a descendant of <SigmaContainer>");return e}function kr(){return vT().sigma}function U4(){const{sigma:e}=vT();return E.useCallback(t=>{e&&Object.keys(t).forEach(n=>{e.setSetting(n,t[n])})},[e])}function Ep(e){return new Set(Object.keys(e))}const ON=Ep({clickNode:!0,rightClickNode:!0,downNode:!0,enterNode:!0,leaveNode:!0,doubleClickNode:!0,wheelNode:!0,clickEdge:!0,rightClickEdge:!0,downEdge:!0,enterEdge:!0,leaveEdge:!0,doubleClickEdge:!0,wheelEdge:!0,clickStage:!0,rightClickStage:!0,downStage:!0,doubleClickStage:!0,wheelStage:!0,beforeRender:!0,afterRender:!0,kill:!0,upStage:!0,upEdge:!0,upNode:!0,enterStage:!0,leaveStage:!0,resize:!0,afterClear:!0,afterProcess:!0,beforeClear:!0,beforeProcess:!0,moveBody:!0}),DN=Ep({click:!0,rightClick:!0,doubleClick:!0,mouseup:!0,mousedown:!0,mousemove:!0,mousemovebody:!0,mouseleave:!0,mouseenter:!0,wheel:!0}),IN=Ep({touchup:!0,touchdown:!0,touchmove:!0,touchmovebody:!0,tap:!0,doubletap:!0}),LN=Ep({updated:!0});function j4(){const e=kr(),t=U4(),[n,r]=E.useState({});return E.useEffect(()=>{if(!e||!n)return;const a=n,o=Object.keys(a);return o.forEach(s=>{const u=a[s];ON.has(s)&&e.on(s,u),DN.has(s)&&e.getMouseCaptor().on(s,u),IN.has(s)&&e.getTouchCaptor().on(s,u),LN.has(s)&&e.getCamera().on(s,u)}),()=>{e&&o.forEach(s=>{const u=a[s];ON.has(s)&&e.off(s,u),DN.has(s)&&e.getMouseCaptor().off(s,u),IN.has(s)&&e.getTouchCaptor().off(s,u),LN.has(s)&&e.getCamera().off(s,u)})}},[e,n,t]),r}function ec(e,t){if(e===t)return!0;if(typeof e=="object"&&e!=null&&typeof t=="object"&&t!=null){if(Object.keys(e).length!=Object.keys(t).length)return!1;for(const n in e)if(!Object.hasOwn(t,n)||!ec(e[n],t[n]))return!1;return!0}return!1}function G4(e){const t=kr(),[n,r]=E.useState(e||{});E.useEffect(()=>{r(d=>ec(d,e||{})?d:e||{})},[e]);const a=E.useCallback(d=>{t.getCamera().animatedZoom(Object.assign(Object.assign({},n),d))},[t,n]),o=E.useCallback(d=>{t.getCamera().animatedUnzoom(Object.assign(Object.assign({},n),d))},[t,n]),s=E.useCallback(d=>{t.getCamera().animatedReset(Object.assign(Object.assign({},n),d))},[t,n]),u=E.useCallback((d,p)=>{t.getCamera().animate(d,Object.assign(Object.assign({},n),p))},[t,n]),c=E.useCallback((d,p)=>{const g=t.getNodeDisplayData(d);g?t.getCamera().animate(g,Object.assign(Object.assign({},n),p)):console.warn(`Node ${d} not found`)},[t,n]);return{zoomIn:a,zoomOut:o,reset:s,goto:u,gotoNode:c}}function Nte(e){const t=vT(),[n,r]=E.useState(!1),[a,o]=E.useState(t.container),s=E.useCallback(()=>r(u=>!u),[]);return E.useEffect(()=>(document.addEventListener("fullscreenchange",s),()=>document.removeEventListener("fullscreenchange",s)),[s]),E.useEffect(()=>{o(t.container)},[e,t.container]),{toggle:E.useCallback(()=>{var u;u=a,document.fullscreenElement!==u?u.requestFullscreen():document.exitFullscreen&&document.exitFullscreen()},[a]),isFullScreen:n}}const Ote=E.forwardRef(({graph:e,id:t,className:n,style:r,settings:a={},children:o},s)=>{const u=E.useRef(null),c=E.useRef(null),d={className:`react-sigma ${n||""}`,id:t,style:r},[p,g]=E.useState(null),[m,b]=E.useState(a);E.useEffect(()=>{b(x=>ec(x,a)?x:a)},[a]),E.useEffect(()=>{g(x=>{let R=null;if(c.current!==null){let k=new _t;e&&(k=typeof e=="function"?new e:e);let A=null;x&&(A=x.getCamera().getState(),x.kill()),R=new _te(k,c.current,m),A&&R.getCamera().setState(A)}return R})},[c,e,m]),E.useImperativeHandle(s,()=>p,[p]);const y=E.useMemo(()=>p&&u.current?{sigma:p,container:u.current}:null,[p,u]),S=y!==null?ye.createElement(Cte,{value:y},o):null;return ye.createElement("div",Object.assign({},d,{ref:u}),ye.createElement("div",{className:"sigma-container",ref:c}),S)});var Dte=`
precision mediump float;

varying vec4 v_color;
varying float v_border;

const float radius = 0.5;
const vec4 transparent = vec4(0.0, 0.0, 0.0, 0.0);

void main(void) {
  vec2 m = gl_PointCoord - vec2(0.5, 0.5);
  float dist = radius - length(m);

  // No antialiasing for picking mode:
  #ifdef PICKING_MODE
  if (dist > v_border)
    gl_FragColor = v_color;
  else
    gl_FragColor = transparent;

  #else
  float t = 0.0;
  if (dist > v_border)
    t = 1.0;
  else if (dist > 0.0)
    t = dist / v_border;

  gl_FragColor = mix(transparent, v_color, t);
  #endif
}
`,Ite=Dte,Lte=`
attribute vec4 a_id;
attribute vec4 a_color;
attribute vec2 a_position;
attribute float a_size;

uniform float u_sizeRatio;
uniform float u_pixelRatio;
uniform mat3 u_matrix;

varying vec4 v_color;
varying float v_border;

const float bias = 255.0 / 254.0;

void main() {
  gl_Position = vec4(
    (u_matrix * vec3(a_position, 1)).xy,
    0,
    1
  );

  // Multiply the point size twice:
  //  - x SCALING_RATIO to correct the canvas scaling
  //  - x 2 to correct the formulae
  gl_PointSize = a_size / u_sizeRatio * u_pixelRatio * 2.0;

  v_border = (0.5 / a_size) * u_sizeRatio;

  #ifdef PICKING_MODE
  // For picking mode, we use the ID as the color:
  v_color = a_id;
  #else
  // For normal mode, we use the color:
  v_color = a_color;
  #endif

  v_color.a *= bias;
}
`,Mte=Lte,H4=WebGLRenderingContext,MN=H4.UNSIGNED_BYTE,FN=H4.FLOAT,Fte=["u_sizeRatio","u_pixelRatio","u_matrix"],Pte=function(e){function t(){return cn(this,t),$n(this,t,arguments)}return qn(t,e),dn(t,[{key:"getDefinition",value:function(){return{VERTICES:1,VERTEX_SHADER_SOURCE:Mte,FRAGMENT_SHADER_SOURCE:Ite,METHOD:WebGLRenderingContext.POINTS,UNIFORMS:Fte,ATTRIBUTES:[{name:"a_position",size:2,type:FN},{name:"a_size",size:1,type:FN},{name:"a_color",size:4,type:MN,normalized:!0},{name:"a_id",size:4,type:MN,normalized:!0}]}}},{key:"processVisibleItem",value:function(r,a,o){var s=this.array;s[a++]=o.x,s[a++]=o.y,s[a++]=o.size,s[a++]=la(o.color),s[a++]=r}},{key:"setUniforms",value:function(r,a){var o=r.sizeRatio,s=r.pixelRatio,u=r.matrix,c=a.gl,d=a.uniformLocations,p=d.u_sizeRatio,g=d.u_pixelRatio,m=d.u_matrix;c.uniform1f(g,s),c.uniform1f(p,o),c.uniformMatrix3fv(m,!1,u)}}])}(hT),zte=`
attribute vec4 a_id;
attribute vec4 a_color;
attribute vec2 a_normal;
attribute float a_normalCoef;
attribute vec2 a_positionStart;
attribute vec2 a_positionEnd;
attribute float a_positionCoef;
attribute float a_sourceRadius;
attribute float a_targetRadius;
attribute float a_sourceRadiusCoef;
attribute float a_targetRadiusCoef;

uniform mat3 u_matrix;
uniform float u_zoomRatio;
uniform float u_sizeRatio;
uniform float u_pixelRatio;
uniform float u_correctionRatio;
uniform float u_minEdgeThickness;
uniform float u_lengthToThicknessRatio;
uniform float u_feather;

varying vec4 v_color;
varying vec2 v_normal;
varying float v_thickness;
varying float v_feather;

const float bias = 255.0 / 254.0;

void main() {
  float minThickness = u_minEdgeThickness;

  vec2 normal = a_normal * a_normalCoef;
  vec2 position = a_positionStart * (1.0 - a_positionCoef) + a_positionEnd * a_positionCoef;

  float normalLength = length(normal);
  vec2 unitNormal = normal / normalLength;

  // These first computations are taken from edge.vert.glsl. Please read it to
  // get better comments on what's happening:
  float pixelsThickness = max(normalLength, minThickness * u_sizeRatio);
  float webGLThickness = pixelsThickness * u_correctionRatio / u_sizeRatio;

  // Here, we move the point to leave space for the arrow heads:
  // Source arrow head
  float sourceRadius = a_sourceRadius * a_sourceRadiusCoef;
  float sourceDirection = sign(sourceRadius);
  float webGLSourceRadius = sourceDirection * sourceRadius * 2.0 * u_correctionRatio / u_sizeRatio;
  float webGLSourceArrowHeadLength = webGLThickness * u_lengthToThicknessRatio * 2.0;
  vec2 sourceCompensationVector =
    vec2(-sourceDirection * unitNormal.y, sourceDirection * unitNormal.x)
    * (webGLSourceRadius + webGLSourceArrowHeadLength);
    
  // Target arrow head
  float targetRadius = a_targetRadius * a_targetRadiusCoef;
  float targetDirection = sign(targetRadius);
  float webGLTargetRadius = targetDirection * targetRadius * 2.0 * u_correctionRatio / u_sizeRatio;
  float webGLTargetArrowHeadLength = webGLThickness * u_lengthToThicknessRatio * 2.0;
  vec2 targetCompensationVector =
  vec2(-targetDirection * unitNormal.y, targetDirection * unitNormal.x)
    * (webGLTargetRadius + webGLTargetArrowHeadLength);

  // Here is the proper position of the vertex
  gl_Position = vec4((u_matrix * vec3(position + unitNormal * webGLThickness + sourceCompensationVector + targetCompensationVector, 1)).xy, 0, 1);

  v_thickness = webGLThickness / u_zoomRatio;

  v_normal = unitNormal;

  v_feather = u_feather * u_correctionRatio / u_zoomRatio / u_pixelRatio * 2.0;

  #ifdef PICKING_MODE
  // For picking mode, we use the ID as the color:
  v_color = a_id;
  #else
  // For normal mode, we use the color:
  v_color = a_color;
  #endif

  v_color.a *= bias;
}
`,Bte=zte,$4=WebGLRenderingContext,PN=$4.UNSIGNED_BYTE,Oa=$4.FLOAT,Ute=["u_matrix","u_zoomRatio","u_sizeRatio","u_correctionRatio","u_pixelRatio","u_feather","u_minEdgeThickness","u_lengthToThicknessRatio"],jte={lengthToThicknessRatio:Ju.lengthToThicknessRatio};function q4(e){var t=je(je({},jte),{});return function(n){function r(){return cn(this,r),$n(this,r,arguments)}return qn(r,n),dn(r,[{key:"getDefinition",value:function(){return{VERTICES:6,VERTEX_SHADER_SOURCE:Bte,FRAGMENT_SHADER_SOURCE:mT,METHOD:WebGLRenderingContext.TRIANGLES,UNIFORMS:Ute,ATTRIBUTES:[{name:"a_positionStart",size:2,type:Oa},{name:"a_positionEnd",size:2,type:Oa},{name:"a_normal",size:2,type:Oa},{name:"a_color",size:4,type:PN,normalized:!0},{name:"a_id",size:4,type:PN,normalized:!0},{name:"a_sourceRadius",size:1,type:Oa},{name:"a_targetRadius",size:1,type:Oa}],CONSTANT_ATTRIBUTES:[{name:"a_positionCoef",size:1,type:Oa},{name:"a_normalCoef",size:1,type:Oa},{name:"a_sourceRadiusCoef",size:1,type:Oa},{name:"a_targetRadiusCoef",size:1,type:Oa}],CONSTANT_DATA:[[0,1,-1,0],[0,-1,1,0],[1,1,0,1],[1,1,0,1],[0,-1,1,0],[1,-1,0,-1]]}}},{key:"processVisibleItem",value:function(o,s,u,c,d){var p=d.size||1,g=u.x,m=u.y,b=c.x,y=c.y,S=la(d.color),x=b-g,R=y-m,k=u.size||1,A=c.size||1,C=x*x+R*R,N=0,_=0;C&&(C=1/Math.sqrt(C),N=-R*C*p,_=x*C*p);var O=this.array;O[s++]=g,O[s++]=m,O[s++]=b,O[s++]=y,O[s++]=N,O[s++]=_,O[s++]=S,O[s++]=o,O[s++]=k,O[s++]=A}},{key:"setUniforms",value:function(o,s){var u=s.gl,c=s.uniformLocations,d=c.u_matrix,p=c.u_zoomRatio,g=c.u_feather,m=c.u_pixelRatio,b=c.u_correctionRatio,y=c.u_sizeRatio,S=c.u_minEdgeThickness,x=c.u_lengthToThicknessRatio;u.uniformMatrix3fv(d,!1,o.matrix),u.uniform1f(p,o.zoomRatio),u.uniform1f(y,o.sizeRatio),u.uniform1f(b,o.correctionRatio),u.uniform1f(m,o.pixelRatio),u.uniform1f(g,o.antiAliasingFeather),u.uniform1f(S,o.minEdgeThickness),u.uniform1f(x,t.lengthToThicknessRatio)}}])}(Zu)}q4();function Gte(e){return A4([q4(),kf(e),kf(je(je({},e),{},{extremity:"source"}))])}Gte();function Hte(e){if(Array.isArray(e))return e}function $te(e,t){var n=e==null?null:typeof Symbol<"u"&&e[Symbol.iterator]||e["@@iterator"];if(n!=null){var r,a,o,s,u=[],c=!0,d=!1;try{if(o=(n=n.call(e)).next,t!==0)for(;!(c=(r=o.call(n)).done)&&(u.push(r.value),u.length!==t);c=!0);}catch(p){d=!0,a=p}finally{try{if(!c&&n.return!=null&&(s=n.return(),Object(s)!==s))return}finally{if(d)throw a}}return u}}function jk(e,t){(t==null||t>e.length)&&(t=e.length);for(var n=0,r=Array(t);n<t;n++)r[n]=e[n];return r}function V4(e,t){if(e){if(typeof e=="string")return jk(e,t);var n={}.toString.call(e).slice(8,-1);return n==="Object"&&e.constructor&&(n=e.constructor.name),n==="Map"||n==="Set"?Array.from(e):n==="Arguments"||/^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n)?jk(e,t):void 0}}function qte(){throw new TypeError(`Invalid attempt to destructure non-iterable instance.
In order to be iterable, non-array objects must have a [Symbol.iterator]() method.`)}function Vte(e,t){return Hte(e)||$te(e,t)||V4(e,t)||qte()}function Wte(e,t){if(!(e instanceof t))throw new TypeError("Cannot call a class as a function")}function Yte(e,t){if(typeof e!="object"||!e)return e;var n=e[Symbol.toPrimitive];if(n!==void 0){var r=n.call(e,t);if(typeof r!="object")return r;throw new TypeError("@@toPrimitive must return a primitive value.")}return(t==="string"?String:Number)(e)}function W4(e){var t=Yte(e,"string");return typeof t=="symbol"?t:t+""}function Kte(e,t){for(var n=0;n<t.length;n++){var r=t[n];r.enumerable=r.enumerable||!1,r.configurable=!0,"value"in r&&(r.writable=!0),Object.defineProperty(e,W4(r.key),r)}}function Xte(e,t,n){return t&&Kte(e.prototype,t),Object.defineProperty(e,"prototype",{writable:!1}),e}function Tf(e){return Tf=Object.setPrototypeOf?Object.getPrototypeOf.bind():function(t){return t.__proto__||Object.getPrototypeOf(t)},Tf(e)}function Y4(){try{var e=!Boolean.prototype.valueOf.call(Reflect.construct(Boolean,[],function(){}))}catch{}return(Y4=function(){return!!e})()}function Gk(e){if(e===void 0)throw new ReferenceError("this hasn't been initialised - super() hasn't been called");return e}function Zte(e,t){if(t&&(typeof t=="object"||typeof t=="function"))return t;if(t!==void 0)throw new TypeError("Derived constructors may only return object or undefined");return Gk(e)}function Qte(e,t,n){return t=Tf(t),Zte(e,Y4()?Reflect.construct(t,n||[],Tf(e).constructor):t.apply(e,n))}function Hk(e,t){return Hk=Object.setPrototypeOf?Object.setPrototypeOf.bind():function(n,r){return n.__proto__=r,n},Hk(e,t)}function Jte(e,t){if(typeof t!="function"&&t!==null)throw new TypeError("Super expression must either be null or a function");e.prototype=Object.create(t&&t.prototype,{constructor:{value:e,writable:!0,configurable:!0}}),Object.defineProperty(e,"prototype",{writable:!1}),t&&Hk(e,t)}function ys(e,t,n){return(t=W4(t))in e?Object.defineProperty(e,t,{value:n,enumerable:!0,configurable:!0,writable:!0}):e[t]=n,e}function ene(e){if(Array.isArray(e))return jk(e)}function tne(e){if(typeof Symbol<"u"&&e[Symbol.iterator]!=null||e["@@iterator"]!=null)return Array.from(e)}function nne(){throw new TypeError(`Invalid attempt to spread non-iterable instance.
In order to be iterable, non-array objects must have a [Symbol.iterator]() method.`)}function xm(e){return ene(e)||tne(e)||V4(e)||nne()}function zN(e,t){var n=Object.keys(e);if(Object.getOwnPropertySymbols){var r=Object.getOwnPropertySymbols(e);t&&(r=r.filter(function(a){return Object.getOwnPropertyDescriptor(e,a).enumerable})),n.push.apply(n,r)}return n}function BN(e){for(var t=1;t<arguments.length;t++){var n=arguments[t]!=null?arguments[t]:{};t%2?zN(Object(n),!0).forEach(function(r){ys(e,r,n[r])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(n)):zN(Object(n)).forEach(function(r){Object.defineProperty(e,r,Object.getOwnPropertyDescriptor(n,r))})}return e}var rne="relative",ane={drawLabel:void 0,drawHover:void 0,borders:[{size:{value:.1},color:{attribute:"borderColor"}},{size:{fill:!0},color:{attribute:"color"}}]},one="#000000";function ine(e){var t=e.borders,n=gN(t.filter(function(a){var o=a.size;return"fill"in o}).length),r=`
precision highp float;

varying vec2 v_diffVector;
varying float v_radius;

#ifdef PICKING_MODE
varying vec4 v_color;
#else
// For normal mode, we use the border colors defined in the program:
`.concat(t.flatMap(function(a,o){var s=a.size;return"attribute"in s?["varying float v_borderSize_".concat(o+1,";")]:[]}).join(`
`),`
`).concat(t.flatMap(function(a,o){var s=a.color;return"attribute"in s?["varying vec4 v_borderColor_".concat(o+1,";")]:"value"in s?["uniform vec4 u_borderColor_".concat(o+1,";")]:[]}).join(`
`),`
#endif

uniform float u_correctionRatio;

const float bias = 255.0 / 254.0;
const vec4 transparent = vec4(0.0, 0.0, 0.0, 0.0);

void main(void) {
  float dist = length(v_diffVector);
  float aaBorder = 2.0 * u_correctionRatio;
  float v_borderSize_0 = v_radius;
  vec4 v_borderColor_0 = transparent;

  // No antialiasing for picking mode:
  #ifdef PICKING_MODE
  if (dist > v_radius)
    gl_FragColor = transparent;
  else {
    gl_FragColor = v_color;
    gl_FragColor.a *= bias;
  }
  #else
  // Sizes:
`).concat(t.flatMap(function(a,o){var s=a.size;if("fill"in s)return[];s=s;var u="attribute"in s?"v_borderSize_".concat(o+1):gN(s.value),c=(s.mode||rne)==="pixels"?"u_correctionRatio":"v_radius";return["  float borderSize_".concat(o+1," = ").concat(c," * ").concat(u,";")]}).join(`
`),`
  // Now, let's split the remaining space between "fill" borders:
  float fillBorderSize = (v_radius - (`).concat(t.flatMap(function(a,o){var s=a.size;return"fill"in s?[]:["borderSize_".concat(o+1)]}).join(" + "),") ) / ").concat(n,`;
`).concat(t.flatMap(function(a,o){var s=a.size;return"fill"in s?["  float borderSize_".concat(o+1," = fillBorderSize;")]:[]}).join(`
`),`

  // Finally, normalize all border sizes, to start from the full size and to end with the smallest:
  float adjustedBorderSize_0 = v_radius;
`).concat(t.map(function(a,o){return"  float adjustedBorderSize_".concat(o+1," = adjustedBorderSize_").concat(o," - borderSize_").concat(o+1,";")}).join(`
`),`

  // Colors:
  vec4 borderColor_0 = transparent;
`).concat(t.map(function(a,o){var s=a.color,u=[];return"attribute"in s?u.push("  vec4 borderColor_".concat(o+1," = v_borderColor_").concat(o+1,";")):"transparent"in s?u.push("  vec4 borderColor_".concat(o+1," = vec4(0.0, 0.0, 0.0, 0.0);")):u.push("  vec4 borderColor_".concat(o+1," = u_borderColor_").concat(o+1,";")),u.push("  borderColor_".concat(o+1,".a *= bias;")),u.push("  if (borderSize_".concat(o+1," <= 1.0 * u_correctionRatio) { borderColor_").concat(o+1," = borderColor_").concat(o,"; }")),u.join(`
`)}).join(`
`),`
  if (dist > adjustedBorderSize_0) {
    gl_FragColor = borderColor_0;
  } else `).concat(t.map(function(a,o){return"if (dist > adjustedBorderSize_".concat(o,` - aaBorder) {
    gl_FragColor = mix(borderColor_`).concat(o+1,", borderColor_").concat(o,", (dist - adjustedBorderSize_").concat(o,` + aaBorder) / aaBorder);
  } else if (dist > adjustedBorderSize_`).concat(o+1,`) {
    gl_FragColor = borderColor_`).concat(o+1,`;
  } else `)}).join(""),` { /* Nothing to add here */ }
  #endif
}
`);return r}function sne(e){var t=e.borders,n=`
attribute vec2 a_position;
attribute float a_size;
attribute float a_angle;

uniform mat3 u_matrix;
uniform float u_sizeRatio;
uniform float u_correctionRatio;

varying vec2 v_diffVector;
varying float v_radius;

#ifdef PICKING_MODE
attribute vec4 a_id;
varying vec4 v_color;
#else
`.concat(t.flatMap(function(r,a){var o=r.size;return"attribute"in o?["attribute float a_borderSize_".concat(a+1,";"),"varying float v_borderSize_".concat(a+1,";")]:[]}).join(`
`),`
`).concat(t.flatMap(function(r,a){var o=r.color;return"attribute"in o?["attribute vec4 a_borderColor_".concat(a+1,";"),"varying vec4 v_borderColor_".concat(a+1,";")]:[]}).join(`
`),`
#endif

const float bias = 255.0 / 254.0;
const vec4 transparent = vec4(0.0, 0.0, 0.0, 0.0);

void main() {
  float size = a_size * u_correctionRatio / u_sizeRatio * 4.0;
  vec2 diffVector = size * vec2(cos(a_angle), sin(a_angle));
  vec2 position = a_position + diffVector;
  gl_Position = vec4(
    (u_matrix * vec3(position, 1)).xy,
    0,
    1
  );

  v_radius = size / 2.0;
  v_diffVector = diffVector;

  #ifdef PICKING_MODE
  v_color = a_id;
  #else
`).concat(t.flatMap(function(r,a){var o=r.size;return"attribute"in o?["  v_borderSize_".concat(a+1," = a_borderSize_").concat(a+1,";")]:[]}).join(`
`),`
`).concat(t.flatMap(function(r,a){var o=r.color;return"attribute"in o?["  v_borderColor_".concat(a+1," = a_borderColor_").concat(a+1,";")]:[]}).join(`
`),`
  #endif
}
`);return n}var K4=WebGLRenderingContext,UN=K4.UNSIGNED_BYTE,Od=K4.FLOAT;function lne(e){var t,n=BN(BN({},ane),{}),r=n.borders,a=n.drawLabel,o=n.drawHover,s=["u_sizeRatio","u_correctionRatio","u_matrix"].concat(xm(r.flatMap(function(u,c){var d=u.color;return"value"in d?["u_borderColor_".concat(c+1)]:[]})));return t=function(u){Jte(c,u);function c(){var d;Wte(this,c);for(var p=arguments.length,g=new Array(p),m=0;m<p;m++)g[m]=arguments[m];return d=Qte(this,c,[].concat(g)),ys(Gk(d),"drawLabel",a),ys(Gk(d),"drawHover",o),d}return Xte(c,[{key:"getDefinition",value:function(){return{VERTICES:3,VERTEX_SHADER_SOURCE:sne(n),FRAGMENT_SHADER_SOURCE:ine(n),METHOD:WebGLRenderingContext.TRIANGLES,UNIFORMS:s,ATTRIBUTES:[{name:"a_position",size:2,type:Od},{name:"a_id",size:4,type:UN,normalized:!0},{name:"a_size",size:1,type:Od}].concat(xm(r.flatMap(function(p,g){var m=p.color;return"attribute"in m?[{name:"a_borderColor_".concat(g+1),size:4,type:UN,normalized:!0}]:[]})),xm(r.flatMap(function(p,g){var m=p.size;return"attribute"in m?[{name:"a_borderSize_".concat(g+1),size:1,type:Od}]:[]}))),CONSTANT_ATTRIBUTES:[{name:"a_angle",size:1,type:Od}],CONSTANT_DATA:[[c.ANGLE_1],[c.ANGLE_2],[c.ANGLE_3]]}}},{key:"processVisibleItem",value:function(p,g,m){var b=this.array;b[g++]=m.x,b[g++]=m.y,b[g++]=p,b[g++]=m.size,r.forEach(function(y){var S=y.color;"attribute"in S&&(b[g++]=la(m[S.attribute]||S.defaultValue||one))}),r.forEach(function(y){var S=y.size;"attribute"in S&&(b[g++]=m[S.attribute]||S.defaultValue)})}},{key:"setUniforms",value:function(p,g){var m=g.gl,b=g.uniformLocations,y=b.u_sizeRatio,S=b.u_correctionRatio,x=b.u_matrix;m.uniform1f(S,p.correctionRatio),m.uniform1f(y,p.sizeRatio),m.uniformMatrix3fv(x,!1,p.matrix),r.forEach(function(R,k){var A=R.color;if("value"in A){var C=b["u_borderColor_".concat(k+1)],N=yee(A.value),_=Vte(N,4),O=_[0],F=_[1],L=_[2],I=_[3];m.uniform4f(C,O/255,F/255,L/255,I/255)}})}}]),c}(hT),ys(t,"ANGLE_1",0),ys(t,"ANGLE_2",2*Math.PI/3),ys(t,"ANGLE_3",4*Math.PI/3),t}var une=lne();function cne(e,t){if(typeof e!="object"||!e)return e;var n=e[Symbol.toPrimitive];if(n!==void 0){var r=n.call(e,t);if(typeof r!="object")return r;throw new TypeError("@@toPrimitive must return a primitive value.")}return(t==="string"?String:Number)(e)}function X4(e){var t=cne(e,"string");return typeof t=="symbol"?t:t+""}function Z4(e,t,n){return(t=X4(t))in e?Object.defineProperty(e,t,{value:n,enumerable:!0,configurable:!0,writable:!0}):e[t]=n,e}function jN(e,t){var n=Object.keys(e);if(Object.getOwnPropertySymbols){var r=Object.getOwnPropertySymbols(e);t&&(r=r.filter(function(a){return Object.getOwnPropertyDescriptor(e,a).enumerable})),n.push.apply(n,r)}return n}function Af(e){for(var t=1;t<arguments.length;t++){var n=arguments[t]!=null?arguments[t]:{};t%2?jN(Object(n),!0).forEach(function(r){Z4(e,r,n[r])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(n)):jN(Object(n)).forEach(function(r){Object.defineProperty(e,r,Object.getOwnPropertyDescriptor(n,r))})}return e}function dne(e,t){if(!(e instanceof t))throw new TypeError("Cannot call a class as a function")}function fne(e,t){for(var n=0;n<t.length;n++){var r=t[n];r.enumerable=r.enumerable||!1,r.configurable=!0,"value"in r&&(r.writable=!0),Object.defineProperty(e,X4(r.key),r)}}function pne(e,t,n){return t&&fne(e.prototype,t),Object.defineProperty(e,"prototype",{writable:!1}),e}function Rf(e){return Rf=Object.setPrototypeOf?Object.getPrototypeOf.bind():function(t){return t.__proto__||Object.getPrototypeOf(t)},Rf(e)}function Q4(){try{var e=!Boolean.prototype.valueOf.call(Reflect.construct(Boolean,[],function(){}))}catch{}return(Q4=function(){return!!e})()}function J4(e){if(e===void 0)throw new ReferenceError("this hasn't been initialised - super() hasn't been called");return e}function gne(e,t){if(t&&(typeof t=="object"||typeof t=="function"))return t;if(t!==void 0)throw new TypeError("Derived constructors may only return object or undefined");return J4(e)}function hne(e,t,n){return t=Rf(t),gne(e,Q4()?Reflect.construct(t,n||[],Rf(e).constructor):t.apply(e,n))}function $k(e,t){return $k=Object.setPrototypeOf?Object.setPrototypeOf.bind():function(n,r){return n.__proto__=r,n},$k(e,t)}function mne(e,t){if(typeof t!="function"&&t!==null)throw new TypeError("Super expression must either be null or a function");e.prototype=Object.create(t&&t.prototype,{constructor:{value:e,writable:!0,configurable:!0}}),Object.defineProperty(e,"prototype",{writable:!1}),t&&$k(e,t)}function qk(e,t){(t==null||t>e.length)&&(t=e.length);for(var n=0,r=Array(t);n<t;n++)r[n]=e[n];return r}function bne(e){if(Array.isArray(e))return qk(e)}function yne(e){if(typeof Symbol<"u"&&e[Symbol.iterator]!=null||e["@@iterator"]!=null)return Array.from(e)}function vne(e,t){if(e){if(typeof e=="string")return qk(e,t);var n={}.toString.call(e).slice(8,-1);return n==="Object"&&e.constructor&&(n=e.constructor.name),n==="Map"||n==="Set"?Array.from(e):n==="Arguments"||/^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n)?qk(e,t):void 0}}function Sne(){throw new TypeError(`Invalid attempt to spread non-iterable instance.
In order to be iterable, non-array objects must have a [Symbol.iterator]() method.`)}function km(e){return bne(e)||yne(e)||vne(e)||Sne()}function e5(e,t,n,r){var a=Math.pow(1-e,2)*t.x+2*(1-e)*e*n.x+Math.pow(e,2)*r.x,o=Math.pow(1-e,2)*t.y+2*(1-e)*e*n.y+Math.pow(e,2)*r.y;return{x:a,y:o}}function Ene(e,t,n){for(var r=20,a=0,o=e,s=0;s<r;s++){var u=e5((s+1)/r,e,t,n);a+=Math.sqrt(Math.pow(o.x-u.x,2)+Math.pow(o.y-u.y,2)),o=u}return a}function wne(e){var t=e.curvatureAttribute,n=e.defaultCurvature,r=e.keepLabelUpright,a=r===void 0?!0:r;return function(o,s,u,c,d){var p=d.edgeLabelSize,g=s[t]||n,m=d.edgeLabelFont,b=d.edgeLabelWeight,y=d.edgeLabelColor.attribute?s[d.edgeLabelColor.attribute]||d.edgeLabelColor.color||"#000":d.edgeLabelColor.color,S=s.label;if(S){o.fillStyle=y,o.font="".concat(b," ").concat(p,"px ").concat(m);var x=!a||u.x<c.x,R=x?u.x:c.x,k=x?u.y:c.y,A=x?c.x:u.x,C=x?c.y:u.y,N=(R+A)/2,_=(k+C)/2,O=A-R,F=C-k,L=Math.sqrt(Math.pow(O,2)+Math.pow(F,2)),I=x?1:-1,H=N+F*g*I,$=_-O*g*I,U=s.size*.7+5,Y={x:$-k,y:-(H-R)},Z=Math.sqrt(Math.pow(Y.x,2)+Math.pow(Y.y,2)),j={x:C-$,y:-(A-H)},G=Math.sqrt(Math.pow(j.x,2)+Math.pow(j.y,2));R+=U*Y.x/Z,k+=U*Y.y/Z,A+=U*j.x/G,C+=U*j.y/G,H+=U*F/L,$-=U*O/L;var z={x:H,y:$},K={x:R,y:k},D={x:A,y:C},V=Ene(K,z,D);if(!(V<u.size+c.size)){var B=o.measureText(S).width,M=V-u.size-c.size;if(B>M){var W="…";for(S=S+W,B=o.measureText(S).width;B>M&&S.length>1;)S=S.slice(0,-2)+W,B=o.measureText(S).width;if(S.length<4)return}for(var Q={},oe=0,re=S.length;oe<re;oe++){var ie=S[oe];Q[ie]||(Q[ie]=o.measureText(ie).width*(1+g*.35))}for(var Se=.5-B/V/2,ae=0,ve=S.length;ae<ve;ae++){var xe=S[ae],Ie=e5(Se,K,z,D),Ce=2*(1-Se)*(H-R)+2*Se*(A-H),ke=2*(1-Se)*($-k)+2*Se*(C-$),J=Math.atan2(ke,Ce);o.save(),o.translate(Ie.x,Ie.y),o.rotate(J),o.fillText(xe,0,0),o.restore(),Se+=Q[xe]/V}}}}}function xne(e){var t=e.arrowHead,n=(t==null?void 0:t.extremity)==="target"||(t==null?void 0:t.extremity)==="both",r=(t==null?void 0:t.extremity)==="source"||(t==null?void 0:t.extremity)==="both",a=`
precision highp float;

varying vec4 v_color;
varying float v_thickness;
varying float v_feather;
varying vec2 v_cpA;
varying vec2 v_cpB;
varying vec2 v_cpC;
`.concat(n?`
varying float v_targetSize;
varying vec2 v_targetPoint;`:"",`
`).concat(r?`
varying float v_sourceSize;
varying vec2 v_sourcePoint;`:"",`
`).concat(t?`
uniform float u_lengthToThicknessRatio;
uniform float u_widenessToThicknessRatio;`:"",`

float det(vec2 a, vec2 b) {
  return a.x * b.y - b.x * a.y;
}

vec2 getDistanceVector(vec2 b0, vec2 b1, vec2 b2) {
  float a = det(b0, b2), b = 2.0 * det(b1, b0), d = 2.0 * det(b2, b1);
  float f = b * d - a * a;
  vec2 d21 = b2 - b1, d10 = b1 - b0, d20 = b2 - b0;
  vec2 gf = 2.0 * (b * d21 + d * d10 + a * d20);
  gf = vec2(gf.y, -gf.x);
  vec2 pp = -f * gf / dot(gf, gf);
  vec2 d0p = b0 - pp;
  float ap = det(d0p, d20), bp = 2.0 * det(d10, d0p);
  float t = clamp((ap + bp) / (2.0 * a + b + d), 0.0, 1.0);
  return mix(mix(b0, b1, t), mix(b1, b2, t), t);
}

float distToQuadraticBezierCurve(vec2 p, vec2 b0, vec2 b1, vec2 b2) {
  return length(getDistanceVector(b0 - p, b1 - p, b2 - p));
}

const vec4 transparent = vec4(0.0, 0.0, 0.0, 0.0);

void main(void) {
  float dist = distToQuadraticBezierCurve(gl_FragCoord.xy, v_cpA, v_cpB, v_cpC);
  float thickness = v_thickness;
`).concat(n?`
  float distToTarget = length(gl_FragCoord.xy - v_targetPoint);
  float targetArrowLength = v_targetSize + thickness * u_lengthToThicknessRatio;
  if (distToTarget < targetArrowLength) {
    thickness = (distToTarget - v_targetSize) / (targetArrowLength - v_targetSize) * u_widenessToThicknessRatio * thickness;
  }`:"",`
`).concat(r?`
  float distToSource = length(gl_FragCoord.xy - v_sourcePoint);
  float sourceArrowLength = v_sourceSize + thickness * u_lengthToThicknessRatio;
  if (distToSource < sourceArrowLength) {
    thickness = (distToSource - v_sourceSize) / (sourceArrowLength - v_sourceSize) * u_widenessToThicknessRatio * thickness;
  }`:"",`

  float halfThickness = thickness / 2.0;
  if (dist < halfThickness) {
    #ifdef PICKING_MODE
    gl_FragColor = v_color;
    #else
    float t = smoothstep(
      halfThickness - v_feather,
      halfThickness,
      dist
    );

    gl_FragColor = mix(v_color, transparent, t);
    #endif
  } else {
    gl_FragColor = transparent;
  }
}
`);return a}function kne(e){var t=e.arrowHead,n=(t==null?void 0:t.extremity)==="target"||(t==null?void 0:t.extremity)==="both",r=(t==null?void 0:t.extremity)==="source"||(t==null?void 0:t.extremity)==="both",a=`
attribute vec4 a_id;
attribute vec4 a_color;
attribute float a_direction;
attribute float a_thickness;
attribute vec2 a_source;
attribute vec2 a_target;
attribute float a_current;
attribute float a_curvature;
`.concat(n?`attribute float a_targetSize;
`:"",`
`).concat(r?`attribute float a_sourceSize;
`:"",`

uniform mat3 u_matrix;
uniform float u_sizeRatio;
uniform float u_pixelRatio;
uniform vec2 u_dimensions;
uniform float u_minEdgeThickness;
uniform float u_feather;

varying vec4 v_color;
varying float v_thickness;
varying float v_feather;
varying vec2 v_cpA;
varying vec2 v_cpB;
varying vec2 v_cpC;
`).concat(n?`
varying float v_targetSize;
varying vec2 v_targetPoint;`:"",`
`).concat(r?`
varying float v_sourceSize;
varying vec2 v_sourcePoint;`:"",`
`).concat(t?`
uniform float u_widenessToThicknessRatio;`:"",`

const float bias = 255.0 / 254.0;
const float epsilon = 0.7;

vec2 clipspaceToViewport(vec2 pos, vec2 dimensions) {
  return vec2(
    (pos.x + 1.0) * dimensions.x / 2.0,
    (pos.y + 1.0) * dimensions.y / 2.0
  );
}

vec2 viewportToClipspace(vec2 pos, vec2 dimensions) {
  return vec2(
    pos.x / dimensions.x * 2.0 - 1.0,
    pos.y / dimensions.y * 2.0 - 1.0
  );
}

void main() {
  float minThickness = u_minEdgeThickness;

  // Selecting the correct position
  // Branchless "position = a_source if a_current == 1.0 else a_target"
  vec2 position = a_source * max(0.0, a_current) + a_target * max(0.0, 1.0 - a_current);
  position = (u_matrix * vec3(position, 1)).xy;

  vec2 source = (u_matrix * vec3(a_source, 1)).xy;
  vec2 target = (u_matrix * vec3(a_target, 1)).xy;

  vec2 viewportPosition = clipspaceToViewport(position, u_dimensions);
  vec2 viewportSource = clipspaceToViewport(source, u_dimensions);
  vec2 viewportTarget = clipspaceToViewport(target, u_dimensions);

  vec2 delta = viewportTarget.xy - viewportSource.xy;
  float len = length(delta);
  vec2 normal = vec2(-delta.y, delta.x) * a_direction;
  vec2 unitNormal = normal / len;
  float boundingBoxThickness = len * a_curvature;

  float curveThickness = max(minThickness, a_thickness / u_sizeRatio);
  v_thickness = curveThickness * u_pixelRatio;
  v_feather = u_feather;

  v_cpA = viewportSource;
  v_cpB = 0.5 * (viewportSource + viewportTarget) + unitNormal * a_direction * boundingBoxThickness;
  v_cpC = viewportTarget;

  vec2 viewportOffsetPosition = (
    viewportPosition +
    unitNormal * (boundingBoxThickness / 2.0 + sign(boundingBoxThickness) * (`).concat(t?"curveThickness * u_widenessToThicknessRatio":"curveThickness",` + epsilon)) *
    max(0.0, a_direction) // NOTE: cutting the bounding box in half to avoid overdraw
  );

  position = viewportToClipspace(viewportOffsetPosition, u_dimensions);
  gl_Position = vec4(position, 0, 1);
    
`).concat(n?`
  v_targetSize = a_targetSize * u_pixelRatio / u_sizeRatio;
  v_targetPoint = viewportTarget;
`:"",`
`).concat(r?`
  v_sourceSize = a_sourceSize * u_pixelRatio / u_sizeRatio;
  v_sourcePoint = viewportSource;
`:"",`

  #ifdef PICKING_MODE
  // For picking mode, we use the ID as the color:
  v_color = a_id;
  #else
  // For normal mode, we use the color:
  v_color = a_color;
  #endif

  v_color.a *= bias;
}
`);return a}var t5=.25,Tne={arrowHead:null,curvatureAttribute:"curvature",defaultCurvature:t5},n5=WebGLRenderingContext,GN=n5.UNSIGNED_BYTE,So=n5.FLOAT;function ST(e){var t=Af(Af({},Tne),e||{}),n=t,r=n.arrowHead,a=n.curvatureAttribute,o=n.drawLabel,s=(r==null?void 0:r.extremity)==="target"||(r==null?void 0:r.extremity)==="both",u=(r==null?void 0:r.extremity)==="source"||(r==null?void 0:r.extremity)==="both",c=["u_matrix","u_sizeRatio","u_dimensions","u_pixelRatio","u_feather","u_minEdgeThickness"].concat(km(r?["u_lengthToThicknessRatio","u_widenessToThicknessRatio"]:[]));return function(d){mne(p,d);function p(){var g;dne(this,p);for(var m=arguments.length,b=new Array(m),y=0;y<m;y++)b[y]=arguments[y];return g=hne(this,p,[].concat(b)),Z4(J4(g),"drawLabel",o||wne(t)),g}return pne(p,[{key:"getDefinition",value:function(){return{VERTICES:6,VERTEX_SHADER_SOURCE:kne(t),FRAGMENT_SHADER_SOURCE:xne(t),METHOD:WebGLRenderingContext.TRIANGLES,UNIFORMS:c,ATTRIBUTES:[{name:"a_source",size:2,type:So},{name:"a_target",size:2,type:So}].concat(km(s?[{name:"a_targetSize",size:1,type:So}]:[]),km(u?[{name:"a_sourceSize",size:1,type:So}]:[]),[{name:"a_thickness",size:1,type:So},{name:"a_curvature",size:1,type:So},{name:"a_color",size:4,type:GN,normalized:!0},{name:"a_id",size:4,type:GN,normalized:!0}]),CONSTANT_ATTRIBUTES:[{name:"a_current",size:1,type:So},{name:"a_direction",size:1,type:So}],CONSTANT_DATA:[[0,1],[0,-1],[1,1],[0,-1],[1,1],[1,-1]]}}},{key:"processVisibleItem",value:function(m,b,y,S,x){var R,k=x.size||1,A=y.x,C=y.y,N=S.x,_=S.y,O=la(x.color),F=(R=x[a])!==null&&R!==void 0?R:t5,L=this.array;L[b++]=A,L[b++]=C,L[b++]=N,L[b++]=_,s&&(L[b++]=S.size),u&&(L[b++]=y.size),L[b++]=k,L[b++]=F,L[b++]=O,L[b++]=m}},{key:"setUniforms",value:function(m,b){var y=b.gl,S=b.uniformLocations,x=S.u_matrix,R=S.u_pixelRatio,k=S.u_feather,A=S.u_sizeRatio,C=S.u_dimensions,N=S.u_minEdgeThickness;if(y.uniformMatrix3fv(x,!1,m.matrix),y.uniform1f(R,m.pixelRatio),y.uniform1f(A,m.sizeRatio),y.uniform1f(k,m.antiAliasingFeather),y.uniform2f(C,m.width*m.pixelRatio,m.height*m.pixelRatio),y.uniform1f(N,m.minEdgeThickness),r){var _=S.u_lengthToThicknessRatio,O=S.u_widenessToThicknessRatio;y.uniform1f(_,r.lengthToThicknessRatio),y.uniform1f(O,r.widenessToThicknessRatio)}}}]),p}(Zu)}var Ane=ST(),Rne=ST({arrowHead:Ju});ST({arrowHead:Af(Af({},Ju),{},{extremity:"both"})});const _ne=({node:e,move:t})=>{const n=kr(),{gotoNode:r}=G4();return E.useEffect(()=>{const a=n.getGraph();if(t){if(e&&a.hasNode(e))try{a.setNodeAttribute(e,"highlighted",!0),r(e)}catch(o){console.error("Error focusing on node:",o)}else n.setCustomBBox(null),n.getCamera().animate({x:.5,y:.5,ratio:1},{duration:0});ze.getState().setMoveToSelectedNode(!1)}else if(e&&a.hasNode(e))try{a.setNodeAttribute(e,"highlighted",!0)}catch(o){console.error("Error highlighting node:",o)}return()=>{if(e&&a.hasNode(e))try{a.setNodeAttribute(e,"highlighted",!1)}catch(o){console.error("Error cleaning up node highlight:",o)}}},[e,t,n,r]),null};function Qs(e,t){const n=kr(),r=E.useRef(t);return ec(r.current,t)||(r.current=t),{positions:E.useCallback(()=>r.current?e(n.getGraph(),r.current):{},[n,r,e]),assign:E.useCallback(()=>{r.current&&e.assign(n.getGraph(),r.current)},[n,r,e])}}function ET(e,t){const n=kr(),[r,a]=E.useState(!1),[o,s]=E.useState(null),u=E.useRef(t);return ec(u.current,t)||(u.current=t),E.useEffect(()=>{a(!1);let c=null;return u.current&&(c=new e(n.getGraph(),u.current)),s(c),()=>{c!==null&&c.kill()}},[n,u,s,a,e]),{stop:E.useCallback(()=>{o&&(o.stop(),a(!1))},[o,a]),start:E.useCallback(()=>{o&&(o.start(),a(!0))},[o,a]),kill:E.useCallback(()=>{o&&o.kill(),a(!1)},[o,a]),isRunning:r}}var Tm,HN;function tc(){if(HN)return Tm;HN=1;function e(n){return!n||typeof n!="object"||typeof n=="function"||Array.isArray(n)||n instanceof Set||n instanceof Map||n instanceof RegExp||n instanceof Date}function t(n,r){n=n||{};var a={};for(var o in r){var s=n[o],u=r[o];if(!e(u)){a[o]=t(s,u);continue}s===void 0?a[o]=u:a[o]=s}return a}return Tm=t,Tm}var Am,$N;function Cne(){if($N)return Am;$N=1;function e(n){return function(r,a){return r+Math.floor(n()*(a-r+1))}}var t=e(Math.random);return t.createRandom=e,Am=t,Am}var Rm,qN;function Nne(){if(qN)return Rm;qN=1;var e=Cne().createRandom;function t(r){var a=e(r);return function(o){for(var s=o.length,u=s-1,c=-1;++c<s;){var d=a(c,u),p=o[d];o[d]=o[c],o[c]=p}}}var n=t(Math.random);return n.createShuffleInPlace=t,Rm=n,Rm}var _m,VN;function One(){if(VN)return _m;VN=1;var e=tc(),t=ua(),n=Nne(),r={attributes:{x:"x",y:"y"},center:0,hierarchyAttributes:[],rng:Math.random,scale:1};function a(O,F,L,I,H){this.wrappedCircle=H||null,this.children={},this.countChildren=0,this.id=O||null,this.next=null,this.previous=null,this.x=F||null,this.y=L||null,H?this.r=1010101:this.r=I||999}a.prototype.hasChildren=function(){return this.countChildren>0},a.prototype.addChild=function(O,F){this.children[O]=F,++this.countChildren},a.prototype.getChild=function(O){if(!this.children.hasOwnProperty(O)){var F=new a;this.children[O]=F,++this.countChildren}return this.children[O]},a.prototype.applyPositionToChildren=function(){if(this.hasChildren()){var O=this;for(var F in O.children){var L=O.children[F];L.x+=O.x,L.y+=O.y,L.applyPositionToChildren()}}};function o(O,F,L){for(var I in F.children){var H=F.children[I];H.hasChildren()?o(O,H,L):L[H.id]={x:H.x,y:H.y}}}function s(O,F){var L=O.r-F.r,I=F.x-O.x,H=F.y-O.y;return L<0||L*L<I*I+H*H}function u(O,F){var L=O.r-F.r+1e-6,I=F.x-O.x,H=F.y-O.y;return L>0&&L*L>I*I+H*H}function c(O,F){for(var L=0;L<F.length;++L)if(!u(O,F[L]))return!1;return!0}function d(O){return new a(null,O.x,O.y,O.r)}function p(O,F){var L=O.x,I=O.y,H=O.r,$=F.x,U=F.y,Y=F.r,Z=$-L,j=U-I,G=Y-H,z=Math.sqrt(Z*Z+j*j);return new a(null,(L+$+Z/z*G)/2,(I+U+j/z*G)/2,(z+H+Y)/2)}function g(O,F,L){var I=O.x,H=O.y,$=O.r,U=F.x,Y=F.y,Z=F.r,j=L.x,G=L.y,z=L.r,K=I-U,D=I-j,V=H-Y,B=H-G,M=Z-$,W=z-$,Q=I*I+H*H-$*$,oe=Q-U*U-Y*Y+Z*Z,re=Q-j*j-G*G+z*z,ie=D*V-K*B,Se=(V*re-B*oe)/(ie*2)-I,ae=(B*M-V*W)/ie,ve=(D*oe-K*re)/(ie*2)-H,xe=(K*W-D*M)/ie,Ie=ae*ae+xe*xe-1,Ce=2*($+Se*ae+ve*xe),ke=Se*Se+ve*ve-$*$,J=-(Ie?(Ce+Math.sqrt(Ce*Ce-4*Ie*ke))/(2*Ie):ke/Ce);return new a(null,I+Se+ae*J,H+ve+xe*J,J)}function m(O){switch(O.length){case 1:return d(O[0]);case 2:return p(O[0],O[1]);case 3:return g(O[0],O[1],O[2]);default:throw new Error("graphology-layout/circlepack: Invalid basis length "+O.length)}}function b(O,F){var L,I;if(c(F,O))return[F];for(L=0;L<O.length;++L)if(s(F,O[L])&&c(p(O[L],F),O))return[O[L],F];for(L=0;L<O.length-1;++L)for(I=L+1;I<O.length;++I)if(s(p(O[L],O[I]),F)&&s(p(O[L],F),O[I])&&s(p(O[I],F),O[L])&&c(g(O[L],O[I],F),O))return[O[L],O[I],F];throw new Error("graphology-layout/circlepack: extendBasis failure !")}function y(O){var F=O.wrappedCircle,L=O.next.wrappedCircle,I=F.r+L.r,H=(F.x*L.r+L.x*F.r)/I,$=(F.y*L.r+L.y*F.r)/I;return H*H+$*$}function S(O,F){var L=0,I=O.slice(),H=O.length,$=[],U,Y;for(F(I);L<H;)U=I[L],Y&&u(Y,U)?++L:($=b($,U),Y=m($),L=0);return Y}function x(O,F,L){var I=O.x-F.x,H,$,U=O.y-F.y,Y,Z,j=I*I+U*U;j?($=F.r+L.r,$*=$,Z=O.r+L.r,Z*=Z,$>Z?(H=(j+Z-$)/(2*j),Y=Math.sqrt(Math.max(0,Z/j-H*H)),L.x=O.x-H*I-Y*U,L.y=O.y-H*U+Y*I):(H=(j+$-Z)/(2*j),Y=Math.sqrt(Math.max(0,$/j-H*H)),L.x=F.x+H*I-Y*U,L.y=F.y+H*U+Y*I)):(L.x=F.x+L.r,L.y=F.y)}function R(O,F){var L=O.r+F.r-1e-6,I=F.x-O.x,H=F.y-O.y;return L>0&&L*L>I*I+H*H}function k(O,F){var L=O.length;if(L===0)return 0;var I,H,$,U,Y,Z,j,G,z,K;if(I=O[0],I.x=0,I.y=0,L<=1)return I.r;if(H=O[1],I.x=-H.r,H.x=I.r,H.y=0,L<=2)return I.r+H.r;$=O[2],x(H,I,$),I=new a(null,null,null,null,I),H=new a(null,null,null,null,H),$=new a(null,null,null,null,$),I.next=$.previous=H,H.next=I.previous=$,$.next=H.previous=I;e:for(Z=3;Z<L;++Z){$=O[Z],x(I.wrappedCircle,H.wrappedCircle,$),$=new a(null,null,null,null,$),j=H.next,G=I.previous,z=H.wrappedCircle.r,K=I.wrappedCircle.r;do if(z<=K){if(R(j.wrappedCircle,$.wrappedCircle)){H=j,I.next=H,H.previous=I,--Z;continue e}z+=j.wrappedCircle.r,j=j.next}else{if(R(G.wrappedCircle,$.wrappedCircle)){I=G,I.next=H,H.previous=I,--Z;continue e}K+=G.wrappedCircle.r,G=G.previous}while(j!==G.next);for($.previous=I,$.next=H,I.next=H.previous=H=$,U=y(I);($=$.next)!==H;)(Y=y($))<U&&(I=$,U=Y);H=I.next}I=[H.wrappedCircle],$=H;for(var D=1e4;($=$.next)!==H&&--D!==0;)I.push($.wrappedCircle);for($=S(I,F),Z=0;Z<L;++Z)I=O[Z],I.x-=$.x,I.y-=$.y;return $.r}function A(O,F){var L=0;if(O.hasChildren()){for(var I in O.children){var H=O.children[I];H.hasChildren()&&(H.r=A(H,F))}L=k(Object.values(O.children),F)}return L}function C(O,F){A(O,F);for(var L in O.children){var I=O.children[L];I.applyPositionToChildren()}}function N(O,F,L){if(!t(F))throw new Error("graphology-layout/circlepack: the given graph is not a valid graphology instance.");L=e(L,r);var I={},H={},$=F.nodes(),U=L.center,Y=L.hierarchyAttributes,Z=n.createShuffleInPlace(L.rng),j=L.scale,G=new a;F.forEachNode(function(M,W){var Q=W.size?W.size:1,oe=new a(M,null,null,Q),re=G;Y.forEach(function(ie){var Se=W[ie];re=re.getChild(Se)}),re.addChild(M,oe)}),C(G,Z),o(F,G,I);var z=$.length,K,D,V;for(V=0;V<z;V++){var B=$[V];K=U+j*I[B].x,D=U+j*I[B].y,H[B]={x:K,y:D},O&&(F.setNodeAttribute(B,L.attributes.x,K),F.setNodeAttribute(B,L.attributes.y,D))}return H}var _=N.bind(null,!1);return _.assign=N.bind(null,!0),_m=_,_m}var Dne=One();const Ine=un(Dne);function Lne(e={}){return Qs(Ine,e)}var Cm,WN;function Mne(){if(WN)return Cm;WN=1;var e=tc(),t=ua(),n={dimensions:["x","y"],center:.5,scale:1};function r(o,s,u){if(!t(s))throw new Error("graphology-layout/random: the given graph is not a valid graphology instance.");u=e(u,n);var c=u.dimensions;if(!Array.isArray(c)||c.length!==2)throw new Error("graphology-layout/random: given dimensions are invalid.");var d=u.center,p=u.scale,g=Math.PI*2,m=(d-.5)*p,b=s.order,y=c[0],S=c[1];function x(A,C){return C[y]=p*Math.cos(A*g/b)+m,C[S]=p*Math.sin(A*g/b)+m,C}var R=0;if(!o){var k={};return s.forEachNode(function(A){k[A]=x(R++,{})}),k}s.updateEachNodeAttributes(function(A,C){return x(R++,C),C},{attributes:c})}var a=r.bind(null,!1);return a.assign=r.bind(null,!0),Cm=a,Cm}var Fne=Mne();const Pne=un(Fne);function zne(e={}){return Qs(Pne,e)}var su={},YN;function wT(){if(YN)return su;YN=1;function e(r){return typeof r!="number"||isNaN(r)?1:r}function t(r,a){var o={},s=function(d){return typeof d>"u"?a:d};typeof a=="function"&&(s=a);var u=function(d){return s(d[r])},c=function(){return s(void 0)};return typeof r=="string"?(o.fromAttributes=u,o.fromGraph=function(d,p){return u(d.getNodeAttributes(p))},o.fromEntry=function(d,p){return u(p)}):typeof r=="function"?(o.fromAttributes=function(){throw new Error("graphology-utils/getters/createNodeValueGetter: irrelevant usage.")},o.fromGraph=function(d,p){return s(r(p,d.getNodeAttributes(p)))},o.fromEntry=function(d,p){return s(r(d,p))}):(o.fromAttributes=c,o.fromGraph=c,o.fromEntry=c),o}function n(r,a){var o={},s=function(d){return typeof d>"u"?a:d};typeof a=="function"&&(s=a);var u=function(d){return s(d[r])},c=function(){return s(void 0)};return typeof r=="string"?(o.fromAttributes=u,o.fromGraph=function(d,p){return u(d.getEdgeAttributes(p))},o.fromEntry=function(d,p){return u(p)},o.fromPartialEntry=o.fromEntry,o.fromMinimalEntry=o.fromEntry):typeof r=="function"?(o.fromAttributes=function(){throw new Error("graphology-utils/getters/createEdgeValueGetter: irrelevant usage.")},o.fromGraph=function(d,p){var g=d.extremities(p);return s(r(p,d.getEdgeAttributes(p),g[0],g[1],d.getNodeAttributes(g[0]),d.getNodeAttributes(g[1]),d.isUndirected(p)))},o.fromEntry=function(d,p,g,m,b,y,S){return s(r(d,p,g,m,b,y,S))},o.fromPartialEntry=function(d,p,g,m){return s(r(d,p,g,m))},o.fromMinimalEntry=function(d,p){return s(r(d,p))}):(o.fromAttributes=c,o.fromGraph=c,o.fromEntry=c,o.fromMinimalEntry=c),o}return su.createNodeValueGetter=t,su.createEdgeValueGetter=n,su.createEdgeWeightGetter=function(r){return n(r,e)},su}var Nm,KN;function r5(){if(KN)return Nm;KN=1;const{createNodeValueGetter:e,createEdgeValueGetter:t}=wT();return Nm=function(r,a,o){const{nodeXAttribute:s,nodeYAttribute:u}=o,{attraction:c,repulsion:d,gravity:p,inertia:g,maxMove:m}=o.settings;let{shouldSkipNode:b,shouldSkipEdge:y,isNodeFixed:S}=o;S=e(S),b=e(b,!1),y=t(y,!1);const x=r.filterNodes((A,C)=>!b.fromEntry(A,C)),R=x.length;for(let A=0;A<R;A++){const C=x[A],N=r.getNodeAttributes(C),_=a[C];_?a[C]={dx:_.dx*g,dy:_.dy*g,x:N[s]||0,y:N[u]||0}:a[C]={dx:0,dy:0,x:N[s]||0,y:N[u]||0}}if(d)for(let A=0;A<R;A++){const C=x[A],N=a[C];for(let _=A+1;_<R;_++){const O=x[_],F=a[O],L=F.x-N.x,I=F.y-N.y,H=Math.sqrt(L*L+I*I)||1,$=d/H*L,U=d/H*I;N.dx-=$,N.dy-=U,F.dx+=$,F.dy+=U}}if(c&&r.forEachEdge((A,C,N,_,O,F,L)=>{if(N===_||b.fromEntry(N,O)||b.fromEntry(_,F)||y.fromEntry(A,C,N,_,O,F,L))return;const I=a[N],H=a[_],$=H.x-I.x,U=H.y-I.y,Y=Math.sqrt($*$+U*U)||1,Z=c*Y*$,j=c*Y*U;I.dx+=Z,I.dy+=j,H.dx-=Z,H.dy-=j}),p)for(let A=0;A<R;A++){const C=x[A],N=a[C],{x:_,y:O}=N,F=Math.sqrt(_*_+O*O)||1;a[C].dx-=_*p*F,a[C].dy-=O*p*F}const k=!1;for(let A=0;A<R;A++){const C=x[A],N=a[C],_=Math.sqrt(N.dx*N.dx+N.dy*N.dy);_>m&&(N.dx*=m/_,N.dy*=m/_),S.fromGraph(r,C)?N.fixed=!0:(N.x+=N.dx,N.y+=N.dy,N.fixed=!1)}return{converged:k}},Nm}var Dd={},XN;function a5(){return XN||(XN=1,Dd.assignLayoutChanges=function(e,t,n){const{nodeXAttribute:r,nodeYAttribute:a}=n;e.updateEachNodeAttributes((o,s)=>{const u=t[o];return!u||u.fixed||(s[r]=u.x,s[a]=u.y),s},{attributes:["x","y"]})},Dd.collectLayoutChanges=function(e){const t={};for(const n in e){const r=e[n];t[n]={x:r.x,y:r.y}}return t}),Dd}var Om,ZN;function o5(){return ZN||(ZN=1,Om={nodeXAttribute:"x",nodeYAttribute:"y",isNodeFixed:"fixed",shouldSkipNode:null,shouldSkipEdge:null,settings:{attraction:5e-4,repulsion:.1,gravity:1e-4,inertia:.6,maxMove:200}}),Om}var Dm,QN;function Bne(){if(QN)return Dm;QN=1;const e=ua(),t=tc(),n=r5(),r=a5(),a=o5();function o(u,c,d){if(!e(c))throw new Error("graphology-layout-force: the given graph is not a valid graphology instance.");typeof d=="number"?d={maxIterations:d}:d=d||{};const p=d.maxIterations;if(d=t(d,a),typeof p!="number"||p<=0)throw new Error("graphology-layout-force: you should provide a positive number of maximum iterations.");const g={};let m=null,b;for(b=0;b<p&&(m=n(c,g,d),!m.converged);b++);if(u){r.assignLayoutChanges(c,g,d);return}return r.collectLayoutChanges(g)}const s=o.bind(null,!1);return s.assign=o.bind(null,!0),Dm=s,Dm}var Une=Bne();const jne=un(Une);var Im,JN;function Gne(){if(JN)return Im;JN=1;const e=ua(),t=tc(),n=r5(),r=a5(),a=o5();function o(s,u){if(!e(s))throw new Error("graphology-layout-force/worker: the given graph is not a valid graphology instance.");u=t(u,a),this.callbacks={},u.onConverged&&(this.callbacks.onConverged=u.onConverged),this.graph=s,this.params=u,this.nodeStates={},this.frameID=null,this.running=!1,this.killed=!1}return o.prototype.isRunning=function(){return this.running},o.prototype.runFrame=function(){let{converged:s}=n(this.graph,this.nodeStates,this.params);r.assignLayoutChanges(this.graph,this.nodeStates,this.params),s=!1,s?(this.callbacks.onConverged&&this.callbacks.onConverged(),this.stop()):this.frameID=window.requestAnimationFrame(()=>this.runFrame())},o.prototype.stop=function(){return this.running=!1,this.frameID!==null&&(window.cancelAnimationFrame(this.frameID),this.frameID=null),this},o.prototype.start=function(){if(this.killed)throw new Error("graphology-layout-force/worker.start: layout was killed.");this.running||(this.running=!0,this.runFrame())},o.prototype.kill=function(){this.stop(),delete this.nodeStates,this.killed=!0},Im=o,Im}var Hne=Gne();const $ne=un(Hne);function qne(e={maxIterations:100}){return Qs(jne,e)}function Vne(e={}){return ET($ne,e)}var Lm,eO;function Wne(){if(eO)return Lm;eO=1;var e=0,t=1,n=2,r=3,a=4,o=5,s=6,u=7,c=8,d=9,p=0,g=1,m=2,b=0,y=1,S=2,x=3,R=4,k=5,A=6,C=7,N=8,_=3,O=10,F=3,L=9,I=10;return Lm=function($,U,Y){var Z,j,G,z,K,D,V,B,M,W,Q=U.length,oe=Y.length,re=$.adjustSizes,ie=$.barnesHutTheta*$.barnesHutTheta,Se,ae,ve,xe,Ie,Ce,ke,J=[];for(G=0;G<Q;G+=O)U[G+a]=U[G+n],U[G+o]=U[G+r],U[G+n]=0,U[G+r]=0;if($.outboundAttractionDistribution){for(Se=0,G=0;G<Q;G+=O)Se+=U[G+s];Se/=Q/O}if($.barnesHutOptimize){var fe=1/0,Te=-1/0,me=1/0,Ee=-1/0,le,Be,Ue;for(G=0;G<Q;G+=O)fe=Math.min(fe,U[G+e]),Te=Math.max(Te,U[G+e]),me=Math.min(me,U[G+t]),Ee=Math.max(Ee,U[G+t]);var he=Te-fe,Ne=Ee-me;for(he>Ne?(me-=(he-Ne)/2,Ee=me+he):(fe-=(Ne-he)/2,Te=fe+Ne),J[0+b]=-1,J[0+y]=(fe+Te)/2,J[0+S]=(me+Ee)/2,J[0+x]=Math.max(Te-fe,Ee-me),J[0+R]=-1,J[0+k]=-1,J[0+A]=0,J[0+C]=0,J[0+N]=0,Z=1,G=0;G<Q;G+=O)for(j=0,Ue=_;;)if(J[j+k]>=0){U[G+e]<J[j+y]?U[G+t]<J[j+S]?le=J[j+k]:le=J[j+k]+L:U[G+t]<J[j+S]?le=J[j+k]+L*2:le=J[j+k]+L*3,J[j+C]=(J[j+C]*J[j+A]+U[G+e]*U[G+s])/(J[j+A]+U[G+s]),J[j+N]=(J[j+N]*J[j+A]+U[G+t]*U[G+s])/(J[j+A]+U[G+s]),J[j+A]+=U[G+s],j=le;continue}else if(J[j+b]<0){J[j+b]=G;break}else{if(J[j+k]=Z*L,B=J[j+x]/2,M=J[j+k],J[M+b]=-1,J[M+y]=J[j+y]-B,J[M+S]=J[j+S]-B,J[M+x]=B,J[M+R]=M+L,J[M+k]=-1,J[M+A]=0,J[M+C]=0,J[M+N]=0,M+=L,J[M+b]=-1,J[M+y]=J[j+y]-B,J[M+S]=J[j+S]+B,J[M+x]=B,J[M+R]=M+L,J[M+k]=-1,J[M+A]=0,J[M+C]=0,J[M+N]=0,M+=L,J[M+b]=-1,J[M+y]=J[j+y]+B,J[M+S]=J[j+S]-B,J[M+x]=B,J[M+R]=M+L,J[M+k]=-1,J[M+A]=0,J[M+C]=0,J[M+N]=0,M+=L,J[M+b]=-1,J[M+y]=J[j+y]+B,J[M+S]=J[j+S]+B,J[M+x]=B,J[M+R]=J[j+R],J[M+k]=-1,J[M+A]=0,J[M+C]=0,J[M+N]=0,Z+=4,U[J[j+b]+e]<J[j+y]?U[J[j+b]+t]<J[j+S]?le=J[j+k]:le=J[j+k]+L:U[J[j+b]+t]<J[j+S]?le=J[j+k]+L*2:le=J[j+k]+L*3,J[j+A]=U[J[j+b]+s],J[j+C]=U[J[j+b]+e],J[j+N]=U[J[j+b]+t],J[le+b]=J[j+b],J[j+b]=-1,U[G+e]<J[j+y]?U[G+t]<J[j+S]?Be=J[j+k]:Be=J[j+k]+L:U[G+t]<J[j+S]?Be=J[j+k]+L*2:Be=J[j+k]+L*3,le===Be)if(Ue--){j=le;continue}else{Ue=_;break}J[Be+b]=G;break}}if($.barnesHutOptimize)for(ae=$.scalingRatio,G=0;G<Q;G+=O)for(j=0;;)if(J[j+k]>=0)if(Ce=Math.pow(U[G+e]-J[j+C],2)+Math.pow(U[G+t]-J[j+N],2),W=J[j+x],4*W*W/Ce<ie){if(ve=U[G+e]-J[j+C],xe=U[G+t]-J[j+N],re===!0?Ce>0?(ke=ae*U[G+s]*J[j+A]/Ce,U[G+n]+=ve*ke,U[G+r]+=xe*ke):Ce<0&&(ke=-ae*U[G+s]*J[j+A]/Math.sqrt(Ce),U[G+n]+=ve*ke,U[G+r]+=xe*ke):Ce>0&&(ke=ae*U[G+s]*J[j+A]/Ce,U[G+n]+=ve*ke,U[G+r]+=xe*ke),j=J[j+R],j<0)break;continue}else{j=J[j+k];continue}else{if(D=J[j+b],D>=0&&D!==G&&(ve=U[G+e]-U[D+e],xe=U[G+t]-U[D+t],Ce=ve*ve+xe*xe,re===!0?Ce>0?(ke=ae*U[G+s]*U[D+s]/Ce,U[G+n]+=ve*ke,U[G+r]+=xe*ke):Ce<0&&(ke=-ae*U[G+s]*U[D+s]/Math.sqrt(Ce),U[G+n]+=ve*ke,U[G+r]+=xe*ke):Ce>0&&(ke=ae*U[G+s]*U[D+s]/Ce,U[G+n]+=ve*ke,U[G+r]+=xe*ke)),j=J[j+R],j<0)break;continue}else for(ae=$.scalingRatio,z=0;z<Q;z+=O)for(K=0;K<z;K+=O)ve=U[z+e]-U[K+e],xe=U[z+t]-U[K+t],re===!0?(Ce=Math.sqrt(ve*ve+xe*xe)-U[z+c]-U[K+c],Ce>0?(ke=ae*U[z+s]*U[K+s]/Ce/Ce,U[z+n]+=ve*ke,U[z+r]+=xe*ke,U[K+n]-=ve*ke,U[K+r]-=xe*ke):Ce<0&&(ke=100*ae*U[z+s]*U[K+s],U[z+n]+=ve*ke,U[z+r]+=xe*ke,U[K+n]-=ve*ke,U[K+r]-=xe*ke)):(Ce=Math.sqrt(ve*ve+xe*xe),Ce>0&&(ke=ae*U[z+s]*U[K+s]/Ce/Ce,U[z+n]+=ve*ke,U[z+r]+=xe*ke,U[K+n]-=ve*ke,U[K+r]-=xe*ke));for(M=$.gravity/$.scalingRatio,ae=$.scalingRatio,G=0;G<Q;G+=O)ke=0,ve=U[G+e],xe=U[G+t],Ce=Math.sqrt(Math.pow(ve,2)+Math.pow(xe,2)),$.strongGravityMode?Ce>0&&(ke=ae*U[G+s]*M):Ce>0&&(ke=ae*U[G+s]*M/Ce),U[G+n]-=ve*ke,U[G+r]-=xe*ke;for(ae=1*($.outboundAttractionDistribution?Se:1),V=0;V<oe;V+=F)z=Y[V+p],K=Y[V+g],B=Y[V+m],Ie=Math.pow(B,$.edgeWeightInfluence),ve=U[z+e]-U[K+e],xe=U[z+t]-U[K+t],re===!0?(Ce=Math.sqrt(ve*ve+xe*xe)-U[z+c]-U[K+c],$.linLogMode?$.outboundAttractionDistribution?Ce>0&&(ke=-ae*Ie*Math.log(1+Ce)/Ce/U[z+s]):Ce>0&&(ke=-ae*Ie*Math.log(1+Ce)/Ce):$.outboundAttractionDistribution?Ce>0&&(ke=-ae*Ie/U[z+s]):Ce>0&&(ke=-ae*Ie)):(Ce=Math.sqrt(Math.pow(ve,2)+Math.pow(xe,2)),$.linLogMode?$.outboundAttractionDistribution?Ce>0&&(ke=-ae*Ie*Math.log(1+Ce)/Ce/U[z+s]):Ce>0&&(ke=-ae*Ie*Math.log(1+Ce)/Ce):$.outboundAttractionDistribution?(Ce=1,ke=-ae*Ie/U[z+s]):(Ce=1,ke=-ae*Ie)),Ce>0&&(U[z+n]+=ve*ke,U[z+r]+=xe*ke,U[K+n]-=ve*ke,U[K+r]-=xe*ke);var ee,ce,_e,Pe,We,yt;if(re===!0)for(G=0;G<Q;G+=O)U[G+d]!==1&&(ee=Math.sqrt(Math.pow(U[G+n],2)+Math.pow(U[G+r],2)),ee>I&&(U[G+n]=U[G+n]*I/ee,U[G+r]=U[G+r]*I/ee),ce=U[G+s]*Math.sqrt((U[G+a]-U[G+n])*(U[G+a]-U[G+n])+(U[G+o]-U[G+r])*(U[G+o]-U[G+r])),_e=Math.sqrt((U[G+a]+U[G+n])*(U[G+a]+U[G+n])+(U[G+o]+U[G+r])*(U[G+o]+U[G+r]))/2,Pe=.1*Math.log(1+_e)/(1+Math.sqrt(ce)),We=U[G+e]+U[G+n]*(Pe/$.slowDown),U[G+e]=We,yt=U[G+t]+U[G+r]*(Pe/$.slowDown),U[G+t]=yt);else for(G=0;G<Q;G+=O)U[G+d]!==1&&(ce=U[G+s]*Math.sqrt((U[G+a]-U[G+n])*(U[G+a]-U[G+n])+(U[G+o]-U[G+r])*(U[G+o]-U[G+r])),_e=Math.sqrt((U[G+a]+U[G+n])*(U[G+a]+U[G+n])+(U[G+o]+U[G+r])*(U[G+o]+U[G+r]))/2,Pe=U[G+u]*Math.log(1+_e)/(1+Math.sqrt(ce)),U[G+u]=Math.min(1,Math.sqrt(Pe*(Math.pow(U[G+n],2)+Math.pow(U[G+r],2))/(1+Math.sqrt(ce)))),We=U[G+e]+U[G+n]*(Pe/$.slowDown),U[G+e]=We,yt=U[G+t]+U[G+r]*(Pe/$.slowDown),U[G+t]=yt);return{}},Lm}var Da={},tO;function i5(){if(tO)return Da;tO=1;var e=10,t=3;return Da.assign=function(n){n=n||{};var r=Array.prototype.slice.call(arguments).slice(1),a,o,s;for(a=0,s=r.length;a<s;a++)if(r[a])for(o in r[a])n[o]=r[a][o];return n},Da.validateSettings=function(n){return"linLogMode"in n&&typeof n.linLogMode!="boolean"?{message:"the `linLogMode` setting should be a boolean."}:"outboundAttractionDistribution"in n&&typeof n.outboundAttractionDistribution!="boolean"?{message:"the `outboundAttractionDistribution` setting should be a boolean."}:"adjustSizes"in n&&typeof n.adjustSizes!="boolean"?{message:"the `adjustSizes` setting should be a boolean."}:"edgeWeightInfluence"in n&&typeof n.edgeWeightInfluence!="number"?{message:"the `edgeWeightInfluence` setting should be a number."}:"scalingRatio"in n&&!(typeof n.scalingRatio=="number"&&n.scalingRatio>=0)?{message:"the `scalingRatio` setting should be a number >= 0."}:"strongGravityMode"in n&&typeof n.strongGravityMode!="boolean"?{message:"the `strongGravityMode` setting should be a boolean."}:"gravity"in n&&!(typeof n.gravity=="number"&&n.gravity>=0)?{message:"the `gravity` setting should be a number >= 0."}:"slowDown"in n&&!(typeof n.slowDown=="number"||n.slowDown>=0)?{message:"the `slowDown` setting should be a number >= 0."}:"barnesHutOptimize"in n&&typeof n.barnesHutOptimize!="boolean"?{message:"the `barnesHutOptimize` setting should be a boolean."}:"barnesHutTheta"in n&&!(typeof n.barnesHutTheta=="number"&&n.barnesHutTheta>=0)?{message:"the `barnesHutTheta` setting should be a number >= 0."}:null},Da.graphToByteArrays=function(n,r){var a=n.order,o=n.size,s={},u,c=new Float32Array(a*e),d=new Float32Array(o*t);return u=0,n.forEachNode(function(p,g){s[p]=u,c[u]=g.x,c[u+1]=g.y,c[u+2]=0,c[u+3]=0,c[u+4]=0,c[u+5]=0,c[u+6]=1,c[u+7]=1,c[u+8]=g.size||1,c[u+9]=g.fixed?1:0,u+=e}),u=0,n.forEachEdge(function(p,g,m,b,y,S,x){var R=s[m],k=s[b],A=r(p,g,m,b,y,S,x);c[R+6]+=A,c[k+6]+=A,d[u]=R,d[u+1]=k,d[u+2]=A,u+=t}),{nodes:c,edges:d}},Da.assignLayoutChanges=function(n,r,a){var o=0;n.updateEachNodeAttributes(function(s,u){return u.x=r[o],u.y=r[o+1],o+=e,a?a(s,u):u})},Da.readGraphPositions=function(n,r){var a=0;n.forEachNode(function(o,s){r[a]=s.x,r[a+1]=s.y,a+=e})},Da.collectLayoutChanges=function(n,r,a){for(var o=n.nodes(),s={},u=0,c=0,d=r.length;u<d;u+=e){if(a){var p=Object.assign({},n.getNodeAttributes(o[c]));p.x=r[u],p.y=r[u+1],p=a(o[c],p),s[o[c]]={x:p.x,y:p.y}}else s[o[c]]={x:r[u],y:r[u+1]};c++}return s},Da.createWorker=function(r){var a=window.URL||window.webkitURL,o=r.toString(),s=a.createObjectURL(new Blob(["("+o+").call(this);"],{type:"text/javascript"})),u=new Worker(s);return a.revokeObjectURL(s),u},Da}var Mm,nO;function s5(){return nO||(nO=1,Mm={linLogMode:!1,outboundAttractionDistribution:!1,adjustSizes:!1,edgeWeightInfluence:1,scalingRatio:1,strongGravityMode:!1,gravity:1,slowDown:1,barnesHutOptimize:!1,barnesHutTheta:.5}),Mm}var Fm,rO;function Yne(){if(rO)return Fm;rO=1;var e=ua(),t=wT().createEdgeWeightGetter,n=Wne(),r=i5(),a=s5();function o(c,d,p){if(!e(d))throw new Error("graphology-layout-forceatlas2: the given graph is not a valid graphology instance.");typeof p=="number"&&(p={iterations:p});var g=p.iterations;if(typeof g!="number")throw new Error("graphology-layout-forceatlas2: invalid number of iterations.");if(g<=0)throw new Error("graphology-layout-forceatlas2: you should provide a positive number of iterations.");var m=t("getEdgeWeight"in p?p.getEdgeWeight:"weight").fromEntry,b=typeof p.outputReducer=="function"?p.outputReducer:null,y=r.assign({},a,p.settings),S=r.validateSettings(y);if(S)throw new Error("graphology-layout-forceatlas2: "+S.message);var x=r.graphToByteArrays(d,m),R;for(R=0;R<g;R++)n(y,x.nodes,x.edges);if(c){r.assignLayoutChanges(d,x.nodes,b);return}return r.collectLayoutChanges(d,x.nodes)}function s(c){var d=typeof c=="number"?c:c.order;return{barnesHutOptimize:d>2e3,strongGravityMode:!0,gravity:.05,scalingRatio:10,slowDown:1+Math.log(d)}}var u=o.bind(null,!1);return u.assign=o.bind(null,!0),u.inferSettings=s,Fm=u,Fm}var Kne=Yne();const Xne=un(Kne);var Pm,aO;function Zne(){return aO||(aO=1,Pm=function(){var t,n,r={};(function(){var o=0,s=1,u=2,c=3,d=4,p=5,g=6,m=7,b=8,y=9,S=0,x=1,R=2,k=0,A=1,C=2,N=3,_=4,O=5,F=6,L=7,I=8,H=3,$=10,U=3,Y=9,Z=10;r.exports=function(G,z,K){var D,V,B,M,W,Q,oe,re,ie,Se,ae=z.length,ve=K.length,xe=G.adjustSizes,Ie=G.barnesHutTheta*G.barnesHutTheta,Ce,ke,J,fe,Te,me,Ee,le=[];for(B=0;B<ae;B+=$)z[B+d]=z[B+u],z[B+p]=z[B+c],z[B+u]=0,z[B+c]=0;if(G.outboundAttractionDistribution){for(Ce=0,B=0;B<ae;B+=$)Ce+=z[B+g];Ce/=ae/$}if(G.barnesHutOptimize){var Be=1/0,Ue=-1/0,he=1/0,Ne=-1/0,ee,ce,_e;for(B=0;B<ae;B+=$)Be=Math.min(Be,z[B+o]),Ue=Math.max(Ue,z[B+o]),he=Math.min(he,z[B+s]),Ne=Math.max(Ne,z[B+s]);var Pe=Ue-Be,We=Ne-he;for(Pe>We?(he-=(Pe-We)/2,Ne=he+Pe):(Be-=(We-Pe)/2,Ue=Be+We),le[0+k]=-1,le[0+A]=(Be+Ue)/2,le[0+C]=(he+Ne)/2,le[0+N]=Math.max(Ue-Be,Ne-he),le[0+_]=-1,le[0+O]=-1,le[0+F]=0,le[0+L]=0,le[0+I]=0,D=1,B=0;B<ae;B+=$)for(V=0,_e=H;;)if(le[V+O]>=0){z[B+o]<le[V+A]?z[B+s]<le[V+C]?ee=le[V+O]:ee=le[V+O]+Y:z[B+s]<le[V+C]?ee=le[V+O]+Y*2:ee=le[V+O]+Y*3,le[V+L]=(le[V+L]*le[V+F]+z[B+o]*z[B+g])/(le[V+F]+z[B+g]),le[V+I]=(le[V+I]*le[V+F]+z[B+s]*z[B+g])/(le[V+F]+z[B+g]),le[V+F]+=z[B+g],V=ee;continue}else if(le[V+k]<0){le[V+k]=B;break}else{if(le[V+O]=D*Y,re=le[V+N]/2,ie=le[V+O],le[ie+k]=-1,le[ie+A]=le[V+A]-re,le[ie+C]=le[V+C]-re,le[ie+N]=re,le[ie+_]=ie+Y,le[ie+O]=-1,le[ie+F]=0,le[ie+L]=0,le[ie+I]=0,ie+=Y,le[ie+k]=-1,le[ie+A]=le[V+A]-re,le[ie+C]=le[V+C]+re,le[ie+N]=re,le[ie+_]=ie+Y,le[ie+O]=-1,le[ie+F]=0,le[ie+L]=0,le[ie+I]=0,ie+=Y,le[ie+k]=-1,le[ie+A]=le[V+A]+re,le[ie+C]=le[V+C]-re,le[ie+N]=re,le[ie+_]=ie+Y,le[ie+O]=-1,le[ie+F]=0,le[ie+L]=0,le[ie+I]=0,ie+=Y,le[ie+k]=-1,le[ie+A]=le[V+A]+re,le[ie+C]=le[V+C]+re,le[ie+N]=re,le[ie+_]=le[V+_],le[ie+O]=-1,le[ie+F]=0,le[ie+L]=0,le[ie+I]=0,D+=4,z[le[V+k]+o]<le[V+A]?z[le[V+k]+s]<le[V+C]?ee=le[V+O]:ee=le[V+O]+Y:z[le[V+k]+s]<le[V+C]?ee=le[V+O]+Y*2:ee=le[V+O]+Y*3,le[V+F]=z[le[V+k]+g],le[V+L]=z[le[V+k]+o],le[V+I]=z[le[V+k]+s],le[ee+k]=le[V+k],le[V+k]=-1,z[B+o]<le[V+A]?z[B+s]<le[V+C]?ce=le[V+O]:ce=le[V+O]+Y:z[B+s]<le[V+C]?ce=le[V+O]+Y*2:ce=le[V+O]+Y*3,ee===ce)if(_e--){V=ee;continue}else{_e=H;break}le[ce+k]=B;break}}if(G.barnesHutOptimize)for(ke=G.scalingRatio,B=0;B<ae;B+=$)for(V=0;;)if(le[V+O]>=0)if(me=Math.pow(z[B+o]-le[V+L],2)+Math.pow(z[B+s]-le[V+I],2),Se=le[V+N],4*Se*Se/me<Ie){if(J=z[B+o]-le[V+L],fe=z[B+s]-le[V+I],xe===!0?me>0?(Ee=ke*z[B+g]*le[V+F]/me,z[B+u]+=J*Ee,z[B+c]+=fe*Ee):me<0&&(Ee=-ke*z[B+g]*le[V+F]/Math.sqrt(me),z[B+u]+=J*Ee,z[B+c]+=fe*Ee):me>0&&(Ee=ke*z[B+g]*le[V+F]/me,z[B+u]+=J*Ee,z[B+c]+=fe*Ee),V=le[V+_],V<0)break;continue}else{V=le[V+O];continue}else{if(Q=le[V+k],Q>=0&&Q!==B&&(J=z[B+o]-z[Q+o],fe=z[B+s]-z[Q+s],me=J*J+fe*fe,xe===!0?me>0?(Ee=ke*z[B+g]*z[Q+g]/me,z[B+u]+=J*Ee,z[B+c]+=fe*Ee):me<0&&(Ee=-ke*z[B+g]*z[Q+g]/Math.sqrt(me),z[B+u]+=J*Ee,z[B+c]+=fe*Ee):me>0&&(Ee=ke*z[B+g]*z[Q+g]/me,z[B+u]+=J*Ee,z[B+c]+=fe*Ee)),V=le[V+_],V<0)break;continue}else for(ke=G.scalingRatio,M=0;M<ae;M+=$)for(W=0;W<M;W+=$)J=z[M+o]-z[W+o],fe=z[M+s]-z[W+s],xe===!0?(me=Math.sqrt(J*J+fe*fe)-z[M+b]-z[W+b],me>0?(Ee=ke*z[M+g]*z[W+g]/me/me,z[M+u]+=J*Ee,z[M+c]+=fe*Ee,z[W+u]-=J*Ee,z[W+c]-=fe*Ee):me<0&&(Ee=100*ke*z[M+g]*z[W+g],z[M+u]+=J*Ee,z[M+c]+=fe*Ee,z[W+u]-=J*Ee,z[W+c]-=fe*Ee)):(me=Math.sqrt(J*J+fe*fe),me>0&&(Ee=ke*z[M+g]*z[W+g]/me/me,z[M+u]+=J*Ee,z[M+c]+=fe*Ee,z[W+u]-=J*Ee,z[W+c]-=fe*Ee));for(ie=G.gravity/G.scalingRatio,ke=G.scalingRatio,B=0;B<ae;B+=$)Ee=0,J=z[B+o],fe=z[B+s],me=Math.sqrt(Math.pow(J,2)+Math.pow(fe,2)),G.strongGravityMode?me>0&&(Ee=ke*z[B+g]*ie):me>0&&(Ee=ke*z[B+g]*ie/me),z[B+u]-=J*Ee,z[B+c]-=fe*Ee;for(ke=1*(G.outboundAttractionDistribution?Ce:1),oe=0;oe<ve;oe+=U)M=K[oe+S],W=K[oe+x],re=K[oe+R],Te=Math.pow(re,G.edgeWeightInfluence),J=z[M+o]-z[W+o],fe=z[M+s]-z[W+s],xe===!0?(me=Math.sqrt(J*J+fe*fe)-z[M+b]-z[W+b],G.linLogMode?G.outboundAttractionDistribution?me>0&&(Ee=-ke*Te*Math.log(1+me)/me/z[M+g]):me>0&&(Ee=-ke*Te*Math.log(1+me)/me):G.outboundAttractionDistribution?me>0&&(Ee=-ke*Te/z[M+g]):me>0&&(Ee=-ke*Te)):(me=Math.sqrt(Math.pow(J,2)+Math.pow(fe,2)),G.linLogMode?G.outboundAttractionDistribution?me>0&&(Ee=-ke*Te*Math.log(1+me)/me/z[M+g]):me>0&&(Ee=-ke*Te*Math.log(1+me)/me):G.outboundAttractionDistribution?(me=1,Ee=-ke*Te/z[M+g]):(me=1,Ee=-ke*Te)),me>0&&(z[M+u]+=J*Ee,z[M+c]+=fe*Ee,z[W+u]-=J*Ee,z[W+c]-=fe*Ee);var yt,kt,ht,et,Tt,ot;if(xe===!0)for(B=0;B<ae;B+=$)z[B+y]!==1&&(yt=Math.sqrt(Math.pow(z[B+u],2)+Math.pow(z[B+c],2)),yt>Z&&(z[B+u]=z[B+u]*Z/yt,z[B+c]=z[B+c]*Z/yt),kt=z[B+g]*Math.sqrt((z[B+d]-z[B+u])*(z[B+d]-z[B+u])+(z[B+p]-z[B+c])*(z[B+p]-z[B+c])),ht=Math.sqrt((z[B+d]+z[B+u])*(z[B+d]+z[B+u])+(z[B+p]+z[B+c])*(z[B+p]+z[B+c]))/2,et=.1*Math.log(1+ht)/(1+Math.sqrt(kt)),Tt=z[B+o]+z[B+u]*(et/G.slowDown),z[B+o]=Tt,ot=z[B+s]+z[B+c]*(et/G.slowDown),z[B+s]=ot);else for(B=0;B<ae;B+=$)z[B+y]!==1&&(kt=z[B+g]*Math.sqrt((z[B+d]-z[B+u])*(z[B+d]-z[B+u])+(z[B+p]-z[B+c])*(z[B+p]-z[B+c])),ht=Math.sqrt((z[B+d]+z[B+u])*(z[B+d]+z[B+u])+(z[B+p]+z[B+c])*(z[B+p]+z[B+c]))/2,et=z[B+m]*Math.log(1+ht)/(1+Math.sqrt(kt)),z[B+m]=Math.min(1,Math.sqrt(et*(Math.pow(z[B+u],2)+Math.pow(z[B+c],2))/(1+Math.sqrt(kt)))),Tt=z[B+o]+z[B+u]*(et/G.slowDown),z[B+o]=Tt,ot=z[B+s]+z[B+c]*(et/G.slowDown),z[B+s]=ot);return{}}})();var a=r.exports;self.addEventListener("message",function(o){var s=o.data;t=new Float32Array(s.nodes),s.edges&&(n=new Float32Array(s.edges)),a(s.settings,t,n),self.postMessage({nodes:t.buffer},[t.buffer])})}),Pm}var zm,oO;function Qne(){if(oO)return zm;oO=1;var e=Zne(),t=ua(),n=wT().createEdgeWeightGetter,r=i5(),a=s5();function o(s,u){if(u=u||{},!t(s))throw new Error("graphology-layout-forceatlas2/worker: the given graph is not a valid graphology instance.");var c=n("getEdgeWeight"in u?u.getEdgeWeight:"weight").fromEntry,d=r.assign({},a,u.settings),p=r.validateSettings(d);if(p)throw new Error("graphology-layout-forceatlas2/worker: "+p.message);this.worker=null,this.graph=s,this.settings=d,this.getEdgeWeight=c,this.matrices=null,this.running=!1,this.killed=!1,this.outputReducer=typeof u.outputReducer=="function"?u.outputReducer:null,this.handleMessage=this.handleMessage.bind(this);var g=void 0,m=this;this.handleGraphUpdate=function(){m.worker&&m.worker.terminate(),g&&clearTimeout(g),g=setTimeout(function(){g=void 0,m.spawnWorker()},0)},s.on("nodeAdded",this.handleGraphUpdate),s.on("edgeAdded",this.handleGraphUpdate),s.on("nodeDropped",this.handleGraphUpdate),s.on("edgeDropped",this.handleGraphUpdate),this.spawnWorker()}return o.prototype.isRunning=function(){return this.running},o.prototype.spawnWorker=function(){this.worker&&this.worker.terminate(),this.worker=r.createWorker(e),this.worker.addEventListener("message",this.handleMessage),this.running&&(this.running=!1,this.start())},o.prototype.handleMessage=function(s){if(this.running){var u=new Float32Array(s.data.nodes);r.assignLayoutChanges(this.graph,u,this.outputReducer),this.outputReducer&&r.readGraphPositions(this.graph,u),this.matrices.nodes=u,this.askForIterations()}},o.prototype.askForIterations=function(s){var u=this.matrices,c={settings:this.settings,nodes:u.nodes.buffer},d=[u.nodes.buffer];return s&&(c.edges=u.edges.buffer,d.push(u.edges.buffer)),this.worker.postMessage(c,d),this},o.prototype.start=function(){if(this.killed)throw new Error("graphology-layout-forceatlas2/worker.start: layout was killed.");return this.running?this:(this.matrices=r.graphToByteArrays(this.graph,this.getEdgeWeight),this.running=!0,this.askForIterations(!0),this)},o.prototype.stop=function(){return this.running=!1,this},o.prototype.kill=function(){if(this.killed)return this;this.running=!1,this.killed=!0,this.matrices=null,this.worker.terminate(),this.graph.removeListener("nodeAdded",this.handleGraphUpdate),this.graph.removeListener("edgeAdded",this.handleGraphUpdate),this.graph.removeListener("nodeDropped",this.handleGraphUpdate),this.graph.removeListener("edgeDropped",this.handleGraphUpdate)},zm=o,zm}var Jne=Qne();const ere=un(Jne);function l5(e={iterations:100}){return Qs(Xne,e)}function tre(e={}){return ET(ere,e)}var Bm,iO;function nre(){if(iO)return Bm;iO=1;var e=0,t=1,n=2,r=3;function a(s,u){return s+"§"+u}function o(){return .01*(.5-Math.random())}return Bm=function(u,c){var d=u.margin,p=u.ratio,g=u.expansion,m=u.gridSize,b=u.speed,y,S,x,R,k,A,C=!0,N=c.length,_=N/r|0,O=new Float32Array(_),F=new Float32Array(_),L=1/0,I=1/0,H=-1/0,$=-1/0;for(y=0;y<N;y+=r)x=c[y+e],R=c[y+t],A=c[y+n]*p+d,L=Math.min(L,x-A),H=Math.max(H,x+A),I=Math.min(I,R-A),$=Math.max($,R+A);var U=H-L,Y=$-I,Z=(L+H)/2,j=(I+$)/2;L=Z-g*U/2,H=Z+g*U/2,I=j-g*Y/2,$=j+g*Y/2;var G=new Array(m*m),z=G.length,K;for(K=0;K<z;K++)G[K]=[];var D,V,B,M,W,Q,oe,re,ie,Se;for(y=0;y<N;y+=r)for(x=c[y+e],R=c[y+t],A=c[y+n]*p+d,D=x-A,V=x+A,B=R-A,M=R+A,W=Math.floor(m*(D-L)/(H-L)),Q=Math.floor(m*(V-L)/(H-L)),oe=Math.floor(m*(B-I)/($-I)),re=Math.floor(m*(M-I)/($-I)),ie=W;ie<=Q;ie++)for(Se=oe;Se<=re;Se++)G[ie*m+Se].push(y);var ae,ve=new Set,xe,Ie,Ce,ke,J,fe,Te,me,Ee,le,Be,Ue,he;for(K=0;K<z;K++)for(ae=G[K],y=0,k=ae.length;y<k;y++)for(xe=ae[y],Ce=c[xe+e],J=c[xe+t],Te=c[xe+n],S=y+1;S<k;S++)Ie=ae[S],Ee=a(xe,Ie),!(z>1&&ve.has(Ee))&&(z>1&&ve.add(Ee),ke=c[Ie+e],fe=c[Ie+t],me=c[Ie+n],le=ke-Ce,Be=fe-J,Ue=Math.sqrt(le*le+Be*Be),he=Ue<Te*p+d+(me*p+d),he&&(C=!1,Ie=Ie/r|0,Ue>0?(O[Ie]+=le/Ue*(1+Te),F[Ie]+=Be/Ue*(1+Te)):(O[Ie]+=U*o(),F[Ie]+=Y*o())));for(y=0,S=0;y<N;y+=r,S++)c[y+e]+=O[S]*.1*b,c[y+t]+=F[S]*.1*b;return{converged:C}},Bm}var ii={},sO;function u5(){if(sO)return ii;sO=1;var e=3;return ii.validateSettings=function(t){return"gridSize"in t&&typeof t.gridSize!="number"||t.gridSize<=0?{message:"the `gridSize` setting should be a positive number."}:"margin"in t&&typeof t.margin!="number"||t.margin<0?{message:"the `margin` setting should be 0 or a positive number."}:"expansion"in t&&typeof t.expansion!="number"||t.expansion<=0?{message:"the `expansion` setting should be a positive number."}:"ratio"in t&&typeof t.ratio!="number"||t.ratio<=0?{message:"the `ratio` setting should be a positive number."}:"speed"in t&&typeof t.speed!="number"||t.speed<=0?{message:"the `speed` setting should be a positive number."}:null},ii.graphToByteArray=function(t,n){var r=t.order,a=new Float32Array(r*e),o=0;return t.forEachNode(function(s,u){typeof n=="function"&&(u=n(s,u)),a[o]=u.x,a[o+1]=u.y,a[o+2]=u.size||1,o+=e}),a},ii.assignLayoutChanges=function(t,n,r){var a=0;t.forEachNode(function(o){var s={x:n[a],y:n[a+1]};typeof r=="function"&&(s=r(o,s)),t.mergeNodeAttributes(o,s),a+=e})},ii.collectLayoutChanges=function(t,n,r){var a={},o=0;return t.forEachNode(function(s){var u={x:n[o],y:n[o+1]};typeof r=="function"&&(u=r(s,u)),a[s]=u,o+=e}),a},ii.createWorker=function(n){var r=window.URL||window.webkitURL,a=n.toString(),o=r.createObjectURL(new Blob(["("+a+").call(this);"],{type:"text/javascript"})),s=new Worker(o);return r.revokeObjectURL(o),s},ii}var Um,lO;function c5(){return lO||(lO=1,Um={gridSize:20,margin:5,expansion:1.1,ratio:1,speed:3}),Um}var jm,uO;function rre(){if(uO)return jm;uO=1;var e=ua(),t=nre(),n=u5(),r=c5(),a=500;function o(u,c,d){if(!e(c))throw new Error("graphology-layout-noverlap: the given graph is not a valid graphology instance.");typeof d=="number"?d={maxIterations:d}:d=d||{};var p=d.maxIterations||a;if(typeof p!="number"||p<=0)throw new Error("graphology-layout-force: you should provide a positive number of maximum iterations.");var g=Object.assign({},r,d.settings),m=n.validateSettings(g);if(m)throw new Error("graphology-layout-noverlap: "+m.message);var b=n.graphToByteArray(c,d.inputReducer),y=!1,S;for(S=0;S<p&&!y;S++)y=t(g,b).converged;if(u){n.assignLayoutChanges(c,b,d.outputReducer);return}return n.collectLayoutChanges(c,b,d.outputReducer)}var s=o.bind(null,!1);return s.assign=o.bind(null,!0),jm=s,jm}var are=rre();const ore=un(are);var Gm,cO;function ire(){return cO||(cO=1,Gm=function(){var t,n={};(function(){var a=0,o=1,s=2,u=3;function c(p,g){return p+"§"+g}function d(){return .01*(.5-Math.random())}n.exports=function(g,m){var b=g.margin,y=g.ratio,S=g.expansion,x=g.gridSize,R=g.speed,k,A,C,N,_,O,F=!0,L=m.length,I=L/u|0,H=new Float32Array(I),$=new Float32Array(I),U=1/0,Y=1/0,Z=-1/0,j=-1/0;for(k=0;k<L;k+=u)C=m[k+a],N=m[k+o],O=m[k+s]*y+b,U=Math.min(U,C-O),Z=Math.max(Z,C+O),Y=Math.min(Y,N-O),j=Math.max(j,N+O);var G=Z-U,z=j-Y,K=(U+Z)/2,D=(Y+j)/2;U=K-S*G/2,Z=K+S*G/2,Y=D-S*z/2,j=D+S*z/2;var V=new Array(x*x),B=V.length,M;for(M=0;M<B;M++)V[M]=[];var W,Q,oe,re,ie,Se,ae,ve,xe,Ie;for(k=0;k<L;k+=u)for(C=m[k+a],N=m[k+o],O=m[k+s]*y+b,W=C-O,Q=C+O,oe=N-O,re=N+O,ie=Math.floor(x*(W-U)/(Z-U)),Se=Math.floor(x*(Q-U)/(Z-U)),ae=Math.floor(x*(oe-Y)/(j-Y)),ve=Math.floor(x*(re-Y)/(j-Y)),xe=ie;xe<=Se;xe++)for(Ie=ae;Ie<=ve;Ie++)V[xe*x+Ie].push(k);var Ce,ke=new Set,J,fe,Te,me,Ee,le,Be,Ue,he,Ne,ee,ce,_e;for(M=0;M<B;M++)for(Ce=V[M],k=0,_=Ce.length;k<_;k++)for(J=Ce[k],Te=m[J+a],Ee=m[J+o],Be=m[J+s],A=k+1;A<_;A++)fe=Ce[A],he=c(J,fe),!(B>1&&ke.has(he))&&(B>1&&ke.add(he),me=m[fe+a],le=m[fe+o],Ue=m[fe+s],Ne=me-Te,ee=le-Ee,ce=Math.sqrt(Ne*Ne+ee*ee),_e=ce<Be*y+b+(Ue*y+b),_e&&(F=!1,fe=fe/u|0,ce>0?(H[fe]+=Ne/ce*(1+Be),$[fe]+=ee/ce*(1+Be)):(H[fe]+=G*d(),$[fe]+=z*d())));for(k=0,A=0;k<L;k+=u,A++)m[k+a]+=H[A]*.1*R,m[k+o]+=$[A]*.1*R;return{converged:F}}})();var r=n.exports;self.addEventListener("message",function(a){var o=a.data;t=new Float32Array(o.nodes);var s=r(o.settings,t);self.postMessage({result:s,nodes:t.buffer},[t.buffer])})}),Gm}var Hm,dO;function sre(){if(dO)return Hm;dO=1;var e=ire(),t=ua(),n=u5(),r=c5();function a(o,s){if(s=s||{},!t(o))throw new Error("graphology-layout-noverlap/worker: the given graph is not a valid graphology instance.");var u=Object.assign({},r,s.settings),c=n.validateSettings(u);if(c)throw new Error("graphology-layout-noverlap/worker: "+c.message);this.worker=null,this.graph=o,this.settings=u,this.matrices=null,this.running=!1,this.killed=!1,this.inputReducer=s.inputReducer,this.outputReducer=s.outputReducer,this.callbacks={onConverged:typeof s.onConverged=="function"?s.onConverged:null},this.handleMessage=this.handleMessage.bind(this);var d=!1,p=this;this.handleAddition=function(){d||(d=!0,p.spawnWorker(),setTimeout(function(){d=!1},0))},o.on("nodeAdded",this.handleAddition),o.on("edgeAdded",this.handleAddition),this.spawnWorker()}return a.prototype.isRunning=function(){return this.running},a.prototype.spawnWorker=function(){this.worker&&this.worker.terminate(),this.worker=n.createWorker(e),this.worker.addEventListener("message",this.handleMessage),this.running&&(this.running=!1,this.start())},a.prototype.handleMessage=function(o){if(this.running){var s=new Float32Array(o.data.nodes);if(n.assignLayoutChanges(this.graph,s,this.outputReducer),this.matrices.nodes=s,o.data.result.converged){this.callbacks.onConverged&&this.callbacks.onConverged(),this.stop();return}this.askForIterations()}},a.prototype.askForIterations=function(){var o=this.matrices,s={settings:this.settings,nodes:o.nodes.buffer},u=[o.nodes.buffer];return this.worker.postMessage(s,u),this},a.prototype.start=function(){if(this.killed)throw new Error("graphology-layout-noverlap/worker.start: layout was killed.");return this.running?this:(this.matrices={nodes:n.graphToByteArray(this.graph,this.inputReducer)},this.running=!0,this.askForIterations(),this)},a.prototype.stop=function(){return this.running=!1,this},a.prototype.kill=function(){if(this.killed)return this;this.running=!1,this.killed=!0,this.matrices=null,this.worker.terminate(),this.graph.removeListener("nodeAdded",this.handleAddition),this.graph.removeListener("edgeAdded",this.handleAddition)},Hm=a,Hm}var lre=sre();const ure=un(lre);function cre(e={}){return Qs(ore,e)}function dre(e={}){return ET(ure,e)}var $m,fO;function fre(){if(fO)return $m;fO=1;var e=tc(),t=ua(),n={dimensions:["x","y"],center:.5,rng:Math.random,scale:1};function r(o,s,u){if(!t(s))throw new Error("graphology-layout/random: the given graph is not a valid graphology instance.");u=e(u,n);var c=u.dimensions;if(!Array.isArray(c)||c.length<1)throw new Error("graphology-layout/random: given dimensions are invalid.");var d=c.length,p=u.center,g=u.rng,m=u.scale,b=(p-.5)*m;function y(x){for(var R=0;R<d;R++)x[c[R]]=g()*m+b;return x}if(!o){var S={};return s.forEachNode(function(x){S[x]=y({})}),S}s.updateEachNodeAttributes(function(x,R){return y(R),R},{attributes:c})}var a=r.bind(null,!1);return a.assign=r.bind(null,!0),$m=a,$m}var pre=fre();const gre=un(pre);function hre(e={}){return Qs(gre,e)}var pO=1,mre=.9,bre=.8,yre=.17,qm=.1,Vm=.999,vre=.9999,Sre=.99,Ere=/[\\\/_+.#"@\[\(\{&]/,wre=/[\\\/_+.#"@\[\(\{&]/g,xre=/[\s-]/,d5=/[\s-]/g;function Vk(e,t,n,r,a,o,s){if(o===t.length)return a===e.length?pO:Sre;var u=`${a},${o}`;if(s[u]!==void 0)return s[u];for(var c=r.charAt(o),d=n.indexOf(c,a),p=0,g,m,b,y;d>=0;)g=Vk(e,t,n,r,d+1,o+1,s),g>p&&(d===a?g*=pO:Ere.test(e.charAt(d-1))?(g*=bre,b=e.slice(a,d-1).match(wre),b&&a>0&&(g*=Math.pow(Vm,b.length))):xre.test(e.charAt(d-1))?(g*=mre,y=e.slice(a,d-1).match(d5),y&&a>0&&(g*=Math.pow(Vm,y.length))):(g*=yre,a>0&&(g*=Math.pow(Vm,d-a))),e.charAt(d)!==t.charAt(o)&&(g*=vre)),(g<qm&&n.charAt(d-1)===r.charAt(o+1)||r.charAt(o+1)===r.charAt(o)&&n.charAt(d-1)!==r.charAt(o))&&(m=Vk(e,t,n,r,d+1,o+2,s),m*qm>g&&(g=m*qm)),g>p&&(p=g),d=n.indexOf(c,d+1);return s[u]=p,p}function gO(e){return e.toLowerCase().replace(d5," ")}function kre(e,t,n){return e=n&&n.length>0?`${e+" "+n.join(" ")}`:e,Vk(e,t,gO(e),gO(t),0,0,{})}var Wm={exports:{}},Ym={};/**
 * @license React
 * use-sync-external-store-shim.production.js
 *
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */var hO;function Tre(){if(hO)return Ym;hO=1;var e=$f();function t(g,m){return g===m&&(g!==0||1/g===1/m)||g!==g&&m!==m}var n=typeof Object.is=="function"?Object.is:t,r=e.useState,a=e.useEffect,o=e.useLayoutEffect,s=e.useDebugValue;function u(g,m){var b=m(),y=r({inst:{value:b,getSnapshot:m}}),S=y[0].inst,x=y[1];return o(function(){S.value=b,S.getSnapshot=m,c(S)&&x({inst:S})},[g,b,m]),a(function(){return c(S)&&x({inst:S}),g(function(){c(S)&&x({inst:S})})},[g]),s(b),b}function c(g){var m=g.getSnapshot;g=g.value;try{var b=m();return!n(g,b)}catch{return!0}}function d(g,m){return m()}var p=typeof window>"u"||typeof window.document>"u"||typeof window.document.createElement>"u"?d:u;return Ym.useSyncExternalStore=e.useSyncExternalStore!==void 0?e.useSyncExternalStore:p,Ym}var mO;function Are(){return mO||(mO=1,Wm.exports=Tre()),Wm.exports}var Rre=Are(),lu='[cmdk-group=""]',Km='[cmdk-group-items=""]',_re='[cmdk-group-heading=""]',xT='[cmdk-item=""]',bO=`${xT}:not([aria-disabled="true"])`,Wk="cmdk-item-select",li="data-value",Cre=(e,t,n)=>kre(e,t,n),f5=E.createContext(void 0),nc=()=>E.useContext(f5),p5=E.createContext(void 0),kT=()=>E.useContext(p5),g5=E.createContext(void 0),h5=E.forwardRef((e,t)=>{let n=vs(()=>{var B,M;return{search:"",value:(M=(B=e.value)!=null?B:e.defaultValue)!=null?M:"",filtered:{count:0,items:new Map,groups:new Set}}}),r=vs(()=>new Set),a=vs(()=>new Map),o=vs(()=>new Map),s=vs(()=>new Set),u=m5(e),{label:c,children:d,value:p,onValueChange:g,filter:m,shouldFilter:b,loop:y,disablePointerSelection:S=!1,vimBindings:x=!0,...R}=e,k=Tn(),A=Tn(),C=Tn(),N=E.useRef(null),_=Ure();yi(()=>{if(p!==void 0){let B=p.trim();n.current.value=B,O.emit()}},[p]),yi(()=>{_(6,U)},[]);let O=E.useMemo(()=>({subscribe:B=>(s.current.add(B),()=>s.current.delete(B)),snapshot:()=>n.current,setState:(B,M,W)=>{var Q,oe,re;if(!Object.is(n.current[B],M)){if(n.current[B]=M,B==="search")$(),I(),_(1,H);else if(B==="value"&&(W||_(5,U),((Q=u.current)==null?void 0:Q.value)!==void 0)){let ie=M??"";(re=(oe=u.current).onValueChange)==null||re.call(oe,ie);return}O.emit()}},emit:()=>{s.current.forEach(B=>B())}}),[]),F=E.useMemo(()=>({value:(B,M,W)=>{var Q;M!==((Q=o.current.get(B))==null?void 0:Q.value)&&(o.current.set(B,{value:M,keywords:W}),n.current.filtered.items.set(B,L(M,W)),_(2,()=>{I(),O.emit()}))},item:(B,M)=>(r.current.add(B),M&&(a.current.has(M)?a.current.get(M).add(B):a.current.set(M,new Set([B]))),_(3,()=>{$(),I(),n.current.value||H(),O.emit()}),()=>{o.current.delete(B),r.current.delete(B),n.current.filtered.items.delete(B);let W=Y();_(4,()=>{$(),(W==null?void 0:W.getAttribute("id"))===B&&H(),O.emit()})}),group:B=>(a.current.has(B)||a.current.set(B,new Set),()=>{o.current.delete(B),a.current.delete(B)}),filter:()=>u.current.shouldFilter,label:c||e["aria-label"],getDisablePointerSelection:()=>u.current.disablePointerSelection,listId:k,inputId:C,labelId:A,listInnerRef:N}),[]);function L(B,M){var W,Q;let oe=(Q=(W=u.current)==null?void 0:W.filter)!=null?Q:Cre;return B?oe(B,n.current.search,M):0}function I(){if(!n.current.search||u.current.shouldFilter===!1)return;let B=n.current.filtered.items,M=[];n.current.filtered.groups.forEach(Q=>{let oe=a.current.get(Q),re=0;oe.forEach(ie=>{let Se=B.get(ie);re=Math.max(Se,re)}),M.push([Q,re])});let W=N.current;Z().sort((Q,oe)=>{var re,ie;let Se=Q.getAttribute("id"),ae=oe.getAttribute("id");return((re=B.get(ae))!=null?re:0)-((ie=B.get(Se))!=null?ie:0)}).forEach(Q=>{let oe=Q.closest(Km);oe?oe.appendChild(Q.parentElement===oe?Q:Q.closest(`${Km} > *`)):W.appendChild(Q.parentElement===W?Q:Q.closest(`${Km} > *`))}),M.sort((Q,oe)=>oe[1]-Q[1]).forEach(Q=>{var oe;let re=(oe=N.current)==null?void 0:oe.querySelector(`${lu}[${li}="${encodeURIComponent(Q[0])}"]`);re==null||re.parentElement.appendChild(re)})}function H(){let B=Z().find(W=>W.getAttribute("aria-disabled")!=="true"),M=B==null?void 0:B.getAttribute(li);O.setState("value",M||void 0)}function $(){var B,M,W,Q;if(!n.current.search||u.current.shouldFilter===!1){n.current.filtered.count=r.current.size;return}n.current.filtered.groups=new Set;let oe=0;for(let re of r.current){let ie=(M=(B=o.current.get(re))==null?void 0:B.value)!=null?M:"",Se=(Q=(W=o.current.get(re))==null?void 0:W.keywords)!=null?Q:[],ae=L(ie,Se);n.current.filtered.items.set(re,ae),ae>0&&oe++}for(let[re,ie]of a.current)for(let Se of ie)if(n.current.filtered.items.get(Se)>0){n.current.filtered.groups.add(re);break}n.current.filtered.count=oe}function U(){var B,M,W;let Q=Y();Q&&(((B=Q.parentElement)==null?void 0:B.firstChild)===Q&&((W=(M=Q.closest(lu))==null?void 0:M.querySelector(_re))==null||W.scrollIntoView({block:"nearest"})),Q.scrollIntoView({block:"nearest"}))}function Y(){var B;return(B=N.current)==null?void 0:B.querySelector(`${xT}[aria-selected="true"]`)}function Z(){var B;return Array.from(((B=N.current)==null?void 0:B.querySelectorAll(bO))||[])}function j(B){let M=Z()[B];M&&O.setState("value",M.getAttribute(li))}function G(B){var M;let W=Y(),Q=Z(),oe=Q.findIndex(ie=>ie===W),re=Q[oe+B];(M=u.current)!=null&&M.loop&&(re=oe+B<0?Q[Q.length-1]:oe+B===Q.length?Q[0]:Q[oe+B]),re&&O.setState("value",re.getAttribute(li))}function z(B){let M=Y(),W=M==null?void 0:M.closest(lu),Q;for(;W&&!Q;)W=B>0?zre(W,lu):Bre(W,lu),Q=W==null?void 0:W.querySelector(bO);Q?O.setState("value",Q.getAttribute(li)):G(B)}let K=()=>j(Z().length-1),D=B=>{B.preventDefault(),B.metaKey?K():B.altKey?z(1):G(1)},V=B=>{B.preventDefault(),B.metaKey?j(0):B.altKey?z(-1):G(-1)};return E.createElement(Ze.div,{ref:t,tabIndex:-1,...R,"cmdk-root":"",onKeyDown:B=>{var M;if((M=R.onKeyDown)==null||M.call(R,B),!B.defaultPrevented)switch(B.key){case"n":case"j":{x&&B.ctrlKey&&D(B);break}case"ArrowDown":{D(B);break}case"p":case"k":{x&&B.ctrlKey&&V(B);break}case"ArrowUp":{V(B);break}case"Home":{B.preventDefault(),j(0);break}case"End":{B.preventDefault(),K();break}case"Enter":if(!B.nativeEvent.isComposing&&B.keyCode!==229){B.preventDefault();let W=Y();if(W){let Q=new Event(Wk);W.dispatchEvent(Q)}}}}},E.createElement("label",{"cmdk-label":"",htmlFor:F.inputId,id:F.labelId,style:Gre},c),wp(e,B=>E.createElement(p5.Provider,{value:O},E.createElement(f5.Provider,{value:F},B))))}),Nre=E.forwardRef((e,t)=>{var n,r;let a=Tn(),o=E.useRef(null),s=E.useContext(g5),u=nc(),c=m5(e),d=(r=(n=c.current)==null?void 0:n.forceMount)!=null?r:s==null?void 0:s.forceMount;yi(()=>{if(!d)return u.item(a,s==null?void 0:s.id)},[d]);let p=b5(a,o,[e.value,e.children,o],e.keywords),g=kT(),m=vi(_=>_.value&&_.value===p.current),b=vi(_=>d||u.filter()===!1?!0:_.search?_.filtered.items.get(a)>0:!0);E.useEffect(()=>{let _=o.current;if(!(!_||e.disabled))return _.addEventListener(Wk,y),()=>_.removeEventListener(Wk,y)},[b,e.onSelect,e.disabled]);function y(){var _,O;S(),(O=(_=c.current).onSelect)==null||O.call(_,p.current)}function S(){g.setState("value",p.current,!0)}if(!b)return null;let{disabled:x,value:R,onSelect:k,forceMount:A,keywords:C,...N}=e;return E.createElement(Ze.div,{ref:_u([o,t]),...N,id:a,"cmdk-item":"",role:"option","aria-disabled":!!x,"aria-selected":!!m,"data-disabled":!!x,"data-selected":!!m,onPointerMove:x||u.getDisablePointerSelection()?void 0:S,onClick:x?void 0:y},e.children)}),Ore=E.forwardRef((e,t)=>{let{heading:n,children:r,forceMount:a,...o}=e,s=Tn(),u=E.useRef(null),c=E.useRef(null),d=Tn(),p=nc(),g=vi(b=>a||p.filter()===!1?!0:b.search?b.filtered.groups.has(s):!0);yi(()=>p.group(s),[]),b5(s,u,[e.value,e.heading,c]);let m=E.useMemo(()=>({id:s,forceMount:a}),[a]);return E.createElement(Ze.div,{ref:_u([u,t]),...o,"cmdk-group":"",role:"presentation",hidden:g?void 0:!0},n&&E.createElement("div",{ref:c,"cmdk-group-heading":"","aria-hidden":!0,id:d},n),wp(e,b=>E.createElement("div",{"cmdk-group-items":"",role:"group","aria-labelledby":n?d:void 0},E.createElement(g5.Provider,{value:m},b))))}),Dre=E.forwardRef((e,t)=>{let{alwaysRender:n,...r}=e,a=E.useRef(null),o=vi(s=>!s.search);return!n&&!o?null:E.createElement(Ze.div,{ref:_u([a,t]),...r,"cmdk-separator":"",role:"separator"})}),Ire=E.forwardRef((e,t)=>{let{onValueChange:n,...r}=e,a=e.value!=null,o=kT(),s=vi(p=>p.search),u=vi(p=>p.value),c=nc(),d=E.useMemo(()=>{var p;let g=(p=c.listInnerRef.current)==null?void 0:p.querySelector(`${xT}[${li}="${encodeURIComponent(u)}"]`);return g==null?void 0:g.getAttribute("id")},[]);return E.useEffect(()=>{e.value!=null&&o.setState("search",e.value)},[e.value]),E.createElement(Ze.input,{ref:t,...r,"cmdk-input":"",autoComplete:"off",autoCorrect:"off",spellCheck:!1,"aria-autocomplete":"list",role:"combobox","aria-expanded":!0,"aria-controls":c.listId,"aria-labelledby":c.labelId,"aria-activedescendant":d,id:c.inputId,type:"text",value:a?e.value:s,onChange:p=>{a||o.setState("search",p.target.value),n==null||n(p.target.value)}})}),Lre=E.forwardRef((e,t)=>{let{children:n,label:r="Suggestions",...a}=e,o=E.useRef(null),s=E.useRef(null),u=nc();return E.useEffect(()=>{if(s.current&&o.current){let c=s.current,d=o.current,p,g=new ResizeObserver(()=>{p=requestAnimationFrame(()=>{let m=c.offsetHeight;d.style.setProperty("--cmdk-list-height",m.toFixed(1)+"px")})});return g.observe(c),()=>{cancelAnimationFrame(p),g.unobserve(c)}}},[]),E.createElement(Ze.div,{ref:_u([o,t]),...a,"cmdk-list":"",role:"listbox","aria-label":r,id:u.listId},wp(e,c=>E.createElement("div",{ref:_u([s,u.listInnerRef]),"cmdk-list-sizer":""},c)))}),Mre=E.forwardRef((e,t)=>{let{open:n,onOpenChange:r,overlayClassName:a,contentClassName:o,container:s,...u}=e;return E.createElement(j0,{open:n,onOpenChange:r},E.createElement(G0,{container:s},E.createElement(ap,{"cmdk-overlay":"",className:a}),E.createElement(op,{"aria-label":e.label,"cmdk-dialog":"",className:o},E.createElement(h5,{ref:t,...u}))))}),Fre=E.forwardRef((e,t)=>vi(n=>n.filtered.count===0)?E.createElement(Ze.div,{ref:t,...e,"cmdk-empty":"",role:"presentation"}):null),Pre=E.forwardRef((e,t)=>{let{progress:n,children:r,label:a="Loading...",...o}=e;return E.createElement(Ze.div,{ref:t,...o,"cmdk-loading":"",role:"progressbar","aria-valuenow":n,"aria-valuemin":0,"aria-valuemax":100,"aria-label":a},wp(e,s=>E.createElement("div",{"aria-hidden":!0},s)))}),Vn=Object.assign(h5,{List:Lre,Item:Nre,Input:Ire,Group:Ore,Separator:Dre,Dialog:Mre,Empty:Fre,Loading:Pre});function zre(e,t){let n=e.nextElementSibling;for(;n;){if(n.matches(t))return n;n=n.nextElementSibling}}function Bre(e,t){let n=e.previousElementSibling;for(;n;){if(n.matches(t))return n;n=n.previousElementSibling}}function m5(e){let t=E.useRef(e);return yi(()=>{t.current=e}),t}var yi=typeof window>"u"?E.useEffect:E.useLayoutEffect;function vs(e){let t=E.useRef();return t.current===void 0&&(t.current=e()),t}function _u(e){return t=>{e.forEach(n=>{typeof n=="function"?n(t):n!=null&&(n.current=t)})}}function vi(e){let t=kT(),n=()=>e(t.snapshot());return Rre.useSyncExternalStore(t.subscribe,n,n)}function b5(e,t,n,r=[]){let a=E.useRef(),o=nc();return yi(()=>{var s;let u=(()=>{var d;for(let p of n){if(typeof p=="string")return p.trim();if(typeof p=="object"&&"current"in p)return p.current?(d=p.current.textContent)==null?void 0:d.trim():a.current}})(),c=r.map(d=>d.trim());o.value(e,u,c),(s=t.current)==null||s.setAttribute(li,u),a.current=u}),a}var Ure=()=>{let[e,t]=E.useState(),n=vs(()=>new Map);return yi(()=>{n.current.forEach(r=>r()),n.current=new Map},[e]),(r,a)=>{n.current.set(r,a),t({})}};function jre(e){let t=e.type;return typeof t=="function"?t(e.props):"render"in t?t.render(e.props):e}function wp({asChild:e,children:t},n){return e&&E.isValidElement(t)?E.cloneElement(jre(t),{ref:t.ref},n(t.props.children)):n(t)}var Gre={position:"absolute",width:"1px",height:"1px",padding:"0",margin:"-1px",overflow:"hidden",clip:"rect(0, 0, 0, 0)",whiteSpace:"nowrap",borderWidth:"0"};const y5=j0,v5=LU,Hre=G0,S5=E.forwardRef(({className:e,...t},n)=>w.jsx(ap,{ref:n,className:Me("data-[state=open]:animate-in data-[state=closed]:animate-out data-[state=closed]:fade-out-0 data-[state=open]:fade-in-0 fixed inset-0 z-50 bg-black/80",e),...t}));S5.displayName=ap.displayName;const TT=E.forwardRef(({className:e,children:t,...n},r)=>w.jsxs(Hre,{children:[w.jsx(S5,{}),w.jsxs(op,{ref:r,className:Me("bg-background data-[state=open]:animate-in data-[state=closed]:animate-out data-[state=closed]:fade-out-0 data-[state=open]:fade-in-0 data-[state=closed]:zoom-out-95 data-[state=open]:zoom-in-95 data-[state=closed]:slide-out-to-top-[48%] data-[state=open]:slide-in-from-top-[48%] fixed top-[50%] left-[50%] z-50 grid w-full max-w-lg translate-x-[-50%] translate-y-[-50%] gap-4 border p-6 shadow-lg duration-200 sm:rounded-lg",e),...n,children:[t,w.jsxs(q0,{className:"ring-offset-background focus:ring-ring data-[state=open]:bg-accent data-[state=open]:text-muted-foreground absolute top-4 right-4 rounded-sm opacity-70 transition-opacity hover:opacity-100 focus:ring-2 focus:ring-offset-2 focus:outline-none disabled:pointer-events-none",children:[w.jsx(rU,{className:"h-4 w-4"}),w.jsx("span",{className:"sr-only",children:"Close"})]})]})]}));TT.displayName=op.displayName;const AT=({className:e,...t})=>w.jsx("div",{className:Me("flex flex-col space-y-1.5 text-center sm:text-left",e),...t});AT.displayName="DialogHeader";const RT=E.forwardRef(({className:e,...t},n)=>w.jsx(H0,{ref:n,className:Me("text-lg leading-none font-semibold tracking-tight",e),...t}));RT.displayName=H0.displayName;const _T=E.forwardRef(({className:e,...t},n)=>w.jsx($0,{ref:n,className:Me("text-muted-foreground text-sm",e),...t}));_T.displayName=$0.displayName;const xp=E.forwardRef(({className:e,...t},n)=>w.jsx(Vn,{ref:n,className:Me("bg-popover text-popover-foreground flex h-full w-full flex-col overflow-hidden rounded-md",e),...t}));xp.displayName=Vn.displayName;const CT=E.forwardRef(({className:e,...t},n)=>w.jsxs("div",{className:"flex items-center border-b px-3","cmdk-input-wrapper":"",children:[w.jsx(dY,{className:"mr-2 h-4 w-4 shrink-0 opacity-50"}),w.jsx(Vn.Input,{ref:n,className:Me("placeholder:text-muted-foreground flex h-11 w-full rounded-md bg-transparent py-3 text-sm outline-none disabled:cursor-not-allowed disabled:opacity-50",e),...t})]}));CT.displayName=Vn.Input.displayName;const kp=E.forwardRef(({className:e,...t},n)=>w.jsx(Vn.List,{ref:n,className:Me("max-h-[300px] overflow-x-hidden overflow-y-auto",e),...t}));kp.displayName=Vn.List.displayName;const NT=E.forwardRef((e,t)=>w.jsx(Vn.Empty,{ref:t,className:"py-6 text-center text-sm",...e}));NT.displayName=Vn.Empty.displayName;const Js=E.forwardRef(({className:e,...t},n)=>w.jsx(Vn.Group,{ref:n,className:Me("text-foreground [&_[cmdk-group-heading]]:text-muted-foreground overflow-hidden p-1 [&_[cmdk-group-heading]]:px-2 [&_[cmdk-group-heading]]:py-1.5 [&_[cmdk-group-heading]]:text-xs [&_[cmdk-group-heading]]:font-medium",e),...t}));Js.displayName=Vn.Group.displayName;const $re=E.forwardRef(({className:e,...t},n)=>w.jsx(Vn.Separator,{ref:n,className:Me("bg-border -mx-1 h-px",e),...t}));$re.displayName=Vn.Separator.displayName;const el=E.forwardRef(({className:e,...t},n)=>w.jsx(Vn.Item,{ref:n,className:Me("data-[selected='true']:bg-accent data-[selected=true]:text-accent-foreground relative flex cursor-default items-center gap-2 rounded-sm px-2 py-1.5 text-sm outline-none select-none data-[disabled=true]:pointer-events-none data-[disabled=true]:opacity-50 [&_svg]:pointer-events-none [&_svg]:size-4 [&_svg]:shrink-0",e),...t}));el.displayName=Vn.Item.displayName;const qre=({layout:e,autoRunFor:t,mainLayout:n})=>{const r=kr(),[a,o]=E.useState(!1),s=E.useRef(null),{t:u}=Pt(),c=E.useCallback(()=>{if(r)try{const p=r.getGraph();if(!p||p.order===0)return;const g=n.positions();F4(p,g,{duration:300})}catch(p){console.error("Error updating positions:",p),s.current&&(window.clearInterval(s.current),s.current=null,o(!1))}},[r,n]),d=E.useCallback(()=>{if(a){console.log("Stopping layout animation"),s.current&&(window.clearInterval(s.current),s.current=null);try{typeof e.kill=="function"?(e.kill(),console.log("Layout algorithm killed")):typeof e.stop=="function"&&(e.stop(),console.log("Layout algorithm stopped"))}catch(p){console.error("Error stopping layout algorithm:",p)}o(!1)}else console.log("Starting layout animation"),c(),s.current=window.setInterval(()=>{c()},200),o(!0),setTimeout(()=>{if(s.current){console.log("Auto-stopping layout animation after 3 seconds"),window.clearInterval(s.current),s.current=null,o(!1);try{typeof e.kill=="function"?e.kill():typeof e.stop=="function"&&e.stop()}catch(p){console.error("Error stopping layout algorithm:",p)}}},3e3)},[a,e,c]);return E.useEffect(()=>{if(!r){console.log("No sigma instance available");return}let p=null;return t!==void 0&&t>-1&&r.getGraph().order>0&&(console.log("Auto-starting layout animation"),c(),s.current=window.setInterval(()=>{c()},200),o(!0),t>0&&(p=window.setTimeout(()=>{console.log("Auto-stopping layout animation after timeout"),s.current&&(window.clearInterval(s.current),s.current=null),o(!1)},t))),()=>{s.current&&(window.clearInterval(s.current),s.current=null),p&&window.clearTimeout(p),o(!1)}},[t,r,c]),w.jsx(xt,{size:"icon",onClick:d,tooltip:u(a?"graphPanel.sideBar.layoutsControl.stopAnimation":"graphPanel.sideBar.layoutsControl.startAnimation"),variant:Fr,children:a?w.jsx(eY,{}):w.jsx(nY,{})})},Vre=()=>{const e=kr(),{t}=Pt(),[n,r]=E.useState("Circular"),[a,o]=E.useState(!1),s=Fe.use.graphLayoutMaxIterations(),u=zne(),c=Lne(),d=hre(),p=cre({maxIterations:s,settings:{margin:5,expansion:1.1,gridSize:1,ratio:1,speed:3}}),g=qne({maxIterations:s,settings:{attraction:3e-4,repulsion:.05,gravity:.01,inertia:.4,maxMove:100}}),m=l5({iterations:s}),b=dre(),y=Vne(),S=tre(),x=E.useMemo(()=>({Circular:{layout:u},Circlepack:{layout:c},Random:{layout:d},Noverlaps:{layout:p,worker:b},"Force Directed":{layout:g,worker:y},"Force Atlas":{layout:m,worker:S}}),[c,u,g,m,p,d,y,b,S]),R=E.useCallback(k=>{console.debug("Running layout:",k);const{positions:A}=x[k].layout;try{const C=e.getGraph();if(!C){console.error("No graph available");return}const N=A();console.log("Positions calculated, animating nodes"),F4(C,N,{duration:400}),r(k)}catch(C){console.error("Error running layout:",C)}},[x,e]);return w.jsxs(w.Fragment,{children:[w.jsx("div",{children:x[n]&&"worker"in x[n]&&w.jsx(qre,{layout:x[n].worker,mainLayout:x[n].layout})}),w.jsx("div",{children:w.jsxs(Yu,{open:a,onOpenChange:o,children:[w.jsx(Ku,{asChild:!0,children:w.jsx(xt,{size:"icon",variant:Fr,onClick:()=>o(k=>!k),tooltip:t("graphPanel.sideBar.layoutsControl.layoutGraph"),children:w.jsx(jW,{})})}),w.jsx(Ws,{side:"right",align:"center",className:"p-1",children:w.jsx(xp,{children:w.jsx(kp,{children:w.jsx(Js,{children:Object.keys(x).map(k=>w.jsx(el,{onSelect:()=>{R(k)},className:"cursor-pointer text-xs",children:t(`graphPanel.sideBar.layoutsControl.layouts.${k}`)},k))})})})})]})})]})},E5=()=>{const e=E.useContext(qB);if(e===void 0)throw new Error("useTheme must be used within a ThemeProvider");return e},Id=e=>!!(e.type.startsWith("mouse")&&e.buttons!==0),Wre=({disableHoverEffect:e})=>{const t=kr(),n=j4(),r=U4(),a=Fe.use.graphLayoutMaxIterations(),{assign:o}=l5({iterations:a}),{theme:s}=E5(),u=Fe.use.enableHideUnselectedEdges(),c=Fe.use.enableEdgeEvents(),d=Fe.use.showEdgeLabel(),p=Fe.use.showNodeLabel(),g=ze.use.selectedNode(),m=ze.use.focusedNode(),b=ze.use.selectedEdge(),y=ze.use.focusedEdge(),S=ze.use.sigmaGraph();return E.useEffect(()=>{if(S&&t){try{typeof t.setGraph=="function"?(t.setGraph(S),console.log("Binding graph to sigma instance")):(t.graph=S,console.warn("Simgma missing setGraph function, set graph property directly"))}catch(x){console.error("Error setting graph on sigma instance:",x)}o(),console.log("Initial layout applied to graph")}},[t,S,o,a]),E.useEffect(()=>{t&&(ze.getState().sigmaInstance||(console.log("Setting sigma instance from GraphControl"),ze.getState().setSigmaInstance(t)))},[t]),E.useEffect(()=>{const{setFocusedNode:x,setSelectedNode:R,setFocusedEdge:k,setSelectedEdge:A,clearSelection:C}=ze.getState(),N={enterNode:_=>{Id(_.event.original)||x(_.node)},leaveNode:_=>{Id(_.event.original)||x(null)},clickNode:_=>{R(_.node),A(null)},clickStage:()=>C()};c&&(N.clickEdge=_=>{A(_.edge),R(null)},N.enterEdge=_=>{Id(_.event.original)||k(_.edge)},N.leaveEdge=_=>{Id(_.event.original)||k(null)}),n(N)},[n,c]),E.useEffect(()=>{const x=s==="dark",R=x?dV:void 0,k=x?hV:void 0;r({enableEdgeEvents:c,renderEdgeLabels:d,renderLabels:p,nodeReducer:(A,C)=>{const N=t.getGraph(),_={...C,highlighted:C.highlighted||!1,labelColor:R};if(!e){_.highlighted=!1;const O=m||g,F=y||b;if(O&&N.hasNode(O))try{(A===O||N.neighbors(O).includes(A))&&(_.highlighted=!0,A===g&&(_.borderColor=gV))}catch(L){console.error("Error in nodeReducer:",L)}else if(F&&N.hasEdge(F))N.extremities(F).includes(A)&&(_.highlighted=!0,_.size=3);else return _;_.highlighted?x&&(_.labelColor=fV):_.color=pV}return _},edgeReducer:(A,C)=>{const N=t.getGraph(),_={...C,hidden:!1,labelColor:R,color:k};if(!e){const O=m||g;if(O&&N.hasNode(O))try{u?N.extremities(A).includes(O)||(_.hidden=!0):N.extremities(A).includes(O)&&(_.color=AC)}catch(F){console.error("Error in edgeReducer:",F)}else{const F=b&&N.hasEdge(b)?b:null,L=y&&N.hasEdge(y)?y:null;(F||L)&&(A===F?_.color=mV:A===L?_.color=AC:u&&(_.hidden=!0))}}return _}})},[g,m,b,y,r,t,e,s,u,c,d,p]),null},Yre=()=>{const{zoomIn:e,zoomOut:t,reset:n}=G4({duration:200,factor:1.5}),r=kr(),{t:a}=Pt(),o=E.useCallback(()=>e(),[e]),s=E.useCallback(()=>t(),[t]),u=E.useCallback(()=>{if(r)try{r.setCustomBBox(null),r.refresh();const p=r.getGraph();if(!(p!=null&&p.order)||p.nodes().length===0){n();return}r.getCamera().animate({x:.5,y:.5,ratio:1.1},{duration:1e3})}catch(p){console.error("Error resetting zoom:",p),n()}},[r,n]),c=E.useCallback(()=>{if(!r)return;const p=r.getCamera(),m=p.angle+Math.PI/8;p.animate({angle:m},{duration:200})},[r]),d=E.useCallback(()=>{if(!r)return;const p=r.getCamera(),m=p.angle-Math.PI/8;p.animate({angle:m},{duration:200})},[r]);return w.jsxs(w.Fragment,{children:[w.jsx(xt,{variant:Fr,onClick:d,tooltip:a("graphPanel.sideBar.zoomControl.rotateCameraCounterClockwise"),size:"icon",children:w.jsx(oY,{})}),w.jsx(xt,{variant:Fr,onClick:c,tooltip:a("graphPanel.sideBar.zoomControl.rotateCamera"),size:"icon",children:w.jsx(sY,{})}),w.jsx(xt,{variant:Fr,onClick:u,tooltip:a("graphPanel.sideBar.zoomControl.resetZoom"),size:"icon",children:w.jsx(MW,{})}),w.jsx(xt,{variant:Fr,onClick:o,tooltip:a("graphPanel.sideBar.zoomControl.zoomIn"),size:"icon",children:w.jsx(SY,{})}),w.jsx(xt,{variant:Fr,onClick:s,tooltip:a("graphPanel.sideBar.zoomControl.zoomOut"),size:"icon",children:w.jsx(wY,{})})]})},Kre=()=>{const{isFullScreen:e,toggle:t}=Nte(),{t:n}=Pt();return w.jsx(w.Fragment,{children:e?w.jsx(xt,{variant:Fr,onClick:t,tooltip:n("graphPanel.sideBar.fullScreenControl.windowed"),size:"icon",children:w.jsx(XW,{})}):w.jsx(xt,{variant:Fr,onClick:t,tooltip:n("graphPanel.sideBar.fullScreenControl.fullScreen"),size:"icon",children:w.jsx(YW,{})})})};var OT="Checkbox",[Xre,Fxe]=wr(OT),[Zre,Qre]=Xre(OT),w5=E.forwardRef((e,t)=>{const{__scopeCheckbox:n,name:r,checked:a,defaultChecked:o,required:s,disabled:u,value:c="on",onCheckedChange:d,form:p,...g}=e,[m,b]=E.useState(null),y=gt(t,C=>b(C)),S=E.useRef(!1),x=m?p||!!m.closest("form"):!0,[R=!1,k]=ja({prop:a,defaultProp:o,onChange:d}),A=E.useRef(R);return E.useEffect(()=>{const C=m==null?void 0:m.form;if(C){const N=()=>k(A.current);return C.addEventListener("reset",N),()=>C.removeEventListener("reset",N)}},[m,k]),w.jsxs(Zre,{scope:n,state:R,disabled:u,children:[w.jsx(Ze.button,{type:"button",role:"checkbox","aria-checked":Ro(R)?"mixed":R,"aria-required":s,"data-state":T5(R),"data-disabled":u?"":void 0,disabled:u,value:c,...g,ref:y,onKeyDown:Ke(e.onKeyDown,C=>{C.key==="Enter"&&C.preventDefault()}),onClick:Ke(e.onClick,C=>{k(N=>Ro(N)?!0:!N),x&&(S.current=C.isPropagationStopped(),S.current||C.stopPropagation())})}),x&&w.jsx(Jre,{control:m,bubbles:!S.current,name:r,value:c,checked:R,required:s,disabled:u,form:p,style:{transform:"translateX(-100%)"},defaultChecked:Ro(o)?!1:o})]})});w5.displayName=OT;var x5="CheckboxIndicator",k5=E.forwardRef((e,t)=>{const{__scopeCheckbox:n,forceMount:r,...a}=e,o=Qre(x5,n);return w.jsx(xr,{present:r||Ro(o.state)||o.state===!0,children:w.jsx(Ze.span,{"data-state":T5(o.state),"data-disabled":o.disabled?"":void 0,...a,ref:t,style:{pointerEvents:"none",...e.style}})})});k5.displayName=x5;var Jre=e=>{const{control:t,checked:n,bubbles:r=!0,defaultChecked:a,...o}=e,s=E.useRef(null),u=X3(n),c=c3(t);E.useEffect(()=>{const p=s.current,g=window.HTMLInputElement.prototype,b=Object.getOwnPropertyDescriptor(g,"checked").set;if(u!==n&&b){const y=new Event("click",{bubbles:r});p.indeterminate=Ro(n),b.call(p,Ro(n)?!1:n),p.dispatchEvent(y)}},[u,n,r]);const d=E.useRef(Ro(n)?!1:n);return w.jsx("input",{type:"checkbox","aria-hidden":!0,defaultChecked:a??d.current,...o,tabIndex:-1,ref:s,style:{...e.style,...c,position:"absolute",pointerEvents:"none",opacity:0,margin:0}})};function Ro(e){return e==="indeterminate"}function T5(e){return Ro(e)?"indeterminate":e?"checked":"unchecked"}var A5=w5,eae=k5;const mu=E.forwardRef(({className:e,...t},n)=>w.jsx(A5,{ref:n,className:Me("peer border-primary ring-offset-background focus-visible:ring-ring data-[state=checked]:bg-primary data-[state=checked]:text-primary-foreground h-4 w-4 shrink-0 rounded-sm border focus-visible:ring-2 focus-visible:ring-offset-2 focus-visible:outline-none disabled:cursor-not-allowed disabled:opacity-50",e),...t,children:w.jsx(eae,{className:Me("flex items-center justify-center text-current"),children:w.jsx(O0,{className:"h-4 w-4"})})}));mu.displayName=A5.displayName;var tae="Separator",yO="horizontal",nae=["horizontal","vertical"],R5=E.forwardRef((e,t)=>{const{decorative:n,orientation:r=yO,...a}=e,o=rae(r)?r:yO,u=n?{role:"none"}:{"aria-orientation":o==="vertical"?o:void 0,role:"separator"};return w.jsx(Ze.div,{"data-orientation":o,...u,...a,ref:t})});R5.displayName=tae;function rae(e){return nae.includes(e)}var _5=R5;const Ss=E.forwardRef(({className:e,orientation:t="horizontal",decorative:n=!0,...r},a)=>w.jsx(_5,{ref:a,decorative:n,orientation:t,className:Me("bg-border shrink-0",t==="horizontal"?"h-[1px] w-full":"h-full w-[1px]",e),...r}));Ss.displayName=_5.displayName;const Eo=({checked:e,onCheckedChange:t,label:n})=>w.jsxs("div",{className:"flex items-center gap-2",children:[w.jsx(mu,{checked:e,onCheckedChange:t}),w.jsx("label",{htmlFor:"terms",className:"text-sm leading-none font-medium peer-disabled:cursor-not-allowed peer-disabled:opacity-70",children:n})]}),Xm=({value:e,onEditFinished:t,label:n,min:r,max:a})=>{const[o,s]=E.useState(e),u=E.useCallback(d=>{const p=d.target.value.trim();if(p.length===0){s(null);return}const g=Number.parseInt(p);if(!isNaN(g)&&g!==o){if(r!==void 0&&g<r||a!==void 0&&g>a)return;s(g)}},[o,r,a]),c=E.useCallback(()=>{o!==null&&e!==o&&t(o)},[e,o,t]);return w.jsxs("div",{className:"flex flex-col gap-2",children:[w.jsx("label",{htmlFor:"terms",className:"text-sm leading-none font-medium peer-disabled:cursor-not-allowed peer-disabled:opacity-70",children:n}),w.jsx($a,{type:"number",value:o===null?"":o,onChange:u,className:"h-6 w-full min-w-0 pr-1",min:r,max:a,onBlur:c,onKeyDown:d=>{d.key==="Enter"&&c()}})]})};function aae(){const[e,t]=E.useState(!1),n=Fe.use.showPropertyPanel(),r=Fe.use.showNodeSearchBar(),a=Fe.use.showNodeLabel(),o=Fe.use.enableEdgeEvents(),s=Fe.use.enableNodeDrag(),u=Fe.use.enableHideUnselectedEdges(),c=Fe.use.showEdgeLabel(),d=Fe.use.graphQueryMaxDepth(),p=Fe.use.graphMinDegree(),g=Fe.use.graphLayoutMaxIterations(),m=Fe.use.enableHealthCheck(),b=E.useCallback(()=>Fe.setState(I=>({enableNodeDrag:!I.enableNodeDrag})),[]),y=E.useCallback(()=>Fe.setState(I=>({enableEdgeEvents:!I.enableEdgeEvents})),[]),S=E.useCallback(()=>Fe.setState(I=>({enableHideUnselectedEdges:!I.enableHideUnselectedEdges})),[]),x=E.useCallback(()=>Fe.setState(I=>({showEdgeLabel:!I.showEdgeLabel})),[]),R=E.useCallback(()=>Fe.setState(I=>({showPropertyPanel:!I.showPropertyPanel})),[]),k=E.useCallback(()=>Fe.setState(I=>({showNodeSearchBar:!I.showNodeSearchBar})),[]),A=E.useCallback(()=>Fe.setState(I=>({showNodeLabel:!I.showNodeLabel})),[]),C=E.useCallback(()=>Fe.setState(I=>({enableHealthCheck:!I.enableHealthCheck})),[]),N=E.useCallback(I=>{if(I<1)return;Fe.setState({graphQueryMaxDepth:I});const H=Fe.getState().queryLabel;Fe.getState().setQueryLabel(""),setTimeout(()=>{Fe.getState().setQueryLabel(H)},300)},[]),_=E.useCallback(I=>{if(I<0)return;Fe.setState({graphMinDegree:I});const H=Fe.getState().queryLabel;Fe.getState().setQueryLabel(""),setTimeout(()=>{Fe.getState().setQueryLabel(H)},300)},[]),O=E.useCallback(I=>{I<1||Fe.setState({graphLayoutMaxIterations:I})},[]),{t:F}=Pt(),L=()=>t(!1);return w.jsx(w.Fragment,{children:w.jsxs(Yu,{open:e,onOpenChange:t,children:[w.jsx(Ku,{asChild:!0,children:w.jsx(xt,{variant:Fr,tooltip:F("graphPanel.sideBar.settings.settings"),size:"icon",children:w.jsx(hY,{})})}),w.jsx(Ws,{side:"right",align:"start",className:"mb-2 p-2",onCloseAutoFocus:I=>I.preventDefault(),children:w.jsxs("div",{className:"flex flex-col gap-2",children:[w.jsx(Eo,{checked:m,onCheckedChange:C,label:F("graphPanel.sideBar.settings.healthCheck")}),w.jsx(Ss,{}),w.jsx(Eo,{checked:n,onCheckedChange:R,label:F("graphPanel.sideBar.settings.showPropertyPanel")}),w.jsx(Eo,{checked:r,onCheckedChange:k,label:F("graphPanel.sideBar.settings.showSearchBar")}),w.jsx(Ss,{}),w.jsx(Eo,{checked:a,onCheckedChange:A,label:F("graphPanel.sideBar.settings.showNodeLabel")}),w.jsx(Eo,{checked:s,onCheckedChange:b,label:F("graphPanel.sideBar.settings.nodeDraggable")}),w.jsx(Ss,{}),w.jsx(Eo,{checked:c,onCheckedChange:x,label:F("graphPanel.sideBar.settings.showEdgeLabel")}),w.jsx(Eo,{checked:u,onCheckedChange:S,label:F("graphPanel.sideBar.settings.hideUnselectedEdges")}),w.jsx(Eo,{checked:o,onCheckedChange:y,label:F("graphPanel.sideBar.settings.edgeEvents")}),w.jsx(Ss,{}),w.jsx(Xm,{label:F("graphPanel.sideBar.settings.maxQueryDepth"),min:1,value:d,onEditFinished:N}),w.jsx(Xm,{label:F("graphPanel.sideBar.settings.minDegree"),min:0,value:p,onEditFinished:_}),w.jsx(Xm,{label:F("graphPanel.sideBar.settings.maxLayoutIterations"),min:1,max:30,value:g,onEditFinished:O}),w.jsx(Ss,{}),w.jsx(xt,{onClick:L,variant:"outline",size:"sm",className:"ml-auto px-4",children:F("graphPanel.sideBar.settings.save")})]})})]})})}const oae="ENTRIES",C5="KEYS",N5="VALUES",mn="";class Zm{constructor(t,n){const r=t._tree,a=Array.from(r.keys());this.set=t,this._type=n,this._path=a.length>0?[{node:r,keys:a}]:[]}next(){const t=this.dive();return this.backtrack(),t}dive(){if(this._path.length===0)return{done:!0,value:void 0};const{node:t,keys:n}=ps(this._path);if(ps(n)===mn)return{done:!1,value:this.result()};const r=t.get(ps(n));return this._path.push({node:r,keys:Array.from(r.keys())}),this.dive()}backtrack(){if(this._path.length===0)return;const t=ps(this._path).keys;t.pop(),!(t.length>0)&&(this._path.pop(),this.backtrack())}key(){return this.set._prefix+this._path.map(({keys:t})=>ps(t)).filter(t=>t!==mn).join("")}value(){return ps(this._path).node.get(mn)}result(){switch(this._type){case N5:return this.value();case C5:return this.key();default:return[this.key(),this.value()]}}[Symbol.iterator](){return this}}const ps=e=>e[e.length-1],iae=(e,t,n)=>{const r=new Map;if(t===void 0)return r;const a=t.length+1,o=a+n,s=new Uint8Array(o*a).fill(n+1);for(let u=0;u<a;++u)s[u]=u;for(let u=1;u<o;++u)s[u*a]=u;return O5(e,t,n,r,s,1,a,""),r},O5=(e,t,n,r,a,o,s,u)=>{const c=o*s;e:for(const d of e.keys())if(d===mn){const p=a[c-1];p<=n&&r.set(u,[e.get(d),p])}else{let p=o;for(let g=0;g<d.length;++g,++p){const m=d[g],b=s*p,y=b-s;let S=a[b];const x=Math.max(0,p-n-1),R=Math.min(s-1,p+n);for(let k=x;k<R;++k){const A=m!==t[k],C=a[y+k]+ +A,N=a[y+k+1]+1,_=a[b+k]+1,O=a[b+k+1]=Math.min(C,N,_);O<S&&(S=O)}if(S>n)continue e}O5(e.get(d),t,n,r,a,p,s,u+d)}};class To{constructor(t=new Map,n=""){this._size=void 0,this._tree=t,this._prefix=n}atPrefix(t){if(!t.startsWith(this._prefix))throw new Error("Mismatched prefix");const[n,r]=_f(this._tree,t.slice(this._prefix.length));if(n===void 0){const[a,o]=DT(r);for(const s of a.keys())if(s!==mn&&s.startsWith(o)){const u=new Map;return u.set(s.slice(o.length),a.get(s)),new To(u,t)}}return new To(n,t)}clear(){this._size=void 0,this._tree.clear()}delete(t){return this._size=void 0,sae(this._tree,t)}entries(){return new Zm(this,oae)}forEach(t){for(const[n,r]of this)t(n,r,this)}fuzzyGet(t,n){return iae(this._tree,t,n)}get(t){const n=Yk(this._tree,t);return n!==void 0?n.get(mn):void 0}has(t){const n=Yk(this._tree,t);return n!==void 0&&n.has(mn)}keys(){return new Zm(this,C5)}set(t,n){if(typeof t!="string")throw new Error("key must be a string");return this._size=void 0,Qm(this._tree,t).set(mn,n),this}get size(){if(this._size)return this._size;this._size=0;const t=this.entries();for(;!t.next().done;)this._size+=1;return this._size}update(t,n){if(typeof t!="string")throw new Error("key must be a string");this._size=void 0;const r=Qm(this._tree,t);return r.set(mn,n(r.get(mn))),this}fetch(t,n){if(typeof t!="string")throw new Error("key must be a string");this._size=void 0;const r=Qm(this._tree,t);let a=r.get(mn);return a===void 0&&r.set(mn,a=n()),a}values(){return new Zm(this,N5)}[Symbol.iterator](){return this.entries()}static from(t){const n=new To;for(const[r,a]of t)n.set(r,a);return n}static fromObject(t){return To.from(Object.entries(t))}}const _f=(e,t,n=[])=>{if(t.length===0||e==null)return[e,n];for(const r of e.keys())if(r!==mn&&t.startsWith(r))return n.push([e,r]),_f(e.get(r),t.slice(r.length),n);return n.push([e,t]),_f(void 0,"",n)},Yk=(e,t)=>{if(t.length===0||e==null)return e;for(const n of e.keys())if(n!==mn&&t.startsWith(n))return Yk(e.get(n),t.slice(n.length))},Qm=(e,t)=>{const n=t.length;e:for(let r=0;e&&r<n;){for(const o of e.keys())if(o!==mn&&t[r]===o[0]){const s=Math.min(n-r,o.length);let u=1;for(;u<s&&t[r+u]===o[u];)++u;const c=e.get(o);if(u===o.length)e=c;else{const d=new Map;d.set(o.slice(u),c),e.set(t.slice(r,r+u),d),e.delete(o),e=d}r+=u;continue e}const a=new Map;return e.set(t.slice(r),a),a}return e},sae=(e,t)=>{const[n,r]=_f(e,t);if(n!==void 0){if(n.delete(mn),n.size===0)D5(r);else if(n.size===1){const[a,o]=n.entries().next().value;I5(r,a,o)}}},D5=e=>{if(e.length===0)return;const[t,n]=DT(e);if(t.delete(n),t.size===0)D5(e.slice(0,-1));else if(t.size===1){const[r,a]=t.entries().next().value;r!==mn&&I5(e.slice(0,-1),r,a)}},I5=(e,t,n)=>{if(e.length===0)return;const[r,a]=DT(e);r.set(a+t,n),r.delete(a)},DT=e=>e[e.length-1],IT="or",L5="and",lae="and_not";class _o{constructor(t){if((t==null?void 0:t.fields)==null)throw new Error('MiniSearch: option "fields" must be provided');const n=t.autoVacuum==null||t.autoVacuum===!0?tb:t.autoVacuum;this._options={...eb,...t,autoVacuum:n,searchOptions:{...vO,...t.searchOptions||{}},autoSuggestOptions:{...pae,...t.autoSuggestOptions||{}}},this._index=new To,this._documentCount=0,this._documentIds=new Map,this._idToShortId=new Map,this._fieldIds={},this._fieldLength=new Map,this._avgFieldLength=[],this._nextId=0,this._storedFields=new Map,this._dirtCount=0,this._currentVacuum=null,this._enqueuedVacuum=null,this._enqueuedVacuumConditions=Xk,this.addFields(this._options.fields)}add(t){const{extractField:n,tokenize:r,processTerm:a,fields:o,idField:s}=this._options,u=n(t,s);if(u==null)throw new Error(`MiniSearch: document does not have ID field "${s}"`);if(this._idToShortId.has(u))throw new Error(`MiniSearch: duplicate ID ${u}`);const c=this.addDocumentId(u);this.saveStoredFields(c,t);for(const d of o){const p=n(t,d);if(p==null)continue;const g=r(p.toString(),d),m=this._fieldIds[d],b=new Set(g).size;this.addFieldLength(c,m,this._documentCount-1,b);for(const y of g){const S=a(y,d);if(Array.isArray(S))for(const x of S)this.addTerm(m,c,x);else S&&this.addTerm(m,c,S)}}}addAll(t){for(const n of t)this.add(n)}addAllAsync(t,n={}){const{chunkSize:r=10}=n,a={chunk:[],promise:Promise.resolve()},{chunk:o,promise:s}=t.reduce(({chunk:u,promise:c},d,p)=>(u.push(d),(p+1)%r===0?{chunk:[],promise:c.then(()=>new Promise(g=>setTimeout(g,0))).then(()=>this.addAll(u))}:{chunk:u,promise:c}),a);return s.then(()=>this.addAll(o))}remove(t){const{tokenize:n,processTerm:r,extractField:a,fields:o,idField:s}=this._options,u=a(t,s);if(u==null)throw new Error(`MiniSearch: document does not have ID field "${s}"`);const c=this._idToShortId.get(u);if(c==null)throw new Error(`MiniSearch: cannot remove document with ID ${u}: it is not in the index`);for(const d of o){const p=a(t,d);if(p==null)continue;const g=n(p.toString(),d),m=this._fieldIds[d],b=new Set(g).size;this.removeFieldLength(c,m,this._documentCount,b);for(const y of g){const S=r(y,d);if(Array.isArray(S))for(const x of S)this.removeTerm(m,c,x);else S&&this.removeTerm(m,c,S)}}this._storedFields.delete(c),this._documentIds.delete(c),this._idToShortId.delete(u),this._fieldLength.delete(c),this._documentCount-=1}removeAll(t){if(t)for(const n of t)this.remove(n);else{if(arguments.length>0)throw new Error("Expected documents to be present. Omit the argument to remove all documents.");this._index=new To,this._documentCount=0,this._documentIds=new Map,this._idToShortId=new Map,this._fieldLength=new Map,this._avgFieldLength=[],this._storedFields=new Map,this._nextId=0}}discard(t){const n=this._idToShortId.get(t);if(n==null)throw new Error(`MiniSearch: cannot discard document with ID ${t}: it is not in the index`);this._idToShortId.delete(t),this._documentIds.delete(n),this._storedFields.delete(n),(this._fieldLength.get(n)||[]).forEach((r,a)=>{this.removeFieldLength(n,a,this._documentCount,r)}),this._fieldLength.delete(n),this._documentCount-=1,this._dirtCount+=1,this.maybeAutoVacuum()}maybeAutoVacuum(){if(this._options.autoVacuum===!1)return;const{minDirtFactor:t,minDirtCount:n,batchSize:r,batchWait:a}=this._options.autoVacuum;this.conditionalVacuum({batchSize:r,batchWait:a},{minDirtCount:n,minDirtFactor:t})}discardAll(t){const n=this._options.autoVacuum;try{this._options.autoVacuum=!1;for(const r of t)this.discard(r)}finally{this._options.autoVacuum=n}this.maybeAutoVacuum()}replace(t){const{idField:n,extractField:r}=this._options,a=r(t,n);this.discard(a),this.add(t)}vacuum(t={}){return this.conditionalVacuum(t)}conditionalVacuum(t,n){return this._currentVacuum?(this._enqueuedVacuumConditions=this._enqueuedVacuumConditions&&n,this._enqueuedVacuum!=null?this._enqueuedVacuum:(this._enqueuedVacuum=this._currentVacuum.then(()=>{const r=this._enqueuedVacuumConditions;return this._enqueuedVacuumConditions=Xk,this.performVacuuming(t,r)}),this._enqueuedVacuum)):this.vacuumConditionsMet(n)===!1?Promise.resolve():(this._currentVacuum=this.performVacuuming(t),this._currentVacuum)}async performVacuuming(t,n){const r=this._dirtCount;if(this.vacuumConditionsMet(n)){const a=t.batchSize||Kk.batchSize,o=t.batchWait||Kk.batchWait;let s=1;for(const[u,c]of this._index){for(const[d,p]of c)for(const[g]of p)this._documentIds.has(g)||(p.size<=1?c.delete(d):p.delete(g));this._index.get(u).size===0&&this._index.delete(u),s%a===0&&await new Promise(d=>setTimeout(d,o)),s+=1}this._dirtCount-=r}await null,this._currentVacuum=this._enqueuedVacuum,this._enqueuedVacuum=null}vacuumConditionsMet(t){if(t==null)return!0;let{minDirtCount:n,minDirtFactor:r}=t;return n=n||tb.minDirtCount,r=r||tb.minDirtFactor,this.dirtCount>=n&&this.dirtFactor>=r}get isVacuuming(){return this._currentVacuum!=null}get dirtCount(){return this._dirtCount}get dirtFactor(){return this._dirtCount/(1+this._documentCount+this._dirtCount)}has(t){return this._idToShortId.has(t)}getStoredFields(t){const n=this._idToShortId.get(t);if(n!=null)return this._storedFields.get(n)}search(t,n={}){const{searchOptions:r}=this._options,a={...r,...n},o=this.executeQuery(t,n),s=[];for(const[u,{score:c,terms:d,match:p}]of o){const g=d.length||1,m={id:this._documentIds.get(u),score:c*g,terms:Object.keys(p),queryTerms:d,match:p};Object.assign(m,this._storedFields.get(u)),(a.filter==null||a.filter(m))&&s.push(m)}return t===_o.wildcard&&a.boostDocument==null||s.sort(EO),s}autoSuggest(t,n={}){n={...this._options.autoSuggestOptions,...n};const r=new Map;for(const{score:o,terms:s}of this.search(t,n)){const u=s.join(" "),c=r.get(u);c!=null?(c.score+=o,c.count+=1):r.set(u,{score:o,terms:s,count:1})}const a=[];for(const[o,{score:s,terms:u,count:c}]of r)a.push({suggestion:o,terms:u,score:s/c});return a.sort(EO),a}get documentCount(){return this._documentCount}get termCount(){return this._index.size}static loadJSON(t,n){if(n==null)throw new Error("MiniSearch: loadJSON should be given the same options used when serializing the index");return this.loadJS(JSON.parse(t),n)}static async loadJSONAsync(t,n){if(n==null)throw new Error("MiniSearch: loadJSON should be given the same options used when serializing the index");return this.loadJSAsync(JSON.parse(t),n)}static getDefault(t){if(eb.hasOwnProperty(t))return Jm(eb,t);throw new Error(`MiniSearch: unknown option "${t}"`)}static loadJS(t,n){const{index:r,documentIds:a,fieldLength:o,storedFields:s,serializationVersion:u}=t,c=this.instantiateMiniSearch(t,n);c._documentIds=Ld(a),c._fieldLength=Ld(o),c._storedFields=Ld(s);for(const[d,p]of c._documentIds)c._idToShortId.set(p,d);for(const[d,p]of r){const g=new Map;for(const m of Object.keys(p)){let b=p[m];u===1&&(b=b.ds),g.set(parseInt(m,10),Ld(b))}c._index.set(d,g)}return c}static async loadJSAsync(t,n){const{index:r,documentIds:a,fieldLength:o,storedFields:s,serializationVersion:u}=t,c=this.instantiateMiniSearch(t,n);c._documentIds=await Md(a),c._fieldLength=await Md(o),c._storedFields=await Md(s);for(const[p,g]of c._documentIds)c._idToShortId.set(g,p);let d=0;for(const[p,g]of r){const m=new Map;for(const b of Object.keys(g)){let y=g[b];u===1&&(y=y.ds),m.set(parseInt(b,10),await Md(y))}++d%1e3===0&&await M5(0),c._index.set(p,m)}return c}static instantiateMiniSearch(t,n){const{documentCount:r,nextId:a,fieldIds:o,averageFieldLength:s,dirtCount:u,serializationVersion:c}=t;if(c!==1&&c!==2)throw new Error("MiniSearch: cannot deserialize an index created with an incompatible version");const d=new _o(n);return d._documentCount=r,d._nextId=a,d._idToShortId=new Map,d._fieldIds=o,d._avgFieldLength=s,d._dirtCount=u||0,d._index=new To,d}executeQuery(t,n={}){if(t===_o.wildcard)return this.executeWildcardQuery(n);if(typeof t!="string"){const m={...n,...t,queries:void 0},b=t.queries.map(y=>this.executeQuery(y,m));return this.combineResults(b,m.combineWith)}const{tokenize:r,processTerm:a,searchOptions:o}=this._options,s={tokenize:r,processTerm:a,...o,...n},{tokenize:u,processTerm:c}=s,g=u(t).flatMap(m=>c(m)).filter(m=>!!m).map(fae(s)).map(m=>this.executeQuerySpec(m,s));return this.combineResults(g,s.combineWith)}executeQuerySpec(t,n){const r={...this._options.searchOptions,...n},a=(r.fields||this._options.fields).reduce((S,x)=>({...S,[x]:Jm(r.boost,x)||1}),{}),{boostDocument:o,weights:s,maxFuzzy:u,bm25:c}=r,{fuzzy:d,prefix:p}={...vO.weights,...s},g=this._index.get(t.term),m=this.termResults(t.term,t.term,1,t.termBoost,g,a,o,c);let b,y;if(t.prefix&&(b=this._index.atPrefix(t.term)),t.fuzzy){const S=t.fuzzy===!0?.2:t.fuzzy,x=S<1?Math.min(u,Math.round(t.term.length*S)):S;x&&(y=this._index.fuzzyGet(t.term,x))}if(b)for(const[S,x]of b){const R=S.length-t.term.length;if(!R)continue;y==null||y.delete(S);const k=p*S.length/(S.length+.3*R);this.termResults(t.term,S,k,t.termBoost,x,a,o,c,m)}if(y)for(const S of y.keys()){const[x,R]=y.get(S);if(!R)continue;const k=d*S.length/(S.length+R);this.termResults(t.term,S,k,t.termBoost,x,a,o,c,m)}return m}executeWildcardQuery(t){const n=new Map,r={...this._options.searchOptions,...t};for(const[a,o]of this._documentIds){const s=r.boostDocument?r.boostDocument(o,"",this._storedFields.get(a)):1;n.set(a,{score:s,terms:[],match:{}})}return n}combineResults(t,n=IT){if(t.length===0)return new Map;const r=n.toLowerCase(),a=uae[r];if(!a)throw new Error(`Invalid combination operator: ${n}`);return t.reduce(a)||new Map}toJSON(){const t=[];for(const[n,r]of this._index){const a={};for(const[o,s]of r)a[o]=Object.fromEntries(s);t.push([n,a])}return{documentCount:this._documentCount,nextId:this._nextId,documentIds:Object.fromEntries(this._documentIds),fieldIds:this._fieldIds,fieldLength:Object.fromEntries(this._fieldLength),averageFieldLength:this._avgFieldLength,storedFields:Object.fromEntries(this._storedFields),dirtCount:this._dirtCount,index:t,serializationVersion:2}}termResults(t,n,r,a,o,s,u,c,d=new Map){if(o==null)return d;for(const p of Object.keys(s)){const g=s[p],m=this._fieldIds[p],b=o.get(m);if(b==null)continue;let y=b.size;const S=this._avgFieldLength[m];for(const x of b.keys()){if(!this._documentIds.has(x)){this.removeTerm(m,x,n),y-=1;continue}const R=u?u(this._documentIds.get(x),n,this._storedFields.get(x)):1;if(!R)continue;const k=b.get(x),A=this._fieldLength.get(x)[m],C=dae(k,y,this._documentCount,A,S,c),N=r*a*g*R*C,_=d.get(x);if(_){_.score+=N,gae(_.terms,t);const O=Jm(_.match,n);O?O.push(p):_.match[n]=[p]}else d.set(x,{score:N,terms:[t],match:{[n]:[p]}})}}return d}addTerm(t,n,r){const a=this._index.fetch(r,wO);let o=a.get(t);if(o==null)o=new Map,o.set(n,1),a.set(t,o);else{const s=o.get(n);o.set(n,(s||0)+1)}}removeTerm(t,n,r){if(!this._index.has(r)){this.warnDocumentChanged(n,t,r);return}const a=this._index.fetch(r,wO),o=a.get(t);o==null||o.get(n)==null?this.warnDocumentChanged(n,t,r):o.get(n)<=1?o.size<=1?a.delete(t):o.delete(n):o.set(n,o.get(n)-1),this._index.get(r).size===0&&this._index.delete(r)}warnDocumentChanged(t,n,r){for(const a of Object.keys(this._fieldIds))if(this._fieldIds[a]===n){this._options.logger("warn",`MiniSearch: document with ID ${this._documentIds.get(t)} has changed before removal: term "${r}" was not present in field "${a}". Removing a document after it has changed can corrupt the index!`,"version_conflict");return}}addDocumentId(t){const n=this._nextId;return this._idToShortId.set(t,n),this._documentIds.set(n,t),this._documentCount+=1,this._nextId+=1,n}addFields(t){for(let n=0;n<t.length;n++)this._fieldIds[t[n]]=n}addFieldLength(t,n,r,a){let o=this._fieldLength.get(t);o==null&&this._fieldLength.set(t,o=[]),o[n]=a;const u=(this._avgFieldLength[n]||0)*r+a;this._avgFieldLength[n]=u/(r+1)}removeFieldLength(t,n,r,a){if(r===1){this._avgFieldLength[n]=0;return}const o=this._avgFieldLength[n]*r-a;this._avgFieldLength[n]=o/(r-1)}saveStoredFields(t,n){const{storeFields:r,extractField:a}=this._options;if(r==null||r.length===0)return;let o=this._storedFields.get(t);o==null&&this._storedFields.set(t,o={});for(const s of r){const u=a(n,s);u!==void 0&&(o[s]=u)}}}_o.wildcard=Symbol("*");const Jm=(e,t)=>Object.prototype.hasOwnProperty.call(e,t)?e[t]:void 0,uae={[IT]:(e,t)=>{for(const n of t.keys()){const r=e.get(n);if(r==null)e.set(n,t.get(n));else{const{score:a,terms:o,match:s}=t.get(n);r.score=r.score+a,r.match=Object.assign(r.match,s),SO(r.terms,o)}}return e},[L5]:(e,t)=>{const n=new Map;for(const r of t.keys()){const a=e.get(r);if(a==null)continue;const{score:o,terms:s,match:u}=t.get(r);SO(a.terms,s),n.set(r,{score:a.score+o,terms:a.terms,match:Object.assign(a.match,u)})}return n},[lae]:(e,t)=>{for(const n of t.keys())e.delete(n);return e}},cae={k:1.2,b:.7,d:.5},dae=(e,t,n,r,a,o)=>{const{k:s,b:u,d:c}=o;return Math.log(1+(n-t+.5)/(t+.5))*(c+e*(s+1)/(e+s*(1-u+u*r/a)))},fae=e=>(t,n,r)=>{const a=typeof e.fuzzy=="function"?e.fuzzy(t,n,r):e.fuzzy||!1,o=typeof e.prefix=="function"?e.prefix(t,n,r):e.prefix===!0,s=typeof e.boostTerm=="function"?e.boostTerm(t,n,r):1;return{term:t,fuzzy:a,prefix:o,termBoost:s}},eb={idField:"id",extractField:(e,t)=>e[t],tokenize:e=>e.split(hae),processTerm:e=>e.toLowerCase(),fields:void 0,searchOptions:void 0,storeFields:[],logger:(e,t)=>{typeof(console==null?void 0:console[e])=="function"&&console[e](t)},autoVacuum:!0},vO={combineWith:IT,prefix:!1,fuzzy:!1,maxFuzzy:6,boost:{},weights:{fuzzy:.45,prefix:.375},bm25:cae},pae={combineWith:L5,prefix:(e,t,n)=>t===n.length-1},Kk={batchSize:1e3,batchWait:10},Xk={minDirtFactor:.1,minDirtCount:20},tb={...Kk,...Xk},gae=(e,t)=>{e.includes(t)||e.push(t)},SO=(e,t)=>{for(const n of t)e.includes(n)||e.push(n)},EO=({score:e},{score:t})=>t-e,wO=()=>new Map,Ld=e=>{const t=new Map;for(const n of Object.keys(e))t.set(parseInt(n,10),e[n]);return t},Md=async e=>{const t=new Map;let n=0;for(const r of Object.keys(e))t.set(parseInt(r,10),e[r]),++n%1e3===0&&await M5(0);return t},M5=e=>new Promise(t=>setTimeout(t,e)),hae=/[\n\r\p{Z}\p{P}]+/u;function Zk(){return Zk=Object.assign?Object.assign.bind():function(e){for(var t=1;t<arguments.length;t++){var n=arguments[t];for(var r in n)({}).hasOwnProperty.call(n,r)&&(e[r]=n[r])}return e},Zk.apply(null,arguments)}function Cu(e){"@babel/helpers - typeof";return Cu=typeof Symbol=="function"&&typeof Symbol.iterator=="symbol"?function(t){return typeof t}:function(t){return t&&typeof Symbol=="function"&&t.constructor===Symbol&&t!==Symbol.prototype?"symbol":typeof t},Cu(e)}function mae(e,t){if(Cu(e)!="object"||!e)return e;var n=e[Symbol.toPrimitive];if(n!==void 0){var r=n.call(e,t);if(Cu(r)!="object")return r;throw new TypeError("@@toPrimitive must return a primitive value.")}return(t==="string"?String:Number)(e)}function bae(e){var t=mae(e,"string");return Cu(t)=="symbol"?t:t+""}function F5(e,t,n){return(t=bae(t))in e?Object.defineProperty(e,t,{value:n,enumerable:!0,configurable:!0,writable:!0}):e[t]=n,e}function Qk(e,t){(t==null||t>e.length)&&(t=e.length);for(var n=0,r=Array(t);n<t;n++)r[n]=e[n];return r}function yae(e){if(Array.isArray(e))return Qk(e)}function vae(e){if(typeof Symbol<"u"&&e[Symbol.iterator]!=null||e["@@iterator"]!=null)return Array.from(e)}function Sae(e,t){if(e){if(typeof e=="string")return Qk(e,t);var n={}.toString.call(e).slice(8,-1);return n==="Object"&&e.constructor&&(n=e.constructor.name),n==="Map"||n==="Set"?Array.from(e):n==="Arguments"||/^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n)?Qk(e,t):void 0}}function Eae(){throw new TypeError(`Invalid attempt to spread non-iterable instance.
In order to be iterable, non-array objects must have a [Symbol.iterator]() method.`)}function Jk(e){return yae(e)||vae(e)||Sae(e)||Eae()}function wae(e,t){if(e==null)return{};var n={};for(var r in e)if({}.hasOwnProperty.call(e,r)){if(t.indexOf(r)!==-1)continue;n[r]=e[r]}return n}function xae(e,t){if(e==null)return{};var n,r,a=wae(e,t);if(Object.getOwnPropertySymbols){var o=Object.getOwnPropertySymbols(e);for(r=0;r<o.length;r++)n=o[r],t.indexOf(n)===-1&&{}.propertyIsEnumerable.call(e,n)&&(a[n]=e[n])}return a}const kae={index:new _o({fields:[]})};E.createContext(kae);const e0=({label:e,color:t,hidden:n,labels:r={}})=>ye.createElement("div",{className:"node"},ye.createElement("span",{className:"render "+(n?"circle":"disc"),style:{backgroundColor:t||"#000"}}),ye.createElement("span",{className:`label ${n?"text-muted":""} ${e?"":"text-italic"}`},e||r.no_label||"No label")),Tae=({id:e,labels:t})=>{const n=kr(),r=E.useMemo(()=>{const a=n.getGraph().getNodeAttributes(e),o=n.getSetting("nodeReducer");return Object.assign(Object.assign({color:n.getSetting("defaultNodeColor")},a),o?o(e,a):{})},[n,e]);return ye.createElement(e0,Object.assign({},r,{labels:t}))},Aae=({label:e,color:t,source:n,target:r,hidden:a,directed:o,labels:s={}})=>ye.createElement("div",{className:"edge"},ye.createElement(e0,Object.assign({},n,{labels:s})),ye.createElement("div",{className:"body"},ye.createElement("div",{className:"render"},ye.createElement("span",{className:a?"dotted":"dash",style:{borderColor:t||"#000"}})," ",o&&ye.createElement("span",{className:"arrow",style:{borderTopColor:t||"#000"}})),ye.createElement("span",{className:`label ${a?"text-muted":""} ${e?"":"fst-italic"}`},e||s.no_label||"No label")),ye.createElement(e0,Object.assign({},r,{labels:s}))),Rae=({id:e,labels:t})=>{const n=kr(),r=E.useMemo(()=>{const a=n.getGraph().getEdgeAttributes(e),o=n.getSetting("nodeReducer"),s=n.getSetting("edgeReducer"),u=n.getGraph().getNodeAttributes(n.getGraph().source(e)),c=n.getGraph().getNodeAttributes(n.getGraph().target(e));return Object.assign(Object.assign(Object.assign({color:n.getSetting("defaultEdgeColor"),directed:n.getGraph().isDirected(e)},a),s?s(e,a):{}),{source:Object.assign(Object.assign({color:n.getSetting("defaultNodeColor")},u),o?o(e,u):{}),target:Object.assign(Object.assign({color:n.getSetting("defaultNodeColor")},c),o?o(e,c):{})})},[n,e]);return ye.createElement(Aae,Object.assign({},r,{labels:t}))};function LT(e,t){const[n,r]=E.useState(e);return E.useEffect(()=>{const a=setTimeout(()=>{r(e)},t);return()=>{clearTimeout(a)}},[e,t]),n}function _ae({fetcher:e,preload:t,filterFn:n,renderOption:r,getOptionValue:a,notFound:o,loadingSkeleton:s,label:u,placeholder:c="Select...",value:d,onChange:p,onFocus:g,disabled:m=!1,className:b,noResultsMessage:y}){const[S,x]=E.useState(!1),[R,k]=E.useState(!1),[A,C]=E.useState([]),[N,_]=E.useState(!1),[O,F]=E.useState(null),[L,I]=E.useState(d),[H,$]=E.useState(null),[U,Y]=E.useState(""),Z=LT(U,t?0:150),[j,G]=E.useState([]);E.useEffect(()=>{x(!0),I(d)},[d]),E.useEffect(()=>{S||(async()=>{try{_(!0),F(null);const V=d!==null?await e(d):[];G(V),C(V)}catch(V){F(V instanceof Error?V.message:"Failed to fetch options")}finally{_(!1)}})()},[S,e,d]),E.useEffect(()=>{const D=async()=>{try{_(!0),F(null);const V=await e(Z);G(V),C(V)}catch(V){F(V instanceof Error?V.message:"Failed to fetch options")}finally{_(!1)}};S&&t?t&&C(Z?j.filter(V=>n?n(V,Z):!0):j):D()},[e,Z,S,t,n]);const z=E.useCallback(D=>{D!==L&&(I(D),p(D)),k(!1)},[L,I,k,p]),K=E.useCallback(D=>{D!==H&&($(D),g(D))},[H,$,g]);return w.jsx("div",{className:Me(m&&"cursor-not-allowed opacity-50",b),onFocus:()=>{k(!0)},onBlur:()=>k(!1),children:w.jsxs(xp,{shouldFilter:!1,className:"bg-transparent",children:[w.jsxs("div",{children:[w.jsx(CT,{placeholder:c,value:U,className:"max-h-8",onValueChange:D=>{Y(D),D&&!R&&k(!0)}}),N&&A.length>0&&w.jsx("div",{className:"absolute top-1/2 right-2 flex -translate-y-1/2 transform items-center",children:w.jsx(tU,{className:"h-4 w-4 animate-spin"})})]}),w.jsxs(kp,{hidden:!R,children:[O&&w.jsx("div",{className:"text-destructive p-4 text-center",children:O}),N&&A.length===0&&(s||w.jsx(Cae,{})),!N&&!O&&A.length===0&&(o||w.jsx(NT,{children:y??`No ${u.toLowerCase()} found.`})),w.jsx(Js,{children:A.map((D,V)=>w.jsxs(ye.Fragment,{children:[w.jsx(el,{value:a(D),onSelect:z,onMouseEnter:()=>K(a(D)),className:"truncate",children:r(D)},a(D)+`${V}`),V!==A.length-1&&w.jsx("div",{className:"bg-foreground/10 h-[1px]"},`divider-${V}`)]},a(D)+`-fragment-${V}`))})]})]})})}function Cae(){return w.jsx(Js,{children:w.jsx(el,{disabled:!0,children:w.jsxs("div",{className:"flex w-full items-center gap-2",children:[w.jsx("div",{className:"bg-muted h-6 w-6 animate-pulse rounded-full"}),w.jsxs("div",{className:"flex flex-1 flex-col gap-1",children:[w.jsx("div",{className:"bg-muted h-4 w-24 animate-pulse rounded"}),w.jsx("div",{className:"bg-muted h-3 w-16 animate-pulse rounded"})]})]})})})}const nb="__message_item",Nae=({id:e})=>{const t=ze.use.sigmaGraph();return t!=null&&t.hasNode(e)?w.jsx(Tae,{id:e}):null};function Oae(e){return w.jsxs("div",{children:[e.type==="nodes"&&w.jsx(Nae,{id:e.id}),e.type==="edges"&&w.jsx(Rae,{id:e.id}),e.type==="message"&&w.jsx("div",{children:e.message})]})}const Dae=({onChange:e,onFocus:t,value:n})=>{const{t:r}=Pt(),a=ze.use.sigmaGraph(),o=ze.use.searchEngine();E.useEffect(()=>{a&&ze.getState().resetSearchEngine()},[a]),E.useEffect(()=>{if(!a||a.nodes().length===0||o)return;const u=new _o({idField:"id",fields:["label"],searchOptions:{prefix:!0,fuzzy:.2,boost:{label:2}}}),c=a.nodes().map(d=>({id:d,label:a.getNodeAttribute(d,"label")}));u.addAll(c),ze.getState().setSearchEngine(u)},[a,o]);const s=E.useCallback(async u=>{if(t&&t(null),!a||!o)return[];if(a.nodes().length===0)return[];if(!u)return a.nodes().filter(p=>a.hasNode(p)).slice(0,md).map(p=>({id:p,type:"nodes"}));const c=o.search(u).filter(d=>a.hasNode(d.id)).map(d=>({id:d.id,type:"nodes"}));return c.length<=md?c:[...c.slice(0,md),{type:"message",id:nb,message:r("graphPanel.search.message",{count:c.length-md})}]},[a,o,t,r]);return w.jsx(_ae,{className:"bg-background/60 w-24 rounded-xl border-1 opacity-60 backdrop-blur-lg transition-all hover:w-fit hover:opacity-100",fetcher:s,renderOption:Oae,getOptionValue:u=>u.id,value:n&&n.type!=="message"?n.id:null,onChange:u=>{u!==nb&&e(u?{id:u,type:"nodes"}:null)},onFocus:u=>{u!==nb&&t&&t(u?{id:u,type:"nodes"}:null)},label:"item",placeholder:r("graphPanel.search.placeholder")})},Iae=({...e})=>w.jsx(Dae,{...e});function Lae({fetcher:e,preload:t,filterFn:n,renderOption:r,getOptionValue:a,getDisplayValue:o,notFound:s,loadingSkeleton:u,label:c,placeholder:d="Select...",value:p,onChange:g,disabled:m=!1,className:b,triggerClassName:y,searchInputClassName:S,noResultsMessage:x,triggerTooltip:R,clearable:k=!0}){const[A,C]=E.useState(!1),[N,_]=E.useState(!1),[O,F]=E.useState([]),[L,I]=E.useState(!1),[H,$]=E.useState(null),[U,Y]=E.useState(p),[Z,j]=E.useState(null),[G,z]=E.useState(""),K=LT(G,t?0:150),[D,V]=E.useState([]);E.useEffect(()=>{C(!0),Y(p)},[p]),E.useEffect(()=>{if(p&&O.length>0){const M=O.find(W=>a(W)===p);M&&j(M)}},[p,O,a]),E.useEffect(()=>{A||(async()=>{try{I(!0),$(null);const W=await e(p);V(W),F(W)}catch(W){$(W instanceof Error?W.message:"Failed to fetch options")}finally{I(!1)}})()},[A,e,p]),E.useEffect(()=>{const M=async()=>{try{I(!0),$(null);const W=await e(K);V(W),F(W)}catch(W){$(W instanceof Error?W.message:"Failed to fetch options")}finally{I(!1)}};A&&t?t&&F(K?D.filter(W=>n?n(W,K):!0):D):M()},[e,K,A,t,n]);const B=E.useCallback(M=>{const W=k&&M===U?"":M;Y(W),j(O.find(Q=>a(Q)===W)||null),g(W),_(!1)},[U,g,k,O,a]);return w.jsxs(Yu,{open:N,onOpenChange:_,children:[w.jsx(Ku,{asChild:!0,children:w.jsxs(xt,{variant:"outline",role:"combobox","aria-expanded":N,className:Me("justify-between",m&&"cursor-not-allowed opacity-50",y),disabled:m,tooltip:R,side:"bottom",children:[Z?o(Z):d,w.jsx(kW,{className:"opacity-50",size:10})]})}),w.jsx(Ws,{className:Me("p-0",b),onCloseAutoFocus:M=>M.preventDefault(),children:w.jsxs(xp,{shouldFilter:!1,children:[w.jsxs("div",{className:"relative w-full border-b",children:[w.jsx(CT,{placeholder:`Search ${c.toLowerCase()}...`,value:G,onValueChange:M=>{z(M)},className:S}),L&&O.length>0&&w.jsx("div",{className:"absolute top-1/2 right-2 flex -translate-y-1/2 transform items-center",children:w.jsx(tU,{className:"h-4 w-4 animate-spin"})})]}),w.jsxs(kp,{children:[H&&w.jsx("div",{className:"text-destructive p-4 text-center",children:H}),L&&O.length===0&&(u||w.jsx(Mae,{})),!L&&!H&&O.length===0&&(s||w.jsx(NT,{children:x??`No ${c.toLowerCase()} found.`})),w.jsx(Js,{children:O.map(M=>w.jsxs(el,{value:a(M),onSelect:B,className:"truncate",children:[r(M),w.jsx(O0,{className:Me("ml-auto h-3 w-3",U===a(M)?"opacity-100":"opacity-0")})]},a(M)))})]})]})})]})}function Mae(){return w.jsx(Js,{children:w.jsx(el,{disabled:!0,children:w.jsxs("div",{className:"flex w-full items-center gap-2",children:[w.jsx("div",{className:"bg-muted h-6 w-6 animate-pulse rounded-full"}),w.jsxs("div",{className:"flex flex-1 flex-col gap-1",children:[w.jsx("div",{className:"bg-muted h-4 w-24 animate-pulse rounded"}),w.jsx("div",{className:"bg-muted h-3 w-16 animate-pulse rounded"})]})]})})})}const Fae=()=>{const{t:e}=Pt(),t=Fe.use.queryLabel(),n=ze.use.allDatabaseLabels(),r=ze.use.rawGraph(),a=E.useRef(!1),o=E.useRef(!1);E.useEffect(()=>{!ze.getState().labelsFetchAttempted&&!o.current&&(o.current=!0,ze.getState().setLabelsFetchAttempted(!0),ze.getState().fetchAllDatabaseLabels().then(()=>{a.current=!0,o.current=!1}).catch(p=>{console.error("Failed to fetch labels:",p),o.current=!1,ze.getState().setLabelsFetchAttempted(!1)}))},[]),E.useEffect(()=>{a.current&&ze.getState().setGraphDataFetchAttempted(!1)},[t]);const s=E.useCallback(()=>{const d=new _o({idField:"id",fields:["value"],searchOptions:{prefix:!0,fuzzy:.2,boost:{label:2}}}),p=n.map((g,m)=>({id:m,value:g}));return d.addAll(p),{labels:n,searchEngine:d}},[n]),u=E.useCallback(async d=>{const{labels:p,searchEngine:g}=s();let m=p;return d&&(m=g.search(d).map(b=>p[b.id])),m.length<=RC?m:[...m.slice(0,RC),"..."]},[s]),c=E.useCallback(()=>{ze.getState().setLabelsFetchAttempted(!1),ze.getState().setGraphDataFetchAttempted(!1),ze.getState().fetchAllDatabaseLabels().then(()=>{const d=Fe.getState().queryLabel;Fe.getState().setQueryLabel(""),setTimeout(()=>{Fe.getState().setQueryLabel(d)},0)}).catch(d=>{console.error("Failed to refresh labels:",d)})},[]);return w.jsxs("div",{className:"flex items-center",children:[r&&w.jsx(xt,{size:"icon",variant:Fr,onClick:c,tooltip:e("graphPanel.graphLabels.refreshTooltip"),className:"mr-1",children:w.jsx(nU,{className:"h-4 w-4"})}),w.jsx(Lae,{className:"ml-2",triggerClassName:"max-h-8",searchInputClassName:"max-h-8",triggerTooltip:e("graphPanel.graphLabels.selectTooltip"),fetcher:u,renderOption:d=>w.jsx("div",{children:d}),getOptionValue:d=>d,getDisplayValue:d=>w.jsx("div",{children:d}),notFound:w.jsx("div",{className:"py-6 text-center text-sm",children:"No labels found"}),label:e("graphPanel.graphLabels.label"),placeholder:e("graphPanel.graphLabels.placeholder"),value:t!==null?t:"*",onChange:d=>{const p=Fe.getState().queryLabel;d==="..."&&(d="*"),d===p&&d!=="*"&&(d="*"),Fe.getState().setQueryLabel(d)},clearable:!1})]})},Bn=({text:e,className:t,tooltipClassName:n,tooltip:r,side:a,onClick:o})=>r?w.jsx(_3,{delayDuration:200,children:w.jsxs(C3,{children:[w.jsx(N3,{asChild:!0,children:w.jsx("label",{className:Me(t,o!==void 0?"cursor-pointer":void 0),onClick:o,children:e})}),w.jsx(iT,{side:a,className:n,children:r})]})}):w.jsx("label",{className:Me(t,o!==void 0?"cursor-pointer":void 0),onClick:o,children:e});var tf={exports:{}},Pae=tf.exports,xO;function zae(){return xO||(xO=1,function(e){(function(t,n,r){function a(c){var d=this,p=u();d.next=function(){var g=2091639*d.s0+d.c*23283064365386963e-26;return d.s0=d.s1,d.s1=d.s2,d.s2=g-(d.c=g|0)},d.c=1,d.s0=p(" "),d.s1=p(" "),d.s2=p(" "),d.s0-=p(c),d.s0<0&&(d.s0+=1),d.s1-=p(c),d.s1<0&&(d.s1+=1),d.s2-=p(c),d.s2<0&&(d.s2+=1),p=null}function o(c,d){return d.c=c.c,d.s0=c.s0,d.s1=c.s1,d.s2=c.s2,d}function s(c,d){var p=new a(c),g=d&&d.state,m=p.next;return m.int32=function(){return p.next()*4294967296|0},m.double=function(){return m()+(m()*2097152|0)*11102230246251565e-32},m.quick=m,g&&(typeof g=="object"&&o(g,p),m.state=function(){return o(p,{})}),m}function u(){var c=4022871197,d=function(p){p=String(p);for(var g=0;g<p.length;g++){c+=p.charCodeAt(g);var m=.02519603282416938*c;c=m>>>0,m-=c,m*=c,c=m>>>0,m-=c,c+=m*4294967296}return(c>>>0)*23283064365386963e-26};return d}n&&n.exports?n.exports=s:this.alea=s})(Pae,e)}(tf)),tf.exports}var nf={exports:{}},Bae=nf.exports,kO;function Uae(){return kO||(kO=1,function(e){(function(t,n,r){function a(u){var c=this,d="";c.x=0,c.y=0,c.z=0,c.w=0,c.next=function(){var g=c.x^c.x<<11;return c.x=c.y,c.y=c.z,c.z=c.w,c.w^=c.w>>>19^g^g>>>8},u===(u|0)?c.x=u:d+=u;for(var p=0;p<d.length+64;p++)c.x^=d.charCodeAt(p)|0,c.next()}function o(u,c){return c.x=u.x,c.y=u.y,c.z=u.z,c.w=u.w,c}function s(u,c){var d=new a(u),p=c&&c.state,g=function(){return(d.next()>>>0)/4294967296};return g.double=function(){do var m=d.next()>>>11,b=(d.next()>>>0)/4294967296,y=(m+b)/(1<<21);while(y===0);return y},g.int32=d.next,g.quick=g,p&&(typeof p=="object"&&o(p,d),g.state=function(){return o(d,{})}),g}n&&n.exports?n.exports=s:this.xor128=s})(Bae,e)}(nf)),nf.exports}var rf={exports:{}},jae=rf.exports,TO;function Gae(){return TO||(TO=1,function(e){(function(t,n,r){function a(u){var c=this,d="";c.next=function(){var g=c.x^c.x>>>2;return c.x=c.y,c.y=c.z,c.z=c.w,c.w=c.v,(c.d=c.d+362437|0)+(c.v=c.v^c.v<<4^(g^g<<1))|0},c.x=0,c.y=0,c.z=0,c.w=0,c.v=0,u===(u|0)?c.x=u:d+=u;for(var p=0;p<d.length+64;p++)c.x^=d.charCodeAt(p)|0,p==d.length&&(c.d=c.x<<10^c.x>>>4),c.next()}function o(u,c){return c.x=u.x,c.y=u.y,c.z=u.z,c.w=u.w,c.v=u.v,c.d=u.d,c}function s(u,c){var d=new a(u),p=c&&c.state,g=function(){return(d.next()>>>0)/4294967296};return g.double=function(){do var m=d.next()>>>11,b=(d.next()>>>0)/4294967296,y=(m+b)/(1<<21);while(y===0);return y},g.int32=d.next,g.quick=g,p&&(typeof p=="object"&&o(p,d),g.state=function(){return o(d,{})}),g}n&&n.exports?n.exports=s:this.xorwow=s})(jae,e)}(rf)),rf.exports}var af={exports:{}},Hae=af.exports,AO;function $ae(){return AO||(AO=1,function(e){(function(t,n,r){function a(u){var c=this;c.next=function(){var p=c.x,g=c.i,m,b;return m=p[g],m^=m>>>7,b=m^m<<24,m=p[g+1&7],b^=m^m>>>10,m=p[g+3&7],b^=m^m>>>3,m=p[g+4&7],b^=m^m<<7,m=p[g+7&7],m=m^m<<13,b^=m^m<<9,p[g]=b,c.i=g+1&7,b};function d(p,g){var m,b=[];if(g===(g|0))b[0]=g;else for(g=""+g,m=0;m<g.length;++m)b[m&7]=b[m&7]<<15^g.charCodeAt(m)+b[m+1&7]<<13;for(;b.length<8;)b.push(0);for(m=0;m<8&&b[m]===0;++m);for(m==8?b[7]=-1:b[m],p.x=b,p.i=0,m=256;m>0;--m)p.next()}d(c,u)}function o(u,c){return c.x=u.x.slice(),c.i=u.i,c}function s(u,c){u==null&&(u=+new Date);var d=new a(u),p=c&&c.state,g=function(){return(d.next()>>>0)/4294967296};return g.double=function(){do var m=d.next()>>>11,b=(d.next()>>>0)/4294967296,y=(m+b)/(1<<21);while(y===0);return y},g.int32=d.next,g.quick=g,p&&(p.x&&o(p,d),g.state=function(){return o(d,{})}),g}n&&n.exports?n.exports=s:this.xorshift7=s})(Hae,e)}(af)),af.exports}var of={exports:{}},qae=of.exports,RO;function Vae(){return RO||(RO=1,function(e){(function(t,n,r){function a(u){var c=this;c.next=function(){var p=c.w,g=c.X,m=c.i,b,y;return c.w=p=p+1640531527|0,y=g[m+34&127],b=g[m=m+1&127],y^=y<<13,b^=b<<17,y^=y>>>15,b^=b>>>12,y=g[m]=y^b,c.i=m,y+(p^p>>>16)|0};function d(p,g){var m,b,y,S,x,R=[],k=128;for(g===(g|0)?(b=g,g=null):(g=g+"\0",b=0,k=Math.max(k,g.length)),y=0,S=-32;S<k;++S)g&&(b^=g.charCodeAt((S+32)%g.length)),S===0&&(x=b),b^=b<<10,b^=b>>>15,b^=b<<4,b^=b>>>13,S>=0&&(x=x+1640531527|0,m=R[S&127]^=b+x,y=m==0?y+1:0);for(y>=128&&(R[(g&&g.length||0)&127]=-1),y=127,S=4*128;S>0;--S)b=R[y+34&127],m=R[y=y+1&127],b^=b<<13,m^=m<<17,b^=b>>>15,m^=m>>>12,R[y]=b^m;p.w=x,p.X=R,p.i=y}d(c,u)}function o(u,c){return c.i=u.i,c.w=u.w,c.X=u.X.slice(),c}function s(u,c){u==null&&(u=+new Date);var d=new a(u),p=c&&c.state,g=function(){return(d.next()>>>0)/4294967296};return g.double=function(){do var m=d.next()>>>11,b=(d.next()>>>0)/4294967296,y=(m+b)/(1<<21);while(y===0);return y},g.int32=d.next,g.quick=g,p&&(p.X&&o(p,d),g.state=function(){return o(d,{})}),g}n&&n.exports?n.exports=s:this.xor4096=s})(qae,e)}(of)),of.exports}var sf={exports:{}},Wae=sf.exports,_O;function Yae(){return _O||(_O=1,function(e){(function(t,n,r){function a(u){var c=this,d="";c.next=function(){var g=c.b,m=c.c,b=c.d,y=c.a;return g=g<<25^g>>>7^m,m=m-b|0,b=b<<24^b>>>8^y,y=y-g|0,c.b=g=g<<20^g>>>12^m,c.c=m=m-b|0,c.d=b<<16^m>>>16^y,c.a=y-g|0},c.a=0,c.b=0,c.c=-1640531527,c.d=1367130551,u===Math.floor(u)?(c.a=u/4294967296|0,c.b=u|0):d+=u;for(var p=0;p<d.length+20;p++)c.b^=d.charCodeAt(p)|0,c.next()}function o(u,c){return c.a=u.a,c.b=u.b,c.c=u.c,c.d=u.d,c}function s(u,c){var d=new a(u),p=c&&c.state,g=function(){return(d.next()>>>0)/4294967296};return g.double=function(){do var m=d.next()>>>11,b=(d.next()>>>0)/4294967296,y=(m+b)/(1<<21);while(y===0);return y},g.int32=d.next,g.quick=g,p&&(typeof p=="object"&&o(p,d),g.state=function(){return o(d,{})}),g}n&&n.exports?n.exports=s:this.tychei=s})(Wae,e)}(sf)),sf.exports}var lf={exports:{}};const Kae={},Xae=Object.freeze(Object.defineProperty({__proto__:null,default:Kae},Symbol.toStringTag,{value:"Module"})),Zae=d9(Xae);var Qae=lf.exports,CO;function Jae(){return CO||(CO=1,function(e){(function(t,n,r){var a=256,o=6,s=52,u="random",c=r.pow(a,o),d=r.pow(2,s),p=d*2,g=a-1,m;function b(C,N,_){var O=[];N=N==!0?{entropy:!0}:N||{};var F=R(x(N.entropy?[C,A(n)]:C??k(),3),O),L=new y(O),I=function(){for(var H=L.g(o),$=c,U=0;H<d;)H=(H+U)*a,$*=a,U=L.g(1);for(;H>=p;)H/=2,$/=2,U>>>=1;return(H+U)/$};return I.int32=function(){return L.g(4)|0},I.quick=function(){return L.g(4)/4294967296},I.double=I,R(A(L.S),n),(N.pass||_||function(H,$,U,Y){return Y&&(Y.S&&S(Y,L),H.state=function(){return S(L,{})}),U?(r[u]=H,$):H})(I,F,"global"in N?N.global:this==r,N.state)}function y(C){var N,_=C.length,O=this,F=0,L=O.i=O.j=0,I=O.S=[];for(_||(C=[_++]);F<a;)I[F]=F++;for(F=0;F<a;F++)I[F]=I[L=g&L+C[F%_]+(N=I[F])],I[L]=N;(O.g=function(H){for(var $,U=0,Y=O.i,Z=O.j,j=O.S;H--;)$=j[Y=g&Y+1],U=U*a+j[g&(j[Y]=j[Z=g&Z+$])+(j[Z]=$)];return O.i=Y,O.j=Z,U})(a)}function S(C,N){return N.i=C.i,N.j=C.j,N.S=C.S.slice(),N}function x(C,N){var _=[],O=typeof C,F;if(N&&O=="object")for(F in C)try{_.push(x(C[F],N-1))}catch{}return _.length?_:O=="string"?C:C+"\0"}function R(C,N){for(var _=C+"",O,F=0;F<_.length;)N[g&F]=g&(O^=N[g&F]*19)+_.charCodeAt(F++);return A(N)}function k(){try{var C;return m&&(C=m.randomBytes)?C=C(a):(C=new Uint8Array(a),(t.crypto||t.msCrypto).getRandomValues(C)),A(C)}catch{var N=t.navigator,_=N&&N.plugins;return[+new Date,t,_,t.screen,A(n)]}}function A(C){return String.fromCharCode.apply(0,C)}if(R(r.random(),n),e.exports){e.exports=b;try{m=Zae}catch{}}else r["seed"+u]=b})(typeof self<"u"?self:Qae,[],Math)}(lf)),lf.exports}var rb,NO;function eoe(){if(NO)return rb;NO=1;var e=zae(),t=Uae(),n=Gae(),r=$ae(),a=Vae(),o=Yae(),s=Jae();return s.alea=e,s.xor128=t,s.xorwow=n,s.xorshift7=r,s.xor4096=a,s.tychei=o,rb=s,rb}var toe=eoe();const Cf=un(toe),noe=e=>{if(!e||!Array.isArray(e.nodes)||!Array.isArray(e.edges))return!1;for(const t of e.nodes)if(!t.id||!t.labels||!t.properties)return!1;for(const t of e.edges)if(!t.id||!t.source||!t.target)return!1;for(const t of e.edges){const n=e.getNode(t.source),r=e.getNode(t.target);if(n==null||r==null)return!1}return!0},roe=async(e,t,n)=>{let r=null;try{r=await $B(e,t,n)}catch(o){return Hn.getState().setErrorMessage(Sr(o),"Query Graphs Error!"),null}let a=null;if(r){const o={},s={};for(let p=0;p<r.nodes.length;p++){const g=r.nodes[p];o[g.id]=p,Cf(g.id,{global:!0}),g.color=fB(),g.x=Math.random(),g.y=Math.random(),g.degree=0,g.size=10}for(let p=0;p<r.edges.length;p++){const g=r.edges[p];s[g.id]=p;const m=o[g.source],b=o[g.target];if(m!==void 0&&m!==void 0){const y=r.nodes[m],S=r.nodes[b];if(!y){console.error(`Source node ${g.source} is undefined`);continue}if(!S){console.error(`Target node ${g.target} is undefined`);continue}y.degree+=1,S.degree+=1}}let u=Number.MAX_SAFE_INTEGER,c=0;for(const p of r.nodes)u=Math.min(u,p.degree),c=Math.max(c,p.degree);const d=c-u;if(d>0){const p=UB-gu;for(const g of r.nodes)g.size=Math.round(gu+p*Math.pow((g.degree-u)/d,.5))}a=new TV,a.nodes=r.nodes,a.edges=r.edges,a.nodeIdMap=o,a.edgeIdMap=s,noe(a)||(a=null,console.error("Invalid graph data")),console.log("Graph data loaded")}return a},aoe=e=>{if(!e||!e.nodes.length)return console.log("No graph data available, skipping sigma graph creation"),null;const t=new Sp;for(const n of(e==null?void 0:e.nodes)??[]){Cf(n.id+Date.now().toString(),{global:!0});const r=Math.random(),a=Math.random();t.addNode(n.id,{label:n.labels.join(", "),color:n.color,x:r,y:a,size:n.size,borderColor:BB,borderSize:.2})}for(const n of(e==null?void 0:e.edges)??[])n.dynamicId=t.addDirectedEdge(n.source,n.target,{label:n.type||void 0});return t},ooe=()=>{const{t:e}=Pt(),t=Fe.use.queryLabel(),n=ze.use.rawGraph(),r=ze.use.sigmaGraph(),a=Fe.use.graphQueryMaxDepth(),o=Fe.use.graphMinDegree(),s=ze.use.isFetching(),u=ze.use.nodeToExpand(),c=ze.use.nodeToPrune(),d=E.useRef(!1),p=E.useRef(!1),g=E.useCallback(x=>(n==null?void 0:n.getNode(x))||null,[n]),m=E.useCallback((x,R=!0)=>(n==null?void 0:n.getEdge(x,R))||null,[n]),b=E.useRef(!1);E.useEffect(()=>{if(!t&&(n!==null||r!==null)){const x=ze.getState();x.reset(),x.setGraphDataFetchAttempted(!1),x.setLabelsFetchAttempted(!1),d.current=!1,p.current=!1}},[t,n,r]),E.useEffect(()=>{if(!(b.current||!t)&&!s&&!ze.getState().graphDataFetchAttempted){b.current=!0,ze.getState().setGraphDataFetchAttempted(!0);const x=ze.getState();x.setIsFetching(!0),x.clearSelection(),x.sigmaGraph&&x.sigmaGraph.forEachNode(C=>{var N;(N=x.sigmaGraph)==null||N.setNodeAttribute(C,"highlighted",!1)}),console.log("Fetching graph data..."),roe(t,a,o).then(C=>{const N=ze.getState();N.reset();const _=aoe(C);C==null||C.buildDynamicMap(),N.setSigmaGraph(_),N.setRawGraph(C),d.current=!0,p.current=!0,b.current=!1,N.setMoveToSelectedNode(!0),N.setIsFetching(!1)}).catch(C=>{console.error("Error fetching graph data:",C);const N=ze.getState();N.setIsFetching(!1),d.current=!1,b.current=!1,N.setGraphDataFetchAttempted(!1)})}},[t,a,o,s]),E.useEffect(()=>{u&&((async R=>{var k,A,C;if(!(!R||!r||!n))try{const N=n.getNode(R);if(!N){console.error("Node not found:",R);return}const _=N.labels[0];if(!_){console.error("Node has no label:",R);return}const O=await $B(_,2,0);if(!O||!O.nodes||!O.edges){console.error("Failed to fetch extended graph");return}const F=[];for(const W of O.nodes){Cf(W.id,{global:!0});const Q=fB();F.push({id:W.id,labels:W.labels,properties:W.properties,size:10,x:Math.random(),y:Math.random(),color:Q,degree:0})}const L=[];for(const W of O.edges)L.push({id:W.id,source:W.source,target:W.target,type:W.type,properties:W.properties,dynamicId:""});const I={};r.forEachNode(W=>{I[W]={x:r.getNodeAttribute(W,"x"),y:r.getNodeAttribute(W,"y")}});const H=new Set(r.nodes()),$=new Set,U=new Set,Y=1;let Z=0;r.forEachNode(W=>{const Q=r.degree(W);Z=Math.max(Z,Q)});const j=Z-Y||1,G=UB-gu;for(const W of F){if(H.has(W.id))continue;L.some(oe=>oe.source===R&&oe.target===W.id||oe.target===R&&oe.source===W.id)&&$.add(W.id)}const z=new Map,K=new Set;for(const W of L){const Q=H.has(W.source)||$.has(W.source),oe=H.has(W.target)||$.has(W.target);Q&&oe?(U.add(W.id),$.has(W.source)&&z.set(W.source,(z.get(W.source)||0)+1),$.has(W.target)&&z.set(W.target,(z.get(W.target)||0)+1)):(r.hasNode(W.source)?K.add(W.source):$.has(W.source)&&(K.add(W.source),z.set(W.source,(z.get(W.source)||0)+1)),r.hasNode(W.target)?K.add(W.target):$.has(W.target)&&(K.add(W.target),z.set(W.target,(z.get(W.target)||0)+1)))}const D=(W,Q,oe,re,ie)=>{for(const Se of Q)if(W.hasNode(Se)){let ae=W.degree(Se);ae+=1;const ve=Math.round(gu+ie*Math.pow((ae-oe)/re,.5)),xe=W.getNodeAttribute(Se,"size");ve>xe&&W.setNodeAttribute(Se,"size",ve)}};if($.size===0){D(r,K,Y,j,G),Ft.info(e("graphPanel.propertiesView.node.noNewNodes"));return}for(const[,W]of z.entries())Z=Math.max(Z,W);const V=((k=ze.getState().sigmaInstance)==null?void 0:k.getCamera().ratio)||1,B=Math.max(Math.sqrt(N.size)*4,Math.sqrt($.size)*3)/V;Cf(Date.now().toString(),{global:!0});const M=Math.random()*2*Math.PI;console.log("nodeSize:",N.size,"nodesToAdd:",$.size),console.log("cameraRatio:",Math.round(V*100)/100,"spreadFactor:",Math.round(B*100)/100);for(const W of $){const Q=F.find(ve=>ve.id===W),oe=z.get(W)||0,re=Math.round(gu+G*Math.pow((oe-Y)/j,.5)),ie=2*Math.PI*(Array.from($).indexOf(W)/$.size),Se=((A=I[W])==null?void 0:A.x)||I[N.id].x+Math.cos(M+ie)*B,ae=((C=I[W])==null?void 0:C.y)||I[N.id].y+Math.sin(M+ie)*B;r.addNode(W,{label:Q.labels.join(", "),color:Q.color,x:Se,y:ae,size:re,borderColor:BB,borderSize:.2}),n.getNode(W)||(Q.size=re,Q.x=Se,Q.y=ae,Q.degree=oe,n.nodes.push(Q),n.nodeIdMap[W]=n.nodes.length-1)}for(const W of U){const Q=L.find(oe=>oe.id===W);r.hasEdge(Q.source,Q.target)||r.hasEdge(Q.target,Q.source)||(Q.dynamicId=r.addDirectedEdge(Q.source,Q.target,{label:Q.type||void 0}),n.getEdge(Q.id,!1)?console.error("Edge already exists in rawGraph:",Q.id):(n.edges.push(Q),n.edgeIdMap[Q.id]=n.edges.length-1,n.edgeDynamicIdMap[Q.dynamicId]=n.edges.length-1))}n.buildDynamicMap(),ze.getState().resetSearchEngine(),D(r,K,Y,j,G)}catch(N){console.error("Error expanding node:",N)}})(u),window.setTimeout(()=>{ze.getState().triggerNodeExpand(null)},0))},[u,r,n,e]);const y=E.useCallback((x,R)=>{const k=new Set([x]);return R.forEachNode(A=>{if(A===x)return;const C=R.neighbors(A);C.length===1&&C[0]===x&&k.add(A)}),k},[]);return E.useEffect(()=>{c&&((R=>{if(!(!R||!r||!n))try{const k=ze.getState();if(!r.hasNode(R)){console.error("Node not found:",R);return}const A=y(R,r);if(A.size===r.nodes().length){Ft.error(e("graphPanel.propertiesView.node.deleteAllNodesError"));return}k.clearSelection();for(const C of A){r.dropNode(C);const N=n.nodeIdMap[C];if(N!==void 0){const _=n.edges.filter(O=>O.source===C||O.target===C);for(const O of _){const F=n.edgeIdMap[O.id];if(F!==void 0){n.edges.splice(F,1);for(const[L,I]of Object.entries(n.edgeIdMap))I>F&&(n.edgeIdMap[L]=I-1);delete n.edgeIdMap[O.id],delete n.edgeDynamicIdMap[O.dynamicId]}}n.nodes.splice(N,1);for(const[O,F]of Object.entries(n.nodeIdMap))F>N&&(n.nodeIdMap[O]=F-1);delete n.nodeIdMap[C]}}n.buildDynamicMap(),ze.getState().resetSearchEngine(),A.size>1&&Ft.info(e("graphPanel.propertiesView.node.nodesRemoved",{count:A.size}))}catch(k){console.error("Error pruning node:",k)}})(c),window.setTimeout(()=>{ze.getState().triggerNodePrune(null)},0))},[c,r,n,y,e]),{lightrageGraph:E.useCallback(()=>{if(r)return r;console.log("Creating new Sigma graph instance");const x=new Sp;return ze.getState().setSigmaGraph(x),x},[r]),getNode:g,getEdge:m}},ioe=()=>{const{getNode:e,getEdge:t}=ooe(),n=ze.use.selectedNode(),r=ze.use.focusedNode(),a=ze.use.selectedEdge(),o=ze.use.focusedEdge(),[s,u]=E.useState(null),[c,d]=E.useState(null);return E.useEffect(()=>{let p=null,g=null;r?(p="node",g=e(r)):n?(p="node",g=e(n)):o?(p="edge",g=t(o,!0)):a&&(p="edge",g=t(a,!0)),g?(p=="node"?u(soe(g)):u(loe(g)),d(p)):(u(null),d(null))},[r,n,o,a,u,d,e,t]),s?w.jsx("div",{className:"bg-background/80 max-w-xs rounded-lg border-2 p-2 text-xs backdrop-blur-lg",children:c=="node"?w.jsx(uoe,{node:s}):w.jsx(coe,{edge:s})}):w.jsx(w.Fragment,{})},soe=e=>{const t=ze.getState(),n=[];if(t.sigmaGraph&&t.rawGraph)try{if(!t.sigmaGraph.hasNode(e.id))return{...e,relationships:[]};const r=t.sigmaGraph.edges(e.id);for(const a of r){if(!t.sigmaGraph.hasEdge(a))continue;const o=t.rawGraph.getEdge(a,!0);if(o){const u=e.id===o.source?o.target:o.source;if(!t.sigmaGraph.hasNode(u))continue;const c=t.rawGraph.getNode(u);c&&n.push({type:"Neighbour",id:u,label:c.properties.entity_id?c.properties.entity_id:c.labels.join(", ")})}}}catch(r){console.error("Error refining node properties:",r)}return{...e,relationships:n}},loe=e=>{const t=ze.getState();let n,r;if(t.sigmaGraph&&t.rawGraph)try{if(!t.sigmaGraph.hasEdge(e.id))return{...e,sourceNode:void 0,targetNode:void 0};t.sigmaGraph.hasNode(e.source)&&(n=t.rawGraph.getNode(e.source)),t.sigmaGraph.hasNode(e.target)&&(r=t.rawGraph.getNode(e.target))}catch(a){console.error("Error refining edge properties:",a)}return{...e,sourceNode:n,targetNode:r}},ta=({name:e,value:t,onClick:n,tooltip:r})=>{const{t:a}=Pt(),o=s=>{const u=`graphPanel.propertiesView.node.propertyNames.${s}`,c=a(u);return c===u?s:c};return w.jsxs("div",{className:"flex items-center gap-2",children:[w.jsx("label",{className:"text-primary/60 tracking-wide whitespace-nowrap",children:o(e)}),":",w.jsx(Bn,{className:"hover:bg-primary/20 rounded p-1 overflow-hidden text-ellipsis",tooltipClassName:"max-w-80",text:t,tooltip:r||(typeof t=="string"?t:JSON.stringify(t,null,2)),side:"left",onClick:n})]})},uoe=({node:e})=>{const{t}=Pt(),n=()=>{ze.getState().triggerNodeExpand(e.id)},r=()=>{ze.getState().triggerNodePrune(e.id)};return w.jsxs("div",{className:"flex flex-col gap-2",children:[w.jsxs("div",{className:"flex justify-between items-center",children:[w.jsx("label",{className:"text-md pl-1 font-bold tracking-wide text-blue-700",children:t("graphPanel.propertiesView.node.title")}),w.jsxs("div",{className:"flex gap-3",children:[w.jsx(xt,{size:"icon",variant:"ghost",className:"h-7 w-7 border border-gray-400 hover:bg-gray-200 dark:border-gray-600 dark:hover:bg-gray-700",onClick:n,tooltip:t("graphPanel.propertiesView.node.expandNode"),children:w.jsx(PW,{className:"h-4 w-4 text-gray-700 dark:text-gray-300"})}),w.jsx(xt,{size:"icon",variant:"ghost",className:"h-7 w-7 border border-gray-400 hover:bg-gray-200 dark:border-gray-600 dark:hover:bg-gray-700",onClick:r,tooltip:t("graphPanel.propertiesView.node.pruneNode"),children:w.jsx(uY,{className:"h-4 w-4 text-gray-900 dark:text-gray-300"})})]})]}),w.jsxs("div",{className:"bg-primary/5 max-h-96 overflow-auto rounded p-1",children:[w.jsx(ta,{name:t("graphPanel.propertiesView.node.id"),value:e.id}),w.jsx(ta,{name:t("graphPanel.propertiesView.node.labels"),value:e.labels.join(", "),onClick:()=>{ze.getState().setSelectedNode(e.id,!0)}}),w.jsx(ta,{name:t("graphPanel.propertiesView.node.degree"),value:e.degree})]}),w.jsx("label",{className:"text-md pl-1 font-bold tracking-wide text-amber-700",children:t("graphPanel.propertiesView.node.properties")}),w.jsx("div",{className:"bg-primary/5 max-h-96 overflow-auto rounded p-1",children:Object.keys(e.properties).sort().map(a=>w.jsx(ta,{name:a,value:e.properties[a]},a))}),e.relationships.length>0&&w.jsxs(w.Fragment,{children:[w.jsx("label",{className:"text-md pl-1 font-bold tracking-wide text-emerald-700",children:t("graphPanel.propertiesView.node.relationships")}),w.jsx("div",{className:"bg-primary/5 max-h-96 overflow-auto rounded p-1",children:e.relationships.map(({type:a,id:o,label:s})=>w.jsx(ta,{name:a,value:s,onClick:()=>{ze.getState().setSelectedNode(o,!0)}},o))})]})]})},coe=({edge:e})=>{const{t}=Pt();return w.jsxs("div",{className:"flex flex-col gap-2",children:[w.jsx("label",{className:"text-md pl-1 font-bold tracking-wide text-violet-700",children:t("graphPanel.propertiesView.edge.title")}),w.jsxs("div",{className:"bg-primary/5 max-h-96 overflow-auto rounded p-1",children:[w.jsx(ta,{name:t("graphPanel.propertiesView.edge.id"),value:e.id}),e.type&&w.jsx(ta,{name:t("graphPanel.propertiesView.edge.type"),value:e.type}),w.jsx(ta,{name:t("graphPanel.propertiesView.edge.source"),value:e.sourceNode?e.sourceNode.labels.join(", "):e.source,onClick:()=>{ze.getState().setSelectedNode(e.source,!0)}}),w.jsx(ta,{name:t("graphPanel.propertiesView.edge.target"),value:e.targetNode?e.targetNode.labels.join(", "):e.target,onClick:()=>{ze.getState().setSelectedNode(e.target,!0)}})]}),w.jsx("label",{className:"text-md pl-1 font-bold tracking-wide text-amber-700",children:t("graphPanel.propertiesView.edge.properties")}),w.jsx("div",{className:"bg-primary/5 max-h-96 overflow-auto rounded p-1",children:Object.keys(e.properties).sort().map(n=>w.jsx(ta,{name:n,value:e.properties[n]},n))})]})},doe=()=>{const{t:e}=Pt(),t=Fe.use.graphQueryMaxDepth(),n=Fe.use.graphMinDegree();return w.jsxs("div",{className:"absolute bottom-4 left-[calc(1rem+2.5rem)] flex items-center gap-2 text-xs text-gray-400",children:[w.jsxs("div",{children:[e("graphPanel.sideBar.settings.depth"),": ",t]}),w.jsxs("div",{children:[e("graphPanel.sideBar.settings.degree"),": ",n]})]})},OO={allowInvalidContainer:!0,defaultNodeType:"default",defaultEdgeType:"curvedArrow",renderEdgeLabels:!1,edgeProgramClasses:{arrow:D4,curvedArrow:Rne,curvedNoArrow:Ane},nodeProgramClasses:{default:une,circel:Qu,point:Pte},labelGridCellSize:60,labelRenderedSizeThreshold:12,enableEdgeEvents:!0,labelColor:{color:"#000",attribute:"labelColor"},edgeLabelColor:{color:"#000",attribute:"labelColor"},edgeLabelSize:8,labelSize:12},foe=()=>{const e=j4(),t=kr(),[n,r]=E.useState(null);return E.useEffect(()=>{e({downNode:a=>{r(a.node),t.getGraph().setNodeAttribute(a.node,"highlighted",!0)},mousemovebody:a=>{if(!n)return;const o=t.viewportToGraph(a);t.getGraph().setNodeAttribute(n,"x",o.x),t.getGraph().setNodeAttribute(n,"y",o.y),a.preventSigmaDefault(),a.original.preventDefault(),a.original.stopPropagation()},mouseup:()=>{n&&(r(null),t.getGraph().removeNodeAttribute(n,"highlighted"))},mousedown:a=>{a.original.buttons!==0&&!t.getCustomBBox()&&t.setCustomBBox(t.getBBox())}})},[e,t,n]),null},poe=()=>{const[e,t]=E.useState(OO),n=E.useRef(null),r=ze.use.selectedNode(),a=ze.use.focusedNode(),o=ze.use.moveToSelectedNode(),s=ze.use.isFetching(),u=Fe.use.showPropertyPanel(),c=Fe.use.showNodeSearchBar(),d=Fe.use.enableNodeDrag();E.useEffect(()=>{t(OO),console.log("Initialized sigma settings")},[]),E.useEffect(()=>()=>{const y=ze.getState().sigmaInstance;if(y)try{y.kill(),ze.getState().setSigmaInstance(null),console.log("Cleared sigma instance on Graphviewer unmount")}catch(S){console.error("Error cleaning up sigma instance:",S)}},[]);const p=E.useCallback(y=>{y===null?ze.getState().setFocusedNode(null):y.type==="nodes"&&ze.getState().setFocusedNode(y.id)},[]),g=E.useCallback(y=>{y===null?ze.getState().setSelectedNode(null):y.type==="nodes"&&ze.getState().setSelectedNode(y.id,!0)},[]),m=E.useMemo(()=>a??r,[a,r]),b=E.useMemo(()=>r?{type:"nodes",id:r}:null,[r]);return w.jsxs("div",{className:"relative h-full w-full overflow-hidden",children:[w.jsxs(Ote,{settings:e,className:"!bg-background !size-full overflow-hidden",ref:n,children:[w.jsx(Wre,{}),d&&w.jsx(foe,{}),w.jsx(_ne,{node:m,move:o}),w.jsxs("div",{className:"absolute top-2 left-2 flex items-start gap-2",children:[w.jsx(Fae,{}),c&&w.jsx(Iae,{value:b,onFocus:p,onChange:g})]}),w.jsxs("div",{className:"bg-background/60 absolute bottom-2 left-2 flex flex-col rounded-xl border-2 backdrop-blur-lg",children:[w.jsx(Vre,{}),w.jsx(Yre,{}),w.jsx(Kre,{}),w.jsx(aae,{})]}),u&&w.jsx("div",{className:"absolute top-2 right-2",children:w.jsx(ioe,{})}),w.jsx(doe,{})]}),s&&w.jsx("div",{className:"absolute inset-0 flex items-center justify-center bg-background/80 z-10",children:w.jsxs("div",{className:"text-center",children:[w.jsx("div",{className:"mb-2 h-8 w-8 animate-spin rounded-full border-4 border-primary border-t-transparent"}),w.jsx("p",{children:"Loading Graph Data..."})]})})]})},P5=E.forwardRef(({className:e,...t},n)=>w.jsx("div",{className:"relative w-full overflow-auto",children:w.jsx("table",{ref:n,className:Me("w-full caption-bottom text-sm",e),...t})}));P5.displayName="Table";const z5=E.forwardRef(({className:e,...t},n)=>w.jsx("thead",{ref:n,className:Me("[&_tr]:border-b",e),...t}));z5.displayName="TableHeader";const B5=E.forwardRef(({className:e,...t},n)=>w.jsx("tbody",{ref:n,className:Me("[&_tr:last-child]:border-0",e),...t}));B5.displayName="TableBody";const goe=E.forwardRef(({className:e,...t},n)=>w.jsx("tfoot",{ref:n,className:Me("bg-muted/50 border-t font-medium [&>tr]:last:border-b-0",e),...t}));goe.displayName="TableFooter";const t0=E.forwardRef(({className:e,...t},n)=>w.jsx("tr",{ref:n,className:Me("hover:bg-muted/50 data-[state=selected]:bg-muted border-b transition-colors",e),...t}));t0.displayName="TableRow";const Ma=E.forwardRef(({className:e,...t},n)=>w.jsx("th",{ref:n,className:Me("text-muted-foreground h-10 px-2 text-left align-middle font-medium [&:has([role=checkbox])]:pr-0 [&>[role=checkbox]]:translate-y-[2px]",e),...t}));Ma.displayName="TableHead";const Fa=E.forwardRef(({className:e,...t},n)=>w.jsx("td",{ref:n,className:Me("p-2 align-middle [&:has([role=checkbox])]:pr-0 [&>[role=checkbox]]:translate-y-[2px]",e),...t}));Fa.displayName="TableCell";const hoe=E.forwardRef(({className:e,...t},n)=>w.jsx("caption",{ref:n,className:Me("text-muted-foreground mt-4 text-sm",e),...t}));hoe.displayName="TableCaption";const Ms=E.forwardRef(({className:e,...t},n)=>w.jsx("div",{ref:n,className:Me("bg-card text-card-foreground rounded-xl border shadow",e),...t}));Ms.displayName="Card";const Nu=E.forwardRef(({className:e,...t},n)=>w.jsx("div",{ref:n,className:Me("flex flex-col space-y-1.5 p-6",e),...t}));Nu.displayName="CardHeader";const Ou=E.forwardRef(({className:e,...t},n)=>w.jsx("div",{ref:n,className:Me("leading-none font-semibold tracking-tight",e),...t}));Ou.displayName="CardTitle";const Tp=E.forwardRef(({className:e,...t},n)=>w.jsx("div",{ref:n,className:Me("text-muted-foreground text-sm",e),...t}));Tp.displayName="CardDescription";const Du=E.forwardRef(({className:e,...t},n)=>w.jsx("div",{ref:n,className:Me("p-6 pt-0",e),...t}));Du.displayName="CardContent";const moe=E.forwardRef(({className:e,...t},n)=>w.jsx("div",{ref:n,className:Me("flex items-center p-6 pt-0",e),...t}));moe.displayName="CardFooter";function boe({title:e,description:t,icon:n=IW,action:r,className:a,...o}){return w.jsxs(Ms,{className:Me("flex w-full flex-col items-center justify-center space-y-6 bg-transparent p-16",a),...o,children:[w.jsx("div",{className:"mr-4 shrink-0 rounded-full border border-dashed p-4",children:w.jsx(n,{className:"text-muted-foreground size-8","aria-hidden":"true"})}),w.jsxs("div",{className:"flex flex-col items-center gap-1.5 text-center",children:[w.jsx(Ou,{children:e}),t?w.jsx(Tp,{children:t}):null]}),r||null]})}var ab={exports:{}},ob,DO;function yoe(){if(DO)return ob;DO=1;var e="SECRET_DO_NOT_PASS_THIS_OR_YOU_WILL_BE_FIRED";return ob=e,ob}var ib,IO;function voe(){if(IO)return ib;IO=1;var e=yoe();function t(){}function n(){}return n.resetWarningCache=t,ib=function(){function r(s,u,c,d,p,g){if(g!==e){var m=new Error("Calling PropTypes validators directly is not supported by the `prop-types` package. Use PropTypes.checkPropTypes() to call them. Read more at http://fb.me/use-check-prop-types");throw m.name="Invariant Violation",m}}r.isRequired=r;function a(){return r}var o={array:r,bigint:r,bool:r,func:r,number:r,object:r,string:r,symbol:r,any:r,arrayOf:a,element:r,elementType:r,instanceOf:a,node:r,objectOf:a,oneOf:a,oneOfType:a,shape:a,exact:a,checkPropTypes:n,resetWarningCache:t};return o.PropTypes=o,o},ib}var LO;function Soe(){return LO||(LO=1,ab.exports=voe()()),ab.exports}var Eoe=Soe();const Ot=un(Eoe),woe=new Map([["1km","application/vnd.1000minds.decision-model+xml"],["3dml","text/vnd.in3d.3dml"],["3ds","image/x-3ds"],["3g2","video/3gpp2"],["3gp","video/3gp"],["3gpp","video/3gpp"],["3mf","model/3mf"],["7z","application/x-7z-compressed"],["7zip","application/x-7z-compressed"],["123","application/vnd.lotus-1-2-3"],["aab","application/x-authorware-bin"],["aac","audio/x-acc"],["aam","application/x-authorware-map"],["aas","application/x-authorware-seg"],["abw","application/x-abiword"],["ac","application/vnd.nokia.n-gage.ac+xml"],["ac3","audio/ac3"],["acc","application/vnd.americandynamics.acc"],["ace","application/x-ace-compressed"],["acu","application/vnd.acucobol"],["acutc","application/vnd.acucorp"],["adp","audio/adpcm"],["aep","application/vnd.audiograph"],["afm","application/x-font-type1"],["afp","application/vnd.ibm.modcap"],["ahead","application/vnd.ahead.space"],["ai","application/pdf"],["aif","audio/x-aiff"],["aifc","audio/x-aiff"],["aiff","audio/x-aiff"],["air","application/vnd.adobe.air-application-installer-package+zip"],["ait","application/vnd.dvb.ait"],["ami","application/vnd.amiga.ami"],["amr","audio/amr"],["apk","application/vnd.android.package-archive"],["apng","image/apng"],["appcache","text/cache-manifest"],["application","application/x-ms-application"],["apr","application/vnd.lotus-approach"],["arc","application/x-freearc"],["arj","application/x-arj"],["asc","application/pgp-signature"],["asf","video/x-ms-asf"],["asm","text/x-asm"],["aso","application/vnd.accpac.simply.aso"],["asx","video/x-ms-asf"],["atc","application/vnd.acucorp"],["atom","application/atom+xml"],["atomcat","application/atomcat+xml"],["atomdeleted","application/atomdeleted+xml"],["atomsvc","application/atomsvc+xml"],["atx","application/vnd.antix.game-component"],["au","audio/x-au"],["avi","video/x-msvideo"],["avif","image/avif"],["aw","application/applixware"],["azf","application/vnd.airzip.filesecure.azf"],["azs","application/vnd.airzip.filesecure.azs"],["azv","image/vnd.airzip.accelerator.azv"],["azw","application/vnd.amazon.ebook"],["b16","image/vnd.pco.b16"],["bat","application/x-msdownload"],["bcpio","application/x-bcpio"],["bdf","application/x-font-bdf"],["bdm","application/vnd.syncml.dm+wbxml"],["bdoc","application/x-bdoc"],["bed","application/vnd.realvnc.bed"],["bh2","application/vnd.fujitsu.oasysprs"],["bin","application/octet-stream"],["blb","application/x-blorb"],["blorb","application/x-blorb"],["bmi","application/vnd.bmi"],["bmml","application/vnd.balsamiq.bmml+xml"],["bmp","image/bmp"],["book","application/vnd.framemaker"],["box","application/vnd.previewsystems.box"],["boz","application/x-bzip2"],["bpk","application/octet-stream"],["bpmn","application/octet-stream"],["bsp","model/vnd.valve.source.compiled-map"],["btif","image/prs.btif"],["buffer","application/octet-stream"],["bz","application/x-bzip"],["bz2","application/x-bzip2"],["c","text/x-c"],["c4d","application/vnd.clonk.c4group"],["c4f","application/vnd.clonk.c4group"],["c4g","application/vnd.clonk.c4group"],["c4p","application/vnd.clonk.c4group"],["c4u","application/vnd.clonk.c4group"],["c11amc","application/vnd.cluetrust.cartomobile-config"],["c11amz","application/vnd.cluetrust.cartomobile-config-pkg"],["cab","application/vnd.ms-cab-compressed"],["caf","audio/x-caf"],["cap","application/vnd.tcpdump.pcap"],["car","application/vnd.curl.car"],["cat","application/vnd.ms-pki.seccat"],["cb7","application/x-cbr"],["cba","application/x-cbr"],["cbr","application/x-cbr"],["cbt","application/x-cbr"],["cbz","application/x-cbr"],["cc","text/x-c"],["cco","application/x-cocoa"],["cct","application/x-director"],["ccxml","application/ccxml+xml"],["cdbcmsg","application/vnd.contact.cmsg"],["cda","application/x-cdf"],["cdf","application/x-netcdf"],["cdfx","application/cdfx+xml"],["cdkey","application/vnd.mediastation.cdkey"],["cdmia","application/cdmi-capability"],["cdmic","application/cdmi-container"],["cdmid","application/cdmi-domain"],["cdmio","application/cdmi-object"],["cdmiq","application/cdmi-queue"],["cdr","application/cdr"],["cdx","chemical/x-cdx"],["cdxml","application/vnd.chemdraw+xml"],["cdy","application/vnd.cinderella"],["cer","application/pkix-cert"],["cfs","application/x-cfs-compressed"],["cgm","image/cgm"],["chat","application/x-chat"],["chm","application/vnd.ms-htmlhelp"],["chrt","application/vnd.kde.kchart"],["cif","chemical/x-cif"],["cii","application/vnd.anser-web-certificate-issue-initiation"],["cil","application/vnd.ms-artgalry"],["cjs","application/node"],["cla","application/vnd.claymore"],["class","application/octet-stream"],["clkk","application/vnd.crick.clicker.keyboard"],["clkp","application/vnd.crick.clicker.palette"],["clkt","application/vnd.crick.clicker.template"],["clkw","application/vnd.crick.clicker.wordbank"],["clkx","application/vnd.crick.clicker"],["clp","application/x-msclip"],["cmc","application/vnd.cosmocaller"],["cmdf","chemical/x-cmdf"],["cml","chemical/x-cml"],["cmp","application/vnd.yellowriver-custom-menu"],["cmx","image/x-cmx"],["cod","application/vnd.rim.cod"],["coffee","text/coffeescript"],["com","application/x-msdownload"],["conf","text/plain"],["cpio","application/x-cpio"],["cpp","text/x-c"],["cpt","application/mac-compactpro"],["crd","application/x-mscardfile"],["crl","application/pkix-crl"],["crt","application/x-x509-ca-cert"],["crx","application/x-chrome-extension"],["cryptonote","application/vnd.rig.cryptonote"],["csh","application/x-csh"],["csl","application/vnd.citationstyles.style+xml"],["csml","chemical/x-csml"],["csp","application/vnd.commonspace"],["csr","application/octet-stream"],["css","text/css"],["cst","application/x-director"],["csv","text/csv"],["cu","application/cu-seeme"],["curl","text/vnd.curl"],["cww","application/prs.cww"],["cxt","application/x-director"],["cxx","text/x-c"],["dae","model/vnd.collada+xml"],["daf","application/vnd.mobius.daf"],["dart","application/vnd.dart"],["dataless","application/vnd.fdsn.seed"],["davmount","application/davmount+xml"],["dbf","application/vnd.dbf"],["dbk","application/docbook+xml"],["dcr","application/x-director"],["dcurl","text/vnd.curl.dcurl"],["dd2","application/vnd.oma.dd2+xml"],["ddd","application/vnd.fujixerox.ddd"],["ddf","application/vnd.syncml.dmddf+xml"],["dds","image/vnd.ms-dds"],["deb","application/x-debian-package"],["def","text/plain"],["deploy","application/octet-stream"],["der","application/x-x509-ca-cert"],["dfac","application/vnd.dreamfactory"],["dgc","application/x-dgc-compressed"],["dic","text/x-c"],["dir","application/x-director"],["dis","application/vnd.mobius.dis"],["disposition-notification","message/disposition-notification"],["dist","application/octet-stream"],["distz","application/octet-stream"],["djv","image/vnd.djvu"],["djvu","image/vnd.djvu"],["dll","application/octet-stream"],["dmg","application/x-apple-diskimage"],["dmn","application/octet-stream"],["dmp","application/vnd.tcpdump.pcap"],["dms","application/octet-stream"],["dna","application/vnd.dna"],["doc","application/msword"],["docm","application/vnd.ms-word.template.macroEnabled.12"],["docx","application/vnd.openxmlformats-officedocument.wordprocessingml.document"],["dot","application/msword"],["dotm","application/vnd.ms-word.template.macroEnabled.12"],["dotx","application/vnd.openxmlformats-officedocument.wordprocessingml.template"],["dp","application/vnd.osgi.dp"],["dpg","application/vnd.dpgraph"],["dra","audio/vnd.dra"],["drle","image/dicom-rle"],["dsc","text/prs.lines.tag"],["dssc","application/dssc+der"],["dtb","application/x-dtbook+xml"],["dtd","application/xml-dtd"],["dts","audio/vnd.dts"],["dtshd","audio/vnd.dts.hd"],["dump","application/octet-stream"],["dvb","video/vnd.dvb.file"],["dvi","application/x-dvi"],["dwd","application/atsc-dwd+xml"],["dwf","model/vnd.dwf"],["dwg","image/vnd.dwg"],["dxf","image/vnd.dxf"],["dxp","application/vnd.spotfire.dxp"],["dxr","application/x-director"],["ear","application/java-archive"],["ecelp4800","audio/vnd.nuera.ecelp4800"],["ecelp7470","audio/vnd.nuera.ecelp7470"],["ecelp9600","audio/vnd.nuera.ecelp9600"],["ecma","application/ecmascript"],["edm","application/vnd.novadigm.edm"],["edx","application/vnd.novadigm.edx"],["efif","application/vnd.picsel"],["ei6","application/vnd.pg.osasli"],["elc","application/octet-stream"],["emf","image/emf"],["eml","message/rfc822"],["emma","application/emma+xml"],["emotionml","application/emotionml+xml"],["emz","application/x-msmetafile"],["eol","audio/vnd.digital-winds"],["eot","application/vnd.ms-fontobject"],["eps","application/postscript"],["epub","application/epub+zip"],["es","application/ecmascript"],["es3","application/vnd.eszigno3+xml"],["esa","application/vnd.osgi.subsystem"],["esf","application/vnd.epson.esf"],["et3","application/vnd.eszigno3+xml"],["etx","text/x-setext"],["eva","application/x-eva"],["evy","application/x-envoy"],["exe","application/octet-stream"],["exi","application/exi"],["exp","application/express"],["exr","image/aces"],["ext","application/vnd.novadigm.ext"],["ez","application/andrew-inset"],["ez2","application/vnd.ezpix-album"],["ez3","application/vnd.ezpix-package"],["f","text/x-fortran"],["f4v","video/mp4"],["f77","text/x-fortran"],["f90","text/x-fortran"],["fbs","image/vnd.fastbidsheet"],["fcdt","application/vnd.adobe.formscentral.fcdt"],["fcs","application/vnd.isac.fcs"],["fdf","application/vnd.fdf"],["fdt","application/fdt+xml"],["fe_launch","application/vnd.denovo.fcselayout-link"],["fg5","application/vnd.fujitsu.oasysgp"],["fgd","application/x-director"],["fh","image/x-freehand"],["fh4","image/x-freehand"],["fh5","image/x-freehand"],["fh7","image/x-freehand"],["fhc","image/x-freehand"],["fig","application/x-xfig"],["fits","image/fits"],["flac","audio/x-flac"],["fli","video/x-fli"],["flo","application/vnd.micrografx.flo"],["flv","video/x-flv"],["flw","application/vnd.kde.kivio"],["flx","text/vnd.fmi.flexstor"],["fly","text/vnd.fly"],["fm","application/vnd.framemaker"],["fnc","application/vnd.frogans.fnc"],["fo","application/vnd.software602.filler.form+xml"],["for","text/x-fortran"],["fpx","image/vnd.fpx"],["frame","application/vnd.framemaker"],["fsc","application/vnd.fsc.weblaunch"],["fst","image/vnd.fst"],["ftc","application/vnd.fluxtime.clip"],["fti","application/vnd.anser-web-funds-transfer-initiation"],["fvt","video/vnd.fvt"],["fxp","application/vnd.adobe.fxp"],["fxpl","application/vnd.adobe.fxp"],["fzs","application/vnd.fuzzysheet"],["g2w","application/vnd.geoplan"],["g3","image/g3fax"],["g3w","application/vnd.geospace"],["gac","application/vnd.groove-account"],["gam","application/x-tads"],["gbr","application/rpki-ghostbusters"],["gca","application/x-gca-compressed"],["gdl","model/vnd.gdl"],["gdoc","application/vnd.google-apps.document"],["geo","application/vnd.dynageo"],["geojson","application/geo+json"],["gex","application/vnd.geometry-explorer"],["ggb","application/vnd.geogebra.file"],["ggt","application/vnd.geogebra.tool"],["ghf","application/vnd.groove-help"],["gif","image/gif"],["gim","application/vnd.groove-identity-message"],["glb","model/gltf-binary"],["gltf","model/gltf+json"],["gml","application/gml+xml"],["gmx","application/vnd.gmx"],["gnumeric","application/x-gnumeric"],["gpg","application/gpg-keys"],["gph","application/vnd.flographit"],["gpx","application/gpx+xml"],["gqf","application/vnd.grafeq"],["gqs","application/vnd.grafeq"],["gram","application/srgs"],["gramps","application/x-gramps-xml"],["gre","application/vnd.geometry-explorer"],["grv","application/vnd.groove-injector"],["grxml","application/srgs+xml"],["gsf","application/x-font-ghostscript"],["gsheet","application/vnd.google-apps.spreadsheet"],["gslides","application/vnd.google-apps.presentation"],["gtar","application/x-gtar"],["gtm","application/vnd.groove-tool-message"],["gtw","model/vnd.gtw"],["gv","text/vnd.graphviz"],["gxf","application/gxf"],["gxt","application/vnd.geonext"],["gz","application/gzip"],["gzip","application/gzip"],["h","text/x-c"],["h261","video/h261"],["h263","video/h263"],["h264","video/h264"],["hal","application/vnd.hal+xml"],["hbci","application/vnd.hbci"],["hbs","text/x-handlebars-template"],["hdd","application/x-virtualbox-hdd"],["hdf","application/x-hdf"],["heic","image/heic"],["heics","image/heic-sequence"],["heif","image/heif"],["heifs","image/heif-sequence"],["hej2","image/hej2k"],["held","application/atsc-held+xml"],["hh","text/x-c"],["hjson","application/hjson"],["hlp","application/winhlp"],["hpgl","application/vnd.hp-hpgl"],["hpid","application/vnd.hp-hpid"],["hps","application/vnd.hp-hps"],["hqx","application/mac-binhex40"],["hsj2","image/hsj2"],["htc","text/x-component"],["htke","application/vnd.kenameaapp"],["htm","text/html"],["html","text/html"],["hvd","application/vnd.yamaha.hv-dic"],["hvp","application/vnd.yamaha.hv-voice"],["hvs","application/vnd.yamaha.hv-script"],["i2g","application/vnd.intergeo"],["icc","application/vnd.iccprofile"],["ice","x-conference/x-cooltalk"],["icm","application/vnd.iccprofile"],["ico","image/x-icon"],["ics","text/calendar"],["ief","image/ief"],["ifb","text/calendar"],["ifm","application/vnd.shana.informed.formdata"],["iges","model/iges"],["igl","application/vnd.igloader"],["igm","application/vnd.insors.igm"],["igs","model/iges"],["igx","application/vnd.micrografx.igx"],["iif","application/vnd.shana.informed.interchange"],["img","application/octet-stream"],["imp","application/vnd.accpac.simply.imp"],["ims","application/vnd.ms-ims"],["in","text/plain"],["ini","text/plain"],["ink","application/inkml+xml"],["inkml","application/inkml+xml"],["install","application/x-install-instructions"],["iota","application/vnd.astraea-software.iota"],["ipfix","application/ipfix"],["ipk","application/vnd.shana.informed.package"],["irm","application/vnd.ibm.rights-management"],["irp","application/vnd.irepository.package+xml"],["iso","application/x-iso9660-image"],["itp","application/vnd.shana.informed.formtemplate"],["its","application/its+xml"],["ivp","application/vnd.immervision-ivp"],["ivu","application/vnd.immervision-ivu"],["jad","text/vnd.sun.j2me.app-descriptor"],["jade","text/jade"],["jam","application/vnd.jam"],["jar","application/java-archive"],["jardiff","application/x-java-archive-diff"],["java","text/x-java-source"],["jhc","image/jphc"],["jisp","application/vnd.jisp"],["jls","image/jls"],["jlt","application/vnd.hp-jlyt"],["jng","image/x-jng"],["jnlp","application/x-java-jnlp-file"],["joda","application/vnd.joost.joda-archive"],["jp2","image/jp2"],["jpe","image/jpeg"],["jpeg","image/jpeg"],["jpf","image/jpx"],["jpg","image/jpeg"],["jpg2","image/jp2"],["jpgm","video/jpm"],["jpgv","video/jpeg"],["jph","image/jph"],["jpm","video/jpm"],["jpx","image/jpx"],["js","application/javascript"],["json","application/json"],["json5","application/json5"],["jsonld","application/ld+json"],["jsonl","application/jsonl"],["jsonml","application/jsonml+json"],["jsx","text/jsx"],["jxr","image/jxr"],["jxra","image/jxra"],["jxrs","image/jxrs"],["jxs","image/jxs"],["jxsc","image/jxsc"],["jxsi","image/jxsi"],["jxss","image/jxss"],["kar","audio/midi"],["karbon","application/vnd.kde.karbon"],["kdb","application/octet-stream"],["kdbx","application/x-keepass2"],["key","application/x-iwork-keynote-sffkey"],["kfo","application/vnd.kde.kformula"],["kia","application/vnd.kidspiration"],["kml","application/vnd.google-earth.kml+xml"],["kmz","application/vnd.google-earth.kmz"],["kne","application/vnd.kinar"],["knp","application/vnd.kinar"],["kon","application/vnd.kde.kontour"],["kpr","application/vnd.kde.kpresenter"],["kpt","application/vnd.kde.kpresenter"],["kpxx","application/vnd.ds-keypoint"],["ksp","application/vnd.kde.kspread"],["ktr","application/vnd.kahootz"],["ktx","image/ktx"],["ktx2","image/ktx2"],["ktz","application/vnd.kahootz"],["kwd","application/vnd.kde.kword"],["kwt","application/vnd.kde.kword"],["lasxml","application/vnd.las.las+xml"],["latex","application/x-latex"],["lbd","application/vnd.llamagraphics.life-balance.desktop"],["lbe","application/vnd.llamagraphics.life-balance.exchange+xml"],["les","application/vnd.hhe.lesson-player"],["less","text/less"],["lgr","application/lgr+xml"],["lha","application/octet-stream"],["link66","application/vnd.route66.link66+xml"],["list","text/plain"],["list3820","application/vnd.ibm.modcap"],["listafp","application/vnd.ibm.modcap"],["litcoffee","text/coffeescript"],["lnk","application/x-ms-shortcut"],["log","text/plain"],["lostxml","application/lost+xml"],["lrf","application/octet-stream"],["lrm","application/vnd.ms-lrm"],["ltf","application/vnd.frogans.ltf"],["lua","text/x-lua"],["luac","application/x-lua-bytecode"],["lvp","audio/vnd.lucent.voice"],["lwp","application/vnd.lotus-wordpro"],["lzh","application/octet-stream"],["m1v","video/mpeg"],["m2a","audio/mpeg"],["m2v","video/mpeg"],["m3a","audio/mpeg"],["m3u","text/plain"],["m3u8","application/vnd.apple.mpegurl"],["m4a","audio/x-m4a"],["m4p","application/mp4"],["m4s","video/iso.segment"],["m4u","application/vnd.mpegurl"],["m4v","video/x-m4v"],["m13","application/x-msmediaview"],["m14","application/x-msmediaview"],["m21","application/mp21"],["ma","application/mathematica"],["mads","application/mads+xml"],["maei","application/mmt-aei+xml"],["mag","application/vnd.ecowin.chart"],["maker","application/vnd.framemaker"],["man","text/troff"],["manifest","text/cache-manifest"],["map","application/json"],["mar","application/octet-stream"],["markdown","text/markdown"],["mathml","application/mathml+xml"],["mb","application/mathematica"],["mbk","application/vnd.mobius.mbk"],["mbox","application/mbox"],["mc1","application/vnd.medcalcdata"],["mcd","application/vnd.mcd"],["mcurl","text/vnd.curl.mcurl"],["md","text/markdown"],["mdb","application/x-msaccess"],["mdi","image/vnd.ms-modi"],["mdx","text/mdx"],["me","text/troff"],["mesh","model/mesh"],["meta4","application/metalink4+xml"],["metalink","application/metalink+xml"],["mets","application/mets+xml"],["mfm","application/vnd.mfmp"],["mft","application/rpki-manifest"],["mgp","application/vnd.osgeo.mapguide.package"],["mgz","application/vnd.proteus.magazine"],["mid","audio/midi"],["midi","audio/midi"],["mie","application/x-mie"],["mif","application/vnd.mif"],["mime","message/rfc822"],["mj2","video/mj2"],["mjp2","video/mj2"],["mjs","application/javascript"],["mk3d","video/x-matroska"],["mka","audio/x-matroska"],["mkd","text/x-markdown"],["mks","video/x-matroska"],["mkv","video/x-matroska"],["mlp","application/vnd.dolby.mlp"],["mmd","application/vnd.chipnuts.karaoke-mmd"],["mmf","application/vnd.smaf"],["mml","text/mathml"],["mmr","image/vnd.fujixerox.edmics-mmr"],["mng","video/x-mng"],["mny","application/x-msmoney"],["mobi","application/x-mobipocket-ebook"],["mods","application/mods+xml"],["mov","video/quicktime"],["movie","video/x-sgi-movie"],["mp2","audio/mpeg"],["mp2a","audio/mpeg"],["mp3","audio/mpeg"],["mp4","video/mp4"],["mp4a","audio/mp4"],["mp4s","application/mp4"],["mp4v","video/mp4"],["mp21","application/mp21"],["mpc","application/vnd.mophun.certificate"],["mpd","application/dash+xml"],["mpe","video/mpeg"],["mpeg","video/mpeg"],["mpg","video/mpeg"],["mpg4","video/mp4"],["mpga","audio/mpeg"],["mpkg","application/vnd.apple.installer+xml"],["mpm","application/vnd.blueice.multipass"],["mpn","application/vnd.mophun.application"],["mpp","application/vnd.ms-project"],["mpt","application/vnd.ms-project"],["mpy","application/vnd.ibm.minipay"],["mqy","application/vnd.mobius.mqy"],["mrc","application/marc"],["mrcx","application/marcxml+xml"],["ms","text/troff"],["mscml","application/mediaservercontrol+xml"],["mseed","application/vnd.fdsn.mseed"],["mseq","application/vnd.mseq"],["msf","application/vnd.epson.msf"],["msg","application/vnd.ms-outlook"],["msh","model/mesh"],["msi","application/x-msdownload"],["msl","application/vnd.mobius.msl"],["msm","application/octet-stream"],["msp","application/octet-stream"],["msty","application/vnd.muvee.style"],["mtl","model/mtl"],["mts","model/vnd.mts"],["mus","application/vnd.musician"],["musd","application/mmt-usd+xml"],["musicxml","application/vnd.recordare.musicxml+xml"],["mvb","application/x-msmediaview"],["mvt","application/vnd.mapbox-vector-tile"],["mwf","application/vnd.mfer"],["mxf","application/mxf"],["mxl","application/vnd.recordare.musicxml"],["mxmf","audio/mobile-xmf"],["mxml","application/xv+xml"],["mxs","application/vnd.triscape.mxs"],["mxu","video/vnd.mpegurl"],["n-gage","application/vnd.nokia.n-gage.symbian.install"],["n3","text/n3"],["nb","application/mathematica"],["nbp","application/vnd.wolfram.player"],["nc","application/x-netcdf"],["ncx","application/x-dtbncx+xml"],["nfo","text/x-nfo"],["ngdat","application/vnd.nokia.n-gage.data"],["nitf","application/vnd.nitf"],["nlu","application/vnd.neurolanguage.nlu"],["nml","application/vnd.enliven"],["nnd","application/vnd.noblenet-directory"],["nns","application/vnd.noblenet-sealer"],["nnw","application/vnd.noblenet-web"],["npx","image/vnd.net-fpx"],["nq","application/n-quads"],["nsc","application/x-conference"],["nsf","application/vnd.lotus-notes"],["nt","application/n-triples"],["ntf","application/vnd.nitf"],["numbers","application/x-iwork-numbers-sffnumbers"],["nzb","application/x-nzb"],["oa2","application/vnd.fujitsu.oasys2"],["oa3","application/vnd.fujitsu.oasys3"],["oas","application/vnd.fujitsu.oasys"],["obd","application/x-msbinder"],["obgx","application/vnd.openblox.game+xml"],["obj","model/obj"],["oda","application/oda"],["odb","application/vnd.oasis.opendocument.database"],["odc","application/vnd.oasis.opendocument.chart"],["odf","application/vnd.oasis.opendocument.formula"],["odft","application/vnd.oasis.opendocument.formula-template"],["odg","application/vnd.oasis.opendocument.graphics"],["odi","application/vnd.oasis.opendocument.image"],["odm","application/vnd.oasis.opendocument.text-master"],["odp","application/vnd.oasis.opendocument.presentation"],["ods","application/vnd.oasis.opendocument.spreadsheet"],["odt","application/vnd.oasis.opendocument.text"],["oga","audio/ogg"],["ogex","model/vnd.opengex"],["ogg","audio/ogg"],["ogv","video/ogg"],["ogx","application/ogg"],["omdoc","application/omdoc+xml"],["onepkg","application/onenote"],["onetmp","application/onenote"],["onetoc","application/onenote"],["onetoc2","application/onenote"],["opf","application/oebps-package+xml"],["opml","text/x-opml"],["oprc","application/vnd.palm"],["opus","audio/ogg"],["org","text/x-org"],["osf","application/vnd.yamaha.openscoreformat"],["osfpvg","application/vnd.yamaha.openscoreformat.osfpvg+xml"],["osm","application/vnd.openstreetmap.data+xml"],["otc","application/vnd.oasis.opendocument.chart-template"],["otf","font/otf"],["otg","application/vnd.oasis.opendocument.graphics-template"],["oth","application/vnd.oasis.opendocument.text-web"],["oti","application/vnd.oasis.opendocument.image-template"],["otp","application/vnd.oasis.opendocument.presentation-template"],["ots","application/vnd.oasis.opendocument.spreadsheet-template"],["ott","application/vnd.oasis.opendocument.text-template"],["ova","application/x-virtualbox-ova"],["ovf","application/x-virtualbox-ovf"],["owl","application/rdf+xml"],["oxps","application/oxps"],["oxt","application/vnd.openofficeorg.extension"],["p","text/x-pascal"],["p7a","application/x-pkcs7-signature"],["p7b","application/x-pkcs7-certificates"],["p7c","application/pkcs7-mime"],["p7m","application/pkcs7-mime"],["p7r","application/x-pkcs7-certreqresp"],["p7s","application/pkcs7-signature"],["p8","application/pkcs8"],["p10","application/x-pkcs10"],["p12","application/x-pkcs12"],["pac","application/x-ns-proxy-autoconfig"],["pages","application/x-iwork-pages-sffpages"],["pas","text/x-pascal"],["paw","application/vnd.pawaafile"],["pbd","application/vnd.powerbuilder6"],["pbm","image/x-portable-bitmap"],["pcap","application/vnd.tcpdump.pcap"],["pcf","application/x-font-pcf"],["pcl","application/vnd.hp-pcl"],["pclxl","application/vnd.hp-pclxl"],["pct","image/x-pict"],["pcurl","application/vnd.curl.pcurl"],["pcx","image/x-pcx"],["pdb","application/x-pilot"],["pde","text/x-processing"],["pdf","application/pdf"],["pem","application/x-x509-user-cert"],["pfa","application/x-font-type1"],["pfb","application/x-font-type1"],["pfm","application/x-font-type1"],["pfr","application/font-tdpfr"],["pfx","application/x-pkcs12"],["pgm","image/x-portable-graymap"],["pgn","application/x-chess-pgn"],["pgp","application/pgp"],["php","application/x-httpd-php"],["php3","application/x-httpd-php"],["php4","application/x-httpd-php"],["phps","application/x-httpd-php-source"],["phtml","application/x-httpd-php"],["pic","image/x-pict"],["pkg","application/octet-stream"],["pki","application/pkixcmp"],["pkipath","application/pkix-pkipath"],["pkpass","application/vnd.apple.pkpass"],["pl","application/x-perl"],["plb","application/vnd.3gpp.pic-bw-large"],["plc","application/vnd.mobius.plc"],["plf","application/vnd.pocketlearn"],["pls","application/pls+xml"],["pm","application/x-perl"],["pml","application/vnd.ctc-posml"],["png","image/png"],["pnm","image/x-portable-anymap"],["portpkg","application/vnd.macports.portpkg"],["pot","application/vnd.ms-powerpoint"],["potm","application/vnd.ms-powerpoint.presentation.macroEnabled.12"],["potx","application/vnd.openxmlformats-officedocument.presentationml.template"],["ppa","application/vnd.ms-powerpoint"],["ppam","application/vnd.ms-powerpoint.addin.macroEnabled.12"],["ppd","application/vnd.cups-ppd"],["ppm","image/x-portable-pixmap"],["pps","application/vnd.ms-powerpoint"],["ppsm","application/vnd.ms-powerpoint.slideshow.macroEnabled.12"],["ppsx","application/vnd.openxmlformats-officedocument.presentationml.slideshow"],["ppt","application/powerpoint"],["pptm","application/vnd.ms-powerpoint.presentation.macroEnabled.12"],["pptx","application/vnd.openxmlformats-officedocument.presentationml.presentation"],["pqa","application/vnd.palm"],["prc","application/x-pilot"],["pre","application/vnd.lotus-freelance"],["prf","application/pics-rules"],["provx","application/provenance+xml"],["ps","application/postscript"],["psb","application/vnd.3gpp.pic-bw-small"],["psd","application/x-photoshop"],["psf","application/x-font-linux-psf"],["pskcxml","application/pskc+xml"],["pti","image/prs.pti"],["ptid","application/vnd.pvi.ptid1"],["pub","application/x-mspublisher"],["pvb","application/vnd.3gpp.pic-bw-var"],["pwn","application/vnd.3m.post-it-notes"],["pya","audio/vnd.ms-playready.media.pya"],["pyv","video/vnd.ms-playready.media.pyv"],["qam","application/vnd.epson.quickanime"],["qbo","application/vnd.intu.qbo"],["qfx","application/vnd.intu.qfx"],["qps","application/vnd.publishare-delta-tree"],["qt","video/quicktime"],["qwd","application/vnd.quark.quarkxpress"],["qwt","application/vnd.quark.quarkxpress"],["qxb","application/vnd.quark.quarkxpress"],["qxd","application/vnd.quark.quarkxpress"],["qxl","application/vnd.quark.quarkxpress"],["qxt","application/vnd.quark.quarkxpress"],["ra","audio/x-realaudio"],["ram","audio/x-pn-realaudio"],["raml","application/raml+yaml"],["rapd","application/route-apd+xml"],["rar","application/x-rar"],["ras","image/x-cmu-raster"],["rcprofile","application/vnd.ipunplugged.rcprofile"],["rdf","application/rdf+xml"],["rdz","application/vnd.data-vision.rdz"],["relo","application/p2p-overlay+xml"],["rep","application/vnd.businessobjects"],["res","application/x-dtbresource+xml"],["rgb","image/x-rgb"],["rif","application/reginfo+xml"],["rip","audio/vnd.rip"],["ris","application/x-research-info-systems"],["rl","application/resource-lists+xml"],["rlc","image/vnd.fujixerox.edmics-rlc"],["rld","application/resource-lists-diff+xml"],["rm","audio/x-pn-realaudio"],["rmi","audio/midi"],["rmp","audio/x-pn-realaudio-plugin"],["rms","application/vnd.jcp.javame.midlet-rms"],["rmvb","application/vnd.rn-realmedia-vbr"],["rnc","application/relax-ng-compact-syntax"],["rng","application/xml"],["roa","application/rpki-roa"],["roff","text/troff"],["rp9","application/vnd.cloanto.rp9"],["rpm","audio/x-pn-realaudio-plugin"],["rpss","application/vnd.nokia.radio-presets"],["rpst","application/vnd.nokia.radio-preset"],["rq","application/sparql-query"],["rs","application/rls-services+xml"],["rsa","application/x-pkcs7"],["rsat","application/atsc-rsat+xml"],["rsd","application/rsd+xml"],["rsheet","application/urc-ressheet+xml"],["rss","application/rss+xml"],["rtf","text/rtf"],["rtx","text/richtext"],["run","application/x-makeself"],["rusd","application/route-usd+xml"],["rv","video/vnd.rn-realvideo"],["s","text/x-asm"],["s3m","audio/s3m"],["saf","application/vnd.yamaha.smaf-audio"],["sass","text/x-sass"],["sbml","application/sbml+xml"],["sc","application/vnd.ibm.secure-container"],["scd","application/x-msschedule"],["scm","application/vnd.lotus-screencam"],["scq","application/scvp-cv-request"],["scs","application/scvp-cv-response"],["scss","text/x-scss"],["scurl","text/vnd.curl.scurl"],["sda","application/vnd.stardivision.draw"],["sdc","application/vnd.stardivision.calc"],["sdd","application/vnd.stardivision.impress"],["sdkd","application/vnd.solent.sdkm+xml"],["sdkm","application/vnd.solent.sdkm+xml"],["sdp","application/sdp"],["sdw","application/vnd.stardivision.writer"],["sea","application/octet-stream"],["see","application/vnd.seemail"],["seed","application/vnd.fdsn.seed"],["sema","application/vnd.sema"],["semd","application/vnd.semd"],["semf","application/vnd.semf"],["senmlx","application/senml+xml"],["sensmlx","application/sensml+xml"],["ser","application/java-serialized-object"],["setpay","application/set-payment-initiation"],["setreg","application/set-registration-initiation"],["sfd-hdstx","application/vnd.hydrostatix.sof-data"],["sfs","application/vnd.spotfire.sfs"],["sfv","text/x-sfv"],["sgi","image/sgi"],["sgl","application/vnd.stardivision.writer-global"],["sgm","text/sgml"],["sgml","text/sgml"],["sh","application/x-sh"],["shar","application/x-shar"],["shex","text/shex"],["shf","application/shf+xml"],["shtml","text/html"],["sid","image/x-mrsid-image"],["sieve","application/sieve"],["sig","application/pgp-signature"],["sil","audio/silk"],["silo","model/mesh"],["sis","application/vnd.symbian.install"],["sisx","application/vnd.symbian.install"],["sit","application/x-stuffit"],["sitx","application/x-stuffitx"],["siv","application/sieve"],["skd","application/vnd.koan"],["skm","application/vnd.koan"],["skp","application/vnd.koan"],["skt","application/vnd.koan"],["sldm","application/vnd.ms-powerpoint.slide.macroenabled.12"],["sldx","application/vnd.openxmlformats-officedocument.presentationml.slide"],["slim","text/slim"],["slm","text/slim"],["sls","application/route-s-tsid+xml"],["slt","application/vnd.epson.salt"],["sm","application/vnd.stepmania.stepchart"],["smf","application/vnd.stardivision.math"],["smi","application/smil"],["smil","application/smil"],["smv","video/x-smv"],["smzip","application/vnd.stepmania.package"],["snd","audio/basic"],["snf","application/x-font-snf"],["so","application/octet-stream"],["spc","application/x-pkcs7-certificates"],["spdx","text/spdx"],["spf","application/vnd.yamaha.smaf-phrase"],["spl","application/x-futuresplash"],["spot","text/vnd.in3d.spot"],["spp","application/scvp-vp-response"],["spq","application/scvp-vp-request"],["spx","audio/ogg"],["sql","application/x-sql"],["src","application/x-wais-source"],["srt","application/x-subrip"],["sru","application/sru+xml"],["srx","application/sparql-results+xml"],["ssdl","application/ssdl+xml"],["sse","application/vnd.kodak-descriptor"],["ssf","application/vnd.epson.ssf"],["ssml","application/ssml+xml"],["sst","application/octet-stream"],["st","application/vnd.sailingtracker.track"],["stc","application/vnd.sun.xml.calc.template"],["std","application/vnd.sun.xml.draw.template"],["stf","application/vnd.wt.stf"],["sti","application/vnd.sun.xml.impress.template"],["stk","application/hyperstudio"],["stl","model/stl"],["stpx","model/step+xml"],["stpxz","model/step-xml+zip"],["stpz","model/step+zip"],["str","application/vnd.pg.format"],["stw","application/vnd.sun.xml.writer.template"],["styl","text/stylus"],["stylus","text/stylus"],["sub","text/vnd.dvb.subtitle"],["sus","application/vnd.sus-calendar"],["susp","application/vnd.sus-calendar"],["sv4cpio","application/x-sv4cpio"],["sv4crc","application/x-sv4crc"],["svc","application/vnd.dvb.service"],["svd","application/vnd.svd"],["svg","image/svg+xml"],["svgz","image/svg+xml"],["swa","application/x-director"],["swf","application/x-shockwave-flash"],["swi","application/vnd.aristanetworks.swi"],["swidtag","application/swid+xml"],["sxc","application/vnd.sun.xml.calc"],["sxd","application/vnd.sun.xml.draw"],["sxg","application/vnd.sun.xml.writer.global"],["sxi","application/vnd.sun.xml.impress"],["sxm","application/vnd.sun.xml.math"],["sxw","application/vnd.sun.xml.writer"],["t","text/troff"],["t3","application/x-t3vm-image"],["t38","image/t38"],["taglet","application/vnd.mynfc"],["tao","application/vnd.tao.intent-module-archive"],["tap","image/vnd.tencent.tap"],["tar","application/x-tar"],["tcap","application/vnd.3gpp2.tcap"],["tcl","application/x-tcl"],["td","application/urc-targetdesc+xml"],["teacher","application/vnd.smart.teacher"],["tei","application/tei+xml"],["teicorpus","application/tei+xml"],["tex","application/x-tex"],["texi","application/x-texinfo"],["texinfo","application/x-texinfo"],["text","text/plain"],["tfi","application/thraud+xml"],["tfm","application/x-tex-tfm"],["tfx","image/tiff-fx"],["tga","image/x-tga"],["tgz","application/x-tar"],["thmx","application/vnd.ms-officetheme"],["tif","image/tiff"],["tiff","image/tiff"],["tk","application/x-tcl"],["tmo","application/vnd.tmobile-livetv"],["toml","application/toml"],["torrent","application/x-bittorrent"],["tpl","application/vnd.groove-tool-template"],["tpt","application/vnd.trid.tpt"],["tr","text/troff"],["tra","application/vnd.trueapp"],["trig","application/trig"],["trm","application/x-msterminal"],["ts","video/mp2t"],["tsd","application/timestamped-data"],["tsv","text/tab-separated-values"],["ttc","font/collection"],["ttf","font/ttf"],["ttl","text/turtle"],["ttml","application/ttml+xml"],["twd","application/vnd.simtech-mindmapper"],["twds","application/vnd.simtech-mindmapper"],["txd","application/vnd.genomatix.tuxedo"],["txf","application/vnd.mobius.txf"],["txt","text/plain"],["u8dsn","message/global-delivery-status"],["u8hdr","message/global-headers"],["u8mdn","message/global-disposition-notification"],["u8msg","message/global"],["u32","application/x-authorware-bin"],["ubj","application/ubjson"],["udeb","application/x-debian-package"],["ufd","application/vnd.ufdl"],["ufdl","application/vnd.ufdl"],["ulx","application/x-glulx"],["umj","application/vnd.umajin"],["unityweb","application/vnd.unity"],["uoml","application/vnd.uoml+xml"],["uri","text/uri-list"],["uris","text/uri-list"],["urls","text/uri-list"],["usdz","model/vnd.usdz+zip"],["ustar","application/x-ustar"],["utz","application/vnd.uiq.theme"],["uu","text/x-uuencode"],["uva","audio/vnd.dece.audio"],["uvd","application/vnd.dece.data"],["uvf","application/vnd.dece.data"],["uvg","image/vnd.dece.graphic"],["uvh","video/vnd.dece.hd"],["uvi","image/vnd.dece.graphic"],["uvm","video/vnd.dece.mobile"],["uvp","video/vnd.dece.pd"],["uvs","video/vnd.dece.sd"],["uvt","application/vnd.dece.ttml+xml"],["uvu","video/vnd.uvvu.mp4"],["uvv","video/vnd.dece.video"],["uvva","audio/vnd.dece.audio"],["uvvd","application/vnd.dece.data"],["uvvf","application/vnd.dece.data"],["uvvg","image/vnd.dece.graphic"],["uvvh","video/vnd.dece.hd"],["uvvi","image/vnd.dece.graphic"],["uvvm","video/vnd.dece.mobile"],["uvvp","video/vnd.dece.pd"],["uvvs","video/vnd.dece.sd"],["uvvt","application/vnd.dece.ttml+xml"],["uvvu","video/vnd.uvvu.mp4"],["uvvv","video/vnd.dece.video"],["uvvx","application/vnd.dece.unspecified"],["uvvz","application/vnd.dece.zip"],["uvx","application/vnd.dece.unspecified"],["uvz","application/vnd.dece.zip"],["vbox","application/x-virtualbox-vbox"],["vbox-extpack","application/x-virtualbox-vbox-extpack"],["vcard","text/vcard"],["vcd","application/x-cdlink"],["vcf","text/x-vcard"],["vcg","application/vnd.groove-vcard"],["vcs","text/x-vcalendar"],["vcx","application/vnd.vcx"],["vdi","application/x-virtualbox-vdi"],["vds","model/vnd.sap.vds"],["vhd","application/x-virtualbox-vhd"],["vis","application/vnd.visionary"],["viv","video/vnd.vivo"],["vlc","application/videolan"],["vmdk","application/x-virtualbox-vmdk"],["vob","video/x-ms-vob"],["vor","application/vnd.stardivision.writer"],["vox","application/x-authorware-bin"],["vrml","model/vrml"],["vsd","application/vnd.visio"],["vsf","application/vnd.vsf"],["vss","application/vnd.visio"],["vst","application/vnd.visio"],["vsw","application/vnd.visio"],["vtf","image/vnd.valve.source.texture"],["vtt","text/vtt"],["vtu","model/vnd.vtu"],["vxml","application/voicexml+xml"],["w3d","application/x-director"],["wad","application/x-doom"],["wadl","application/vnd.sun.wadl+xml"],["war","application/java-archive"],["wasm","application/wasm"],["wav","audio/x-wav"],["wax","audio/x-ms-wax"],["wbmp","image/vnd.wap.wbmp"],["wbs","application/vnd.criticaltools.wbs+xml"],["wbxml","application/wbxml"],["wcm","application/vnd.ms-works"],["wdb","application/vnd.ms-works"],["wdp","image/vnd.ms-photo"],["weba","audio/webm"],["webapp","application/x-web-app-manifest+json"],["webm","video/webm"],["webmanifest","application/manifest+json"],["webp","image/webp"],["wg","application/vnd.pmi.widget"],["wgt","application/widget"],["wks","application/vnd.ms-works"],["wm","video/x-ms-wm"],["wma","audio/x-ms-wma"],["wmd","application/x-ms-wmd"],["wmf","image/wmf"],["wml","text/vnd.wap.wml"],["wmlc","application/wmlc"],["wmls","text/vnd.wap.wmlscript"],["wmlsc","application/vnd.wap.wmlscriptc"],["wmv","video/x-ms-wmv"],["wmx","video/x-ms-wmx"],["wmz","application/x-msmetafile"],["woff","font/woff"],["woff2","font/woff2"],["word","application/msword"],["wpd","application/vnd.wordperfect"],["wpl","application/vnd.ms-wpl"],["wps","application/vnd.ms-works"],["wqd","application/vnd.wqd"],["wri","application/x-mswrite"],["wrl","model/vrml"],["wsc","message/vnd.wfa.wsc"],["wsdl","application/wsdl+xml"],["wspolicy","application/wspolicy+xml"],["wtb","application/vnd.webturbo"],["wvx","video/x-ms-wvx"],["x3d","model/x3d+xml"],["x3db","model/x3d+fastinfoset"],["x3dbz","model/x3d+binary"],["x3dv","model/x3d-vrml"],["x3dvz","model/x3d+vrml"],["x3dz","model/x3d+xml"],["x32","application/x-authorware-bin"],["x_b","model/vnd.parasolid.transmit.binary"],["x_t","model/vnd.parasolid.transmit.text"],["xaml","application/xaml+xml"],["xap","application/x-silverlight-app"],["xar","application/vnd.xara"],["xav","application/xcap-att+xml"],["xbap","application/x-ms-xbap"],["xbd","application/vnd.fujixerox.docuworks.binder"],["xbm","image/x-xbitmap"],["xca","application/xcap-caps+xml"],["xcs","application/calendar+xml"],["xdf","application/xcap-diff+xml"],["xdm","application/vnd.syncml.dm+xml"],["xdp","application/vnd.adobe.xdp+xml"],["xdssc","application/dssc+xml"],["xdw","application/vnd.fujixerox.docuworks"],["xel","application/xcap-el+xml"],["xenc","application/xenc+xml"],["xer","application/patch-ops-error+xml"],["xfdf","application/vnd.adobe.xfdf"],["xfdl","application/vnd.xfdl"],["xht","application/xhtml+xml"],["xhtml","application/xhtml+xml"],["xhvml","application/xv+xml"],["xif","image/vnd.xiff"],["xl","application/excel"],["xla","application/vnd.ms-excel"],["xlam","application/vnd.ms-excel.addin.macroEnabled.12"],["xlc","application/vnd.ms-excel"],["xlf","application/xliff+xml"],["xlm","application/vnd.ms-excel"],["xls","application/vnd.ms-excel"],["xlsb","application/vnd.ms-excel.sheet.binary.macroEnabled.12"],["xlsm","application/vnd.ms-excel.sheet.macroEnabled.12"],["xlsx","application/vnd.openxmlformats-officedocument.spreadsheetml.sheet"],["xlt","application/vnd.ms-excel"],["xltm","application/vnd.ms-excel.template.macroEnabled.12"],["xltx","application/vnd.openxmlformats-officedocument.spreadsheetml.template"],["xlw","application/vnd.ms-excel"],["xm","audio/xm"],["xml","application/xml"],["xns","application/xcap-ns+xml"],["xo","application/vnd.olpc-sugar"],["xop","application/xop+xml"],["xpi","application/x-xpinstall"],["xpl","application/xproc+xml"],["xpm","image/x-xpixmap"],["xpr","application/vnd.is-xpr"],["xps","application/vnd.ms-xpsdocument"],["xpw","application/vnd.intercon.formnet"],["xpx","application/vnd.intercon.formnet"],["xsd","application/xml"],["xsl","application/xml"],["xslt","application/xslt+xml"],["xsm","application/vnd.syncml+xml"],["xspf","application/xspf+xml"],["xul","application/vnd.mozilla.xul+xml"],["xvm","application/xv+xml"],["xvml","application/xv+xml"],["xwd","image/x-xwindowdump"],["xyz","chemical/x-xyz"],["xz","application/x-xz"],["yaml","text/yaml"],["yang","application/yang"],["yin","application/yin+xml"],["yml","text/yaml"],["ymp","text/x-suse-ymp"],["z","application/x-compress"],["z1","application/x-zmachine"],["z2","application/x-zmachine"],["z3","application/x-zmachine"],["z4","application/x-zmachine"],["z5","application/x-zmachine"],["z6","application/x-zmachine"],["z7","application/x-zmachine"],["z8","application/x-zmachine"],["zaz","application/vnd.zzazz.deck+xml"],["zip","application/zip"],["zir","application/vnd.zul"],["zirz","application/vnd.zul"],["zmm","application/vnd.handheld-entertainment+xml"],["zsh","text/x-scriptzsh"]]);function Fs(e,t,n){const r=xoe(e),{webkitRelativePath:a}=e,o=typeof t=="string"?t:typeof a=="string"&&a.length>0?a:`./${e.name}`;return typeof r.path!="string"&&MO(r,"path",o),MO(r,"relativePath",o),r}function xoe(e){const{name:t}=e;if(t&&t.lastIndexOf(".")!==-1&&!e.type){const r=t.split(".").pop().toLowerCase(),a=woe.get(r);a&&Object.defineProperty(e,"type",{value:a,writable:!1,configurable:!1,enumerable:!0})}return e}function MO(e,t,n){Object.defineProperty(e,t,{value:n,writable:!1,configurable:!1,enumerable:!0})}const koe=[".DS_Store","Thumbs.db"];function Toe(e){return xi(this,void 0,void 0,function*(){return Nf(e)&&Aoe(e.dataTransfer)?Noe(e.dataTransfer,e.type):Roe(e)?_oe(e):Array.isArray(e)&&e.every(t=>"getFile"in t&&typeof t.getFile=="function")?Coe(e):[]})}function Aoe(e){return Nf(e)}function Roe(e){return Nf(e)&&Nf(e.target)}function Nf(e){return typeof e=="object"&&e!==null}function _oe(e){return n0(e.target.files).map(t=>Fs(t))}function Coe(e){return xi(this,void 0,void 0,function*(){return(yield Promise.all(e.map(n=>n.getFile()))).map(n=>Fs(n))})}function Noe(e,t){return xi(this,void 0,void 0,function*(){if(e.items){const n=n0(e.items).filter(a=>a.kind==="file");if(t!=="drop")return n;const r=yield Promise.all(n.map(Ooe));return FO(U5(r))}return FO(n0(e.files).map(n=>Fs(n)))})}function FO(e){return e.filter(t=>koe.indexOf(t.name)===-1)}function n0(e){if(e===null)return[];const t=[];for(let n=0;n<e.length;n++){const r=e[n];t.push(r)}return t}function Ooe(e){if(typeof e.webkitGetAsEntry!="function")return PO(e);const t=e.webkitGetAsEntry();return t&&t.isDirectory?j5(t):PO(e,t)}function U5(e){return e.reduce((t,n)=>[...t,...Array.isArray(n)?U5(n):[n]],[])}function PO(e,t){return xi(this,void 0,void 0,function*(){var n;if(globalThis.isSecureContext&&typeof e.getAsFileSystemHandle=="function"){const o=yield e.getAsFileSystemHandle();if(o===null)throw new Error(`${e} is not a File`);if(o!==void 0){const s=yield o.getFile();return s.handle=o,Fs(s)}}const r=e.getAsFile();if(!r)throw new Error(`${e} is not a File`);return Fs(r,(n=t==null?void 0:t.fullPath)!==null&&n!==void 0?n:void 0)})}function Doe(e){return xi(this,void 0,void 0,function*(){return e.isDirectory?j5(e):Ioe(e)})}function j5(e){const t=e.createReader();return new Promise((n,r)=>{const a=[];function o(){t.readEntries(s=>xi(this,void 0,void 0,function*(){if(s.length){const u=Promise.all(s.map(Doe));a.push(u),o()}else try{const u=yield Promise.all(a);n(u)}catch(u){r(u)}}),s=>{r(s)})}o()})}function Ioe(e){return xi(this,void 0,void 0,function*(){return new Promise((t,n)=>{e.file(r=>{const a=Fs(r,e.fullPath);t(a)},r=>{n(r)})})})}var Fd={},zO;function Loe(){return zO||(zO=1,Fd.__esModule=!0,Fd.default=function(e,t){if(e&&t){var n=Array.isArray(t)?t:t.split(",");if(n.length===0)return!0;var r=e.name||"",a=(e.type||"").toLowerCase(),o=a.replace(/\/.*$/,"");return n.some(function(s){var u=s.trim().toLowerCase();return u.charAt(0)==="."?r.toLowerCase().endsWith(u):u.endsWith("/*")?o===u.replace(/\/.*$/,""):a===u})}return!0}),Fd}var Moe=Loe();const sb=un(Moe);function BO(e){return zoe(e)||Poe(e)||H5(e)||Foe()}function Foe(){throw new TypeError(`Invalid attempt to spread non-iterable instance.
In order to be iterable, non-array objects must have a [Symbol.iterator]() method.`)}function Poe(e){if(typeof Symbol<"u"&&e[Symbol.iterator]!=null||e["@@iterator"]!=null)return Array.from(e)}function zoe(e){if(Array.isArray(e))return r0(e)}function UO(e,t){var n=Object.keys(e);if(Object.getOwnPropertySymbols){var r=Object.getOwnPropertySymbols(e);t&&(r=r.filter(function(a){return Object.getOwnPropertyDescriptor(e,a).enumerable})),n.push.apply(n,r)}return n}function jO(e){for(var t=1;t<arguments.length;t++){var n=arguments[t]!=null?arguments[t]:{};t%2?UO(Object(n),!0).forEach(function(r){G5(e,r,n[r])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(n)):UO(Object(n)).forEach(function(r){Object.defineProperty(e,r,Object.getOwnPropertyDescriptor(n,r))})}return e}function G5(e,t,n){return t in e?Object.defineProperty(e,t,{value:n,enumerable:!0,configurable:!0,writable:!0}):e[t]=n,e}function Iu(e,t){return joe(e)||Uoe(e,t)||H5(e,t)||Boe()}function Boe(){throw new TypeError(`Invalid attempt to destructure non-iterable instance.
In order to be iterable, non-array objects must have a [Symbol.iterator]() method.`)}function H5(e,t){if(e){if(typeof e=="string")return r0(e,t);var n=Object.prototype.toString.call(e).slice(8,-1);if(n==="Object"&&e.constructor&&(n=e.constructor.name),n==="Map"||n==="Set")return Array.from(e);if(n==="Arguments"||/^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n))return r0(e,t)}}function r0(e,t){(t==null||t>e.length)&&(t=e.length);for(var n=0,r=new Array(t);n<t;n++)r[n]=e[n];return r}function Uoe(e,t){var n=e==null?null:typeof Symbol<"u"&&e[Symbol.iterator]||e["@@iterator"];if(n!=null){var r=[],a=!0,o=!1,s,u;try{for(n=n.call(e);!(a=(s=n.next()).done)&&(r.push(s.value),!(t&&r.length===t));a=!0);}catch(c){o=!0,u=c}finally{try{!a&&n.return!=null&&n.return()}finally{if(o)throw u}}return r}}function joe(e){if(Array.isArray(e))return e}var Goe=typeof sb=="function"?sb:sb.default,Hoe="file-invalid-type",$oe="file-too-large",qoe="file-too-small",Voe="too-many-files",Woe=function(){var t=arguments.length>0&&arguments[0]!==void 0?arguments[0]:"",n=t.split(","),r=n.length>1?"one of ".concat(n.join(", ")):n[0];return{code:Hoe,message:"File type must be ".concat(r)}},GO=function(t){return{code:$oe,message:"File is larger than ".concat(t," ").concat(t===1?"byte":"bytes")}},HO=function(t){return{code:qoe,message:"File is smaller than ".concat(t," ").concat(t===1?"byte":"bytes")}},Yoe={code:Voe,message:"Too many files"};function $5(e,t){var n=e.type==="application/x-moz-file"||Goe(e,t);return[n,n?null:Woe(t)]}function q5(e,t,n){if(ui(e.size))if(ui(t)&&ui(n)){if(e.size>n)return[!1,GO(n)];if(e.size<t)return[!1,HO(t)]}else{if(ui(t)&&e.size<t)return[!1,HO(t)];if(ui(n)&&e.size>n)return[!1,GO(n)]}return[!0,null]}function ui(e){return e!=null}function Koe(e){var t=e.files,n=e.accept,r=e.minSize,a=e.maxSize,o=e.multiple,s=e.maxFiles,u=e.validator;return!o&&t.length>1||o&&s>=1&&t.length>s?!1:t.every(function(c){var d=$5(c,n),p=Iu(d,1),g=p[0],m=q5(c,r,a),b=Iu(m,1),y=b[0],S=u?u(c):null;return g&&y&&!S})}function Of(e){return typeof e.isPropagationStopped=="function"?e.isPropagationStopped():typeof e.cancelBubble<"u"?e.cancelBubble:!1}function Pd(e){return e.dataTransfer?Array.prototype.some.call(e.dataTransfer.types,function(t){return t==="Files"||t==="application/x-moz-file"}):!!e.target&&!!e.target.files}function $O(e){e.preventDefault()}function Xoe(e){return e.indexOf("MSIE")!==-1||e.indexOf("Trident/")!==-1}function Zoe(e){return e.indexOf("Edge/")!==-1}function Qoe(){var e=arguments.length>0&&arguments[0]!==void 0?arguments[0]:window.navigator.userAgent;return Xoe(e)||Zoe(e)}function Qr(){for(var e=arguments.length,t=new Array(e),n=0;n<e;n++)t[n]=arguments[n];return function(r){for(var a=arguments.length,o=new Array(a>1?a-1:0),s=1;s<a;s++)o[s-1]=arguments[s];return t.some(function(u){return!Of(r)&&u&&u.apply(void 0,[r].concat(o)),Of(r)})}}function Joe(){return"showOpenFilePicker"in window}function eie(e){if(ui(e)){var t=Object.entries(e).filter(function(n){var r=Iu(n,2),a=r[0],o=r[1],s=!0;return V5(a)||(console.warn('Skipped "'.concat(a,'" because it is not a valid MIME type. Check https://developer.mozilla.org/en-US/docs/Web/HTTP/Basics_of_HTTP/MIME_types/Common_types for a list of valid MIME types.')),s=!1),(!Array.isArray(o)||!o.every(W5))&&(console.warn('Skipped "'.concat(a,'" because an invalid file extension was provided.')),s=!1),s}).reduce(function(n,r){var a=Iu(r,2),o=a[0],s=a[1];return jO(jO({},n),{},G5({},o,s))},{});return[{description:"Files",accept:t}]}return e}function tie(e){if(ui(e))return Object.entries(e).reduce(function(t,n){var r=Iu(n,2),a=r[0],o=r[1];return[].concat(BO(t),[a],BO(o))},[]).filter(function(t){return V5(t)||W5(t)}).join(",")}function nie(e){return e instanceof DOMException&&(e.name==="AbortError"||e.code===e.ABORT_ERR)}function rie(e){return e instanceof DOMException&&(e.name==="SecurityError"||e.code===e.SECURITY_ERR)}function V5(e){return e==="audio/*"||e==="video/*"||e==="image/*"||e==="text/*"||e==="application/*"||/\w+\/[-+.\w]+/g.test(e)}function W5(e){return/^.*\.[\w]+$/.test(e)}var aie=["children"],oie=["open"],iie=["refKey","role","onKeyDown","onFocus","onBlur","onClick","onDragEnter","onDragOver","onDragLeave","onDrop"],sie=["refKey","onChange","onClick"];function lie(e){return die(e)||cie(e)||Y5(e)||uie()}function uie(){throw new TypeError(`Invalid attempt to spread non-iterable instance.
In order to be iterable, non-array objects must have a [Symbol.iterator]() method.`)}function cie(e){if(typeof Symbol<"u"&&e[Symbol.iterator]!=null||e["@@iterator"]!=null)return Array.from(e)}function die(e){if(Array.isArray(e))return a0(e)}function lb(e,t){return gie(e)||pie(e,t)||Y5(e,t)||fie()}function fie(){throw new TypeError(`Invalid attempt to destructure non-iterable instance.
In order to be iterable, non-array objects must have a [Symbol.iterator]() method.`)}function Y5(e,t){if(e){if(typeof e=="string")return a0(e,t);var n=Object.prototype.toString.call(e).slice(8,-1);if(n==="Object"&&e.constructor&&(n=e.constructor.name),n==="Map"||n==="Set")return Array.from(e);if(n==="Arguments"||/^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n))return a0(e,t)}}function a0(e,t){(t==null||t>e.length)&&(t=e.length);for(var n=0,r=new Array(t);n<t;n++)r[n]=e[n];return r}function pie(e,t){var n=e==null?null:typeof Symbol<"u"&&e[Symbol.iterator]||e["@@iterator"];if(n!=null){var r=[],a=!0,o=!1,s,u;try{for(n=n.call(e);!(a=(s=n.next()).done)&&(r.push(s.value),!(t&&r.length===t));a=!0);}catch(c){o=!0,u=c}finally{try{!a&&n.return!=null&&n.return()}finally{if(o)throw u}}return r}}function gie(e){if(Array.isArray(e))return e}function qO(e,t){var n=Object.keys(e);if(Object.getOwnPropertySymbols){var r=Object.getOwnPropertySymbols(e);t&&(r=r.filter(function(a){return Object.getOwnPropertyDescriptor(e,a).enumerable})),n.push.apply(n,r)}return n}function jt(e){for(var t=1;t<arguments.length;t++){var n=arguments[t]!=null?arguments[t]:{};t%2?qO(Object(n),!0).forEach(function(r){o0(e,r,n[r])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(n)):qO(Object(n)).forEach(function(r){Object.defineProperty(e,r,Object.getOwnPropertyDescriptor(n,r))})}return e}function o0(e,t,n){return t in e?Object.defineProperty(e,t,{value:n,enumerable:!0,configurable:!0,writable:!0}):e[t]=n,e}function Df(e,t){if(e==null)return{};var n=hie(e,t),r,a;if(Object.getOwnPropertySymbols){var o=Object.getOwnPropertySymbols(e);for(a=0;a<o.length;a++)r=o[a],!(t.indexOf(r)>=0)&&Object.prototype.propertyIsEnumerable.call(e,r)&&(n[r]=e[r])}return n}function hie(e,t){if(e==null)return{};var n={},r=Object.keys(e),a,o;for(o=0;o<r.length;o++)a=r[o],!(t.indexOf(a)>=0)&&(n[a]=e[a]);return n}var Ap=E.forwardRef(function(e,t){var n=e.children,r=Df(e,aie),a=mie(r),o=a.open,s=Df(a,oie);return E.useImperativeHandle(t,function(){return{open:o}},[o]),ye.createElement(E.Fragment,null,n(jt(jt({},s),{},{open:o})))});Ap.displayName="Dropzone";var K5={disabled:!1,getFilesFromEvent:Toe,maxSize:1/0,minSize:0,multiple:!0,maxFiles:0,preventDropOnDocument:!0,noClick:!1,noKeyboard:!1,noDrag:!1,noDragEventsBubbling:!1,validator:null,useFsAccessApi:!1,autoFocus:!1};Ap.defaultProps=K5;Ap.propTypes={children:Ot.func,accept:Ot.objectOf(Ot.arrayOf(Ot.string)),multiple:Ot.bool,preventDropOnDocument:Ot.bool,noClick:Ot.bool,noKeyboard:Ot.bool,noDrag:Ot.bool,noDragEventsBubbling:Ot.bool,minSize:Ot.number,maxSize:Ot.number,maxFiles:Ot.number,disabled:Ot.bool,getFilesFromEvent:Ot.func,onFileDialogCancel:Ot.func,onFileDialogOpen:Ot.func,useFsAccessApi:Ot.bool,autoFocus:Ot.bool,onDragEnter:Ot.func,onDragLeave:Ot.func,onDragOver:Ot.func,onDrop:Ot.func,onDropAccepted:Ot.func,onDropRejected:Ot.func,onError:Ot.func,validator:Ot.func};var i0={isFocused:!1,isFileDialogActive:!1,isDragActive:!1,isDragAccept:!1,isDragReject:!1,acceptedFiles:[],fileRejections:[]};function mie(){var e=arguments.length>0&&arguments[0]!==void 0?arguments[0]:{},t=jt(jt({},K5),e),n=t.accept,r=t.disabled,a=t.getFilesFromEvent,o=t.maxSize,s=t.minSize,u=t.multiple,c=t.maxFiles,d=t.onDragEnter,p=t.onDragLeave,g=t.onDragOver,m=t.onDrop,b=t.onDropAccepted,y=t.onDropRejected,S=t.onFileDialogCancel,x=t.onFileDialogOpen,R=t.useFsAccessApi,k=t.autoFocus,A=t.preventDropOnDocument,C=t.noClick,N=t.noKeyboard,_=t.noDrag,O=t.noDragEventsBubbling,F=t.onError,L=t.validator,I=E.useMemo(function(){return tie(n)},[n]),H=E.useMemo(function(){return eie(n)},[n]),$=E.useMemo(function(){return typeof x=="function"?x:VO},[x]),U=E.useMemo(function(){return typeof S=="function"?S:VO},[S]),Y=E.useRef(null),Z=E.useRef(null),j=E.useReducer(bie,i0),G=lb(j,2),z=G[0],K=G[1],D=z.isFocused,V=z.isFileDialogActive,B=E.useRef(typeof window<"u"&&window.isSecureContext&&R&&Joe()),M=function(){!B.current&&V&&setTimeout(function(){if(Z.current){var Ne=Z.current.files;Ne.length||(K({type:"closeDialog"}),U())}},300)};E.useEffect(function(){return window.addEventListener("focus",M,!1),function(){window.removeEventListener("focus",M,!1)}},[Z,V,U,B]);var W=E.useRef([]),Q=function(Ne){Y.current&&Y.current.contains(Ne.target)||(Ne.preventDefault(),W.current=[])};E.useEffect(function(){return A&&(document.addEventListener("dragover",$O,!1),document.addEventListener("drop",Q,!1)),function(){A&&(document.removeEventListener("dragover",$O),document.removeEventListener("drop",Q))}},[Y,A]),E.useEffect(function(){return!r&&k&&Y.current&&Y.current.focus(),function(){}},[Y,k,r]);var oe=E.useCallback(function(he){F?F(he):console.error(he)},[F]),re=E.useCallback(function(he){he.preventDefault(),he.persist(),Ee(he),W.current=[].concat(lie(W.current),[he.target]),Pd(he)&&Promise.resolve(a(he)).then(function(Ne){if(!(Of(he)&&!O)){var ee=Ne.length,ce=ee>0&&Koe({files:Ne,accept:I,minSize:s,maxSize:o,multiple:u,maxFiles:c,validator:L}),_e=ee>0&&!ce;K({isDragAccept:ce,isDragReject:_e,isDragActive:!0,type:"setDraggedFiles"}),d&&d(he)}}).catch(function(Ne){return oe(Ne)})},[a,d,oe,O,I,s,o,u,c,L]),ie=E.useCallback(function(he){he.preventDefault(),he.persist(),Ee(he);var Ne=Pd(he);if(Ne&&he.dataTransfer)try{he.dataTransfer.dropEffect="copy"}catch{}return Ne&&g&&g(he),!1},[g,O]),Se=E.useCallback(function(he){he.preventDefault(),he.persist(),Ee(he);var Ne=W.current.filter(function(ce){return Y.current&&Y.current.contains(ce)}),ee=Ne.indexOf(he.target);ee!==-1&&Ne.splice(ee,1),W.current=Ne,!(Ne.length>0)&&(K({type:"setDraggedFiles",isDragActive:!1,isDragAccept:!1,isDragReject:!1}),Pd(he)&&p&&p(he))},[Y,p,O]),ae=E.useCallback(function(he,Ne){var ee=[],ce=[];he.forEach(function(_e){var Pe=$5(_e,I),We=lb(Pe,2),yt=We[0],kt=We[1],ht=q5(_e,s,o),et=lb(ht,2),Tt=et[0],ot=et[1],St=L?L(_e):null;if(yt&&Tt&&!St)ee.push(_e);else{var Ht=[kt,ot];St&&(Ht=Ht.concat(St)),ce.push({file:_e,errors:Ht.filter(function(fn){return fn})})}}),(!u&&ee.length>1||u&&c>=1&&ee.length>c)&&(ee.forEach(function(_e){ce.push({file:_e,errors:[Yoe]})}),ee.splice(0)),K({acceptedFiles:ee,fileRejections:ce,isDragReject:ce.length>0,type:"setFiles"}),m&&m(ee,ce,Ne),ce.length>0&&y&&y(ce,Ne),ee.length>0&&b&&b(ee,Ne)},[K,u,I,s,o,c,m,b,y,L]),ve=E.useCallback(function(he){he.preventDefault(),he.persist(),Ee(he),W.current=[],Pd(he)&&Promise.resolve(a(he)).then(function(Ne){Of(he)&&!O||ae(Ne,he)}).catch(function(Ne){return oe(Ne)}),K({type:"reset"})},[a,ae,oe,O]),xe=E.useCallback(function(){if(B.current){K({type:"openDialog"}),$();var he={multiple:u,types:H};window.showOpenFilePicker(he).then(function(Ne){return a(Ne)}).then(function(Ne){ae(Ne,null),K({type:"closeDialog"})}).catch(function(Ne){nie(Ne)?(U(Ne),K({type:"closeDialog"})):rie(Ne)?(B.current=!1,Z.current?(Z.current.value=null,Z.current.click()):oe(new Error("Cannot open the file picker because the https://developer.mozilla.org/en-US/docs/Web/API/File_System_Access_API is not supported and no <input> was provided."))):oe(Ne)});return}Z.current&&(K({type:"openDialog"}),$(),Z.current.value=null,Z.current.click())},[K,$,U,R,ae,oe,H,u]),Ie=E.useCallback(function(he){!Y.current||!Y.current.isEqualNode(he.target)||(he.key===" "||he.key==="Enter"||he.keyCode===32||he.keyCode===13)&&(he.preventDefault(),xe())},[Y,xe]),Ce=E.useCallback(function(){K({type:"focus"})},[]),ke=E.useCallback(function(){K({type:"blur"})},[]),J=E.useCallback(function(){C||(Qoe()?setTimeout(xe,0):xe())},[C,xe]),fe=function(Ne){return r?null:Ne},Te=function(Ne){return N?null:fe(Ne)},me=function(Ne){return _?null:fe(Ne)},Ee=function(Ne){O&&Ne.stopPropagation()},le=E.useMemo(function(){return function(){var he=arguments.length>0&&arguments[0]!==void 0?arguments[0]:{},Ne=he.refKey,ee=Ne===void 0?"ref":Ne,ce=he.role,_e=he.onKeyDown,Pe=he.onFocus,We=he.onBlur,yt=he.onClick,kt=he.onDragEnter,ht=he.onDragOver,et=he.onDragLeave,Tt=he.onDrop,ot=Df(he,iie);return jt(jt(o0({onKeyDown:Te(Qr(_e,Ie)),onFocus:Te(Qr(Pe,Ce)),onBlur:Te(Qr(We,ke)),onClick:fe(Qr(yt,J)),onDragEnter:me(Qr(kt,re)),onDragOver:me(Qr(ht,ie)),onDragLeave:me(Qr(et,Se)),onDrop:me(Qr(Tt,ve)),role:typeof ce=="string"&&ce!==""?ce:"presentation"},ee,Y),!r&&!N?{tabIndex:0}:{}),ot)}},[Y,Ie,Ce,ke,J,re,ie,Se,ve,N,_,r]),Be=E.useCallback(function(he){he.stopPropagation()},[]),Ue=E.useMemo(function(){return function(){var he=arguments.length>0&&arguments[0]!==void 0?arguments[0]:{},Ne=he.refKey,ee=Ne===void 0?"ref":Ne,ce=he.onChange,_e=he.onClick,Pe=Df(he,sie),We=o0({accept:I,multiple:u,type:"file",style:{border:0,clip:"rect(0, 0, 0, 0)",clipPath:"inset(50%)",height:"1px",margin:"0 -1px -1px 0",overflow:"hidden",padding:0,position:"absolute",width:"1px",whiteSpace:"nowrap"},onChange:fe(Qr(ce,ve)),onClick:fe(Qr(_e,Be)),tabIndex:-1},ee,Z);return jt(jt({},We),Pe)}},[Z,n,u,ve,r]);return jt(jt({},z),{},{isFocused:D&&!r,getRootProps:le,getInputProps:Ue,rootRef:Y,inputRef:Z,open:fe(xe)})}function bie(e,t){switch(t.type){case"focus":return jt(jt({},e),{},{isFocused:!0});case"blur":return jt(jt({},e),{},{isFocused:!1});case"openDialog":return jt(jt({},i0),{},{isFileDialogActive:!0});case"closeDialog":return jt(jt({},e),{},{isFileDialogActive:!1});case"setDraggedFiles":return jt(jt({},e),{},{isDragActive:t.isDragActive,isDragAccept:t.isDragAccept,isDragReject:t.isDragReject});case"setFiles":return jt(jt({},e),{},{acceptedFiles:t.acceptedFiles,fileRejections:t.fileRejections,isDragReject:t.isDragReject});case"reset":return jt({},i0);default:return e}}function VO(){}var MT="Progress",FT=100,[yie,Pxe]=wr(MT),[vie,Sie]=yie(MT),X5=E.forwardRef((e,t)=>{const{__scopeProgress:n,value:r=null,max:a,getValueLabel:o=Eie,...s}=e;(a||a===0)&&!WO(a)&&console.error(wie(`${a}`,"Progress"));const u=WO(a)?a:FT;r!==null&&!YO(r,u)&&console.error(xie(`${r}`,"Progress"));const c=YO(r,u)?r:null,d=If(c)?o(c,u):void 0;return w.jsx(vie,{scope:n,value:c,max:u,children:w.jsx(Ze.div,{"aria-valuemax":u,"aria-valuemin":0,"aria-valuenow":If(c)?c:void 0,"aria-valuetext":d,role:"progressbar","data-state":J5(c,u),"data-value":c??void 0,"data-max":u,...s,ref:t})})});X5.displayName=MT;var Z5="ProgressIndicator",Q5=E.forwardRef((e,t)=>{const{__scopeProgress:n,...r}=e,a=Sie(Z5,n);return w.jsx(Ze.div,{"data-state":J5(a.value,a.max),"data-value":a.value??void 0,"data-max":a.max,...r,ref:t})});Q5.displayName=Z5;function Eie(e,t){return`${Math.round(e/t*100)}%`}function J5(e,t){return e==null?"indeterminate":e===t?"complete":"loading"}function If(e){return typeof e=="number"}function WO(e){return If(e)&&!isNaN(e)&&e>0}function YO(e,t){return If(e)&&!isNaN(e)&&e<=t&&e>=0}function wie(e,t){return`Invalid prop \`max\` of value \`${e}\` supplied to \`${t}\`. Only numbers greater than 0 are valid max values. Defaulting to \`${FT}\`.`}function xie(e,t){return`Invalid prop \`value\` of value \`${e}\` supplied to \`${t}\`. The \`value\` prop must be:
  - a positive number
  - less than the value passed to \`max\` (or ${FT} if no \`max\` prop is set)
  - \`null\` or \`undefined\` if the progress is indeterminate.

Defaulting to \`null\`.`}var eG=X5,kie=Q5;const tG=E.forwardRef(({className:e,value:t,...n},r)=>w.jsx(eG,{ref:r,className:Me("bg-secondary relative h-4 w-full overflow-hidden rounded-full",e),...n,children:w.jsx(kie,{className:"bg-primary h-full w-full flex-1 transition-all",style:{transform:`translateX(-${100-(t||0)}%)`}})}));tG.displayName=eG.displayName;function Tie(e,t){return E.useReducer((n,r)=>t[n][r]??n,e)}var PT="ScrollArea",[nG,zxe]=wr(PT),[Aie,Tr]=nG(PT),rG=E.forwardRef((e,t)=>{const{__scopeScrollArea:n,type:r="hover",dir:a,scrollHideDelay:o=600,...s}=e,[u,c]=E.useState(null),[d,p]=E.useState(null),[g,m]=E.useState(null),[b,y]=E.useState(null),[S,x]=E.useState(null),[R,k]=E.useState(0),[A,C]=E.useState(0),[N,_]=E.useState(!1),[O,F]=E.useState(!1),L=gt(t,H=>c(H)),I=gp(a);return w.jsx(Aie,{scope:n,type:r,dir:I,scrollHideDelay:o,scrollArea:u,viewport:d,onViewportChange:p,content:g,onContentChange:m,scrollbarX:b,onScrollbarXChange:y,scrollbarXEnabled:N,onScrollbarXEnabledChange:_,scrollbarY:S,onScrollbarYChange:x,scrollbarYEnabled:O,onScrollbarYEnabledChange:F,onCornerWidthChange:k,onCornerHeightChange:C,children:w.jsx(Ze.div,{dir:I,...s,ref:L,style:{position:"relative","--radix-scroll-area-corner-width":R+"px","--radix-scroll-area-corner-height":A+"px",...e.style}})})});rG.displayName=PT;var aG="ScrollAreaViewport",oG=E.forwardRef((e,t)=>{const{__scopeScrollArea:n,children:r,nonce:a,...o}=e,s=Tr(aG,n),u=E.useRef(null),c=gt(t,u,s.onViewportChange);return w.jsxs(w.Fragment,{children:[w.jsx("style",{dangerouslySetInnerHTML:{__html:"[data-radix-scroll-area-viewport]{scrollbar-width:none;-ms-overflow-style:none;-webkit-overflow-scrolling:touch;}[data-radix-scroll-area-viewport]::-webkit-scrollbar{display:none}"},nonce:a}),w.jsx(Ze.div,{"data-radix-scroll-area-viewport":"",...o,ref:c,style:{overflowX:s.scrollbarXEnabled?"scroll":"hidden",overflowY:s.scrollbarYEnabled?"scroll":"hidden",...e.style},children:w.jsx("div",{ref:s.onContentChange,style:{minWidth:"100%",display:"table"},children:r})})]})});oG.displayName=aG;var ca="ScrollAreaScrollbar",zT=E.forwardRef((e,t)=>{const{forceMount:n,...r}=e,a=Tr(ca,e.__scopeScrollArea),{onScrollbarXEnabledChange:o,onScrollbarYEnabledChange:s}=a,u=e.orientation==="horizontal";return E.useEffect(()=>(u?o(!0):s(!0),()=>{u?o(!1):s(!1)}),[u,o,s]),a.type==="hover"?w.jsx(Rie,{...r,ref:t,forceMount:n}):a.type==="scroll"?w.jsx(_ie,{...r,ref:t,forceMount:n}):a.type==="auto"?w.jsx(iG,{...r,ref:t,forceMount:n}):a.type==="always"?w.jsx(BT,{...r,ref:t}):null});zT.displayName=ca;var Rie=E.forwardRef((e,t)=>{const{forceMount:n,...r}=e,a=Tr(ca,e.__scopeScrollArea),[o,s]=E.useState(!1);return E.useEffect(()=>{const u=a.scrollArea;let c=0;if(u){const d=()=>{window.clearTimeout(c),s(!0)},p=()=>{c=window.setTimeout(()=>s(!1),a.scrollHideDelay)};return u.addEventListener("pointerenter",d),u.addEventListener("pointerleave",p),()=>{window.clearTimeout(c),u.removeEventListener("pointerenter",d),u.removeEventListener("pointerleave",p)}}},[a.scrollArea,a.scrollHideDelay]),w.jsx(xr,{present:n||o,children:w.jsx(iG,{"data-state":o?"visible":"hidden",...r,ref:t})})}),_ie=E.forwardRef((e,t)=>{const{forceMount:n,...r}=e,a=Tr(ca,e.__scopeScrollArea),o=e.orientation==="horizontal",s=_p(()=>c("SCROLL_END"),100),[u,c]=Tie("hidden",{hidden:{SCROLL:"scrolling"},scrolling:{SCROLL_END:"idle",POINTER_ENTER:"interacting"},interacting:{SCROLL:"interacting",POINTER_LEAVE:"idle"},idle:{HIDE:"hidden",SCROLL:"scrolling",POINTER_ENTER:"interacting"}});return E.useEffect(()=>{if(u==="idle"){const d=window.setTimeout(()=>c("HIDE"),a.scrollHideDelay);return()=>window.clearTimeout(d)}},[u,a.scrollHideDelay,c]),E.useEffect(()=>{const d=a.viewport,p=o?"scrollLeft":"scrollTop";if(d){let g=d[p];const m=()=>{const b=d[p];g!==b&&(c("SCROLL"),s()),g=b};return d.addEventListener("scroll",m),()=>d.removeEventListener("scroll",m)}},[a.viewport,o,c,s]),w.jsx(xr,{present:n||u!=="hidden",children:w.jsx(BT,{"data-state":u==="hidden"?"hidden":"visible",...r,ref:t,onPointerEnter:Ke(e.onPointerEnter,()=>c("POINTER_ENTER")),onPointerLeave:Ke(e.onPointerLeave,()=>c("POINTER_LEAVE"))})})}),iG=E.forwardRef((e,t)=>{const n=Tr(ca,e.__scopeScrollArea),{forceMount:r,...a}=e,[o,s]=E.useState(!1),u=e.orientation==="horizontal",c=_p(()=>{if(n.viewport){const d=n.viewport.offsetWidth<n.viewport.scrollWidth,p=n.viewport.offsetHeight<n.viewport.scrollHeight;s(u?d:p)}},10);return Ps(n.viewport,c),Ps(n.content,c),w.jsx(xr,{present:r||o,children:w.jsx(BT,{"data-state":o?"visible":"hidden",...a,ref:t})})}),BT=E.forwardRef((e,t)=>{const{orientation:n="vertical",...r}=e,a=Tr(ca,e.__scopeScrollArea),o=E.useRef(null),s=E.useRef(0),[u,c]=E.useState({content:0,viewport:0,scrollbar:{size:0,paddingStart:0,paddingEnd:0}}),d=dG(u.viewport,u.content),p={...r,sizes:u,onSizesChange:c,hasThumb:d>0&&d<1,onThumbChange:m=>o.current=m,onThumbPointerUp:()=>s.current=0,onThumbPointerDown:m=>s.current=m};function g(m,b){return Lie(m,s.current,u,b)}return n==="horizontal"?w.jsx(Cie,{...p,ref:t,onThumbPositionChange:()=>{if(a.viewport&&o.current){const m=a.viewport.scrollLeft,b=KO(m,u,a.dir);o.current.style.transform=`translate3d(${b}px, 0, 0)`}},onWheelScroll:m=>{a.viewport&&(a.viewport.scrollLeft=m)},onDragScroll:m=>{a.viewport&&(a.viewport.scrollLeft=g(m,a.dir))}}):n==="vertical"?w.jsx(Nie,{...p,ref:t,onThumbPositionChange:()=>{if(a.viewport&&o.current){const m=a.viewport.scrollTop,b=KO(m,u);o.current.style.transform=`translate3d(0, ${b}px, 0)`}},onWheelScroll:m=>{a.viewport&&(a.viewport.scrollTop=m)},onDragScroll:m=>{a.viewport&&(a.viewport.scrollTop=g(m))}}):null}),Cie=E.forwardRef((e,t)=>{const{sizes:n,onSizesChange:r,...a}=e,o=Tr(ca,e.__scopeScrollArea),[s,u]=E.useState(),c=E.useRef(null),d=gt(t,c,o.onScrollbarXChange);return E.useEffect(()=>{c.current&&u(getComputedStyle(c.current))},[c]),w.jsx(lG,{"data-orientation":"horizontal",...a,ref:d,sizes:n,style:{bottom:0,left:o.dir==="rtl"?"var(--radix-scroll-area-corner-width)":0,right:o.dir==="ltr"?"var(--radix-scroll-area-corner-width)":0,"--radix-scroll-area-thumb-width":Rp(n)+"px",...e.style},onThumbPointerDown:p=>e.onThumbPointerDown(p.x),onDragScroll:p=>e.onDragScroll(p.x),onWheelScroll:(p,g)=>{if(o.viewport){const m=o.viewport.scrollLeft+p.deltaX;e.onWheelScroll(m),pG(m,g)&&p.preventDefault()}},onResize:()=>{c.current&&o.viewport&&s&&r({content:o.viewport.scrollWidth,viewport:o.viewport.offsetWidth,scrollbar:{size:c.current.clientWidth,paddingStart:Mf(s.paddingLeft),paddingEnd:Mf(s.paddingRight)}})}})}),Nie=E.forwardRef((e,t)=>{const{sizes:n,onSizesChange:r,...a}=e,o=Tr(ca,e.__scopeScrollArea),[s,u]=E.useState(),c=E.useRef(null),d=gt(t,c,o.onScrollbarYChange);return E.useEffect(()=>{c.current&&u(getComputedStyle(c.current))},[c]),w.jsx(lG,{"data-orientation":"vertical",...a,ref:d,sizes:n,style:{top:0,right:o.dir==="ltr"?0:void 0,left:o.dir==="rtl"?0:void 0,bottom:"var(--radix-scroll-area-corner-height)","--radix-scroll-area-thumb-height":Rp(n)+"px",...e.style},onThumbPointerDown:p=>e.onThumbPointerDown(p.y),onDragScroll:p=>e.onDragScroll(p.y),onWheelScroll:(p,g)=>{if(o.viewport){const m=o.viewport.scrollTop+p.deltaY;e.onWheelScroll(m),pG(m,g)&&p.preventDefault()}},onResize:()=>{c.current&&o.viewport&&s&&r({content:o.viewport.scrollHeight,viewport:o.viewport.offsetHeight,scrollbar:{size:c.current.clientHeight,paddingStart:Mf(s.paddingTop),paddingEnd:Mf(s.paddingBottom)}})}})}),[Oie,sG]=nG(ca),lG=E.forwardRef((e,t)=>{const{__scopeScrollArea:n,sizes:r,hasThumb:a,onThumbChange:o,onThumbPointerUp:s,onThumbPointerDown:u,onThumbPositionChange:c,onDragScroll:d,onWheelScroll:p,onResize:g,...m}=e,b=Tr(ca,n),[y,S]=E.useState(null),x=gt(t,L=>S(L)),R=E.useRef(null),k=E.useRef(""),A=b.viewport,C=r.content-r.viewport,N=bn(p),_=bn(c),O=_p(g,10);function F(L){if(R.current){const I=L.clientX-R.current.left,H=L.clientY-R.current.top;d({x:I,y:H})}}return E.useEffect(()=>{const L=I=>{const H=I.target;(y==null?void 0:y.contains(H))&&N(I,C)};return document.addEventListener("wheel",L,{passive:!1}),()=>document.removeEventListener("wheel",L,{passive:!1})},[A,y,C,N]),E.useEffect(_,[r,_]),Ps(y,O),Ps(b.content,O),w.jsx(Oie,{scope:n,scrollbar:y,hasThumb:a,onThumbChange:bn(o),onThumbPointerUp:bn(s),onThumbPositionChange:_,onThumbPointerDown:bn(u),children:w.jsx(Ze.div,{...m,ref:x,style:{position:"absolute",...m.style},onPointerDown:Ke(e.onPointerDown,L=>{L.button===0&&(L.target.setPointerCapture(L.pointerId),R.current=y.getBoundingClientRect(),k.current=document.body.style.webkitUserSelect,document.body.style.webkitUserSelect="none",b.viewport&&(b.viewport.style.scrollBehavior="auto"),F(L))}),onPointerMove:Ke(e.onPointerMove,F),onPointerUp:Ke(e.onPointerUp,L=>{const I=L.target;I.hasPointerCapture(L.pointerId)&&I.releasePointerCapture(L.pointerId),document.body.style.webkitUserSelect=k.current,b.viewport&&(b.viewport.style.scrollBehavior=""),R.current=null})})})}),Lf="ScrollAreaThumb",uG=E.forwardRef((e,t)=>{const{forceMount:n,...r}=e,a=sG(Lf,e.__scopeScrollArea);return w.jsx(xr,{present:n||a.hasThumb,children:w.jsx(Die,{ref:t,...r})})}),Die=E.forwardRef((e,t)=>{const{__scopeScrollArea:n,style:r,...a}=e,o=Tr(Lf,n),s=sG(Lf,n),{onThumbPositionChange:u}=s,c=gt(t,g=>s.onThumbChange(g)),d=E.useRef(void 0),p=_p(()=>{d.current&&(d.current(),d.current=void 0)},100);return E.useEffect(()=>{const g=o.viewport;if(g){const m=()=>{if(p(),!d.current){const b=Mie(g,u);d.current=b,u()}};return u(),g.addEventListener("scroll",m),()=>g.removeEventListener("scroll",m)}},[o.viewport,p,u]),w.jsx(Ze.div,{"data-state":s.hasThumb?"visible":"hidden",...a,ref:c,style:{width:"var(--radix-scroll-area-thumb-width)",height:"var(--radix-scroll-area-thumb-height)",...r},onPointerDownCapture:Ke(e.onPointerDownCapture,g=>{const b=g.target.getBoundingClientRect(),y=g.clientX-b.left,S=g.clientY-b.top;s.onThumbPointerDown({x:y,y:S})}),onPointerUp:Ke(e.onPointerUp,s.onThumbPointerUp)})});uG.displayName=Lf;var UT="ScrollAreaCorner",cG=E.forwardRef((e,t)=>{const n=Tr(UT,e.__scopeScrollArea),r=!!(n.scrollbarX&&n.scrollbarY);return n.type!=="scroll"&&r?w.jsx(Iie,{...e,ref:t}):null});cG.displayName=UT;var Iie=E.forwardRef((e,t)=>{const{__scopeScrollArea:n,...r}=e,a=Tr(UT,n),[o,s]=E.useState(0),[u,c]=E.useState(0),d=!!(o&&u);return Ps(a.scrollbarX,()=>{var g;const p=((g=a.scrollbarX)==null?void 0:g.offsetHeight)||0;a.onCornerHeightChange(p),c(p)}),Ps(a.scrollbarY,()=>{var g;const p=((g=a.scrollbarY)==null?void 0:g.offsetWidth)||0;a.onCornerWidthChange(p),s(p)}),d?w.jsx(Ze.div,{...r,ref:t,style:{width:o,height:u,position:"absolute",right:a.dir==="ltr"?0:void 0,left:a.dir==="rtl"?0:void 0,bottom:0,...e.style}}):null});function Mf(e){return e?parseInt(e,10):0}function dG(e,t){const n=e/t;return isNaN(n)?0:n}function Rp(e){const t=dG(e.viewport,e.content),n=e.scrollbar.paddingStart+e.scrollbar.paddingEnd,r=(e.scrollbar.size-n)*t;return Math.max(r,18)}function Lie(e,t,n,r="ltr"){const a=Rp(n),o=a/2,s=t||o,u=a-s,c=n.scrollbar.paddingStart+s,d=n.scrollbar.size-n.scrollbar.paddingEnd-u,p=n.content-n.viewport,g=r==="ltr"?[0,p]:[p*-1,0];return fG([c,d],g)(e)}function KO(e,t,n="ltr"){const r=Rp(t),a=t.scrollbar.paddingStart+t.scrollbar.paddingEnd,o=t.scrollbar.size-a,s=t.content-t.viewport,u=o-r,c=n==="ltr"?[0,s]:[s*-1,0],d=Ck(e,c);return fG([0,s],[0,u])(d)}function fG(e,t){return n=>{if(e[0]===e[1]||t[0]===t[1])return t[0];const r=(t[1]-t[0])/(e[1]-e[0]);return t[0]+r*(n-e[0])}}function pG(e,t){return e>0&&e<t}var Mie=(e,t=()=>{})=>{let n={left:e.scrollLeft,top:e.scrollTop},r=0;return function a(){const o={left:e.scrollLeft,top:e.scrollTop},s=n.left!==o.left,u=n.top!==o.top;(s||u)&&t(),n=o,r=window.requestAnimationFrame(a)}(),()=>window.cancelAnimationFrame(r)};function _p(e,t){const n=bn(e),r=E.useRef(0);return E.useEffect(()=>()=>window.clearTimeout(r.current),[]),E.useCallback(()=>{window.clearTimeout(r.current),r.current=window.setTimeout(n,t)},[n,t])}function Ps(e,t){const n=bn(t);An(()=>{let r=0;if(e){const a=new ResizeObserver(()=>{cancelAnimationFrame(r),r=window.requestAnimationFrame(n)});return a.observe(e),()=>{window.cancelAnimationFrame(r),a.unobserve(e)}}},[e,n])}var gG=rG,Fie=oG,Pie=cG;const hG=E.forwardRef(({className:e,children:t,...n},r)=>w.jsxs(gG,{ref:r,className:Me("relative overflow-hidden",e),...n,children:[w.jsx(Fie,{className:"h-full w-full rounded-[inherit]",children:t}),w.jsx(mG,{}),w.jsx(Pie,{})]}));hG.displayName=gG.displayName;const mG=E.forwardRef(({className:e,orientation:t="vertical",...n},r)=>w.jsx(zT,{ref:r,orientation:t,className:Me("flex touch-none transition-colors select-none",t==="vertical"&&"h-full w-2.5 border-l border-l-transparent p-[1px]",t==="horizontal"&&"h-2.5 flex-col border-t border-t-transparent p-[1px]",e),...n,children:w.jsx(uG,{className:"bg-border relative flex-1 rounded-full"})}));mG.displayName=zT.displayName;function s0(e,t={}){const{decimals:n=0,sizeType:r="normal"}=t,a=["Bytes","KB","MB","GB","TB"],o=["Bytes","KiB","MiB","GiB","TiB"];if(e===0)return"0 Byte";const s=Math.floor(Math.log(e)/Math.log(1024));return`${(e/Math.pow(1024,s)).toFixed(n)} ${r==="accurate"?o[s]??"Bytes":a[s]??"Bytes"}`}function zie(e){const{value:t,onValueChange:n,onUpload:r,progresses:a,accept:o=yV,maxSize:s=1024*1024*200,maxFileCount:u=1,multiple:c=!1,disabled:d=!1,description:p,className:g,...m}=e,[b,y]=ja({prop:t,onChange:n}),S=E.useCallback((k,A)=>{if(!c&&u===1&&k.length>1){Ft.error("Cannot upload more than 1 file at a time");return}if(((b==null?void 0:b.length)??0)+k.length>u){Ft.error(`Cannot upload more than ${u} files`);return}const C=k.map(_=>Object.assign(_,{preview:URL.createObjectURL(_)})),N=b?[...b,...C]:C;if(y(N),A.length>0&&A.forEach(({file:_})=>{Ft.error(`File ${_.name} was rejected`)}),r&&N.length>0&&N.length<=u){const _=N.length>0?`${N.length} files`:"file";Ft.promise(r(N),{loading:`Uploading ${_}...`,success:()=>(y([]),`${_} uploaded`),error:`Failed to upload ${_}`})}},[b,u,c,r,y]);function x(k){if(!b)return;const A=b.filter((C,N)=>N!==k);y(A),n==null||n(A)}E.useEffect(()=>()=>{b&&b.forEach(k=>{bG(k)&&URL.revokeObjectURL(k.preview)})},[]);const R=d||((b==null?void 0:b.length)??0)>=u;return w.jsxs("div",{className:"relative flex flex-col gap-6 overflow-hidden",children:[w.jsx(Ap,{onDrop:S,accept:o,maxSize:s,maxFiles:u,multiple:u>1||c,disabled:R,children:({getRootProps:k,getInputProps:A,isDragActive:C})=>w.jsxs("div",{...k(),className:Me("group border-muted-foreground/25 hover:bg-muted/25 relative grid h-52 w-full cursor-pointer place-items-center rounded-lg border-2 border-dashed px-5 py-2.5 text-center transition","ring-offset-background focus-visible:ring-ring focus-visible:ring-2 focus-visible:ring-offset-2 focus-visible:outline-none",C&&"border-muted-foreground/50",R&&"pointer-events-none opacity-60",g),...m,children:[w.jsx("input",{...A()}),C?w.jsxs("div",{className:"flex flex-col items-center justify-center gap-4 sm:px-5",children:[w.jsx("div",{className:"rounded-full border border-dashed p-3",children:w.jsx(bk,{className:"text-muted-foreground size-7","aria-hidden":"true"})}),w.jsx("p",{className:"text-muted-foreground font-medium",children:"Drop the files here"})]}):w.jsxs("div",{className:"flex flex-col items-center justify-center gap-4 sm:px-5",children:[w.jsx("div",{className:"rounded-full border border-dashed p-3",children:w.jsx(bk,{className:"text-muted-foreground size-7","aria-hidden":"true"})}),w.jsxs("div",{className:"flex flex-col gap-px",children:[w.jsx("p",{className:"text-muted-foreground font-medium",children:"Drag and drop files here, or click to select files"}),p?w.jsx("p",{className:"text-muted-foreground/70 text-sm",children:p}):w.jsxs("p",{className:"text-muted-foreground/70 text-sm",children:["You can upload",u>1?` ${u===1/0?"multiple":u}
                      files (up to ${s0(s)} each)`:` a file with ${s0(s)}`,"Supported formats: TXT, MD, DOCX, PDF, PPTX, RTF, ODT, EPUB, HTML, HTM, TEX, JSON, XML, YAML, YML, CSV, LOG, CONF, INI, PROPERTIES, SQL, BAT, SH, C, CPP, PY, JAVA, JS, TS, SWIFT, GO, RB, PHP, CSS, SCSS, LESS"]})]})]})]})}),b!=null&&b.length?w.jsx(hG,{className:"h-fit w-full px-3",children:w.jsx("div",{className:"flex max-h-48 flex-col gap-4",children:b==null?void 0:b.map((k,A)=>w.jsx(Bie,{file:k,onRemove:()=>x(A),progress:a==null?void 0:a[k.name]},A))})}):null]})}function Bie({file:e,progress:t,onRemove:n}){return w.jsxs("div",{className:"relative flex items-center gap-2.5",children:[w.jsxs("div",{className:"flex flex-1 gap-2.5",children:[bG(e)?w.jsx(Uie,{file:e}):null,w.jsxs("div",{className:"flex w-full flex-col gap-2",children:[w.jsxs("div",{className:"flex flex-col gap-px",children:[w.jsx("p",{className:"text-foreground/80 line-clamp-1 text-sm font-medium",children:e.name}),w.jsx("p",{className:"text-muted-foreground text-xs",children:s0(e.size)})]}),t?w.jsx(tG,{value:t}):null]})]}),w.jsx("div",{className:"flex items-center gap-2",children:w.jsxs(xt,{type:"button",variant:"outline",size:"icon",className:"size-7",onClick:n,children:[w.jsx(rU,{className:"size-4","aria-hidden":"true"}),w.jsx("span",{className:"sr-only",children:"Remove file"})]})})]})}function bG(e){return"preview"in e&&typeof e.preview=="string"}function Uie({file:e}){return e.type.startsWith("image/")?w.jsx("div",{className:"aspect-square shrink-0 rounded-md object-cover"}):w.jsx(OW,{className:"text-muted-foreground size-10","aria-hidden":"true"})}function jie(){const{t:e}=Pt(),[t,n]=E.useState(!1),[r,a]=E.useState(!1),[o,s]=E.useState({}),u=E.useCallback(async c=>{a(!0);try{await Promise.all(c.map(async d=>{try{const p=await LV(d,g=>{console.debug(e("documentPanel.uploadDocuments.uploading",{name:d.name,percent:g})),s(m=>({...m,[d.name]:g}))});p.status==="success"?Ft.success(e("documentPanel.uploadDocuments.success",{name:d.name})):Ft.error(e("documentPanel.uploadDocuments.failed",{name:d.name,message:p.message}))}catch(p){Ft.error(e("documentPanel.uploadDocuments.error",{name:d.name,error:Sr(p)}))}}))}catch(d){Ft.error(e("documentPanel.uploadDocuments.generalError",{error:Sr(d)}))}finally{a(!1)}},[a,s]);return w.jsxs(y5,{open:t,onOpenChange:c=>{r&&!c||n(c)},children:[w.jsx(v5,{asChild:!0,children:w.jsxs(xt,{variant:"default",side:"bottom",tooltip:e("documentPanel.uploadDocuments.tooltip"),size:"sm",children:[w.jsx(bk,{})," ",e("documentPanel.uploadDocuments.button")]})}),w.jsxs(TT,{className:"sm:max-w-xl",onCloseAutoFocus:c=>c.preventDefault(),children:[w.jsxs(AT,{children:[w.jsx(RT,{children:e("documentPanel.uploadDocuments.title")}),w.jsx(_T,{children:e("documentPanel.uploadDocuments.description")})]}),w.jsx(zie,{maxFileCount:1/0,maxSize:200*1024*1024,description:e("documentPanel.uploadDocuments.fileTypes"),onUpload:u,progresses:o,disabled:r})]})]})}function Gie(){const{t:e}=Pt(),[t,n]=E.useState(!1),r=E.useCallback(async()=>{try{const a=await MV();a.status==="success"?(Ft.success(e("documentPanel.clearDocuments.success")),n(!1)):Ft.error(e("documentPanel.clearDocuments.failed",{message:a.message}))}catch(a){Ft.error(e("documentPanel.clearDocuments.error",{error:Sr(a)}))}},[n]);return w.jsxs(y5,{open:t,onOpenChange:n,children:[w.jsx(v5,{asChild:!0,children:w.jsxs(xt,{variant:"outline",side:"bottom",tooltip:e("documentPanel.clearDocuments.tooltip"),size:"sm",children:[w.jsx(eU,{})," ",e("documentPanel.clearDocuments.button")]})}),w.jsxs(TT,{className:"sm:max-w-xl",onCloseAutoFocus:a=>a.preventDefault(),children:[w.jsxs(AT,{children:[w.jsx(RT,{children:e("documentPanel.clearDocuments.title")}),w.jsx(_T,{children:e("documentPanel.clearDocuments.confirm")})]}),w.jsx(xt,{variant:"destructive",onClick:r,children:e("documentPanel.clearDocuments.confirmButton")})]})]})}function Hie(){const{t:e}=Pt(),t=Hn.use.health(),[n,r]=E.useState(null),a=Fe.use.currentTab(),o=E.useCallback(async()=>{try{const u=await NV();u&&u.statuses&&Object.values(u.statuses).reduce((d,p)=>d+p.length,0)>0?r(u):r(null)}catch(u){Ft.error(e("documentPanel.documentManager.errors.loadFailed",{error:Sr(u)}))}},[r,e]);E.useEffect(()=>{a==="documents"&&o()},[a,o]);const s=E.useCallback(async()=>{try{const{status:u}=await OV();Ft.message(u)}catch(u){Ft.error(e("documentPanel.documentManager.errors.scanFailed",{error:Sr(u)}))}},[e]);return E.useEffect(()=>{if(a!=="documents"||!t)return;const u=setInterval(async()=>{try{await o()}catch(c){Ft.error(e("documentPanel.documentManager.errors.scanProgressFailed",{error:Sr(c)}))}},5e3);return()=>clearInterval(u)},[t,o,e,a]),w.jsxs(Ms,{className:"!size-full !rounded-none !border-none",children:[w.jsx(Nu,{children:w.jsx(Ou,{className:"text-lg",children:e("documentPanel.documentManager.title")})}),w.jsxs(Du,{className:"space-y-4",children:[w.jsxs("div",{className:"flex gap-2",children:[w.jsxs(xt,{variant:"outline",onClick:s,side:"bottom",tooltip:e("documentPanel.documentManager.scanTooltip"),size:"sm",children:[w.jsx(nU,{})," ",e("documentPanel.documentManager.scanButton")]}),w.jsx("div",{className:"flex-1"}),w.jsx(Gie,{}),w.jsx(jie,{})]}),w.jsxs(Ms,{children:[w.jsxs(Nu,{children:[w.jsx(Ou,{children:e("documentPanel.documentManager.uploadedTitle")}),w.jsx(Tp,{children:e("documentPanel.documentManager.uploadedDescription")})]}),w.jsxs(Du,{children:[!n&&w.jsx(boe,{title:e("documentPanel.documentManager.emptyTitle"),description:e("documentPanel.documentManager.emptyDescription")}),n&&w.jsxs(P5,{children:[w.jsx(z5,{children:w.jsxs(t0,{children:[w.jsx(Ma,{children:e("documentPanel.documentManager.columns.id")}),w.jsx(Ma,{children:e("documentPanel.documentManager.columns.summary")}),w.jsx(Ma,{children:e("documentPanel.documentManager.columns.status")}),w.jsx(Ma,{children:e("documentPanel.documentManager.columns.length")}),w.jsx(Ma,{children:e("documentPanel.documentManager.columns.chunks")}),w.jsx(Ma,{children:e("documentPanel.documentManager.columns.created")}),w.jsx(Ma,{children:e("documentPanel.documentManager.columns.updated")}),w.jsx(Ma,{children:e("documentPanel.documentManager.columns.metadata")})]})}),w.jsx(B5,{className:"text-sm",children:Object.entries(n.statuses).map(([u,c])=>c.map(d=>w.jsxs(t0,{children:[w.jsx(Fa,{className:"truncate font-mono",children:d.id}),w.jsx(Fa,{className:"max-w-xs min-w-24 truncate",children:w.jsx(Bn,{text:d.content_summary,tooltip:d.content_summary,tooltipClassName:"max-w-none overflow-visible block"})}),w.jsxs(Fa,{children:[u==="processed"&&w.jsx("span",{className:"text-green-600",children:e("documentPanel.documentManager.status.completed")}),u==="processing"&&w.jsx("span",{className:"text-blue-600",children:e("documentPanel.documentManager.status.processing")}),u==="pending"&&w.jsx("span",{className:"text-yellow-600",children:e("documentPanel.documentManager.status.pending")}),u==="failed"&&w.jsx("span",{className:"text-red-600",children:e("documentPanel.documentManager.status.failed")}),d.error&&w.jsx("span",{className:"ml-2 text-red-500",title:d.error,children:"⚠️"})]}),w.jsx(Fa,{children:d.content_length??"-"}),w.jsx(Fa,{children:d.chunks_count??"-"}),w.jsx(Fa,{className:"truncate",children:new Date(d.created_at).toLocaleString()}),w.jsx(Fa,{className:"truncate",children:new Date(d.updated_at).toLocaleString()}),w.jsx(Fa,{className:"max-w-xs truncate",children:d.metadata?JSON.stringify(d.metadata):"-"})]},d.id)))})]})]})]})]})]})}function yG(e,t){var n={};for(var r in e)Object.prototype.hasOwnProperty.call(e,r)&&t.indexOf(r)<0&&(n[r]=e[r]);if(e!=null&&typeof Object.getOwnPropertySymbols=="function")for(var a=0,r=Object.getOwnPropertySymbols(e);a<r.length;a++)t.indexOf(r[a])<0&&Object.prototype.propertyIsEnumerable.call(e,r[a])&&(n[r[a]]=e[r[a]]);return n}var Lu;(function(e){e.event="event",e.props="prop"})(Lu||(Lu={}));function ko(){}function $ie(e){var t,n=void 0;return function(){for(var r=[],a=arguments.length;a--;)r[a]=arguments[a];return t&&r.length===t.length&&r.every(function(o,s){return o===t[s]})||(t=r,n=e.apply(void 0,r)),n}}function Mu(e){return!!(e||"").match(/\d/)}function _s(e){return e==null}function qie(e){return typeof e=="number"&&isNaN(e)}function vG(e){return _s(e)||qie(e)||typeof e=="number"&&!isFinite(e)}function SG(e){return e.replace(/[-[\]/{}()*+?.\\^$|]/g,"\\$&")}function Vie(e){switch(e){case"lakh":return/(\d+?)(?=(\d\d)+(\d)(?!\d))(\.\d+)?/g;case"wan":return/(\d)(?=(\d{4})+(?!\d))/g;case"thousand":default:return/(\d)(?=(\d{3})+(?!\d))/g}}function Wie(e,t,n){var r=Vie(n),a=e.search(/[1-9]/);return a=a===-1?e.length:a,e.substring(0,a)+e.substring(a,e.length).replace(r,"$1"+t)}function Yie(e){var t=E.useRef(e);t.current=e;var n=E.useRef(function(){for(var r=[],a=arguments.length;a--;)r[a]=arguments[a];return t.current.apply(t,r)});return n.current}function jT(e,t){t===void 0&&(t=!0);var n=e[0]==="-",r=n&&t;e=e.replace("-","");var a=e.split("."),o=a[0],s=a[1]||"";return{beforeDecimal:o,afterDecimal:s,hasNegation:n,addNegation:r}}function Kie(e){if(!e)return e;var t=e[0]==="-";t&&(e=e.substring(1,e.length));var n=e.split("."),r=n[0].replace(/^0+/,"")||"0",a=n[1]||"";return(t?"-":"")+r+(a?"."+a:"")}function EG(e,t,n){for(var r="",a=n?"0":"",o=0;o<=t-1;o++)r+=e[o]||a;return r}function XO(e,t){return Array(t+1).join(e)}function wG(e){var t=e+"",n=t[0]==="-"?"-":"";n&&(t=t.substring(1));var r=t.split(/[eE]/g),a=r[0],o=r[1];if(o=Number(o),!o)return n+a;a=a.replace(".","");var s=1+o,u=a.length;return s<0?a="0."+XO("0",Math.abs(s))+a:s>=u?a=a+XO("0",s-u):a=(a.substring(0,s)||"0")+"."+a.substring(s),n+a}function ZO(e,t,n){if(["","-"].indexOf(e)!==-1)return e;var r=(e.indexOf(".")!==-1||n)&&t,a=jT(e),o=a.beforeDecimal,s=a.afterDecimal,u=a.hasNegation,c=parseFloat("0."+(s||"0")),d=s.length<=t?"0."+s:c.toFixed(t),p=d.split("."),g=o;o&&Number(p[0])&&(g=o.split("").reverse().reduce(function(S,x,R){return S.length>R?(Number(S[0])+Number(x)).toString()+S.substring(1,S.length):x+S},p[0]));var m=EG(p[1]||"",t,n),b=u?"-":"",y=r?".":"";return""+b+g+y+m}function ci(e,t){if(e.value=e.value,e!==null){if(e.createTextRange){var n=e.createTextRange();return n.move("character",t),n.select(),!0}return e.selectionStart||e.selectionStart===0?(e.focus(),e.setSelectionRange(t,t),!0):(e.focus(),!1)}}var xG=$ie(function(e,t){for(var n=0,r=0,a=e.length,o=t.length;e[n]===t[n]&&n<a;)n++;for(;e[a-1-r]===t[o-1-r]&&o-r>n&&a-r>n;)r++;return{from:{start:n,end:a-r},to:{start:n,end:o-r}}}),Xie=function(e,t){var n=Math.min(e.selectionStart,t);return{from:{start:n,end:e.selectionEnd},to:{start:n,end:t}}};function Zie(e,t,n){return Math.min(Math.max(e,t),n)}function ub(e){return Math.max(e.selectionStart,e.selectionEnd)}function Qie(){return typeof navigator<"u"&&!(navigator.platform&&/iPhone|iPod/.test(navigator.platform))}function Jie(e){return{from:{start:0,end:0},to:{start:0,end:e.length},lastValue:""}}function ese(e){var t=e.currentValue,n=e.formattedValue,r=e.currentValueIndex,a=e.formattedValueIndex;return t[r]===n[a]}function tse(e,t,n,r,a,o,s){s===void 0&&(s=ese);var u=a.findIndex(function(C){return C}),c=e.slice(0,u);!t&&!n.startsWith(c)&&(t=c,n=c+n,r=r+c.length);for(var d=n.length,p=e.length,g={},m=new Array(d),b=0;b<d;b++){m[b]=-1;for(var y=0,S=p;y<S;y++){var x=s({currentValue:n,lastValue:t,formattedValue:e,currentValueIndex:b,formattedValueIndex:y});if(x&&g[y]!==!0){m[b]=y,g[y]=!0;break}}}for(var R=r;R<d&&(m[R]===-1||!o(n[R]));)R++;var k=R===d||m[R]===-1?p:m[R];for(R=r-1;R>0&&m[R]===-1;)R--;var A=R===-1||m[R]===-1?0:m[R]+1;return A>k?k:r-A<k-r?A:k}function QO(e,t,n,r){var a=e.length;if(t=Zie(t,0,a),r==="left"){for(;t>=0&&!n[t];)t--;t===-1&&(t=n.indexOf(!0))}else{for(;t<=a&&!n[t];)t++;t>a&&(t=n.lastIndexOf(!0))}return t===-1&&(t=a),t}function nse(e){for(var t=Array.from({length:e.length+1}).map(function(){return!0}),n=0,r=t.length;n<r;n++)t[n]=!!(Mu(e[n])||Mu(e[n-1]));return t}function kG(e,t,n,r,a,o){o===void 0&&(o=ko);var s=Yie(function(y,S){var x,R;return vG(y)?(R="",x=""):typeof y=="number"||S?(R=typeof y=="number"?wG(y):y,x=r(R)):(R=a(y,void 0),x=r(R)),{formattedValue:x,numAsString:R}}),u=E.useState(function(){return s(_s(e)?t:e,n)}),c=u[0],d=u[1],p=function(y,S){y.formattedValue!==c.formattedValue&&d({formattedValue:y.formattedValue,numAsString:y.value}),o(y,S)},g=e,m=n;_s(e)&&(g=c.numAsString,m=!0);var b=s(g,m);return E.useMemo(function(){d(b)},[b.formattedValue]),[c,p]}function rse(e){return e.replace(/[^0-9]/g,"")}function ase(e){return e}function ose(e){var t=e.type;t===void 0&&(t="text");var n=e.displayType;n===void 0&&(n="input");var r=e.customInput,a=e.renderText,o=e.getInputRef,s=e.format;s===void 0&&(s=ase);var u=e.removeFormatting;u===void 0&&(u=rse);var c=e.defaultValue,d=e.valueIsNumericString,p=e.onValueChange,g=e.isAllowed,m=e.onChange;m===void 0&&(m=ko);var b=e.onKeyDown;b===void 0&&(b=ko);var y=e.onMouseUp;y===void 0&&(y=ko);var S=e.onFocus;S===void 0&&(S=ko);var x=e.onBlur;x===void 0&&(x=ko);var R=e.value,k=e.getCaretBoundary;k===void 0&&(k=nse);var A=e.isValidInputCharacter;A===void 0&&(A=Mu);var C=e.isCharacterSame,N=yG(e,["type","displayType","customInput","renderText","getInputRef","format","removeFormatting","defaultValue","valueIsNumericString","onValueChange","isAllowed","onChange","onKeyDown","onMouseUp","onFocus","onBlur","value","getCaretBoundary","isValidInputCharacter","isCharacterSame"]),_=kG(R,c,!!d,s,u,p),O=_[0],F=O.formattedValue,L=O.numAsString,I=_[1],H=E.useRef(),$=E.useRef({formattedValue:F,numAsString:L}),U=function(fe,Te){$.current={formattedValue:fe.formattedValue,numAsString:fe.value},I(fe,Te)},Y=E.useState(!1),Z=Y[0],j=Y[1],G=E.useRef(null),z=E.useRef({setCaretTimeout:null,focusTimeout:null});E.useEffect(function(){return j(!0),function(){clearTimeout(z.current.setCaretTimeout),clearTimeout(z.current.focusTimeout)}},[]);var K=s,D=function(fe,Te){var me=parseFloat(Te);return{formattedValue:fe,value:Te,floatValue:isNaN(me)?void 0:me}},V=function(fe,Te,me){fe.selectionStart===0&&fe.selectionEnd===fe.value.length||(ci(fe,Te),z.current.setCaretTimeout=setTimeout(function(){fe.value===me&&fe.selectionStart!==Te&&ci(fe,Te)},0))},B=function(fe,Te,me){return QO(fe,Te,k(fe),me)},M=function(fe,Te,me){var Ee=k(Te),le=tse(Te,F,fe,me,Ee,A,C);return le=QO(Te,le,Ee),le},W=function(fe){var Te=fe.formattedValue;Te===void 0&&(Te="");var me=fe.input,Ee=fe.source,le=fe.event,Be=fe.numAsString,Ue;if(me){var he=fe.inputValue||me.value,Ne=ub(me);me.value=Te,Ue=M(he,Te,Ne),Ue!==void 0&&V(me,Ue,Te)}Te!==F&&U(D(Te,Be),{event:le,source:Ee})};E.useEffect(function(){var fe=$.current,Te=fe.formattedValue,me=fe.numAsString;(F!==Te||L!==me)&&U(D(F,L),{event:void 0,source:Lu.props})},[F,L]);var Q=G.current?ub(G.current):void 0,oe=typeof window<"u"?E.useLayoutEffect:E.useEffect;oe(function(){var fe=G.current;if(F!==$.current.formattedValue&&fe){var Te=M($.current.formattedValue,F,Q);fe.value=F,V(fe,Te,F)}},[F]);var re=function(fe,Te,me){var Ee=Te.target,le=H.current?Xie(H.current,Ee.selectionEnd):xG(F,fe),Be=Object.assign(Object.assign({},le),{lastValue:F}),Ue=u(fe,Be),he=K(Ue);if(Ue=u(he,void 0),g&&!g(D(he,Ue))){var Ne=Te.target,ee=ub(Ne),ce=M(fe,F,ee);return Ne.value=F,V(Ne,ce,F),!1}return W({formattedValue:he,numAsString:Ue,inputValue:fe,event:Te,source:me,input:Te.target}),!0},ie=function(fe,Te){Te===void 0&&(Te=0);var me=fe.selectionStart,Ee=fe.selectionEnd;H.current={selectionStart:me,selectionEnd:Ee+Te}},Se=function(fe){var Te=fe.target,me=Te.value,Ee=re(me,fe,Lu.event);Ee&&m(fe),H.current=void 0},ae=function(fe){var Te=fe.target,me=fe.key,Ee=Te.selectionStart,le=Te.selectionEnd,Be=Te.value;Be===void 0&&(Be="");var Ue;me==="ArrowLeft"||me==="Backspace"?Ue=Math.max(Ee-1,0):me==="ArrowRight"?Ue=Math.min(Ee+1,Be.length):me==="Delete"&&(Ue=Ee);var he=0;me==="Delete"&&Ee===le&&(he=1);var Ne=me==="ArrowLeft"||me==="ArrowRight";if(Ue===void 0||Ee!==le&&!Ne){b(fe),ie(Te,he);return}var ee=Ue;if(Ne){var ce=me==="ArrowLeft"?"left":"right";ee=B(Be,Ue,ce),ee!==Ue&&fe.preventDefault()}else me==="Delete"&&!A(Be[Ue])?ee=B(Be,Ue,"right"):me==="Backspace"&&!A(Be[Ue])&&(ee=B(Be,Ue,"left"));ee!==Ue&&V(Te,ee,Be),b(fe),ie(Te,he)},ve=function(fe){var Te=fe.target,me=function(){var Ee=Te.selectionStart,le=Te.selectionEnd,Be=Te.value;if(Be===void 0&&(Be=""),Ee===le){var Ue=B(Be,Ee);Ue!==Ee&&V(Te,Ue,Be)}};me(),requestAnimationFrame(function(){me()}),y(fe),ie(Te)},xe=function(fe){fe.persist&&fe.persist();var Te=fe.target,me=fe.currentTarget;G.current=Te,z.current.focusTimeout=setTimeout(function(){var Ee=Te.selectionStart,le=Te.selectionEnd,Be=Te.value;Be===void 0&&(Be="");var Ue=B(Be,Ee);Ue!==Ee&&!(Ee===0&&le===Be.length)&&V(Te,Ue,Be),S(Object.assign(Object.assign({},fe),{currentTarget:me}))},0)},Ie=function(fe){G.current=null,clearTimeout(z.current.focusTimeout),clearTimeout(z.current.setCaretTimeout),x(fe)},Ce=Z&&Qie()?"numeric":void 0,ke=Object.assign({inputMode:Ce},N,{type:t,value:F,onChange:Se,onKeyDown:ae,onMouseUp:ve,onFocus:xe,onBlur:Ie});if(n==="text")return a?ye.createElement(ye.Fragment,null,a(F,N)||null):ye.createElement("span",Object.assign({},N,{ref:o}),F);if(r){var J=r;return ye.createElement(J,Object.assign({},ke,{ref:o}))}return ye.createElement("input",Object.assign({},ke,{ref:o}))}function JO(e,t){var n=t.decimalScale,r=t.fixedDecimalScale,a=t.prefix;a===void 0&&(a="");var o=t.suffix;o===void 0&&(o="");var s=t.allowNegative,u=t.thousandsGroupStyle;if(u===void 0&&(u="thousand"),e===""||e==="-")return e;var c=Cp(t),d=c.thousandSeparator,p=c.decimalSeparator,g=n!==0&&e.indexOf(".")!==-1||n&&r,m=jT(e,s),b=m.beforeDecimal,y=m.afterDecimal,S=m.addNegation;return n!==void 0&&(y=EG(y,n,!!r)),d&&(b=Wie(b,d,u)),a&&(b=a+b),o&&(y=y+o),S&&(b="-"+b),e=b+(g&&p||"")+y,e}function Cp(e){var t=e.decimalSeparator;t===void 0&&(t=".");var n=e.thousandSeparator,r=e.allowedDecimalSeparators;return n===!0&&(n=","),r||(r=[t,"."]),{decimalSeparator:t,thousandSeparator:n,allowedDecimalSeparators:r}}function ise(e,t){e===void 0&&(e="");var n=new RegExp("(-)"),r=new RegExp("(-)(.)*(-)"),a=n.test(e),o=r.test(e);return e=e.replace(/-/g,""),a&&!o&&t&&(e="-"+e),e}function sse(e,t){return new RegExp("(^-)|[0-9]|"+SG(e),"g")}function lse(e,t,n){return e===""?!0:!(t!=null&&t.match(/\d/))&&!(n!=null&&n.match(/\d/))&&typeof e=="string"&&!isNaN(Number(e))}function use(e,t,n){var r;t===void 0&&(t=Jie(e));var a=n.allowNegative,o=n.prefix;o===void 0&&(o="");var s=n.suffix;s===void 0&&(s="");var u=n.decimalScale,c=t.from,d=t.to,p=d.start,g=d.end,m=Cp(n),b=m.allowedDecimalSeparators,y=m.decimalSeparator,S=e[g]===y;if(Mu(e)&&(e===o||e===s)&&t.lastValue==="")return e;if(g-p===1&&b.indexOf(e[p])!==-1){var x=u===0?"":y;e=e.substring(0,p)+x+e.substring(p+1,e.length)}var R=function(G,z,K){var D=!1,V=!1;o.startsWith("-")?D=!1:G.startsWith("--")?(D=!1,V=!0):s.startsWith("-")&&G.length===s.length?D=!1:G[0]==="-"&&(D=!0);var B=D?1:0;return V&&(B=2),B&&(G=G.substring(B),z-=B,K-=B),{value:G,start:z,end:K,hasNegation:D}},k=R(e,p,g),A=k.hasNegation;r=k,e=r.value,p=r.start,g=r.end;var C=R(t.lastValue,c.start,c.end),N=C.start,_=C.end,O=C.value,F=e.substring(p,g);e.length&&O.length&&(N>O.length-s.length||_<o.length)&&!(F&&s.startsWith(F))&&(e=O);var L=0;e.startsWith(o)?L+=o.length:p<o.length&&(L=p),e=e.substring(L),g-=L;var I=e.length,H=e.length-s.length;e.endsWith(s)?I=H:(g>H||g>e.length-s.length)&&(I=g),e=e.substring(0,I),e=ise(A?"-"+e:e,a),e=(e.match(sse(y))||[]).join("");var $=e.indexOf(y);e=e.replace(new RegExp(SG(y),"g"),function(G,z){return z===$?".":""});var U=jT(e,a),Y=U.beforeDecimal,Z=U.afterDecimal,j=U.addNegation;return d.end-d.start<c.end-c.start&&Y===""&&S&&!parseFloat(Z)&&(e=j?"-":""),e}function cse(e,t){var n=t.prefix;n===void 0&&(n="");var r=t.suffix;r===void 0&&(r="");var a=Array.from({length:e.length+1}).map(function(){return!0}),o=e[0]==="-";a.fill(!1,0,n.length+(o?1:0));var s=e.length;return a.fill(!1,s-r.length+1,s+1),a}function dse(e){var t=Cp(e),n=t.thousandSeparator,r=t.decimalSeparator,a=e.prefix;a===void 0&&(a="");var o=e.allowNegative;if(o===void 0&&(o=!0),n===r)throw new Error(`
        Decimal separator can't be same as thousand separator.
        thousandSeparator: `+n+` (thousandSeparator = {true} is same as thousandSeparator = ",")
        decimalSeparator: `+r+` (default value for decimalSeparator is .)
     `);return a.startsWith("-")&&o&&(console.error(`
      Prefix can't start with '-' when allowNegative is true.
      prefix: `+a+`
      allowNegative: `+o+`
    `),o=!1),Object.assign(Object.assign({},e),{allowNegative:o})}function fse(e){e=dse(e),e.decimalSeparator,e.allowedDecimalSeparators,e.thousandsGroupStyle;var t=e.suffix,n=e.allowNegative,r=e.allowLeadingZeros,a=e.onKeyDown;a===void 0&&(a=ko);var o=e.onBlur;o===void 0&&(o=ko);var s=e.thousandSeparator,u=e.decimalScale,c=e.fixedDecimalScale,d=e.prefix;d===void 0&&(d="");var p=e.defaultValue,g=e.value,m=e.valueIsNumericString,b=e.onValueChange,y=yG(e,["decimalSeparator","allowedDecimalSeparators","thousandsGroupStyle","suffix","allowNegative","allowLeadingZeros","onKeyDown","onBlur","thousandSeparator","decimalScale","fixedDecimalScale","prefix","defaultValue","value","valueIsNumericString","onValueChange"]),S=Cp(e),x=S.decimalSeparator,R=S.allowedDecimalSeparators,k=function(j){return JO(j,e)},A=function(j,G){return use(j,G,e)},C=_s(g)?p:g,N=m??lse(C,d,t);_s(g)?_s(p)||(N=N||typeof p=="number"):N=N||typeof g=="number";var _=function(j){return vG(j)?j:(typeof j=="number"&&(j=wG(j)),N&&typeof u=="number"?ZO(j,u,!!c):j)},O=kG(_(g),_(p),!!N,k,A,b),F=O[0],L=F.numAsString,I=F.formattedValue,H=O[1],$=function(j){var G=j.target,z=j.key,K=G.selectionStart,D=G.selectionEnd,V=G.value;if(V===void 0&&(V=""),(z==="Backspace"||z==="Delete")&&D<d.length){j.preventDefault();return}if(K!==D){a(j);return}z==="Backspace"&&V[0]==="-"&&K===d.length+1&&n&&ci(G,1),u&&c&&(z==="Backspace"&&V[K-1]===x?(ci(G,K-1),j.preventDefault()):z==="Delete"&&V[K]===x&&j.preventDefault()),R!=null&&R.includes(z)&&V[K]===x&&ci(G,K+1);var B=s===!0?",":s;z==="Backspace"&&V[K-1]===B&&ci(G,K-1),z==="Delete"&&V[K]===B&&ci(G,K+1),a(j)},U=function(j){var G=L;if(G.match(/\d/g)||(G=""),r||(G=Kie(G)),c&&u&&(G=ZO(G,u,c)),G!==L){var z=JO(G,e);H({formattedValue:z,value:G,floatValue:parseFloat(G)},{event:j,source:Lu.event})}o(j)},Y=function(j){return j===x?!0:Mu(j)},Z=function(j){var G=j.currentValue,z=j.lastValue,K=j.formattedValue,D=j.currentValueIndex,V=j.formattedValueIndex,B=G[D],M=K[V],W=xG(z,G),Q=W.to,oe=function(re){return A(re).indexOf(".")+d.length};return g===0&&c&&u&&G[Q.start]===x&&oe(G)<D&&oe(K)>V?!1:D>=Q.start&&D<Q.end&&R&&R.includes(B)&&M===x?!0:B===M};return Object.assign(Object.assign({},y),{value:I,valueIsNumericString:!1,isValidInputCharacter:Y,isCharacterSame:Z,onValueChange:H,format:k,removeFormatting:A,getCaretBoundary:function(j){return cse(j,e)},onKeyDown:$,onBlur:U})}function pse(e){var t=fse(e);return ye.createElement(ose,Object.assign({},t))}const Es=E.forwardRef(({stepper:e,thousandSeparator:t,placeholder:n,defaultValue:r,min:a=-1/0,max:o=1/0,onValueChange:s,fixedDecimalScale:u=!1,decimalScale:c=0,className:d=void 0,suffix:p,prefix:g,value:m,...b},y)=>{const[S,x]=E.useState(m??r),R=E.useCallback(()=>{x(N=>N===void 0?e??1:Math.min(N+(e??1),o))},[e,o]),k=E.useCallback(()=>{x(N=>N===void 0?-(e??1):Math.max(N-(e??1),a))},[e,a]);E.useEffect(()=>{m!==void 0&&x(m)},[m]);const A=N=>{const _=N.floatValue===void 0?void 0:N.floatValue;x(_),s&&s(_)},C=()=>{S!==void 0&&(S<a?(x(a),y.current.value=String(a)):S>o&&(x(o),y.current.value=String(o)))};return w.jsxs("div",{className:"relative flex",children:[w.jsx(pse,{value:S,onValueChange:A,thousandSeparator:t,decimalScale:c,fixedDecimalScale:u,allowNegative:a<0,valueIsNumericString:!0,onBlur:C,max:o,min:a,suffix:p,prefix:g,customInput:N=>w.jsx($a,{...N,className:Me("w-full",d)}),placeholder:n,className:"[appearance:textfield] [&::-webkit-inner-spin-button]:appearance-none [&::-webkit-outer-spin-button]:appearance-none",getInputRef:y,...b}),w.jsxs("div",{className:"absolute top-0 right-0 bottom-0 flex flex-col",children:[w.jsx(xt,{"aria-label":"Increase value",className:"border-input h-1/2 rounded-l-none rounded-br-none border-b border-l px-2 focus-visible:relative",variant:"outline",onClick:R,disabled:S===o,children:w.jsx(JB,{size:15})}),w.jsx(xt,{"aria-label":"Decrease value",className:"border-input h-1/2 rounded-l-none rounded-tr-none border-b border-l px-2 focus-visible:relative",variant:"outline",onClick:k,disabled:S===a,children:w.jsx(D0,{size:15})})]})]})});Es.displayName="NumberInput";function gse(){var r,a;const{t:e}=Pt(),t=Fe(o=>o.querySettings),n=E.useCallback((o,s)=>{Fe.getState().updateQuerySettings({[o]:s})},[]);return w.jsxs(Ms,{className:"flex shrink-0 flex-col min-w-[180px]",children:[w.jsxs(Nu,{className:"px-4 pt-4 pb-2",children:[w.jsx(Ou,{children:e("retrievePanel.querySettings.parametersTitle")}),w.jsx(Tp,{children:e("retrievePanel.querySettings.parametersDescription")})]}),w.jsx(Du,{className:"m-0 flex grow flex-col p-0 text-xs",children:w.jsx("div",{className:"relative size-full",children:w.jsxs("div",{className:"absolute inset-0 flex flex-col gap-2 overflow-auto px-2",children:[w.jsxs(w.Fragment,{children:[w.jsx(Bn,{className:"ml-1",text:e("retrievePanel.querySettings.queryMode"),tooltip:e("retrievePanel.querySettings.queryModeTooltip"),side:"left"}),w.jsxs(wf,{value:t.mode,onValueChange:o=>n("mode",o),children:[w.jsx(Au,{className:"hover:bg-primary/5 h-9 cursor-pointer focus:ring-0 focus:ring-offset-0 focus:outline-0 active:right-0",children:w.jsx(xf,{})}),w.jsx(Ru,{children:w.jsxs(aN,{children:[w.jsx(jn,{value:"naive",children:e("retrievePanel.querySettings.queryModeOptions.naive")}),w.jsx(jn,{value:"local",children:e("retrievePanel.querySettings.queryModeOptions.local")}),w.jsx(jn,{value:"global",children:e("retrievePanel.querySettings.queryModeOptions.global")}),w.jsx(jn,{value:"hybrid",children:e("retrievePanel.querySettings.queryModeOptions.hybrid")}),w.jsx(jn,{value:"mix",children:e("retrievePanel.querySettings.queryModeOptions.mix")})]})})]})]}),w.jsxs(w.Fragment,{children:[w.jsx(Bn,{className:"ml-1",text:e("retrievePanel.querySettings.responseFormat"),tooltip:e("retrievePanel.querySettings.responseFormatTooltip"),side:"left"}),w.jsxs(wf,{value:t.response_type,onValueChange:o=>n("response_type",o),children:[w.jsx(Au,{className:"hover:bg-primary/5 h-9 cursor-pointer focus:ring-0 focus:ring-offset-0 focus:outline-0 active:right-0",children:w.jsx(xf,{})}),w.jsx(Ru,{children:w.jsxs(aN,{children:[w.jsx(jn,{value:"Multiple Paragraphs",children:e("retrievePanel.querySettings.responseFormatOptions.multipleParagraphs")}),w.jsx(jn,{value:"Single Paragraph",children:e("retrievePanel.querySettings.responseFormatOptions.singleParagraph")}),w.jsx(jn,{value:"Bullet Points",children:e("retrievePanel.querySettings.responseFormatOptions.bulletPoints")})]})})]})]}),w.jsxs(w.Fragment,{children:[w.jsx(Bn,{className:"ml-1",text:e("retrievePanel.querySettings.topK"),tooltip:e("retrievePanel.querySettings.topKTooltip"),side:"left"}),w.jsx(Es,{id:"top_k",stepper:1,value:t.top_k,onValueChange:o=>n("top_k",o),min:1,placeholder:e("retrievePanel.querySettings.topKPlaceholder")})]}),w.jsxs(w.Fragment,{children:[w.jsxs(w.Fragment,{children:[w.jsx(Bn,{className:"ml-1",text:e("retrievePanel.querySettings.maxTokensTextUnit"),tooltip:e("retrievePanel.querySettings.maxTokensTextUnitTooltip"),side:"left"}),w.jsx(Es,{id:"max_token_for_text_unit",stepper:500,value:t.max_token_for_text_unit,onValueChange:o=>n("max_token_for_text_unit",o),min:1,placeholder:e("retrievePanel.querySettings.maxTokensTextUnit")})]}),w.jsxs(w.Fragment,{children:[w.jsx(Bn,{text:e("retrievePanel.querySettings.maxTokensGlobalContext"),tooltip:e("retrievePanel.querySettings.maxTokensGlobalContextTooltip"),side:"left"}),w.jsx(Es,{id:"max_token_for_global_context",stepper:500,value:t.max_token_for_global_context,onValueChange:o=>n("max_token_for_global_context",o),min:1,placeholder:e("retrievePanel.querySettings.maxTokensGlobalContext")})]}),w.jsxs(w.Fragment,{children:[w.jsx(Bn,{className:"ml-1",text:e("retrievePanel.querySettings.maxTokensLocalContext"),tooltip:e("retrievePanel.querySettings.maxTokensLocalContextTooltip"),side:"left"}),w.jsx(Es,{id:"max_token_for_local_context",stepper:500,value:t.max_token_for_local_context,onValueChange:o=>n("max_token_for_local_context",o),min:1,placeholder:e("retrievePanel.querySettings.maxTokensLocalContext")})]})]}),w.jsxs(w.Fragment,{children:[w.jsx(Bn,{className:"ml-1",text:e("retrievePanel.querySettings.historyTurns"),tooltip:e("retrievePanel.querySettings.historyTurnsTooltip"),side:"left"}),w.jsx(Es,{className:"!border-input",id:"history_turns",stepper:1,type:"text",value:t.history_turns,onValueChange:o=>n("history_turns",o),min:0,placeholder:e("retrievePanel.querySettings.historyTurnsPlaceholder")})]}),w.jsxs(w.Fragment,{children:[w.jsxs(w.Fragment,{children:[w.jsx(Bn,{className:"ml-1",text:e("retrievePanel.querySettings.hlKeywords"),tooltip:e("retrievePanel.querySettings.hlKeywordsTooltip"),side:"left"}),w.jsx($a,{id:"hl_keywords",type:"text",value:(r=t.hl_keywords)==null?void 0:r.join(", "),onChange:o=>{const s=o.target.value.split(",").map(u=>u.trim()).filter(u=>u!=="");n("hl_keywords",s)},placeholder:e("retrievePanel.querySettings.hlkeywordsPlaceHolder")})]}),w.jsxs(w.Fragment,{children:[w.jsx(Bn,{className:"ml-1",text:e("retrievePanel.querySettings.llKeywords"),tooltip:e("retrievePanel.querySettings.llKeywordsTooltip"),side:"left"}),w.jsx($a,{id:"ll_keywords",type:"text",value:(a=t.ll_keywords)==null?void 0:a.join(", "),onChange:o=>{const s=o.target.value.split(",").map(u=>u.trim()).filter(u=>u!=="");n("ll_keywords",s)},placeholder:e("retrievePanel.querySettings.hlkeywordsPlaceHolder")})]})]}),w.jsxs(w.Fragment,{children:[w.jsxs("div",{className:"flex items-center gap-2",children:[w.jsx(Bn,{className:"ml-1",text:e("retrievePanel.querySettings.onlyNeedContext"),tooltip:e("retrievePanel.querySettings.onlyNeedContextTooltip"),side:"left"}),w.jsx("div",{className:"grow"}),w.jsx(mu,{className:"mr-1 cursor-pointer",id:"only_need_context",checked:t.only_need_context,onCheckedChange:o=>n("only_need_context",o)})]}),w.jsxs("div",{className:"flex items-center gap-2",children:[w.jsx(Bn,{className:"ml-1",text:e("retrievePanel.querySettings.onlyNeedPrompt"),tooltip:e("retrievePanel.querySettings.onlyNeedPromptTooltip"),side:"left"}),w.jsx("div",{className:"grow"}),w.jsx(mu,{className:"mr-1 cursor-pointer",id:"only_need_prompt",checked:t.only_need_prompt,onCheckedChange:o=>n("only_need_prompt",o)})]}),w.jsxs("div",{className:"flex items-center gap-2",children:[w.jsx(Bn,{className:"ml-1",text:e("retrievePanel.querySettings.streamResponse"),tooltip:e("retrievePanel.querySettings.streamResponseTooltip"),side:"left"}),w.jsx("div",{className:"grow"}),w.jsx(mu,{className:"mr-1 cursor-pointer",id:"stream",checked:t.stream,onCheckedChange:o=>n("stream",o)})]})]})]})})})]})}function hse(e,t){const n={};return(e[e.length-1]===""?[...e,""]:e).join((n.padRight?" ":"")+","+(n.padLeft===!1?"":" ")).trim()}const mse=/^[$_\p{ID_Start}][$_\u{200C}\u{200D}\p{ID_Continue}]*$/u,bse=/^[$_\p{ID_Start}][-$_\u{200C}\u{200D}\p{ID_Continue}]*$/u,yse={};function eD(e,t){return(yse.jsx?bse:mse).test(e)}const vse=/[ \t\n\f\r]/g;function Sse(e){return typeof e=="object"?e.type==="text"?tD(e.value):!1:tD(e)}function tD(e){return e.replace(vse,"")===""}class rc{constructor(t,n,r){this.property=t,this.normal=n,r&&(this.space=r)}}rc.prototype.property={};rc.prototype.normal={};rc.prototype.space=null;function TG(e,t){const n={},r={};let a=-1;for(;++a<e.length;)Object.assign(n,e[a].property),Object.assign(r,e[a].normal);return new rc(n,r,t)}function l0(e){return e.toLowerCase()}class Ar{constructor(t,n){this.property=t,this.attribute=n}}Ar.prototype.space=null;Ar.prototype.boolean=!1;Ar.prototype.booleanish=!1;Ar.prototype.overloadedBoolean=!1;Ar.prototype.number=!1;Ar.prototype.commaSeparated=!1;Ar.prototype.spaceSeparated=!1;Ar.prototype.commaOrSpaceSeparated=!1;Ar.prototype.mustUseProperty=!1;Ar.prototype.defined=!1;let Ese=0;const rt=Ti(),Qt=Ti(),AG=Ti(),Oe=Ti(),Mt=Ti(),Cs=Ti(),Jn=Ti();function Ti(){return 2**++Ese}const u0=Object.freeze(Object.defineProperty({__proto__:null,boolean:rt,booleanish:Qt,commaOrSpaceSeparated:Jn,commaSeparated:Cs,number:Oe,overloadedBoolean:AG,spaceSeparated:Mt},Symbol.toStringTag,{value:"Module"})),cb=Object.keys(u0);class GT extends Ar{constructor(t,n,r,a){let o=-1;if(super(t,n),nD(this,"space",a),typeof r=="number")for(;++o<cb.length;){const s=cb[o];nD(this,cb[o],(r&u0[s])===u0[s])}}}GT.prototype.defined=!0;function nD(e,t,n){n&&(e[t]=n)}const wse={}.hasOwnProperty;function tl(e){const t={},n={};let r;for(r in e.properties)if(wse.call(e.properties,r)){const a=e.properties[r],o=new GT(r,e.transform(e.attributes||{},r),a,e.space);e.mustUseProperty&&e.mustUseProperty.includes(r)&&(o.mustUseProperty=!0),t[r]=o,n[l0(r)]=r,n[l0(o.attribute)]=r}return new rc(t,n,e.space)}const RG=tl({space:"xlink",transform(e,t){return"xlink:"+t.slice(5).toLowerCase()},properties:{xLinkActuate:null,xLinkArcRole:null,xLinkHref:null,xLinkRole:null,xLinkShow:null,xLinkTitle:null,xLinkType:null}}),_G=tl({space:"xml",transform(e,t){return"xml:"+t.slice(3).toLowerCase()},properties:{xmlLang:null,xmlBase:null,xmlSpace:null}});function CG(e,t){return t in e?e[t]:t}function NG(e,t){return CG(e,t.toLowerCase())}const OG=tl({space:"xmlns",attributes:{xmlnsxlink:"xmlns:xlink"},transform:NG,properties:{xmlns:null,xmlnsXLink:null}}),DG=tl({transform(e,t){return t==="role"?t:"aria-"+t.slice(4).toLowerCase()},properties:{ariaActiveDescendant:null,ariaAtomic:Qt,ariaAutoComplete:null,ariaBusy:Qt,ariaChecked:Qt,ariaColCount:Oe,ariaColIndex:Oe,ariaColSpan:Oe,ariaControls:Mt,ariaCurrent:null,ariaDescribedBy:Mt,ariaDetails:null,ariaDisabled:Qt,ariaDropEffect:Mt,ariaErrorMessage:null,ariaExpanded:Qt,ariaFlowTo:Mt,ariaGrabbed:Qt,ariaHasPopup:null,ariaHidden:Qt,ariaInvalid:null,ariaKeyShortcuts:null,ariaLabel:null,ariaLabelledBy:Mt,ariaLevel:Oe,ariaLive:null,ariaModal:Qt,ariaMultiLine:Qt,ariaMultiSelectable:Qt,ariaOrientation:null,ariaOwns:Mt,ariaPlaceholder:null,ariaPosInSet:Oe,ariaPressed:Qt,ariaReadOnly:Qt,ariaRelevant:null,ariaRequired:Qt,ariaRoleDescription:Mt,ariaRowCount:Oe,ariaRowIndex:Oe,ariaRowSpan:Oe,ariaSelected:Qt,ariaSetSize:Oe,ariaSort:null,ariaValueMax:Oe,ariaValueMin:Oe,ariaValueNow:Oe,ariaValueText:null,role:null}}),xse=tl({space:"html",attributes:{acceptcharset:"accept-charset",classname:"class",htmlfor:"for",httpequiv:"http-equiv"},transform:NG,mustUseProperty:["checked","multiple","muted","selected"],properties:{abbr:null,accept:Cs,acceptCharset:Mt,accessKey:Mt,action:null,allow:null,allowFullScreen:rt,allowPaymentRequest:rt,allowUserMedia:rt,alt:null,as:null,async:rt,autoCapitalize:null,autoComplete:Mt,autoFocus:rt,autoPlay:rt,blocking:Mt,capture:null,charSet:null,checked:rt,cite:null,className:Mt,cols:Oe,colSpan:null,content:null,contentEditable:Qt,controls:rt,controlsList:Mt,coords:Oe|Cs,crossOrigin:null,data:null,dateTime:null,decoding:null,default:rt,defer:rt,dir:null,dirName:null,disabled:rt,download:AG,draggable:Qt,encType:null,enterKeyHint:null,fetchPriority:null,form:null,formAction:null,formEncType:null,formMethod:null,formNoValidate:rt,formTarget:null,headers:Mt,height:Oe,hidden:rt,high:Oe,href:null,hrefLang:null,htmlFor:Mt,httpEquiv:Mt,id:null,imageSizes:null,imageSrcSet:null,inert:rt,inputMode:null,integrity:null,is:null,isMap:rt,itemId:null,itemProp:Mt,itemRef:Mt,itemScope:rt,itemType:Mt,kind:null,label:null,lang:null,language:null,list:null,loading:null,loop:rt,low:Oe,manifest:null,max:null,maxLength:Oe,media:null,method:null,min:null,minLength:Oe,multiple:rt,muted:rt,name:null,nonce:null,noModule:rt,noValidate:rt,onAbort:null,onAfterPrint:null,onAuxClick:null,onBeforeMatch:null,onBeforePrint:null,onBeforeToggle:null,onBeforeUnload:null,onBlur:null,onCancel:null,onCanPlay:null,onCanPlayThrough:null,onChange:null,onClick:null,onClose:null,onContextLost:null,onContextMenu:null,onContextRestored:null,onCopy:null,onCueChange:null,onCut:null,onDblClick:null,onDrag:null,onDragEnd:null,onDragEnter:null,onDragExit:null,onDragLeave:null,onDragOver:null,onDragStart:null,onDrop:null,onDurationChange:null,onEmptied:null,onEnded:null,onError:null,onFocus:null,onFormData:null,onHashChange:null,onInput:null,onInvalid:null,onKeyDown:null,onKeyPress:null,onKeyUp:null,onLanguageChange:null,onLoad:null,onLoadedData:null,onLoadedMetadata:null,onLoadEnd:null,onLoadStart:null,onMessage:null,onMessageError:null,onMouseDown:null,onMouseEnter:null,onMouseLeave:null,onMouseMove:null,onMouseOut:null,onMouseOver:null,onMouseUp:null,onOffline:null,onOnline:null,onPageHide:null,onPageShow:null,onPaste:null,onPause:null,onPlay:null,onPlaying:null,onPopState:null,onProgress:null,onRateChange:null,onRejectionHandled:null,onReset:null,onResize:null,onScroll:null,onScrollEnd:null,onSecurityPolicyViolation:null,onSeeked:null,onSeeking:null,onSelect:null,onSlotChange:null,onStalled:null,onStorage:null,onSubmit:null,onSuspend:null,onTimeUpdate:null,onToggle:null,onUnhandledRejection:null,onUnload:null,onVolumeChange:null,onWaiting:null,onWheel:null,open:rt,optimum:Oe,pattern:null,ping:Mt,placeholder:null,playsInline:rt,popover:null,popoverTarget:null,popoverTargetAction:null,poster:null,preload:null,readOnly:rt,referrerPolicy:null,rel:Mt,required:rt,reversed:rt,rows:Oe,rowSpan:Oe,sandbox:Mt,scope:null,scoped:rt,seamless:rt,selected:rt,shadowRootClonable:rt,shadowRootDelegatesFocus:rt,shadowRootMode:null,shape:null,size:Oe,sizes:null,slot:null,span:Oe,spellCheck:Qt,src:null,srcDoc:null,srcLang:null,srcSet:null,start:Oe,step:null,style:null,tabIndex:Oe,target:null,title:null,translate:null,type:null,typeMustMatch:rt,useMap:null,value:Qt,width:Oe,wrap:null,writingSuggestions:null,align:null,aLink:null,archive:Mt,axis:null,background:null,bgColor:null,border:Oe,borderColor:null,bottomMargin:Oe,cellPadding:null,cellSpacing:null,char:null,charOff:null,classId:null,clear:null,code:null,codeBase:null,codeType:null,color:null,compact:rt,declare:rt,event:null,face:null,frame:null,frameBorder:null,hSpace:Oe,leftMargin:Oe,link:null,longDesc:null,lowSrc:null,marginHeight:Oe,marginWidth:Oe,noResize:rt,noHref:rt,noShade:rt,noWrap:rt,object:null,profile:null,prompt:null,rev:null,rightMargin:Oe,rules:null,scheme:null,scrolling:Qt,standby:null,summary:null,text:null,topMargin:Oe,valueType:null,version:null,vAlign:null,vLink:null,vSpace:Oe,allowTransparency:null,autoCorrect:null,autoSave:null,disablePictureInPicture:rt,disableRemotePlayback:rt,prefix:null,property:null,results:Oe,security:null,unselectable:null}}),kse=tl({space:"svg",attributes:{accentHeight:"accent-height",alignmentBaseline:"alignment-baseline",arabicForm:"arabic-form",baselineShift:"baseline-shift",capHeight:"cap-height",className:"class",clipPath:"clip-path",clipRule:"clip-rule",colorInterpolation:"color-interpolation",colorInterpolationFilters:"color-interpolation-filters",colorProfile:"color-profile",colorRendering:"color-rendering",crossOrigin:"crossorigin",dataType:"datatype",dominantBaseline:"dominant-baseline",enableBackground:"enable-background",fillOpacity:"fill-opacity",fillRule:"fill-rule",floodColor:"flood-color",floodOpacity:"flood-opacity",fontFamily:"font-family",fontSize:"font-size",fontSizeAdjust:"font-size-adjust",fontStretch:"font-stretch",fontStyle:"font-style",fontVariant:"font-variant",fontWeight:"font-weight",glyphName:"glyph-name",glyphOrientationHorizontal:"glyph-orientation-horizontal",glyphOrientationVertical:"glyph-orientation-vertical",hrefLang:"hreflang",horizAdvX:"horiz-adv-x",horizOriginX:"horiz-origin-x",horizOriginY:"horiz-origin-y",imageRendering:"image-rendering",letterSpacing:"letter-spacing",lightingColor:"lighting-color",markerEnd:"marker-end",markerMid:"marker-mid",markerStart:"marker-start",navDown:"nav-down",navDownLeft:"nav-down-left",navDownRight:"nav-down-right",navLeft:"nav-left",navNext:"nav-next",navPrev:"nav-prev",navRight:"nav-right",navUp:"nav-up",navUpLeft:"nav-up-left",navUpRight:"nav-up-right",onAbort:"onabort",onActivate:"onactivate",onAfterPrint:"onafterprint",onBeforePrint:"onbeforeprint",onBegin:"onbegin",onCancel:"oncancel",onCanPlay:"oncanplay",onCanPlayThrough:"oncanplaythrough",onChange:"onchange",onClick:"onclick",onClose:"onclose",onCopy:"oncopy",onCueChange:"oncuechange",onCut:"oncut",onDblClick:"ondblclick",onDrag:"ondrag",onDragEnd:"ondragend",onDragEnter:"ondragenter",onDragExit:"ondragexit",onDragLeave:"ondragleave",onDragOver:"ondragover",onDragStart:"ondragstart",onDrop:"ondrop",onDurationChange:"ondurationchange",onEmptied:"onemptied",onEnd:"onend",onEnded:"onended",onError:"onerror",onFocus:"onfocus",onFocusIn:"onfocusin",onFocusOut:"onfocusout",onHashChange:"onhashchange",onInput:"oninput",onInvalid:"oninvalid",onKeyDown:"onkeydown",onKeyPress:"onkeypress",onKeyUp:"onkeyup",onLoad:"onload",onLoadedData:"onloadeddata",onLoadedMetadata:"onloadedmetadata",onLoadStart:"onloadstart",onMessage:"onmessage",onMouseDown:"onmousedown",onMouseEnter:"onmouseenter",onMouseLeave:"onmouseleave",onMouseMove:"onmousemove",onMouseOut:"onmouseout",onMouseOver:"onmouseover",onMouseUp:"onmouseup",onMouseWheel:"onmousewheel",onOffline:"onoffline",onOnline:"ononline",onPageHide:"onpagehide",onPageShow:"onpageshow",onPaste:"onpaste",onPause:"onpause",onPlay:"onplay",onPlaying:"onplaying",onPopState:"onpopstate",onProgress:"onprogress",onRateChange:"onratechange",onRepeat:"onrepeat",onReset:"onreset",onResize:"onresize",onScroll:"onscroll",onSeeked:"onseeked",onSeeking:"onseeking",onSelect:"onselect",onShow:"onshow",onStalled:"onstalled",onStorage:"onstorage",onSubmit:"onsubmit",onSuspend:"onsuspend",onTimeUpdate:"ontimeupdate",onToggle:"ontoggle",onUnload:"onunload",onVolumeChange:"onvolumechange",onWaiting:"onwaiting",onZoom:"onzoom",overlinePosition:"overline-position",overlineThickness:"overline-thickness",paintOrder:"paint-order",panose1:"panose-1",pointerEvents:"pointer-events",referrerPolicy:"referrerpolicy",renderingIntent:"rendering-intent",shapeRendering:"shape-rendering",stopColor:"stop-color",stopOpacity:"stop-opacity",strikethroughPosition:"strikethrough-position",strikethroughThickness:"strikethrough-thickness",strokeDashArray:"stroke-dasharray",strokeDashOffset:"stroke-dashoffset",strokeLineCap:"stroke-linecap",strokeLineJoin:"stroke-linejoin",strokeMiterLimit:"stroke-miterlimit",strokeOpacity:"stroke-opacity",strokeWidth:"stroke-width",tabIndex:"tabindex",textAnchor:"text-anchor",textDecoration:"text-decoration",textRendering:"text-rendering",transformOrigin:"transform-origin",typeOf:"typeof",underlinePosition:"underline-position",underlineThickness:"underline-thickness",unicodeBidi:"unicode-bidi",unicodeRange:"unicode-range",unitsPerEm:"units-per-em",vAlphabetic:"v-alphabetic",vHanging:"v-hanging",vIdeographic:"v-ideographic",vMathematical:"v-mathematical",vectorEffect:"vector-effect",vertAdvY:"vert-adv-y",vertOriginX:"vert-origin-x",vertOriginY:"vert-origin-y",wordSpacing:"word-spacing",writingMode:"writing-mode",xHeight:"x-height",playbackOrder:"playbackorder",timelineBegin:"timelinebegin"},transform:CG,properties:{about:Jn,accentHeight:Oe,accumulate:null,additive:null,alignmentBaseline:null,alphabetic:Oe,amplitude:Oe,arabicForm:null,ascent:Oe,attributeName:null,attributeType:null,azimuth:Oe,bandwidth:null,baselineShift:null,baseFrequency:null,baseProfile:null,bbox:null,begin:null,bias:Oe,by:null,calcMode:null,capHeight:Oe,className:Mt,clip:null,clipPath:null,clipPathUnits:null,clipRule:null,color:null,colorInterpolation:null,colorInterpolationFilters:null,colorProfile:null,colorRendering:null,content:null,contentScriptType:null,contentStyleType:null,crossOrigin:null,cursor:null,cx:null,cy:null,d:null,dataType:null,defaultAction:null,descent:Oe,diffuseConstant:Oe,direction:null,display:null,dur:null,divisor:Oe,dominantBaseline:null,download:rt,dx:null,dy:null,edgeMode:null,editable:null,elevation:Oe,enableBackground:null,end:null,event:null,exponent:Oe,externalResourcesRequired:null,fill:null,fillOpacity:Oe,fillRule:null,filter:null,filterRes:null,filterUnits:null,floodColor:null,floodOpacity:null,focusable:null,focusHighlight:null,fontFamily:null,fontSize:null,fontSizeAdjust:null,fontStretch:null,fontStyle:null,fontVariant:null,fontWeight:null,format:null,fr:null,from:null,fx:null,fy:null,g1:Cs,g2:Cs,glyphName:Cs,glyphOrientationHorizontal:null,glyphOrientationVertical:null,glyphRef:null,gradientTransform:null,gradientUnits:null,handler:null,hanging:Oe,hatchContentUnits:null,hatchUnits:null,height:null,href:null,hrefLang:null,horizAdvX:Oe,horizOriginX:Oe,horizOriginY:Oe,id:null,ideographic:Oe,imageRendering:null,initialVisibility:null,in:null,in2:null,intercept:Oe,k:Oe,k1:Oe,k2:Oe,k3:Oe,k4:Oe,kernelMatrix:Jn,kernelUnitLength:null,keyPoints:null,keySplines:null,keyTimes:null,kerning:null,lang:null,lengthAdjust:null,letterSpacing:null,lightingColor:null,limitingConeAngle:Oe,local:null,markerEnd:null,markerMid:null,markerStart:null,markerHeight:null,markerUnits:null,markerWidth:null,mask:null,maskContentUnits:null,maskUnits:null,mathematical:null,max:null,media:null,mediaCharacterEncoding:null,mediaContentEncodings:null,mediaSize:Oe,mediaTime:null,method:null,min:null,mode:null,name:null,navDown:null,navDownLeft:null,navDownRight:null,navLeft:null,navNext:null,navPrev:null,navRight:null,navUp:null,navUpLeft:null,navUpRight:null,numOctaves:null,observer:null,offset:null,onAbort:null,onActivate:null,onAfterPrint:null,onBeforePrint:null,onBegin:null,onCancel:null,onCanPlay:null,onCanPlayThrough:null,onChange:null,onClick:null,onClose:null,onCopy:null,onCueChange:null,onCut:null,onDblClick:null,onDrag:null,onDragEnd:null,onDragEnter:null,onDragExit:null,onDragLeave:null,onDragOver:null,onDragStart:null,onDrop:null,onDurationChange:null,onEmptied:null,onEnd:null,onEnded:null,onError:null,onFocus:null,onFocusIn:null,onFocusOut:null,onHashChange:null,onInput:null,onInvalid:null,onKeyDown:null,onKeyPress:null,onKeyUp:null,onLoad:null,onLoadedData:null,onLoadedMetadata:null,onLoadStart:null,onMessage:null,onMouseDown:null,onMouseEnter:null,onMouseLeave:null,onMouseMove:null,onMouseOut:null,onMouseOver:null,onMouseUp:null,onMouseWheel:null,onOffline:null,onOnline:null,onPageHide:null,onPageShow:null,onPaste:null,onPause:null,onPlay:null,onPlaying:null,onPopState:null,onProgress:null,onRateChange:null,onRepeat:null,onReset:null,onResize:null,onScroll:null,onSeeked:null,onSeeking:null,onSelect:null,onShow:null,onStalled:null,onStorage:null,onSubmit:null,onSuspend:null,onTimeUpdate:null,onToggle:null,onUnload:null,onVolumeChange:null,onWaiting:null,onZoom:null,opacity:null,operator:null,order:null,orient:null,orientation:null,origin:null,overflow:null,overlay:null,overlinePosition:Oe,overlineThickness:Oe,paintOrder:null,panose1:null,path:null,pathLength:Oe,patternContentUnits:null,patternTransform:null,patternUnits:null,phase:null,ping:Mt,pitch:null,playbackOrder:null,pointerEvents:null,points:null,pointsAtX:Oe,pointsAtY:Oe,pointsAtZ:Oe,preserveAlpha:null,preserveAspectRatio:null,primitiveUnits:null,propagate:null,property:Jn,r:null,radius:null,referrerPolicy:null,refX:null,refY:null,rel:Jn,rev:Jn,renderingIntent:null,repeatCount:null,repeatDur:null,requiredExtensions:Jn,requiredFeatures:Jn,requiredFonts:Jn,requiredFormats:Jn,resource:null,restart:null,result:null,rotate:null,rx:null,ry:null,scale:null,seed:null,shapeRendering:null,side:null,slope:null,snapshotTime:null,specularConstant:Oe,specularExponent:Oe,spreadMethod:null,spacing:null,startOffset:null,stdDeviation:null,stemh:null,stemv:null,stitchTiles:null,stopColor:null,stopOpacity:null,strikethroughPosition:Oe,strikethroughThickness:Oe,string:null,stroke:null,strokeDashArray:Jn,strokeDashOffset:null,strokeLineCap:null,strokeLineJoin:null,strokeMiterLimit:Oe,strokeOpacity:Oe,strokeWidth:null,style:null,surfaceScale:Oe,syncBehavior:null,syncBehaviorDefault:null,syncMaster:null,syncTolerance:null,syncToleranceDefault:null,systemLanguage:Jn,tabIndex:Oe,tableValues:null,target:null,targetX:Oe,targetY:Oe,textAnchor:null,textDecoration:null,textRendering:null,textLength:null,timelineBegin:null,title:null,transformBehavior:null,type:null,typeOf:Jn,to:null,transform:null,transformOrigin:null,u1:null,u2:null,underlinePosition:Oe,underlineThickness:Oe,unicode:null,unicodeBidi:null,unicodeRange:null,unitsPerEm:Oe,values:null,vAlphabetic:Oe,vMathematical:Oe,vectorEffect:null,vHanging:Oe,vIdeographic:Oe,version:null,vertAdvY:Oe,vertOriginX:Oe,vertOriginY:Oe,viewBox:null,viewTarget:null,visibility:null,width:null,widths:null,wordSpacing:null,writingMode:null,x:null,x1:null,x2:null,xChannelSelector:null,xHeight:Oe,y:null,y1:null,y2:null,yChannelSelector:null,z:null,zoomAndPan:null}}),Tse=/^data[-\w.:]+$/i,rD=/-[a-z]/g,Ase=/[A-Z]/g;function Rse(e,t){const n=l0(t);let r=t,a=Ar;if(n in e.normal)return e.property[e.normal[n]];if(n.length>4&&n.slice(0,4)==="data"&&Tse.test(t)){if(t.charAt(4)==="-"){const o=t.slice(5).replace(rD,Cse);r="data"+o.charAt(0).toUpperCase()+o.slice(1)}else{const o=t.slice(4);if(!rD.test(o)){let s=o.replace(Ase,_se);s.charAt(0)!=="-"&&(s="-"+s),t="data"+s}}a=GT}return new a(r,t)}function _se(e){return"-"+e.toLowerCase()}function Cse(e){return e.charAt(1).toUpperCase()}const Nse={classId:"classID",dataType:"datatype",itemId:"itemID",strokeDashArray:"strokeDasharray",strokeDashOffset:"strokeDashoffset",strokeLineCap:"strokeLinecap",strokeLineJoin:"strokeLinejoin",strokeMiterLimit:"strokeMiterlimit",typeOf:"typeof",xLinkActuate:"xlinkActuate",xLinkArcRole:"xlinkArcrole",xLinkHref:"xlinkHref",xLinkRole:"xlinkRole",xLinkShow:"xlinkShow",xLinkTitle:"xlinkTitle",xLinkType:"xlinkType",xmlnsXLink:"xmlnsXlink"},Ose=TG([_G,RG,OG,DG,xse],"html"),HT=TG([_G,RG,OG,DG,kse],"svg");function Dse(e){return e.join(" ").trim()}var gs={},db,aD;function Ise(){if(aD)return db;aD=1;var e=/\/\*[^*]*\*+([^/*][^*]*\*+)*\//g,t=/\n/g,n=/^\s*/,r=/^(\*?[-#/*\\\w]+(\[[0-9a-z_-]+\])?)\s*/,a=/^:\s*/,o=/^((?:'(?:\\'|.)*?'|"(?:\\"|.)*?"|\([^)]*?\)|[^};])+)/,s=/^[;\s]*/,u=/^\s+|\s+$/g,c=`
`,d="/",p="*",g="",m="comment",b="declaration";db=function(S,x){if(typeof S!="string")throw new TypeError("First argument must be a string");if(!S)return[];x=x||{};var R=1,k=1;function A(U){var Y=U.match(t);Y&&(R+=Y.length);var Z=U.lastIndexOf(c);k=~Z?U.length-Z:k+U.length}function C(){var U={line:R,column:k};return function(Y){return Y.position=new N(U),F(),Y}}function N(U){this.start=U,this.end={line:R,column:k},this.source=x.source}N.prototype.content=S;function _(U){var Y=new Error(x.source+":"+R+":"+k+": "+U);if(Y.reason=U,Y.filename=x.source,Y.line=R,Y.column=k,Y.source=S,!x.silent)throw Y}function O(U){var Y=U.exec(S);if(Y){var Z=Y[0];return A(Z),S=S.slice(Z.length),Y}}function F(){O(n)}function L(U){var Y;for(U=U||[];Y=I();)Y!==!1&&U.push(Y);return U}function I(){var U=C();if(!(d!=S.charAt(0)||p!=S.charAt(1))){for(var Y=2;g!=S.charAt(Y)&&(p!=S.charAt(Y)||d!=S.charAt(Y+1));)++Y;if(Y+=2,g===S.charAt(Y-1))return _("End of comment missing");var Z=S.slice(2,Y-2);return k+=2,A(Z),S=S.slice(Y),k+=2,U({type:m,comment:Z})}}function H(){var U=C(),Y=O(r);if(Y){if(I(),!O(a))return _("property missing ':'");var Z=O(o),j=U({type:b,property:y(Y[0].replace(e,g)),value:Z?y(Z[0].replace(e,g)):g});return O(s),j}}function $(){var U=[];L(U);for(var Y;Y=H();)Y!==!1&&(U.push(Y),L(U));return U}return F(),$()};function y(S){return S?S.replace(u,g):g}return db}var oD;function Lse(){if(oD)return gs;oD=1;var e=gs&&gs.__importDefault||function(r){return r&&r.__esModule?r:{default:r}};Object.defineProperty(gs,"__esModule",{value:!0}),gs.default=n;var t=e(Ise());function n(r,a){var o=null;if(!r||typeof r!="string")return o;var s=(0,t.default)(r),u=typeof a=="function";return s.forEach(function(c){if(c.type==="declaration"){var d=c.property,p=c.value;u?a(d,p,c):p&&(o=o||{},o[d]=p)}}),o}return gs}var Mse=Lse();const iD=un(Mse),Fse=iD.default||iD,IG=LG("end"),$T=LG("start");function LG(e){return t;function t(n){const r=n&&n.position&&n.position[e]||{};if(typeof r.line=="number"&&r.line>0&&typeof r.column=="number"&&r.column>0)return{line:r.line,column:r.column,offset:typeof r.offset=="number"&&r.offset>-1?r.offset:void 0}}}function Pse(e){const t=$T(e),n=IG(e);if(t&&n)return{start:t,end:n}}function bu(e){return!e||typeof e!="object"?"":"position"in e||"type"in e?sD(e.position):"start"in e||"end"in e?sD(e):"line"in e||"column"in e?c0(e):""}function c0(e){return lD(e&&e.line)+":"+lD(e&&e.column)}function sD(e){return c0(e&&e.start)+"-"+c0(e&&e.end)}function lD(e){return e&&typeof e=="number"?e:1}class Rn extends Error{constructor(t,n,r){super(),typeof n=="string"&&(r=n,n=void 0);let a="",o={},s=!1;if(n&&("line"in n&&"column"in n?o={place:n}:"start"in n&&"end"in n?o={place:n}:"type"in n?o={ancestors:[n],place:n.position}:o={...n}),typeof t=="string"?a=t:!o.cause&&t&&(s=!0,a=t.message,o.cause=t),!o.ruleId&&!o.source&&typeof r=="string"){const c=r.indexOf(":");c===-1?o.ruleId=r:(o.source=r.slice(0,c),o.ruleId=r.slice(c+1))}if(!o.place&&o.ancestors&&o.ancestors){const c=o.ancestors[o.ancestors.length-1];c&&(o.place=c.position)}const u=o.place&&"start"in o.place?o.place.start:o.place;this.ancestors=o.ancestors||void 0,this.cause=o.cause||void 0,this.column=u?u.column:void 0,this.fatal=void 0,this.file,this.message=a,this.line=u?u.line:void 0,this.name=bu(o.place)||"1:1",this.place=o.place||void 0,this.reason=this.message,this.ruleId=o.ruleId||void 0,this.source=o.source||void 0,this.stack=s&&o.cause&&typeof o.cause.stack=="string"?o.cause.stack:"",this.actual,this.expected,this.note,this.url}}Rn.prototype.file="";Rn.prototype.name="";Rn.prototype.reason="";Rn.prototype.message="";Rn.prototype.stack="";Rn.prototype.column=void 0;Rn.prototype.line=void 0;Rn.prototype.ancestors=void 0;Rn.prototype.cause=void 0;Rn.prototype.fatal=void 0;Rn.prototype.place=void 0;Rn.prototype.ruleId=void 0;Rn.prototype.source=void 0;const qT={}.hasOwnProperty,zse=new Map,Bse=/[A-Z]/g,Use=/-([a-z])/g,jse=new Set(["table","tbody","thead","tfoot","tr"]),Gse=new Set(["td","th"]),MG="https://github.com/syntax-tree/hast-util-to-jsx-runtime";function FG(e,t){if(!t||t.Fragment===void 0)throw new TypeError("Expected `Fragment` in options");const n=t.filePath||void 0;let r;if(t.development){if(typeof t.jsxDEV!="function")throw new TypeError("Expected `jsxDEV` in options when `development: true`");r=Xse(n,t.jsxDEV)}else{if(typeof t.jsx!="function")throw new TypeError("Expected `jsx` in production options");if(typeof t.jsxs!="function")throw new TypeError("Expected `jsxs` in production options");r=Kse(n,t.jsx,t.jsxs)}const a={Fragment:t.Fragment,ancestors:[],components:t.components||{},create:r,elementAttributeNameCase:t.elementAttributeNameCase||"react",evaluater:t.createEvaluater?t.createEvaluater():void 0,filePath:n,ignoreInvalidStyle:t.ignoreInvalidStyle||!1,passKeys:t.passKeys!==!1,passNode:t.passNode||!1,schema:t.space==="svg"?HT:Ose,stylePropertyNameCase:t.stylePropertyNameCase||"dom",tableCellAlignToStyle:t.tableCellAlignToStyle!==!1},o=PG(a,e,void 0);return o&&typeof o!="string"?o:a.create(e,a.Fragment,{children:o||void 0},void 0)}function PG(e,t,n){if(t.type==="element")return Hse(e,t,n);if(t.type==="mdxFlowExpression"||t.type==="mdxTextExpression")return $se(e,t);if(t.type==="mdxJsxFlowElement"||t.type==="mdxJsxTextElement")return Vse(e,t,n);if(t.type==="mdxjsEsm")return qse(e,t);if(t.type==="root")return Wse(e,t,n);if(t.type==="text")return Yse(e,t)}function Hse(e,t,n){const r=e.schema;let a=r;t.tagName.toLowerCase()==="svg"&&r.space==="html"&&(a=HT,e.schema=a),e.ancestors.push(t);const o=BG(e,t.tagName,!1),s=Zse(e,t);let u=WT(e,t);return jse.has(t.tagName)&&(u=u.filter(function(c){return typeof c=="string"?!Sse(c):!0})),zG(e,s,o,t),VT(s,u),e.ancestors.pop(),e.schema=r,e.create(t,o,s,n)}function $se(e,t){if(t.data&&t.data.estree&&e.evaluater){const r=t.data.estree.body[0];return r.type,e.evaluater.evaluateExpression(r.expression)}Fu(e,t.position)}function qse(e,t){if(t.data&&t.data.estree&&e.evaluater)return e.evaluater.evaluateProgram(t.data.estree);Fu(e,t.position)}function Vse(e,t,n){const r=e.schema;let a=r;t.name==="svg"&&r.space==="html"&&(a=HT,e.schema=a),e.ancestors.push(t);const o=t.name===null?e.Fragment:BG(e,t.name,!0),s=Qse(e,t),u=WT(e,t);return zG(e,s,o,t),VT(s,u),e.ancestors.pop(),e.schema=r,e.create(t,o,s,n)}function Wse(e,t,n){const r={};return VT(r,WT(e,t)),e.create(t,e.Fragment,r,n)}function Yse(e,t){return t.value}function zG(e,t,n,r){typeof n!="string"&&n!==e.Fragment&&e.passNode&&(t.node=r)}function VT(e,t){if(t.length>0){const n=t.length>1?t:t[0];n&&(e.children=n)}}function Kse(e,t,n){return r;function r(a,o,s,u){const d=Array.isArray(s.children)?n:t;return u?d(o,s,u):d(o,s)}}function Xse(e,t){return n;function n(r,a,o,s){const u=Array.isArray(o.children),c=$T(r);return t(a,o,s,u,{columnNumber:c?c.column-1:void 0,fileName:e,lineNumber:c?c.line:void 0},void 0)}}function Zse(e,t){const n={};let r,a;for(a in t.properties)if(a!=="children"&&qT.call(t.properties,a)){const o=Jse(e,a,t.properties[a]);if(o){const[s,u]=o;e.tableCellAlignToStyle&&s==="align"&&typeof u=="string"&&Gse.has(t.tagName)?r=u:n[s]=u}}if(r){const o=n.style||(n.style={});o[e.stylePropertyNameCase==="css"?"text-align":"textAlign"]=r}return n}function Qse(e,t){const n={};for(const r of t.attributes)if(r.type==="mdxJsxExpressionAttribute")if(r.data&&r.data.estree&&e.evaluater){const o=r.data.estree.body[0];o.type;const s=o.expression;s.type;const u=s.properties[0];u.type,Object.assign(n,e.evaluater.evaluateExpression(u.argument))}else Fu(e,t.position);else{const a=r.name;let o;if(r.value&&typeof r.value=="object")if(r.value.data&&r.value.data.estree&&e.evaluater){const u=r.value.data.estree.body[0];u.type,o=e.evaluater.evaluateExpression(u.expression)}else Fu(e,t.position);else o=r.value===null?!0:r.value;n[a]=o}return n}function WT(e,t){const n=[];let r=-1;const a=e.passKeys?new Map:zse;for(;++r<t.children.length;){const o=t.children[r];let s;if(e.passKeys){const c=o.type==="element"?o.tagName:o.type==="mdxJsxFlowElement"||o.type==="mdxJsxTextElement"?o.name:void 0;if(c){const d=a.get(c)||0;s=c+"-"+d,a.set(c,d+1)}}const u=PG(e,o,s);u!==void 0&&n.push(u)}return n}function Jse(e,t,n){const r=Rse(e.schema,t);if(!(n==null||typeof n=="number"&&Number.isNaN(n))){if(Array.isArray(n)&&(n=r.commaSeparated?hse(n):Dse(n)),r.property==="style"){let a=typeof n=="object"?n:ele(e,String(n));return e.stylePropertyNameCase==="css"&&(a=tle(a)),["style",a]}return[e.elementAttributeNameCase==="react"&&r.space?Nse[r.property]||r.property:r.attribute,n]}}function ele(e,t){const n={};try{Fse(t,r)}catch(a){if(!e.ignoreInvalidStyle){const o=a,s=new Rn("Cannot parse `style` attribute",{ancestors:e.ancestors,cause:o,ruleId:"style",source:"hast-util-to-jsx-runtime"});throw s.file=e.filePath||void 0,s.url=MG+"#cannot-parse-style-attribute",s}}return n;function r(a,o){let s=a;s.slice(0,2)!=="--"&&(s.slice(0,4)==="-ms-"&&(s="ms-"+s.slice(4)),s=s.replace(Use,rle)),n[s]=o}}function BG(e,t,n){let r;if(!n)r={type:"Literal",value:t};else if(t.includes(".")){const a=t.split(".");let o=-1,s;for(;++o<a.length;){const u=eD(a[o])?{type:"Identifier",name:a[o]}:{type:"Literal",value:a[o]};s=s?{type:"MemberExpression",object:s,property:u,computed:!!(o&&u.type==="Literal"),optional:!1}:u}r=s}else r=eD(t)&&!/^[a-z]/.test(t)?{type:"Identifier",name:t}:{type:"Literal",value:t};if(r.type==="Literal"){const a=r.value;return qT.call(e.components,a)?e.components[a]:a}if(e.evaluater)return e.evaluater.evaluateExpression(r);Fu(e)}function Fu(e,t){const n=new Rn("Cannot handle MDX estrees without `createEvaluater`",{ancestors:e.ancestors,place:t,ruleId:"mdx-estree",source:"hast-util-to-jsx-runtime"});throw n.file=e.filePath||void 0,n.url=MG+"#cannot-handle-mdx-estrees-without-createevaluater",n}function tle(e){const t={};let n;for(n in e)qT.call(e,n)&&(t[nle(n)]=e[n]);return t}function nle(e){let t=e.replace(Bse,ale);return t.slice(0,3)==="ms-"&&(t="-"+t),t}function rle(e,t){return t.toUpperCase()}function ale(e){return"-"+e.toLowerCase()}const fb={action:["form"],cite:["blockquote","del","ins","q"],data:["object"],formAction:["button","input"],href:["a","area","base","link"],icon:["menuitem"],itemId:null,manifest:["html"],ping:["a","area"],poster:["video"],src:["audio","embed","iframe","img","input","script","source","track","video"]},ole={};function YT(e,t){const n=ole,r=typeof n.includeImageAlt=="boolean"?n.includeImageAlt:!0,a=typeof n.includeHtml=="boolean"?n.includeHtml:!0;return UG(e,r,a)}function UG(e,t,n){if(ile(e)){if("value"in e)return e.type==="html"&&!n?"":e.value;if(t&&"alt"in e&&e.alt)return e.alt;if("children"in e)return uD(e.children,t,n)}return Array.isArray(e)?uD(e,t,n):""}function uD(e,t,n){const r=[];let a=-1;for(;++a<e.length;)r[a]=UG(e[a],t,n);return r.join("")}function ile(e){return!!(e&&typeof e=="object")}const cD=document.createElement("i");function KT(e){const t="&"+e+";";cD.innerHTML=t;const n=cD.textContent;return n.charCodeAt(n.length-1)===59&&e!=="semi"||n===t?!1:n}function rr(e,t,n,r){const a=e.length;let o=0,s;if(t<0?t=-t>a?0:a+t:t=t>a?a:t,n=n>0?n:0,r.length<1e4)s=Array.from(r),s.unshift(t,n),e.splice(...s);else for(n&&e.splice(t,n);o<r.length;)s=r.slice(o,o+1e4),s.unshift(t,0),e.splice(...s),o+=1e4,t+=1e4}function vr(e,t){return e.length>0?(rr(e,e.length,0,t),e):t}const dD={}.hasOwnProperty;function jG(e){const t={};let n=-1;for(;++n<e.length;)sle(t,e[n]);return t}function sle(e,t){let n;for(n in t){const a=(dD.call(e,n)?e[n]:void 0)||(e[n]={}),o=t[n];let s;if(o)for(s in o){dD.call(a,s)||(a[s]=[]);const u=o[s];lle(a[s],Array.isArray(u)?u:u?[u]:[])}}}function lle(e,t){let n=-1;const r=[];for(;++n<t.length;)(t[n].add==="after"?e:r).push(t[n]);rr(e,0,0,r)}function GG(e,t){const n=Number.parseInt(e,t);return n<9||n===11||n>13&&n<32||n>126&&n<160||n>55295&&n<57344||n>64975&&n<65008||(n&65535)===65535||(n&65535)===65534||n>1114111?"�":String.fromCodePoint(n)}function zr(e){return e.replace(/[\t\n\r ]+/g," ").replace(/^ | $/g,"").toLowerCase().toUpperCase()}const On=Mo(/[A-Za-z]/),kn=Mo(/[\dA-Za-z]/),ule=Mo(/[#-'*+\--9=?A-Z^-~]/);function Ff(e){return e!==null&&(e<32||e===127)}const d0=Mo(/\d/),cle=Mo(/[\dA-Fa-f]/),dle=Mo(/[!-/:-@[-`{-~]/);function Ve(e){return e!==null&&e<-2}function Dt(e){return e!==null&&(e<0||e===32)}function dt(e){return e===-2||e===-1||e===32}const Np=Mo(new RegExp("\\p{P}|\\p{S}","u")),Si=Mo(/\s/);function Mo(e){return t;function t(n){return n!==null&&n>-1&&e.test(String.fromCharCode(n))}}function nl(e){const t=[];let n=-1,r=0,a=0;for(;++n<e.length;){const o=e.charCodeAt(n);let s="";if(o===37&&kn(e.charCodeAt(n+1))&&kn(e.charCodeAt(n+2)))a=2;else if(o<128)/[!#$&-;=?-Z_a-z~]/.test(String.fromCharCode(o))||(s=String.fromCharCode(o));else if(o>55295&&o<57344){const u=e.charCodeAt(n+1);o<56320&&u>56319&&u<57344?(s=String.fromCharCode(o,u),a=1):s="�"}else s=String.fromCharCode(o);s&&(t.push(e.slice(r,n),encodeURIComponent(s)),r=n+a+1,s=""),a&&(n+=a,a=0)}return t.join("")+e.slice(r)}function ut(e,t,n,r){const a=r?r-1:Number.POSITIVE_INFINITY;let o=0;return s;function s(c){return dt(c)?(e.enter(n),u(c)):t(c)}function u(c){return dt(c)&&o++<a?(e.consume(c),u):(e.exit(n),t(c))}}const fle={tokenize:ple};function ple(e){const t=e.attempt(this.parser.constructs.contentInitial,r,a);let n;return t;function r(u){if(u===null){e.consume(u);return}return e.enter("lineEnding"),e.consume(u),e.exit("lineEnding"),ut(e,t,"linePrefix")}function a(u){return e.enter("paragraph"),o(u)}function o(u){const c=e.enter("chunkText",{contentType:"text",previous:n});return n&&(n.next=c),n=c,s(u)}function s(u){if(u===null){e.exit("chunkText"),e.exit("paragraph"),e.consume(u);return}return Ve(u)?(e.consume(u),e.exit("chunkText"),o):(e.consume(u),s)}}const gle={tokenize:hle},fD={tokenize:mle};function hle(e){const t=this,n=[];let r=0,a,o,s;return u;function u(A){if(r<n.length){const C=n[r];return t.containerState=C[1],e.attempt(C[0].continuation,c,d)(A)}return d(A)}function c(A){if(r++,t.containerState._closeFlow){t.containerState._closeFlow=void 0,a&&k();const C=t.events.length;let N=C,_;for(;N--;)if(t.events[N][0]==="exit"&&t.events[N][1].type==="chunkFlow"){_=t.events[N][1].end;break}R(r);let O=C;for(;O<t.events.length;)t.events[O][1].end={..._},O++;return rr(t.events,N+1,0,t.events.slice(C)),t.events.length=O,d(A)}return u(A)}function d(A){if(r===n.length){if(!a)return m(A);if(a.currentConstruct&&a.currentConstruct.concrete)return y(A);t.interrupt=!!(a.currentConstruct&&!a._gfmTableDynamicInterruptHack)}return t.containerState={},e.check(fD,p,g)(A)}function p(A){return a&&k(),R(r),m(A)}function g(A){return t.parser.lazy[t.now().line]=r!==n.length,s=t.now().offset,y(A)}function m(A){return t.containerState={},e.attempt(fD,b,y)(A)}function b(A){return r++,n.push([t.currentConstruct,t.containerState]),m(A)}function y(A){if(A===null){a&&k(),R(0),e.consume(A);return}return a=a||t.parser.flow(t.now()),e.enter("chunkFlow",{_tokenizer:a,contentType:"flow",previous:o}),S(A)}function S(A){if(A===null){x(e.exit("chunkFlow"),!0),R(0),e.consume(A);return}return Ve(A)?(e.consume(A),x(e.exit("chunkFlow")),r=0,t.interrupt=void 0,u):(e.consume(A),S)}function x(A,C){const N=t.sliceStream(A);if(C&&N.push(null),A.previous=o,o&&(o.next=A),o=A,a.defineSkip(A.start),a.write(N),t.parser.lazy[A.start.line]){let _=a.events.length;for(;_--;)if(a.events[_][1].start.offset<s&&(!a.events[_][1].end||a.events[_][1].end.offset>s))return;const O=t.events.length;let F=O,L,I;for(;F--;)if(t.events[F][0]==="exit"&&t.events[F][1].type==="chunkFlow"){if(L){I=t.events[F][1].end;break}L=!0}for(R(r),_=O;_<t.events.length;)t.events[_][1].end={...I},_++;rr(t.events,F+1,0,t.events.slice(O)),t.events.length=_}}function R(A){let C=n.length;for(;C-- >A;){const N=n[C];t.containerState=N[1],N[0].exit.call(t,e)}n.length=A}function k(){a.write([null]),o=void 0,a=void 0,t.containerState._closeFlow=void 0}}function mle(e,t,n){return ut(e,e.attempt(this.parser.constructs.document,t,n),"linePrefix",this.parser.constructs.disable.null.includes("codeIndented")?void 0:4)}function zs(e){if(e===null||Dt(e)||Si(e))return 1;if(Np(e))return 2}function Op(e,t,n){const r=[];let a=-1;for(;++a<e.length;){const o=e[a].resolveAll;o&&!r.includes(o)&&(t=o(t,n),r.push(o))}return t}const f0={name:"attention",resolveAll:ble,tokenize:yle};function ble(e,t){let n=-1,r,a,o,s,u,c,d,p;for(;++n<e.length;)if(e[n][0]==="enter"&&e[n][1].type==="attentionSequence"&&e[n][1]._close){for(r=n;r--;)if(e[r][0]==="exit"&&e[r][1].type==="attentionSequence"&&e[r][1]._open&&t.sliceSerialize(e[r][1]).charCodeAt(0)===t.sliceSerialize(e[n][1]).charCodeAt(0)){if((e[r][1]._close||e[n][1]._open)&&(e[n][1].end.offset-e[n][1].start.offset)%3&&!((e[r][1].end.offset-e[r][1].start.offset+e[n][1].end.offset-e[n][1].start.offset)%3))continue;c=e[r][1].end.offset-e[r][1].start.offset>1&&e[n][1].end.offset-e[n][1].start.offset>1?2:1;const g={...e[r][1].end},m={...e[n][1].start};pD(g,-c),pD(m,c),s={type:c>1?"strongSequence":"emphasisSequence",start:g,end:{...e[r][1].end}},u={type:c>1?"strongSequence":"emphasisSequence",start:{...e[n][1].start},end:m},o={type:c>1?"strongText":"emphasisText",start:{...e[r][1].end},end:{...e[n][1].start}},a={type:c>1?"strong":"emphasis",start:{...s.start},end:{...u.end}},e[r][1].end={...s.start},e[n][1].start={...u.end},d=[],e[r][1].end.offset-e[r][1].start.offset&&(d=vr(d,[["enter",e[r][1],t],["exit",e[r][1],t]])),d=vr(d,[["enter",a,t],["enter",s,t],["exit",s,t],["enter",o,t]]),d=vr(d,Op(t.parser.constructs.insideSpan.null,e.slice(r+1,n),t)),d=vr(d,[["exit",o,t],["enter",u,t],["exit",u,t],["exit",a,t]]),e[n][1].end.offset-e[n][1].start.offset?(p=2,d=vr(d,[["enter",e[n][1],t],["exit",e[n][1],t]])):p=0,rr(e,r-1,n-r+3,d),n=r+d.length-p-2;break}}for(n=-1;++n<e.length;)e[n][1].type==="attentionSequence"&&(e[n][1].type="data");return e}function yle(e,t){const n=this.parser.constructs.attentionMarkers.null,r=this.previous,a=zs(r);let o;return s;function s(c){return o=c,e.enter("attentionSequence"),u(c)}function u(c){if(c===o)return e.consume(c),u;const d=e.exit("attentionSequence"),p=zs(c),g=!p||p===2&&a||n.includes(c),m=!a||a===2&&p||n.includes(r);return d._open=!!(o===42?g:g&&(a||!m)),d._close=!!(o===42?m:m&&(p||!g)),t(c)}}function pD(e,t){e.column+=t,e.offset+=t,e._bufferIndex+=t}const vle={name:"autolink",tokenize:Sle};function Sle(e,t,n){let r=0;return a;function a(b){return e.enter("autolink"),e.enter("autolinkMarker"),e.consume(b),e.exit("autolinkMarker"),e.enter("autolinkProtocol"),o}function o(b){return On(b)?(e.consume(b),s):b===64?n(b):d(b)}function s(b){return b===43||b===45||b===46||kn(b)?(r=1,u(b)):d(b)}function u(b){return b===58?(e.consume(b),r=0,c):(b===43||b===45||b===46||kn(b))&&r++<32?(e.consume(b),u):(r=0,d(b))}function c(b){return b===62?(e.exit("autolinkProtocol"),e.enter("autolinkMarker"),e.consume(b),e.exit("autolinkMarker"),e.exit("autolink"),t):b===null||b===32||b===60||Ff(b)?n(b):(e.consume(b),c)}function d(b){return b===64?(e.consume(b),p):ule(b)?(e.consume(b),d):n(b)}function p(b){return kn(b)?g(b):n(b)}function g(b){return b===46?(e.consume(b),r=0,p):b===62?(e.exit("autolinkProtocol").type="autolinkEmail",e.enter("autolinkMarker"),e.consume(b),e.exit("autolinkMarker"),e.exit("autolink"),t):m(b)}function m(b){if((b===45||kn(b))&&r++<63){const y=b===45?m:g;return e.consume(b),y}return n(b)}}const ac={partial:!0,tokenize:Ele};function Ele(e,t,n){return r;function r(o){return dt(o)?ut(e,a,"linePrefix")(o):a(o)}function a(o){return o===null||Ve(o)?t(o):n(o)}}const HG={continuation:{tokenize:xle},exit:kle,name:"blockQuote",tokenize:wle};function wle(e,t,n){const r=this;return a;function a(s){if(s===62){const u=r.containerState;return u.open||(e.enter("blockQuote",{_container:!0}),u.open=!0),e.enter("blockQuotePrefix"),e.enter("blockQuoteMarker"),e.consume(s),e.exit("blockQuoteMarker"),o}return n(s)}function o(s){return dt(s)?(e.enter("blockQuotePrefixWhitespace"),e.consume(s),e.exit("blockQuotePrefixWhitespace"),e.exit("blockQuotePrefix"),t):(e.exit("blockQuotePrefix"),t(s))}}function xle(e,t,n){const r=this;return a;function a(s){return dt(s)?ut(e,o,"linePrefix",r.parser.constructs.disable.null.includes("codeIndented")?void 0:4)(s):o(s)}function o(s){return e.attempt(HG,t,n)(s)}}function kle(e){e.exit("blockQuote")}const $G={name:"characterEscape",tokenize:Tle};function Tle(e,t,n){return r;function r(o){return e.enter("characterEscape"),e.enter("escapeMarker"),e.consume(o),e.exit("escapeMarker"),a}function a(o){return dle(o)?(e.enter("characterEscapeValue"),e.consume(o),e.exit("characterEscapeValue"),e.exit("characterEscape"),t):n(o)}}const qG={name:"characterReference",tokenize:Ale};function Ale(e,t,n){const r=this;let a=0,o,s;return u;function u(g){return e.enter("characterReference"),e.enter("characterReferenceMarker"),e.consume(g),e.exit("characterReferenceMarker"),c}function c(g){return g===35?(e.enter("characterReferenceMarkerNumeric"),e.consume(g),e.exit("characterReferenceMarkerNumeric"),d):(e.enter("characterReferenceValue"),o=31,s=kn,p(g))}function d(g){return g===88||g===120?(e.enter("characterReferenceMarkerHexadecimal"),e.consume(g),e.exit("characterReferenceMarkerHexadecimal"),e.enter("characterReferenceValue"),o=6,s=cle,p):(e.enter("characterReferenceValue"),o=7,s=d0,p(g))}function p(g){if(g===59&&a){const m=e.exit("characterReferenceValue");return s===kn&&!KT(r.sliceSerialize(m))?n(g):(e.enter("characterReferenceMarker"),e.consume(g),e.exit("characterReferenceMarker"),e.exit("characterReference"),t)}return s(g)&&a++<o?(e.consume(g),p):n(g)}}const gD={partial:!0,tokenize:_le},hD={concrete:!0,name:"codeFenced",tokenize:Rle};function Rle(e,t,n){const r=this,a={partial:!0,tokenize:N};let o=0,s=0,u;return c;function c(_){return d(_)}function d(_){const O=r.events[r.events.length-1];return o=O&&O[1].type==="linePrefix"?O[2].sliceSerialize(O[1],!0).length:0,u=_,e.enter("codeFenced"),e.enter("codeFencedFence"),e.enter("codeFencedFenceSequence"),p(_)}function p(_){return _===u?(s++,e.consume(_),p):s<3?n(_):(e.exit("codeFencedFenceSequence"),dt(_)?ut(e,g,"whitespace")(_):g(_))}function g(_){return _===null||Ve(_)?(e.exit("codeFencedFence"),r.interrupt?t(_):e.check(gD,S,C)(_)):(e.enter("codeFencedFenceInfo"),e.enter("chunkString",{contentType:"string"}),m(_))}function m(_){return _===null||Ve(_)?(e.exit("chunkString"),e.exit("codeFencedFenceInfo"),g(_)):dt(_)?(e.exit("chunkString"),e.exit("codeFencedFenceInfo"),ut(e,b,"whitespace")(_)):_===96&&_===u?n(_):(e.consume(_),m)}function b(_){return _===null||Ve(_)?g(_):(e.enter("codeFencedFenceMeta"),e.enter("chunkString",{contentType:"string"}),y(_))}function y(_){return _===null||Ve(_)?(e.exit("chunkString"),e.exit("codeFencedFenceMeta"),g(_)):_===96&&_===u?n(_):(e.consume(_),y)}function S(_){return e.attempt(a,C,x)(_)}function x(_){return e.enter("lineEnding"),e.consume(_),e.exit("lineEnding"),R}function R(_){return o>0&&dt(_)?ut(e,k,"linePrefix",o+1)(_):k(_)}function k(_){return _===null||Ve(_)?e.check(gD,S,C)(_):(e.enter("codeFlowValue"),A(_))}function A(_){return _===null||Ve(_)?(e.exit("codeFlowValue"),k(_)):(e.consume(_),A)}function C(_){return e.exit("codeFenced"),t(_)}function N(_,O,F){let L=0;return I;function I(Z){return _.enter("lineEnding"),_.consume(Z),_.exit("lineEnding"),H}function H(Z){return _.enter("codeFencedFence"),dt(Z)?ut(_,$,"linePrefix",r.parser.constructs.disable.null.includes("codeIndented")?void 0:4)(Z):$(Z)}function $(Z){return Z===u?(_.enter("codeFencedFenceSequence"),U(Z)):F(Z)}function U(Z){return Z===u?(L++,_.consume(Z),U):L>=s?(_.exit("codeFencedFenceSequence"),dt(Z)?ut(_,Y,"whitespace")(Z):Y(Z)):F(Z)}function Y(Z){return Z===null||Ve(Z)?(_.exit("codeFencedFence"),O(Z)):F(Z)}}}function _le(e,t,n){const r=this;return a;function a(s){return s===null?n(s):(e.enter("lineEnding"),e.consume(s),e.exit("lineEnding"),o)}function o(s){return r.parser.lazy[r.now().line]?n(s):t(s)}}const pb={name:"codeIndented",tokenize:Nle},Cle={partial:!0,tokenize:Ole};function Nle(e,t,n){const r=this;return a;function a(d){return e.enter("codeIndented"),ut(e,o,"linePrefix",5)(d)}function o(d){const p=r.events[r.events.length-1];return p&&p[1].type==="linePrefix"&&p[2].sliceSerialize(p[1],!0).length>=4?s(d):n(d)}function s(d){return d===null?c(d):Ve(d)?e.attempt(Cle,s,c)(d):(e.enter("codeFlowValue"),u(d))}function u(d){return d===null||Ve(d)?(e.exit("codeFlowValue"),s(d)):(e.consume(d),u)}function c(d){return e.exit("codeIndented"),t(d)}}function Ole(e,t,n){const r=this;return a;function a(s){return r.parser.lazy[r.now().line]?n(s):Ve(s)?(e.enter("lineEnding"),e.consume(s),e.exit("lineEnding"),a):ut(e,o,"linePrefix",5)(s)}function o(s){const u=r.events[r.events.length-1];return u&&u[1].type==="linePrefix"&&u[2].sliceSerialize(u[1],!0).length>=4?t(s):Ve(s)?a(s):n(s)}}const Dle={name:"codeText",previous:Lle,resolve:Ile,tokenize:Mle};function Ile(e){let t=e.length-4,n=3,r,a;if((e[n][1].type==="lineEnding"||e[n][1].type==="space")&&(e[t][1].type==="lineEnding"||e[t][1].type==="space")){for(r=n;++r<t;)if(e[r][1].type==="codeTextData"){e[n][1].type="codeTextPadding",e[t][1].type="codeTextPadding",n+=2,t-=2;break}}for(r=n-1,t++;++r<=t;)a===void 0?r!==t&&e[r][1].type!=="lineEnding"&&(a=r):(r===t||e[r][1].type==="lineEnding")&&(e[a][1].type="codeTextData",r!==a+2&&(e[a][1].end=e[r-1][1].end,e.splice(a+2,r-a-2),t-=r-a-2,r=a+2),a=void 0);return e}function Lle(e){return e!==96||this.events[this.events.length-1][1].type==="characterEscape"}function Mle(e,t,n){let r=0,a,o;return s;function s(g){return e.enter("codeText"),e.enter("codeTextSequence"),u(g)}function u(g){return g===96?(e.consume(g),r++,u):(e.exit("codeTextSequence"),c(g))}function c(g){return g===null?n(g):g===32?(e.enter("space"),e.consume(g),e.exit("space"),c):g===96?(o=e.enter("codeTextSequence"),a=0,p(g)):Ve(g)?(e.enter("lineEnding"),e.consume(g),e.exit("lineEnding"),c):(e.enter("codeTextData"),d(g))}function d(g){return g===null||g===32||g===96||Ve(g)?(e.exit("codeTextData"),c(g)):(e.consume(g),d)}function p(g){return g===96?(e.consume(g),a++,p):a===r?(e.exit("codeTextSequence"),e.exit("codeText"),t(g)):(o.type="codeTextData",d(g))}}class Fle{constructor(t){this.left=t?[...t]:[],this.right=[]}get(t){if(t<0||t>=this.left.length+this.right.length)throw new RangeError("Cannot access index `"+t+"` in a splice buffer of size `"+(this.left.length+this.right.length)+"`");return t<this.left.length?this.left[t]:this.right[this.right.length-t+this.left.length-1]}get length(){return this.left.length+this.right.length}shift(){return this.setCursor(0),this.right.pop()}slice(t,n){const r=n??Number.POSITIVE_INFINITY;return r<this.left.length?this.left.slice(t,r):t>this.left.length?this.right.slice(this.right.length-r+this.left.length,this.right.length-t+this.left.length).reverse():this.left.slice(t).concat(this.right.slice(this.right.length-r+this.left.length).reverse())}splice(t,n,r){const a=n||0;this.setCursor(Math.trunc(t));const o=this.right.splice(this.right.length-a,Number.POSITIVE_INFINITY);return r&&uu(this.left,r),o.reverse()}pop(){return this.setCursor(Number.POSITIVE_INFINITY),this.left.pop()}push(t){this.setCursor(Number.POSITIVE_INFINITY),this.left.push(t)}pushMany(t){this.setCursor(Number.POSITIVE_INFINITY),uu(this.left,t)}unshift(t){this.setCursor(0),this.right.push(t)}unshiftMany(t){this.setCursor(0),uu(this.right,t.reverse())}setCursor(t){if(!(t===this.left.length||t>this.left.length&&this.right.length===0||t<0&&this.left.length===0))if(t<this.left.length){const n=this.left.splice(t,Number.POSITIVE_INFINITY);uu(this.right,n.reverse())}else{const n=this.right.splice(this.left.length+this.right.length-t,Number.POSITIVE_INFINITY);uu(this.left,n.reverse())}}}function uu(e,t){let n=0;if(t.length<1e4)e.push(...t);else for(;n<t.length;)e.push(...t.slice(n,n+1e4)),n+=1e4}function VG(e){const t={};let n=-1,r,a,o,s,u,c,d;const p=new Fle(e);for(;++n<p.length;){for(;n in t;)n=t[n];if(r=p.get(n),n&&r[1].type==="chunkFlow"&&p.get(n-1)[1].type==="listItemPrefix"&&(c=r[1]._tokenizer.events,o=0,o<c.length&&c[o][1].type==="lineEndingBlank"&&(o+=2),o<c.length&&c[o][1].type==="content"))for(;++o<c.length&&c[o][1].type!=="content";)c[o][1].type==="chunkText"&&(c[o][1]._isInFirstContentOfListItem=!0,o++);if(r[0]==="enter")r[1].contentType&&(Object.assign(t,Ple(p,n)),n=t[n],d=!0);else if(r[1]._container){for(o=n,a=void 0;o--;)if(s=p.get(o),s[1].type==="lineEnding"||s[1].type==="lineEndingBlank")s[0]==="enter"&&(a&&(p.get(a)[1].type="lineEndingBlank"),s[1].type="lineEnding",a=o);else if(s[1].type!=="linePrefix")break;a&&(r[1].end={...p.get(a)[1].start},u=p.slice(a,n),u.unshift(r),p.splice(a,n-a+1,u))}}return rr(e,0,Number.POSITIVE_INFINITY,p.slice(0)),!d}function Ple(e,t){const n=e.get(t)[1],r=e.get(t)[2];let a=t-1;const o=[],s=n._tokenizer||r.parser[n.contentType](n.start),u=s.events,c=[],d={};let p,g,m=-1,b=n,y=0,S=0;const x=[S];for(;b;){for(;e.get(++a)[1]!==b;);o.push(a),b._tokenizer||(p=r.sliceStream(b),b.next||p.push(null),g&&s.defineSkip(b.start),b._isInFirstContentOfListItem&&(s._gfmTasklistFirstContentOfListItem=!0),s.write(p),b._isInFirstContentOfListItem&&(s._gfmTasklistFirstContentOfListItem=void 0)),g=b,b=b.next}for(b=n;++m<u.length;)u[m][0]==="exit"&&u[m-1][0]==="enter"&&u[m][1].type===u[m-1][1].type&&u[m][1].start.line!==u[m][1].end.line&&(S=m+1,x.push(S),b._tokenizer=void 0,b.previous=void 0,b=b.next);for(s.events=[],b?(b._tokenizer=void 0,b.previous=void 0):x.pop(),m=x.length;m--;){const R=u.slice(x[m],x[m+1]),k=o.pop();c.push([k,k+R.length-1]),e.splice(k,2,R)}for(c.reverse(),m=-1;++m<c.length;)d[y+c[m][0]]=y+c[m][1],y+=c[m][1]-c[m][0]-1;return d}const zle={resolve:Ule,tokenize:jle},Ble={partial:!0,tokenize:Gle};function Ule(e){return VG(e),e}function jle(e,t){let n;return r;function r(u){return e.enter("content"),n=e.enter("chunkContent",{contentType:"content"}),a(u)}function a(u){return u===null?o(u):Ve(u)?e.check(Ble,s,o)(u):(e.consume(u),a)}function o(u){return e.exit("chunkContent"),e.exit("content"),t(u)}function s(u){return e.consume(u),e.exit("chunkContent"),n.next=e.enter("chunkContent",{contentType:"content",previous:n}),n=n.next,a}}function Gle(e,t,n){const r=this;return a;function a(s){return e.exit("chunkContent"),e.enter("lineEnding"),e.consume(s),e.exit("lineEnding"),ut(e,o,"linePrefix")}function o(s){if(s===null||Ve(s))return n(s);const u=r.events[r.events.length-1];return!r.parser.constructs.disable.null.includes("codeIndented")&&u&&u[1].type==="linePrefix"&&u[2].sliceSerialize(u[1],!0).length>=4?t(s):e.interrupt(r.parser.constructs.flow,n,t)(s)}}function WG(e,t,n,r,a,o,s,u,c){const d=c||Number.POSITIVE_INFINITY;let p=0;return g;function g(R){return R===60?(e.enter(r),e.enter(a),e.enter(o),e.consume(R),e.exit(o),m):R===null||R===32||R===41||Ff(R)?n(R):(e.enter(r),e.enter(s),e.enter(u),e.enter("chunkString",{contentType:"string"}),S(R))}function m(R){return R===62?(e.enter(o),e.consume(R),e.exit(o),e.exit(a),e.exit(r),t):(e.enter(u),e.enter("chunkString",{contentType:"string"}),b(R))}function b(R){return R===62?(e.exit("chunkString"),e.exit(u),m(R)):R===null||R===60||Ve(R)?n(R):(e.consume(R),R===92?y:b)}function y(R){return R===60||R===62||R===92?(e.consume(R),b):b(R)}function S(R){return!p&&(R===null||R===41||Dt(R))?(e.exit("chunkString"),e.exit(u),e.exit(s),e.exit(r),t(R)):p<d&&R===40?(e.consume(R),p++,S):R===41?(e.consume(R),p--,S):R===null||R===32||R===40||Ff(R)?n(R):(e.consume(R),R===92?x:S)}function x(R){return R===40||R===41||R===92?(e.consume(R),S):S(R)}}function YG(e,t,n,r,a,o){const s=this;let u=0,c;return d;function d(b){return e.enter(r),e.enter(a),e.consume(b),e.exit(a),e.enter(o),p}function p(b){return u>999||b===null||b===91||b===93&&!c||b===94&&!u&&"_hiddenFootnoteSupport"in s.parser.constructs?n(b):b===93?(e.exit(o),e.enter(a),e.consume(b),e.exit(a),e.exit(r),t):Ve(b)?(e.enter("lineEnding"),e.consume(b),e.exit("lineEnding"),p):(e.enter("chunkString",{contentType:"string"}),g(b))}function g(b){return b===null||b===91||b===93||Ve(b)||u++>999?(e.exit("chunkString"),p(b)):(e.consume(b),c||(c=!dt(b)),b===92?m:g)}function m(b){return b===91||b===92||b===93?(e.consume(b),u++,g):g(b)}}function KG(e,t,n,r,a,o){let s;return u;function u(m){return m===34||m===39||m===40?(e.enter(r),e.enter(a),e.consume(m),e.exit(a),s=m===40?41:m,c):n(m)}function c(m){return m===s?(e.enter(a),e.consume(m),e.exit(a),e.exit(r),t):(e.enter(o),d(m))}function d(m){return m===s?(e.exit(o),c(s)):m===null?n(m):Ve(m)?(e.enter("lineEnding"),e.consume(m),e.exit("lineEnding"),ut(e,d,"linePrefix")):(e.enter("chunkString",{contentType:"string"}),p(m))}function p(m){return m===s||m===null||Ve(m)?(e.exit("chunkString"),d(m)):(e.consume(m),m===92?g:p)}function g(m){return m===s||m===92?(e.consume(m),p):p(m)}}function yu(e,t){let n;return r;function r(a){return Ve(a)?(e.enter("lineEnding"),e.consume(a),e.exit("lineEnding"),n=!0,r):dt(a)?ut(e,r,n?"linePrefix":"lineSuffix")(a):t(a)}}const Hle={name:"definition",tokenize:qle},$le={partial:!0,tokenize:Vle};function qle(e,t,n){const r=this;let a;return o;function o(b){return e.enter("definition"),s(b)}function s(b){return YG.call(r,e,u,n,"definitionLabel","definitionLabelMarker","definitionLabelString")(b)}function u(b){return a=zr(r.sliceSerialize(r.events[r.events.length-1][1]).slice(1,-1)),b===58?(e.enter("definitionMarker"),e.consume(b),e.exit("definitionMarker"),c):n(b)}function c(b){return Dt(b)?yu(e,d)(b):d(b)}function d(b){return WG(e,p,n,"definitionDestination","definitionDestinationLiteral","definitionDestinationLiteralMarker","definitionDestinationRaw","definitionDestinationString")(b)}function p(b){return e.attempt($le,g,g)(b)}function g(b){return dt(b)?ut(e,m,"whitespace")(b):m(b)}function m(b){return b===null||Ve(b)?(e.exit("definition"),r.parser.defined.push(a),t(b)):n(b)}}function Vle(e,t,n){return r;function r(u){return Dt(u)?yu(e,a)(u):n(u)}function a(u){return KG(e,o,n,"definitionTitle","definitionTitleMarker","definitionTitleString")(u)}function o(u){return dt(u)?ut(e,s,"whitespace")(u):s(u)}function s(u){return u===null||Ve(u)?t(u):n(u)}}const Wle={name:"hardBreakEscape",tokenize:Yle};function Yle(e,t,n){return r;function r(o){return e.enter("hardBreakEscape"),e.consume(o),a}function a(o){return Ve(o)?(e.exit("hardBreakEscape"),t(o)):n(o)}}const Kle={name:"headingAtx",resolve:Xle,tokenize:Zle};function Xle(e,t){let n=e.length-2,r=3,a,o;return e[r][1].type==="whitespace"&&(r+=2),n-2>r&&e[n][1].type==="whitespace"&&(n-=2),e[n][1].type==="atxHeadingSequence"&&(r===n-1||n-4>r&&e[n-2][1].type==="whitespace")&&(n-=r+1===n?2:4),n>r&&(a={type:"atxHeadingText",start:e[r][1].start,end:e[n][1].end},o={type:"chunkText",start:e[r][1].start,end:e[n][1].end,contentType:"text"},rr(e,r,n-r+1,[["enter",a,t],["enter",o,t],["exit",o,t],["exit",a,t]])),e}function Zle(e,t,n){let r=0;return a;function a(p){return e.enter("atxHeading"),o(p)}function o(p){return e.enter("atxHeadingSequence"),s(p)}function s(p){return p===35&&r++<6?(e.consume(p),s):p===null||Dt(p)?(e.exit("atxHeadingSequence"),u(p)):n(p)}function u(p){return p===35?(e.enter("atxHeadingSequence"),c(p)):p===null||Ve(p)?(e.exit("atxHeading"),t(p)):dt(p)?ut(e,u,"whitespace")(p):(e.enter("atxHeadingText"),d(p))}function c(p){return p===35?(e.consume(p),c):(e.exit("atxHeadingSequence"),u(p))}function d(p){return p===null||p===35||Dt(p)?(e.exit("atxHeadingText"),u(p)):(e.consume(p),d)}}const Qle=["address","article","aside","base","basefont","blockquote","body","caption","center","col","colgroup","dd","details","dialog","dir","div","dl","dt","fieldset","figcaption","figure","footer","form","frame","frameset","h1","h2","h3","h4","h5","h6","head","header","hr","html","iframe","legend","li","link","main","menu","menuitem","nav","noframes","ol","optgroup","option","p","param","search","section","summary","table","tbody","td","tfoot","th","thead","title","tr","track","ul"],mD=["pre","script","style","textarea"],Jle={concrete:!0,name:"htmlFlow",resolveTo:nue,tokenize:rue},eue={partial:!0,tokenize:oue},tue={partial:!0,tokenize:aue};function nue(e){let t=e.length;for(;t--&&!(e[t][0]==="enter"&&e[t][1].type==="htmlFlow"););return t>1&&e[t-2][1].type==="linePrefix"&&(e[t][1].start=e[t-2][1].start,e[t+1][1].start=e[t-2][1].start,e.splice(t-2,2)),e}function rue(e,t,n){const r=this;let a,o,s,u,c;return d;function d(M){return p(M)}function p(M){return e.enter("htmlFlow"),e.enter("htmlFlowData"),e.consume(M),g}function g(M){return M===33?(e.consume(M),m):M===47?(e.consume(M),o=!0,S):M===63?(e.consume(M),a=3,r.interrupt?t:D):On(M)?(e.consume(M),s=String.fromCharCode(M),x):n(M)}function m(M){return M===45?(e.consume(M),a=2,b):M===91?(e.consume(M),a=5,u=0,y):On(M)?(e.consume(M),a=4,r.interrupt?t:D):n(M)}function b(M){return M===45?(e.consume(M),r.interrupt?t:D):n(M)}function y(M){const W="CDATA[";return M===W.charCodeAt(u++)?(e.consume(M),u===W.length?r.interrupt?t:$:y):n(M)}function S(M){return On(M)?(e.consume(M),s=String.fromCharCode(M),x):n(M)}function x(M){if(M===null||M===47||M===62||Dt(M)){const W=M===47,Q=s.toLowerCase();return!W&&!o&&mD.includes(Q)?(a=1,r.interrupt?t(M):$(M)):Qle.includes(s.toLowerCase())?(a=6,W?(e.consume(M),R):r.interrupt?t(M):$(M)):(a=7,r.interrupt&&!r.parser.lazy[r.now().line]?n(M):o?k(M):A(M))}return M===45||kn(M)?(e.consume(M),s+=String.fromCharCode(M),x):n(M)}function R(M){return M===62?(e.consume(M),r.interrupt?t:$):n(M)}function k(M){return dt(M)?(e.consume(M),k):I(M)}function A(M){return M===47?(e.consume(M),I):M===58||M===95||On(M)?(e.consume(M),C):dt(M)?(e.consume(M),A):I(M)}function C(M){return M===45||M===46||M===58||M===95||kn(M)?(e.consume(M),C):N(M)}function N(M){return M===61?(e.consume(M),_):dt(M)?(e.consume(M),N):A(M)}function _(M){return M===null||M===60||M===61||M===62||M===96?n(M):M===34||M===39?(e.consume(M),c=M,O):dt(M)?(e.consume(M),_):F(M)}function O(M){return M===c?(e.consume(M),c=null,L):M===null||Ve(M)?n(M):(e.consume(M),O)}function F(M){return M===null||M===34||M===39||M===47||M===60||M===61||M===62||M===96||Dt(M)?N(M):(e.consume(M),F)}function L(M){return M===47||M===62||dt(M)?A(M):n(M)}function I(M){return M===62?(e.consume(M),H):n(M)}function H(M){return M===null||Ve(M)?$(M):dt(M)?(e.consume(M),H):n(M)}function $(M){return M===45&&a===2?(e.consume(M),j):M===60&&a===1?(e.consume(M),G):M===62&&a===4?(e.consume(M),V):M===63&&a===3?(e.consume(M),D):M===93&&a===5?(e.consume(M),K):Ve(M)&&(a===6||a===7)?(e.exit("htmlFlowData"),e.check(eue,B,U)(M)):M===null||Ve(M)?(e.exit("htmlFlowData"),U(M)):(e.consume(M),$)}function U(M){return e.check(tue,Y,B)(M)}function Y(M){return e.enter("lineEnding"),e.consume(M),e.exit("lineEnding"),Z}function Z(M){return M===null||Ve(M)?U(M):(e.enter("htmlFlowData"),$(M))}function j(M){return M===45?(e.consume(M),D):$(M)}function G(M){return M===47?(e.consume(M),s="",z):$(M)}function z(M){if(M===62){const W=s.toLowerCase();return mD.includes(W)?(e.consume(M),V):$(M)}return On(M)&&s.length<8?(e.consume(M),s+=String.fromCharCode(M),z):$(M)}function K(M){return M===93?(e.consume(M),D):$(M)}function D(M){return M===62?(e.consume(M),V):M===45&&a===2?(e.consume(M),D):$(M)}function V(M){return M===null||Ve(M)?(e.exit("htmlFlowData"),B(M)):(e.consume(M),V)}function B(M){return e.exit("htmlFlow"),t(M)}}function aue(e,t,n){const r=this;return a;function a(s){return Ve(s)?(e.enter("lineEnding"),e.consume(s),e.exit("lineEnding"),o):n(s)}function o(s){return r.parser.lazy[r.now().line]?n(s):t(s)}}function oue(e,t,n){return r;function r(a){return e.enter("lineEnding"),e.consume(a),e.exit("lineEnding"),e.attempt(ac,t,n)}}const iue={name:"htmlText",tokenize:sue};function sue(e,t,n){const r=this;let a,o,s;return u;function u(D){return e.enter("htmlText"),e.enter("htmlTextData"),e.consume(D),c}function c(D){return D===33?(e.consume(D),d):D===47?(e.consume(D),N):D===63?(e.consume(D),A):On(D)?(e.consume(D),F):n(D)}function d(D){return D===45?(e.consume(D),p):D===91?(e.consume(D),o=0,y):On(D)?(e.consume(D),k):n(D)}function p(D){return D===45?(e.consume(D),b):n(D)}function g(D){return D===null?n(D):D===45?(e.consume(D),m):Ve(D)?(s=g,G(D)):(e.consume(D),g)}function m(D){return D===45?(e.consume(D),b):g(D)}function b(D){return D===62?j(D):D===45?m(D):g(D)}function y(D){const V="CDATA[";return D===V.charCodeAt(o++)?(e.consume(D),o===V.length?S:y):n(D)}function S(D){return D===null?n(D):D===93?(e.consume(D),x):Ve(D)?(s=S,G(D)):(e.consume(D),S)}function x(D){return D===93?(e.consume(D),R):S(D)}function R(D){return D===62?j(D):D===93?(e.consume(D),R):S(D)}function k(D){return D===null||D===62?j(D):Ve(D)?(s=k,G(D)):(e.consume(D),k)}function A(D){return D===null?n(D):D===63?(e.consume(D),C):Ve(D)?(s=A,G(D)):(e.consume(D),A)}function C(D){return D===62?j(D):A(D)}function N(D){return On(D)?(e.consume(D),_):n(D)}function _(D){return D===45||kn(D)?(e.consume(D),_):O(D)}function O(D){return Ve(D)?(s=O,G(D)):dt(D)?(e.consume(D),O):j(D)}function F(D){return D===45||kn(D)?(e.consume(D),F):D===47||D===62||Dt(D)?L(D):n(D)}function L(D){return D===47?(e.consume(D),j):D===58||D===95||On(D)?(e.consume(D),I):Ve(D)?(s=L,G(D)):dt(D)?(e.consume(D),L):j(D)}function I(D){return D===45||D===46||D===58||D===95||kn(D)?(e.consume(D),I):H(D)}function H(D){return D===61?(e.consume(D),$):Ve(D)?(s=H,G(D)):dt(D)?(e.consume(D),H):L(D)}function $(D){return D===null||D===60||D===61||D===62||D===96?n(D):D===34||D===39?(e.consume(D),a=D,U):Ve(D)?(s=$,G(D)):dt(D)?(e.consume(D),$):(e.consume(D),Y)}function U(D){return D===a?(e.consume(D),a=void 0,Z):D===null?n(D):Ve(D)?(s=U,G(D)):(e.consume(D),U)}function Y(D){return D===null||D===34||D===39||D===60||D===61||D===96?n(D):D===47||D===62||Dt(D)?L(D):(e.consume(D),Y)}function Z(D){return D===47||D===62||Dt(D)?L(D):n(D)}function j(D){return D===62?(e.consume(D),e.exit("htmlTextData"),e.exit("htmlText"),t):n(D)}function G(D){return e.exit("htmlTextData"),e.enter("lineEnding"),e.consume(D),e.exit("lineEnding"),z}function z(D){return dt(D)?ut(e,K,"linePrefix",r.parser.constructs.disable.null.includes("codeIndented")?void 0:4)(D):K(D)}function K(D){return e.enter("htmlTextData"),s(D)}}const XT={name:"labelEnd",resolveAll:due,resolveTo:fue,tokenize:pue},lue={tokenize:gue},uue={tokenize:hue},cue={tokenize:mue};function due(e){let t=-1;const n=[];for(;++t<e.length;){const r=e[t][1];if(n.push(e[t]),r.type==="labelImage"||r.type==="labelLink"||r.type==="labelEnd"){const a=r.type==="labelImage"?4:2;r.type="data",t+=a}}return e.length!==n.length&&rr(e,0,e.length,n),e}function fue(e,t){let n=e.length,r=0,a,o,s,u;for(;n--;)if(a=e[n][1],o){if(a.type==="link"||a.type==="labelLink"&&a._inactive)break;e[n][0]==="enter"&&a.type==="labelLink"&&(a._inactive=!0)}else if(s){if(e[n][0]==="enter"&&(a.type==="labelImage"||a.type==="labelLink")&&!a._balanced&&(o=n,a.type!=="labelLink")){r=2;break}}else a.type==="labelEnd"&&(s=n);const c={type:e[o][1].type==="labelLink"?"link":"image",start:{...e[o][1].start},end:{...e[e.length-1][1].end}},d={type:"label",start:{...e[o][1].start},end:{...e[s][1].end}},p={type:"labelText",start:{...e[o+r+2][1].end},end:{...e[s-2][1].start}};return u=[["enter",c,t],["enter",d,t]],u=vr(u,e.slice(o+1,o+r+3)),u=vr(u,[["enter",p,t]]),u=vr(u,Op(t.parser.constructs.insideSpan.null,e.slice(o+r+4,s-3),t)),u=vr(u,[["exit",p,t],e[s-2],e[s-1],["exit",d,t]]),u=vr(u,e.slice(s+1)),u=vr(u,[["exit",c,t]]),rr(e,o,e.length,u),e}function pue(e,t,n){const r=this;let a=r.events.length,o,s;for(;a--;)if((r.events[a][1].type==="labelImage"||r.events[a][1].type==="labelLink")&&!r.events[a][1]._balanced){o=r.events[a][1];break}return u;function u(m){return o?o._inactive?g(m):(s=r.parser.defined.includes(zr(r.sliceSerialize({start:o.end,end:r.now()}))),e.enter("labelEnd"),e.enter("labelMarker"),e.consume(m),e.exit("labelMarker"),e.exit("labelEnd"),c):n(m)}function c(m){return m===40?e.attempt(lue,p,s?p:g)(m):m===91?e.attempt(uue,p,s?d:g)(m):s?p(m):g(m)}function d(m){return e.attempt(cue,p,g)(m)}function p(m){return t(m)}function g(m){return o._balanced=!0,n(m)}}function gue(e,t,n){return r;function r(g){return e.enter("resource"),e.enter("resourceMarker"),e.consume(g),e.exit("resourceMarker"),a}function a(g){return Dt(g)?yu(e,o)(g):o(g)}function o(g){return g===41?p(g):WG(e,s,u,"resourceDestination","resourceDestinationLiteral","resourceDestinationLiteralMarker","resourceDestinationRaw","resourceDestinationString",32)(g)}function s(g){return Dt(g)?yu(e,c)(g):p(g)}function u(g){return n(g)}function c(g){return g===34||g===39||g===40?KG(e,d,n,"resourceTitle","resourceTitleMarker","resourceTitleString")(g):p(g)}function d(g){return Dt(g)?yu(e,p)(g):p(g)}function p(g){return g===41?(e.enter("resourceMarker"),e.consume(g),e.exit("resourceMarker"),e.exit("resource"),t):n(g)}}function hue(e,t,n){const r=this;return a;function a(u){return YG.call(r,e,o,s,"reference","referenceMarker","referenceString")(u)}function o(u){return r.parser.defined.includes(zr(r.sliceSerialize(r.events[r.events.length-1][1]).slice(1,-1)))?t(u):n(u)}function s(u){return n(u)}}function mue(e,t,n){return r;function r(o){return e.enter("reference"),e.enter("referenceMarker"),e.consume(o),e.exit("referenceMarker"),a}function a(o){return o===93?(e.enter("referenceMarker"),e.consume(o),e.exit("referenceMarker"),e.exit("reference"),t):n(o)}}const bue={name:"labelStartImage",resolveAll:XT.resolveAll,tokenize:yue};function yue(e,t,n){const r=this;return a;function a(u){return e.enter("labelImage"),e.enter("labelImageMarker"),e.consume(u),e.exit("labelImageMarker"),o}function o(u){return u===91?(e.enter("labelMarker"),e.consume(u),e.exit("labelMarker"),e.exit("labelImage"),s):n(u)}function s(u){return u===94&&"_hiddenFootnoteSupport"in r.parser.constructs?n(u):t(u)}}const vue={name:"labelStartLink",resolveAll:XT.resolveAll,tokenize:Sue};function Sue(e,t,n){const r=this;return a;function a(s){return e.enter("labelLink"),e.enter("labelMarker"),e.consume(s),e.exit("labelMarker"),e.exit("labelLink"),o}function o(s){return s===94&&"_hiddenFootnoteSupport"in r.parser.constructs?n(s):t(s)}}const gb={name:"lineEnding",tokenize:Eue};function Eue(e,t){return n;function n(r){return e.enter("lineEnding"),e.consume(r),e.exit("lineEnding"),ut(e,t,"linePrefix")}}const uf={name:"thematicBreak",tokenize:wue};function wue(e,t,n){let r=0,a;return o;function o(d){return e.enter("thematicBreak"),s(d)}function s(d){return a=d,u(d)}function u(d){return d===a?(e.enter("thematicBreakSequence"),c(d)):r>=3&&(d===null||Ve(d))?(e.exit("thematicBreak"),t(d)):n(d)}function c(d){return d===a?(e.consume(d),r++,c):(e.exit("thematicBreakSequence"),dt(d)?ut(e,u,"whitespace")(d):u(d))}}const zn={continuation:{tokenize:Aue},exit:_ue,name:"list",tokenize:Tue},xue={partial:!0,tokenize:Cue},kue={partial:!0,tokenize:Rue};function Tue(e,t,n){const r=this,a=r.events[r.events.length-1];let o=a&&a[1].type==="linePrefix"?a[2].sliceSerialize(a[1],!0).length:0,s=0;return u;function u(b){const y=r.containerState.type||(b===42||b===43||b===45?"listUnordered":"listOrdered");if(y==="listUnordered"?!r.containerState.marker||b===r.containerState.marker:d0(b)){if(r.containerState.type||(r.containerState.type=y,e.enter(y,{_container:!0})),y==="listUnordered")return e.enter("listItemPrefix"),b===42||b===45?e.check(uf,n,d)(b):d(b);if(!r.interrupt||b===49)return e.enter("listItemPrefix"),e.enter("listItemValue"),c(b)}return n(b)}function c(b){return d0(b)&&++s<10?(e.consume(b),c):(!r.interrupt||s<2)&&(r.containerState.marker?b===r.containerState.marker:b===41||b===46)?(e.exit("listItemValue"),d(b)):n(b)}function d(b){return e.enter("listItemMarker"),e.consume(b),e.exit("listItemMarker"),r.containerState.marker=r.containerState.marker||b,e.check(ac,r.interrupt?n:p,e.attempt(xue,m,g))}function p(b){return r.containerState.initialBlankLine=!0,o++,m(b)}function g(b){return dt(b)?(e.enter("listItemPrefixWhitespace"),e.consume(b),e.exit("listItemPrefixWhitespace"),m):n(b)}function m(b){return r.containerState.size=o+r.sliceSerialize(e.exit("listItemPrefix"),!0).length,t(b)}}function Aue(e,t,n){const r=this;return r.containerState._closeFlow=void 0,e.check(ac,a,o);function a(u){return r.containerState.furtherBlankLines=r.containerState.furtherBlankLines||r.containerState.initialBlankLine,ut(e,t,"listItemIndent",r.containerState.size+1)(u)}function o(u){return r.containerState.furtherBlankLines||!dt(u)?(r.containerState.furtherBlankLines=void 0,r.containerState.initialBlankLine=void 0,s(u)):(r.containerState.furtherBlankLines=void 0,r.containerState.initialBlankLine=void 0,e.attempt(kue,t,s)(u))}function s(u){return r.containerState._closeFlow=!0,r.interrupt=void 0,ut(e,e.attempt(zn,t,n),"linePrefix",r.parser.constructs.disable.null.includes("codeIndented")?void 0:4)(u)}}function Rue(e,t,n){const r=this;return ut(e,a,"listItemIndent",r.containerState.size+1);function a(o){const s=r.events[r.events.length-1];return s&&s[1].type==="listItemIndent"&&s[2].sliceSerialize(s[1],!0).length===r.containerState.size?t(o):n(o)}}function _ue(e){e.exit(this.containerState.type)}function Cue(e,t,n){const r=this;return ut(e,a,"listItemPrefixWhitespace",r.parser.constructs.disable.null.includes("codeIndented")?void 0:5);function a(o){const s=r.events[r.events.length-1];return!dt(o)&&s&&s[1].type==="listItemPrefixWhitespace"?t(o):n(o)}}const bD={name:"setextUnderline",resolveTo:Nue,tokenize:Oue};function Nue(e,t){let n=e.length,r,a,o;for(;n--;)if(e[n][0]==="enter"){if(e[n][1].type==="content"){r=n;break}e[n][1].type==="paragraph"&&(a=n)}else e[n][1].type==="content"&&e.splice(n,1),!o&&e[n][1].type==="definition"&&(o=n);const s={type:"setextHeading",start:{...e[a][1].start},end:{...e[e.length-1][1].end}};return e[a][1].type="setextHeadingText",o?(e.splice(a,0,["enter",s,t]),e.splice(o+1,0,["exit",e[r][1],t]),e[r][1].end={...e[o][1].end}):e[r][1]=s,e.push(["exit",s,t]),e}function Oue(e,t,n){const r=this;let a;return o;function o(d){let p=r.events.length,g;for(;p--;)if(r.events[p][1].type!=="lineEnding"&&r.events[p][1].type!=="linePrefix"&&r.events[p][1].type!=="content"){g=r.events[p][1].type==="paragraph";break}return!r.parser.lazy[r.now().line]&&(r.interrupt||g)?(e.enter("setextHeadingLine"),a=d,s(d)):n(d)}function s(d){return e.enter("setextHeadingLineSequence"),u(d)}function u(d){return d===a?(e.consume(d),u):(e.exit("setextHeadingLineSequence"),dt(d)?ut(e,c,"lineSuffix")(d):c(d))}function c(d){return d===null||Ve(d)?(e.exit("setextHeadingLine"),t(d)):n(d)}}const Due={tokenize:Iue};function Iue(e){const t=this,n=e.attempt(ac,r,e.attempt(this.parser.constructs.flowInitial,a,ut(e,e.attempt(this.parser.constructs.flow,a,e.attempt(zle,a)),"linePrefix")));return n;function r(o){if(o===null){e.consume(o);return}return e.enter("lineEndingBlank"),e.consume(o),e.exit("lineEndingBlank"),t.currentConstruct=void 0,n}function a(o){if(o===null){e.consume(o);return}return e.enter("lineEnding"),e.consume(o),e.exit("lineEnding"),t.currentConstruct=void 0,n}}const Lue={resolveAll:ZG()},Mue=XG("string"),Fue=XG("text");function XG(e){return{resolveAll:ZG(e==="text"?Pue:void 0),tokenize:t};function t(n){const r=this,a=this.parser.constructs[e],o=n.attempt(a,s,u);return s;function s(p){return d(p)?o(p):u(p)}function u(p){if(p===null){n.consume(p);return}return n.enter("data"),n.consume(p),c}function c(p){return d(p)?(n.exit("data"),o(p)):(n.consume(p),c)}function d(p){if(p===null)return!0;const g=a[p];let m=-1;if(g)for(;++m<g.length;){const b=g[m];if(!b.previous||b.previous.call(r,r.previous))return!0}return!1}}}function ZG(e){return t;function t(n,r){let a=-1,o;for(;++a<=n.length;)o===void 0?n[a]&&n[a][1].type==="data"&&(o=a,a++):(!n[a]||n[a][1].type!=="data")&&(a!==o+2&&(n[o][1].end=n[a-1][1].end,n.splice(o+2,a-o-2),a=o+2),o=void 0);return e?e(n,r):n}}function Pue(e,t){let n=0;for(;++n<=e.length;)if((n===e.length||e[n][1].type==="lineEnding")&&e[n-1][1].type==="data"){const r=e[n-1][1],a=t.sliceStream(r);let o=a.length,s=-1,u=0,c;for(;o--;){const d=a[o];if(typeof d=="string"){for(s=d.length;d.charCodeAt(s-1)===32;)u++,s--;if(s)break;s=-1}else if(d===-2)c=!0,u++;else if(d!==-1){o++;break}}if(u){const d={type:n===e.length||c||u<2?"lineSuffix":"hardBreakTrailing",start:{_bufferIndex:o?s:r.start._bufferIndex+s,_index:r.start._index+o,line:r.end.line,column:r.end.column-u,offset:r.end.offset-u},end:{...r.end}};r.end={...d.start},r.start.offset===r.end.offset?Object.assign(r,d):(e.splice(n,0,["enter",d,t],["exit",d,t]),n+=2)}n++}return e}const zue={42:zn,43:zn,45:zn,48:zn,49:zn,50:zn,51:zn,52:zn,53:zn,54:zn,55:zn,56:zn,57:zn,62:HG},Bue={91:Hle},Uue={[-2]:pb,[-1]:pb,32:pb},jue={35:Kle,42:uf,45:[bD,uf],60:Jle,61:bD,95:uf,96:hD,126:hD},Gue={38:qG,92:$G},Hue={[-5]:gb,[-4]:gb,[-3]:gb,33:bue,38:qG,42:f0,60:[vle,iue],91:vue,92:[Wle,$G],93:XT,95:f0,96:Dle},$ue={null:[f0,Lue]},que={null:[42,95]},Vue={null:[]},Wue=Object.freeze(Object.defineProperty({__proto__:null,attentionMarkers:que,contentInitial:Bue,disable:Vue,document:zue,flow:jue,flowInitial:Uue,insideSpan:$ue,string:Gue,text:Hue},Symbol.toStringTag,{value:"Module"}));function Yue(e,t,n){let r={_bufferIndex:-1,_index:0,line:n&&n.line||1,column:n&&n.column||1,offset:n&&n.offset||0};const a={},o=[];let s=[],u=[];const c={attempt:O(N),check:O(_),consume:k,enter:A,exit:C,interrupt:O(_,{interrupt:!0})},d={code:null,containerState:{},defineSkip:S,events:[],now:y,parser:e,previous:null,sliceSerialize:m,sliceStream:b,write:g};let p=t.tokenize.call(d,c);return t.resolveAll&&o.push(t),d;function g(H){return s=vr(s,H),x(),s[s.length-1]!==null?[]:(F(t,0),d.events=Op(o,d.events,d),d.events)}function m(H,$){return Xue(b(H),$)}function b(H){return Kue(s,H)}function y(){const{_bufferIndex:H,_index:$,line:U,column:Y,offset:Z}=r;return{_bufferIndex:H,_index:$,line:U,column:Y,offset:Z}}function S(H){a[H.line]=H.column,I()}function x(){let H;for(;r._index<s.length;){const $=s[r._index];if(typeof $=="string")for(H=r._index,r._bufferIndex<0&&(r._bufferIndex=0);r._index===H&&r._bufferIndex<$.length;)R($.charCodeAt(r._bufferIndex));else R($)}}function R(H){p=p(H)}function k(H){Ve(H)?(r.line++,r.column=1,r.offset+=H===-3?2:1,I()):H!==-1&&(r.column++,r.offset++),r._bufferIndex<0?r._index++:(r._bufferIndex++,r._bufferIndex===s[r._index].length&&(r._bufferIndex=-1,r._index++)),d.previous=H}function A(H,$){const U=$||{};return U.type=H,U.start=y(),d.events.push(["enter",U,d]),u.push(U),U}function C(H){const $=u.pop();return $.end=y(),d.events.push(["exit",$,d]),$}function N(H,$){F(H,$.from)}function _(H,$){$.restore()}function O(H,$){return U;function U(Y,Z,j){let G,z,K,D;return Array.isArray(Y)?B(Y):"tokenize"in Y?B([Y]):V(Y);function V(oe){return re;function re(ie){const Se=ie!==null&&oe[ie],ae=ie!==null&&oe.null,ve=[...Array.isArray(Se)?Se:Se?[Se]:[],...Array.isArray(ae)?ae:ae?[ae]:[]];return B(ve)(ie)}}function B(oe){return G=oe,z=0,oe.length===0?j:M(oe[z])}function M(oe){return re;function re(ie){return D=L(),K=oe,oe.partial||(d.currentConstruct=oe),oe.name&&d.parser.constructs.disable.null.includes(oe.name)?Q():oe.tokenize.call($?Object.assign(Object.create(d),$):d,c,W,Q)(ie)}}function W(oe){return H(K,D),Z}function Q(oe){return D.restore(),++z<G.length?M(G[z]):j}}}function F(H,$){H.resolveAll&&!o.includes(H)&&o.push(H),H.resolve&&rr(d.events,$,d.events.length-$,H.resolve(d.events.slice($),d)),H.resolveTo&&(d.events=H.resolveTo(d.events,d))}function L(){const H=y(),$=d.previous,U=d.currentConstruct,Y=d.events.length,Z=Array.from(u);return{from:Y,restore:j};function j(){r=H,d.previous=$,d.currentConstruct=U,d.events.length=Y,u=Z,I()}}function I(){r.line in a&&r.column<2&&(r.column=a[r.line],r.offset+=a[r.line]-1)}}function Kue(e,t){const n=t.start._index,r=t.start._bufferIndex,a=t.end._index,o=t.end._bufferIndex;let s;if(n===a)s=[e[n].slice(r,o)];else{if(s=e.slice(n,a),r>-1){const u=s[0];typeof u=="string"?s[0]=u.slice(r):s.shift()}o>0&&s.push(e[a].slice(0,o))}return s}function Xue(e,t){let n=-1;const r=[];let a;for(;++n<e.length;){const o=e[n];let s;if(typeof o=="string")s=o;else switch(o){case-5:{s="\r";break}case-4:{s=`
`;break}case-3:{s=`\r
`;break}case-2:{s=t?" ":"	";break}case-1:{if(!t&&a)continue;s=" ";break}default:s=String.fromCharCode(o)}a=o===-2,r.push(s)}return r.join("")}function Zue(e){const r={constructs:jG([Wue,...(e||{}).extensions||[]]),content:a(fle),defined:[],document:a(gle),flow:a(Due),lazy:{},string:a(Mue),text:a(Fue)};return r;function a(o){return s;function s(u){return Yue(r,o,u)}}}function Que(e){for(;!VG(e););return e}const yD=/[\0\t\n\r]/g;function Jue(){let e=1,t="",n=!0,r;return a;function a(o,s,u){const c=[];let d,p,g,m,b;for(o=t+(typeof o=="string"?o.toString():new TextDecoder(s||void 0).decode(o)),g=0,t="",n&&(o.charCodeAt(0)===65279&&g++,n=void 0);g<o.length;){if(yD.lastIndex=g,d=yD.exec(o),m=d&&d.index!==void 0?d.index:o.length,b=o.charCodeAt(m),!d){t=o.slice(g);break}if(b===10&&g===m&&r)c.push(-3),r=void 0;else switch(r&&(c.push(-5),r=void 0),g<m&&(c.push(o.slice(g,m)),e+=m-g),b){case 0:{c.push(65533),e++;break}case 9:{for(p=Math.ceil(e/4)*4,c.push(-2);e++<p;)c.push(-1);break}case 10:{c.push(-4),e=1;break}default:r=!0,e=1}g=m+1}return u&&(r&&c.push(-5),t&&c.push(t),c.push(null)),c}}const ece=/\\([!-/:-@[-`{-~])|&(#(?:\d{1,7}|x[\da-f]{1,6})|[\da-z]{1,31});/gi;function tce(e){return e.replace(ece,nce)}function nce(e,t,n){if(t)return t;if(n.charCodeAt(0)===35){const a=n.charCodeAt(1),o=a===120||a===88;return GG(n.slice(o?2:1),o?16:10)}return KT(n)||e}const QG={}.hasOwnProperty;function rce(e,t,n){return typeof t!="string"&&(n=t,t=void 0),ace(n)(Que(Zue(n).document().write(Jue()(e,t,!0))))}function ace(e){const t={transforms:[],canContainEols:["emphasis","fragment","heading","paragraph","strong"],enter:{autolink:o(me),autolinkProtocol:L,autolinkEmail:L,atxHeading:o(ke),blockQuote:o(ae),characterEscape:L,characterReference:L,codeFenced:o(ve),codeFencedFenceInfo:s,codeFencedFenceMeta:s,codeIndented:o(ve,s),codeText:o(xe,s),codeTextData:L,data:L,codeFlowValue:L,definition:o(Ie),definitionDestinationString:s,definitionLabelString:s,definitionTitleString:s,emphasis:o(Ce),hardBreakEscape:o(J),hardBreakTrailing:o(J),htmlFlow:o(fe,s),htmlFlowData:L,htmlText:o(fe,s),htmlTextData:L,image:o(Te),label:s,link:o(me),listItem:o(le),listItemValue:m,listOrdered:o(Ee,g),listUnordered:o(Ee),paragraph:o(Be),reference:M,referenceString:s,resourceDestinationString:s,resourceTitleString:s,setextHeading:o(ke),strong:o(Ue),thematicBreak:o(Ne)},exit:{atxHeading:c(),atxHeadingSequence:N,autolink:c(),autolinkEmail:Se,autolinkProtocol:ie,blockQuote:c(),characterEscapeValue:I,characterReferenceMarkerHexadecimal:Q,characterReferenceMarkerNumeric:Q,characterReferenceValue:oe,characterReference:re,codeFenced:c(x),codeFencedFence:S,codeFencedFenceInfo:b,codeFencedFenceMeta:y,codeFlowValue:I,codeIndented:c(R),codeText:c(Z),codeTextData:I,data:I,definition:c(),definitionDestinationString:C,definitionLabelString:k,definitionTitleString:A,emphasis:c(),hardBreakEscape:c($),hardBreakTrailing:c($),htmlFlow:c(U),htmlFlowData:I,htmlText:c(Y),htmlTextData:I,image:c(G),label:K,labelText:z,lineEnding:H,link:c(j),listItem:c(),listOrdered:c(),listUnordered:c(),paragraph:c(),referenceString:W,resourceDestinationString:D,resourceTitleString:V,resource:B,setextHeading:c(F),setextHeadingLineSequence:O,setextHeadingText:_,strong:c(),thematicBreak:c()}};JG(t,(e||{}).mdastExtensions||[]);const n={};return r;function r(ee){let ce={type:"root",children:[]};const _e={stack:[ce],tokenStack:[],config:t,enter:u,exit:d,buffer:s,resume:p,data:n},Pe=[];let We=-1;for(;++We<ee.length;)if(ee[We][1].type==="listOrdered"||ee[We][1].type==="listUnordered")if(ee[We][0]==="enter")Pe.push(We);else{const yt=Pe.pop();We=a(ee,yt,We)}for(We=-1;++We<ee.length;){const yt=t[ee[We][0]];QG.call(yt,ee[We][1].type)&&yt[ee[We][1].type].call(Object.assign({sliceSerialize:ee[We][2].sliceSerialize},_e),ee[We][1])}if(_e.tokenStack.length>0){const yt=_e.tokenStack[_e.tokenStack.length-1];(yt[1]||vD).call(_e,void 0,yt[0])}for(ce.position={start:wo(ee.length>0?ee[0][1].start:{line:1,column:1,offset:0}),end:wo(ee.length>0?ee[ee.length-2][1].end:{line:1,column:1,offset:0})},We=-1;++We<t.transforms.length;)ce=t.transforms[We](ce)||ce;return ce}function a(ee,ce,_e){let Pe=ce-1,We=-1,yt=!1,kt,ht,et,Tt;for(;++Pe<=_e;){const ot=ee[Pe];switch(ot[1].type){case"listUnordered":case"listOrdered":case"blockQuote":{ot[0]==="enter"?We++:We--,Tt=void 0;break}case"lineEndingBlank":{ot[0]==="enter"&&(kt&&!Tt&&!We&&!et&&(et=Pe),Tt=void 0);break}case"linePrefix":case"listItemValue":case"listItemMarker":case"listItemPrefix":case"listItemPrefixWhitespace":break;default:Tt=void 0}if(!We&&ot[0]==="enter"&&ot[1].type==="listItemPrefix"||We===-1&&ot[0]==="exit"&&(ot[1].type==="listUnordered"||ot[1].type==="listOrdered")){if(kt){let St=Pe;for(ht=void 0;St--;){const Ht=ee[St];if(Ht[1].type==="lineEnding"||Ht[1].type==="lineEndingBlank"){if(Ht[0]==="exit")continue;ht&&(ee[ht][1].type="lineEndingBlank",yt=!0),Ht[1].type="lineEnding",ht=St}else if(!(Ht[1].type==="linePrefix"||Ht[1].type==="blockQuotePrefix"||Ht[1].type==="blockQuotePrefixWhitespace"||Ht[1].type==="blockQuoteMarker"||Ht[1].type==="listItemIndent"))break}et&&(!ht||et<ht)&&(kt._spread=!0),kt.end=Object.assign({},ht?ee[ht][1].start:ot[1].end),ee.splice(ht||Pe,0,["exit",kt,ot[2]]),Pe++,_e++}if(ot[1].type==="listItemPrefix"){const St={type:"listItem",_spread:!1,start:Object.assign({},ot[1].start),end:void 0};kt=St,ee.splice(Pe,0,["enter",St,ot[2]]),Pe++,_e++,et=void 0,Tt=!0}}}return ee[ce][1]._spread=yt,_e}function o(ee,ce){return _e;function _e(Pe){u.call(this,ee(Pe),Pe),ce&&ce.call(this,Pe)}}function s(){this.stack.push({type:"fragment",children:[]})}function u(ee,ce,_e){this.stack[this.stack.length-1].children.push(ee),this.stack.push(ee),this.tokenStack.push([ce,_e||void 0]),ee.position={start:wo(ce.start),end:void 0}}function c(ee){return ce;function ce(_e){ee&&ee.call(this,_e),d.call(this,_e)}}function d(ee,ce){const _e=this.stack.pop(),Pe=this.tokenStack.pop();if(Pe)Pe[0].type!==ee.type&&(ce?ce.call(this,ee,Pe[0]):(Pe[1]||vD).call(this,ee,Pe[0]));else throw new Error("Cannot close `"+ee.type+"` ("+bu({start:ee.start,end:ee.end})+"): it’s not open");_e.position.end=wo(ee.end)}function p(){return YT(this.stack.pop())}function g(){this.data.expectingFirstListItemValue=!0}function m(ee){if(this.data.expectingFirstListItemValue){const ce=this.stack[this.stack.length-2];ce.start=Number.parseInt(this.sliceSerialize(ee),10),this.data.expectingFirstListItemValue=void 0}}function b(){const ee=this.resume(),ce=this.stack[this.stack.length-1];ce.lang=ee}function y(){const ee=this.resume(),ce=this.stack[this.stack.length-1];ce.meta=ee}function S(){this.data.flowCodeInside||(this.buffer(),this.data.flowCodeInside=!0)}function x(){const ee=this.resume(),ce=this.stack[this.stack.length-1];ce.value=ee.replace(/^(\r?\n|\r)|(\r?\n|\r)$/g,""),this.data.flowCodeInside=void 0}function R(){const ee=this.resume(),ce=this.stack[this.stack.length-1];ce.value=ee.replace(/(\r?\n|\r)$/g,"")}function k(ee){const ce=this.resume(),_e=this.stack[this.stack.length-1];_e.label=ce,_e.identifier=zr(this.sliceSerialize(ee)).toLowerCase()}function A(){const ee=this.resume(),ce=this.stack[this.stack.length-1];ce.title=ee}function C(){const ee=this.resume(),ce=this.stack[this.stack.length-1];ce.url=ee}function N(ee){const ce=this.stack[this.stack.length-1];if(!ce.depth){const _e=this.sliceSerialize(ee).length;ce.depth=_e}}function _(){this.data.setextHeadingSlurpLineEnding=!0}function O(ee){const ce=this.stack[this.stack.length-1];ce.depth=this.sliceSerialize(ee).codePointAt(0)===61?1:2}function F(){this.data.setextHeadingSlurpLineEnding=void 0}function L(ee){const _e=this.stack[this.stack.length-1].children;let Pe=_e[_e.length-1];(!Pe||Pe.type!=="text")&&(Pe=he(),Pe.position={start:wo(ee.start),end:void 0},_e.push(Pe)),this.stack.push(Pe)}function I(ee){const ce=this.stack.pop();ce.value+=this.sliceSerialize(ee),ce.position.end=wo(ee.end)}function H(ee){const ce=this.stack[this.stack.length-1];if(this.data.atHardBreak){const _e=ce.children[ce.children.length-1];_e.position.end=wo(ee.end),this.data.atHardBreak=void 0;return}!this.data.setextHeadingSlurpLineEnding&&t.canContainEols.includes(ce.type)&&(L.call(this,ee),I.call(this,ee))}function $(){this.data.atHardBreak=!0}function U(){const ee=this.resume(),ce=this.stack[this.stack.length-1];ce.value=ee}function Y(){const ee=this.resume(),ce=this.stack[this.stack.length-1];ce.value=ee}function Z(){const ee=this.resume(),ce=this.stack[this.stack.length-1];ce.value=ee}function j(){const ee=this.stack[this.stack.length-1];if(this.data.inReference){const ce=this.data.referenceType||"shortcut";ee.type+="Reference",ee.referenceType=ce,delete ee.url,delete ee.title}else delete ee.identifier,delete ee.label;this.data.referenceType=void 0}function G(){const ee=this.stack[this.stack.length-1];if(this.data.inReference){const ce=this.data.referenceType||"shortcut";ee.type+="Reference",ee.referenceType=ce,delete ee.url,delete ee.title}else delete ee.identifier,delete ee.label;this.data.referenceType=void 0}function z(ee){const ce=this.sliceSerialize(ee),_e=this.stack[this.stack.length-2];_e.label=tce(ce),_e.identifier=zr(ce).toLowerCase()}function K(){const ee=this.stack[this.stack.length-1],ce=this.resume(),_e=this.stack[this.stack.length-1];if(this.data.inReference=!0,_e.type==="link"){const Pe=ee.children;_e.children=Pe}else _e.alt=ce}function D(){const ee=this.resume(),ce=this.stack[this.stack.length-1];ce.url=ee}function V(){const ee=this.resume(),ce=this.stack[this.stack.length-1];ce.title=ee}function B(){this.data.inReference=void 0}function M(){this.data.referenceType="collapsed"}function W(ee){const ce=this.resume(),_e=this.stack[this.stack.length-1];_e.label=ce,_e.identifier=zr(this.sliceSerialize(ee)).toLowerCase(),this.data.referenceType="full"}function Q(ee){this.data.characterReferenceType=ee.type}function oe(ee){const ce=this.sliceSerialize(ee),_e=this.data.characterReferenceType;let Pe;_e?(Pe=GG(ce,_e==="characterReferenceMarkerNumeric"?10:16),this.data.characterReferenceType=void 0):Pe=KT(ce);const We=this.stack[this.stack.length-1];We.value+=Pe}function re(ee){const ce=this.stack.pop();ce.position.end=wo(ee.end)}function ie(ee){I.call(this,ee);const ce=this.stack[this.stack.length-1];ce.url=this.sliceSerialize(ee)}function Se(ee){I.call(this,ee);const ce=this.stack[this.stack.length-1];ce.url="mailto:"+this.sliceSerialize(ee)}function ae(){return{type:"blockquote",children:[]}}function ve(){return{type:"code",lang:null,meta:null,value:""}}function xe(){return{type:"inlineCode",value:""}}function Ie(){return{type:"definition",identifier:"",label:null,title:null,url:""}}function Ce(){return{type:"emphasis",children:[]}}function ke(){return{type:"heading",depth:0,children:[]}}function J(){return{type:"break"}}function fe(){return{type:"html",value:""}}function Te(){return{type:"image",title:null,url:"",alt:null}}function me(){return{type:"link",title:null,url:"",children:[]}}function Ee(ee){return{type:"list",ordered:ee.type==="listOrdered",start:null,spread:ee._spread,children:[]}}function le(ee){return{type:"listItem",spread:ee._spread,checked:null,children:[]}}function Be(){return{type:"paragraph",children:[]}}function Ue(){return{type:"strong",children:[]}}function he(){return{type:"text",value:""}}function Ne(){return{type:"thematicBreak"}}}function wo(e){return{line:e.line,column:e.column,offset:e.offset}}function JG(e,t){let n=-1;for(;++n<t.length;){const r=t[n];Array.isArray(r)?JG(e,r):oce(e,r)}}function oce(e,t){let n;for(n in t)if(QG.call(t,n))switch(n){case"canContainEols":{const r=t[n];r&&e[n].push(...r);break}case"transforms":{const r=t[n];r&&e[n].push(...r);break}case"enter":case"exit":{const r=t[n];r&&Object.assign(e[n],r);break}}}function vD(e,t){throw e?new Error("Cannot close `"+e.type+"` ("+bu({start:e.start,end:e.end})+"): a different token (`"+t.type+"`, "+bu({start:t.start,end:t.end})+") is open"):new Error("Cannot close document, a token (`"+t.type+"`, "+bu({start:t.start,end:t.end})+") is still open")}function ice(e){const t=this;t.parser=n;function n(r){return rce(r,{...t.data("settings"),...e,extensions:t.data("micromarkExtensions")||[],mdastExtensions:t.data("fromMarkdownExtensions")||[]})}}function sce(e,t){const n={type:"element",tagName:"blockquote",properties:{},children:e.wrap(e.all(t),!0)};return e.patch(t,n),e.applyData(t,n)}function lce(e,t){const n={type:"element",tagName:"br",properties:{},children:[]};return e.patch(t,n),[e.applyData(t,n),{type:"text",value:`
`}]}function uce(e,t){const n=t.value?t.value+`
`:"",r={};t.lang&&(r.className=["language-"+t.lang]);let a={type:"element",tagName:"code",properties:r,children:[{type:"text",value:n}]};return t.meta&&(a.data={meta:t.meta}),e.patch(t,a),a=e.applyData(t,a),a={type:"element",tagName:"pre",properties:{},children:[a]},e.patch(t,a),a}function cce(e,t){const n={type:"element",tagName:"del",properties:{},children:e.all(t)};return e.patch(t,n),e.applyData(t,n)}function dce(e,t){const n={type:"element",tagName:"em",properties:{},children:e.all(t)};return e.patch(t,n),e.applyData(t,n)}function fce(e,t){const n=typeof e.options.clobberPrefix=="string"?e.options.clobberPrefix:"user-content-",r=String(t.identifier).toUpperCase(),a=nl(r.toLowerCase()),o=e.footnoteOrder.indexOf(r);let s,u=e.footnoteCounts.get(r);u===void 0?(u=0,e.footnoteOrder.push(r),s=e.footnoteOrder.length):s=o+1,u+=1,e.footnoteCounts.set(r,u);const c={type:"element",tagName:"a",properties:{href:"#"+n+"fn-"+a,id:n+"fnref-"+a+(u>1?"-"+u:""),dataFootnoteRef:!0,ariaDescribedBy:["footnote-label"]},children:[{type:"text",value:String(s)}]};e.patch(t,c);const d={type:"element",tagName:"sup",properties:{},children:[c]};return e.patch(t,d),e.applyData(t,d)}function pce(e,t){const n={type:"element",tagName:"h"+t.depth,properties:{},children:e.all(t)};return e.patch(t,n),e.applyData(t,n)}function gce(e,t){if(e.options.allowDangerousHtml){const n={type:"raw",value:t.value};return e.patch(t,n),e.applyData(t,n)}}function eH(e,t){const n=t.referenceType;let r="]";if(n==="collapsed"?r+="[]":n==="full"&&(r+="["+(t.label||t.identifier)+"]"),t.type==="imageReference")return[{type:"text",value:"!["+t.alt+r}];const a=e.all(t),o=a[0];o&&o.type==="text"?o.value="["+o.value:a.unshift({type:"text",value:"["});const s=a[a.length-1];return s&&s.type==="text"?s.value+=r:a.push({type:"text",value:r}),a}function hce(e,t){const n=String(t.identifier).toUpperCase(),r=e.definitionById.get(n);if(!r)return eH(e,t);const a={src:nl(r.url||""),alt:t.alt};r.title!==null&&r.title!==void 0&&(a.title=r.title);const o={type:"element",tagName:"img",properties:a,children:[]};return e.patch(t,o),e.applyData(t,o)}function mce(e,t){const n={src:nl(t.url)};t.alt!==null&&t.alt!==void 0&&(n.alt=t.alt),t.title!==null&&t.title!==void 0&&(n.title=t.title);const r={type:"element",tagName:"img",properties:n,children:[]};return e.patch(t,r),e.applyData(t,r)}function bce(e,t){const n={type:"text",value:t.value.replace(/\r?\n|\r/g," ")};e.patch(t,n);const r={type:"element",tagName:"code",properties:{},children:[n]};return e.patch(t,r),e.applyData(t,r)}function yce(e,t){const n=String(t.identifier).toUpperCase(),r=e.definitionById.get(n);if(!r)return eH(e,t);const a={href:nl(r.url||"")};r.title!==null&&r.title!==void 0&&(a.title=r.title);const o={type:"element",tagName:"a",properties:a,children:e.all(t)};return e.patch(t,o),e.applyData(t,o)}function vce(e,t){const n={href:nl(t.url)};t.title!==null&&t.title!==void 0&&(n.title=t.title);const r={type:"element",tagName:"a",properties:n,children:e.all(t)};return e.patch(t,r),e.applyData(t,r)}function Sce(e,t,n){const r=e.all(t),a=n?Ece(n):tH(t),o={},s=[];if(typeof t.checked=="boolean"){const p=r[0];let g;p&&p.type==="element"&&p.tagName==="p"?g=p:(g={type:"element",tagName:"p",properties:{},children:[]},r.unshift(g)),g.children.length>0&&g.children.unshift({type:"text",value:" "}),g.children.unshift({type:"element",tagName:"input",properties:{type:"checkbox",checked:t.checked,disabled:!0},children:[]}),o.className=["task-list-item"]}let u=-1;for(;++u<r.length;){const p=r[u];(a||u!==0||p.type!=="element"||p.tagName!=="p")&&s.push({type:"text",value:`
`}),p.type==="element"&&p.tagName==="p"&&!a?s.push(...p.children):s.push(p)}const c=r[r.length-1];c&&(a||c.type!=="element"||c.tagName!=="p")&&s.push({type:"text",value:`
`});const d={type:"element",tagName:"li",properties:o,children:s};return e.patch(t,d),e.applyData(t,d)}function Ece(e){let t=!1;if(e.type==="list"){t=e.spread||!1;const n=e.children;let r=-1;for(;!t&&++r<n.length;)t=tH(n[r])}return t}function tH(e){const t=e.spread;return t??e.children.length>1}function wce(e,t){const n={},r=e.all(t);let a=-1;for(typeof t.start=="number"&&t.start!==1&&(n.start=t.start);++a<r.length;){const s=r[a];if(s.type==="element"&&s.tagName==="li"&&s.properties&&Array.isArray(s.properties.className)&&s.properties.className.includes("task-list-item")){n.className=["contains-task-list"];break}}const o={type:"element",tagName:t.ordered?"ol":"ul",properties:n,children:e.wrap(r,!0)};return e.patch(t,o),e.applyData(t,o)}function xce(e,t){const n={type:"element",tagName:"p",properties:{},children:e.all(t)};return e.patch(t,n),e.applyData(t,n)}function kce(e,t){const n={type:"root",children:e.wrap(e.all(t))};return e.patch(t,n),e.applyData(t,n)}function Tce(e,t){const n={type:"element",tagName:"strong",properties:{},children:e.all(t)};return e.patch(t,n),e.applyData(t,n)}function Ace(e,t){const n=e.all(t),r=n.shift(),a=[];if(r){const s={type:"element",tagName:"thead",properties:{},children:e.wrap([r],!0)};e.patch(t.children[0],s),a.push(s)}if(n.length>0){const s={type:"element",tagName:"tbody",properties:{},children:e.wrap(n,!0)},u=$T(t.children[1]),c=IG(t.children[t.children.length-1]);u&&c&&(s.position={start:u,end:c}),a.push(s)}const o={type:"element",tagName:"table",properties:{},children:e.wrap(a,!0)};return e.patch(t,o),e.applyData(t,o)}function Rce(e,t,n){const r=n?n.children:void 0,o=(r?r.indexOf(t):1)===0?"th":"td",s=n&&n.type==="table"?n.align:void 0,u=s?s.length:t.children.length;let c=-1;const d=[];for(;++c<u;){const g=t.children[c],m={},b=s?s[c]:void 0;b&&(m.align=b);let y={type:"element",tagName:o,properties:m,children:[]};g&&(y.children=e.all(g),e.patch(g,y),y=e.applyData(g,y)),d.push(y)}const p={type:"element",tagName:"tr",properties:{},children:e.wrap(d,!0)};return e.patch(t,p),e.applyData(t,p)}function _ce(e,t){const n={type:"element",tagName:"td",properties:{},children:e.all(t)};return e.patch(t,n),e.applyData(t,n)}const SD=9,ED=32;function Cce(e){const t=String(e),n=/\r?\n|\r/g;let r=n.exec(t),a=0;const o=[];for(;r;)o.push(wD(t.slice(a,r.index),a>0,!0),r[0]),a=r.index+r[0].length,r=n.exec(t);return o.push(wD(t.slice(a),a>0,!1)),o.join("")}function wD(e,t,n){let r=0,a=e.length;if(t){let o=e.codePointAt(r);for(;o===SD||o===ED;)r++,o=e.codePointAt(r)}if(n){let o=e.codePointAt(a-1);for(;o===SD||o===ED;)a--,o=e.codePointAt(a-1)}return a>r?e.slice(r,a):""}function Nce(e,t){const n={type:"text",value:Cce(String(t.value))};return e.patch(t,n),e.applyData(t,n)}function Oce(e,t){const n={type:"element",tagName:"hr",properties:{},children:[]};return e.patch(t,n),e.applyData(t,n)}const Dce={blockquote:sce,break:lce,code:uce,delete:cce,emphasis:dce,footnoteReference:fce,heading:pce,html:gce,imageReference:hce,image:mce,inlineCode:bce,linkReference:yce,link:vce,listItem:Sce,list:wce,paragraph:xce,root:kce,strong:Tce,table:Ace,tableCell:_ce,tableRow:Rce,text:Nce,thematicBreak:Oce,toml:zd,yaml:zd,definition:zd,footnoteDefinition:zd};function zd(){}const nH=-1,Dp=0,vu=1,Pf=2,ZT=3,QT=4,JT=5,eA=6,rH=7,aH=8,xD=typeof self=="object"?self:globalThis,Ice=(e,t)=>{const n=(a,o)=>(e.set(o,a),a),r=a=>{if(e.has(a))return e.get(a);const[o,s]=t[a];switch(o){case Dp:case nH:return n(s,a);case vu:{const u=n([],a);for(const c of s)u.push(r(c));return u}case Pf:{const u=n({},a);for(const[c,d]of s)u[r(c)]=r(d);return u}case ZT:return n(new Date(s),a);case QT:{const{source:u,flags:c}=s;return n(new RegExp(u,c),a)}case JT:{const u=n(new Map,a);for(const[c,d]of s)u.set(r(c),r(d));return u}case eA:{const u=n(new Set,a);for(const c of s)u.add(r(c));return u}case rH:{const{name:u,message:c}=s;return n(new xD[u](c),a)}case aH:return n(BigInt(s),a);case"BigInt":return n(Object(BigInt(s)),a);case"ArrayBuffer":return n(new Uint8Array(s).buffer,s);case"DataView":{const{buffer:u}=new Uint8Array(s);return n(new DataView(u),s)}}return n(new xD[o](s),a)};return r},kD=e=>Ice(new Map,e)(0),hs="",{toString:Lce}={},{keys:Mce}=Object,cu=e=>{const t=typeof e;if(t!=="object"||!e)return[Dp,t];const n=Lce.call(e).slice(8,-1);switch(n){case"Array":return[vu,hs];case"Object":return[Pf,hs];case"Date":return[ZT,hs];case"RegExp":return[QT,hs];case"Map":return[JT,hs];case"Set":return[eA,hs];case"DataView":return[vu,n]}return n.includes("Array")?[vu,n]:n.includes("Error")?[rH,n]:[Pf,n]},Bd=([e,t])=>e===Dp&&(t==="function"||t==="symbol"),Fce=(e,t,n,r)=>{const a=(s,u)=>{const c=r.push(s)-1;return n.set(u,c),c},o=s=>{if(n.has(s))return n.get(s);let[u,c]=cu(s);switch(u){case Dp:{let p=s;switch(c){case"bigint":u=aH,p=s.toString();break;case"function":case"symbol":if(e)throw new TypeError("unable to serialize "+c);p=null;break;case"undefined":return a([nH],s)}return a([u,p],s)}case vu:{if(c){let m=s;return c==="DataView"?m=new Uint8Array(s.buffer):c==="ArrayBuffer"&&(m=new Uint8Array(s)),a([c,[...m]],s)}const p=[],g=a([u,p],s);for(const m of s)p.push(o(m));return g}case Pf:{if(c)switch(c){case"BigInt":return a([c,s.toString()],s);case"Boolean":case"Number":case"String":return a([c,s.valueOf()],s)}if(t&&"toJSON"in s)return o(s.toJSON());const p=[],g=a([u,p],s);for(const m of Mce(s))(e||!Bd(cu(s[m])))&&p.push([o(m),o(s[m])]);return g}case ZT:return a([u,s.toISOString()],s);case QT:{const{source:p,flags:g}=s;return a([u,{source:p,flags:g}],s)}case JT:{const p=[],g=a([u,p],s);for(const[m,b]of s)(e||!(Bd(cu(m))||Bd(cu(b))))&&p.push([o(m),o(b)]);return g}case eA:{const p=[],g=a([u,p],s);for(const m of s)(e||!Bd(cu(m)))&&p.push(o(m));return g}}const{message:d}=s;return a([u,{name:c,message:d}],s)};return o},TD=(e,{json:t,lossy:n}={})=>{const r=[];return Fce(!(t||n),!!t,new Map,r)(e),r},zf=typeof structuredClone=="function"?(e,t)=>t&&("json"in t||"lossy"in t)?kD(TD(e,t)):structuredClone(e):(e,t)=>kD(TD(e,t));function Pce(e,t){const n=[{type:"text",value:"↩"}];return t>1&&n.push({type:"element",tagName:"sup",properties:{},children:[{type:"text",value:String(t)}]}),n}function zce(e,t){return"Back to reference "+(e+1)+(t>1?"-"+t:"")}function Bce(e){const t=typeof e.options.clobberPrefix=="string"?e.options.clobberPrefix:"user-content-",n=e.options.footnoteBackContent||Pce,r=e.options.footnoteBackLabel||zce,a=e.options.footnoteLabel||"Footnotes",o=e.options.footnoteLabelTagName||"h2",s=e.options.footnoteLabelProperties||{className:["sr-only"]},u=[];let c=-1;for(;++c<e.footnoteOrder.length;){const d=e.footnoteById.get(e.footnoteOrder[c]);if(!d)continue;const p=e.all(d),g=String(d.identifier).toUpperCase(),m=nl(g.toLowerCase());let b=0;const y=[],S=e.footnoteCounts.get(g);for(;S!==void 0&&++b<=S;){y.length>0&&y.push({type:"text",value:" "});let k=typeof n=="string"?n:n(c,b);typeof k=="string"&&(k={type:"text",value:k}),y.push({type:"element",tagName:"a",properties:{href:"#"+t+"fnref-"+m+(b>1?"-"+b:""),dataFootnoteBackref:"",ariaLabel:typeof r=="string"?r:r(c,b),className:["data-footnote-backref"]},children:Array.isArray(k)?k:[k]})}const x=p[p.length-1];if(x&&x.type==="element"&&x.tagName==="p"){const k=x.children[x.children.length-1];k&&k.type==="text"?k.value+=" ":x.children.push({type:"text",value:" "}),x.children.push(...y)}else p.push(...y);const R={type:"element",tagName:"li",properties:{id:t+"fn-"+m},children:e.wrap(p,!0)};e.patch(d,R),u.push(R)}if(u.length!==0)return{type:"element",tagName:"section",properties:{dataFootnotes:!0,className:["footnotes"]},children:[{type:"element",tagName:o,properties:{...zf(s),id:"footnote-label"},children:[{type:"text",value:a}]},{type:"text",value:`
`},{type:"element",tagName:"ol",properties:{},children:e.wrap(u,!0)},{type:"text",value:`
`}]}}const Ip=function(e){if(e==null)return Hce;if(typeof e=="function")return Lp(e);if(typeof e=="object")return Array.isArray(e)?Uce(e):jce(e);if(typeof e=="string")return Gce(e);throw new Error("Expected function, string, or object as test")};function Uce(e){const t=[];let n=-1;for(;++n<e.length;)t[n]=Ip(e[n]);return Lp(r);function r(...a){let o=-1;for(;++o<t.length;)if(t[o].apply(this,a))return!0;return!1}}function jce(e){const t=e;return Lp(n);function n(r){const a=r;let o;for(o in e)if(a[o]!==t[o])return!1;return!0}}function Gce(e){return Lp(t);function t(n){return n&&n.type===e}}function Lp(e){return t;function t(n,r,a){return!!($ce(n)&&e.call(this,n,typeof r=="number"?r:void 0,a||void 0))}}function Hce(){return!0}function $ce(e){return e!==null&&typeof e=="object"&&"type"in e}const oH=[],qce=!0,p0=!1,Vce="skip";function iH(e,t,n,r){let a;typeof t=="function"&&typeof n!="function"?(r=n,n=t):a=t;const o=Ip(a),s=r?-1:1;u(e,void 0,[])();function u(c,d,p){const g=c&&typeof c=="object"?c:{};if(typeof g.type=="string"){const b=typeof g.tagName=="string"?g.tagName:typeof g.name=="string"?g.name:void 0;Object.defineProperty(m,"name",{value:"node ("+(c.type+(b?"<"+b+">":""))+")"})}return m;function m(){let b=oH,y,S,x;if((!t||o(c,d,p[p.length-1]||void 0))&&(b=Wce(n(c,p)),b[0]===p0))return b;if("children"in c&&c.children){const R=c;if(R.children&&b[0]!==Vce)for(S=(r?R.children.length:-1)+s,x=p.concat(R);S>-1&&S<R.children.length;){const k=R.children[S];if(y=u(k,S,x)(),y[0]===p0)return y;S=typeof y[1]=="number"?y[1]:S+s}}return b}}}function Wce(e){return Array.isArray(e)?e:typeof e=="number"?[qce,e]:e==null?oH:[e]}function tA(e,t,n,r){let a,o,s;typeof t=="function"?(o=void 0,s=t,a=n):(o=t,s=n,a=r),iH(e,o,u,a);function u(c,d){const p=d[d.length-1],g=p?p.children.indexOf(c):void 0;return s(c,g,p)}}const g0={}.hasOwnProperty,Yce={};function Kce(e,t){const n=t||Yce,r=new Map,a=new Map,o=new Map,s={...Dce,...n.handlers},u={all:d,applyData:Zce,definitionById:r,footnoteById:a,footnoteCounts:o,footnoteOrder:[],handlers:s,one:c,options:n,patch:Xce,wrap:Jce};return tA(e,function(p){if(p.type==="definition"||p.type==="footnoteDefinition"){const g=p.type==="definition"?r:a,m=String(p.identifier).toUpperCase();g.has(m)||g.set(m,p)}}),u;function c(p,g){const m=p.type,b=u.handlers[m];if(g0.call(u.handlers,m)&&b)return b(u,p,g);if(u.options.passThrough&&u.options.passThrough.includes(m)){if("children"in p){const{children:S,...x}=p,R=zf(x);return R.children=u.all(p),R}return zf(p)}return(u.options.unknownHandler||Qce)(u,p,g)}function d(p){const g=[];if("children"in p){const m=p.children;let b=-1;for(;++b<m.length;){const y=u.one(m[b],p);if(y){if(b&&m[b-1].type==="break"&&(!Array.isArray(y)&&y.type==="text"&&(y.value=AD(y.value)),!Array.isArray(y)&&y.type==="element")){const S=y.children[0];S&&S.type==="text"&&(S.value=AD(S.value))}Array.isArray(y)?g.push(...y):g.push(y)}}}return g}}function Xce(e,t){e.position&&(t.position=Pse(e))}function Zce(e,t){let n=t;if(e&&e.data){const r=e.data.hName,a=e.data.hChildren,o=e.data.hProperties;if(typeof r=="string")if(n.type==="element")n.tagName=r;else{const s="children"in n?n.children:[n];n={type:"element",tagName:r,properties:{},children:s}}n.type==="element"&&o&&Object.assign(n.properties,zf(o)),"children"in n&&n.children&&a!==null&&a!==void 0&&(n.children=a)}return n}function Qce(e,t){const n=t.data||{},r="value"in t&&!(g0.call(n,"hProperties")||g0.call(n,"hChildren"))?{type:"text",value:t.value}:{type:"element",tagName:"div",properties:{},children:e.all(t)};return e.patch(t,r),e.applyData(t,r)}function Jce(e,t){const n=[];let r=-1;for(t&&n.push({type:"text",value:`
`});++r<e.length;)r&&n.push({type:"text",value:`
`}),n.push(e[r]);return t&&e.length>0&&n.push({type:"text",value:`
`}),n}function AD(e){let t=0,n=e.charCodeAt(t);for(;n===9||n===32;)t++,n=e.charCodeAt(t);return e.slice(t)}function RD(e,t){const n=Kce(e,t),r=n.one(e,void 0),a=Bce(n),o=Array.isArray(r)?{type:"root",children:r}:r||{type:"root",children:[]};return a&&o.children.push({type:"text",value:`
`},a),o}function ede(e,t){return e&&"run"in e?async function(n,r){const a=RD(n,{file:r,...t});await e.run(a,r)}:function(n,r){return RD(n,{file:r,...e||t})}}function _D(e){if(e)throw e}var hb,CD;function tde(){if(CD)return hb;CD=1;var e=Object.prototype.hasOwnProperty,t=Object.prototype.toString,n=Object.defineProperty,r=Object.getOwnPropertyDescriptor,a=function(d){return typeof Array.isArray=="function"?Array.isArray(d):t.call(d)==="[object Array]"},o=function(d){if(!d||t.call(d)!=="[object Object]")return!1;var p=e.call(d,"constructor"),g=d.constructor&&d.constructor.prototype&&e.call(d.constructor.prototype,"isPrototypeOf");if(d.constructor&&!p&&!g)return!1;var m;for(m in d);return typeof m>"u"||e.call(d,m)},s=function(d,p){n&&p.name==="__proto__"?n(d,p.name,{enumerable:!0,configurable:!0,value:p.newValue,writable:!0}):d[p.name]=p.newValue},u=function(d,p){if(p==="__proto__")if(e.call(d,p)){if(r)return r(d,p).value}else return;return d[p]};return hb=function c(){var d,p,g,m,b,y,S=arguments[0],x=1,R=arguments.length,k=!1;for(typeof S=="boolean"&&(k=S,S=arguments[1]||{},x=2),(S==null||typeof S!="object"&&typeof S!="function")&&(S={});x<R;++x)if(d=arguments[x],d!=null)for(p in d)g=u(S,p),m=u(d,p),S!==m&&(k&&m&&(o(m)||(b=a(m)))?(b?(b=!1,y=g&&a(g)?g:[]):y=g&&o(g)?g:{},s(S,{name:p,newValue:c(k,y,m)})):typeof m<"u"&&s(S,{name:p,newValue:m}));return S},hb}var nde=tde();const mb=un(nde);function h0(e){if(typeof e!="object"||e===null)return!1;const t=Object.getPrototypeOf(e);return(t===null||t===Object.prototype||Object.getPrototypeOf(t)===null)&&!(Symbol.toStringTag in e)&&!(Symbol.iterator in e)}function rde(){const e=[],t={run:n,use:r};return t;function n(...a){let o=-1;const s=a.pop();if(typeof s!="function")throw new TypeError("Expected function as last argument, not "+s);u(null,...a);function u(c,...d){const p=e[++o];let g=-1;if(c){s(c);return}for(;++g<a.length;)(d[g]===null||d[g]===void 0)&&(d[g]=a[g]);a=d,p?ade(p,u)(...d):s(null,...d)}}function r(a){if(typeof a!="function")throw new TypeError("Expected `middelware` to be a function, not "+a);return e.push(a),t}}function ade(e,t){let n;return r;function r(...s){const u=e.length>s.length;let c;u&&s.push(a);try{c=e.apply(this,s)}catch(d){const p=d;if(u&&n)throw p;return a(p)}u||(c&&c.then&&typeof c.then=="function"?c.then(o,a):c instanceof Error?a(c):o(c))}function a(s,...u){n||(n=!0,t(s,...u))}function o(s){a(null,s)}}const Jr={basename:ode,dirname:ide,extname:sde,join:lde,sep:"/"};function ode(e,t){if(t!==void 0&&typeof t!="string")throw new TypeError('"ext" argument must be a string');oc(e);let n=0,r=-1,a=e.length,o;if(t===void 0||t.length===0||t.length>e.length){for(;a--;)if(e.codePointAt(a)===47){if(o){n=a+1;break}}else r<0&&(o=!0,r=a+1);return r<0?"":e.slice(n,r)}if(t===e)return"";let s=-1,u=t.length-1;for(;a--;)if(e.codePointAt(a)===47){if(o){n=a+1;break}}else s<0&&(o=!0,s=a+1),u>-1&&(e.codePointAt(a)===t.codePointAt(u--)?u<0&&(r=a):(u=-1,r=s));return n===r?r=s:r<0&&(r=e.length),e.slice(n,r)}function ide(e){if(oc(e),e.length===0)return".";let t=-1,n=e.length,r;for(;--n;)if(e.codePointAt(n)===47){if(r){t=n;break}}else r||(r=!0);return t<0?e.codePointAt(0)===47?"/":".":t===1&&e.codePointAt(0)===47?"//":e.slice(0,t)}function sde(e){oc(e);let t=e.length,n=-1,r=0,a=-1,o=0,s;for(;t--;){const u=e.codePointAt(t);if(u===47){if(s){r=t+1;break}continue}n<0&&(s=!0,n=t+1),u===46?a<0?a=t:o!==1&&(o=1):a>-1&&(o=-1)}return a<0||n<0||o===0||o===1&&a===n-1&&a===r+1?"":e.slice(a,n)}function lde(...e){let t=-1,n;for(;++t<e.length;)oc(e[t]),e[t]&&(n=n===void 0?e[t]:n+"/"+e[t]);return n===void 0?".":ude(n)}function ude(e){oc(e);const t=e.codePointAt(0)===47;let n=cde(e,!t);return n.length===0&&!t&&(n="."),n.length>0&&e.codePointAt(e.length-1)===47&&(n+="/"),t?"/"+n:n}function cde(e,t){let n="",r=0,a=-1,o=0,s=-1,u,c;for(;++s<=e.length;){if(s<e.length)u=e.codePointAt(s);else{if(u===47)break;u=47}if(u===47){if(!(a===s-1||o===1))if(a!==s-1&&o===2){if(n.length<2||r!==2||n.codePointAt(n.length-1)!==46||n.codePointAt(n.length-2)!==46){if(n.length>2){if(c=n.lastIndexOf("/"),c!==n.length-1){c<0?(n="",r=0):(n=n.slice(0,c),r=n.length-1-n.lastIndexOf("/")),a=s,o=0;continue}}else if(n.length>0){n="",r=0,a=s,o=0;continue}}t&&(n=n.length>0?n+"/..":"..",r=2)}else n.length>0?n+="/"+e.slice(a+1,s):n=e.slice(a+1,s),r=s-a-1;a=s,o=0}else u===46&&o>-1?o++:o=-1}return n}function oc(e){if(typeof e!="string")throw new TypeError("Path must be a string. Received "+JSON.stringify(e))}const dde={cwd:fde};function fde(){return"/"}function m0(e){return!!(e!==null&&typeof e=="object"&&"href"in e&&e.href&&"protocol"in e&&e.protocol&&e.auth===void 0)}function pde(e){if(typeof e=="string")e=new URL(e);else if(!m0(e)){const t=new TypeError('The "path" argument must be of type string or an instance of URL. Received `'+e+"`");throw t.code="ERR_INVALID_ARG_TYPE",t}if(e.protocol!=="file:"){const t=new TypeError("The URL must be of scheme file");throw t.code="ERR_INVALID_URL_SCHEME",t}return gde(e)}function gde(e){if(e.hostname!==""){const r=new TypeError('File URL host must be "localhost" or empty on darwin');throw r.code="ERR_INVALID_FILE_URL_HOST",r}const t=e.pathname;let n=-1;for(;++n<t.length;)if(t.codePointAt(n)===37&&t.codePointAt(n+1)===50){const r=t.codePointAt(n+2);if(r===70||r===102){const a=new TypeError("File URL path must not include encoded / characters");throw a.code="ERR_INVALID_FILE_URL_PATH",a}}return decodeURIComponent(t)}const bb=["history","path","basename","stem","extname","dirname"];class sH{constructor(t){let n;t?m0(t)?n={path:t}:typeof t=="string"||hde(t)?n={value:t}:n=t:n={},this.cwd="cwd"in n?"":dde.cwd(),this.data={},this.history=[],this.messages=[],this.value,this.map,this.result,this.stored;let r=-1;for(;++r<bb.length;){const o=bb[r];o in n&&n[o]!==void 0&&n[o]!==null&&(this[o]=o==="history"?[...n[o]]:n[o])}let a;for(a in n)bb.includes(a)||(this[a]=n[a])}get basename(){return typeof this.path=="string"?Jr.basename(this.path):void 0}set basename(t){vb(t,"basename"),yb(t,"basename"),this.path=Jr.join(this.dirname||"",t)}get dirname(){return typeof this.path=="string"?Jr.dirname(this.path):void 0}set dirname(t){ND(this.basename,"dirname"),this.path=Jr.join(t||"",this.basename)}get extname(){return typeof this.path=="string"?Jr.extname(this.path):void 0}set extname(t){if(yb(t,"extname"),ND(this.dirname,"extname"),t){if(t.codePointAt(0)!==46)throw new Error("`extname` must start with `.`");if(t.includes(".",1))throw new Error("`extname` cannot contain multiple dots")}this.path=Jr.join(this.dirname,this.stem+(t||""))}get path(){return this.history[this.history.length-1]}set path(t){m0(t)&&(t=pde(t)),vb(t,"path"),this.path!==t&&this.history.push(t)}get stem(){return typeof this.path=="string"?Jr.basename(this.path,this.extname):void 0}set stem(t){vb(t,"stem"),yb(t,"stem"),this.path=Jr.join(this.dirname||"",t+(this.extname||""))}fail(t,n,r){const a=this.message(t,n,r);throw a.fatal=!0,a}info(t,n,r){const a=this.message(t,n,r);return a.fatal=void 0,a}message(t,n,r){const a=new Rn(t,n,r);return this.path&&(a.name=this.path+":"+a.name,a.file=this.path),a.fatal=!1,this.messages.push(a),a}toString(t){return this.value===void 0?"":typeof this.value=="string"?this.value:new TextDecoder(t||void 0).decode(this.value)}}function yb(e,t){if(e&&e.includes(Jr.sep))throw new Error("`"+t+"` cannot be a path: did not expect `"+Jr.sep+"`")}function vb(e,t){if(!e)throw new Error("`"+t+"` cannot be empty")}function ND(e,t){if(!e)throw new Error("Setting `"+t+"` requires `path` to be set too")}function hde(e){return!!(e&&typeof e=="object"&&"byteLength"in e&&"byteOffset"in e)}const mde=function(e){const r=this.constructor.prototype,a=r[e],o=function(){return a.apply(o,arguments)};return Object.setPrototypeOf(o,r),o},bde={}.hasOwnProperty;class nA extends mde{constructor(){super("copy"),this.Compiler=void 0,this.Parser=void 0,this.attachers=[],this.compiler=void 0,this.freezeIndex=-1,this.frozen=void 0,this.namespace={},this.parser=void 0,this.transformers=rde()}copy(){const t=new nA;let n=-1;for(;++n<this.attachers.length;){const r=this.attachers[n];t.use(...r)}return t.data(mb(!0,{},this.namespace)),t}data(t,n){return typeof t=="string"?arguments.length===2?(wb("data",this.frozen),this.namespace[t]=n,this):bde.call(this.namespace,t)&&this.namespace[t]||void 0:t?(wb("data",this.frozen),this.namespace=t,this):this.namespace}freeze(){if(this.frozen)return this;const t=this;for(;++this.freezeIndex<this.attachers.length;){const[n,...r]=this.attachers[this.freezeIndex];if(r[0]===!1)continue;r[0]===!0&&(r[0]=void 0);const a=n.call(t,...r);typeof a=="function"&&this.transformers.use(a)}return this.frozen=!0,this.freezeIndex=Number.POSITIVE_INFINITY,this}parse(t){this.freeze();const n=Ud(t),r=this.parser||this.Parser;return Sb("parse",r),r(String(n),n)}process(t,n){const r=this;return this.freeze(),Sb("process",this.parser||this.Parser),Eb("process",this.compiler||this.Compiler),n?a(void 0,n):new Promise(a);function a(o,s){const u=Ud(t),c=r.parse(u);r.run(c,u,function(p,g,m){if(p||!g||!m)return d(p);const b=g,y=r.stringify(b,m);Sde(y)?m.value=y:m.result=y,d(p,m)});function d(p,g){p||!g?s(p):o?o(g):n(void 0,g)}}}processSync(t){let n=!1,r;return this.freeze(),Sb("processSync",this.parser||this.Parser),Eb("processSync",this.compiler||this.Compiler),this.process(t,a),DD("processSync","process",n),r;function a(o,s){n=!0,_D(o),r=s}}run(t,n,r){OD(t),this.freeze();const a=this.transformers;return!r&&typeof n=="function"&&(r=n,n=void 0),r?o(void 0,r):new Promise(o);function o(s,u){const c=Ud(n);a.run(t,c,d);function d(p,g,m){const b=g||t;p?u(p):s?s(b):r(void 0,b,m)}}}runSync(t,n){let r=!1,a;return this.run(t,n,o),DD("runSync","run",r),a;function o(s,u){_D(s),a=u,r=!0}}stringify(t,n){this.freeze();const r=Ud(n),a=this.compiler||this.Compiler;return Eb("stringify",a),OD(t),a(t,r)}use(t,...n){const r=this.attachers,a=this.namespace;if(wb("use",this.frozen),t!=null)if(typeof t=="function")c(t,n);else if(typeof t=="object")Array.isArray(t)?u(t):s(t);else throw new TypeError("Expected usable value, not `"+t+"`");return this;function o(d){if(typeof d=="function")c(d,[]);else if(typeof d=="object")if(Array.isArray(d)){const[p,...g]=d;c(p,g)}else s(d);else throw new TypeError("Expected usable value, not `"+d+"`")}function s(d){if(!("plugins"in d)&&!("settings"in d))throw new Error("Expected usable value but received an empty preset, which is probably a mistake: presets typically come with `plugins` and sometimes with `settings`, but this has neither");u(d.plugins),d.settings&&(a.settings=mb(!0,a.settings,d.settings))}function u(d){let p=-1;if(d!=null)if(Array.isArray(d))for(;++p<d.length;){const g=d[p];o(g)}else throw new TypeError("Expected a list of plugins, not `"+d+"`")}function c(d,p){let g=-1,m=-1;for(;++g<r.length;)if(r[g][0]===d){m=g;break}if(m===-1)r.push([d,...p]);else if(p.length>0){let[b,...y]=p;const S=r[m][1];h0(S)&&h0(b)&&(b=mb(!0,S,b)),r[m]=[d,b,...y]}}}}const yde=new nA().freeze();function Sb(e,t){if(typeof t!="function")throw new TypeError("Cannot `"+e+"` without `parser`")}function Eb(e,t){if(typeof t!="function")throw new TypeError("Cannot `"+e+"` without `compiler`")}function wb(e,t){if(t)throw new Error("Cannot call `"+e+"` on a frozen processor.\nCreate a new processor first, by calling it: use `processor()` instead of `processor`.")}function OD(e){if(!h0(e)||typeof e.type!="string")throw new TypeError("Expected node, got `"+e+"`")}function DD(e,t,n){if(!n)throw new Error("`"+e+"` finished async. Use `"+t+"` instead")}function Ud(e){return vde(e)?e:new sH(e)}function vde(e){return!!(e&&typeof e=="object"&&"message"in e&&"messages"in e)}function Sde(e){return typeof e=="string"||Ede(e)}function Ede(e){return!!(e&&typeof e=="object"&&"byteLength"in e&&"byteOffset"in e)}const wde="https://github.com/remarkjs/react-markdown/blob/main/changelog.md",ID=[],LD={allowDangerousHtml:!0},xde=/^(https?|ircs?|mailto|xmpp)$/i,kde=[{from:"astPlugins",id:"remove-buggy-html-in-markdown-parser"},{from:"allowDangerousHtml",id:"remove-buggy-html-in-markdown-parser"},{from:"allowNode",id:"replace-allownode-allowedtypes-and-disallowedtypes",to:"allowElement"},{from:"allowedTypes",id:"replace-allownode-allowedtypes-and-disallowedtypes",to:"allowedElements"},{from:"disallowedTypes",id:"replace-allownode-allowedtypes-and-disallowedtypes",to:"disallowedElements"},{from:"escapeHtml",id:"remove-buggy-html-in-markdown-parser"},{from:"includeElementIndex",id:"#remove-includeelementindex"},{from:"includeNodeIndex",id:"change-includenodeindex-to-includeelementindex"},{from:"linkTarget",id:"remove-linktarget"},{from:"plugins",id:"change-plugins-to-remarkplugins",to:"remarkPlugins"},{from:"rawSourcePos",id:"#remove-rawsourcepos"},{from:"renderers",id:"change-renderers-to-components",to:"components"},{from:"source",id:"change-source-to-children",to:"children"},{from:"sourcePos",id:"#remove-sourcepos"},{from:"transformImageUri",id:"#add-urltransform",to:"urlTransform"},{from:"transformLinkUri",id:"#add-urltransform",to:"urlTransform"}];function Tde(e){const t=Ade(e),n=Rde(e);return _de(t.runSync(t.parse(n),n),e)}function Ade(e){const t=e.rehypePlugins||ID,n=e.remarkPlugins||ID,r=e.remarkRehypeOptions?{...e.remarkRehypeOptions,...LD}:LD;return yde().use(ice).use(n).use(ede,r).use(t)}function Rde(e){const t=e.children||"",n=new sH;return typeof t=="string"&&(n.value=t),n}function _de(e,t){const n=t.allowedElements,r=t.allowElement,a=t.components,o=t.disallowedElements,s=t.skipHtml,u=t.unwrapDisallowed,c=t.urlTransform||Cde;for(const p of kde)Object.hasOwn(t,p.from)&&(""+p.from+(p.to?"use `"+p.to+"` instead":"remove it")+wde+p.id,void 0);return t.className&&(e={type:"element",tagName:"div",properties:{className:t.className},children:e.type==="root"?e.children:[e]}),tA(e,d),FG(e,{Fragment:w.Fragment,components:a,ignoreInvalidStyle:!0,jsx:w.jsx,jsxs:w.jsxs,passKeys:!0,passNode:!0});function d(p,g,m){if(p.type==="raw"&&m&&typeof g=="number")return s?m.children.splice(g,1):m.children[g]={type:"text",value:p.value},g;if(p.type==="element"){let b;for(b in fb)if(Object.hasOwn(fb,b)&&Object.hasOwn(p.properties,b)){const y=p.properties[b],S=fb[b];(S===null||S.includes(p.tagName))&&(p.properties[b]=c(String(y||""),b,p))}}if(p.type==="element"){let b=n?!n.includes(p.tagName):o?o.includes(p.tagName):!1;if(!b&&r&&typeof g=="number"&&(b=!r(p,g,m)),b&&m&&typeof g=="number")return u&&p.children?m.children.splice(g,1,...p.children):m.children.splice(g,1),g}}}function Cde(e){const t=e.indexOf(":"),n=e.indexOf("?"),r=e.indexOf("#"),a=e.indexOf("/");return t===-1||a!==-1&&t>a||n!==-1&&t>n||r!==-1&&t>r||xde.test(e.slice(0,t))?e:""}function MD(e,t){const n=String(e);if(typeof t!="string")throw new TypeError("Expected character");let r=0,a=n.indexOf(t);for(;a!==-1;)r++,a=n.indexOf(t,a+t.length);return r}function Nde(e){if(typeof e!="string")throw new TypeError("Expected a string");return e.replace(/[|\\{}()[\]^$+*?.]/g,"\\$&").replace(/-/g,"\\x2d")}function Ode(e,t,n){const a=Ip((n||{}).ignore||[]),o=Dde(t);let s=-1;for(;++s<o.length;)iH(e,"text",u);function u(d,p){let g=-1,m;for(;++g<p.length;){const b=p[g],y=m?m.children:void 0;if(a(b,y?y.indexOf(b):void 0,m))return;m=b}if(m)return c(d,p)}function c(d,p){const g=p[p.length-1],m=o[s][0],b=o[s][1];let y=0;const x=g.children.indexOf(d);let R=!1,k=[];m.lastIndex=0;let A=m.exec(d.value);for(;A;){const C=A.index,N={index:A.index,input:A.input,stack:[...p,d]};let _=b(...A,N);if(typeof _=="string"&&(_=_.length>0?{type:"text",value:_}:void 0),_===!1?m.lastIndex=C+1:(y!==C&&k.push({type:"text",value:d.value.slice(y,C)}),Array.isArray(_)?k.push(..._):_&&k.push(_),y=C+A[0].length,R=!0),!m.global)break;A=m.exec(d.value)}return R?(y<d.value.length&&k.push({type:"text",value:d.value.slice(y)}),g.children.splice(x,1,...k)):k=[d],x+k.length}}function Dde(e){const t=[];if(!Array.isArray(e))throw new TypeError("Expected find and replace tuple or list of tuples");const n=!e[0]||Array.isArray(e[0])?e:[e];let r=-1;for(;++r<n.length;){const a=n[r];t.push([Ide(a[0]),Lde(a[1])])}return t}function Ide(e){return typeof e=="string"?new RegExp(Nde(e),"g"):e}function Lde(e){return typeof e=="function"?e:function(){return e}}const xb="phrasing",kb=["autolink","link","image","label"];function Mde(){return{transforms:[Gde],enter:{literalAutolink:Pde,literalAutolinkEmail:Tb,literalAutolinkHttp:Tb,literalAutolinkWww:Tb},exit:{literalAutolink:jde,literalAutolinkEmail:Ude,literalAutolinkHttp:zde,literalAutolinkWww:Bde}}}function Fde(){return{unsafe:[{character:"@",before:"[+\\-.\\w]",after:"[\\-.\\w]",inConstruct:xb,notInConstruct:kb},{character:".",before:"[Ww]",after:"[\\-.\\w]",inConstruct:xb,notInConstruct:kb},{character:":",before:"[ps]",after:"\\/",inConstruct:xb,notInConstruct:kb}]}}function Pde(e){this.enter({type:"link",title:null,url:"",children:[]},e)}function Tb(e){this.config.enter.autolinkProtocol.call(this,e)}function zde(e){this.config.exit.autolinkProtocol.call(this,e)}function Bde(e){this.config.exit.data.call(this,e);const t=this.stack[this.stack.length-1];t.type,t.url="http://"+this.sliceSerialize(e)}function Ude(e){this.config.exit.autolinkEmail.call(this,e)}function jde(e){this.exit(e)}function Gde(e){Ode(e,[[/(https?:\/\/|www(?=\.))([-.\w]+)([^ \t\r\n]*)/gi,Hde],[new RegExp("(?<=^|\\s|\\p{P}|\\p{S})([-.\\w+]+)@([-\\w]+(?:\\.[-\\w]+)+)","gu"),$de]],{ignore:["link","linkReference"]})}function Hde(e,t,n,r,a){let o="";if(!lH(a)||(/^w/i.test(t)&&(n=t+n,t="",o="http://"),!qde(n)))return!1;const s=Vde(n+r);if(!s[0])return!1;const u={type:"link",title:null,url:o+t+s[0],children:[{type:"text",value:t+s[0]}]};return s[1]?[u,{type:"text",value:s[1]}]:u}function $de(e,t,n,r){return!lH(r,!0)||/[-\d_]$/.test(n)?!1:{type:"link",title:null,url:"mailto:"+t+"@"+n,children:[{type:"text",value:t+"@"+n}]}}function qde(e){const t=e.split(".");return!(t.length<2||t[t.length-1]&&(/_/.test(t[t.length-1])||!/[a-zA-Z\d]/.test(t[t.length-1]))||t[t.length-2]&&(/_/.test(t[t.length-2])||!/[a-zA-Z\d]/.test(t[t.length-2])))}function Vde(e){const t=/[!"&'),.:;<>?\]}]+$/.exec(e);if(!t)return[e,void 0];e=e.slice(0,t.index);let n=t[0],r=n.indexOf(")");const a=MD(e,"(");let o=MD(e,")");for(;r!==-1&&a>o;)e+=n.slice(0,r+1),n=n.slice(r+1),r=n.indexOf(")"),o++;return[e,n]}function lH(e,t){const n=e.input.charCodeAt(e.index-1);return(e.index===0||Si(n)||Np(n))&&(!t||n!==47)}uH.peek=tfe;function Wde(){this.buffer()}function Yde(e){this.enter({type:"footnoteReference",identifier:"",label:""},e)}function Kde(){this.buffer()}function Xde(e){this.enter({type:"footnoteDefinition",identifier:"",label:"",children:[]},e)}function Zde(e){const t=this.resume(),n=this.stack[this.stack.length-1];n.type,n.identifier=zr(this.sliceSerialize(e)).toLowerCase(),n.label=t}function Qde(e){this.exit(e)}function Jde(e){const t=this.resume(),n=this.stack[this.stack.length-1];n.type,n.identifier=zr(this.sliceSerialize(e)).toLowerCase(),n.label=t}function efe(e){this.exit(e)}function tfe(){return"["}function uH(e,t,n,r){const a=n.createTracker(r);let o=a.move("[^");const s=n.enter("footnoteReference"),u=n.enter("reference");return o+=a.move(n.safe(n.associationId(e),{after:"]",before:o})),u(),s(),o+=a.move("]"),o}function nfe(){return{enter:{gfmFootnoteCallString:Wde,gfmFootnoteCall:Yde,gfmFootnoteDefinitionLabelString:Kde,gfmFootnoteDefinition:Xde},exit:{gfmFootnoteCallString:Zde,gfmFootnoteCall:Qde,gfmFootnoteDefinitionLabelString:Jde,gfmFootnoteDefinition:efe}}}function rfe(e){let t=!1;return e&&e.firstLineBlank&&(t=!0),{handlers:{footnoteDefinition:n,footnoteReference:uH},unsafe:[{character:"[",inConstruct:["label","phrasing","reference"]}]};function n(r,a,o,s){const u=o.createTracker(s);let c=u.move("[^");const d=o.enter("footnoteDefinition"),p=o.enter("label");return c+=u.move(o.safe(o.associationId(r),{before:c,after:"]"})),p(),c+=u.move("]:"),r.children&&r.children.length>0&&(u.shift(4),c+=u.move((t?`
`:" ")+o.indentLines(o.containerFlow(r,u.current()),t?cH:afe))),d(),c}}function afe(e,t,n){return t===0?e:cH(e,t,n)}function cH(e,t,n){return(n?"":"    ")+e}const ofe=["autolink","destinationLiteral","destinationRaw","reference","titleQuote","titleApostrophe"];dH.peek=cfe;function ife(){return{canContainEols:["delete"],enter:{strikethrough:lfe},exit:{strikethrough:ufe}}}function sfe(){return{unsafe:[{character:"~",inConstruct:"phrasing",notInConstruct:ofe}],handlers:{delete:dH}}}function lfe(e){this.enter({type:"delete",children:[]},e)}function ufe(e){this.exit(e)}function dH(e,t,n,r){const a=n.createTracker(r),o=n.enter("strikethrough");let s=a.move("~~");return s+=n.containerPhrasing(e,{...a.current(),before:s,after:"~"}),s+=a.move("~~"),o(),s}function cfe(){return"~"}function dfe(e){return e.length}function ffe(e,t){const n=t||{},r=(n.align||[]).concat(),a=n.stringLength||dfe,o=[],s=[],u=[],c=[];let d=0,p=-1;for(;++p<e.length;){const S=[],x=[];let R=-1;for(e[p].length>d&&(d=e[p].length);++R<e[p].length;){const k=pfe(e[p][R]);if(n.alignDelimiters!==!1){const A=a(k);x[R]=A,(c[R]===void 0||A>c[R])&&(c[R]=A)}S.push(k)}s[p]=S,u[p]=x}let g=-1;if(typeof r=="object"&&"length"in r)for(;++g<d;)o[g]=FD(r[g]);else{const S=FD(r);for(;++g<d;)o[g]=S}g=-1;const m=[],b=[];for(;++g<d;){const S=o[g];let x="",R="";S===99?(x=":",R=":"):S===108?x=":":S===114&&(R=":");let k=n.alignDelimiters===!1?1:Math.max(1,c[g]-x.length-R.length);const A=x+"-".repeat(k)+R;n.alignDelimiters!==!1&&(k=x.length+k+R.length,k>c[g]&&(c[g]=k),b[g]=k),m[g]=A}s.splice(1,0,m),u.splice(1,0,b),p=-1;const y=[];for(;++p<s.length;){const S=s[p],x=u[p];g=-1;const R=[];for(;++g<d;){const k=S[g]||"";let A="",C="";if(n.alignDelimiters!==!1){const N=c[g]-(x[g]||0),_=o[g];_===114?A=" ".repeat(N):_===99?N%2?(A=" ".repeat(N/2+.5),C=" ".repeat(N/2-.5)):(A=" ".repeat(N/2),C=A):C=" ".repeat(N)}n.delimiterStart!==!1&&!g&&R.push("|"),n.padding!==!1&&!(n.alignDelimiters===!1&&k==="")&&(n.delimiterStart!==!1||g)&&R.push(" "),n.alignDelimiters!==!1&&R.push(A),R.push(k),n.alignDelimiters!==!1&&R.push(C),n.padding!==!1&&R.push(" "),(n.delimiterEnd!==!1||g!==d-1)&&R.push("|")}y.push(n.delimiterEnd===!1?R.join("").replace(/ +$/,""):R.join(""))}return y.join(`
`)}function pfe(e){return e==null?"":String(e)}function FD(e){const t=typeof e=="string"?e.codePointAt(0):0;return t===67||t===99?99:t===76||t===108?108:t===82||t===114?114:0}function gfe(e,t,n,r){const a=n.enter("blockquote"),o=n.createTracker(r);o.move("> "),o.shift(2);const s=n.indentLines(n.containerFlow(e,o.current()),hfe);return a(),s}function hfe(e,t,n){return">"+(n?"":" ")+e}function mfe(e,t){return PD(e,t.inConstruct,!0)&&!PD(e,t.notInConstruct,!1)}function PD(e,t,n){if(typeof t=="string"&&(t=[t]),!t||t.length===0)return n;let r=-1;for(;++r<t.length;)if(e.includes(t[r]))return!0;return!1}function zD(e,t,n,r){let a=-1;for(;++a<n.unsafe.length;)if(n.unsafe[a].character===`
`&&mfe(n.stack,n.unsafe[a]))return/[ \t]/.test(r.before)?"":" ";return`\\
`}function fH(e,t){const n=String(e);let r=n.indexOf(t),a=r,o=0,s=0;if(typeof t!="string")throw new TypeError("Expected substring");for(;r!==-1;)r===a?++o>s&&(s=o):o=1,a=r+t.length,r=n.indexOf(t,a);return s}function bfe(e,t){return!!(t.options.fences===!1&&e.value&&!e.lang&&/[^ \r\n]/.test(e.value)&&!/^[\t ]*(?:[\r\n]|$)|(?:^|[\r\n])[\t ]*$/.test(e.value))}function yfe(e){const t=e.options.fence||"`";if(t!=="`"&&t!=="~")throw new Error("Cannot serialize code with `"+t+"` for `options.fence`, expected `` ` `` or `~`");return t}function vfe(e,t,n,r){const a=yfe(n),o=e.value||"",s=a==="`"?"GraveAccent":"Tilde";if(bfe(e,n)){const g=n.enter("codeIndented"),m=n.indentLines(o,Sfe);return g(),m}const u=n.createTracker(r),c=a.repeat(Math.max(fH(o,a)+1,3)),d=n.enter("codeFenced");let p=u.move(c);if(e.lang){const g=n.enter(`codeFencedLang${s}`);p+=u.move(n.safe(e.lang,{before:p,after:" ",encode:["`"],...u.current()})),g()}if(e.lang&&e.meta){const g=n.enter(`codeFencedMeta${s}`);p+=u.move(" "),p+=u.move(n.safe(e.meta,{before:p,after:`
`,encode:["`"],...u.current()})),g()}return p+=u.move(`
`),o&&(p+=u.move(o+`
`)),p+=u.move(c),d(),p}function Sfe(e,t,n){return(n?"":"    ")+e}function rA(e){const t=e.options.quote||'"';if(t!=='"'&&t!=="'")throw new Error("Cannot serialize title with `"+t+"` for `options.quote`, expected `\"`, or `'`");return t}function Efe(e,t,n,r){const a=rA(n),o=a==='"'?"Quote":"Apostrophe",s=n.enter("definition");let u=n.enter("label");const c=n.createTracker(r);let d=c.move("[");return d+=c.move(n.safe(n.associationId(e),{before:d,after:"]",...c.current()})),d+=c.move("]: "),u(),!e.url||/[\0- \u007F]/.test(e.url)?(u=n.enter("destinationLiteral"),d+=c.move("<"),d+=c.move(n.safe(e.url,{before:d,after:">",...c.current()})),d+=c.move(">")):(u=n.enter("destinationRaw"),d+=c.move(n.safe(e.url,{before:d,after:e.title?" ":`
`,...c.current()}))),u(),e.title&&(u=n.enter(`title${o}`),d+=c.move(" "+a),d+=c.move(n.safe(e.title,{before:d,after:a,...c.current()})),d+=c.move(a),u()),s(),d}function wfe(e){const t=e.options.emphasis||"*";if(t!=="*"&&t!=="_")throw new Error("Cannot serialize emphasis with `"+t+"` for `options.emphasis`, expected `*`, or `_`");return t}function Pu(e){return"&#x"+e.toString(16).toUpperCase()+";"}function Bf(e,t,n){const r=zs(e),a=zs(t);return r===void 0?a===void 0?n==="_"?{inside:!0,outside:!0}:{inside:!1,outside:!1}:a===1?{inside:!0,outside:!0}:{inside:!1,outside:!0}:r===1?a===void 0?{inside:!1,outside:!1}:a===1?{inside:!0,outside:!0}:{inside:!1,outside:!1}:a===void 0?{inside:!1,outside:!1}:a===1?{inside:!0,outside:!1}:{inside:!1,outside:!1}}pH.peek=xfe;function pH(e,t,n,r){const a=wfe(n),o=n.enter("emphasis"),s=n.createTracker(r),u=s.move(a);let c=s.move(n.containerPhrasing(e,{after:a,before:u,...s.current()}));const d=c.charCodeAt(0),p=Bf(r.before.charCodeAt(r.before.length-1),d,a);p.inside&&(c=Pu(d)+c.slice(1));const g=c.charCodeAt(c.length-1),m=Bf(r.after.charCodeAt(0),g,a);m.inside&&(c=c.slice(0,-1)+Pu(g));const b=s.move(a);return o(),n.attentionEncodeSurroundingInfo={after:m.outside,before:p.outside},u+c+b}function xfe(e,t,n){return n.options.emphasis||"*"}function kfe(e,t){let n=!1;return tA(e,function(r){if("value"in r&&/\r?\n|\r/.test(r.value)||r.type==="break")return n=!0,p0}),!!((!e.depth||e.depth<3)&&YT(e)&&(t.options.setext||n))}function Tfe(e,t,n,r){const a=Math.max(Math.min(6,e.depth||1),1),o=n.createTracker(r);if(kfe(e,n)){const p=n.enter("headingSetext"),g=n.enter("phrasing"),m=n.containerPhrasing(e,{...o.current(),before:`
`,after:`
`});return g(),p(),m+`
`+(a===1?"=":"-").repeat(m.length-(Math.max(m.lastIndexOf("\r"),m.lastIndexOf(`
`))+1))}const s="#".repeat(a),u=n.enter("headingAtx"),c=n.enter("phrasing");o.move(s+" ");let d=n.containerPhrasing(e,{before:"# ",after:`
`,...o.current()});return/^[\t ]/.test(d)&&(d=Pu(d.charCodeAt(0))+d.slice(1)),d=d?s+" "+d:s,n.options.closeAtx&&(d+=" "+s),c(),u(),d}gH.peek=Afe;function gH(e){return e.value||""}function Afe(){return"<"}hH.peek=Rfe;function hH(e,t,n,r){const a=rA(n),o=a==='"'?"Quote":"Apostrophe",s=n.enter("image");let u=n.enter("label");const c=n.createTracker(r);let d=c.move("![");return d+=c.move(n.safe(e.alt,{before:d,after:"]",...c.current()})),d+=c.move("]("),u(),!e.url&&e.title||/[\0- \u007F]/.test(e.url)?(u=n.enter("destinationLiteral"),d+=c.move("<"),d+=c.move(n.safe(e.url,{before:d,after:">",...c.current()})),d+=c.move(">")):(u=n.enter("destinationRaw"),d+=c.move(n.safe(e.url,{before:d,after:e.title?" ":")",...c.current()}))),u(),e.title&&(u=n.enter(`title${o}`),d+=c.move(" "+a),d+=c.move(n.safe(e.title,{before:d,after:a,...c.current()})),d+=c.move(a),u()),d+=c.move(")"),s(),d}function Rfe(){return"!"}mH.peek=_fe;function mH(e,t,n,r){const a=e.referenceType,o=n.enter("imageReference");let s=n.enter("label");const u=n.createTracker(r);let c=u.move("![");const d=n.safe(e.alt,{before:c,after:"]",...u.current()});c+=u.move(d+"]["),s();const p=n.stack;n.stack=[],s=n.enter("reference");const g=n.safe(n.associationId(e),{before:c,after:"]",...u.current()});return s(),n.stack=p,o(),a==="full"||!d||d!==g?c+=u.move(g+"]"):a==="shortcut"?c=c.slice(0,-1):c+=u.move("]"),c}function _fe(){return"!"}bH.peek=Cfe;function bH(e,t,n){let r=e.value||"",a="`",o=-1;for(;new RegExp("(^|[^`])"+a+"([^`]|$)").test(r);)a+="`";for(/[^ \r\n]/.test(r)&&(/^[ \r\n]/.test(r)&&/[ \r\n]$/.test(r)||/^`|`$/.test(r))&&(r=" "+r+" ");++o<n.unsafe.length;){const s=n.unsafe[o],u=n.compilePattern(s);let c;if(s.atBreak)for(;c=u.exec(r);){let d=c.index;r.charCodeAt(d)===10&&r.charCodeAt(d-1)===13&&d--,r=r.slice(0,d)+" "+r.slice(c.index+1)}}return a+r+a}function Cfe(){return"`"}function yH(e,t){const n=YT(e);return!!(!t.options.resourceLink&&e.url&&!e.title&&e.children&&e.children.length===1&&e.children[0].type==="text"&&(n===e.url||"mailto:"+n===e.url)&&/^[a-z][a-z+.-]+:/i.test(e.url)&&!/[\0- <>\u007F]/.test(e.url))}vH.peek=Nfe;function vH(e,t,n,r){const a=rA(n),o=a==='"'?"Quote":"Apostrophe",s=n.createTracker(r);let u,c;if(yH(e,n)){const p=n.stack;n.stack=[],u=n.enter("autolink");let g=s.move("<");return g+=s.move(n.containerPhrasing(e,{before:g,after:">",...s.current()})),g+=s.move(">"),u(),n.stack=p,g}u=n.enter("link"),c=n.enter("label");let d=s.move("[");return d+=s.move(n.containerPhrasing(e,{before:d,after:"](",...s.current()})),d+=s.move("]("),c(),!e.url&&e.title||/[\0- \u007F]/.test(e.url)?(c=n.enter("destinationLiteral"),d+=s.move("<"),d+=s.move(n.safe(e.url,{before:d,after:">",...s.current()})),d+=s.move(">")):(c=n.enter("destinationRaw"),d+=s.move(n.safe(e.url,{before:d,after:e.title?" ":")",...s.current()}))),c(),e.title&&(c=n.enter(`title${o}`),d+=s.move(" "+a),d+=s.move(n.safe(e.title,{before:d,after:a,...s.current()})),d+=s.move(a),c()),d+=s.move(")"),u(),d}function Nfe(e,t,n){return yH(e,n)?"<":"["}SH.peek=Ofe;function SH(e,t,n,r){const a=e.referenceType,o=n.enter("linkReference");let s=n.enter("label");const u=n.createTracker(r);let c=u.move("[");const d=n.containerPhrasing(e,{before:c,after:"]",...u.current()});c+=u.move(d+"]["),s();const p=n.stack;n.stack=[],s=n.enter("reference");const g=n.safe(n.associationId(e),{before:c,after:"]",...u.current()});return s(),n.stack=p,o(),a==="full"||!d||d!==g?c+=u.move(g+"]"):a==="shortcut"?c=c.slice(0,-1):c+=u.move("]"),c}function Ofe(){return"["}function aA(e){const t=e.options.bullet||"*";if(t!=="*"&&t!=="+"&&t!=="-")throw new Error("Cannot serialize items with `"+t+"` for `options.bullet`, expected `*`, `+`, or `-`");return t}function Dfe(e){const t=aA(e),n=e.options.bulletOther;if(!n)return t==="*"?"-":"*";if(n!=="*"&&n!=="+"&&n!=="-")throw new Error("Cannot serialize items with `"+n+"` for `options.bulletOther`, expected `*`, `+`, or `-`");if(n===t)throw new Error("Expected `bullet` (`"+t+"`) and `bulletOther` (`"+n+"`) to be different");return n}function Ife(e){const t=e.options.bulletOrdered||".";if(t!=="."&&t!==")")throw new Error("Cannot serialize items with `"+t+"` for `options.bulletOrdered`, expected `.` or `)`");return t}function EH(e){const t=e.options.rule||"*";if(t!=="*"&&t!=="-"&&t!=="_")throw new Error("Cannot serialize rules with `"+t+"` for `options.rule`, expected `*`, `-`, or `_`");return t}function Lfe(e,t,n,r){const a=n.enter("list"),o=n.bulletCurrent;let s=e.ordered?Ife(n):aA(n);const u=e.ordered?s==="."?")":".":Dfe(n);let c=t&&n.bulletLastUsed?s===n.bulletLastUsed:!1;if(!e.ordered){const p=e.children?e.children[0]:void 0;if((s==="*"||s==="-")&&p&&(!p.children||!p.children[0])&&n.stack[n.stack.length-1]==="list"&&n.stack[n.stack.length-2]==="listItem"&&n.stack[n.stack.length-3]==="list"&&n.stack[n.stack.length-4]==="listItem"&&n.indexStack[n.indexStack.length-1]===0&&n.indexStack[n.indexStack.length-2]===0&&n.indexStack[n.indexStack.length-3]===0&&(c=!0),EH(n)===s&&p){let g=-1;for(;++g<e.children.length;){const m=e.children[g];if(m&&m.type==="listItem"&&m.children&&m.children[0]&&m.children[0].type==="thematicBreak"){c=!0;break}}}}c&&(s=u),n.bulletCurrent=s;const d=n.containerFlow(e,r);return n.bulletLastUsed=s,n.bulletCurrent=o,a(),d}function Mfe(e){const t=e.options.listItemIndent||"one";if(t!=="tab"&&t!=="one"&&t!=="mixed")throw new Error("Cannot serialize items with `"+t+"` for `options.listItemIndent`, expected `tab`, `one`, or `mixed`");return t}function Ffe(e,t,n,r){const a=Mfe(n);let o=n.bulletCurrent||aA(n);t&&t.type==="list"&&t.ordered&&(o=(typeof t.start=="number"&&t.start>-1?t.start:1)+(n.options.incrementListMarker===!1?0:t.children.indexOf(e))+o);let s=o.length+1;(a==="tab"||a==="mixed"&&(t&&t.type==="list"&&t.spread||e.spread))&&(s=Math.ceil(s/4)*4);const u=n.createTracker(r);u.move(o+" ".repeat(s-o.length)),u.shift(s);const c=n.enter("listItem"),d=n.indentLines(n.containerFlow(e,u.current()),p);return c(),d;function p(g,m,b){return m?(b?"":" ".repeat(s))+g:(b?o:o+" ".repeat(s-o.length))+g}}function Pfe(e,t,n,r){const a=n.enter("paragraph"),o=n.enter("phrasing"),s=n.containerPhrasing(e,r);return o(),a(),s}const zfe=Ip(["break","delete","emphasis","footnote","footnoteReference","image","imageReference","inlineCode","inlineMath","link","linkReference","mdxJsxTextElement","mdxTextExpression","strong","text","textDirective"]);function Bfe(e,t,n,r){return(e.children.some(function(s){return zfe(s)})?n.containerPhrasing:n.containerFlow).call(n,e,r)}function Ufe(e){const t=e.options.strong||"*";if(t!=="*"&&t!=="_")throw new Error("Cannot serialize strong with `"+t+"` for `options.strong`, expected `*`, or `_`");return t}wH.peek=jfe;function wH(e,t,n,r){const a=Ufe(n),o=n.enter("strong"),s=n.createTracker(r),u=s.move(a+a);let c=s.move(n.containerPhrasing(e,{after:a,before:u,...s.current()}));const d=c.charCodeAt(0),p=Bf(r.before.charCodeAt(r.before.length-1),d,a);p.inside&&(c=Pu(d)+c.slice(1));const g=c.charCodeAt(c.length-1),m=Bf(r.after.charCodeAt(0),g,a);m.inside&&(c=c.slice(0,-1)+Pu(g));const b=s.move(a+a);return o(),n.attentionEncodeSurroundingInfo={after:m.outside,before:p.outside},u+c+b}function jfe(e,t,n){return n.options.strong||"*"}function Gfe(e,t,n,r){return n.safe(e.value,r)}function Hfe(e){const t=e.options.ruleRepetition||3;if(t<3)throw new Error("Cannot serialize rules with repetition `"+t+"` for `options.ruleRepetition`, expected `3` or more");return t}function $fe(e,t,n){const r=(EH(n)+(n.options.ruleSpaces?" ":"")).repeat(Hfe(n));return n.options.ruleSpaces?r.slice(0,-1):r}const xH={blockquote:gfe,break:zD,code:vfe,definition:Efe,emphasis:pH,hardBreak:zD,heading:Tfe,html:gH,image:hH,imageReference:mH,inlineCode:bH,link:vH,linkReference:SH,list:Lfe,listItem:Ffe,paragraph:Pfe,root:Bfe,strong:wH,text:Gfe,thematicBreak:$fe};function qfe(){return{enter:{table:Vfe,tableData:BD,tableHeader:BD,tableRow:Yfe},exit:{codeText:Kfe,table:Wfe,tableData:Ab,tableHeader:Ab,tableRow:Ab}}}function Vfe(e){const t=e._align;this.enter({type:"table",align:t.map(function(n){return n==="none"?null:n}),children:[]},e),this.data.inTable=!0}function Wfe(e){this.exit(e),this.data.inTable=void 0}function Yfe(e){this.enter({type:"tableRow",children:[]},e)}function Ab(e){this.exit(e)}function BD(e){this.enter({type:"tableCell",children:[]},e)}function Kfe(e){let t=this.resume();this.data.inTable&&(t=t.replace(/\\([\\|])/g,Xfe));const n=this.stack[this.stack.length-1];n.type,n.value=t,this.exit(e)}function Xfe(e,t){return t==="|"?t:e}function Zfe(e){const t=e||{},n=t.tableCellPadding,r=t.tablePipeAlign,a=t.stringLength,o=n?" ":"|";return{unsafe:[{character:"\r",inConstruct:"tableCell"},{character:`
`,inConstruct:"tableCell"},{atBreak:!0,character:"|",after:"[	 :-]"},{character:"|",inConstruct:"tableCell"},{atBreak:!0,character:":",after:"-"},{atBreak:!0,character:"-",after:"[:|-]"}],handlers:{inlineCode:m,table:s,tableCell:c,tableRow:u}};function s(b,y,S,x){return d(p(b,S,x),b.align)}function u(b,y,S,x){const R=g(b,S,x),k=d([R]);return k.slice(0,k.indexOf(`
`))}function c(b,y,S,x){const R=S.enter("tableCell"),k=S.enter("phrasing"),A=S.containerPhrasing(b,{...x,before:o,after:o});return k(),R(),A}function d(b,y){return ffe(b,{align:y,alignDelimiters:r,padding:n,stringLength:a})}function p(b,y,S){const x=b.children;let R=-1;const k=[],A=y.enter("table");for(;++R<x.length;)k[R]=g(x[R],y,S);return A(),k}function g(b,y,S){const x=b.children;let R=-1;const k=[],A=y.enter("tableRow");for(;++R<x.length;)k[R]=c(x[R],b,y,S);return A(),k}function m(b,y,S){let x=xH.inlineCode(b,y,S);return S.stack.includes("tableCell")&&(x=x.replace(/\|/g,"\\$&")),x}}function Qfe(){return{exit:{taskListCheckValueChecked:UD,taskListCheckValueUnchecked:UD,paragraph:epe}}}function Jfe(){return{unsafe:[{atBreak:!0,character:"-",after:"[:|-]"}],handlers:{listItem:tpe}}}function UD(e){const t=this.stack[this.stack.length-2];t.type,t.checked=e.type==="taskListCheckValueChecked"}function epe(e){const t=this.stack[this.stack.length-2];if(t&&t.type==="listItem"&&typeof t.checked=="boolean"){const n=this.stack[this.stack.length-1];n.type;const r=n.children[0];if(r&&r.type==="text"){const a=t.children;let o=-1,s;for(;++o<a.length;){const u=a[o];if(u.type==="paragraph"){s=u;break}}s===n&&(r.value=r.value.slice(1),r.value.length===0?n.children.shift():n.position&&r.position&&typeof r.position.start.offset=="number"&&(r.position.start.column++,r.position.start.offset++,n.position.start=Object.assign({},r.position.start)))}}this.exit(e)}function tpe(e,t,n,r){const a=e.children[0],o=typeof e.checked=="boolean"&&a&&a.type==="paragraph",s="["+(e.checked?"x":" ")+"] ",u=n.createTracker(r);o&&u.move(s);let c=xH.listItem(e,t,n,{...r,...u.current()});return o&&(c=c.replace(/^(?:[*+-]|\d+\.)([\r\n]| {1,3})/,d)),c;function d(p){return p+s}}function npe(){return[Mde(),nfe(),ife(),qfe(),Qfe()]}function rpe(e){return{extensions:[Fde(),rfe(e),sfe(),Zfe(e),Jfe()]}}const ape={tokenize:cpe,partial:!0},kH={tokenize:dpe,partial:!0},TH={tokenize:fpe,partial:!0},AH={tokenize:ppe,partial:!0},ope={tokenize:gpe,partial:!0},RH={name:"wwwAutolink",tokenize:lpe,previous:CH},_H={name:"protocolAutolink",tokenize:upe,previous:NH},Ya={name:"emailAutolink",tokenize:spe,previous:OH},da={};function ipe(){return{text:da}}let si=48;for(;si<123;)da[si]=Ya,si++,si===58?si=65:si===91&&(si=97);da[43]=Ya;da[45]=Ya;da[46]=Ya;da[95]=Ya;da[72]=[Ya,_H];da[104]=[Ya,_H];da[87]=[Ya,RH];da[119]=[Ya,RH];function spe(e,t,n){const r=this;let a,o;return s;function s(g){return!b0(g)||!OH.call(r,r.previous)||oA(r.events)?n(g):(e.enter("literalAutolink"),e.enter("literalAutolinkEmail"),u(g))}function u(g){return b0(g)?(e.consume(g),u):g===64?(e.consume(g),c):n(g)}function c(g){return g===46?e.check(ope,p,d)(g):g===45||g===95||kn(g)?(o=!0,e.consume(g),c):p(g)}function d(g){return e.consume(g),a=!0,c}function p(g){return o&&a&&On(r.previous)?(e.exit("literalAutolinkEmail"),e.exit("literalAutolink"),t(g)):n(g)}}function lpe(e,t,n){const r=this;return a;function a(s){return s!==87&&s!==119||!CH.call(r,r.previous)||oA(r.events)?n(s):(e.enter("literalAutolink"),e.enter("literalAutolinkWww"),e.check(ape,e.attempt(kH,e.attempt(TH,o),n),n)(s))}function o(s){return e.exit("literalAutolinkWww"),e.exit("literalAutolink"),t(s)}}function upe(e,t,n){const r=this;let a="",o=!1;return s;function s(g){return(g===72||g===104)&&NH.call(r,r.previous)&&!oA(r.events)?(e.enter("literalAutolink"),e.enter("literalAutolinkHttp"),a+=String.fromCodePoint(g),e.consume(g),u):n(g)}function u(g){if(On(g)&&a.length<5)return a+=String.fromCodePoint(g),e.consume(g),u;if(g===58){const m=a.toLowerCase();if(m==="http"||m==="https")return e.consume(g),c}return n(g)}function c(g){return g===47?(e.consume(g),o?d:(o=!0,c)):n(g)}function d(g){return g===null||Ff(g)||Dt(g)||Si(g)||Np(g)?n(g):e.attempt(kH,e.attempt(TH,p),n)(g)}function p(g){return e.exit("literalAutolinkHttp"),e.exit("literalAutolink"),t(g)}}function cpe(e,t,n){let r=0;return a;function a(s){return(s===87||s===119)&&r<3?(r++,e.consume(s),a):s===46&&r===3?(e.consume(s),o):n(s)}function o(s){return s===null?n(s):t(s)}}function dpe(e,t,n){let r,a,o;return s;function s(d){return d===46||d===95?e.check(AH,c,u)(d):d===null||Dt(d)||Si(d)||d!==45&&Np(d)?c(d):(o=!0,e.consume(d),s)}function u(d){return d===95?r=!0:(a=r,r=void 0),e.consume(d),s}function c(d){return a||r||!o?n(d):t(d)}}function fpe(e,t){let n=0,r=0;return a;function a(s){return s===40?(n++,e.consume(s),a):s===41&&r<n?o(s):s===33||s===34||s===38||s===39||s===41||s===42||s===44||s===46||s===58||s===59||s===60||s===63||s===93||s===95||s===126?e.check(AH,t,o)(s):s===null||Dt(s)||Si(s)?t(s):(e.consume(s),a)}function o(s){return s===41&&r++,e.consume(s),a}}function ppe(e,t,n){return r;function r(u){return u===33||u===34||u===39||u===41||u===42||u===44||u===46||u===58||u===59||u===63||u===95||u===126?(e.consume(u),r):u===38?(e.consume(u),o):u===93?(e.consume(u),a):u===60||u===null||Dt(u)||Si(u)?t(u):n(u)}function a(u){return u===null||u===40||u===91||Dt(u)||Si(u)?t(u):r(u)}function o(u){return On(u)?s(u):n(u)}function s(u){return u===59?(e.consume(u),r):On(u)?(e.consume(u),s):n(u)}}function gpe(e,t,n){return r;function r(o){return e.consume(o),a}function a(o){return kn(o)?n(o):t(o)}}function CH(e){return e===null||e===40||e===42||e===95||e===91||e===93||e===126||Dt(e)}function NH(e){return!On(e)}function OH(e){return!(e===47||b0(e))}function b0(e){return e===43||e===45||e===46||e===95||kn(e)}function oA(e){let t=e.length,n=!1;for(;t--;){const r=e[t][1];if((r.type==="labelLink"||r.type==="labelImage")&&!r._balanced){n=!0;break}if(r._gfmAutolinkLiteralWalkedInto){n=!1;break}}return e.length>0&&!n&&(e[e.length-1][1]._gfmAutolinkLiteralWalkedInto=!0),n}const hpe={tokenize:xpe,partial:!0};function mpe(){return{document:{91:{name:"gfmFootnoteDefinition",tokenize:Spe,continuation:{tokenize:Epe},exit:wpe}},text:{91:{name:"gfmFootnoteCall",tokenize:vpe},93:{name:"gfmPotentialFootnoteCall",add:"after",tokenize:bpe,resolveTo:ype}}}}function bpe(e,t,n){const r=this;let a=r.events.length;const o=r.parser.gfmFootnotes||(r.parser.gfmFootnotes=[]);let s;for(;a--;){const c=r.events[a][1];if(c.type==="labelImage"){s=c;break}if(c.type==="gfmFootnoteCall"||c.type==="labelLink"||c.type==="label"||c.type==="image"||c.type==="link")break}return u;function u(c){if(!s||!s._balanced)return n(c);const d=zr(r.sliceSerialize({start:s.end,end:r.now()}));return d.codePointAt(0)!==94||!o.includes(d.slice(1))?n(c):(e.enter("gfmFootnoteCallLabelMarker"),e.consume(c),e.exit("gfmFootnoteCallLabelMarker"),t(c))}}function ype(e,t){let n=e.length;for(;n--;)if(e[n][1].type==="labelImage"&&e[n][0]==="enter"){e[n][1];break}e[n+1][1].type="data",e[n+3][1].type="gfmFootnoteCallLabelMarker";const r={type:"gfmFootnoteCall",start:Object.assign({},e[n+3][1].start),end:Object.assign({},e[e.length-1][1].end)},a={type:"gfmFootnoteCallMarker",start:Object.assign({},e[n+3][1].end),end:Object.assign({},e[n+3][1].end)};a.end.column++,a.end.offset++,a.end._bufferIndex++;const o={type:"gfmFootnoteCallString",start:Object.assign({},a.end),end:Object.assign({},e[e.length-1][1].start)},s={type:"chunkString",contentType:"string",start:Object.assign({},o.start),end:Object.assign({},o.end)},u=[e[n+1],e[n+2],["enter",r,t],e[n+3],e[n+4],["enter",a,t],["exit",a,t],["enter",o,t],["enter",s,t],["exit",s,t],["exit",o,t],e[e.length-2],e[e.length-1],["exit",r,t]];return e.splice(n,e.length-n+1,...u),e}function vpe(e,t,n){const r=this,a=r.parser.gfmFootnotes||(r.parser.gfmFootnotes=[]);let o=0,s;return u;function u(g){return e.enter("gfmFootnoteCall"),e.enter("gfmFootnoteCallLabelMarker"),e.consume(g),e.exit("gfmFootnoteCallLabelMarker"),c}function c(g){return g!==94?n(g):(e.enter("gfmFootnoteCallMarker"),e.consume(g),e.exit("gfmFootnoteCallMarker"),e.enter("gfmFootnoteCallString"),e.enter("chunkString").contentType="string",d)}function d(g){if(o>999||g===93&&!s||g===null||g===91||Dt(g))return n(g);if(g===93){e.exit("chunkString");const m=e.exit("gfmFootnoteCallString");return a.includes(zr(r.sliceSerialize(m)))?(e.enter("gfmFootnoteCallLabelMarker"),e.consume(g),e.exit("gfmFootnoteCallLabelMarker"),e.exit("gfmFootnoteCall"),t):n(g)}return Dt(g)||(s=!0),o++,e.consume(g),g===92?p:d}function p(g){return g===91||g===92||g===93?(e.consume(g),o++,d):d(g)}}function Spe(e,t,n){const r=this,a=r.parser.gfmFootnotes||(r.parser.gfmFootnotes=[]);let o,s=0,u;return c;function c(y){return e.enter("gfmFootnoteDefinition")._container=!0,e.enter("gfmFootnoteDefinitionLabel"),e.enter("gfmFootnoteDefinitionLabelMarker"),e.consume(y),e.exit("gfmFootnoteDefinitionLabelMarker"),d}function d(y){return y===94?(e.enter("gfmFootnoteDefinitionMarker"),e.consume(y),e.exit("gfmFootnoteDefinitionMarker"),e.enter("gfmFootnoteDefinitionLabelString"),e.enter("chunkString").contentType="string",p):n(y)}function p(y){if(s>999||y===93&&!u||y===null||y===91||Dt(y))return n(y);if(y===93){e.exit("chunkString");const S=e.exit("gfmFootnoteDefinitionLabelString");return o=zr(r.sliceSerialize(S)),e.enter("gfmFootnoteDefinitionLabelMarker"),e.consume(y),e.exit("gfmFootnoteDefinitionLabelMarker"),e.exit("gfmFootnoteDefinitionLabel"),m}return Dt(y)||(u=!0),s++,e.consume(y),y===92?g:p}function g(y){return y===91||y===92||y===93?(e.consume(y),s++,p):p(y)}function m(y){return y===58?(e.enter("definitionMarker"),e.consume(y),e.exit("definitionMarker"),a.includes(o)||a.push(o),ut(e,b,"gfmFootnoteDefinitionWhitespace")):n(y)}function b(y){return t(y)}}function Epe(e,t,n){return e.check(ac,t,e.attempt(hpe,t,n))}function wpe(e){e.exit("gfmFootnoteDefinition")}function xpe(e,t,n){const r=this;return ut(e,a,"gfmFootnoteDefinitionIndent",5);function a(o){const s=r.events[r.events.length-1];return s&&s[1].type==="gfmFootnoteDefinitionIndent"&&s[2].sliceSerialize(s[1],!0).length===4?t(o):n(o)}}function kpe(e){let n=(e||{}).singleTilde;const r={name:"strikethrough",tokenize:o,resolveAll:a};return n==null&&(n=!0),{text:{126:r},insideSpan:{null:[r]},attentionMarkers:{null:[126]}};function a(s,u){let c=-1;for(;++c<s.length;)if(s[c][0]==="enter"&&s[c][1].type==="strikethroughSequenceTemporary"&&s[c][1]._close){let d=c;for(;d--;)if(s[d][0]==="exit"&&s[d][1].type==="strikethroughSequenceTemporary"&&s[d][1]._open&&s[c][1].end.offset-s[c][1].start.offset===s[d][1].end.offset-s[d][1].start.offset){s[c][1].type="strikethroughSequence",s[d][1].type="strikethroughSequence";const p={type:"strikethrough",start:Object.assign({},s[d][1].start),end:Object.assign({},s[c][1].end)},g={type:"strikethroughText",start:Object.assign({},s[d][1].end),end:Object.assign({},s[c][1].start)},m=[["enter",p,u],["enter",s[d][1],u],["exit",s[d][1],u],["enter",g,u]],b=u.parser.constructs.insideSpan.null;b&&rr(m,m.length,0,Op(b,s.slice(d+1,c),u)),rr(m,m.length,0,[["exit",g,u],["enter",s[c][1],u],["exit",s[c][1],u],["exit",p,u]]),rr(s,d-1,c-d+3,m),c=d+m.length-2;break}}for(c=-1;++c<s.length;)s[c][1].type==="strikethroughSequenceTemporary"&&(s[c][1].type="data");return s}function o(s,u,c){const d=this.previous,p=this.events;let g=0;return m;function m(y){return d===126&&p[p.length-1][1].type!=="characterEscape"?c(y):(s.enter("strikethroughSequenceTemporary"),b(y))}function b(y){const S=zs(d);if(y===126)return g>1?c(y):(s.consume(y),g++,b);if(g<2&&!n)return c(y);const x=s.exit("strikethroughSequenceTemporary"),R=zs(y);return x._open=!R||R===2&&!!S,x._close=!S||S===2&&!!R,u(y)}}}class Tpe{constructor(){this.map=[]}add(t,n,r){Ape(this,t,n,r)}consume(t){if(this.map.sort(function(o,s){return o[0]-s[0]}),this.map.length===0)return;let n=this.map.length;const r=[];for(;n>0;)n-=1,r.push(t.slice(this.map[n][0]+this.map[n][1]),this.map[n][2]),t.length=this.map[n][0];r.push(t.slice()),t.length=0;let a=r.pop();for(;a;){for(const o of a)t.push(o);a=r.pop()}this.map.length=0}}function Ape(e,t,n,r){let a=0;if(!(n===0&&r.length===0)){for(;a<e.map.length;){if(e.map[a][0]===t){e.map[a][1]+=n,e.map[a][2].push(...r);return}a+=1}e.map.push([t,n,r])}}function Rpe(e,t){let n=!1;const r=[];for(;t<e.length;){const a=e[t];if(n){if(a[0]==="enter")a[1].type==="tableContent"&&r.push(e[t+1][1].type==="tableDelimiterMarker"?"left":"none");else if(a[1].type==="tableContent"){if(e[t-1][1].type==="tableDelimiterMarker"){const o=r.length-1;r[o]=r[o]==="left"?"center":"right"}}else if(a[1].type==="tableDelimiterRow")break}else a[0]==="enter"&&a[1].type==="tableDelimiterRow"&&(n=!0);t+=1}return r}function _pe(){return{flow:{null:{name:"table",tokenize:Cpe,resolveAll:Npe}}}}function Cpe(e,t,n){const r=this;let a=0,o=0,s;return u;function u(I){let H=r.events.length-1;for(;H>-1;){const Y=r.events[H][1].type;if(Y==="lineEnding"||Y==="linePrefix")H--;else break}const $=H>-1?r.events[H][1].type:null,U=$==="tableHead"||$==="tableRow"?_:c;return U===_&&r.parser.lazy[r.now().line]?n(I):U(I)}function c(I){return e.enter("tableHead"),e.enter("tableRow"),d(I)}function d(I){return I===124||(s=!0,o+=1),p(I)}function p(I){return I===null?n(I):Ve(I)?o>1?(o=0,r.interrupt=!0,e.exit("tableRow"),e.enter("lineEnding"),e.consume(I),e.exit("lineEnding"),b):n(I):dt(I)?ut(e,p,"whitespace")(I):(o+=1,s&&(s=!1,a+=1),I===124?(e.enter("tableCellDivider"),e.consume(I),e.exit("tableCellDivider"),s=!0,p):(e.enter("data"),g(I)))}function g(I){return I===null||I===124||Dt(I)?(e.exit("data"),p(I)):(e.consume(I),I===92?m:g)}function m(I){return I===92||I===124?(e.consume(I),g):g(I)}function b(I){return r.interrupt=!1,r.parser.lazy[r.now().line]?n(I):(e.enter("tableDelimiterRow"),s=!1,dt(I)?ut(e,y,"linePrefix",r.parser.constructs.disable.null.includes("codeIndented")?void 0:4)(I):y(I))}function y(I){return I===45||I===58?x(I):I===124?(s=!0,e.enter("tableCellDivider"),e.consume(I),e.exit("tableCellDivider"),S):N(I)}function S(I){return dt(I)?ut(e,x,"whitespace")(I):x(I)}function x(I){return I===58?(o+=1,s=!0,e.enter("tableDelimiterMarker"),e.consume(I),e.exit("tableDelimiterMarker"),R):I===45?(o+=1,R(I)):I===null||Ve(I)?C(I):N(I)}function R(I){return I===45?(e.enter("tableDelimiterFiller"),k(I)):N(I)}function k(I){return I===45?(e.consume(I),k):I===58?(s=!0,e.exit("tableDelimiterFiller"),e.enter("tableDelimiterMarker"),e.consume(I),e.exit("tableDelimiterMarker"),A):(e.exit("tableDelimiterFiller"),A(I))}function A(I){return dt(I)?ut(e,C,"whitespace")(I):C(I)}function C(I){return I===124?y(I):I===null||Ve(I)?!s||a!==o?N(I):(e.exit("tableDelimiterRow"),e.exit("tableHead"),t(I)):N(I)}function N(I){return n(I)}function _(I){return e.enter("tableRow"),O(I)}function O(I){return I===124?(e.enter("tableCellDivider"),e.consume(I),e.exit("tableCellDivider"),O):I===null||Ve(I)?(e.exit("tableRow"),t(I)):dt(I)?ut(e,O,"whitespace")(I):(e.enter("data"),F(I))}function F(I){return I===null||I===124||Dt(I)?(e.exit("data"),O(I)):(e.consume(I),I===92?L:F)}function L(I){return I===92||I===124?(e.consume(I),F):F(I)}}function Npe(e,t){let n=-1,r=!0,a=0,o=[0,0,0,0],s=[0,0,0,0],u=!1,c=0,d,p,g;const m=new Tpe;for(;++n<e.length;){const b=e[n],y=b[1];b[0]==="enter"?y.type==="tableHead"?(u=!1,c!==0&&(jD(m,t,c,d,p),p=void 0,c=0),d={type:"table",start:Object.assign({},y.start),end:Object.assign({},y.end)},m.add(n,0,[["enter",d,t]])):y.type==="tableRow"||y.type==="tableDelimiterRow"?(r=!0,g=void 0,o=[0,0,0,0],s=[0,n+1,0,0],u&&(u=!1,p={type:"tableBody",start:Object.assign({},y.start),end:Object.assign({},y.end)},m.add(n,0,[["enter",p,t]])),a=y.type==="tableDelimiterRow"?2:p?3:1):a&&(y.type==="data"||y.type==="tableDelimiterMarker"||y.type==="tableDelimiterFiller")?(r=!1,s[2]===0&&(o[1]!==0&&(s[0]=s[1],g=jd(m,t,o,a,void 0,g),o=[0,0,0,0]),s[2]=n)):y.type==="tableCellDivider"&&(r?r=!1:(o[1]!==0&&(s[0]=s[1],g=jd(m,t,o,a,void 0,g)),o=s,s=[o[1],n,0,0])):y.type==="tableHead"?(u=!0,c=n):y.type==="tableRow"||y.type==="tableDelimiterRow"?(c=n,o[1]!==0?(s[0]=s[1],g=jd(m,t,o,a,n,g)):s[1]!==0&&(g=jd(m,t,s,a,n,g)),a=0):a&&(y.type==="data"||y.type==="tableDelimiterMarker"||y.type==="tableDelimiterFiller")&&(s[3]=n)}for(c!==0&&jD(m,t,c,d,p),m.consume(t.events),n=-1;++n<t.events.length;){const b=t.events[n];b[0]==="enter"&&b[1].type==="table"&&(b[1]._align=Rpe(t.events,n))}return e}function jd(e,t,n,r,a,o){const s=r===1?"tableHeader":r===2?"tableDelimiter":"tableData",u="tableContent";n[0]!==0&&(o.end=Object.assign({},ws(t.events,n[0])),e.add(n[0],0,[["exit",o,t]]));const c=ws(t.events,n[1]);if(o={type:s,start:Object.assign({},c),end:Object.assign({},c)},e.add(n[1],0,[["enter",o,t]]),n[2]!==0){const d=ws(t.events,n[2]),p=ws(t.events,n[3]),g={type:u,start:Object.assign({},d),end:Object.assign({},p)};if(e.add(n[2],0,[["enter",g,t]]),r!==2){const m=t.events[n[2]],b=t.events[n[3]];if(m[1].end=Object.assign({},b[1].end),m[1].type="chunkText",m[1].contentType="text",n[3]>n[2]+1){const y=n[2]+1,S=n[3]-n[2]-1;e.add(y,S,[])}}e.add(n[3]+1,0,[["exit",g,t]])}return a!==void 0&&(o.end=Object.assign({},ws(t.events,a)),e.add(a,0,[["exit",o,t]]),o=void 0),o}function jD(e,t,n,r,a){const o=[],s=ws(t.events,n);a&&(a.end=Object.assign({},s),o.push(["exit",a,t])),r.end=Object.assign({},s),o.push(["exit",r,t]),e.add(n+1,0,o)}function ws(e,t){const n=e[t],r=n[0]==="enter"?"start":"end";return n[1][r]}const Ope={name:"tasklistCheck",tokenize:Ipe};function Dpe(){return{text:{91:Ope}}}function Ipe(e,t,n){const r=this;return a;function a(c){return r.previous!==null||!r._gfmTasklistFirstContentOfListItem?n(c):(e.enter("taskListCheck"),e.enter("taskListCheckMarker"),e.consume(c),e.exit("taskListCheckMarker"),o)}function o(c){return Dt(c)?(e.enter("taskListCheckValueUnchecked"),e.consume(c),e.exit("taskListCheckValueUnchecked"),s):c===88||c===120?(e.enter("taskListCheckValueChecked"),e.consume(c),e.exit("taskListCheckValueChecked"),s):n(c)}function s(c){return c===93?(e.enter("taskListCheckMarker"),e.consume(c),e.exit("taskListCheckMarker"),e.exit("taskListCheck"),u):n(c)}function u(c){return Ve(c)?t(c):dt(c)?e.check({tokenize:Lpe},t,n)(c):n(c)}}function Lpe(e,t,n){return ut(e,r,"whitespace");function r(a){return a===null?n(a):t(a)}}function Mpe(e){return jG([ipe(),mpe(),kpe(e),_pe(),Dpe()])}const Fpe={};function Ppe(e){const t=this,n=e||Fpe,r=t.data(),a=r.micromarkExtensions||(r.micromarkExtensions=[]),o=r.fromMarkdownExtensions||(r.fromMarkdownExtensions=[]),s=r.toMarkdownExtensions||(r.toMarkdownExtensions=[]);a.push(Mpe(n)),o.push(npe()),s.push(rpe(n))}function zpe(e){const t=this;t.compiler=n;function n(r,a){return FG(r,{filePath:a.path,...e})}}function Bpe(){return{enter:{mathFlow:e,mathFlowFenceMeta:t,mathText:o},exit:{mathFlow:a,mathFlowFence:r,mathFlowFenceMeta:n,mathFlowValue:u,mathText:s,mathTextData:u}};function e(c){const d={type:"element",tagName:"code",properties:{className:["language-math","math-display"]},children:[]};this.enter({type:"math",meta:null,value:"",data:{hName:"pre",hChildren:[d]}},c)}function t(){this.buffer()}function n(){const c=this.resume(),d=this.stack[this.stack.length-1];d.type,d.meta=c}function r(){this.data.mathFlowInside||(this.buffer(),this.data.mathFlowInside=!0)}function a(c){const d=this.resume().replace(/^(\r?\n|\r)|(\r?\n|\r)$/g,""),p=this.stack[this.stack.length-1];p.type,this.exit(c),p.value=d;const g=p.data.hChildren[0];g.type,g.tagName,g.children.push({type:"text",value:d}),this.data.mathFlowInside=void 0}function o(c){this.enter({type:"inlineMath",value:"",data:{hName:"code",hProperties:{className:["language-math","math-inline"]},hChildren:[]}},c),this.buffer()}function s(c){const d=this.resume(),p=this.stack[this.stack.length-1];p.type,this.exit(c),p.value=d,p.data.hChildren.push({type:"text",value:d})}function u(c){this.config.enter.data.call(this,c),this.config.exit.data.call(this,c)}}function Upe(e){let t=(e||{}).singleDollarTextMath;return t==null&&(t=!0),r.peek=a,{unsafe:[{character:"\r",inConstruct:"mathFlowMeta"},{character:`
`,inConstruct:"mathFlowMeta"},{character:"$",after:t?void 0:"\\$",inConstruct:"phrasing"},{character:"$",inConstruct:"mathFlowMeta"},{atBreak:!0,character:"$",after:"\\$"}],handlers:{math:n,inlineMath:r}};function n(o,s,u,c){const d=o.value||"",p=u.createTracker(c),g="$".repeat(Math.max(fH(d,"$")+1,2)),m=u.enter("mathFlow");let b=p.move(g);if(o.meta){const y=u.enter("mathFlowMeta");b+=p.move(u.safe(o.meta,{after:`
`,before:b,encode:["$"],...p.current()})),y()}return b+=p.move(`
`),d&&(b+=p.move(d+`
`)),b+=p.move(g),m(),b}function r(o,s,u){let c=o.value||"",d=1;for(t||d++;new RegExp("(^|[^$])"+"\\$".repeat(d)+"([^$]|$)").test(c);)d++;const p="$".repeat(d);/[^ \r\n]/.test(c)&&(/^[ \r\n]/.test(c)&&/[ \r\n]$/.test(c)||/^\$|\$$/.test(c))&&(c=" "+c+" ");let g=-1;for(;++g<u.unsafe.length;){const m=u.unsafe[g];if(!m.atBreak)continue;const b=u.compilePattern(m);let y;for(;y=b.exec(c);){let S=y.index;c.codePointAt(S)===10&&c.codePointAt(S-1)===13&&S--,c=c.slice(0,S)+" "+c.slice(y.index+1)}}return p+c+p}function a(){return"$"}}const jpe={tokenize:Gpe,concrete:!0,name:"mathFlow"},GD={tokenize:Hpe,partial:!0};function Gpe(e,t,n){const r=this,a=r.events[r.events.length-1],o=a&&a[1].type==="linePrefix"?a[2].sliceSerialize(a[1],!0).length:0;let s=0;return u;function u(k){return e.enter("mathFlow"),e.enter("mathFlowFence"),e.enter("mathFlowFenceSequence"),c(k)}function c(k){return k===36?(e.consume(k),s++,c):s<2?n(k):(e.exit("mathFlowFenceSequence"),ut(e,d,"whitespace")(k))}function d(k){return k===null||Ve(k)?g(k):(e.enter("mathFlowFenceMeta"),e.enter("chunkString",{contentType:"string"}),p(k))}function p(k){return k===null||Ve(k)?(e.exit("chunkString"),e.exit("mathFlowFenceMeta"),g(k)):k===36?n(k):(e.consume(k),p)}function g(k){return e.exit("mathFlowFence"),r.interrupt?t(k):e.attempt(GD,m,x)(k)}function m(k){return e.attempt({tokenize:R,partial:!0},x,b)(k)}function b(k){return(o?ut(e,y,"linePrefix",o+1):y)(k)}function y(k){return k===null?x(k):Ve(k)?e.attempt(GD,m,x)(k):(e.enter("mathFlowValue"),S(k))}function S(k){return k===null||Ve(k)?(e.exit("mathFlowValue"),y(k)):(e.consume(k),S)}function x(k){return e.exit("mathFlow"),t(k)}function R(k,A,C){let N=0;return ut(k,_,"linePrefix",r.parser.constructs.disable.null.includes("codeIndented")?void 0:4);function _(L){return k.enter("mathFlowFence"),k.enter("mathFlowFenceSequence"),O(L)}function O(L){return L===36?(N++,k.consume(L),O):N<s?C(L):(k.exit("mathFlowFenceSequence"),ut(k,F,"whitespace")(L))}function F(L){return L===null||Ve(L)?(k.exit("mathFlowFence"),A(L)):C(L)}}}function Hpe(e,t,n){const r=this;return a;function a(s){return s===null?t(s):(e.enter("lineEnding"),e.consume(s),e.exit("lineEnding"),o)}function o(s){return r.parser.lazy[r.now().line]?n(s):t(s)}}function $pe(e){let n=(e||{}).singleDollarTextMath;return n==null&&(n=!0),{tokenize:r,resolve:qpe,previous:Vpe,name:"mathText"};function r(a,o,s){let u=0,c,d;return p;function p(S){return a.enter("mathText"),a.enter("mathTextSequence"),g(S)}function g(S){return S===36?(a.consume(S),u++,g):u<2&&!n?s(S):(a.exit("mathTextSequence"),m(S))}function m(S){return S===null?s(S):S===36?(d=a.enter("mathTextSequence"),c=0,y(S)):S===32?(a.enter("space"),a.consume(S),a.exit("space"),m):Ve(S)?(a.enter("lineEnding"),a.consume(S),a.exit("lineEnding"),m):(a.enter("mathTextData"),b(S))}function b(S){return S===null||S===32||S===36||Ve(S)?(a.exit("mathTextData"),m(S)):(a.consume(S),b)}function y(S){return S===36?(a.consume(S),c++,y):c===u?(a.exit("mathTextSequence"),a.exit("mathText"),o(S)):(d.type="mathTextData",b(S))}}}function qpe(e){let t=e.length-4,n=3,r,a;if((e[n][1].type==="lineEnding"||e[n][1].type==="space")&&(e[t][1].type==="lineEnding"||e[t][1].type==="space")){for(r=n;++r<t;)if(e[r][1].type==="mathTextData"){e[t][1].type="mathTextPadding",e[n][1].type="mathTextPadding",n+=2,t-=2;break}}for(r=n-1,t++;++r<=t;)a===void 0?r!==t&&e[r][1].type!=="lineEnding"&&(a=r):(r===t||e[r][1].type==="lineEnding")&&(e[a][1].type="mathTextData",r!==a+2&&(e[a][1].end=e[r-1][1].end,e.splice(a+2,r-a-2),t-=r-a-2,r=a+2),a=void 0);return e}function Vpe(e){return e!==36||this.events[this.events.length-1][1].type==="characterEscape"}function Wpe(e){return{flow:{36:jpe},text:{36:$pe(e)}}}const Ype={};function Kpe(e){const t=this,n=e||Ype,r=t.data(),a=r.micromarkExtensions||(r.micromarkExtensions=[]),o=r.fromMarkdownExtensions||(r.fromMarkdownExtensions=[]),s=r.toMarkdownExtensions||(r.toMarkdownExtensions=[]);a.push(Wpe(n)),o.push(Bpe()),s.push(Upe(n))}function HD(e,t){var n=Object.keys(e);if(Object.getOwnPropertySymbols){var r=Object.getOwnPropertySymbols(e);t&&(r=r.filter(function(a){return Object.getOwnPropertyDescriptor(e,a).enumerable})),n.push.apply(n,r)}return n}function xs(e){for(var t=1;t<arguments.length;t++){var n=arguments[t]!=null?arguments[t]:{};t%2?HD(Object(n),!0).forEach(function(r){F5(e,r,n[r])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(n)):HD(Object(n)).forEach(function(r){Object.defineProperty(e,r,Object.getOwnPropertyDescriptor(n,r))})}return e}function Xpe(e){var t=e.length;if(t===0||t===1)return e;if(t===2)return[e[0],e[1],"".concat(e[0],".").concat(e[1]),"".concat(e[1],".").concat(e[0])];if(t===3)return[e[0],e[1],e[2],"".concat(e[0],".").concat(e[1]),"".concat(e[0],".").concat(e[2]),"".concat(e[1],".").concat(e[0]),"".concat(e[1],".").concat(e[2]),"".concat(e[2],".").concat(e[0]),"".concat(e[2],".").concat(e[1]),"".concat(e[0],".").concat(e[1],".").concat(e[2]),"".concat(e[0],".").concat(e[2],".").concat(e[1]),"".concat(e[1],".").concat(e[0],".").concat(e[2]),"".concat(e[1],".").concat(e[2],".").concat(e[0]),"".concat(e[2],".").concat(e[0],".").concat(e[1]),"".concat(e[2],".").concat(e[1],".").concat(e[0])];if(t>=4)return[e[0],e[1],e[2],e[3],"".concat(e[0],".").concat(e[1]),"".concat(e[0],".").concat(e[2]),"".concat(e[0],".").concat(e[3]),"".concat(e[1],".").concat(e[0]),"".concat(e[1],".").concat(e[2]),"".concat(e[1],".").concat(e[3]),"".concat(e[2],".").concat(e[0]),"".concat(e[2],".").concat(e[1]),"".concat(e[2],".").concat(e[3]),"".concat(e[3],".").concat(e[0]),"".concat(e[3],".").concat(e[1]),"".concat(e[3],".").concat(e[2]),"".concat(e[0],".").concat(e[1],".").concat(e[2]),"".concat(e[0],".").concat(e[1],".").concat(e[3]),"".concat(e[0],".").concat(e[2],".").concat(e[1]),"".concat(e[0],".").concat(e[2],".").concat(e[3]),"".concat(e[0],".").concat(e[3],".").concat(e[1]),"".concat(e[0],".").concat(e[3],".").concat(e[2]),"".concat(e[1],".").concat(e[0],".").concat(e[2]),"".concat(e[1],".").concat(e[0],".").concat(e[3]),"".concat(e[1],".").concat(e[2],".").concat(e[0]),"".concat(e[1],".").concat(e[2],".").concat(e[3]),"".concat(e[1],".").concat(e[3],".").concat(e[0]),"".concat(e[1],".").concat(e[3],".").concat(e[2]),"".concat(e[2],".").concat(e[0],".").concat(e[1]),"".concat(e[2],".").concat(e[0],".").concat(e[3]),"".concat(e[2],".").concat(e[1],".").concat(e[0]),"".concat(e[2],".").concat(e[1],".").concat(e[3]),"".concat(e[2],".").concat(e[3],".").concat(e[0]),"".concat(e[2],".").concat(e[3],".").concat(e[1]),"".concat(e[3],".").concat(e[0],".").concat(e[1]),"".concat(e[3],".").concat(e[0],".").concat(e[2]),"".concat(e[3],".").concat(e[1],".").concat(e[0]),"".concat(e[3],".").concat(e[1],".").concat(e[2]),"".concat(e[3],".").concat(e[2],".").concat(e[0]),"".concat(e[3],".").concat(e[2],".").concat(e[1]),"".concat(e[0],".").concat(e[1],".").concat(e[2],".").concat(e[3]),"".concat(e[0],".").concat(e[1],".").concat(e[3],".").concat(e[2]),"".concat(e[0],".").concat(e[2],".").concat(e[1],".").concat(e[3]),"".concat(e[0],".").concat(e[2],".").concat(e[3],".").concat(e[1]),"".concat(e[0],".").concat(e[3],".").concat(e[1],".").concat(e[2]),"".concat(e[0],".").concat(e[3],".").concat(e[2],".").concat(e[1]),"".concat(e[1],".").concat(e[0],".").concat(e[2],".").concat(e[3]),"".concat(e[1],".").concat(e[0],".").concat(e[3],".").concat(e[2]),"".concat(e[1],".").concat(e[2],".").concat(e[0],".").concat(e[3]),"".concat(e[1],".").concat(e[2],".").concat(e[3],".").concat(e[0]),"".concat(e[1],".").concat(e[3],".").concat(e[0],".").concat(e[2]),"".concat(e[1],".").concat(e[3],".").concat(e[2],".").concat(e[0]),"".concat(e[2],".").concat(e[0],".").concat(e[1],".").concat(e[3]),"".concat(e[2],".").concat(e[0],".").concat(e[3],".").concat(e[1]),"".concat(e[2],".").concat(e[1],".").concat(e[0],".").concat(e[3]),"".concat(e[2],".").concat(e[1],".").concat(e[3],".").concat(e[0]),"".concat(e[2],".").concat(e[3],".").concat(e[0],".").concat(e[1]),"".concat(e[2],".").concat(e[3],".").concat(e[1],".").concat(e[0]),"".concat(e[3],".").concat(e[0],".").concat(e[1],".").concat(e[2]),"".concat(e[3],".").concat(e[0],".").concat(e[2],".").concat(e[1]),"".concat(e[3],".").concat(e[1],".").concat(e[0],".").concat(e[2]),"".concat(e[3],".").concat(e[1],".").concat(e[2],".").concat(e[0]),"".concat(e[3],".").concat(e[2],".").concat(e[0],".").concat(e[1]),"".concat(e[3],".").concat(e[2],".").concat(e[1],".").concat(e[0])]}var Rb={};function Zpe(e){if(e.length===0||e.length===1)return e;var t=e.join(".");return Rb[t]||(Rb[t]=Xpe(e)),Rb[t]}function Qpe(e){var t=arguments.length>1&&arguments[1]!==void 0?arguments[1]:{},n=arguments.length>2?arguments[2]:void 0,r=e.filter(function(o){return o!=="token"}),a=Zpe(r);return a.reduce(function(o,s){return xs(xs({},o),n[s])},t)}function $D(e){return e.join(" ")}function Jpe(e,t){var n=0;return function(r){return n+=1,r.map(function(a,o){return DH({node:a,stylesheet:e,useInlineStyles:t,key:"code-segment-".concat(n,"-").concat(o)})})}}function DH(e){var t=e.node,n=e.stylesheet,r=e.style,a=r===void 0?{}:r,o=e.useInlineStyles,s=e.key,u=t.properties,c=t.type,d=t.tagName,p=t.value;if(c==="text")return p;if(d){var g=Jpe(n,o),m;if(!o)m=xs(xs({},u),{},{className:$D(u.className)});else{var b=Object.keys(n).reduce(function(R,k){return k.split(".").forEach(function(A){R.includes(A)||R.push(A)}),R},[]),y=u.className&&u.className.includes("token")?["token"]:[],S=u.className&&y.concat(u.className.filter(function(R){return!b.includes(R)}));m=xs(xs({},u),{},{className:$D(S)||void 0,style:Qpe(u.className,Object.assign({},u.style,a),n)})}var x=g(t.children);return ye.createElement(d,Zk({key:s},m),x)}}const ege=function(e,t){var n=e.listLanguages();return n.indexOf(t)!==-1};var tge=["language","children","style","customStyle","codeTagProps","useInlineStyles","showLineNumbers","showInlineLineNumbers","startingLineNumber","lineNumberContainerStyle","lineNumberStyle","wrapLines","wrapLongLines","lineProps","renderer","PreTag","CodeTag","code","astGenerator"];function qD(e,t){var n=Object.keys(e);if(Object.getOwnPropertySymbols){var r=Object.getOwnPropertySymbols(e);t&&(r=r.filter(function(a){return Object.getOwnPropertyDescriptor(e,a).enumerable})),n.push.apply(n,r)}return n}function Ao(e){for(var t=1;t<arguments.length;t++){var n=arguments[t]!=null?arguments[t]:{};t%2?qD(Object(n),!0).forEach(function(r){F5(e,r,n[r])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(n)):qD(Object(n)).forEach(function(r){Object.defineProperty(e,r,Object.getOwnPropertyDescriptor(n,r))})}return e}var nge=/\n/g;function rge(e){return e.match(nge)}function age(e){var t=e.lines,n=e.startingLineNumber,r=e.style;return t.map(function(a,o){var s=o+n;return ye.createElement("span",{key:"line-".concat(o),className:"react-syntax-highlighter-line-number",style:typeof r=="function"?r(s):r},"".concat(s,`
`))})}function oge(e){var t=e.codeString,n=e.codeStyle,r=e.containerStyle,a=r===void 0?{float:"left",paddingRight:"10px"}:r,o=e.numberStyle,s=o===void 0?{}:o,u=e.startingLineNumber;return ye.createElement("code",{style:Object.assign({},n,a)},age({lines:t.replace(/\n$/,"").split(`
`),style:s,startingLineNumber:u}))}function ige(e){return"".concat(e.toString().length,".25em")}function IH(e,t){return{type:"element",tagName:"span",properties:{key:"line-number--".concat(e),className:["comment","linenumber","react-syntax-highlighter-line-number"],style:t},children:[{type:"text",value:e}]}}function LH(e,t,n){var r={display:"inline-block",minWidth:ige(n),paddingRight:"1em",textAlign:"right",userSelect:"none"},a=typeof e=="function"?e(t):e,o=Ao(Ao({},r),a);return o}function cf(e){var t=e.children,n=e.lineNumber,r=e.lineNumberStyle,a=e.largestLineNumber,o=e.showInlineLineNumbers,s=e.lineProps,u=s===void 0?{}:s,c=e.className,d=c===void 0?[]:c,p=e.showLineNumbers,g=e.wrapLongLines,m=e.wrapLines,b=m===void 0?!1:m,y=b?Ao({},typeof u=="function"?u(n):u):{};if(y.className=y.className?[].concat(Jk(y.className.trim().split(/\s+/)),Jk(d)):d,n&&o){var S=LH(r,n,a);t.unshift(IH(n,S))}return g&p&&(y.style=Ao({display:"flex"},y.style)),{type:"element",tagName:"span",properties:y,children:t}}function MH(e){for(var t=arguments.length>1&&arguments[1]!==void 0?arguments[1]:[],n=arguments.length>2&&arguments[2]!==void 0?arguments[2]:[],r=0;r<e.length;r++){var a=e[r];if(a.type==="text")n.push(cf({children:[a],className:Jk(new Set(t))}));else if(a.children){var o=t.concat(a.properties.className);MH(a.children,o).forEach(function(s){return n.push(s)})}}return n}function sge(e,t,n,r,a,o,s,u,c){var d,p=MH(e.value),g=[],m=-1,b=0;function y(N,_){var O=arguments.length>2&&arguments[2]!==void 0?arguments[2]:[];return cf({children:N,lineNumber:_,lineNumberStyle:u,largestLineNumber:s,showInlineLineNumbers:a,lineProps:n,className:O,showLineNumbers:r,wrapLongLines:c,wrapLines:t})}function S(N,_){if(r&&_&&a){var O=LH(u,_,s);N.unshift(IH(_,O))}return N}function x(N,_){var O=arguments.length>2&&arguments[2]!==void 0?arguments[2]:[];return t||O.length>0?y(N,_,O):S(N,_)}for(var R=function(){var _=p[b],O=_.children[0].value,F=rge(O);if(F){var L=O.split(`
`);L.forEach(function(I,H){var $=r&&g.length+o,U={type:"text",value:"".concat(I,`
`)};if(H===0){var Y=p.slice(m+1,b).concat(cf({children:[U],className:_.properties.className})),Z=x(Y,$);g.push(Z)}else if(H===L.length-1){var j=p[b+1]&&p[b+1].children&&p[b+1].children[0],G={type:"text",value:"".concat(I)};if(j){var z=cf({children:[G],className:_.properties.className});p.splice(b+1,0,z)}else{var K=[G],D=x(K,$,_.properties.className);g.push(D)}}else{var V=[U],B=x(V,$,_.properties.className);g.push(B)}}),m=b}b++};b<p.length;)R();if(m!==p.length-1){var k=p.slice(m+1,p.length);if(k&&k.length){var A=r&&g.length+o,C=x(k,A);g.push(C)}}return t?g:(d=[]).concat.apply(d,g)}function lge(e){var t=e.rows,n=e.stylesheet,r=e.useInlineStyles;return t.map(function(a,o){return DH({node:a,stylesheet:n,useInlineStyles:r,key:"code-segement".concat(o)})})}function FH(e){return e&&typeof e.highlightAuto<"u"}function uge(e){var t=e.astGenerator,n=e.language,r=e.code,a=e.defaultCodeValue;if(FH(t)){var o=ege(t,n);return n==="text"?{value:a,language:"text"}:o?t.highlight(n,r):t.highlightAuto(r)}try{return n&&n!=="text"?{value:t.highlight(r,n)}:{value:a}}catch{return{value:a}}}function cge(e,t){return function(r){var a=r.language,o=r.children,s=r.style,u=s===void 0?t:s,c=r.customStyle,d=c===void 0?{}:c,p=r.codeTagProps,g=p===void 0?{className:a?"language-".concat(a):void 0,style:Ao(Ao({},u['code[class*="language-"]']),u['code[class*="language-'.concat(a,'"]')])}:p,m=r.useInlineStyles,b=m===void 0?!0:m,y=r.showLineNumbers,S=y===void 0?!1:y,x=r.showInlineLineNumbers,R=x===void 0?!0:x,k=r.startingLineNumber,A=k===void 0?1:k,C=r.lineNumberContainerStyle,N=r.lineNumberStyle,_=N===void 0?{}:N,O=r.wrapLines,F=r.wrapLongLines,L=F===void 0?!1:F,I=r.lineProps,H=I===void 0?{}:I,$=r.renderer,U=r.PreTag,Y=U===void 0?"pre":U,Z=r.CodeTag,j=Z===void 0?"code":Z,G=r.code,z=G===void 0?(Array.isArray(o)?o[0]:o)||"":G,K=r.astGenerator,D=xae(r,tge);K=K||e;var V=S?ye.createElement(oge,{containerStyle:C,codeStyle:g.style||{},numberStyle:_,startingLineNumber:A,codeString:z}):null,B=u.hljs||u['pre[class*="language-"]']||{backgroundColor:"#fff"},M=FH(K)?"hljs":"prismjs",W=b?Object.assign({},D,{style:Object.assign({},B,d)}):Object.assign({},D,{className:D.className?"".concat(M," ").concat(D.className):M,style:Object.assign({},d)});if(L?g.style=Ao({whiteSpace:"pre-wrap"},g.style):g.style=Ao({whiteSpace:"pre"},g.style),!K)return ye.createElement(Y,W,V,ye.createElement(j,g,z));(O===void 0&&$||L)&&(O=!0),$=$||lge;var Q=[{type:"text",value:z}],oe=uge({astGenerator:K,language:a,code:z,defaultCodeValue:Q});oe.language===null&&(oe.value=Q);var re=oe.value.length;re===1&&oe.value[0].type==="text"&&(re=oe.value[0].value.split(`
`).length);var ie=re+A,Se=sge(oe,O,H,S,R,A,ie,_,L);return ye.createElement(Y,W,ye.createElement(j,g,!R&&V,$({rows:Se,stylesheet:u,useInlineStyles:b})))}}const dge=["abap","abnf","actionscript","ada","agda","al","antlr4","apacheconf","apex","apl","applescript","aql","arduino","arff","asciidoc","asm6502","asmatmel","aspnet","autohotkey","autoit","avisynth","avro-idl","bash","basic","batch","bbcode","bicep","birb","bison","bnf","brainfuck","brightscript","bro","bsl","c","cfscript","chaiscript","cil","clike","clojure","cmake","cobol","coffeescript","concurnas","coq","cpp","crystal","csharp","cshtml","csp","css-extras","css","csv","cypher","d","dart","dataweave","dax","dhall","diff","django","dns-zone-file","docker","dot","ebnf","editorconfig","eiffel","ejs","elixir","elm","erb","erlang","etlua","excel-formula","factor","false","firestore-security-rules","flow","fortran","fsharp","ftl","gap","gcode","gdscript","gedcom","gherkin","git","glsl","gml","gn","go-module","go","graphql","groovy","haml","handlebars","haskell","haxe","hcl","hlsl","hoon","hpkp","hsts","http","ichigojam","icon","icu-message-format","idris","iecst","ignore","inform7","ini","io","j","java","javadoc","javadoclike","javascript","javastacktrace","jexl","jolie","jq","js-extras","js-templates","jsdoc","json","json5","jsonp","jsstacktrace","jsx","julia","keepalived","keyman","kotlin","kumir","kusto","latex","latte","less","lilypond","liquid","lisp","livescript","llvm","log","lolcode","lua","magma","makefile","markdown","markup-templating","markup","matlab","maxscript","mel","mermaid","mizar","mongodb","monkey","moonscript","n1ql","n4js","nand2tetris-hdl","naniscript","nasm","neon","nevod","nginx","nim","nix","nsis","objectivec","ocaml","opencl","openqasm","oz","parigp","parser","pascal","pascaligo","pcaxis","peoplecode","perl","php-extras","php","phpdoc","plsql","powerquery","powershell","processing","prolog","promql","properties","protobuf","psl","pug","puppet","pure","purebasic","purescript","python","q","qml","qore","qsharp","r","racket","reason","regex","rego","renpy","rest","rip","roboconf","robotframework","ruby","rust","sas","sass","scala","scheme","scss","shell-session","smali","smalltalk","smarty","sml","solidity","solution-file","soy","sparql","splunk-spl","sqf","sql","squirrel","stan","stylus","swift","systemd","t4-cs","t4-templating","t4-vb","tap","tcl","textile","toml","tremor","tsx","tt2","turtle","twig","typescript","typoscript","unrealscript","uorazor","uri","v","vala","vbnet","velocity","verilog","vhdl","vim","visual-basic","warpscript","wasm","web-idl","wiki","wolfram","wren","xeora","xml-doc","xojo","xquery","yaml","yang","zig"];var _b,VD;function fge(){if(VD)return _b;VD=1,_b=t;var e=Object.prototype.hasOwnProperty;function t(){for(var n={},r=0;r<arguments.length;r++){var a=arguments[r];for(var o in a)e.call(a,o)&&(n[o]=a[o])}return n}return _b}var Cb,WD;function PH(){if(WD)return Cb;WD=1,Cb=t;var e=t.prototype;e.space=null,e.normal={},e.property={};function t(n,r,a){this.property=n,this.normal=r,a&&(this.space=a)}return Cb}var Nb,YD;function pge(){if(YD)return Nb;YD=1;var e=fge(),t=PH();Nb=n;function n(r){for(var a=r.length,o=[],s=[],u=-1,c,d;++u<a;)c=r[u],o.push(c.property),s.push(c.normal),d=c.space;return new t(e.apply(null,o),e.apply(null,s),d)}return Nb}var Ob,KD;function iA(){if(KD)return Ob;KD=1,Ob=e;function e(t){return t.toLowerCase()}return Ob}var Db,XD;function zH(){if(XD)return Db;XD=1,Db=t;var e=t.prototype;e.space=null,e.attribute=null,e.property=null,e.boolean=!1,e.booleanish=!1,e.overloadedBoolean=!1,e.number=!1,e.commaSeparated=!1,e.spaceSeparated=!1,e.commaOrSpaceSeparated=!1,e.mustUseProperty=!1,e.defined=!1;function t(n,r){this.property=n,this.attribute=r}return Db}var Ia={},ZD;function sA(){if(ZD)return Ia;ZD=1;var e=0;Ia.boolean=t(),Ia.booleanish=t(),Ia.overloadedBoolean=t(),Ia.number=t(),Ia.spaceSeparated=t(),Ia.commaSeparated=t(),Ia.commaOrSpaceSeparated=t();function t(){return Math.pow(2,++e)}return Ia}var Ib,QD;function BH(){if(QD)return Ib;QD=1;var e=zH(),t=sA();Ib=a,a.prototype=new e,a.prototype.defined=!0;var n=["boolean","booleanish","overloadedBoolean","number","commaSeparated","spaceSeparated","commaOrSpaceSeparated"],r=n.length;function a(s,u,c,d){var p=-1,g;for(o(this,"space",d),e.call(this,s,u);++p<r;)g=n[p],o(this,g,(c&t[g])===t[g])}function o(s,u,c){c&&(s[u]=c)}return Ib}var Lb,JD;function ic(){if(JD)return Lb;JD=1;var e=iA(),t=PH(),n=BH();Lb=r;function r(a){var o=a.space,s=a.mustUseProperty||[],u=a.attributes||{},c=a.properties,d=a.transform,p={},g={},m,b;for(m in c)b=new n(m,d(u,m),c[m],o),s.indexOf(m)!==-1&&(b.mustUseProperty=!0),p[m]=b,g[e(m)]=m,g[e(b.attribute)]=m;return new t(p,g,o)}return Lb}var Mb,eI;function gge(){if(eI)return Mb;eI=1;var e=ic();Mb=e({space:"xlink",transform:t,properties:{xLinkActuate:null,xLinkArcRole:null,xLinkHref:null,xLinkRole:null,xLinkShow:null,xLinkTitle:null,xLinkType:null}});function t(n,r){return"xlink:"+r.slice(5).toLowerCase()}return Mb}var Fb,tI;function hge(){if(tI)return Fb;tI=1;var e=ic();Fb=e({space:"xml",transform:t,properties:{xmlLang:null,xmlBase:null,xmlSpace:null}});function t(n,r){return"xml:"+r.slice(3).toLowerCase()}return Fb}var Pb,nI;function mge(){if(nI)return Pb;nI=1,Pb=e;function e(t,n){return n in t?t[n]:n}return Pb}var zb,rI;function UH(){if(rI)return zb;rI=1;var e=mge();zb=t;function t(n,r){return e(n,r.toLowerCase())}return zb}var Bb,aI;function bge(){if(aI)return Bb;aI=1;var e=ic(),t=UH();return Bb=e({space:"xmlns",attributes:{xmlnsxlink:"xmlns:xlink"},transform:t,properties:{xmlns:null,xmlnsXLink:null}}),Bb}var Ub,oI;function yge(){if(oI)return Ub;oI=1;var e=sA(),t=ic(),n=e.booleanish,r=e.number,a=e.spaceSeparated;Ub=t({transform:o,properties:{ariaActiveDescendant:null,ariaAtomic:n,ariaAutoComplete:null,ariaBusy:n,ariaChecked:n,ariaColCount:r,ariaColIndex:r,ariaColSpan:r,ariaControls:a,ariaCurrent:null,ariaDescribedBy:a,ariaDetails:null,ariaDisabled:n,ariaDropEffect:a,ariaErrorMessage:null,ariaExpanded:n,ariaFlowTo:a,ariaGrabbed:n,ariaHasPopup:null,ariaHidden:n,ariaInvalid:null,ariaKeyShortcuts:null,ariaLabel:null,ariaLabelledBy:a,ariaLevel:r,ariaLive:null,ariaModal:n,ariaMultiLine:n,ariaMultiSelectable:n,ariaOrientation:null,ariaOwns:a,ariaPlaceholder:null,ariaPosInSet:r,ariaPressed:n,ariaReadOnly:n,ariaRelevant:null,ariaRequired:n,ariaRoleDescription:a,ariaRowCount:r,ariaRowIndex:r,ariaRowSpan:r,ariaSelected:n,ariaSetSize:r,ariaSort:null,ariaValueMax:r,ariaValueMin:r,ariaValueNow:r,ariaValueText:null,role:null}});function o(s,u){return u==="role"?u:"aria-"+u.slice(4).toLowerCase()}return Ub}var jb,iI;function vge(){if(iI)return jb;iI=1;var e=sA(),t=ic(),n=UH(),r=e.boolean,a=e.overloadedBoolean,o=e.booleanish,s=e.number,u=e.spaceSeparated,c=e.commaSeparated;return jb=t({space:"html",attributes:{acceptcharset:"accept-charset",classname:"class",htmlfor:"for",httpequiv:"http-equiv"},transform:n,mustUseProperty:["checked","multiple","muted","selected"],properties:{abbr:null,accept:c,acceptCharset:u,accessKey:u,action:null,allow:null,allowFullScreen:r,allowPaymentRequest:r,allowUserMedia:r,alt:null,as:null,async:r,autoCapitalize:null,autoComplete:u,autoFocus:r,autoPlay:r,capture:r,charSet:null,checked:r,cite:null,className:u,cols:s,colSpan:null,content:null,contentEditable:o,controls:r,controlsList:u,coords:s|c,crossOrigin:null,data:null,dateTime:null,decoding:null,default:r,defer:r,dir:null,dirName:null,disabled:r,download:a,draggable:o,encType:null,enterKeyHint:null,form:null,formAction:null,formEncType:null,formMethod:null,formNoValidate:r,formTarget:null,headers:u,height:s,hidden:r,high:s,href:null,hrefLang:null,htmlFor:u,httpEquiv:u,id:null,imageSizes:null,imageSrcSet:c,inputMode:null,integrity:null,is:null,isMap:r,itemId:null,itemProp:u,itemRef:u,itemScope:r,itemType:u,kind:null,label:null,lang:null,language:null,list:null,loading:null,loop:r,low:s,manifest:null,max:null,maxLength:s,media:null,method:null,min:null,minLength:s,multiple:r,muted:r,name:null,nonce:null,noModule:r,noValidate:r,onAbort:null,onAfterPrint:null,onAuxClick:null,onBeforePrint:null,onBeforeUnload:null,onBlur:null,onCancel:null,onCanPlay:null,onCanPlayThrough:null,onChange:null,onClick:null,onClose:null,onContextMenu:null,onCopy:null,onCueChange:null,onCut:null,onDblClick:null,onDrag:null,onDragEnd:null,onDragEnter:null,onDragExit:null,onDragLeave:null,onDragOver:null,onDragStart:null,onDrop:null,onDurationChange:null,onEmptied:null,onEnded:null,onError:null,onFocus:null,onFormData:null,onHashChange:null,onInput:null,onInvalid:null,onKeyDown:null,onKeyPress:null,onKeyUp:null,onLanguageChange:null,onLoad:null,onLoadedData:null,onLoadedMetadata:null,onLoadEnd:null,onLoadStart:null,onMessage:null,onMessageError:null,onMouseDown:null,onMouseEnter:null,onMouseLeave:null,onMouseMove:null,onMouseOut:null,onMouseOver:null,onMouseUp:null,onOffline:null,onOnline:null,onPageHide:null,onPageShow:null,onPaste:null,onPause:null,onPlay:null,onPlaying:null,onPopState:null,onProgress:null,onRateChange:null,onRejectionHandled:null,onReset:null,onResize:null,onScroll:null,onSecurityPolicyViolation:null,onSeeked:null,onSeeking:null,onSelect:null,onSlotChange:null,onStalled:null,onStorage:null,onSubmit:null,onSuspend:null,onTimeUpdate:null,onToggle:null,onUnhandledRejection:null,onUnload:null,onVolumeChange:null,onWaiting:null,onWheel:null,open:r,optimum:s,pattern:null,ping:u,placeholder:null,playsInline:r,poster:null,preload:null,readOnly:r,referrerPolicy:null,rel:u,required:r,reversed:r,rows:s,rowSpan:s,sandbox:u,scope:null,scoped:r,seamless:r,selected:r,shape:null,size:s,sizes:null,slot:null,span:s,spellCheck:o,src:null,srcDoc:null,srcLang:null,srcSet:c,start:s,step:null,style:null,tabIndex:s,target:null,title:null,translate:null,type:null,typeMustMatch:r,useMap:null,value:o,width:s,wrap:null,align:null,aLink:null,archive:u,axis:null,background:null,bgColor:null,border:s,borderColor:null,bottomMargin:s,cellPadding:null,cellSpacing:null,char:null,charOff:null,classId:null,clear:null,code:null,codeBase:null,codeType:null,color:null,compact:r,declare:r,event:null,face:null,frame:null,frameBorder:null,hSpace:s,leftMargin:s,link:null,longDesc:null,lowSrc:null,marginHeight:s,marginWidth:s,noResize:r,noHref:r,noShade:r,noWrap:r,object:null,profile:null,prompt:null,rev:null,rightMargin:s,rules:null,scheme:null,scrolling:o,standby:null,summary:null,text:null,topMargin:s,valueType:null,version:null,vAlign:null,vLink:null,vSpace:s,allowTransparency:null,autoCorrect:null,autoSave:null,disablePictureInPicture:r,disableRemotePlayback:r,prefix:null,property:null,results:s,security:null,unselectable:null}}),jb}var Gb,sI;function Sge(){if(sI)return Gb;sI=1;var e=pge(),t=gge(),n=hge(),r=bge(),a=yge(),o=vge();return Gb=e([n,t,r,a,o]),Gb}var Hb,lI;function Ege(){if(lI)return Hb;lI=1;var e=iA(),t=BH(),n=zH(),r="data";Hb=u;var a=/^data[-\w.:]+$/i,o=/-[a-z]/g,s=/[A-Z]/g;function u(m,b){var y=e(b),S=b,x=n;return y in m.normal?m.property[m.normal[y]]:(y.length>4&&y.slice(0,4)===r&&a.test(b)&&(b.charAt(4)==="-"?S=c(b):b=d(b),x=t),new x(S,b))}function c(m){var b=m.slice(5).replace(o,g);return r+b.charAt(0).toUpperCase()+b.slice(1)}function d(m){var b=m.slice(4);return o.test(b)?m:(b=b.replace(s,p),b.charAt(0)!=="-"&&(b="-"+b),r+b)}function p(m){return"-"+m.toLowerCase()}function g(m){return m.charAt(1).toUpperCase()}return Hb}var $b,uI;function wge(){if(uI)return $b;uI=1,$b=t;var e=/[#.]/g;function t(n,r){for(var a=n||"",o=r||"div",s={},u=0,c,d,p;u<a.length;)e.lastIndex=u,p=e.exec(a),c=a.slice(u,p?p.index:a.length),c&&(d?d==="#"?s.id=c:s.className?s.className.push(c):s.className=[c]:o=c,u+=c.length),p&&(d=p[0],u++);return{type:"element",tagName:o,properties:s,children:[]}}return $b}var Gd={},cI;function xge(){if(cI)return Gd;cI=1,Gd.parse=r,Gd.stringify=a;var e="",t=" ",n=/[ \t\n\r\f]+/g;function r(o){var s=String(o||e).trim();return s===e?[]:s.split(n)}function a(o){return o.join(t).trim()}return Gd}var Hd={},dI;function kge(){if(dI)return Hd;dI=1,Hd.parse=r,Hd.stringify=a;var e=",",t=" ",n="";function r(o){for(var s=[],u=String(o||n),c=u.indexOf(e),d=0,p=!1,g;!p;)c===-1&&(c=u.length,p=!0),g=u.slice(d,c).trim(),(g||!p)&&s.push(g),d=c+1,c=u.indexOf(e,d);return s}function a(o,s){var u=s||{},c=u.padLeft===!1?n:t,d=u.padRight?t:n;return o[o.length-1]===n&&(o=o.concat(n)),o.join(d+e+c).trim()}return Hd}var qb,fI;function Tge(){if(fI)return qb;fI=1;var e=Ege(),t=iA(),n=wge(),r=xge().parse,a=kge().parse;qb=s;var o={}.hasOwnProperty;function s(y,S,x){var R=x?b(x):null;return k;function k(C,N){var _=n(C,S),O=Array.prototype.slice.call(arguments,2),F=_.tagName.toLowerCase(),L;if(_.tagName=R&&o.call(R,F)?R[F]:F,N&&u(N,_)&&(O.unshift(N),N=null),N)for(L in N)A(_.properties,L,N[L]);return d(_.children,O),_.tagName==="template"&&(_.content={type:"root",children:_.children},_.children=[]),_}function A(C,N,_){var O,F,L;_==null||_!==_||(O=e(y,N),F=O.property,L=_,typeof L=="string"&&(O.spaceSeparated?L=r(L):O.commaSeparated?L=a(L):O.commaOrSpaceSeparated&&(L=r(a(L).join(" ")))),F==="style"&&typeof _!="string"&&(L=m(L)),F==="className"&&C.className&&(L=C.className.concat(L)),C[F]=p(O,F,L))}}function u(y,S){return typeof y=="string"||"length"in y||c(S.tagName,y)}function c(y,S){var x=S.type;return y==="input"||!x||typeof x!="string"?!1:typeof S.children=="object"&&"length"in S.children?!0:(x=x.toLowerCase(),y==="button"?x!=="menu"&&x!=="submit"&&x!=="reset"&&x!=="button":"value"in S)}function d(y,S){var x,R;if(typeof S=="string"||typeof S=="number"){y.push({type:"text",value:String(S)});return}if(typeof S=="object"&&"length"in S){for(x=-1,R=S.length;++x<R;)d(y,S[x]);return}if(typeof S!="object"||!("type"in S))throw new Error("Expected node, nodes, or string, got `"+S+"`");y.push(S)}function p(y,S,x){var R,k,A;if(typeof x!="object"||!("length"in x))return g(y,S,x);for(k=x.length,R=-1,A=[];++R<k;)A[R]=g(y,S,x[R]);return A}function g(y,S,x){var R=x;return y.number||y.positiveNumber?!isNaN(R)&&R!==""&&(R=Number(R)):(y.boolean||y.overloadedBoolean)&&typeof R=="string"&&(R===""||t(x)===t(S))&&(R=!0),R}function m(y){var S=[],x;for(x in y)S.push([x,y[x]].join(": "));return S.join("; ")}function b(y){for(var S=y.length,x=-1,R={},k;++x<S;)k=y[x],R[k.toLowerCase()]=k;return R}return qb}var Vb,pI;function Age(){if(pI)return Vb;pI=1;var e=Sge(),t=Tge(),n=t(e,"div");return n.displayName="html",Vb=n,Vb}var Wb,gI;function Rge(){return gI||(gI=1,Wb=Age()),Wb}const _ge="Æ",Cge="&",Nge="Á",Oge="Â",Dge="À",Ige="Å",Lge="Ã",Mge="Ä",Fge="©",Pge="Ç",zge="Ð",Bge="É",Uge="Ê",jge="È",Gge="Ë",Hge=">",$ge="Í",qge="Î",Vge="Ì",Wge="Ï",Yge="<",Kge="Ñ",Xge="Ó",Zge="Ô",Qge="Ò",Jge="Ø",ehe="Õ",the="Ö",nhe='"',rhe="®",ahe="Þ",ohe="Ú",ihe="Û",she="Ù",lhe="Ü",uhe="Ý",che="á",dhe="â",fhe="´",phe="æ",ghe="à",hhe="&",mhe="å",bhe="ã",yhe="ä",vhe="¦",She="ç",Ehe="¸",whe="¢",xhe="©",khe="¤",The="°",Ahe="÷",Rhe="é",_he="ê",Che="è",Nhe="ð",Ohe="ë",Dhe="½",Ihe="¼",Lhe="¾",Mhe=">",Fhe="í",Phe="î",zhe="¡",Bhe="ì",Uhe="¿",jhe="ï",Ghe="«",Hhe="<",$he="¯",qhe="µ",Vhe="·",Whe=" ",Yhe="¬",Khe="ñ",Xhe="ó",Zhe="ô",Qhe="ò",Jhe="ª",eme="º",tme="ø",nme="õ",rme="ö",ame="¶",ome="±",ime="£",sme='"',lme="»",ume="®",cme="§",dme="­",fme="¹",pme="²",gme="³",hme="ß",mme="þ",bme="×",yme="ú",vme="û",Sme="ù",Eme="¨",wme="ü",xme="ý",kme="¥",Tme="ÿ",Ame={AElig:_ge,AMP:Cge,Aacute:Nge,Acirc:Oge,Agrave:Dge,Aring:Ige,Atilde:Lge,Auml:Mge,COPY:Fge,Ccedil:Pge,ETH:zge,Eacute:Bge,Ecirc:Uge,Egrave:jge,Euml:Gge,GT:Hge,Iacute:$ge,Icirc:qge,Igrave:Vge,Iuml:Wge,LT:Yge,Ntilde:Kge,Oacute:Xge,Ocirc:Zge,Ograve:Qge,Oslash:Jge,Otilde:ehe,Ouml:the,QUOT:nhe,REG:rhe,THORN:ahe,Uacute:ohe,Ucirc:ihe,Ugrave:she,Uuml:lhe,Yacute:uhe,aacute:che,acirc:dhe,acute:fhe,aelig:phe,agrave:ghe,amp:hhe,aring:mhe,atilde:bhe,auml:yhe,brvbar:vhe,ccedil:She,cedil:Ehe,cent:whe,copy:xhe,curren:khe,deg:The,divide:Ahe,eacute:Rhe,ecirc:_he,egrave:Che,eth:Nhe,euml:Ohe,frac12:Dhe,frac14:Ihe,frac34:Lhe,gt:Mhe,iacute:Fhe,icirc:Phe,iexcl:zhe,igrave:Bhe,iquest:Uhe,iuml:jhe,laquo:Ghe,lt:Hhe,macr:$he,micro:qhe,middot:Vhe,nbsp:Whe,not:Yhe,ntilde:Khe,oacute:Xhe,ocirc:Zhe,ograve:Qhe,ordf:Jhe,ordm:eme,oslash:tme,otilde:nme,ouml:rme,para:ame,plusmn:ome,pound:ime,quot:sme,raquo:lme,reg:ume,sect:cme,shy:dme,sup1:fme,sup2:pme,sup3:gme,szlig:hme,thorn:mme,times:bme,uacute:yme,ucirc:vme,ugrave:Sme,uml:Eme,uuml:wme,yacute:xme,yen:kme,yuml:Tme},Rme={0:"�",128:"€",130:"‚",131:"ƒ",132:"„",133:"…",134:"†",135:"‡",136:"ˆ",137:"‰",138:"Š",139:"‹",140:"Œ",142:"Ž",145:"‘",146:"’",147:"“",148:"”",149:"•",150:"–",151:"—",152:"˜",153:"™",154:"š",155:"›",156:"œ",158:"ž",159:"Ÿ"};var Yb,hI;function jH(){if(hI)return Yb;hI=1,Yb=e;function e(t){var n=typeof t=="string"?t.charCodeAt(0):t;return n>=48&&n<=57}return Yb}var Kb,mI;function _me(){if(mI)return Kb;mI=1,Kb=e;function e(t){var n=typeof t=="string"?t.charCodeAt(0):t;return n>=97&&n<=102||n>=65&&n<=70||n>=48&&n<=57}return Kb}var Xb,bI;function Cme(){if(bI)return Xb;bI=1,Xb=e;function e(t){var n=typeof t=="string"?t.charCodeAt(0):t;return n>=97&&n<=122||n>=65&&n<=90}return Xb}var Zb,yI;function Nme(){if(yI)return Zb;yI=1;var e=Cme(),t=jH();Zb=n;function n(r){return e(r)||t(r)}return Zb}var Qb,vI;function Ome(){if(vI)return Qb;vI=1;var e,t=59;Qb=n;function n(r){var a="&"+r+";",o;return e=e||document.createElement("i"),e.innerHTML=a,o=e.textContent,o.charCodeAt(o.length-1)===t&&r!=="semi"||o===a?!1:o}return Qb}var Jb,SI;function Dme(){if(SI)return Jb;SI=1;var e=Ame,t=Rme,n=jH(),r=_me(),a=Nme(),o=Ome();Jb=K;var s={}.hasOwnProperty,u=String.fromCharCode,c=Function.prototype,d={warning:null,reference:null,text:null,warningContext:null,referenceContext:null,textContext:null,position:{},additional:null,attribute:!1,nonTerminated:!0},p=9,g=10,m=12,b=32,y=38,S=59,x=60,R=61,k=35,A=88,C=120,N=65533,_="named",O="hexadecimal",F="decimal",L={};L[O]=16,L[F]=10;var I={};I[_]=a,I[F]=n,I[O]=r;var H=1,$=2,U=3,Y=4,Z=5,j=6,G=7,z={};z[H]="Named character references must be terminated by a semicolon",z[$]="Numeric character references must be terminated by a semicolon",z[U]="Named character references cannot be empty",z[Y]="Numeric character references cannot be empty",z[Z]="Named character references must be known",z[j]="Numeric character references cannot be disallowed",z[G]="Numeric character references cannot be outside the permissible Unicode range";function K(M,W){var Q={},oe,re;W||(W={});for(re in d)oe=W[re],Q[re]=oe??d[re];return(Q.position.indent||Q.position.start)&&(Q.indent=Q.position.indent||[],Q.position=Q.position.start),D(M,Q)}function D(M,W){var Q=W.additional,oe=W.nonTerminated,re=W.text,ie=W.reference,Se=W.warning,ae=W.textContext,ve=W.referenceContext,xe=W.warningContext,Ie=W.position,Ce=W.indent||[],ke=M.length,J=0,fe=-1,Te=Ie.column||1,me=Ie.line||1,Ee="",le=[],Be,Ue,he,Ne,ee,ce,_e,Pe,We,yt,kt,ht,et,Tt,ot,St,Ht,fn,zt;for(typeof Q=="string"&&(Q=Q.charCodeAt(0)),St=or(),Pe=Se?$r:c,J--,ke++;++J<ke;)if(ee===g&&(Te=Ce[fe]||1),ee=M.charCodeAt(J),ee===y){if(_e=M.charCodeAt(J+1),_e===p||_e===g||_e===m||_e===b||_e===y||_e===x||_e!==_e||Q&&_e===Q){Ee+=u(ee),Te++;continue}for(et=J+1,ht=et,zt=et,_e===k?(zt=++ht,_e=M.charCodeAt(zt),_e===A||_e===C?(Tt=O,zt=++ht):Tt=F):Tt=_,Be="",kt="",Ne="",ot=I[Tt],zt--;++zt<ke&&(_e=M.charCodeAt(zt),!!ot(_e));)Ne+=u(_e),Tt===_&&s.call(e,Ne)&&(Be=Ne,kt=e[Ne]);he=M.charCodeAt(zt)===S,he&&(zt++,Ue=Tt===_?o(Ne):!1,Ue&&(Be=Ne,kt=Ue)),fn=1+zt-et,!he&&!oe||(Ne?Tt===_?(he&&!kt?Pe(Z,1):(Be!==Ne&&(zt=ht+Be.length,fn=1+zt-ht,he=!1),he||(We=Be?H:U,W.attribute?(_e=M.charCodeAt(zt),_e===R?(Pe(We,fn),kt=null):a(_e)?kt=null:Pe(We,fn)):Pe(We,fn))),ce=kt):(he||Pe($,fn),ce=parseInt(Ne,L[Tt]),V(ce)?(Pe(G,fn),ce=u(N)):ce in t?(Pe(j,fn),ce=t[ce]):(yt="",B(ce)&&Pe(j,fn),ce>65535&&(ce-=65536,yt+=u(ce>>>10|55296),ce=56320|ce&1023),ce=yt+u(ce))):Tt!==_&&Pe(Y,fn)),ce?(Jt(),St=or(),J=zt-1,Te+=zt-et+1,le.push(ce),Ht=or(),Ht.offset++,ie&&ie.call(ve,ce,{start:St,end:Ht},M.slice(et-1,zt)),St=Ht):(Ne=M.slice(et-1,zt),Ee+=Ne,Te+=Ne.length,J=zt-1)}else ee===10&&(me++,fe++,Te=0),ee===ee?(Ee+=u(ee),Te++):Jt();return le.join("");function or(){return{line:me,column:Te,offset:J+(Ie.offset||0)}}function $r(fa,Xe){var mt=or();mt.column+=Xe,mt.offset+=Xe,Se.call(xe,z[fa],mt,fa)}function Jt(){Ee&&(le.push(Ee),re&&re.call(ae,Ee,{start:St,end:or()}),Ee="")}}function V(M){return M>=55296&&M<=57343||M>1114111}function B(M){return M>=1&&M<=8||M===11||M>=13&&M<=31||M>=127&&M<=159||M>=64976&&M<=65007||(M&65535)===65535||(M&65535)===65534}return Jb}var ey={exports:{}},EI;function Ime(){return EI||(EI=1,function(e){var t=typeof window<"u"?window:typeof WorkerGlobalScope<"u"&&self instanceof WorkerGlobalScope?self:{};/**
 * Prism: Lightweight, robust, elegant syntax highlighting
 *
 * @license MIT <https://opensource.org/licenses/MIT>
 * @author Lea Verou <https://lea.verou.me>
 * @namespace
 * @public
 */var n=function(r){var a=/(?:^|\s)lang(?:uage)?-([\w-]+)(?=\s|$)/i,o=0,s={},u={manual:r.Prism&&r.Prism.manual,disableWorkerMessageHandler:r.Prism&&r.Prism.disableWorkerMessageHandler,util:{encode:function k(A){return A instanceof c?new c(A.type,k(A.content),A.alias):Array.isArray(A)?A.map(k):A.replace(/&/g,"&amp;").replace(/</g,"&lt;").replace(/\u00a0/g," ")},type:function(k){return Object.prototype.toString.call(k).slice(8,-1)},objId:function(k){return k.__id||Object.defineProperty(k,"__id",{value:++o}),k.__id},clone:function k(A,C){C=C||{};var N,_;switch(u.util.type(A)){case"Object":if(_=u.util.objId(A),C[_])return C[_];N={},C[_]=N;for(var O in A)A.hasOwnProperty(O)&&(N[O]=k(A[O],C));return N;case"Array":return _=u.util.objId(A),C[_]?C[_]:(N=[],C[_]=N,A.forEach(function(F,L){N[L]=k(F,C)}),N);default:return A}},getLanguage:function(k){for(;k;){var A=a.exec(k.className);if(A)return A[1].toLowerCase();k=k.parentElement}return"none"},setLanguage:function(k,A){k.className=k.className.replace(RegExp(a,"gi"),""),k.classList.add("language-"+A)},currentScript:function(){if(typeof document>"u")return null;if("currentScript"in document)return document.currentScript;try{throw new Error}catch(N){var k=(/at [^(\r\n]*\((.*):[^:]+:[^:]+\)$/i.exec(N.stack)||[])[1];if(k){var A=document.getElementsByTagName("script");for(var C in A)if(A[C].src==k)return A[C]}return null}},isActive:function(k,A,C){for(var N="no-"+A;k;){var _=k.classList;if(_.contains(A))return!0;if(_.contains(N))return!1;k=k.parentElement}return!!C}},languages:{plain:s,plaintext:s,text:s,txt:s,extend:function(k,A){var C=u.util.clone(u.languages[k]);for(var N in A)C[N]=A[N];return C},insertBefore:function(k,A,C,N){N=N||u.languages;var _=N[k],O={};for(var F in _)if(_.hasOwnProperty(F)){if(F==A)for(var L in C)C.hasOwnProperty(L)&&(O[L]=C[L]);C.hasOwnProperty(F)||(O[F]=_[F])}var I=N[k];return N[k]=O,u.languages.DFS(u.languages,function(H,$){$===I&&H!=k&&(this[H]=O)}),O},DFS:function k(A,C,N,_){_=_||{};var O=u.util.objId;for(var F in A)if(A.hasOwnProperty(F)){C.call(A,F,A[F],N||F);var L=A[F],I=u.util.type(L);I==="Object"&&!_[O(L)]?(_[O(L)]=!0,k(L,C,null,_)):I==="Array"&&!_[O(L)]&&(_[O(L)]=!0,k(L,C,F,_))}}},plugins:{},highlightAll:function(k,A){u.highlightAllUnder(document,k,A)},highlightAllUnder:function(k,A,C){var N={callback:C,container:k,selector:'code[class*="language-"], [class*="language-"] code, code[class*="lang-"], [class*="lang-"] code'};u.hooks.run("before-highlightall",N),N.elements=Array.prototype.slice.apply(N.container.querySelectorAll(N.selector)),u.hooks.run("before-all-elements-highlight",N);for(var _=0,O;O=N.elements[_++];)u.highlightElement(O,A===!0,N.callback)},highlightElement:function(k,A,C){var N=u.util.getLanguage(k),_=u.languages[N];u.util.setLanguage(k,N);var O=k.parentElement;O&&O.nodeName.toLowerCase()==="pre"&&u.util.setLanguage(O,N);var F=k.textContent,L={element:k,language:N,grammar:_,code:F};function I($){L.highlightedCode=$,u.hooks.run("before-insert",L),L.element.innerHTML=L.highlightedCode,u.hooks.run("after-highlight",L),u.hooks.run("complete",L),C&&C.call(L.element)}if(u.hooks.run("before-sanity-check",L),O=L.element.parentElement,O&&O.nodeName.toLowerCase()==="pre"&&!O.hasAttribute("tabindex")&&O.setAttribute("tabindex","0"),!L.code){u.hooks.run("complete",L),C&&C.call(L.element);return}if(u.hooks.run("before-highlight",L),!L.grammar){I(u.util.encode(L.code));return}if(A&&r.Worker){var H=new Worker(u.filename);H.onmessage=function($){I($.data)},H.postMessage(JSON.stringify({language:L.language,code:L.code,immediateClose:!0}))}else I(u.highlight(L.code,L.grammar,L.language))},highlight:function(k,A,C){var N={code:k,grammar:A,language:C};if(u.hooks.run("before-tokenize",N),!N.grammar)throw new Error('The language "'+N.language+'" has no grammar.');return N.tokens=u.tokenize(N.code,N.grammar),u.hooks.run("after-tokenize",N),c.stringify(u.util.encode(N.tokens),N.language)},tokenize:function(k,A){var C=A.rest;if(C){for(var N in C)A[N]=C[N];delete A.rest}var _=new g;return m(_,_.head,k),p(k,_,A,_.head,0),y(_)},hooks:{all:{},add:function(k,A){var C=u.hooks.all;C[k]=C[k]||[],C[k].push(A)},run:function(k,A){var C=u.hooks.all[k];if(!(!C||!C.length))for(var N=0,_;_=C[N++];)_(A)}},Token:c};r.Prism=u;function c(k,A,C,N){this.type=k,this.content=A,this.alias=C,this.length=(N||"").length|0}c.stringify=function k(A,C){if(typeof A=="string")return A;if(Array.isArray(A)){var N="";return A.forEach(function(I){N+=k(I,C)}),N}var _={type:A.type,content:k(A.content,C),tag:"span",classes:["token",A.type],attributes:{},language:C},O=A.alias;O&&(Array.isArray(O)?Array.prototype.push.apply(_.classes,O):_.classes.push(O)),u.hooks.run("wrap",_);var F="";for(var L in _.attributes)F+=" "+L+'="'+(_.attributes[L]||"").replace(/"/g,"&quot;")+'"';return"<"+_.tag+' class="'+_.classes.join(" ")+'"'+F+">"+_.content+"</"+_.tag+">"};function d(k,A,C,N){k.lastIndex=A;var _=k.exec(C);if(_&&N&&_[1]){var O=_[1].length;_.index+=O,_[0]=_[0].slice(O)}return _}function p(k,A,C,N,_,O){for(var F in C)if(!(!C.hasOwnProperty(F)||!C[F])){var L=C[F];L=Array.isArray(L)?L:[L];for(var I=0;I<L.length;++I){if(O&&O.cause==F+","+I)return;var H=L[I],$=H.inside,U=!!H.lookbehind,Y=!!H.greedy,Z=H.alias;if(Y&&!H.pattern.global){var j=H.pattern.toString().match(/[imsuy]*$/)[0];H.pattern=RegExp(H.pattern.source,j+"g")}for(var G=H.pattern||H,z=N.next,K=_;z!==A.tail&&!(O&&K>=O.reach);K+=z.value.length,z=z.next){var D=z.value;if(A.length>k.length)return;if(!(D instanceof c)){var V=1,B;if(Y){if(B=d(G,K,k,U),!B||B.index>=k.length)break;var oe=B.index,M=B.index+B[0].length,W=K;for(W+=z.value.length;oe>=W;)z=z.next,W+=z.value.length;if(W-=z.value.length,K=W,z.value instanceof c)continue;for(var Q=z;Q!==A.tail&&(W<M||typeof Q.value=="string");Q=Q.next)V++,W+=Q.value.length;V--,D=k.slice(K,W),B.index-=K}else if(B=d(G,0,D,U),!B)continue;var oe=B.index,re=B[0],ie=D.slice(0,oe),Se=D.slice(oe+re.length),ae=K+D.length;O&&ae>O.reach&&(O.reach=ae);var ve=z.prev;ie&&(ve=m(A,ve,ie),K+=ie.length),b(A,ve,V);var xe=new c(F,$?u.tokenize(re,$):re,Z,re);if(z=m(A,ve,xe),Se&&m(A,z,Se),V>1){var Ie={cause:F+","+I,reach:ae};p(k,A,C,z.prev,K,Ie),O&&Ie.reach>O.reach&&(O.reach=Ie.reach)}}}}}}function g(){var k={value:null,prev:null,next:null},A={value:null,prev:k,next:null};k.next=A,this.head=k,this.tail=A,this.length=0}function m(k,A,C){var N=A.next,_={value:C,prev:A,next:N};return A.next=_,N.prev=_,k.length++,_}function b(k,A,C){for(var N=A.next,_=0;_<C&&N!==k.tail;_++)N=N.next;A.next=N,N.prev=A,k.length-=_}function y(k){for(var A=[],C=k.head.next;C!==k.tail;)A.push(C.value),C=C.next;return A}if(!r.document)return r.addEventListener&&(u.disableWorkerMessageHandler||r.addEventListener("message",function(k){var A=JSON.parse(k.data),C=A.language,N=A.code,_=A.immediateClose;r.postMessage(u.highlight(N,u.languages[C],C)),_&&r.close()},!1)),u;var S=u.util.currentScript();S&&(u.filename=S.src,S.hasAttribute("data-manual")&&(u.manual=!0));function x(){u.manual||u.highlightAll()}if(!u.manual){var R=document.readyState;R==="loading"||R==="interactive"&&S&&S.defer?document.addEventListener("DOMContentLoaded",x):window.requestAnimationFrame?window.requestAnimationFrame(x):window.setTimeout(x,16)}return u}(t);e.exports&&(e.exports=n),typeof df<"u"&&(df.Prism=n)}(ey)),ey.exports}var ty,wI;function Lme(){if(wI)return ty;wI=1,ty=e,e.displayName="markup",e.aliases=["html","mathml","svg","xml","ssml","atom","rss"];function e(t){t.languages.markup={comment:{pattern:/<!--(?:(?!<!--)[\s\S])*?-->/,greedy:!0},prolog:{pattern:/<\?[\s\S]+?\?>/,greedy:!0},doctype:{pattern:/<!DOCTYPE(?:[^>"'[\]]|"[^"]*"|'[^']*')+(?:\[(?:[^<"'\]]|"[^"]*"|'[^']*'|<(?!!--)|<!--(?:[^-]|-(?!->))*-->)*\]\s*)?>/i,greedy:!0,inside:{"internal-subset":{pattern:/(^[^\[]*\[)[\s\S]+(?=\]>$)/,lookbehind:!0,greedy:!0,inside:null},string:{pattern:/"[^"]*"|'[^']*'/,greedy:!0},punctuation:/^<!|>$|[[\]]/,"doctype-tag":/^DOCTYPE/i,name:/[^\s<>'"]+/}},cdata:{pattern:/<!\[CDATA\[[\s\S]*?\]\]>/i,greedy:!0},tag:{pattern:/<\/?(?!\d)[^\s>\/=$<%]+(?:\s(?:\s*[^\s>\/=]+(?:\s*=\s*(?:"[^"]*"|'[^']*'|[^\s'">=]+(?=[\s>]))|(?=[\s/>])))+)?\s*\/?>/,greedy:!0,inside:{tag:{pattern:/^<\/?[^\s>\/]+/,inside:{punctuation:/^<\/?/,namespace:/^[^\s>\/:]+:/}},"special-attr":[],"attr-value":{pattern:/=\s*(?:"[^"]*"|'[^']*'|[^\s'">=]+)/,inside:{punctuation:[{pattern:/^=/,alias:"attr-equals"},/"|'/]}},punctuation:/\/?>/,"attr-name":{pattern:/[^\s>\/]+/,inside:{namespace:/^[^\s>\/:]+:/}}}},entity:[{pattern:/&[\da-z]{1,8};/i,alias:"named-entity"},/&#x?[\da-f]{1,8};/i]},t.languages.markup.tag.inside["attr-value"].inside.entity=t.languages.markup.entity,t.languages.markup.doctype.inside["internal-subset"].inside=t.languages.markup,t.hooks.add("wrap",function(n){n.type==="entity"&&(n.attributes.title=n.content.value.replace(/&amp;/,"&"))}),Object.defineProperty(t.languages.markup.tag,"addInlined",{value:function(r,a){var o={};o["language-"+a]={pattern:/(^<!\[CDATA\[)[\s\S]+?(?=\]\]>$)/i,lookbehind:!0,inside:t.languages[a]},o.cdata=/^<!\[CDATA\[|\]\]>$/i;var s={"included-cdata":{pattern:/<!\[CDATA\[[\s\S]*?\]\]>/i,inside:o}};s["language-"+a]={pattern:/[\s\S]+/,inside:t.languages[a]};var u={};u[r]={pattern:RegExp(/(<__[^>]*>)(?:<!\[CDATA\[(?:[^\]]|\](?!\]>))*\]\]>|(?!<!\[CDATA\[)[\s\S])*?(?=<\/__>)/.source.replace(/__/g,function(){return r}),"i"),lookbehind:!0,greedy:!0,inside:s},t.languages.insertBefore("markup","cdata",u)}}),Object.defineProperty(t.languages.markup.tag,"addAttribute",{value:function(n,r){t.languages.markup.tag.inside["special-attr"].push({pattern:RegExp(/(^|["'\s])/.source+"(?:"+n+")"+/\s*=\s*(?:"[^"]*"|'[^']*'|[^\s'">=]+(?=[\s>]))/.source,"i"),lookbehind:!0,inside:{"attr-name":/^[^\s=]+/,"attr-value":{pattern:/=[\s\S]+/,inside:{value:{pattern:/(^=\s*(["']|(?!["'])))\S[\s\S]*(?=\2$)/,lookbehind:!0,alias:[r,"language-"+r],inside:t.languages[r]},punctuation:[{pattern:/^=/,alias:"attr-equals"},/"|'/]}}}})}}),t.languages.html=t.languages.markup,t.languages.mathml=t.languages.markup,t.languages.svg=t.languages.markup,t.languages.xml=t.languages.extend("markup",{}),t.languages.ssml=t.languages.xml,t.languages.atom=t.languages.xml,t.languages.rss=t.languages.xml}return ty}var ny,xI;function Mme(){if(xI)return ny;xI=1,ny=e,e.displayName="css",e.aliases=[];function e(t){(function(n){var r=/(?:"(?:\\(?:\r\n|[\s\S])|[^"\\\r\n])*"|'(?:\\(?:\r\n|[\s\S])|[^'\\\r\n])*')/;n.languages.css={comment:/\/\*[\s\S]*?\*\//,atrule:{pattern:/@[\w-](?:[^;{\s]|\s+(?![\s{]))*(?:;|(?=\s*\{))/,inside:{rule:/^@[\w-]+/,"selector-function-argument":{pattern:/(\bselector\s*\(\s*(?![\s)]))(?:[^()\s]|\s+(?![\s)])|\((?:[^()]|\([^()]*\))*\))+(?=\s*\))/,lookbehind:!0,alias:"selector"},keyword:{pattern:/(^|[^\w-])(?:and|not|only|or)(?![\w-])/,lookbehind:!0}}},url:{pattern:RegExp("\\burl\\((?:"+r.source+"|"+/(?:[^\\\r\n()"']|\\[\s\S])*/.source+")\\)","i"),greedy:!0,inside:{function:/^url/i,punctuation:/^\(|\)$/,string:{pattern:RegExp("^"+r.source+"$"),alias:"url"}}},selector:{pattern:RegExp(`(^|[{}\\s])[^{}\\s](?:[^{};"'\\s]|\\s+(?![\\s{])|`+r.source+")*(?=\\s*\\{)"),lookbehind:!0},string:{pattern:r,greedy:!0},property:{pattern:/(^|[^-\w\xA0-\uFFFF])(?!\s)[-_a-z\xA0-\uFFFF](?:(?!\s)[-\w\xA0-\uFFFF])*(?=\s*:)/i,lookbehind:!0},important:/!important\b/i,function:{pattern:/(^|[^-a-z0-9])[-a-z0-9]+(?=\()/i,lookbehind:!0},punctuation:/[(){};:,]/},n.languages.css.atrule.inside.rest=n.languages.css;var a=n.languages.markup;a&&(a.tag.addInlined("style","css"),a.tag.addAttribute("style","css"))})(t)}return ny}var ry,kI;function Fme(){if(kI)return ry;kI=1,ry=e,e.displayName="clike",e.aliases=[];function e(t){t.languages.clike={comment:[{pattern:/(^|[^\\])\/\*[\s\S]*?(?:\*\/|$)/,lookbehind:!0,greedy:!0},{pattern:/(^|[^\\:])\/\/.*/,lookbehind:!0,greedy:!0}],string:{pattern:/(["'])(?:\\(?:\r\n|[\s\S])|(?!\1)[^\\\r\n])*\1/,greedy:!0},"class-name":{pattern:/(\b(?:class|extends|implements|instanceof|interface|new|trait)\s+|\bcatch\s+\()[\w.\\]+/i,lookbehind:!0,inside:{punctuation:/[.\\]/}},keyword:/\b(?:break|catch|continue|do|else|finally|for|function|if|in|instanceof|new|null|return|throw|try|while)\b/,boolean:/\b(?:false|true)\b/,function:/\b\w+(?=\()/,number:/\b0x[\da-f]+\b|(?:\b\d+(?:\.\d*)?|\B\.\d+)(?:e[+-]?\d+)?/i,operator:/[<>]=?|[!=]=?=?|--?|\+\+?|&&?|\|\|?|[?*/~^%]/,punctuation:/[{}[\];(),.:]/}}return ry}var ay,TI;function Pme(){if(TI)return ay;TI=1,ay=e,e.displayName="javascript",e.aliases=["js"];function e(t){t.languages.javascript=t.languages.extend("clike",{"class-name":[t.languages.clike["class-name"],{pattern:/(^|[^$\w\xA0-\uFFFF])(?!\s)[_$A-Z\xA0-\uFFFF](?:(?!\s)[$\w\xA0-\uFFFF])*(?=\.(?:constructor|prototype))/,lookbehind:!0}],keyword:[{pattern:/((?:^|\})\s*)catch\b/,lookbehind:!0},{pattern:/(^|[^.]|\.\.\.\s*)\b(?:as|assert(?=\s*\{)|async(?=\s*(?:function\b|\(|[$\w\xA0-\uFFFF]|$))|await|break|case|class|const|continue|debugger|default|delete|do|else|enum|export|extends|finally(?=\s*(?:\{|$))|for|from(?=\s*(?:['"]|$))|function|(?:get|set)(?=\s*(?:[#\[$\w\xA0-\uFFFF]|$))|if|implements|import|in|instanceof|interface|let|new|null|of|package|private|protected|public|return|static|super|switch|this|throw|try|typeof|undefined|var|void|while|with|yield)\b/,lookbehind:!0}],function:/#?(?!\s)[_$a-zA-Z\xA0-\uFFFF](?:(?!\s)[$\w\xA0-\uFFFF])*(?=\s*(?:\.\s*(?:apply|bind|call)\s*)?\()/,number:{pattern:RegExp(/(^|[^\w$])/.source+"(?:"+(/NaN|Infinity/.source+"|"+/0[bB][01]+(?:_[01]+)*n?/.source+"|"+/0[oO][0-7]+(?:_[0-7]+)*n?/.source+"|"+/0[xX][\dA-Fa-f]+(?:_[\dA-Fa-f]+)*n?/.source+"|"+/\d+(?:_\d+)*n/.source+"|"+/(?:\d+(?:_\d+)*(?:\.(?:\d+(?:_\d+)*)?)?|\.\d+(?:_\d+)*)(?:[Ee][+-]?\d+(?:_\d+)*)?/.source)+")"+/(?![\w$])/.source),lookbehind:!0},operator:/--|\+\+|\*\*=?|=>|&&=?|\|\|=?|[!=]==|<<=?|>>>?=?|[-+*/%&|^!=<>]=?|\.{3}|\?\?=?|\?\.?|[~:]/}),t.languages.javascript["class-name"][0].pattern=/(\b(?:class|extends|implements|instanceof|interface|new)\s+)[\w.\\]+/,t.languages.insertBefore("javascript","keyword",{regex:{pattern:/((?:^|[^$\w\xA0-\uFFFF."'\])\s]|\b(?:return|yield))\s*)\/(?:\[(?:[^\]\\\r\n]|\\.)*\]|\\.|[^/\\\[\r\n])+\/[dgimyus]{0,7}(?=(?:\s|\/\*(?:[^*]|\*(?!\/))*\*\/)*(?:$|[\r\n,.;:})\]]|\/\/))/,lookbehind:!0,greedy:!0,inside:{"regex-source":{pattern:/^(\/)[\s\S]+(?=\/[a-z]*$)/,lookbehind:!0,alias:"language-regex",inside:t.languages.regex},"regex-delimiter":/^\/|\/$/,"regex-flags":/^[a-z]+$/}},"function-variable":{pattern:/#?(?!\s)[_$a-zA-Z\xA0-\uFFFF](?:(?!\s)[$\w\xA0-\uFFFF])*(?=\s*[=:]\s*(?:async\s*)?(?:\bfunction\b|(?:\((?:[^()]|\([^()]*\))*\)|(?!\s)[_$a-zA-Z\xA0-\uFFFF](?:(?!\s)[$\w\xA0-\uFFFF])*)\s*=>))/,alias:"function"},parameter:[{pattern:/(function(?:\s+(?!\s)[_$a-zA-Z\xA0-\uFFFF](?:(?!\s)[$\w\xA0-\uFFFF])*)?\s*\(\s*)(?!\s)(?:[^()\s]|\s+(?![\s)])|\([^()]*\))+(?=\s*\))/,lookbehind:!0,inside:t.languages.javascript},{pattern:/(^|[^$\w\xA0-\uFFFF])(?!\s)[_$a-z\xA0-\uFFFF](?:(?!\s)[$\w\xA0-\uFFFF])*(?=\s*=>)/i,lookbehind:!0,inside:t.languages.javascript},{pattern:/(\(\s*)(?!\s)(?:[^()\s]|\s+(?![\s)])|\([^()]*\))+(?=\s*\)\s*=>)/,lookbehind:!0,inside:t.languages.javascript},{pattern:/((?:\b|\s|^)(?!(?:as|async|await|break|case|catch|class|const|continue|debugger|default|delete|do|else|enum|export|extends|finally|for|from|function|get|if|implements|import|in|instanceof|interface|let|new|null|of|package|private|protected|public|return|set|static|super|switch|this|throw|try|typeof|undefined|var|void|while|with|yield)(?![$\w\xA0-\uFFFF]))(?:(?!\s)[_$a-zA-Z\xA0-\uFFFF](?:(?!\s)[$\w\xA0-\uFFFF])*\s*)\(\s*|\]\s*\(\s*)(?!\s)(?:[^()\s]|\s+(?![\s)])|\([^()]*\))+(?=\s*\)\s*\{)/,lookbehind:!0,inside:t.languages.javascript}],constant:/\b[A-Z](?:[A-Z_]|\dx?)*\b/}),t.languages.insertBefore("javascript","string",{hashbang:{pattern:/^#!.*/,greedy:!0,alias:"comment"},"template-string":{pattern:/`(?:\\[\s\S]|\$\{(?:[^{}]|\{(?:[^{}]|\{[^}]*\})*\})+\}|(?!\$\{)[^\\`])*`/,greedy:!0,inside:{"template-punctuation":{pattern:/^`|`$/,alias:"string"},interpolation:{pattern:/((?:^|[^\\])(?:\\{2})*)\$\{(?:[^{}]|\{(?:[^{}]|\{[^}]*\})*\})+\}/,lookbehind:!0,inside:{"interpolation-punctuation":{pattern:/^\$\{|\}$/,alias:"punctuation"},rest:t.languages.javascript}},string:/[\s\S]+/}},"string-property":{pattern:/((?:^|[,{])[ \t]*)(["'])(?:\\(?:\r\n|[\s\S])|(?!\2)[^\\\r\n])*\2(?=\s*:)/m,lookbehind:!0,greedy:!0,alias:"property"}}),t.languages.insertBefore("javascript","operator",{"literal-property":{pattern:/((?:^|[,{])[ \t]*)(?!\s)[_$a-zA-Z\xA0-\uFFFF](?:(?!\s)[$\w\xA0-\uFFFF])*(?=\s*:)/m,lookbehind:!0,alias:"property"}}),t.languages.markup&&(t.languages.markup.tag.addInlined("script","javascript"),t.languages.markup.tag.addAttribute(/on(?:abort|blur|change|click|composition(?:end|start|update)|dblclick|error|focus(?:in|out)?|key(?:down|up)|load|mouse(?:down|enter|leave|move|out|over|up)|reset|resize|scroll|select|slotchange|submit|unload|wheel)/.source,"javascript")),t.languages.js=t.languages.javascript}return ay}var oy,AI;function zme(){if(AI)return oy;AI=1;var e=typeof globalThis=="object"?globalThis:typeof self=="object"?self:typeof window=="object"?window:typeof df=="object"?df:{},t=N();e.Prism={manual:!0,disableWorkerMessageHandler:!0};var n=Rge(),r=Dme(),a=Ime(),o=Lme(),s=Mme(),u=Fme(),c=Pme();t();var d={}.hasOwnProperty;function p(){}p.prototype=a;var g=new p;oy=g,g.highlight=y,g.register=m,g.alias=b,g.registered=S,g.listLanguages=x,m(o),m(s),m(u),m(c),g.util.encode=A,g.Token.stringify=R;function m(_){if(typeof _!="function"||!_.displayName)throw new Error("Expected `function` for `grammar`, got `"+_+"`");g.languages[_.displayName]===void 0&&_(g)}function b(_,O){var F=g.languages,L=_,I,H,$,U;O&&(L={},L[_]=O);for(I in L)for(H=L[I],H=typeof H=="string"?[H]:H,$=H.length,U=-1;++U<$;)F[H[U]]=F[I]}function y(_,O){var F=a.highlight,L;if(typeof _!="string")throw new Error("Expected `string` for `value`, got `"+_+"`");if(g.util.type(O)==="Object")L=O,O=null;else{if(typeof O!="string")throw new Error("Expected `string` for `name`, got `"+O+"`");if(d.call(g.languages,O))L=g.languages[O];else throw new Error("Unknown language: `"+O+"` is not registered")}return F.call(this,_,L,O)}function S(_){if(typeof _!="string")throw new Error("Expected `string` for `language`, got `"+_+"`");return d.call(g.languages,_)}function x(){var _=g.languages,O=[],F;for(F in _)d.call(_,F)&&typeof _[F]=="object"&&O.push(F);return O}function R(_,O,F){var L;return typeof _=="string"?{type:"text",value:_}:g.util.type(_)==="Array"?k(_,O):(L={type:_.type,content:g.Token.stringify(_.content,O,F),tag:"span",classes:["token",_.type],attributes:{},language:O,parent:F},_.alias&&(L.classes=L.classes.concat(_.alias)),g.hooks.run("wrap",L),n(L.tag+"."+L.classes.join("."),C(L.attributes),L.content))}function k(_,O){for(var F=[],L=_.length,I=-1,H;++I<L;)H=_[I],H!==""&&H!==null&&H!==void 0&&F.push(H);for(I=-1,L=F.length;++I<L;)H=F[I],F[I]=g.Token.stringify(H,O,F);return F}function A(_){return _}function C(_){var O;for(O in _)_[O]=r(_[O]);return _}function N(){var _="Prism"in e,O=_?e.Prism:void 0;return F;function F(){_?e.Prism=O:delete e.Prism,_=void 0,O=void 0}}return oy}const Bme={'code[class*="language-"]':{color:"black",background:"none",textShadow:"0 1px white",fontFamily:"Consolas, Monaco, 'Andale Mono', 'Ubuntu Mono', monospace",fontSize:"1em",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",wordWrap:"normal",lineHeight:"1.5",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none"},'pre[class*="language-"]':{color:"black",background:"#f5f2f0",textShadow:"0 1px white",fontFamily:"Consolas, Monaco, 'Andale Mono', 'Ubuntu Mono', monospace",fontSize:"1em",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",wordWrap:"normal",lineHeight:"1.5",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none",padding:"1em",margin:".5em 0",overflow:"auto"},'pre[class*="language-"]::-moz-selection':{textShadow:"none",background:"#b3d4fc"},'pre[class*="language-"] ::-moz-selection':{textShadow:"none",background:"#b3d4fc"},'code[class*="language-"]::-moz-selection':{textShadow:"none",background:"#b3d4fc"},'code[class*="language-"] ::-moz-selection':{textShadow:"none",background:"#b3d4fc"},'pre[class*="language-"]::selection':{textShadow:"none",background:"#b3d4fc"},'pre[class*="language-"] ::selection':{textShadow:"none",background:"#b3d4fc"},'code[class*="language-"]::selection':{textShadow:"none",background:"#b3d4fc"},'code[class*="language-"] ::selection':{textShadow:"none",background:"#b3d4fc"},':not(pre) > code[class*="language-"]':{background:"#f5f2f0",padding:".1em",borderRadius:".3em",whiteSpace:"normal"},comment:{color:"slategray"},prolog:{color:"slategray"},doctype:{color:"slategray"},cdata:{color:"slategray"},punctuation:{color:"#999"},namespace:{Opacity:".7"},property:{color:"#905"},tag:{color:"#905"},boolean:{color:"#905"},number:{color:"#905"},constant:{color:"#905"},symbol:{color:"#905"},deleted:{color:"#905"},selector:{color:"#690"},"attr-name":{color:"#690"},string:{color:"#690"},char:{color:"#690"},builtin:{color:"#690"},inserted:{color:"#690"},operator:{color:"#9a6e3a",background:"hsla(0, 0%, 100%, .5)"},entity:{color:"#9a6e3a",background:"hsla(0, 0%, 100%, .5)",cursor:"help"},url:{color:"#9a6e3a",background:"hsla(0, 0%, 100%, .5)"},".language-css .token.string":{color:"#9a6e3a",background:"hsla(0, 0%, 100%, .5)"},".style .token.string":{color:"#9a6e3a",background:"hsla(0, 0%, 100%, .5)"},atrule:{color:"#07a"},"attr-value":{color:"#07a"},keyword:{color:"#07a"},function:{color:"#DD4A68"},"class-name":{color:"#DD4A68"},regex:{color:"#e90"},important:{color:"#e90",fontWeight:"bold"},variable:{color:"#e90"},bold:{fontWeight:"bold"},italic:{fontStyle:"italic"}};var iy,RI;function Ume(){if(RI)return iy;RI=1,iy=e,e.displayName="abap",e.aliases=[];function e(t){t.languages.abap={comment:/^\*.*/m,string:/(`|')(?:\\.|(?!\1)[^\\\r\n])*\1/,"string-template":{pattern:/([|}])(?:\\.|[^\\|{\r\n])*(?=[|{])/,lookbehind:!0,alias:"string"},"eol-comment":{pattern:/(^|\s)".*/m,lookbehind:!0,alias:"comment"},keyword:{pattern:/(\s|\.|^)(?:SCIENTIFIC_WITH_LEADING_ZERO|SCALE_PRESERVING_SCIENTIFIC|RMC_COMMUNICATION_FAILURE|END-ENHANCEMENT-SECTION|MULTIPLY-CORRESPONDING|SUBTRACT-CORRESPONDING|VERIFICATION-MESSAGE|DIVIDE-CORRESPONDING|ENHANCEMENT-SECTION|CURRENCY_CONVERSION|RMC_SYSTEM_FAILURE|START-OF-SELECTION|MOVE-CORRESPONDING|RMC_INVALID_STATUS|CUSTOMER-FUNCTION|END-OF-DEFINITION|ENHANCEMENT-POINT|SYSTEM-EXCEPTIONS|ADD-CORRESPONDING|SCALE_PRESERVING|SELECTION-SCREEN|CURSOR-SELECTION|END-OF-SELECTION|LOAD-OF-PROGRAM|SCROLL-BOUNDARY|SELECTION-TABLE|EXCEPTION-TABLE|IMPLEMENTATIONS|PARAMETER-TABLE|RIGHT-JUSTIFIED|UNIT_CONVERSION|AUTHORITY-CHECK|LIST-PROCESSING|SIGN_AS_POSTFIX|COL_BACKGROUND|IMPLEMENTATION|INTERFACE-POOL|TRANSFORMATION|IDENTIFICATION|ENDENHANCEMENT|LINE-SELECTION|INITIALIZATION|LEFT-JUSTIFIED|SELECT-OPTIONS|SELECTION-SETS|COMMUNICATION|CORRESPONDING|DECIMAL_SHIFT|PRINT-CONTROL|VALUE-REQUEST|CHAIN-REQUEST|FUNCTION-POOL|FIELD-SYMBOLS|FUNCTIONALITY|INVERTED-DATE|SELECTION-SET|CLASS-METHODS|OUTPUT-LENGTH|CLASS-CODING|COL_NEGATIVE|ERRORMESSAGE|FIELD-GROUPS|HELP-REQUEST|NO-EXTENSION|NO-TOPOFPAGE|REDEFINITION|DISPLAY-MODE|ENDINTERFACE|EXIT-COMMAND|FIELD-SYMBOL|NO-SCROLLING|SHORTDUMP-ID|ACCESSPOLICY|CLASS-EVENTS|COL_POSITIVE|DECLARATIONS|ENHANCEMENTS|FILTER-TABLE|SWITCHSTATES|SYNTAX-CHECK|TRANSPORTING|ASYNCHRONOUS|SYNTAX-TRACE|TOKENIZATION|USER-COMMAND|WITH-HEADING|ABAP-SOURCE|BREAK-POINT|CHAIN-INPUT|COMPRESSION|FIXED-POINT|NEW-SECTION|NON-UNICODE|OCCURRENCES|RESPONSIBLE|SYSTEM-CALL|TRACE-TABLE|ABBREVIATED|CHAR-TO-HEX|END-OF-FILE|ENDFUNCTION|ENVIRONMENT|ASSOCIATION|COL_HEADING|EDITOR-CALL|END-OF-PAGE|ENGINEERING|IMPLEMENTED|INTENSIFIED|RADIOBUTTON|SYSTEM-EXIT|TOP-OF-PAGE|TRANSACTION|APPLICATION|CONCATENATE|DESTINATION|ENHANCEMENT|IMMEDIATELY|NO-GROUPING|PRECOMPILED|REPLACEMENT|TITLE-LINES|ACTIVATION|BYTE-ORDER|CLASS-POOL|CONNECTION|CONVERSION|DEFINITION|DEPARTMENT|EXPIRATION|INHERITING|MESSAGE-ID|NO-HEADING|PERFORMING|QUEUE-ONLY|RIGHTSPACE|SCIENTIFIC|STATUSINFO|STRUCTURES|SYNCPOINTS|WITH-TITLE|ATTRIBUTES|BOUNDARIES|CLASS-DATA|COL_NORMAL|DD\/MM\/YYYY|DESCENDING|INTERFACES|LINE-COUNT|MM\/DD\/YYYY|NON-UNIQUE|PRESERVING|SELECTIONS|STATEMENTS|SUBROUTINE|TRUNCATION|TYPE-POOLS|ARITHMETIC|BACKGROUND|ENDPROVIDE|EXCEPTIONS|IDENTIFIER|INDEX-LINE|OBLIGATORY|PARAMETERS|PERCENTAGE|PUSHBUTTON|RESOLUTION|COMPONENTS|DEALLOCATE|DISCONNECT|DUPLICATES|FIRST-LINE|HEAD-LINES|NO-DISPLAY|OCCURRENCE|RESPECTING|RETURNCODE|SUBMATCHES|TRACE-FILE|ASCENDING|BYPASSING|ENDMODULE|EXCEPTION|EXCLUDING|EXPORTING|INCREMENT|MATCHCODE|PARAMETER|PARTIALLY|PREFERRED|REFERENCE|REPLACING|RETURNING|SELECTION|SEPARATED|SPECIFIED|STATEMENT|TIMESTAMP|TYPE-POOL|ACCEPTING|APPENDAGE|ASSIGNING|COL_GROUP|COMPARING|CONSTANTS|DANGEROUS|IMPORTING|INSTANCES|LEFTSPACE|LOG-POINT|QUICKINFO|READ-ONLY|SCROLLING|SQLSCRIPT|STEP-LOOP|TOP-LINES|TRANSLATE|APPENDING|AUTHORITY|CHARACTER|COMPONENT|CONDITION|DIRECTORY|DUPLICATE|MESSAGING|RECEIVING|SUBSCREEN|ACCORDING|COL_TOTAL|END-LINES|ENDMETHOD|ENDSELECT|EXPANDING|EXTENSION|INCLUDING|INFOTYPES|INTERFACE|INTERVALS|LINE-SIZE|PF-STATUS|PROCEDURE|PROTECTED|REQUESTED|RESUMABLE|RIGHTPLUS|SAP-SPOOL|SECONDARY|STRUCTURE|SUBSTRING|TABLEVIEW|NUMOFCHAR|ADJACENT|ANALYSIS|ASSIGNED|BACKWARD|CHANNELS|CHECKBOX|CONTINUE|CRITICAL|DATAINFO|DD\/MM\/YY|DURATION|ENCODING|ENDCLASS|FUNCTION|LEFTPLUS|LINEFEED|MM\/DD\/YY|OVERFLOW|RECEIVED|SKIPPING|SORTABLE|STANDARD|SUBTRACT|SUPPRESS|TABSTRIP|TITLEBAR|TRUNCATE|UNASSIGN|WHENEVER|ANALYZER|COALESCE|COMMENTS|CONDENSE|DECIMALS|DEFERRED|ENDWHILE|EXPLICIT|KEYWORDS|MESSAGES|POSITION|PRIORITY|RECEIVER|RENAMING|TIMEZONE|TRAILING|ALLOCATE|CENTERED|CIRCULAR|CONTROLS|CURRENCY|DELETING|DESCRIBE|DISTANCE|ENDCATCH|EXPONENT|EXTENDED|GENERATE|IGNORING|INCLUDES|INTERNAL|MAJOR-ID|MODIFIER|NEW-LINE|OPTIONAL|PROPERTY|ROLLBACK|STARTING|SUPPLIED|ABSTRACT|CHANGING|CONTEXTS|CREATING|CUSTOMER|DATABASE|DAYLIGHT|DEFINING|DISTINCT|DIVISION|ENABLING|ENDCHAIN|ESCAPING|HARMLESS|IMPLICIT|INACTIVE|LANGUAGE|MINOR-ID|MULTIPLY|NEW-PAGE|NO-TITLE|POS_HIGH|SEPARATE|TEXTPOOL|TRANSFER|SELECTOR|DBMAXLEN|ITERATOR|ARCHIVE|BIT-XOR|BYTE-CO|COLLECT|COMMENT|CURRENT|DEFAULT|DISPLAY|ENDFORM|EXTRACT|LEADING|LISTBOX|LOCATOR|MEMBERS|METHODS|NESTING|POS_LOW|PROCESS|PROVIDE|RAISING|RESERVE|SECONDS|SUMMARY|VISIBLE|BETWEEN|BIT-AND|BYTE-CS|CLEANUP|COMPUTE|CONTROL|CONVERT|DATASET|ENDCASE|FORWARD|HEADERS|HOTSPOT|INCLUDE|INVERSE|KEEPING|NO-ZERO|OBJECTS|OVERLAY|PADDING|PATTERN|PROGRAM|REFRESH|SECTION|SUMMING|TESTING|VERSION|WINDOWS|WITHOUT|BIT-NOT|BYTE-CA|BYTE-NA|CASTING|CONTEXT|COUNTRY|DYNAMIC|ENABLED|ENDLOOP|EXECUTE|FRIENDS|HANDLER|HEADING|INITIAL|\*-INPUT|LOGFILE|MAXIMUM|MINIMUM|NO-GAPS|NO-SIGN|PRAGMAS|PRIMARY|PRIVATE|REDUCED|REPLACE|REQUEST|RESULTS|UNICODE|WARNING|ALIASES|BYTE-CN|BYTE-NS|CALLING|COL_KEY|COLUMNS|CONNECT|ENDEXEC|ENTRIES|EXCLUDE|FILTERS|FURTHER|HELP-ID|LOGICAL|MAPPING|MESSAGE|NAMETAB|OPTIONS|PACKAGE|PERFORM|RECEIVE|STATICS|VARYING|BINDING|CHARLEN|GREATER|XSTRLEN|ACCEPT|APPEND|DETAIL|ELSEIF|ENDING|ENDTRY|FORMAT|FRAMES|GIVING|HASHED|HEADER|IMPORT|INSERT|MARGIN|MODULE|NATIVE|OBJECT|OFFSET|REMOTE|RESUME|SAVING|SIMPLE|SUBMIT|TABBED|TOKENS|UNIQUE|UNPACK|UPDATE|WINDOW|YELLOW|ACTUAL|ASPECT|CENTER|CURSOR|DELETE|DIALOG|DIVIDE|DURING|ERRORS|EVENTS|EXTEND|FILTER|HANDLE|HAVING|IGNORE|LITTLE|MEMORY|NO-GAP|OCCURS|OPTION|PERSON|PLACES|PUBLIC|REDUCE|REPORT|RESULT|SINGLE|SORTED|SWITCH|SYNTAX|TARGET|VALUES|WRITER|ASSERT|BLOCKS|BOUNDS|BUFFER|CHANGE|COLUMN|COMMIT|CONCAT|COPIES|CREATE|DDMMYY|DEFINE|ENDIAN|ESCAPE|EXPAND|KERNEL|LAYOUT|LEGACY|LEVELS|MMDDYY|NUMBER|OUTPUT|RANGES|READER|RETURN|SCREEN|SEARCH|SELECT|SHARED|SOURCE|STABLE|STATIC|SUBKEY|SUFFIX|TABLES|UNWIND|YYMMDD|ASSIGN|BACKUP|BEFORE|BINARY|BIT-OR|BLANKS|CLIENT|CODING|COMMON|DEMAND|DYNPRO|EXCEPT|EXISTS|EXPORT|FIELDS|GLOBAL|GROUPS|LENGTH|LOCALE|MEDIUM|METHOD|MODIFY|NESTED|OTHERS|REJECT|SCROLL|SUPPLY|SYMBOL|ENDFOR|STRLEN|ALIGN|BEGIN|BOUND|ENDAT|ENTRY|EVENT|FINAL|FLUSH|GRANT|INNER|SHORT|USING|WRITE|AFTER|BLACK|BLOCK|CLOCK|COLOR|COUNT|DUMMY|EMPTY|ENDDO|ENDON|GREEN|INDEX|INOUT|LEAVE|LEVEL|LINES|MODIF|ORDER|OUTER|RANGE|RESET|RETRY|RIGHT|SMART|SPLIT|STYLE|TABLE|THROW|UNDER|UNTIL|UPPER|UTF-8|WHERE|ALIAS|BLANK|CLEAR|CLOSE|EXACT|FETCH|FIRST|FOUND|GROUP|LLANG|LOCAL|OTHER|REGEX|SPOOL|TITLE|TYPES|VALID|WHILE|ALPHA|BOXED|CATCH|CHAIN|CHECK|CLASS|COVER|ENDIF|EQUIV|FIELD|FLOOR|FRAME|INPUT|LOWER|MATCH|NODES|PAGES|PRINT|RAISE|ROUND|SHIFT|SPACE|SPOTS|STAMP|STATE|TASKS|TIMES|TRMAC|ULINE|UNION|VALUE|WIDTH|EQUAL|LOG10|TRUNC|BLOB|CASE|CEIL|CLOB|COND|EXIT|FILE|GAPS|HOLD|INCL|INTO|KEEP|KEYS|LAST|LINE|LONG|LPAD|MAIL|MODE|OPEN|PINK|READ|ROWS|TEST|THEN|ZERO|AREA|BACK|BADI|BYTE|CAST|EDIT|EXEC|FAIL|FIND|FKEQ|FONT|FREE|GKEQ|HIDE|INIT|ITNO|LATE|LOOP|MAIN|MARK|MOVE|NEXT|NULL|RISK|ROLE|UNIT|WAIT|ZONE|BASE|CALL|CODE|DATA|DATE|FKGE|GKGE|HIGH|KIND|LEFT|LIST|MASK|MESH|NAME|NODE|PACK|PAGE|POOL|SEND|SIGN|SIZE|SOME|STOP|TASK|TEXT|TIME|USER|VARY|WITH|WORD|BLUE|CONV|COPY|DEEP|ELSE|FORM|FROM|HINT|ICON|JOIN|LIKE|LOAD|ONLY|PART|SCAN|SKIP|SORT|TYPE|UNIX|VIEW|WHEN|WORK|ACOS|ASIN|ATAN|COSH|EACH|FRAC|LESS|RTTI|SINH|SQRT|TANH|AVG|BIT|DIV|ISO|LET|OUT|PAD|SQL|ALL|CI_|CPI|END|LOB|LPI|MAX|MIN|NEW|OLE|RUN|SET|\?TO|YES|ABS|ADD|AND|BIG|FOR|HDB|JOB|LOW|NOT|SAP|TRY|VIA|XML|ANY|GET|IDS|KEY|MOD|OFF|PUT|RAW|RED|REF|SUM|TAB|XSD|CNT|COS|EXP|LOG|SIN|TAN|XOR|AT|CO|CP|DO|GT|ID|IF|NS|OR|BT|CA|CS|GE|NA|NB|EQ|IN|LT|NE|NO|OF|ON|PF|TO|AS|BY|CN|IS|LE|NP|UP|E|I|M|O|Z|C|X)\b/i,lookbehind:!0},number:/\b\d+\b/,operator:{pattern:/(\s)(?:\*\*?|<[=>]?|>=?|\?=|[-+\/=])(?=\s)/,lookbehind:!0},"string-operator":{pattern:/(\s)&&?(?=\s)/,lookbehind:!0,alias:"keyword"},"token-operator":[{pattern:/(\w)(?:->?|=>|[~|{}])(?=\w)/,lookbehind:!0,alias:"punctuation"},{pattern:/[|{}]/,alias:"punctuation"}],punctuation:/[,.:()]/}}return iy}var sy,_I;function jme(){if(_I)return sy;_I=1,sy=e,e.displayName="abnf",e.aliases=[];function e(t){(function(n){var r="(?:ALPHA|BIT|CHAR|CR|CRLF|CTL|DIGIT|DQUOTE|HEXDIG|HTAB|LF|LWSP|OCTET|SP|VCHAR|WSP)";n.languages.abnf={comment:/;.*/,string:{pattern:/(?:%[is])?"[^"\n\r]*"/,greedy:!0,inside:{punctuation:/^%[is]/}},range:{pattern:/%(?:b[01]+-[01]+|d\d+-\d+|x[A-F\d]+-[A-F\d]+)/i,alias:"number"},terminal:{pattern:/%(?:b[01]+(?:\.[01]+)*|d\d+(?:\.\d+)*|x[A-F\d]+(?:\.[A-F\d]+)*)/i,alias:"number"},repetition:{pattern:/(^|[^\w-])(?:\d*\*\d*|\d+)/,lookbehind:!0,alias:"operator"},definition:{pattern:/(^[ \t]*)(?:[a-z][\w-]*|<[^<>\r\n]*>)(?=\s*=)/m,lookbehind:!0,alias:"keyword",inside:{punctuation:/<|>/}},"core-rule":{pattern:RegExp("(?:(^|[^<\\w-])"+r+"|<"+r+">)(?![\\w-])","i"),lookbehind:!0,alias:["rule","constant"],inside:{punctuation:/<|>/}},rule:{pattern:/(^|[^<\w-])[a-z][\w-]*|<[^<>\r\n]*>/i,lookbehind:!0,inside:{punctuation:/<|>/}},operator:/=\/?|\//,punctuation:/[()\[\]]/}})(t)}return sy}var ly,CI;function Gme(){if(CI)return ly;CI=1,ly=e,e.displayName="actionscript",e.aliases=[];function e(t){t.languages.actionscript=t.languages.extend("javascript",{keyword:/\b(?:as|break|case|catch|class|const|default|delete|do|dynamic|each|else|extends|final|finally|for|function|get|if|implements|import|in|include|instanceof|interface|internal|is|namespace|native|new|null|override|package|private|protected|public|return|set|static|super|switch|this|throw|try|typeof|use|var|void|while|with)\b/,operator:/\+\+|--|(?:[+\-*\/%^]|&&?|\|\|?|<<?|>>?>?|[!=]=?)=?|[~?@]/}),t.languages.actionscript["class-name"].alias="function",delete t.languages.actionscript.parameter,delete t.languages.actionscript["literal-property"],t.languages.markup&&t.languages.insertBefore("actionscript","string",{xml:{pattern:/(^|[^.])<\/?\w+(?:\s+[^\s>\/=]+=("|')(?:\\[\s\S]|(?!\2)[^\\])*\2)*\s*\/?>/,lookbehind:!0,inside:t.languages.markup}})}return ly}var uy,NI;function Hme(){if(NI)return uy;NI=1,uy=e,e.displayName="ada",e.aliases=[];function e(t){t.languages.ada={comment:/--.*/,string:/"(?:""|[^"\r\f\n])*"/,number:[{pattern:/\b\d(?:_?\d)*#[\dA-F](?:_?[\dA-F])*(?:\.[\dA-F](?:_?[\dA-F])*)?#(?:E[+-]?\d(?:_?\d)*)?/i},{pattern:/\b\d(?:_?\d)*(?:\.\d(?:_?\d)*)?(?:E[+-]?\d(?:_?\d)*)?\b/i}],"attr-name":/\b'\w+/,keyword:/\b(?:abort|abs|abstract|accept|access|aliased|all|and|array|at|begin|body|case|constant|declare|delay|delta|digits|do|else|elsif|end|entry|exception|exit|for|function|generic|goto|if|in|interface|is|limited|loop|mod|new|not|null|of|others|out|overriding|package|pragma|private|procedure|protected|raise|range|record|rem|renames|requeue|return|reverse|select|separate|some|subtype|synchronized|tagged|task|terminate|then|type|until|use|when|while|with|xor)\b/i,boolean:/\b(?:false|true)\b/i,operator:/<[=>]?|>=?|=>?|:=|\/=?|\*\*?|[&+-]/,punctuation:/\.\.?|[,;():]/,char:/'.'/,variable:/\b[a-z](?:\w)*\b/i}}return uy}var cy,OI;function $me(){if(OI)return cy;OI=1,cy=e,e.displayName="agda",e.aliases=[];function e(t){(function(n){n.languages.agda={comment:/\{-[\s\S]*?(?:-\}|$)|--.*/,string:{pattern:/"(?:\\(?:\r\n|[\s\S])|[^\\\r\n"])*"/,greedy:!0},punctuation:/[(){}⦃⦄.;@]/,"class-name":{pattern:/((?:data|record) +)\S+/,lookbehind:!0},function:{pattern:/(^[ \t]*)(?!\s)[^:\r\n]+(?=:)/m,lookbehind:!0},operator:{pattern:/(^\s*|\s)(?:[=|:∀→λ\\?_]|->)(?=\s)/,lookbehind:!0},keyword:/\b(?:Set|abstract|constructor|data|eta-equality|field|forall|hiding|import|in|inductive|infix|infixl|infixr|instance|let|macro|module|mutual|no-eta-equality|open|overlap|pattern|postulate|primitive|private|public|quote|quoteContext|quoteGoal|quoteTerm|record|renaming|rewrite|syntax|tactic|unquote|unquoteDecl|unquoteDef|using|variable|where|with)\b/}})(t)}return cy}var dy,DI;function qme(){if(DI)return dy;DI=1,dy=e,e.displayName="al",e.aliases=[];function e(t){t.languages.al={comment:/\/\/.*|\/\*[\s\S]*?\*\//,string:{pattern:/'(?:''|[^'\r\n])*'(?!')|"(?:""|[^"\r\n])*"(?!")/,greedy:!0},function:{pattern:/(\b(?:event|procedure|trigger)\s+|(?:^|[^.])\.\s*)[a-z_]\w*(?=\s*\()/i,lookbehind:!0},keyword:[/\b(?:array|asserterror|begin|break|case|do|downto|else|end|event|exit|for|foreach|function|if|implements|in|indataset|interface|internal|local|of|procedure|program|protected|repeat|runonclient|securityfiltering|suppressdispose|temporary|then|to|trigger|until|var|while|with|withevents)\b/i,/\b(?:action|actions|addafter|addbefore|addfirst|addlast|area|assembly|chartpart|codeunit|column|controladdin|cuegroup|customizes|dataitem|dataset|dotnet|elements|enum|enumextension|extends|field|fieldattribute|fieldelement|fieldgroup|fieldgroups|fields|filter|fixed|grid|group|key|keys|label|labels|layout|modify|moveafter|movebefore|movefirst|movelast|page|pagecustomization|pageextension|part|profile|query|repeater|report|requestpage|schema|separator|systempart|table|tableelement|tableextension|textattribute|textelement|type|usercontrol|value|xmlport)\b/i],number:/\b(?:0x[\da-f]+|(?:\d+(?:\.\d*)?|\.\d+)(?:e[+-]?\d+)?)(?:F|LL?|U(?:LL?)?)?\b/i,boolean:/\b(?:false|true)\b/i,variable:/\b(?:Curr(?:FieldNo|Page|Report)|x?Rec|RequestOptionsPage)\b/,"class-name":/\b(?:automation|biginteger|bigtext|blob|boolean|byte|char|clienttype|code|completiontriggererrorlevel|connectiontype|database|dataclassification|datascope|date|dateformula|datetime|decimal|defaultlayout|dialog|dictionary|dotnetassembly|dotnettypedeclaration|duration|errorinfo|errortype|executioncontext|executionmode|fieldclass|fieldref|fieldtype|file|filterpagebuilder|guid|httpclient|httpcontent|httpheaders|httprequestmessage|httpresponsemessage|instream|integer|joker|jsonarray|jsonobject|jsontoken|jsonvalue|keyref|list|moduledependencyinfo|moduleinfo|none|notification|notificationscope|objecttype|option|outstream|pageresult|record|recordid|recordref|reportformat|securityfilter|sessionsettings|tableconnectiontype|tablefilter|testaction|testfield|testfilterfield|testpage|testpermissions|testrequestpage|text|textbuilder|textconst|textencoding|time|transactionmodel|transactiontype|variant|verbosity|version|view|views|webserviceactioncontext|webserviceactionresultcode|xmlattribute|xmlattributecollection|xmlcdata|xmlcomment|xmldeclaration|xmldocument|xmldocumenttype|xmlelement|xmlnamespacemanager|xmlnametable|xmlnode|xmlnodelist|xmlprocessinginstruction|xmlreadoptions|xmltext|xmlwriteoptions)\b/i,operator:/\.\.|:[=:]|[-+*/]=?|<>|[<>]=?|=|\b(?:and|div|mod|not|or|xor)\b/i,punctuation:/[()\[\]{}:.;,]/}}return dy}var fy,II;function Vme(){if(II)return fy;II=1,fy=e,e.displayName="antlr4",e.aliases=["g4"];function e(t){t.languages.antlr4={comment:/\/\/.*|\/\*[\s\S]*?(?:\*\/|$)/,string:{pattern:/'(?:\\.|[^\\'\r\n])*'/,greedy:!0},"character-class":{pattern:/\[(?:\\.|[^\\\]\r\n])*\]/,greedy:!0,alias:"regex",inside:{range:{pattern:/([^[]|(?:^|[^\\])(?:\\\\)*\\\[)-(?!\])/,lookbehind:!0,alias:"punctuation"},escape:/\\(?:u(?:[a-fA-F\d]{4}|\{[a-fA-F\d]+\})|[pP]\{[=\w-]+\}|[^\r\nupP])/,punctuation:/[\[\]]/}},action:{pattern:/\{(?:[^{}]|\{(?:[^{}]|\{(?:[^{}]|\{[^{}]*\})*\})*\})*\}/,greedy:!0,inside:{content:{pattern:/(\{)[\s\S]+(?=\})/,lookbehind:!0},punctuation:/[{}]/}},command:{pattern:/(->\s*(?!\s))(?:\s*(?:,\s*)?\b[a-z]\w*(?:\s*\([^()\r\n]*\))?)+(?=\s*;)/i,lookbehind:!0,inside:{function:/\b\w+(?=\s*(?:[,(]|$))/,punctuation:/[,()]/}},annotation:{pattern:/@\w+(?:::\w+)*/,alias:"keyword"},label:{pattern:/#[ \t]*\w+/,alias:"punctuation"},keyword:/\b(?:catch|channels|finally|fragment|grammar|import|lexer|locals|mode|options|parser|returns|throws|tokens)\b/,definition:[{pattern:/\b[a-z]\w*(?=\s*:)/,alias:["rule","class-name"]},{pattern:/\b[A-Z]\w*(?=\s*:)/,alias:["token","constant"]}],constant:/\b[A-Z][A-Z_]*\b/,operator:/\.\.|->|[|~]|[*+?]\??/,punctuation:/[;:()=]/},t.languages.g4=t.languages.antlr4}return fy}var py,LI;function Wme(){if(LI)return py;LI=1,py=e,e.displayName="apacheconf",e.aliases=[];function e(t){t.languages.apacheconf={comment:/#.*/,"directive-inline":{pattern:/(^[\t ]*)\b(?:AcceptFilter|AcceptPathInfo|AccessFileName|Action|Add(?:Alt|AltByEncoding|AltByType|Charset|DefaultCharset|Description|Encoding|Handler|Icon|IconByEncoding|IconByType|InputFilter|Language|ModuleInfo|OutputFilter|OutputFilterByType|Type)|Alias|AliasMatch|Allow(?:CONNECT|EncodedSlashes|Methods|Override|OverrideList)?|Anonymous(?:_LogEmail|_MustGiveEmail|_NoUserID|_VerifyEmail)?|AsyncRequestWorkerFactor|Auth(?:BasicAuthoritative|BasicFake|BasicProvider|BasicUseDigestAlgorithm|DBDUserPWQuery|DBDUserRealmQuery|DBMGroupFile|DBMType|DBMUserFile|Digest(?:Algorithm|Domain|NonceLifetime|Provider|Qop|ShmemSize)|Form(?:Authoritative|Body|DisableNoStore|FakeBasicAuth|Location|LoginRequiredLocation|LoginSuccessLocation|LogoutLocation|Method|Mimetype|Password|Provider|SitePassphrase|Size|Username)|GroupFile|LDAP(?:AuthorizePrefix|BindAuthoritative|BindDN|BindPassword|CharsetConfig|CompareAsUser|CompareDNOnServer|DereferenceAliases|GroupAttribute|GroupAttributeIsDN|InitialBindAsUser|InitialBindPattern|MaxSubGroupDepth|RemoteUserAttribute|RemoteUserIsDN|SearchAsUser|SubGroupAttribute|SubGroupClass|Url)|Merging|Name|nCache(?:Context|Enable|ProvideFor|SOCache|Timeout)|nzFcgiCheckAuthnProvider|nzFcgiDefineProvider|Type|UserFile|zDBDLoginToReferer|zDBDQuery|zDBDRedirectQuery|zDBMType|zSendForbiddenOnFailure)|BalancerGrowth|BalancerInherit|BalancerMember|BalancerPersist|BrowserMatch|BrowserMatchNoCase|BufferedLogs|BufferSize|Cache(?:DefaultExpire|DetailHeader|DirLength|DirLevels|Disable|Enable|File|Header|IgnoreCacheControl|IgnoreHeaders|IgnoreNoLastMod|IgnoreQueryString|IgnoreURLSessionIdentifiers|KeyBaseURL|LastModifiedFactor|Lock|LockMaxAge|LockPath|MaxExpire|MaxFileSize|MinExpire|MinFileSize|NegotiatedDocs|QuickHandler|ReadSize|ReadTime|Root|Socache(?:MaxSize|MaxTime|MinTime|ReadSize|ReadTime)?|StaleOnError|StoreExpired|StoreNoStore|StorePrivate)|CGIDScriptTimeout|CGIMapExtension|CharsetDefault|CharsetOptions|CharsetSourceEnc|CheckCaseOnly|CheckSpelling|ChrootDir|ContentDigest|CookieDomain|CookieExpires|CookieName|CookieStyle|CookieTracking|CoreDumpDirectory|CustomLog|Dav|DavDepthInfinity|DavGenericLockDB|DavLockDB|DavMinTimeout|DBDExptime|DBDInitSQL|DBDKeep|DBDMax|DBDMin|DBDParams|DBDPersist|DBDPrepareSQL|DBDriver|DefaultIcon|DefaultLanguage|DefaultRuntimeDir|DefaultType|Define|Deflate(?:BufferSize|CompressionLevel|FilterNote|InflateLimitRequestBody|InflateRatio(?:Burst|Limit)|MemLevel|WindowSize)|Deny|DirectoryCheckHandler|DirectoryIndex|DirectoryIndexRedirect|DirectorySlash|DocumentRoot|DTracePrivileges|DumpIOInput|DumpIOOutput|EnableExceptionHook|EnableMMAP|EnableSendfile|Error|ErrorDocument|ErrorLog|ErrorLogFormat|Example|ExpiresActive|ExpiresByType|ExpiresDefault|ExtendedStatus|ExtFilterDefine|ExtFilterOptions|FallbackResource|FileETag|FilterChain|FilterDeclare|FilterProtocol|FilterProvider|FilterTrace|ForceLanguagePriority|ForceType|ForensicLog|GprofDir|GracefulShutdownTimeout|Group|Header|HeaderName|Heartbeat(?:Address|Listen|MaxServers|Storage)|HostnameLookups|IdentityCheck|IdentityCheckTimeout|ImapBase|ImapDefault|ImapMenu|Include|IncludeOptional|Index(?:HeadInsert|Ignore|IgnoreReset|Options|OrderDefault|StyleSheet)|InputSed|ISAPI(?:AppendLogToErrors|AppendLogToQuery|CacheFile|FakeAsync|LogNotSupported|ReadAheadBuffer)|KeepAlive|KeepAliveTimeout|KeptBodySize|LanguagePriority|LDAP(?:CacheEntries|CacheTTL|ConnectionPoolTTL|ConnectionTimeout|LibraryDebug|OpCacheEntries|OpCacheTTL|ReferralHopLimit|Referrals|Retries|RetryDelay|SharedCacheFile|SharedCacheSize|Timeout|TrustedClientCert|TrustedGlobalCert|TrustedMode|VerifyServerCert)|Limit(?:InternalRecursion|Request(?:Body|Fields|FieldSize|Line)|XMLRequestBody)|Listen|ListenBackLog|LoadFile|LoadModule|LogFormat|LogLevel|LogMessage|LuaAuthzProvider|LuaCodeCache|Lua(?:Hook(?:AccessChecker|AuthChecker|CheckUserID|Fixups|InsertFilter|Log|MapToStorage|TranslateName|TypeChecker)|Inherit|InputFilter|MapHandler|OutputFilter|PackageCPath|PackagePath|QuickHandler|Root|Scope)|Max(?:ConnectionsPerChild|KeepAliveRequests|MemFree|RangeOverlaps|RangeReversals|Ranges|RequestWorkers|SpareServers|SpareThreads|Threads)|MergeTrailers|MetaDir|MetaFiles|MetaSuffix|MimeMagicFile|MinSpareServers|MinSpareThreads|MMapFile|ModemStandard|ModMimeUsePathInfo|MultiviewsMatch|Mutex|NameVirtualHost|NoProxy|NWSSLTrustedCerts|NWSSLUpgradeable|Options|Order|OutputSed|PassEnv|PidFile|PrivilegesMode|Protocol|ProtocolEcho|Proxy(?:AddHeaders|BadHeader|Block|Domain|ErrorOverride|ExpressDBMFile|ExpressDBMType|ExpressEnable|FtpDirCharset|FtpEscapeWildcards|FtpListOnWildcard|HTML(?:BufSize|CharsetOut|DocType|Enable|Events|Extended|Fixups|Interp|Links|Meta|StripComments|URLMap)|IOBufferSize|MaxForwards|Pass(?:Inherit|InterpolateEnv|Match|Reverse|ReverseCookieDomain|ReverseCookiePath)?|PreserveHost|ReceiveBufferSize|Remote|RemoteMatch|Requests|SCGIInternalRedirect|SCGISendfile|Set|SourceAddress|Status|Timeout|Via)|ReadmeName|ReceiveBufferSize|Redirect|RedirectMatch|RedirectPermanent|RedirectTemp|ReflectorHeader|RemoteIP(?:Header|InternalProxy|InternalProxyList|ProxiesHeader|TrustedProxy|TrustedProxyList)|RemoveCharset|RemoveEncoding|RemoveHandler|RemoveInputFilter|RemoveLanguage|RemoveOutputFilter|RemoveType|RequestHeader|RequestReadTimeout|Require|Rewrite(?:Base|Cond|Engine|Map|Options|Rule)|RLimitCPU|RLimitMEM|RLimitNPROC|Satisfy|ScoreBoardFile|Script(?:Alias|AliasMatch|InterpreterSource|Log|LogBuffer|LogLength|Sock)?|SecureListen|SeeRequestTail|SendBufferSize|Server(?:Admin|Alias|Limit|Name|Path|Root|Signature|Tokens)|Session(?:Cookie(?:Name|Name2|Remove)|Crypto(?:Cipher|Driver|Passphrase|PassphraseFile)|DBD(?:CookieName|CookieName2|CookieRemove|DeleteLabel|InsertLabel|PerUser|SelectLabel|UpdateLabel)|Env|Exclude|Header|Include|MaxAge)?|SetEnv|SetEnvIf|SetEnvIfExpr|SetEnvIfNoCase|SetHandler|SetInputFilter|SetOutputFilter|SSIEndTag|SSIErrorMsg|SSIETag|SSILastModified|SSILegacyExprParser|SSIStartTag|SSITimeFormat|SSIUndefinedEcho|SSL(?:CACertificateFile|CACertificatePath|CADNRequestFile|CADNRequestPath|CARevocationCheck|CARevocationFile|CARevocationPath|CertificateChainFile|CertificateFile|CertificateKeyFile|CipherSuite|Compression|CryptoDevice|Engine|FIPS|HonorCipherOrder|InsecureRenegotiation|OCSP(?:DefaultResponder|Enable|OverrideResponder|ResponderTimeout|ResponseMaxAge|ResponseTimeSkew|UseRequestNonce)|OpenSSLConfCmd|Options|PassPhraseDialog|Protocol|Proxy(?:CACertificateFile|CACertificatePath|CARevocation(?:Check|File|Path)|CheckPeer(?:CN|Expire|Name)|CipherSuite|Engine|MachineCertificate(?:ChainFile|File|Path)|Protocol|Verify|VerifyDepth)|RandomSeed|RenegBufferSize|Require|RequireSSL|Session(?:Cache|CacheTimeout|TicketKeyFile|Tickets)|SRPUnknownUserSeed|SRPVerifierFile|Stapling(?:Cache|ErrorCacheTimeout|FakeTryLater|ForceURL|ResponderTimeout|ResponseMaxAge|ResponseTimeSkew|ReturnResponderErrors|StandardCacheTimeout)|StrictSNIVHostCheck|UserName|UseStapling|VerifyClient|VerifyDepth)|StartServers|StartThreads|Substitute|Suexec|SuexecUserGroup|ThreadLimit|ThreadsPerChild|ThreadStackSize|TimeOut|TraceEnable|TransferLog|TypesConfig|UnDefine|UndefMacro|UnsetEnv|Use|UseCanonicalName|UseCanonicalPhysicalPort|User|UserDir|VHostCGIMode|VHostCGIPrivs|VHostGroup|VHostPrivs|VHostSecure|VHostUser|Virtual(?:DocumentRoot|ScriptAlias)(?:IP)?|WatchdogInterval|XBitHack|xml2EncAlias|xml2EncDefault|xml2StartParse)\b/im,lookbehind:!0,alias:"property"},"directive-block":{pattern:/<\/?\b(?:Auth[nz]ProviderAlias|Directory|DirectoryMatch|Else|ElseIf|Files|FilesMatch|If|IfDefine|IfModule|IfVersion|Limit|LimitExcept|Location|LocationMatch|Macro|Proxy|Require(?:All|Any|None)|VirtualHost)\b.*>/i,inside:{"directive-block":{pattern:/^<\/?\w+/,inside:{punctuation:/^<\/?/},alias:"tag"},"directive-block-parameter":{pattern:/.*[^>]/,inside:{punctuation:/:/,string:{pattern:/("|').*\1/,inside:{variable:/[$%]\{?(?:\w\.?[-+:]?)+\}?/}}},alias:"attr-value"},punctuation:/>/},alias:"tag"},"directive-flags":{pattern:/\[(?:[\w=],?)+\]/,alias:"keyword"},string:{pattern:/("|').*\1/,inside:{variable:/[$%]\{?(?:\w\.?[-+:]?)+\}?/}},variable:/[$%]\{?(?:\w\.?[-+:]?)+\}?/,regex:/\^?.*\$|\^.*\$?/}}return py}var gy,MI;function lA(){if(MI)return gy;MI=1,gy=e,e.displayName="sql",e.aliases=[];function e(t){t.languages.sql={comment:{pattern:/(^|[^\\])(?:\/\*[\s\S]*?\*\/|(?:--|\/\/|#).*)/,lookbehind:!0},variable:[{pattern:/@(["'`])(?:\\[\s\S]|(?!\1)[^\\])+\1/,greedy:!0},/@[\w.$]+/],string:{pattern:/(^|[^@\\])("|')(?:\\[\s\S]|(?!\2)[^\\]|\2\2)*\2/,greedy:!0,lookbehind:!0},identifier:{pattern:/(^|[^@\\])`(?:\\[\s\S]|[^`\\]|``)*`/,greedy:!0,lookbehind:!0,inside:{punctuation:/^`|`$/}},function:/\b(?:AVG|COUNT|FIRST|FORMAT|LAST|LCASE|LEN|MAX|MID|MIN|MOD|NOW|ROUND|SUM|UCASE)(?=\s*\()/i,keyword:/\b(?:ACTION|ADD|AFTER|ALGORITHM|ALL|ALTER|ANALYZE|ANY|APPLY|AS|ASC|AUTHORIZATION|AUTO_INCREMENT|BACKUP|BDB|BEGIN|BERKELEYDB|BIGINT|BINARY|BIT|BLOB|BOOL|BOOLEAN|BREAK|BROWSE|BTREE|BULK|BY|CALL|CASCADED?|CASE|CHAIN|CHAR(?:ACTER|SET)?|CHECK(?:POINT)?|CLOSE|CLUSTERED|COALESCE|COLLATE|COLUMNS?|COMMENT|COMMIT(?:TED)?|COMPUTE|CONNECT|CONSISTENT|CONSTRAINT|CONTAINS(?:TABLE)?|CONTINUE|CONVERT|CREATE|CROSS|CURRENT(?:_DATE|_TIME|_TIMESTAMP|_USER)?|CURSOR|CYCLE|DATA(?:BASES?)?|DATE(?:TIME)?|DAY|DBCC|DEALLOCATE|DEC|DECIMAL|DECLARE|DEFAULT|DEFINER|DELAYED|DELETE|DELIMITERS?|DENY|DESC|DESCRIBE|DETERMINISTIC|DISABLE|DISCARD|DISK|DISTINCT|DISTINCTROW|DISTRIBUTED|DO|DOUBLE|DROP|DUMMY|DUMP(?:FILE)?|DUPLICATE|ELSE(?:IF)?|ENABLE|ENCLOSED|END|ENGINE|ENUM|ERRLVL|ERRORS|ESCAPED?|EXCEPT|EXEC(?:UTE)?|EXISTS|EXIT|EXPLAIN|EXTENDED|FETCH|FIELDS|FILE|FILLFACTOR|FIRST|FIXED|FLOAT|FOLLOWING|FOR(?: EACH ROW)?|FORCE|FOREIGN|FREETEXT(?:TABLE)?|FROM|FULL|FUNCTION|GEOMETRY(?:COLLECTION)?|GLOBAL|GOTO|GRANT|GROUP|HANDLER|HASH|HAVING|HOLDLOCK|HOUR|IDENTITY(?:COL|_INSERT)?|IF|IGNORE|IMPORT|INDEX|INFILE|INNER|INNODB|INOUT|INSERT|INT|INTEGER|INTERSECT|INTERVAL|INTO|INVOKER|ISOLATION|ITERATE|JOIN|KEYS?|KILL|LANGUAGE|LAST|LEAVE|LEFT|LEVEL|LIMIT|LINENO|LINES|LINESTRING|LOAD|LOCAL|LOCK|LONG(?:BLOB|TEXT)|LOOP|MATCH(?:ED)?|MEDIUM(?:BLOB|INT|TEXT)|MERGE|MIDDLEINT|MINUTE|MODE|MODIFIES|MODIFY|MONTH|MULTI(?:LINESTRING|POINT|POLYGON)|NATIONAL|NATURAL|NCHAR|NEXT|NO|NONCLUSTERED|NULLIF|NUMERIC|OFF?|OFFSETS?|ON|OPEN(?:DATASOURCE|QUERY|ROWSET)?|OPTIMIZE|OPTION(?:ALLY)?|ORDER|OUT(?:ER|FILE)?|OVER|PARTIAL|PARTITION|PERCENT|PIVOT|PLAN|POINT|POLYGON|PRECEDING|PRECISION|PREPARE|PREV|PRIMARY|PRINT|PRIVILEGES|PROC(?:EDURE)?|PUBLIC|PURGE|QUICK|RAISERROR|READS?|REAL|RECONFIGURE|REFERENCES|RELEASE|RENAME|REPEAT(?:ABLE)?|REPLACE|REPLICATION|REQUIRE|RESIGNAL|RESTORE|RESTRICT|RETURN(?:ING|S)?|REVOKE|RIGHT|ROLLBACK|ROUTINE|ROW(?:COUNT|GUIDCOL|S)?|RTREE|RULE|SAVE(?:POINT)?|SCHEMA|SECOND|SELECT|SERIAL(?:IZABLE)?|SESSION(?:_USER)?|SET(?:USER)?|SHARE|SHOW|SHUTDOWN|SIMPLE|SMALLINT|SNAPSHOT|SOME|SONAME|SQL|START(?:ING)?|STATISTICS|STATUS|STRIPED|SYSTEM_USER|TABLES?|TABLESPACE|TEMP(?:ORARY|TABLE)?|TERMINATED|TEXT(?:SIZE)?|THEN|TIME(?:STAMP)?|TINY(?:BLOB|INT|TEXT)|TOP?|TRAN(?:SACTIONS?)?|TRIGGER|TRUNCATE|TSEQUAL|TYPES?|UNBOUNDED|UNCOMMITTED|UNDEFINED|UNION|UNIQUE|UNLOCK|UNPIVOT|UNSIGNED|UPDATE(?:TEXT)?|USAGE|USE|USER|USING|VALUES?|VAR(?:BINARY|CHAR|CHARACTER|YING)|VIEW|WAITFOR|WARNINGS|WHEN|WHERE|WHILE|WITH(?: ROLLUP|IN)?|WORK|WRITE(?:TEXT)?|YEAR)\b/i,boolean:/\b(?:FALSE|NULL|TRUE)\b/i,number:/\b0x[\da-f]+\b|\b\d+(?:\.\d*)?|\B\.\d+\b/i,operator:/[-+*\/=%^~]|&&?|\|\|?|!=?|<(?:=>?|<|>)?|>[>=]?|\b(?:AND|BETWEEN|DIV|ILIKE|IN|IS|LIKE|NOT|OR|REGEXP|RLIKE|SOUNDS LIKE|XOR)\b/i,punctuation:/[;[\]()`,.]/}}return gy}var hy,FI;function Yme(){if(FI)return hy;FI=1;var e=lA();hy=t,t.displayName="apex",t.aliases=[];function t(n){n.register(e),function(r){var a=/\b(?:(?:after|before)(?=\s+[a-z])|abstract|activate|and|any|array|as|asc|autonomous|begin|bigdecimal|blob|boolean|break|bulk|by|byte|case|cast|catch|char|class|collect|commit|const|continue|currency|date|datetime|decimal|default|delete|desc|do|double|else|end|enum|exception|exit|export|extends|final|finally|float|for|from|get(?=\s*[{};])|global|goto|group|having|hint|if|implements|import|in|inner|insert|instanceof|int|integer|interface|into|join|like|limit|list|long|loop|map|merge|new|not|null|nulls|number|object|of|on|or|outer|override|package|parallel|pragma|private|protected|public|retrieve|return|rollback|select|set|short|sObject|sort|static|string|super|switch|synchronized|system|testmethod|then|this|throw|time|transaction|transient|trigger|try|undelete|update|upsert|using|virtual|void|webservice|when|where|while|(?:inherited|with|without)\s+sharing)\b/i,o=/\b(?:(?=[a-z_]\w*\s*[<\[])|(?!<keyword>))[A-Z_]\w*(?:\s*\.\s*[A-Z_]\w*)*\b(?:\s*(?:\[\s*\]|<(?:[^<>]|<(?:[^<>]|<[^<>]*>)*>)*>))*/.source.replace(/<keyword>/g,function(){return a.source});function s(c){return RegExp(c.replace(/<CLASS-NAME>/g,function(){return o}),"i")}var u={keyword:a,punctuation:/[()\[\]{};,:.<>]/};r.languages.apex={comment:r.languages.clike.comment,string:r.languages.clike.string,sql:{pattern:/((?:[=,({:]|\breturn)\s*)\[[^\[\]]*\]/i,lookbehind:!0,greedy:!0,alias:"language-sql",inside:r.languages.sql},annotation:{pattern:/@\w+\b/,alias:"punctuation"},"class-name":[{pattern:s(/(\b(?:class|enum|extends|implements|instanceof|interface|new|trigger\s+\w+\s+on)\s+)<CLASS-NAME>/.source),lookbehind:!0,inside:u},{pattern:s(/(\(\s*)<CLASS-NAME>(?=\s*\)\s*[\w(])/.source),lookbehind:!0,inside:u},{pattern:s(/<CLASS-NAME>(?=\s*\w+\s*[;=,(){:])/.source),inside:u}],trigger:{pattern:/(\btrigger\s+)\w+\b/i,lookbehind:!0,alias:"class-name"},keyword:a,function:/\b[a-z_]\w*(?=\s*\()/i,boolean:/\b(?:false|true)\b/i,number:/(?:\B\.\d+|\b\d+(?:\.\d+|L)?)\b/i,operator:/[!=](?:==?)?|\?\.?|&&|\|\||--|\+\+|[-+*/^&|]=?|:|<<?=?|>{1,3}=?/,punctuation:/[()\[\]{};,.]/}}(n)}return hy}var my,PI;function Kme(){if(PI)return my;PI=1,my=e,e.displayName="apl",e.aliases=[];function e(t){t.languages.apl={comment:/(?:⍝|#[! ]).*$/m,string:{pattern:/'(?:[^'\r\n]|'')*'/,greedy:!0},number:/¯?(?:\d*\.?\b\d+(?:e[+¯]?\d+)?|¯|∞)(?:j¯?(?:(?:\d+(?:\.\d+)?|\.\d+)(?:e[+¯]?\d+)?|¯|∞))?/i,statement:/:[A-Z][a-z][A-Za-z]*\b/,"system-function":{pattern:/⎕[A-Z]+/i,alias:"function"},constant:/[⍬⌾#⎕⍞]/,function:/[-+×÷⌈⌊∣|⍳⍸?*⍟○!⌹<≤=>≥≠≡≢∊⍷∪∩~∨∧⍱⍲⍴,⍪⌽⊖⍉↑↓⊂⊃⊆⊇⌷⍋⍒⊤⊥⍕⍎⊣⊢⍁⍂≈⍯↗¤→]/,"monadic-operator":{pattern:/[\\\/⌿⍀¨⍨⌶&∥]/,alias:"operator"},"dyadic-operator":{pattern:/[.⍣⍠⍤∘⌸@⌺⍥]/,alias:"operator"},assignment:{pattern:/←/,alias:"keyword"},punctuation:/[\[;\]()◇⋄]/,dfn:{pattern:/[{}⍺⍵⍶⍹∇⍫:]/,alias:"builtin"}}}return my}var by,zI;function Xme(){if(zI)return by;zI=1,by=e,e.displayName="applescript",e.aliases=[];function e(t){t.languages.applescript={comment:[/\(\*(?:\(\*(?:[^*]|\*(?!\)))*\*\)|(?!\(\*)[\s\S])*?\*\)/,/--.+/,/#.+/],string:/"(?:\\.|[^"\\\r\n])*"/,number:/(?:\b\d+(?:\.\d*)?|\B\.\d+)(?:e-?\d+)?\b/i,operator:[/[&=≠≤≥*+\-\/÷^]|[<>]=?/,/\b(?:(?:begin|end|start)s? with|(?:contains?|(?:does not|doesn't) contain)|(?:is|isn't|is not) (?:contained by|in)|(?:(?:is|isn't|is not) )?(?:greater|less) than(?: or equal)?(?: to)?|(?:comes|(?:does not|doesn't) come) (?:after|before)|(?:is|isn't|is not) equal(?: to)?|(?:(?:does not|doesn't) equal|equal to|equals|is not|isn't)|(?:a )?(?:ref(?: to)?|reference to)|(?:and|as|div|mod|not|or))\b/],keyword:/\b(?:about|above|after|against|apart from|around|aside from|at|back|before|beginning|behind|below|beneath|beside|between|but|by|considering|continue|copy|does|eighth|else|end|equal|error|every|exit|false|fifth|first|for|fourth|from|front|get|given|global|if|ignoring|in|instead of|into|is|it|its|last|local|me|middle|my|ninth|of|on|onto|out of|over|prop|property|put|repeat|return|returning|second|set|seventh|since|sixth|some|tell|tenth|that|the|then|third|through|thru|timeout|times|to|transaction|true|try|until|where|while|whose|with|without)\b/,"class-name":/\b(?:POSIX file|RGB color|alias|application|boolean|centimeters|centimetres|class|constant|cubic centimeters|cubic centimetres|cubic feet|cubic inches|cubic meters|cubic metres|cubic yards|date|degrees Celsius|degrees Fahrenheit|degrees Kelvin|feet|file|gallons|grams|inches|integer|kilograms|kilometers|kilometres|list|liters|litres|meters|metres|miles|number|ounces|pounds|quarts|real|record|reference|script|square feet|square kilometers|square kilometres|square meters|square metres|square miles|square yards|text|yards)\b/,punctuation:/[{}():,¬«»《》]/}}return by}var yy,BI;function Zme(){if(BI)return yy;BI=1,yy=e,e.displayName="aql",e.aliases=[];function e(t){t.languages.aql={comment:/\/\/.*|\/\*[\s\S]*?\*\//,property:{pattern:/([{,]\s*)(?:(?!\d)\w+|(["'´`])(?:(?!\2)[^\\\r\n]|\\.)*\2)(?=\s*:)/,lookbehind:!0,greedy:!0},string:{pattern:/(["'])(?:(?!\1)[^\\\r\n]|\\.)*\1/,greedy:!0},identifier:{pattern:/([´`])(?:(?!\1)[^\\\r\n]|\\.)*\1/,greedy:!0},variable:/@@?\w+/,keyword:[{pattern:/(\bWITH\s+)COUNT(?=\s+INTO\b)/i,lookbehind:!0},/\b(?:AGGREGATE|ALL|AND|ANY|ASC|COLLECT|DESC|DISTINCT|FILTER|FOR|GRAPH|IN|INBOUND|INSERT|INTO|K_PATHS|K_SHORTEST_PATHS|LET|LIKE|LIMIT|NONE|NOT|NULL|OR|OUTBOUND|REMOVE|REPLACE|RETURN|SHORTEST_PATH|SORT|UPDATE|UPSERT|WINDOW|WITH)\b/i,{pattern:/(^|[^\w.[])(?:KEEP|PRUNE|SEARCH|TO)\b/i,lookbehind:!0},{pattern:/(^|[^\w.[])(?:CURRENT|NEW|OLD)\b/,lookbehind:!0},{pattern:/\bOPTIONS(?=\s*\{)/i}],function:/\b(?!\d)\w+(?=\s*\()/,boolean:/\b(?:false|true)\b/i,range:{pattern:/\.\./,alias:"operator"},number:[/\b0b[01]+/i,/\b0x[0-9a-f]+/i,/(?:\B\.\d+|\b(?:0|[1-9]\d*)(?:\.\d+)?)(?:e[+-]?\d+)?/i],operator:/\*{2,}|[=!]~|[!=<>]=?|&&|\|\||[-+*/%]/,punctuation:/::|[?.:,;()[\]{}]/}}return yy}var vy,UI;function Ai(){if(UI)return vy;UI=1,vy=e,e.displayName="c",e.aliases=[];function e(t){t.languages.c=t.languages.extend("clike",{comment:{pattern:/\/\/(?:[^\r\n\\]|\\(?:\r\n?|\n|(?![\r\n])))*|\/\*[\s\S]*?(?:\*\/|$)/,greedy:!0},string:{pattern:/"(?:\\(?:\r\n|[\s\S])|[^"\\\r\n])*"/,greedy:!0},"class-name":{pattern:/(\b(?:enum|struct)\s+(?:__attribute__\s*\(\([\s\S]*?\)\)\s*)?)\w+|\b[a-z]\w*_t\b/,lookbehind:!0},keyword:/\b(?:_Alignas|_Alignof|_Atomic|_Bool|_Complex|_Generic|_Imaginary|_Noreturn|_Static_assert|_Thread_local|__attribute__|asm|auto|break|case|char|const|continue|default|do|double|else|enum|extern|float|for|goto|if|inline|int|long|register|return|short|signed|sizeof|static|struct|switch|typedef|typeof|union|unsigned|void|volatile|while)\b/,function:/\b[a-z_]\w*(?=\s*\()/i,number:/(?:\b0x(?:[\da-f]+(?:\.[\da-f]*)?|\.[\da-f]+)(?:p[+-]?\d+)?|(?:\b\d+(?:\.\d*)?|\B\.\d+)(?:e[+-]?\d+)?)[ful]{0,4}/i,operator:/>>=?|<<=?|->|([-+&|:])\1|[?:~]|[-+*/%&|^!=<>]=?/}),t.languages.insertBefore("c","string",{char:{pattern:/'(?:\\(?:\r\n|[\s\S])|[^'\\\r\n]){0,32}'/,greedy:!0}}),t.languages.insertBefore("c","string",{macro:{pattern:/(^[\t ]*)#\s*[a-z](?:[^\r\n\\/]|\/(?!\*)|\/\*(?:[^*]|\*(?!\/))*\*\/|\\(?:\r\n|[\s\S]))*/im,lookbehind:!0,greedy:!0,alias:"property",inside:{string:[{pattern:/^(#\s*include\s*)<[^>]+>/,lookbehind:!0},t.languages.c.string],char:t.languages.c.char,comment:t.languages.c.comment,"macro-name":[{pattern:/(^#\s*define\s+)\w+\b(?!\()/i,lookbehind:!0},{pattern:/(^#\s*define\s+)\w+\b(?=\()/i,lookbehind:!0,alias:"function"}],directive:{pattern:/^(#\s*)[a-z]+/,lookbehind:!0,alias:"keyword"},"directive-hash":/^#/,punctuation:/##|\\(?=[\r\n])/,expression:{pattern:/\S[\s\S]*/,inside:t.languages.c}}}}),t.languages.insertBefore("c","function",{constant:/\b(?:EOF|NULL|SEEK_CUR|SEEK_END|SEEK_SET|__DATE__|__FILE__|__LINE__|__TIMESTAMP__|__TIME__|__func__|stderr|stdin|stdout)\b/}),delete t.languages.c.boolean}return vy}var Sy,jI;function uA(){if(jI)return Sy;jI=1;var e=Ai();Sy=t,t.displayName="cpp",t.aliases=[];function t(n){n.register(e),function(r){var a=/\b(?:alignas|alignof|asm|auto|bool|break|case|catch|char|char16_t|char32_t|char8_t|class|co_await|co_return|co_yield|compl|concept|const|const_cast|consteval|constexpr|constinit|continue|decltype|default|delete|do|double|dynamic_cast|else|enum|explicit|export|extern|final|float|for|friend|goto|if|import|inline|int|int16_t|int32_t|int64_t|int8_t|long|module|mutable|namespace|new|noexcept|nullptr|operator|override|private|protected|public|register|reinterpret_cast|requires|return|short|signed|sizeof|static|static_assert|static_cast|struct|switch|template|this|thread_local|throw|try|typedef|typeid|typename|uint16_t|uint32_t|uint64_t|uint8_t|union|unsigned|using|virtual|void|volatile|wchar_t|while)\b/,o=/\b(?!<keyword>)\w+(?:\s*\.\s*\w+)*\b/.source.replace(/<keyword>/g,function(){return a.source});r.languages.cpp=r.languages.extend("c",{"class-name":[{pattern:RegExp(/(\b(?:class|concept|enum|struct|typename)\s+)(?!<keyword>)\w+/.source.replace(/<keyword>/g,function(){return a.source})),lookbehind:!0},/\b[A-Z]\w*(?=\s*::\s*\w+\s*\()/,/\b[A-Z_]\w*(?=\s*::\s*~\w+\s*\()/i,/\b\w+(?=\s*<(?:[^<>]|<(?:[^<>]|<[^<>]*>)*>)*>\s*::\s*\w+\s*\()/],keyword:a,number:{pattern:/(?:\b0b[01']+|\b0x(?:[\da-f']+(?:\.[\da-f']*)?|\.[\da-f']+)(?:p[+-]?[\d']+)?|(?:\b[\d']+(?:\.[\d']*)?|\B\.[\d']+)(?:e[+-]?[\d']+)?)[ful]{0,4}/i,greedy:!0},operator:/>>=?|<<=?|->|--|\+\+|&&|\|\||[?:~]|<=>|[-+*/%&|^!=<>]=?|\b(?:and|and_eq|bitand|bitor|not|not_eq|or|or_eq|xor|xor_eq)\b/,boolean:/\b(?:false|true)\b/}),r.languages.insertBefore("cpp","string",{module:{pattern:RegExp(/(\b(?:import|module)\s+)/.source+"(?:"+/"(?:\\(?:\r\n|[\s\S])|[^"\\\r\n])*"|<[^<>\r\n]*>/.source+"|"+/<mod-name>(?:\s*:\s*<mod-name>)?|:\s*<mod-name>/.source.replace(/<mod-name>/g,function(){return o})+")"),lookbehind:!0,greedy:!0,inside:{string:/^[<"][\s\S]+/,operator:/:/,punctuation:/\./}},"raw-string":{pattern:/R"([^()\\ ]{0,16})\([\s\S]*?\)\1"/,alias:"string",greedy:!0}}),r.languages.insertBefore("cpp","keyword",{"generic-function":{pattern:/\b(?!operator\b)[a-z_]\w*\s*<(?:[^<>]|<[^<>]*>)*>(?=\s*\()/i,inside:{function:/^\w+/,generic:{pattern:/<[\s\S]+/,alias:"class-name",inside:r.languages.cpp}}}}),r.languages.insertBefore("cpp","operator",{"double-colon":{pattern:/::/,alias:"punctuation"}}),r.languages.insertBefore("cpp","class-name",{"base-clause":{pattern:/(\b(?:class|struct)\s+\w+\s*:\s*)[^;{}"'\s]+(?:\s+[^;{}"'\s]+)*(?=\s*[;{])/,lookbehind:!0,greedy:!0,inside:r.languages.extend("cpp",{})}}),r.languages.insertBefore("inside","double-colon",{"class-name":/\b[a-z_]\w*\b(?!\s*::)/i},r.languages.cpp["base-clause"])}(n)}return Sy}var Ey,GI;function Qme(){if(GI)return Ey;GI=1;var e=uA();Ey=t,t.displayName="arduino",t.aliases=["ino"];function t(n){n.register(e),n.languages.arduino=n.languages.extend("cpp",{keyword:/\b(?:String|array|bool|boolean|break|byte|case|catch|continue|default|do|double|else|finally|for|function|goto|if|in|instanceof|int|integer|long|loop|new|null|return|setup|string|switch|throw|try|void|while|word)\b/,constant:/\b(?:ANALOG_MESSAGE|DEFAULT|DIGITAL_MESSAGE|EXTERNAL|FIRMATA_STRING|HIGH|INPUT|INPUT_PULLUP|INTERNAL|INTERNAL1V1|INTERNAL2V56|LED_BUILTIN|LOW|OUTPUT|REPORT_ANALOG|REPORT_DIGITAL|SET_PIN_MODE|SYSEX_START|SYSTEM_RESET)\b/,builtin:/\b(?:Audio|BSSID|Bridge|Client|Console|EEPROM|Esplora|EsploraTFT|Ethernet|EthernetClient|EthernetServer|EthernetUDP|File|FileIO|FileSystem|Firmata|GPRS|GSM|GSMBand|GSMClient|GSMModem|GSMPIN|GSMScanner|GSMServer|GSMVoiceCall|GSM_SMS|HttpClient|IPAddress|IRread|Keyboard|KeyboardController|LiquidCrystal|LiquidCrystal_I2C|Mailbox|Mouse|MouseController|PImage|Process|RSSI|RobotControl|RobotMotor|SD|SPI|SSID|Scheduler|Serial|Server|Servo|SoftwareSerial|Stepper|Stream|TFT|Task|USBHost|WiFi|WiFiClient|WiFiServer|WiFiUDP|Wire|YunClient|YunServer|abs|addParameter|analogRead|analogReadResolution|analogReference|analogWrite|analogWriteResolution|answerCall|attach|attachGPRS|attachInterrupt|attached|autoscroll|available|background|beep|begin|beginPacket|beginSD|beginSMS|beginSpeaker|beginTFT|beginTransmission|beginWrite|bit|bitClear|bitRead|bitSet|bitWrite|blink|blinkVersion|buffer|changePIN|checkPIN|checkPUK|checkReg|circle|cityNameRead|cityNameWrite|clear|clearScreen|click|close|compassRead|config|connect|connected|constrain|cos|countryNameRead|countryNameWrite|createChar|cursor|debugPrint|delay|delayMicroseconds|detach|detachInterrupt|digitalRead|digitalWrite|disconnect|display|displayLogos|drawBMP|drawCompass|encryptionType|end|endPacket|endSMS|endTransmission|endWrite|exists|exitValue|fill|find|findUntil|flush|gatewayIP|get|getAsynchronously|getBand|getButton|getCurrentCarrier|getIMEI|getKey|getModifiers|getOemKey|getPINUsed|getResult|getSignalStrength|getSocket|getVoiceCallStatus|getXChange|getYChange|hangCall|height|highByte|home|image|interrupts|isActionDone|isDirectory|isListening|isPIN|isPressed|isValid|keyPressed|keyReleased|keyboardRead|knobRead|leftToRight|line|lineFollowConfig|listen|listenOnLocalhost|loadImage|localIP|lowByte|macAddress|maintain|map|max|messageAvailable|micros|millis|min|mkdir|motorsStop|motorsWrite|mouseDragged|mouseMoved|mousePressed|mouseReleased|move|noAutoscroll|noBlink|noBuffer|noCursor|noDisplay|noFill|noInterrupts|noListenOnLocalhost|noStroke|noTone|onReceive|onRequest|open|openNextFile|overflow|parseCommand|parseFloat|parseInt|parsePacket|pauseMode|peek|pinMode|playFile|playMelody|point|pointTo|position|pow|prepare|press|print|printFirmwareVersion|printVersion|println|process|processInput|pulseIn|put|random|randomSeed|read|readAccelerometer|readBlue|readButton|readBytes|readBytesUntil|readGreen|readJoystickButton|readJoystickSwitch|readJoystickX|readJoystickY|readLightSensor|readMessage|readMicrophone|readNetworks|readRed|readSlider|readString|readStringUntil|readTemperature|ready|rect|release|releaseAll|remoteIP|remoteNumber|remotePort|remove|requestFrom|retrieveCallingNumber|rewindDirectory|rightToLeft|rmdir|robotNameRead|robotNameWrite|run|runAsynchronously|runShellCommand|runShellCommandAsynchronously|running|scanNetworks|scrollDisplayLeft|scrollDisplayRight|seek|sendAnalog|sendDigitalPortPair|sendDigitalPorts|sendString|sendSysex|serialEvent|setBand|setBitOrder|setClockDivider|setCursor|setDNS|setDataMode|setFirmwareVersion|setMode|setPINUsed|setSpeed|setTextSize|setTimeout|shiftIn|shiftOut|shutdown|sin|size|sqrt|startLoop|step|stop|stroke|subnetMask|switchPIN|tan|tempoWrite|text|tone|transfer|tuneWrite|turn|updateIR|userNameRead|userNameWrite|voiceCall|waitContinue|width|write|writeBlue|writeGreen|writeJSON|writeMessage|writeMicroseconds|writeRGB|writeRed|yield)\b/}),n.languages.ino=n.languages.arduino}return Ey}var wy,HI;function Jme(){if(HI)return wy;HI=1,wy=e,e.displayName="arff",e.aliases=[];function e(t){t.languages.arff={comment:/%.*/,string:{pattern:/(["'])(?:\\.|(?!\1)[^\\\r\n])*\1/,greedy:!0},keyword:/@(?:attribute|data|end|relation)\b/i,number:/\b\d+(?:\.\d+)?\b/,punctuation:/[{},]/}}return wy}var xy,$I;function ebe(){if($I)return xy;$I=1,xy=e,e.displayName="asciidoc",e.aliases=["adoc"];function e(t){(function(n){var r={pattern:/(^[ \t]*)\[(?!\[)(?:(["'$`])(?:(?!\2)[^\\]|\\.)*\2|\[(?:[^\[\]\\]|\\.)*\]|[^\[\]\\"'$`]|\\.)*\]/m,lookbehind:!0,inside:{quoted:{pattern:/([$`])(?:(?!\1)[^\\]|\\.)*\1/,inside:{punctuation:/^[$`]|[$`]$/}},interpreted:{pattern:/'(?:[^'\\]|\\.)*'/,inside:{punctuation:/^'|'$/}},string:/"(?:[^"\\]|\\.)*"/,variable:/\w+(?==)/,punctuation:/^\[|\]$|,/,operator:/=/,"attr-value":/(?!^\s+$).+/}},a=n.languages.asciidoc={"comment-block":{pattern:/^(\/{4,})(?:\r?\n|\r)(?:[\s\S]*(?:\r?\n|\r))??\1/m,alias:"comment"},table:{pattern:/^\|={3,}(?:(?:\r?\n|\r(?!\n)).*)*?(?:\r?\n|\r)\|={3,}$/m,inside:{specifiers:{pattern:/(?:(?:(?:\d+(?:\.\d+)?|\.\d+)[+*](?:[<^>](?:\.[<^>])?|\.[<^>])?|[<^>](?:\.[<^>])?|\.[<^>])[a-z]*|[a-z]+)(?=\|)/,alias:"attr-value"},punctuation:{pattern:/(^|[^\\])[|!]=*/,lookbehind:!0}}},"passthrough-block":{pattern:/^(\+{4,})(?:\r?\n|\r)(?:[\s\S]*(?:\r?\n|\r))??\1$/m,inside:{punctuation:/^\++|\++$/}},"literal-block":{pattern:/^(-{4,}|\.{4,})(?:\r?\n|\r)(?:[\s\S]*(?:\r?\n|\r))??\1$/m,inside:{punctuation:/^(?:-+|\.+)|(?:-+|\.+)$/}},"other-block":{pattern:/^(--|\*{4,}|_{4,}|={4,})(?:\r?\n|\r)(?:[\s\S]*(?:\r?\n|\r))??\1$/m,inside:{punctuation:/^(?:-+|\*+|_+|=+)|(?:-+|\*+|_+|=+)$/}},"list-punctuation":{pattern:/(^[ \t]*)(?:-|\*{1,5}|\.{1,5}|(?:[a-z]|\d+)\.|[xvi]+\))(?= )/im,lookbehind:!0,alias:"punctuation"},"list-label":{pattern:/(^[ \t]*)[a-z\d].+(?::{2,4}|;;)(?=\s)/im,lookbehind:!0,alias:"symbol"},"indented-block":{pattern:/((\r?\n|\r)\2)([ \t]+)\S.*(?:(?:\r?\n|\r)\3.+)*(?=\2{2}|$)/,lookbehind:!0},comment:/^\/\/.*/m,title:{pattern:/^.+(?:\r?\n|\r)(?:={3,}|-{3,}|~{3,}|\^{3,}|\+{3,})$|^={1,5} .+|^\.(?![\s.]).*/m,alias:"important",inside:{punctuation:/^(?:\.|=+)|(?:=+|-+|~+|\^+|\++)$/}},"attribute-entry":{pattern:/^:[^:\r\n]+:(?: .*?(?: \+(?:\r?\n|\r).*?)*)?$/m,alias:"tag"},attributes:r,hr:{pattern:/^'{3,}$/m,alias:"punctuation"},"page-break":{pattern:/^<{3,}$/m,alias:"punctuation"},admonition:{pattern:/^(?:CAUTION|IMPORTANT|NOTE|TIP|WARNING):/m,alias:"keyword"},callout:[{pattern:/(^[ \t]*)<?\d*>/m,lookbehind:!0,alias:"symbol"},{pattern:/<\d+>/,alias:"symbol"}],macro:{pattern:/\b[a-z\d][a-z\d-]*::?(?:[^\s\[\]]*\[(?:[^\]\\"']|(["'])(?:(?!\1)[^\\]|\\.)*\1|\\.)*\])/,inside:{function:/^[a-z\d-]+(?=:)/,punctuation:/^::?/,attributes:{pattern:/(?:\[(?:[^\]\\"']|(["'])(?:(?!\1)[^\\]|\\.)*\1|\\.)*\])/,inside:r.inside}}},inline:{pattern:/(^|[^\\])(?:(?:\B\[(?:[^\]\\"']|(["'])(?:(?!\2)[^\\]|\\.)*\2|\\.)*\])?(?:\b_(?!\s)(?: _|[^_\\\r\n]|\\.)+(?:(?:\r?\n|\r)(?: _|[^_\\\r\n]|\\.)+)*_\b|\B``(?!\s).+?(?:(?:\r?\n|\r).+?)*''\B|\B`(?!\s)(?:[^`'\s]|\s+\S)+['`]\B|\B(['*+#])(?!\s)(?: \3|(?!\3)[^\\\r\n]|\\.)+(?:(?:\r?\n|\r)(?: \3|(?!\3)[^\\\r\n]|\\.)+)*\3\B)|(?:\[(?:[^\]\\"']|(["'])(?:(?!\4)[^\\]|\\.)*\4|\\.)*\])?(?:(__|\*\*|\+\+\+?|##|\$\$|[~^]).+?(?:(?:\r?\n|\r).+?)*\5|\{[^}\r\n]+\}|\[\[\[?.+?(?:(?:\r?\n|\r).+?)*\]?\]\]|<<.+?(?:(?:\r?\n|\r).+?)*>>|\(\(\(?.+?(?:(?:\r?\n|\r).+?)*\)?\)\)))/m,lookbehind:!0,inside:{attributes:r,url:{pattern:/^(?:\[\[\[?.+?\]?\]\]|<<.+?>>)$/,inside:{punctuation:/^(?:\[\[\[?|<<)|(?:\]\]\]?|>>)$/}},"attribute-ref":{pattern:/^\{.+\}$/,inside:{variable:{pattern:/(^\{)[a-z\d,+_-]+/,lookbehind:!0},operator:/^[=?!#%@$]|!(?=[:}])/,punctuation:/^\{|\}$|::?/}},italic:{pattern:/^(['_])[\s\S]+\1$/,inside:{punctuation:/^(?:''?|__?)|(?:''?|__?)$/}},bold:{pattern:/^\*[\s\S]+\*$/,inside:{punctuation:/^\*\*?|\*\*?$/}},punctuation:/^(?:``?|\+{1,3}|##?|\$\$|[~^]|\(\(\(?)|(?:''?|\+{1,3}|##?|\$\$|[~^`]|\)?\)\))$/}},replacement:{pattern:/\((?:C|R|TM)\)/,alias:"builtin"},entity:/&#?[\da-z]{1,8};/i,"line-continuation":{pattern:/(^| )\+$/m,lookbehind:!0,alias:"punctuation"}};function o(s){s=s.split(" ");for(var u={},c=0,d=s.length;c<d;c++)u[s[c]]=a[s[c]];return u}r.inside.interpreted.inside.rest=o("macro inline replacement entity"),a["passthrough-block"].inside.rest=o("macro"),a["literal-block"].inside.rest=o("callout"),a.table.inside.rest=o("comment-block passthrough-block literal-block other-block list-punctuation indented-block comment title attribute-entry attributes hr page-break admonition list-label callout macro inline replacement entity line-continuation"),a["other-block"].inside.rest=o("table list-punctuation indented-block comment attribute-entry attributes hr page-break admonition list-label macro inline replacement entity line-continuation"),a.title.inside.rest=o("macro inline replacement entity"),n.hooks.add("wrap",function(s){s.type==="entity"&&(s.attributes.title=s.content.value.replace(/&amp;/,"&"))}),n.languages.adoc=n.languages.asciidoc})(t)}return xy}var ky,qI;function tbe(){if(qI)return ky;qI=1,ky=e,e.displayName="asm6502",e.aliases=[];function e(t){t.languages.asm6502={comment:/;.*/,directive:{pattern:/\.\w+(?= )/,alias:"property"},string:/(["'`])(?:\\.|(?!\1)[^\\\r\n])*\1/,"op-code":{pattern:/\b(?:ADC|AND|ASL|BCC|BCS|BEQ|BIT|BMI|BNE|BPL|BRK|BVC|BVS|CLC|CLD|CLI|CLV|CMP|CPX|CPY|DEC|DEX|DEY|EOR|INC|INX|INY|JMP|JSR|LDA|LDX|LDY|LSR|NOP|ORA|PHA|PHP|PLA|PLP|ROL|ROR|RTI|RTS|SBC|SEC|SED|SEI|STA|STX|STY|TAX|TAY|TSX|TXA|TXS|TYA|adc|and|asl|bcc|bcs|beq|bit|bmi|bne|bpl|brk|bvc|bvs|clc|cld|cli|clv|cmp|cpx|cpy|dec|dex|dey|eor|inc|inx|iny|jmp|jsr|lda|ldx|ldy|lsr|nop|ora|pha|php|pla|plp|rol|ror|rti|rts|sbc|sec|sed|sei|sta|stx|sty|tax|tay|tsx|txa|txs|tya)\b/,alias:"keyword"},"hex-number":{pattern:/#?\$[\da-f]{1,4}\b/i,alias:"number"},"binary-number":{pattern:/#?%[01]+\b/,alias:"number"},"decimal-number":{pattern:/#?\b\d+\b/,alias:"number"},register:{pattern:/\b[xya]\b/i,alias:"variable"},punctuation:/[(),:]/}}return ky}var Ty,VI;function nbe(){if(VI)return Ty;VI=1,Ty=e,e.displayName="asmatmel",e.aliases=[];function e(t){t.languages.asmatmel={comment:{pattern:/;.*/,greedy:!0},string:{pattern:/(["'`])(?:\\.|(?!\1)[^\\\r\n])*\1/,greedy:!0},constant:/\b(?:PORT[A-Z]|DDR[A-Z]|(?:DD|P)[A-Z](?:\d|[0-2]\d|3[01]))\b/,directive:{pattern:/\.\w+(?= )/,alias:"property"},"r-register":{pattern:/\br(?:\d|[12]\d|3[01])\b/,alias:"variable"},"op-code":{pattern:/\b(?:ADC|ADD|ADIW|AND|ANDI|ASR|BCLR|BLD|BRBC|BRBS|BRCC|BRCS|BREAK|BREQ|BRGE|BRHC|BRHS|BRID|BRIE|BRLO|BRLT|BRMI|BRNE|BRPL|BRSH|BRTC|BRTS|BRVC|BRVS|BSET|BST|CALL|CBI|CBR|CLC|CLH|CLI|CLN|CLR|CLS|CLT|CLV|CLZ|COM|CP|CPC|CPI|CPSE|DEC|DES|EICALL|EIJMP|ELPM|EOR|FMUL|FMULS|FMULSU|ICALL|IJMP|IN|INC|JMP|LAC|LAS|LAT|LD|LD[A-Za-z0-9]|LPM|LSL|LSR|MOV|MOVW|MUL|MULS|MULSU|NEG|NOP|OR|ORI|OUT|POP|PUSH|RCALL|RET|RETI|RJMP|ROL|ROR|SBC|SBCI|SBI|SBIC|SBIS|SBIW|SBR|SBRC|SBRS|SEC|SEH|SEI|SEN|SER|SES|SET|SEV|SEZ|SLEEP|SPM|ST|ST[A-Z0-9]|SUB|SUBI|SWAP|TST|WDR|XCH|adc|add|adiw|and|andi|asr|bclr|bld|brbc|brbs|brcc|brcs|break|breq|brge|brhc|brhs|brid|brie|brlo|brlt|brmi|brne|brpl|brsh|brtc|brts|brvc|brvs|bset|bst|call|cbi|cbr|clc|clh|cli|cln|clr|cls|clt|clv|clz|com|cp|cpc|cpi|cpse|dec|des|eicall|eijmp|elpm|eor|fmul|fmuls|fmulsu|icall|ijmp|in|inc|jmp|lac|las|lat|ld|ld[a-z0-9]|lpm|lsl|lsr|mov|movw|mul|muls|mulsu|neg|nop|or|ori|out|pop|push|rcall|ret|reti|rjmp|rol|ror|sbc|sbci|sbi|sbic|sbis|sbiw|sbr|sbrc|sbrs|sec|seh|sei|sen|ser|ses|set|sev|sez|sleep|spm|st|st[a-zA-Z0-9]|sub|subi|swap|tst|wdr|xch)\b/,alias:"keyword"},"hex-number":{pattern:/#?\$[\da-f]{2,4}\b/i,alias:"number"},"binary-number":{pattern:/#?%[01]+\b/,alias:"number"},"decimal-number":{pattern:/#?\b\d+\b/,alias:"number"},register:{pattern:/\b[acznvshtixy]\b/i,alias:"variable"},operator:/>>=?|<<=?|&&?|\|\|?|[-+*/%&|^!=<>?]=?/,punctuation:/[(),:]/}}return Ty}var Ay,WI;function Mp(){if(WI)return Ay;WI=1,Ay=e,e.displayName="csharp",e.aliases=["dotnet","cs"];function e(t){(function(n){function r(V,B){return V.replace(/<<(\d+)>>/g,function(M,W){return"(?:"+B[+W]+")"})}function a(V,B,M){return RegExp(r(V,B),"")}function o(V,B){for(var M=0;M<B;M++)V=V.replace(/<<self>>/g,function(){return"(?:"+V+")"});return V.replace(/<<self>>/g,"[^\\s\\S]")}var s={type:"bool byte char decimal double dynamic float int long object sbyte short string uint ulong ushort var void",typeDeclaration:"class enum interface record struct",contextual:"add alias and ascending async await by descending from(?=\\s*(?:\\w|$)) get global group into init(?=\\s*;) join let nameof not notnull on or orderby partial remove select set unmanaged value when where with(?=\\s*{)",other:"abstract as base break case catch checked const continue default delegate do else event explicit extern finally fixed for foreach goto if implicit in internal is lock namespace new null operator out override params private protected public readonly ref return sealed sizeof stackalloc static switch this throw try typeof unchecked unsafe using virtual volatile while yield"};function u(V){return"\\b(?:"+V.trim().replace(/ /g,"|")+")\\b"}var c=u(s.typeDeclaration),d=RegExp(u(s.type+" "+s.typeDeclaration+" "+s.contextual+" "+s.other)),p=u(s.typeDeclaration+" "+s.contextual+" "+s.other),g=u(s.type+" "+s.typeDeclaration+" "+s.other),m=o(/<(?:[^<>;=+\-*/%&|^]|<<self>>)*>/.source,2),b=o(/\((?:[^()]|<<self>>)*\)/.source,2),y=/@?\b[A-Za-z_]\w*\b/.source,S=r(/<<0>>(?:\s*<<1>>)?/.source,[y,m]),x=r(/(?!<<0>>)<<1>>(?:\s*\.\s*<<1>>)*/.source,[p,S]),R=/\[\s*(?:,\s*)*\]/.source,k=r(/<<0>>(?:\s*(?:\?\s*)?<<1>>)*(?:\s*\?)?/.source,[x,R]),A=r(/[^,()<>[\];=+\-*/%&|^]|<<0>>|<<1>>|<<2>>/.source,[m,b,R]),C=r(/\(<<0>>+(?:,<<0>>+)+\)/.source,[A]),N=r(/(?:<<0>>|<<1>>)(?:\s*(?:\?\s*)?<<2>>)*(?:\s*\?)?/.source,[C,x,R]),_={keyword:d,punctuation:/[<>()?,.:[\]]/},O=/'(?:[^\r\n'\\]|\\.|\\[Uux][\da-fA-F]{1,8})'/.source,F=/"(?:\\.|[^\\"\r\n])*"/.source,L=/@"(?:""|\\[\s\S]|[^\\"])*"(?!")/.source;n.languages.csharp=n.languages.extend("clike",{string:[{pattern:a(/(^|[^$\\])<<0>>/.source,[L]),lookbehind:!0,greedy:!0},{pattern:a(/(^|[^@$\\])<<0>>/.source,[F]),lookbehind:!0,greedy:!0}],"class-name":[{pattern:a(/(\busing\s+static\s+)<<0>>(?=\s*;)/.source,[x]),lookbehind:!0,inside:_},{pattern:a(/(\busing\s+<<0>>\s*=\s*)<<1>>(?=\s*;)/.source,[y,N]),lookbehind:!0,inside:_},{pattern:a(/(\busing\s+)<<0>>(?=\s*=)/.source,[y]),lookbehind:!0},{pattern:a(/(\b<<0>>\s+)<<1>>/.source,[c,S]),lookbehind:!0,inside:_},{pattern:a(/(\bcatch\s*\(\s*)<<0>>/.source,[x]),lookbehind:!0,inside:_},{pattern:a(/(\bwhere\s+)<<0>>/.source,[y]),lookbehind:!0},{pattern:a(/(\b(?:is(?:\s+not)?|as)\s+)<<0>>/.source,[k]),lookbehind:!0,inside:_},{pattern:a(/\b<<0>>(?=\s+(?!<<1>>|with\s*\{)<<2>>(?:\s*[=,;:{)\]]|\s+(?:in|when)\b))/.source,[N,g,y]),inside:_}],keyword:d,number:/(?:\b0(?:x[\da-f_]*[\da-f]|b[01_]*[01])|(?:\B\.\d+(?:_+\d+)*|\b\d+(?:_+\d+)*(?:\.\d+(?:_+\d+)*)?)(?:e[-+]?\d+(?:_+\d+)*)?)(?:[dflmu]|lu|ul)?\b/i,operator:/>>=?|<<=?|[-=]>|([-+&|])\1|~|\?\?=?|[-+*/%&|^!=<>]=?/,punctuation:/\?\.?|::|[{}[\];(),.:]/}),n.languages.insertBefore("csharp","number",{range:{pattern:/\.\./,alias:"operator"}}),n.languages.insertBefore("csharp","punctuation",{"named-parameter":{pattern:a(/([(,]\s*)<<0>>(?=\s*:)/.source,[y]),lookbehind:!0,alias:"punctuation"}}),n.languages.insertBefore("csharp","class-name",{namespace:{pattern:a(/(\b(?:namespace|using)\s+)<<0>>(?:\s*\.\s*<<0>>)*(?=\s*[;{])/.source,[y]),lookbehind:!0,inside:{punctuation:/\./}},"type-expression":{pattern:a(/(\b(?:default|sizeof|typeof)\s*\(\s*(?!\s))(?:[^()\s]|\s(?!\s)|<<0>>)*(?=\s*\))/.source,[b]),lookbehind:!0,alias:"class-name",inside:_},"return-type":{pattern:a(/<<0>>(?=\s+(?:<<1>>\s*(?:=>|[({]|\.\s*this\s*\[)|this\s*\[))/.source,[N,x]),inside:_,alias:"class-name"},"constructor-invocation":{pattern:a(/(\bnew\s+)<<0>>(?=\s*[[({])/.source,[N]),lookbehind:!0,inside:_,alias:"class-name"},"generic-method":{pattern:a(/<<0>>\s*<<1>>(?=\s*\()/.source,[y,m]),inside:{function:a(/^<<0>>/.source,[y]),generic:{pattern:RegExp(m),alias:"class-name",inside:_}}},"type-list":{pattern:a(/\b((?:<<0>>\s+<<1>>|record\s+<<1>>\s*<<5>>|where\s+<<2>>)\s*:\s*)(?:<<3>>|<<4>>|<<1>>\s*<<5>>|<<6>>)(?:\s*,\s*(?:<<3>>|<<4>>|<<6>>))*(?=\s*(?:where|[{;]|=>|$))/.source,[c,S,y,N,d.source,b,/\bnew\s*\(\s*\)/.source]),lookbehind:!0,inside:{"record-arguments":{pattern:a(/(^(?!new\s*\()<<0>>\s*)<<1>>/.source,[S,b]),lookbehind:!0,greedy:!0,inside:n.languages.csharp},keyword:d,"class-name":{pattern:RegExp(N),greedy:!0,inside:_},punctuation:/[,()]/}},preprocessor:{pattern:/(^[\t ]*)#.*/m,lookbehind:!0,alias:"property",inside:{directive:{pattern:/(#)\b(?:define|elif|else|endif|endregion|error|if|line|nullable|pragma|region|undef|warning)\b/,lookbehind:!0,alias:"keyword"}}}});var I=F+"|"+O,H=r(/\/(?![*/])|\/\/[^\r\n]*[\r\n]|\/\*(?:[^*]|\*(?!\/))*\*\/|<<0>>/.source,[I]),$=o(r(/[^"'/()]|<<0>>|\(<<self>>*\)/.source,[H]),2),U=/\b(?:assembly|event|field|method|module|param|property|return|type)\b/.source,Y=r(/<<0>>(?:\s*\(<<1>>*\))?/.source,[x,$]);n.languages.insertBefore("csharp","class-name",{attribute:{pattern:a(/((?:^|[^\s\w>)?])\s*\[\s*)(?:<<0>>\s*:\s*)?<<1>>(?:\s*,\s*<<1>>)*(?=\s*\])/.source,[U,Y]),lookbehind:!0,greedy:!0,inside:{target:{pattern:a(/^<<0>>(?=\s*:)/.source,[U]),alias:"keyword"},"attribute-arguments":{pattern:a(/\(<<0>>*\)/.source,[$]),inside:n.languages.csharp},"class-name":{pattern:RegExp(x),inside:{punctuation:/\./}},punctuation:/[:,]/}}});var Z=/:[^}\r\n]+/.source,j=o(r(/[^"'/()]|<<0>>|\(<<self>>*\)/.source,[H]),2),G=r(/\{(?!\{)(?:(?![}:])<<0>>)*<<1>>?\}/.source,[j,Z]),z=o(r(/[^"'/()]|\/(?!\*)|\/\*(?:[^*]|\*(?!\/))*\*\/|<<0>>|\(<<self>>*\)/.source,[I]),2),K=r(/\{(?!\{)(?:(?![}:])<<0>>)*<<1>>?\}/.source,[z,Z]);function D(V,B){return{interpolation:{pattern:a(/((?:^|[^{])(?:\{\{)*)<<0>>/.source,[V]),lookbehind:!0,inside:{"format-string":{pattern:a(/(^\{(?:(?![}:])<<0>>)*)<<1>>(?=\}$)/.source,[B,Z]),lookbehind:!0,inside:{punctuation:/^:/}},punctuation:/^\{|\}$/,expression:{pattern:/[\s\S]+/,alias:"language-csharp",inside:n.languages.csharp}}},string:/[\s\S]+/}}n.languages.insertBefore("csharp","string",{"interpolation-string":[{pattern:a(/(^|[^\\])(?:\$@|@\$)"(?:""|\\[\s\S]|\{\{|<<0>>|[^\\{"])*"/.source,[G]),lookbehind:!0,greedy:!0,inside:D(G,j)},{pattern:a(/(^|[^@\\])\$"(?:\\.|\{\{|<<0>>|[^\\"{])*"/.source,[K]),lookbehind:!0,greedy:!0,inside:D(K,z)}],char:{pattern:RegExp(O),greedy:!0}}),n.languages.dotnet=n.languages.cs=n.languages.csharp})(t)}return Ay}var Ry,YI;function rbe(){if(YI)return Ry;YI=1;var e=Mp();Ry=t,t.displayName="aspnet",t.aliases=[];function t(n){n.register(e),n.languages.aspnet=n.languages.extend("markup",{"page-directive":{pattern:/<%\s*@.*%>/,alias:"tag",inside:{"page-directive":{pattern:/<%\s*@\s*(?:Assembly|Control|Implements|Import|Master(?:Type)?|OutputCache|Page|PreviousPageType|Reference|Register)?|%>/i,alias:"tag"},rest:n.languages.markup.tag.inside}},directive:{pattern:/<%.*%>/,alias:"tag",inside:{directive:{pattern:/<%\s*?[$=%#:]{0,2}|%>/,alias:"tag"},rest:n.languages.csharp}}}),n.languages.aspnet.tag.pattern=/<(?!%)\/?[^\s>\/]+(?:\s+[^\s>\/=]+(?:=(?:("|')(?:\\[\s\S]|(?!\1)[^\\])*\1|[^\s'">=]+))?)*\s*\/?>/,n.languages.insertBefore("inside","punctuation",{directive:n.languages.aspnet.directive},n.languages.aspnet.tag.inside["attr-value"]),n.languages.insertBefore("aspnet","comment",{"asp-comment":{pattern:/<%--[\s\S]*?--%>/,alias:["asp","comment"]}}),n.languages.insertBefore("aspnet",n.languages.javascript?"script":"tag",{"asp-script":{pattern:/(<script(?=.*runat=['"]?server\b)[^>]*>)[\s\S]*?(?=<\/script>)/i,lookbehind:!0,alias:["asp","script"],inside:n.languages.csharp||{}}})}return Ry}var _y,KI;function abe(){if(KI)return _y;KI=1,_y=e,e.displayName="autohotkey",e.aliases=[];function e(t){t.languages.autohotkey={comment:[{pattern:/(^|\s);.*/,lookbehind:!0},{pattern:/(^[\t ]*)\/\*(?:[\r\n](?![ \t]*\*\/)|[^\r\n])*(?:[\r\n][ \t]*\*\/)?/m,lookbehind:!0,greedy:!0}],tag:{pattern:/^([ \t]*)[^\s,`":]+(?=:[ \t]*$)/m,lookbehind:!0},string:/"(?:[^"\n\r]|"")*"/,variable:/%\w+%/,number:/\b0x[\dA-Fa-f]+\b|(?:\b\d+(?:\.\d*)?|\B\.\d+)(?:[Ee]-?\d+)?/,operator:/\?|\/\/?=?|:=|\|[=|]?|&[=&]?|\+[=+]?|-[=-]?|\*[=*]?|<(?:<=?|>|=)?|>>?=?|[.^!=~]=?|\b(?:AND|NOT|OR)\b/,boolean:/\b(?:false|true)\b/,selector:/\b(?:AutoTrim|BlockInput|Break|Click|ClipWait|Continue|Control|ControlClick|ControlFocus|ControlGet|ControlGetFocus|ControlGetPos|ControlGetText|ControlMove|ControlSend|ControlSendRaw|ControlSetText|CoordMode|Critical|DetectHiddenText|DetectHiddenWindows|Drive|DriveGet|DriveSpaceFree|EnvAdd|EnvDiv|EnvGet|EnvMult|EnvSet|EnvSub|EnvUpdate|Exit|ExitApp|FileAppend|FileCopy|FileCopyDir|FileCreateDir|FileCreateShortcut|FileDelete|FileEncoding|FileGetAttrib|FileGetShortcut|FileGetSize|FileGetTime|FileGetVersion|FileInstall|FileMove|FileMoveDir|FileRead|FileReadLine|FileRecycle|FileRecycleEmpty|FileRemoveDir|FileSelectFile|FileSelectFolder|FileSetAttrib|FileSetTime|FormatTime|GetKeyState|Gosub|Goto|GroupActivate|GroupAdd|GroupClose|GroupDeactivate|Gui|GuiControl|GuiControlGet|Hotkey|ImageSearch|IniDelete|IniRead|IniWrite|Input|InputBox|KeyWait|ListHotkeys|ListLines|ListVars|Loop|Menu|MouseClick|MouseClickDrag|MouseGetPos|MouseMove|MsgBox|OnExit|OutputDebug|Pause|PixelGetColor|PixelSearch|PostMessage|Process|Progress|Random|RegDelete|RegRead|RegWrite|Reload|Repeat|Return|Run|RunAs|RunWait|Send|SendEvent|SendInput|SendMessage|SendMode|SendPlay|SendRaw|SetBatchLines|SetCapslockState|SetControlDelay|SetDefaultMouseSpeed|SetEnv|SetFormat|SetKeyDelay|SetMouseDelay|SetNumlockState|SetRegView|SetScrollLockState|SetStoreCapslockMode|SetTimer|SetTitleMatchMode|SetWinDelay|SetWorkingDir|Shutdown|Sleep|Sort|SoundBeep|SoundGet|SoundGetWaveVolume|SoundPlay|SoundSet|SoundSetWaveVolume|SplashImage|SplashTextOff|SplashTextOn|SplitPath|StatusBarGetText|StatusBarWait|StringCaseSense|StringGetPos|StringLeft|StringLen|StringLower|StringMid|StringReplace|StringRight|StringSplit|StringTrimLeft|StringTrimRight|StringUpper|Suspend|SysGet|Thread|ToolTip|Transform|TrayTip|URLDownloadToFile|WinActivate|WinActivateBottom|WinClose|WinGet|WinGetActiveStats|WinGetActiveTitle|WinGetClass|WinGetPos|WinGetText|WinGetTitle|WinHide|WinKill|WinMaximize|WinMenuSelectItem|WinMinimize|WinMinimizeAll|WinMinimizeAllUndo|WinMove|WinRestore|WinSet|WinSetTitle|WinShow|WinWait|WinWaitActive|WinWaitClose|WinWaitNotActive)\b/i,constant:/\b(?:a_ahkpath|a_ahkversion|a_appdata|a_appdatacommon|a_autotrim|a_batchlines|a_caretx|a_carety|a_computername|a_controldelay|a_cursor|a_dd|a_ddd|a_dddd|a_defaultmousespeed|a_desktop|a_desktopcommon|a_detecthiddentext|a_detecthiddenwindows|a_endchar|a_eventinfo|a_exitreason|a_fileencoding|a_formatfloat|a_formatinteger|a_gui|a_guicontrol|a_guicontrolevent|a_guievent|a_guiheight|a_guiwidth|a_guix|a_guiy|a_hour|a_iconfile|a_iconhidden|a_iconnumber|a_icontip|a_index|a_ipaddress1|a_ipaddress2|a_ipaddress3|a_ipaddress4|a_is64bitos|a_isadmin|a_iscompiled|a_iscritical|a_ispaused|a_issuspended|a_isunicode|a_keydelay|a_language|a_lasterror|a_linefile|a_linenumber|a_loopfield|a_loopfileattrib|a_loopfiledir|a_loopfileext|a_loopfilefullpath|a_loopfilelongpath|a_loopfilename|a_loopfileshortname|a_loopfileshortpath|a_loopfilesize|a_loopfilesizekb|a_loopfilesizemb|a_loopfiletimeaccessed|a_loopfiletimecreated|a_loopfiletimemodified|a_loopreadline|a_loopregkey|a_loopregname|a_loopregsubkey|a_loopregtimemodified|a_loopregtype|a_mday|a_min|a_mm|a_mmm|a_mmmm|a_mon|a_mousedelay|a_msec|a_mydocuments|a_now|a_nowutc|a_numbatchlines|a_ostype|a_osversion|a_priorhotkey|a_priorkey|a_programfiles|a_programs|a_programscommon|a_ptrsize|a_regview|a_screendpi|a_screenheight|a_screenwidth|a_scriptdir|a_scriptfullpath|a_scripthwnd|a_scriptname|a_sec|a_space|a_startmenu|a_startmenucommon|a_startup|a_startupcommon|a_stringcasesense|a_tab|a_temp|a_thisfunc|a_thishotkey|a_thislabel|a_thismenu|a_thismenuitem|a_thismenuitempos|a_tickcount|a_timeidle|a_timeidlephysical|a_timesincepriorhotkey|a_timesincethishotkey|a_titlematchmode|a_titlematchmodespeed|a_username|a_wday|a_windelay|a_windir|a_workingdir|a_yday|a_year|a_yweek|a_yyyy|clipboard|clipboardall|comspec|errorlevel|programfiles)\b/i,builtin:/\b(?:abs|acos|asc|asin|atan|ceil|chr|class|comobjactive|comobjarray|comobjconnect|comobjcreate|comobjerror|comobjflags|comobjget|comobjquery|comobjtype|comobjvalue|cos|dllcall|exp|fileexist|Fileopen|floor|format|il_add|il_create|il_destroy|instr|isfunc|islabel|IsObject|ln|log|ltrim|lv_add|lv_delete|lv_deletecol|lv_getcount|lv_getnext|lv_gettext|lv_insert|lv_insertcol|lv_modify|lv_modifycol|lv_setimagelist|mod|numget|numput|onmessage|regexmatch|regexreplace|registercallback|round|rtrim|sb_seticon|sb_setparts|sb_settext|sin|sqrt|strlen|strreplace|strsplit|substr|tan|tv_add|tv_delete|tv_get|tv_getchild|tv_getcount|tv_getnext|tv_getparent|tv_getprev|tv_getselection|tv_gettext|tv_modify|varsetcapacity|winactive|winexist|__Call|__Get|__New|__Set)\b/i,symbol:/\b(?:alt|altdown|altup|appskey|backspace|browser_back|browser_favorites|browser_forward|browser_home|browser_refresh|browser_search|browser_stop|bs|capslock|ctrl|ctrlbreak|ctrldown|ctrlup|del|delete|down|end|enter|esc|escape|f1|f10|f11|f12|f13|f14|f15|f16|f17|f18|f19|f2|f20|f21|f22|f23|f24|f3|f4|f5|f6|f7|f8|f9|home|ins|insert|joy1|joy10|joy11|joy12|joy13|joy14|joy15|joy16|joy17|joy18|joy19|joy2|joy20|joy21|joy22|joy23|joy24|joy25|joy26|joy27|joy28|joy29|joy3|joy30|joy31|joy32|joy4|joy5|joy6|joy7|joy8|joy9|joyaxes|joybuttons|joyinfo|joyname|joypov|joyr|joyu|joyv|joyx|joyy|joyz|lalt|launch_app1|launch_app2|launch_mail|launch_media|lbutton|lcontrol|lctrl|left|lshift|lwin|lwindown|lwinup|mbutton|media_next|media_play_pause|media_prev|media_stop|numlock|numpad0|numpad1|numpad2|numpad3|numpad4|numpad5|numpad6|numpad7|numpad8|numpad9|numpadadd|numpadclear|numpaddel|numpaddiv|numpaddot|numpaddown|numpadend|numpadenter|numpadhome|numpadins|numpadleft|numpadmult|numpadpgdn|numpadpgup|numpadright|numpadsub|numpadup|pgdn|pgup|printscreen|ralt|rbutton|rcontrol|rctrl|right|rshift|rwin|rwindown|rwinup|scrolllock|shift|shiftdown|shiftup|space|tab|up|volume_down|volume_mute|volume_up|wheeldown|wheelleft|wheelright|wheelup|xbutton1|xbutton2)\b/i,important:/#\b(?:AllowSameLineComments|ClipboardTimeout|CommentFlag|DerefChar|ErrorStdOut|EscapeChar|HotkeyInterval|HotkeyModifierTimeout|Hotstring|If|IfTimeout|IfWinActive|IfWinExist|IfWinNotActive|IfWinNotExist|Include|IncludeAgain|InputLevel|InstallKeybdHook|InstallMouseHook|KeyHistory|MaxHotkeysPerInterval|MaxMem|MaxThreads|MaxThreadsBuffer|MaxThreadsPerHotkey|MenuMaskKey|NoEnv|NoTrayIcon|Persistent|SingleInstance|UseHook|Warn|WinActivateForce)\b/i,keyword:/\b(?:Abort|AboveNormal|Add|ahk_class|ahk_exe|ahk_group|ahk_id|ahk_pid|All|Alnum|Alpha|AltSubmit|AltTab|AltTabAndMenu|AltTabMenu|AltTabMenuDismiss|AlwaysOnTop|AutoSize|Background|BackgroundTrans|BelowNormal|between|BitAnd|BitNot|BitOr|BitShiftLeft|BitShiftRight|BitXOr|Bold|Border|Button|ByRef|Catch|Checkbox|Checked|CheckedGray|Choose|ChooseString|Close|Color|ComboBox|Contains|ControlList|Count|Date|DateTime|Days|DDL|Default|DeleteAll|Delimiter|Deref|Destroy|Digit|Disable|Disabled|DropDownList|Edit|Eject|Else|Enable|Enabled|Error|Exist|Expand|ExStyle|FileSystem|Finally|First|Flash|Float|FloatFast|Focus|Font|for|global|Grid|Group|GroupBox|GuiClose|GuiContextMenu|GuiDropFiles|GuiEscape|GuiSize|Hdr|Hidden|Hide|High|HKCC|HKCR|HKCU|HKEY_CLASSES_ROOT|HKEY_CURRENT_CONFIG|HKEY_CURRENT_USER|HKEY_LOCAL_MACHINE|HKEY_USERS|HKLM|HKU|Hours|HScroll|Icon|IconSmall|ID|IDLast|If|IfEqual|IfExist|IfGreater|IfGreaterOrEqual|IfInString|IfLess|IfLessOrEqual|IfMsgBox|IfNotEqual|IfNotExist|IfNotInString|IfWinActive|IfWinExist|IfWinNotActive|IfWinNotExist|Ignore|ImageList|in|Integer|IntegerFast|Interrupt|is|italic|Join|Label|LastFound|LastFoundExist|Limit|Lines|List|ListBox|ListView|local|Lock|Logoff|Low|Lower|Lowercase|MainWindow|Margin|Maximize|MaximizeBox|MaxSize|Minimize|MinimizeBox|MinMax|MinSize|Minutes|MonthCal|Mouse|Move|Multi|NA|No|NoActivate|NoDefault|NoHide|NoIcon|NoMainWindow|norm|Normal|NoSort|NoSortHdr|NoStandard|Not|NoTab|NoTimers|Number|Off|Ok|On|OwnDialogs|Owner|Parse|Password|Picture|Pixel|Pos|Pow|Priority|ProcessName|Radio|Range|Read|ReadOnly|Realtime|Redraw|Region|REG_BINARY|REG_DWORD|REG_EXPAND_SZ|REG_MULTI_SZ|REG_SZ|Relative|Rename|Report|Resize|Restore|Retry|RGB|Screen|Seconds|Section|Serial|SetLabel|ShiftAltTab|Show|Single|Slider|SortDesc|Standard|static|Status|StatusBar|StatusCD|strike|Style|Submit|SysMenu|Tab2|TabStop|Text|Theme|Throw|Tile|ToggleCheck|ToggleEnable|ToolWindow|Top|Topmost|TransColor|Transparent|Tray|TreeView|Try|TryAgain|Type|UnCheck|underline|Unicode|Unlock|Until|UpDown|Upper|Uppercase|UseErrorLevel|Vis|VisFirst|Visible|VScroll|Wait|WaitClose|WantCtrlA|WantF2|WantReturn|While|Wrap|Xdigit|xm|xp|xs|Yes|ym|yp|ys)\b/i,function:/[^(); \t,\n+*\-=?>:\\\/<&%\[\]]+(?=\()/,punctuation:/[{}[\]():,]/}}return _y}var Cy,XI;function obe(){if(XI)return Cy;XI=1,Cy=e,e.displayName="autoit",e.aliases=[];function e(t){t.languages.autoit={comment:[/;.*/,{pattern:/(^[\t ]*)#(?:comments-start|cs)[\s\S]*?^[ \t]*#(?:ce|comments-end)/m,lookbehind:!0}],url:{pattern:/(^[\t ]*#include\s+)(?:<[^\r\n>]+>|"[^\r\n"]+")/m,lookbehind:!0},string:{pattern:/(["'])(?:\1\1|(?!\1)[^\r\n])*\1/,greedy:!0,inside:{variable:/([%$@])\w+\1/}},directive:{pattern:/(^[\t ]*)#[\w-]+/m,lookbehind:!0,alias:"keyword"},function:/\b\w+(?=\()/,variable:/[$@]\w+/,keyword:/\b(?:Case|Const|Continue(?:Case|Loop)|Default|Dim|Do|Else(?:If)?|End(?:Func|If|Select|Switch|With)|Enum|Exit(?:Loop)?|For|Func|Global|If|In|Local|Next|Null|ReDim|Select|Static|Step|Switch|Then|To|Until|Volatile|WEnd|While|With)\b/i,number:/\b(?:0x[\da-f]+|\d+(?:\.\d+)?(?:e[+-]?\d+)?)\b/i,boolean:/\b(?:False|True)\b/i,operator:/<[=>]?|[-+*\/=&>]=?|[?^]|\b(?:And|Not|Or)\b/i,punctuation:/[\[\]().,:]/}}return Cy}var Ny,ZI;function ibe(){if(ZI)return Ny;ZI=1,Ny=e,e.displayName="avisynth",e.aliases=["avs"];function e(t){(function(n){function r(p,g){return p.replace(/<<(\d+)>>/g,function(m,b){return g[+b]})}function a(p,g,m){return RegExp(r(p,g),m)}var o=/bool|clip|float|int|string|val/.source,s=[/is(?:bool|clip|float|int|string)|defined|(?:(?:internal)?function|var)?exists?/.source,/apply|assert|default|eval|import|nop|select|undefined/.source,/opt_(?:allowfloataudio|avipadscanlines|dwchannelmask|enable_(?:b64a|planartopackedrgb|v210|y3_10_10|y3_10_16)|usewaveextensible|vdubplanarhack)|set(?:cachemode|maxcpu|memorymax|planarlegacyalignment|workingdir)/.source,/hex(?:value)?|value/.source,/abs|ceil|continued(?:denominator|numerator)?|exp|floor|fmod|frac|log(?:10)?|max|min|muldiv|pi|pow|rand|round|sign|spline|sqrt/.source,/a?sinh?|a?cosh?|a?tan[2h]?/.source,/(?:bit(?:and|not|x?or|[lr]?shift[aslu]?|sh[lr]|sa[lr]|[lr]rotatel?|ro[rl]|te?st|set(?:count)?|cl(?:ea)?r|ch(?:an)?ge?))/.source,/average(?:[bgr]|chroma[uv]|luma)|(?:[rgb]|chroma[uv]|luma|rgb|[yuv](?=difference(?:fromprevious|tonext)))difference(?:fromprevious|tonext)?|[yuvrgb]plane(?:median|min|max|minmaxdifference)/.source,/getprocessinfo|logmsg|script(?:dir(?:utf8)?|file(?:utf8)?|name(?:utf8)?)|setlogparams/.source,/chr|(?:fill|find|left|mid|replace|rev|right)str|format|[lu]case|ord|str(?:cmpi?|fromutf8|len|toutf8)|time|trim(?:all|left|right)/.source,/isversionorgreater|version(?:number|string)/.source,/buildpixeltype|colorspacenametopixeltype/.source,/addautoloaddir|on(?:cpu|cuda)|prefetch|setfiltermtmode/.source].join("|"),u=[/has(?:audio|video)/.source,/height|width/.source,/frame(?:count|rate)|framerate(?:denominator|numerator)/.source,/getparity|is(?:field|frame)based/.source,/bitspercomponent|componentsize|hasalpha|is(?:planar(?:rgba?)?|interleaved|rgb(?:24|32|48|64)?|y(?:8|u(?:va?|y2))?|yv(?:12|16|24|411)|420|422|444|packedrgb)|numcomponents|pixeltype/.source,/audio(?:bits|channels|duration|length(?:[fs]|hi|lo)?|rate)|isaudio(?:float|int)/.source].join("|"),c=[/avi(?:file)?source|directshowsource|image(?:reader|source|sourceanim)|opendmlsource|segmented(?:avisource|directshowsource)|wavsource/.source,/coloryuv|convertbacktoyuy2|convertto(?:RGB(?:24|32|48|64)|(?:planar)?RGBA?|Y8?|YV(?:12|16|24|411)|YUVA?(?:411|420|422|444)|YUY2)|fixluminance|gr[ae]yscale|invert|levels|limiter|mergea?rgb|merge(?:chroma|luma)|rgbadjust|show(?:alpha|blue|green|red)|swapuv|tweak|[uv]toy8?|ytouv/.source,/(?:colorkey|reset)mask|layer|mask(?:hs)?|merge|overlay|subtract/.source,/addborders|(?:bicubic|bilinear|blackman|gauss|lanczos4|lanczos|point|sinc|spline(?:16|36|64))resize|crop(?:bottom)?|flip(?:horizontal|vertical)|(?:horizontal|vertical)?reduceby2|letterbox|skewrows|turn(?:180|left|right)/.source,/blur|fixbrokenchromaupsampling|generalconvolution|(?:spatial|temporal)soften|sharpen/.source,/trim|(?:un)?alignedsplice|(?:assume|assumescaled|change|convert)FPS|(?:delete|duplicate)frame|dissolve|fade(?:in|io|out)[02]?|freezeframe|interleave|loop|reverse|select(?:even|odd|(?:range)?every)/.source,/assume[bt]ff|assume(?:field|frame)based|bob|complementparity|doubleweave|peculiarblend|pulldown|separate(?:columns|fields|rows)|swapfields|weave(?:columns|rows)?/.source,/amplify(?:db)?|assumesamplerate|audiodub(?:ex)?|audiotrim|convertaudioto(?:(?:8|16|24|32)bit|float)|converttomono|delayaudio|ensurevbrmp3sync|get(?:left|right)?channel|kill(?:audio|video)|mergechannels|mixaudio|monotostereo|normalize|resampleaudio|ssrc|supereq|timestretch/.source,/animate|applyrange|conditional(?:filter|reader|select)|frameevaluate|scriptclip|tcp(?:server|source)|writefile(?:end|if|start)?/.source,/imagewriter/.source,/blackness|blankclip|colorbars(?:hd)?|compare|dumpfiltergraph|echo|histogram|info|messageclip|preroll|setgraphanalysis|show(?:framenumber|smpte|time)|showfiveversions|stack(?:horizontal|vertical)|subtitle|tone|version/.source].join("|"),d=[s,u,c].join("|");n.languages.avisynth={comment:[{pattern:/(^|[^\\])\[\*(?:[^\[*]|\[(?!\*)|\*(?!\])|\[\*(?:[^\[*]|\[(?!\*)|\*(?!\]))*\*\])*\*\]/,lookbehind:!0,greedy:!0},{pattern:/(^|[^\\])\/\*[\s\S]*?(?:\*\/|$)/,lookbehind:!0,greedy:!0},{pattern:/(^|[^\\$])#.*/,lookbehind:!0,greedy:!0}],argument:{pattern:a(/\b(?:<<0>>)\s+("?)\w+\1/.source,[o],"i"),inside:{keyword:/^\w+/}},"argument-label":{pattern:/([,(][\s\\]*)\w+\s*=(?!=)/,lookbehind:!0,inside:{"argument-name":{pattern:/^\w+/,alias:"punctuation"},punctuation:/=$/}},string:[{pattern:/"""[\s\S]*?"""/,greedy:!0},{pattern:/"(?:\\(?:\r\n|[\s\S])|[^"\\\r\n])*"/,greedy:!0,inside:{constant:{pattern:/\b(?:DEFAULT_MT_MODE|(?:MAINSCRIPT|PROGRAM|SCRIPT)DIR|(?:MACHINE|USER)_(?:CLASSIC|PLUS)_PLUGINS)\b/}}}],variable:/\b(?:last)\b/i,boolean:/\b(?:false|no|true|yes)\b/i,keyword:/\b(?:catch|else|for|function|global|if|return|try|while|__END__)\b/i,constant:/\bMT_(?:MULTI_INSTANCE|NICE_FILTER|SERIALIZED|SPECIAL_MT)\b/,"builtin-function":{pattern:a(/\b(?:<<0>>)\b/.source,[d],"i"),alias:"function"},"type-cast":{pattern:a(/\b(?:<<0>>)(?=\s*\()/.source,[o],"i"),alias:"keyword"},function:{pattern:/\b[a-z_]\w*(?=\s*\()|(\.)[a-z_]\w*\b/i,lookbehind:!0},"line-continuation":{pattern:/(^[ \t]*)\\|\\(?=[ \t]*$)/m,lookbehind:!0,alias:"punctuation"},number:/\B\$(?:[\da-f]{6}|[\da-f]{8})\b|(?:(?:\b|\B-)\d+(?:\.\d*)?\b|\B\.\d+\b)/i,operator:/\+\+?|[!=<>]=?|&&|\|\||[?:*/%-]/,punctuation:/[{}\[\]();,.]/},n.languages.avs=n.languages.avisynth})(t)}return Ny}var Oy,QI;function sbe(){if(QI)return Oy;QI=1,Oy=e,e.displayName="avroIdl",e.aliases=[];function e(t){t.languages["avro-idl"]={comment:{pattern:/\/\/.*|\/\*[\s\S]*?\*\//,greedy:!0},string:{pattern:/(^|[^\\])"(?:[^\r\n"\\]|\\.)*"/,lookbehind:!0,greedy:!0},annotation:{pattern:/@(?:[$\w.-]|`[^\r\n`]+`)+/,greedy:!0,alias:"function"},"function-identifier":{pattern:/`[^\r\n`]+`(?=\s*\()/,greedy:!0,alias:"function"},identifier:{pattern:/`[^\r\n`]+`/,greedy:!0},"class-name":{pattern:/(\b(?:enum|error|protocol|record|throws)\b\s+)[$\w]+/,lookbehind:!0,greedy:!0},keyword:/\b(?:array|boolean|bytes|date|decimal|double|enum|error|false|fixed|float|idl|import|int|local_timestamp_ms|long|map|null|oneway|protocol|record|schema|string|throws|time_ms|timestamp_ms|true|union|uuid|void)\b/,function:/\b[a-z_]\w*(?=\s*\()/i,number:[{pattern:/(^|[^\w.])-?(?:(?:\d+(?:\.\d*)?|\.\d+)(?:e[+-]?\d+)?|0x(?:[a-f0-9]+(?:\.[a-f0-9]*)?|\.[a-f0-9]+)(?:p[+-]?\d+)?)[dfl]?(?![\w.])/i,lookbehind:!0},/-?\b(?:Infinity|NaN)\b/],operator:/=/,punctuation:/[()\[\]{}<>.:,;-]/},t.languages.avdl=t.languages["avro-idl"]}return Oy}var Dy,JI;function GH(){if(JI)return Dy;JI=1,Dy=e,e.displayName="bash",e.aliases=["shell"];function e(t){(function(n){var r="\\b(?:BASH|BASHOPTS|BASH_ALIASES|BASH_ARGC|BASH_ARGV|BASH_CMDS|BASH_COMPLETION_COMPAT_DIR|BASH_LINENO|BASH_REMATCH|BASH_SOURCE|BASH_VERSINFO|BASH_VERSION|COLORTERM|COLUMNS|COMP_WORDBREAKS|DBUS_SESSION_BUS_ADDRESS|DEFAULTS_PATH|DESKTOP_SESSION|DIRSTACK|DISPLAY|EUID|GDMSESSION|GDM_LANG|GNOME_KEYRING_CONTROL|GNOME_KEYRING_PID|GPG_AGENT_INFO|GROUPS|HISTCONTROL|HISTFILE|HISTFILESIZE|HISTSIZE|HOME|HOSTNAME|HOSTTYPE|IFS|INSTANCE|JOB|LANG|LANGUAGE|LC_ADDRESS|LC_ALL|LC_IDENTIFICATION|LC_MEASUREMENT|LC_MONETARY|LC_NAME|LC_NUMERIC|LC_PAPER|LC_TELEPHONE|LC_TIME|LESSCLOSE|LESSOPEN|LINES|LOGNAME|LS_COLORS|MACHTYPE|MAILCHECK|MANDATORY_PATH|NO_AT_BRIDGE|OLDPWD|OPTERR|OPTIND|ORBIT_SOCKETDIR|OSTYPE|PAPERSIZE|PATH|PIPESTATUS|PPID|PS1|PS2|PS3|PS4|PWD|RANDOM|REPLY|SECONDS|SELINUX_INIT|SESSION|SESSIONTYPE|SESSION_MANAGER|SHELL|SHELLOPTS|SHLVL|SSH_AUTH_SOCK|TERM|UID|UPSTART_EVENTS|UPSTART_INSTANCE|UPSTART_JOB|UPSTART_SESSION|USER|WINDOWID|XAUTHORITY|XDG_CONFIG_DIRS|XDG_CURRENT_DESKTOP|XDG_DATA_DIRS|XDG_GREETER_DATA_DIR|XDG_MENU_PREFIX|XDG_RUNTIME_DIR|XDG_SEAT|XDG_SEAT_PATH|XDG_SESSION_DESKTOP|XDG_SESSION_ID|XDG_SESSION_PATH|XDG_SESSION_TYPE|XDG_VTNR|XMODIFIERS)\\b",a={pattern:/(^(["']?)\w+\2)[ \t]+\S.*/,lookbehind:!0,alias:"punctuation",inside:null},o={bash:a,environment:{pattern:RegExp("\\$"+r),alias:"constant"},variable:[{pattern:/\$?\(\([\s\S]+?\)\)/,greedy:!0,inside:{variable:[{pattern:/(^\$\(\([\s\S]+)\)\)/,lookbehind:!0},/^\$\(\(/],number:/\b0x[\dA-Fa-f]+\b|(?:\b\d+(?:\.\d*)?|\B\.\d+)(?:[Ee]-?\d+)?/,operator:/--|\+\+|\*\*=?|<<=?|>>=?|&&|\|\||[=!+\-*/%<>^&|]=?|[?~:]/,punctuation:/\(\(?|\)\)?|,|;/}},{pattern:/\$\((?:\([^)]+\)|[^()])+\)|`[^`]+`/,greedy:!0,inside:{variable:/^\$\(|^`|\)$|`$/}},{pattern:/\$\{[^}]+\}/,greedy:!0,inside:{operator:/:[-=?+]?|[!\/]|##?|%%?|\^\^?|,,?/,punctuation:/[\[\]]/,environment:{pattern:RegExp("(\\{)"+r),lookbehind:!0,alias:"constant"}}},/\$(?:\w+|[#?*!@$])/],entity:/\\(?:[abceEfnrtv\\"]|O?[0-7]{1,3}|U[0-9a-fA-F]{8}|u[0-9a-fA-F]{4}|x[0-9a-fA-F]{1,2})/};n.languages.bash={shebang:{pattern:/^#!\s*\/.*/,alias:"important"},comment:{pattern:/(^|[^"{\\$])#.*/,lookbehind:!0},"function-name":[{pattern:/(\bfunction\s+)[\w-]+(?=(?:\s*\(?:\s*\))?\s*\{)/,lookbehind:!0,alias:"function"},{pattern:/\b[\w-]+(?=\s*\(\s*\)\s*\{)/,alias:"function"}],"for-or-select":{pattern:/(\b(?:for|select)\s+)\w+(?=\s+in\s)/,alias:"variable",lookbehind:!0},"assign-left":{pattern:/(^|[\s;|&]|[<>]\()\w+(?=\+?=)/,inside:{environment:{pattern:RegExp("(^|[\\s;|&]|[<>]\\()"+r),lookbehind:!0,alias:"constant"}},alias:"variable",lookbehind:!0},string:[{pattern:/((?:^|[^<])<<-?\s*)(\w+)\s[\s\S]*?(?:\r?\n|\r)\2/,lookbehind:!0,greedy:!0,inside:o},{pattern:/((?:^|[^<])<<-?\s*)(["'])(\w+)\2\s[\s\S]*?(?:\r?\n|\r)\3/,lookbehind:!0,greedy:!0,inside:{bash:a}},{pattern:/(^|[^\\](?:\\\\)*)"(?:\\[\s\S]|\$\([^)]+\)|\$(?!\()|`[^`]+`|[^"\\`$])*"/,lookbehind:!0,greedy:!0,inside:o},{pattern:/(^|[^$\\])'[^']*'/,lookbehind:!0,greedy:!0},{pattern:/\$'(?:[^'\\]|\\[\s\S])*'/,greedy:!0,inside:{entity:o.entity}}],environment:{pattern:RegExp("\\$?"+r),alias:"constant"},variable:o.variable,function:{pattern:/(^|[\s;|&]|[<>]\()(?:add|apropos|apt|apt-cache|apt-get|aptitude|aspell|automysqlbackup|awk|basename|bash|bc|bconsole|bg|bzip2|cal|cat|cfdisk|chgrp|chkconfig|chmod|chown|chroot|cksum|clear|cmp|column|comm|composer|cp|cron|crontab|csplit|curl|cut|date|dc|dd|ddrescue|debootstrap|df|diff|diff3|dig|dir|dircolors|dirname|dirs|dmesg|docker|docker-compose|du|egrep|eject|env|ethtool|expand|expect|expr|fdformat|fdisk|fg|fgrep|file|find|fmt|fold|format|free|fsck|ftp|fuser|gawk|git|gparted|grep|groupadd|groupdel|groupmod|groups|grub-mkconfig|gzip|halt|head|hg|history|host|hostname|htop|iconv|id|ifconfig|ifdown|ifup|import|install|ip|jobs|join|kill|killall|less|link|ln|locate|logname|logrotate|look|lpc|lpr|lprint|lprintd|lprintq|lprm|ls|lsof|lynx|make|man|mc|mdadm|mkconfig|mkdir|mke2fs|mkfifo|mkfs|mkisofs|mknod|mkswap|mmv|more|most|mount|mtools|mtr|mutt|mv|nano|nc|netstat|nice|nl|node|nohup|notify-send|npm|nslookup|op|open|parted|passwd|paste|pathchk|ping|pkill|pnpm|podman|podman-compose|popd|pr|printcap|printenv|ps|pushd|pv|quota|quotacheck|quotactl|ram|rar|rcp|reboot|remsync|rename|renice|rev|rm|rmdir|rpm|rsync|scp|screen|sdiff|sed|sendmail|seq|service|sftp|sh|shellcheck|shuf|shutdown|sleep|slocate|sort|split|ssh|stat|strace|su|sudo|sum|suspend|swapon|sync|tac|tail|tar|tee|time|timeout|top|touch|tr|traceroute|tsort|tty|umount|uname|unexpand|uniq|units|unrar|unshar|unzip|update-grub|uptime|useradd|userdel|usermod|users|uudecode|uuencode|v|vcpkg|vdir|vi|vim|virsh|vmstat|wait|watch|wc|wget|whereis|which|who|whoami|write|xargs|xdg-open|yarn|yes|zenity|zip|zsh|zypper)(?=$|[)\s;|&])/,lookbehind:!0},keyword:{pattern:/(^|[\s;|&]|[<>]\()(?:case|do|done|elif|else|esac|fi|for|function|if|in|select|then|until|while)(?=$|[)\s;|&])/,lookbehind:!0},builtin:{pattern:/(^|[\s;|&]|[<>]\()(?:\.|:|alias|bind|break|builtin|caller|cd|command|continue|declare|echo|enable|eval|exec|exit|export|getopts|hash|help|let|local|logout|mapfile|printf|pwd|read|readarray|readonly|return|set|shift|shopt|source|test|times|trap|type|typeset|ulimit|umask|unalias|unset)(?=$|[)\s;|&])/,lookbehind:!0,alias:"class-name"},boolean:{pattern:/(^|[\s;|&]|[<>]\()(?:false|true)(?=$|[)\s;|&])/,lookbehind:!0},"file-descriptor":{pattern:/\B&\d\b/,alias:"important"},operator:{pattern:/\d?<>|>\||\+=|=[=~]?|!=?|<<[<-]?|[&\d]?>>|\d[<>]&?|[<>][&=]?|&[>&]?|\|[&|]?/,inside:{"file-descriptor":{pattern:/^\d/,alias:"important"}}},punctuation:/\$?\(\(?|\)\)?|\.\.|[{}[\];\\]/,number:{pattern:/(^|\s)(?:[1-9]\d*|0)(?:[.,]\d+)?\b/,lookbehind:!0}},a.inside=n.languages.bash;for(var s=["comment","function-name","for-or-select","assign-left","string","environment","function","keyword","builtin","boolean","file-descriptor","operator","punctuation","number"],u=o.variable[1].inside,c=0;c<s.length;c++)u[s[c]]=n.languages.bash[s[c]];n.languages.shell=n.languages.bash})(t)}return Dy}var Iy,eL;function HH(){if(eL)return Iy;eL=1,Iy=e,e.displayName="basic",e.aliases=[];function e(t){t.languages.basic={comment:{pattern:/(?:!|REM\b).+/i,inside:{keyword:/^REM/i}},string:{pattern:/"(?:""|[!#$%&'()*,\/:;<=>?^\w +\-.])*"/,greedy:!0},number:/(?:\b\d+(?:\.\d*)?|\B\.\d+)(?:E[+-]?\d+)?/i,keyword:/\b(?:AS|BEEP|BLOAD|BSAVE|CALL(?: ABSOLUTE)?|CASE|CHAIN|CHDIR|CLEAR|CLOSE|CLS|COM|COMMON|CONST|DATA|DECLARE|DEF(?: FN| SEG|DBL|INT|LNG|SNG|STR)|DIM|DO|DOUBLE|ELSE|ELSEIF|END|ENVIRON|ERASE|ERROR|EXIT|FIELD|FILES|FOR|FUNCTION|GET|GOSUB|GOTO|IF|INPUT|INTEGER|IOCTL|KEY|KILL|LINE INPUT|LOCATE|LOCK|LONG|LOOP|LSET|MKDIR|NAME|NEXT|OFF|ON(?: COM| ERROR| KEY| TIMER)?|OPEN|OPTION BASE|OUT|POKE|PUT|READ|REDIM|REM|RESTORE|RESUME|RETURN|RMDIR|RSET|RUN|SELECT CASE|SHARED|SHELL|SINGLE|SLEEP|STATIC|STEP|STOP|STRING|SUB|SWAP|SYSTEM|THEN|TIMER|TO|TROFF|TRON|TYPE|UNLOCK|UNTIL|USING|VIEW PRINT|WAIT|WEND|WHILE|WRITE)(?:\$|\b)/i,function:/\b(?:ABS|ACCESS|ACOS|ANGLE|AREA|ARITHMETIC|ARRAY|ASIN|ASK|AT|ATN|BASE|BEGIN|BREAK|CAUSE|CEIL|CHR|CLIP|COLLATE|COLOR|CON|COS|COSH|COT|CSC|DATE|DATUM|DEBUG|DECIMAL|DEF|DEG|DEGREES|DELETE|DET|DEVICE|DISPLAY|DOT|ELAPSED|EPS|ERASABLE|EXLINE|EXP|EXTERNAL|EXTYPE|FILETYPE|FIXED|FP|GO|GRAPH|HANDLER|IDN|IMAGE|IN|INT|INTERNAL|IP|IS|KEYED|LBOUND|LCASE|LEFT|LEN|LENGTH|LET|LINE|LINES|LOG|LOG10|LOG2|LTRIM|MARGIN|MAT|MAX|MAXNUM|MID|MIN|MISSING|MOD|NATIVE|NUL|NUMERIC|OF|OPTION|ORD|ORGANIZATION|OUTIN|OUTPUT|PI|POINT|POINTER|POINTS|POS|PRINT|PROGRAM|PROMPT|RAD|RADIANS|RANDOMIZE|RECORD|RECSIZE|RECTYPE|RELATIVE|REMAINDER|REPEAT|REST|RETRY|REWRITE|RIGHT|RND|ROUND|RTRIM|SAME|SEC|SELECT|SEQUENTIAL|SET|SETTER|SGN|SIN|SINH|SIZE|SKIP|SQR|STANDARD|STATUS|STR|STREAM|STYLE|TAB|TAN|TANH|TEMPLATE|TEXT|THERE|TIME|TIMEOUT|TRACE|TRANSFORM|TRUNCATE|UBOUND|UCASE|USE|VAL|VARIABLE|VIEWPORT|WHEN|WINDOW|WITH|ZER|ZONEWIDTH)(?:\$|\b)/i,operator:/<[=>]?|>=?|[+\-*\/^=&]|\b(?:AND|EQV|IMP|NOT|OR|XOR)\b/i,punctuation:/[,;:()]/}}return Iy}var Ly,tL;function lbe(){if(tL)return Ly;tL=1,Ly=e,e.displayName="batch",e.aliases=[];function e(t){(function(n){var r=/%%?[~:\w]+%?|!\S+!/,a={pattern:/\/[a-z?]+(?=[ :]|$):?|-[a-z]\b|--[a-z-]+\b/im,alias:"attr-name",inside:{punctuation:/:/}},o=/"(?:[\\"]"|[^"])*"(?!")/,s=/(?:\b|-)\d+\b/;n.languages.batch={comment:[/^::.*/m,{pattern:/((?:^|[&(])[ \t]*)rem\b(?:[^^&)\r\n]|\^(?:\r\n|[\s\S]))*/im,lookbehind:!0}],label:{pattern:/^:.*/m,alias:"property"},command:[{pattern:/((?:^|[&(])[ \t]*)for(?: \/[a-z?](?:[ :](?:"[^"]*"|[^\s"/]\S*))?)* \S+ in \([^)]+\) do/im,lookbehind:!0,inside:{keyword:/\b(?:do|in)\b|^for\b/i,string:o,parameter:a,variable:r,number:s,punctuation:/[()',]/}},{pattern:/((?:^|[&(])[ \t]*)if(?: \/[a-z?](?:[ :](?:"[^"]*"|[^\s"/]\S*))?)* (?:not )?(?:cmdextversion \d+|defined \w+|errorlevel \d+|exist \S+|(?:"[^"]*"|(?!")(?:(?!==)\S)+)?(?:==| (?:equ|geq|gtr|leq|lss|neq) )(?:"[^"]*"|[^\s"]\S*))/im,lookbehind:!0,inside:{keyword:/\b(?:cmdextversion|defined|errorlevel|exist|not)\b|^if\b/i,string:o,parameter:a,variable:r,number:s,operator:/\^|==|\b(?:equ|geq|gtr|leq|lss|neq)\b/i}},{pattern:/((?:^|[&()])[ \t]*)else\b/im,lookbehind:!0,inside:{keyword:/^else\b/i}},{pattern:/((?:^|[&(])[ \t]*)set(?: \/[a-z](?:[ :](?:"[^"]*"|[^\s"/]\S*))?)* (?:[^^&)\r\n]|\^(?:\r\n|[\s\S]))*/im,lookbehind:!0,inside:{keyword:/^set\b/i,string:o,parameter:a,variable:[r,/\w+(?=(?:[*\/%+\-&^|]|<<|>>)?=)/],number:s,operator:/[*\/%+\-&^|]=?|<<=?|>>=?|[!~_=]/,punctuation:/[()',]/}},{pattern:/((?:^|[&(])[ \t]*@?)\w+\b(?:"(?:[\\"]"|[^"])*"(?!")|[^"^&)\r\n]|\^(?:\r\n|[\s\S]))*/m,lookbehind:!0,inside:{keyword:/^\w+\b/,string:o,parameter:a,label:{pattern:/(^\s*):\S+/m,lookbehind:!0,alias:"property"},variable:r,number:s,operator:/\^/}}],operator:/[&@]/,punctuation:/[()']/}})(t)}return Ly}var My,nL;function ube(){if(nL)return My;nL=1,My=e,e.displayName="bbcode",e.aliases=["shortcode"];function e(t){t.languages.bbcode={tag:{pattern:/\[\/?[^\s=\]]+(?:\s*=\s*(?:"[^"]*"|'[^']*'|[^\s'"\]=]+))?(?:\s+[^\s=\]]+\s*=\s*(?:"[^"]*"|'[^']*'|[^\s'"\]=]+))*\s*\]/,inside:{tag:{pattern:/^\[\/?[^\s=\]]+/,inside:{punctuation:/^\[\/?/}},"attr-value":{pattern:/=\s*(?:"[^"]*"|'[^']*'|[^\s'"\]=]+)/,inside:{punctuation:[/^=/,{pattern:/^(\s*)["']|["']$/,lookbehind:!0}]}},punctuation:/\]/,"attr-name":/[^\s=\]]+/}}},t.languages.shortcode=t.languages.bbcode}return My}var Fy,rL;function cbe(){if(rL)return Fy;rL=1,Fy=e,e.displayName="bicep",e.aliases=[];function e(t){t.languages.bicep={comment:[{pattern:/(^|[^\\])\/\*[\s\S]*?(?:\*\/|$)/,lookbehind:!0,greedy:!0},{pattern:/(^|[^\\:])\/\/.*/,lookbehind:!0,greedy:!0}],property:[{pattern:/([\r\n][ \t]*)[a-z_]\w*(?=[ \t]*:)/i,lookbehind:!0},{pattern:/([\r\n][ \t]*)'(?:\\.|\$(?!\{)|[^'\\\r\n$])*'(?=[ \t]*:)/,lookbehind:!0,greedy:!0}],string:[{pattern:/'''[^'][\s\S]*?'''/,greedy:!0},{pattern:/(^|[^\\'])'(?:\\.|\$(?!\{)|[^'\\\r\n$])*'/,lookbehind:!0,greedy:!0}],"interpolated-string":{pattern:/(^|[^\\'])'(?:\\.|\$(?:(?!\{)|\{[^{}\r\n]*\})|[^'\\\r\n$])*'/,lookbehind:!0,greedy:!0,inside:{interpolation:{pattern:/\$\{[^{}\r\n]*\}/,inside:{expression:{pattern:/(^\$\{)[\s\S]+(?=\}$)/,lookbehind:!0},punctuation:/^\$\{|\}$/}},string:/[\s\S]+/}},datatype:{pattern:/(\b(?:output|param)\b[ \t]+\w+[ \t]+)\w+\b/,lookbehind:!0,alias:"class-name"},boolean:/\b(?:false|true)\b/,keyword:/\b(?:existing|for|if|in|module|null|output|param|resource|targetScope|var)\b/,decorator:/@\w+\b/,function:/\b[a-z_]\w*(?=[ \t]*\()/i,number:/(?:\b\d+(?:\.\d*)?|\B\.\d+)(?:E[+-]?\d+)?/i,operator:/--|\+\+|\*\*=?|=>|&&=?|\|\|=?|[!=]==|<<=?|>>>?=?|[-+*/%&|^!=<>]=?|\.{3}|\?\?=?|\?\.?|[~:]/,punctuation:/[{}[\];(),.:]/},t.languages.bicep["interpolated-string"].inside.interpolation.inside.expression.inside=t.languages.bicep}return Fy}var Py,aL;function dbe(){if(aL)return Py;aL=1,Py=e,e.displayName="birb",e.aliases=[];function e(t){t.languages.birb=t.languages.extend("clike",{string:{pattern:/r?("|')(?:\\.|(?!\1)[^\\])*\1/,greedy:!0},"class-name":[/\b[A-Z](?:[\d_]*[a-zA-Z]\w*)?\b/,/\b(?:[A-Z]\w*|(?!(?:var|void)\b)[a-z]\w*)(?=\s+\w+\s*[;,=()])/],keyword:/\b(?:assert|break|case|class|const|default|else|enum|final|follows|for|grab|if|nest|new|next|noSeeb|return|static|switch|throw|var|void|while)\b/,operator:/\+\+|--|&&|\|\||<<=?|>>=?|~(?:\/=?)?|[+\-*\/%&^|=!<>]=?|\?|:/,variable:/\b[a-z_]\w*\b/}),t.languages.insertBefore("birb","function",{metadata:{pattern:/<\w+>/,greedy:!0,alias:"symbol"}})}return Py}var zy,oL;function fbe(){if(oL)return zy;oL=1;var e=Ai();zy=t,t.displayName="bison",t.aliases=[];function t(n){n.register(e),n.languages.bison=n.languages.extend("c",{}),n.languages.insertBefore("bison","comment",{bison:{pattern:/^(?:[^%]|%(?!%))*%%[\s\S]*?%%/,inside:{c:{pattern:/%\{[\s\S]*?%\}|\{(?:\{[^}]*\}|[^{}])*\}/,inside:{delimiter:{pattern:/^%?\{|%?\}$/,alias:"punctuation"},"bison-variable":{pattern:/[$@](?:<[^\s>]+>)?[\w$]+/,alias:"variable",inside:{punctuation:/<|>/}},rest:n.languages.c}},comment:n.languages.c.comment,string:n.languages.c.string,property:/\S+(?=:)/,keyword:/%\w+/,number:{pattern:/(^|[^@])\b(?:0x[\da-f]+|\d+)/i,lookbehind:!0},punctuation:/%[%?]|[|:;\[\]<>]/}}})}return zy}var By,iL;function pbe(){if(iL)return By;iL=1,By=e,e.displayName="bnf",e.aliases=["rbnf"];function e(t){t.languages.bnf={string:{pattern:/"[^\r\n"]*"|'[^\r\n']*'/},definition:{pattern:/<[^<>\r\n\t]+>(?=\s*::=)/,alias:["rule","keyword"],inside:{punctuation:/^<|>$/}},rule:{pattern:/<[^<>\r\n\t]+>/,inside:{punctuation:/^<|>$/}},operator:/::=|[|()[\]{}*+?]|\.{3}/},t.languages.rbnf=t.languages.bnf}return By}var Uy,sL;function gbe(){if(sL)return Uy;sL=1,Uy=e,e.displayName="brainfuck",e.aliases=[];function e(t){t.languages.brainfuck={pointer:{pattern:/<|>/,alias:"keyword"},increment:{pattern:/\+/,alias:"inserted"},decrement:{pattern:/-/,alias:"deleted"},branching:{pattern:/\[|\]/,alias:"important"},operator:/[.,]/,comment:/\S+/}}return Uy}var jy,lL;function hbe(){if(lL)return jy;lL=1,jy=e,e.displayName="brightscript",e.aliases=[];function e(t){t.languages.brightscript={comment:/(?:\brem|').*/i,"directive-statement":{pattern:/(^[\t ]*)#(?:const|else(?:[\t ]+if)?|end[\t ]+if|error|if).*/im,lookbehind:!0,alias:"property",inside:{"error-message":{pattern:/(^#error).+/,lookbehind:!0},directive:{pattern:/^#(?:const|else(?:[\t ]+if)?|end[\t ]+if|error|if)/,alias:"keyword"},expression:{pattern:/[\s\S]+/,inside:null}}},property:{pattern:/([\r\n{,][\t ]*)(?:(?!\d)\w+|"(?:[^"\r\n]|"")*"(?!"))(?=[ \t]*:)/,lookbehind:!0,greedy:!0},string:{pattern:/"(?:[^"\r\n]|"")*"(?!")/,greedy:!0},"class-name":{pattern:/(\bAs[\t ]+)\w+/i,lookbehind:!0},keyword:/\b(?:As|Dim|Each|Else|Elseif|End|Exit|For|Function|Goto|If|In|Print|Return|Step|Stop|Sub|Then|To|While)\b/i,boolean:/\b(?:false|true)\b/i,function:/\b(?!\d)\w+(?=[\t ]*\()/,number:/(?:\b\d+(?:\.\d+)?(?:[ed][+-]\d+)?|&h[a-f\d]+)\b[%&!#]?/i,operator:/--|\+\+|>>=?|<<=?|<>|[-+*/\\<>]=?|[:^=?]|\b(?:and|mod|not|or)\b/i,punctuation:/[.,;()[\]{}]/,constant:/\b(?:LINE_NUM)\b/i},t.languages.brightscript["directive-statement"].inside.expression.inside=t.languages.brightscript}return jy}var Gy,uL;function mbe(){if(uL)return Gy;uL=1,Gy=e,e.displayName="bro",e.aliases=[];function e(t){t.languages.bro={comment:{pattern:/(^|[^\\$])#.*/,lookbehind:!0,inside:{italic:/\b(?:FIXME|TODO|XXX)\b/}},string:{pattern:/(["'])(?:\\(?:\r\n|[\s\S])|(?!\1)[^\\\r\n])*\1/,greedy:!0},boolean:/\b[TF]\b/,function:{pattern:/(\b(?:event|function|hook)[ \t]+)\w+(?:::\w+)?/,lookbehind:!0},builtin:/(?:@(?:load(?:-(?:plugin|sigs))?|unload|prefixes|ifn?def|else|(?:end)?if|DIR|FILENAME))|(?:&?(?:add_func|create_expire|default|delete_func|encrypt|error_handler|expire_func|group|log|mergeable|optional|persistent|priority|raw_output|read_expire|redef|rotate_interval|rotate_size|synchronized|type_column|write_expire))/,constant:{pattern:/(\bconst[ \t]+)\w+/i,lookbehind:!0},keyword:/\b(?:add|addr|alarm|any|bool|break|const|continue|count|delete|double|else|enum|event|export|file|for|function|global|hook|if|in|int|interval|local|module|next|of|opaque|pattern|port|print|record|return|schedule|set|string|subnet|table|time|timeout|using|vector|when)\b/,operator:/--?|\+\+?|!=?=?|<=?|>=?|==?=?|&&|\|\|?|\?|\*|\/|~|\^|%/,number:/\b0x[\da-f]+\b|(?:\b\d+(?:\.\d*)?|\B\.\d+)(?:e[+-]?\d+)?/i,punctuation:/[{}[\];(),.:]/}}return Gy}var Hy,cL;function bbe(){if(cL)return Hy;cL=1,Hy=e,e.displayName="bsl",e.aliases=[];function e(t){t.languages.bsl={comment:/\/\/.*/,string:[{pattern:/"(?:[^"]|"")*"(?!")/,greedy:!0},{pattern:/'(?:[^'\r\n\\]|\\.)*'/}],keyword:[{pattern:/(^|[^\w\u0400-\u0484\u0487-\u052f\u1d2b\u1d78\u2de0-\u2dff\ua640-\ua69f\ufe2e\ufe2f])(?:пока|для|новый|прервать|попытка|исключение|вызватьисключение|иначе|конецпопытки|неопределено|функция|перем|возврат|конецфункции|если|иначеесли|процедура|конецпроцедуры|тогда|знач|экспорт|конецесли|из|каждого|истина|ложь|по|цикл|конеццикла|выполнить)(?![\w\u0400-\u0484\u0487-\u052f\u1d2b\u1d78\u2de0-\u2dff\ua640-\ua69f\ufe2e\ufe2f])/i,lookbehind:!0},{pattern:/\b(?:break|do|each|else|elseif|enddo|endfunction|endif|endprocedure|endtry|except|execute|export|false|for|function|if|in|new|null|procedure|raise|return|then|to|true|try|undefined|val|var|while)\b/i}],number:{pattern:/(^(?=\d)|[^\w\u0400-\u0484\u0487-\u052f\u1d2b\u1d78\u2de0-\u2dff\ua640-\ua69f\ufe2e\ufe2f])(?:\d+(?:\.\d*)?|\.\d+)(?:E[+-]?\d+)?/i,lookbehind:!0},operator:[/[<>+\-*/]=?|[%=]/,{pattern:/(^|[^\w\u0400-\u0484\u0487-\u052f\u1d2b\u1d78\u2de0-\u2dff\ua640-\ua69f\ufe2e\ufe2f])(?:и|или|не)(?![\w\u0400-\u0484\u0487-\u052f\u1d2b\u1d78\u2de0-\u2dff\ua640-\ua69f\ufe2e\ufe2f])/i,lookbehind:!0},{pattern:/\b(?:and|not|or)\b/i}],punctuation:/\(\.|\.\)|[()\[\]:;,.]/,directive:[{pattern:/^([ \t]*)&.*/m,lookbehind:!0,greedy:!0,alias:"important"},{pattern:/^([ \t]*)#.*/gm,lookbehind:!0,greedy:!0,alias:"important"}]},t.languages.oscript=t.languages.bsl}return Hy}var $y,dL;function ybe(){if(dL)return $y;dL=1,$y=e,e.displayName="cfscript",e.aliases=[];function e(t){t.languages.cfscript=t.languages.extend("clike",{comment:[{pattern:/(^|[^\\])\/\*[\s\S]*?(?:\*\/|$)/,lookbehind:!0,inside:{annotation:{pattern:/(?:^|[^.])@[\w\.]+/,alias:"punctuation"}}},{pattern:/(^|[^\\:])\/\/.*/,lookbehind:!0,greedy:!0}],keyword:/\b(?:abstract|break|catch|component|continue|default|do|else|extends|final|finally|for|function|if|in|include|package|private|property|public|remote|required|rethrow|return|static|switch|throw|try|var|while|xml)\b(?!\s*=)/,operator:[/\+\+|--|&&|\|\||::|=>|[!=]==|<=?|>=?|[-+*/%&|^!=<>]=?|\?(?:\.|:)?|[?:]/,/\b(?:and|contains|eq|equal|eqv|gt|gte|imp|is|lt|lte|mod|not|or|xor)\b/],scope:{pattern:/\b(?:application|arguments|cgi|client|cookie|local|session|super|this|variables)\b/,alias:"global"},type:{pattern:/\b(?:any|array|binary|boolean|date|guid|numeric|query|string|struct|uuid|void|xml)\b/,alias:"builtin"}}),t.languages.insertBefore("cfscript","keyword",{"function-variable":{pattern:/[_$a-zA-Z\xA0-\uFFFF](?:(?!\s)[$\w\xA0-\uFFFF])*(?=\s*[=:]\s*(?:\bfunction\b|(?:\((?:[^()]|\([^()]*\))*\)|(?!\s)[_$a-zA-Z\xA0-\uFFFF](?:(?!\s)[$\w\xA0-\uFFFF])*)\s*=>))/,alias:"function"}}),delete t.languages.cfscript["class-name"],t.languages.cfc=t.languages.cfscript}return $y}var qy,fL;function vbe(){if(fL)return qy;fL=1;var e=uA();qy=t,t.displayName="chaiscript",t.aliases=[];function t(n){n.register(e),n.languages.chaiscript=n.languages.extend("clike",{string:{pattern:/(^|[^\\])'(?:[^'\\]|\\[\s\S])*'/,lookbehind:!0,greedy:!0},"class-name":[{pattern:/(\bclass\s+)\w+/,lookbehind:!0},{pattern:/(\b(?:attr|def)\s+)\w+(?=\s*::)/,lookbehind:!0}],keyword:/\b(?:attr|auto|break|case|catch|class|continue|def|default|else|finally|for|fun|global|if|return|switch|this|try|var|while)\b/,number:[n.languages.cpp.number,/\b(?:Infinity|NaN)\b/],operator:/>>=?|<<=?|\|\||&&|:[:=]?|--|\+\+|[=!<>+\-*/%|&^]=?|[?~]|`[^`\r\n]{1,4}`/}),n.languages.insertBefore("chaiscript","operator",{"parameter-type":{pattern:/([,(]\s*)\w+(?=\s+\w)/,lookbehind:!0,alias:"class-name"}}),n.languages.insertBefore("chaiscript","string",{"string-interpolation":{pattern:/(^|[^\\])"(?:[^"$\\]|\\[\s\S]|\$(?!\{)|\$\{(?:[^{}]|\{(?:[^{}]|\{[^{}]*\})*\})*\})*"/,lookbehind:!0,greedy:!0,inside:{interpolation:{pattern:/((?:^|[^\\])(?:\\{2})*)\$\{(?:[^{}]|\{(?:[^{}]|\{[^{}]*\})*\})*\}/,lookbehind:!0,inside:{"interpolation-expression":{pattern:/(^\$\{)[\s\S]+(?=\}$)/,lookbehind:!0,inside:n.languages.chaiscript},"interpolation-punctuation":{pattern:/^\$\{|\}$/,alias:"punctuation"}}},string:/[\s\S]+/}}})}return qy}var Vy,pL;function Sbe(){if(pL)return Vy;pL=1,Vy=e,e.displayName="cil",e.aliases=[];function e(t){t.languages.cil={comment:/\/\/.*/,string:{pattern:/(["'])(?:\\(?:\r\n|[\s\S])|(?!\1)[^\\\r\n])*\1/,greedy:!0},directive:{pattern:/(^|\W)\.[a-z]+(?=\s)/,lookbehind:!0,alias:"class-name"},variable:/\[[\w\.]+\]/,keyword:/\b(?:abstract|ansi|assembly|auto|autochar|beforefieldinit|bool|bstr|byvalstr|catch|char|cil|class|currency|date|decimal|default|enum|error|explicit|extends|extern|famandassem|family|famorassem|final(?:ly)?|float32|float64|hidebysig|u?int(?:8|16|32|64)?|iant|idispatch|implements|import|initonly|instance|interface|iunknown|literal|lpstr|lpstruct|lptstr|lpwstr|managed|method|native(?:Type)?|nested|newslot|object(?:ref)?|pinvokeimpl|private|privatescope|public|reqsecobj|rtspecialname|runtime|sealed|sequential|serializable|specialname|static|string|struct|syschar|tbstr|unicode|unmanagedexp|unsigned|value(?:type)?|variant|virtual|void)\b/,function:/\b(?:(?:constrained|no|readonly|tail|unaligned|volatile)\.)?(?:conv\.(?:[iu][1248]?|ovf\.[iu][1248]?(?:\.un)?|r\.un|r4|r8)|ldc\.(?:i4(?:\.\d+|\.[mM]1|\.s)?|i8|r4|r8)|ldelem(?:\.[iu][1248]?|\.r[48]|\.ref|a)?|ldind\.(?:[iu][1248]?|r[48]|ref)|stelem\.?(?:i[1248]?|r[48]|ref)?|stind\.(?:i[1248]?|r[48]|ref)?|end(?:fault|filter|finally)|ldarg(?:\.[0-3s]|a(?:\.s)?)?|ldloc(?:\.\d+|\.s)?|sub(?:\.ovf(?:\.un)?)?|mul(?:\.ovf(?:\.un)?)?|add(?:\.ovf(?:\.un)?)?|stloc(?:\.[0-3s])?|refany(?:type|val)|blt(?:\.un)?(?:\.s)?|ble(?:\.un)?(?:\.s)?|bgt(?:\.un)?(?:\.s)?|bge(?:\.un)?(?:\.s)?|unbox(?:\.any)?|init(?:blk|obj)|call(?:i|virt)?|brfalse(?:\.s)?|bne\.un(?:\.s)?|ldloca(?:\.s)?|brzero(?:\.s)?|brtrue(?:\.s)?|brnull(?:\.s)?|brinst(?:\.s)?|starg(?:\.s)?|leave(?:\.s)?|shr(?:\.un)?|rem(?:\.un)?|div(?:\.un)?|clt(?:\.un)?|alignment|castclass|ldvirtftn|beq(?:\.s)?|ckfinite|ldsflda|ldtoken|localloc|mkrefany|rethrow|cgt\.un|arglist|switch|stsfld|sizeof|newobj|newarr|ldsfld|ldnull|ldflda|isinst|throw|stobj|stfld|ldstr|ldobj|ldlen|ldftn|ldfld|cpobj|cpblk|break|br\.s|xor|shl|ret|pop|not|nop|neg|jmp|dup|cgt|ceq|box|and|or|br)\b/,boolean:/\b(?:false|true)\b/,number:/\b-?(?:0x[0-9a-f]+|\d+)(?:\.[0-9a-f]+)?\b/i,punctuation:/[{}[\];(),:=]|IL_[0-9A-Za-z]+/}}return Vy}var Wy,gL;function Ebe(){if(gL)return Wy;gL=1,Wy=e,e.displayName="clojure",e.aliases=[];function e(t){t.languages.clojure={comment:{pattern:/;.*/,greedy:!0},string:{pattern:/"(?:[^"\\]|\\.)*"/,greedy:!0},char:/\\\w+/,symbol:{pattern:/(^|[\s()\[\]{},])::?[\w*+!?'<>=/.-]+/,lookbehind:!0},keyword:{pattern:/(\()(?:-|->|->>|\.|\.\.|\*|\/|\+|<|<=|=|==|>|>=|accessor|agent|agent-errors|aget|alength|all-ns|alter|and|append-child|apply|array-map|aset|aset-boolean|aset-byte|aset-char|aset-double|aset-float|aset-int|aset-long|aset-short|assert|assoc|await|await-for|bean|binding|bit-and|bit-not|bit-or|bit-shift-left|bit-shift-right|bit-xor|boolean|branch\?|butlast|byte|cast|char|children|class|clear-agent-errors|comment|commute|comp|comparator|complement|concat|cond|conj|cons|constantly|construct-proxy|contains\?|count|create-ns|create-struct|cycle|dec|declare|def|def-|definline|definterface|defmacro|defmethod|defmulti|defn|defn-|defonce|defproject|defprotocol|defrecord|defstruct|deftype|deref|difference|disj|dissoc|distinct|do|doall|doc|dorun|doseq|dosync|dotimes|doto|double|down|drop|drop-while|edit|end\?|ensure|eval|every\?|false\?|ffirst|file-seq|filter|find|find-doc|find-ns|find-var|first|float|flush|fn|fnseq|for|frest|gensym|get|get-proxy-class|hash-map|hash-set|identical\?|identity|if|if-let|if-not|import|in-ns|inc|index|insert-child|insert-left|insert-right|inspect-table|inspect-tree|instance\?|int|interleave|intersection|into|into-array|iterate|join|key|keys|keyword|keyword\?|last|lazy-cat|lazy-cons|left|lefts|let|line-seq|list|list\*|load|load-file|locking|long|loop|macroexpand|macroexpand-1|make-array|make-node|map|map-invert|map\?|mapcat|max|max-key|memfn|merge|merge-with|meta|min|min-key|monitor-enter|name|namespace|neg\?|new|newline|next|nil\?|node|not|not-any\?|not-every\?|not=|ns|ns-imports|ns-interns|ns-map|ns-name|ns-publics|ns-refers|ns-resolve|ns-unmap|nth|nthrest|or|parse|partial|path|peek|pop|pos\?|pr|pr-str|print|print-str|println|println-str|prn|prn-str|project|proxy|proxy-mappings|quot|quote|rand|rand-int|range|re-find|re-groups|re-matcher|re-matches|re-pattern|re-seq|read|read-line|recur|reduce|ref|ref-set|refer|rem|remove|remove-method|remove-ns|rename|rename-keys|repeat|replace|replicate|resolve|rest|resultset-seq|reverse|rfirst|right|rights|root|rrest|rseq|second|select|select-keys|send|send-off|seq|seq-zip|seq\?|set|set!|short|slurp|some|sort|sort-by|sorted-map|sorted-map-by|sorted-set|special-symbol\?|split-at|split-with|str|string\?|struct|struct-map|subs|subvec|symbol|symbol\?|sync|take|take-nth|take-while|test|throw|time|to-array|to-array-2d|tree-seq|true\?|try|union|up|update-proxy|val|vals|var|var-get|var-set|var\?|vector|vector-zip|vector\?|when|when-first|when-let|when-not|with-local-vars|with-meta|with-open|with-out-str|xml-seq|xml-zip|zero\?|zipmap|zipper)(?=[\s)]|$)/,lookbehind:!0},boolean:/\b(?:false|nil|true)\b/,number:{pattern:/(^|[^\w$@])(?:\d+(?:[/.]\d+)?(?:e[+-]?\d+)?|0x[a-f0-9]+|[1-9]\d?r[a-z0-9]+)[lmn]?(?![\w$@])/i,lookbehind:!0},function:{pattern:/((?:^|[^'])\()[\w*+!?'<>=/.-]+(?=[\s)]|$)/,lookbehind:!0},operator:/[#@^`~]/,punctuation:/[{}\[\](),]/}}return Wy}var Yy,hL;function wbe(){if(hL)return Yy;hL=1,Yy=e,e.displayName="cmake",e.aliases=[];function e(t){t.languages.cmake={comment:/#.*/,string:{pattern:/"(?:[^\\"]|\\.)*"/,greedy:!0,inside:{interpolation:{pattern:/\$\{(?:[^{}$]|\$\{[^{}$]*\})*\}/,inside:{punctuation:/\$\{|\}/,variable:/\w+/}}}},variable:/\b(?:CMAKE_\w+|\w+_(?:(?:BINARY|SOURCE)_DIR|DESCRIPTION|HOMEPAGE_URL|ROOT|VERSION(?:_MAJOR|_MINOR|_PATCH|_TWEAK)?)|(?:ANDROID|APPLE|BORLAND|BUILD_SHARED_LIBS|CACHE|CPACK_(?:ABSOLUTE_DESTINATION_FILES|COMPONENT_INCLUDE_TOPLEVEL_DIRECTORY|ERROR_ON_ABSOLUTE_INSTALL_DESTINATION|INCLUDE_TOPLEVEL_DIRECTORY|INSTALL_DEFAULT_DIRECTORY_PERMISSIONS|INSTALL_SCRIPT|PACKAGING_INSTALL_PREFIX|SET_DESTDIR|WARN_ON_ABSOLUTE_INSTALL_DESTINATION)|CTEST_(?:BINARY_DIRECTORY|BUILD_COMMAND|BUILD_NAME|BZR_COMMAND|BZR_UPDATE_OPTIONS|CHANGE_ID|CHECKOUT_COMMAND|CONFIGURATION_TYPE|CONFIGURE_COMMAND|COVERAGE_COMMAND|COVERAGE_EXTRA_FLAGS|CURL_OPTIONS|CUSTOM_(?:COVERAGE_EXCLUDE|ERROR_EXCEPTION|ERROR_MATCH|ERROR_POST_CONTEXT|ERROR_PRE_CONTEXT|MAXIMUM_FAILED_TEST_OUTPUT_SIZE|MAXIMUM_NUMBER_OF_(?:ERRORS|WARNINGS)|MAXIMUM_PASSED_TEST_OUTPUT_SIZE|MEMCHECK_IGNORE|POST_MEMCHECK|POST_TEST|PRE_MEMCHECK|PRE_TEST|TESTS_IGNORE|WARNING_EXCEPTION|WARNING_MATCH)|CVS_CHECKOUT|CVS_COMMAND|CVS_UPDATE_OPTIONS|DROP_LOCATION|DROP_METHOD|DROP_SITE|DROP_SITE_CDASH|DROP_SITE_PASSWORD|DROP_SITE_USER|EXTRA_COVERAGE_GLOB|GIT_COMMAND|GIT_INIT_SUBMODULES|GIT_UPDATE_CUSTOM|GIT_UPDATE_OPTIONS|HG_COMMAND|HG_UPDATE_OPTIONS|LABELS_FOR_SUBPROJECTS|MEMORYCHECK_(?:COMMAND|COMMAND_OPTIONS|SANITIZER_OPTIONS|SUPPRESSIONS_FILE|TYPE)|NIGHTLY_START_TIME|P4_CLIENT|P4_COMMAND|P4_OPTIONS|P4_UPDATE_OPTIONS|RUN_CURRENT_SCRIPT|SCP_COMMAND|SITE|SOURCE_DIRECTORY|SUBMIT_URL|SVN_COMMAND|SVN_OPTIONS|SVN_UPDATE_OPTIONS|TEST_LOAD|TEST_TIMEOUT|TRIGGER_SITE|UPDATE_COMMAND|UPDATE_OPTIONS|UPDATE_VERSION_ONLY|USE_LAUNCHERS)|CYGWIN|ENV|EXECUTABLE_OUTPUT_PATH|GHS-MULTI|IOS|LIBRARY_OUTPUT_PATH|MINGW|MSVC(?:10|11|12|14|60|70|71|80|90|_IDE|_TOOLSET_VERSION|_VERSION)?|MSYS|PROJECT_(?:BINARY_DIR|DESCRIPTION|HOMEPAGE_URL|NAME|SOURCE_DIR|VERSION|VERSION_(?:MAJOR|MINOR|PATCH|TWEAK))|UNIX|WIN32|WINCE|WINDOWS_PHONE|WINDOWS_STORE|XCODE|XCODE_VERSION))\b/,property:/\b(?:cxx_\w+|(?:ARCHIVE_OUTPUT_(?:DIRECTORY|NAME)|COMPILE_DEFINITIONS|COMPILE_PDB_NAME|COMPILE_PDB_OUTPUT_DIRECTORY|EXCLUDE_FROM_DEFAULT_BUILD|IMPORTED_(?:IMPLIB|LIBNAME|LINK_DEPENDENT_LIBRARIES|LINK_INTERFACE_LANGUAGES|LINK_INTERFACE_LIBRARIES|LINK_INTERFACE_MULTIPLICITY|LOCATION|NO_SONAME|OBJECTS|SONAME)|INTERPROCEDURAL_OPTIMIZATION|LIBRARY_OUTPUT_DIRECTORY|LIBRARY_OUTPUT_NAME|LINK_FLAGS|LINK_INTERFACE_LIBRARIES|LINK_INTERFACE_MULTIPLICITY|LOCATION|MAP_IMPORTED_CONFIG|OSX_ARCHITECTURES|OUTPUT_NAME|PDB_NAME|PDB_OUTPUT_DIRECTORY|RUNTIME_OUTPUT_DIRECTORY|RUNTIME_OUTPUT_NAME|STATIC_LIBRARY_FLAGS|VS_CSHARP|VS_DOTNET_REFERENCEPROP|VS_DOTNET_REFERENCE|VS_GLOBAL_SECTION_POST|VS_GLOBAL_SECTION_PRE|VS_GLOBAL|XCODE_ATTRIBUTE)_\w+|\w+_(?:CLANG_TIDY|COMPILER_LAUNCHER|CPPCHECK|CPPLINT|INCLUDE_WHAT_YOU_USE|OUTPUT_NAME|POSTFIX|VISIBILITY_PRESET)|ABSTRACT|ADDITIONAL_MAKE_CLEAN_FILES|ADVANCED|ALIASED_TARGET|ALLOW_DUPLICATE_CUSTOM_TARGETS|ANDROID_(?:ANT_ADDITIONAL_OPTIONS|API|API_MIN|ARCH|ASSETS_DIRECTORIES|GUI|JAR_DEPENDENCIES|NATIVE_LIB_DEPENDENCIES|NATIVE_LIB_DIRECTORIES|PROCESS_MAX|PROGUARD|PROGUARD_CONFIG_PATH|SECURE_PROPS_PATH|SKIP_ANT_STEP|STL_TYPE)|ARCHIVE_OUTPUT_DIRECTORY|ATTACHED_FILES|ATTACHED_FILES_ON_FAIL|AUTOGEN_(?:BUILD_DIR|ORIGIN_DEPENDS|PARALLEL|SOURCE_GROUP|TARGETS_FOLDER|TARGET_DEPENDS)|AUTOMOC|AUTOMOC_(?:COMPILER_PREDEFINES|DEPEND_FILTERS|EXECUTABLE|MACRO_NAMES|MOC_OPTIONS|SOURCE_GROUP|TARGETS_FOLDER)|AUTORCC|AUTORCC_EXECUTABLE|AUTORCC_OPTIONS|AUTORCC_SOURCE_GROUP|AUTOUIC|AUTOUIC_EXECUTABLE|AUTOUIC_OPTIONS|AUTOUIC_SEARCH_PATHS|BINARY_DIR|BUILDSYSTEM_TARGETS|BUILD_RPATH|BUILD_RPATH_USE_ORIGIN|BUILD_WITH_INSTALL_NAME_DIR|BUILD_WITH_INSTALL_RPATH|BUNDLE|BUNDLE_EXTENSION|CACHE_VARIABLES|CLEAN_NO_CUSTOM|COMMON_LANGUAGE_RUNTIME|COMPATIBLE_INTERFACE_(?:BOOL|NUMBER_MAX|NUMBER_MIN|STRING)|COMPILE_(?:DEFINITIONS|FEATURES|FLAGS|OPTIONS|PDB_NAME|PDB_OUTPUT_DIRECTORY)|COST|CPACK_DESKTOP_SHORTCUTS|CPACK_NEVER_OVERWRITE|CPACK_PERMANENT|CPACK_STARTUP_SHORTCUTS|CPACK_START_MENU_SHORTCUTS|CPACK_WIX_ACL|CROSSCOMPILING_EMULATOR|CUDA_EXTENSIONS|CUDA_PTX_COMPILATION|CUDA_RESOLVE_DEVICE_SYMBOLS|CUDA_SEPARABLE_COMPILATION|CUDA_STANDARD|CUDA_STANDARD_REQUIRED|CXX_EXTENSIONS|CXX_STANDARD|CXX_STANDARD_REQUIRED|C_EXTENSIONS|C_STANDARD|C_STANDARD_REQUIRED|DEBUG_CONFIGURATIONS|DEFINE_SYMBOL|DEFINITIONS|DEPENDS|DEPLOYMENT_ADDITIONAL_FILES|DEPLOYMENT_REMOTE_DIRECTORY|DISABLED|DISABLED_FEATURES|ECLIPSE_EXTRA_CPROJECT_CONTENTS|ECLIPSE_EXTRA_NATURES|ENABLED_FEATURES|ENABLED_LANGUAGES|ENABLE_EXPORTS|ENVIRONMENT|EXCLUDE_FROM_ALL|EXCLUDE_FROM_DEFAULT_BUILD|EXPORT_NAME|EXPORT_PROPERTIES|EXTERNAL_OBJECT|EchoString|FAIL_REGULAR_EXPRESSION|FIND_LIBRARY_USE_LIB32_PATHS|FIND_LIBRARY_USE_LIB64_PATHS|FIND_LIBRARY_USE_LIBX32_PATHS|FIND_LIBRARY_USE_OPENBSD_VERSIONING|FIXTURES_CLEANUP|FIXTURES_REQUIRED|FIXTURES_SETUP|FOLDER|FRAMEWORK|Fortran_FORMAT|Fortran_MODULE_DIRECTORY|GENERATED|GENERATOR_FILE_NAME|GENERATOR_IS_MULTI_CONFIG|GHS_INTEGRITY_APP|GHS_NO_SOURCE_GROUP_FILE|GLOBAL_DEPENDS_DEBUG_MODE|GLOBAL_DEPENDS_NO_CYCLES|GNUtoMS|HAS_CXX|HEADER_FILE_ONLY|HELPSTRING|IMPLICIT_DEPENDS_INCLUDE_TRANSFORM|IMPORTED|IMPORTED_(?:COMMON_LANGUAGE_RUNTIME|CONFIGURATIONS|GLOBAL|IMPLIB|LIBNAME|LINK_DEPENDENT_LIBRARIES|LINK_INTERFACE_(?:LANGUAGES|LIBRARIES|MULTIPLICITY)|LOCATION|NO_SONAME|OBJECTS|SONAME)|IMPORT_PREFIX|IMPORT_SUFFIX|INCLUDE_DIRECTORIES|INCLUDE_REGULAR_EXPRESSION|INSTALL_NAME_DIR|INSTALL_RPATH|INSTALL_RPATH_USE_LINK_PATH|INTERFACE_(?:AUTOUIC_OPTIONS|COMPILE_DEFINITIONS|COMPILE_FEATURES|COMPILE_OPTIONS|INCLUDE_DIRECTORIES|LINK_DEPENDS|LINK_DIRECTORIES|LINK_LIBRARIES|LINK_OPTIONS|POSITION_INDEPENDENT_CODE|SOURCES|SYSTEM_INCLUDE_DIRECTORIES)|INTERPROCEDURAL_OPTIMIZATION|IN_TRY_COMPILE|IOS_INSTALL_COMBINED|JOB_POOLS|JOB_POOL_COMPILE|JOB_POOL_LINK|KEEP_EXTENSION|LABELS|LANGUAGE|LIBRARY_OUTPUT_DIRECTORY|LINKER_LANGUAGE|LINK_(?:DEPENDS|DEPENDS_NO_SHARED|DIRECTORIES|FLAGS|INTERFACE_LIBRARIES|INTERFACE_MULTIPLICITY|LIBRARIES|OPTIONS|SEARCH_END_STATIC|SEARCH_START_STATIC|WHAT_YOU_USE)|LISTFILE_STACK|LOCATION|MACOSX_BUNDLE|MACOSX_BUNDLE_INFO_PLIST|MACOSX_FRAMEWORK_INFO_PLIST|MACOSX_PACKAGE_LOCATION|MACOSX_RPATH|MACROS|MANUALLY_ADDED_DEPENDENCIES|MEASUREMENT|MODIFIED|NAME|NO_SONAME|NO_SYSTEM_FROM_IMPORTED|OBJECT_DEPENDS|OBJECT_OUTPUTS|OSX_ARCHITECTURES|OUTPUT_NAME|PACKAGES_FOUND|PACKAGES_NOT_FOUND|PARENT_DIRECTORY|PASS_REGULAR_EXPRESSION|PDB_NAME|PDB_OUTPUT_DIRECTORY|POSITION_INDEPENDENT_CODE|POST_INSTALL_SCRIPT|PREDEFINED_TARGETS_FOLDER|PREFIX|PRE_INSTALL_SCRIPT|PRIVATE_HEADER|PROCESSORS|PROCESSOR_AFFINITY|PROJECT_LABEL|PUBLIC_HEADER|REPORT_UNDEFINED_PROPERTIES|REQUIRED_FILES|RESOURCE|RESOURCE_LOCK|RULE_LAUNCH_COMPILE|RULE_LAUNCH_CUSTOM|RULE_LAUNCH_LINK|RULE_MESSAGES|RUNTIME_OUTPUT_DIRECTORY|RUN_SERIAL|SKIP_AUTOGEN|SKIP_AUTOMOC|SKIP_AUTORCC|SKIP_AUTOUIC|SKIP_BUILD_RPATH|SKIP_RETURN_CODE|SOURCES|SOURCE_DIR|SOVERSION|STATIC_LIBRARY_FLAGS|STATIC_LIBRARY_OPTIONS|STRINGS|SUBDIRECTORIES|SUFFIX|SYMBOLIC|TARGET_ARCHIVES_MAY_BE_SHARED_LIBS|TARGET_MESSAGES|TARGET_SUPPORTS_SHARED_LIBS|TESTS|TEST_INCLUDE_FILE|TEST_INCLUDE_FILES|TIMEOUT|TIMEOUT_AFTER_MATCH|TYPE|USE_FOLDERS|VALUE|VARIABLES|VERSION|VISIBILITY_INLINES_HIDDEN|VS_(?:CONFIGURATION_TYPE|COPY_TO_OUT_DIR|DEBUGGER_(?:COMMAND|COMMAND_ARGUMENTS|ENVIRONMENT|WORKING_DIRECTORY)|DEPLOYMENT_CONTENT|DEPLOYMENT_LOCATION|DOTNET_REFERENCES|DOTNET_REFERENCES_COPY_LOCAL|GLOBAL_KEYWORD|GLOBAL_PROJECT_TYPES|GLOBAL_ROOTNAMESPACE|INCLUDE_IN_VSIX|IOT_STARTUP_TASK|KEYWORD|RESOURCE_GENERATOR|SCC_AUXPATH|SCC_LOCALPATH|SCC_PROJECTNAME|SCC_PROVIDER|SDK_REFERENCES|SHADER_(?:DISABLE_OPTIMIZATIONS|ENABLE_DEBUG|ENTRYPOINT|FLAGS|MODEL|OBJECT_FILE_NAME|OUTPUT_HEADER_FILE|TYPE|VARIABLE_NAME)|STARTUP_PROJECT|TOOL_OVERRIDE|USER_PROPS|WINRT_COMPONENT|WINRT_EXTENSIONS|WINRT_REFERENCES|XAML_TYPE)|WILL_FAIL|WIN32_EXECUTABLE|WINDOWS_EXPORT_ALL_SYMBOLS|WORKING_DIRECTORY|WRAP_EXCLUDE|XCODE_(?:EMIT_EFFECTIVE_PLATFORM_NAME|EXPLICIT_FILE_TYPE|FILE_ATTRIBUTES|LAST_KNOWN_FILE_TYPE|PRODUCT_TYPE|SCHEME_(?:ADDRESS_SANITIZER|ADDRESS_SANITIZER_USE_AFTER_RETURN|ARGUMENTS|DISABLE_MAIN_THREAD_CHECKER|DYNAMIC_LIBRARY_LOADS|DYNAMIC_LINKER_API_USAGE|ENVIRONMENT|EXECUTABLE|GUARD_MALLOC|MAIN_THREAD_CHECKER_STOP|MALLOC_GUARD_EDGES|MALLOC_SCRIBBLE|MALLOC_STACK|THREAD_SANITIZER(?:_STOP)?|UNDEFINED_BEHAVIOUR_SANITIZER(?:_STOP)?|ZOMBIE_OBJECTS))|XCTEST)\b/,keyword:/\b(?:add_compile_definitions|add_compile_options|add_custom_command|add_custom_target|add_definitions|add_dependencies|add_executable|add_library|add_link_options|add_subdirectory|add_test|aux_source_directory|break|build_command|build_name|cmake_host_system_information|cmake_minimum_required|cmake_parse_arguments|cmake_policy|configure_file|continue|create_test_sourcelist|ctest_build|ctest_configure|ctest_coverage|ctest_empty_binary_directory|ctest_memcheck|ctest_read_custom_files|ctest_run_script|ctest_sleep|ctest_start|ctest_submit|ctest_test|ctest_update|ctest_upload|define_property|else|elseif|enable_language|enable_testing|endforeach|endfunction|endif|endmacro|endwhile|exec_program|execute_process|export|export_library_dependencies|file|find_file|find_library|find_package|find_path|find_program|fltk_wrap_ui|foreach|function|get_cmake_property|get_directory_property|get_filename_component|get_property|get_source_file_property|get_target_property|get_test_property|if|include|include_directories|include_external_msproject|include_guard|include_regular_expression|install|install_files|install_programs|install_targets|link_directories|link_libraries|list|load_cache|load_command|macro|make_directory|mark_as_advanced|math|message|option|output_required_files|project|qt_wrap_cpp|qt_wrap_ui|remove|remove_definitions|return|separate_arguments|set|set_directory_properties|set_property|set_source_files_properties|set_target_properties|set_tests_properties|site_name|source_group|string|subdir_depends|subdirs|target_compile_definitions|target_compile_features|target_compile_options|target_include_directories|target_link_directories|target_link_libraries|target_link_options|target_sources|try_compile|try_run|unset|use_mangled_mesa|utility_source|variable_requires|variable_watch|while|write_file)(?=\s*\()\b/,boolean:/\b(?:FALSE|OFF|ON|TRUE)\b/,namespace:/\b(?:INTERFACE|PRIVATE|PROPERTIES|PUBLIC|SHARED|STATIC|TARGET_OBJECTS)\b/,operator:/\b(?:AND|DEFINED|EQUAL|GREATER|LESS|MATCHES|NOT|OR|STREQUAL|STRGREATER|STRLESS|VERSION_EQUAL|VERSION_GREATER|VERSION_LESS)\b/,inserted:{pattern:/\b\w+::\w+\b/,alias:"class-name"},number:/\b\d+(?:\.\d+)*\b/,function:/\b[a-z_]\w*(?=\s*\()\b/i,punctuation:/[()>}]|\$[<{]/}}return Yy}var Ky,mL;function xbe(){if(mL)return Ky;mL=1,Ky=e,e.displayName="cobol",e.aliases=[];function e(t){t.languages.cobol={comment:{pattern:/\*>.*|(^[ \t]*)\*.*/m,lookbehind:!0,greedy:!0},string:{pattern:/[xzgn]?(?:"(?:[^\r\n"]|"")*"(?!")|'(?:[^\r\n']|'')*'(?!'))/i,greedy:!0},level:{pattern:/(^[ \t]*)\d+\b/m,lookbehind:!0,greedy:!0,alias:"number"},"class-name":{pattern:/(\bpic(?:ture)?\s+)(?:(?:[-\w$/,:*+<>]|\.(?!\s|$))(?:\(\d+\))?)+/i,lookbehind:!0,inside:{number:{pattern:/(\()\d+/,lookbehind:!0},punctuation:/[()]/}},keyword:{pattern:/(^|[^\w-])(?:ABORT|ACCEPT|ACCESS|ADD|ADDRESS|ADVANCING|AFTER|ALIGNED|ALL|ALPHABET|ALPHABETIC|ALPHABETIC-LOWER|ALPHABETIC-UPPER|ALPHANUMERIC|ALPHANUMERIC-EDITED|ALSO|ALTER|ALTERNATE|ANY|ARE|AREA|AREAS|AS|ASCENDING|ASCII|ASSIGN|ASSOCIATED-DATA|ASSOCIATED-DATA-LENGTH|AT|ATTRIBUTE|AUTHOR|AUTO|AUTO-SKIP|BACKGROUND-COLOR|BACKGROUND-COLOUR|BASIS|BEEP|BEFORE|BEGINNING|BELL|BINARY|BIT|BLANK|BLINK|BLOCK|BOTTOM|BOUNDS|BY|BYFUNCTION|BYTITLE|CALL|CANCEL|CAPABLE|CCSVERSION|CD|CF|CH|CHAINING|CHANGED|CHANNEL|CHARACTER|CHARACTERS|CLASS|CLASS-ID|CLOCK-UNITS|CLOSE|CLOSE-DISPOSITION|COBOL|CODE|CODE-SET|COL|COLLATING|COLUMN|COM-REG|COMMA|COMMITMENT|COMMON|COMMUNICATION|COMP|COMP-1|COMP-2|COMP-3|COMP-4|COMP-5|COMPUTATIONAL|COMPUTATIONAL-1|COMPUTATIONAL-2|COMPUTATIONAL-3|COMPUTATIONAL-4|COMPUTATIONAL-5|COMPUTE|CONFIGURATION|CONTAINS|CONTENT|CONTINUE|CONTROL|CONTROL-POINT|CONTROLS|CONVENTION|CONVERTING|COPY|CORR|CORRESPONDING|COUNT|CRUNCH|CURRENCY|CURSOR|DATA|DATA-BASE|DATE|DATE-COMPILED|DATE-WRITTEN|DAY|DAY-OF-WEEK|DBCS|DE|DEBUG-CONTENTS|DEBUG-ITEM|DEBUG-LINE|DEBUG-NAME|DEBUG-SUB-1|DEBUG-SUB-2|DEBUG-SUB-3|DEBUGGING|DECIMAL-POINT|DECLARATIVES|DEFAULT|DEFAULT-DISPLAY|DEFINITION|DELETE|DELIMITED|DELIMITER|DEPENDING|DESCENDING|DESTINATION|DETAIL|DFHRESP|DFHVALUE|DISABLE|DISK|DISPLAY|DISPLAY-1|DIVIDE|DIVISION|DONTCARE|DOUBLE|DOWN|DUPLICATES|DYNAMIC|EBCDIC|EGCS|EGI|ELSE|EMI|EMPTY-CHECK|ENABLE|END|END-ACCEPT|END-ADD|END-CALL|END-COMPUTE|END-DELETE|END-DIVIDE|END-EVALUATE|END-IF|END-MULTIPLY|END-OF-PAGE|END-PERFORM|END-READ|END-RECEIVE|END-RETURN|END-REWRITE|END-SEARCH|END-START|END-STRING|END-SUBTRACT|END-UNSTRING|END-WRITE|ENDING|ENTER|ENTRY|ENTRY-PROCEDURE|ENVIRONMENT|EOL|EOP|EOS|ERASE|ERROR|ESCAPE|ESI|EVALUATE|EVENT|EVERY|EXCEPTION|EXCLUSIVE|EXHIBIT|EXIT|EXPORT|EXTEND|EXTENDED|EXTERNAL|FD|FILE|FILE-CONTROL|FILLER|FINAL|FIRST|FOOTING|FOR|FOREGROUND-COLOR|FOREGROUND-COLOUR|FROM|FULL|FUNCTION|FUNCTION-POINTER|FUNCTIONNAME|GENERATE|GIVING|GLOBAL|GO|GOBACK|GRID|GROUP|HEADING|HIGH-VALUE|HIGH-VALUES|HIGHLIGHT|I-O|I-O-CONTROL|ID|IDENTIFICATION|IF|IMPLICIT|IMPORT|IN|INDEX|INDEXED|INDICATE|INITIAL|INITIALIZE|INITIATE|INPUT|INPUT-OUTPUT|INSPECT|INSTALLATION|INTEGER|INTO|INVALID|INVOKE|IS|JUST|JUSTIFIED|KANJI|KEPT|KEY|KEYBOARD|LABEL|LANGUAGE|LAST|LB|LD|LEADING|LEFT|LEFTLINE|LENGTH|LENGTH-CHECK|LIBACCESS|LIBPARAMETER|LIBRARY|LIMIT|LIMITS|LINAGE|LINAGE-COUNTER|LINE|LINE-COUNTER|LINES|LINKAGE|LIST|LOCAL|LOCAL-STORAGE|LOCK|LONG-DATE|LONG-TIME|LOW-VALUE|LOW-VALUES|LOWER|LOWLIGHT|MEMORY|MERGE|MESSAGE|MMDDYYYY|MODE|MODULES|MORE-LABELS|MOVE|MULTIPLE|MULTIPLY|NAMED|NATIONAL|NATIONAL-EDITED|NATIVE|NEGATIVE|NETWORK|NEXT|NO|NO-ECHO|NULL|NULLS|NUMBER|NUMERIC|NUMERIC-DATE|NUMERIC-EDITED|NUMERIC-TIME|OBJECT-COMPUTER|OCCURS|ODT|OF|OFF|OMITTED|ON|OPEN|OPTIONAL|ORDER|ORDERLY|ORGANIZATION|OTHER|OUTPUT|OVERFLOW|OVERLINE|OWN|PACKED-DECIMAL|PADDING|PAGE|PAGE-COUNTER|PASSWORD|PERFORM|PF|PH|PIC|PICTURE|PLUS|POINTER|PORT|POSITION|POSITIVE|PRINTER|PRINTING|PRIVATE|PROCEDURE|PROCEDURE-POINTER|PROCEDURES|PROCEED|PROCESS|PROGRAM|PROGRAM-ID|PROGRAM-LIBRARY|PROMPT|PURGE|QUEUE|QUOTE|QUOTES|RANDOM|RD|READ|READER|REAL|RECEIVE|RECEIVED|RECORD|RECORDING|RECORDS|RECURSIVE|REDEFINES|REEL|REF|REFERENCE|REFERENCES|RELATIVE|RELEASE|REMAINDER|REMARKS|REMOTE|REMOVAL|REMOVE|RENAMES|REPLACE|REPLACING|REPORT|REPORTING|REPORTS|REQUIRED|RERUN|RESERVE|RESET|RETURN|RETURN-CODE|RETURNING|REVERSE-VIDEO|REVERSED|REWIND|REWRITE|RF|RH|RIGHT|ROUNDED|RUN|SAME|SAVE|SCREEN|SD|SEARCH|SECTION|SECURE|SECURITY|SEGMENT|SEGMENT-LIMIT|SELECT|SEND|SENTENCE|SEPARATE|SEQUENCE|SEQUENTIAL|SET|SHARED|SHAREDBYALL|SHAREDBYRUNUNIT|SHARING|SHIFT-IN|SHIFT-OUT|SHORT-DATE|SIGN|SIZE|SORT|SORT-CONTROL|SORT-CORE-SIZE|SORT-FILE-SIZE|SORT-MERGE|SORT-MESSAGE|SORT-MODE-SIZE|SORT-RETURN|SOURCE|SOURCE-COMPUTER|SPACE|SPACES|SPECIAL-NAMES|STANDARD|STANDARD-1|STANDARD-2|START|STATUS|STOP|STRING|SUB-QUEUE-1|SUB-QUEUE-2|SUB-QUEUE-3|SUBTRACT|SUM|SUPPRESS|SYMBOL|SYMBOLIC|SYNC|SYNCHRONIZED|TABLE|TALLY|TALLYING|TAPE|TASK|TERMINAL|TERMINATE|TEST|TEXT|THEN|THREAD|THREAD-LOCAL|THROUGH|THRU|TIME|TIMER|TIMES|TITLE|TO|TODAYS-DATE|TODAYS-NAME|TOP|TRAILING|TRUNCATED|TYPE|TYPEDEF|UNDERLINE|UNIT|UNSTRING|UNTIL|UP|UPON|USAGE|USE|USING|VALUE|VALUES|VARYING|VIRTUAL|WAIT|WHEN|WHEN-COMPILED|WITH|WORDS|WORKING-STORAGE|WRITE|YEAR|YYYYDDD|YYYYMMDD|ZERO-FILL|ZEROES|ZEROS)(?![\w-])/i,lookbehind:!0},boolean:{pattern:/(^|[^\w-])(?:false|true)(?![\w-])/i,lookbehind:!0},number:{pattern:/(^|[^\w-])(?:[+-]?(?:(?:\d+(?:[.,]\d+)?|[.,]\d+)(?:e[+-]?\d+)?|zero))(?![\w-])/i,lookbehind:!0},operator:[/<>|[<>]=?|[=+*/&]/,{pattern:/(^|[^\w-])(?:-|and|equal|greater|less|not|or|than)(?![\w-])/i,lookbehind:!0}],punctuation:/[.:,()]/}}return Ky}var Xy,bL;function kbe(){if(bL)return Xy;bL=1,Xy=e,e.displayName="coffeescript",e.aliases=["coffee"];function e(t){(function(n){var r=/#(?!\{).+/,a={pattern:/#\{[^}]+\}/,alias:"variable"};n.languages.coffeescript=n.languages.extend("javascript",{comment:r,string:[{pattern:/'(?:\\[\s\S]|[^\\'])*'/,greedy:!0},{pattern:/"(?:\\[\s\S]|[^\\"])*"/,greedy:!0,inside:{interpolation:a}}],keyword:/\b(?:and|break|by|catch|class|continue|debugger|delete|do|each|else|extend|extends|false|finally|for|if|in|instanceof|is|isnt|let|loop|namespace|new|no|not|null|of|off|on|or|own|return|super|switch|then|this|throw|true|try|typeof|undefined|unless|until|when|while|window|with|yes|yield)\b/,"class-member":{pattern:/@(?!\d)\w+/,alias:"variable"}}),n.languages.insertBefore("coffeescript","comment",{"multiline-comment":{pattern:/###[\s\S]+?###/,alias:"comment"},"block-regex":{pattern:/\/{3}[\s\S]*?\/{3}/,alias:"regex",inside:{comment:r,interpolation:a}}}),n.languages.insertBefore("coffeescript","string",{"inline-javascript":{pattern:/`(?:\\[\s\S]|[^\\`])*`/,inside:{delimiter:{pattern:/^`|`$/,alias:"punctuation"},script:{pattern:/[\s\S]+/,alias:"language-javascript",inside:n.languages.javascript}}},"multiline-string":[{pattern:/'''[\s\S]*?'''/,greedy:!0,alias:"string"},{pattern:/"""[\s\S]*?"""/,greedy:!0,alias:"string",inside:{interpolation:a}}]}),n.languages.insertBefore("coffeescript","keyword",{property:/(?!\d)\w+(?=\s*:(?!:))/}),delete n.languages.coffeescript["template-string"],n.languages.coffee=n.languages.coffeescript})(t)}return Xy}var Zy,yL;function Tbe(){if(yL)return Zy;yL=1,Zy=e,e.displayName="concurnas",e.aliases=["conc"];function e(t){t.languages.concurnas={comment:{pattern:/(^|[^\\])(?:\/\*[\s\S]*?(?:\*\/|$)|\/\/.*)/,lookbehind:!0,greedy:!0},langext:{pattern:/\b\w+\s*\|\|[\s\S]+?\|\|/,greedy:!0,inside:{"class-name":/^\w+/,string:{pattern:/(^\s*\|\|)[\s\S]+(?=\|\|$)/,lookbehind:!0},punctuation:/\|\|/}},function:{pattern:/((?:^|\s)def[ \t]+)[a-zA-Z_]\w*(?=\s*\()/,lookbehind:!0},keyword:/\b(?:abstract|actor|also|annotation|assert|async|await|bool|boolean|break|byte|case|catch|changed|char|class|closed|constant|continue|def|default|del|double|elif|else|enum|every|extends|false|finally|float|for|from|global|gpudef|gpukernel|if|import|in|init|inject|int|lambda|local|long|loop|match|new|nodefault|null|of|onchange|open|out|override|package|parfor|parforsync|post|pre|private|protected|provide|provider|public|return|shared|short|single|size_t|sizeof|super|sync|this|throw|trait|trans|transient|true|try|typedef|unchecked|using|val|var|void|while|with)\b/,boolean:/\b(?:false|true)\b/,number:/\b0b[01][01_]*L?\b|\b0x(?:[\da-f_]*\.)?[\da-f_p+-]+\b|(?:\b\d[\d_]*(?:\.[\d_]*)?|\B\.\d[\d_]*)(?:e[+-]?\d[\d_]*)?[dfls]?/i,punctuation:/[{}[\];(),.:]/,operator:/<==|>==|=>|->|<-|<>|&==|&<>|\?:?|\.\?|\+\+|--|[-+*/=<>]=?|[!^~]|\b(?:and|as|band|bor|bxor|comp|is|isnot|mod|or)\b=?/,annotation:{pattern:/@(?:\w+:)?(?:\w+|\[[^\]]+\])?/,alias:"builtin"}},t.languages.insertBefore("concurnas","langext",{"regex-literal":{pattern:/\br("|')(?:\\.|(?!\1)[^\\\r\n])*\1/,greedy:!0,inside:{interpolation:{pattern:/((?:^|[^\\])(?:\\{2})*)\{(?:[^{}]|\{(?:[^{}]|\{[^}]*\})*\})+\}/,lookbehind:!0,inside:t.languages.concurnas},regex:/[\s\S]+/}},"string-literal":{pattern:/(?:\B|\bs)("|')(?:\\.|(?!\1)[^\\\r\n])*\1/,greedy:!0,inside:{interpolation:{pattern:/((?:^|[^\\])(?:\\{2})*)\{(?:[^{}]|\{(?:[^{}]|\{[^}]*\})*\})+\}/,lookbehind:!0,inside:t.languages.concurnas},string:/[\s\S]+/}}}),t.languages.conc=t.languages.concurnas}return Zy}var Qy,vL;function Abe(){if(vL)return Qy;vL=1,Qy=e,e.displayName="coq",e.aliases=[];function e(t){(function(n){for(var r=/\(\*(?:[^(*]|\((?!\*)|\*(?!\))|<self>)*\*\)/.source,a=0;a<2;a++)r=r.replace(/<self>/g,function(){return r});r=r.replace(/<self>/g,"[]"),n.languages.coq={comment:RegExp(r),string:{pattern:/"(?:[^"]|"")*"(?!")/,greedy:!0},attribute:[{pattern:RegExp(/#\[(?:[^\[\]("]|"(?:[^"]|"")*"(?!")|\((?!\*)|<comment>)*\]/.source.replace(/<comment>/g,function(){return r})),greedy:!0,alias:"attr-name",inside:{comment:RegExp(r),string:{pattern:/"(?:[^"]|"")*"(?!")/,greedy:!0},operator:/=/,punctuation:/^#\[|\]$|[,()]/}},{pattern:/\b(?:Cumulative|Global|Local|Monomorphic|NonCumulative|Polymorphic|Private|Program)\b/,alias:"attr-name"}],keyword:/\b(?:Abort|About|Add|Admit|Admitted|All|Arguments|As|Assumptions|Axiom|Axioms|Back|BackTo|Backtrace|BinOp|BinOpSpec|BinRel|Bind|Blacklist|Canonical|Case|Cd|Check|Class|Classes|Close|CoFixpoint|CoInductive|Coercion|Coercions|Collection|Combined|Compute|Conjecture|Conjectures|Constant|Constants|Constraint|Constructors|Context|Corollary|Create|CstOp|Custom|Cut|Debug|Declare|Defined|Definition|Delimit|Dependencies|Dependent|Derive|Diffs|Drop|Elimination|End|Entry|Equality|Eval|Example|Existential|Existentials|Existing|Export|Extern|Extraction|Fact|Fail|Field|File|Firstorder|Fixpoint|Flags|Focus|From|Funclass|Function|Functional|GC|Generalizable|Goal|Grab|Grammar|Graph|Guarded|Haskell|Heap|Hide|Hint|HintDb|Hints|Hypotheses|Hypothesis|IF|Identity|Immediate|Implicit|Implicits|Import|Include|Induction|Inductive|Infix|Info|Initial|InjTyp|Inline|Inspect|Instance|Instances|Intro|Intros|Inversion|Inversion_clear|JSON|Language|Left|Lemma|Let|Lia|Libraries|Library|Load|LoadPath|Locate|Ltac|Ltac2|ML|Match|Method|Minimality|Module|Modules|Morphism|Next|NoInline|Notation|Number|OCaml|Obligation|Obligations|Opaque|Open|Optimize|Parameter|Parameters|Parametric|Path|Paths|Prenex|Preterm|Primitive|Print|Profile|Projections|Proof|Prop|PropBinOp|PropOp|PropUOp|Property|Proposition|Pwd|Qed|Quit|Rec|Record|Recursive|Redirect|Reduction|Register|Relation|Remark|Remove|Require|Reserved|Reset|Resolve|Restart|Rewrite|Right|Ring|Rings|SProp|Saturate|Save|Scheme|Scope|Scopes|Search|SearchHead|SearchPattern|SearchRewrite|Section|Separate|Set|Setoid|Show|Signatures|Solve|Solver|Sort|Sortclass|Sorted|Spec|Step|Strategies|Strategy|String|Structure|SubClass|Subgraph|SuchThat|Tactic|Term|TestCompile|Theorem|Time|Timeout|To|Transparent|Type|Typeclasses|Types|Typing|UnOp|UnOpSpec|Undelimit|Undo|Unfocus|Unfocused|Unfold|Universe|Universes|Unshelve|Variable|Variables|Variant|Verbose|View|Visibility|Zify|_|apply|as|at|by|cofix|else|end|exists|exists2|fix|for|forall|fun|if|in|let|match|measure|move|removed|return|struct|then|using|wf|where|with)\b/,number:/\b(?:0x[a-f0-9][a-f0-9_]*(?:\.[a-f0-9_]+)?(?:p[+-]?\d[\d_]*)?|\d[\d_]*(?:\.[\d_]+)?(?:e[+-]?\d[\d_]*)?)\b/i,punct:{pattern:/@\{|\{\||\[=|:>/,alias:"punctuation"},operator:/\/\\|\\\/|\.{2,3}|:{1,2}=|\*\*|[-=]>|<(?:->?|[+:=>]|<:)|>(?:=|->)|\|[-|]?|[-!%&*+/<=>?@^~']/,punctuation:/\.\(|`\(|@\{|`\{|\{\||\[=|:>|[:.,;(){}\[\]]/}})(t)}return Qy}var Jy,SL;function Fp(){if(SL)return Jy;SL=1,Jy=e,e.displayName="ruby",e.aliases=["rb"];function e(t){(function(n){n.languages.ruby=n.languages.extend("clike",{comment:{pattern:/#.*|^=begin\s[\s\S]*?^=end/m,greedy:!0},"class-name":{pattern:/(\b(?:class|module)\s+|\bcatch\s+\()[\w.\\]+|\b[A-Z_]\w*(?=\s*\.\s*new\b)/,lookbehind:!0,inside:{punctuation:/[.\\]/}},keyword:/\b(?:BEGIN|END|alias|and|begin|break|case|class|def|define_method|defined|do|each|else|elsif|end|ensure|extend|for|if|in|include|module|new|next|nil|not|or|prepend|private|protected|public|raise|redo|require|rescue|retry|return|self|super|then|throw|undef|unless|until|when|while|yield)\b/,operator:/\.{2,3}|&\.|===|<?=>|[!=]?~|(?:&&|\|\||<<|>>|\*\*|[+\-*/%<>!^&|=])=?|[?:]/,punctuation:/[(){}[\].,;]/}),n.languages.insertBefore("ruby","operator",{"double-colon":{pattern:/::/,alias:"punctuation"}});var r={pattern:/((?:^|[^\\])(?:\\{2})*)#\{(?:[^{}]|\{[^{}]*\})*\}/,lookbehind:!0,inside:{content:{pattern:/^(#\{)[\s\S]+(?=\}$)/,lookbehind:!0,inside:n.languages.ruby},delimiter:{pattern:/^#\{|\}$/,alias:"punctuation"}}};delete n.languages.ruby.function;var a="(?:"+[/([^a-zA-Z0-9\s{(\[<=])(?:(?!\1)[^\\]|\\[\s\S])*\1/.source,/\((?:[^()\\]|\\[\s\S]|\((?:[^()\\]|\\[\s\S])*\))*\)/.source,/\{(?:[^{}\\]|\\[\s\S]|\{(?:[^{}\\]|\\[\s\S])*\})*\}/.source,/\[(?:[^\[\]\\]|\\[\s\S]|\[(?:[^\[\]\\]|\\[\s\S])*\])*\]/.source,/<(?:[^<>\\]|\\[\s\S]|<(?:[^<>\\]|\\[\s\S])*>)*>/.source].join("|")+")",o=/(?:"(?:\\.|[^"\\\r\n])*"|(?:\b[a-zA-Z_]\w*|[^\s\0-\x7F]+)[?!]?|\$.)/.source;n.languages.insertBefore("ruby","keyword",{"regex-literal":[{pattern:RegExp(/%r/.source+a+/[egimnosux]{0,6}/.source),greedy:!0,inside:{interpolation:r,regex:/[\s\S]+/}},{pattern:/(^|[^/])\/(?!\/)(?:\[[^\r\n\]]+\]|\\.|[^[/\\\r\n])+\/[egimnosux]{0,6}(?=\s*(?:$|[\r\n,.;})#]))/,lookbehind:!0,greedy:!0,inside:{interpolation:r,regex:/[\s\S]+/}}],variable:/[@$]+[a-zA-Z_]\w*(?:[?!]|\b)/,symbol:[{pattern:RegExp(/(^|[^:]):/.source+o),lookbehind:!0,greedy:!0},{pattern:RegExp(/([\r\n{(,][ \t]*)/.source+o+/(?=:(?!:))/.source),lookbehind:!0,greedy:!0}],"method-definition":{pattern:/(\bdef\s+)\w+(?:\s*\.\s*\w+)?/,lookbehind:!0,inside:{function:/\b\w+$/,keyword:/^self\b/,"class-name":/^\w+/,punctuation:/\./}}}),n.languages.insertBefore("ruby","string",{"string-literal":[{pattern:RegExp(/%[qQiIwWs]?/.source+a),greedy:!0,inside:{interpolation:r,string:/[\s\S]+/}},{pattern:/("|')(?:#\{[^}]+\}|#(?!\{)|\\(?:\r\n|[\s\S])|(?!\1)[^\\#\r\n])*\1/,greedy:!0,inside:{interpolation:r,string:/[\s\S]+/}},{pattern:/<<[-~]?([a-z_]\w*)[\r\n](?:.*[\r\n])*?[\t ]*\1/i,alias:"heredoc-string",greedy:!0,inside:{delimiter:{pattern:/^<<[-~]?[a-z_]\w*|\b[a-z_]\w*$/i,inside:{symbol:/\b\w+/,punctuation:/^<<[-~]?/}},interpolation:r,string:/[\s\S]+/}},{pattern:/<<[-~]?'([a-z_]\w*)'[\r\n](?:.*[\r\n])*?[\t ]*\1/i,alias:"heredoc-string",greedy:!0,inside:{delimiter:{pattern:/^<<[-~]?'[a-z_]\w*'|\b[a-z_]\w*$/i,inside:{symbol:/\b\w+/,punctuation:/^<<[-~]?'|'$/}},string:/[\s\S]+/}}],"command-literal":[{pattern:RegExp(/%x/.source+a),greedy:!0,inside:{interpolation:r,command:{pattern:/[\s\S]+/,alias:"string"}}},{pattern:/`(?:#\{[^}]+\}|#(?!\{)|\\(?:\r\n|[\s\S])|[^\\`#\r\n])*`/,greedy:!0,inside:{interpolation:r,command:{pattern:/[\s\S]+/,alias:"string"}}}]}),delete n.languages.ruby.string,n.languages.insertBefore("ruby","number",{builtin:/\b(?:Array|Bignum|Binding|Class|Continuation|Dir|Exception|FalseClass|File|Fixnum|Float|Hash|IO|Integer|MatchData|Method|Module|NilClass|Numeric|Object|Proc|Range|Regexp|Stat|String|Struct|Symbol|TMS|Thread|ThreadGroup|Time|TrueClass)\b/,constant:/\b[A-Z][A-Z0-9_]*(?:[?!]|\b)/}),n.languages.rb=n.languages.ruby})(t)}return Jy}var ev,EL;function Rbe(){if(EL)return ev;EL=1;var e=Fp();ev=t,t.displayName="crystal",t.aliases=[];function t(n){n.register(e),function(r){r.languages.crystal=r.languages.extend("ruby",{keyword:[/\b(?:__DIR__|__END_LINE__|__FILE__|__LINE__|abstract|alias|annotation|as|asm|begin|break|case|class|def|do|else|elsif|end|ensure|enum|extend|for|fun|if|ifdef|include|instance_sizeof|lib|macro|module|next|of|out|pointerof|private|protected|ptr|require|rescue|return|select|self|sizeof|struct|super|then|type|typeof|undef|uninitialized|union|unless|until|when|while|with|yield)\b/,{pattern:/(\.\s*)(?:is_a|responds_to)\?/,lookbehind:!0}],number:/\b(?:0b[01_]*[01]|0o[0-7_]*[0-7]|0x[\da-fA-F_]*[\da-fA-F]|(?:\d(?:[\d_]*\d)?)(?:\.[\d_]*\d)?(?:[eE][+-]?[\d_]*\d)?)(?:_(?:[uif](?:8|16|32|64))?)?\b/,operator:[/->/,r.languages.ruby.operator],punctuation:/[(){}[\].,;\\]/}),r.languages.insertBefore("crystal","string-literal",{attribute:{pattern:/@\[.*?\]/,inside:{delimiter:{pattern:/^@\[|\]$/,alias:"punctuation"},attribute:{pattern:/^(\s*)\w+/,lookbehind:!0,alias:"class-name"},args:{pattern:/\S(?:[\s\S]*\S)?/,inside:r.languages.crystal}}},expansion:{pattern:/\{(?:\{.*?\}|%.*?%)\}/,inside:{content:{pattern:/^(\{.)[\s\S]+(?=.\}$)/,lookbehind:!0,inside:r.languages.crystal},delimiter:{pattern:/^\{[\{%]|[\}%]\}$/,alias:"operator"}}},char:{pattern:/'(?:[^\\\r\n]{1,2}|\\(?:.|u(?:[A-Fa-f0-9]{1,4}|\{[A-Fa-f0-9]{1,6}\})))'/,greedy:!0}})}(n)}return ev}var tv,wL;function _be(){if(wL)return tv;wL=1;var e=Mp();tv=t,t.displayName="cshtml",t.aliases=["razor"];function t(n){n.register(e),function(r){var a=/\/(?![/*])|\/\/.*[\r\n]|\/\*[^*]*(?:\*(?!\/)[^*]*)*\*\//.source,o=/@(?!")|"(?:[^\r\n\\"]|\\.)*"|@"(?:[^\\"]|""|\\[\s\S])*"(?!")/.source+"|"+/'(?:(?:[^\r\n'\\]|\\.|\\[Uux][\da-fA-F]{1,8})'|(?=[^\\](?!')))/.source;function s(x,R){for(var k=0;k<R;k++)x=x.replace(/<self>/g,function(){return"(?:"+x+")"});return x.replace(/<self>/g,"[^\\s\\S]").replace(/<str>/g,"(?:"+o+")").replace(/<comment>/g,"(?:"+a+")")}var u=s(/\((?:[^()'"@/]|<str>|<comment>|<self>)*\)/.source,2),c=s(/\[(?:[^\[\]'"@/]|<str>|<comment>|<self>)*\]/.source,2),d=s(/\{(?:[^{}'"@/]|<str>|<comment>|<self>)*\}/.source,2),p=s(/<(?:[^<>'"@/]|<str>|<comment>|<self>)*>/.source,2),g=/(?:\s(?:\s*[^\s>\/=]+(?:\s*=\s*(?:"[^"]*"|'[^']*'|[^\s'">=]+(?=[\s>]))|(?=[\s/>])))+)?/.source,m=/(?!\d)[^\s>\/=$<%]+/.source+g+/\s*\/?>/.source,b=/\B@?/.source+"(?:"+/<([a-zA-Z][\w:]*)/.source+g+/\s*>/.source+"(?:"+(/[^<]/.source+"|"+/<\/?(?!\1\b)/.source+m+"|"+s(/<\1/.source+g+/\s*>/.source+"(?:"+(/[^<]/.source+"|"+/<\/?(?!\1\b)/.source+m+"|<self>")+")*"+/<\/\1\s*>/.source,2))+")*"+/<\/\1\s*>/.source+"|"+/</.source+m+")";r.languages.cshtml=r.languages.extend("markup",{});var y=r.languages.insertBefore("csharp","string",{html:{pattern:RegExp(b),greedy:!0,inside:r.languages.cshtml}},{csharp:r.languages.extend("csharp",{})}),S={pattern:/\S[\s\S]*/,alias:"language-csharp",inside:y};r.languages.insertBefore("cshtml","prolog",{"razor-comment":{pattern:/@\*[\s\S]*?\*@/,greedy:!0,alias:"comment"},block:{pattern:RegExp(/(^|[^@])@/.source+"(?:"+[d,/(?:code|functions)\s*/.source+d,/(?:for|foreach|lock|switch|using|while)\s*/.source+u+/\s*/.source+d,/do\s*/.source+d+/\s*while\s*/.source+u+/(?:\s*;)?/.source,/try\s*/.source+d+/\s*catch\s*/.source+u+/\s*/.source+d+/\s*finally\s*/.source+d,/if\s*/.source+u+/\s*/.source+d+"(?:"+/\s*else/.source+"(?:"+/\s+if\s*/.source+u+")?"+/\s*/.source+d+")*"].join("|")+")"),lookbehind:!0,greedy:!0,inside:{keyword:/^@\w*/,csharp:S}},directive:{pattern:/^([ \t]*)@(?:addTagHelper|attribute|implements|inherits|inject|layout|model|namespace|page|preservewhitespace|removeTagHelper|section|tagHelperPrefix|using)(?=\s).*/m,lookbehind:!0,greedy:!0,inside:{keyword:/^@\w+/,csharp:S}},value:{pattern:RegExp(/(^|[^@])@/.source+/(?:await\b\s*)?/.source+"(?:"+/\w+\b/.source+"|"+u+")(?:"+/[?!]?\.\w+\b/.source+"|"+u+"|"+c+"|"+p+u+")*"),lookbehind:!0,greedy:!0,alias:"variable",inside:{keyword:/^@/,csharp:S}},"delegate-operator":{pattern:/(^|[^@])@(?=<)/,lookbehind:!0,alias:"operator"}}),r.languages.razor=r.languages.cshtml}(n)}return tv}var nv,xL;function Cbe(){if(xL)return nv;xL=1,nv=e,e.displayName="csp",e.aliases=[];function e(t){(function(n){function r(a){return RegExp(/([ \t])/.source+"(?:"+a+")"+/(?=[\s;]|$)/.source,"i")}n.languages.csp={directive:{pattern:/(^|[\s;])(?:base-uri|block-all-mixed-content|(?:child|connect|default|font|frame|img|manifest|media|object|prefetch|script|style|worker)-src|disown-opener|form-action|frame-(?:ancestors|options)|input-protection(?:-(?:clip|selectors))?|navigate-to|plugin-types|policy-uri|referrer|reflected-xss|report-(?:to|uri)|require-sri-for|sandbox|(?:script|style)-src-(?:attr|elem)|upgrade-insecure-requests)(?=[\s;]|$)/i,lookbehind:!0,alias:"property"},scheme:{pattern:r(/[a-z][a-z0-9.+-]*:/.source),lookbehind:!0},none:{pattern:r(/'none'/.source),lookbehind:!0,alias:"keyword"},nonce:{pattern:r(/'nonce-[-+/\w=]+'/.source),lookbehind:!0,alias:"number"},hash:{pattern:r(/'sha(?:256|384|512)-[-+/\w=]+'/.source),lookbehind:!0,alias:"number"},host:{pattern:r(/[a-z][a-z0-9.+-]*:\/\/[^\s;,']*/.source+"|"+/\*[^\s;,']*/.source+"|"+/[a-z0-9-]+(?:\.[a-z0-9-]+)+(?::[\d*]+)?(?:\/[^\s;,']*)?/.source),lookbehind:!0,alias:"url",inside:{important:/\*/}},keyword:[{pattern:r(/'unsafe-[a-z-]+'/.source),lookbehind:!0,alias:"unsafe"},{pattern:r(/'[a-z-]+'/.source),lookbehind:!0,alias:"safe"}],punctuation:/;/}})(t)}return nv}var rv,kL;function Nbe(){if(kL)return rv;kL=1,rv=e,e.displayName="cssExtras",e.aliases=[];function e(t){(function(n){var r=/("|')(?:\\(?:\r\n|[\s\S])|(?!\1)[^\\\r\n])*\1/,a;n.languages.css.selector={pattern:n.languages.css.selector.pattern,lookbehind:!0,inside:a={"pseudo-element":/:(?:after|before|first-letter|first-line|selection)|::[-\w]+/,"pseudo-class":/:[-\w]+/,class:/\.[-\w]+/,id:/#[-\w]+/,attribute:{pattern:RegExp(`\\[(?:[^[\\]"']|`+r.source+")*\\]"),greedy:!0,inside:{punctuation:/^\[|\]$/,"case-sensitivity":{pattern:/(\s)[si]$/i,lookbehind:!0,alias:"keyword"},namespace:{pattern:/^(\s*)(?:(?!\s)[-*\w\xA0-\uFFFF])*\|(?!=)/,lookbehind:!0,inside:{punctuation:/\|$/}},"attr-name":{pattern:/^(\s*)(?:(?!\s)[-\w\xA0-\uFFFF])+/,lookbehind:!0},"attr-value":[r,{pattern:/(=\s*)(?:(?!\s)[-\w\xA0-\uFFFF])+(?=\s*$)/,lookbehind:!0}],operator:/[|~*^$]?=/}},"n-th":[{pattern:/(\(\s*)[+-]?\d*[\dn](?:\s*[+-]\s*\d+)?(?=\s*\))/,lookbehind:!0,inside:{number:/[\dn]+/,operator:/[+-]/}},{pattern:/(\(\s*)(?:even|odd)(?=\s*\))/i,lookbehind:!0}],combinator:/>|\+|~|\|\|/,punctuation:/[(),]/}},n.languages.css.atrule.inside["selector-function-argument"].inside=a,n.languages.insertBefore("css","property",{variable:{pattern:/(^|[^-\w\xA0-\uFFFF])--(?!\s)[-_a-z\xA0-\uFFFF](?:(?!\s)[-\w\xA0-\uFFFF])*/i,lookbehind:!0}});var o={pattern:/(\b\d+)(?:%|[a-z]+(?![\w-]))/,lookbehind:!0},s={pattern:/(^|[^\w.-])-?(?:\d+(?:\.\d+)?|\.\d+)/,lookbehind:!0};n.languages.insertBefore("css","function",{operator:{pattern:/(\s)[+\-*\/](?=\s)/,lookbehind:!0},hexcode:{pattern:/\B#[\da-f]{3,8}\b/i,alias:"color"},color:[{pattern:/(^|[^\w-])(?:AliceBlue|AntiqueWhite|Aqua|Aquamarine|Azure|Beige|Bisque|Black|BlanchedAlmond|Blue|BlueViolet|Brown|BurlyWood|CadetBlue|Chartreuse|Chocolate|Coral|CornflowerBlue|Cornsilk|Crimson|Cyan|DarkBlue|DarkCyan|DarkGoldenRod|DarkGr[ae]y|DarkGreen|DarkKhaki|DarkMagenta|DarkOliveGreen|DarkOrange|DarkOrchid|DarkRed|DarkSalmon|DarkSeaGreen|DarkSlateBlue|DarkSlateGr[ae]y|DarkTurquoise|DarkViolet|DeepPink|DeepSkyBlue|DimGr[ae]y|DodgerBlue|FireBrick|FloralWhite|ForestGreen|Fuchsia|Gainsboro|GhostWhite|Gold|GoldenRod|Gr[ae]y|Green|GreenYellow|HoneyDew|HotPink|IndianRed|Indigo|Ivory|Khaki|Lavender|LavenderBlush|LawnGreen|LemonChiffon|LightBlue|LightCoral|LightCyan|LightGoldenRodYellow|LightGr[ae]y|LightGreen|LightPink|LightSalmon|LightSeaGreen|LightSkyBlue|LightSlateGr[ae]y|LightSteelBlue|LightYellow|Lime|LimeGreen|Linen|Magenta|Maroon|MediumAquaMarine|MediumBlue|MediumOrchid|MediumPurple|MediumSeaGreen|MediumSlateBlue|MediumSpringGreen|MediumTurquoise|MediumVioletRed|MidnightBlue|MintCream|MistyRose|Moccasin|NavajoWhite|Navy|OldLace|Olive|OliveDrab|Orange|OrangeRed|Orchid|PaleGoldenRod|PaleGreen|PaleTurquoise|PaleVioletRed|PapayaWhip|PeachPuff|Peru|Pink|Plum|PowderBlue|Purple|Red|RosyBrown|RoyalBlue|SaddleBrown|Salmon|SandyBrown|SeaGreen|SeaShell|Sienna|Silver|SkyBlue|SlateBlue|SlateGr[ae]y|Snow|SpringGreen|SteelBlue|Tan|Teal|Thistle|Tomato|Transparent|Turquoise|Violet|Wheat|White|WhiteSmoke|Yellow|YellowGreen)(?![\w-])/i,lookbehind:!0},{pattern:/\b(?:hsl|rgb)\(\s*\d{1,3}\s*,\s*\d{1,3}%?\s*,\s*\d{1,3}%?\s*\)\B|\b(?:hsl|rgb)a\(\s*\d{1,3}\s*,\s*\d{1,3}%?\s*,\s*\d{1,3}%?\s*,\s*(?:0|0?\.\d+|1)\s*\)\B/i,inside:{unit:o,number:s,function:/[\w-]+(?=\()/,punctuation:/[(),]/}}],entity:/\\[\da-f]{1,8}/i,unit:o,number:s})})(t)}return rv}var av,TL;function Obe(){if(TL)return av;TL=1,av=e,e.displayName="csv",e.aliases=[];function e(t){t.languages.csv={value:/[^\r\n,"]+|"(?:[^"]|"")*"(?!")/,punctuation:/,/}}return av}var ov,AL;function Dbe(){if(AL)return ov;AL=1,ov=e,e.displayName="cypher",e.aliases=[];function e(t){t.languages.cypher={comment:/\/\/.*/,string:{pattern:/"(?:[^"\\\r\n]|\\.)*"|'(?:[^'\\\r\n]|\\.)*'/,greedy:!0},"class-name":{pattern:/(:\s*)(?:\w+|`(?:[^`\\\r\n])*`)(?=\s*[{):])/,lookbehind:!0,greedy:!0},relationship:{pattern:/(-\[\s*(?:\w+\s*|`(?:[^`\\\r\n])*`\s*)?:\s*|\|\s*:\s*)(?:\w+|`(?:[^`\\\r\n])*`)/,lookbehind:!0,greedy:!0,alias:"property"},identifier:{pattern:/`(?:[^`\\\r\n])*`/,greedy:!0},variable:/\$\w+/,keyword:/\b(?:ADD|ALL|AND|AS|ASC|ASCENDING|ASSERT|BY|CALL|CASE|COMMIT|CONSTRAINT|CONTAINS|CREATE|CSV|DELETE|DESC|DESCENDING|DETACH|DISTINCT|DO|DROP|ELSE|END|ENDS|EXISTS|FOR|FOREACH|IN|INDEX|IS|JOIN|KEY|LIMIT|LOAD|MANDATORY|MATCH|MERGE|NODE|NOT|OF|ON|OPTIONAL|OR|ORDER(?=\s+BY)|PERIODIC|REMOVE|REQUIRE|RETURN|SCALAR|SCAN|SET|SKIP|START|STARTS|THEN|UNION|UNIQUE|UNWIND|USING|WHEN|WHERE|WITH|XOR|YIELD)\b/i,function:/\b\w+\b(?=\s*\()/,boolean:/\b(?:false|null|true)\b/i,number:/\b(?:0x[\da-fA-F]+|\d+(?:\.\d+)?(?:[eE][+-]?\d+)?)\b/,operator:/:|<--?|--?>?|<>|=~?|[<>]=?|[+*/%^|]|\.\.\.?/,punctuation:/[()[\]{},;.]/}}return ov}var iv,RL;function Ibe(){if(RL)return iv;RL=1,iv=e,e.displayName="d",e.aliases=[];function e(t){t.languages.d=t.languages.extend("clike",{comment:[{pattern:/^\s*#!.+/,greedy:!0},{pattern:RegExp(/(^|[^\\])/.source+"(?:"+[/\/\+(?:\/\+(?:[^+]|\+(?!\/))*\+\/|(?!\/\+)[\s\S])*?\+\//.source,/\/\/.*/.source,/\/\*[\s\S]*?\*\//.source].join("|")+")"),lookbehind:!0,greedy:!0}],string:[{pattern:RegExp([/\b[rx]"(?:\\[\s\S]|[^\\"])*"[cwd]?/.source,/\bq"(?:\[[\s\S]*?\]|\([\s\S]*?\)|<[\s\S]*?>|\{[\s\S]*?\})"/.source,/\bq"((?!\d)\w+)$[\s\S]*?^\1"/.source,/\bq"(.)[\s\S]*?\2"/.source,/(["`])(?:\\[\s\S]|(?!\3)[^\\])*\3[cwd]?/.source].join("|"),"m"),greedy:!0},{pattern:/\bq\{(?:\{[^{}]*\}|[^{}])*\}/,greedy:!0,alias:"token-string"}],keyword:/\$|\b(?:__(?:(?:DATE|EOF|FILE|FUNCTION|LINE|MODULE|PRETTY_FUNCTION|TIMESTAMP|TIME|VENDOR|VERSION)__|gshared|parameters|traits|vector)|abstract|alias|align|asm|assert|auto|body|bool|break|byte|case|cast|catch|cdouble|cent|cfloat|char|class|const|continue|creal|dchar|debug|default|delegate|delete|deprecated|do|double|dstring|else|enum|export|extern|false|final|finally|float|for|foreach|foreach_reverse|function|goto|idouble|if|ifloat|immutable|import|inout|int|interface|invariant|ireal|lazy|long|macro|mixin|module|new|nothrow|null|out|override|package|pragma|private|protected|ptrdiff_t|public|pure|real|ref|return|scope|shared|short|size_t|static|string|struct|super|switch|synchronized|template|this|throw|true|try|typedef|typeid|typeof|ubyte|ucent|uint|ulong|union|unittest|ushort|version|void|volatile|wchar|while|with|wstring)\b/,number:[/\b0x\.?[a-f\d_]+(?:(?!\.\.)\.[a-f\d_]*)?(?:p[+-]?[a-f\d_]+)?[ulfi]{0,4}/i,{pattern:/((?:\.\.)?)(?:\b0b\.?|\b|\.)\d[\d_]*(?:(?!\.\.)\.[\d_]*)?(?:e[+-]?\d[\d_]*)?[ulfi]{0,4}/i,lookbehind:!0}],operator:/\|[|=]?|&[&=]?|\+[+=]?|-[-=]?|\.?\.\.|=[>=]?|!(?:i[ns]\b|<>?=?|>=?|=)?|\bi[ns]\b|(?:<[<>]?|>>?>?|\^\^|[*\/%^~])=?/}),t.languages.insertBefore("d","string",{char:/'(?:\\(?:\W|\w+)|[^\\])'/}),t.languages.insertBefore("d","keyword",{property:/\B@\w*/}),t.languages.insertBefore("d","function",{register:{pattern:/\b(?:[ABCD][LHX]|E?(?:BP|DI|SI|SP)|[BS]PL|[ECSDGF]S|CR[0234]|[DS]IL|DR[012367]|E[ABCD]X|X?MM[0-7]|R(?:1[0-5]|[89])[BWD]?|R[ABCD]X|R[BS]P|R[DS]I|TR[3-7]|XMM(?:1[0-5]|[89])|YMM(?:1[0-5]|\d))\b|\bST(?:\([0-7]\)|\b)/,alias:"variable"}})}return iv}var sv,_L;function Lbe(){if(_L)return sv;_L=1,sv=e,e.displayName="dart",e.aliases=[];function e(t){(function(n){var r=[/\b(?:async|sync|yield)\*/,/\b(?:abstract|assert|async|await|break|case|catch|class|const|continue|covariant|default|deferred|do|dynamic|else|enum|export|extends|extension|external|factory|final|finally|for|get|hide|if|implements|import|in|interface|library|mixin|new|null|on|operator|part|rethrow|return|set|show|static|super|switch|sync|this|throw|try|typedef|var|void|while|with|yield)\b/],a=/(^|[^\w.])(?:[a-z]\w*\s*\.\s*)*(?:[A-Z]\w*\s*\.\s*)*/.source,o={pattern:RegExp(a+/[A-Z](?:[\d_A-Z]*[a-z]\w*)?\b/.source),lookbehind:!0,inside:{namespace:{pattern:/^[a-z]\w*(?:\s*\.\s*[a-z]\w*)*(?:\s*\.)?/,inside:{punctuation:/\./}}}};n.languages.dart=n.languages.extend("clike",{"class-name":[o,{pattern:RegExp(a+/[A-Z]\w*(?=\s+\w+\s*[;,=()])/.source),lookbehind:!0,inside:o.inside}],keyword:r,operator:/\bis!|\b(?:as|is)\b|\+\+|--|&&|\|\||<<=?|>>=?|~(?:\/=?)?|[+\-*\/%&^|=!<>]=?|\?/}),n.languages.insertBefore("dart","string",{"string-literal":{pattern:/r?(?:("""|''')[\s\S]*?\1|(["'])(?:\\.|(?!\2)[^\\\r\n])*\2(?!\2))/,greedy:!0,inside:{interpolation:{pattern:/((?:^|[^\\])(?:\\{2})*)\$(?:\w+|\{(?:[^{}]|\{[^{}]*\})*\})/,lookbehind:!0,inside:{punctuation:/^\$\{?|\}$/,expression:{pattern:/[\s\S]+/,inside:n.languages.dart}}},string:/[\s\S]+/}},string:void 0}),n.languages.insertBefore("dart","class-name",{metadata:{pattern:/@\w+/,alias:"function"}}),n.languages.insertBefore("dart","class-name",{generics:{pattern:/<(?:[\w\s,.&?]|<(?:[\w\s,.&?]|<(?:[\w\s,.&?]|<[\w\s,.&?]*>)*>)*>)*>/,inside:{"class-name":o,keyword:r,punctuation:/[<>(),.:]/,operator:/[?&|]/}}})})(t)}return sv}var lv,CL;function Mbe(){if(CL)return lv;CL=1,lv=e,e.displayName="dataweave",e.aliases=[];function e(t){(function(n){n.languages.dataweave={url:/\b[A-Za-z]+:\/\/[\w/:.?=&-]+|\burn:[\w:.?=&-]+/,property:{pattern:/(?:\b\w+#)?(?:"(?:\\.|[^\\"\r\n])*"|\b\w+)(?=\s*[:@])/,greedy:!0},string:{pattern:/(["'`])(?:\\[\s\S]|(?!\1)[^\\])*\1/,greedy:!0},"mime-type":/\b(?:application|audio|image|multipart|text|video)\/[\w+-]+/,date:{pattern:/\|[\w:+-]+\|/,greedy:!0},comment:[{pattern:/(^|[^\\])\/\*[\s\S]*?(?:\*\/|$)/,lookbehind:!0,greedy:!0},{pattern:/(^|[^\\:])\/\/.*/,lookbehind:!0,greedy:!0}],regex:{pattern:/\/(?:[^\\\/\r\n]|\\[^\r\n])+\//,greedy:!0},keyword:/\b(?:and|as|at|case|do|else|fun|if|input|is|match|not|ns|null|or|output|type|unless|update|using|var)\b/,function:/\b[A-Z_]\w*(?=\s*\()/i,number:/-?\b\d+(?:\.\d+)?(?:e[+-]?\d+)?\b/i,punctuation:/[{}[\];(),.:@]/,operator:/<<|>>|->|[<>~=]=?|!=|--?-?|\+\+?|!|\?/,boolean:/\b(?:false|true)\b/}})(t)}return lv}var uv,NL;function Fbe(){if(NL)return uv;NL=1,uv=e,e.displayName="dax",e.aliases=[];function e(t){t.languages.dax={comment:{pattern:/(^|[^\\])(?:\/\*[\s\S]*?\*\/|(?:--|\/\/).*)/,lookbehind:!0},"data-field":{pattern:/'(?:[^']|'')*'(?!')(?:\[[ \w\xA0-\uFFFF]+\])?|\w+\[[ \w\xA0-\uFFFF]+\]/,alias:"symbol"},measure:{pattern:/\[[ \w\xA0-\uFFFF]+\]/,alias:"constant"},string:{pattern:/"(?:[^"]|"")*"(?!")/,greedy:!0},function:/\b(?:ABS|ACOS|ACOSH|ACOT|ACOTH|ADDCOLUMNS|ADDMISSINGITEMS|ALL|ALLCROSSFILTERED|ALLEXCEPT|ALLNOBLANKROW|ALLSELECTED|AND|APPROXIMATEDISTINCTCOUNT|ASIN|ASINH|ATAN|ATANH|AVERAGE|AVERAGEA|AVERAGEX|BETA\.DIST|BETA\.INV|BLANK|CALCULATE|CALCULATETABLE|CALENDAR|CALENDARAUTO|CEILING|CHISQ\.DIST|CHISQ\.DIST\.RT|CHISQ\.INV|CHISQ\.INV\.RT|CLOSINGBALANCEMONTH|CLOSINGBALANCEQUARTER|CLOSINGBALANCEYEAR|COALESCE|COMBIN|COMBINA|COMBINEVALUES|CONCATENATE|CONCATENATEX|CONFIDENCE\.NORM|CONFIDENCE\.T|CONTAINS|CONTAINSROW|CONTAINSSTRING|CONTAINSSTRINGEXACT|CONVERT|COS|COSH|COT|COTH|COUNT|COUNTA|COUNTAX|COUNTBLANK|COUNTROWS|COUNTX|CROSSFILTER|CROSSJOIN|CURRENCY|CURRENTGROUP|CUSTOMDATA|DATATABLE|DATE|DATEADD|DATEDIFF|DATESBETWEEN|DATESINPERIOD|DATESMTD|DATESQTD|DATESYTD|DATEVALUE|DAY|DEGREES|DETAILROWS|DISTINCT|DISTINCTCOUNT|DISTINCTCOUNTNOBLANK|DIVIDE|EARLIER|EARLIEST|EDATE|ENDOFMONTH|ENDOFQUARTER|ENDOFYEAR|EOMONTH|ERROR|EVEN|EXACT|EXCEPT|EXP|EXPON\.DIST|FACT|FALSE|FILTER|FILTERS|FIND|FIRSTDATE|FIRSTNONBLANK|FIRSTNONBLANKVALUE|FIXED|FLOOR|FORMAT|GCD|GENERATE|GENERATEALL|GENERATESERIES|GEOMEAN|GEOMEANX|GROUPBY|HASONEFILTER|HASONEVALUE|HOUR|IF|IF\.EAGER|IFERROR|IGNORE|INT|INTERSECT|ISBLANK|ISCROSSFILTERED|ISEMPTY|ISERROR|ISEVEN|ISFILTERED|ISINSCOPE|ISLOGICAL|ISNONTEXT|ISNUMBER|ISO\.CEILING|ISODD|ISONORAFTER|ISSELECTEDMEASURE|ISSUBTOTAL|ISTEXT|KEEPFILTERS|KEYWORDMATCH|LASTDATE|LASTNONBLANK|LASTNONBLANKVALUE|LCM|LEFT|LEN|LN|LOG|LOG10|LOOKUPVALUE|LOWER|MAX|MAXA|MAXX|MEDIAN|MEDIANX|MID|MIN|MINA|MINUTE|MINX|MOD|MONTH|MROUND|NATURALINNERJOIN|NATURALLEFTOUTERJOIN|NEXTDAY|NEXTMONTH|NEXTQUARTER|NEXTYEAR|NONVISUAL|NORM\.DIST|NORM\.INV|NORM\.S\.DIST|NORM\.S\.INV|NOT|NOW|ODD|OPENINGBALANCEMONTH|OPENINGBALANCEQUARTER|OPENINGBALANCEYEAR|OR|PARALLELPERIOD|PATH|PATHCONTAINS|PATHITEM|PATHITEMREVERSE|PATHLENGTH|PERCENTILE\.EXC|PERCENTILE\.INC|PERCENTILEX\.EXC|PERCENTILEX\.INC|PERMUT|PI|POISSON\.DIST|POWER|PREVIOUSDAY|PREVIOUSMONTH|PREVIOUSQUARTER|PREVIOUSYEAR|PRODUCT|PRODUCTX|QUARTER|QUOTIENT|RADIANS|RAND|RANDBETWEEN|RANK\.EQ|RANKX|RELATED|RELATEDTABLE|REMOVEFILTERS|REPLACE|REPT|RIGHT|ROLLUP|ROLLUPADDISSUBTOTAL|ROLLUPGROUP|ROLLUPISSUBTOTAL|ROUND|ROUNDDOWN|ROUNDUP|ROW|SAMEPERIODLASTYEAR|SAMPLE|SEARCH|SECOND|SELECTCOLUMNS|SELECTEDMEASURE|SELECTEDMEASUREFORMATSTRING|SELECTEDMEASURENAME|SELECTEDVALUE|SIGN|SIN|SINH|SQRT|SQRTPI|STARTOFMONTH|STARTOFQUARTER|STARTOFYEAR|STDEV\.P|STDEV\.S|STDEVX\.P|STDEVX\.S|SUBSTITUTE|SUBSTITUTEWITHINDEX|SUM|SUMMARIZE|SUMMARIZECOLUMNS|SUMX|SWITCH|T\.DIST|T\.DIST\.2T|T\.DIST\.RT|T\.INV|T\.INV\.2T|TAN|TANH|TIME|TIMEVALUE|TODAY|TOPN|TOPNPERLEVEL|TOPNSKIP|TOTALMTD|TOTALQTD|TOTALYTD|TREATAS|TRIM|TRUE|TRUNC|UNICHAR|UNICODE|UNION|UPPER|USERELATIONSHIP|USERNAME|USEROBJECTID|USERPRINCIPALNAME|UTCNOW|UTCTODAY|VALUE|VALUES|VAR\.P|VAR\.S|VARX\.P|VARX\.S|WEEKDAY|WEEKNUM|XIRR|XNPV|YEAR|YEARFRAC)(?=\s*\()/i,keyword:/\b(?:DEFINE|EVALUATE|MEASURE|ORDER\s+BY|RETURN|VAR|START\s+AT|ASC|DESC)\b/i,boolean:{pattern:/\b(?:FALSE|NULL|TRUE)\b/i,alias:"constant"},number:/\b\d+(?:\.\d*)?|\B\.\d+\b/,operator:/:=|[-+*\/=^]|&&?|\|\||<(?:=>?|<|>)?|>[>=]?|\b(?:IN|NOT)\b/i,punctuation:/[;\[\](){}`,.]/}}return uv}var cv,OL;function Pbe(){if(OL)return cv;OL=1,cv=e,e.displayName="dhall",e.aliases=[];function e(t){t.languages.dhall={comment:/--.*|\{-(?:[^-{]|-(?!\})|\{(?!-)|\{-(?:[^-{]|-(?!\})|\{(?!-))*-\})*-\}/,string:{pattern:/"(?:[^"\\]|\\.)*"|''(?:[^']|'(?!')|'''|''\$\{)*''(?!'|\$)/,greedy:!0,inside:{interpolation:{pattern:/\$\{[^{}]*\}/,inside:{expression:{pattern:/(^\$\{)[\s\S]+(?=\}$)/,lookbehind:!0,alias:"language-dhall",inside:null},punctuation:/\$\{|\}/}}}},label:{pattern:/`[^`]*`/,greedy:!0},url:{pattern:/\bhttps?:\/\/[\w.:%!$&'*+;=@~-]+(?:\/[\w.:%!$&'*+;=@~-]*)*(?:\?[/?\w.:%!$&'*+;=@~-]*)?/,greedy:!0},env:{pattern:/\benv:(?:(?!\d)\w+|"(?:[^"\\=]|\\.)*")/,greedy:!0,inside:{function:/^env/,operator:/^:/,variable:/[\s\S]+/}},hash:{pattern:/\bsha256:[\da-fA-F]{64}\b/,inside:{function:/sha256/,operator:/:/,number:/[\da-fA-F]{64}/}},keyword:/\b(?:as|assert|else|forall|if|in|let|merge|missing|then|toMap|using|with)\b|\u2200/,builtin:/\b(?:None|Some)\b/,boolean:/\b(?:False|True)\b/,number:/\bNaN\b|-?\bInfinity\b|[+-]?\b(?:0x[\da-fA-F]+|\d+(?:\.\d+)?(?:e[+-]?\d+)?)\b/,operator:/\/\\|\/\/\\\\|&&|\|\||===|[!=]=|\/\/|->|\+\+|::|[+*#@=:?<>|\\\u2227\u2a53\u2261\u2afd\u03bb\u2192]/,punctuation:/\.\.|[{}\[\](),./]/,"class-name":/\b[A-Z]\w*\b/},t.languages.dhall.string.inside.interpolation.inside.expression.inside=t.languages.dhall}return cv}var dv,DL;function zbe(){if(DL)return dv;DL=1,dv=e,e.displayName="diff",e.aliases=[];function e(t){(function(n){n.languages.diff={coord:[/^(?:\*{3}|-{3}|\+{3}).*$/m,/^@@.*@@$/m,/^\d.*$/m]};var r={"deleted-sign":"-","deleted-arrow":"<","inserted-sign":"+","inserted-arrow":">",unchanged:" ",diff:"!"};Object.keys(r).forEach(function(a){var o=r[a],s=[];/^\w+$/.test(a)||s.push(/\w+/.exec(a)[0]),a==="diff"&&s.push("bold"),n.languages.diff[a]={pattern:RegExp("^(?:["+o+`].*(?:\r
?|
|(?![\\s\\S])))+`,"m"),alias:s,inside:{line:{pattern:/(.)(?=[\s\S]).*(?:\r\n?|\n)?/,lookbehind:!0},prefix:{pattern:/[\s\S]/,alias:/\w+/.exec(a)[0]}}}}),Object.defineProperty(n.languages.diff,"PREFIXES",{value:r})})(t)}return dv}var fv,IL;function Wn(){if(IL)return fv;IL=1,fv=e,e.displayName="markupTemplating",e.aliases=[];function e(t){(function(n){function r(a,o){return"___"+a.toUpperCase()+o+"___"}Object.defineProperties(n.languages["markup-templating"]={},{buildPlaceholders:{value:function(a,o,s,u){if(a.language===o){var c=a.tokenStack=[];a.code=a.code.replace(s,function(d){if(typeof u=="function"&&!u(d))return d;for(var p=c.length,g;a.code.indexOf(g=r(o,p))!==-1;)++p;return c[p]=d,g}),a.grammar=n.languages.markup}}},tokenizePlaceholders:{value:function(a,o){if(a.language!==o||!a.tokenStack)return;a.grammar=n.languages[o];var s=0,u=Object.keys(a.tokenStack);function c(d){for(var p=0;p<d.length&&!(s>=u.length);p++){var g=d[p];if(typeof g=="string"||g.content&&typeof g.content=="string"){var m=u[s],b=a.tokenStack[m],y=typeof g=="string"?g:g.content,S=r(o,m),x=y.indexOf(S);if(x>-1){++s;var R=y.substring(0,x),k=new n.Token(o,n.tokenize(b,a.grammar),"language-"+o,b),A=y.substring(x+S.length),C=[];R&&C.push.apply(C,c([R])),C.push(k),A&&C.push.apply(C,c([A])),typeof g=="string"?d.splice.apply(d,[p,1].concat(C)):g.content=C}}else g.content&&c(g.content)}return d}c(a.tokens)}}})})(t)}return fv}var pv,LL;function Bbe(){if(LL)return pv;LL=1;var e=Wn();pv=t,t.displayName="django",t.aliases=["jinja2"];function t(n){n.register(e),function(r){r.languages.django={comment:/^\{#[\s\S]*?#\}$/,tag:{pattern:/(^\{%[+-]?\s*)\w+/,lookbehind:!0,alias:"keyword"},delimiter:{pattern:/^\{[{%][+-]?|[+-]?[}%]\}$/,alias:"punctuation"},string:{pattern:/("|')(?:\\.|(?!\1)[^\\\r\n])*\1/,greedy:!0},filter:{pattern:/(\|)\w+/,lookbehind:!0,alias:"function"},test:{pattern:/(\bis\s+(?:not\s+)?)(?!not\b)\w+/,lookbehind:!0,alias:"function"},function:/\b[a-z_]\w+(?=\s*\()/i,keyword:/\b(?:and|as|by|else|for|if|import|in|is|loop|not|or|recursive|with|without)\b/,operator:/[-+%=]=?|!=|\*\*?=?|\/\/?=?|<[<=>]?|>[=>]?|[&|^~]/,number:/\b\d+(?:\.\d+)?\b/,boolean:/[Ff]alse|[Nn]one|[Tt]rue/,variable:/\b\w+\b/,punctuation:/[{}[\](),.:;]/};var a=/\{\{[\s\S]*?\}\}|\{%[\s\S]*?%\}|\{#[\s\S]*?#\}/g,o=r.languages["markup-templating"];r.hooks.add("before-tokenize",function(s){o.buildPlaceholders(s,"django",a)}),r.hooks.add("after-tokenize",function(s){o.tokenizePlaceholders(s,"django")}),r.languages.jinja2=r.languages.django,r.hooks.add("before-tokenize",function(s){o.buildPlaceholders(s,"jinja2",a)}),r.hooks.add("after-tokenize",function(s){o.tokenizePlaceholders(s,"jinja2")})}(n)}return pv}var gv,ML;function Ube(){if(ML)return gv;ML=1,gv=e,e.displayName="dnsZoneFile",e.aliases=[];function e(t){t.languages["dns-zone-file"]={comment:/;.*/,string:{pattern:/"(?:\\.|[^"\\\r\n])*"/,greedy:!0},variable:[{pattern:/(^\$ORIGIN[ \t]+)\S+/m,lookbehind:!0},{pattern:/(^|\s)@(?=\s|$)/,lookbehind:!0}],keyword:/^\$(?:INCLUDE|ORIGIN|TTL)(?=\s|$)/m,class:{pattern:/(^|\s)(?:CH|CS|HS|IN)(?=\s|$)/,lookbehind:!0,alias:"keyword"},type:{pattern:/(^|\s)(?:A|A6|AAAA|AFSDB|APL|ATMA|CAA|CDNSKEY|CDS|CERT|CNAME|DHCID|DLV|DNAME|DNSKEY|DS|EID|GID|GPOS|HINFO|HIP|IPSECKEY|ISDN|KEY|KX|LOC|MAILA|MAILB|MB|MD|MF|MG|MINFO|MR|MX|NAPTR|NB|NBSTAT|NIMLOC|NINFO|NS|NSAP|NSAP-PTR|NSEC|NSEC3|NSEC3PARAM|NULL|NXT|OPENPGPKEY|PTR|PX|RKEY|RP|RRSIG|RT|SIG|SINK|SMIMEA|SOA|SPF|SRV|SSHFP|TA|TKEY|TLSA|TSIG|TXT|UID|UINFO|UNSPEC|URI|WKS|X25)(?=\s|$)/,lookbehind:!0,alias:"keyword"},punctuation:/[()]/},t.languages["dns-zone"]=t.languages["dns-zone-file"]}return gv}var hv,FL;function jbe(){if(FL)return hv;FL=1,hv=e,e.displayName="docker",e.aliases=["dockerfile"];function e(t){(function(n){var r=/\\[\r\n](?:\s|\\[\r\n]|#.*(?!.))*(?![\s#]|\\[\r\n])/.source,a=/(?:[ \t]+(?![ \t])(?:<SP_BS>)?|<SP_BS>)/.source.replace(/<SP_BS>/g,function(){return r}),o=/"(?:[^"\\\r\n]|\\(?:\r\n|[\s\S]))*"|'(?:[^'\\\r\n]|\\(?:\r\n|[\s\S]))*'/.source,s=/--[\w-]+=(?:<STR>|(?!["'])(?:[^\s\\]|\\.)+)/.source.replace(/<STR>/g,function(){return o}),u={pattern:RegExp(o),greedy:!0},c={pattern:/(^[ \t]*)#.*/m,lookbehind:!0,greedy:!0};function d(p,g){return p=p.replace(/<OPT>/g,function(){return s}).replace(/<SP>/g,function(){return a}),RegExp(p,g)}n.languages.docker={instruction:{pattern:/(^[ \t]*)(?:ADD|ARG|CMD|COPY|ENTRYPOINT|ENV|EXPOSE|FROM|HEALTHCHECK|LABEL|MAINTAINER|ONBUILD|RUN|SHELL|STOPSIGNAL|USER|VOLUME|WORKDIR)(?=\s)(?:\\.|[^\r\n\\])*(?:\\$(?:\s|#.*$)*(?![\s#])(?:\\.|[^\r\n\\])*)*/im,lookbehind:!0,greedy:!0,inside:{options:{pattern:d(/(^(?:ONBUILD<SP>)?\w+<SP>)<OPT>(?:<SP><OPT>)*/.source,"i"),lookbehind:!0,greedy:!0,inside:{property:{pattern:/(^|\s)--[\w-]+/,lookbehind:!0},string:[u,{pattern:/(=)(?!["'])(?:[^\s\\]|\\.)+/,lookbehind:!0}],operator:/\\$/m,punctuation:/=/}},keyword:[{pattern:d(/(^(?:ONBUILD<SP>)?HEALTHCHECK<SP>(?:<OPT><SP>)*)(?:CMD|NONE)\b/.source,"i"),lookbehind:!0,greedy:!0},{pattern:d(/(^(?:ONBUILD<SP>)?FROM<SP>(?:<OPT><SP>)*(?!--)[^ \t\\]+<SP>)AS/.source,"i"),lookbehind:!0,greedy:!0},{pattern:d(/(^ONBUILD<SP>)\w+/.source,"i"),lookbehind:!0,greedy:!0},{pattern:/^\w+/,greedy:!0}],comment:c,string:u,variable:/\$(?:\w+|\{[^{}"'\\]*\})/,operator:/\\$/m}},comment:c},n.languages.dockerfile=n.languages.docker})(t)}return hv}var mv,PL;function Gbe(){if(PL)return mv;PL=1,mv=e,e.displayName="dot",e.aliases=["gv"];function e(t){(function(n){var r="(?:"+[/[a-zA-Z_\x80-\uFFFF][\w\x80-\uFFFF]*/.source,/-?(?:\.\d+|\d+(?:\.\d*)?)/.source,/"[^"\\]*(?:\\[\s\S][^"\\]*)*"/.source,/<(?:[^<>]|(?!<!--)<(?:[^<>"']|"[^"]*"|'[^']*')+>|<!--(?:[^-]|-(?!->))*-->)*>/.source].join("|")+")",a={markup:{pattern:/(^<)[\s\S]+(?=>$)/,lookbehind:!0,alias:["language-markup","language-html","language-xml"],inside:n.languages.markup}};function o(s,u){return RegExp(s.replace(/<ID>/g,function(){return r}),u)}n.languages.dot={comment:{pattern:/\/\/.*|\/\*[\s\S]*?\*\/|^#.*/m,greedy:!0},"graph-name":{pattern:o(/(\b(?:digraph|graph|subgraph)[ \t\r\n]+)<ID>/.source,"i"),lookbehind:!0,greedy:!0,alias:"class-name",inside:a},"attr-value":{pattern:o(/(=[ \t\r\n]*)<ID>/.source),lookbehind:!0,greedy:!0,inside:a},"attr-name":{pattern:o(/([\[;, \t\r\n])<ID>(?=[ \t\r\n]*=)/.source),lookbehind:!0,greedy:!0,inside:a},keyword:/\b(?:digraph|edge|graph|node|strict|subgraph)\b/i,"compass-point":{pattern:/(:[ \t\r\n]*)(?:[ewc_]|[ns][ew]?)(?![\w\x80-\uFFFF])/,lookbehind:!0,alias:"builtin"},node:{pattern:o(/(^|[^-.\w\x80-\uFFFF\\])<ID>/.source),lookbehind:!0,greedy:!0,inside:a},operator:/[=:]|-[->]/,punctuation:/[\[\]{};,]/},n.languages.gv=n.languages.dot})(t)}return mv}var bv,zL;function Hbe(){if(zL)return bv;zL=1,bv=e,e.displayName="ebnf",e.aliases=[];function e(t){t.languages.ebnf={comment:/\(\*[\s\S]*?\*\)/,string:{pattern:/"[^"\r\n]*"|'[^'\r\n]*'/,greedy:!0},special:{pattern:/\?[^?\r\n]*\?/,greedy:!0,alias:"class-name"},definition:{pattern:/^([\t ]*)[a-z]\w*(?:[ \t]+[a-z]\w*)*(?=\s*=)/im,lookbehind:!0,alias:["rule","keyword"]},rule:/\b[a-z]\w*(?:[ \t]+[a-z]\w*)*\b/i,punctuation:/\([:/]|[:/]\)|[.,;()[\]{}]/,operator:/[-=|*/!]/}}return bv}var yv,BL;function $be(){if(BL)return yv;BL=1,yv=e,e.displayName="editorconfig",e.aliases=[];function e(t){t.languages.editorconfig={comment:/[;#].*/,section:{pattern:/(^[ \t]*)\[.+\]/m,lookbehind:!0,alias:"selector",inside:{regex:/\\\\[\[\]{},!?.*]/,operator:/[!?]|\.\.|\*{1,2}/,punctuation:/[\[\]{},]/}},key:{pattern:/(^[ \t]*)[^\s=]+(?=[ \t]*=)/m,lookbehind:!0,alias:"attr-name"},value:{pattern:/=.*/,alias:"attr-value",inside:{punctuation:/^=/}}}}return yv}var vv,UL;function qbe(){if(UL)return vv;UL=1,vv=e,e.displayName="eiffel",e.aliases=[];function e(t){t.languages.eiffel={comment:/--.*/,string:[{pattern:/"([^[]*)\[[\s\S]*?\]\1"/,greedy:!0},{pattern:/"([^{]*)\{[\s\S]*?\}\1"/,greedy:!0},{pattern:/"(?:%(?:(?!\n)\s)*\n\s*%|%\S|[^%"\r\n])*"/,greedy:!0}],char:/'(?:%.|[^%'\r\n])+'/,keyword:/\b(?:across|agent|alias|all|and|as|assign|attached|attribute|check|class|convert|create|Current|debug|deferred|detachable|do|else|elseif|end|ensure|expanded|export|external|feature|from|frozen|if|implies|inherit|inspect|invariant|like|local|loop|not|note|obsolete|old|once|or|Precursor|redefine|rename|require|rescue|Result|retry|select|separate|some|then|undefine|until|variant|Void|when|xor)\b/i,boolean:/\b(?:False|True)\b/i,"class-name":/\b[A-Z][\dA-Z_]*\b/,number:[/\b0[xcb][\da-f](?:_*[\da-f])*\b/i,/(?:\b\d(?:_*\d)*)?\.(?:(?:\d(?:_*\d)*)?e[+-]?)?\d(?:_*\d)*\b|\b\d(?:_*\d)*\b\.?/i],punctuation:/:=|<<|>>|\(\||\|\)|->|\.(?=\w)|[{}[\];(),:?]/,operator:/\\\\|\|\.\.\||\.\.|\/[~\/=]?|[><]=?|[-+*^=~]/}}return vv}var Sv,jL;function Vbe(){if(jL)return Sv;jL=1;var e=Wn();Sv=t,t.displayName="ejs",t.aliases=["eta"];function t(n){n.register(e),function(r){r.languages.ejs={delimiter:{pattern:/^<%[-_=]?|[-_]?%>$/,alias:"punctuation"},comment:/^#[\s\S]*/,"language-javascript":{pattern:/[\s\S]+/,inside:r.languages.javascript}},r.hooks.add("before-tokenize",function(a){var o=/<%(?!%)[\s\S]+?%>/g;r.languages["markup-templating"].buildPlaceholders(a,"ejs",o)}),r.hooks.add("after-tokenize",function(a){r.languages["markup-templating"].tokenizePlaceholders(a,"ejs")}),r.languages.eta=r.languages.ejs}(n)}return Sv}var Ev,GL;function Wbe(){if(GL)return Ev;GL=1,Ev=e,e.displayName="elixir",e.aliases=[];function e(t){t.languages.elixir={doc:{pattern:/@(?:doc|moduledoc)\s+(?:("""|''')[\s\S]*?\1|("|')(?:\\(?:\r\n|[\s\S])|(?!\2)[^\\\r\n])*\2)/,inside:{attribute:/^@\w+/,string:/['"][\s\S]+/}},comment:{pattern:/#.*/,greedy:!0},regex:{pattern:/~[rR](?:("""|''')(?:\\[\s\S]|(?!\1)[^\\])+\1|([\/|"'])(?:\\.|(?!\2)[^\\\r\n])+\2|\((?:\\.|[^\\)\r\n])+\)|\[(?:\\.|[^\\\]\r\n])+\]|\{(?:\\.|[^\\}\r\n])+\}|<(?:\\.|[^\\>\r\n])+>)[uismxfr]*/,greedy:!0},string:[{pattern:/~[cCsSwW](?:("""|''')(?:\\[\s\S]|(?!\1)[^\\])+\1|([\/|"'])(?:\\.|(?!\2)[^\\\r\n])+\2|\((?:\\.|[^\\)\r\n])+\)|\[(?:\\.|[^\\\]\r\n])+\]|\{(?:\\.|#\{[^}]+\}|#(?!\{)|[^#\\}\r\n])+\}|<(?:\\.|[^\\>\r\n])+>)[csa]?/,greedy:!0,inside:{}},{pattern:/("""|''')[\s\S]*?\1/,greedy:!0,inside:{}},{pattern:/("|')(?:\\(?:\r\n|[\s\S])|(?!\1)[^\\\r\n])*\1/,greedy:!0,inside:{}}],atom:{pattern:/(^|[^:]):\w+/,lookbehind:!0,alias:"symbol"},module:{pattern:/\b[A-Z]\w*\b/,alias:"class-name"},"attr-name":/\b\w+\??:(?!:)/,argument:{pattern:/(^|[^&])&\d+/,lookbehind:!0,alias:"variable"},attribute:{pattern:/@\w+/,alias:"variable"},function:/\b[_a-zA-Z]\w*[?!]?(?:(?=\s*(?:\.\s*)?\()|(?=\/\d))/,number:/\b(?:0[box][a-f\d_]+|\d[\d_]*)(?:\.[\d_]+)?(?:e[+-]?[\d_]+)?\b/i,keyword:/\b(?:after|alias|and|case|catch|cond|def(?:callback|delegate|exception|impl|macro|module|n|np|p|protocol|struct)?|do|else|end|fn|for|if|import|not|or|quote|raise|require|rescue|try|unless|unquote|use|when)\b/,boolean:/\b(?:false|nil|true)\b/,operator:[/\bin\b|&&?|\|[|>]?|\\\\|::|\.\.\.?|\+\+?|-[->]?|<[-=>]|>=|!==?|\B!|=(?:==?|[>~])?|[*\/^]/,{pattern:/([^<])<(?!<)/,lookbehind:!0},{pattern:/([^>])>(?!>)/,lookbehind:!0}],punctuation:/<<|>>|[.,%\[\]{}()]/},t.languages.elixir.string.forEach(function(n){n.inside={interpolation:{pattern:/#\{[^}]+\}/,inside:{delimiter:{pattern:/^#\{|\}$/,alias:"punctuation"},rest:t.languages.elixir}}}})}return Ev}var wv,HL;function Ybe(){if(HL)return wv;HL=1,wv=e,e.displayName="elm",e.aliases=[];function e(t){t.languages.elm={comment:/--.*|\{-[\s\S]*?-\}/,char:{pattern:/'(?:[^\\'\r\n]|\\(?:[abfnrtv\\']|\d+|x[0-9a-fA-F]+|u\{[0-9a-fA-F]+\}))'/,greedy:!0},string:[{pattern:/"""[\s\S]*?"""/,greedy:!0},{pattern:/"(?:[^\\"\r\n]|\\.)*"/,greedy:!0}],"import-statement":{pattern:/(^[\t ]*)import\s+[A-Z]\w*(?:\.[A-Z]\w*)*(?:\s+as\s+(?:[A-Z]\w*)(?:\.[A-Z]\w*)*)?(?:\s+exposing\s+)?/m,lookbehind:!0,inside:{keyword:/\b(?:as|exposing|import)\b/}},keyword:/\b(?:alias|as|case|else|exposing|if|in|infixl|infixr|let|module|of|then|type)\b/,builtin:/\b(?:abs|acos|always|asin|atan|atan2|ceiling|clamp|compare|cos|curry|degrees|e|flip|floor|fromPolar|identity|isInfinite|isNaN|logBase|max|min|negate|never|not|pi|radians|rem|round|sin|sqrt|tan|toFloat|toPolar|toString|truncate|turns|uncurry|xor)\b/,number:/\b(?:\d+(?:\.\d+)?(?:e[+-]?\d+)?|0x[0-9a-f]+)\b/i,operator:/\s\.\s|[+\-/*=.$<>:&|^?%#@~!]{2,}|[+\-/*=$<>:&|^?%#@~!]/,hvariable:/\b(?:[A-Z]\w*\.)*[a-z]\w*\b/,constant:/\b(?:[A-Z]\w*\.)*[A-Z]\w*\b/,punctuation:/[{}[\]|(),.:]/}}return wv}var xv,$L;function Kbe(){if($L)return xv;$L=1;var e=Fp(),t=Wn();xv=n,n.displayName="erb",n.aliases=[];function n(r){r.register(e),r.register(t),function(a){a.languages.erb={delimiter:{pattern:/^(\s*)<%=?|%>(?=\s*$)/,lookbehind:!0,alias:"punctuation"},ruby:{pattern:/\s*\S[\s\S]*/,alias:"language-ruby",inside:a.languages.ruby}},a.hooks.add("before-tokenize",function(o){var s=/<%=?(?:[^\r\n]|[\r\n](?!=begin)|[\r\n]=begin\s(?:[^\r\n]|[\r\n](?!=end))*[\r\n]=end)+?%>/g;a.languages["markup-templating"].buildPlaceholders(o,"erb",s)}),a.hooks.add("after-tokenize",function(o){a.languages["markup-templating"].tokenizePlaceholders(o,"erb")})}(r)}return xv}var kv,qL;function Xbe(){if(qL)return kv;qL=1,kv=e,e.displayName="erlang",e.aliases=[];function e(t){t.languages.erlang={comment:/%.+/,string:{pattern:/"(?:\\.|[^\\"\r\n])*"/,greedy:!0},"quoted-function":{pattern:/'(?:\\.|[^\\'\r\n])+'(?=\()/,alias:"function"},"quoted-atom":{pattern:/'(?:\\.|[^\\'\r\n])+'/,alias:"atom"},boolean:/\b(?:false|true)\b/,keyword:/\b(?:after|case|catch|end|fun|if|of|receive|try|when)\b/,number:[/\$\\?./,/\b\d+#[a-z0-9]+/i,/(?:\b\d+(?:\.\d*)?|\B\.\d+)(?:e[+-]?\d+)?/i],function:/\b[a-z][\w@]*(?=\()/,variable:{pattern:/(^|[^@])(?:\b|\?)[A-Z_][\w@]*/,lookbehind:!0},operator:[/[=\/<>:]=|=[:\/]=|\+\+?|--?|[=*\/!]|\b(?:and|andalso|band|bnot|bor|bsl|bsr|bxor|div|not|or|orelse|rem|xor)\b/,{pattern:/(^|[^<])<(?!<)/,lookbehind:!0},{pattern:/(^|[^>])>(?!>)/,lookbehind:!0}],atom:/\b[a-z][\w@]*/,punctuation:/[()[\]{}:;,.#|]|<<|>>/}}return kv}var Tv,VL;function $H(){if(VL)return Tv;VL=1,Tv=e,e.displayName="lua",e.aliases=[];function e(t){t.languages.lua={comment:/^#!.+|--(?:\[(=*)\[[\s\S]*?\]\1\]|.*)/m,string:{pattern:/(["'])(?:(?!\1)[^\\\r\n]|\\z(?:\r\n|\s)|\\(?:\r\n|[^z]))*\1|\[(=*)\[[\s\S]*?\]\2\]/,greedy:!0},number:/\b0x[a-f\d]+(?:\.[a-f\d]*)?(?:p[+-]?\d+)?\b|\b\d+(?:\.\B|(?:\.\d*)?(?:e[+-]?\d+)?\b)|\B\.\d+(?:e[+-]?\d+)?\b/i,keyword:/\b(?:and|break|do|else|elseif|end|false|for|function|goto|if|in|local|nil|not|or|repeat|return|then|true|until|while)\b/,function:/(?!\d)\w+(?=\s*(?:[({]))/,operator:[/[-+*%^&|#]|\/\/?|<[<=]?|>[>=]?|[=~]=?/,{pattern:/(^|[^.])\.\.(?!\.)/,lookbehind:!0}],punctuation:/[\[\](){},;]|\.+|:+/}}return Tv}var Av,WL;function Zbe(){if(WL)return Av;WL=1;var e=$H(),t=Wn();Av=n,n.displayName="etlua",n.aliases=[];function n(r){r.register(e),r.register(t),function(a){a.languages.etlua={delimiter:{pattern:/^<%[-=]?|-?%>$/,alias:"punctuation"},"language-lua":{pattern:/[\s\S]+/,inside:a.languages.lua}},a.hooks.add("before-tokenize",function(o){var s=/<%[\s\S]+?%>/g;a.languages["markup-templating"].buildPlaceholders(o,"etlua",s)}),a.hooks.add("after-tokenize",function(o){a.languages["markup-templating"].tokenizePlaceholders(o,"etlua")})}(r)}return Av}var Rv,YL;function Qbe(){if(YL)return Rv;YL=1,Rv=e,e.displayName="excelFormula",e.aliases=[];function e(t){t.languages["excel-formula"]={comment:{pattern:/(\bN\(\s*)"(?:[^"]|"")*"(?=\s*\))/i,lookbehind:!0,greedy:!0},string:{pattern:/"(?:[^"]|"")*"(?!")/,greedy:!0},reference:{pattern:/(?:'[^']*'|(?:[^\s()[\]{}<>*?"';,$&]*\[[^^\s()[\]{}<>*?"']+\])?\w+)!/,greedy:!0,alias:"string",inside:{operator:/!$/,punctuation:/'/,sheet:{pattern:/[^[\]]+$/,alias:"function"},file:{pattern:/\[[^[\]]+\]$/,inside:{punctuation:/[[\]]/}},path:/[\s\S]+/}},"function-name":{pattern:/\b[A-Z]\w*(?=\()/i,alias:"keyword"},range:{pattern:/\$?\b(?:[A-Z]+\$?\d+:\$?[A-Z]+\$?\d+|[A-Z]+:\$?[A-Z]+|\d+:\$?\d+)\b/i,alias:"property",inside:{operator:/:/,cell:/\$?[A-Z]+\$?\d+/i,column:/\$?[A-Z]+/i,row:/\$?\d+/}},cell:{pattern:/\b[A-Z]+\d+\b|\$[A-Za-z]+\$?\d+\b|\b[A-Za-z]+\$\d+\b/,alias:"property"},number:/(?:\b\d+(?:\.\d+)?|\B\.\d+)(?:e[+-]?\d+)?\b/i,boolean:/\b(?:FALSE|TRUE)\b/i,operator:/[-+*/^%=&,]|<[=>]?|>=?/,punctuation:/[[\]();{}|]/},t.languages.xlsx=t.languages.xls=t.languages["excel-formula"]}return Rv}var _v,KL;function Jbe(){if(KL)return _v;KL=1,_v=e,e.displayName="factor",e.aliases=[];function e(t){(function(n){var r={function:/\b(?:BUGS?|FIX(?:MES?)?|NOTES?|TODOS?|XX+|HACKS?|WARN(?:ING)?|\?{2,}|!{2,})\b/},a={number:/\\[^\s']|%\w/},o={comment:[{pattern:/(^|\s)(?:! .*|!$)/,lookbehind:!0,inside:r},{pattern:/(^|\s)\/\*\s[\s\S]*?\*\/(?=\s|$)/,lookbehind:!0,greedy:!0,inside:r},{pattern:/(^|\s)!\[(={0,6})\[\s[\s\S]*?\]\2\](?=\s|$)/,lookbehind:!0,greedy:!0,inside:r}],number:[{pattern:/(^|\s)[+-]?\d+(?=\s|$)/,lookbehind:!0},{pattern:/(^|\s)[+-]?0(?:b[01]+|o[0-7]+|d\d+|x[\dA-F]+)(?=\s|$)/i,lookbehind:!0},{pattern:/(^|\s)[+-]?\d+\/\d+\.?(?=\s|$)/,lookbehind:!0},{pattern:/(^|\s)\+?\d+\+\d+\/\d+(?=\s|$)/,lookbehind:!0},{pattern:/(^|\s)-\d+-\d+\/\d+(?=\s|$)/,lookbehind:!0},{pattern:/(^|\s)[+-]?(?:\d*\.\d+|\d+\.\d*|\d+)(?:e[+-]?\d+)?(?=\s|$)/i,lookbehind:!0},{pattern:/(^|\s)NAN:\s+[\da-fA-F]+(?=\s|$)/,lookbehind:!0},{pattern:/(^|\s)[+-]?0(?:b1\.[01]*|o1\.[0-7]*|d1\.\d*|x1\.[\dA-F]*)p\d+(?=\s|$)/i,lookbehind:!0}],regexp:{pattern:/(^|\s)R\/\s(?:\\\S|[^\\/])*\/(?:[idmsr]*|[idmsr]+-[idmsr]+)(?=\s|$)/,lookbehind:!0,alias:"number",inside:{variable:/\\\S/,keyword:/[+?*\[\]^$(){}.|]/,operator:{pattern:/(\/)[idmsr]+(?:-[idmsr]+)?/,lookbehind:!0}}},boolean:{pattern:/(^|\s)[tf](?=\s|$)/,lookbehind:!0},"custom-string":{pattern:/(^|\s)[A-Z0-9\-]+"\s(?:\\\S|[^"\\])*"/,lookbehind:!0,greedy:!0,alias:"string",inside:{number:/\\\S|%\w|\//}},"multiline-string":[{pattern:/(^|\s)STRING:\s+\S+(?:\n|\r\n).*(?:\n|\r\n)\s*;(?=\s|$)/,lookbehind:!0,greedy:!0,alias:"string",inside:{number:a.number,"semicolon-or-setlocal":{pattern:/([\r\n][ \t]*);(?=\s|$)/,lookbehind:!0,alias:"function"}}},{pattern:/(^|\s)HEREDOC:\s+\S+(?:\n|\r\n).*(?:\n|\r\n)\s*\S+(?=\s|$)/,lookbehind:!0,greedy:!0,alias:"string",inside:a},{pattern:/(^|\s)\[(={0,6})\[\s[\s\S]*?\]\2\](?=\s|$)/,lookbehind:!0,greedy:!0,alias:"string",inside:a}],"special-using":{pattern:/(^|\s)USING:(?:\s\S+)*(?=\s+;(?:\s|$))/,lookbehind:!0,alias:"function",inside:{string:{pattern:/(\s)[^:\s]+/,lookbehind:!0}}},"stack-effect-delimiter":[{pattern:/(^|\s)(?:call|eval|execute)?\((?=\s)/,lookbehind:!0,alias:"operator"},{pattern:/(\s)--(?=\s)/,lookbehind:!0,alias:"operator"},{pattern:/(\s)\)(?=\s|$)/,lookbehind:!0,alias:"operator"}],combinators:{pattern:null,lookbehind:!0,alias:"keyword"},"kernel-builtin":{pattern:null,lookbehind:!0,alias:"variable"},"sequences-builtin":{pattern:null,lookbehind:!0,alias:"variable"},"math-builtin":{pattern:null,lookbehind:!0,alias:"variable"},"constructor-word":{pattern:/(^|\s)<(?!=+>|-+>)\S+>(?=\s|$)/,lookbehind:!0,alias:"keyword"},"other-builtin-syntax":{pattern:null,lookbehind:!0,alias:"operator"},"conventionally-named-word":{pattern:/(^|\s)(?!")(?:(?:change|new|set|with)-\S+|\$\S+|>[^>\s]+|[^:>\s]+>|[^>\s]+>[^>\s]+|\+[^+\s]+\+|[^?\s]+\?|\?[^?\s]+|[^>\s]+>>|>>[^>\s]+|[^<\s]+<<|\([^()\s]+\)|[^!\s]+!|[^*\s]\S*\*|[^.\s]\S*\.)(?=\s|$)/,lookbehind:!0,alias:"keyword"},"colon-syntax":{pattern:/(^|\s)(?:[A-Z0-9\-]+#?)?:{1,2}\s+(?:;\S+|(?!;)\S+)(?=\s|$)/,lookbehind:!0,greedy:!0,alias:"function"},"semicolon-or-setlocal":{pattern:/(\s)(?:;|:>)(?=\s|$)/,lookbehind:!0,alias:"function"},"curly-brace-literal-delimiter":[{pattern:/(^|\s)[a-z]*\{(?=\s)/i,lookbehind:!0,alias:"operator"},{pattern:/(\s)\}(?=\s|$)/,lookbehind:!0,alias:"operator"}],"quotation-delimiter":[{pattern:/(^|\s)\[(?=\s)/,lookbehind:!0,alias:"operator"},{pattern:/(\s)\](?=\s|$)/,lookbehind:!0,alias:"operator"}],"normal-word":{pattern:/(^|\s)[^"\s]\S*(?=\s|$)/,lookbehind:!0},string:{pattern:/"(?:\\\S|[^"\\])*"/,greedy:!0,inside:a}},s=function(p){return(p+"").replace(/([.?*+\^$\[\]\\(){}|\-])/g,"\\$1")},u=function(p){return new RegExp("(^|\\s)(?:"+p.map(s).join("|")+")(?=\\s|$)")},c={"kernel-builtin":["or","2nipd","4drop","tuck","wrapper","nip","wrapper?","callstack>array","die","dupd","callstack","callstack?","3dup","hashcode","pick","4nip","build",">boolean","nipd","clone","5nip","eq?","?","=","swapd","2over","clear","2dup","get-retainstack","not","tuple?","dup","3nipd","call","-rotd","object","drop","assert=","assert?","-rot","execute","boa","get-callstack","curried?","3drop","pickd","overd","over","roll","3nip","swap","and","2nip","rotd","throw","(clone)","hashcode*","spin","reach","4dup","equal?","get-datastack","assert","2drop","<wrapper>","boolean?","identity-hashcode","identity-tuple?","null","composed?","new","5drop","rot","-roll","xor","identity-tuple","boolean"],"other-builtin-syntax":["=======","recursive","flushable",">>","<<<<<<","M\\","B","PRIVATE>","\\","======","final","inline","delimiter","deprecated","<PRIVATE",">>>>>>","<<<<<<<","parse-complex","malformed-complex","read-only",">>>>>>>","call-next-method","<<","foldable","$","$[","${"],"sequences-builtin":["member-eq?","mismatch","append","assert-sequence=","longer","repetition","clone-like","3sequence","assert-sequence?","last-index-from","reversed","index-from","cut*","pad-tail","join-as","remove-eq!","concat-as","but-last","snip","nths","nth","sequence","longest","slice?","<slice>","remove-nth","tail-slice","empty?","tail*","member?","virtual-sequence?","set-length","drop-prefix","iota","unclip","bounds-error?","unclip-last-slice","non-negative-integer-expected","non-negative-integer-expected?","midpoint@","longer?","?set-nth","?first","rest-slice","prepend-as","prepend","fourth","sift","subseq-start","new-sequence","?last","like","first4","1sequence","reverse","slice","virtual@","repetition?","set-last","index","4sequence","max-length","set-second","immutable-sequence","first2","first3","supremum","unclip-slice","suffix!","insert-nth","tail","3append","short","suffix","concat","flip","immutable?","reverse!","2sequence","sum","delete-all","indices","snip-slice","<iota>","check-slice","sequence?","head","append-as","halves","sequence=","collapse-slice","?second","slice-error?","product","bounds-check?","bounds-check","immutable","virtual-exemplar","harvest","remove","pad-head","last","set-fourth","cartesian-product","remove-eq","shorten","shorter","reversed?","shorter?","shortest","head-slice","pop*","tail-slice*","but-last-slice","iota?","append!","cut-slice","new-resizable","head-slice*","sequence-hashcode","pop","set-nth","?nth","second","join","immutable-sequence?","<reversed>","3append-as","virtual-sequence","subseq?","remove-nth!","length","last-index","lengthen","assert-sequence","copy","move","third","first","tail?","set-first","prefix","bounds-error","<repetition>","exchange","surround","cut","min-length","set-third","push-all","head?","subseq-start-from","delete-slice","rest","sum-lengths","head*","infimum","remove!","glue","slice-error","subseq","push","replace-slice","subseq-as","unclip-last"],"math-builtin":["number=","next-power-of-2","?1+","fp-special?","imaginary-part","float>bits","number?","fp-infinity?","bignum?","fp-snan?","denominator","gcd","*","+","fp-bitwise=","-","u>=","/",">=","bitand","power-of-2?","log2-expects-positive","neg?","<","log2",">","integer?","number","bits>double","2/","zero?","bits>float","float?","shift","ratio?","rect>","even?","ratio","fp-sign","bitnot",">fixnum","complex?","/i","integer>fixnum","/f","sgn",">bignum","next-float","u<","u>","mod","recip","rational",">float","2^","integer","fixnum?","neg","fixnum","sq","bignum",">rect","bit?","fp-qnan?","simple-gcd","complex","<fp-nan>","real",">fraction","double>bits","bitor","rem","fp-nan-payload","real-part","log2-expects-positive?","prev-float","align","unordered?","float","fp-nan?","abs","bitxor","integer>fixnum-strict","u<=","odd?","<=","/mod",">integer","real?","rational?","numerator"]};Object.keys(c).forEach(function(p){o[p].pattern=u(c[p])});var d=["2bi","while","2tri","bi*","4dip","both?","same?","tri@","curry","prepose","3bi","?if","tri*","2keep","3keep","curried","2keepd","when","2bi*","2tri*","4keep","bi@","keepdd","do","unless*","tri-curry","if*","loop","bi-curry*","when*","2bi@","2tri@","with","2with","either?","bi","until","3dip","3curry","tri-curry*","tri-curry@","bi-curry","keepd","compose","2dip","if","3tri","unless","tuple","keep","2curry","tri","most","while*","dip","composed","bi-curry@","find-last-from","trim-head-slice","map-as","each-from","none?","trim-tail","partition","if-empty","accumulate*","reject!","find-from","accumulate-as","collector-for-as","reject","map","map-sum","accumulate!","2each-from","follow","supremum-by","map!","unless-empty","collector","padding","reduce-index","replicate-as","infimum-by","trim-tail-slice","count","find-index","filter","accumulate*!","reject-as","map-integers","map-find","reduce","selector","interleave","2map","filter-as","binary-reduce","map-index-as","find","produce","filter!","replicate","cartesian-map","cartesian-each","find-index-from","map-find-last","3map-as","3map","find-last","selector-as","2map-as","2map-reduce","accumulate","each","each-index","accumulate*-as","when-empty","all?","collector-as","push-either","new-like","collector-for","2selector","push-if","2all?","map-reduce","3each","any?","trim-slice","2reduce","change-nth","produce-as","2each","trim","trim-head","cartesian-find","map-index","if-zero","each-integer","unless-zero","(find-integer)","when-zero","find-last-integer","(all-integers?)","times","(each-integer)","find-integer","all-integers?","unless-negative","if-positive","when-positive","when-negative","unless-positive","if-negative","case","2cleave","cond>quot","case>quot","3cleave","wrong-values","to-fixed-point","alist>quot","cond","cleave","call-effect","recursive-hashcode","spread","deep-spread>quot","2||","0||","n||","0&&","2&&","3||","1||","1&&","n&&","3&&","smart-unless*","keep-inputs","reduce-outputs","smart-when*","cleave>array","smart-with","smart-apply","smart-if","inputs/outputs","output>sequence-n","map-outputs","map-reduce-outputs","dropping","output>array","smart-map-reduce","smart-2map-reduce","output>array-n","nullary","input<sequence","append-outputs","drop-inputs","inputs","smart-2reduce","drop-outputs","smart-reduce","preserving","smart-when","outputs","append-outputs-as","smart-unless","smart-if*","sum-outputs","input<sequence-unsafe","output>sequence"];o.combinators.pattern=u(d),n.languages.factor=o})(t)}return _v}var Cv,XL;function eye(){if(XL)return Cv;XL=1,Cv=e,e.displayName="$false",e.aliases=[];function e(t){(function(n){n.languages.false={comment:{pattern:/\{[^}]*\}/},string:{pattern:/"[^"]*"/,greedy:!0},"character-code":{pattern:/'(?:[^\r]|\r\n?)/,alias:"number"},"assembler-code":{pattern:/\d+`/,alias:"important"},number:/\d+/,operator:/[-!#$%&'*+,./:;=>?@\\^_`|~ßø]/,punctuation:/\[|\]/,variable:/[a-z]/,"non-standard":{pattern:/[()<BDO®]/,alias:"bold"}}})(t)}return Cv}var Nv,ZL;function tye(){if(ZL)return Nv;ZL=1,Nv=e,e.displayName="firestoreSecurityRules",e.aliases=[];function e(t){t.languages["firestore-security-rules"]=t.languages.extend("clike",{comment:/\/\/.*/,keyword:/\b(?:allow|function|if|match|null|return|rules_version|service)\b/,operator:/&&|\|\||[<>!=]=?|[-+*/%]|\b(?:in|is)\b/}),delete t.languages["firestore-security-rules"]["class-name"],t.languages.insertBefore("firestore-security-rules","keyword",{path:{pattern:/(^|[\s(),])(?:\/(?:[\w\xA0-\uFFFF]+|\{[\w\xA0-\uFFFF]+(?:=\*\*)?\}|\$\([\w\xA0-\uFFFF.]+\)))+/,lookbehind:!0,greedy:!0,inside:{variable:{pattern:/\{[\w\xA0-\uFFFF]+(?:=\*\*)?\}|\$\([\w\xA0-\uFFFF.]+\)/,inside:{operator:/=/,keyword:/\*\*/,punctuation:/[.$(){}]/}},punctuation:/\//}},method:{pattern:/(\ballow\s+)[a-z]+(?:\s*,\s*[a-z]+)*(?=\s*[:;])/,lookbehind:!0,alias:"builtin",inside:{punctuation:/,/}}})}return Nv}var Ov,QL;function nye(){if(QL)return Ov;QL=1,Ov=e,e.displayName="flow",e.aliases=[];function e(t){(function(n){n.languages.flow=n.languages.extend("javascript",{}),n.languages.insertBefore("flow","keyword",{type:[{pattern:/\b(?:[Bb]oolean|Function|[Nn]umber|[Ss]tring|any|mixed|null|void)\b/,alias:"tag"}]}),n.languages.flow["function-variable"].pattern=/(?!\s)[_$a-z\xA0-\uFFFF](?:(?!\s)[$\w\xA0-\uFFFF])*(?=\s*=\s*(?:function\b|(?:\([^()]*\)(?:\s*:\s*\w+)?|(?!\s)[_$a-z\xA0-\uFFFF](?:(?!\s)[$\w\xA0-\uFFFF])*)\s*=>))/i,delete n.languages.flow.parameter,n.languages.insertBefore("flow","operator",{"flow-punctuation":{pattern:/\{\||\|\}/,alias:"punctuation"}}),Array.isArray(n.languages.flow.keyword)||(n.languages.flow.keyword=[n.languages.flow.keyword]),n.languages.flow.keyword.unshift({pattern:/(^|[^$]\b)(?:Class|declare|opaque|type)\b(?!\$)/,lookbehind:!0},{pattern:/(^|[^$]\B)\$(?:Diff|Enum|Exact|Keys|ObjMap|PropertyType|Record|Shape|Subtype|Supertype|await)\b(?!\$)/,lookbehind:!0})})(t)}return Ov}var Dv,JL;function rye(){if(JL)return Dv;JL=1,Dv=e,e.displayName="fortran",e.aliases=[];function e(t){t.languages.fortran={"quoted-number":{pattern:/[BOZ](['"])[A-F0-9]+\1/i,alias:"number"},string:{pattern:/(?:\b\w+_)?(['"])(?:\1\1|&(?:\r\n?|\n)(?:[ \t]*!.*(?:\r\n?|\n)|(?![ \t]*!))|(?!\1).)*(?:\1|&)/,inside:{comment:{pattern:/(&(?:\r\n?|\n)\s*)!.*/,lookbehind:!0}}},comment:{pattern:/!.*/,greedy:!0},boolean:/\.(?:FALSE|TRUE)\.(?:_\w+)?/i,number:/(?:\b\d+(?:\.\d*)?|\B\.\d+)(?:[ED][+-]?\d+)?(?:_\w+)?/i,keyword:[/\b(?:CHARACTER|COMPLEX|DOUBLE ?PRECISION|INTEGER|LOGICAL|REAL)\b/i,/\b(?:END ?)?(?:BLOCK ?DATA|DO|FILE|FORALL|FUNCTION|IF|INTERFACE|MODULE(?! PROCEDURE)|PROGRAM|SELECT|SUBROUTINE|TYPE|WHERE)\b/i,/\b(?:ALLOCATABLE|ALLOCATE|BACKSPACE|CALL|CASE|CLOSE|COMMON|CONTAINS|CONTINUE|CYCLE|DATA|DEALLOCATE|DIMENSION|DO|END|EQUIVALENCE|EXIT|EXTERNAL|FORMAT|GO ?TO|IMPLICIT(?: NONE)?|INQUIRE|INTENT|INTRINSIC|MODULE PROCEDURE|NAMELIST|NULLIFY|OPEN|OPTIONAL|PARAMETER|POINTER|PRINT|PRIVATE|PUBLIC|READ|RETURN|REWIND|SAVE|SELECT|STOP|TARGET|WHILE|WRITE)\b/i,/\b(?:ASSIGNMENT|DEFAULT|ELEMENTAL|ELSE|ELSEIF|ELSEWHERE|ENTRY|IN|INCLUDE|INOUT|KIND|NULL|ONLY|OPERATOR|OUT|PURE|RECURSIVE|RESULT|SEQUENCE|STAT|THEN|USE)\b/i],operator:[/\*\*|\/\/|=>|[=\/]=|[<>]=?|::|[+\-*=%]|\.[A-Z]+\./i,{pattern:/(^|(?!\().)\/(?!\))/,lookbehind:!0}],punctuation:/\(\/|\/\)|[(),;:&]/}}return Dv}var Iv,e2;function aye(){if(e2)return Iv;e2=1,Iv=e,e.displayName="fsharp",e.aliases=[];function e(t){t.languages.fsharp=t.languages.extend("clike",{comment:[{pattern:/(^|[^\\])\(\*(?!\))[\s\S]*?\*\)/,lookbehind:!0,greedy:!0},{pattern:/(^|[^\\:])\/\/.*/,lookbehind:!0,greedy:!0}],string:{pattern:/(?:"""[\s\S]*?"""|@"(?:""|[^"])*"|"(?:\\[\s\S]|[^\\"])*")B?/,greedy:!0},"class-name":{pattern:/(\b(?:exception|inherit|interface|new|of|type)\s+|\w\s*:\s*|\s:\??>\s*)[.\w]+\b(?:\s*(?:->|\*)\s*[.\w]+\b)*(?!\s*[:.])/,lookbehind:!0,inside:{operator:/->|\*/,punctuation:/\./}},keyword:/\b(?:let|return|use|yield)(?:!\B|\b)|\b(?:abstract|and|as|asr|assert|atomic|base|begin|break|checked|class|component|const|constraint|constructor|continue|default|delegate|do|done|downcast|downto|eager|elif|else|end|event|exception|extern|external|false|finally|fixed|for|fun|function|functor|global|if|in|include|inherit|inline|interface|internal|land|lazy|lor|lsl|lsr|lxor|match|member|method|mixin|mod|module|mutable|namespace|new|not|null|object|of|open|or|override|parallel|private|process|protected|public|pure|rec|sealed|select|sig|static|struct|tailcall|then|to|trait|true|try|type|upcast|val|virtual|void|volatile|when|while|with)\b/,number:[/\b0x[\da-fA-F]+(?:LF|lf|un)?\b/,/\b0b[01]+(?:uy|y)?\b/,/(?:\b\d+(?:\.\d*)?|\B\.\d+)(?:[fm]|e[+-]?\d+)?\b/i,/\b\d+(?:[IlLsy]|UL|u[lsy]?)?\b/],operator:/([<>~&^])\1\1|([*.:<>&])\2|<-|->|[!=:]=|<?\|{1,3}>?|\??(?:<=|>=|<>|[-+*/%=<>])\??|[!?^&]|~[+~-]|:>|:\?>?/}),t.languages.insertBefore("fsharp","keyword",{preprocessor:{pattern:/(^[\t ]*)#.*/m,lookbehind:!0,alias:"property",inside:{directive:{pattern:/(^#)\b(?:else|endif|if|light|line|nowarn)\b/,lookbehind:!0,alias:"keyword"}}}}),t.languages.insertBefore("fsharp","punctuation",{"computation-expression":{pattern:/\b[_a-z]\w*(?=\s*\{)/i,alias:"keyword"}}),t.languages.insertBefore("fsharp","string",{annotation:{pattern:/\[<.+?>\]/,greedy:!0,inside:{punctuation:/^\[<|>\]$/,"class-name":{pattern:/^\w+$|(^|;\s*)[A-Z]\w*(?=\()/,lookbehind:!0},"annotation-content":{pattern:/[\s\S]+/,inside:t.languages.fsharp}}},char:{pattern:/'(?:[^\\']|\\(?:.|\d{3}|x[a-fA-F\d]{2}|u[a-fA-F\d]{4}|U[a-fA-F\d]{8}))'B?/,greedy:!0}})}return Iv}var Lv,t2;function oye(){if(t2)return Lv;t2=1;var e=Wn();Lv=t,t.displayName="ftl",t.aliases=[];function t(n){n.register(e),function(r){for(var a=/[^<()"']|\((?:<expr>)*\)|<(?!#--)|<#--(?:[^-]|-(?!->))*-->|"(?:[^\\"]|\\.)*"|'(?:[^\\']|\\.)*'/.source,o=0;o<2;o++)a=a.replace(/<expr>/g,function(){return a});a=a.replace(/<expr>/g,/[^\s\S]/.source);var s={comment:/<#--[\s\S]*?-->/,string:[{pattern:/\br("|')(?:(?!\1)[^\\]|\\.)*\1/,greedy:!0},{pattern:RegExp(/("|')(?:(?!\1|\$\{)[^\\]|\\.|\$\{(?:(?!\})(?:<expr>))*\})*\1/.source.replace(/<expr>/g,function(){return a})),greedy:!0,inside:{interpolation:{pattern:RegExp(/((?:^|[^\\])(?:\\\\)*)\$\{(?:(?!\})(?:<expr>))*\}/.source.replace(/<expr>/g,function(){return a})),lookbehind:!0,inside:{"interpolation-punctuation":{pattern:/^\$\{|\}$/,alias:"punctuation"},rest:null}}}}],keyword:/\b(?:as)\b/,boolean:/\b(?:false|true)\b/,"builtin-function":{pattern:/((?:^|[^?])\?\s*)\w+/,lookbehind:!0,alias:"function"},function:/\b\w+(?=\s*\()/,number:/\b\d+(?:\.\d+)?\b/,operator:/\.\.[<*!]?|->|--|\+\+|&&|\|\||\?{1,2}|[-+*/%!=<>]=?|\b(?:gt|gte|lt|lte)\b/,punctuation:/[,;.:()[\]{}]/};s.string[1].inside.interpolation.inside.rest=s,r.languages.ftl={"ftl-comment":{pattern:/^<#--[\s\S]*/,alias:"comment"},"ftl-directive":{pattern:/^<[\s\S]+>$/,inside:{directive:{pattern:/(^<\/?)[#@][a-z]\w*/i,lookbehind:!0,alias:"keyword"},punctuation:/^<\/?|\/?>$/,content:{pattern:/\s*\S[\s\S]*/,alias:"ftl",inside:s}}},"ftl-interpolation":{pattern:/^\$\{[\s\S]*\}$/,inside:{punctuation:/^\$\{|\}$/,content:{pattern:/\s*\S[\s\S]*/,alias:"ftl",inside:s}}}},r.hooks.add("before-tokenize",function(u){var c=RegExp(/<#--[\s\S]*?-->|<\/?[#@][a-zA-Z](?:<expr>)*?>|\$\{(?:<expr>)*?\}/.source.replace(/<expr>/g,function(){return a}),"gi");r.languages["markup-templating"].buildPlaceholders(u,"ftl",c)}),r.hooks.add("after-tokenize",function(u){r.languages["markup-templating"].tokenizePlaceholders(u,"ftl")})}(n)}return Lv}var Mv,n2;function iye(){if(n2)return Mv;n2=1,Mv=e,e.displayName="gap",e.aliases=[];function e(t){t.languages.gap={shell:{pattern:/^gap>[\s\S]*?(?=^gap>|$(?![\s\S]))/m,greedy:!0,inside:{gap:{pattern:/^(gap>).+(?:(?:\r(?:\n|(?!\n))|\n)>.*)*/,lookbehind:!0,inside:null},punctuation:/^gap>/}},comment:{pattern:/#.*/,greedy:!0},string:{pattern:/(^|[^\\'"])(?:'(?:[^\r\n\\']|\\.){1,10}'|"(?:[^\r\n\\"]|\\.)*"(?!")|"""[\s\S]*?""")/,lookbehind:!0,greedy:!0,inside:{continuation:{pattern:/([\r\n])>/,lookbehind:!0,alias:"punctuation"}}},keyword:/\b(?:Assert|Info|IsBound|QUIT|TryNextMethod|Unbind|and|atomic|break|continue|do|elif|else|end|fi|for|function|if|in|local|mod|not|od|or|quit|readonly|readwrite|rec|repeat|return|then|until|while)\b/,boolean:/\b(?:false|true)\b/,function:/\b[a-z_]\w*(?=\s*\()/i,number:{pattern:/(^|[^\w.]|\.\.)(?:\d+(?:\.\d*)?|\.\d+)(?:[eE][+-]?\d+)?(?:_[a-z]?)?(?=$|[^\w.]|\.\.)/,lookbehind:!0},continuation:{pattern:/([\r\n])>/,lookbehind:!0,alias:"punctuation"},operator:/->|[-+*/^~=!]|<>|[<>]=?|:=|\.\./,punctuation:/[()[\]{},;.:]/},t.languages.gap.shell.inside.gap.inside=t.languages.gap}return Mv}var Fv,r2;function sye(){if(r2)return Fv;r2=1,Fv=e,e.displayName="gcode",e.aliases=[];function e(t){t.languages.gcode={comment:/;.*|\B\(.*?\)\B/,string:{pattern:/"(?:""|[^"])*"/,greedy:!0},keyword:/\b[GM]\d+(?:\.\d+)?\b/,property:/\b[A-Z]/,checksum:{pattern:/(\*)\d+/,lookbehind:!0,alias:"number"},punctuation:/[:*]/}}return Fv}var Pv,a2;function lye(){if(a2)return Pv;a2=1,Pv=e,e.displayName="gdscript",e.aliases=[];function e(t){t.languages.gdscript={comment:/#.*/,string:{pattern:/@?(?:("|')(?:(?!\1)[^\n\\]|\\[\s\S])*\1(?!"|')|"""(?:[^\\]|\\[\s\S])*?""")/,greedy:!0},"class-name":{pattern:/(^(?:class|class_name|extends)[ \t]+|^export\([ \t]*|\bas[ \t]+|(?:\b(?:const|var)[ \t]|[,(])[ \t]*\w+[ \t]*:[ \t]*|->[ \t]*)[a-zA-Z_]\w*/m,lookbehind:!0},keyword:/\b(?:and|as|assert|break|breakpoint|class|class_name|const|continue|elif|else|enum|export|extends|for|func|if|in|is|master|mastersync|match|not|null|onready|or|pass|preload|puppet|puppetsync|remote|remotesync|return|self|setget|signal|static|tool|var|while|yield)\b/,function:/\b[a-z_]\w*(?=[ \t]*\()/i,variable:/\$\w+/,number:[/\b0b[01_]+\b|\b0x[\da-fA-F_]+\b|(?:\b\d[\d_]*(?:\.[\d_]*)?|\B\.[\d_]+)(?:e[+-]?[\d_]+)?\b/,/\b(?:INF|NAN|PI|TAU)\b/],constant:/\b[A-Z][A-Z_\d]*\b/,boolean:/\b(?:false|true)\b/,operator:/->|:=|&&|\|\||<<|>>|[-+*/%&|!<>=]=?|[~^]/,punctuation:/[.:,;()[\]{}]/}}return Pv}var zv,o2;function uye(){if(o2)return zv;o2=1,zv=e,e.displayName="gedcom",e.aliases=[];function e(t){t.languages.gedcom={"line-value":{pattern:/(^[\t ]*\d+ +(?:@\w[\w!"$%&'()*+,\-./:;<=>?[\\\]^`{|}~\x80-\xfe #]*@ +)?\w+ ).+/m,lookbehind:!0,inside:{pointer:{pattern:/^@\w[\w!"$%&'()*+,\-./:;<=>?[\\\]^`{|}~\x80-\xfe #]*@$/,alias:"variable"}}},tag:{pattern:/(^[\t ]*\d+ +(?:@\w[\w!"$%&'()*+,\-./:;<=>?[\\\]^`{|}~\x80-\xfe #]*@ +)?)\w+/m,lookbehind:!0,alias:"string"},level:{pattern:/(^[\t ]*)\d+/m,lookbehind:!0,alias:"number"},pointer:{pattern:/@\w[\w!"$%&'()*+,\-./:;<=>?[\\\]^`{|}~\x80-\xfe #]*@/,alias:"variable"}}}return zv}var Bv,i2;function cye(){if(i2)return Bv;i2=1,Bv=e,e.displayName="gherkin",e.aliases=[];function e(t){(function(n){var r=/(?:\r?\n|\r)[ \t]*\|.+\|(?:(?!\|).)*/.source;n.languages.gherkin={pystring:{pattern:/("""|''')[\s\S]+?\1/,alias:"string"},comment:{pattern:/(^[ \t]*)#.*/m,lookbehind:!0},tag:{pattern:/(^[ \t]*)@\S*/m,lookbehind:!0},feature:{pattern:/((?:^|\r?\n|\r)[ \t]*)(?:Ability|Ahoy matey!|Arwedd|Aspekt|Besigheid Behoefte|Business Need|Caracteristica|Característica|Egenskab|Egenskap|Eiginleiki|Feature|Fīča|Fitur|Fonctionnalité|Fonksyonalite|Funcionalidade|Funcionalitat|Functionalitate|Funcţionalitate|Funcționalitate|Functionaliteit|Fungsi|Funkcia|Funkcija|Funkcionalitāte|Funkcionalnost|Funkcja|Funksie|Funktionalität|Funktionalitéit|Funzionalità|Hwaet|Hwæt|Jellemző|Karakteristik|Lastnost|Mak|Mogucnost|laH|Mogućnost|Moznosti|Možnosti|OH HAI|Omadus|Ominaisuus|Osobina|Özellik|Potrzeba biznesowa|perbogh|poQbogh malja'|Požadavek|Požiadavka|Pretty much|Qap|Qu'meH 'ut|Savybė|Tính năng|Trajto|Vermoë|Vlastnosť|Właściwość|Značilnost|Δυνατότητα|Λειτουργία|Могућност|Мөмкинлек|Особина|Свойство|Үзенчәлеклелек|Функционал|Функционалност|Функция|Функціонал|תכונה|خاصية|خصوصیت|صلاحیت|کاروبار کی ضرورت|وِیژگی|रूप लेख|ਖਾਸੀਅਤ|ਨਕਸ਼ ਨੁਹਾਰ|ਮੁਹਾਂਦਰਾ|గుణము|ಹೆಚ್ಚಳ|ความต้องการทางธุรกิจ|ความสามารถ|โครงหลัก|기능|フィーチャ|功能|機能):(?:[^:\r\n]+(?:\r?\n|\r|$))*/,lookbehind:!0,inside:{important:{pattern:/(:)[^\r\n]+/,lookbehind:!0},keyword:/[^:\r\n]+:/}},scenario:{pattern:/(^[ \t]*)(?:Abstract Scenario|Abstrakt Scenario|Achtergrond|Aer|Ær|Agtergrond|All y'all|Antecedentes|Antecedents|Atburðarás|Atburðarásir|Awww, look mate|B4|Background|Baggrund|Bakgrund|Bakgrunn|Bakgrunnur|Beispiele|Beispiller|Bối cảnh|Cefndir|Cenario|Cenário|Cenario de Fundo|Cenário de Fundo|Cenarios|Cenários|Contesto|Context|Contexte|Contexto|Conto|Contoh|Contone|Dæmi|Dasar|Dead men tell no tales|Delineacao do Cenario|Delineação do Cenário|Dis is what went down|Dữ liệu|Dyagram Senaryo|Dyagram senaryo|Egzanp|Ejemplos|Eksempler|Ekzemploj|Enghreifftiau|Esbozo do escenario|Escenari|Escenario|Esempi|Esquema de l'escenari|Esquema del escenario|Esquema do Cenario|Esquema do Cenário|EXAMPLZ|Examples|Exempel|Exemple|Exemples|Exemplos|First off|Fono|Forgatókönyv|Forgatókönyv vázlat|Fundo|Geçmiş|Grundlage|Hannergrond|ghantoH|Háttér|Heave to|Istorik|Juhtumid|Keadaan|Khung kịch bản|Khung tình huống|Kịch bản|Koncept|Konsep skenario|Kontèks|Kontekst|Kontekstas|Konteksts|Kontext|Konturo de la scenaro|Latar Belakang|lut chovnatlh|lut|lutmey|Lýsing Atburðarásar|Lýsing Dæma|MISHUN SRSLY|MISHUN|Menggariskan Senario|mo'|Náčrt Scenára|Náčrt Scénáře|Náčrt Scenáru|Oris scenarija|Örnekler|Osnova|Osnova Scenára|Osnova scénáře|Osnutek|Ozadje|Paraugs|Pavyzdžiai|Példák|Piemēri|Plan du scénario|Plan du Scénario|Plan Senaryo|Plan senaryo|Plang vum Szenario|Pozadí|Pozadie|Pozadina|Príklady|Příklady|Primer|Primeri|Primjeri|Przykłady|Raamstsenaarium|Reckon it's like|Rerefons|Scenár|Scénář|Scenarie|Scenarij|Scenarijai|Scenarijaus šablonas|Scenariji|Scenārijs|Scenārijs pēc parauga|Scenarijus|Scenario|Scénario|Scenario Amlinellol|Scenario Outline|Scenario Template|Scenariomal|Scenariomall|Scenarios|Scenariu|Scenariusz|Scenaro|Schema dello scenario|Se ðe|Se the|Se þe|Senario|Senaryo Deskripsyon|Senaryo deskripsyon|Senaryo|Senaryo taslağı|Shiver me timbers|Situācija|Situai|Situasie Uiteensetting|Situasie|Skenario konsep|Skenario|Skica|Structura scenariu|Structură scenariu|Struktura scenarija|Stsenaarium|Swa hwaer swa|Swa|Swa hwær swa|Szablon scenariusza|Szenario|Szenariogrundriss|Tapaukset|Tapaus|Tapausaihio|Taust|Tausta|Template Keadaan|Template Senario|Template Situai|The thing of it is|Tình huống|Variantai|Voorbeelde|Voorbeelden|Wharrimean is|Yo-ho-ho|You'll wanna|Założenia|Παραδείγματα|Περιγραφή Σεναρίου|Σενάρια|Σενάριο|Υπόβαθρο|Кереш|Контекст|Концепт|Мисаллар|Мисоллар|Основа|Передумова|Позадина|Предистория|Предыстория|Приклади|Пример|Примери|Примеры|Рамка на сценарий|Скица|Структура сценарија|Структура сценария|Структура сценарію|Сценарий|Сценарий структураси|Сценарийның төзелеше|Сценарији|Сценарио|Сценарій|Тарих|Үрнәкләр|דוגמאות|רקע|תבנית תרחיש|תרחיש|الخلفية|الگوی سناریو|امثلة|پس منظر|زمینه|سناریو|سيناريو|سيناريو مخطط|مثالیں|منظر نامے کا خاکہ|منظرنامہ|نمونه ها|उदाहरण|परिदृश्य|परिदृश्य रूपरेखा|पृष्ठभूमि|ਉਦਾਹਰਨਾਂ|ਪਟਕਥਾ|ਪਟਕਥਾ ਢਾਂਚਾ|ਪਟਕਥਾ ਰੂਪ ਰੇਖਾ|ਪਿਛੋਕੜ|ఉదాహరణలు|కథనం|నేపథ్యం|సన్నివేశం|ಉದಾಹರಣೆಗಳು|ಕಥಾಸಾರಾಂಶ|ವಿವರಣೆ|ಹಿನ್ನೆಲೆ|โครงสร้างของเหตุการณ์|ชุดของตัวอย่าง|ชุดของเหตุการณ์|แนวคิด|สรุปเหตุการณ์|เหตุการณ์|배경|시나리오|시나리오 개요|예|サンプル|シナリオ|シナリオアウトライン|シナリオテンプレ|シナリオテンプレート|テンプレ|例|例子|剧本|剧本大纲|劇本|劇本大綱|场景|场景大纲|場景|場景大綱|背景):[^:\r\n]*/m,lookbehind:!0,inside:{important:{pattern:/(:)[^\r\n]*/,lookbehind:!0},keyword:/[^:\r\n]+:/}},"table-body":{pattern:RegExp("("+r+")(?:"+r+")+"),lookbehind:!0,inside:{outline:{pattern:/<[^>]+>/,alias:"variable"},td:{pattern:/\s*[^\s|][^|]*/,alias:"string"},punctuation:/\|/}},"table-head":{pattern:RegExp(r),inside:{th:{pattern:/\s*[^\s|][^|]*/,alias:"variable"},punctuation:/\|/}},atrule:{pattern:/(^[ \t]+)(?:'a|'ach|'ej|7|a|A také|A taktiež|A tiež|A zároveň|Aber|Ac|Adott|Akkor|Ak|Aleshores|Ale|Ali|Allora|Alors|Als|Ama|Amennyiben|Amikor|Ampak|an|AN|Ananging|And y'all|And|Angenommen|Anrhegedig a|An|Apabila|Atès|Atesa|Atunci|Avast!|Aye|A|awer|Bagi|Banjur|Bet|Biết|Blimey!|Buh|But at the end of the day I reckon|But y'all|But|BUT|Cal|Când|Cand|Cando|Ce|Cuando|Če|Ða ðe|Ða|Dadas|Dada|Dados|Dado|DaH ghu' bejlu'|dann|Dann|Dano|Dan|Dar|Dat fiind|Data|Date fiind|Date|Dati fiind|Dati|Daţi fiind|Dați fiind|DEN|Dato|De|Den youse gotta|Dengan|Diberi|Diyelim ki|Donada|Donat|Donitaĵo|Do|Dun|Duota|Ðurh|Eeldades|Ef|Eğer ki|Entao|Então|Entón|E|En|Entonces|Epi|És|Etant donnée|Etant donné|Et|Étant données|Étant donnée|Étant donné|Etant données|Etant donnés|Étant donnés|Fakat|Gangway!|Gdy|Gegeben seien|Gegeben sei|Gegeven|Gegewe|ghu' noblu'|Gitt|Given y'all|Given|Givet|Givun|Ha|Cho|I CAN HAZ|In|Ir|It's just unbelievable|I|Ja|Jeśli|Jeżeli|Kad|Kada|Kadar|Kai|Kaj|Když|Keď|Kemudian|Ketika|Khi|Kiedy|Ko|Kuid|Kui|Kun|Lan|latlh|Le sa a|Let go and haul|Le|Lè sa a|Lè|Logo|Lorsqu'<|Lorsque|mä|Maar|Mais|Mając|Ma|Majd|Maka|Manawa|Mas|Men|Menawa|Mutta|Nalika|Nalikaning|Nanging|Når|När|Nato|Nhưng|Niin|Njuk|O zaman|Och|Og|Oletetaan|Ond|Onda|Oraz|Pak|Pero|Però|Podano|Pokiaľ|Pokud|Potem|Potom|Privzeto|Pryd|Quan|Quand|Quando|qaSDI'|Så|Sed|Se|Siis|Sipoze ke|Sipoze Ke|Sipoze|Si|Şi|Și|Soit|Stel|Tada|Tad|Takrat|Tak|Tapi|Ter|Tetapi|Tha the|Tha|Then y'all|Then|Thì|Thurh|Toda|Too right|Un|Und|ugeholl|Và|vaj|Vendar|Ve|wann|Wanneer|WEN|Wenn|When y'all|When|Wtedy|Wun|Y'know|Yeah nah|Yna|Youse know like when|Youse know when youse got|Y|Za predpokladu|Za předpokladu|Zadan|Zadani|Zadano|Zadate|Zadato|Zakładając|Zaradi|Zatati|Þa þe|Þa|Þá|Þegar|Þurh|Αλλά|Δεδομένου|Και|Όταν|Τότε|А також|Агар|Але|Али|Аммо|А|Әгәр|Әйтик|Әмма|Бирок|Ва|Вә|Дадено|Дано|Допустим|Если|Задате|Задати|Задато|И|І|К тому же|Када|Кад|Когато|Когда|Коли|Ләкин|Лекин|Нәтиҗәдә|Нехай|Но|Онда|Припустимо, що|Припустимо|Пусть|Также|Та|Тогда|Тоді|То|Унда|Һәм|Якщо|אבל|אזי|אז|בהינתן|וגם|כאשר|آنگاه|اذاً|اگر|اما|اور|با فرض|بالفرض|بفرض|پھر|تب|ثم|جب|عندما|فرض کیا|لكن|لیکن|متى|هنگامی|و|अगर|और|कदा|किन्तु|चूंकि|जब|तथा|तदा|तब|परन्तु|पर|यदि|ਅਤੇ|ਜਦੋਂ|ਜਿਵੇਂ ਕਿ|ਜੇਕਰ|ਤਦ|ਪਰ|అప్పుడు|ఈ పరిస్థితిలో|కాని|చెప్పబడినది|మరియు|ಆದರೆ|ನಂತರ|ನೀಡಿದ|ಮತ್ತು|ಸ್ಥಿತಿಯನ್ನು|กำหนดให้|ดังนั้น|แต่|เมื่อ|และ|그러면<|그리고<|단<|만약<|만일<|먼저<|조건<|하지만<|かつ<|しかし<|ただし<|ならば<|もし<|並且<|但し<|但是<|假如<|假定<|假設<|假设<|前提<|同时<|同時<|并且<|当<|當<|而且<|那么<|那麼<)(?=[ \t])/m,lookbehind:!0},string:{pattern:/"(?:\\.|[^"\\\r\n])*"|'(?:\\.|[^'\\\r\n])*'/,inside:{outline:{pattern:/<[^>]+>/,alias:"variable"}}},outline:{pattern:/<[^>]+>/,alias:"variable"}}})(t)}return Bv}var Uv,s2;function dye(){if(s2)return Uv;s2=1,Uv=e,e.displayName="git",e.aliases=[];function e(t){t.languages.git={comment:/^#.*/m,deleted:/^[-–].*/m,inserted:/^\+.*/m,string:/("|')(?:\\.|(?!\1)[^\\\r\n])*\1/,command:{pattern:/^.*\$ git .*$/m,inside:{parameter:/\s--?\w+/}},coord:/^@@.*@@$/m,"commit-sha1":/^commit \w{40}$/m}}return Uv}var jv,l2;function fye(){if(l2)return jv;l2=1;var e=Ai();jv=t,t.displayName="glsl",t.aliases=[];function t(n){n.register(e),n.languages.glsl=n.languages.extend("c",{keyword:/\b(?:active|asm|atomic_uint|attribute|[ibdu]?vec[234]|bool|break|buffer|case|cast|centroid|class|coherent|common|const|continue|d?mat[234](?:x[234])?|default|discard|do|double|else|enum|extern|external|false|filter|fixed|flat|float|for|fvec[234]|goto|half|highp|hvec[234]|[iu]?sampler2DMS(?:Array)?|[iu]?sampler2DRect|[iu]?samplerBuffer|[iu]?samplerCube|[iu]?samplerCubeArray|[iu]?sampler[123]D|[iu]?sampler[12]DArray|[iu]?image2DMS(?:Array)?|[iu]?image2DRect|[iu]?imageBuffer|[iu]?imageCube|[iu]?imageCubeArray|[iu]?image[123]D|[iu]?image[12]DArray|if|in|inline|inout|input|int|interface|invariant|layout|long|lowp|mediump|namespace|noinline|noperspective|out|output|partition|patch|precise|precision|public|readonly|resource|restrict|return|sample|sampler[12]DArrayShadow|sampler[12]DShadow|sampler2DRectShadow|sampler3DRect|samplerCubeArrayShadow|samplerCubeShadow|shared|short|sizeof|smooth|static|struct|subroutine|superp|switch|template|this|true|typedef|uint|uniform|union|unsigned|using|varying|void|volatile|while|writeonly)\b/})}return jv}var Gv,u2;function pye(){if(u2)return Gv;u2=1,Gv=e,e.displayName="gml",e.aliases=[];function e(t){t.languages.gamemakerlanguage=t.languages.gml=t.languages.extend("clike",{keyword:/\b(?:break|case|continue|default|do|else|enum|exit|for|globalvar|if|repeat|return|switch|until|var|while)\b/,number:/(?:\b0x[\da-f]+|(?:\b\d+(?:\.\d*)?|\B\.\d+)(?:e[+-]?\d+)?)[ulf]{0,4}/i,operator:/--|\+\+|[-+%/=]=?|!=|\*\*?=?|<[<=>]?|>[=>]?|&&?|\^\^?|\|\|?|~|\b(?:and|at|not|or|with|xor)\b/,constant:/\b(?:GM_build_date|GM_version|action_(?:continue|restart|reverse|stop)|all|gamespeed_(?:fps|microseconds)|global|local|noone|other|pi|pointer_(?:invalid|null)|self|timezone_(?:local|utc)|undefined|ev_(?:create|destroy|step|alarm|keyboard|mouse|collision|other|draw|draw_(?:begin|end|post|pre)|keypress|keyrelease|trigger|(?:left|middle|no|right)_button|(?:left|middle|right)_press|(?:left|middle|right)_release|mouse_(?:enter|leave|wheel_down|wheel_up)|global_(?:left|middle|right)_button|global_(?:left|middle|right)_press|global_(?:left|middle|right)_release|joystick(?:1|2)_(?:button1|button2|button3|button4|button5|button6|button7|button8|down|left|right|up)|outside|boundary|game_start|game_end|room_start|room_end|no_more_lives|animation_end|end_of_path|no_more_health|user\d|gui|gui_begin|gui_end|step_(?:begin|end|normal))|vk_(?:alt|anykey|backspace|control|delete|down|end|enter|escape|home|insert|left|nokey|pagedown|pageup|pause|printscreen|return|right|shift|space|tab|up|f\d|numpad\d|add|decimal|divide|lalt|lcontrol|lshift|multiply|ralt|rcontrol|rshift|subtract)|achievement_(?:filter_(?:all_players|favorites_only|friends_only)|friends_info|info|leaderboard_info|our_info|pic_loaded|show_(?:achievement|bank|friend_picker|leaderboard|profile|purchase_prompt|ui)|type_challenge|type_score_challenge)|asset_(?:font|object|path|room|script|shader|sound|sprite|tiles|timeline|unknown)|audio_(?:3d|falloff_(?:exponent_distance|exponent_distance_clamped|inverse_distance|inverse_distance_clamped|linear_distance|linear_distance_clamped|none)|mono|new_system|old_system|stereo)|bm_(?:add|complex|dest_alpha|dest_color|dest_colour|inv_dest_alpha|inv_dest_color|inv_dest_colour|inv_src_alpha|inv_src_color|inv_src_colour|max|normal|one|src_alpha|src_alpha_sat|src_color|src_colour|subtract|zero)|browser_(?:chrome|firefox|ie|ie_mobile|not_a_browser|opera|safari|safari_mobile|tizen|unknown|windows_store)|buffer_(?:bool|f16|f32|f64|fast|fixed|generalerror|grow|invalidtype|network|outofbounds|outofspace|s16|s32|s8|seek_end|seek_relative|seek_start|string|text|u16|u32|u64|u8|vbuffer|wrap)|c_(?:aqua|black|blue|dkgray|fuchsia|gray|green|lime|ltgray|maroon|navy|olive|orange|purple|red|silver|teal|white|yellow)|cmpfunc_(?:always|equal|greater|greaterequal|less|lessequal|never|notequal)|cr_(?:appstart|arrow|beam|cross|default|drag|handpoint|hourglass|none|size_all|size_nesw|size_ns|size_nwse|size_we|uparrow)|cull_(?:clockwise|counterclockwise|noculling)|device_(?:emulator|tablet)|device_ios_(?:ipad|ipad_retina|iphone|iphone5|iphone6|iphone6plus|iphone_retina|unknown)|display_(?:landscape|landscape_flipped|portrait|portrait_flipped)|dll_(?:cdecl|cdel|stdcall)|ds_type_(?:grid|list|map|priority|queue|stack)|ef_(?:cloud|ellipse|explosion|firework|flare|rain|ring|smoke|smokeup|snow|spark|star)|fa_(?:archive|bottom|center|directory|hidden|left|middle|readonly|right|sysfile|top|volumeid)|fb_login_(?:default|fallback_to_webview|forcing_safari|forcing_webview|no_fallback_to_webview|use_system_account)|iap_(?:available|canceled|ev_consume|ev_product|ev_purchase|ev_restore|ev_storeload|failed|purchased|refunded|status_available|status_loading|status_processing|status_restoring|status_unavailable|status_uninitialised|storeload_failed|storeload_ok|unavailable)|leaderboard_type_(?:number|time_mins_secs)|lighttype_(?:dir|point)|matrix_(?:projection|view|world)|mb_(?:any|left|middle|none|right)|network_(?:config_(?:connect_timeout|disable_reliable_udp|enable_reliable_udp|use_non_blocking_socket)|socket_(?:bluetooth|tcp|udp)|type_(?:connect|data|disconnect|non_blocking_connect))|of_challenge_(?:lose|tie|win)|os_(?:android|ios|linux|macosx|ps3|ps4|psvita|unknown|uwp|win32|win8native|windows|winphone|xboxone)|phy_debug_render_(?:aabb|collision_pairs|coms|core_shapes|joints|obb|shapes)|phy_joint_(?:anchor_1_x|anchor_1_y|anchor_2_x|anchor_2_y|angle|angle_limits|damping_ratio|frequency|length_1|length_2|lower_angle_limit|max_force|max_length|max_motor_force|max_motor_torque|max_torque|motor_force|motor_speed|motor_torque|reaction_force_x|reaction_force_y|reaction_torque|speed|translation|upper_angle_limit)|phy_particle_data_flag_(?:category|color|colour|position|typeflags|velocity)|phy_particle_flag_(?:colormixing|colourmixing|elastic|powder|spring|tensile|viscous|wall|water|zombie)|phy_particle_group_flag_(?:rigid|solid)|pr_(?:linelist|linestrip|pointlist|trianglefan|trianglelist|trianglestrip)|ps_(?:distr|shape)_(?:diamond|ellipse|gaussian|invgaussian|line|linear|rectangle)|pt_shape_(?:circle|cloud|disk|explosion|flare|line|pixel|ring|smoke|snow|spark|sphere|square|star)|ty_(?:real|string)|gp_(?:face\d|axislh|axislv|axisrh|axisrv|padd|padl|padr|padu|select|shoulderl|shoulderlb|shoulderr|shoulderrb|start|stickl|stickr)|lb_disp_(?:none|numeric|time_ms|time_sec)|lb_sort_(?:ascending|descending|none)|ov_(?:achievements|community|friends|gamegroup|players|settings)|ugc_(?:filetype_(?:community|microtrans)|list_(?:Favorited|Followed|Published|Subscribed|UsedOrPlayed|VotedDown|VotedOn|VotedUp|WillVoteLater)|match_(?:AllGuides|Artwork|Collections|ControllerBindings|IntegratedGuides|Items|Items_Mtx|Items_ReadyToUse|Screenshots|UsableInGame|Videos|WebGuides)|query_(?:AcceptedForGameRankedByAcceptanceDate|CreatedByFriendsRankedByPublicationDate|FavoritedByFriendsRankedByPublicationDate|NotYetRated)|query_RankedBy(?:NumTimesReported|PublicationDate|TextSearch|TotalVotesAsc|Trend|Vote|VotesUp)|result_success|sortorder_CreationOrder(?:Asc|Desc)|sortorder_(?:ForModeration|LastUpdatedDesc|SubscriptionDateDesc|TitleAsc|VoteScoreDesc)|visibility_(?:friends_only|private|public))|vertex_usage_(?:binormal|blendindices|blendweight|color|colour|depth|fog|normal|position|psize|sample|tangent|texcoord|textcoord)|vertex_type_(?:float\d|color|colour|ubyte4)|input_type|layerelementtype_(?:background|instance|oldtilemap|particlesystem|sprite|tile|tilemap|undefined)|se_(?:chorus|compressor|echo|equalizer|flanger|gargle|none|reverb)|text_type|tile_(?:flip|index_mask|mirror|rotate)|(?:obj|rm|scr|spr)\w+)\b/,variable:/\b(?:alarm|application_surface|async_load|background_(?:alpha|blend|color|colour|foreground|height|hspeed|htiled|index|showcolor|showcolour|visible|vspeed|vtiled|width|x|xscale|y|yscale)|bbox_(?:bottom|left|right|top)|browser_(?:height|width)|caption_(?:health|lives|score)|current_(?:day|hour|minute|month|second|time|weekday|year)|cursor_sprite|debug_mode|delta_time|direction|display_aa|error_(?:last|occurred)|event_(?:action|number|object|type)|fps|fps_real|friction|game_(?:display|project|save)_(?:id|name)|gamemaker_(?:pro|registered|version)|gravity|gravity_direction|(?:h|v)speed|health|iap_data|id|image_(?:alpha|angle|blend|depth|index|number|speed|xscale|yscale)|instance_(?:count|id)|keyboard_(?:key|lastchar|lastkey|string)|layer|lives|mask_index|mouse_(?:button|lastbutton|x|y)|object_index|os_(?:browser|device|type|version)|path_(?:endaction|index|orientation|position|positionprevious|scale|speed)|persistent|phy_(?:rotation|(?:col_normal|collision|com|linear_velocity|position|speed)_(?:x|y)|angular_(?:damping|velocity)|position_(?:x|y)previous|speed|linear_damping|bullet|fixed_rotation|active|mass|inertia|dynamic|kinematic|sleeping|collision_points)|pointer_(?:invalid|null)|room|room_(?:caption|first|height|last|persistent|speed|width)|score|secure_mode|show_(?:health|lives|score)|solid|speed|sprite_(?:height|index|width|xoffset|yoffset)|temp_directory|timeline_(?:index|loop|position|running|speed)|transition_(?:color|kind|steps)|undefined|view_(?:angle|current|enabled|(?:h|v)(?:border|speed)|(?:h|w|x|y)port|(?:h|w|x|y)view|object|surface_id|visible)|visible|webgl_enabled|working_directory|(?:x|y)(?:previous|start)|x|y|argument(?:_relitive|_count|\d)|argument|global|local|other|self)\b/})}return Gv}var Hv,c2;function gye(){if(c2)return Hv;c2=1,Hv=e,e.displayName="gn",e.aliases=["gni"];function e(t){t.languages.gn={comment:{pattern:/#.*/,greedy:!0},"string-literal":{pattern:/(^|[^\\"])"(?:[^\r\n"\\]|\\.)*"/,lookbehind:!0,greedy:!0,inside:{interpolation:{pattern:/((?:^|[^\\])(?:\\{2})*)\$(?:\{[\s\S]*?\}|[a-zA-Z_]\w*|0x[a-fA-F0-9]{2})/,lookbehind:!0,inside:{number:/^\$0x[\s\S]{2}$/,variable:/^\$\w+$/,"interpolation-punctuation":{pattern:/^\$\{|\}$/,alias:"punctuation"},expression:{pattern:/[\s\S]+/,inside:null}}},string:/[\s\S]+/}},keyword:/\b(?:else|if)\b/,boolean:/\b(?:false|true)\b/,"builtin-function":{pattern:/\b(?:assert|defined|foreach|import|pool|print|template|tool|toolchain)(?=\s*\()/i,alias:"keyword"},function:/\b[a-z_]\w*(?=\s*\()/i,constant:/\b(?:current_cpu|current_os|current_toolchain|default_toolchain|host_cpu|host_os|root_build_dir|root_gen_dir|root_out_dir|target_cpu|target_gen_dir|target_os|target_out_dir)\b/,number:/-?\b\d+\b/,operator:/[-+!=<>]=?|&&|\|\|/,punctuation:/[(){}[\],.]/},t.languages.gn["string-literal"].inside.interpolation.inside.expression.inside=t.languages.gn,t.languages.gni=t.languages.gn}return Hv}var $v,d2;function hye(){if(d2)return $v;d2=1,$v=e,e.displayName="goModule",e.aliases=[];function e(t){t.languages["go-mod"]=t.languages["go-module"]={comment:{pattern:/\/\/.*/,greedy:!0},version:{pattern:/(^|[\s()[\],])v\d+\.\d+\.\d+(?:[+-][-+.\w]*)?(?![^\s()[\],])/,lookbehind:!0,alias:"number"},"go-version":{pattern:/((?:^|\s)go\s+)\d+(?:\.\d+){1,2}/,lookbehind:!0,alias:"number"},keyword:{pattern:/^([ \t]*)(?:exclude|go|module|replace|require|retract)\b/m,lookbehind:!0},operator:/=>/,punctuation:/[()[\],]/}}return $v}var qv,f2;function mye(){if(f2)return qv;f2=1,qv=e,e.displayName="go",e.aliases=[];function e(t){t.languages.go=t.languages.extend("clike",{string:{pattern:/(^|[^\\])"(?:\\.|[^"\\\r\n])*"|`[^`]*`/,lookbehind:!0,greedy:!0},keyword:/\b(?:break|case|chan|const|continue|default|defer|else|fallthrough|for|func|go(?:to)?|if|import|interface|map|package|range|return|select|struct|switch|type|var)\b/,boolean:/\b(?:_|false|iota|nil|true)\b/,number:[/\b0(?:b[01_]+|o[0-7_]+)i?\b/i,/\b0x(?:[a-f\d_]+(?:\.[a-f\d_]*)?|\.[a-f\d_]+)(?:p[+-]?\d+(?:_\d+)*)?i?(?!\w)/i,/(?:\b\d[\d_]*(?:\.[\d_]*)?|\B\.\d[\d_]*)(?:e[+-]?[\d_]+)?i?(?!\w)/i],operator:/[*\/%^!=]=?|\+[=+]?|-[=-]?|\|[=|]?|&(?:=|&|\^=?)?|>(?:>=?|=)?|<(?:<=?|=|-)?|:=|\.\.\./,builtin:/\b(?:append|bool|byte|cap|close|complex|complex(?:64|128)|copy|delete|error|float(?:32|64)|u?int(?:8|16|32|64)?|imag|len|make|new|panic|print(?:ln)?|real|recover|rune|string|uintptr)\b/}),t.languages.insertBefore("go","string",{char:{pattern:/'(?:\\.|[^'\\\r\n]){0,10}'/,greedy:!0}}),delete t.languages.go["class-name"]}return qv}var Vv,p2;function bye(){if(p2)return Vv;p2=1,Vv=e,e.displayName="graphql",e.aliases=[];function e(t){t.languages.graphql={comment:/#.*/,description:{pattern:/(?:"""(?:[^"]|(?!""")")*"""|"(?:\\.|[^\\"\r\n])*")(?=\s*[a-z_])/i,greedy:!0,alias:"string",inside:{"language-markdown":{pattern:/(^"(?:"")?)(?!\1)[\s\S]+(?=\1$)/,lookbehind:!0,inside:t.languages.markdown}}},string:{pattern:/"""(?:[^"]|(?!""")")*"""|"(?:\\.|[^\\"\r\n])*"/,greedy:!0},number:/(?:\B-|\b)\d+(?:\.\d+)?(?:e[+-]?\d+)?\b/i,boolean:/\b(?:false|true)\b/,variable:/\$[a-z_]\w*/i,directive:{pattern:/@[a-z_]\w*/i,alias:"function"},"attr-name":{pattern:/\b[a-z_]\w*(?=\s*(?:\((?:[^()"]|"(?:\\.|[^\\"\r\n])*")*\))?:)/i,greedy:!0},"atom-input":{pattern:/\b[A-Z]\w*Input\b/,alias:"class-name"},scalar:/\b(?:Boolean|Float|ID|Int|String)\b/,constant:/\b[A-Z][A-Z_\d]*\b/,"class-name":{pattern:/(\b(?:enum|implements|interface|on|scalar|type|union)\s+|&\s*|:\s*|\[)[A-Z_]\w*/,lookbehind:!0},fragment:{pattern:/(\bfragment\s+|\.{3}\s*(?!on\b))[a-zA-Z_]\w*/,lookbehind:!0,alias:"function"},"definition-mutation":{pattern:/(\bmutation\s+)[a-zA-Z_]\w*/,lookbehind:!0,alias:"function"},"definition-query":{pattern:/(\bquery\s+)[a-zA-Z_]\w*/,lookbehind:!0,alias:"function"},keyword:/\b(?:directive|enum|extend|fragment|implements|input|interface|mutation|on|query|repeatable|scalar|schema|subscription|type|union)\b/,operator:/[!=|&]|\.{3}/,"property-query":/\w+(?=\s*\()/,object:/\w+(?=\s*\{)/,punctuation:/[!(){}\[\]:=,]/,property:/\w+/},t.hooks.add("after-tokenize",function(r){if(r.language!=="graphql")return;var a=r.tokens.filter(function(R){return typeof R!="string"&&R.type!=="comment"&&R.type!=="scalar"}),o=0;function s(R){return a[o+R]}function u(R,k){k=k||0;for(var A=0;A<R.length;A++){var C=s(A+k);if(!C||C.type!==R[A])return!1}return!0}function c(R,k){for(var A=1,C=o;C<a.length;C++){var N=a[C],_=N.content;if(N.type==="punctuation"&&typeof _=="string"){if(R.test(_))A++;else if(k.test(_)&&(A--,A===0))return C}}return-1}function d(R,k){var A=R.alias;A?Array.isArray(A)||(R.alias=A=[A]):R.alias=A=[],A.push(k)}for(;o<a.length;){var p=a[o++];if(p.type==="keyword"&&p.content==="mutation"){var g=[];if(u(["definition-mutation","punctuation"])&&s(1).content==="("){o+=2;var m=c(/^\($/,/^\)$/);if(m===-1)continue;for(;o<m;o++){var b=s(0);b.type==="variable"&&(d(b,"variable-input"),g.push(b.content))}o=m+1}if(u(["punctuation","property-query"])&&s(0).content==="{"&&(o++,d(s(0),"property-mutation"),g.length>0)){var y=c(/^\{$/,/^\}$/);if(y===-1)continue;for(var S=o;S<y;S++){var x=a[S];x.type==="variable"&&g.indexOf(x.content)>=0&&d(x,"variable-input")}}}}})}return Vv}var Wv,g2;function yye(){if(g2)return Wv;g2=1,Wv=e,e.displayName="groovy",e.aliases=[];function e(t){t.languages.groovy=t.languages.extend("clike",{string:[{pattern:/("""|''')(?:[^\\]|\\[\s\S])*?\1|\$\/(?:[^/$]|\$(?:[/$]|(?![/$]))|\/(?!\$))*\/\$/,greedy:!0},{pattern:/(["'/])(?:\\.|(?!\1)[^\\\r\n])*\1/,greedy:!0}],keyword:/\b(?:abstract|as|assert|boolean|break|byte|case|catch|char|class|const|continue|def|default|do|double|else|enum|extends|final|finally|float|for|goto|if|implements|import|in|instanceof|int|interface|long|native|new|package|private|protected|public|return|short|static|strictfp|super|switch|synchronized|this|throw|throws|trait|transient|try|void|volatile|while)\b/,number:/\b(?:0b[01_]+|0x[\da-f_]+(?:\.[\da-f_p\-]+)?|[\d_]+(?:\.[\d_]+)?(?:e[+-]?\d+)?)[glidf]?\b/i,operator:{pattern:/(^|[^.])(?:~|==?~?|\?[.:]?|\*(?:[.=]|\*=?)?|\.[@&]|\.\.<|\.\.(?!\.)|-[-=>]?|\+[+=]?|!=?|<(?:<=?|=>?)?|>(?:>>?=?|=)?|&[&=]?|\|[|=]?|\/=?|\^=?|%=?)/,lookbehind:!0},punctuation:/\.+|[{}[\];(),:$]/}),t.languages.insertBefore("groovy","string",{shebang:{pattern:/#!.+/,alias:"comment"}}),t.languages.insertBefore("groovy","punctuation",{"spock-block":/\b(?:and|cleanup|expect|given|setup|then|when|where):/}),t.languages.insertBefore("groovy","function",{annotation:{pattern:/(^|[^.])@\w+/,lookbehind:!0,alias:"punctuation"}}),t.hooks.add("wrap",function(n){if(n.language==="groovy"&&n.type==="string"){var r=n.content.value[0];if(r!="'"){var a=/([^\\])(?:\$(?:\{.*?\}|[\w.]+))/;r==="$"&&(a=/([^\$])(?:\$(?:\{.*?\}|[\w.]+))/),n.content.value=n.content.value.replace(/&lt;/g,"<").replace(/&amp;/g,"&"),n.content=t.highlight(n.content.value,{expression:{pattern:a,lookbehind:!0,inside:t.languages.groovy}}),n.classes.push(r==="/"?"regex":"gstring")}}})}return Wv}var Yv,h2;function vye(){if(h2)return Yv;h2=1;var e=Fp();Yv=t,t.displayName="haml",t.aliases=[];function t(n){n.register(e),function(r){r.languages.haml={"multiline-comment":{pattern:/((?:^|\r?\n|\r)([\t ]*))(?:\/|-#).*(?:(?:\r?\n|\r)\2[\t ].+)*/,lookbehind:!0,alias:"comment"},"multiline-code":[{pattern:/((?:^|\r?\n|\r)([\t ]*)(?:[~-]|[&!]?=)).*,[\t ]*(?:(?:\r?\n|\r)\2[\t ].*,[\t ]*)*(?:(?:\r?\n|\r)\2[\t ].+)/,lookbehind:!0,inside:r.languages.ruby},{pattern:/((?:^|\r?\n|\r)([\t ]*)(?:[~-]|[&!]?=)).*\|[\t ]*(?:(?:\r?\n|\r)\2[\t ].*\|[\t ]*)*/,lookbehind:!0,inside:r.languages.ruby}],filter:{pattern:/((?:^|\r?\n|\r)([\t ]*)):[\w-]+(?:(?:\r?\n|\r)(?:\2[\t ].+|\s*?(?=\r?\n|\r)))+/,lookbehind:!0,inside:{"filter-name":{pattern:/^:[\w-]+/,alias:"symbol"}}},markup:{pattern:/((?:^|\r?\n|\r)[\t ]*)<.+/,lookbehind:!0,inside:r.languages.markup},doctype:{pattern:/((?:^|\r?\n|\r)[\t ]*)!!!(?: .+)?/,lookbehind:!0},tag:{pattern:/((?:^|\r?\n|\r)[\t ]*)[%.#][\w\-#.]*[\w\-](?:\([^)]+\)|\{(?:\{[^}]+\}|[^{}])+\}|\[[^\]]+\])*[\/<>]*/,lookbehind:!0,inside:{attributes:[{pattern:/(^|[^#])\{(?:\{[^}]+\}|[^{}])+\}/,lookbehind:!0,inside:r.languages.ruby},{pattern:/\([^)]+\)/,inside:{"attr-value":{pattern:/(=\s*)(?:"(?:\\.|[^\\"\r\n])*"|[^)\s]+)/,lookbehind:!0},"attr-name":/[\w:-]+(?=\s*!?=|\s*[,)])/,punctuation:/[=(),]/}},{pattern:/\[[^\]]+\]/,inside:r.languages.ruby}],punctuation:/[<>]/}},code:{pattern:/((?:^|\r?\n|\r)[\t ]*(?:[~-]|[&!]?=)).+/,lookbehind:!0,inside:r.languages.ruby},interpolation:{pattern:/#\{[^}]+\}/,inside:{delimiter:{pattern:/^#\{|\}$/,alias:"punctuation"},ruby:{pattern:/[\s\S]+/,inside:r.languages.ruby}}},punctuation:{pattern:/((?:^|\r?\n|\r)[\t ]*)[~=\-&!]+/,lookbehind:!0}};for(var a="((?:^|\\r?\\n|\\r)([\\t ]*)):{{filter_name}}(?:(?:\\r?\\n|\\r)(?:\\2[\\t ].+|\\s*?(?=\\r?\\n|\\r)))+",o=["css",{filter:"coffee",language:"coffeescript"},"erb","javascript","less","markdown","ruby","scss","textile"],s={},u=0,c=o.length;u<c;u++){var d=o[u];d=typeof d=="string"?{filter:d,language:d}:d,r.languages[d.language]&&(s["filter-"+d.filter]={pattern:RegExp(a.replace("{{filter_name}}",function(){return d.filter})),lookbehind:!0,inside:{"filter-name":{pattern:/^:[\w-]+/,alias:"symbol"},text:{pattern:/[\s\S]+/,alias:[d.language,"language-"+d.language],inside:r.languages[d.language]}}})}r.languages.insertBefore("haml","filter",s)}(n)}return Yv}var Kv,m2;function Sye(){if(m2)return Kv;m2=1;var e=Wn();Kv=t,t.displayName="handlebars",t.aliases=["hbs"];function t(n){n.register(e),function(r){r.languages.handlebars={comment:/\{\{![\s\S]*?\}\}/,delimiter:{pattern:/^\{\{\{?|\}\}\}?$/,alias:"punctuation"},string:/(["'])(?:\\.|(?!\1)[^\\\r\n])*\1/,number:/\b0x[\dA-Fa-f]+\b|(?:\b\d+(?:\.\d*)?|\B\.\d+)(?:[Ee][+-]?\d+)?/,boolean:/\b(?:false|true)\b/,block:{pattern:/^(\s*(?:~\s*)?)[#\/]\S+?(?=\s*(?:~\s*)?$|\s)/,lookbehind:!0,alias:"keyword"},brackets:{pattern:/\[[^\]]+\]/,inside:{punctuation:/\[|\]/,variable:/[\s\S]+/}},punctuation:/[!"#%&':()*+,.\/;<=>@\[\\\]^`{|}~]/,variable:/[^!"#%&'()*+,\/;<=>@\[\\\]^`{|}~\s]+/},r.hooks.add("before-tokenize",function(a){var o=/\{\{\{[\s\S]+?\}\}\}|\{\{[\s\S]+?\}\}/g;r.languages["markup-templating"].buildPlaceholders(a,"handlebars",o)}),r.hooks.add("after-tokenize",function(a){r.languages["markup-templating"].tokenizePlaceholders(a,"handlebars")}),r.languages.hbs=r.languages.handlebars}(n)}return Kv}var Xv,b2;function cA(){if(b2)return Xv;b2=1,Xv=e,e.displayName="haskell",e.aliases=["hs"];function e(t){t.languages.haskell={comment:{pattern:/(^|[^-!#$%*+=?&@|~.:<>^\\\/])(?:--(?:(?=.)[^-!#$%*+=?&@|~.:<>^\\\/].*|$)|\{-[\s\S]*?-\})/m,lookbehind:!0},char:{pattern:/'(?:[^\\']|\\(?:[abfnrtv\\"'&]|\^[A-Z@[\]^_]|ACK|BEL|BS|CAN|CR|DC1|DC2|DC3|DC4|DEL|DLE|EM|ENQ|EOT|ESC|ETB|ETX|FF|FS|GS|HT|LF|NAK|NUL|RS|SI|SO|SOH|SP|STX|SUB|SYN|US|VT|\d+|o[0-7]+|x[0-9a-fA-F]+))'/,alias:"string"},string:{pattern:/"(?:[^\\"]|\\(?:\S|\s+\\))*"/,greedy:!0},keyword:/\b(?:case|class|data|deriving|do|else|if|in|infixl|infixr|instance|let|module|newtype|of|primitive|then|type|where)\b/,"import-statement":{pattern:/(^[\t ]*)import\s+(?:qualified\s+)?(?:[A-Z][\w']*)(?:\.[A-Z][\w']*)*(?:\s+as\s+(?:[A-Z][\w']*)(?:\.[A-Z][\w']*)*)?(?:\s+hiding\b)?/m,lookbehind:!0,inside:{keyword:/\b(?:as|hiding|import|qualified)\b/,punctuation:/\./}},builtin:/\b(?:abs|acos|acosh|all|and|any|appendFile|approxRational|asTypeOf|asin|asinh|atan|atan2|atanh|basicIORun|break|catch|ceiling|chr|compare|concat|concatMap|const|cos|cosh|curry|cycle|decodeFloat|denominator|digitToInt|div|divMod|drop|dropWhile|either|elem|encodeFloat|enumFrom|enumFromThen|enumFromThenTo|enumFromTo|error|even|exp|exponent|fail|filter|flip|floatDigits|floatRadix|floatRange|floor|fmap|foldl|foldl1|foldr|foldr1|fromDouble|fromEnum|fromInt|fromInteger|fromIntegral|fromRational|fst|gcd|getChar|getContents|getLine|group|head|id|inRange|index|init|intToDigit|interact|ioError|isAlpha|isAlphaNum|isAscii|isControl|isDenormalized|isDigit|isHexDigit|isIEEE|isInfinite|isLower|isNaN|isNegativeZero|isOctDigit|isPrint|isSpace|isUpper|iterate|last|lcm|length|lex|lexDigits|lexLitChar|lines|log|logBase|lookup|map|mapM|mapM_|max|maxBound|maximum|maybe|min|minBound|minimum|mod|negate|not|notElem|null|numerator|odd|or|ord|otherwise|pack|pi|pred|primExitWith|print|product|properFraction|putChar|putStr|putStrLn|quot|quotRem|range|rangeSize|read|readDec|readFile|readFloat|readHex|readIO|readInt|readList|readLitChar|readLn|readOct|readParen|readSigned|reads|readsPrec|realToFrac|recip|rem|repeat|replicate|return|reverse|round|scaleFloat|scanl|scanl1|scanr|scanr1|seq|sequence|sequence_|show|showChar|showInt|showList|showLitChar|showParen|showSigned|showString|shows|showsPrec|significand|signum|sin|sinh|snd|sort|span|splitAt|sqrt|subtract|succ|sum|tail|take|takeWhile|tan|tanh|threadToIOResult|toEnum|toInt|toInteger|toLower|toRational|toUpper|truncate|uncurry|undefined|unlines|until|unwords|unzip|unzip3|userError|words|writeFile|zip|zip3|zipWith|zipWith3)\b/,number:/\b(?:\d+(?:\.\d+)?(?:e[+-]?\d+)?|0o[0-7]+|0x[0-9a-f]+)\b/i,operator:[{pattern:/`(?:[A-Z][\w']*\.)*[_a-z][\w']*`/,greedy:!0},{pattern:/(\s)\.(?=\s)/,lookbehind:!0},/[-!#$%*+=?&@|~:<>^\\\/][-!#$%*+=?&@|~.:<>^\\\/]*|\.[-!#$%*+=?&@|~.:<>^\\\/]+/],hvariable:{pattern:/\b(?:[A-Z][\w']*\.)*[_a-z][\w']*/,inside:{punctuation:/\./}},constant:{pattern:/\b(?:[A-Z][\w']*\.)*[A-Z][\w']*/,inside:{punctuation:/\./}},punctuation:/[{}[\];(),.:]/},t.languages.hs=t.languages.haskell}return Xv}var Zv,y2;function Eye(){if(y2)return Zv;y2=1,Zv=e,e.displayName="haxe",e.aliases=[];function e(t){t.languages.haxe=t.languages.extend("clike",{string:{pattern:/"(?:[^"\\]|\\[\s\S])*"/,greedy:!0},"class-name":[{pattern:/(\b(?:abstract|class|enum|extends|implements|interface|new|typedef)\s+)[A-Z_]\w*/,lookbehind:!0},/\b[A-Z]\w*/],keyword:/\bthis\b|\b(?:abstract|as|break|case|cast|catch|class|continue|default|do|dynamic|else|enum|extends|extern|final|for|from|function|if|implements|import|in|inline|interface|macro|new|null|operator|overload|override|package|private|public|return|static|super|switch|throw|to|try|typedef|untyped|using|var|while)(?!\.)\b/,function:{pattern:/\b[a-z_]\w*(?=\s*(?:<[^<>]*>\s*)?\()/i,greedy:!0},operator:/\.{3}|\+\+|--|&&|\|\||->|=>|(?:<<?|>{1,3}|[-+*/%!=&|^])=?|[?:~]/}),t.languages.insertBefore("haxe","string",{"string-interpolation":{pattern:/'(?:[^'\\]|\\[\s\S])*'/,greedy:!0,inside:{interpolation:{pattern:/(^|[^\\])\$(?:\w+|\{[^{}]+\})/,lookbehind:!0,inside:{"interpolation-punctuation":{pattern:/^\$\{?|\}$/,alias:"punctuation"},expression:{pattern:/[\s\S]+/,inside:t.languages.haxe}}},string:/[\s\S]+/}}}),t.languages.insertBefore("haxe","class-name",{regex:{pattern:/~\/(?:[^\/\\\r\n]|\\.)+\/[a-z]*/,greedy:!0,inside:{"regex-flags":/\b[a-z]+$/,"regex-source":{pattern:/^(~\/)[\s\S]+(?=\/$)/,lookbehind:!0,alias:"language-regex",inside:t.languages.regex},"regex-delimiter":/^~\/|\/$/}}}),t.languages.insertBefore("haxe","keyword",{preprocessor:{pattern:/#(?:else|elseif|end|if)\b.*/,alias:"property"},metadata:{pattern:/@:?[\w.]+/,alias:"symbol"},reification:{pattern:/\$(?:\w+|(?=\{))/,alias:"important"}})}return Zv}var Qv,v2;function wye(){if(v2)return Qv;v2=1,Qv=e,e.displayName="hcl",e.aliases=[];function e(t){t.languages.hcl={comment:/(?:\/\/|#).*|\/\*[\s\S]*?(?:\*\/|$)/,heredoc:{pattern:/<<-?(\w+\b)[\s\S]*?^[ \t]*\1/m,greedy:!0,alias:"string"},keyword:[{pattern:/(?:data|resource)\s+(?:"(?:\\[\s\S]|[^\\"])*")(?=\s+"[\w-]+"\s+\{)/i,inside:{type:{pattern:/(resource|data|\s+)(?:"(?:\\[\s\S]|[^\\"])*")/i,lookbehind:!0,alias:"variable"}}},{pattern:/(?:backend|module|output|provider|provisioner|variable)\s+(?:[\w-]+|"(?:\\[\s\S]|[^\\"])*")\s+(?=\{)/i,inside:{type:{pattern:/(backend|module|output|provider|provisioner|variable)\s+(?:[\w-]+|"(?:\\[\s\S]|[^\\"])*")\s+/i,lookbehind:!0,alias:"variable"}}},/[\w-]+(?=\s+\{)/],property:[/[-\w\.]+(?=\s*=(?!=))/,/"(?:\\[\s\S]|[^\\"])+"(?=\s*[:=])/],string:{pattern:/"(?:[^\\$"]|\\[\s\S]|\$(?:(?=")|\$+(?!\$)|[^"${])|\$\{(?:[^{}"]|"(?:[^\\"]|\\[\s\S])*")*\})*"/,greedy:!0,inside:{interpolation:{pattern:/(^|[^$])\$\{(?:[^{}"]|"(?:[^\\"]|\\[\s\S])*")*\}/,lookbehind:!0,inside:{type:{pattern:/(\b(?:count|data|local|module|path|self|terraform|var)\b\.)[\w\*]+/i,lookbehind:!0,alias:"variable"},keyword:/\b(?:count|data|local|module|path|self|terraform|var)\b/i,function:/\w+(?=\()/,string:{pattern:/"(?:\\[\s\S]|[^\\"])*"/,greedy:!0},number:/\b0x[\da-f]+\b|\b\d+(?:\.\d*)?(?:e[+-]?\d+)?/i,punctuation:/[!\$#%&'()*+,.\/;<=>@\[\\\]^`{|}~?:]/}}}},number:/\b0x[\da-f]+\b|\b\d+(?:\.\d*)?(?:e[+-]?\d+)?/i,boolean:/\b(?:false|true)\b/i,punctuation:/[=\[\]{}]/}}return Qv}var Jv,S2;function xye(){if(S2)return Jv;S2=1;var e=Ai();Jv=t,t.displayName="hlsl",t.aliases=[];function t(n){n.register(e),n.languages.hlsl=n.languages.extend("c",{"class-name":[n.languages.c["class-name"],/\b(?:AppendStructuredBuffer|BlendState|Buffer|ByteAddressBuffer|CompileShader|ComputeShader|ConsumeStructuredBuffer|DepthStencilState|DepthStencilView|DomainShader|GeometryShader|Hullshader|InputPatch|LineStream|OutputPatch|PixelShader|PointStream|RWBuffer|RWByteAddressBuffer|RWStructuredBuffer|RWTexture(?:1D|1DArray|2D|2DArray|3D)|RasterizerState|RenderTargetView|SamplerComparisonState|SamplerState|StructuredBuffer|Texture(?:1D|1DArray|2D|2DArray|2DMS|2DMSArray|3D|Cube|CubeArray)|TriangleStream|VertexShader)\b/],keyword:[/\b(?:asm|asm_fragment|auto|break|case|catch|cbuffer|centroid|char|class|column_major|compile|compile_fragment|const|const_cast|continue|default|delete|discard|do|dynamic_cast|else|enum|explicit|export|extern|for|friend|fxgroup|goto|groupshared|if|in|inline|inout|interface|line|lineadj|linear|long|matrix|mutable|namespace|new|nointerpolation|noperspective|operator|out|packoffset|pass|pixelfragment|point|precise|private|protected|public|register|reinterpret_cast|return|row_major|sample|sampler|shared|short|signed|sizeof|snorm|stateblock|stateblock_state|static|static_cast|string|struct|switch|tbuffer|technique|technique10|technique11|template|texture|this|throw|triangle|triangleadj|try|typedef|typename|uniform|union|unorm|unsigned|using|vector|vertexfragment|virtual|void|volatile|while)\b/,/\b(?:bool|double|dword|float|half|int|min(?:10float|12int|16(?:float|int|uint))|uint)(?:[1-4](?:x[1-4])?)?\b/],number:/(?:(?:\b\d+(?:\.\d*)?|\B\.\d+)(?:[eE][+-]?\d+)?|\b0x[\da-fA-F]+)[fFhHlLuU]?\b/,boolean:/\b(?:false|true)\b/})}return Jv}var eS,E2;function kye(){if(E2)return eS;E2=1,eS=e,e.displayName="hoon",e.aliases=[];function e(t){t.languages.hoon={comment:{pattern:/::.*/,greedy:!0},string:{pattern:/"[^"]*"|'[^']*'/,greedy:!0},constant:/%(?:\.[ny]|[\w-]+)/,"class-name":/@(?:[a-z0-9-]*[a-z0-9])?|\*/i,function:/(?:\+[-+] {2})?(?:[a-z](?:[a-z0-9-]*[a-z0-9])?)/,keyword:/\.[\^\+\*=\?]|![><:\.=\?!]|=[>|:,\.\-\^<+;/~\*\?]|\?[>|:\.\-\^<\+&~=@!]|\|[\$_%:\.\-\^~\*=@\?]|\+[|\$\+\*]|:[_\-\^\+~\*]|%[_:\.\-\^\+~\*=]|\^[|:\.\-\+&~\*=\?]|\$[|_%:<>\-\^&~@=\?]|;[:<\+;\/~\*=]|~[>|\$_%<\+\/&=\?!]|--|==/}}return eS}var tS,w2;function Tye(){if(w2)return tS;w2=1,tS=e,e.displayName="hpkp",e.aliases=[];function e(t){t.languages.hpkp={directive:{pattern:/\b(?:includeSubDomains|max-age|pin-sha256|preload|report-to|report-uri|strict)(?=[\s;=]|$)/i,alias:"property"},operator:/=/,punctuation:/;/}}return tS}var nS,x2;function Aye(){if(x2)return nS;x2=1,nS=e,e.displayName="hsts",e.aliases=[];function e(t){t.languages.hsts={directive:{pattern:/\b(?:includeSubDomains|max-age|preload)(?=[\s;=]|$)/i,alias:"property"},operator:/=/,punctuation:/;/}}return nS}var rS,k2;function Rye(){if(k2)return rS;k2=1,rS=e,e.displayName="http",e.aliases=[];function e(t){(function(n){function r(g){return RegExp("(^(?:"+g+"):[ 	]*(?![ 	]))[^]+","i")}n.languages.http={"request-line":{pattern:/^(?:CONNECT|DELETE|GET|HEAD|OPTIONS|PATCH|POST|PRI|PUT|SEARCH|TRACE)\s(?:https?:\/\/|\/)\S*\sHTTP\/[\d.]+/m,inside:{method:{pattern:/^[A-Z]+\b/,alias:"property"},"request-target":{pattern:/^(\s)(?:https?:\/\/|\/)\S*(?=\s)/,lookbehind:!0,alias:"url",inside:n.languages.uri},"http-version":{pattern:/^(\s)HTTP\/[\d.]+/,lookbehind:!0,alias:"property"}}},"response-status":{pattern:/^HTTP\/[\d.]+ \d+ .+/m,inside:{"http-version":{pattern:/^HTTP\/[\d.]+/,alias:"property"},"status-code":{pattern:/^(\s)\d+(?=\s)/,lookbehind:!0,alias:"number"},"reason-phrase":{pattern:/^(\s).+/,lookbehind:!0,alias:"string"}}},header:{pattern:/^[\w-]+:.+(?:(?:\r\n?|\n)[ \t].+)*/m,inside:{"header-value":[{pattern:r(/Content-Security-Policy/.source),lookbehind:!0,alias:["csp","languages-csp"],inside:n.languages.csp},{pattern:r(/Public-Key-Pins(?:-Report-Only)?/.source),lookbehind:!0,alias:["hpkp","languages-hpkp"],inside:n.languages.hpkp},{pattern:r(/Strict-Transport-Security/.source),lookbehind:!0,alias:["hsts","languages-hsts"],inside:n.languages.hsts},{pattern:r(/[^:]+/.source),lookbehind:!0}],"header-name":{pattern:/^[^:]+/,alias:"keyword"},punctuation:/^:/}}};var a=n.languages,o={"application/javascript":a.javascript,"application/json":a.json||a.javascript,"application/xml":a.xml,"text/xml":a.xml,"text/html":a.html,"text/css":a.css,"text/plain":a.plain},s={"application/json":!0,"application/xml":!0};function u(g){var m=g.replace(/^[a-z]+\//,""),b="\\w+/(?:[\\w.-]+\\+)+"+m+"(?![+\\w.-])";return"(?:"+g+"|"+b+")"}var c;for(var d in o)if(o[d]){c=c||{};var p=s[d]?u(d):d;c[d.replace(/\//g,"-")]={pattern:RegExp("("+/content-type:\s*/.source+p+/(?:(?:\r\n?|\n)[\w-].*)*(?:\r(?:\n|(?!\n))|\n)/.source+")"+/[^ \t\w-][\s\S]*/.source,"i"),lookbehind:!0,inside:o[d]}}c&&n.languages.insertBefore("http","header",c)})(t)}return rS}var aS,T2;function _ye(){if(T2)return aS;T2=1,aS=e,e.displayName="ichigojam",e.aliases=[];function e(t){t.languages.ichigojam={comment:/(?:\B'|REM)(?:[^\n\r]*)/i,string:{pattern:/"(?:""|[!#$%&'()*,\/:;<=>?^\w +\-.])*"/,greedy:!0},number:/\B#[0-9A-F]+|\B`[01]+|(?:\b\d+(?:\.\d*)?|\B\.\d+)(?:E[+-]?\d+)?/i,keyword:/\b(?:BEEP|BPS|CASE|CLEAR|CLK|CLO|CLP|CLS|CLT|CLV|CONT|COPY|ELSE|END|FILE|FILES|FOR|GOSUB|GOTO|GSB|IF|INPUT|KBD|LED|LET|LIST|LOAD|LOCATE|LRUN|NEW|NEXT|OUT|PLAY|POKE|PRINT|PWM|REM|RENUM|RESET|RETURN|RIGHT|RTN|RUN|SAVE|SCROLL|SLEEP|SRND|STEP|STOP|SUB|TEMPO|THEN|TO|UART|VIDEO|WAIT)(?:\$|\b)/i,function:/\b(?:ABS|ANA|ASC|BIN|BTN|DEC|END|FREE|HELP|HEX|I2CR|I2CW|IN|INKEY|LEN|LINE|PEEK|RND|SCR|SOUND|STR|TICK|USR|VER|VPEEK|ZER)(?:\$|\b)/i,label:/(?:\B@\S+)/,operator:/<[=>]?|>=?|\|\||&&|[+\-*\/=|&^~!]|\b(?:AND|NOT|OR)\b/i,punctuation:/[\[,;:()\]]/}}return aS}var oS,A2;function Cye(){if(A2)return oS;A2=1,oS=e,e.displayName="icon",e.aliases=[];function e(t){t.languages.icon={comment:/#.*/,string:{pattern:/(["'])(?:(?!\1)[^\\\r\n_]|\\.|_(?!\1)(?:\r\n|[\s\S]))*\1/,greedy:!0},number:/\b(?:\d+r[a-z\d]+|\d+(?:\.\d+)?(?:e[+-]?\d+)?)\b|\.\d+\b/i,"builtin-keyword":{pattern:/&(?:allocated|ascii|clock|collections|cset|current|date|dateline|digits|dump|e|error(?:number|text|value)?|errout|fail|features|file|host|input|lcase|letters|level|line|main|null|output|phi|pi|pos|progname|random|regions|source|storage|subject|time|trace|ucase|version)\b/,alias:"variable"},directive:{pattern:/\$\w+/,alias:"builtin"},keyword:/\b(?:break|by|case|create|default|do|else|end|every|fail|global|if|initial|invocable|link|local|next|not|of|procedure|record|repeat|return|static|suspend|then|to|until|while)\b/,function:/\b(?!\d)\w+(?=\s*[({]|\s*!\s*\[)/,operator:/[+-]:(?!=)|(?:[\/?@^%&]|\+\+?|--?|==?=?|~==?=?|\*\*?|\|\|\|?|<(?:->?|<?=?)|>>?=?)(?::=)?|:(?:=:?)?|[!.\\|~]/,punctuation:/[\[\](){},;]/}}return oS}var iS,R2;function Nye(){if(R2)return iS;R2=1,iS=e,e.displayName="icuMessageFormat",e.aliases=[];function e(t){(function(n){function r(d,p){return p<=0?/[]/.source:d.replace(/<SELF>/g,function(){return r(d,p-1)})}var a=/'[{}:=,](?:[^']|'')*'(?!')/,o={pattern:/''/,greedy:!0,alias:"operator"},s={pattern:a,greedy:!0,inside:{escape:o}},u=r(/\{(?:[^{}']|'(?![{},'])|''|<STR>|<SELF>)*\}/.source.replace(/<STR>/g,function(){return a.source}),8),c={pattern:RegExp(u),inside:{message:{pattern:/^(\{)[\s\S]+(?=\}$)/,lookbehind:!0,inside:null},"message-delimiter":{pattern:/./,alias:"punctuation"}}};n.languages["icu-message-format"]={argument:{pattern:RegExp(u),greedy:!0,inside:{content:{pattern:/^(\{)[\s\S]+(?=\}$)/,lookbehind:!0,inside:{"argument-name":{pattern:/^(\s*)[^{}:=,\s]+/,lookbehind:!0},"choice-style":{pattern:/^(\s*,\s*choice\s*,\s*)\S(?:[\s\S]*\S)?/,lookbehind:!0,inside:{punctuation:/\|/,range:{pattern:/^(\s*)[+-]?(?:\d+(?:\.\d*)?|\u221e)\s*[<#\u2264]/,lookbehind:!0,inside:{operator:/[<#\u2264]/,number:/\S+/}},rest:null}},"plural-style":{pattern:/^(\s*,\s*(?:plural|selectordinal)\s*,\s*)\S(?:[\s\S]*\S)?/,lookbehind:!0,inside:{offset:/^offset:\s*\d+/,"nested-message":c,selector:{pattern:/=\d+|[^{}:=,\s]+/,inside:{keyword:/^(?:few|many|one|other|two|zero)$/}}}},"select-style":{pattern:/^(\s*,\s*select\s*,\s*)\S(?:[\s\S]*\S)?/,lookbehind:!0,inside:{"nested-message":c,selector:{pattern:/[^{}:=,\s]+/,inside:{keyword:/^other$/}}}},keyword:/\b(?:choice|plural|select|selectordinal)\b/,"arg-type":{pattern:/\b(?:date|duration|number|ordinal|spellout|time)\b/,alias:"keyword"},"arg-skeleton":{pattern:/(,\s*)::[^{}:=,\s]+/,lookbehind:!0},"arg-style":{pattern:/(,\s*)(?:currency|full|integer|long|medium|percent|short)(?=\s*$)/,lookbehind:!0},"arg-style-text":{pattern:RegExp(/(^\s*,\s*(?=\S))/.source+r(/(?:[^{}']|'[^']*'|\{(?:<SELF>)?\})+/.source,8)+"$"),lookbehind:!0,alias:"string"},punctuation:/,/}},"argument-delimiter":{pattern:/./,alias:"operator"}}},escape:o,string:s},c.inside.message.inside=n.languages["icu-message-format"],n.languages["icu-message-format"].argument.inside.content.inside["choice-style"].inside.rest=n.languages["icu-message-format"]})(t)}return iS}var sS,_2;function Oye(){if(_2)return sS;_2=1;var e=cA();sS=t,t.displayName="idris",t.aliases=["idr"];function t(n){n.register(e),n.languages.idris=n.languages.extend("haskell",{comment:{pattern:/(?:(?:--|\|\|\|).*$|\{-[\s\S]*?-\})/m},keyword:/\b(?:Type|case|class|codata|constructor|corecord|data|do|dsl|else|export|if|implementation|implicit|import|impossible|in|infix|infixl|infixr|instance|interface|let|module|mutual|namespace|of|parameters|partial|postulate|private|proof|public|quoteGoal|record|rewrite|syntax|then|total|using|where|with)\b/,builtin:void 0}),n.languages.insertBefore("idris","keyword",{"import-statement":{pattern:/(^\s*import\s+)(?:[A-Z][\w']*)(?:\.[A-Z][\w']*)*/m,lookbehind:!0,inside:{punctuation:/\./}}}),n.languages.idr=n.languages.idris}return sS}var lS,C2;function Dye(){if(C2)return lS;C2=1,lS=e,e.displayName="iecst",e.aliases=[];function e(t){t.languages.iecst={comment:[{pattern:/(^|[^\\])(?:\/\*[\s\S]*?(?:\*\/|$)|\(\*[\s\S]*?(?:\*\)|$)|\{[\s\S]*?(?:\}|$))/,lookbehind:!0,greedy:!0},{pattern:/(^|[^\\:])\/\/.*/,lookbehind:!0,greedy:!0}],string:{pattern:/(["'])(?:\\(?:\r\n|[\s\S])|(?!\1)[^\\\r\n])*\1/,greedy:!0},keyword:[/\b(?:END_)?(?:PROGRAM|CONFIGURATION|INTERFACE|FUNCTION_BLOCK|FUNCTION|ACTION|TRANSITION|TYPE|STRUCT|(?:INITIAL_)?STEP|NAMESPACE|LIBRARY|CHANNEL|FOLDER|RESOURCE|VAR_(?:ACCESS|CONFIG|EXTERNAL|GLOBAL|INPUT|IN_OUT|OUTPUT|TEMP)|VAR|METHOD|PROPERTY)\b/i,/\b(?:AT|BY|(?:END_)?(?:CASE|FOR|IF|REPEAT|WHILE)|CONSTANT|CONTINUE|DO|ELSE|ELSIF|EXIT|EXTENDS|FROM|GET|GOTO|IMPLEMENTS|JMP|NON_RETAIN|OF|PRIVATE|PROTECTED|PUBLIC|RETAIN|RETURN|SET|TASK|THEN|TO|UNTIL|USING|WITH|__CATCH|__ENDTRY|__FINALLY|__TRY)\b/],"class-name":/\b(?:ANY|ARRAY|BOOL|BYTE|U?(?:D|L|S)?INT|(?:D|L)?WORD|DATE(?:_AND_TIME)?|DT|L?REAL|POINTER|STRING|TIME(?:_OF_DAY)?|TOD)\b/,address:{pattern:/%[IQM][XBWDL][\d.]*|%[IQ][\d.]*/,alias:"symbol"},number:/\b(?:16#[\da-f]+|2#[01_]+|0x[\da-f]+)\b|\b(?:D|DT|T|TOD)#[\d_shmd:]*|\b[A-Z]*#[\d.,_]*|(?:\b\d+(?:\.\d*)?|\B\.\d+)(?:e[+-]?\d+)?/i,boolean:/\b(?:FALSE|NULL|TRUE)\b/,operator:/S?R?:?=>?|&&?|\*\*?|<[=>]?|>=?|[-:^/+#]|\b(?:AND|EQ|EXPT|GE|GT|LE|LT|MOD|NE|NOT|OR|XOR)\b/,function:/\b[a-z_]\w*(?=\s*\()/i,punctuation:/[()[\].,;]/}}return lS}var uS,N2;function Iye(){if(N2)return uS;N2=1,uS=e,e.displayName="ignore",e.aliases=["gitignore","hgignore","npmignore"];function e(t){(function(n){n.languages.ignore={comment:/^#.*/m,entry:{pattern:/\S(?:.*(?:(?:\\ )|\S))?/,alias:"string",inside:{operator:/^!|\*\*?|\?/,regex:{pattern:/(^|[^\\])\[[^\[\]]*\]/,lookbehind:!0},punctuation:/\//}}},n.languages.gitignore=n.languages.ignore,n.languages.hgignore=n.languages.ignore,n.languages.npmignore=n.languages.ignore})(t)}return uS}var cS,O2;function Lye(){if(O2)return cS;O2=1,cS=e,e.displayName="inform7",e.aliases=[];function e(t){t.languages.inform7={string:{pattern:/"[^"]*"/,inside:{substitution:{pattern:/\[[^\[\]]+\]/,inside:{delimiter:{pattern:/\[|\]/,alias:"punctuation"}}}}},comment:{pattern:/\[[^\[\]]+\]/,greedy:!0},title:{pattern:/^[ \t]*(?:book|chapter|part(?! of)|section|table|volume)\b.+/im,alias:"important"},number:{pattern:/(^|[^-])(?:\b\d+(?:\.\d+)?(?:\^\d+)?(?:(?!\d)\w+)?|\b(?:eight|eleven|five|four|nine|one|seven|six|ten|three|twelve|two))\b(?!-)/i,lookbehind:!0},verb:{pattern:/(^|[^-])\b(?:answering|applying to|are|asking|attacking|be(?:ing)?|burning|buying|called|carries|carry(?! out)|carrying|climbing|closing|conceal(?:ing|s)?|consulting|contain(?:ing|s)?|cutting|drinking|dropping|eating|enclos(?:es?|ing)|entering|examining|exiting|getting|giving|going|ha(?:s|ve|ving)|hold(?:ing|s)?|impl(?:ies|y)|incorporat(?:es?|ing)|inserting|is|jumping|kissing|listening|locking|looking|mean(?:ing|s)?|opening|provid(?:es?|ing)|pulling|pushing|putting|relat(?:es?|ing)|removing|searching|see(?:ing|s)?|setting|showing|singing|sleeping|smelling|squeezing|support(?:ing|s)?|swearing|switching|taking|tasting|telling|thinking|throwing|touching|turning|tying|unlock(?:ing|s)?|var(?:ies|y|ying)|waiting|waking|waving|wear(?:ing|s)?)\b(?!-)/i,lookbehind:!0,alias:"operator"},keyword:{pattern:/(^|[^-])\b(?:after|before|carry out|check|continue the action|definition(?= *:)|do nothing|else|end (?:if|the story|unless)|every turn|if|include|instead(?: of)?|let|move|no|now|otherwise|repeat|report|resume the story|rule for|running through|say(?:ing)?|stop the action|test|try(?:ing)?|understand|unless|use|when|while|yes)\b(?!-)/i,lookbehind:!0},property:{pattern:/(^|[^-])\b(?:adjacent(?! to)|carried|closed|concealed|contained|dark|described|edible|empty|enclosed|enterable|even|female|fixed in place|full|handled|held|improper-named|incorporated|inedible|invisible|lighted|lit|lock(?:able|ed)|male|marked for listing|mentioned|negative|neuter|non-(?:empty|full|recurring)|odd|opaque|open(?:able)?|plural-named|portable|positive|privately-named|proper-named|provided|publically-named|pushable between rooms|recurring|related|rubbing|scenery|seen|singular-named|supported|swinging|switch(?:able|ed(?: off| on)?)|touch(?:able|ed)|transparent|unconcealed|undescribed|unlit|unlocked|unmarked for listing|unmentioned|unopenable|untouchable|unvisited|variable|visible|visited|wearable|worn)\b(?!-)/i,lookbehind:!0,alias:"symbol"},position:{pattern:/(^|[^-])\b(?:above|adjacent to|back side of|below|between|down|east|everywhere|front side|here|in|inside(?: from)?|north(?:east|west)?|nowhere|on(?: top of)?|other side|outside(?: from)?|parts? of|regionally in|south(?:east|west)?|through|up|west|within)\b(?!-)/i,lookbehind:!0,alias:"keyword"},type:{pattern:/(^|[^-])\b(?:actions?|activit(?:ies|y)|actors?|animals?|backdrops?|containers?|devices?|directions?|doors?|holders?|kinds?|lists?|m[ae]n|nobody|nothing|nouns?|numbers?|objects?|people|persons?|player(?:'s holdall)?|regions?|relations?|rooms?|rule(?:book)?s?|scenes?|someone|something|supporters?|tables?|texts?|things?|time|vehicles?|wom[ae]n)\b(?!-)/i,lookbehind:!0,alias:"variable"},punctuation:/[.,:;(){}]/},t.languages.inform7.string.inside.substitution.inside.rest=t.languages.inform7,t.languages.inform7.string.inside.substitution.inside.rest.text={pattern:/\S(?:\s*\S)*/,alias:"comment"}}return cS}var dS,D2;function Mye(){if(D2)return dS;D2=1,dS=e,e.displayName="ini",e.aliases=[];function e(t){t.languages.ini={comment:{pattern:/(^[ \f\t\v]*)[#;][^\n\r]*/m,lookbehind:!0},section:{pattern:/(^[ \f\t\v]*)\[[^\n\r\]]*\]?/m,lookbehind:!0,inside:{"section-name":{pattern:/(^\[[ \f\t\v]*)[^ \f\t\v\]]+(?:[ \f\t\v]+[^ \f\t\v\]]+)*/,lookbehind:!0,alias:"selector"},punctuation:/\[|\]/}},key:{pattern:/(^[ \f\t\v]*)[^ \f\n\r\t\v=]+(?:[ \f\t\v]+[^ \f\n\r\t\v=]+)*(?=[ \f\t\v]*=)/m,lookbehind:!0,alias:"attr-name"},value:{pattern:/(=[ \f\t\v]*)[^ \f\n\r\t\v]+(?:[ \f\t\v]+[^ \f\n\r\t\v]+)*/,lookbehind:!0,alias:"attr-value",inside:{"inner-value":{pattern:/^("|').+(?=\1$)/,lookbehind:!0}}},punctuation:/=/}}return dS}var fS,I2;function Fye(){if(I2)return fS;I2=1,fS=e,e.displayName="io",e.aliases=[];function e(t){t.languages.io={comment:{pattern:/(^|[^\\])(?:\/\*[\s\S]*?(?:\*\/|$)|\/\/.*|#.*)/,lookbehind:!0,greedy:!0},"triple-quoted-string":{pattern:/"""(?:\\[\s\S]|(?!""")[^\\])*"""/,greedy:!0,alias:"string"},string:{pattern:/"(?:\\.|[^\\\r\n"])*"/,greedy:!0},keyword:/\b(?:activate|activeCoroCount|asString|block|break|call|catch|clone|collectGarbage|compileString|continue|do|doFile|doMessage|doString|else|elseif|exit|for|foreach|forward|getEnvironmentVariable|getSlot|hasSlot|if|ifFalse|ifNil|ifNilEval|ifTrue|isActive|isNil|isResumable|list|message|method|parent|pass|pause|perform|performWithArgList|print|println|proto|raise|raiseResumable|removeSlot|resend|resume|schedulerSleepSeconds|self|sender|setSchedulerSleepSeconds|setSlot|shallowCopy|slotNames|super|system|then|thisBlock|thisContext|try|type|uniqueId|updateSlot|wait|while|write|yield)\b/,builtin:/\b(?:Array|AudioDevice|AudioMixer|BigNum|Block|Box|Buffer|CFunction|CGI|Color|Curses|DBM|DNSResolver|DOConnection|DOProxy|DOServer|Date|Directory|Duration|DynLib|Error|Exception|FFT|File|Fnmatch|Font|Future|GL|GLE|GLScissor|GLU|GLUCylinder|GLUQuadric|GLUSphere|GLUT|Host|Image|Importer|LinkList|List|Lobby|Locals|MD5|MP3Decoder|MP3Encoder|Map|Message|Movie|Notification|Number|Object|OpenGL|Point|Protos|Random|Regex|SGML|SGMLElement|SGMLParser|SQLite|Sequence|Server|ShowMessage|SleepyCat|SleepyCatCursor|Socket|SocketManager|Sound|Soup|Store|String|Tree|UDPSender|UPDReceiver|URL|User|Warning|WeakLink)\b/,boolean:/\b(?:false|nil|true)\b/,number:/\b0x[\da-f]+\b|(?:\b\d+(?:\.\d*)?|\B\.\d+)(?:e-?\d+)?/i,operator:/[=!*/%+\-^&|]=|>>?=?|<<?=?|:?:?=|\+\+?|--?|\*\*?|\/\/?|%|\|\|?|&&?|\b(?:and|not|or|return)\b|@@?|\?\??|\.\./,punctuation:/[{}[\];(),.:]/}}return fS}var pS,L2;function Pye(){if(L2)return pS;L2=1,pS=e,e.displayName="j",e.aliases=[];function e(t){t.languages.j={comment:{pattern:/\bNB\..*/,greedy:!0},string:{pattern:/'(?:''|[^'\r\n])*'/,greedy:!0},keyword:/\b(?:(?:CR|LF|adverb|conjunction|def|define|dyad|monad|noun|verb)\b|(?:assert|break|case|catch[dt]?|continue|do|else|elseif|end|fcase|for|for_\w+|goto_\w+|if|label_\w+|return|select|throw|try|while|whilst)\.)/,verb:{pattern:/(?!\^:|;\.|[=!][.:])(?:\{(?:\.|::?)?|p(?:\.\.?|:)|[=!\]]|[<>+*\-%$|,#][.:]?|[?^]\.?|[;\[]:?|[~}"i][.:]|[ACeEIjLor]\.|(?:[_\/\\qsux]|_?\d):)/,alias:"keyword"},number:/\b_?(?:(?!\d:)\d+(?:\.\d+)?(?:(?:ad|ar|[ejpx])_?\d+(?:\.\d+)?)*(?:b_?[\da-z]+(?:\.[\da-z]+)?)?|_\b(?!\.))/,adverb:{pattern:/[~}]|[\/\\]\.?|[bfM]\.|t[.:]/,alias:"builtin"},operator:/[=a][.:]|_\./,conjunction:{pattern:/&(?:\.:?|:)?|[.:@][.:]?|[!D][.:]|[;dHT]\.|`:?|[\^LS]:|"/,alias:"variable"},punctuation:/[()]/}}return pS}var gS,M2;function dA(){if(M2)return gS;M2=1,gS=e,e.displayName="java",e.aliases=[];function e(t){(function(n){var r=/\b(?:abstract|assert|boolean|break|byte|case|catch|char|class|const|continue|default|do|double|else|enum|exports|extends|final|finally|float|for|goto|if|implements|import|instanceof|int|interface|long|module|native|new|non-sealed|null|open|opens|package|permits|private|protected|provides|public|record|requires|return|sealed|short|static|strictfp|super|switch|synchronized|this|throw|throws|to|transient|transitive|try|uses|var|void|volatile|while|with|yield)\b/,a=/(^|[^\w.])(?:[a-z]\w*\s*\.\s*)*(?:[A-Z]\w*\s*\.\s*)*/.source,o={pattern:RegExp(a+/[A-Z](?:[\d_A-Z]*[a-z]\w*)?\b/.source),lookbehind:!0,inside:{namespace:{pattern:/^[a-z]\w*(?:\s*\.\s*[a-z]\w*)*(?:\s*\.)?/,inside:{punctuation:/\./}},punctuation:/\./}};n.languages.java=n.languages.extend("clike",{string:{pattern:/(^|[^\\])"(?:\\.|[^"\\\r\n])*"/,lookbehind:!0,greedy:!0},"class-name":[o,{pattern:RegExp(a+/[A-Z]\w*(?=\s+\w+\s*[;,=()])/.source),lookbehind:!0,inside:o.inside}],keyword:r,function:[n.languages.clike.function,{pattern:/(::\s*)[a-z_]\w*/,lookbehind:!0}],number:/\b0b[01][01_]*L?\b|\b0x(?:\.[\da-f_p+-]+|[\da-f_]+(?:\.[\da-f_p+-]+)?)\b|(?:\b\d[\d_]*(?:\.[\d_]*)?|\B\.\d[\d_]*)(?:e[+-]?\d[\d_]*)?[dfl]?/i,operator:{pattern:/(^|[^.])(?:<<=?|>>>?=?|->|--|\+\+|&&|\|\||::|[?:~]|[-+*/%&|^!=<>]=?)/m,lookbehind:!0}}),n.languages.insertBefore("java","string",{"triple-quoted-string":{pattern:/"""[ \t]*[\r\n](?:(?:"|"")?(?:\\.|[^"\\]))*"""/,greedy:!0,alias:"string"},char:{pattern:/'(?:\\.|[^'\\\r\n]){1,6}'/,greedy:!0}}),n.languages.insertBefore("java","class-name",{annotation:{pattern:/(^|[^.])@\w+(?:\s*\.\s*\w+)*/,lookbehind:!0,alias:"punctuation"},generics:{pattern:/<(?:[\w\s,.?]|&(?!&)|<(?:[\w\s,.?]|&(?!&)|<(?:[\w\s,.?]|&(?!&)|<(?:[\w\s,.?]|&(?!&))*>)*>)*>)*>/,inside:{"class-name":o,keyword:r,punctuation:/[<>(),.:]/,operator:/[?&|]/}},namespace:{pattern:RegExp(/(\b(?:exports|import(?:\s+static)?|module|open|opens|package|provides|requires|to|transitive|uses|with)\s+)(?!<keyword>)[a-z]\w*(?:\.[a-z]\w*)*\.?/.source.replace(/<keyword>/g,function(){return r.source})),lookbehind:!0,inside:{punctuation:/\./}}})})(t)}return gS}var hS,F2;function Pp(){if(F2)return hS;F2=1,hS=e,e.displayName="javadoclike",e.aliases=[];function e(t){(function(n){var r=n.languages.javadoclike={parameter:{pattern:/(^[\t ]*(?:\/{3}|\*|\/\*\*)\s*@(?:arg|arguments|param)\s+)\w+/m,lookbehind:!0},keyword:{pattern:/(^[\t ]*(?:\/{3}|\*|\/\*\*)\s*|\{)@[a-z][a-zA-Z-]+\b/m,lookbehind:!0},punctuation:/[{}]/};function a(s,u){var c="doc-comment",d=n.languages[s];if(d){var p=d[c];if(!p){var g={};g[c]={pattern:/(^|[^\\])\/\*\*[^/][\s\S]*?(?:\*\/|$)/,lookbehind:!0,alias:"comment"},d=n.languages.insertBefore(s,"comment",g),p=d[c]}if(p instanceof RegExp&&(p=d[c]={pattern:p}),Array.isArray(p))for(var m=0,b=p.length;m<b;m++)p[m]instanceof RegExp&&(p[m]={pattern:p[m]}),u(p[m]);else u(p)}}function o(s,u){typeof s=="string"&&(s=[s]),s.forEach(function(c){a(c,function(d){d.inside||(d.inside={}),d.inside.rest=u})})}Object.defineProperty(r,"addSupport",{value:o}),r.addSupport(["java","javascript","php"],r)})(t)}return hS}var mS,P2;function zye(){if(P2)return mS;P2=1;var e=dA(),t=Pp();mS=n,n.displayName="javadoc",n.aliases=[];function n(r){r.register(e),r.register(t),function(a){var o=/(^(?:[\t ]*(?:\*\s*)*))[^*\s].*$/m,s=/#\s*\w+(?:\s*\([^()]*\))?/.source,u=/(?:\b[a-zA-Z]\w+\s*\.\s*)*\b[A-Z]\w*(?:\s*<mem>)?|<mem>/.source.replace(/<mem>/g,function(){return s});a.languages.javadoc=a.languages.extend("javadoclike",{}),a.languages.insertBefore("javadoc","keyword",{reference:{pattern:RegExp(/(@(?:exception|link|linkplain|see|throws|value)\s+(?:\*\s*)?)/.source+"(?:"+u+")"),lookbehind:!0,inside:{function:{pattern:/(#\s*)\w+(?=\s*\()/,lookbehind:!0},field:{pattern:/(#\s*)\w+/,lookbehind:!0},namespace:{pattern:/\b(?:[a-z]\w*\s*\.\s*)+/,inside:{punctuation:/\./}},"class-name":/\b[A-Z]\w*/,keyword:a.languages.java.keyword,punctuation:/[#()[\],.]/}},"class-name":{pattern:/(@param\s+)<[A-Z]\w*>/,lookbehind:!0,inside:{punctuation:/[.<>]/}},"code-section":[{pattern:/(\{@code\s+(?!\s))(?:[^\s{}]|\s+(?![\s}])|\{(?:[^{}]|\{(?:[^{}]|\{(?:[^{}]|\{[^{}]*\})*\})*\})*\})+(?=\s*\})/,lookbehind:!0,inside:{code:{pattern:o,lookbehind:!0,inside:a.languages.java,alias:"language-java"}}},{pattern:/(<(code|pre|tt)>(?!<code>)\s*)\S(?:\S|\s+\S)*?(?=\s*<\/\2>)/,lookbehind:!0,inside:{line:{pattern:o,lookbehind:!0,inside:{tag:a.languages.markup.tag,entity:a.languages.markup.entity,code:{pattern:/.+/,inside:a.languages.java,alias:"language-java"}}}}}],tag:a.languages.markup.tag,entity:a.languages.markup.entity}),a.languages.javadoclike.addSupport("java",a.languages.javadoc)}(r)}return mS}var bS,z2;function Bye(){if(z2)return bS;z2=1,bS=e,e.displayName="javastacktrace",e.aliases=[];function e(t){t.languages.javastacktrace={summary:{pattern:/^([\t ]*)(?:(?:Caused by:|Suppressed:|Exception in thread "[^"]*")[\t ]+)?[\w$.]+(?::.*)?$/m,lookbehind:!0,inside:{keyword:{pattern:/^([\t ]*)(?:(?:Caused by|Suppressed)(?=:)|Exception in thread)/m,lookbehind:!0},string:{pattern:/^(\s*)"[^"]*"/,lookbehind:!0},exceptions:{pattern:/^(:?\s*)[\w$.]+(?=:|$)/,lookbehind:!0,inside:{"class-name":/[\w$]+$/,namespace:/\b[a-z]\w*\b/,punctuation:/\./}},message:{pattern:/(:\s*)\S.*/,lookbehind:!0,alias:"string"},punctuation:/:/}},"stack-frame":{pattern:/^([\t ]*)at (?:[\w$./]|@[\w$.+-]*\/)+(?:<init>)?\([^()]*\)/m,lookbehind:!0,inside:{keyword:{pattern:/^(\s*)at(?= )/,lookbehind:!0},source:[{pattern:/(\()\w+\.\w+:\d+(?=\))/,lookbehind:!0,inside:{file:/^\w+\.\w+/,punctuation:/:/,"line-number":{pattern:/\b\d+\b/,alias:"number"}}},{pattern:/(\()[^()]*(?=\))/,lookbehind:!0,inside:{keyword:/^(?:Native Method|Unknown Source)$/}}],"class-name":/[\w$]+(?=\.(?:<init>|[\w$]+)\()/,function:/(?:<init>|[\w$]+)(?=\()/,"class-loader":{pattern:/(\s)[a-z]\w*(?:\.[a-z]\w*)*(?=\/[\w@$.]*\/)/,lookbehind:!0,alias:"namespace",inside:{punctuation:/\./}},module:{pattern:/([\s/])[a-z]\w*(?:\.[a-z]\w*)*(?:@[\w$.+-]*)?(?=\/)/,lookbehind:!0,inside:{version:{pattern:/(@)[\s\S]+/,lookbehind:!0,alias:"number"},punctuation:/[@.]/}},namespace:{pattern:/(?:\b[a-z]\w*\.)+/,inside:{punctuation:/\./}},punctuation:/[()/.]/}},more:{pattern:/^([\t ]*)\.{3} \d+ [a-z]+(?: [a-z]+)*/m,lookbehind:!0,inside:{punctuation:/\.{3}/,number:/\d+/,keyword:/\b[a-z]+(?: [a-z]+)*\b/}}}}return bS}var yS,B2;function Uye(){if(B2)return yS;B2=1,yS=e,e.displayName="jexl",e.aliases=[];function e(t){t.languages.jexl={string:/(["'])(?:\\[\s\S]|(?!\1)[^\\])*\1/,transform:{pattern:/(\|\s*)[a-zA-Zа-яА-Я_\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u00FF$][\wа-яА-Я\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u00FF$]*/,alias:"function",lookbehind:!0},function:/[a-zA-Zа-яА-Я_\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u00FF$][\wа-яА-Я\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u00FF$]*\s*(?=\()/,number:/\b\d+(?:\.\d+)?\b|\B\.\d+\b/,operator:/[<>!]=?|-|\+|&&|==|\|\|?|\/\/?|[?:*^%]/,boolean:/\b(?:false|true)\b/,keyword:/\bin\b/,punctuation:/[{}[\](),.]/}}return yS}var vS,U2;function jye(){if(U2)return vS;U2=1,vS=e,e.displayName="jolie",e.aliases=[];function e(t){t.languages.jolie=t.languages.extend("clike",{string:{pattern:/(^|[^\\])"(?:\\[\s\S]|[^"\\])*"/,lookbehind:!0,greedy:!0},"class-name":{pattern:/((?:\b(?:as|courier|embed|in|inputPort|outputPort|service)\b|@)[ \t]*)\w+/,lookbehind:!0},keyword:/\b(?:as|cH|comp|concurrent|constants|courier|cset|csets|default|define|else|embed|embedded|execution|exit|extender|for|foreach|forward|from|global|if|import|in|include|init|inputPort|install|instanceof|interface|is_defined|linkIn|linkOut|main|new|nullProcess|outputPort|over|private|provide|public|scope|sequential|service|single|spawn|synchronized|this|throw|throws|type|undef|until|while|with)\b/,function:/\b[a-z_]\w*(?=[ \t]*[@(])/i,number:/(?:\b\d+(?:\.\d*)?|\B\.\d+)(?:e[+-]?\d+)?l?/i,operator:/-[-=>]?|\+[+=]?|<[<=]?|[>=*!]=?|&&|\|\||[?\/%^@|]/,punctuation:/[()[\]{},;.:]/,builtin:/\b(?:Byte|any|bool|char|double|enum|float|int|length|long|ranges|regex|string|undefined|void)\b/}),t.languages.insertBefore("jolie","keyword",{aggregates:{pattern:/(\bAggregates\s*:\s*)(?:\w+(?:\s+with\s+\w+)?\s*,\s*)*\w+(?:\s+with\s+\w+)?/,lookbehind:!0,inside:{keyword:/\bwith\b/,"class-name":/\w+/,punctuation:/,/}},redirects:{pattern:/(\bRedirects\s*:\s*)(?:\w+\s*=>\s*\w+\s*,\s*)*(?:\w+\s*=>\s*\w+)/,lookbehind:!0,inside:{punctuation:/,/,"class-name":/\w+/,operator:/=>/}},property:{pattern:/\b(?:Aggregates|[Ii]nterfaces|Java|Javascript|Jolie|[Ll]ocation|OneWay|[Pp]rotocol|Redirects|RequestResponse)\b(?=[ \t]*:)/}})}return vS}var SS,j2;function Gye(){if(j2)return SS;j2=1,SS=e,e.displayName="jq",e.aliases=[];function e(t){(function(n){var r=/\\\((?:[^()]|\([^()]*\))*\)/.source,a=RegExp(/(^|[^\\])"(?:[^"\r\n\\]|\\[^\r\n(]|__)*"/.source.replace(/__/g,function(){return r})),o={interpolation:{pattern:RegExp(/((?:^|[^\\])(?:\\{2})*)/.source+r),lookbehind:!0,inside:{content:{pattern:/^(\\\()[\s\S]+(?=\)$)/,lookbehind:!0,inside:null},punctuation:/^\\\(|\)$/}}},s=n.languages.jq={comment:/#.*/,property:{pattern:RegExp(a.source+/(?=\s*:(?!:))/.source),lookbehind:!0,greedy:!0,inside:o},string:{pattern:a,lookbehind:!0,greedy:!0,inside:o},function:{pattern:/(\bdef\s+)[a-z_]\w+/i,lookbehind:!0},variable:/\B\$\w+/,"property-literal":{pattern:/\b[a-z_]\w*(?=\s*:(?!:))/i,alias:"property"},keyword:/\b(?:as|break|catch|def|elif|else|end|foreach|if|import|include|label|module|modulemeta|null|reduce|then|try|while)\b/,boolean:/\b(?:false|true)\b/,number:/(?:\b\d+\.|\B\.)?\b\d+(?:[eE][+-]?\d+)?\b/,operator:[{pattern:/\|=?/,alias:"pipe"},/\.\.|[!=<>]?=|\?\/\/|\/\/=?|[-+*/%]=?|[<>?]|\b(?:and|not|or)\b/],"c-style-function":{pattern:/\b[a-z_]\w*(?=\s*\()/i,alias:"function"},punctuation:/::|[()\[\]{},:;]|\.(?=\s*[\[\w$])/,dot:{pattern:/\./,alias:"important"}};o.interpolation.inside.content.inside=s})(t)}return SS}var ES,G2;function Hye(){if(G2)return ES;G2=1,ES=e,e.displayName="jsExtras",e.aliases=[];function e(t){(function(n){n.languages.insertBefore("javascript","function-variable",{"method-variable":{pattern:RegExp("(\\.\\s*)"+n.languages.javascript["function-variable"].pattern.source),lookbehind:!0,alias:["function-variable","method","function","property-access"]}}),n.languages.insertBefore("javascript","function",{method:{pattern:RegExp("(\\.\\s*)"+n.languages.javascript.function.source),lookbehind:!0,alias:["function","property-access"]}}),n.languages.insertBefore("javascript","constant",{"known-class-name":[{pattern:/\b(?:(?:Float(?:32|64)|(?:Int|Uint)(?:8|16|32)|Uint8Clamped)?Array|ArrayBuffer|BigInt|Boolean|DataView|Date|Error|Function|Intl|JSON|(?:Weak)?(?:Map|Set)|Math|Number|Object|Promise|Proxy|Reflect|RegExp|String|Symbol|WebAssembly)\b/,alias:"class-name"},{pattern:/\b(?:[A-Z]\w*)Error\b/,alias:"class-name"}]});function r(d,p){return RegExp(d.replace(/<ID>/g,function(){return/(?!\s)[_$a-zA-Z\xA0-\uFFFF](?:(?!\s)[$\w\xA0-\uFFFF])*/.source}),p)}n.languages.insertBefore("javascript","keyword",{imports:{pattern:r(/(\bimport\b\s*)(?:<ID>(?:\s*,\s*(?:\*\s*as\s+<ID>|\{[^{}]*\}))?|\*\s*as\s+<ID>|\{[^{}]*\})(?=\s*\bfrom\b)/.source),lookbehind:!0,inside:n.languages.javascript},exports:{pattern:r(/(\bexport\b\s*)(?:\*(?:\s*as\s+<ID>)?(?=\s*\bfrom\b)|\{[^{}]*\})/.source),lookbehind:!0,inside:n.languages.javascript}}),n.languages.javascript.keyword.unshift({pattern:/\b(?:as|default|export|from|import)\b/,alias:"module"},{pattern:/\b(?:await|break|catch|continue|do|else|finally|for|if|return|switch|throw|try|while|yield)\b/,alias:"control-flow"},{pattern:/\bnull\b/,alias:["null","nil"]},{pattern:/\bundefined\b/,alias:"nil"}),n.languages.insertBefore("javascript","operator",{spread:{pattern:/\.{3}/,alias:"operator"},arrow:{pattern:/=>/,alias:"operator"}}),n.languages.insertBefore("javascript","punctuation",{"property-access":{pattern:r(/(\.\s*)#?<ID>/.source),lookbehind:!0},"maybe-class-name":{pattern:/(^|[^$\w\xA0-\uFFFF])[A-Z][$\w\xA0-\uFFFF]+/,lookbehind:!0},dom:{pattern:/\b(?:document|(?:local|session)Storage|location|navigator|performance|window)\b/,alias:"variable"},console:{pattern:/\bconsole(?=\s*\.)/,alias:"class-name"}});for(var a=["function","function-variable","method","method-variable","property-access"],o=0;o<a.length;o++){var s=a[o],u=n.languages.javascript[s];n.util.type(u)==="RegExp"&&(u=n.languages.javascript[s]={pattern:u});var c=u.inside||{};u.inside=c,c["maybe-class-name"]=/^[A-Z][\s\S]*/}})(t)}return ES}var wS,H2;function $ye(){if(H2)return wS;H2=1,wS=e,e.displayName="jsTemplates",e.aliases=[];function e(t){(function(n){var r=n.languages.javascript["template-string"],a=r.pattern.source,o=r.inside.interpolation,s=o.inside["interpolation-punctuation"],u=o.pattern.source;function c(S,x){if(n.languages[S])return{pattern:RegExp("((?:"+x+")\\s*)"+a),lookbehind:!0,greedy:!0,inside:{"template-punctuation":{pattern:/^`|`$/,alias:"string"},"embedded-code":{pattern:/[\s\S]+/,alias:S}}}}n.languages.javascript["template-string"]=[c("css",/\b(?:styled(?:\([^)]*\))?(?:\s*\.\s*\w+(?:\([^)]*\))*)*|css(?:\s*\.\s*(?:global|resolve))?|createGlobalStyle|keyframes)/.source),c("html",/\bhtml|\.\s*(?:inner|outer)HTML\s*\+?=/.source),c("svg",/\bsvg/.source),c("markdown",/\b(?:markdown|md)/.source),c("graphql",/\b(?:gql|graphql(?:\s*\.\s*experimental)?)/.source),c("sql",/\bsql/.source),r].filter(Boolean);function d(S,x){return"___"+x.toUpperCase()+"_"+S+"___"}function p(S,x,R){var k={code:S,grammar:x,language:R};return n.hooks.run("before-tokenize",k),k.tokens=n.tokenize(k.code,k.grammar),n.hooks.run("after-tokenize",k),k.tokens}function g(S){var x={};x["interpolation-punctuation"]=s;var R=n.tokenize(S,x);if(R.length===3){var k=[1,1];k.push.apply(k,p(R[1],n.languages.javascript,"javascript")),R.splice.apply(R,k)}return new n.Token("interpolation",R,o.alias,S)}function m(S,x,R){var k=n.tokenize(S,{interpolation:{pattern:RegExp(u),lookbehind:!0}}),A=0,C={},N=k.map(function(L){if(typeof L=="string")return L;for(var I=L.content,H;S.indexOf(H=d(A++,R))!==-1;);return C[H]=I,H}).join(""),_=p(N,x,R),O=Object.keys(C);A=0;function F(L){for(var I=0;I<L.length;I++){if(A>=O.length)return;var H=L[I];if(typeof H=="string"||typeof H.content=="string"){var $=O[A],U=typeof H=="string"?H:H.content,Y=U.indexOf($);if(Y!==-1){++A;var Z=U.substring(0,Y),j=g(C[$]),G=U.substring(Y+$.length),z=[];if(Z&&z.push(Z),z.push(j),G){var K=[G];F(K),z.push.apply(z,K)}typeof H=="string"?(L.splice.apply(L,[I,1].concat(z)),I+=z.length-1):H.content=z}}else{var D=H.content;Array.isArray(D)?F(D):F([D])}}}return F(_),new n.Token(R,_,"language-"+R,S)}var b={javascript:!0,js:!0,typescript:!0,ts:!0,jsx:!0,tsx:!0};n.hooks.add("after-tokenize",function(S){if(!(S.language in b))return;function x(R){for(var k=0,A=R.length;k<A;k++){var C=R[k];if(typeof C!="string"){var N=C.content;if(!Array.isArray(N)){typeof N!="string"&&x([N]);continue}if(C.type==="template-string"){var _=N[1];if(N.length===3&&typeof _!="string"&&_.type==="embedded-code"){var O=y(_),F=_.alias,L=Array.isArray(F)?F[0]:F,I=n.languages[L];if(!I)continue;N[1]=m(O,I,L)}}else x(N)}}}x(S.tokens)});function y(S){return typeof S=="string"?S:Array.isArray(S)?S.map(y).join(""):y(S.content)}})(t)}return wS}var xS,$2;function fA(){if($2)return xS;$2=1,xS=e,e.displayName="typescript",e.aliases=["ts"];function e(t){(function(n){n.languages.typescript=n.languages.extend("javascript",{"class-name":{pattern:/(\b(?:class|extends|implements|instanceof|interface|new|type)\s+)(?!keyof\b)(?!\s)[_$a-zA-Z\xA0-\uFFFF](?:(?!\s)[$\w\xA0-\uFFFF])*(?:\s*<(?:[^<>]|<(?:[^<>]|<[^<>]*>)*>)*>)?/,lookbehind:!0,greedy:!0,inside:null},builtin:/\b(?:Array|Function|Promise|any|boolean|console|never|number|string|symbol|unknown)\b/}),n.languages.typescript.keyword.push(/\b(?:abstract|declare|is|keyof|readonly|require)\b/,/\b(?:asserts|infer|interface|module|namespace|type)\b(?=\s*(?:[{_$a-zA-Z\xA0-\uFFFF]|$))/,/\btype\b(?=\s*(?:[\{*]|$))/),delete n.languages.typescript.parameter,delete n.languages.typescript["literal-property"];var r=n.languages.extend("typescript",{});delete r["class-name"],n.languages.typescript["class-name"].inside=r,n.languages.insertBefore("typescript","function",{decorator:{pattern:/@[$\w\xA0-\uFFFF]+/,inside:{at:{pattern:/^@/,alias:"operator"},function:/^[\s\S]+/}},"generic-function":{pattern:/#?(?!\s)[_$a-zA-Z\xA0-\uFFFF](?:(?!\s)[$\w\xA0-\uFFFF])*\s*<(?:[^<>]|<(?:[^<>]|<[^<>]*>)*>)*>(?=\s*\()/,greedy:!0,inside:{function:/^#?(?!\s)[_$a-zA-Z\xA0-\uFFFF](?:(?!\s)[$\w\xA0-\uFFFF])*/,generic:{pattern:/<[\s\S]+/,alias:"class-name",inside:r}}}}),n.languages.ts=n.languages.typescript})(t)}return xS}var kS,q2;function qye(){if(q2)return kS;q2=1;var e=Pp(),t=fA();kS=n,n.displayName="jsdoc",n.aliases=[];function n(r){r.register(e),r.register(t),function(a){var o=a.languages.javascript,s=/\{(?:[^{}]|\{(?:[^{}]|\{[^{}]*\})*\})+\}/.source,u="(@(?:arg|argument|param|property)\\s+(?:"+s+"\\s+)?)";a.languages.jsdoc=a.languages.extend("javadoclike",{parameter:{pattern:RegExp(u+/(?:(?!\s)[$\w\xA0-\uFFFF.])+(?=\s|$)/.source),lookbehind:!0,inside:{punctuation:/\./}}}),a.languages.insertBefore("jsdoc","keyword",{"optional-parameter":{pattern:RegExp(u+/\[(?:(?!\s)[$\w\xA0-\uFFFF.])+(?:=[^[\]]+)?\](?=\s|$)/.source),lookbehind:!0,inside:{parameter:{pattern:/(^\[)[$\w\xA0-\uFFFF\.]+/,lookbehind:!0,inside:{punctuation:/\./}},code:{pattern:/(=)[\s\S]*(?=\]$)/,lookbehind:!0,inside:o,alias:"language-javascript"},punctuation:/[=[\]]/}},"class-name":[{pattern:RegExp(/(@(?:augments|class|extends|interface|memberof!?|template|this|typedef)\s+(?:<TYPE>\s+)?)[A-Z]\w*(?:\.[A-Z]\w*)*/.source.replace(/<TYPE>/g,function(){return s})),lookbehind:!0,inside:{punctuation:/\./}},{pattern:RegExp("(@[a-z]+\\s+)"+s),lookbehind:!0,inside:{string:o.string,number:o.number,boolean:o.boolean,keyword:a.languages.typescript.keyword,operator:/=>|\.\.\.|[&|?:*]/,punctuation:/[.,;=<>{}()[\]]/}}],example:{pattern:/(@example\s+(?!\s))(?:[^@\s]|\s+(?!\s))+?(?=\s*(?:\*\s*)?(?:@\w|\*\/))/,lookbehind:!0,inside:{code:{pattern:/^([\t ]*(?:\*\s*)?)\S.*$/m,lookbehind:!0,inside:o,alias:"language-javascript"}}}}),a.languages.javadoclike.addSupport("javascript",a.languages.jsdoc)}(r)}return kS}var TS,V2;function pA(){if(V2)return TS;V2=1,TS=e,e.displayName="json",e.aliases=["webmanifest"];function e(t){t.languages.json={property:{pattern:/(^|[^\\])"(?:\\.|[^\\"\r\n])*"(?=\s*:)/,lookbehind:!0,greedy:!0},string:{pattern:/(^|[^\\])"(?:\\.|[^\\"\r\n])*"(?!\s*:)/,lookbehind:!0,greedy:!0},comment:{pattern:/\/\/.*|\/\*[\s\S]*?(?:\*\/|$)/,greedy:!0},number:/-?\b\d+(?:\.\d+)?(?:e[+-]?\d+)?\b/i,punctuation:/[{}[\],]/,operator:/:/,boolean:/\b(?:false|true)\b/,null:{pattern:/\bnull\b/,alias:"keyword"}},t.languages.webmanifest=t.languages.json}return TS}var AS,W2;function Vye(){if(W2)return AS;W2=1;var e=pA();AS=t,t.displayName="json5",t.aliases=[];function t(n){n.register(e),function(r){var a=/("|')(?:\\(?:\r\n?|\n|.)|(?!\1)[^\\\r\n])*\1/;r.languages.json5=r.languages.extend("json",{property:[{pattern:RegExp(a.source+"(?=\\s*:)"),greedy:!0},{pattern:/(?!\s)[_$a-zA-Z\xA0-\uFFFF](?:(?!\s)[$\w\xA0-\uFFFF])*(?=\s*:)/,alias:"unquoted"}],string:{pattern:a,greedy:!0},number:/[+-]?\b(?:NaN|Infinity|0x[a-fA-F\d]+)\b|[+-]?(?:\b\d+(?:\.\d*)?|\B\.\d+)(?:[eE][+-]?\d+\b)?/})}(n)}return AS}var RS,Y2;function Wye(){if(Y2)return RS;Y2=1;var e=pA();RS=t,t.displayName="jsonp",t.aliases=[];function t(n){n.register(e),n.languages.jsonp=n.languages.extend("json",{punctuation:/[{}[\]();,.]/}),n.languages.insertBefore("jsonp","punctuation",{function:/(?!\s)[_$a-zA-Z\xA0-\uFFFF](?:(?!\s)[$\w\xA0-\uFFFF])*(?=\s*\()/})}return RS}var _S,K2;function Yye(){if(K2)return _S;K2=1,_S=e,e.displayName="jsstacktrace",e.aliases=[];function e(t){t.languages.jsstacktrace={"error-message":{pattern:/^\S.*/m,alias:"string"},"stack-frame":{pattern:/(^[ \t]+)at[ \t].*/m,lookbehind:!0,inside:{"not-my-code":{pattern:/^at[ \t]+(?!\s)(?:node\.js|<unknown>|.*(?:node_modules|\(<anonymous>\)|\(<unknown>|<anonymous>$|\(internal\/|\(node\.js)).*/m,alias:"comment"},filename:{pattern:/(\bat\s+(?!\s)|\()(?:[a-zA-Z]:)?[^():]+(?=:)/,lookbehind:!0,alias:"url"},function:{pattern:/(\bat\s+(?:new\s+)?)(?!\s)[_$a-zA-Z\xA0-\uFFFF<][.$\w\xA0-\uFFFF<>]*/,lookbehind:!0,inside:{punctuation:/\./}},punctuation:/[()]/,keyword:/\b(?:at|new)\b/,alias:{pattern:/\[(?:as\s+)?(?!\s)[_$a-zA-Z\xA0-\uFFFF][$\w\xA0-\uFFFF]*\]/,alias:"variable"},"line-number":{pattern:/:\d+(?::\d+)?\b/,alias:"number",inside:{punctuation:/:/}}}}}}return _S}var CS,X2;function qH(){if(X2)return CS;X2=1,CS=e,e.displayName="jsx",e.aliases=[];function e(t){(function(n){var r=n.util.clone(n.languages.javascript),a=/(?:\s|\/\/.*(?!.)|\/\*(?:[^*]|\*(?!\/))\*\/)/.source,o=/(?:\{(?:\{(?:\{[^{}]*\}|[^{}])*\}|[^{}])*\})/.source,s=/(?:\{<S>*\.{3}(?:[^{}]|<BRACES>)*\})/.source;function u(p,g){return p=p.replace(/<S>/g,function(){return a}).replace(/<BRACES>/g,function(){return o}).replace(/<SPREAD>/g,function(){return s}),RegExp(p,g)}s=u(s).source,n.languages.jsx=n.languages.extend("markup",r),n.languages.jsx.tag.pattern=u(/<\/?(?:[\w.:-]+(?:<S>+(?:[\w.:$-]+(?:=(?:"(?:\\[\s\S]|[^\\"])*"|'(?:\\[\s\S]|[^\\'])*'|[^\s{'"/>=]+|<BRACES>))?|<SPREAD>))*<S>*\/?)?>/.source),n.languages.jsx.tag.inside.tag.pattern=/^<\/?[^\s>\/]*/,n.languages.jsx.tag.inside["attr-value"].pattern=/=(?!\{)(?:"(?:\\[\s\S]|[^\\"])*"|'(?:\\[\s\S]|[^\\'])*'|[^\s'">]+)/,n.languages.jsx.tag.inside.tag.inside["class-name"]=/^[A-Z]\w*(?:\.[A-Z]\w*)*$/,n.languages.jsx.tag.inside.comment=r.comment,n.languages.insertBefore("inside","attr-name",{spread:{pattern:u(/<SPREAD>/.source),inside:n.languages.jsx}},n.languages.jsx.tag),n.languages.insertBefore("inside","special-attr",{script:{pattern:u(/=<BRACES>/.source),alias:"language-javascript",inside:{"script-punctuation":{pattern:/^=(?=\{)/,alias:"punctuation"},rest:n.languages.jsx}}},n.languages.jsx.tag);var c=function(p){return p?typeof p=="string"?p:typeof p.content=="string"?p.content:p.content.map(c).join(""):""},d=function(p){for(var g=[],m=0;m<p.length;m++){var b=p[m],y=!1;if(typeof b!="string"&&(b.type==="tag"&&b.content[0]&&b.content[0].type==="tag"?b.content[0].content[0].content==="</"?g.length>0&&g[g.length-1].tagName===c(b.content[0].content[1])&&g.pop():b.content[b.content.length-1].content==="/>"||g.push({tagName:c(b.content[0].content[1]),openedBraces:0}):g.length>0&&b.type==="punctuation"&&b.content==="{"?g[g.length-1].openedBraces++:g.length>0&&g[g.length-1].openedBraces>0&&b.type==="punctuation"&&b.content==="}"?g[g.length-1].openedBraces--:y=!0),(y||typeof b=="string")&&g.length>0&&g[g.length-1].openedBraces===0){var S=c(b);m<p.length-1&&(typeof p[m+1]=="string"||p[m+1].type==="plain-text")&&(S+=c(p[m+1]),p.splice(m+1,1)),m>0&&(typeof p[m-1]=="string"||p[m-1].type==="plain-text")&&(S=c(p[m-1])+S,p.splice(m-1,1),m--),p[m]=new n.Token("plain-text",S,null,S)}b.content&&typeof b.content!="string"&&d(b.content)}};n.hooks.add("after-tokenize",function(p){p.language!=="jsx"&&p.language!=="tsx"||d(p.tokens)})})(t)}return CS}var NS,Z2;function Kye(){if(Z2)return NS;Z2=1,NS=e,e.displayName="julia",e.aliases=[];function e(t){t.languages.julia={comment:{pattern:/(^|[^\\])(?:#=(?:[^#=]|=(?!#)|#(?!=)|#=(?:[^#=]|=(?!#)|#(?!=))*=#)*=#|#.*)/,lookbehind:!0},regex:{pattern:/r"(?:\\.|[^"\\\r\n])*"[imsx]{0,4}/,greedy:!0},string:{pattern:/"""[\s\S]+?"""|(?:\b\w+)?"(?:\\.|[^"\\\r\n])*"|`(?:[^\\`\r\n]|\\.)*`/,greedy:!0},char:{pattern:/(^|[^\w'])'(?:\\[^\r\n][^'\r\n]*|[^\\\r\n])'/,lookbehind:!0,greedy:!0},keyword:/\b(?:abstract|baremodule|begin|bitstype|break|catch|ccall|const|continue|do|else|elseif|end|export|finally|for|function|global|if|immutable|import|importall|in|let|local|macro|module|print|println|quote|return|struct|try|type|typealias|using|while)\b/,boolean:/\b(?:false|true)\b/,number:/(?:\b(?=\d)|\B(?=\.))(?:0[box])?(?:[\da-f]+(?:_[\da-f]+)*(?:\.(?:\d+(?:_\d+)*)?)?|\.\d+(?:_\d+)*)(?:[efp][+-]?\d+(?:_\d+)*)?j?/i,operator:/&&|\|\||[-+*^%÷⊻&$\\]=?|\/[\/=]?|!=?=?|\|[=>]?|<(?:<=?|[=:|])?|>(?:=|>>?=?)?|==?=?|[~≠≤≥'√∛]/,punctuation:/::?|[{}[\]();,.?]/,constant:/\b(?:(?:Inf|NaN)(?:16|32|64)?|im|pi)\b|[πℯ]/}}return NS}var OS,Q2;function Xye(){if(Q2)return OS;Q2=1,OS=e,e.displayName="keepalived",e.aliases=[];function e(t){t.languages.keepalived={comment:{pattern:/[#!].*/,greedy:!0},string:{pattern:/(^|[^\\])(?:"(?:\\(?:\r\n|[\s\S])|[^"\\\r\n])*"|'(?:\\(?:\r\n|[\s\S])|[^'\\\r\n])*')/,lookbehind:!0,greedy:!0},ip:{pattern:RegExp(/\b(?:(?:(?:[\da-f]{1,4}:){7}[\da-f]{1,4}|(?:[\da-f]{1,4}:){6}:[\da-f]{1,4}|(?:[\da-f]{1,4}:){5}:(?:[\da-f]{1,4}:)?[\da-f]{1,4}|(?:[\da-f]{1,4}:){4}:(?:[\da-f]{1,4}:){0,2}[\da-f]{1,4}|(?:[\da-f]{1,4}:){3}:(?:[\da-f]{1,4}:){0,3}[\da-f]{1,4}|(?:[\da-f]{1,4}:){2}:(?:[\da-f]{1,4}:){0,4}[\da-f]{1,4}|(?:[\da-f]{1,4}:){6}<ipv4>|(?:[\da-f]{1,4}:){0,5}:<ipv4>|::(?:[\da-f]{1,4}:){0,5}<ipv4>|[\da-f]{1,4}::(?:[\da-f]{1,4}:){0,5}[\da-f]{1,4}|::(?:[\da-f]{1,4}:){0,6}[\da-f]{1,4}|(?:[\da-f]{1,4}:){1,7}:)(?:\/\d{1,3})?|<ipv4>(?:\/\d{1,2})?)\b/.source.replace(/<ipv4>/g,function(){return/(?:(?:(?:25[0-5]|2[0-4]\d|1\d\d|[1-9]\d|\d)\.){3}(?:25[0-5]|2[0-4]\d|1\d\d|[1-9]\d|\d))/.source}),"i"),alias:"number"},path:{pattern:/(\s)\/(?:[^\/\s]+\/)*[^\/\s]*|\b[a-zA-Z]:\\(?:[^\\\s]+\\)*[^\\\s]*/,lookbehind:!0,alias:"string"},variable:/\$\{?\w+\}?/,email:{pattern:/[\w-]+@[\w-]+(?:\.[\w-]{2,3}){1,2}/,alias:"string"},"conditional-configuration":{pattern:/@\^?[\w-]+/,alias:"variable"},operator:/=/,property:/\b(?:BFD_CHECK|DNS_CHECK|FILE_CHECK|HTTP_GET|MISC_CHECK|NAME|PING_CHECK|SCRIPTS|SMTP_CHECK|SSL|SSL_GET|TCP_CHECK|UDP_CHECK|accept|advert_int|alpha|auth_pass|auth_type|authentication|bfd_cpu_affinity|bfd_instance|bfd_no_swap|bfd_priority|bfd_process_name|bfd_rlimit_rttime|bfd_rt_priority|bind_if|bind_port|bindto|ca|certificate|check_unicast_src|checker|checker_cpu_affinity|checker_log_all_failures|checker_no_swap|checker_priority|checker_rlimit_rttime|checker_rt_priority|child_wait_time|connect_ip|connect_port|connect_timeout|dbus_service_name|debug|default_interface|delay|delay_before_retry|delay_loop|digest|dont_track_primary|dynamic|dynamic_interfaces|enable_(?:dbus|script_security|sni|snmp_checker|snmp_rfc|snmp_rfcv2|snmp_rfcv3|snmp_vrrp|traps)|end|fall|fast_recovery|file|flag-[123]|fork_delay|full_command|fwmark|garp_group|garp_interval|garp_lower_prio_delay|garp_lower_prio_repeat|garp_master_delay|garp_master_refresh|garp_master_refresh_repeat|garp_master_repeat|global_defs|global_tracking|gna_interval|group|ha_suspend|hashed|helo_name|higher_prio_send_advert|hoplimit|http_protocol|hysteresis|idle_tx|include|inhibit_on_failure|init_fail|init_file|instance|interface|interfaces|interval|ip_family|ipvs_process_name|keepalived.conf|kernel_rx_buf_size|key|linkbeat_interfaces|linkbeat_use_polling|log_all_failures|log_unknown_vrids|lower_prio_no_advert|lthreshold|lvs_flush|lvs_flush_onstop|lvs_method|lvs_netlink_cmd_rcv_bufs|lvs_netlink_cmd_rcv_bufs_force|lvs_netlink_monitor_rcv_bufs|lvs_netlink_monitor_rcv_bufs_force|lvs_notify_fifo|lvs_notify_fifo_script|lvs_sched|lvs_sync_daemon|max_auto_priority|max_hops|mcast_src_ip|mh-fallback|mh-port|min_auto_priority_delay|min_rx|min_tx|misc_dynamic|misc_path|misc_timeout|multiplier|name|namespace_with_ipsets|native_ipv6|neighbor_ip|net_namespace|net_namespace_ipvs|nftables|nftables_counters|nftables_ifindex|nftables_priority|no_accept|no_checker_emails|no_email_faults|nopreempt|notification_email|notification_email_from|notify|notify_backup|notify_deleted|notify_down|notify_fault|notify_fifo|notify_fifo_script|notify_master|notify_master_rx_lower_pri|notify_priority_changes|notify_stop|notify_up|old_unicast_checksum|omega|ops|param_match|passive|password|path|persistence_engine|persistence_granularity|persistence_timeout|preempt|preempt_delay|priority|process|process_monitor_rcv_bufs|process_monitor_rcv_bufs_force|process_name|process_names|promote_secondaries|protocol|proxy_arp|proxy_arp_pvlan|quorum|quorum_down|quorum_max|quorum_up|random_seed|real_server|regex|regex_max_offset|regex_min_offset|regex_no_match|regex_options|regex_stack|reload_repeat|reload_time_file|require_reply|retry|rise|router_id|rs_init_notifies|script|script_user|sh-fallback|sh-port|shutdown_script|shutdown_script_timeout|skip_check_adv_addr|smtp_alert|smtp_alert_checker|smtp_alert_vrrp|smtp_connect_timeout|smtp_helo_name|smtp_server|snmp_socket|sorry_server|sorry_server_inhibit|sorry_server_lvs_method|source_ip|start|startup_script|startup_script_timeout|state|static_ipaddress|static_routes|static_rules|status_code|step|strict_mode|sync_group_tracking_weight|terminate_delay|timeout|track_bfd|track_file|track_group|track_interface|track_process|track_script|track_src_ip|ttl|type|umask|unicast_peer|unicast_src_ip|unicast_ttl|url|use_ipvlan|use_pid_dir|use_vmac|user|uthreshold|val[123]|version|virtual_ipaddress|virtual_ipaddress_excluded|virtual_router_id|virtual_routes|virtual_rules|virtual_server|virtual_server_group|virtualhost|vmac_xmit_base|vrrp|vrrp_(?:check_unicast_src|cpu_affinity|garp_interval|garp_lower_prio_delay|garp_lower_prio_repeat|garp_master_delay|garp_master_refresh|garp_master_refresh_repeat|garp_master_repeat|gna_interval|higher_prio_send_advert|instance|ipsets|iptables|lower_prio_no_advert|mcast_group4|mcast_group6|min_garp|netlink_cmd_rcv_bufs|netlink_cmd_rcv_bufs_force|netlink_monitor_rcv_bufs|netlink_monitor_rcv_bufs_force|no_swap|notify_fifo|notify_fifo_script|notify_priority_changes|priority|process_name|rlimit_rttime|rt_priority|rx_bufs_multiplier|rx_bufs_policy|script|skip_check_adv_addr|startup_delay|strict|sync_group|track_process|version)|warmup|weight)\b/,constant:/\b(?:A|AAAA|AH|BACKUP|CNAME|DR|MASTER|MX|NAT|NS|PASS|SCTP|SOA|TCP|TUN|TXT|UDP|dh|fo|lblc|lblcr|lc|mh|nq|ovf|rr|sed|sh|wlc|wrr)\b/,number:{pattern:/(^|[^\w.-])-?\d+(?:\.\d+)?/,lookbehind:!0},boolean:/\b(?:false|no|off|on|true|yes)\b/,punctuation:/[\{\}]/}}return OS}var DS,J2;function Zye(){if(J2)return DS;J2=1,DS=e,e.displayName="keyman",e.aliases=[];function e(t){t.languages.keyman={comment:{pattern:/\bc .*/i,greedy:!0},string:{pattern:/"[^"\r\n]*"|'[^'\r\n]*'/,greedy:!0},"virtual-key":{pattern:/\[\s*(?:(?:ALT|CAPS|CTRL|LALT|LCTRL|NCAPS|RALT|RCTRL|SHIFT)\s+)*(?:[TKU]_[\w?]+|[A-E]\d\d?|"[^"\r\n]*"|'[^'\r\n]*')\s*\]/i,greedy:!0,alias:"function"},"header-keyword":{pattern:/&\w+/,alias:"bold"},"header-statement":{pattern:/\b(?:bitmap|bitmaps|caps always off|caps on only|copyright|hotkey|language|layout|message|name|shift frees caps|version)\b/i,alias:"bold"},"rule-keyword":{pattern:/\b(?:any|baselayout|beep|call|context|deadkey|dk|if|index|layer|notany|nul|outs|platform|reset|return|save|set|store|use)\b/i,alias:"keyword"},"structural-keyword":{pattern:/\b(?:ansi|begin|group|match|nomatch|unicode|using keys)\b/i,alias:"keyword"},"compile-target":{pattern:/\$(?:keyman|keymanonly|keymanweb|kmfl|weaver):/i,alias:"property"},number:/\b(?:U\+[\dA-F]+|d\d+|x[\da-f]+|\d+)\b/i,operator:/[+>\\$]|\.\./,punctuation:/[()=,]/}}return DS}var IS,eM;function Qye(){if(eM)return IS;eM=1,IS=e,e.displayName="kotlin",e.aliases=["kt","kts"];function e(t){(function(n){n.languages.kotlin=n.languages.extend("clike",{keyword:{pattern:/(^|[^.])\b(?:abstract|actual|annotation|as|break|by|catch|class|companion|const|constructor|continue|crossinline|data|do|dynamic|else|enum|expect|external|final|finally|for|fun|get|if|import|in|infix|init|inline|inner|interface|internal|is|lateinit|noinline|null|object|open|operator|out|override|package|private|protected|public|reified|return|sealed|set|super|suspend|tailrec|this|throw|to|try|typealias|val|var|vararg|when|where|while)\b/,lookbehind:!0},function:[{pattern:/(?:`[^\r\n`]+`|\b\w+)(?=\s*\()/,greedy:!0},{pattern:/(\.)(?:`[^\r\n`]+`|\w+)(?=\s*\{)/,lookbehind:!0,greedy:!0}],number:/\b(?:0[xX][\da-fA-F]+(?:_[\da-fA-F]+)*|0[bB][01]+(?:_[01]+)*|\d+(?:_\d+)*(?:\.\d+(?:_\d+)*)?(?:[eE][+-]?\d+(?:_\d+)*)?[fFL]?)\b/,operator:/\+[+=]?|-[-=>]?|==?=?|!(?:!|==?)?|[\/*%<>]=?|[?:]:?|\.\.|&&|\|\||\b(?:and|inv|or|shl|shr|ushr|xor)\b/}),delete n.languages.kotlin["class-name"];var r={"interpolation-punctuation":{pattern:/^\$\{?|\}$/,alias:"punctuation"},expression:{pattern:/[\s\S]+/,inside:n.languages.kotlin}};n.languages.insertBefore("kotlin","string",{"string-literal":[{pattern:/"""(?:[^$]|\$(?:(?!\{)|\{[^{}]*\}))*?"""/,alias:"multiline",inside:{interpolation:{pattern:/\$(?:[a-z_]\w*|\{[^{}]*\})/i,inside:r},string:/[\s\S]+/}},{pattern:/"(?:[^"\\\r\n$]|\\.|\$(?:(?!\{)|\{[^{}]*\}))*"/,alias:"singleline",inside:{interpolation:{pattern:/((?:^|[^\\])(?:\\{2})*)\$(?:[a-z_]\w*|\{[^{}]*\})/i,lookbehind:!0,inside:r},string:/[\s\S]+/}}],char:{pattern:/'(?:[^'\\\r\n]|\\(?:.|u[a-fA-F0-9]{0,4}))'/,greedy:!0}}),delete n.languages.kotlin.string,n.languages.insertBefore("kotlin","keyword",{annotation:{pattern:/\B@(?:\w+:)?(?:[A-Z]\w*|\[[^\]]+\])/,alias:"builtin"}}),n.languages.insertBefore("kotlin","function",{label:{pattern:/\b\w+@|@\w+\b/,alias:"symbol"}}),n.languages.kt=n.languages.kotlin,n.languages.kts=n.languages.kotlin})(t)}return IS}var LS,tM;function Jye(){if(tM)return LS;tM=1,LS=e,e.displayName="kumir",e.aliases=["kum"];function e(t){(function(n){var r=/\s\x00-\x1f\x22-\x2f\x3a-\x3f\x5b-\x5e\x60\x7b-\x7e/.source;function a(o,s){return RegExp(o.replace(/<nonId>/g,r),s)}n.languages.kumir={comment:{pattern:/\|.*/},prolog:{pattern:/#.*/,greedy:!0},string:{pattern:/"[^\n\r"]*"|'[^\n\r']*'/,greedy:!0},boolean:{pattern:a(/(^|[<nonId>])(?:да|нет)(?=[<nonId>]|$)/.source),lookbehind:!0},"operator-word":{pattern:a(/(^|[<nonId>])(?:и|или|не)(?=[<nonId>]|$)/.source),lookbehind:!0,alias:"keyword"},"system-variable":{pattern:a(/(^|[<nonId>])знач(?=[<nonId>]|$)/.source),lookbehind:!0,alias:"keyword"},type:[{pattern:a(/(^|[<nonId>])(?:вещ|лит|лог|сим|цел)(?:\x20*таб)?(?=[<nonId>]|$)/.source),lookbehind:!0,alias:"builtin"},{pattern:a(/(^|[<nonId>])(?:компл|сканкод|файл|цвет)(?=[<nonId>]|$)/.source),lookbehind:!0,alias:"important"}],keyword:{pattern:a(/(^|[<nonId>])(?:алг|арг(?:\x20*рез)?|ввод|ВКЛЮЧИТЬ|вс[её]|выбор|вывод|выход|дано|для|до|дс|если|иначе|исп|использовать|кон(?:(?:\x20+|_)исп)?|кц(?:(?:\x20+|_)при)?|надо|нач|нс|нц|от|пауза|пока|при|раза?|рез|стоп|таб|то|утв|шаг)(?=[<nonId>]|$)/.source),lookbehind:!0},name:{pattern:a(/(^|[<nonId>])[^\d<nonId>][^<nonId>]*(?:\x20+[^<nonId>]+)*(?=[<nonId>]|$)/.source),lookbehind:!0},number:{pattern:a(/(^|[<nonId>])(?:\B\$[\da-f]+\b|(?:\b\d+(?:\.\d*)?|\B\.\d+)(?:e[+-]?\d+)?)(?=[<nonId>]|$)/.source,"i"),lookbehind:!0},punctuation:/:=|[(),:;\[\]]/,"operator-char":{pattern:/\*\*?|<[=>]?|>=?|[-+/=]/,alias:"operator"}},n.languages.kum=n.languages.kumir})(t)}return LS}var MS,nM;function eve(){if(nM)return MS;nM=1,MS=e,e.displayName="kusto",e.aliases=[];function e(t){t.languages.kusto={comment:{pattern:/\/\/.*/,greedy:!0},string:{pattern:/```[\s\S]*?```|[hH]?(?:"(?:[^\r\n\\"]|\\.)*"|'(?:[^\r\n\\']|\\.)*'|@(?:"[^\r\n"]*"|'[^\r\n']*'))/,greedy:!0},verb:{pattern:/(\|\s*)[a-z][\w-]*/i,lookbehind:!0,alias:"keyword"},command:{pattern:/\.[a-z][a-z\d-]*\b/,alias:"keyword"},"class-name":/\b(?:bool|datetime|decimal|dynamic|guid|int|long|real|string|timespan)\b/,keyword:/\b(?:access|alias|and|anti|as|asc|auto|between|by|(?:contains|(?:ends|starts)with|has(?:perfix|suffix)?)(?:_cs)?|database|declare|desc|external|from|fullouter|has_all|in|ingestion|inline|inner|innerunique|into|(?:left|right)(?:anti(?:semi)?|inner|outer|semi)?|let|like|local|not|of|on|or|pattern|print|query_parameters|range|restrict|schema|set|step|table|tables|to|view|where|with|matches\s+regex|nulls\s+(?:first|last))(?![\w-])/,boolean:/\b(?:false|null|true)\b/,function:/\b[a-z_]\w*(?=\s*\()/,datetime:[{pattern:/\b(?:(?:Fri|Friday|Mon|Monday|Sat|Saturday|Sun|Sunday|Thu|Thursday|Tue|Tuesday|Wed|Wednesday)\s*,\s*)?\d{1,2}(?:\s+|-)(?:Apr|Aug|Dec|Feb|Jan|Jul|Jun|Mar|May|Nov|Oct|Sep)(?:\s+|-)\d{2}\s+\d{2}:\d{2}(?::\d{2})?(?:\s*(?:\b(?:[A-Z]|(?:[ECMT][DS]|GM|U)T)|[+-]\d{4}))?\b/,alias:"number"},{pattern:/[+-]?\b(?:\d{4}-\d{2}-\d{2}(?:[ T]\d{2}:\d{2}(?::\d{2}(?:\.\d+)?)?)?|\d{2}:\d{2}(?::\d{2}(?:\.\d+)?)?)Z?/,alias:"number"}],number:/\b(?:0x[0-9A-Fa-f]+|\d+(?:\.\d+)?(?:[Ee][+-]?\d+)?)(?:(?:min|sec|[mnµ]s|[dhms]|microsecond|tick)\b)?|[+-]?\binf\b/,operator:/=>|[!=]~|[!=<>]=?|[-+*/%|]|\.\./,punctuation:/[()\[\]{},;.:]/}}return MS}var FS,rM;function tve(){if(rM)return FS;rM=1,FS=e,e.displayName="latex",e.aliases=["tex","context"];function e(t){(function(n){var r=/\\(?:[^a-z()[\]]|[a-z*]+)/i,a={"equation-command":{pattern:r,alias:"regex"}};n.languages.latex={comment:/%.*/,cdata:{pattern:/(\\begin\{((?:lstlisting|verbatim)\*?)\})[\s\S]*?(?=\\end\{\2\})/,lookbehind:!0},equation:[{pattern:/\$\$(?:\\[\s\S]|[^\\$])+\$\$|\$(?:\\[\s\S]|[^\\$])+\$|\\\([\s\S]*?\\\)|\\\[[\s\S]*?\\\]/,inside:a,alias:"string"},{pattern:/(\\begin\{((?:align|eqnarray|equation|gather|math|multline)\*?)\})[\s\S]*?(?=\\end\{\2\})/,lookbehind:!0,inside:a,alias:"string"}],keyword:{pattern:/(\\(?:begin|cite|documentclass|end|label|ref|usepackage)(?:\[[^\]]+\])?\{)[^}]+(?=\})/,lookbehind:!0},url:{pattern:/(\\url\{)[^}]+(?=\})/,lookbehind:!0},headline:{pattern:/(\\(?:chapter|frametitle|paragraph|part|section|subparagraph|subsection|subsubparagraph|subsubsection|subsubsubparagraph)\*?(?:\[[^\]]+\])?\{)[^}]+(?=\})/,lookbehind:!0,alias:"class-name"},function:{pattern:r,alias:"selector"},punctuation:/[[\]{}&]/},n.languages.tex=n.languages.latex,n.languages.context=n.languages.latex})(t)}return FS}var PS,aM;function zp(){if(aM)return PS;aM=1;var e=Wn();PS=t,t.displayName="php",t.aliases=[];function t(n){n.register(e),function(r){var a=/\/\*[\s\S]*?\*\/|\/\/.*|#(?!\[).*/,o=[{pattern:/\b(?:false|true)\b/i,alias:"boolean"},{pattern:/(::\s*)\b[a-z_]\w*\b(?!\s*\()/i,greedy:!0,lookbehind:!0},{pattern:/(\b(?:case|const)\s+)\b[a-z_]\w*(?=\s*[;=])/i,greedy:!0,lookbehind:!0},/\b(?:null)\b/i,/\b[A-Z_][A-Z0-9_]*\b(?!\s*\()/],s=/\b0b[01]+(?:_[01]+)*\b|\b0o[0-7]+(?:_[0-7]+)*\b|\b0x[\da-f]+(?:_[\da-f]+)*\b|(?:\b\d+(?:_\d+)*\.?(?:\d+(?:_\d+)*)?|\B\.\d+)(?:e[+-]?\d+)?/i,u=/<?=>|\?\?=?|\.{3}|\??->|[!=]=?=?|::|\*\*=?|--|\+\+|&&|\|\||<<|>>|[?~]|[/^|%*&<>.+-]=?/,c=/[{}\[\](),:;]/;r.languages.php={delimiter:{pattern:/\?>$|^<\?(?:php(?=\s)|=)?/i,alias:"important"},comment:a,variable:/\$+(?:\w+\b|(?=\{))/,package:{pattern:/(namespace\s+|use\s+(?:function\s+)?)(?:\\?\b[a-z_]\w*)+\b(?!\\)/i,lookbehind:!0,inside:{punctuation:/\\/}},"class-name-definition":{pattern:/(\b(?:class|enum|interface|trait)\s+)\b[a-z_]\w*(?!\\)\b/i,lookbehind:!0,alias:"class-name"},"function-definition":{pattern:/(\bfunction\s+)[a-z_]\w*(?=\s*\()/i,lookbehind:!0,alias:"function"},keyword:[{pattern:/(\(\s*)\b(?:array|bool|boolean|float|int|integer|object|string)\b(?=\s*\))/i,alias:"type-casting",greedy:!0,lookbehind:!0},{pattern:/([(,?]\s*)\b(?:array(?!\s*\()|bool|callable|(?:false|null)(?=\s*\|)|float|int|iterable|mixed|object|self|static|string)\b(?=\s*\$)/i,alias:"type-hint",greedy:!0,lookbehind:!0},{pattern:/(\)\s*:\s*(?:\?\s*)?)\b(?:array(?!\s*\()|bool|callable|(?:false|null)(?=\s*\|)|float|int|iterable|mixed|object|self|static|string|void)\b/i,alias:"return-type",greedy:!0,lookbehind:!0},{pattern:/\b(?:array(?!\s*\()|bool|float|int|iterable|mixed|object|string|void)\b/i,alias:"type-declaration",greedy:!0},{pattern:/(\|\s*)(?:false|null)\b|\b(?:false|null)(?=\s*\|)/i,alias:"type-declaration",greedy:!0,lookbehind:!0},{pattern:/\b(?:parent|self|static)(?=\s*::)/i,alias:"static-context",greedy:!0},{pattern:/(\byield\s+)from\b/i,lookbehind:!0},/\bclass\b/i,{pattern:/((?:^|[^\s>:]|(?:^|[^-])>|(?:^|[^:]):)\s*)\b(?:abstract|and|array|as|break|callable|case|catch|clone|const|continue|declare|default|die|do|echo|else|elseif|empty|enddeclare|endfor|endforeach|endif|endswitch|endwhile|enum|eval|exit|extends|final|finally|fn|for|foreach|function|global|goto|if|implements|include|include_once|instanceof|insteadof|interface|isset|list|match|namespace|new|or|parent|print|private|protected|public|require|require_once|return|self|static|switch|throw|trait|try|unset|use|var|while|xor|yield|__halt_compiler)\b/i,lookbehind:!0}],"argument-name":{pattern:/([(,]\s+)\b[a-z_]\w*(?=\s*:(?!:))/i,lookbehind:!0},"class-name":[{pattern:/(\b(?:extends|implements|instanceof|new(?!\s+self|\s+static))\s+|\bcatch\s*\()\b[a-z_]\w*(?!\\)\b/i,greedy:!0,lookbehind:!0},{pattern:/(\|\s*)\b[a-z_]\w*(?!\\)\b/i,greedy:!0,lookbehind:!0},{pattern:/\b[a-z_]\w*(?!\\)\b(?=\s*\|)/i,greedy:!0},{pattern:/(\|\s*)(?:\\?\b[a-z_]\w*)+\b/i,alias:"class-name-fully-qualified",greedy:!0,lookbehind:!0,inside:{punctuation:/\\/}},{pattern:/(?:\\?\b[a-z_]\w*)+\b(?=\s*\|)/i,alias:"class-name-fully-qualified",greedy:!0,inside:{punctuation:/\\/}},{pattern:/(\b(?:extends|implements|instanceof|new(?!\s+self\b|\s+static\b))\s+|\bcatch\s*\()(?:\\?\b[a-z_]\w*)+\b(?!\\)/i,alias:"class-name-fully-qualified",greedy:!0,lookbehind:!0,inside:{punctuation:/\\/}},{pattern:/\b[a-z_]\w*(?=\s*\$)/i,alias:"type-declaration",greedy:!0},{pattern:/(?:\\?\b[a-z_]\w*)+(?=\s*\$)/i,alias:["class-name-fully-qualified","type-declaration"],greedy:!0,inside:{punctuation:/\\/}},{pattern:/\b[a-z_]\w*(?=\s*::)/i,alias:"static-context",greedy:!0},{pattern:/(?:\\?\b[a-z_]\w*)+(?=\s*::)/i,alias:["class-name-fully-qualified","static-context"],greedy:!0,inside:{punctuation:/\\/}},{pattern:/([(,?]\s*)[a-z_]\w*(?=\s*\$)/i,alias:"type-hint",greedy:!0,lookbehind:!0},{pattern:/([(,?]\s*)(?:\\?\b[a-z_]\w*)+(?=\s*\$)/i,alias:["class-name-fully-qualified","type-hint"],greedy:!0,lookbehind:!0,inside:{punctuation:/\\/}},{pattern:/(\)\s*:\s*(?:\?\s*)?)\b[a-z_]\w*(?!\\)\b/i,alias:"return-type",greedy:!0,lookbehind:!0},{pattern:/(\)\s*:\s*(?:\?\s*)?)(?:\\?\b[a-z_]\w*)+\b(?!\\)/i,alias:["class-name-fully-qualified","return-type"],greedy:!0,lookbehind:!0,inside:{punctuation:/\\/}}],constant:o,function:{pattern:/(^|[^\\\w])\\?[a-z_](?:[\w\\]*\w)?(?=\s*\()/i,lookbehind:!0,inside:{punctuation:/\\/}},property:{pattern:/(->\s*)\w+/,lookbehind:!0},number:s,operator:u,punctuation:c};var d={pattern:/\{\$(?:\{(?:\{[^{}]+\}|[^{}]+)\}|[^{}])+\}|(^|[^\\{])\$+(?:\w+(?:\[[^\r\n\[\]]+\]|->\w+)?)/,lookbehind:!0,inside:r.languages.php},p=[{pattern:/<<<'([^']+)'[\r\n](?:.*[\r\n])*?\1;/,alias:"nowdoc-string",greedy:!0,inside:{delimiter:{pattern:/^<<<'[^']+'|[a-z_]\w*;$/i,alias:"symbol",inside:{punctuation:/^<<<'?|[';]$/}}}},{pattern:/<<<(?:"([^"]+)"[\r\n](?:.*[\r\n])*?\1;|([a-z_]\w*)[\r\n](?:.*[\r\n])*?\2;)/i,alias:"heredoc-string",greedy:!0,inside:{delimiter:{pattern:/^<<<(?:"[^"]+"|[a-z_]\w*)|[a-z_]\w*;$/i,alias:"symbol",inside:{punctuation:/^<<<"?|[";]$/}},interpolation:d}},{pattern:/`(?:\\[\s\S]|[^\\`])*`/,alias:"backtick-quoted-string",greedy:!0},{pattern:/'(?:\\[\s\S]|[^\\'])*'/,alias:"single-quoted-string",greedy:!0},{pattern:/"(?:\\[\s\S]|[^\\"])*"/,alias:"double-quoted-string",greedy:!0,inside:{interpolation:d}}];r.languages.insertBefore("php","variable",{string:p,attribute:{pattern:/#\[(?:[^"'\/#]|\/(?![*/])|\/\/.*$|#(?!\[).*$|\/\*(?:[^*]|\*(?!\/))*\*\/|"(?:\\[\s\S]|[^\\"])*"|'(?:\\[\s\S]|[^\\'])*')+\](?=\s*[a-z$#])/im,greedy:!0,inside:{"attribute-content":{pattern:/^(#\[)[\s\S]+(?=\]$)/,lookbehind:!0,inside:{comment:a,string:p,"attribute-class-name":[{pattern:/([^:]|^)\b[a-z_]\w*(?!\\)\b/i,alias:"class-name",greedy:!0,lookbehind:!0},{pattern:/([^:]|^)(?:\\?\b[a-z_]\w*)+/i,alias:["class-name","class-name-fully-qualified"],greedy:!0,lookbehind:!0,inside:{punctuation:/\\/}}],constant:o,number:s,operator:u,punctuation:c}},delimiter:{pattern:/^#\[|\]$/,alias:"punctuation"}}}}),r.hooks.add("before-tokenize",function(g){if(/<\?/.test(g.code)){var m=/<\?(?:[^"'/#]|\/(?![*/])|("|')(?:\\[\s\S]|(?!\1)[^\\])*\1|(?:\/\/|#(?!\[))(?:[^?\n\r]|\?(?!>))*(?=$|\?>|[\r\n])|#\[|\/\*(?:[^*]|\*(?!\/))*(?:\*\/|$))*?(?:\?>|$)/g;r.languages["markup-templating"].buildPlaceholders(g,"php",m)}}),r.hooks.add("after-tokenize",function(g){r.languages["markup-templating"].tokenizePlaceholders(g,"php")})}(n)}return PS}var zS,oM;function nve(){if(oM)return zS;oM=1;var e=Wn(),t=zp();zS=n,n.displayName="latte",n.aliases=[];function n(r){r.register(e),r.register(t),function(a){a.languages.latte={comment:/^\{\*[\s\S]*/,"latte-tag":{pattern:/(^\{(?:\/(?=[a-z]))?)(?:[=_]|[a-z]\w*\b(?!\())/i,lookbehind:!0,alias:"important"},delimiter:{pattern:/^\{\/?|\}$/,alias:"punctuation"},php:{pattern:/\S(?:[\s\S]*\S)?/,alias:"language-php",inside:a.languages.php}};var o=a.languages.extend("markup",{});a.languages.insertBefore("inside","attr-value",{"n-attr":{pattern:/n:[\w-]+(?:\s*=\s*(?:"[^"]*"|'[^']*'|[^\s'">=]+))?/,inside:{"attr-name":{pattern:/^[^\s=]+/,alias:"important"},"attr-value":{pattern:/=[\s\S]+/,inside:{punctuation:[/^=/,{pattern:/^(\s*)["']|["']$/,lookbehind:!0}],php:{pattern:/\S(?:[\s\S]*\S)?/,inside:a.languages.php}}}}}},o.tag),a.hooks.add("before-tokenize",function(s){if(s.language==="latte"){var u=/\{\*[\s\S]*?\*\}|\{[^'"\s{}*](?:[^"'/{}]|\/(?![*/])|("|')(?:\\[\s\S]|(?!\1)[^\\])*\1|\/\*(?:[^*]|\*(?!\/))*\*\/)*\}/g;a.languages["markup-templating"].buildPlaceholders(s,"latte",u),s.grammar=o}}),a.hooks.add("after-tokenize",function(s){a.languages["markup-templating"].tokenizePlaceholders(s,"latte")})}(r)}return zS}var BS,iM;function rve(){if(iM)return BS;iM=1,BS=e,e.displayName="less",e.aliases=[];function e(t){t.languages.less=t.languages.extend("css",{comment:[/\/\*[\s\S]*?\*\//,{pattern:/(^|[^\\])\/\/.*/,lookbehind:!0}],atrule:{pattern:/@[\w-](?:\((?:[^(){}]|\([^(){}]*\))*\)|[^(){};\s]|\s+(?!\s))*?(?=\s*\{)/,inside:{punctuation:/[:()]/}},selector:{pattern:/(?:@\{[\w-]+\}|[^{};\s@])(?:@\{[\w-]+\}|\((?:[^(){}]|\([^(){}]*\))*\)|[^(){};@\s]|\s+(?!\s))*?(?=\s*\{)/,inside:{variable:/@+[\w-]+/}},property:/(?:@\{[\w-]+\}|[\w-])+(?:\+_?)?(?=\s*:)/,operator:/[+\-*\/]/}),t.languages.insertBefore("less","property",{variable:[{pattern:/@[\w-]+\s*:/,inside:{punctuation:/:/}},/@@?[\w-]+/],"mixin-usage":{pattern:/([{;]\s*)[.#](?!\d)[\w-].*?(?=[(;])/,lookbehind:!0,alias:"function"}})}return BS}var US,sM;function gA(){if(sM)return US;sM=1,US=e,e.displayName="scheme",e.aliases=[];function e(t){(function(n){n.languages.scheme={comment:/;.*|#;\s*(?:\((?:[^()]|\([^()]*\))*\)|\[(?:[^\[\]]|\[[^\[\]]*\])*\])|#\|(?:[^#|]|#(?!\|)|\|(?!#)|#\|(?:[^#|]|#(?!\|)|\|(?!#))*\|#)*\|#/,string:{pattern:/"(?:[^"\\]|\\.)*"/,greedy:!0},symbol:{pattern:/'[^()\[\]#'\s]+/,greedy:!0},char:{pattern:/#\\(?:[ux][a-fA-F\d]+\b|[-a-zA-Z]+\b|[\uD800-\uDBFF][\uDC00-\uDFFF]|\S)/,greedy:!0},"lambda-parameter":[{pattern:/((?:^|[^'`#])[(\[]lambda\s+)(?:[^|()\[\]'\s]+|\|(?:[^\\|]|\\.)*\|)/,lookbehind:!0},{pattern:/((?:^|[^'`#])[(\[]lambda\s+[(\[])[^()\[\]']+/,lookbehind:!0}],keyword:{pattern:/((?:^|[^'`#])[(\[])(?:begin|case(?:-lambda)?|cond(?:-expand)?|define(?:-library|-macro|-record-type|-syntax|-values)?|defmacro|delay(?:-force)?|do|else|except|export|guard|if|import|include(?:-ci|-library-declarations)?|lambda|let(?:rec)?(?:-syntax|-values|\*)?|let\*-values|only|parameterize|prefix|(?:quasi-?)?quote|rename|set!|syntax-(?:case|rules)|unless|unquote(?:-splicing)?|when)(?=[()\[\]\s]|$)/,lookbehind:!0},builtin:{pattern:/((?:^|[^'`#])[(\[])(?:abs|and|append|apply|assoc|ass[qv]|binary-port\?|boolean=?\?|bytevector(?:-append|-copy|-copy!|-length|-u8-ref|-u8-set!|\?)?|caar|cadr|call-with-(?:current-continuation|port|values)|call\/cc|car|cdar|cddr|cdr|ceiling|char(?:->integer|-ready\?|\?|<\?|<=\?|=\?|>\?|>=\?)|close-(?:input-port|output-port|port)|complex\?|cons|current-(?:error|input|output)-port|denominator|dynamic-wind|eof-object\??|eq\?|equal\?|eqv\?|error|error-object(?:-irritants|-message|\?)|eval|even\?|exact(?:-integer-sqrt|-integer\?|\?)?|expt|features|file-error\?|floor(?:-quotient|-remainder|\/)?|flush-output-port|for-each|gcd|get-output-(?:bytevector|string)|inexact\??|input-port(?:-open\?|\?)|integer(?:->char|\?)|lcm|length|list(?:->string|->vector|-copy|-ref|-set!|-tail|\?)?|make-(?:bytevector|list|parameter|string|vector)|map|max|member|memq|memv|min|modulo|negative\?|newline|not|null\?|number(?:->string|\?)|numerator|odd\?|open-(?:input|output)-(?:bytevector|string)|or|output-port(?:-open\?|\?)|pair\?|peek-char|peek-u8|port\?|positive\?|procedure\?|quotient|raise|raise-continuable|rational\?|rationalize|read-(?:bytevector|bytevector!|char|error\?|line|string|u8)|real\?|remainder|reverse|round|set-c[ad]r!|square|string(?:->list|->number|->symbol|->utf8|->vector|-append|-copy|-copy!|-fill!|-for-each|-length|-map|-ref|-set!|\?|<\?|<=\?|=\?|>\?|>=\?)?|substring|symbol(?:->string|\?|=\?)|syntax-error|textual-port\?|truncate(?:-quotient|-remainder|\/)?|u8-ready\?|utf8->string|values|vector(?:->list|->string|-append|-copy|-copy!|-fill!|-for-each|-length|-map|-ref|-set!|\?)?|with-exception-handler|write-(?:bytevector|char|string|u8)|zero\?)(?=[()\[\]\s]|$)/,lookbehind:!0},operator:{pattern:/((?:^|[^'`#])[(\[])(?:[-+*%/]|[<>]=?|=>?)(?=[()\[\]\s]|$)/,lookbehind:!0},number:{pattern:RegExp(r({"<ureal dec>":/\d+(?:\/\d+)|(?:\d+(?:\.\d*)?|\.\d+)(?:[esfdl][+-]?\d+)?/.source,"<real dec>":/[+-]?<ureal dec>|[+-](?:inf|nan)\.0/.source,"<imaginary dec>":/[+-](?:<ureal dec>|(?:inf|nan)\.0)?i/.source,"<complex dec>":/<real dec>(?:@<real dec>|<imaginary dec>)?|<imaginary dec>/.source,"<num dec>":/(?:#d(?:#[ei])?|#[ei](?:#d)?)?<complex dec>/.source,"<ureal box>":/[0-9a-f]+(?:\/[0-9a-f]+)?/.source,"<real box>":/[+-]?<ureal box>|[+-](?:inf|nan)\.0/.source,"<imaginary box>":/[+-](?:<ureal box>|(?:inf|nan)\.0)?i/.source,"<complex box>":/<real box>(?:@<real box>|<imaginary box>)?|<imaginary box>/.source,"<num box>":/#[box](?:#[ei])?|(?:#[ei])?#[box]<complex box>/.source,"<number>":/(^|[()\[\]\s])(?:<num dec>|<num box>)(?=[()\[\]\s]|$)/.source}),"i"),lookbehind:!0},boolean:{pattern:/(^|[()\[\]\s])#(?:[ft]|false|true)(?=[()\[\]\s]|$)/,lookbehind:!0},function:{pattern:/((?:^|[^'`#])[(\[])(?:[^|()\[\]'\s]+|\|(?:[^\\|]|\\.)*\|)(?=[()\[\]\s]|$)/,lookbehind:!0},identifier:{pattern:/(^|[()\[\]\s])\|(?:[^\\|]|\\.)*\|(?=[()\[\]\s]|$)/,lookbehind:!0,greedy:!0},punctuation:/[()\[\]']/};function r(a){for(var o in a)a[o]=a[o].replace(/<[\w\s]+>/g,function(s){return"(?:"+a[s].trim()+")"});return a[o]}})(t)}return US}var jS,lM;function ave(){if(lM)return jS;lM=1;var e=gA();jS=t,t.displayName="lilypond",t.aliases=[];function t(n){n.register(e),function(r){for(var a=/\((?:[^();"#\\]|\\[\s\S]|;.*(?!.)|"(?:[^"\\]|\\.)*"|#(?:\{(?:(?!#\})[\s\S])*#\}|[^{])|<expr>)*\)/.source,o=5,s=0;s<o;s++)a=a.replace(/<expr>/g,function(){return a});a=a.replace(/<expr>/g,/[^\s\S]/.source);var u=r.languages.lilypond={comment:/%(?:(?!\{).*|\{[\s\S]*?%\})/,"embedded-scheme":{pattern:RegExp(/(^|[=\s])#(?:"(?:[^"\\]|\\.)*"|[^\s()"]*(?:[^\s()]|<expr>))/.source.replace(/<expr>/g,function(){return a}),"m"),lookbehind:!0,greedy:!0,inside:{scheme:{pattern:/^(#)[\s\S]+$/,lookbehind:!0,alias:"language-scheme",inside:{"embedded-lilypond":{pattern:/#\{[\s\S]*?#\}/,greedy:!0,inside:{punctuation:/^#\{|#\}$/,lilypond:{pattern:/[\s\S]+/,alias:"language-lilypond",inside:null}}},rest:r.languages.scheme}},punctuation:/#/}},string:{pattern:/"(?:[^"\\]|\\.)*"/,greedy:!0},"class-name":{pattern:/(\\new\s+)[\w-]+/,lookbehind:!0},keyword:{pattern:/\\[a-z][-\w]*/i,inside:{punctuation:/^\\/}},operator:/[=|]|<<|>>/,punctuation:{pattern:/(^|[a-z\d])(?:'+|,+|[_^]?-[_^]?(?:[-+^!>._]|(?=\d))|[_^]\.?|[.!])|[{}()[\]<>^~]|\\[()[\]<>\\!]|--|__/,lookbehind:!0},number:/\b\d+(?:\/\d+)?\b/};u["embedded-scheme"].inside.scheme.inside["embedded-lilypond"].inside.lilypond.inside=u,r.languages.ly=u}(n)}return jS}var GS,uM;function ove(){if(uM)return GS;uM=1;var e=Wn();GS=t,t.displayName="liquid",t.aliases=[];function t(n){n.register(e),n.languages.liquid={comment:{pattern:/(^\{%\s*comment\s*%\})[\s\S]+(?=\{%\s*endcomment\s*%\}$)/,lookbehind:!0},delimiter:{pattern:/^\{(?:\{\{|[%\{])-?|-?(?:\}\}|[%\}])\}$/,alias:"punctuation"},string:{pattern:/"[^"]*"|'[^']*'/,greedy:!0},keyword:/\b(?:as|assign|break|(?:end)?(?:capture|case|comment|for|form|if|paginate|raw|style|tablerow|unless)|continue|cycle|decrement|echo|else|elsif|in|include|increment|limit|liquid|offset|range|render|reversed|section|when|with)\b/,object:/\b(?:address|all_country_option_tags|article|block|blog|cart|checkout|collection|color|country|country_option_tags|currency|current_page|current_tags|customer|customer_address|date|discount_allocation|discount_application|external_video|filter|filter_value|font|forloop|fulfillment|generic_file|gift_card|group|handle|image|line_item|link|linklist|localization|location|measurement|media|metafield|model|model_source|order|page|page_description|page_image|page_title|part|policy|product|product_option|recommendations|request|robots|routes|rule|script|search|selling_plan|selling_plan_allocation|selling_plan_group|shipping_method|shop|shop_locale|sitemap|store_availability|tax_line|template|theme|transaction|unit_price_measurement|user_agent|variant|video|video_source)\b/,function:[{pattern:/(\|\s*)\w+/,lookbehind:!0,alias:"filter"},{pattern:/(\.\s*)(?:first|last|size)/,lookbehind:!0}],boolean:/\b(?:false|nil|true)\b/,range:{pattern:/\.\./,alias:"operator"},number:/\b\d+(?:\.\d+)?\b/,operator:/[!=]=|<>|[<>]=?|[|?:=-]|\b(?:and|contains(?=\s)|or)\b/,punctuation:/[.,\[\]()]/,empty:{pattern:/\bempty\b/,alias:"keyword"}},n.hooks.add("before-tokenize",function(r){var a=/\{%\s*comment\s*%\}[\s\S]*?\{%\s*endcomment\s*%\}|\{(?:%[\s\S]*?%|\{\{[\s\S]*?\}\}|\{[\s\S]*?\})\}/g,o=!1;n.languages["markup-templating"].buildPlaceholders(r,"liquid",a,function(s){var u=/^\{%-?\s*(\w+)/.exec(s);if(u){var c=u[1];if(c==="raw"&&!o)return o=!0,!0;if(c==="endraw")return o=!1,!0}return!o})}),n.hooks.add("after-tokenize",function(r){n.languages["markup-templating"].tokenizePlaceholders(r,"liquid")})}return GS}var HS,cM;function ive(){if(cM)return HS;cM=1,HS=e,e.displayName="lisp",e.aliases=[];function e(t){(function(n){function r(S){return RegExp(/(\()/.source+"(?:"+S+")"+/(?=[\s\)])/.source)}function a(S){return RegExp(/([\s([])/.source+"(?:"+S+")"+/(?=[\s)])/.source)}var o=/(?!\d)[-+*/~!@$%^=<>{}\w]+/.source,s="&"+o,u="(\\()",c="(?=\\))",d="(?=\\s)",p=/(?:[^()]|\((?:[^()]|\((?:[^()]|\((?:[^()]|\((?:[^()]|\([^()]*\))*\))*\))*\))*\))*/.source,g={heading:{pattern:/;;;.*/,alias:["comment","title"]},comment:/;.*/,string:{pattern:/"(?:[^"\\]|\\.)*"/,greedy:!0,inside:{argument:/[-A-Z]+(?=[.,\s])/,symbol:RegExp("`"+o+"'")}},"quoted-symbol":{pattern:RegExp("#?'"+o),alias:["variable","symbol"]},"lisp-property":{pattern:RegExp(":"+o),alias:"property"},splice:{pattern:RegExp(",@?"+o),alias:["symbol","variable"]},keyword:[{pattern:RegExp(u+"(?:and|(?:cl-)?letf|cl-loop|cond|cons|error|if|(?:lexical-)?let\\*?|message|not|null|or|provide|require|setq|unless|use-package|when|while)"+d),lookbehind:!0},{pattern:RegExp(u+"(?:append|by|collect|concat|do|finally|for|in|return)"+d),lookbehind:!0}],declare:{pattern:r(/declare/.source),lookbehind:!0,alias:"keyword"},interactive:{pattern:r(/interactive/.source),lookbehind:!0,alias:"keyword"},boolean:{pattern:a(/nil|t/.source),lookbehind:!0},number:{pattern:a(/[-+]?\d+(?:\.\d*)?/.source),lookbehind:!0},defvar:{pattern:RegExp(u+"def(?:const|custom|group|var)\\s+"+o),lookbehind:!0,inside:{keyword:/^def[a-z]+/,variable:RegExp(o)}},defun:{pattern:RegExp(u+/(?:cl-)?(?:defmacro|defun\*?)\s+/.source+o+/\s+\(/.source+p+/\)/.source),lookbehind:!0,greedy:!0,inside:{keyword:/^(?:cl-)?def\S+/,arguments:null,function:{pattern:RegExp("(^\\s)"+o),lookbehind:!0},punctuation:/[()]/}},lambda:{pattern:RegExp(u+"lambda\\s+\\(\\s*(?:&?"+o+"(?:\\s+&?"+o+")*\\s*)?\\)"),lookbehind:!0,greedy:!0,inside:{keyword:/^lambda/,arguments:null,punctuation:/[()]/}},car:{pattern:RegExp(u+o),lookbehind:!0},punctuation:[/(?:['`,]?\(|[)\[\]])/,{pattern:/(\s)\.(?=\s)/,lookbehind:!0}]},m={"lisp-marker":RegExp(s),varform:{pattern:RegExp(/\(/.source+o+/\s+(?=\S)/.source+p+/\)/.source),inside:g},argument:{pattern:RegExp(/(^|[\s(])/.source+o),lookbehind:!0,alias:"variable"},rest:g},b="\\S+(?:\\s+\\S+)*",y={pattern:RegExp(u+p+c),lookbehind:!0,inside:{"rest-vars":{pattern:RegExp("&(?:body|rest)\\s+"+b),inside:m},"other-marker-vars":{pattern:RegExp("&(?:aux|optional)\\s+"+b),inside:m},keys:{pattern:RegExp("&key\\s+"+b+"(?:\\s+&allow-other-keys)?"),inside:m},argument:{pattern:RegExp(o),alias:"variable"},punctuation:/[()]/}};g.lambda.inside.arguments=y,g.defun.inside.arguments=n.util.clone(y),g.defun.inside.arguments.inside.sublist=y,n.languages.lisp=g,n.languages.elisp=g,n.languages.emacs=g,n.languages["emacs-lisp"]=g})(t)}return HS}var $S,dM;function sve(){if(dM)return $S;dM=1,$S=e,e.displayName="livescript",e.aliases=[];function e(t){t.languages.livescript={comment:[{pattern:/(^|[^\\])\/\*[\s\S]*?\*\//,lookbehind:!0},{pattern:/(^|[^\\])#.*/,lookbehind:!0}],"interpolated-string":{pattern:/(^|[^"])("""|")(?:\\[\s\S]|(?!\2)[^\\])*\2(?!")/,lookbehind:!0,greedy:!0,inside:{variable:{pattern:/(^|[^\\])#[a-z_](?:-?[a-z]|[\d_])*/m,lookbehind:!0},interpolation:{pattern:/(^|[^\\])#\{[^}]+\}/m,lookbehind:!0,inside:{"interpolation-punctuation":{pattern:/^#\{|\}$/,alias:"variable"}}},string:/[\s\S]+/}},string:[{pattern:/('''|')(?:\\[\s\S]|(?!\1)[^\\])*\1/,greedy:!0},{pattern:/<\[[\s\S]*?\]>/,greedy:!0},/\\[^\s,;\])}]+/],regex:[{pattern:/\/\/(?:\[[^\r\n\]]*\]|\\.|(?!\/\/)[^\\\[])+\/\/[gimyu]{0,5}/,greedy:!0,inside:{comment:{pattern:/(^|[^\\])#.*/,lookbehind:!0}}},{pattern:/\/(?:\[[^\r\n\]]*\]|\\.|[^/\\\r\n\[])+\/[gimyu]{0,5}/,greedy:!0}],keyword:{pattern:/(^|(?!-).)\b(?:break|case|catch|class|const|continue|default|do|else|extends|fallthrough|finally|for(?: ever)?|function|if|implements|it|let|loop|new|null|otherwise|own|return|super|switch|that|then|this|throw|try|unless|until|var|void|when|while|yield)(?!-)\b/m,lookbehind:!0},"keyword-operator":{pattern:/(^|[^-])\b(?:(?:delete|require|typeof)!|(?:and|by|delete|export|from|import(?: all)?|in|instanceof|is(?: not|nt)?|not|of|or|til|to|typeof|with|xor)(?!-)\b)/m,lookbehind:!0,alias:"operator"},boolean:{pattern:/(^|[^-])\b(?:false|no|off|on|true|yes)(?!-)\b/m,lookbehind:!0},argument:{pattern:/(^|(?!\.&\.)[^&])&(?!&)\d*/m,lookbehind:!0,alias:"variable"},number:/\b(?:\d+~[\da-z]+|\d[\d_]*(?:\.\d[\d_]*)?(?:[a-z]\w*)?)/i,identifier:/[a-z_](?:-?[a-z]|[\d_])*/i,operator:[{pattern:/( )\.(?= )/,lookbehind:!0},/\.(?:[=~]|\.\.?)|\.(?:[&|^]|<<|>>>?)\.|:(?:=|:=?)|&&|\|[|>]|<(?:<<?<?|--?!?|~~?!?|[|=?])?|>[>=?]?|-(?:->?|>)?|\+\+?|@@?|%%?|\*\*?|!(?:~?=|--?>|~?~>)?|~(?:~?>|=)?|==?|\^\^?|[\/?]/],punctuation:/[(){}\[\]|.,:;`]/},t.languages.livescript["interpolated-string"].inside.interpolation.inside.rest=t.languages.livescript}return $S}var qS,fM;function lve(){if(fM)return qS;fM=1,qS=e,e.displayName="llvm",e.aliases=[];function e(t){(function(n){n.languages.llvm={comment:/;.*/,string:{pattern:/"[^"]*"/,greedy:!0},boolean:/\b(?:false|true)\b/,variable:/[%@!#](?:(?!\d)(?:[-$.\w]|\\[a-f\d]{2})+|\d+)/i,label:/(?!\d)(?:[-$.\w]|\\[a-f\d]{2})+:/i,type:{pattern:/\b(?:double|float|fp128|half|i[1-9]\d*|label|metadata|ppc_fp128|token|void|x86_fp80|x86_mmx)\b/,alias:"class-name"},keyword:/\b[a-z_][a-z_0-9]*\b/,number:/[+-]?\b\d+(?:\.\d+)?(?:[eE][+-]?\d+)?\b|\b0x[\dA-Fa-f]+\b|\b0xK[\dA-Fa-f]{20}\b|\b0x[ML][\dA-Fa-f]{32}\b|\b0xH[\dA-Fa-f]{4}\b/,punctuation:/[{}[\];(),.!*=<>]/}})(t)}return qS}var VS,pM;function uve(){if(pM)return VS;pM=1,VS=e,e.displayName="log",e.aliases=[];function e(t){t.languages.log={string:{pattern:/"(?:[^"\\\r\n]|\\.)*"|'(?![st] | \w)(?:[^'\\\r\n]|\\.)*'/,greedy:!0},exception:{pattern:/(^|[^\w.])[a-z][\w.]*(?:Error|Exception):.*(?:(?:\r\n?|\n)[ \t]*(?:at[ \t].+|\.{3}.*|Caused by:.*))+(?:(?:\r\n?|\n)[ \t]*\.\.\. .*)?/,lookbehind:!0,greedy:!0,alias:["javastacktrace","language-javastacktrace"],inside:t.languages.javastacktrace||{keyword:/\bat\b/,function:/[a-z_][\w$]*(?=\()/,punctuation:/[.:()]/}},level:[{pattern:/\b(?:ALERT|CRIT|CRITICAL|EMERG|EMERGENCY|ERR|ERROR|FAILURE|FATAL|SEVERE)\b/,alias:["error","important"]},{pattern:/\b(?:WARN|WARNING|WRN)\b/,alias:["warning","important"]},{pattern:/\b(?:DISPLAY|INF|INFO|NOTICE|STATUS)\b/,alias:["info","keyword"]},{pattern:/\b(?:DBG|DEBUG|FINE)\b/,alias:["debug","keyword"]},{pattern:/\b(?:FINER|FINEST|TRACE|TRC|VERBOSE|VRB)\b/,alias:["trace","comment"]}],property:{pattern:/((?:^|[\]|])[ \t]*)[a-z_](?:[\w-]|\b\/\b)*(?:[. ]\(?\w(?:[\w-]|\b\/\b)*\)?)*:(?=\s)/im,lookbehind:!0},separator:{pattern:/(^|[^-+])-{3,}|={3,}|\*{3,}|- - /m,lookbehind:!0,alias:"comment"},url:/\b(?:file|ftp|https?):\/\/[^\s|,;'"]*[^\s|,;'">.]/,email:{pattern:/(^|\s)[-\w+.]+@[a-z][a-z0-9-]*(?:\.[a-z][a-z0-9-]*)+(?=\s)/,lookbehind:!0,alias:"url"},"ip-address":{pattern:/\b(?:\d{1,3}(?:\.\d{1,3}){3})\b/,alias:"constant"},"mac-address":{pattern:/\b[a-f0-9]{2}(?::[a-f0-9]{2}){5}\b/i,alias:"constant"},domain:{pattern:/(^|\s)[a-z][a-z0-9-]*(?:\.[a-z][a-z0-9-]*)*\.[a-z][a-z0-9-]+(?=\s)/,lookbehind:!0,alias:"constant"},uuid:{pattern:/\b[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}\b/i,alias:"constant"},hash:{pattern:/\b(?:[a-f0-9]{32}){1,2}\b/i,alias:"constant"},"file-path":{pattern:/\b[a-z]:[\\/][^\s|,;:(){}\[\]"']+|(^|[\s:\[\](>|])\.{0,2}\/\w[^\s|,;:(){}\[\]"']*/i,lookbehind:!0,greedy:!0,alias:"string"},date:{pattern:RegExp(/\b\d{4}[-/]\d{2}[-/]\d{2}(?:T(?=\d{1,2}:)|(?=\s\d{1,2}:))/.source+"|"+/\b\d{1,4}[-/ ](?:\d{1,2}|Apr|Aug|Dec|Feb|Jan|Jul|Jun|Mar|May|Nov|Oct|Sep)[-/ ]\d{2,4}T?\b/.source+"|"+/\b(?:(?:Fri|Mon|Sat|Sun|Thu|Tue|Wed)(?:\s{1,2}(?:Apr|Aug|Dec|Feb|Jan|Jul|Jun|Mar|May|Nov|Oct|Sep))?|Apr|Aug|Dec|Feb|Jan|Jul|Jun|Mar|May|Nov|Oct|Sep)\s{1,2}\d{1,2}\b/.source,"i"),alias:"number"},time:{pattern:/\b\d{1,2}:\d{1,2}:\d{1,2}(?:[.,:]\d+)?(?:\s?[+-]\d{2}:?\d{2}|Z)?\b/,alias:"number"},boolean:/\b(?:false|null|true)\b/i,number:{pattern:/(^|[^.\w])(?:0x[a-f0-9]+|0o[0-7]+|0b[01]+|v?\d[\da-f]*(?:\.\d+)*(?:e[+-]?\d+)?[a-z]{0,3}\b)\b(?!\.\w)/i,lookbehind:!0},operator:/[;:?<=>~/@!$%&+\-|^(){}*#]/,punctuation:/[\[\].,]/}}return VS}var WS,gM;function cve(){if(gM)return WS;gM=1,WS=e,e.displayName="lolcode",e.aliases=[];function e(t){t.languages.lolcode={comment:[/\bOBTW\s[\s\S]*?\sTLDR\b/,/\bBTW.+/],string:{pattern:/"(?::.|[^":])*"/,inside:{variable:/:\{[^}]+\}/,symbol:[/:\([a-f\d]+\)/i,/:\[[^\]]+\]/,/:[)>o":]/]},greedy:!0},number:/(?:\B-)?(?:\b\d+(?:\.\d*)?|\B\.\d+)/,symbol:{pattern:/(^|\s)(?:A )?(?:BUKKIT|NOOB|NUMBAR|NUMBR|TROOF|YARN)(?=\s|,|$)/,lookbehind:!0,inside:{keyword:/A(?=\s)/}},label:{pattern:/((?:^|\s)(?:IM IN YR|IM OUTTA YR) )[a-zA-Z]\w*/,lookbehind:!0,alias:"string"},function:{pattern:/((?:^|\s)(?:HOW IZ I|I IZ|IZ) )[a-zA-Z]\w*/,lookbehind:!0},keyword:[{pattern:/(^|\s)(?:AN|FOUND YR|GIMMEH|GTFO|HAI|HAS A|HOW IZ I|I HAS A|I IZ|IF U SAY SO|IM IN YR|IM OUTTA YR|IS NOW(?: A)?|ITZ(?: A)?|IZ|KTHX|KTHXBYE|LIEK(?: A)?|MAEK|MEBBE|MKAY|NERFIN|NO WAI|O HAI IM|O RLY\?|OIC|OMG|OMGWTF|R|SMOOSH|SRS|TIL|UPPIN|VISIBLE|WILE|WTF\?|YA RLY|YR)(?=\s|,|$)/,lookbehind:!0},/'Z(?=\s|,|$)/],boolean:{pattern:/(^|\s)(?:FAIL|WIN)(?=\s|,|$)/,lookbehind:!0},variable:{pattern:/(^|\s)IT(?=\s|,|$)/,lookbehind:!0},operator:{pattern:/(^|\s)(?:NOT|BOTH SAEM|DIFFRINT|(?:ALL|ANY|BIGGR|BOTH|DIFF|EITHER|MOD|PRODUKT|QUOSHUNT|SMALLR|SUM|WON) OF)(?=\s|,|$)/,lookbehind:!0},punctuation:/\.{3}|…|,|!/}}return WS}var YS,hM;function dve(){if(hM)return YS;hM=1,YS=e,e.displayName="magma",e.aliases=[];function e(t){t.languages.magma={output:{pattern:/^(>.*(?:\r(?:\n|(?!\n))|\n))(?!>)(?:.+|(?:\r(?:\n|(?!\n))|\n)(?!>).*)(?:(?:\r(?:\n|(?!\n))|\n)(?!>).*)*/m,lookbehind:!0,greedy:!0},comment:{pattern:/\/\/.*|\/\*[\s\S]*?\*\//,greedy:!0},string:{pattern:/(^|[^\\"])"(?:[^\r\n\\"]|\\.)*"/,lookbehind:!0,greedy:!0},keyword:/\b(?:_|adj|and|assert|assert2|assert3|assigned|break|by|case|cat|catch|clear|cmpeq|cmpne|continue|declare|default|delete|diff|div|do|elif|else|end|eq|error|eval|exists|exit|for|forall|forward|fprintf|freeze|function|ge|gt|if|iload|import|in|intrinsic|is|join|le|load|local|lt|meet|mod|ne|not|notadj|notin|notsubset|or|print|printf|procedure|quit|random|read|readi|repeat|require|requirege|requirerange|restore|return|save|sdiff|select|subset|then|time|to|try|until|vprint|vprintf|vtime|when|where|while|xor)\b/,boolean:/\b(?:false|true)\b/,generator:{pattern:/\b[a-z_]\w*(?=\s*<)/i,alias:"class-name"},function:/\b[a-z_]\w*(?=\s*\()/i,number:{pattern:/(^|[^\w.]|\.\.)(?:\d+(?:\.\d*)?|\.\d+)(?:[eE][+-]?\d+)?(?:_[a-z]?)?(?=$|[^\w.]|\.\.)/,lookbehind:!0},operator:/->|[-+*/^~!|#=]|:=|\.\./,punctuation:/[()[\]{}<>,;.:]/}}return YS}var KS,mM;function fve(){if(mM)return KS;mM=1,KS=e,e.displayName="makefile",e.aliases=[];function e(t){t.languages.makefile={comment:{pattern:/(^|[^\\])#(?:\\(?:\r\n|[\s\S])|[^\\\r\n])*/,lookbehind:!0},string:{pattern:/(["'])(?:\\(?:\r\n|[\s\S])|(?!\1)[^\\\r\n])*\1/,greedy:!0},"builtin-target":{pattern:/\.[A-Z][^:#=\s]+(?=\s*:(?!=))/,alias:"builtin"},target:{pattern:/^(?:[^:=\s]|[ \t]+(?![\s:]))+(?=\s*:(?!=))/m,alias:"symbol",inside:{variable:/\$+(?:(?!\$)[^(){}:#=\s]+|(?=[({]))/}},variable:/\$+(?:(?!\$)[^(){}:#=\s]+|\([@*%<^+?][DF]\)|(?=[({]))/,keyword:/-include\b|\b(?:define|else|endef|endif|export|ifn?def|ifn?eq|include|override|private|sinclude|undefine|unexport|vpath)\b/,function:{pattern:/(\()(?:abspath|addsuffix|and|basename|call|dir|error|eval|file|filter(?:-out)?|findstring|firstword|flavor|foreach|guile|if|info|join|lastword|load|notdir|or|origin|patsubst|realpath|shell|sort|strip|subst|suffix|value|warning|wildcard|word(?:list|s)?)(?=[ \t])/,lookbehind:!0},operator:/(?:::|[?:+!])?=|[|@]/,punctuation:/[:;(){}]/}}return KS}var XS,bM;function pve(){if(bM)return XS;bM=1,XS=e,e.displayName="markdown",e.aliases=["md"];function e(t){(function(n){var r=/(?:\\.|[^\\\n\r]|(?:\n|\r\n?)(?![\r\n]))/.source;function a(m){return m=m.replace(/<inner>/g,function(){return r}),RegExp(/((?:^|[^\\])(?:\\{2})*)/.source+"(?:"+m+")")}var o=/(?:\\.|``(?:[^`\r\n]|`(?!`))+``|`[^`\r\n]+`|[^\\|\r\n`])+/.source,s=/\|?__(?:\|__)+\|?(?:(?:\n|\r\n?)|(?![\s\S]))/.source.replace(/__/g,function(){return o}),u=/\|?[ \t]*:?-{3,}:?[ \t]*(?:\|[ \t]*:?-{3,}:?[ \t]*)+\|?(?:\n|\r\n?)/.source;n.languages.markdown=n.languages.extend("markup",{}),n.languages.insertBefore("markdown","prolog",{"front-matter-block":{pattern:/(^(?:\s*[\r\n])?)---(?!.)[\s\S]*?[\r\n]---(?!.)/,lookbehind:!0,greedy:!0,inside:{punctuation:/^---|---$/,"front-matter":{pattern:/\S+(?:\s+\S+)*/,alias:["yaml","language-yaml"],inside:n.languages.yaml}}},blockquote:{pattern:/^>(?:[\t ]*>)*/m,alias:"punctuation"},table:{pattern:RegExp("^"+s+u+"(?:"+s+")*","m"),inside:{"table-data-rows":{pattern:RegExp("^("+s+u+")(?:"+s+")*$"),lookbehind:!0,inside:{"table-data":{pattern:RegExp(o),inside:n.languages.markdown},punctuation:/\|/}},"table-line":{pattern:RegExp("^("+s+")"+u+"$"),lookbehind:!0,inside:{punctuation:/\||:?-{3,}:?/}},"table-header-row":{pattern:RegExp("^"+s+"$"),inside:{"table-header":{pattern:RegExp(o),alias:"important",inside:n.languages.markdown},punctuation:/\|/}}}},code:[{pattern:/((?:^|\n)[ \t]*\n|(?:^|\r\n?)[ \t]*\r\n?)(?: {4}|\t).+(?:(?:\n|\r\n?)(?: {4}|\t).+)*/,lookbehind:!0,alias:"keyword"},{pattern:/^```[\s\S]*?^```$/m,greedy:!0,inside:{"code-block":{pattern:/^(```.*(?:\n|\r\n?))[\s\S]+?(?=(?:\n|\r\n?)^```$)/m,lookbehind:!0},"code-language":{pattern:/^(```).+/,lookbehind:!0},punctuation:/```/}}],title:[{pattern:/\S.*(?:\n|\r\n?)(?:==+|--+)(?=[ \t]*$)/m,alias:"important",inside:{punctuation:/==+$|--+$/}},{pattern:/(^\s*)#.+/m,lookbehind:!0,alias:"important",inside:{punctuation:/^#+|#+$/}}],hr:{pattern:/(^\s*)([*-])(?:[\t ]*\2){2,}(?=\s*$)/m,lookbehind:!0,alias:"punctuation"},list:{pattern:/(^\s*)(?:[*+-]|\d+\.)(?=[\t ].)/m,lookbehind:!0,alias:"punctuation"},"url-reference":{pattern:/!?\[[^\]]+\]:[\t ]+(?:\S+|<(?:\\.|[^>\\])+>)(?:[\t ]+(?:"(?:\\.|[^"\\])*"|'(?:\\.|[^'\\])*'|\((?:\\.|[^)\\])*\)))?/,inside:{variable:{pattern:/^(!?\[)[^\]]+/,lookbehind:!0},string:/(?:"(?:\\.|[^"\\])*"|'(?:\\.|[^'\\])*'|\((?:\\.|[^)\\])*\))$/,punctuation:/^[\[\]!:]|[<>]/},alias:"url"},bold:{pattern:a(/\b__(?:(?!_)<inner>|_(?:(?!_)<inner>)+_)+__\b|\*\*(?:(?!\*)<inner>|\*(?:(?!\*)<inner>)+\*)+\*\*/.source),lookbehind:!0,greedy:!0,inside:{content:{pattern:/(^..)[\s\S]+(?=..$)/,lookbehind:!0,inside:{}},punctuation:/\*\*|__/}},italic:{pattern:a(/\b_(?:(?!_)<inner>|__(?:(?!_)<inner>)+__)+_\b|\*(?:(?!\*)<inner>|\*\*(?:(?!\*)<inner>)+\*\*)+\*/.source),lookbehind:!0,greedy:!0,inside:{content:{pattern:/(^.)[\s\S]+(?=.$)/,lookbehind:!0,inside:{}},punctuation:/[*_]/}},strike:{pattern:a(/(~~?)(?:(?!~)<inner>)+\2/.source),lookbehind:!0,greedy:!0,inside:{content:{pattern:/(^~~?)[\s\S]+(?=\1$)/,lookbehind:!0,inside:{}},punctuation:/~~?/}},"code-snippet":{pattern:/(^|[^\\`])(?:``[^`\r\n]+(?:`[^`\r\n]+)*``(?!`)|`[^`\r\n]+`(?!`))/,lookbehind:!0,greedy:!0,alias:["code","keyword"]},url:{pattern:a(/!?\[(?:(?!\])<inner>)+\](?:\([^\s)]+(?:[\t ]+"(?:\\.|[^"\\])*")?\)|[ \t]?\[(?:(?!\])<inner>)+\])/.source),lookbehind:!0,greedy:!0,inside:{operator:/^!/,content:{pattern:/(^\[)[^\]]+(?=\])/,lookbehind:!0,inside:{}},variable:{pattern:/(^\][ \t]?\[)[^\]]+(?=\]$)/,lookbehind:!0},url:{pattern:/(^\]\()[^\s)]+/,lookbehind:!0},string:{pattern:/(^[ \t]+)"(?:\\.|[^"\\])*"(?=\)$)/,lookbehind:!0}}}}),["url","bold","italic","strike"].forEach(function(m){["url","bold","italic","strike","code-snippet"].forEach(function(b){m!==b&&(n.languages.markdown[m].inside.content.inside[b]=n.languages.markdown[b])})}),n.hooks.add("after-tokenize",function(m){if(m.language!=="markdown"&&m.language!=="md")return;function b(y){if(!(!y||typeof y=="string"))for(var S=0,x=y.length;S<x;S++){var R=y[S];if(R.type!=="code"){b(R.content);continue}var k=R.content[1],A=R.content[3];if(k&&A&&k.type==="code-language"&&A.type==="code-block"&&typeof k.content=="string"){var C=k.content.replace(/\b#/g,"sharp").replace(/\b\+\+/g,"pp");C=(/[a-z][\w-]*/i.exec(C)||[""])[0].toLowerCase();var N="language-"+C;A.alias?typeof A.alias=="string"?A.alias=[A.alias,N]:A.alias.push(N):A.alias=[N]}}}b(m.tokens)}),n.hooks.add("wrap",function(m){if(m.type==="code-block"){for(var b="",y=0,S=m.classes.length;y<S;y++){var x=m.classes[y],R=/language-(.+)/.exec(x);if(R){b=R[1];break}}var k=n.languages[b];if(k)m.content=n.highlight(g(m.content.value),k,b);else if(b&&b!=="none"&&n.plugins.autoloader){var A="md-"+new Date().valueOf()+"-"+Math.floor(Math.random()*1e16);m.attributes.id=A,n.plugins.autoloader.loadLanguages(b,function(){var C=document.getElementById(A);C&&(C.innerHTML=n.highlight(C.textContent,n.languages[b],b))})}}});var c=RegExp(n.languages.markup.tag.pattern.source,"gi"),d={amp:"&",lt:"<",gt:">",quot:'"'},p=String.fromCodePoint||String.fromCharCode;function g(m){var b=m.replace(c,"");return b=b.replace(/&(\w{1,8}|#x?[\da-f]{1,8});/gi,function(y,S){if(S=S.toLowerCase(),S[0]==="#"){var x;return S[1]==="x"?x=parseInt(S.slice(2),16):x=Number(S.slice(1)),p(x)}else{var R=d[S];return R||y}}),b}n.languages.md=n.languages.markdown})(t)}return XS}var ZS,yM;function gve(){if(yM)return ZS;yM=1,ZS=e,e.displayName="matlab",e.aliases=[];function e(t){t.languages.matlab={comment:[/%\{[\s\S]*?\}%/,/%.+/],string:{pattern:/\B'(?:''|[^'\r\n])*'/,greedy:!0},number:/(?:\b\d+(?:\.\d*)?|\B\.\d+)(?:[eE][+-]?\d+)?(?:[ij])?|\b[ij]\b/,keyword:/\b(?:NaN|break|case|catch|continue|else|elseif|end|for|function|if|inf|otherwise|parfor|pause|pi|return|switch|try|while)\b/,function:/\b(?!\d)\w+(?=\s*\()/,operator:/\.?[*^\/\\']|[+\-:@]|[<>=~]=?|&&?|\|\|?/,punctuation:/\.{3}|[.,;\[\](){}!]/}}return ZS}var QS,vM;function hve(){if(vM)return QS;vM=1,QS=e,e.displayName="maxscript",e.aliases=[];function e(t){(function(n){var r=/\b(?:about|and|animate|as|at|attributes|by|case|catch|collect|continue|coordsys|do|else|exit|fn|for|from|function|global|if|in|local|macroscript|mapped|max|not|of|off|on|or|parameters|persistent|plugin|rcmenu|return|rollout|set|struct|then|throw|to|tool|try|undo|utility|when|where|while|with)\b/i;n.languages.maxscript={comment:{pattern:/\/\*[\s\S]*?(?:\*\/|$)|--.*/,greedy:!0},string:{pattern:/(^|[^"\\@])(?:"(?:[^"\\]|\\[\s\S])*"|@"[^"]*")/,lookbehind:!0,greedy:!0},path:{pattern:/\$(?:[\w/\\.*?]|'[^']*')*/,greedy:!0,alias:"string"},"function-call":{pattern:RegExp("((?:"+(/^/.source+"|"+/[;=<>+\-*/^({\[]/.source+"|"+/\b(?:and|by|case|catch|collect|do|else|if|in|not|or|return|then|to|try|where|while|with)\b/.source)+")[ 	]*)(?!"+r.source+")"+/[a-z_]\w*\b/.source+"(?=[ 	]*(?:"+("(?!"+r.source+")"+/[a-z_]/.source+"|"+/\d|-\.?\d/.source+"|"+/[({'"$@#?]/.source)+"))","im"),lookbehind:!0,greedy:!0,alias:"function"},"function-definition":{pattern:/(\b(?:fn|function)\s+)\w+\b/i,lookbehind:!0,alias:"function"},argument:{pattern:/\b[a-z_]\w*(?=:)/i,alias:"attr-name"},keyword:r,boolean:/\b(?:false|true)\b/,time:{pattern:/(^|[^\w.])(?:(?:(?:\d+(?:\.\d*)?|\.\d+)(?:[eEdD][+-]\d+|[LP])?[msft])+|\d+:\d+(?:\.\d*)?)(?![\w.:])/,lookbehind:!0,alias:"number"},number:[{pattern:/(^|[^\w.])(?:(?:\d+(?:\.\d*)?|\.\d+)(?:[eEdD][+-]\d+|[LP])?|0x[a-fA-F0-9]+)(?![\w.:])/,lookbehind:!0},/\b(?:e|pi)\b/],constant:/\b(?:dontcollect|ok|silentValue|undefined|unsupplied)\b/,color:{pattern:/\b(?:black|blue|brown|gray|green|orange|red|white|yellow)\b/i,alias:"constant"},operator:/[-+*/<>=!]=?|[&^?]|#(?!\()/,punctuation:/[()\[\]{}.:,;]|#(?=\()|\\$/m}})(t)}return QS}var JS,SM;function mve(){if(SM)return JS;SM=1,JS=e,e.displayName="mel",e.aliases=[];function e(t){t.languages.mel={comment:/\/\/.*/,code:{pattern:/`(?:\\.|[^\\`\r\n])*`/,greedy:!0,alias:"italic",inside:{delimiter:{pattern:/^`|`$/,alias:"punctuation"}}},string:{pattern:/"(?:\\.|[^\\"\r\n])*"/,greedy:!0},variable:/\$\w+/,number:/\b0x[\da-fA-F]+\b|\b\d+(?:\.\d*)?|\B\.\d+/,flag:{pattern:/-[^\d\W]\w*/,alias:"operator"},keyword:/\b(?:break|case|continue|default|do|else|float|for|global|if|in|int|matrix|proc|return|string|switch|vector|while)\b/,function:/\b\w+(?=\()|\b(?:CBG|HfAddAttractorToAS|HfAssignAS|HfBuildEqualMap|HfBuildFurFiles|HfBuildFurImages|HfCancelAFR|HfConnectASToHF|HfCreateAttractor|HfDeleteAS|HfEditAS|HfPerformCreateAS|HfRemoveAttractorFromAS|HfSelectAttached|HfSelectAttractors|HfUnAssignAS|Mayatomr|about|abs|addAttr|addAttributeEditorNodeHelp|addDynamic|addNewShelfTab|addPP|addPanelCategory|addPrefixToName|advanceToNextDrivenKey|affectedNet|affects|aimConstraint|air|alias|aliasAttr|align|alignCtx|alignCurve|alignSurface|allViewFit|ambientLight|angle|angleBetween|animCone|animCurveEditor|animDisplay|animView|annotate|appendStringArray|applicationName|applyAttrPreset|applyTake|arcLenDimContext|arcLengthDimension|arclen|arrayMapper|art3dPaintCtx|artAttrCtx|artAttrPaintVertexCtx|artAttrSkinPaintCtx|artAttrTool|artBuildPaintMenu|artFluidAttrCtx|artPuttyCtx|artSelectCtx|artSetPaintCtx|artUserPaintCtx|assignCommand|assignInputDevice|assignViewportFactories|attachCurve|attachDeviceAttr|attachSurface|attrColorSliderGrp|attrCompatibility|attrControlGrp|attrEnumOptionMenu|attrEnumOptionMenuGrp|attrFieldGrp|attrFieldSliderGrp|attrNavigationControlGrp|attrPresetEditWin|attributeExists|attributeInfo|attributeMenu|attributeQuery|autoKeyframe|autoPlace|bakeClip|bakeFluidShading|bakePartialHistory|bakeResults|bakeSimulation|basename|basenameEx|batchRender|bessel|bevel|bevelPlus|binMembership|bindSkin|blend2|blendShape|blendShapeEditor|blendShapePanel|blendTwoAttr|blindDataType|boneLattice|boundary|boxDollyCtx|boxZoomCtx|bufferCurve|buildBookmarkMenu|buildKeyframeMenu|button|buttonManip|cacheFile|cacheFileCombine|cacheFileMerge|cacheFileTrack|camera|cameraView|canCreateManip|canvas|capitalizeString|catch|catchQuiet|ceil|changeSubdivComponentDisplayLevel|changeSubdivRegion|channelBox|character|characterMap|characterOutlineEditor|characterize|chdir|checkBox|checkBoxGrp|checkDefaultRenderGlobals|choice|circle|circularFillet|clamp|clear|clearCache|clip|clipEditor|clipEditorCurrentTimeCtx|clipSchedule|clipSchedulerOutliner|clipTrimBefore|closeCurve|closeSurface|cluster|cmdFileOutput|cmdScrollFieldExecuter|cmdScrollFieldReporter|cmdShell|coarsenSubdivSelectionList|collision|color|colorAtPoint|colorEditor|colorIndex|colorIndexSliderGrp|colorSliderButtonGrp|colorSliderGrp|columnLayout|commandEcho|commandLine|commandPort|compactHairSystem|componentEditor|compositingInterop|computePolysetVolume|condition|cone|confirmDialog|connectAttr|connectControl|connectDynamic|connectJoint|connectionInfo|constrain|constrainValue|constructionHistory|container|containsMultibyte|contextInfo|control|convertFromOldLayers|convertIffToPsd|convertLightmap|convertSolidTx|convertTessellation|convertUnit|copyArray|copyFlexor|copyKey|copySkinWeights|cos|cpButton|cpCache|cpClothSet|cpCollision|cpConstraint|cpConvClothToMesh|cpForces|cpGetSolverAttr|cpPanel|cpProperty|cpRigidCollisionFilter|cpSeam|cpSetEdit|cpSetSolverAttr|cpSolver|cpSolverTypes|cpTool|cpUpdateClothUVs|createDisplayLayer|createDrawCtx|createEditor|createLayeredPsdFile|createMotionField|createNewShelf|createNode|createRenderLayer|createSubdivRegion|cross|crossProduct|ctxAbort|ctxCompletion|ctxEditMode|ctxTraverse|currentCtx|currentTime|currentTimeCtx|currentUnit|curve|curveAddPtCtx|curveCVCtx|curveEPCtx|curveEditorCtx|curveIntersect|curveMoveEPCtx|curveOnSurface|curveSketchCtx|cutKey|cycleCheck|cylinder|dagPose|date|defaultLightListCheckBox|defaultNavigation|defineDataServer|defineVirtualDevice|deformer|deg_to_rad|delete|deleteAttr|deleteShadingGroupsAndMaterials|deleteShelfTab|deleteUI|deleteUnusedBrushes|delrandstr|detachCurve|detachDeviceAttr|detachSurface|deviceEditor|devicePanel|dgInfo|dgdirty|dgeval|dgtimer|dimWhen|directKeyCtx|directionalLight|dirmap|dirname|disable|disconnectAttr|disconnectJoint|diskCache|displacementToPoly|displayAffected|displayColor|displayCull|displayLevelOfDetail|displayPref|displayRGBColor|displaySmoothness|displayStats|displayString|displaySurface|distanceDimContext|distanceDimension|doBlur|dolly|dollyCtx|dopeSheetEditor|dot|dotProduct|doubleProfileBirailSurface|drag|dragAttrContext|draggerContext|dropoffLocator|duplicate|duplicateCurve|duplicateSurface|dynCache|dynControl|dynExport|dynExpression|dynGlobals|dynPaintEditor|dynParticleCtx|dynPref|dynRelEdPanel|dynRelEditor|dynamicLoad|editAttrLimits|editDisplayLayerGlobals|editDisplayLayerMembers|editRenderLayerAdjustment|editRenderLayerGlobals|editRenderLayerMembers|editor|editorTemplate|effector|emit|emitter|enableDevice|encodeString|endString|endsWith|env|equivalent|equivalentTol|erf|error|eval|evalDeferred|evalEcho|event|exactWorldBoundingBox|exclusiveLightCheckBox|exec|executeForEachObject|exists|exp|expression|expressionEditorListen|extendCurve|extendSurface|extrude|fcheck|fclose|feof|fflush|fgetline|fgetword|file|fileBrowserDialog|fileDialog|fileExtension|fileInfo|filetest|filletCurve|filter|filterCurve|filterExpand|filterStudioImport|findAllIntersections|findAnimCurves|findKeyframe|findMenuItem|findRelatedSkinCluster|finder|firstParentOf|fitBspline|flexor|floatEq|floatField|floatFieldGrp|floatScrollBar|floatSlider|floatSlider2|floatSliderButtonGrp|floatSliderGrp|floor|flow|fluidCacheInfo|fluidEmitter|fluidVoxelInfo|flushUndo|fmod|fontDialog|fopen|formLayout|format|fprint|frameLayout|fread|freeFormFillet|frewind|fromNativePath|fwrite|gamma|gauss|geometryConstraint|getApplicationVersionAsFloat|getAttr|getClassification|getDefaultBrush|getFileList|getFluidAttr|getInputDeviceRange|getMayaPanelTypes|getModifiers|getPanel|getParticleAttr|getPluginResource|getenv|getpid|glRender|glRenderEditor|globalStitch|gmatch|goal|gotoBindPose|grabColor|gradientControl|gradientControlNoAttr|graphDollyCtx|graphSelectContext|graphTrackCtx|gravity|grid|gridLayout|group|groupObjectsByName|hardenPointCurve|hardware|hardwareRenderPanel|headsUpDisplay|headsUpMessage|help|helpLine|hermite|hide|hilite|hitTest|hotBox|hotkey|hotkeyCheck|hsv_to_rgb|hudButton|hudSlider|hudSliderButton|hwReflectionMap|hwRender|hwRenderLoad|hyperGraph|hyperPanel|hyperShade|hypot|iconTextButton|iconTextCheckBox|iconTextRadioButton|iconTextRadioCollection|iconTextScrollList|iconTextStaticLabel|ikHandle|ikHandleCtx|ikHandleDisplayScale|ikSolver|ikSplineHandleCtx|ikSystem|ikSystemInfo|ikfkDisplayMethod|illustratorCurves|image|imfPlugins|inheritTransform|insertJoint|insertJointCtx|insertKeyCtx|insertKnotCurve|insertKnotSurface|instance|instanceable|instancer|intField|intFieldGrp|intScrollBar|intSlider|intSliderGrp|interToUI|internalVar|intersect|iprEngine|isAnimCurve|isConnected|isDirty|isParentOf|isSameObject|isTrue|isValidObjectName|isValidString|isValidUiName|isolateSelect|itemFilter|itemFilterAttr|itemFilterRender|itemFilterType|joint|jointCluster|jointCtx|jointDisplayScale|jointLattice|keyTangent|keyframe|keyframeOutliner|keyframeRegionCurrentTimeCtx|keyframeRegionDirectKeyCtx|keyframeRegionDollyCtx|keyframeRegionInsertKeyCtx|keyframeRegionMoveKeyCtx|keyframeRegionScaleKeyCtx|keyframeRegionSelectKeyCtx|keyframeRegionSetKeyCtx|keyframeRegionTrackCtx|keyframeStats|lassoContext|lattice|latticeDeformKeyCtx|launch|launchImageEditor|layerButton|layeredShaderPort|layeredTexturePort|layout|layoutDialog|lightList|lightListEditor|lightListPanel|lightlink|lineIntersection|linearPrecision|linstep|listAnimatable|listAttr|listCameras|listConnections|listDeviceAttachments|listHistory|listInputDeviceAxes|listInputDeviceButtons|listInputDevices|listMenuAnnotation|listNodeTypes|listPanelCategories|listRelatives|listSets|listTransforms|listUnselected|listerEditor|loadFluid|loadNewShelf|loadPlugin|loadPluginLanguageResources|loadPrefObjects|localizedPanelLabel|lockNode|loft|log|longNameOf|lookThru|ls|lsThroughFilter|lsType|lsUI|mag|makeIdentity|makeLive|makePaintable|makeRoll|makeSingleSurface|makeTubeOn|makebot|manipMoveContext|manipMoveLimitsCtx|manipOptions|manipRotateContext|manipRotateLimitsCtx|manipScaleContext|manipScaleLimitsCtx|marker|match|max|memory|menu|menuBarLayout|menuEditor|menuItem|menuItemToShelf|menuSet|menuSetPref|messageLine|min|minimizeApp|mirrorJoint|modelCurrentTimeCtx|modelEditor|modelPanel|mouse|movIn|movOut|move|moveIKtoFK|moveKeyCtx|moveVertexAlongDirection|multiProfileBirailSurface|mute|nParticle|nameCommand|nameField|namespace|namespaceInfo|newPanelItems|newton|nodeCast|nodeIconButton|nodeOutliner|nodePreset|nodeType|noise|nonLinear|normalConstraint|normalize|nurbsBoolean|nurbsCopyUVSet|nurbsCube|nurbsEditUV|nurbsPlane|nurbsSelect|nurbsSquare|nurbsToPoly|nurbsToPolygonsPref|nurbsToSubdiv|nurbsToSubdivPref|nurbsUVSet|nurbsViewDirectionVector|objExists|objectCenter|objectLayer|objectType|objectTypeUI|obsoleteProc|oceanNurbsPreviewPlane|offsetCurve|offsetCurveOnSurface|offsetSurface|openGLExtension|openMayaPref|optionMenu|optionMenuGrp|optionVar|orbit|orbitCtx|orientConstraint|outlinerEditor|outlinerPanel|overrideModifier|paintEffectsDisplay|pairBlend|palettePort|paneLayout|panel|panelConfiguration|panelHistory|paramDimContext|paramDimension|paramLocator|parent|parentConstraint|particle|particleExists|particleInstancer|particleRenderInfo|partition|pasteKey|pathAnimation|pause|pclose|percent|performanceOptions|pfxstrokes|pickWalk|picture|pixelMove|planarSrf|plane|play|playbackOptions|playblast|plugAttr|plugNode|pluginInfo|pluginResourceUtil|pointConstraint|pointCurveConstraint|pointLight|pointMatrixMult|pointOnCurve|pointOnSurface|pointPosition|poleVectorConstraint|polyAppend|polyAppendFacetCtx|polyAppendVertex|polyAutoProjection|polyAverageNormal|polyAverageVertex|polyBevel|polyBlendColor|polyBlindData|polyBoolOp|polyBridgeEdge|polyCacheMonitor|polyCheck|polyChipOff|polyClipboard|polyCloseBorder|polyCollapseEdge|polyCollapseFacet|polyColorBlindData|polyColorDel|polyColorPerVertex|polyColorSet|polyCompare|polyCone|polyCopyUV|polyCrease|polyCreaseCtx|polyCreateFacet|polyCreateFacetCtx|polyCube|polyCut|polyCutCtx|polyCylinder|polyCylindricalProjection|polyDelEdge|polyDelFacet|polyDelVertex|polyDuplicateAndConnect|polyDuplicateEdge|polyEditUV|polyEditUVShell|polyEvaluate|polyExtrudeEdge|polyExtrudeFacet|polyExtrudeVertex|polyFlipEdge|polyFlipUV|polyForceUV|polyGeoSampler|polyHelix|polyInfo|polyInstallAction|polyLayoutUV|polyListComponentConversion|polyMapCut|polyMapDel|polyMapSew|polyMapSewMove|polyMergeEdge|polyMergeEdgeCtx|polyMergeFacet|polyMergeFacetCtx|polyMergeUV|polyMergeVertex|polyMirrorFace|polyMoveEdge|polyMoveFacet|polyMoveFacetUV|polyMoveUV|polyMoveVertex|polyNormal|polyNormalPerVertex|polyNormalizeUV|polyOptUvs|polyOptions|polyOutput|polyPipe|polyPlanarProjection|polyPlane|polyPlatonicSolid|polyPoke|polyPrimitive|polyPrism|polyProjection|polyPyramid|polyQuad|polyQueryBlindData|polyReduce|polySelect|polySelectConstraint|polySelectConstraintMonitor|polySelectCtx|polySelectEditCtx|polySeparate|polySetToFaceNormal|polySewEdge|polyShortestPathCtx|polySmooth|polySoftEdge|polySphere|polySphericalProjection|polySplit|polySplitCtx|polySplitEdge|polySplitRing|polySplitVertex|polyStraightenUVBorder|polySubdivideEdge|polySubdivideFacet|polyToSubdiv|polyTorus|polyTransfer|polyTriangulate|polyUVSet|polyUnite|polyWedgeFace|popen|popupMenu|pose|pow|preloadRefEd|print|progressBar|progressWindow|projFileViewer|projectCurve|projectTangent|projectionContext|projectionManip|promptDialog|propModCtx|propMove|psdChannelOutliner|psdEditTextureFile|psdExport|psdTextureFile|putenv|pwd|python|querySubdiv|quit|rad_to_deg|radial|radioButton|radioButtonGrp|radioCollection|radioMenuItemCollection|rampColorPort|rand|randomizeFollicles|randstate|rangeControl|readTake|rebuildCurve|rebuildSurface|recordAttr|recordDevice|redo|reference|referenceEdit|referenceQuery|refineSubdivSelectionList|refresh|refreshAE|registerPluginResource|rehash|reloadImage|removeJoint|removeMultiInstance|removePanelCategory|rename|renameAttr|renameSelectionList|renameUI|render|renderGlobalsNode|renderInfo|renderLayerButton|renderLayerParent|renderLayerPostProcess|renderLayerUnparent|renderManip|renderPartition|renderQualityNode|renderSettings|renderThumbnailUpdate|renderWindowEditor|renderWindowSelectContext|renderer|reorder|reorderDeformers|requires|reroot|resampleFluid|resetAE|resetPfxToPolyCamera|resetTool|resolutionNode|retarget|reverseCurve|reverseSurface|revolve|rgb_to_hsv|rigidBody|rigidSolver|roll|rollCtx|rootOf|rot|rotate|rotationInterpolation|roundConstantRadius|rowColumnLayout|rowLayout|runTimeCommand|runup|sampleImage|saveAllShelves|saveAttrPreset|saveFluid|saveImage|saveInitialState|saveMenu|savePrefObjects|savePrefs|saveShelf|saveToolSettings|scale|scaleBrushBrightness|scaleComponents|scaleConstraint|scaleKey|scaleKeyCtx|sceneEditor|sceneUIReplacement|scmh|scriptCtx|scriptEditorInfo|scriptJob|scriptNode|scriptTable|scriptToShelf|scriptedPanel|scriptedPanelType|scrollField|scrollLayout|sculpt|searchPathArray|seed|selLoadSettings|select|selectContext|selectCurveCV|selectKey|selectKeyCtx|selectKeyframeRegionCtx|selectMode|selectPref|selectPriority|selectType|selectedNodes|selectionConnection|separator|setAttr|setAttrEnumResource|setAttrMapping|setAttrNiceNameResource|setConstraintRestPosition|setDefaultShadingGroup|setDrivenKeyframe|setDynamic|setEditCtx|setEditor|setFluidAttr|setFocus|setInfinity|setInputDeviceMapping|setKeyCtx|setKeyPath|setKeyframe|setKeyframeBlendshapeTargetWts|setMenuMode|setNodeNiceNameResource|setNodeTypeFlag|setParent|setParticleAttr|setPfxToPolyCamera|setPluginResource|setProject|setStampDensity|setStartupMessage|setState|setToolTo|setUITemplate|setXformManip|sets|shadingConnection|shadingGeometryRelCtx|shadingLightRelCtx|shadingNetworkCompare|shadingNode|shapeCompare|shelfButton|shelfLayout|shelfTabLayout|shellField|shortNameOf|showHelp|showHidden|showManipCtx|showSelectionInTitle|showShadingGroupAttrEditor|showWindow|sign|simplify|sin|singleProfileBirailSurface|size|sizeBytes|skinCluster|skinPercent|smoothCurve|smoothTangentSurface|smoothstep|snap2to2|snapKey|snapMode|snapTogetherCtx|snapshot|soft|softMod|softModCtx|sort|sound|soundControl|source|spaceLocator|sphere|sphrand|spotLight|spotLightPreviewPort|spreadSheetEditor|spring|sqrt|squareSurface|srtContext|stackTrace|startString|startsWith|stitchAndExplodeShell|stitchSurface|stitchSurfacePoints|strcmp|stringArrayCatenate|stringArrayContains|stringArrayCount|stringArrayInsertAtIndex|stringArrayIntersector|stringArrayRemove|stringArrayRemoveAtIndex|stringArrayRemoveDuplicates|stringArrayRemoveExact|stringArrayToString|stringToStringArray|strip|stripPrefixFromName|stroke|subdAutoProjection|subdCleanTopology|subdCollapse|subdDuplicateAndConnect|subdEditUV|subdListComponentConversion|subdMapCut|subdMapSewMove|subdMatchTopology|subdMirror|subdToBlind|subdToPoly|subdTransferUVsToCache|subdiv|subdivCrease|subdivDisplaySmoothness|substitute|substituteAllString|substituteGeometry|substring|surface|surfaceSampler|surfaceShaderList|swatchDisplayPort|switchTable|symbolButton|symbolCheckBox|sysFile|system|tabLayout|tan|tangentConstraint|texLatticeDeformContext|texManipContext|texMoveContext|texMoveUVShellContext|texRotateContext|texScaleContext|texSelectContext|texSelectShortestPathCtx|texSmudgeUVContext|texWinToolCtx|text|textCurves|textField|textFieldButtonGrp|textFieldGrp|textManip|textScrollList|textToShelf|textureDisplacePlane|textureHairColor|texturePlacementContext|textureWindow|threadCount|threePointArcCtx|timeControl|timePort|timerX|toNativePath|toggle|toggleAxis|toggleWindowVisibility|tokenize|tokenizeList|tolerance|tolower|toolButton|toolCollection|toolDropped|toolHasOptions|toolPropertyWindow|torus|toupper|trace|track|trackCtx|transferAttributes|transformCompare|transformLimits|translator|trim|trunc|truncateFluidCache|truncateHairCache|tumble|tumbleCtx|turbulence|twoPointArcCtx|uiRes|uiTemplate|unassignInputDevice|undo|undoInfo|ungroup|uniform|unit|unloadPlugin|untangleUV|untitledFileName|untrim|upAxis|updateAE|userCtx|uvLink|uvSnapshot|validateShelfName|vectorize|view2dToolCtx|viewCamera|viewClipPlane|viewFit|viewHeadOn|viewLookAt|viewManip|viewPlace|viewSet|visor|volumeAxis|vortex|waitCursor|warning|webBrowser|webBrowserPrefs|whatIs|window|windowPref|wire|wireContext|workspace|wrinkle|wrinkleContext|writeTake|xbmLangPathList|xform)\b/,operator:[/\+[+=]?|-[-=]?|&&|\|\||[<>]=|[*\/!=]=?|[%^]/,{pattern:/(^|[^<])<(?!<)/,lookbehind:!0},{pattern:/(^|[^>])>(?!>)/,lookbehind:!0}],punctuation:/<<|>>|[.,:;?\[\](){}]/},t.languages.mel.code.inside.rest=t.languages.mel}return JS}var eE,EM;function bve(){if(EM)return eE;EM=1,eE=e,e.displayName="mermaid",e.aliases=[];function e(t){t.languages.mermaid={comment:{pattern:/%%.*/,greedy:!0},style:{pattern:/^([ \t]*(?:classDef|linkStyle|style)[ \t]+[\w$-]+[ \t]+)\w.*[^\s;]/m,lookbehind:!0,inside:{property:/\b\w[\w-]*(?=[ \t]*:)/,operator:/:/,punctuation:/,/}},"inter-arrow-label":{pattern:/([^<>ox.=-])(?:-[-.]|==)(?![<>ox.=-])[ \t]*(?:"[^"\r\n]*"|[^\s".=-](?:[^\r\n.=-]*[^\s.=-])?)[ \t]*(?:\.+->?|--+[->]|==+[=>])(?![<>ox.=-])/,lookbehind:!0,greedy:!0,inside:{arrow:{pattern:/(?:\.+->?|--+[->]|==+[=>])$/,alias:"operator"},label:{pattern:/^([\s\S]{2}[ \t]*)\S(?:[\s\S]*\S)?/,lookbehind:!0,alias:"property"},"arrow-head":{pattern:/^\S+/,alias:["arrow","operator"]}}},arrow:[{pattern:/(^|[^{}|o.-])[|}][|o](?:--|\.\.)[|o][|{](?![{}|o.-])/,lookbehind:!0,alias:"operator"},{pattern:/(^|[^<>ox.=-])(?:[<ox](?:==+|--+|-\.*-)[>ox]?|(?:==+|--+|-\.*-)[>ox]|===+|---+|-\.+-)(?![<>ox.=-])/,lookbehind:!0,alias:"operator"},{pattern:/(^|[^<>()x-])(?:--?(?:>>|[x>)])(?![<>()x])|(?:<<|[x<(])--?(?!-))/,lookbehind:!0,alias:"operator"},{pattern:/(^|[^<>|*o.-])(?:[*o]--|--[*o]|<\|?(?:--|\.\.)|(?:--|\.\.)\|?>|--|\.\.)(?![<>|*o.-])/,lookbehind:!0,alias:"operator"}],label:{pattern:/(^|[^|<])\|(?:[^\r\n"|]|"[^"\r\n]*")+\|/,lookbehind:!0,greedy:!0,alias:"property"},text:{pattern:/(?:[(\[{]+|\b>)(?:[^\r\n"()\[\]{}]|"[^"\r\n]*")+(?:[)\]}]+|>)/,alias:"string"},string:{pattern:/"[^"\r\n]*"/,greedy:!0},annotation:{pattern:/<<(?:abstract|choice|enumeration|fork|interface|join|service)>>|\[\[(?:choice|fork|join)\]\]/i,alias:"important"},keyword:[{pattern:/(^[ \t]*)(?:action|callback|class|classDef|classDiagram|click|direction|erDiagram|flowchart|gantt|gitGraph|graph|journey|link|linkStyle|pie|requirementDiagram|sequenceDiagram|stateDiagram|stateDiagram-v2|style|subgraph)(?![\w$-])/m,lookbehind:!0,greedy:!0},{pattern:/(^[ \t]*)(?:activate|alt|and|as|autonumber|deactivate|else|end(?:[ \t]+note)?|loop|opt|par|participant|rect|state|note[ \t]+(?:over|(?:left|right)[ \t]+of))(?![\w$-])/im,lookbehind:!0,greedy:!0}],entity:/#[a-z0-9]+;/,operator:{pattern:/(\w[ \t]*)&(?=[ \t]*\w)|:::|:/,lookbehind:!0},punctuation:/[(){};]/}}return eE}var tE,wM;function yve(){if(wM)return tE;wM=1,tE=e,e.displayName="mizar",e.aliases=[];function e(t){t.languages.mizar={comment:/::.+/,keyword:/@proof\b|\b(?:according|aggregate|all|and|antonym|are|as|associativity|assume|asymmetry|attr|be|begin|being|by|canceled|case|cases|clusters?|coherence|commutativity|compatibility|connectedness|consider|consistency|constructors|contradiction|correctness|def|deffunc|define|definitions?|defpred|do|does|end|environ|equals|ex|exactly|existence|for|from|func|given|hence|hereby|holds|idempotence|identity|iff?|implies|involutiveness|irreflexivity|is|it|let|means|mode|non|not|notations?|now|of|or|otherwise|over|per|pred|prefix|projectivity|proof|provided|qua|reconsider|redefine|reduce|reducibility|reflexivity|registrations?|requirements|reserve|sch|schemes?|section|selector|set|sethood|st|struct|such|suppose|symmetry|synonym|take|that|the|then|theorems?|thesis|thus|to|transitivity|uniqueness|vocabular(?:ies|y)|when|where|with|wrt)\b/,parameter:{pattern:/\$(?:10|\d)/,alias:"variable"},variable:/\b\w+(?=:)/,number:/(?:\b|-)\d+\b/,operator:/\.\.\.|->|&|\.?=/,punctuation:/\(#|#\)|[,:;\[\](){}]/}}return tE}var nE,xM;function vve(){if(xM)return nE;xM=1,nE=e,e.displayName="mongodb",e.aliases=[];function e(t){(function(n){var r=["$eq","$gt","$gte","$in","$lt","$lte","$ne","$nin","$and","$not","$nor","$or","$exists","$type","$expr","$jsonSchema","$mod","$regex","$text","$where","$geoIntersects","$geoWithin","$near","$nearSphere","$all","$elemMatch","$size","$bitsAllClear","$bitsAllSet","$bitsAnyClear","$bitsAnySet","$comment","$elemMatch","$meta","$slice","$currentDate","$inc","$min","$max","$mul","$rename","$set","$setOnInsert","$unset","$addToSet","$pop","$pull","$push","$pullAll","$each","$position","$slice","$sort","$bit","$addFields","$bucket","$bucketAuto","$collStats","$count","$currentOp","$facet","$geoNear","$graphLookup","$group","$indexStats","$limit","$listLocalSessions","$listSessions","$lookup","$match","$merge","$out","$planCacheStats","$project","$redact","$replaceRoot","$replaceWith","$sample","$set","$skip","$sort","$sortByCount","$unionWith","$unset","$unwind","$setWindowFields","$abs","$accumulator","$acos","$acosh","$add","$addToSet","$allElementsTrue","$and","$anyElementTrue","$arrayElemAt","$arrayToObject","$asin","$asinh","$atan","$atan2","$atanh","$avg","$binarySize","$bsonSize","$ceil","$cmp","$concat","$concatArrays","$cond","$convert","$cos","$dateFromParts","$dateToParts","$dateFromString","$dateToString","$dayOfMonth","$dayOfWeek","$dayOfYear","$degreesToRadians","$divide","$eq","$exp","$filter","$first","$floor","$function","$gt","$gte","$hour","$ifNull","$in","$indexOfArray","$indexOfBytes","$indexOfCP","$isArray","$isNumber","$isoDayOfWeek","$isoWeek","$isoWeekYear","$last","$last","$let","$literal","$ln","$log","$log10","$lt","$lte","$ltrim","$map","$max","$mergeObjects","$meta","$min","$millisecond","$minute","$mod","$month","$multiply","$ne","$not","$objectToArray","$or","$pow","$push","$radiansToDegrees","$range","$reduce","$regexFind","$regexFindAll","$regexMatch","$replaceOne","$replaceAll","$reverseArray","$round","$rtrim","$second","$setDifference","$setEquals","$setIntersection","$setIsSubset","$setUnion","$size","$sin","$slice","$split","$sqrt","$stdDevPop","$stdDevSamp","$strcasecmp","$strLenBytes","$strLenCP","$substr","$substrBytes","$substrCP","$subtract","$sum","$switch","$tan","$toBool","$toDate","$toDecimal","$toDouble","$toInt","$toLong","$toObjectId","$toString","$toLower","$toUpper","$trim","$trunc","$type","$week","$year","$zip","$count","$dateAdd","$dateDiff","$dateSubtract","$dateTrunc","$getField","$rand","$sampleRate","$setField","$unsetField","$comment","$explain","$hint","$max","$maxTimeMS","$min","$orderby","$query","$returnKey","$showDiskLoc","$natural"],a=["ObjectId","Code","BinData","DBRef","Timestamp","NumberLong","NumberDecimal","MaxKey","MinKey","RegExp","ISODate","UUID"];r=r.map(function(s){return s.replace("$","\\$")});var o="(?:"+r.join("|")+")\\b";n.languages.mongodb=n.languages.extend("javascript",{}),n.languages.insertBefore("mongodb","string",{property:{pattern:/(?:(["'])(?:\\(?:\r\n|[\s\S])|(?!\1)[^\\\r\n])*\1|(?!\s)[_$a-zA-Z\xA0-\uFFFF](?:(?!\s)[$\w\xA0-\uFFFF])*)(?=\s*:)/,greedy:!0,inside:{keyword:RegExp(`^(['"])?`+o+"(?:\\1)?$")}}}),n.languages.mongodb.string.inside={url:{pattern:/https?:\/\/[-\w@:%.+~#=]{1,256}\.[a-z0-9()]{1,6}\b[-\w()@:%+.~#?&/=]*/i,greedy:!0},entity:{pattern:/\b(?:(?:[01]?\d\d?|2[0-4]\d|25[0-5])\.){3}(?:[01]?\d\d?|2[0-4]\d|25[0-5])\b/,greedy:!0}},n.languages.insertBefore("mongodb","constant",{builtin:{pattern:RegExp("\\b(?:"+a.join("|")+")\\b"),alias:"keyword"}})})(t)}return nE}var rE,kM;function Sve(){if(kM)return rE;kM=1,rE=e,e.displayName="monkey",e.aliases=[];function e(t){t.languages.monkey={comment:{pattern:/^#Rem\s[\s\S]*?^#End|'.+/im,greedy:!0},string:{pattern:/"[^"\r\n]*"/,greedy:!0},preprocessor:{pattern:/(^[ \t]*)#.+/m,lookbehind:!0,greedy:!0,alias:"property"},function:/\b\w+(?=\()/,"type-char":{pattern:/\b[?%#$]/,alias:"class-name"},number:{pattern:/((?:\.\.)?)(?:(?:\b|\B-\.?|\B\.)\d+(?:(?!\.\.)\.\d*)?|\$[\da-f]+)/i,lookbehind:!0},keyword:/\b(?:Abstract|Array|Bool|Case|Catch|Class|Const|Continue|Default|Eachin|Else|ElseIf|End|EndIf|Exit|Extends|Extern|False|Field|Final|Float|For|Forever|Function|Global|If|Implements|Import|Inline|Int|Interface|Local|Method|Module|New|Next|Null|Object|Private|Property|Public|Repeat|Return|Select|Self|Step|Strict|String|Super|Then|Throw|To|True|Try|Until|Void|Wend|While)\b/i,operator:/\.\.|<[=>]?|>=?|:?=|(?:[+\-*\/&~|]|\b(?:Mod|Shl|Shr)\b)=?|\b(?:And|Not|Or)\b/i,punctuation:/[.,:;()\[\]]/}}return rE}var aE,TM;function Eve(){if(TM)return aE;TM=1,aE=e,e.displayName="moonscript",e.aliases=["moon"];function e(t){t.languages.moonscript={comment:/--.*/,string:[{pattern:/'[^']*'|\[(=*)\[[\s\S]*?\]\1\]/,greedy:!0},{pattern:/"[^"]*"/,greedy:!0,inside:{interpolation:{pattern:/#\{[^{}]*\}/,inside:{moonscript:{pattern:/(^#\{)[\s\S]+(?=\})/,lookbehind:!0,inside:null},"interpolation-punctuation":{pattern:/#\{|\}/,alias:"punctuation"}}}}}],"class-name":[{pattern:/(\b(?:class|extends)[ \t]+)\w+/,lookbehind:!0},/\b[A-Z]\w*/],keyword:/\b(?:class|continue|do|else|elseif|export|extends|for|from|if|import|in|local|nil|return|self|super|switch|then|unless|using|when|while|with)\b/,variable:/@@?\w*/,property:{pattern:/\b(?!\d)\w+(?=:)|(:)(?!\d)\w+/,lookbehind:!0},function:{pattern:/\b(?:_G|_VERSION|assert|collectgarbage|coroutine\.(?:create|resume|running|status|wrap|yield)|debug\.(?:debug|getfenv|gethook|getinfo|getlocal|getmetatable|getregistry|getupvalue|setfenv|sethook|setlocal|setmetatable|setupvalue|traceback)|dofile|error|getfenv|getmetatable|io\.(?:close|flush|input|lines|open|output|popen|read|stderr|stdin|stdout|tmpfile|type|write)|ipairs|load|loadfile|loadstring|math\.(?:abs|acos|asin|atan|atan2|ceil|cos|cosh|deg|exp|floor|fmod|frexp|ldexp|log|log10|max|min|modf|pi|pow|rad|random|randomseed|sin|sinh|sqrt|tan|tanh)|module|next|os\.(?:clock|date|difftime|execute|exit|getenv|remove|rename|setlocale|time|tmpname)|package\.(?:cpath|loaded|loadlib|path|preload|seeall)|pairs|pcall|print|rawequal|rawget|rawset|require|select|setfenv|setmetatable|string\.(?:byte|char|dump|find|format|gmatch|gsub|len|lower|match|rep|reverse|sub|upper)|table\.(?:concat|insert|maxn|remove|sort)|tonumber|tostring|type|unpack|xpcall)\b/,inside:{punctuation:/\./}},boolean:/\b(?:false|true)\b/,number:/(?:\B\.\d+|\b\d+\.\d+|\b\d+(?=[eE]))(?:[eE][-+]?\d+)?\b|\b(?:0x[a-fA-F\d]+|\d+)(?:U?LL)?\b/,operator:/\.{3}|[-=]>|~=|(?:[-+*/%<>!=]|\.\.)=?|[:#^]|\b(?:and|or)\b=?|\b(?:not)\b/,punctuation:/[.,()[\]{}\\]/},t.languages.moonscript.string[1].inside.interpolation.inside.moonscript.inside=t.languages.moonscript,t.languages.moon=t.languages.moonscript}return aE}var oE,AM;function wve(){if(AM)return oE;AM=1,oE=e,e.displayName="n1ql",e.aliases=[];function e(t){t.languages.n1ql={comment:{pattern:/\/\*[\s\S]*?(?:$|\*\/)|--.*/,greedy:!0},string:{pattern:/(["'])(?:\\[\s\S]|(?!\1)[^\\]|\1\1)*\1/,greedy:!0},identifier:{pattern:/`(?:\\[\s\S]|[^\\`]|``)*`/,greedy:!0},parameter:/\$[\w.]+/,keyword:/\b(?:ADVISE|ALL|ALTER|ANALYZE|AS|ASC|AT|BEGIN|BINARY|BOOLEAN|BREAK|BUCKET|BUILD|BY|CALL|CAST|CLUSTER|COLLATE|COLLECTION|COMMIT|COMMITTED|CONNECT|CONTINUE|CORRELATE|CORRELATED|COVER|CREATE|CURRENT|DATABASE|DATASET|DATASTORE|DECLARE|DECREMENT|DELETE|DERIVED|DESC|DESCRIBE|DISTINCT|DO|DROP|EACH|ELEMENT|EXCEPT|EXCLUDE|EXECUTE|EXPLAIN|FETCH|FILTER|FLATTEN|FLUSH|FOLLOWING|FOR|FORCE|FROM|FTS|FUNCTION|GOLANG|GRANT|GROUP|GROUPS|GSI|HASH|HAVING|IF|IGNORE|ILIKE|INCLUDE|INCREMENT|INDEX|INFER|INLINE|INNER|INSERT|INTERSECT|INTO|IS|ISOLATION|JAVASCRIPT|JOIN|KEY|KEYS|KEYSPACE|KNOWN|LANGUAGE|LAST|LEFT|LET|LETTING|LEVEL|LIMIT|LSM|MAP|MAPPING|MATCHED|MATERIALIZED|MERGE|MINUS|MISSING|NAMESPACE|NEST|NL|NO|NTH_VALUE|NULL|NULLS|NUMBER|OBJECT|OFFSET|ON|OPTION|OPTIONS|ORDER|OTHERS|OUTER|OVER|PARSE|PARTITION|PASSWORD|PATH|POOL|PRECEDING|PREPARE|PRIMARY|PRIVATE|PRIVILEGE|PROBE|PROCEDURE|PUBLIC|RANGE|RAW|REALM|REDUCE|RENAME|RESPECT|RETURN|RETURNING|REVOKE|RIGHT|ROLE|ROLLBACK|ROW|ROWS|SATISFIES|SAVEPOINT|SCHEMA|SCOPE|SELECT|SELF|SEMI|SET|SHOW|SOME|START|STATISTICS|STRING|SYSTEM|TIES|TO|TRAN|TRANSACTION|TRIGGER|TRUNCATE|UNBOUNDED|UNDER|UNION|UNIQUE|UNKNOWN|UNNEST|UNSET|UPDATE|UPSERT|USE|USER|USING|VALIDATE|VALUE|VALUES|VIA|VIEW|WHERE|WHILE|WINDOW|WITH|WORK|XOR)\b/i,function:/\b[a-z_]\w*(?=\s*\()/i,boolean:/\b(?:FALSE|TRUE)\b/i,number:/(?:\b\d+\.|\B\.)\d+e[+\-]?\d+\b|\b\d+(?:\.\d*)?|\B\.\d+\b/i,operator:/[-+*\/%]|!=|==?|\|\||<[>=]?|>=?|\b(?:AND|ANY|ARRAY|BETWEEN|CASE|ELSE|END|EVERY|EXISTS|FIRST|IN|LIKE|NOT|OR|THEN|VALUED|WHEN|WITHIN)\b/i,punctuation:/[;[\](),.{}:]/}}return oE}var iE,RM;function xve(){if(RM)return iE;RM=1,iE=e,e.displayName="n4js",e.aliases=["n4jsd"];function e(t){t.languages.n4js=t.languages.extend("javascript",{keyword:/\b(?:Array|any|boolean|break|case|catch|class|const|constructor|continue|debugger|declare|default|delete|do|else|enum|export|extends|false|finally|for|from|function|get|if|implements|import|in|instanceof|interface|let|module|new|null|number|package|private|protected|public|return|set|static|string|super|switch|this|throw|true|try|typeof|var|void|while|with|yield)\b/}),t.languages.insertBefore("n4js","constant",{annotation:{pattern:/@+\w+/,alias:"operator"}}),t.languages.n4jsd=t.languages.n4js}return iE}var sE,_M;function kve(){if(_M)return sE;_M=1,sE=e,e.displayName="nand2tetrisHdl",e.aliases=[];function e(t){t.languages["nand2tetris-hdl"]={comment:/\/\/.*|\/\*[\s\S]*?(?:\*\/|$)/,keyword:/\b(?:BUILTIN|CHIP|CLOCKED|IN|OUT|PARTS)\b/,boolean:/\b(?:false|true)\b/,function:/\b[A-Za-z][A-Za-z0-9]*(?=\()/,number:/\b\d+\b/,operator:/=|\.\./,punctuation:/[{}[\];(),:]/}}return sE}var lE,CM;function Tve(){if(CM)return lE;CM=1,lE=e,e.displayName="naniscript",e.aliases=[];function e(t){(function(n){var r=/\{[^\r\n\[\]{}]*\}/,a={"quoted-string":{pattern:/"(?:[^"\\]|\\.)*"/,alias:"operator"},"command-param-id":{pattern:/(\s)\w+:/,lookbehind:!0,alias:"property"},"command-param-value":[{pattern:r,alias:"selector"},{pattern:/([\t ])\S+/,lookbehind:!0,greedy:!0,alias:"operator"},{pattern:/\S(?:.*\S)?/,alias:"operator"}]};n.languages.naniscript={comment:{pattern:/^([\t ]*);.*/m,lookbehind:!0},define:{pattern:/^>.+/m,alias:"tag",inside:{value:{pattern:/(^>\w+[\t ]+)(?!\s)[^{}\r\n]+/,lookbehind:!0,alias:"operator"},key:{pattern:/(^>)\w+/,lookbehind:!0}}},label:{pattern:/^([\t ]*)#[\t ]*\w+[\t ]*$/m,lookbehind:!0,alias:"regex"},command:{pattern:/^([\t ]*)@\w+(?=[\t ]|$).*/m,lookbehind:!0,alias:"function",inside:{"command-name":/^@\w+/,expression:{pattern:r,greedy:!0,alias:"selector"},"command-params":{pattern:/\s*\S[\s\S]*/,inside:a}}},"generic-text":{pattern:/(^[ \t]*)[^#@>;\s].*/m,lookbehind:!0,alias:"punctuation",inside:{"escaped-char":/\\[{}\[\]"]/,expression:{pattern:r,greedy:!0,alias:"selector"},"inline-command":{pattern:/\[[\t ]*\w[^\r\n\[\]]*\]/,greedy:!0,alias:"function",inside:{"command-params":{pattern:/(^\[[\t ]*\w+\b)[\s\S]+(?=\]$)/,lookbehind:!0,inside:a},"command-param-name":{pattern:/^(\[[\t ]*)\w+/,lookbehind:!0,alias:"name"},"start-stop-char":/[\[\]]/}}}}},n.languages.nani=n.languages.naniscript,n.hooks.add("after-tokenize",function(u){var c=u.tokens;c.forEach(function(d){if(typeof d!="string"&&d.type==="generic-text"){var p=s(d);o(p)||(d.type="bad-line",d.content=p)}})});function o(u){for(var c="[]{}",d=[],p=0;p<u.length;p++){var g=u[p],m=c.indexOf(g);if(m!==-1){if(m%2===0)d.push(m+1);else if(d.pop()!==m)return!1}}return d.length===0}function s(u){return typeof u=="string"?u:Array.isArray(u)?u.map(s).join(""):s(u.content)}})(t)}return lE}var uE,NM;function Ave(){if(NM)return uE;NM=1,uE=e,e.displayName="nasm",e.aliases=[];function e(t){t.languages.nasm={comment:/;.*$/m,string:/(["'`])(?:\\.|(?!\1)[^\\\r\n])*\1/,label:{pattern:/(^\s*)[A-Za-z._?$][\w.?$@~#]*:/m,lookbehind:!0,alias:"function"},keyword:[/\[?BITS (?:16|32|64)\]?/,{pattern:/(^\s*)section\s*[a-z.]+:?/im,lookbehind:!0},/(?:extern|global)[^;\r\n]*/i,/(?:CPU|DEFAULT|FLOAT).*$/m],register:{pattern:/\b(?:st\d|[xyz]mm\d\d?|[cdt]r\d|r\d\d?[bwd]?|[er]?[abcd]x|[abcd][hl]|[er]?(?:bp|di|si|sp)|[cdefgs]s)\b/i,alias:"variable"},number:/(?:\b|(?=\$))(?:0[hx](?:\.[\da-f]+|[\da-f]+(?:\.[\da-f]+)?)(?:p[+-]?\d+)?|\d[\da-f]+[hx]|\$\d[\da-f]*|0[oq][0-7]+|[0-7]+[oq]|0[by][01]+|[01]+[by]|0[dt]\d+|(?:\d+(?:\.\d+)?|\.\d+)(?:\.?e[+-]?\d+)?[dt]?)\b/i,operator:/[\[\]*+\-\/%<>=&|$!]/}}return uE}var cE,OM;function Rve(){if(OM)return cE;OM=1,cE=e,e.displayName="neon",e.aliases=[];function e(t){t.languages.neon={comment:{pattern:/#.*/,greedy:!0},datetime:{pattern:/(^|[[{(=:,\s])\d\d\d\d-\d\d?-\d\d?(?:(?:[Tt]| +)\d\d?:\d\d:\d\d(?:\.\d*)? *(?:Z|[-+]\d\d?(?::?\d\d)?)?)?(?=$|[\]}),\s])/,lookbehind:!0,alias:"number"},key:{pattern:/(^|[[{(,\s])[^,:=[\]{}()'"\s]+(?=\s*:(?:$|[\]}),\s])|\s*=)/,lookbehind:!0,alias:"atrule"},number:{pattern:/(^|[[{(=:,\s])[+-]?(?:0x[\da-fA-F]+|0o[0-7]+|0b[01]+|(?:\d+(?:\.\d*)?|\.?\d+)(?:[eE][+-]?\d+)?)(?=$|[\]}),:=\s])/,lookbehind:!0},boolean:{pattern:/(^|[[{(=:,\s])(?:false|no|true|yes)(?=$|[\]}),:=\s])/i,lookbehind:!0},null:{pattern:/(^|[[{(=:,\s])(?:null)(?=$|[\]}),:=\s])/i,lookbehind:!0,alias:"keyword"},string:{pattern:/(^|[[{(=:,\s])(?:('''|""")\r?\n(?:(?:[^\r\n]|\r?\n(?![\t ]*\2))*\r?\n)?[\t ]*\2|'[^'\r\n]*'|"(?:\\.|[^\\"\r\n])*")/,lookbehind:!0,greedy:!0},literal:{pattern:/(^|[[{(=:,\s])(?:[^#"',:=[\]{}()\s`-]|[:-][^"',=[\]{}()\s])(?:[^,:=\]})(\s]|:(?![\s,\]})]|$)|[ \t]+[^#,:=\]})(\s])*/,lookbehind:!0,alias:"string"},punctuation:/[,:=[\]{}()-]/}}return cE}var dE,DM;function _ve(){if(DM)return dE;DM=1,dE=e,e.displayName="nevod",e.aliases=[];function e(t){t.languages.nevod={comment:/\/\/.*|(?:\/\*[\s\S]*?(?:\*\/|$))/,string:{pattern:/(?:"(?:""|[^"])*"(?!")|'(?:''|[^'])*'(?!'))!?\*?/,greedy:!0,inside:{"string-attrs":/!$|!\*$|\*$/}},namespace:{pattern:/(@namespace\s+)[a-zA-Z0-9\-.]+(?=\s*\{)/,lookbehind:!0},pattern:{pattern:/(@pattern\s+)?#?[a-zA-Z0-9\-.]+(?:\s*\(\s*(?:~\s*)?[a-zA-Z0-9\-.]+\s*(?:,\s*(?:~\s*)?[a-zA-Z0-9\-.]*)*\))?(?=\s*=)/,lookbehind:!0,inside:{"pattern-name":{pattern:/^#?[a-zA-Z0-9\-.]+/,alias:"class-name"},fields:{pattern:/\(.*\)/,inside:{"field-name":{pattern:/[a-zA-Z0-9\-.]+/,alias:"variable"},punctuation:/[,()]/,operator:{pattern:/~/,alias:"field-hidden-mark"}}}}},search:{pattern:/(@search\s+|#)[a-zA-Z0-9\-.]+(?:\.\*)?(?=\s*;)/,alias:"function",lookbehind:!0},keyword:/@(?:having|inside|namespace|outside|pattern|require|search|where)\b/,"standard-pattern":{pattern:/\b(?:Alpha|AlphaNum|Any|Blank|End|LineBreak|Num|NumAlpha|Punct|Space|Start|Symbol|Word|WordBreak)\b(?:\([a-zA-Z0-9\-.,\s+]*\))?/,inside:{"standard-pattern-name":{pattern:/^[a-zA-Z0-9\-.]+/,alias:"builtin"},quantifier:{pattern:/\b\d+(?:\s*\+|\s*-\s*\d+)?(?!\w)/,alias:"number"},"standard-pattern-attr":{pattern:/[a-zA-Z0-9\-.]+/,alias:"builtin"},punctuation:/[,()]/}},quantifier:{pattern:/\b\d+(?:\s*\+|\s*-\s*\d+)?(?!\w)/,alias:"number"},operator:[{pattern:/=/,alias:"pattern-def"},{pattern:/&/,alias:"conjunction"},{pattern:/~/,alias:"exception"},{pattern:/\?/,alias:"optionality"},{pattern:/[[\]]/,alias:"repetition"},{pattern:/[{}]/,alias:"variation"},{pattern:/[+_]/,alias:"sequence"},{pattern:/\.{2,3}/,alias:"span"}],"field-capture":[{pattern:/([a-zA-Z0-9\-.]+\s*\()\s*[a-zA-Z0-9\-.]+\s*:\s*[a-zA-Z0-9\-.]+(?:\s*,\s*[a-zA-Z0-9\-.]+\s*:\s*[a-zA-Z0-9\-.]+)*(?=\s*\))/,lookbehind:!0,inside:{"field-name":{pattern:/[a-zA-Z0-9\-.]+/,alias:"variable"},colon:/:/}},{pattern:/[a-zA-Z0-9\-.]+\s*:/,inside:{"field-name":{pattern:/[a-zA-Z0-9\-.]+/,alias:"variable"},colon:/:/}}],punctuation:/[:;,()]/,name:/[a-zA-Z0-9\-.]+/}}return dE}var fE,IM;function Cve(){if(IM)return fE;IM=1,fE=e,e.displayName="nginx",e.aliases=[];function e(t){(function(n){var r=/\$(?:\w[a-z\d]*(?:_[^\x00-\x1F\s"'\\()$]*)?|\{[^}\s"'\\]+\})/i;n.languages.nginx={comment:{pattern:/(^|[\s{};])#.*/,lookbehind:!0,greedy:!0},directive:{pattern:/(^|\s)\w(?:[^;{}"'\\\s]|\\.|"(?:[^"\\]|\\.)*"|'(?:[^'\\]|\\.)*'|\s+(?:#.*(?!.)|(?![#\s])))*?(?=\s*[;{])/,lookbehind:!0,greedy:!0,inside:{string:{pattern:/((?:^|[^\\])(?:\\\\)*)(?:"(?:[^"\\]|\\.)*"|'(?:[^'\\]|\\.)*')/,lookbehind:!0,greedy:!0,inside:{escape:{pattern:/\\["'\\nrt]/,alias:"entity"},variable:r}},comment:{pattern:/(\s)#.*/,lookbehind:!0,greedy:!0},keyword:{pattern:/^\S+/,greedy:!0},boolean:{pattern:/(\s)(?:off|on)(?!\S)/,lookbehind:!0},number:{pattern:/(\s)\d+[a-z]*(?!\S)/i,lookbehind:!0},variable:r}},punctuation:/[{};]/}})(t)}return fE}var pE,LM;function Nve(){if(LM)return pE;LM=1,pE=e,e.displayName="nim",e.aliases=[];function e(t){t.languages.nim={comment:{pattern:/#.*/,greedy:!0},string:{pattern:/(?:\b(?!\d)(?:\w|\\x[89a-fA-F][0-9a-fA-F])+)?(?:"""[\s\S]*?"""(?!")|"(?:\\[\s\S]|""|[^"\\])*")/,greedy:!0},char:{pattern:/'(?:\\(?:\d+|x[\da-fA-F]{0,2}|.)|[^'])'/,greedy:!0},function:{pattern:/(?:(?!\d)(?:\w|\\x[89a-fA-F][0-9a-fA-F])+|`[^`\r\n]+`)\*?(?:\[[^\]]+\])?(?=\s*\()/,greedy:!0,inside:{operator:/\*$/}},identifier:{pattern:/`[^`\r\n]+`/,greedy:!0,inside:{punctuation:/`/}},number:/\b(?:0[xXoObB][\da-fA-F_]+|\d[\d_]*(?:(?!\.\.)\.[\d_]*)?(?:[eE][+-]?\d[\d_]*)?)(?:'?[iuf]\d*)?/,keyword:/\b(?:addr|as|asm|atomic|bind|block|break|case|cast|concept|const|continue|converter|defer|discard|distinct|do|elif|else|end|enum|except|export|finally|for|from|func|generic|if|import|include|interface|iterator|let|macro|method|mixin|nil|object|out|proc|ptr|raise|ref|return|static|template|try|tuple|type|using|var|when|while|with|without|yield)\b/,operator:{pattern:/(^|[({\[](?=\.\.)|(?![({\[]\.).)(?:(?:[=+\-*\/<>@$~&%|!?^:\\]|\.\.|\.(?![)}\]]))+|\b(?:and|div|in|is|isnot|mod|not|notin|of|or|shl|shr|xor)\b)/m,lookbehind:!0},punctuation:/[({\[]\.|\.[)}\]]|[`(){}\[\],:]/}}return pE}var gE,MM;function Ove(){if(MM)return gE;MM=1,gE=e,e.displayName="nix",e.aliases=[];function e(t){t.languages.nix={comment:{pattern:/\/\*[\s\S]*?\*\/|#.*/,greedy:!0},string:{pattern:/"(?:[^"\\]|\\[\s\S])*"|''(?:(?!'')[\s\S]|''(?:'|\\|\$\{))*''/,greedy:!0,inside:{interpolation:{pattern:/(^|(?:^|(?!'').)[^\\])\$\{(?:[^{}]|\{[^}]*\})*\}/,lookbehind:!0,inside:null}}},url:[/\b(?:[a-z]{3,7}:\/\/)[\w\-+%~\/.:#=?&]+/,{pattern:/([^\/])(?:[\w\-+%~.:#=?&]*(?!\/\/)[\w\-+%~\/.:#=?&])?(?!\/\/)\/[\w\-+%~\/.:#=?&]*/,lookbehind:!0}],antiquotation:{pattern:/\$(?=\{)/,alias:"important"},number:/\b\d+\b/,keyword:/\b(?:assert|builtins|else|if|in|inherit|let|null|or|then|with)\b/,function:/\b(?:abort|add|all|any|attrNames|attrValues|baseNameOf|compareVersions|concatLists|currentSystem|deepSeq|derivation|dirOf|div|elem(?:At)?|fetch(?:Tarball|url)|filter(?:Source)?|fromJSON|genList|getAttr|getEnv|hasAttr|hashString|head|import|intersectAttrs|is(?:Attrs|Bool|Function|Int|List|Null|String)|length|lessThan|listToAttrs|map|mul|parseDrvName|pathExists|read(?:Dir|File)|removeAttrs|replaceStrings|seq|sort|stringLength|sub(?:string)?|tail|throw|to(?:File|JSON|Path|String|XML)|trace|typeOf)\b|\bfoldl'\B/,boolean:/\b(?:false|true)\b/,operator:/[=!<>]=?|\+\+?|\|\||&&|\/\/|->?|[?@]/,punctuation:/[{}()[\].,:;]/},t.languages.nix.string.inside.interpolation.inside=t.languages.nix}return gE}var hE,FM;function Dve(){if(FM)return hE;FM=1,hE=e,e.displayName="nsis",e.aliases=[];function e(t){t.languages.nsis={comment:{pattern:/(^|[^\\])(?:\/\*[\s\S]*?\*\/|[#;].*)/,lookbehind:!0,greedy:!0},string:{pattern:/("|')(?:\\.|(?!\1)[^\\\r\n])*\1/,greedy:!0},keyword:{pattern:/(^[\t ]*)(?:Abort|Add(?:BrandingImage|Size)|AdvSplash|Allow(?:RootDirInstall|SkipFiles)|AutoCloseWindow|BG(?:Font|Gradient|Image)|Banner|BrandingText|BringToFront|CRCCheck|Call(?:InstDLL)?|Caption|ChangeUI|CheckBitmap|ClearErrors|CompletedText|ComponentText|CopyFiles|Create(?:Directory|Font|ShortCut)|Delete(?:INISec|INIStr|RegKey|RegValue)?|Detail(?:Print|sButtonText)|Dialer|Dir(?:Text|Var|Verify)|EnableWindow|Enum(?:RegKey|RegValue)|Exch|Exec(?:Shell(?:Wait)?|Wait)?|ExpandEnvStrings|File(?:BufSize|Close|ErrorText|Open|Read|ReadByte|ReadUTF16LE|ReadWord|Seek|Write|WriteByte|WriteUTF16LE|WriteWord)?|Find(?:Close|First|Next|Window)|FlushINI|Get(?:CurInstType|CurrentAddress|DLLVersion(?:Local)?|DlgItem|ErrorLevel|FileTime(?:Local)?|FullPathName|Function(?:Address|End)?|InstDirError|LabelAddress|TempFileName)|Goto|HideWindow|Icon|If(?:Abort|Errors|FileExists|RebootFlag|Silent)|InitPluginsDir|InstProgressFlags|Inst(?:Type(?:GetText|SetText)?)|Install(?:ButtonText|Colors|Dir(?:RegKey)?)|Int(?:64|Ptr)?CmpU?|Int(?:64)?Fmt|Int(?:Ptr)?Op|IsWindow|Lang(?:DLL|String)|License(?:BkColor|Data|ForceSelection|LangString|Text)|LoadLanguageFile|LockWindow|Log(?:Set|Text)|Manifest(?:DPIAware|SupportedOS)|Math|MessageBox|MiscButtonText|NSISdl|Name|Nop|OutFile|PE(?:DllCharacteristics|SubsysVer)|Page(?:Callbacks)?|Pop|Push|Quit|RMDir|Read(?:EnvStr|INIStr|RegDWORD|RegStr)|Reboot|RegDLL|Rename|RequestExecutionLevel|ReserveFile|Return|SearchPath|Section(?:End|GetFlags|GetInstTypes|GetSize|GetText|Group|In|SetFlags|SetInstTypes|SetSize|SetText)?|SendMessage|Set(?:AutoClose|BrandingImage|Compress|Compressor(?:DictSize)?|CtlColors|CurInstType|DatablockOptimize|DateSave|Details(?:Print|View)|ErrorLevel|Errors|FileAttributes|Font|OutPath|Overwrite|PluginUnload|RebootFlag|RegView|ShellVarContext|Silent)|Show(?:InstDetails|UninstDetails|Window)|Silent(?:Install|UnInstall)|Sleep|SpaceTexts|Splash|StartMenu|Str(?:CmpS?|Cpy|Len)|SubCaption|System|UnRegDLL|Unicode|UninstPage|Uninstall(?:ButtonText|Caption|Icon|SubCaption|Text)|UserInfo|VI(?:AddVersionKey|FileVersion|ProductVersion)|VPatch|Var|WindowIcon|Write(?:INIStr|Reg(?:Bin|DWORD|ExpandStr|MultiStr|None|Str)|Uninstaller)|XPStyle|ns(?:Dialogs|Exec))\b/m,lookbehind:!0},property:/\b(?:ARCHIVE|FILE_(?:ATTRIBUTE_ARCHIVE|ATTRIBUTE_NORMAL|ATTRIBUTE_OFFLINE|ATTRIBUTE_READONLY|ATTRIBUTE_SYSTEM|ATTRIBUTE_TEMPORARY)|HK(?:(?:CR|CU|LM)(?:32|64)?|DD|PD|U)|HKEY_(?:CLASSES_ROOT|CURRENT_CONFIG|CURRENT_USER|DYN_DATA|LOCAL_MACHINE|PERFORMANCE_DATA|USERS)|ID(?:ABORT|CANCEL|IGNORE|NO|OK|RETRY|YES)|MB_(?:ABORTRETRYIGNORE|DEFBUTTON1|DEFBUTTON2|DEFBUTTON3|DEFBUTTON4|ICONEXCLAMATION|ICONINFORMATION|ICONQUESTION|ICONSTOP|OK|OKCANCEL|RETRYCANCEL|RIGHT|RTLREADING|SETFOREGROUND|TOPMOST|USERICON|YESNO)|NORMAL|OFFLINE|READONLY|SHCTX|SHELL_CONTEXT|SYSTEM|TEMPORARY|admin|all|auto|both|colored|false|force|hide|highest|lastused|leave|listonly|none|normal|notset|off|on|open|print|show|silent|silentlog|smooth|textonly|true|user)\b/,constant:/\$\{[!\w\.:\^-]+\}|\$\([!\w\.:\^-]+\)/,variable:/\$\w[\w\.]*/,number:/\b0x[\dA-Fa-f]+\b|(?:\b\d+(?:\.\d*)?|\B\.\d+)(?:[Ee]-?\d+)?/,operator:/--?|\+\+?|<=?|>=?|==?=?|&&?|\|\|?|[?*\/~^%]/,punctuation:/[{}[\];(),.:]/,important:{pattern:/(^[\t ]*)!(?:addincludedir|addplugindir|appendfile|cd|define|delfile|echo|else|endif|error|execute|finalize|getdllversion|gettlbversion|if|ifdef|ifmacrodef|ifmacrondef|ifndef|include|insertmacro|macro|macroend|makensis|packhdr|pragma|searchparse|searchreplace|system|tempfile|undef|verbose|warning)\b/im,lookbehind:!0}}}return hE}var mE,PM;function Ive(){if(PM)return mE;PM=1;var e=Ai();mE=t,t.displayName="objectivec",t.aliases=["objc"];function t(n){n.register(e),n.languages.objectivec=n.languages.extend("c",{string:{pattern:/@?"(?:\\(?:\r\n|[\s\S])|[^"\\\r\n])*"/,greedy:!0},keyword:/\b(?:asm|auto|break|case|char|const|continue|default|do|double|else|enum|extern|float|for|goto|if|in|inline|int|long|register|return|self|short|signed|sizeof|static|struct|super|switch|typedef|typeof|union|unsigned|void|volatile|while)\b|(?:@interface|@end|@implementation|@protocol|@class|@public|@protected|@private|@property|@try|@catch|@finally|@throw|@synthesize|@dynamic|@selector)\b/,operator:/-[->]?|\+\+?|!=?|<<?=?|>>?=?|==?|&&?|\|\|?|[~^%?*\/@]/}),delete n.languages.objectivec["class-name"],n.languages.objc=n.languages.objectivec}return mE}var bE,zM;function Lve(){if(zM)return bE;zM=1,bE=e,e.displayName="ocaml",e.aliases=[];function e(t){t.languages.ocaml={comment:{pattern:/\(\*[\s\S]*?\*\)/,greedy:!0},char:{pattern:/'(?:[^\\\r\n']|\\(?:.|[ox]?[0-9a-f]{1,3}))'/i,greedy:!0},string:[{pattern:/"(?:\\(?:[\s\S]|\r\n)|[^\\\r\n"])*"/,greedy:!0},{pattern:/\{([a-z_]*)\|[\s\S]*?\|\1\}/,greedy:!0}],number:[/\b(?:0b[01][01_]*|0o[0-7][0-7_]*)\b/i,/\b0x[a-f0-9][a-f0-9_]*(?:\.[a-f0-9_]*)?(?:p[+-]?\d[\d_]*)?(?!\w)/i,/\b\d[\d_]*(?:\.[\d_]*)?(?:e[+-]?\d[\d_]*)?(?!\w)/i],directive:{pattern:/\B#\w+/,alias:"property"},label:{pattern:/\B~\w+/,alias:"property"},"type-variable":{pattern:/\B'\w+/,alias:"function"},variant:{pattern:/`\w+/,alias:"symbol"},keyword:/\b(?:as|assert|begin|class|constraint|do|done|downto|else|end|exception|external|for|fun|function|functor|if|in|include|inherit|initializer|lazy|let|match|method|module|mutable|new|nonrec|object|of|open|private|rec|sig|struct|then|to|try|type|val|value|virtual|when|where|while|with)\b/,boolean:/\b(?:false|true)\b/,"operator-like-punctuation":{pattern:/\[[<>|]|[>|]\]|\{<|>\}/,alias:"punctuation"},operator:/\.[.~]|:[=>]|[=<>@^|&+\-*\/$%!?~][!$%&*+\-.\/:<=>?@^|~]*|\b(?:and|asr|land|lor|lsl|lsr|lxor|mod|or)\b/,punctuation:/;;|::|[(){}\[\].,:;#]|\b_\b/}}return bE}var yE,BM;function Mve(){if(BM)return yE;BM=1;var e=Ai();yE=t,t.displayName="opencl",t.aliases=[];function t(n){n.register(e),function(r){r.languages.opencl=r.languages.extend("c",{keyword:/\b(?:(?:__)?(?:constant|global|kernel|local|private|read_only|read_write|write_only)|__attribute__|auto|(?:bool|u?(?:char|int|long|short)|half|quad)(?:2|3|4|8|16)?|break|case|complex|const|continue|(?:double|float)(?:16(?:x(?:1|2|4|8|16))?|1x(?:1|2|4|8|16)|2(?:x(?:1|2|4|8|16))?|3|4(?:x(?:1|2|4|8|16))?|8(?:x(?:1|2|4|8|16))?)?|default|do|else|enum|extern|for|goto|if|imaginary|inline|packed|pipe|register|restrict|return|signed|sizeof|static|struct|switch|typedef|uniform|union|unsigned|void|volatile|while)\b/,number:/(?:\b0x(?:[\da-f]+(?:\.[\da-f]*)?|\.[\da-f]+)(?:p[+-]?\d+)?|(?:\b\d+(?:\.\d*)?|\B\.\d+)(?:e[+-]?\d+)?)[fuhl]{0,4}/i,boolean:/\b(?:false|true)\b/,"constant-opencl-kernel":{pattern:/\b(?:CHAR_(?:BIT|MAX|MIN)|CLK_(?:ADDRESS_(?:CLAMP(?:_TO_EDGE)?|NONE|REPEAT)|FILTER_(?:LINEAR|NEAREST)|(?:GLOBAL|LOCAL)_MEM_FENCE|NORMALIZED_COORDS_(?:FALSE|TRUE))|CL_(?:BGRA|(?:HALF_)?FLOAT|INTENSITY|LUMINANCE|A?R?G?B?[Ax]?|(?:(?:UN)?SIGNED|[US]NORM)_(?:INT(?:8|16|32))|UNORM_(?:INT_101010|SHORT_(?:555|565)))|(?:DBL|FLT|HALF)_(?:DIG|EPSILON|(?:MAX|MIN)(?:(?:_10)?_EXP)?|MANT_DIG)|FLT_RADIX|HUGE_VALF?|(?:INT|LONG|SCHAR|SHRT)_(?:MAX|MIN)|INFINITY|MAXFLOAT|M_(?:[12]_PI|2_SQRTPI|E|LN(?:2|10)|LOG(?:2|10)E?|PI(?:_[24])?|SQRT(?:1_2|2))(?:_F|_H)?|NAN|(?:UCHAR|UINT|ULONG|USHRT)_MAX)\b/,alias:"constant"}}),r.languages.insertBefore("opencl","class-name",{"builtin-type":{pattern:/\b(?:_cl_(?:command_queue|context|device_id|event|kernel|mem|platform_id|program|sampler)|cl_(?:image_format|mem_fence_flags)|clk_event_t|event_t|image(?:1d_(?:array_|buffer_)?t|2d_(?:array_(?:depth_|msaa_depth_|msaa_)?|depth_|msaa_depth_|msaa_)?t|3d_t)|intptr_t|ndrange_t|ptrdiff_t|queue_t|reserve_id_t|sampler_t|size_t|uintptr_t)\b/,alias:"keyword"}});var a={"type-opencl-host":{pattern:/\b(?:cl_(?:GLenum|GLint|GLuin|addressing_mode|bitfield|bool|buffer_create_type|build_status|channel_(?:order|type)|(?:u?(?:char|int|long|short)|double|float)(?:2|3|4|8|16)?|command_(?:queue(?:_info|_properties)?|type)|context(?:_info|_properties)?|device_(?:exec_capabilities|fp_config|id|info|local_mem_type|mem_cache_type|type)|(?:event|sampler)(?:_info)?|filter_mode|half|image_info|kernel(?:_info|_work_group_info)?|map_flags|mem(?:_flags|_info|_object_type)?|platform_(?:id|info)|profiling_info|program(?:_build_info|_info)?))\b/,alias:"keyword"},"boolean-opencl-host":{pattern:/\bCL_(?:FALSE|TRUE)\b/,alias:"boolean"},"constant-opencl-host":{pattern:/\bCL_(?:A|ABGR|ADDRESS_(?:CLAMP(?:_TO_EDGE)?|MIRRORED_REPEAT|NONE|REPEAT)|ARGB|BGRA|BLOCKING|BUFFER_CREATE_TYPE_REGION|BUILD_(?:ERROR|IN_PROGRESS|NONE|PROGRAM_FAILURE|SUCCESS)|COMMAND_(?:ACQUIRE_GL_OBJECTS|BARRIER|COPY_(?:BUFFER(?:_RECT|_TO_IMAGE)?|IMAGE(?:_TO_BUFFER)?)|FILL_(?:BUFFER|IMAGE)|MAP(?:_BUFFER|_IMAGE)|MARKER|MIGRATE(?:_SVM)?_MEM_OBJECTS|NATIVE_KERNEL|NDRANGE_KERNEL|READ_(?:BUFFER(?:_RECT)?|IMAGE)|RELEASE_GL_OBJECTS|SVM_(?:FREE|MAP|MEMCPY|MEMFILL|UNMAP)|TASK|UNMAP_MEM_OBJECT|USER|WRITE_(?:BUFFER(?:_RECT)?|IMAGE))|COMPILER_NOT_AVAILABLE|COMPILE_PROGRAM_FAILURE|COMPLETE|CONTEXT_(?:DEVICES|INTEROP_USER_SYNC|NUM_DEVICES|PLATFORM|PROPERTIES|REFERENCE_COUNT)|DEPTH(?:_STENCIL)?|DEVICE_(?:ADDRESS_BITS|AFFINITY_DOMAIN_(?:L[1-4]_CACHE|NEXT_PARTITIONABLE|NUMA)|AVAILABLE|BUILT_IN_KERNELS|COMPILER_AVAILABLE|DOUBLE_FP_CONFIG|ENDIAN_LITTLE|ERROR_CORRECTION_SUPPORT|EXECUTION_CAPABILITIES|EXTENSIONS|GLOBAL_(?:MEM_(?:CACHELINE_SIZE|CACHE_SIZE|CACHE_TYPE|SIZE)|VARIABLE_PREFERRED_TOTAL_SIZE)|HOST_UNIFIED_MEMORY|IL_VERSION|IMAGE(?:2D_MAX_(?:HEIGHT|WIDTH)|3D_MAX_(?:DEPTH|HEIGHT|WIDTH)|_BASE_ADDRESS_ALIGNMENT|_MAX_ARRAY_SIZE|_MAX_BUFFER_SIZE|_PITCH_ALIGNMENT|_SUPPORT)|LINKER_AVAILABLE|LOCAL_MEM_SIZE|LOCAL_MEM_TYPE|MAX_(?:CLOCK_FREQUENCY|COMPUTE_UNITS|CONSTANT_ARGS|CONSTANT_BUFFER_SIZE|GLOBAL_VARIABLE_SIZE|MEM_ALLOC_SIZE|NUM_SUB_GROUPS|ON_DEVICE_(?:EVENTS|QUEUES)|PARAMETER_SIZE|PIPE_ARGS|READ_IMAGE_ARGS|READ_WRITE_IMAGE_ARGS|SAMPLERS|WORK_GROUP_SIZE|WORK_ITEM_DIMENSIONS|WORK_ITEM_SIZES|WRITE_IMAGE_ARGS)|MEM_BASE_ADDR_ALIGN|MIN_DATA_TYPE_ALIGN_SIZE|NAME|NATIVE_VECTOR_WIDTH_(?:CHAR|DOUBLE|FLOAT|HALF|INT|LONG|SHORT)|NOT_(?:AVAILABLE|FOUND)|OPENCL_C_VERSION|PARENT_DEVICE|PARTITION_(?:AFFINITY_DOMAIN|BY_AFFINITY_DOMAIN|BY_COUNTS|BY_COUNTS_LIST_END|EQUALLY|FAILED|MAX_SUB_DEVICES|PROPERTIES|TYPE)|PIPE_MAX_(?:ACTIVE_RESERVATIONS|PACKET_SIZE)|PLATFORM|PREFERRED_(?:GLOBAL_ATOMIC_ALIGNMENT|INTEROP_USER_SYNC|LOCAL_ATOMIC_ALIGNMENT|PLATFORM_ATOMIC_ALIGNMENT|VECTOR_WIDTH_(?:CHAR|DOUBLE|FLOAT|HALF|INT|LONG|SHORT))|PRINTF_BUFFER_SIZE|PROFILE|PROFILING_TIMER_RESOLUTION|QUEUE_(?:ON_(?:DEVICE_(?:MAX_SIZE|PREFERRED_SIZE|PROPERTIES)|HOST_PROPERTIES)|PROPERTIES)|REFERENCE_COUNT|SINGLE_FP_CONFIG|SUB_GROUP_INDEPENDENT_FORWARD_PROGRESS|SVM_(?:ATOMICS|CAPABILITIES|COARSE_GRAIN_BUFFER|FINE_GRAIN_BUFFER|FINE_GRAIN_SYSTEM)|TYPE(?:_ACCELERATOR|_ALL|_CPU|_CUSTOM|_DEFAULT|_GPU)?|VENDOR(?:_ID)?|VERSION)|DRIVER_VERSION|EVENT_(?:COMMAND_(?:EXECUTION_STATUS|QUEUE|TYPE)|CONTEXT|REFERENCE_COUNT)|EXEC_(?:KERNEL|NATIVE_KERNEL|STATUS_ERROR_FOR_EVENTS_IN_WAIT_LIST)|FILTER_(?:LINEAR|NEAREST)|FLOAT|FP_(?:CORRECTLY_ROUNDED_DIVIDE_SQRT|DENORM|FMA|INF_NAN|ROUND_TO_INF|ROUND_TO_NEAREST|ROUND_TO_ZERO|SOFT_FLOAT)|GLOBAL|HALF_FLOAT|IMAGE_(?:ARRAY_SIZE|BUFFER|DEPTH|ELEMENT_SIZE|FORMAT|FORMAT_MISMATCH|FORMAT_NOT_SUPPORTED|HEIGHT|NUM_MIP_LEVELS|NUM_SAMPLES|ROW_PITCH|SLICE_PITCH|WIDTH)|INTENSITY|INVALID_(?:ARG_INDEX|ARG_SIZE|ARG_VALUE|BINARY|BUFFER_SIZE|BUILD_OPTIONS|COMMAND_QUEUE|COMPILER_OPTIONS|CONTEXT|DEVICE|DEVICE_PARTITION_COUNT|DEVICE_QUEUE|DEVICE_TYPE|EVENT|EVENT_WAIT_LIST|GLOBAL_OFFSET|GLOBAL_WORK_SIZE|GL_OBJECT|HOST_PTR|IMAGE_DESCRIPTOR|IMAGE_FORMAT_DESCRIPTOR|IMAGE_SIZE|KERNEL|KERNEL_ARGS|KERNEL_DEFINITION|KERNEL_NAME|LINKER_OPTIONS|MEM_OBJECT|MIP_LEVEL|OPERATION|PIPE_SIZE|PLATFORM|PROGRAM|PROGRAM_EXECUTABLE|PROPERTY|QUEUE_PROPERTIES|SAMPLER|VALUE|WORK_DIMENSION|WORK_GROUP_SIZE|WORK_ITEM_SIZE)|KERNEL_(?:ARG_(?:ACCESS_(?:NONE|QUALIFIER|READ_ONLY|READ_WRITE|WRITE_ONLY)|ADDRESS_(?:CONSTANT|GLOBAL|LOCAL|PRIVATE|QUALIFIER)|INFO_NOT_AVAILABLE|NAME|TYPE_(?:CONST|NAME|NONE|PIPE|QUALIFIER|RESTRICT|VOLATILE))|ATTRIBUTES|COMPILE_NUM_SUB_GROUPS|COMPILE_WORK_GROUP_SIZE|CONTEXT|EXEC_INFO_SVM_FINE_GRAIN_SYSTEM|EXEC_INFO_SVM_PTRS|FUNCTION_NAME|GLOBAL_WORK_SIZE|LOCAL_MEM_SIZE|LOCAL_SIZE_FOR_SUB_GROUP_COUNT|MAX_NUM_SUB_GROUPS|MAX_SUB_GROUP_SIZE_FOR_NDRANGE|NUM_ARGS|PREFERRED_WORK_GROUP_SIZE_MULTIPLE|PRIVATE_MEM_SIZE|PROGRAM|REFERENCE_COUNT|SUB_GROUP_COUNT_FOR_NDRANGE|WORK_GROUP_SIZE)|LINKER_NOT_AVAILABLE|LINK_PROGRAM_FAILURE|LOCAL|LUMINANCE|MAP_(?:FAILURE|READ|WRITE|WRITE_INVALIDATE_REGION)|MEM_(?:ALLOC_HOST_PTR|ASSOCIATED_MEMOBJECT|CONTEXT|COPY_HOST_PTR|COPY_OVERLAP|FLAGS|HOST_NO_ACCESS|HOST_PTR|HOST_READ_ONLY|HOST_WRITE_ONLY|KERNEL_READ_AND_WRITE|MAP_COUNT|OBJECT_(?:ALLOCATION_FAILURE|BUFFER|IMAGE1D|IMAGE1D_ARRAY|IMAGE1D_BUFFER|IMAGE2D|IMAGE2D_ARRAY|IMAGE3D|PIPE)|OFFSET|READ_ONLY|READ_WRITE|REFERENCE_COUNT|SIZE|SVM_ATOMICS|SVM_FINE_GRAIN_BUFFER|TYPE|USES_SVM_POINTER|USE_HOST_PTR|WRITE_ONLY)|MIGRATE_MEM_OBJECT_(?:CONTENT_UNDEFINED|HOST)|MISALIGNED_SUB_BUFFER_OFFSET|NONE|NON_BLOCKING|OUT_OF_(?:HOST_MEMORY|RESOURCES)|PIPE_(?:MAX_PACKETS|PACKET_SIZE)|PLATFORM_(?:EXTENSIONS|HOST_TIMER_RESOLUTION|NAME|PROFILE|VENDOR|VERSION)|PROFILING_(?:COMMAND_(?:COMPLETE|END|QUEUED|START|SUBMIT)|INFO_NOT_AVAILABLE)|PROGRAM_(?:BINARIES|BINARY_SIZES|BINARY_TYPE(?:_COMPILED_OBJECT|_EXECUTABLE|_LIBRARY|_NONE)?|BUILD_(?:GLOBAL_VARIABLE_TOTAL_SIZE|LOG|OPTIONS|STATUS)|CONTEXT|DEVICES|IL|KERNEL_NAMES|NUM_DEVICES|NUM_KERNELS|REFERENCE_COUNT|SOURCE)|QUEUED|QUEUE_(?:CONTEXT|DEVICE|DEVICE_DEFAULT|ON_DEVICE|ON_DEVICE_DEFAULT|OUT_OF_ORDER_EXEC_MODE_ENABLE|PROFILING_ENABLE|PROPERTIES|REFERENCE_COUNT|SIZE)|R|RA|READ_(?:ONLY|WRITE)_CACHE|RG|RGB|RGBA|RGBx|RGx|RUNNING|Rx|SAMPLER_(?:ADDRESSING_MODE|CONTEXT|FILTER_MODE|LOD_MAX|LOD_MIN|MIP_FILTER_MODE|NORMALIZED_COORDS|REFERENCE_COUNT)|(?:UN)?SIGNED_INT(?:8|16|32)|SNORM_INT(?:8|16)|SUBMITTED|SUCCESS|UNORM_INT(?:8|16|24|_101010|_101010_2)|UNORM_SHORT_(?:555|565)|VERSION_(?:1_0|1_1|1_2|2_0|2_1)|sBGRA|sRGB|sRGBA|sRGBx)\b/,alias:"constant"},"function-opencl-host":{pattern:/\bcl(?:BuildProgram|CloneKernel|CompileProgram|Create(?:Buffer|CommandQueue(?:WithProperties)?|Context|ContextFromType|Image|Image2D|Image3D|Kernel|KernelsInProgram|Pipe|ProgramWith(?:Binary|BuiltInKernels|IL|Source)|Sampler|SamplerWithProperties|SubBuffer|SubDevices|UserEvent)|Enqueue(?:(?:Barrier|Marker)(?:WithWaitList)?|Copy(?:Buffer(?:Rect|ToImage)?|Image(?:ToBuffer)?)|(?:Fill|Map)(?:Buffer|Image)|MigrateMemObjects|NDRangeKernel|NativeKernel|(?:Read|Write)(?:Buffer(?:Rect)?|Image)|SVM(?:Free|Map|MemFill|Memcpy|MigrateMem|Unmap)|Task|UnmapMemObject|WaitForEvents)|Finish|Flush|Get(?:CommandQueueInfo|ContextInfo|Device(?:AndHostTimer|IDs|Info)|Event(?:Profiling)?Info|ExtensionFunctionAddress(?:ForPlatform)?|HostTimer|ImageInfo|Kernel(?:ArgInfo|Info|SubGroupInfo|WorkGroupInfo)|MemObjectInfo|PipeInfo|Platform(?:IDs|Info)|Program(?:Build)?Info|SamplerInfo|SupportedImageFormats)|LinkProgram|(?:Release|Retain)(?:CommandQueue|Context|Device|Event|Kernel|MemObject|Program|Sampler)|SVM(?:Alloc|Free)|Set(?:CommandQueueProperty|DefaultDeviceCommandQueue|EventCallback|Kernel|Kernel(?:Arg(?:SVMPointer)?|ExecInfo)|MemObjectDestructorCallback|UserEventStatus)|Unload(?:Platform)?Compiler|WaitForEvents)\b/,alias:"function"}};r.languages.insertBefore("c","keyword",a),r.languages.cpp&&(a["type-opencl-host-cpp"]={pattern:/\b(?:Buffer|BufferGL|BufferRenderGL|CommandQueue|Context|Device|DeviceCommandQueue|EnqueueArgs|Event|Image|Image1D|Image1DArray|Image1DBuffer|Image2D|Image2DArray|Image2DGL|Image3D|Image3DGL|ImageFormat|ImageGL|Kernel|KernelFunctor|LocalSpaceArg|Memory|NDRange|Pipe|Platform|Program|SVMAllocator|SVMTraitAtomic|SVMTraitCoarse|SVMTraitFine|SVMTraitReadOnly|SVMTraitReadWrite|SVMTraitWriteOnly|Sampler|UserEvent)\b/,alias:"keyword"},r.languages.insertBefore("cpp","keyword",a))}(n)}return yE}var vE,UM;function Fve(){if(UM)return vE;UM=1,vE=e,e.displayName="openqasm",e.aliases=["qasm"];function e(t){t.languages.openqasm={comment:/\/\*[\s\S]*?\*\/|\/\/.*/,string:{pattern:/"[^"\r\n\t]*"|'[^'\r\n\t]*'/,greedy:!0},keyword:/\b(?:CX|OPENQASM|U|barrier|boxas|boxto|break|const|continue|ctrl|def|defcal|defcalgrammar|delay|else|end|for|gate|gphase|if|in|include|inv|kernel|lengthof|let|measure|pow|reset|return|rotary|stretchinf|while)\b|#pragma\b/,"class-name":/\b(?:angle|bit|bool|creg|fixed|float|int|length|qreg|qubit|stretch|uint)\b/,function:/\b(?:cos|exp|ln|popcount|rotl|rotr|sin|sqrt|tan)\b(?=\s*\()/,constant:/\b(?:euler|pi|tau)\b|π|𝜏|ℇ/,number:{pattern:/(^|[^.\w$])(?:\d+(?:\.\d*)?|\.\d+)(?:e[+-]?\d+)?(?:dt|ns|us|µs|ms|s)?/i,lookbehind:!0},operator:/->|>>=?|<<=?|&&|\|\||\+\+|--|[!=<>&|~^+\-*/%]=?|@/,punctuation:/[(){}\[\];,:.]/},t.languages.qasm=t.languages.openqasm}return vE}var SE,jM;function Pve(){if(jM)return SE;jM=1,SE=e,e.displayName="oz",e.aliases=[];function e(t){t.languages.oz={comment:{pattern:/\/\*[\s\S]*?\*\/|%.*/,greedy:!0},string:{pattern:/"(?:[^"\\]|\\[\s\S])*"/,greedy:!0},atom:{pattern:/'(?:[^'\\]|\\[\s\S])*'/,greedy:!0,alias:"builtin"},keyword:/\$|\[\]|\b(?:_|at|attr|case|catch|choice|class|cond|declare|define|dis|else(?:case|if)?|end|export|fail|false|feat|finally|from|fun|functor|if|import|in|local|lock|meth|nil|not|of|or|prepare|proc|prop|raise|require|self|skip|then|thread|true|try|unit)\b/,function:[/\b[a-z][A-Za-z\d]*(?=\()/,{pattern:/(\{)[A-Z][A-Za-z\d]*\b/,lookbehind:!0}],number:/\b(?:0[bx][\da-f]+|\d+(?:\.\d*)?(?:e~?\d+)?)\b|&(?:[^\\]|\\(?:\d{3}|.))/i,variable:/`(?:[^`\\]|\\.)+`/,"attr-name":/\b\w+(?=[ \t]*:(?![:=]))/,operator:/:(?:=|::?)|<[-:=]?|=(?:=|<?:?)|>=?:?|\\=:?|!!?|[|#+\-*\/,~^@]|\b(?:andthen|div|mod|orelse)\b/,punctuation:/[\[\](){}.:;?]/}}return SE}var EE,GM;function zve(){if(GM)return EE;GM=1,EE=e,e.displayName="parigp",e.aliases=[];function e(t){t.languages.parigp={comment:/\/\*[\s\S]*?\*\/|\\\\.*/,string:{pattern:/"(?:[^"\\\r\n]|\\.)*"/,greedy:!0},keyword:function(){var n=["breakpoint","break","dbg_down","dbg_err","dbg_up","dbg_x","forcomposite","fordiv","forell","forpart","forprime","forstep","forsubgroup","forvec","for","iferr","if","local","my","next","return","until","while"];return n=n.map(function(r){return r.split("").join(" *")}).join("|"),RegExp("\\b(?:"+n+")\\b")}(),function:/\b\w(?:[\w ]*\w)?(?= *\()/,number:{pattern:/((?:\. *\. *)?)(?:\b\d(?: *\d)*(?: *(?!\. *\.)\.(?: *\d)*)?|\. *\d(?: *\d)*)(?: *e *(?:[+-] *)?\d(?: *\d)*)?/i,lookbehind:!0},operator:/\. *\.|[*\/!](?: *=)?|%(?: *=|(?: *#)?(?: *')*)?|\+(?: *[+=])?|-(?: *[-=>])?|<(?: *>|(?: *<)?(?: *=)?)?|>(?: *>)?(?: *=)?|=(?: *=){0,2}|\\(?: *\/)?(?: *=)?|&(?: *&)?|\| *\||['#~^]/,punctuation:/[\[\]{}().,:;|]/}}return EE}var wE,HM;function Bve(){if(HM)return wE;HM=1,wE=e,e.displayName="parser",e.aliases=[];function e(t){(function(n){var r=n.languages.parser=n.languages.extend("markup",{keyword:{pattern:/(^|[^^])(?:\^(?:case|eval|for|if|switch|throw)\b|@(?:BASE|CLASS|GET(?:_DEFAULT)?|OPTIONS|SET_DEFAULT|USE)\b)/,lookbehind:!0},variable:{pattern:/(^|[^^])\B\$(?:\w+|(?=[.{]))(?:(?:\.|::?)\w+)*(?:\.|::?)?/,lookbehind:!0,inside:{punctuation:/\.|:+/}},function:{pattern:/(^|[^^])\B[@^]\w+(?:(?:\.|::?)\w+)*(?:\.|::?)?/,lookbehind:!0,inside:{keyword:{pattern:/(^@)(?:GET_|SET_)/,lookbehind:!0},punctuation:/\.|:+/}},escape:{pattern:/\^(?:[$^;@()\[\]{}"':]|#[a-f\d]*)/i,alias:"builtin"},punctuation:/[\[\](){};]/});r=n.languages.insertBefore("parser","keyword",{"parser-comment":{pattern:/(\s)#.*/,lookbehind:!0,alias:"comment"},expression:{pattern:/(^|[^^])\((?:[^()]|\((?:[^()]|\((?:[^()])*\))*\))*\)/,greedy:!0,lookbehind:!0,inside:{string:{pattern:/(^|[^^])(["'])(?:(?!\2)[^^]|\^[\s\S])*\2/,lookbehind:!0},keyword:r.keyword,variable:r.variable,function:r.function,boolean:/\b(?:false|true)\b/,number:/\b(?:0x[a-f\d]+|\d+(?:\.\d*)?(?:e[+-]?\d+)?)\b/i,escape:r.escape,operator:/[~+*\/\\%]|!(?:\|\|?|=)?|&&?|\|\|?|==|<[<=]?|>[>=]?|-[fd]?|\b(?:def|eq|ge|gt|in|is|le|lt|ne)\b/,punctuation:r.punctuation}}}),n.languages.insertBefore("inside","punctuation",{expression:r.expression,keyword:r.keyword,variable:r.variable,function:r.function,escape:r.escape,"parser-punctuation":{pattern:r.punctuation,alias:"punctuation"}},r.tag.inside["attr-value"])})(t)}return wE}var xE,$M;function Uve(){if($M)return xE;$M=1,xE=e,e.displayName="pascal",e.aliases=["objectpascal"];function e(t){t.languages.pascal={directive:{pattern:/\{\$[\s\S]*?\}/,greedy:!0,alias:["marco","property"]},comment:{pattern:/\(\*[\s\S]*?\*\)|\{[\s\S]*?\}|\/\/.*/,greedy:!0},string:{pattern:/(?:'(?:''|[^'\r\n])*'(?!')|#[&$%]?[a-f\d]+)+|\^[a-z]/i,greedy:!0},asm:{pattern:/(\basm\b)[\s\S]+?(?=\bend\s*[;[])/i,lookbehind:!0,greedy:!0,inside:null},keyword:[{pattern:/(^|[^&])\b(?:absolute|array|asm|begin|case|const|constructor|destructor|do|downto|else|end|file|for|function|goto|if|implementation|inherited|inline|interface|label|nil|object|of|operator|packed|procedure|program|record|reintroduce|repeat|self|set|string|then|to|type|unit|until|uses|var|while|with)\b/i,lookbehind:!0},{pattern:/(^|[^&])\b(?:dispose|exit|false|new|true)\b/i,lookbehind:!0},{pattern:/(^|[^&])\b(?:class|dispinterface|except|exports|finalization|finally|initialization|inline|library|on|out|packed|property|raise|resourcestring|threadvar|try)\b/i,lookbehind:!0},{pattern:/(^|[^&])\b(?:absolute|abstract|alias|assembler|bitpacked|break|cdecl|continue|cppdecl|cvar|default|deprecated|dynamic|enumerator|experimental|export|external|far|far16|forward|generic|helper|implements|index|interrupt|iochecks|local|message|name|near|nodefault|noreturn|nostackframe|oldfpccall|otherwise|overload|override|pascal|platform|private|protected|public|published|read|register|reintroduce|result|safecall|saveregisters|softfloat|specialize|static|stdcall|stored|strict|unaligned|unimplemented|varargs|virtual|write)\b/i,lookbehind:!0}],number:[/(?:[&%]\d+|\$[a-f\d]+)/i,/\b\d+(?:\.\d+)?(?:e[+-]?\d+)?/i],operator:[/\.\.|\*\*|:=|<[<=>]?|>[>=]?|[+\-*\/]=?|[@^=]/,{pattern:/(^|[^&])\b(?:and|as|div|exclude|in|include|is|mod|not|or|shl|shr|xor)\b/,lookbehind:!0}],punctuation:/\(\.|\.\)|[()\[\]:;,.]/},t.languages.pascal.asm.inside=t.languages.extend("pascal",{asm:void 0,keyword:void 0,operator:void 0}),t.languages.objectpascal=t.languages.pascal}return xE}var kE,qM;function jve(){if(qM)return kE;qM=1,kE=e,e.displayName="pascaligo",e.aliases=[];function e(t){(function(n){var r=/\((?:[^()]|\((?:[^()]|\([^()]*\))*\))*\)/.source,a=/(?:\b\w+(?:<braces>)?|<braces>)/.source.replace(/<braces>/g,function(){return r}),o=n.languages.pascaligo={comment:/\(\*[\s\S]+?\*\)|\/\/.*/,string:{pattern:/(["'`])(?:\\[\s\S]|(?!\1)[^\\])*\1|\^[a-z]/i,greedy:!0},"class-name":[{pattern:RegExp(/(\btype\s+\w+\s+is\s+)<type>/.source.replace(/<type>/g,function(){return a}),"i"),lookbehind:!0,inside:null},{pattern:RegExp(/<type>(?=\s+is\b)/.source.replace(/<type>/g,function(){return a}),"i"),inside:null},{pattern:RegExp(/(:\s*)<type>/.source.replace(/<type>/g,function(){return a})),lookbehind:!0,inside:null}],keyword:{pattern:/(^|[^&])\b(?:begin|block|case|const|else|end|fail|for|from|function|if|is|nil|of|remove|return|skip|then|type|var|while|with)\b/i,lookbehind:!0},boolean:{pattern:/(^|[^&])\b(?:False|True)\b/i,lookbehind:!0},builtin:{pattern:/(^|[^&])\b(?:bool|int|list|map|nat|record|string|unit)\b/i,lookbehind:!0},function:/\b\w+(?=\s*\()/,number:[/%[01]+|&[0-7]+|\$[a-f\d]+/i,/\b\d+(?:\.\d+)?(?:e[+-]?\d+)?(?:mtz|n)?/i],operator:/->|=\/=|\.\.|\*\*|:=|<[<=>]?|>[>=]?|[+\-*\/]=?|[@^=|]|\b(?:and|mod|or)\b/,punctuation:/\(\.|\.\)|[()\[\]:;,.{}]/},s=["comment","keyword","builtin","operator","punctuation"].reduce(function(u,c){return u[c]=o[c],u},{});o["class-name"].forEach(function(u){u.inside=s})})(t)}return kE}var TE,VM;function Gve(){if(VM)return TE;VM=1,TE=e,e.displayName="pcaxis",e.aliases=["px"];function e(t){t.languages.pcaxis={string:/"[^"]*"/,keyword:{pattern:/((?:^|;)\s*)[-A-Z\d]+(?:\s*\[[-\w]+\])?(?:\s*\("[^"]*"(?:,\s*"[^"]*")*\))?(?=\s*=)/,lookbehind:!0,greedy:!0,inside:{keyword:/^[-A-Z\d]+/,language:{pattern:/^(\s*)\[[-\w]+\]/,lookbehind:!0,inside:{punctuation:/^\[|\]$/,property:/[-\w]+/}},"sub-key":{pattern:/^(\s*)\S[\s\S]*/,lookbehind:!0,inside:{parameter:{pattern:/"[^"]*"/,alias:"property"},punctuation:/^\(|\)$|,/}}}},operator:/=/,tlist:{pattern:/TLIST\s*\(\s*\w+(?:(?:\s*,\s*"[^"]*")+|\s*,\s*"[^"]*"-"[^"]*")?\s*\)/,greedy:!0,inside:{function:/^TLIST/,property:{pattern:/^(\s*\(\s*)\w+/,lookbehind:!0},string:/"[^"]*"/,punctuation:/[(),]/,operator:/-/}},punctuation:/[;,]/,number:{pattern:/(^|\s)\d+(?:\.\d+)?(?!\S)/,lookbehind:!0},boolean:/NO|YES/},t.languages.px=t.languages.pcaxis}return TE}var AE,WM;function Hve(){if(WM)return AE;WM=1,AE=e,e.displayName="peoplecode",e.aliases=["pcode"];function e(t){t.languages.peoplecode={comment:RegExp([/\/\*[\s\S]*?\*\//.source,/\bREM[^;]*;/.source,/<\*(?:[^<*]|\*(?!>)|<(?!\*)|<\*(?:(?!\*>)[\s\S])*\*>)*\*>/.source,/\/\+[\s\S]*?\+\//.source].join("|")),string:{pattern:/'(?:''|[^'\r\n])*'(?!')|"(?:""|[^"\r\n])*"(?!")/,greedy:!0},variable:/%\w+/,"function-definition":{pattern:/((?:^|[^\w-])(?:function|method)\s+)\w+/i,lookbehind:!0,alias:"function"},"class-name":{pattern:/((?:^|[^-\w])(?:as|catch|class|component|create|extends|global|implements|instance|local|of|property|returns)\s+)\w+(?::\w+)*/i,lookbehind:!0,inside:{punctuation:/:/}},keyword:/\b(?:abstract|alias|as|catch|class|component|constant|create|declare|else|end-(?:class|evaluate|for|function|get|if|method|set|try|while)|evaluate|extends|for|function|get|global|if|implements|import|instance|library|local|method|null|of|out|peopleCode|private|program|property|protected|readonly|ref|repeat|returns?|set|step|then|throw|to|try|until|value|when(?:-other)?|while)\b/i,"operator-keyword":{pattern:/\b(?:and|not|or)\b/i,alias:"operator"},function:/[_a-z]\w*(?=\s*\()/i,boolean:/\b(?:false|true)\b/i,number:/\b\d+(?:\.\d+)?\b/,operator:/<>|[<>]=?|!=|\*\*|[-+*/|=@]/,punctuation:/[:.;,()[\]]/},t.languages.pcode=t.languages.peoplecode}return AE}var RE,YM;function $ve(){if(YM)return RE;YM=1,RE=e,e.displayName="perl",e.aliases=[];function e(t){(function(n){var r=/(?:\((?:[^()\\]|\\[\s\S])*\)|\{(?:[^{}\\]|\\[\s\S])*\}|\[(?:[^[\]\\]|\\[\s\S])*\]|<(?:[^<>\\]|\\[\s\S])*>)/.source;n.languages.perl={comment:[{pattern:/(^\s*)=\w[\s\S]*?=cut.*/m,lookbehind:!0,greedy:!0},{pattern:/(^|[^\\$])#.*/,lookbehind:!0,greedy:!0}],string:[{pattern:RegExp(/\b(?:q|qq|qw|qx)(?![a-zA-Z0-9])\s*/.source+"(?:"+[/([^a-zA-Z0-9\s{(\[<])(?:(?!\1)[^\\]|\\[\s\S])*\1/.source,/([a-zA-Z0-9])(?:(?!\2)[^\\]|\\[\s\S])*\2/.source,r].join("|")+")"),greedy:!0},{pattern:/("|`)(?:(?!\1)[^\\]|\\[\s\S])*\1/,greedy:!0},{pattern:/'(?:[^'\\\r\n]|\\.)*'/,greedy:!0}],regex:[{pattern:RegExp(/\b(?:m|qr)(?![a-zA-Z0-9])\s*/.source+"(?:"+[/([^a-zA-Z0-9\s{(\[<])(?:(?!\1)[^\\]|\\[\s\S])*\1/.source,/([a-zA-Z0-9])(?:(?!\2)[^\\]|\\[\s\S])*\2/.source,r].join("|")+")"+/[msixpodualngc]*/.source),greedy:!0},{pattern:RegExp(/(^|[^-])\b(?:s|tr|y)(?![a-zA-Z0-9])\s*/.source+"(?:"+[/([^a-zA-Z0-9\s{(\[<])(?:(?!\2)[^\\]|\\[\s\S])*\2(?:(?!\2)[^\\]|\\[\s\S])*\2/.source,/([a-zA-Z0-9])(?:(?!\3)[^\\]|\\[\s\S])*\3(?:(?!\3)[^\\]|\\[\s\S])*\3/.source,r+/\s*/.source+r].join("|")+")"+/[msixpodualngcer]*/.source),lookbehind:!0,greedy:!0},{pattern:/\/(?:[^\/\\\r\n]|\\.)*\/[msixpodualngc]*(?=\s*(?:$|[\r\n,.;})&|\-+*~<>!?^]|(?:and|cmp|eq|ge|gt|le|lt|ne|not|or|x|xor)\b))/,greedy:!0}],variable:[/[&*$@%]\{\^[A-Z]+\}/,/[&*$@%]\^[A-Z_]/,/[&*$@%]#?(?=\{)/,/[&*$@%]#?(?:(?:::)*'?(?!\d)[\w$]+(?![\w$]))+(?:::)*/,/[&*$@%]\d+/,/(?!%=)[$@%][!"#$%&'()*+,\-.\/:;<=>?@[\\\]^_`{|}~]/],filehandle:{pattern:/<(?![<=])\S*?>|\b_\b/,alias:"symbol"},"v-string":{pattern:/v\d+(?:\.\d+)*|\d+(?:\.\d+){2,}/,alias:"string"},function:{pattern:/(\bsub[ \t]+)\w+/,lookbehind:!0},keyword:/\b(?:any|break|continue|default|delete|die|do|else|elsif|eval|for|foreach|given|goto|if|last|local|my|next|our|package|print|redo|require|return|say|state|sub|switch|undef|unless|until|use|when|while)\b/,number:/\b(?:0x[\dA-Fa-f](?:_?[\dA-Fa-f])*|0b[01](?:_?[01])*|(?:(?:\d(?:_?\d)*)?\.)?\d(?:_?\d)*(?:[Ee][+-]?\d+)?)\b/,operator:/-[rwxoRWXOezsfdlpSbctugkTBMAC]\b|\+[+=]?|-[-=>]?|\*\*?=?|\/\/?=?|=[=~>]?|~[~=]?|\|\|?=?|&&?=?|<(?:=>?|<=?)?|>>?=?|![~=]?|[%^]=?|\.(?:=|\.\.?)?|[\\?]|\bx(?:=|\b)|\b(?:and|cmp|eq|ge|gt|le|lt|ne|not|or|xor)\b/,punctuation:/[{}[\];(),:]/}})(t)}return RE}var _E,KM;function qve(){if(KM)return _E;KM=1;var e=zp();_E=t,t.displayName="phpExtras",t.aliases=[];function t(n){n.register(e),n.languages.insertBefore("php","variable",{this:{pattern:/\$this\b/,alias:"keyword"},global:/\$(?:GLOBALS|HTTP_RAW_POST_DATA|_(?:COOKIE|ENV|FILES|GET|POST|REQUEST|SERVER|SESSION)|argc|argv|http_response_header|php_errormsg)\b/,scope:{pattern:/\b[\w\\]+::/,inside:{keyword:/\b(?:parent|self|static)\b/,punctuation:/::|\\/}}})}return _E}var CE,XM;function Vve(){if(XM)return CE;XM=1;var e=zp(),t=Pp();CE=n,n.displayName="phpdoc",n.aliases=[];function n(r){r.register(e),r.register(t),function(a){var o=/(?:\b[a-zA-Z]\w*|[|\\[\]])+/.source;a.languages.phpdoc=a.languages.extend("javadoclike",{parameter:{pattern:RegExp("(@(?:global|param|property(?:-read|-write)?|var)\\s+(?:"+o+"\\s+)?)\\$\\w+"),lookbehind:!0}}),a.languages.insertBefore("phpdoc","keyword",{"class-name":[{pattern:RegExp("(@(?:global|package|param|property(?:-read|-write)?|return|subpackage|throws|var)\\s+)"+o),lookbehind:!0,inside:{keyword:/\b(?:array|bool|boolean|callback|double|false|float|int|integer|mixed|null|object|resource|self|string|true|void)\b/,punctuation:/[|\\[\]()]/}}]}),a.languages.javadoclike.addSupport("php",a.languages.phpdoc)}(r)}return CE}var NE,ZM;function Wve(){if(ZM)return NE;ZM=1;var e=lA();NE=t,t.displayName="plsql",t.aliases=[];function t(n){n.register(e),n.languages.plsql=n.languages.extend("sql",{comment:{pattern:/\/\*[\s\S]*?\*\/|--.*/,greedy:!0},keyword:/\b(?:A|ACCESSIBLE|ADD|AGENT|AGGREGATE|ALL|ALTER|AND|ANY|ARRAY|AS|ASC|AT|ATTRIBUTE|AUTHID|AVG|BEGIN|BETWEEN|BFILE_BASE|BINARY|BLOB_BASE|BLOCK|BODY|BOTH|BOUND|BULK|BY|BYTE|C|CALL|CALLING|CASCADE|CASE|CHAR|CHARACTER|CHARSET|CHARSETFORM|CHARSETID|CHAR_BASE|CHECK|CLOB_BASE|CLONE|CLOSE|CLUSTER|CLUSTERS|COLAUTH|COLLECT|COLUMNS|COMMENT|COMMIT|COMMITTED|COMPILED|COMPRESS|CONNECT|CONSTANT|CONSTRUCTOR|CONTEXT|CONTINUE|CONVERT|COUNT|CRASH|CREATE|CREDENTIAL|CURRENT|CURSOR|CUSTOMDATUM|DANGLING|DATA|DATE|DATE_BASE|DAY|DECLARE|DEFAULT|DEFINE|DELETE|DESC|DETERMINISTIC|DIRECTORY|DISTINCT|DOUBLE|DROP|DURATION|ELEMENT|ELSE|ELSIF|EMPTY|END|ESCAPE|EXCEPT|EXCEPTION|EXCEPTIONS|EXCLUSIVE|EXECUTE|EXISTS|EXIT|EXTERNAL|FETCH|FINAL|FIRST|FIXED|FLOAT|FOR|FORALL|FORCE|FROM|FUNCTION|GENERAL|GOTO|GRANT|GROUP|HASH|HAVING|HEAP|HIDDEN|HOUR|IDENTIFIED|IF|IMMEDIATE|IMMUTABLE|IN|INCLUDING|INDEX|INDEXES|INDICATOR|INDICES|INFINITE|INSERT|INSTANTIABLE|INT|INTERFACE|INTERSECT|INTERVAL|INTO|INVALIDATE|IS|ISOLATION|JAVA|LANGUAGE|LARGE|LEADING|LENGTH|LEVEL|LIBRARY|LIKE|LIKE2|LIKE4|LIKEC|LIMIT|LIMITED|LOCAL|LOCK|LONG|LOOP|MAP|MAX|MAXLEN|MEMBER|MERGE|MIN|MINUS|MINUTE|MOD|MODE|MODIFY|MONTH|MULTISET|MUTABLE|NAME|NAN|NATIONAL|NATIVE|NCHAR|NEW|NOCOMPRESS|NOCOPY|NOT|NOWAIT|NULL|NUMBER_BASE|OBJECT|OCICOLL|OCIDATE|OCIDATETIME|OCIDURATION|OCIINTERVAL|OCILOBLOCATOR|OCINUMBER|OCIRAW|OCIREF|OCIREFCURSOR|OCIROWID|OCISTRING|OCITYPE|OF|OLD|ON|ONLY|OPAQUE|OPEN|OPERATOR|OPTION|OR|ORACLE|ORADATA|ORDER|ORGANIZATION|ORLANY|ORLVARY|OTHERS|OUT|OVERLAPS|OVERRIDING|PACKAGE|PARALLEL_ENABLE|PARAMETER|PARAMETERS|PARENT|PARTITION|PASCAL|PERSISTABLE|PIPE|PIPELINED|PLUGGABLE|POLYMORPHIC|PRAGMA|PRECISION|PRIOR|PRIVATE|PROCEDURE|PUBLIC|RAISE|RANGE|RAW|READ|RECORD|REF|REFERENCE|RELIES_ON|REM|REMAINDER|RENAME|RESOURCE|RESULT|RESULT_CACHE|RETURN|RETURNING|REVERSE|REVOKE|ROLLBACK|ROW|SAMPLE|SAVE|SAVEPOINT|SB1|SB2|SB4|SECOND|SEGMENT|SELECT|SELF|SEPARATE|SEQUENCE|SERIALIZABLE|SET|SHARE|SHORT|SIZE|SIZE_T|SOME|SPARSE|SQL|SQLCODE|SQLDATA|SQLNAME|SQLSTATE|STANDARD|START|STATIC|STDDEV|STORED|STRING|STRUCT|STYLE|SUBMULTISET|SUBPARTITION|SUBSTITUTABLE|SUBTYPE|SUM|SYNONYM|TABAUTH|TABLE|TDO|THE|THEN|TIME|TIMESTAMP|TIMEZONE_ABBR|TIMEZONE_HOUR|TIMEZONE_MINUTE|TIMEZONE_REGION|TO|TRAILING|TRANSACTION|TRANSACTIONAL|TRUSTED|TYPE|UB1|UB2|UB4|UNDER|UNION|UNIQUE|UNPLUG|UNSIGNED|UNTRUSTED|UPDATE|USE|USING|VALIST|VALUE|VALUES|VARIABLE|VARIANCE|VARRAY|VARYING|VIEW|VIEWS|VOID|WHEN|WHERE|WHILE|WITH|WORK|WRAPPED|WRITE|YEAR|ZONE)\b/i,operator:/:=?|=>|[<>^~!]=|\.\.|\|\||\*\*|[-+*/%<>=@]/}),n.languages.insertBefore("plsql","operator",{label:{pattern:/<<\s*\w+\s*>>/,alias:"symbol"}})}return NE}var OE,QM;function Yve(){if(QM)return OE;QM=1,OE=e,e.displayName="powerquery",e.aliases=[];function e(t){t.languages.powerquery={comment:{pattern:/(^|[^\\])(?:\/\*[\s\S]*?\*\/|\/\/.*)/,lookbehind:!0,greedy:!0},"quoted-identifier":{pattern:/#"(?:[^"\r\n]|"")*"(?!")/,greedy:!0},string:{pattern:/(?:#!)?"(?:[^"\r\n]|"")*"(?!")/,greedy:!0},constant:[/\bDay\.(?:Friday|Monday|Saturday|Sunday|Thursday|Tuesday|Wednesday)\b/,/\bTraceLevel\.(?:Critical|Error|Information|Verbose|Warning)\b/,/\bOccurrence\.(?:All|First|Last)\b/,/\bOrder\.(?:Ascending|Descending)\b/,/\bRoundingMode\.(?:AwayFromZero|Down|ToEven|TowardZero|Up)\b/,/\bMissingField\.(?:Error|Ignore|UseNull)\b/,/\bQuoteStyle\.(?:Csv|None)\b/,/\bJoinKind\.(?:FullOuter|Inner|LeftAnti|LeftOuter|RightAnti|RightOuter)\b/,/\bGroupKind\.(?:Global|Local)\b/,/\bExtraValues\.(?:Error|Ignore|List)\b/,/\bJoinAlgorithm\.(?:Dynamic|LeftHash|LeftIndex|PairwiseHash|RightHash|RightIndex|SortMerge)\b/,/\bJoinSide\.(?:Left|Right)\b/,/\bPrecision\.(?:Decimal|Double)\b/,/\bRelativePosition\.From(?:End|Start)\b/,/\bTextEncoding\.(?:Ascii|BigEndianUnicode|Unicode|Utf16|Utf8|Windows)\b/,/\b(?:Any|Binary|Date|DateTime|DateTimeZone|Duration|Function|Int16|Int32|Int64|Int8|List|Logical|None|Number|Record|Table|Text|Time)\.Type\b/,/\bnull\b/],boolean:/\b(?:false|true)\b/,keyword:/\b(?:and|as|each|else|error|if|in|is|let|meta|not|nullable|optional|or|otherwise|section|shared|then|try|type)\b|#(?:binary|date|datetime|datetimezone|duration|infinity|nan|sections|shared|table|time)\b/,function:{pattern:/(^|[^#\w.])[a-z_][\w.]*(?=\s*\()/i,lookbehind:!0},"data-type":{pattern:/\b(?:any|anynonnull|binary|date|datetime|datetimezone|duration|function|list|logical|none|number|record|table|text|time)\b/,alias:"class-name"},number:{pattern:/\b0x[\da-f]+\b|(?:[+-]?(?:\b\d+\.)?\b\d+|[+-]\.\d+|(^|[^.])\B\.\d+)(?:e[+-]?\d+)?\b/i,lookbehind:!0},operator:/[-+*\/&?@^]|<(?:=>?|>)?|>=?|=>?|\.\.\.?/,punctuation:/[,;\[\](){}]/},t.languages.pq=t.languages.powerquery,t.languages.mscript=t.languages.powerquery}return OE}var DE,JM;function Kve(){if(JM)return DE;JM=1,DE=e,e.displayName="powershell",e.aliases=[];function e(t){(function(n){var r=n.languages.powershell={comment:[{pattern:/(^|[^`])<#[\s\S]*?#>/,lookbehind:!0},{pattern:/(^|[^`])#.*/,lookbehind:!0}],string:[{pattern:/"(?:`[\s\S]|[^`"])*"/,greedy:!0,inside:null},{pattern:/'(?:[^']|'')*'/,greedy:!0}],namespace:/\[[a-z](?:\[(?:\[[^\]]*\]|[^\[\]])*\]|[^\[\]])*\]/i,boolean:/\$(?:false|true)\b/i,variable:/\$\w+\b/,function:[/\b(?:Add|Approve|Assert|Backup|Block|Checkpoint|Clear|Close|Compare|Complete|Compress|Confirm|Connect|Convert|ConvertFrom|ConvertTo|Copy|Debug|Deny|Disable|Disconnect|Dismount|Edit|Enable|Enter|Exit|Expand|Export|Find|ForEach|Format|Get|Grant|Group|Hide|Import|Initialize|Install|Invoke|Join|Limit|Lock|Measure|Merge|Move|New|Open|Optimize|Out|Ping|Pop|Protect|Publish|Push|Read|Receive|Redo|Register|Remove|Rename|Repair|Request|Reset|Resize|Resolve|Restart|Restore|Resume|Revoke|Save|Search|Select|Send|Set|Show|Skip|Sort|Split|Start|Step|Stop|Submit|Suspend|Switch|Sync|Tee|Test|Trace|Unblock|Undo|Uninstall|Unlock|Unprotect|Unpublish|Unregister|Update|Use|Wait|Watch|Where|Write)-[a-z]+\b/i,/\b(?:ac|cat|chdir|clc|cli|clp|clv|compare|copy|cp|cpi|cpp|cvpa|dbp|del|diff|dir|ebp|echo|epal|epcsv|epsn|erase|fc|fl|ft|fw|gal|gbp|gc|gci|gcs|gdr|gi|gl|gm|gp|gps|group|gsv|gu|gv|gwmi|iex|ii|ipal|ipcsv|ipsn|irm|iwmi|iwr|kill|lp|ls|measure|mi|mount|move|mp|mv|nal|ndr|ni|nv|ogv|popd|ps|pushd|pwd|rbp|rd|rdr|ren|ri|rm|rmdir|rni|rnp|rp|rv|rvpa|rwmi|sal|saps|sasv|sbp|sc|select|set|shcm|si|sl|sleep|sls|sort|sp|spps|spsv|start|sv|swmi|tee|trcm|type|write)\b/i],keyword:/\b(?:Begin|Break|Catch|Class|Continue|Data|Define|Do|DynamicParam|Else|ElseIf|End|Exit|Filter|Finally|For|ForEach|From|Function|If|InlineScript|Parallel|Param|Process|Return|Sequence|Switch|Throw|Trap|Try|Until|Using|Var|While|Workflow)\b/i,operator:{pattern:/(^|\W)(?:!|-(?:b?(?:and|x?or)|as|(?:Not)?(?:Contains|In|Like|Match)|eq|ge|gt|is(?:Not)?|Join|le|lt|ne|not|Replace|sh[lr])\b|-[-=]?|\+[+=]?|[*\/%]=?)/i,lookbehind:!0},punctuation:/[|{}[\];(),.]/};r.string[0].inside={function:{pattern:/(^|[^`])\$\((?:\$\([^\r\n()]*\)|(?!\$\()[^\r\n)])*\)/,lookbehind:!0,inside:r},boolean:r.boolean,variable:r.variable}})(t)}return DE}var IE,eF;function Xve(){if(eF)return IE;eF=1,IE=e,e.displayName="processing",e.aliases=[];function e(t){t.languages.processing=t.languages.extend("clike",{keyword:/\b(?:break|case|catch|class|continue|default|else|extends|final|for|if|implements|import|new|null|private|public|return|static|super|switch|this|try|void|while)\b/,function:/\b\w+(?=\s*\()/,operator:/<[<=]?|>[>=]?|&&?|\|\|?|[%?]|[!=+\-*\/]=?/}),t.languages.insertBefore("processing","number",{constant:/\b(?!XML\b)[A-Z][A-Z\d_]+\b/,type:{pattern:/\b(?:boolean|byte|char|color|double|float|int|[A-Z]\w*)\b/,alias:"class-name"}})}return IE}var LE,tF;function Zve(){if(tF)return LE;tF=1,LE=e,e.displayName="prolog",e.aliases=[];function e(t){t.languages.prolog={comment:{pattern:/\/\*[\s\S]*?\*\/|%.*/,greedy:!0},string:{pattern:/(["'])(?:\1\1|\\(?:\r\n|[\s\S])|(?!\1)[^\\\r\n])*\1(?!\1)/,greedy:!0},builtin:/\b(?:fx|fy|xf[xy]?|yfx?)\b/,function:/\b[a-z]\w*(?:(?=\()|\/\d+)/,number:/\b\d+(?:\.\d*)?/,operator:/[:\\=><\-?*@\/;+^|!$.]+|\b(?:is|mod|not|xor)\b/,punctuation:/[(){}\[\],]/}}return LE}var ME,nF;function Qve(){if(nF)return ME;nF=1,ME=e,e.displayName="promql",e.aliases=[];function e(t){(function(n){var r=["sum","min","max","avg","group","stddev","stdvar","count","count_values","bottomk","topk","quantile"],a=["on","ignoring","group_right","group_left","by","without"],o=["offset"],s=r.concat(a,o);n.languages.promql={comment:{pattern:/(^[ \t]*)#.*/m,lookbehind:!0},"vector-match":{pattern:new RegExp("((?:"+a.join("|")+")\\s*)\\([^)]*\\)"),lookbehind:!0,inside:{"label-key":{pattern:/\b[^,]+\b/,alias:"attr-name"},punctuation:/[(),]/}},"context-labels":{pattern:/\{[^{}]*\}/,inside:{"label-key":{pattern:/\b[a-z_]\w*(?=\s*(?:=|![=~]))/,alias:"attr-name"},"label-value":{pattern:/(["'`])(?:\\[\s\S]|(?!\1)[^\\])*\1/,greedy:!0,alias:"attr-value"},punctuation:/\{|\}|=~?|![=~]|,/}},"context-range":[{pattern:/\[[\w\s:]+\]/,inside:{punctuation:/\[|\]|:/,"range-duration":{pattern:/\b(?:\d+(?:[smhdwy]|ms))+\b/i,alias:"number"}}},{pattern:/(\boffset\s+)\w+/,lookbehind:!0,inside:{"range-duration":{pattern:/\b(?:\d+(?:[smhdwy]|ms))+\b/i,alias:"number"}}}],keyword:new RegExp("\\b(?:"+s.join("|")+")\\b","i"),function:/\b[a-z_]\w*(?=\s*\()/i,number:/[-+]?(?:(?:\b\d+(?:\.\d+)?|\B\.\d+)(?:e[-+]?\d+)?\b|\b(?:0x[0-9a-f]+|nan|inf)\b)/i,operator:/[\^*/%+-]|==|!=|<=|<|>=|>|\b(?:and|or|unless)\b/i,punctuation:/[{};()`,.[\]]/}})(t)}return ME}var FE,rF;function Jve(){if(rF)return FE;rF=1,FE=e,e.displayName="properties",e.aliases=[];function e(t){t.languages.properties={comment:/^[ \t]*[#!].*$/m,"attr-value":{pattern:/(^[ \t]*(?:\\(?:\r\n|[\s\S])|[^\\\s:=])+(?: *[=:] *(?! )| ))(?:\\(?:\r\n|[\s\S])|[^\\\r\n])+/m,lookbehind:!0},"attr-name":/^[ \t]*(?:\\(?:\r\n|[\s\S])|[^\\\s:=])+(?= *[=:]| )/m,punctuation:/[=:]/}}return FE}var PE,aF;function eSe(){if(aF)return PE;aF=1,PE=e,e.displayName="protobuf",e.aliases=[];function e(t){(function(n){var r=/\b(?:bool|bytes|double|s?fixed(?:32|64)|float|[su]?int(?:32|64)|string)\b/;n.languages.protobuf=n.languages.extend("clike",{"class-name":[{pattern:/(\b(?:enum|extend|message|service)\s+)[A-Za-z_]\w*(?=\s*\{)/,lookbehind:!0},{pattern:/(\b(?:rpc\s+\w+|returns)\s*\(\s*(?:stream\s+)?)\.?[A-Za-z_]\w*(?:\.[A-Za-z_]\w*)*(?=\s*\))/,lookbehind:!0}],keyword:/\b(?:enum|extend|extensions|import|message|oneof|option|optional|package|public|repeated|required|reserved|returns|rpc(?=\s+\w)|service|stream|syntax|to)\b(?!\s*=\s*\d)/,function:/\b[a-z_]\w*(?=\s*\()/i}),n.languages.insertBefore("protobuf","operator",{map:{pattern:/\bmap<\s*[\w.]+\s*,\s*[\w.]+\s*>(?=\s+[a-z_]\w*\s*[=;])/i,alias:"class-name",inside:{punctuation:/[<>.,]/,builtin:r}},builtin:r,"positional-class-name":{pattern:/(?:\b|\B\.)[a-z_]\w*(?:\.[a-z_]\w*)*(?=\s+[a-z_]\w*\s*[=;])/i,alias:"class-name",inside:{punctuation:/\./}},annotation:{pattern:/(\[\s*)[a-z_]\w*(?=\s*=)/i,lookbehind:!0}})})(t)}return PE}var zE,oF;function tSe(){if(oF)return zE;oF=1,zE=e,e.displayName="psl",e.aliases=[];function e(t){t.languages.psl={comment:{pattern:/#.*/,greedy:!0},string:{pattern:/"(?:\\.|[^\\"])*"/,greedy:!0,inside:{symbol:/\\[ntrbA-Z"\\]/}},"heredoc-string":{pattern:/<<<([a-zA-Z_]\w*)[\r\n](?:.*[\r\n])*?\1\b/,alias:"string",greedy:!0},keyword:/\b(?:__multi|__single|case|default|do|else|elsif|exit|export|for|foreach|function|if|last|line|local|next|requires|return|switch|until|while|word)\b/,constant:/\b(?:ALARM|CHART_ADD_GRAPH|CHART_DELETE_GRAPH|CHART_DESTROY|CHART_LOAD|CHART_PRINT|EOF|OFFLINE|OK|PSL_PROF_LOG|R_CHECK_HORIZ|R_CHECK_VERT|R_CLICKER|R_COLUMN|R_FRAME|R_ICON|R_LABEL|R_LABEL_CENTER|R_LIST_MULTIPLE|R_LIST_MULTIPLE_ND|R_LIST_SINGLE|R_LIST_SINGLE_ND|R_MENU|R_POPUP|R_POPUP_SCROLLED|R_RADIO_HORIZ|R_RADIO_VERT|R_ROW|R_SCALE_HORIZ|R_SCALE_VERT|R_SEP_HORIZ|R_SEP_VERT|R_SPINNER|R_TEXT_FIELD|R_TEXT_FIELD_LABEL|R_TOGGLE|TRIM_LEADING|TRIM_LEADING_AND_TRAILING|TRIM_REDUNDANT|TRIM_TRAILING|VOID|WARN)\b/,boolean:/\b(?:FALSE|False|NO|No|TRUE|True|YES|Yes|false|no|true|yes)\b/,variable:/\b(?:PslDebug|errno|exit_status)\b/,builtin:{pattern:/\b(?:PslExecute|PslFunctionCall|PslFunctionExists|PslSetOptions|_snmp_debug|acos|add_diary|annotate|annotate_get|ascii_to_ebcdic|asctime|asin|atan|atexit|batch_set|blackout|cat|ceil|chan_exists|change_state|close|code_cvt|cond_signal|cond_wait|console_type|convert_base|convert_date|convert_locale_date|cos|cosh|create|date|dcget_text|destroy|destroy_lock|dget_text|difference|dump_hist|ebcdic_to_ascii|encrypt|event_archive|event_catalog_get|event_check|event_query|event_range_manage|event_range_query|event_report|event_schedule|event_trigger|event_trigger2|execute|exists|exp|fabs|file|floor|fmod|fopen|fseek|ftell|full_discovery|get|get_chan_info|get_ranges|get_text|get_vars|getenv|gethostinfo|getpid|getpname|grep|history|history_get_retention|in_transition|index|int|internal|intersection|is_var|isnumber|join|kill|length|lines|lock|lock_info|log|log10|loge|matchline|msg_check|msg_get_format|msg_get_severity|msg_printf|msg_sprintf|ntharg|nthargf|nthline|nthlinef|num_bytes|num_consoles|pconfig|popen|poplines|pow|print|printf|proc_exists|process|random|read|readln|refresh_parameters|remote_check|remote_close|remote_event_query|remote_event_trigger|remote_file_send|remote_open|remove|replace|rindex|sec_check_priv|sec_store_get|sec_store_set|set|set_alarm_ranges|set_locale|share|sin|sinh|sleep|snmp_agent_config|snmp_agent_start|snmp_agent_stop|snmp_close|snmp_config|snmp_get|snmp_get_next|snmp_h_get|snmp_h_get_next|snmp_h_set|snmp_open|snmp_set|snmp_trap_ignore|snmp_trap_listen|snmp_trap_raise_std_trap|snmp_trap_receive|snmp_trap_register_im|snmp_trap_send|snmp_walk|sopen|sort|splitline|sprintf|sqrt|srandom|str_repeat|strcasecmp|subset|substr|system|tail|tan|tanh|text_domain|time|tmpnam|tolower|toupper|trace_psl_process|trim|union|unique|unlock|unset|va_arg|va_start|write)\b/,alias:"builtin-function"},"foreach-variable":{pattern:/(\bforeach\s+(?:(?:\w+\b|"(?:\\.|[^\\"])*")\s+){0,2})[_a-zA-Z]\w*(?=\s*\()/,lookbehind:!0,greedy:!0},function:/\b[_a-z]\w*\b(?=\s*\()/i,number:/\b(?:0x[0-9a-f]+|\d+(?:\.\d+)?)\b/i,operator:/--|\+\+|&&=?|\|\|=?|<<=?|>>=?|[=!]~|[-+*/%&|^!=<>]=?|\.|[:?]/,punctuation:/[(){}\[\];,]/}}return zE}var BE,iF;function nSe(){if(iF)return BE;iF=1,BE=e,e.displayName="pug",e.aliases=[];function e(t){(function(n){n.languages.pug={comment:{pattern:/(^([\t ]*))\/\/.*(?:(?:\r?\n|\r)\2[\t ].+)*/m,lookbehind:!0},"multiline-script":{pattern:/(^([\t ]*)script\b.*\.[\t ]*)(?:(?:\r?\n|\r(?!\n))(?:\2[\t ].+|\s*?(?=\r?\n|\r)))+/m,lookbehind:!0,inside:n.languages.javascript},filter:{pattern:/(^([\t ]*)):.+(?:(?:\r?\n|\r(?!\n))(?:\2[\t ].+|\s*?(?=\r?\n|\r)))+/m,lookbehind:!0,inside:{"filter-name":{pattern:/^:[\w-]+/,alias:"variable"},text:/\S[\s\S]*/}},"multiline-plain-text":{pattern:/(^([\t ]*)[\w\-#.]+\.[\t ]*)(?:(?:\r?\n|\r(?!\n))(?:\2[\t ].+|\s*?(?=\r?\n|\r)))+/m,lookbehind:!0},markup:{pattern:/(^[\t ]*)<.+/m,lookbehind:!0,inside:n.languages.markup},doctype:{pattern:/((?:^|\n)[\t ]*)doctype(?: .+)?/,lookbehind:!0},"flow-control":{pattern:/(^[\t ]*)(?:case|default|each|else|if|unless|when|while)\b(?: .+)?/m,lookbehind:!0,inside:{each:{pattern:/^each .+? in\b/,inside:{keyword:/\b(?:each|in)\b/,punctuation:/,/}},branch:{pattern:/^(?:case|default|else|if|unless|when|while)\b/,alias:"keyword"},rest:n.languages.javascript}},keyword:{pattern:/(^[\t ]*)(?:append|block|extends|include|prepend)\b.+/m,lookbehind:!0},mixin:[{pattern:/(^[\t ]*)mixin .+/m,lookbehind:!0,inside:{keyword:/^mixin/,function:/\w+(?=\s*\(|\s*$)/,punctuation:/[(),.]/}},{pattern:/(^[\t ]*)\+.+/m,lookbehind:!0,inside:{name:{pattern:/^\+\w+/,alias:"function"},rest:n.languages.javascript}}],script:{pattern:/(^[\t ]*script(?:(?:&[^(]+)?\([^)]+\))*[\t ]).+/m,lookbehind:!0,inside:n.languages.javascript},"plain-text":{pattern:/(^[\t ]*(?!-)[\w\-#.]*[\w\-](?:(?:&[^(]+)?\([^)]+\))*\/?[\t ]).+/m,lookbehind:!0},tag:{pattern:/(^[\t ]*)(?!-)[\w\-#.]*[\w\-](?:(?:&[^(]+)?\([^)]+\))*\/?:?/m,lookbehind:!0,inside:{attributes:[{pattern:/&[^(]+\([^)]+\)/,inside:n.languages.javascript},{pattern:/\([^)]+\)/,inside:{"attr-value":{pattern:/(=\s*(?!\s))(?:\{[^}]*\}|[^,)\r\n]+)/,lookbehind:!0,inside:n.languages.javascript},"attr-name":/[\w-]+(?=\s*!?=|\s*[,)])/,punctuation:/[!=(),]+/}}],punctuation:/:/,"attr-id":/#[\w\-]+/,"attr-class":/\.[\w\-]+/}},code:[{pattern:/(^[\t ]*(?:-|!?=)).+/m,lookbehind:!0,inside:n.languages.javascript}],punctuation:/[.\-!=|]+/};for(var r=/(^([\t ]*)):<filter_name>(?:(?:\r?\n|\r(?!\n))(?:\2[\t ].+|\s*?(?=\r?\n|\r)))+/.source,a=[{filter:"atpl",language:"twig"},{filter:"coffee",language:"coffeescript"},"ejs","handlebars","less","livescript","markdown",{filter:"sass",language:"scss"},"stylus"],o={},s=0,u=a.length;s<u;s++){var c=a[s];c=typeof c=="string"?{filter:c,language:c}:c,n.languages[c.language]&&(o["filter-"+c.filter]={pattern:RegExp(r.replace("<filter_name>",function(){return c.filter}),"m"),lookbehind:!0,inside:{"filter-name":{pattern:/^:[\w-]+/,alias:"variable"},text:{pattern:/\S[\s\S]*/,alias:[c.language,"language-"+c.language],inside:n.languages[c.language]}}})}n.languages.insertBefore("pug","filter",o)})(t)}return BE}var UE,sF;function rSe(){if(sF)return UE;sF=1,UE=e,e.displayName="puppet",e.aliases=[];function e(t){(function(n){n.languages.puppet={heredoc:[{pattern:/(@\("([^"\r\n\/):]+)"(?:\/[nrts$uL]*)?\).*(?:\r?\n|\r))(?:.*(?:\r?\n|\r(?!\n)))*?[ \t]*(?:\|[ \t]*)?(?:-[ \t]*)?\2/,lookbehind:!0,alias:"string",inside:{punctuation:/(?=\S).*\S(?= *$)/}},{pattern:/(@\(([^"\r\n\/):]+)(?:\/[nrts$uL]*)?\).*(?:\r?\n|\r))(?:.*(?:\r?\n|\r(?!\n)))*?[ \t]*(?:\|[ \t]*)?(?:-[ \t]*)?\2/,lookbehind:!0,greedy:!0,alias:"string",inside:{punctuation:/(?=\S).*\S(?= *$)/}},{pattern:/@\("?(?:[^"\r\n\/):]+)"?(?:\/[nrts$uL]*)?\)/,alias:"string",inside:{punctuation:{pattern:/(\().+?(?=\))/,lookbehind:!0}}}],"multiline-comment":{pattern:/(^|[^\\])\/\*[\s\S]*?\*\//,lookbehind:!0,greedy:!0,alias:"comment"},regex:{pattern:/((?:\bnode\s+|[~=\(\[\{,]\s*|[=+]>\s*|^\s*))\/(?:[^\/\\]|\\[\s\S])+\/(?:[imx]+\b|\B)/,lookbehind:!0,greedy:!0,inside:{"extended-regex":{pattern:/^\/(?:[^\/\\]|\\[\s\S])+\/[im]*x[im]*$/,inside:{comment:/#.*/}}}},comment:{pattern:/(^|[^\\])#.*/,lookbehind:!0,greedy:!0},string:{pattern:/(["'])(?:\$\{(?:[^'"}]|(["'])(?:(?!\2)[^\\]|\\[\s\S])*\2)+\}|\$(?!\{)|(?!\1)[^\\$]|\\[\s\S])*\1/,greedy:!0,inside:{"double-quoted":{pattern:/^"[\s\S]*"$/,inside:{}}}},variable:{pattern:/\$(?:::)?\w+(?:::\w+)*/,inside:{punctuation:/::/}},"attr-name":/(?:\b\w+|\*)(?=\s*=>)/,function:[{pattern:/(\.)(?!\d)\w+/,lookbehind:!0},/\b(?:contain|debug|err|fail|include|info|notice|realize|require|tag|warning)\b|\b(?!\d)\w+(?=\()/],number:/\b(?:0x[a-f\d]+|\d+(?:\.\d+)?(?:e-?\d+)?)\b/i,boolean:/\b(?:false|true)\b/,keyword:/\b(?:application|attr|case|class|consumes|default|define|else|elsif|function|if|import|inherits|node|private|produces|type|undef|unless)\b/,datatype:{pattern:/\b(?:Any|Array|Boolean|Callable|Catalogentry|Class|Collection|Data|Default|Enum|Float|Hash|Integer|NotUndef|Numeric|Optional|Pattern|Regexp|Resource|Runtime|Scalar|String|Struct|Tuple|Type|Undef|Variant)\b/,alias:"symbol"},operator:/=[=~>]?|![=~]?|<(?:<\|?|[=~|-])?|>[>=]?|->?|~>|\|>?>?|[*\/%+?]|\b(?:and|in|or)\b/,punctuation:/[\[\]{}().,;]|:+/};var r=[{pattern:/(^|[^\\])\$\{(?:[^'"{}]|\{[^}]*\}|(["'])(?:(?!\2)[^\\]|\\[\s\S])*\2)+\}/,lookbehind:!0,inside:{"short-variable":{pattern:/(^\$\{)(?!\w+\()(?:::)?\w+(?:::\w+)*/,lookbehind:!0,alias:"variable",inside:{punctuation:/::/}},delimiter:{pattern:/^\$/,alias:"variable"},rest:n.languages.puppet}},{pattern:/(^|[^\\])\$(?:::)?\w+(?:::\w+)*/,lookbehind:!0,alias:"variable",inside:{punctuation:/::/}}];n.languages.puppet.heredoc[0].inside.interpolation=r,n.languages.puppet.string.inside["double-quoted"].inside.interpolation=r})(t)}return UE}var jE,lF;function aSe(){if(lF)return jE;lF=1,jE=e,e.displayName="pure",e.aliases=[];function e(t){(function(n){n.languages.pure={comment:[{pattern:/(^|[^\\])\/\*[\s\S]*?\*\//,lookbehind:!0},{pattern:/(^|[^\\:])\/\/.*/,lookbehind:!0},/#!.+/],"inline-lang":{pattern:/%<[\s\S]+?%>/,greedy:!0,inside:{lang:{pattern:/(^%< *)-\*-.+?-\*-/,lookbehind:!0,alias:"comment"},delimiter:{pattern:/^%<.*|%>$/,alias:"punctuation"}}},string:{pattern:/"(?:\\.|[^"\\\r\n])*"/,greedy:!0},number:{pattern:/((?:\.\.)?)(?:\b(?:inf|nan)\b|\b0x[\da-f]+|(?:\b(?:0b)?\d+(?:\.\d+)?|\B\.\d+)(?:e[+-]?\d+)?L?)/i,lookbehind:!0},keyword:/\b(?:NULL|ans|break|bt|case|catch|cd|clear|const|def|del|dump|else|end|exit|extern|false|force|help|if|infix[lr]?|interface|let|ls|mem|namespace|nonfix|of|otherwise|outfix|override|postfix|prefix|private|public|pwd|quit|run|save|show|stats|then|throw|trace|true|type|underride|using|when|with)\b/,function:/\b(?:abs|add_(?:addr|constdef|(?:fundef|interface|macdef|typedef)(?:_at)?|vardef)|all|any|applp?|arity|bigintp?|blob(?:_crc|_size|p)?|boolp?|byte_c?string(?:_pointer)?|byte_(?:matrix|pointer)|calloc|cat|catmap|ceil|char[ps]?|check_ptrtag|chr|clear_sentry|clearsym|closurep?|cmatrixp?|cols?|colcat(?:map)?|colmap|colrev|colvector(?:p|seq)?|complex(?:_float_(?:matrix|pointer)|_matrix(?:_view)?|_pointer|p)?|conj|cookedp?|cst|cstring(?:_(?:dup|list|vector))?|curry3?|cyclen?|del_(?:constdef|fundef|interface|macdef|typedef|vardef)|delete|diag(?:mat)?|dim|dmatrixp?|do|double(?:_matrix(?:_view)?|_pointer|p)?|dowith3?|drop|dropwhile|eval(?:cmd)?|exactp|filter|fix|fixity|flip|float(?:_matrix|_pointer)|floor|fold[lr]1?|frac|free|funp?|functionp?|gcd|get(?:_(?:byte|constdef|double|float|fundef|int(?:64)?|interface(?:_typedef)?|long|macdef|pointer|ptrtag|sentry|short|string|typedef|vardef))?|globsym|hash|head|id|im|imatrixp?|index|inexactp|infp|init|insert|int(?:_matrix(?:_view)?|_pointer|p)?|int64_(?:matrix|pointer)|integerp?|iteraten?|iterwhile|join|keys?|lambdap?|last(?:err(?:pos)?)?|lcd|list[2p]?|listmap|make_ptrtag|malloc|map|matcat|matrixp?|max|member|min|nanp|nargs|nmatrixp?|null|numberp?|ord|pack(?:ed)?|pointer(?:_cast|_tag|_type|p)?|pow|pred|ptrtag|put(?:_(?:byte|double|float|int(?:64)?|long|pointer|short|string))?|rationalp?|re|realp?|realloc|recordp?|redim|reduce(?:_with)?|refp?|repeatn?|reverse|rlistp?|round|rows?|rowcat(?:map)?|rowmap|rowrev|rowvector(?:p|seq)?|same|scan[lr]1?|sentry|sgn|short_(?:matrix|pointer)|slice|smatrixp?|sort|split|str|strcat|stream|stride|string(?:_(?:dup|list|vector)|p)?|subdiag(?:mat)?|submat|subseq2?|substr|succ|supdiag(?:mat)?|symbolp?|tail|take|takewhile|thunkp?|transpose|trunc|tuplep?|typep|ubyte|uint(?:64)?|ulong|uncurry3?|unref|unzip3?|update|ushort|vals?|varp?|vector(?:p|seq)?|void|zip3?|zipwith3?)\b/,special:{pattern:/\b__[a-z]+__\b/i,alias:"builtin"},operator:/(?:[!"#$%&'*+,\-.\/:<=>?@\\^`|~\u00a1-\u00bf\u00d7-\u00f7\u20d0-\u2bff]|\b_+\b)+|\b(?:and|div|mod|not|or)\b/,punctuation:/[(){}\[\];,|]/};var r=["c",{lang:"c++",alias:"cpp"},"fortran"],a=/%< *-\*- *<lang>\d* *-\*-[\s\S]+?%>/.source;r.forEach(function(o){var s=o;if(typeof o!="string"&&(s=o.alias,o=o.lang),n.languages[s]){var u={};u["inline-lang-"+s]={pattern:RegExp(a.replace("<lang>",o.replace(/([.+*?\/\\(){}\[\]])/g,"\\$1")),"i"),inside:n.util.clone(n.languages.pure["inline-lang"].inside)},u["inline-lang-"+s].inside.rest=n.util.clone(n.languages[s]),n.languages.insertBefore("pure","inline-lang",u)}}),n.languages.c&&(n.languages.pure["inline-lang"].inside.rest=n.util.clone(n.languages.c))})(t)}return jE}var GE,uF;function oSe(){if(uF)return GE;uF=1,GE=e,e.displayName="purebasic",e.aliases=[];function e(t){t.languages.purebasic=t.languages.extend("clike",{comment:/;.*/,keyword:/\b(?:align|and|as|break|calldebugger|case|compilercase|compilerdefault|compilerelse|compilerelseif|compilerendif|compilerendselect|compilererror|compilerif|compilerselect|continue|data|datasection|debug|debuglevel|declare|declarec|declarecdll|declaredll|declaremodule|default|define|dim|disableasm|disabledebugger|disableexplicit|else|elseif|enableasm|enabledebugger|enableexplicit|end|enddatasection|enddeclaremodule|endenumeration|endif|endimport|endinterface|endmacro|endmodule|endprocedure|endselect|endstructure|endstructureunion|endwith|enumeration|extends|fakereturn|for|foreach|forever|global|gosub|goto|if|import|importc|includebinary|includefile|includepath|interface|macro|module|newlist|newmap|next|not|or|procedure|procedurec|procedurecdll|proceduredll|procedurereturn|protected|prototype|prototypec|read|redim|repeat|restore|return|runtime|select|shared|static|step|structure|structureunion|swap|threaded|to|until|wend|while|with|xincludefile|xor)\b/i,function:/\b\w+(?:\.\w+)?\s*(?=\()/,number:/(?:\$[\da-f]+|\b-?(?:\d+(?:\.\d+)?|\.\d+)(?:e[+-]?\d+)?)\b/i,operator:/(?:@\*?|\?|\*)\w+|-[>-]?|\+\+?|!=?|<<?=?|>>?=?|==?|&&?|\|?\||[~^%?*/@]/}),t.languages.insertBefore("purebasic","keyword",{tag:/#\w+\$?/,asm:{pattern:/(^[\t ]*)!.*/m,lookbehind:!0,alias:"tag",inside:{comment:/;.*/,string:{pattern:/(["'`])(?:\\.|(?!\1)[^\\\r\n])*\1/,greedy:!0},"label-reference-anonymous":{pattern:/(!\s*j[a-z]+\s+)@[fb]/i,lookbehind:!0,alias:"fasm-label"},"label-reference-addressed":{pattern:/(!\s*j[a-z]+\s+)[A-Z._?$@][\w.?$@~#]*/i,lookbehind:!0,alias:"fasm-label"},keyword:[/\b(?:extern|global)\b[^;\r\n]*/i,/\b(?:CPU|DEFAULT|FLOAT)\b.*/],function:{pattern:/^([\t ]*!\s*)[\da-z]+(?=\s|$)/im,lookbehind:!0},"function-inline":{pattern:/(:\s*)[\da-z]+(?=\s)/i,lookbehind:!0,alias:"function"},label:{pattern:/^([\t ]*!\s*)[A-Za-z._?$@][\w.?$@~#]*(?=:)/m,lookbehind:!0,alias:"fasm-label"},register:/\b(?:st\d|[xyz]mm\d\d?|[cdt]r\d|r\d\d?[bwd]?|[er]?[abcd]x|[abcd][hl]|[er]?(?:bp|di|si|sp)|[cdefgs]s|mm\d+)\b/i,number:/(?:\b|-|(?=\$))(?:0[hx](?:[\da-f]*\.)?[\da-f]+(?:p[+-]?\d+)?|\d[\da-f]+[hx]|\$\d[\da-f]*|0[oq][0-7]+|[0-7]+[oq]|0[by][01]+|[01]+[by]|0[dt]\d+|(?:\d+(?:\.\d+)?|\.\d+)(?:\.?e[+-]?\d+)?[dt]?)\b/i,operator:/[\[\]*+\-/%<>=&|$!,.:]/}}}),delete t.languages.purebasic["class-name"],delete t.languages.purebasic.boolean,t.languages.pbfasm=t.languages.purebasic}return GE}var HE,cF;function iSe(){if(cF)return HE;cF=1;var e=cA();HE=t,t.displayName="purescript",t.aliases=["purs"];function t(n){n.register(e),n.languages.purescript=n.languages.extend("haskell",{keyword:/\b(?:ado|case|class|data|derive|do|else|forall|if|in|infixl|infixr|instance|let|module|newtype|of|primitive|then|type|where)\b|∀/,"import-statement":{pattern:/(^[\t ]*)import\s+[A-Z][\w']*(?:\.[A-Z][\w']*)*(?:\s+as\s+[A-Z][\w']*(?:\.[A-Z][\w']*)*)?(?:\s+hiding\b)?/m,lookbehind:!0,inside:{keyword:/\b(?:as|hiding|import)\b/,punctuation:/\./}},builtin:/\b(?:absurd|add|ap|append|apply|between|bind|bottom|clamp|compare|comparing|compose|conj|const|degree|discard|disj|div|eq|flap|flip|gcd|identity|ifM|join|lcm|liftA1|liftM1|map|max|mempty|min|mod|mul|negate|not|notEq|one|otherwise|recip|show|sub|top|unit|unless|unlessM|void|when|whenM|zero)\b/,operator:[n.languages.haskell.operator[0],n.languages.haskell.operator[2],/[\xa2-\xa6\xa8\xa9\xac\xae-\xb1\xb4\xb8\xd7\xf7\u02c2-\u02c5\u02d2-\u02df\u02e5-\u02eb\u02ed\u02ef-\u02ff\u0375\u0384\u0385\u03f6\u0482\u058d-\u058f\u0606-\u0608\u060b\u060e\u060f\u06de\u06e9\u06fd\u06fe\u07f6\u07fe\u07ff\u09f2\u09f3\u09fa\u09fb\u0af1\u0b70\u0bf3-\u0bfa\u0c7f\u0d4f\u0d79\u0e3f\u0f01-\u0f03\u0f13\u0f15-\u0f17\u0f1a-\u0f1f\u0f34\u0f36\u0f38\u0fbe-\u0fc5\u0fc7-\u0fcc\u0fce\u0fcf\u0fd5-\u0fd8\u109e\u109f\u1390-\u1399\u166d\u17db\u1940\u19de-\u19ff\u1b61-\u1b6a\u1b74-\u1b7c\u1fbd\u1fbf-\u1fc1\u1fcd-\u1fcf\u1fdd-\u1fdf\u1fed-\u1fef\u1ffd\u1ffe\u2044\u2052\u207a-\u207c\u208a-\u208c\u20a0-\u20bf\u2100\u2101\u2103-\u2106\u2108\u2109\u2114\u2116-\u2118\u211e-\u2123\u2125\u2127\u2129\u212e\u213a\u213b\u2140-\u2144\u214a-\u214d\u214f\u218a\u218b\u2190-\u2307\u230c-\u2328\u232b-\u2426\u2440-\u244a\u249c-\u24e9\u2500-\u2767\u2794-\u27c4\u27c7-\u27e5\u27f0-\u2982\u2999-\u29d7\u29dc-\u29fb\u29fe-\u2b73\u2b76-\u2b95\u2b97-\u2bff\u2ce5-\u2cea\u2e50\u2e51\u2e80-\u2e99\u2e9b-\u2ef3\u2f00-\u2fd5\u2ff0-\u2ffb\u3004\u3012\u3013\u3020\u3036\u3037\u303e\u303f\u309b\u309c\u3190\u3191\u3196-\u319f\u31c0-\u31e3\u3200-\u321e\u322a-\u3247\u3250\u3260-\u327f\u328a-\u32b0\u32c0-\u33ff\u4dc0-\u4dff\ua490-\ua4c6\ua700-\ua716\ua720\ua721\ua789\ua78a\ua828-\ua82b\ua836-\ua839\uaa77-\uaa79\uab5b\uab6a\uab6b\ufb29\ufbb2-\ufbc1\ufdfc\ufdfd\ufe62\ufe64-\ufe66\ufe69\uff04\uff0b\uff1c-\uff1e\uff3e\uff40\uff5c\uff5e\uffe0-\uffe6\uffe8-\uffee\ufffc\ufffd]/]}),n.languages.purs=n.languages.purescript}return HE}var $E,dF;function sSe(){if(dF)return $E;dF=1,$E=e,e.displayName="python",e.aliases=["py"];function e(t){t.languages.python={comment:{pattern:/(^|[^\\])#.*/,lookbehind:!0,greedy:!0},"string-interpolation":{pattern:/(?:f|fr|rf)(?:("""|''')[\s\S]*?\1|("|')(?:\\.|(?!\2)[^\\\r\n])*\2)/i,greedy:!0,inside:{interpolation:{pattern:/((?:^|[^{])(?:\{\{)*)\{(?!\{)(?:[^{}]|\{(?!\{)(?:[^{}]|\{(?!\{)(?:[^{}])+\})+\})+\}/,lookbehind:!0,inside:{"format-spec":{pattern:/(:)[^:(){}]+(?=\}$)/,lookbehind:!0},"conversion-option":{pattern:/![sra](?=[:}]$)/,alias:"punctuation"},rest:null}},string:/[\s\S]+/}},"triple-quoted-string":{pattern:/(?:[rub]|br|rb)?("""|''')[\s\S]*?\1/i,greedy:!0,alias:"string"},string:{pattern:/(?:[rub]|br|rb)?("|')(?:\\.|(?!\1)[^\\\r\n])*\1/i,greedy:!0},function:{pattern:/((?:^|\s)def[ \t]+)[a-zA-Z_]\w*(?=\s*\()/g,lookbehind:!0},"class-name":{pattern:/(\bclass\s+)\w+/i,lookbehind:!0},decorator:{pattern:/(^[\t ]*)@\w+(?:\.\w+)*/m,lookbehind:!0,alias:["annotation","punctuation"],inside:{punctuation:/\./}},keyword:/\b(?:_(?=\s*:)|and|as|assert|async|await|break|case|class|continue|def|del|elif|else|except|exec|finally|for|from|global|if|import|in|is|lambda|match|nonlocal|not|or|pass|print|raise|return|try|while|with|yield)\b/,builtin:/\b(?:__import__|abs|all|any|apply|ascii|basestring|bin|bool|buffer|bytearray|bytes|callable|chr|classmethod|cmp|coerce|compile|complex|delattr|dict|dir|divmod|enumerate|eval|execfile|file|filter|float|format|frozenset|getattr|globals|hasattr|hash|help|hex|id|input|int|intern|isinstance|issubclass|iter|len|list|locals|long|map|max|memoryview|min|next|object|oct|open|ord|pow|property|range|raw_input|reduce|reload|repr|reversed|round|set|setattr|slice|sorted|staticmethod|str|sum|super|tuple|type|unichr|unicode|vars|xrange|zip)\b/,boolean:/\b(?:False|None|True)\b/,number:/\b0(?:b(?:_?[01])+|o(?:_?[0-7])+|x(?:_?[a-f0-9])+)\b|(?:\b\d+(?:_\d+)*(?:\.(?:\d+(?:_\d+)*)?)?|\B\.\d+(?:_\d+)*)(?:e[+-]?\d+(?:_\d+)*)?j?(?!\w)/i,operator:/[-+%=]=?|!=|:=|\*\*?=?|\/\/?=?|<[<=>]?|>[=>]?|[&|^~]/,punctuation:/[{}[\];(),.:]/},t.languages.python["string-interpolation"].inside.interpolation.inside.rest=t.languages.python,t.languages.py=t.languages.python}return $E}var qE,fF;function lSe(){if(fF)return qE;fF=1,qE=e,e.displayName="q",e.aliases=[];function e(t){t.languages.q={string:/"(?:\\.|[^"\\\r\n])*"/,comment:[{pattern:/([\t )\]}])\/.*/,lookbehind:!0,greedy:!0},{pattern:/(^|\r?\n|\r)\/[\t ]*(?:(?:\r?\n|\r)(?:.*(?:\r?\n|\r(?!\n)))*?(?:\\(?=[\t ]*(?:\r?\n|\r))|$)|\S.*)/,lookbehind:!0,greedy:!0},{pattern:/^\\[\t ]*(?:\r?\n|\r)[\s\S]+/m,greedy:!0},{pattern:/^#!.+/m,greedy:!0}],symbol:/`(?::\S+|[\w.]*)/,datetime:{pattern:/0N[mdzuvt]|0W[dtz]|\d{4}\.\d\d(?:m|\.\d\d(?:T(?:\d\d(?::\d\d(?::\d\d(?:[.:]\d\d\d)?)?)?)?)?[dz]?)|\d\d:\d\d(?::\d\d(?:[.:]\d\d\d)?)?[uvt]?/,alias:"number"},number:/\b(?![01]:)(?:0N[hje]?|0W[hj]?|0[wn]|0x[\da-fA-F]+|\d+(?:\.\d*)?(?:e[+-]?\d+)?[hjfeb]?)/,keyword:/\\\w+\b|\b(?:abs|acos|aj0?|all|and|any|asc|asin|asof|atan|attr|avgs?|binr?|by|ceiling|cols|cor|cos|count|cov|cross|csv|cut|delete|deltas|desc|dev|differ|distinct|div|do|dsave|ej|enlist|eval|except|exec|exit|exp|fby|fills|first|fkeys|flip|floor|from|get|getenv|group|gtime|hclose|hcount|hdel|hopen|hsym|iasc|identity|idesc|if|ij|in|insert|inter|inv|keys?|last|like|list|ljf?|load|log|lower|lsq|ltime|ltrim|mavg|maxs?|mcount|md5|mdev|med|meta|mins?|mmax|mmin|mmu|mod|msum|neg|next|not|null|or|over|parse|peach|pj|plist|prds?|prev|prior|rand|rank|ratios|raze|read0|read1|reciprocal|reval|reverse|rload|rotate|rsave|rtrim|save|scan|scov|sdev|select|set|setenv|show|signum|sin|sqrt|ssr?|string|sublist|sums?|sv|svar|system|tables|tan|til|trim|txf|type|uj|ungroup|union|update|upper|upsert|value|var|views?|vs|wavg|where|while|within|wj1?|wsum|ww|xasc|xbar|xcols?|xdesc|xexp|xgroup|xkey|xlog|xprev|xrank)\b/,adverb:{pattern:/['\/\\]:?|\beach\b/,alias:"function"},verb:{pattern:/(?:\B\.\B|\b[01]:|<[=>]?|>=?|[:+\-*%,!?~=|$&#@^]):?|\b_\b:?/,alias:"operator"},punctuation:/[(){}\[\];.]/}}return qE}var VE,pF;function uSe(){if(pF)return VE;pF=1,VE=e,e.displayName="qml",e.aliases=[];function e(t){(function(n){for(var r=/"(?:\\.|[^\\"\r\n])*"|'(?:\\.|[^\\'\r\n])*'/.source,a=/\/\/.*(?!.)|\/\*(?:[^*]|\*(?!\/))*\*\//.source,o=/(?:[^\\()[\]{}"'/]|<string>|\/(?![*/])|<comment>|\(<expr>*\)|\[<expr>*\]|\{<expr>*\}|\\[\s\S])/.source.replace(/<string>/g,function(){return r}).replace(/<comment>/g,function(){return a}),s=0;s<2;s++)o=o.replace(/<expr>/g,function(){return o});o=o.replace(/<expr>/g,"[^\\s\\S]"),n.languages.qml={comment:{pattern:/\/\/.*|\/\*[\s\S]*?\*\//,greedy:!0},"javascript-function":{pattern:RegExp(/((?:^|;)[ \t]*)function\s+(?!\s)[_$a-zA-Z\xA0-\uFFFF](?:(?!\s)[$\w\xA0-\uFFFF])*\s*\(<js>*\)\s*\{<js>*\}/.source.replace(/<js>/g,function(){return o}),"m"),lookbehind:!0,greedy:!0,alias:"language-javascript",inside:n.languages.javascript},"class-name":{pattern:/((?:^|[:;])[ \t]*)(?!\d)\w+(?=[ \t]*\{|[ \t]+on\b)/m,lookbehind:!0},property:[{pattern:/((?:^|[;{])[ \t]*)(?!\d)\w+(?:\.\w+)*(?=[ \t]*:)/m,lookbehind:!0},{pattern:/((?:^|[;{])[ \t]*)property[ \t]+(?!\d)\w+(?:\.\w+)*[ \t]+(?!\d)\w+(?:\.\w+)*(?=[ \t]*:)/m,lookbehind:!0,inside:{keyword:/^property/,property:/\w+(?:\.\w+)*/}}],"javascript-expression":{pattern:RegExp(/(:[ \t]*)(?![\s;}[])(?:(?!$|[;}])<js>)+/.source.replace(/<js>/g,function(){return o}),"m"),lookbehind:!0,greedy:!0,alias:"language-javascript",inside:n.languages.javascript},string:{pattern:/"(?:\\.|[^\\"\r\n])*"/,greedy:!0},keyword:/\b(?:as|import|on)\b/,punctuation:/[{}[\]:;,]/}})(t)}return VE}var WE,gF;function cSe(){if(gF)return WE;gF=1,WE=e,e.displayName="qore",e.aliases=[];function e(t){t.languages.qore=t.languages.extend("clike",{comment:{pattern:/(^|[^\\])(?:\/\*[\s\S]*?\*\/|(?:\/\/|#).*)/,lookbehind:!0},string:{pattern:/("|')(?:\\[\s\S]|(?!\1)[^\\])*\1/,greedy:!0},keyword:/\b(?:abstract|any|assert|binary|bool|boolean|break|byte|case|catch|char|class|code|const|continue|data|default|do|double|else|enum|extends|final|finally|float|for|goto|hash|if|implements|import|inherits|instanceof|int|interface|long|my|native|new|nothing|null|object|our|own|private|reference|rethrow|return|short|soft(?:bool|date|float|int|list|number|string)|static|strictfp|string|sub|super|switch|synchronized|this|throw|throws|transient|try|void|volatile|while)\b/,boolean:/\b(?:false|true)\b/i,function:/\$?\b(?!\d)\w+(?=\()/,number:/\b(?:0b[01]+|0x(?:[\da-f]*\.)?[\da-fp\-]+|(?:\d+(?:\.\d+)?|\.\d+)(?:e\d+)?[df]|(?:\d+(?:\.\d+)?|\.\d+))\b/i,operator:{pattern:/(^|[^.])(?:\+[+=]?|-[-=]?|[!=](?:==?|~)?|>>?=?|<(?:=>?|<=?)?|&[&=]?|\|[|=]?|[*\/%^]=?|[~?])/,lookbehind:!0},variable:/\$(?!\d)\w+\b/})}return WE}var YE,hF;function dSe(){if(hF)return YE;hF=1,YE=e,e.displayName="qsharp",e.aliases=["qs"];function e(t){(function(n){function r(y,S){return y.replace(/<<(\d+)>>/g,function(x,R){return"(?:"+S[+R]+")"})}function a(y,S,x){return RegExp(r(y,S),"")}function o(y,S){for(var x=0;x<S;x++)y=y.replace(/<<self>>/g,function(){return"(?:"+y+")"});return y.replace(/<<self>>/g,"[^\\s\\S]")}var s={type:"Adj BigInt Bool Ctl Double false Int One Pauli PauliI PauliX PauliY PauliZ Qubit Range Result String true Unit Zero",other:"Adjoint adjoint apply as auto body borrow borrowing Controlled controlled distribute elif else fail fixup for function if in internal intrinsic invert is let mutable namespace new newtype open operation repeat return self set until use using while within"};function u(y){return"\\b(?:"+y.trim().replace(/ /g,"|")+")\\b"}var c=RegExp(u(s.type+" "+s.other)),d=/\b[A-Za-z_]\w*\b/.source,p=r(/<<0>>(?:\s*\.\s*<<0>>)*/.source,[d]),g={keyword:c,punctuation:/[<>()?,.:[\]]/},m=/"(?:\\.|[^\\"])*"/.source;n.languages.qsharp=n.languages.extend("clike",{comment:/\/\/.*/,string:[{pattern:a(/(^|[^$\\])<<0>>/.source,[m]),lookbehind:!0,greedy:!0}],"class-name":[{pattern:a(/(\b(?:as|open)\s+)<<0>>(?=\s*(?:;|as\b))/.source,[p]),lookbehind:!0,inside:g},{pattern:a(/(\bnamespace\s+)<<0>>(?=\s*\{)/.source,[p]),lookbehind:!0,inside:g}],keyword:c,number:/(?:\b0(?:x[\da-f]+|b[01]+|o[0-7]+)|(?:\B\.\d+|\b\d+(?:\.\d*)?)(?:e[-+]?\d+)?)l?\b/i,operator:/\band=|\bor=|\band\b|\bnot\b|\bor\b|<[-=]|[-=]>|>>>=?|<<<=?|\^\^\^=?|\|\|\|=?|&&&=?|w\/=?|~~~|[*\/+\-^=!%]=?/,punctuation:/::|[{}[\];(),.:]/}),n.languages.insertBefore("qsharp","number",{range:{pattern:/\.\./,alias:"operator"}});var b=o(r(/\{(?:[^"{}]|<<0>>|<<self>>)*\}/.source,[m]),2);n.languages.insertBefore("qsharp","string",{"interpolation-string":{pattern:a(/\$"(?:\\.|<<0>>|[^\\"{])*"/.source,[b]),greedy:!0,inside:{interpolation:{pattern:a(/((?:^|[^\\])(?:\\\\)*)<<0>>/.source,[b]),lookbehind:!0,inside:{punctuation:/^\{|\}$/,expression:{pattern:/[\s\S]+/,alias:"language-qsharp",inside:n.languages.qsharp}}},string:/[\s\S]+/}}})})(t),t.languages.qs=t.languages.qsharp}return YE}var KE,mF;function fSe(){if(mF)return KE;mF=1,KE=e,e.displayName="r",e.aliases=[];function e(t){t.languages.r={comment:/#.*/,string:{pattern:/(['"])(?:\\.|(?!\1)[^\\\r\n])*\1/,greedy:!0},"percent-operator":{pattern:/%[^%\s]*%/,alias:"operator"},boolean:/\b(?:FALSE|TRUE)\b/,ellipsis:/\.\.(?:\.|\d+)/,number:[/\b(?:Inf|NaN)\b/,/(?:\b0x[\dA-Fa-f]+(?:\.\d*)?|\b\d+(?:\.\d*)?|\B\.\d+)(?:[EePp][+-]?\d+)?[iL]?/],keyword:/\b(?:NA|NA_character_|NA_complex_|NA_integer_|NA_real_|NULL|break|else|for|function|if|in|next|repeat|while)\b/,operator:/->?>?|<(?:=|<?-)?|[>=!]=?|::?|&&?|\|\|?|[+*\/^$@~]/,punctuation:/[(){}\[\],;]/}}return KE}var XE,bF;function pSe(){if(bF)return XE;bF=1;var e=gA();XE=t,t.displayName="racket",t.aliases=["rkt"];function t(n){n.register(e),n.languages.racket=n.languages.extend("scheme",{"lambda-parameter":{pattern:/([(\[]lambda\s+[(\[])[^()\[\]'\s]+/,lookbehind:!0}}),n.languages.insertBefore("racket","string",{lang:{pattern:/^#lang.+/m,greedy:!0,alias:"keyword"}}),n.languages.rkt=n.languages.racket}return XE}var ZE,yF;function gSe(){if(yF)return ZE;yF=1,ZE=e,e.displayName="reason",e.aliases=[];function e(t){t.languages.reason=t.languages.extend("clike",{string:{pattern:/"(?:\\(?:\r\n|[\s\S])|[^\\\r\n"])*"/,greedy:!0},"class-name":/\b[A-Z]\w*/,keyword:/\b(?:and|as|assert|begin|class|constraint|do|done|downto|else|end|exception|external|for|fun|function|functor|if|in|include|inherit|initializer|lazy|let|method|module|mutable|new|nonrec|object|of|open|or|private|rec|sig|struct|switch|then|to|try|type|val|virtual|when|while|with)\b/,operator:/\.{3}|:[:=]|\|>|->|=(?:==?|>)?|<=?|>=?|[|^?'#!~`]|[+\-*\/]\.?|\b(?:asr|land|lor|lsl|lsr|lxor|mod)\b/}),t.languages.insertBefore("reason","class-name",{char:{pattern:/'(?:\\x[\da-f]{2}|\\o[0-3][0-7][0-7]|\\\d{3}|\\.|[^'\\\r\n])'/,greedy:!0},constructor:/\b[A-Z]\w*\b(?!\s*\.)/,label:{pattern:/\b[a-z]\w*(?=::)/,alias:"symbol"}}),delete t.languages.reason.function}return ZE}var QE,vF;function hSe(){if(vF)return QE;vF=1,QE=e,e.displayName="regex",e.aliases=[];function e(t){(function(n){var r={pattern:/\\[\\(){}[\]^$+*?|.]/,alias:"escape"},a=/\\(?:x[\da-fA-F]{2}|u[\da-fA-F]{4}|u\{[\da-fA-F]+\}|0[0-7]{0,2}|[123][0-7]{2}|c[a-zA-Z]|.)/,o={pattern:/\.|\\[wsd]|\\p\{[^{}]+\}/i,alias:"class-name"},s={pattern:/\\[wsd]|\\p\{[^{}]+\}/i,alias:"class-name"},u="(?:[^\\\\-]|"+a.source+")",c=RegExp(u+"-"+u),d={pattern:/(<|')[^<>']+(?=[>']$)/,lookbehind:!0,alias:"variable"};n.languages.regex={"char-class":{pattern:/((?:^|[^\\])(?:\\\\)*)\[(?:[^\\\]]|\\[\s\S])*\]/,lookbehind:!0,inside:{"char-class-negation":{pattern:/(^\[)\^/,lookbehind:!0,alias:"operator"},"char-class-punctuation":{pattern:/^\[|\]$/,alias:"punctuation"},range:{pattern:c,inside:{escape:a,"range-punctuation":{pattern:/-/,alias:"operator"}}},"special-escape":r,"char-set":s,escape:a}},"special-escape":r,"char-set":o,backreference:[{pattern:/\\(?![123][0-7]{2})[1-9]/,alias:"keyword"},{pattern:/\\k<[^<>']+>/,alias:"keyword",inside:{"group-name":d}}],anchor:{pattern:/[$^]|\\[ABbGZz]/,alias:"function"},escape:a,group:[{pattern:/\((?:\?(?:<[^<>']+>|'[^<>']+'|[>:]|<?[=!]|[idmnsuxU]+(?:-[idmnsuxU]+)?:?))?/,alias:"punctuation",inside:{"group-name":d}},{pattern:/\)/,alias:"punctuation"}],quantifier:{pattern:/(?:[+*?]|\{\d+(?:,\d*)?\})[?+]?/,alias:"number"},alternation:{pattern:/\|/,alias:"keyword"}}})(t)}return QE}var JE,SF;function mSe(){if(SF)return JE;SF=1,JE=e,e.displayName="rego",e.aliases=[];function e(t){t.languages.rego={comment:/#.*/,property:{pattern:/(^|[^\\.])(?:"(?:\\.|[^\\"\r\n])*"|`[^`]*`|\b[a-z_]\w*\b)(?=\s*:(?!=))/i,lookbehind:!0,greedy:!0},string:{pattern:/(^|[^\\])"(?:\\.|[^\\"\r\n])*"|`[^`]*`/,lookbehind:!0,greedy:!0},keyword:/\b(?:as|default|else|import|not|null|package|set(?=\s*\()|some|with)\b/,boolean:/\b(?:false|true)\b/,function:{pattern:/\b[a-z_]\w*\b(?:\s*\.\s*\b[a-z_]\w*\b)*(?=\s*\()/i,inside:{namespace:/\b\w+\b(?=\s*\.)/,punctuation:/\./}},number:/-?\b\d+(?:\.\d+)?(?:e[+-]?\d+)?\b/i,operator:/[-+*/%|&]|[<>:=]=?|!=|\b_\b/,punctuation:/[,;.\[\]{}()]/}}return JE}var ew,EF;function bSe(){if(EF)return ew;EF=1,ew=e,e.displayName="renpy",e.aliases=["rpy"];function e(t){t.languages.renpy={comment:{pattern:/(^|[^\\])#.+/,lookbehind:!0},string:{pattern:/("""|''')[\s\S]+?\1|("|')(?:\\.|(?!\2)[^\\])*\2|(?:^#?(?:(?:[0-9a-fA-F]){3}|[0-9a-fA-F]{6})$)/m,greedy:!0},function:/\b[a-z_]\w*(?=\()/i,property:/\b(?:Update|UpdateVersion|action|activate_sound|adv_nvl_transition|after_load_transition|align|alpha|alt|anchor|antialias|area|auto|background|bar_invert|bar_resizing|bar_vertical|black_color|bold|bottom_bar|bottom_gutter|bottom_margin|bottom_padding|box_reverse|box_wrap|can_update|caret|child|color|crop|default_afm_enable|default_afm_time|default_fullscreen|default_text_cps|developer|directory_name|drag_handle|drag_joined|drag_name|drag_raise|draggable|dragged|drop_shadow|drop_shadow_color|droppable|dropped|easein|easeout|edgescroll|end_game_transition|end_splash_transition|enter_replay_transition|enter_sound|enter_transition|enter_yesno_transition|executable_name|exit_replay_transition|exit_sound|exit_transition|exit_yesno_transition|fadein|fadeout|first_indent|first_spacing|fit_first|focus|focus_mask|font|foreground|game_main_transition|get_installed_packages|google_play_key|google_play_salt|ground|has_music|has_sound|has_voice|height|help|hinting|hover|hover_background|hover_color|hover_sound|hovered|hyperlink_functions|idle|idle_color|image_style|include_update|insensitive|insensitive_background|insensitive_color|inside|intra_transition|italic|justify|kerning|keyboard_focus|language|layer_clipping|layers|layout|left_bar|left_gutter|left_margin|left_padding|length|line_leading|line_overlap_split|line_spacing|linear|main_game_transition|main_menu_music|maximum|min_width|minimum|minwidth|modal|mouse|mousewheel|name|narrator_menu|newline_indent|nvl_adv_transition|offset|order_reverse|outlines|overlay_functions|pos|position|prefix|radius|range|rest_indent|right_bar|right_gutter|right_margin|right_padding|rotate|rotate_pad|ruby_style|sample_sound|save_directory|say_attribute_transition|screen_height|screen_width|scrollbars|selected_hover|selected_hover_color|selected_idle|selected_idle_color|selected_insensitive|show_side_image|show_two_window|side_spacing|side_xpos|side_ypos|size|size_group|slow_cps|slow_cps_multiplier|spacing|strikethrough|subpixel|text_align|text_style|text_xpos|text_y_fudge|text_ypos|thumb|thumb_offset|thumb_shadow|thumbnail_height|thumbnail_width|time|top_bar|top_gutter|top_margin|top_padding|translations|underline|unscrollable|update|value|version|version_name|version_tuple|vertical|width|window_hide_transition|window_icon|window_left_padding|window_show_transition|window_title|windows_icon|xadjustment|xalign|xanchor|xanchoraround|xaround|xcenter|xfill|xinitial|xmargin|xmaximum|xminimum|xoffset|xofsset|xpadding|xpos|xsize|xzoom|yadjustment|yalign|yanchor|yanchoraround|yaround|ycenter|yfill|yinitial|ymargin|ymaximum|yminimum|yoffset|ypadding|ypos|ysize|ysizexysize|yzoom|zoom|zorder)\b/,tag:/\b(?:bar|block|button|buttoscreenn|drag|draggroup|fixed|frame|grid|[hv]box|hotbar|hotspot|image|imagebutton|imagemap|input|key|label|menu|mm_menu_frame|mousearea|nvl|parallel|screen|self|side|tag|text|textbutton|timer|vbar|viewport|window)\b|\$/,keyword:/\b(?:None|add|adjustment|alignaround|allow|angle|animation|around|as|assert|behind|box_layout|break|build|cache|call|center|changed|child_size|choice|circles|class|clear|clicked|clipping|clockwise|config|contains|continue|corner1|corner2|counterclockwise|def|default|define|del|delay|disabled|disabled_text|dissolve|elif|else|event|except|exclude|exec|expression|fade|finally|for|from|function|global|gm_root|has|hide|id|if|import|in|init|is|jump|knot|lambda|left|less_rounded|mm_root|movie|music|null|on|onlayer|pass|pause|persistent|play|print|python|queue|raise|random|renpy|repeat|return|right|rounded_window|scene|scope|set|show|slow|slow_abortable|slow_done|sound|stop|store|style|style_group|substitute|suffix|theme|transform|transform_anchor|transpose|try|ui|unhovered|updater|use|voice|while|widget|widget_hover|widget_selected|widget_text|yield)\b/,boolean:/\b(?:[Ff]alse|[Tt]rue)\b/,number:/(?:\b(?:0[bo])?(?:(?:\d|0x[\da-f])[\da-f]*(?:\.\d*)?)|\B\.\d+)(?:e[+-]?\d+)?j?/i,operator:/[-+%=]=?|!=|\*\*?=?|\/\/?=?|<[<=>]?|>[=>]?|[&|^~]|\b(?:and|at|not|or|with)\b/,punctuation:/[{}[\];(),.:]/},t.languages.rpy=t.languages.renpy}return ew}var tw,wF;function ySe(){if(wF)return tw;wF=1,tw=e,e.displayName="rest",e.aliases=[];function e(t){t.languages.rest={table:[{pattern:/(^[\t ]*)(?:\+[=-]+)+\+(?:\r?\n|\r)(?:\1[+|].+[+|](?:\r?\n|\r))+\1(?:\+[=-]+)+\+/m,lookbehind:!0,inside:{punctuation:/\||(?:\+[=-]+)+\+/}},{pattern:/(^[\t ]*)=+ [ =]*=(?:(?:\r?\n|\r)\1.+)+(?:\r?\n|\r)\1=+ [ =]*=(?=(?:\r?\n|\r){2}|\s*$)/m,lookbehind:!0,inside:{punctuation:/[=-]+/}}],"substitution-def":{pattern:/(^[\t ]*\.\. )\|(?:[^|\s](?:[^|]*[^|\s])?)\| [^:]+::/m,lookbehind:!0,inside:{substitution:{pattern:/^\|(?:[^|\s]|[^|\s][^|]*[^|\s])\|/,alias:"attr-value",inside:{punctuation:/^\||\|$/}},directive:{pattern:/( )(?! )[^:]+::/,lookbehind:!0,alias:"function",inside:{punctuation:/::$/}}}},"link-target":[{pattern:/(^[\t ]*\.\. )\[[^\]]+\]/m,lookbehind:!0,alias:"string",inside:{punctuation:/^\[|\]$/}},{pattern:/(^[\t ]*\.\. )_(?:`[^`]+`|(?:[^:\\]|\\.)+):/m,lookbehind:!0,alias:"string",inside:{punctuation:/^_|:$/}}],directive:{pattern:/(^[\t ]*\.\. )[^:]+::/m,lookbehind:!0,alias:"function",inside:{punctuation:/::$/}},comment:{pattern:/(^[\t ]*\.\.)(?:(?: .+)?(?:(?:\r?\n|\r).+)+| .+)(?=(?:\r?\n|\r){2}|$)/m,lookbehind:!0},title:[{pattern:/^(([!"#$%&'()*+,\-.\/:;<=>?@\[\\\]^_`{|}~])\2+)(?:\r?\n|\r).+(?:\r?\n|\r)\1$/m,inside:{punctuation:/^[!"#$%&'()*+,\-.\/:;<=>?@\[\\\]^_`{|}~]+|[!"#$%&'()*+,\-.\/:;<=>?@\[\\\]^_`{|}~]+$/,important:/.+/}},{pattern:/(^|(?:\r?\n|\r){2}).+(?:\r?\n|\r)([!"#$%&'()*+,\-.\/:;<=>?@\[\\\]^_`{|}~])\2+(?=\r?\n|\r|$)/,lookbehind:!0,inside:{punctuation:/[!"#$%&'()*+,\-.\/:;<=>?@\[\\\]^_`{|}~]+$/,important:/.+/}}],hr:{pattern:/((?:\r?\n|\r){2})([!"#$%&'()*+,\-.\/:;<=>?@\[\\\]^_`{|}~])\2{3,}(?=(?:\r?\n|\r){2})/,lookbehind:!0,alias:"punctuation"},field:{pattern:/(^[\t ]*):[^:\r\n]+:(?= )/m,lookbehind:!0,alias:"attr-name"},"command-line-option":{pattern:/(^[\t ]*)(?:[+-][a-z\d]|(?:--|\/)[a-z\d-]+)(?:[ =](?:[a-z][\w-]*|<[^<>]+>))?(?:, (?:[+-][a-z\d]|(?:--|\/)[a-z\d-]+)(?:[ =](?:[a-z][\w-]*|<[^<>]+>))?)*(?=(?:\r?\n|\r)? {2,}\S)/im,lookbehind:!0,alias:"symbol"},"literal-block":{pattern:/::(?:\r?\n|\r){2}([ \t]+)(?![ \t]).+(?:(?:\r?\n|\r)\1.+)*/,inside:{"literal-block-punctuation":{pattern:/^::/,alias:"punctuation"}}},"quoted-literal-block":{pattern:/::(?:\r?\n|\r){2}([!"#$%&'()*+,\-.\/:;<=>?@\[\\\]^_`{|}~]).*(?:(?:\r?\n|\r)\1.*)*/,inside:{"literal-block-punctuation":{pattern:/^(?:::|([!"#$%&'()*+,\-.\/:;<=>?@\[\\\]^_`{|}~])\1*)/m,alias:"punctuation"}}},"list-bullet":{pattern:/(^[\t ]*)(?:[*+\-•‣⁃]|\(?(?:\d+|[a-z]|[ivxdclm]+)\)|(?:\d+|[a-z]|[ivxdclm]+)\.)(?= )/im,lookbehind:!0,alias:"punctuation"},"doctest-block":{pattern:/(^[\t ]*)>>> .+(?:(?:\r?\n|\r).+)*/m,lookbehind:!0,inside:{punctuation:/^>>>/}},inline:[{pattern:/(^|[\s\-:\/'"<(\[{])(?::[^:]+:`.*?`|`.*?`:[^:]+:|(\*\*?|``?|\|)(?!\s)(?:(?!\2).)*\S\2(?=[\s\-.,:;!?\\\/'")\]}]|$))/m,lookbehind:!0,inside:{bold:{pattern:/(^\*\*).+(?=\*\*$)/,lookbehind:!0},italic:{pattern:/(^\*).+(?=\*$)/,lookbehind:!0},"inline-literal":{pattern:/(^``).+(?=``$)/,lookbehind:!0,alias:"symbol"},role:{pattern:/^:[^:]+:|:[^:]+:$/,alias:"function",inside:{punctuation:/^:|:$/}},"interpreted-text":{pattern:/(^`).+(?=`$)/,lookbehind:!0,alias:"attr-value"},substitution:{pattern:/(^\|).+(?=\|$)/,lookbehind:!0,alias:"attr-value"},punctuation:/\*\*?|``?|\|/}}],link:[{pattern:/\[[^\[\]]+\]_(?=[\s\-.,:;!?\\\/'")\]}]|$)/,alias:"string",inside:{punctuation:/^\[|\]_$/}},{pattern:/(?:\b[a-z\d]+(?:[_.:+][a-z\d]+)*_?_|`[^`]+`_?_|_`[^`]+`)(?=[\s\-.,:;!?\\\/'")\]}]|$)/i,alias:"string",inside:{punctuation:/^_?`|`$|`?_?_$/}}],punctuation:{pattern:/(^[\t ]*)(?:\|(?= |$)|(?:---?|—|\.\.|__)(?= )|\.\.$)/m,lookbehind:!0}}}return tw}var nw,xF;function vSe(){if(xF)return nw;xF=1,nw=e,e.displayName="rip",e.aliases=[];function e(t){t.languages.rip={comment:{pattern:/#.*/,greedy:!0},char:{pattern:/\B`[^\s`'",.:;#\/\\()<>\[\]{}]\b/,greedy:!0},string:{pattern:/("|')(?:\\.|(?!\1)[^\\\r\n])*\1/,greedy:!0},regex:{pattern:/(^|[^/])\/(?!\/)(?:\[[^\n\r\]]*\]|\\.|[^/\\\r\n\[])+\/(?=\s*(?:$|[\r\n,.;})]))/,lookbehind:!0,greedy:!0},keyword:/(?:=>|->)|\b(?:case|catch|class|else|exit|finally|if|raise|return|switch|try)\b/,builtin:/@|\bSystem\b/,boolean:/\b(?:false|true)\b/,date:/\b\d{4}-\d{2}-\d{2}\b/,time:/\b\d{2}:\d{2}:\d{2}\b/,datetime:/\b\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\b/,symbol:/:[^\d\s`'",.:;#\/\\()<>\[\]{}][^\s`'",.:;#\/\\()<>\[\]{}]*/,number:/[+-]?\b(?:\d+\.\d+|\d+)\b/,punctuation:/(?:\.{2,3})|[`,.:;=\/\\()<>\[\]{}]/,reference:/[^\d\s`'",.:;#\/\\()<>\[\]{}][^\s`'",.:;#\/\\()<>\[\]{}]*/}}return nw}var rw,kF;function SSe(){if(kF)return rw;kF=1,rw=e,e.displayName="roboconf",e.aliases=[];function e(t){t.languages.roboconf={comment:/#.*/,keyword:{pattern:/(^|\s)(?:(?:external|import)\b|(?:facet|instance of)(?=[ \t]+[\w-]+[ \t]*\{))/,lookbehind:!0},component:{pattern:/[\w-]+(?=[ \t]*\{)/,alias:"variable"},property:/[\w.-]+(?=[ \t]*:)/,value:{pattern:/(=[ \t]*(?![ \t]))[^,;]+/,lookbehind:!0,alias:"attr-value"},optional:{pattern:/\(optional\)/,alias:"builtin"},wildcard:{pattern:/(\.)\*/,lookbehind:!0,alias:"operator"},punctuation:/[{},.;:=]/}}return rw}var aw,TF;function ESe(){if(TF)return aw;TF=1,aw=e,e.displayName="robotframework",e.aliases=[];function e(t){(function(n){var r={pattern:/(^[ \t]*| {2}|\t)#.*/m,lookbehind:!0,greedy:!0},a={pattern:/((?:^|[^\\])(?:\\{2})*)[$@&%]\{(?:[^{}\r\n]|\{[^{}\r\n]*\})*\}/,lookbehind:!0,inside:{punctuation:/^[$@&%]\{|\}$/}};function o(d,p){var g={};g["section-header"]={pattern:/^ ?\*{3}.+?\*{3}/,alias:"keyword"};for(var m in p)g[m]=p[m];return g.tag={pattern:/([\r\n](?: {2}|\t)[ \t]*)\[[-\w]+\]/,lookbehind:!0,inside:{punctuation:/\[|\]/}},g.variable=a,g.comment=r,{pattern:RegExp(/^ ?\*{3}[ \t]*<name>[ \t]*\*{3}(?:.|[\r\n](?!\*{3}))*/.source.replace(/<name>/g,function(){return d}),"im"),alias:"section",inside:g}}var s={pattern:/(\[Documentation\](?: {2}|\t)[ \t]*)(?![ \t]|#)(?:.|(?:\r\n?|\n)[ \t]*\.{3})+/,lookbehind:!0,alias:"string"},u={pattern:/([\r\n] ?)(?!#)(?:\S(?:[ \t]\S)*)+/,lookbehind:!0,alias:"function",inside:{variable:a}},c={pattern:/([\r\n](?: {2}|\t)[ \t]*)(?!\[|\.{3}|#)(?:\S(?:[ \t]\S)*)+/,lookbehind:!0,inside:{variable:a}};n.languages.robotframework={settings:o("Settings",{documentation:{pattern:/([\r\n] ?Documentation(?: {2}|\t)[ \t]*)(?![ \t]|#)(?:.|(?:\r\n?|\n)[ \t]*\.{3})+/,lookbehind:!0,alias:"string"},property:{pattern:/([\r\n] ?)(?!\.{3}|#)(?:\S(?:[ \t]\S)*)+/,lookbehind:!0}}),variables:o("Variables"),"test-cases":o("Test Cases",{"test-name":u,documentation:s,property:c}),keywords:o("Keywords",{"keyword-name":u,documentation:s,property:c}),tasks:o("Tasks",{"task-name":u,documentation:s,property:c}),comment:r},n.languages.robot=n.languages.robotframework})(t)}return aw}var ow,AF;function wSe(){if(AF)return ow;AF=1,ow=e,e.displayName="rust",e.aliases=[];function e(t){(function(n){for(var r=/\/\*(?:[^*/]|\*(?!\/)|\/(?!\*)|<self>)*\*\//.source,a=0;a<2;a++)r=r.replace(/<self>/g,function(){return r});r=r.replace(/<self>/g,function(){return/[^\s\S]/.source}),n.languages.rust={comment:[{pattern:RegExp(/(^|[^\\])/.source+r),lookbehind:!0,greedy:!0},{pattern:/(^|[^\\:])\/\/.*/,lookbehind:!0,greedy:!0}],string:{pattern:/b?"(?:\\[\s\S]|[^\\"])*"|b?r(#*)"(?:[^"]|"(?!\1))*"\1/,greedy:!0},char:{pattern:/b?'(?:\\(?:x[0-7][\da-fA-F]|u\{(?:[\da-fA-F]_*){1,6}\}|.)|[^\\\r\n\t'])'/,greedy:!0},attribute:{pattern:/#!?\[(?:[^\[\]"]|"(?:\\[\s\S]|[^\\"])*")*\]/,greedy:!0,alias:"attr-name",inside:{string:null}},"closure-params":{pattern:/([=(,:]\s*|\bmove\s*)\|[^|]*\||\|[^|]*\|(?=\s*(?:\{|->))/,lookbehind:!0,greedy:!0,inside:{"closure-punctuation":{pattern:/^\||\|$/,alias:"punctuation"},rest:null}},"lifetime-annotation":{pattern:/'\w+/,alias:"symbol"},"fragment-specifier":{pattern:/(\$\w+:)[a-z]+/,lookbehind:!0,alias:"punctuation"},variable:/\$\w+/,"function-definition":{pattern:/(\bfn\s+)\w+/,lookbehind:!0,alias:"function"},"type-definition":{pattern:/(\b(?:enum|struct|trait|type|union)\s+)\w+/,lookbehind:!0,alias:"class-name"},"module-declaration":[{pattern:/(\b(?:crate|mod)\s+)[a-z][a-z_\d]*/,lookbehind:!0,alias:"namespace"},{pattern:/(\b(?:crate|self|super)\s*)::\s*[a-z][a-z_\d]*\b(?:\s*::(?:\s*[a-z][a-z_\d]*\s*::)*)?/,lookbehind:!0,alias:"namespace",inside:{punctuation:/::/}}],keyword:[/\b(?:Self|abstract|as|async|await|become|box|break|const|continue|crate|do|dyn|else|enum|extern|final|fn|for|if|impl|in|let|loop|macro|match|mod|move|mut|override|priv|pub|ref|return|self|static|struct|super|trait|try|type|typeof|union|unsafe|unsized|use|virtual|where|while|yield)\b/,/\b(?:bool|char|f(?:32|64)|[ui](?:8|16|32|64|128|size)|str)\b/],function:/\b[a-z_]\w*(?=\s*(?:::\s*<|\())/,macro:{pattern:/\b\w+!/,alias:"property"},constant:/\b[A-Z_][A-Z_\d]+\b/,"class-name":/\b[A-Z]\w*\b/,namespace:{pattern:/(?:\b[a-z][a-z_\d]*\s*::\s*)*\b[a-z][a-z_\d]*\s*::(?!\s*<)/,inside:{punctuation:/::/}},number:/\b(?:0x[\dA-Fa-f](?:_?[\dA-Fa-f])*|0o[0-7](?:_?[0-7])*|0b[01](?:_?[01])*|(?:(?:\d(?:_?\d)*)?\.)?\d(?:_?\d)*(?:[Ee][+-]?\d+)?)(?:_?(?:f32|f64|[iu](?:8|16|32|64|size)?))?\b/,boolean:/\b(?:false|true)\b/,punctuation:/->|\.\.=|\.{1,3}|::|[{}[\];(),:]/,operator:/[-+*\/%!^]=?|=[=>]?|&[&=]?|\|[|=]?|<<?=?|>>?=?|[@?]/},n.languages.rust["closure-params"].inside.rest=n.languages.rust,n.languages.rust.attribute.inside.string=n.languages.rust.string})(t)}return ow}var iw,RF;function xSe(){if(RF)return iw;RF=1,iw=e,e.displayName="sas",e.aliases=[];function e(t){(function(n){var r=/(?:"(?:""|[^"])*"(?!")|'(?:''|[^'])*'(?!'))/.source,a=/\b(?:\d[\da-f]*x|\d+(?:\.\d+)?(?:e[+-]?\d+)?)\b/i,o={pattern:RegExp(r+"[bx]"),alias:"number"},s={pattern:/&[a-z_]\w*/i},u={pattern:/((?:^|\s|=|\())%(?:ABORT|BY|CMS|COPY|DISPLAY|DO|ELSE|END|EVAL|GLOBAL|GO|GOTO|IF|INC|INCLUDE|INDEX|INPUT|KTRIM|LENGTH|LET|LIST|LOCAL|PUT|QKTRIM|QSCAN|QSUBSTR|QSYSFUNC|QUPCASE|RETURN|RUN|SCAN|SUBSTR|SUPERQ|SYMDEL|SYMEXIST|SYMGLOBL|SYMLOCAL|SYSCALL|SYSEVALF|SYSEXEC|SYSFUNC|SYSGET|SYSRPUT|THEN|TO|TSO|UNQUOTE|UNTIL|UPCASE|WHILE|WINDOW)\b/i,lookbehind:!0,alias:"keyword"},c={pattern:/(^|\s)(?:proc\s+\w+|data(?!=)|quit|run)\b/i,alias:"keyword",lookbehind:!0},d=[/\/\*[\s\S]*?\*\//,{pattern:/(^[ \t]*|;\s*)\*[^;]*;/m,lookbehind:!0}],p={pattern:RegExp(r),greedy:!0},g=/[$%@.(){}\[\];,\\]/,m={pattern:/%?\b\w+(?=\()/,alias:"keyword"},b={function:m,"arg-value":{pattern:/(=\s*)[A-Z\.]+/i,lookbehind:!0},operator:/=/,"macro-variable":s,arg:{pattern:/[A-Z]+/i,alias:"keyword"},number:a,"numeric-constant":o,punctuation:g,string:p},y={pattern:/\b(?:format|put)\b=?[\w'$.]+/i,inside:{keyword:/^(?:format|put)(?==)/i,equals:/=/,format:{pattern:/(?:\w|\$\d)+\.\d?/,alias:"number"}}},S={pattern:/\b(?:format|put)\s+[\w']+(?:\s+[$.\w]+)+(?=;)/i,inside:{keyword:/^(?:format|put)/i,format:{pattern:/[\w$]+\.\d?/,alias:"number"}}},x={pattern:/((?:^|\s)=?)(?:catname|checkpoint execute_always|dm|endsas|filename|footnote|%include|libname|%list|lock|missing|options|page|resetline|%run|sasfile|skip|sysecho|title\d?)\b/i,lookbehind:!0,alias:"keyword"},R={pattern:/(^|\s)(?:submit(?:\s+(?:load|norun|parseonly))?|endsubmit)\b/i,lookbehind:!0,alias:"keyword"},k=/aStore|accessControl|aggregation|audio|autotune|bayesianNetClassifier|bioMedImage|boolRule|builtins|cardinality|cdm|clustering|conditionalRandomFields|configuration|copula|countreg|dataDiscovery|dataPreprocess|dataSciencePilot|dataStep|decisionTree|deduplication|deepLearn|deepNeural|deepRnn|ds2|ecm|entityRes|espCluster|explainModel|factmac|fastKnn|fcmpact|fedSql|freqTab|gVarCluster|gam|gleam|graphSemiSupLearn|hiddenMarkovModel|hyperGroup|ica|image|iml|kernalPca|langModel|ldaTopic|loadStreams|mbc|mixed|mlTools|modelPublishing|network|neuralNet|nmf|nonParametricBayes|nonlinear|optNetwork|optimization|panel|pca|percentile|phreg|pls|qkb|qlim|quantreg|recommend|regression|reinforcementLearn|robustPca|ruleMining|sampling|sandwich|sccasl|search(?:Analytics)?|sentimentAnalysis|sequence|session(?:Prop)?|severity|simSystem|simple|smartData|sparkEmbeddedProcess|sparseML|spatialreg|spc|stabilityMonitoring|svDataDescription|svm|table|text(?:Filters|Frequency|Mining|Parse|Rule(?:Develop|Score)|Topic|Util)|timeData|transpose|tsInfo|tsReconcile|uniTimeSeries|varReduce/.source,A={pattern:RegExp(/(^|\s)(?:action\s+)?(?:<act>)\.[a-z]+\b[^;]+/.source.replace(/<act>/g,function(){return k}),"i"),lookbehind:!0,inside:{keyword:RegExp(/(?:<act>)\.[a-z]+\b/.source.replace(/<act>/g,function(){return k}),"i"),action:{pattern:/(?:action)/i,alias:"keyword"},comment:d,function:m,"arg-value":b["arg-value"],operator:b.operator,argument:b.arg,number:a,"numeric-constant":o,punctuation:g,string:p}},C={pattern:/((?:^|\s)=?)(?:after|analysis|and|array|barchart|barwidth|begingraph|by|call|cas|cbarline|cfill|class(?:lev)?|close|column|computed?|contains|continue|data(?==)|define|delete|describe|document|do\s+over|do|dol|drop|dul|else|end(?:comp|source)?|entryTitle|eval(?:uate)?|exec(?:ute)?|exit|file(?:name)?|fill(?:attrs)?|flist|fnc|function(?:list)?|global|goto|group(?:by)?|headline|headskip|histogram|if|infile|keep|keylabel|keyword|label|layout|leave|legendlabel|length|libname|loadactionset|merge|midpoints|_?null_|name|noobs|nowd|ods|options|or|otherwise|out(?:put)?|over(?:lay)?|plot|print|put|raise|ranexp|rannor|rbreak|retain|return|select|session|sessref|set|source|statgraph|sum|summarize|table|temp|terminate|then\s+do|then|title\d?|to|var|when|where|xaxisopts|y2axisopts|yaxisopts)\b/i,lookbehind:!0};n.languages.sas={datalines:{pattern:/^([ \t]*)(?:cards|(?:data)?lines);[\s\S]+?^[ \t]*;/im,lookbehind:!0,alias:"string",inside:{keyword:{pattern:/^(?:cards|(?:data)?lines)/i},punctuation:/;/}},"proc-sql":{pattern:/(^proc\s+(?:fed)?sql(?:\s+[\w|=]+)?;)[\s\S]+?(?=^(?:proc\s+\w+|data|quit|run);|(?![\s\S]))/im,lookbehind:!0,inside:{sql:{pattern:RegExp(/^[ \t]*(?:select|alter\s+table|(?:create|describe|drop)\s+(?:index|table(?:\s+constraints)?|view)|create\s+unique\s+index|insert\s+into|update)(?:<str>|[^;"'])+;/.source.replace(/<str>/g,function(){return r}),"im"),alias:"language-sql",inside:n.languages.sql},"global-statements":x,"sql-statements":{pattern:/(^|\s)(?:disconnect\s+from|begin|commit|exec(?:ute)?|reset|rollback|validate)\b/i,lookbehind:!0,alias:"keyword"},number:a,"numeric-constant":o,punctuation:g,string:p}},"proc-groovy":{pattern:/(^proc\s+groovy(?:\s+[\w|=]+)?;)[\s\S]+?(?=^(?:proc\s+\w+|data|quit|run);|(?![\s\S]))/im,lookbehind:!0,inside:{comment:d,groovy:{pattern:RegExp(/(^[ \t]*submit(?:\s+(?:load|norun|parseonly))?)(?:<str>|[^"'])+?(?=endsubmit;)/.source.replace(/<str>/g,function(){return r}),"im"),lookbehind:!0,alias:"language-groovy",inside:n.languages.groovy},keyword:C,"submit-statement":R,"global-statements":x,number:a,"numeric-constant":o,punctuation:g,string:p}},"proc-lua":{pattern:/(^proc\s+lua(?:\s+[\w|=]+)?;)[\s\S]+?(?=^(?:proc\s+\w+|data|quit|run);|(?![\s\S]))/im,lookbehind:!0,inside:{comment:d,lua:{pattern:RegExp(/(^[ \t]*submit(?:\s+(?:load|norun|parseonly))?)(?:<str>|[^"'])+?(?=endsubmit;)/.source.replace(/<str>/g,function(){return r}),"im"),lookbehind:!0,alias:"language-lua",inside:n.languages.lua},keyword:C,"submit-statement":R,"global-statements":x,number:a,"numeric-constant":o,punctuation:g,string:p}},"proc-cas":{pattern:/(^proc\s+cas(?:\s+[\w|=]+)?;)[\s\S]+?(?=^(?:proc\s+\w+|quit|data);|(?![\s\S]))/im,lookbehind:!0,inside:{comment:d,"statement-var":{pattern:/((?:^|\s)=?)saveresult\s[^;]+/im,lookbehind:!0,inside:{statement:{pattern:/^saveresult\s+\S+/i,inside:{keyword:/^(?:saveresult)/i}},rest:b}},"cas-actions":A,statement:{pattern:/((?:^|\s)=?)(?:default|(?:un)?set|on|output|upload)[^;]+/im,lookbehind:!0,inside:b},step:c,keyword:C,function:m,format:y,altformat:S,"global-statements":x,number:a,"numeric-constant":o,punctuation:g,string:p}},"proc-args":{pattern:RegExp(/(^proc\s+\w+\s+)(?!\s)(?:[^;"']|<str>)+;/.source.replace(/<str>/g,function(){return r}),"im"),lookbehind:!0,inside:b},"macro-keyword":u,"macro-variable":s,"macro-string-functions":{pattern:/((?:^|\s|=))%(?:BQUOTE|NRBQUOTE|NRQUOTE|NRSTR|QUOTE|STR)\(.*?(?:[^%]\))/i,lookbehind:!0,inside:{function:{pattern:/%(?:BQUOTE|NRBQUOTE|NRQUOTE|NRSTR|QUOTE|STR)/i,alias:"keyword"},"macro-keyword":u,"macro-variable":s,"escaped-char":{pattern:/%['"()<>=¬^~;,#]/},punctuation:g}},"macro-declaration":{pattern:/^%macro[^;]+(?=;)/im,inside:{keyword:/%macro/i}},"macro-end":{pattern:/^%mend[^;]+(?=;)/im,inside:{keyword:/%mend/i}},macro:{pattern:/%_\w+(?=\()/,alias:"keyword"},input:{pattern:/\binput\s[-\w\s/*.$&]+;/i,inside:{input:{alias:"keyword",pattern:/^input/i},comment:d,number:a,"numeric-constant":o}},"options-args":{pattern:/(^options)[-'"|/\\<>*+=:()\w\s]*(?=;)/im,lookbehind:!0,inside:b},"cas-actions":A,comment:d,function:m,format:y,altformat:S,"numeric-constant":o,datetime:{pattern:RegExp(r+"(?:dt?|t)"),alias:"number"},string:p,step:c,keyword:C,"operator-keyword":{pattern:/\b(?:eq|ge|gt|in|le|lt|ne|not)\b/i,alias:"operator"},number:a,operator:/\*\*?|\|\|?|!!?|¦¦?|<[>=]?|>[<=]?|[-+\/=&]|[~¬^]=?/,punctuation:g}})(t)}return iw}var sw,_F;function kSe(){if(_F)return sw;_F=1,sw=e,e.displayName="sass",e.aliases=[];function e(t){(function(n){n.languages.sass=n.languages.extend("css",{comment:{pattern:/^([ \t]*)\/[\/*].*(?:(?:\r?\n|\r)\1[ \t].+)*/m,lookbehind:!0,greedy:!0}}),n.languages.insertBefore("sass","atrule",{"atrule-line":{pattern:/^(?:[ \t]*)[@+=].+/m,greedy:!0,inside:{atrule:/(?:@[\w-]+|[+=])/}}}),delete n.languages.sass.atrule;var r=/\$[-\w]+|#\{\$[-\w]+\}/,a=[/[+*\/%]|[=!]=|<=?|>=?|\b(?:and|not|or)\b/,{pattern:/(\s)-(?=\s)/,lookbehind:!0}];n.languages.insertBefore("sass","property",{"variable-line":{pattern:/^[ \t]*\$.+/m,greedy:!0,inside:{punctuation:/:/,variable:r,operator:a}},"property-line":{pattern:/^[ \t]*(?:[^:\s]+ *:.*|:[^:\s].*)/m,greedy:!0,inside:{property:[/[^:\s]+(?=\s*:)/,{pattern:/(:)[^:\s]+/,lookbehind:!0}],punctuation:/:/,variable:r,operator:a,important:n.languages.sass.important}}}),delete n.languages.sass.property,delete n.languages.sass.important,n.languages.insertBefore("sass","punctuation",{selector:{pattern:/^([ \t]*)\S(?:,[^,\r\n]+|[^,\r\n]*)(?:,[^,\r\n]+)*(?:,(?:\r?\n|\r)\1[ \t]+\S(?:,[^,\r\n]+|[^,\r\n]*)(?:,[^,\r\n]+)*)*/m,lookbehind:!0,greedy:!0}})})(t)}return sw}var lw,CF;function TSe(){if(CF)return lw;CF=1;var e=dA();lw=t,t.displayName="scala",t.aliases=[];function t(n){n.register(e),n.languages.scala=n.languages.extend("java",{"triple-quoted-string":{pattern:/"""[\s\S]*?"""/,greedy:!0,alias:"string"},string:{pattern:/("|')(?:\\.|(?!\1)[^\\\r\n])*\1/,greedy:!0},keyword:/<-|=>|\b(?:abstract|case|catch|class|def|do|else|extends|final|finally|for|forSome|if|implicit|import|lazy|match|new|null|object|override|package|private|protected|return|sealed|self|super|this|throw|trait|try|type|val|var|while|with|yield)\b/,number:/\b0x(?:[\da-f]*\.)?[\da-f]+|(?:\b\d+(?:\.\d*)?|\B\.\d+)(?:e\d+)?[dfl]?/i,builtin:/\b(?:Any|AnyRef|AnyVal|Boolean|Byte|Char|Double|Float|Int|Long|Nothing|Short|String|Unit)\b/,symbol:/'[^\d\s\\]\w*/}),n.languages.insertBefore("scala","triple-quoted-string",{"string-interpolation":{pattern:/\b[a-z]\w*(?:"""(?:[^$]|\$(?:[^{]|\{(?:[^{}]|\{[^{}]*\})*\}))*?"""|"(?:[^$"\r\n]|\$(?:[^{]|\{(?:[^{}]|\{[^{}]*\})*\}))*")/i,greedy:!0,inside:{id:{pattern:/^\w+/,greedy:!0,alias:"function"},escape:{pattern:/\\\$"|\$[$"]/,greedy:!0,alias:"symbol"},interpolation:{pattern:/\$(?:\w+|\{(?:[^{}]|\{[^{}]*\})*\})/,greedy:!0,inside:{punctuation:/^\$\{?|\}$/,expression:{pattern:/[\s\S]+/,inside:n.languages.scala}}},string:/[\s\S]+/}}}),delete n.languages.scala["class-name"],delete n.languages.scala.function}return lw}var uw,NF;function ASe(){if(NF)return uw;NF=1,uw=e,e.displayName="scss",e.aliases=[];function e(t){t.languages.scss=t.languages.extend("css",{comment:{pattern:/(^|[^\\])(?:\/\*[\s\S]*?\*\/|\/\/.*)/,lookbehind:!0},atrule:{pattern:/@[\w-](?:\([^()]+\)|[^()\s]|\s+(?!\s))*?(?=\s+[{;])/,inside:{rule:/@[\w-]+/}},url:/(?:[-a-z]+-)?url(?=\()/i,selector:{pattern:/(?=\S)[^@;{}()]?(?:[^@;{}()\s]|\s+(?!\s)|#\{\$[-\w]+\})+(?=\s*\{(?:\}|\s|[^}][^:{}]*[:{][^}]))/,inside:{parent:{pattern:/&/,alias:"important"},placeholder:/%[-\w]+/,variable:/\$[-\w]+|#\{\$[-\w]+\}/}},property:{pattern:/(?:[-\w]|\$[-\w]|#\{\$[-\w]+\})+(?=\s*:)/,inside:{variable:/\$[-\w]+|#\{\$[-\w]+\}/}}}),t.languages.insertBefore("scss","atrule",{keyword:[/@(?:content|debug|each|else(?: if)?|extend|for|forward|function|if|import|include|mixin|return|use|warn|while)\b/i,{pattern:/( )(?:from|through)(?= )/,lookbehind:!0}]}),t.languages.insertBefore("scss","important",{variable:/\$[-\w]+|#\{\$[-\w]+\}/}),t.languages.insertBefore("scss","function",{"module-modifier":{pattern:/\b(?:as|hide|show|with)\b/i,alias:"keyword"},placeholder:{pattern:/%[-\w]+/,alias:"selector"},statement:{pattern:/\B!(?:default|optional)\b/i,alias:"keyword"},boolean:/\b(?:false|true)\b/,null:{pattern:/\bnull\b/,alias:"keyword"},operator:{pattern:/(\s)(?:[-+*\/%]|[=!]=|<=?|>=?|and|not|or)(?=\s)/,lookbehind:!0}}),t.languages.scss.atrule.inside.rest=t.languages.scss}return uw}var cw,OF;function RSe(){if(OF)return cw;OF=1;var e=GH();cw=t,t.displayName="shellSession",t.aliases=[];function t(n){n.register(e),function(r){var a=[/"(?:\\[\s\S]|\$\([^)]+\)|\$(?!\()|`[^`]+`|[^"\\`$])*"/.source,/'[^']*'/.source,/\$'(?:[^'\\]|\\[\s\S])*'/.source,/<<-?\s*(["']?)(\w+)\1\s[\s\S]*?[\r\n]\2/.source].join("|");r.languages["shell-session"]={command:{pattern:RegExp(/^/.source+"(?:"+(/[^\s@:$#%*!/\\]+@[^\r\n@:$#%*!/\\]+(?::[^\0-\x1F$#%*?"<>:;|]+)?/.source+"|"+/[/~.][^\0-\x1F$#%*?"<>@:;|]*/.source)+")?"+/[$#%](?=\s)/.source+/(?:[^\\\r\n \t'"<$]|[ \t](?:(?!#)|#.*$)|\\(?:[^\r]|\r\n?)|\$(?!')|<(?!<)|<<str>>)+/.source.replace(/<<str>>/g,function(){return a}),"m"),greedy:!0,inside:{info:{pattern:/^[^#$%]+/,alias:"punctuation",inside:{user:/^[^\s@:$#%*!/\\]+@[^\r\n@:$#%*!/\\]+/,punctuation:/:/,path:/[\s\S]+/}},bash:{pattern:/(^[$#%]\s*)\S[\s\S]*/,lookbehind:!0,alias:"language-bash",inside:r.languages.bash},"shell-symbol":{pattern:/^[$#%]/,alias:"important"}}},output:/.(?:.*(?:[\r\n]|.$))*/},r.languages["sh-session"]=r.languages.shellsession=r.languages["shell-session"]}(n)}return cw}var dw,DF;function _Se(){if(DF)return dw;DF=1,dw=e,e.displayName="smali",e.aliases=[];function e(t){t.languages.smali={comment:/#.*/,string:{pattern:/"(?:[^\r\n\\"]|\\.)*"|'(?:[^\r\n\\']|\\(?:.|u[\da-fA-F]{4}))'/,greedy:!0},"class-name":{pattern:/(^|[^L])L(?:(?:\w+|`[^`\r\n]*`)\/)*(?:[\w$]+|`[^`\r\n]*`)(?=\s*;)/,lookbehind:!0,inside:{"class-name":{pattern:/(^L|\/)(?:[\w$]+|`[^`\r\n]*`)$/,lookbehind:!0},namespace:{pattern:/^(L)(?:(?:\w+|`[^`\r\n]*`)\/)+/,lookbehind:!0,inside:{punctuation:/\//}},builtin:/^L/}},builtin:[{pattern:/([();\[])[BCDFIJSVZ]+/,lookbehind:!0},{pattern:/([\w$>]:)[BCDFIJSVZ]/,lookbehind:!0}],keyword:[{pattern:/(\.end\s+)[\w-]+/,lookbehind:!0},{pattern:/(^|[^\w.-])\.(?!\d)[\w-]+/,lookbehind:!0},{pattern:/(^|[^\w.-])(?:abstract|annotation|bridge|constructor|enum|final|interface|private|protected|public|runtime|static|synthetic|system|transient)(?![\w.-])/,lookbehind:!0}],function:{pattern:/(^|[^\w.-])(?:\w+|<[\w$-]+>)(?=\()/,lookbehind:!0},field:{pattern:/[\w$]+(?=:)/,alias:"variable"},register:{pattern:/(^|[^\w.-])[vp]\d(?![\w.-])/,lookbehind:!0,alias:"variable"},boolean:{pattern:/(^|[^\w.-])(?:false|true)(?![\w.-])/,lookbehind:!0},number:{pattern:/(^|[^/\w.-])-?(?:NAN|INFINITY|0x(?:[\dA-F]+(?:\.[\dA-F]*)?|\.[\dA-F]+)(?:p[+-]?[\dA-F]+)?|(?:\d+(?:\.\d*)?|\.\d+)(?:e[+-]?\d+)?)[dflst]?(?![\w.-])/i,lookbehind:!0},label:{pattern:/(:)\w+/,lookbehind:!0,alias:"property"},operator:/->|\.\.|[\[=]/,punctuation:/[{}(),;:]/}}return dw}var fw,IF;function CSe(){if(IF)return fw;IF=1,fw=e,e.displayName="smalltalk",e.aliases=[];function e(t){t.languages.smalltalk={comment:{pattern:/"(?:""|[^"])*"/,greedy:!0},char:{pattern:/\$./,greedy:!0},string:{pattern:/'(?:''|[^'])*'/,greedy:!0},symbol:/#[\da-z]+|#(?:-|([+\/\\*~<>=@%|&?!])\1?)|#(?=\()/i,"block-arguments":{pattern:/(\[\s*):[^\[|]*\|/,lookbehind:!0,inside:{variable:/:[\da-z]+/i,punctuation:/\|/}},"temporary-variables":{pattern:/\|[^|]+\|/,inside:{variable:/[\da-z]+/i,punctuation:/\|/}},keyword:/\b(?:new|nil|self|super)\b/,boolean:/\b(?:false|true)\b/,number:[/\d+r-?[\dA-Z]+(?:\.[\dA-Z]+)?(?:e-?\d+)?/,/\b\d+(?:\.\d+)?(?:e-?\d+)?/],operator:/[<=]=?|:=|~[~=]|\/\/?|\\\\|>[>=]?|[!^+\-*&|,@]/,punctuation:/[.;:?\[\](){}]/}}return fw}var pw,LF;function NSe(){if(LF)return pw;LF=1;var e=Wn();pw=t,t.displayName="smarty",t.aliases=[];function t(n){n.register(e),function(r){r.languages.smarty={comment:{pattern:/^\{\*[\s\S]*?\*\}/,greedy:!0},"embedded-php":{pattern:/^\{php\}[\s\S]*?\{\/php\}/,greedy:!0,inside:{smarty:{pattern:/^\{php\}|\{\/php\}$/,inside:null},php:{pattern:/[\s\S]+/,alias:"language-php",inside:r.languages.php}}},string:[{pattern:/"(?:\\.|[^"\\\r\n])*"/,greedy:!0,inside:{interpolation:{pattern:/\{[^{}]*\}|`[^`]*`/,inside:{"interpolation-punctuation":{pattern:/^[{`]|[`}]$/,alias:"punctuation"},expression:{pattern:/[\s\S]+/,inside:null}}},variable:/\$\w+/}},{pattern:/'(?:\\.|[^'\\\r\n])*'/,greedy:!0}],keyword:{pattern:/(^\{\/?)[a-z_]\w*\b(?!\()/i,lookbehind:!0,greedy:!0},delimiter:{pattern:/^\{\/?|\}$/,greedy:!0,alias:"punctuation"},number:/\b0x[\dA-Fa-f]+|(?:\b\d+(?:\.\d*)?|\B\.\d+)(?:[Ee][-+]?\d+)?/,variable:[/\$(?!\d)\w+/,/#(?!\d)\w+#/,{pattern:/(\.|->|\w\s*=)(?!\d)\w+\b(?!\()/,lookbehind:!0},{pattern:/(\[)(?!\d)\w+(?=\])/,lookbehind:!0}],function:{pattern:/(\|\s*)@?[a-z_]\w*|\b[a-z_]\w*(?=\()/i,lookbehind:!0},"attr-name":/\b[a-z_]\w*(?=\s*=)/i,boolean:/\b(?:false|no|off|on|true|yes)\b/,punctuation:/[\[\](){}.,:`]|->/,operator:[/[+\-*\/%]|==?=?|[!<>]=?|&&|\|\|?/,/\bis\s+(?:not\s+)?(?:div|even|odd)(?:\s+by)?\b/,/\b(?:and|eq|gt?e|gt|lt?e|lt|mod|neq?|not|or)\b/]},r.languages.smarty["embedded-php"].inside.smarty.inside=r.languages.smarty,r.languages.smarty.string[0].inside.interpolation.inside.expression.inside=r.languages.smarty;var a=/"(?:\\.|[^"\\\r\n])*"|'(?:\\.|[^'\\\r\n])*'/,o=RegExp(/\{\*[\s\S]*?\*\}/.source+"|"+/\{php\}[\s\S]*?\{\/php\}/.source+"|"+/\{(?:[^{}"']|<str>|\{(?:[^{}"']|<str>|\{(?:[^{}"']|<str>)*\})*\})*\}/.source.replace(/<str>/g,function(){return a.source}),"g");r.hooks.add("before-tokenize",function(s){var u="{literal}",c="{/literal}",d=!1;r.languages["markup-templating"].buildPlaceholders(s,"smarty",o,function(p){return p===c&&(d=!1),d?!1:(p===u&&(d=!0),!0)})}),r.hooks.add("after-tokenize",function(s){r.languages["markup-templating"].tokenizePlaceholders(s,"smarty")})}(n)}return pw}var gw,MF;function OSe(){if(MF)return gw;MF=1,gw=e,e.displayName="sml",e.aliases=["smlnj"];function e(t){(function(n){var r=/\b(?:abstype|and|andalso|as|case|datatype|do|else|end|eqtype|exception|fn|fun|functor|handle|if|in|include|infix|infixr|let|local|nonfix|of|op|open|orelse|raise|rec|sharing|sig|signature|struct|structure|then|type|val|where|while|with|withtype)\b/i;n.languages.sml={comment:/\(\*(?:[^*(]|\*(?!\))|\((?!\*)|\(\*(?:[^*(]|\*(?!\))|\((?!\*))*\*\))*\*\)/,string:{pattern:/#?"(?:[^"\\]|\\.)*"/,greedy:!0},"class-name":[{pattern:RegExp(/((?:^|[^:]):\s*)<TERMINAL>(?:\s*(?:(?:\*|->)\s*<TERMINAL>|,\s*<TERMINAL>(?:(?=<NOT-LAST>)|(?!<NOT-LAST>)\s+<LONG-ID>)))*/.source.replace(/<NOT-LAST>/g,function(){return/\s*(?:[*,]|->)/.source}).replace(/<TERMINAL>/g,function(){return/(?:'[\w']*|<LONG-ID>|\((?:[^()]|\([^()]*\))*\)|\{(?:[^{}]|\{[^{}]*\})*\})(?:\s+<LONG-ID>)*/.source}).replace(/<LONG-ID>/g,function(){return/(?!<KEYWORD>)[a-z\d_][\w'.]*/.source}).replace(/<KEYWORD>/g,function(){return r.source}),"i"),lookbehind:!0,greedy:!0,inside:null},{pattern:/((?:^|[^\w'])(?:datatype|exception|functor|signature|structure|type)\s+)[a-z_][\w'.]*/i,lookbehind:!0}],function:{pattern:/((?:^|[^\w'])fun\s+)[a-z_][\w'.]*/i,lookbehind:!0},keyword:r,variable:{pattern:/(^|[^\w'])'[\w']*/,lookbehind:!0},number:/~?\b(?:\d+(?:\.\d+)?(?:e~?\d+)?|0x[\da-f]+)\b/i,word:{pattern:/\b0w(?:\d+|x[\da-f]+)\b/i,alias:"constant"},boolean:/\b(?:false|true)\b/i,operator:/\.\.\.|:[>=:]|=>?|->|[<>]=?|[!+\-*/^#|@~]/,punctuation:/[(){}\[\].:,;]/},n.languages.sml["class-name"][0].inside=n.languages.sml,n.languages.smlnj=n.languages.sml})(t)}return gw}var hw,FF;function DSe(){if(FF)return hw;FF=1,hw=e,e.displayName="solidity",e.aliases=["sol"];function e(t){t.languages.solidity=t.languages.extend("clike",{"class-name":{pattern:/(\b(?:contract|enum|interface|library|new|struct|using)\s+)(?!\d)[\w$]+/,lookbehind:!0},keyword:/\b(?:_|anonymous|as|assembly|assert|break|calldata|case|constant|constructor|continue|contract|default|delete|do|else|emit|enum|event|external|for|from|function|if|import|indexed|inherited|interface|internal|is|let|library|mapping|memory|modifier|new|payable|pragma|private|public|pure|require|returns?|revert|selfdestruct|solidity|storage|struct|suicide|switch|this|throw|using|var|view|while)\b/,operator:/=>|->|:=|=:|\*\*|\+\+|--|\|\||&&|<<=?|>>=?|[-+*/%^&|<>!=]=?|[~?]/}),t.languages.insertBefore("solidity","keyword",{builtin:/\b(?:address|bool|byte|u?int(?:8|16|24|32|40|48|56|64|72|80|88|96|104|112|120|128|136|144|152|160|168|176|184|192|200|208|216|224|232|240|248|256)?|string|bytes(?:[1-9]|[12]\d|3[0-2])?)\b/}),t.languages.insertBefore("solidity","number",{version:{pattern:/([<>]=?|\^)\d+\.\d+\.\d+\b/,lookbehind:!0,alias:"number"}}),t.languages.sol=t.languages.solidity}return hw}var mw,PF;function ISe(){if(PF)return mw;PF=1,mw=e,e.displayName="solutionFile",e.aliases=[];function e(t){(function(n){var r={pattern:/\{[\da-f]{8}-[\da-f]{4}-[\da-f]{4}-[\da-f]{4}-[\da-f]{12}\}/i,alias:"constant",inside:{punctuation:/[{}]/}};n.languages["solution-file"]={comment:{pattern:/#.*/,greedy:!0},string:{pattern:/"[^"\r\n]*"|'[^'\r\n]*'/,greedy:!0,inside:{guid:r}},object:{pattern:/^([ \t]*)(?:([A-Z]\w*)\b(?=.*(?:\r\n?|\n)(?:\1[ \t].*(?:\r\n?|\n))*\1End\2(?=[ \t]*$))|End[A-Z]\w*(?=[ \t]*$))/m,lookbehind:!0,greedy:!0,alias:"keyword"},property:{pattern:/^([ \t]*)(?!\s)[^\r\n"#=()]*[^\s"#=()](?=\s*=)/m,lookbehind:!0,inside:{guid:r}},guid:r,number:/\b\d+(?:\.\d+)*\b/,boolean:/\b(?:FALSE|TRUE)\b/,operator:/=/,punctuation:/[(),]/},n.languages.sln=n.languages["solution-file"]})(t)}return mw}var bw,zF;function LSe(){if(zF)return bw;zF=1;var e=Wn();bw=t,t.displayName="soy",t.aliases=[];function t(n){n.register(e),function(r){var a=/(["'])(?:\\(?:\r\n|[\s\S])|(?!\1)[^\\\r\n])*\1/,o=/\b\d+(?:\.\d+)?(?:[eE][+-]?\d+)?\b|\b0x[\dA-F]+\b/;r.languages.soy={comment:[/\/\*[\s\S]*?\*\//,{pattern:/(\s)\/\/.*/,lookbehind:!0,greedy:!0}],"command-arg":{pattern:/(\{+\/?\s*(?:alias|call|delcall|delpackage|deltemplate|namespace|template)\s+)\.?[\w.]+/,lookbehind:!0,alias:"string",inside:{punctuation:/\./}},parameter:{pattern:/(\{+\/?\s*@?param\??\s+)\.?[\w.]+/,lookbehind:!0,alias:"variable"},keyword:[{pattern:/(\{+\/?[^\S\r\n]*)(?:\\[nrt]|alias|call|case|css|default|delcall|delpackage|deltemplate|else(?:if)?|fallbackmsg|for(?:each)?|if(?:empty)?|lb|let|literal|msg|namespace|nil|@?param\??|rb|sp|switch|template|xid)/,lookbehind:!0},/\b(?:any|as|attributes|bool|css|float|html|in|int|js|list|map|null|number|string|uri)\b/],delimiter:{pattern:/^\{+\/?|\/?\}+$/,alias:"punctuation"},property:/\w+(?==)/,variable:{pattern:/\$[^\W\d]\w*(?:\??(?:\.\w+|\[[^\]]+\]))*/,inside:{string:{pattern:a,greedy:!0},number:o,punctuation:/[\[\].?]/}},string:{pattern:a,greedy:!0},function:[/\w+(?=\()/,{pattern:/(\|[^\S\r\n]*)\w+/,lookbehind:!0}],boolean:/\b(?:false|true)\b/,number:o,operator:/\?:?|<=?|>=?|==?|!=|[+*/%-]|\b(?:and|not|or)\b/,punctuation:/[{}()\[\]|.,:]/},r.hooks.add("before-tokenize",function(s){var u=/\{\{.+?\}\}|\{.+?\}|\s\/\/.*|\/\*[\s\S]*?\*\//g,c="{literal}",d="{/literal}",p=!1;r.languages["markup-templating"].buildPlaceholders(s,"soy",u,function(g){return g===d&&(p=!1),p?!1:(g===c&&(p=!0),!0)})}),r.hooks.add("after-tokenize",function(s){r.languages["markup-templating"].tokenizePlaceholders(s,"soy")})}(n)}return bw}var yw,BF;function VH(){if(BF)return yw;BF=1,yw=e,e.displayName="turtle",e.aliases=[];function e(t){t.languages.turtle={comment:{pattern:/#.*/,greedy:!0},"multiline-string":{pattern:/"""(?:(?:""?)?(?:[^"\\]|\\.))*"""|'''(?:(?:''?)?(?:[^'\\]|\\.))*'''/,greedy:!0,alias:"string",inside:{comment:/#.*/}},string:{pattern:/"(?:[^\\"\r\n]|\\.)*"|'(?:[^\\'\r\n]|\\.)*'/,greedy:!0},url:{pattern:/<(?:[^\x00-\x20<>"{}|^`\\]|\\(?:u[\da-fA-F]{4}|U[\da-fA-F]{8}))*>/,greedy:!0,inside:{punctuation:/[<>]/}},function:{pattern:/(?:(?![-.\d\xB7])[-.\w\xB7\xC0-\uFFFD]+)?:(?:(?![-.])(?:[-.:\w\xC0-\uFFFD]|%[\da-f]{2}|\\.)+)?/i,inside:{"local-name":{pattern:/([^:]*:)[\s\S]+/,lookbehind:!0},prefix:{pattern:/[\s\S]+/,inside:{punctuation:/:/}}}},number:/[+-]?\b\d+(?:\.\d*)?(?:e[+-]?\d+)?/i,punctuation:/[{}.,;()[\]]|\^\^/,boolean:/\b(?:false|true)\b/,keyword:[/(?:\ba|@prefix|@base)\b|=/,/\b(?:base|graph|prefix)\b/i],tag:{pattern:/@[a-z]+(?:-[a-z\d]+)*/i,inside:{punctuation:/@/}}},t.languages.trig=t.languages.turtle}return yw}var vw,UF;function MSe(){if(UF)return vw;UF=1;var e=VH();vw=t,t.displayName="sparql",t.aliases=["rq"];function t(n){n.register(e),n.languages.sparql=n.languages.extend("turtle",{boolean:/\b(?:false|true)\b/i,variable:{pattern:/[?$]\w+/,greedy:!0}}),n.languages.insertBefore("sparql","punctuation",{keyword:[/\b(?:A|ADD|ALL|AS|ASC|ASK|BNODE|BY|CLEAR|CONSTRUCT|COPY|CREATE|DATA|DEFAULT|DELETE|DESC|DESCRIBE|DISTINCT|DROP|EXISTS|FILTER|FROM|GROUP|HAVING|INSERT|INTO|LIMIT|LOAD|MINUS|MOVE|NAMED|NOT|NOW|OFFSET|OPTIONAL|ORDER|RAND|REDUCED|SELECT|SEPARATOR|SERVICE|SILENT|STRUUID|UNION|USING|UUID|VALUES|WHERE)\b/i,/\b(?:ABS|AVG|BIND|BOUND|CEIL|COALESCE|CONCAT|CONTAINS|COUNT|DATATYPE|DAY|ENCODE_FOR_URI|FLOOR|GROUP_CONCAT|HOURS|IF|IRI|isBLANK|isIRI|isLITERAL|isNUMERIC|isURI|LANG|LANGMATCHES|LCASE|MAX|MD5|MIN|MINUTES|MONTH|REGEX|REPLACE|ROUND|sameTerm|SAMPLE|SECONDS|SHA1|SHA256|SHA384|SHA512|STR|STRAFTER|STRBEFORE|STRDT|STRENDS|STRLANG|STRLEN|STRSTARTS|SUBSTR|SUM|TIMEZONE|TZ|UCASE|URI|YEAR)\b(?=\s*\()/i,/\b(?:BASE|GRAPH|PREFIX)\b/i]}),n.languages.rq=n.languages.sparql}return vw}var Sw,jF;function FSe(){if(jF)return Sw;jF=1,Sw=e,e.displayName="splunkSpl",e.aliases=[];function e(t){t.languages["splunk-spl"]={comment:/`comment\("(?:\\.|[^\\"])*"\)`/,string:{pattern:/"(?:\\.|[^\\"])*"/,greedy:!0},keyword:/\b(?:abstract|accum|addcoltotals|addinfo|addtotals|analyzefields|anomalies|anomalousvalue|anomalydetection|append|appendcols|appendcsv|appendlookup|appendpipe|arules|associate|audit|autoregress|bin|bucket|bucketdir|chart|cluster|cofilter|collect|concurrency|contingency|convert|correlate|datamodel|dbinspect|dedup|delete|delta|diff|erex|eval|eventcount|eventstats|extract|fieldformat|fields|fieldsummary|filldown|fillnull|findtypes|folderize|foreach|format|from|gauge|gentimes|geom|geomfilter|geostats|head|highlight|history|iconify|input|inputcsv|inputlookup|iplocation|join|kmeans|kv|kvform|loadjob|localize|localop|lookup|makecontinuous|makemv|makeresults|map|mcollect|metadata|metasearch|meventcollect|mstats|multikv|multisearch|mvcombine|mvexpand|nomv|outlier|outputcsv|outputlookup|outputtext|overlap|pivot|predict|rangemap|rare|regex|relevancy|reltime|rename|replace|rest|return|reverse|rex|rtorder|run|savedsearch|script|scrub|search|searchtxn|selfjoin|sendemail|set|setfields|sichart|sirare|sistats|sitimechart|sitop|sort|spath|stats|strcat|streamstats|table|tags|tail|timechart|timewrap|top|transaction|transpose|trendline|tscollect|tstats|typeahead|typelearner|typer|union|uniq|untable|where|x11|xmlkv|xmlunescape|xpath|xyseries)\b/i,"operator-word":{pattern:/\b(?:and|as|by|not|or|xor)\b/i,alias:"operator"},function:/\b\w+(?=\s*\()/,property:/\b\w+(?=\s*=(?!=))/,date:{pattern:/\b\d{1,2}\/\d{1,2}\/\d{1,4}(?:(?::\d{1,2}){3})?\b/,alias:"number"},number:/\b\d+(?:\.\d+)?\b/,boolean:/\b(?:f|false|t|true)\b/i,operator:/[<>=]=?|[-+*/%|]/,punctuation:/[()[\],]/}}return Sw}var Ew,GF;function PSe(){if(GF)return Ew;GF=1,Ew=e,e.displayName="sqf",e.aliases=[];function e(t){t.languages.sqf=t.languages.extend("clike",{string:{pattern:/"(?:(?:"")?[^"])*"(?!")|'(?:[^'])*'/,greedy:!0},keyword:/\b(?:breakOut|breakTo|call|case|catch|default|do|echo|else|execFSM|execVM|exitWith|for|forEach|forEachMember|forEachMemberAgent|forEachMemberTeam|from|goto|if|nil|preprocessFile|preprocessFileLineNumbers|private|scopeName|spawn|step|switch|then|throw|to|try|while|with)\b/i,boolean:/\b(?:false|true)\b/i,function:/\b(?:abs|accTime|acos|action|actionIDs|actionKeys|actionKeysImages|actionKeysNames|actionKeysNamesArray|actionName|actionParams|activateAddons|activatedAddons|activateKey|add3DENConnection|add3DENEventHandler|add3DENLayer|addAction|addBackpack|addBackpackCargo|addBackpackCargoGlobal|addBackpackGlobal|addCamShake|addCuratorAddons|addCuratorCameraArea|addCuratorEditableObjects|addCuratorEditingArea|addCuratorPoints|addEditorObject|addEventHandler|addForce|addForceGeneratorRTD|addGoggles|addGroupIcon|addHandgunItem|addHeadgear|addItem|addItemCargo|addItemCargoGlobal|addItemPool|addItemToBackpack|addItemToUniform|addItemToVest|addLiveStats|addMagazine|addMagazineAmmoCargo|addMagazineCargo|addMagazineCargoGlobal|addMagazineGlobal|addMagazinePool|addMagazines|addMagazineTurret|addMenu|addMenuItem|addMissionEventHandler|addMPEventHandler|addMusicEventHandler|addOwnedMine|addPlayerScores|addPrimaryWeaponItem|addPublicVariableEventHandler|addRating|addResources|addScore|addScoreSide|addSecondaryWeaponItem|addSwitchableUnit|addTeamMember|addToRemainsCollector|addTorque|addUniform|addVehicle|addVest|addWaypoint|addWeapon|addWeaponCargo|addWeaponCargoGlobal|addWeaponGlobal|addWeaponItem|addWeaponPool|addWeaponTurret|admin|agent|agents|AGLToASL|aimedAtTarget|aimPos|airDensityCurveRTD|airDensityRTD|airplaneThrottle|airportSide|AISFinishHeal|alive|all3DENEntities|allAirports|allControls|allCurators|allCutLayers|allDead|allDeadMen|allDisplays|allGroups|allMapMarkers|allMines|allMissionObjects|allow3DMode|allowCrewInImmobile|allowCuratorLogicIgnoreAreas|allowDamage|allowDammage|allowFileOperations|allowFleeing|allowGetIn|allowSprint|allPlayers|allSimpleObjects|allSites|allTurrets|allUnits|allUnitsUAV|allVariables|ammo|ammoOnPylon|animate|animateBay|animateDoor|animatePylon|animateSource|animationNames|animationPhase|animationSourcePhase|animationState|append|apply|armoryPoints|arrayIntersect|asin|ASLToAGL|ASLToATL|assert|assignAsCargo|assignAsCargoIndex|assignAsCommander|assignAsDriver|assignAsGunner|assignAsTurret|assignCurator|assignedCargo|assignedCommander|assignedDriver|assignedGunner|assignedItems|assignedTarget|assignedTeam|assignedVehicle|assignedVehicleRole|assignItem|assignTeam|assignToAirport|atan|atan2|atg|ATLToASL|attachedObject|attachedObjects|attachedTo|attachObject|attachTo|attackEnabled|backpack|backpackCargo|backpackContainer|backpackItems|backpackMagazines|backpackSpaceFor|behaviour|benchmark|binocular|blufor|boundingBox|boundingBoxReal|boundingCenter|briefingName|buildingExit|buildingPos|buldozer_EnableRoadDiag|buldozer_IsEnabledRoadDiag|buldozer_LoadNewRoads|buldozer_reloadOperMap|buttonAction|buttonSetAction|cadetMode|callExtension|camCommand|camCommit|camCommitPrepared|camCommitted|camConstuctionSetParams|camCreate|camDestroy|cameraEffect|cameraEffectEnableHUD|cameraInterest|cameraOn|cameraView|campaignConfigFile|camPreload|camPreloaded|camPrepareBank|camPrepareDir|camPrepareDive|camPrepareFocus|camPrepareFov|camPrepareFovRange|camPreparePos|camPrepareRelPos|camPrepareTarget|camSetBank|camSetDir|camSetDive|camSetFocus|camSetFov|camSetFovRange|camSetPos|camSetRelPos|camSetTarget|camTarget|camUseNVG|canAdd|canAddItemToBackpack|canAddItemToUniform|canAddItemToVest|cancelSimpleTaskDestination|canFire|canMove|canSlingLoad|canStand|canSuspend|canTriggerDynamicSimulation|canUnloadInCombat|canVehicleCargo|captive|captiveNum|cbChecked|cbSetChecked|ceil|channelEnabled|cheatsEnabled|checkAIFeature|checkVisibility|civilian|className|clear3DENAttribute|clear3DENInventory|clearAllItemsFromBackpack|clearBackpackCargo|clearBackpackCargoGlobal|clearForcesRTD|clearGroupIcons|clearItemCargo|clearItemCargoGlobal|clearItemPool|clearMagazineCargo|clearMagazineCargoGlobal|clearMagazinePool|clearOverlay|clearRadio|clearVehicleInit|clearWeaponCargo|clearWeaponCargoGlobal|clearWeaponPool|clientOwner|closeDialog|closeDisplay|closeOverlay|collapseObjectTree|collect3DENHistory|collectiveRTD|combatMode|commandArtilleryFire|commandChat|commander|commandFire|commandFollow|commandFSM|commandGetOut|commandingMenu|commandMove|commandRadio|commandStop|commandSuppressiveFire|commandTarget|commandWatch|comment|commitOverlay|compile|compileFinal|completedFSM|composeText|configClasses|configFile|configHierarchy|configName|configNull|configProperties|configSourceAddonList|configSourceMod|configSourceModList|confirmSensorTarget|connectTerminalToUAV|controlNull|controlsGroupCtrl|copyFromClipboard|copyToClipboard|copyWaypoints|cos|count|countEnemy|countFriendly|countSide|countType|countUnknown|create3DENComposition|create3DENEntity|createAgent|createCenter|createDialog|createDiaryLink|createDiaryRecord|createDiarySubject|createDisplay|createGearDialog|createGroup|createGuardedPoint|createLocation|createMarker|createMarkerLocal|createMenu|createMine|createMissionDisplay|createMPCampaignDisplay|createSimpleObject|createSimpleTask|createSite|createSoundSource|createTask|createTeam|createTrigger|createUnit|createVehicle|createVehicleCrew|createVehicleLocal|crew|ctAddHeader|ctAddRow|ctClear|ctCurSel|ctData|ctFindHeaderRows|ctFindRowHeader|ctHeaderControls|ctHeaderCount|ctRemoveHeaders|ctRemoveRows|ctrlActivate|ctrlAddEventHandler|ctrlAngle|ctrlAutoScrollDelay|ctrlAutoScrollRewind|ctrlAutoScrollSpeed|ctrlChecked|ctrlClassName|ctrlCommit|ctrlCommitted|ctrlCreate|ctrlDelete|ctrlEnable|ctrlEnabled|ctrlFade|ctrlHTMLLoaded|ctrlIDC|ctrlIDD|ctrlMapAnimAdd|ctrlMapAnimClear|ctrlMapAnimCommit|ctrlMapAnimDone|ctrlMapCursor|ctrlMapMouseOver|ctrlMapScale|ctrlMapScreenToWorld|ctrlMapWorldToScreen|ctrlModel|ctrlModelDirAndUp|ctrlModelScale|ctrlParent|ctrlParentControlsGroup|ctrlPosition|ctrlRemoveAllEventHandlers|ctrlRemoveEventHandler|ctrlScale|ctrlSetActiveColor|ctrlSetAngle|ctrlSetAutoScrollDelay|ctrlSetAutoScrollRewind|ctrlSetAutoScrollSpeed|ctrlSetBackgroundColor|ctrlSetChecked|ctrlSetDisabledColor|ctrlSetEventHandler|ctrlSetFade|ctrlSetFocus|ctrlSetFont|ctrlSetFontH1|ctrlSetFontH1B|ctrlSetFontH2|ctrlSetFontH2B|ctrlSetFontH3|ctrlSetFontH3B|ctrlSetFontH4|ctrlSetFontH4B|ctrlSetFontH5|ctrlSetFontH5B|ctrlSetFontH6|ctrlSetFontH6B|ctrlSetFontHeight|ctrlSetFontHeightH1|ctrlSetFontHeightH2|ctrlSetFontHeightH3|ctrlSetFontHeightH4|ctrlSetFontHeightH5|ctrlSetFontHeightH6|ctrlSetFontHeightSecondary|ctrlSetFontP|ctrlSetFontPB|ctrlSetFontSecondary|ctrlSetForegroundColor|ctrlSetModel|ctrlSetModelDirAndUp|ctrlSetModelScale|ctrlSetPixelPrecision|ctrlSetPosition|ctrlSetScale|ctrlSetStructuredText|ctrlSetText|ctrlSetTextColor|ctrlSetTextColorSecondary|ctrlSetTextSecondary|ctrlSetTooltip|ctrlSetTooltipColorBox|ctrlSetTooltipColorShade|ctrlSetTooltipColorText|ctrlShow|ctrlShown|ctrlText|ctrlTextHeight|ctrlTextSecondary|ctrlTextWidth|ctrlType|ctrlVisible|ctRowControls|ctRowCount|ctSetCurSel|ctSetData|ctSetHeaderTemplate|ctSetRowTemplate|ctSetValue|ctValue|curatorAddons|curatorCamera|curatorCameraArea|curatorCameraAreaCeiling|curatorCoef|curatorEditableObjects|curatorEditingArea|curatorEditingAreaType|curatorMouseOver|curatorPoints|curatorRegisteredObjects|curatorSelected|curatorWaypointCost|current3DENOperation|currentChannel|currentCommand|currentMagazine|currentMagazineDetail|currentMagazineDetailTurret|currentMagazineTurret|currentMuzzle|currentNamespace|currentTask|currentTasks|currentThrowable|currentVisionMode|currentWaypoint|currentWeapon|currentWeaponMode|currentWeaponTurret|currentZeroing|cursorObject|cursorTarget|customChat|customRadio|cutFadeOut|cutObj|cutRsc|cutText|damage|date|dateToNumber|daytime|deActivateKey|debriefingText|debugFSM|debugLog|deg|delete3DENEntities|deleteAt|deleteCenter|deleteCollection|deleteEditorObject|deleteGroup|deleteGroupWhenEmpty|deleteIdentity|deleteLocation|deleteMarker|deleteMarkerLocal|deleteRange|deleteResources|deleteSite|deleteStatus|deleteTeam|deleteVehicle|deleteVehicleCrew|deleteWaypoint|detach|detectedMines|diag_activeMissionFSMs|diag_activeScripts|diag_activeSQFScripts|diag_activeSQSScripts|diag_captureFrame|diag_captureFrameToFile|diag_captureSlowFrame|diag_codePerformance|diag_drawMode|diag_dynamicSimulationEnd|diag_enable|diag_enabled|diag_fps|diag_fpsMin|diag_frameNo|diag_lightNewLoad|diag_list|diag_log|diag_logSlowFrame|diag_mergeConfigFile|diag_recordTurretLimits|diag_setLightNew|diag_tickTime|diag_toggle|dialog|diarySubjectExists|didJIP|didJIPOwner|difficulty|difficultyEnabled|difficultyEnabledRTD|difficultyOption|direction|directSay|disableAI|disableCollisionWith|disableConversation|disableDebriefingStats|disableMapIndicators|disableNVGEquipment|disableRemoteSensors|disableSerialization|disableTIEquipment|disableUAVConnectability|disableUserInput|displayAddEventHandler|displayCtrl|displayNull|displayParent|displayRemoveAllEventHandlers|displayRemoveEventHandler|displaySetEventHandler|dissolveTeam|distance|distance2D|distanceSqr|distributionRegion|do3DENAction|doArtilleryFire|doFire|doFollow|doFSM|doGetOut|doMove|doorPhase|doStop|doSuppressiveFire|doTarget|doWatch|drawArrow|drawEllipse|drawIcon|drawIcon3D|drawLine|drawLine3D|drawLink|drawLocation|drawPolygon|drawRectangle|drawTriangle|driver|drop|dynamicSimulationDistance|dynamicSimulationDistanceCoef|dynamicSimulationEnabled|dynamicSimulationSystemEnabled|east|edit3DENMissionAttributes|editObject|editorSetEventHandler|effectiveCommander|emptyPositions|enableAI|enableAIFeature|enableAimPrecision|enableAttack|enableAudioFeature|enableAutoStartUpRTD|enableAutoTrimRTD|enableCamShake|enableCaustics|enableChannel|enableCollisionWith|enableCopilot|enableDebriefingStats|enableDiagLegend|enableDynamicSimulation|enableDynamicSimulationSystem|enableEndDialog|enableEngineArtillery|enableEnvironment|enableFatigue|enableGunLights|enableInfoPanelComponent|enableIRLasers|enableMimics|enablePersonTurret|enableRadio|enableReload|enableRopeAttach|enableSatNormalOnDetail|enableSaving|enableSentences|enableSimulation|enableSimulationGlobal|enableStamina|enableStressDamage|enableTeamSwitch|enableTraffic|enableUAVConnectability|enableUAVWaypoints|enableVehicleCargo|enableVehicleSensor|enableWeaponDisassembly|endl|endLoadingScreen|endMission|engineOn|enginesIsOnRTD|enginesPowerRTD|enginesRpmRTD|enginesTorqueRTD|entities|environmentEnabled|estimatedEndServerTime|estimatedTimeLeft|evalObjectArgument|everyBackpack|everyContainer|exec|execEditorScript|exp|expectedDestination|exportJIPMessages|eyeDirection|eyePos|face|faction|fadeMusic|fadeRadio|fadeSound|fadeSpeech|failMission|fillWeaponsFromPool|find|findCover|findDisplay|findEditorObject|findEmptyPosition|findEmptyPositionReady|findIf|findNearestEnemy|finishMissionInit|finite|fire|fireAtTarget|firstBackpack|flag|flagAnimationPhase|flagOwner|flagSide|flagTexture|fleeing|floor|flyInHeight|flyInHeightASL|fog|fogForecast|fogParams|forceAddUniform|forceAtPositionRTD|forcedMap|forceEnd|forceFlagTexture|forceFollowRoad|forceGeneratorRTD|forceMap|forceRespawn|forceSpeed|forceWalk|forceWeaponFire|forceWeatherChange|forgetTarget|format|formation|formationDirection|formationLeader|formationMembers|formationPosition|formationTask|formatText|formLeader|freeLook|fromEditor|fuel|fullCrew|gearIDCAmmoCount|gearSlotAmmoCount|gearSlotData|get3DENActionState|get3DENAttribute|get3DENCamera|get3DENConnections|get3DENEntity|get3DENEntityID|get3DENGrid|get3DENIconsVisible|get3DENLayerEntities|get3DENLinesVisible|get3DENMissionAttribute|get3DENMouseOver|get3DENSelected|getAimingCoef|getAllEnvSoundControllers|getAllHitPointsDamage|getAllOwnedMines|getAllSoundControllers|getAmmoCargo|getAnimAimPrecision|getAnimSpeedCoef|getArray|getArtilleryAmmo|getArtilleryComputerSettings|getArtilleryETA|getAssignedCuratorLogic|getAssignedCuratorUnit|getBackpackCargo|getBleedingRemaining|getBurningValue|getCameraViewDirection|getCargoIndex|getCenterOfMass|getClientState|getClientStateNumber|getCompatiblePylonMagazines|getConnectedUAV|getContainerMaxLoad|getCursorObjectParams|getCustomAimCoef|getDammage|getDescription|getDir|getDirVisual|getDLCAssetsUsage|getDLCAssetsUsageByName|getDLCs|getDLCUsageTime|getEditorCamera|getEditorMode|getEditorObjectScope|getElevationOffset|getEngineTargetRpmRTD|getEnvSoundController|getFatigue|getFieldManualStartPage|getForcedFlagTexture|getFriend|getFSMVariable|getFuelCargo|getGroupIcon|getGroupIconParams|getGroupIcons|getHideFrom|getHit|getHitIndex|getHitPointDamage|getItemCargo|getMagazineCargo|getMarkerColor|getMarkerPos|getMarkerSize|getMarkerType|getMass|getMissionConfig|getMissionConfigValue|getMissionDLCs|getMissionLayerEntities|getMissionLayers|getModelInfo|getMousePosition|getMusicPlayedTime|getNumber|getObjectArgument|getObjectChildren|getObjectDLC|getObjectMaterials|getObjectProxy|getObjectTextures|getObjectType|getObjectViewDistance|getOxygenRemaining|getPersonUsedDLCs|getPilotCameraDirection|getPilotCameraPosition|getPilotCameraRotation|getPilotCameraTarget|getPlateNumber|getPlayerChannel|getPlayerScores|getPlayerUID|getPlayerUIDOld|getPos|getPosASL|getPosASLVisual|getPosASLW|getPosATL|getPosATLVisual|getPosVisual|getPosWorld|getPylonMagazines|getRelDir|getRelPos|getRemoteSensorsDisabled|getRepairCargo|getResolution|getRotorBrakeRTD|getShadowDistance|getShotParents|getSlingLoad|getSoundController|getSoundControllerResult|getSpeed|getStamina|getStatValue|getSuppression|getTerrainGrid|getTerrainHeightASL|getText|getTotalDLCUsageTime|getTrimOffsetRTD|getUnitLoadout|getUnitTrait|getUserMFDText|getUserMFDValue|getVariable|getVehicleCargo|getWeaponCargo|getWeaponSway|getWingsOrientationRTD|getWingsPositionRTD|getWPPos|glanceAt|globalChat|globalRadio|goggles|group|groupChat|groupFromNetId|groupIconSelectable|groupIconsVisible|groupId|groupOwner|groupRadio|groupSelectedUnits|groupSelectUnit|grpNull|gunner|gusts|halt|handgunItems|handgunMagazine|handgunWeapon|handsHit|hasInterface|hasPilotCamera|hasWeapon|hcAllGroups|hcGroupParams|hcLeader|hcRemoveAllGroups|hcRemoveGroup|hcSelected|hcSelectGroup|hcSetGroup|hcShowBar|hcShownBar|headgear|hideBody|hideObject|hideObjectGlobal|hideSelection|hint|hintC|hintCadet|hintSilent|hmd|hostMission|htmlLoad|HUDMovementLevels|humidity|image|importAllGroups|importance|in|inArea|inAreaArray|incapacitatedState|independent|inflame|inflamed|infoPanel|infoPanelComponentEnabled|infoPanelComponents|infoPanels|inGameUISetEventHandler|inheritsFrom|initAmbientLife|inPolygon|inputAction|inRangeOfArtillery|insertEditorObject|intersect|is3DEN|is3DENMultiplayer|isAbleToBreathe|isAgent|isAimPrecisionEnabled|isArray|isAutoHoverOn|isAutonomous|isAutoStartUpEnabledRTD|isAutotest|isAutoTrimOnRTD|isBleeding|isBurning|isClass|isCollisionLightOn|isCopilotEnabled|isDamageAllowed|isDedicated|isDLCAvailable|isEngineOn|isEqualTo|isEqualType|isEqualTypeAll|isEqualTypeAny|isEqualTypeArray|isEqualTypeParams|isFilePatchingEnabled|isFlashlightOn|isFlatEmpty|isForcedWalk|isFormationLeader|isGroupDeletedWhenEmpty|isHidden|isInRemainsCollector|isInstructorFigureEnabled|isIRLaserOn|isKeyActive|isKindOf|isLaserOn|isLightOn|isLocalized|isManualFire|isMarkedForCollection|isMultiplayer|isMultiplayerSolo|isNil|isNull|isNumber|isObjectHidden|isObjectRTD|isOnRoad|isPipEnabled|isPlayer|isRealTime|isRemoteExecuted|isRemoteExecutedJIP|isServer|isShowing3DIcons|isSimpleObject|isSprintAllowed|isStaminaEnabled|isSteamMission|isStreamFriendlyUIEnabled|isStressDamageEnabled|isText|isTouchingGround|isTurnedOut|isTutHintsEnabled|isUAVConnectable|isUAVConnected|isUIContext|isUniformAllowed|isVehicleCargo|isVehicleRadarOn|isVehicleSensorEnabled|isWalking|isWeaponDeployed|isWeaponRested|itemCargo|items|itemsWithMagazines|join|joinAs|joinAsSilent|joinSilent|joinString|kbAddDatabase|kbAddDatabaseTargets|kbAddTopic|kbHasTopic|kbReact|kbRemoveTopic|kbTell|kbWasSaid|keyImage|keyName|knowsAbout|land|landAt|landResult|language|laserTarget|lbAdd|lbClear|lbColor|lbColorRight|lbCurSel|lbData|lbDelete|lbIsSelected|lbPicture|lbPictureRight|lbSelection|lbSetColor|lbSetColorRight|lbSetCurSel|lbSetData|lbSetPicture|lbSetPictureColor|lbSetPictureColorDisabled|lbSetPictureColorSelected|lbSetPictureRight|lbSetPictureRightColor|lbSetPictureRightColorDisabled|lbSetPictureRightColorSelected|lbSetSelectColor|lbSetSelectColorRight|lbSetSelected|lbSetText|lbSetTextRight|lbSetTooltip|lbSetValue|lbSize|lbSort|lbSortByValue|lbText|lbTextRight|lbValue|leader|leaderboardDeInit|leaderboardGetRows|leaderboardInit|leaderboardRequestRowsFriends|leaderboardRequestRowsGlobal|leaderboardRequestRowsGlobalAroundUser|leaderboardsRequestUploadScore|leaderboardsRequestUploadScoreKeepBest|leaderboardState|leaveVehicle|libraryCredits|libraryDisclaimers|lifeState|lightAttachObject|lightDetachObject|lightIsOn|lightnings|limitSpeed|linearConversion|lineBreak|lineIntersects|lineIntersectsObjs|lineIntersectsSurfaces|lineIntersectsWith|linkItem|list|listObjects|listRemoteTargets|listVehicleSensors|ln|lnbAddArray|lnbAddColumn|lnbAddRow|lnbClear|lnbColor|lnbColorRight|lnbCurSelRow|lnbData|lnbDeleteColumn|lnbDeleteRow|lnbGetColumnsPosition|lnbPicture|lnbPictureRight|lnbSetColor|lnbSetColorRight|lnbSetColumnsPos|lnbSetCurSelRow|lnbSetData|lnbSetPicture|lnbSetPictureColor|lnbSetPictureColorRight|lnbSetPictureColorSelected|lnbSetPictureColorSelectedRight|lnbSetPictureRight|lnbSetText|lnbSetTextRight|lnbSetValue|lnbSize|lnbSort|lnbSortByValue|lnbText|lnbTextRight|lnbValue|load|loadAbs|loadBackpack|loadFile|loadGame|loadIdentity|loadMagazine|loadOverlay|loadStatus|loadUniform|loadVest|local|localize|locationNull|locationPosition|lock|lockCameraTo|lockCargo|lockDriver|locked|lockedCargo|lockedDriver|lockedTurret|lockIdentity|lockTurret|lockWP|log|logEntities|logNetwork|logNetworkTerminate|lookAt|lookAtPos|magazineCargo|magazines|magazinesAllTurrets|magazinesAmmo|magazinesAmmoCargo|magazinesAmmoFull|magazinesDetail|magazinesDetailBackpack|magazinesDetailUniform|magazinesDetailVest|magazinesTurret|magazineTurretAmmo|mapAnimAdd|mapAnimClear|mapAnimCommit|mapAnimDone|mapCenterOnCamera|mapGridPosition|markAsFinishedOnSteam|markerAlpha|markerBrush|markerColor|markerDir|markerPos|markerShape|markerSize|markerText|markerType|max|members|menuAction|menuAdd|menuChecked|menuClear|menuCollapse|menuData|menuDelete|menuEnable|menuEnabled|menuExpand|menuHover|menuPicture|menuSetAction|menuSetCheck|menuSetData|menuSetPicture|menuSetValue|menuShortcut|menuShortcutText|menuSize|menuSort|menuText|menuURL|menuValue|min|mineActive|mineDetectedBy|missionConfigFile|missionDifficulty|missionName|missionNamespace|missionStart|missionVersion|modelToWorld|modelToWorldVisual|modelToWorldVisualWorld|modelToWorldWorld|modParams|moonIntensity|moonPhase|morale|move|move3DENCamera|moveInAny|moveInCargo|moveInCommander|moveInDriver|moveInGunner|moveInTurret|moveObjectToEnd|moveOut|moveTime|moveTo|moveToCompleted|moveToFailed|musicVolume|name|nameSound|nearEntities|nearestBuilding|nearestLocation|nearestLocations|nearestLocationWithDubbing|nearestObject|nearestObjects|nearestTerrainObjects|nearObjects|nearObjectsReady|nearRoads|nearSupplies|nearTargets|needReload|netId|netObjNull|newOverlay|nextMenuItemIndex|nextWeatherChange|nMenuItems|numberOfEnginesRTD|numberToDate|objectCurators|objectFromNetId|objectParent|objNull|objStatus|onBriefingGear|onBriefingGroup|onBriefingNotes|onBriefingPlan|onBriefingTeamSwitch|onCommandModeChanged|onDoubleClick|onEachFrame|onGroupIconClick|onGroupIconOverEnter|onGroupIconOverLeave|onHCGroupSelectionChanged|onMapSingleClick|onPlayerConnected|onPlayerDisconnected|onPreloadFinished|onPreloadStarted|onShowNewObject|onTeamSwitch|openCuratorInterface|openDLCPage|openDSInterface|openMap|openSteamApp|openYoutubeVideo|opfor|orderGetIn|overcast|overcastForecast|owner|param|params|parseNumber|parseSimpleArray|parseText|parsingNamespace|particlesQuality|pi|pickWeaponPool|pitch|pixelGrid|pixelGridBase|pixelGridNoUIScale|pixelH|pixelW|playableSlotsNumber|playableUnits|playAction|playActionNow|player|playerRespawnTime|playerSide|playersNumber|playGesture|playMission|playMove|playMoveNow|playMusic|playScriptedMission|playSound|playSound3D|position|positionCameraToWorld|posScreenToWorld|posWorldToScreen|ppEffectAdjust|ppEffectCommit|ppEffectCommitted|ppEffectCreate|ppEffectDestroy|ppEffectEnable|ppEffectEnabled|ppEffectForceInNVG|precision|preloadCamera|preloadObject|preloadSound|preloadTitleObj|preloadTitleRsc|primaryWeapon|primaryWeaponItems|primaryWeaponMagazine|priority|processDiaryLink|processInitCommands|productVersion|profileName|profileNamespace|profileNameSteam|progressLoadingScreen|progressPosition|progressSetPosition|publicVariable|publicVariableClient|publicVariableServer|pushBack|pushBackUnique|putWeaponPool|queryItemsPool|queryMagazinePool|queryWeaponPool|rad|radioChannelAdd|radioChannelCreate|radioChannelRemove|radioChannelSetCallSign|radioChannelSetLabel|radioVolume|rain|rainbow|random|rank|rankId|rating|rectangular|registeredTasks|registerTask|reload|reloadEnabled|remoteControl|remoteExec|remoteExecCall|remoteExecutedOwner|remove3DENConnection|remove3DENEventHandler|remove3DENLayer|removeAction|removeAll3DENEventHandlers|removeAllActions|removeAllAssignedItems|removeAllContainers|removeAllCuratorAddons|removeAllCuratorCameraAreas|removeAllCuratorEditingAreas|removeAllEventHandlers|removeAllHandgunItems|removeAllItems|removeAllItemsWithMagazines|removeAllMissionEventHandlers|removeAllMPEventHandlers|removeAllMusicEventHandlers|removeAllOwnedMines|removeAllPrimaryWeaponItems|removeAllWeapons|removeBackpack|removeBackpackGlobal|removeCuratorAddons|removeCuratorCameraArea|removeCuratorEditableObjects|removeCuratorEditingArea|removeDrawIcon|removeDrawLinks|removeEventHandler|removeFromRemainsCollector|removeGoggles|removeGroupIcon|removeHandgunItem|removeHeadgear|removeItem|removeItemFromBackpack|removeItemFromUniform|removeItemFromVest|removeItems|removeMagazine|removeMagazineGlobal|removeMagazines|removeMagazinesTurret|removeMagazineTurret|removeMenuItem|removeMissionEventHandler|removeMPEventHandler|removeMusicEventHandler|removeOwnedMine|removePrimaryWeaponItem|removeSecondaryWeaponItem|removeSimpleTask|removeSwitchableUnit|removeTeamMember|removeUniform|removeVest|removeWeapon|removeWeaponAttachmentCargo|removeWeaponCargo|removeWeaponGlobal|removeWeaponTurret|reportRemoteTarget|requiredVersion|resetCamShake|resetSubgroupDirection|resistance|resize|resources|respawnVehicle|restartEditorCamera|reveal|revealMine|reverse|reversedMouseY|roadAt|roadsConnectedTo|roleDescription|ropeAttachedObjects|ropeAttachedTo|ropeAttachEnabled|ropeAttachTo|ropeCreate|ropeCut|ropeDestroy|ropeDetach|ropeEndPosition|ropeLength|ropes|ropeUnwind|ropeUnwound|rotorsForcesRTD|rotorsRpmRTD|round|runInitScript|safeZoneH|safeZoneW|safeZoneWAbs|safeZoneX|safeZoneXAbs|safeZoneY|save3DENInventory|saveGame|saveIdentity|saveJoysticks|saveOverlay|saveProfileNamespace|saveStatus|saveVar|savingEnabled|say|say2D|say3D|score|scoreSide|screenshot|screenToWorld|scriptDone|scriptName|scriptNull|scudState|secondaryWeapon|secondaryWeaponItems|secondaryWeaponMagazine|select|selectBestPlaces|selectDiarySubject|selectedEditorObjects|selectEditorObject|selectionNames|selectionPosition|selectLeader|selectMax|selectMin|selectNoPlayer|selectPlayer|selectRandom|selectRandomWeighted|selectWeapon|selectWeaponTurret|sendAUMessage|sendSimpleCommand|sendTask|sendTaskResult|sendUDPMessage|serverCommand|serverCommandAvailable|serverCommandExecutable|serverName|serverTime|set|set3DENAttribute|set3DENAttributes|set3DENGrid|set3DENIconsVisible|set3DENLayer|set3DENLinesVisible|set3DENLogicType|set3DENMissionAttribute|set3DENMissionAttributes|set3DENModelsVisible|set3DENObjectType|set3DENSelected|setAccTime|setActualCollectiveRTD|setAirplaneThrottle|setAirportSide|setAmmo|setAmmoCargo|setAmmoOnPylon|setAnimSpeedCoef|setAperture|setApertureNew|setArmoryPoints|setAttributes|setAutonomous|setBehaviour|setBleedingRemaining|setBrakesRTD|setCameraInterest|setCamShakeDefParams|setCamShakeParams|setCamUseTI|setCaptive|setCenterOfMass|setCollisionLight|setCombatMode|setCompassOscillation|setConvoySeparation|setCuratorCameraAreaCeiling|setCuratorCoef|setCuratorEditingAreaType|setCuratorWaypointCost|setCurrentChannel|setCurrentTask|setCurrentWaypoint|setCustomAimCoef|setCustomWeightRTD|setDamage|setDammage|setDate|setDebriefingText|setDefaultCamera|setDestination|setDetailMapBlendPars|setDir|setDirection|setDrawIcon|setDriveOnPath|setDropInterval|setDynamicSimulationDistance|setDynamicSimulationDistanceCoef|setEditorMode|setEditorObjectScope|setEffectCondition|setEngineRpmRTD|setFace|setFaceAnimation|setFatigue|setFeatureType|setFlagAnimationPhase|setFlagOwner|setFlagSide|setFlagTexture|setFog|setForceGeneratorRTD|setFormation|setFormationTask|setFormDir|setFriend|setFromEditor|setFSMVariable|setFuel|setFuelCargo|setGroupIcon|setGroupIconParams|setGroupIconsSelectable|setGroupIconsVisible|setGroupId|setGroupIdGlobal|setGroupOwner|setGusts|setHideBehind|setHit|setHitIndex|setHitPointDamage|setHorizonParallaxCoef|setHUDMovementLevels|setIdentity|setImportance|setInfoPanel|setLeader|setLightAmbient|setLightAttenuation|setLightBrightness|setLightColor|setLightDayLight|setLightFlareMaxDistance|setLightFlareSize|setLightIntensity|setLightnings|setLightUseFlare|setLocalWindParams|setMagazineTurretAmmo|setMarkerAlpha|setMarkerAlphaLocal|setMarkerBrush|setMarkerBrushLocal|setMarkerColor|setMarkerColorLocal|setMarkerDir|setMarkerDirLocal|setMarkerPos|setMarkerPosLocal|setMarkerShape|setMarkerShapeLocal|setMarkerSize|setMarkerSizeLocal|setMarkerText|setMarkerTextLocal|setMarkerType|setMarkerTypeLocal|setMass|setMimic|setMousePosition|setMusicEffect|setMusicEventHandler|setName|setNameSound|setObjectArguments|setObjectMaterial|setObjectMaterialGlobal|setObjectProxy|setObjectTexture|setObjectTextureGlobal|setObjectViewDistance|setOvercast|setOwner|setOxygenRemaining|setParticleCircle|setParticleClass|setParticleFire|setParticleParams|setParticleRandom|setPilotCameraDirection|setPilotCameraRotation|setPilotCameraTarget|setPilotLight|setPiPEffect|setPitch|setPlateNumber|setPlayable|setPlayerRespawnTime|setPos|setPosASL|setPosASL2|setPosASLW|setPosATL|setPosition|setPosWorld|setPylonLoadOut|setPylonsPriority|setRadioMsg|setRain|setRainbow|setRandomLip|setRank|setRectangular|setRepairCargo|setRotorBrakeRTD|setShadowDistance|setShotParents|setSide|setSimpleTaskAlwaysVisible|setSimpleTaskCustomData|setSimpleTaskDescription|setSimpleTaskDestination|setSimpleTaskTarget|setSimpleTaskType|setSimulWeatherLayers|setSize|setSkill|setSlingLoad|setSoundEffect|setSpeaker|setSpeech|setSpeedMode|setStamina|setStaminaScheme|setStatValue|setSuppression|setSystemOfUnits|setTargetAge|setTaskMarkerOffset|setTaskResult|setTaskState|setTerrainGrid|setText|setTimeMultiplier|setTitleEffect|setToneMapping|setToneMappingParams|setTrafficDensity|setTrafficDistance|setTrafficGap|setTrafficSpeed|setTriggerActivation|setTriggerArea|setTriggerStatements|setTriggerText|setTriggerTimeout|setTriggerType|setType|setUnconscious|setUnitAbility|setUnitLoadout|setUnitPos|setUnitPosWeak|setUnitRank|setUnitRecoilCoefficient|setUnitTrait|setUnloadInCombat|setUserActionText|setUserMFDText|setUserMFDValue|setVariable|setVectorDir|setVectorDirAndUp|setVectorUp|setVehicleAmmo|setVehicleAmmoDef|setVehicleArmor|setVehicleCargo|setVehicleId|setVehicleInit|setVehicleLock|setVehiclePosition|setVehicleRadar|setVehicleReceiveRemoteTargets|setVehicleReportOwnPosition|setVehicleReportRemoteTargets|setVehicleTIPars|setVehicleVarName|setVelocity|setVelocityModelSpace|setVelocityTransformation|setViewDistance|setVisibleIfTreeCollapsed|setWantedRpmRTD|setWaves|setWaypointBehaviour|setWaypointCombatMode|setWaypointCompletionRadius|setWaypointDescription|setWaypointForceBehaviour|setWaypointFormation|setWaypointHousePosition|setWaypointLoiterRadius|setWaypointLoiterType|setWaypointName|setWaypointPosition|setWaypointScript|setWaypointSpeed|setWaypointStatements|setWaypointTimeout|setWaypointType|setWaypointVisible|setWeaponReloadingTime|setWind|setWindDir|setWindForce|setWindStr|setWingForceScaleRTD|setWPPos|show3DIcons|showChat|showCinemaBorder|showCommandingMenu|showCompass|showCuratorCompass|showGPS|showHUD|showLegend|showMap|shownArtilleryComputer|shownChat|shownCompass|shownCuratorCompass|showNewEditorObject|shownGPS|shownHUD|shownMap|shownPad|shownRadio|shownScoretable|shownUAVFeed|shownWarrant|shownWatch|showPad|showRadio|showScoretable|showSubtitles|showUAVFeed|showWarrant|showWatch|showWaypoint|showWaypoints|side|sideAmbientLife|sideChat|sideEmpty|sideEnemy|sideFriendly|sideLogic|sideRadio|sideUnknown|simpleTasks|simulationEnabled|simulCloudDensity|simulCloudOcclusion|simulInClouds|simulWeatherSync|sin|size|sizeOf|skill|skillFinal|skipTime|sleep|sliderPosition|sliderRange|sliderSetPosition|sliderSetRange|sliderSetSpeed|sliderSpeed|slingLoadAssistantShown|soldierMagazines|someAmmo|sort|soundVolume|speaker|speed|speedMode|splitString|sqrt|squadParams|stance|startLoadingScreen|stop|stopEngineRTD|stopped|str|sunOrMoon|supportInfo|suppressFor|surfaceIsWater|surfaceNormal|surfaceType|swimInDepth|switchableUnits|switchAction|switchCamera|switchGesture|switchLight|switchMove|synchronizedObjects|synchronizedTriggers|synchronizedWaypoints|synchronizeObjectsAdd|synchronizeObjectsRemove|synchronizeTrigger|synchronizeWaypoint|systemChat|systemOfUnits|tan|targetKnowledge|targets|targetsAggregate|targetsQuery|taskAlwaysVisible|taskChildren|taskCompleted|taskCustomData|taskDescription|taskDestination|taskHint|taskMarkerOffset|taskNull|taskParent|taskResult|taskState|taskType|teamMember|teamMemberNull|teamName|teams|teamSwitch|teamSwitchEnabled|teamType|terminate|terrainIntersect|terrainIntersectASL|terrainIntersectAtASL|text|textLog|textLogFormat|tg|time|timeMultiplier|titleCut|titleFadeOut|titleObj|titleRsc|titleText|toArray|toFixed|toLower|toString|toUpper|triggerActivated|triggerActivation|triggerArea|triggerAttachedVehicle|triggerAttachObject|triggerAttachVehicle|triggerDynamicSimulation|triggerStatements|triggerText|triggerTimeout|triggerTimeoutCurrent|triggerType|turretLocal|turretOwner|turretUnit|tvAdd|tvClear|tvCollapse|tvCollapseAll|tvCount|tvCurSel|tvData|tvDelete|tvExpand|tvExpandAll|tvPicture|tvPictureRight|tvSetColor|tvSetCurSel|tvSetData|tvSetPicture|tvSetPictureColor|tvSetPictureColorDisabled|tvSetPictureColorSelected|tvSetPictureRight|tvSetPictureRightColor|tvSetPictureRightColorDisabled|tvSetPictureRightColorSelected|tvSetSelectColor|tvSetText|tvSetTooltip|tvSetValue|tvSort|tvSortByValue|tvText|tvTooltip|tvValue|type|typeName|typeOf|UAVControl|uiNamespace|uiSleep|unassignCurator|unassignItem|unassignTeam|unassignVehicle|underwater|uniform|uniformContainer|uniformItems|uniformMagazines|unitAddons|unitAimPosition|unitAimPositionVisual|unitBackpack|unitIsUAV|unitPos|unitReady|unitRecoilCoefficient|units|unitsBelowHeight|unlinkItem|unlockAchievement|unregisterTask|updateDrawIcon|updateMenuItem|updateObjectTree|useAIOperMapObstructionTest|useAISteeringComponent|useAudioTimeForMoves|userInputDisabled|vectorAdd|vectorCos|vectorCrossProduct|vectorDiff|vectorDir|vectorDirVisual|vectorDistance|vectorDistanceSqr|vectorDotProduct|vectorFromTo|vectorMagnitude|vectorMagnitudeSqr|vectorModelToWorld|vectorModelToWorldVisual|vectorMultiply|vectorNormalized|vectorUp|vectorUpVisual|vectorWorldToModel|vectorWorldToModelVisual|vehicle|vehicleCargoEnabled|vehicleChat|vehicleRadio|vehicleReceiveRemoteTargets|vehicleReportOwnPosition|vehicleReportRemoteTargets|vehicles|vehicleVarName|velocity|velocityModelSpace|verifySignature|vest|vestContainer|vestItems|vestMagazines|viewDistance|visibleCompass|visibleGPS|visibleMap|visiblePosition|visiblePositionASL|visibleScoretable|visibleWatch|waitUntil|waves|waypointAttachedObject|waypointAttachedVehicle|waypointAttachObject|waypointAttachVehicle|waypointBehaviour|waypointCombatMode|waypointCompletionRadius|waypointDescription|waypointForceBehaviour|waypointFormation|waypointHousePosition|waypointLoiterRadius|waypointLoiterType|waypointName|waypointPosition|waypoints|waypointScript|waypointsEnabledUAV|waypointShow|waypointSpeed|waypointStatements|waypointTimeout|waypointTimeoutCurrent|waypointType|waypointVisible|weaponAccessories|weaponAccessoriesCargo|weaponCargo|weaponDirection|weaponInertia|weaponLowered|weapons|weaponsItems|weaponsItemsCargo|weaponState|weaponsTurret|weightRTD|west|WFSideText|wind|windDir|windRTD|windStr|wingsForcesRTD|worldName|worldSize|worldToModel|worldToModelVisual|worldToScreen)\b/i,number:/(?:\$|\b0x)[\da-f]+\b|(?:\B\.\d+|\b\d+(?:\.\d+)?)(?:e[+-]?\d+)?\b/i,operator:/##|>>|&&|\|\||[!=<>]=?|[-+*/%#^]|\b(?:and|mod|not|or)\b/i,"magic-variable":{pattern:/\b(?:this|thisList|thisTrigger|_exception|_fnc_scriptName|_fnc_scriptNameParent|_forEachIndex|_this|_thisEventHandler|_thisFSM|_thisScript|_x)\b/i,alias:"keyword"},constant:/\bDIK(?:_[a-z\d]+)+\b/i}),t.languages.insertBefore("sqf","string",{macro:{pattern:/(^[ \t]*)#[a-z](?:[^\r\n\\]|\\(?:\r\n|[\s\S]))*/im,lookbehind:!0,greedy:!0,alias:"property",inside:{directive:{pattern:/#[a-z]+\b/i,alias:"keyword"},comment:t.languages.sqf.comment}}}),delete t.languages.sqf["class-name"]}return Ew}var ww,HF;function zSe(){if(HF)return ww;HF=1,ww=e,e.displayName="squirrel",e.aliases=[];function e(t){t.languages.squirrel=t.languages.extend("clike",{comment:[t.languages.clike.comment[0],{pattern:/(^|[^\\:])(?:\/\/|#).*/,lookbehind:!0,greedy:!0}],string:{pattern:/(^|[^\\"'@])(?:@"(?:[^"]|"")*"(?!")|"(?:[^\\\r\n"]|\\.)*")/,lookbehind:!0,greedy:!0},"class-name":{pattern:/(\b(?:class|enum|extends|instanceof)\s+)\w+(?:\.\w+)*/,lookbehind:!0,inside:{punctuation:/\./}},keyword:/\b(?:__FILE__|__LINE__|base|break|case|catch|class|clone|const|constructor|continue|default|delete|else|enum|extends|for|foreach|function|if|in|instanceof|local|null|resume|return|static|switch|this|throw|try|typeof|while|yield)\b/,number:/\b(?:0x[0-9a-fA-F]+|\d+(?:\.(?:\d+|[eE][+-]?\d+))?)\b/,operator:/\+\+|--|<=>|<[-<]|>>>?|&&?|\|\|?|[-+*/%!=<>]=?|[~^]|::?/,punctuation:/[(){}\[\],;.]/}),t.languages.insertBefore("squirrel","string",{char:{pattern:/(^|[^\\"'])'(?:[^\\']|\\(?:[xuU][0-9a-fA-F]{0,8}|[\s\S]))'/,lookbehind:!0,greedy:!0}}),t.languages.insertBefore("squirrel","operator",{"attribute-punctuation":{pattern:/<\/|\/>/,alias:"important"},lambda:{pattern:/@(?=\()/,alias:"operator"}})}return ww}var xw,$F;function BSe(){if($F)return xw;$F=1,xw=e,e.displayName="stan",e.aliases=[];function e(t){(function(n){var r=/\b(?:algebra_solver|algebra_solver_newton|integrate_1d|integrate_ode|integrate_ode_bdf|integrate_ode_rk45|map_rect|ode_(?:adams|bdf|ckrk|rk45)(?:_tol)?|ode_adjoint_tol_ctl|reduce_sum|reduce_sum_static)\b/;n.languages.stan={comment:/\/\/.*|\/\*[\s\S]*?\*\/|#(?!include).*/,string:{pattern:/"[\x20\x21\x23-\x5B\x5D-\x7E]*"/,greedy:!0},directive:{pattern:/^([ \t]*)#include\b.*/m,lookbehind:!0,alias:"property"},"function-arg":{pattern:RegExp("("+r.source+/\s*\(\s*/.source+")"+/[a-zA-Z]\w*/.source),lookbehind:!0,alias:"function"},constraint:{pattern:/(\b(?:int|matrix|real|row_vector|vector)\s*)<[^<>]*>/,lookbehind:!0,inside:{expression:{pattern:/(=\s*)\S(?:\S|\s+(?!\s))*?(?=\s*(?:>$|,\s*\w+\s*=))/,lookbehind:!0,inside:null},property:/\b[a-z]\w*(?=\s*=)/i,operator:/=/,punctuation:/^<|>$|,/}},keyword:[{pattern:/\bdata(?=\s*\{)|\b(?:functions|generated|model|parameters|quantities|transformed)\b/,alias:"program-block"},/\b(?:array|break|cholesky_factor_corr|cholesky_factor_cov|complex|continue|corr_matrix|cov_matrix|data|else|for|if|in|increment_log_prob|int|matrix|ordered|positive_ordered|print|real|reject|return|row_vector|simplex|target|unit_vector|vector|void|while)\b/,r],function:/\b[a-z]\w*(?=\s*\()/i,number:/(?:\b\d+(?:_\d+)*(?:\.(?:\d+(?:_\d+)*)?)?|\B\.\d+(?:_\d+)*)(?:E[+-]?\d+(?:_\d+)*)?i?(?!\w)/i,boolean:/\b(?:false|true)\b/,operator:/<-|\.[*/]=?|\|\|?|&&|[!=<>+\-*/]=?|['^%~?:]/,punctuation:/[()\[\]{},;]/},n.languages.stan.constraint.inside.expression.inside=n.languages.stan})(t)}return xw}var kw,qF;function USe(){if(qF)return kw;qF=1,kw=e,e.displayName="stylus",e.aliases=[];function e(t){(function(n){var r={pattern:/(\b\d+)(?:%|[a-z]+)/,lookbehind:!0},a={pattern:/(^|[^\w.-])-?(?:\d+(?:\.\d+)?|\.\d+)/,lookbehind:!0},o={comment:{pattern:/(^|[^\\])(?:\/\*[\s\S]*?\*\/|\/\/.*)/,lookbehind:!0},url:{pattern:/\burl\((["']?).*?\1\)/i,greedy:!0},string:{pattern:/("|')(?:(?!\1)[^\\\r\n]|\\(?:\r\n|[\s\S]))*\1/,greedy:!0},interpolation:null,func:null,important:/\B!(?:important|optional)\b/i,keyword:{pattern:/(^|\s+)(?:(?:else|for|if|return|unless)(?=\s|$)|@[\w-]+)/,lookbehind:!0},hexcode:/#[\da-f]{3,6}/i,color:[/\b(?:AliceBlue|AntiqueWhite|Aqua|Aquamarine|Azure|Beige|Bisque|Black|BlanchedAlmond|Blue|BlueViolet|Brown|BurlyWood|CadetBlue|Chartreuse|Chocolate|Coral|CornflowerBlue|Cornsilk|Crimson|Cyan|DarkBlue|DarkCyan|DarkGoldenRod|DarkGr[ae]y|DarkGreen|DarkKhaki|DarkMagenta|DarkOliveGreen|DarkOrange|DarkOrchid|DarkRed|DarkSalmon|DarkSeaGreen|DarkSlateBlue|DarkSlateGr[ae]y|DarkTurquoise|DarkViolet|DeepPink|DeepSkyBlue|DimGr[ae]y|DodgerBlue|FireBrick|FloralWhite|ForestGreen|Fuchsia|Gainsboro|GhostWhite|Gold|GoldenRod|Gr[ae]y|Green|GreenYellow|HoneyDew|HotPink|IndianRed|Indigo|Ivory|Khaki|Lavender|LavenderBlush|LawnGreen|LemonChiffon|LightBlue|LightCoral|LightCyan|LightGoldenRodYellow|LightGr[ae]y|LightGreen|LightPink|LightSalmon|LightSeaGreen|LightSkyBlue|LightSlateGr[ae]y|LightSteelBlue|LightYellow|Lime|LimeGreen|Linen|Magenta|Maroon|MediumAquaMarine|MediumBlue|MediumOrchid|MediumPurple|MediumSeaGreen|MediumSlateBlue|MediumSpringGreen|MediumTurquoise|MediumVioletRed|MidnightBlue|MintCream|MistyRose|Moccasin|NavajoWhite|Navy|OldLace|Olive|OliveDrab|Orange|OrangeRed|Orchid|PaleGoldenRod|PaleGreen|PaleTurquoise|PaleVioletRed|PapayaWhip|PeachPuff|Peru|Pink|Plum|PowderBlue|Purple|Red|RosyBrown|RoyalBlue|SaddleBrown|Salmon|SandyBrown|SeaGreen|SeaShell|Sienna|Silver|SkyBlue|SlateBlue|SlateGr[ae]y|Snow|SpringGreen|SteelBlue|Tan|Teal|Thistle|Tomato|Transparent|Turquoise|Violet|Wheat|White|WhiteSmoke|Yellow|YellowGreen)\b/i,{pattern:/\b(?:hsl|rgb)\(\s*\d{1,3}\s*,\s*\d{1,3}%?\s*,\s*\d{1,3}%?\s*\)\B|\b(?:hsl|rgb)a\(\s*\d{1,3}\s*,\s*\d{1,3}%?\s*,\s*\d{1,3}%?\s*,\s*(?:0|0?\.\d+|1)\s*\)\B/i,inside:{unit:r,number:a,function:/[\w-]+(?=\()/,punctuation:/[(),]/}}],entity:/\\[\da-f]{1,8}/i,unit:r,boolean:/\b(?:false|true)\b/,operator:[/~|[+!\/%<>?=]=?|[-:]=|\*[*=]?|\.{2,3}|&&|\|\||\B-\B|\b(?:and|in|is(?: a| defined| not|nt)?|not|or)\b/],number:a,punctuation:/[{}()\[\];:,]/};o.interpolation={pattern:/\{[^\r\n}:]+\}/,alias:"variable",inside:{delimiter:{pattern:/^\{|\}$/,alias:"punctuation"},rest:o}},o.func={pattern:/[\w-]+\([^)]*\).*/,inside:{function:/^[^(]+/,rest:o}},n.languages.stylus={"atrule-declaration":{pattern:/(^[ \t]*)@.+/m,lookbehind:!0,inside:{atrule:/^@[\w-]+/,rest:o}},"variable-declaration":{pattern:/(^[ \t]*)[\w$-]+\s*.?=[ \t]*(?:\{[^{}]*\}|\S.*|$)/m,lookbehind:!0,inside:{variable:/^\S+/,rest:o}},statement:{pattern:/(^[ \t]*)(?:else|for|if|return|unless)[ \t].+/m,lookbehind:!0,inside:{keyword:/^\S+/,rest:o}},"property-declaration":{pattern:/((?:^|\{)([ \t]*))(?:[\w-]|\{[^}\r\n]+\})+(?:\s*:\s*|[ \t]+)(?!\s)[^{\r\n]*(?:;|[^{\r\n,]$(?!(?:\r?\n|\r)(?:\{|\2[ \t])))/m,lookbehind:!0,inside:{property:{pattern:/^[^\s:]+/,inside:{interpolation:o.interpolation}},rest:o}},selector:{pattern:/(^[ \t]*)(?:(?=\S)(?:[^{}\r\n:()]|::?[\w-]+(?:\([^)\r\n]*\)|(?![\w-]))|\{[^}\r\n]+\})+)(?:(?:\r?\n|\r)(?:\1(?:(?=\S)(?:[^{}\r\n:()]|::?[\w-]+(?:\([^)\r\n]*\)|(?![\w-]))|\{[^}\r\n]+\})+)))*(?:,$|\{|(?=(?:\r?\n|\r)(?:\{|\1[ \t])))/m,lookbehind:!0,inside:{interpolation:o.interpolation,comment:o.comment,punctuation:/[{},]/}},func:o.func,string:o.string,comment:{pattern:/(^|[^\\])(?:\/\*[\s\S]*?\*\/|\/\/.*)/,lookbehind:!0,greedy:!0},interpolation:o.interpolation,punctuation:/[{}()\[\];:.]/}})(t)}return kw}var Tw,VF;function jSe(){if(VF)return Tw;VF=1,Tw=e,e.displayName="swift",e.aliases=[];function e(t){t.languages.swift={comment:{pattern:/(^|[^\\:])(?:\/\/.*|\/\*(?:[^/*]|\/(?!\*)|\*(?!\/)|\/\*(?:[^*]|\*(?!\/))*\*\/)*\*\/)/,lookbehind:!0,greedy:!0},"string-literal":[{pattern:RegExp(/(^|[^"#])/.source+"(?:"+/"(?:\\(?:\((?:[^()]|\([^()]*\))*\)|\r\n|[^(])|[^\\\r\n"])*"/.source+"|"+/"""(?:\\(?:\((?:[^()]|\([^()]*\))*\)|[^(])|[^\\"]|"(?!""))*"""/.source+")"+/(?!["#])/.source),lookbehind:!0,greedy:!0,inside:{interpolation:{pattern:/(\\\()(?:[^()]|\([^()]*\))*(?=\))/,lookbehind:!0,inside:null},"interpolation-punctuation":{pattern:/^\)|\\\($/,alias:"punctuation"},punctuation:/\\(?=[\r\n])/,string:/[\s\S]+/}},{pattern:RegExp(/(^|[^"#])(#+)/.source+"(?:"+/"(?:\\(?:#+\((?:[^()]|\([^()]*\))*\)|\r\n|[^#])|[^\\\r\n])*?"/.source+"|"+/"""(?:\\(?:#+\((?:[^()]|\([^()]*\))*\)|[^#])|[^\\])*?"""/.source+")\\2"),lookbehind:!0,greedy:!0,inside:{interpolation:{pattern:/(\\#+\()(?:[^()]|\([^()]*\))*(?=\))/,lookbehind:!0,inside:null},"interpolation-punctuation":{pattern:/^\)|\\#+\($/,alias:"punctuation"},string:/[\s\S]+/}}],directive:{pattern:RegExp(/#/.source+"(?:"+(/(?:elseif|if)\b/.source+"(?:[ 	]*"+/(?:![ \t]*)?(?:\b\w+\b(?:[ \t]*\((?:[^()]|\([^()]*\))*\))?|\((?:[^()]|\([^()]*\))*\))(?:[ \t]*(?:&&|\|\|))?/.source+")+")+"|"+/(?:else|endif)\b/.source+")"),alias:"property",inside:{"directive-name":/^#\w+/,boolean:/\b(?:false|true)\b/,number:/\b\d+(?:\.\d+)*\b/,operator:/!|&&|\|\||[<>]=?/,punctuation:/[(),]/}},literal:{pattern:/#(?:colorLiteral|column|dsohandle|file(?:ID|Literal|Path)?|function|imageLiteral|line)\b/,alias:"constant"},"other-directive":{pattern:/#\w+\b/,alias:"property"},attribute:{pattern:/@\w+/,alias:"atrule"},"function-definition":{pattern:/(\bfunc\s+)\w+/,lookbehind:!0,alias:"function"},label:{pattern:/\b(break|continue)\s+\w+|\b[a-zA-Z_]\w*(?=\s*:\s*(?:for|repeat|while)\b)/,lookbehind:!0,alias:"important"},keyword:/\b(?:Any|Protocol|Self|Type|actor|as|assignment|associatedtype|associativity|async|await|break|case|catch|class|continue|convenience|default|defer|deinit|didSet|do|dynamic|else|enum|extension|fallthrough|fileprivate|final|for|func|get|guard|higherThan|if|import|in|indirect|infix|init|inout|internal|is|isolated|lazy|left|let|lowerThan|mutating|none|nonisolated|nonmutating|open|operator|optional|override|postfix|precedencegroup|prefix|private|protocol|public|repeat|required|rethrows|return|right|safe|self|set|some|static|struct|subscript|super|switch|throw|throws|try|typealias|unowned|unsafe|var|weak|where|while|willSet)\b/,boolean:/\b(?:false|true)\b/,nil:{pattern:/\bnil\b/,alias:"constant"},"short-argument":/\$\d+\b/,omit:{pattern:/\b_\b/,alias:"keyword"},number:/\b(?:[\d_]+(?:\.[\de_]+)?|0x[a-f0-9_]+(?:\.[a-f0-9p_]+)?|0b[01_]+|0o[0-7_]+)\b/i,"class-name":/\b[A-Z](?:[A-Z_\d]*[a-z]\w*)?\b/,function:/\b[a-z_]\w*(?=\s*\()/i,constant:/\b(?:[A-Z_]{2,}|k[A-Z][A-Za-z_]+)\b/,operator:/[-+*/%=!<>&|^~?]+|\.[.\-+*/%=!<>&|^~?]+/,punctuation:/[{}[\]();,.:\\]/},t.languages.swift["string-literal"].forEach(function(n){n.inside.interpolation.inside=t.languages.swift})}return Tw}var Aw,WF;function GSe(){if(WF)return Aw;WF=1,Aw=e,e.displayName="systemd",e.aliases=[];function e(t){(function(n){var r={pattern:/^[;#].*/m,greedy:!0},a=/"(?:[^\r\n"\\]|\\(?:[^\r]|\r\n?))*"(?!\S)/.source;n.languages.systemd={comment:r,section:{pattern:/^\[[^\n\r\[\]]*\](?=[ \t]*$)/m,greedy:!0,inside:{punctuation:/^\[|\]$/,"section-name":{pattern:/[\s\S]+/,alias:"selector"}}},key:{pattern:/^[^\s=]+(?=[ \t]*=)/m,greedy:!0,alias:"attr-name"},value:{pattern:RegExp(/(=[ \t]*(?!\s))/.source+"(?:"+a+`|(?=[^"\r
]))(?:`+(/[^\s\\]/.source+'|[ 	]+(?:(?![ 	"])|'+a+")|"+/\\[\r\n]+(?:[#;].*[\r\n]+)*(?![#;])/.source)+")*"),lookbehind:!0,greedy:!0,alias:"attr-value",inside:{comment:r,quoted:{pattern:RegExp(/(^|\s)/.source+a),lookbehind:!0,greedy:!0},punctuation:/\\$/m,boolean:{pattern:/^(?:false|no|off|on|true|yes)$/,greedy:!0}}},punctuation:/=/}})(t)}return Aw}var Rw,YF;function hA(){if(YF)return Rw;YF=1,Rw=e,e.displayName="t4Templating",e.aliases=[];function e(t){(function(n){function r(o,s,u){return{pattern:RegExp("<#"+o+"[\\s\\S]*?#>"),alias:"block",inside:{delimiter:{pattern:RegExp("^<#"+o+"|#>$"),alias:"important"},content:{pattern:/[\s\S]+/,inside:s,alias:u}}}}function a(o){var s=n.languages[o],u="language-"+o;return{block:{pattern:/<#[\s\S]+?#>/,inside:{directive:r("@",{"attr-value":{pattern:/=(?:("|')(?:\\[\s\S]|(?!\1)[^\\])*\1|[^\s'">=]+)/,inside:{punctuation:/^=|^["']|["']$/}},keyword:/\b\w+(?=\s)/,"attr-name":/\b\w+/}),expression:r("=",s,u),"class-feature":r("\\+",s,u),standard:r("",s,u)}}}}n.languages["t4-templating"]=Object.defineProperty({},"createT4",{value:a})})(t)}return Rw}var _w,KF;function HSe(){if(KF)return _w;KF=1;var e=hA(),t=Mp();_w=n,n.displayName="t4Cs",n.aliases=[];function n(r){r.register(e),r.register(t),r.languages.t4=r.languages["t4-cs"]=r.languages["t4-templating"].createT4("csharp")}return _w}var Cw,XF;function WH(){if(XF)return Cw;XF=1;var e=HH();Cw=t,t.displayName="vbnet",t.aliases=[];function t(n){n.register(e),n.languages.vbnet=n.languages.extend("basic",{comment:[{pattern:/(?:!|REM\b).+/i,inside:{keyword:/^REM/i}},{pattern:/(^|[^\\:])'.*/,lookbehind:!0,greedy:!0}],string:{pattern:/(^|[^"])"(?:""|[^"])*"(?!")/,lookbehind:!0,greedy:!0},keyword:/(?:\b(?:ADDHANDLER|ADDRESSOF|ALIAS|AND|ANDALSO|AS|BEEP|BLOAD|BOOLEAN|BSAVE|BYREF|BYTE|BYVAL|CALL(?: ABSOLUTE)?|CASE|CATCH|CBOOL|CBYTE|CCHAR|CDATE|CDBL|CDEC|CHAIN|CHAR|CHDIR|CINT|CLASS|CLEAR|CLNG|CLOSE|CLS|COBJ|COM|COMMON|CONST|CONTINUE|CSBYTE|CSHORT|CSNG|CSTR|CTYPE|CUINT|CULNG|CUSHORT|DATA|DATE|DECIMAL|DECLARE|DEF(?: FN| SEG|DBL|INT|LNG|SNG|STR)|DEFAULT|DELEGATE|DIM|DIRECTCAST|DO|DOUBLE|ELSE|ELSEIF|END|ENUM|ENVIRON|ERASE|ERROR|EVENT|EXIT|FALSE|FIELD|FILES|FINALLY|FOR(?: EACH)?|FRIEND|FUNCTION|GET|GETTYPE|GETXMLNAMESPACE|GLOBAL|GOSUB|GOTO|HANDLES|IF|IMPLEMENTS|IMPORTS|IN|INHERITS|INPUT|INTEGER|INTERFACE|IOCTL|IS|ISNOT|KEY|KILL|LET|LIB|LIKE|LINE INPUT|LOCATE|LOCK|LONG|LOOP|LSET|ME|MKDIR|MOD|MODULE|MUSTINHERIT|MUSTOVERRIDE|MYBASE|MYCLASS|NAME|NAMESPACE|NARROWING|NEW|NEXT|NOT|NOTHING|NOTINHERITABLE|NOTOVERRIDABLE|OBJECT|OF|OFF|ON(?: COM| ERROR| KEY| TIMER)?|OPEN|OPERATOR|OPTION(?: BASE)?|OPTIONAL|OR|ORELSE|OUT|OVERLOADS|OVERRIDABLE|OVERRIDES|PARAMARRAY|PARTIAL|POKE|PRIVATE|PROPERTY|PROTECTED|PUBLIC|PUT|RAISEEVENT|READ|READONLY|REDIM|REM|REMOVEHANDLER|RESTORE|RESUME|RETURN|RMDIR|RSET|RUN|SBYTE|SELECT(?: CASE)?|SET|SHADOWS|SHARED|SHELL|SHORT|SINGLE|SLEEP|STATIC|STEP|STOP|STRING|STRUCTURE|SUB|SWAP|SYNCLOCK|SYSTEM|THEN|THROW|TIMER|TO|TROFF|TRON|TRUE|TRY|TRYCAST|TYPE|TYPEOF|UINTEGER|ULONG|UNLOCK|UNTIL|USHORT|USING|VIEW PRINT|WAIT|WEND|WHEN|WHILE|WIDENING|WITH|WITHEVENTS|WRITE|WRITEONLY|XOR)|\B(?:#CONST|#ELSE|#ELSEIF|#END|#IF))(?:\$|\b)/i,punctuation:/[,;:(){}]/})}return Cw}var Nw,ZF;function $Se(){if(ZF)return Nw;ZF=1;var e=hA(),t=WH();Nw=n,n.displayName="t4Vb",n.aliases=[];function n(r){r.register(e),r.register(t),r.languages["t4-vb"]=r.languages["t4-templating"].createT4("vbnet")}return Nw}var Ow,QF;function YH(){if(QF)return Ow;QF=1,Ow=e,e.displayName="yaml",e.aliases=["yml"];function e(t){(function(n){var r=/[*&][^\s[\]{},]+/,a=/!(?:<[\w\-%#;/?:@&=+$,.!~*'()[\]]+>|(?:[a-zA-Z\d-]*!)?[\w\-%#;/?:@&=+$.~*'()]+)?/,o="(?:"+a.source+"(?:[ 	]+"+r.source+")?|"+r.source+"(?:[ 	]+"+a.source+")?)",s=/(?:[^\s\x00-\x08\x0e-\x1f!"#%&'*,\-:>?@[\]`{|}\x7f-\x84\x86-\x9f\ud800-\udfff\ufffe\uffff]|[?:-]<PLAIN>)(?:[ \t]*(?:(?![#:])<PLAIN>|:<PLAIN>))*/.source.replace(/<PLAIN>/g,function(){return/[^\s\x00-\x08\x0e-\x1f,[\]{}\x7f-\x84\x86-\x9f\ud800-\udfff\ufffe\uffff]/.source}),u=/"(?:[^"\\\r\n]|\\.)*"|'(?:[^'\\\r\n]|\\.)*'/.source;function c(d,p){p=(p||"").replace(/m/g,"")+"m";var g=/([:\-,[{]\s*(?:\s<<prop>>[ \t]+)?)(?:<<value>>)(?=[ \t]*(?:$|,|\]|\}|(?:[\r\n]\s*)?#))/.source.replace(/<<prop>>/g,function(){return o}).replace(/<<value>>/g,function(){return d});return RegExp(g,p)}n.languages.yaml={scalar:{pattern:RegExp(/([\-:]\s*(?:\s<<prop>>[ \t]+)?[|>])[ \t]*(?:((?:\r?\n|\r)[ \t]+)\S[^\r\n]*(?:\2[^\r\n]+)*)/.source.replace(/<<prop>>/g,function(){return o})),lookbehind:!0,alias:"string"},comment:/#.*/,key:{pattern:RegExp(/((?:^|[:\-,[{\r\n?])[ \t]*(?:<<prop>>[ \t]+)?)<<key>>(?=\s*:\s)/.source.replace(/<<prop>>/g,function(){return o}).replace(/<<key>>/g,function(){return"(?:"+s+"|"+u+")"})),lookbehind:!0,greedy:!0,alias:"atrule"},directive:{pattern:/(^[ \t]*)%.+/m,lookbehind:!0,alias:"important"},datetime:{pattern:c(/\d{4}-\d\d?-\d\d?(?:[tT]|[ \t]+)\d\d?:\d{2}:\d{2}(?:\.\d*)?(?:[ \t]*(?:Z|[-+]\d\d?(?::\d{2})?))?|\d{4}-\d{2}-\d{2}|\d\d?:\d{2}(?::\d{2}(?:\.\d*)?)?/.source),lookbehind:!0,alias:"number"},boolean:{pattern:c(/false|true/.source,"i"),lookbehind:!0,alias:"important"},null:{pattern:c(/null|~/.source,"i"),lookbehind:!0,alias:"important"},string:{pattern:c(u),lookbehind:!0,greedy:!0},number:{pattern:c(/[+-]?(?:0x[\da-f]+|0o[0-7]+|(?:\d+(?:\.\d*)?|\.\d+)(?:e[+-]?\d+)?|\.inf|\.nan)/.source,"i"),lookbehind:!0},tag:a,important:r,punctuation:/---|[:[\]{}\-,|>?]|\.\.\./},n.languages.yml=n.languages.yaml})(t)}return Ow}var Dw,JF;function qSe(){if(JF)return Dw;JF=1;var e=YH();Dw=t,t.displayName="tap",t.aliases=[];function t(n){n.register(e),n.languages.tap={fail:/not ok[^#{\n\r]*/,pass:/ok[^#{\n\r]*/,pragma:/pragma [+-][a-z]+/,bailout:/bail out!.*/i,version:/TAP version \d+/i,plan:/\b\d+\.\.\d+(?: +#.*)?/,subtest:{pattern:/# Subtest(?:: .*)?/,greedy:!0},punctuation:/[{}]/,directive:/#.*/,yamlish:{pattern:/(^[ \t]*)---[\s\S]*?[\r\n][ \t]*\.\.\.$/m,lookbehind:!0,inside:n.languages.yaml,alias:"language-yaml"}}}return Dw}var Iw,eP;function VSe(){if(eP)return Iw;eP=1,Iw=e,e.displayName="tcl",e.aliases=[];function e(t){t.languages.tcl={comment:{pattern:/(^|[^\\])#.*/,lookbehind:!0},string:{pattern:/"(?:[^"\\\r\n]|\\(?:\r\n|[\s\S]))*"/,greedy:!0},variable:[{pattern:/(\$)(?:::)?(?:[a-zA-Z0-9]+::)*\w+/,lookbehind:!0},{pattern:/(\$)\{[^}]+\}/,lookbehind:!0},{pattern:/(^[\t ]*set[ \t]+)(?:::)?(?:[a-zA-Z0-9]+::)*\w+/m,lookbehind:!0}],function:{pattern:/(^[\t ]*proc[ \t]+)\S+/m,lookbehind:!0},builtin:[{pattern:/(^[\t ]*)(?:break|class|continue|error|eval|exit|for|foreach|if|proc|return|switch|while)\b/m,lookbehind:!0},/\b(?:else|elseif)\b/],scope:{pattern:/(^[\t ]*)(?:global|upvar|variable)\b/m,lookbehind:!0,alias:"constant"},keyword:{pattern:/(^[\t ]*|\[)(?:Safe_Base|Tcl|after|append|apply|array|auto_(?:execok|import|load|mkindex|qualify|reset)|automkindex_old|bgerror|binary|catch|cd|chan|clock|close|concat|dde|dict|encoding|eof|exec|expr|fblocked|fconfigure|fcopy|file(?:event|name)?|flush|gets|glob|history|http|incr|info|interp|join|lappend|lassign|lindex|linsert|list|llength|load|lrange|lrepeat|lreplace|lreverse|lsearch|lset|lsort|math(?:func|op)|memory|msgcat|namespace|open|package|parray|pid|pkg_mkIndex|platform|puts|pwd|re_syntax|read|refchan|regexp|registry|regsub|rename|scan|seek|set|socket|source|split|string|subst|tcl(?:_endOfWord|_findLibrary|startOf(?:Next|Previous)Word|test|vars|wordBreak(?:After|Before))|tell|time|tm|trace|unknown|unload|unset|update|uplevel|vwait)\b/m,lookbehind:!0},operator:/!=?|\*\*?|==|&&?|\|\|?|<[=<]?|>[=>]?|[-+~\/%?^]|\b(?:eq|in|ne|ni)\b/,punctuation:/[{}()\[\]]/}}return Iw}var Lw,tP;function WSe(){if(tP)return Lw;tP=1,Lw=e,e.displayName="textile",e.aliases=[];function e(t){(function(n){var r=/\([^|()\n]+\)|\[[^\]\n]+\]|\{[^}\n]+\}/.source,a=/\)|\((?![^|()\n]+\))/.source;function o(m,b){return RegExp(m.replace(/<MOD>/g,function(){return"(?:"+r+")"}).replace(/<PAR>/g,function(){return"(?:"+a+")"}),b||"")}var s={css:{pattern:/\{[^{}]+\}/,inside:{rest:n.languages.css}},"class-id":{pattern:/(\()[^()]+(?=\))/,lookbehind:!0,alias:"attr-value"},lang:{pattern:/(\[)[^\[\]]+(?=\])/,lookbehind:!0,alias:"attr-value"},punctuation:/[\\\/]\d+|\S/},u=n.languages.textile=n.languages.extend("markup",{phrase:{pattern:/(^|\r|\n)\S[\s\S]*?(?=$|\r?\n\r?\n|\r\r)/,lookbehind:!0,inside:{"block-tag":{pattern:o(/^[a-z]\w*(?:<MOD>|<PAR>|[<>=])*\./.source),inside:{modifier:{pattern:o(/(^[a-z]\w*)(?:<MOD>|<PAR>|[<>=])+(?=\.)/.source),lookbehind:!0,inside:s},tag:/^[a-z]\w*/,punctuation:/\.$/}},list:{pattern:o(/^[*#]+<MOD>*\s+\S.*/.source,"m"),inside:{modifier:{pattern:o(/(^[*#]+)<MOD>+/.source),lookbehind:!0,inside:s},punctuation:/^[*#]+/}},table:{pattern:o(/^(?:(?:<MOD>|<PAR>|[<>=^~])+\.\s*)?(?:\|(?:(?:<MOD>|<PAR>|[<>=^~_]|[\\/]\d+)+\.|(?!(?:<MOD>|<PAR>|[<>=^~_]|[\\/]\d+)+\.))[^|]*)+\|/.source,"m"),inside:{modifier:{pattern:o(/(^|\|(?:\r?\n|\r)?)(?:<MOD>|<PAR>|[<>=^~_]|[\\/]\d+)+(?=\.)/.source),lookbehind:!0,inside:s},punctuation:/\||^\./}},inline:{pattern:o(/(^|[^a-zA-Z\d])(\*\*|__|\?\?|[*_%@+\-^~])<MOD>*.+?\2(?![a-zA-Z\d])/.source),lookbehind:!0,inside:{bold:{pattern:o(/(^(\*\*?)<MOD>*).+?(?=\2)/.source),lookbehind:!0},italic:{pattern:o(/(^(__?)<MOD>*).+?(?=\2)/.source),lookbehind:!0},cite:{pattern:o(/(^\?\?<MOD>*).+?(?=\?\?)/.source),lookbehind:!0,alias:"string"},code:{pattern:o(/(^@<MOD>*).+?(?=@)/.source),lookbehind:!0,alias:"keyword"},inserted:{pattern:o(/(^\+<MOD>*).+?(?=\+)/.source),lookbehind:!0},deleted:{pattern:o(/(^-<MOD>*).+?(?=-)/.source),lookbehind:!0},span:{pattern:o(/(^%<MOD>*).+?(?=%)/.source),lookbehind:!0},modifier:{pattern:o(/(^\*\*|__|\?\?|[*_%@+\-^~])<MOD>+/.source),lookbehind:!0,inside:s},punctuation:/[*_%?@+\-^~]+/}},"link-ref":{pattern:/^\[[^\]]+\]\S+$/m,inside:{string:{pattern:/(^\[)[^\]]+(?=\])/,lookbehind:!0},url:{pattern:/(^\])\S+$/,lookbehind:!0},punctuation:/[\[\]]/}},link:{pattern:o(/"<MOD>*[^"]+":.+?(?=[^\w/]?(?:\s|$))/.source),inside:{text:{pattern:o(/(^"<MOD>*)[^"]+(?=")/.source),lookbehind:!0},modifier:{pattern:o(/(^")<MOD>+/.source),lookbehind:!0,inside:s},url:{pattern:/(:).+/,lookbehind:!0},punctuation:/[":]/}},image:{pattern:o(/!(?:<MOD>|<PAR>|[<>=])*(?![<>=])[^!\s()]+(?:\([^)]+\))?!(?::.+?(?=[^\w/]?(?:\s|$)))?/.source),inside:{source:{pattern:o(/(^!(?:<MOD>|<PAR>|[<>=])*)(?![<>=])[^!\s()]+(?:\([^)]+\))?(?=!)/.source),lookbehind:!0,alias:"url"},modifier:{pattern:o(/(^!)(?:<MOD>|<PAR>|[<>=])+/.source),lookbehind:!0,inside:s},url:{pattern:/(:).+/,lookbehind:!0},punctuation:/[!:]/}},footnote:{pattern:/\b\[\d+\]/,alias:"comment",inside:{punctuation:/\[|\]/}},acronym:{pattern:/\b[A-Z\d]+\([^)]+\)/,inside:{comment:{pattern:/(\()[^()]+(?=\))/,lookbehind:!0},punctuation:/[()]/}},mark:{pattern:/\b\((?:C|R|TM)\)/,alias:"comment",inside:{punctuation:/[()]/}}}}}),c=u.phrase.inside,d={inline:c.inline,link:c.link,image:c.image,footnote:c.footnote,acronym:c.acronym,mark:c.mark};u.tag.pattern=/<\/?(?!\d)[a-z0-9]+(?:\s+[^\s>\/=]+(?:=(?:("|')(?:\\[\s\S]|(?!\1)[^\\])*\1|[^\s'">=]+))?)*\s*\/?>/i;var p=c.inline.inside;p.bold.inside=d,p.italic.inside=d,p.inserted.inside=d,p.deleted.inside=d,p.span.inside=d;var g=c.table.inside;g.inline=d.inline,g.link=d.link,g.image=d.image,g.footnote=d.footnote,g.acronym=d.acronym,g.mark=d.mark})(t)}return Lw}var Mw,nP;function YSe(){if(nP)return Mw;nP=1,Mw=e,e.displayName="toml",e.aliases=[];function e(t){(function(n){var r=/(?:[\w-]+|'[^'\n\r]*'|"(?:\\.|[^\\"\r\n])*")/.source;function a(o){return o.replace(/__/g,function(){return r})}n.languages.toml={comment:{pattern:/#.*/,greedy:!0},table:{pattern:RegExp(a(/(^[\t ]*\[\s*(?:\[\s*)?)__(?:\s*\.\s*__)*(?=\s*\])/.source),"m"),lookbehind:!0,greedy:!0,alias:"class-name"},key:{pattern:RegExp(a(/(^[\t ]*|[{,]\s*)__(?:\s*\.\s*__)*(?=\s*=)/.source),"m"),lookbehind:!0,greedy:!0,alias:"property"},string:{pattern:/"""(?:\\[\s\S]|[^\\])*?"""|'''[\s\S]*?'''|'[^'\n\r]*'|"(?:\\.|[^\\"\r\n])*"/,greedy:!0},date:[{pattern:/\b\d{4}-\d{2}-\d{2}(?:[T\s]\d{2}:\d{2}:\d{2}(?:\.\d+)?(?:Z|[+-]\d{2}:\d{2})?)?\b/i,alias:"number"},{pattern:/\b\d{2}:\d{2}:\d{2}(?:\.\d+)?\b/,alias:"number"}],number:/(?:\b0(?:x[\da-zA-Z]+(?:_[\da-zA-Z]+)*|o[0-7]+(?:_[0-7]+)*|b[10]+(?:_[10]+)*))\b|[-+]?\b\d+(?:_\d+)*(?:\.\d+(?:_\d+)*)?(?:[eE][+-]?\d+(?:_\d+)*)?\b|[-+]?\b(?:inf|nan)\b/,boolean:/\b(?:false|true)\b/,punctuation:/[.,=[\]{}]/}})(t)}return Mw}var Fw,rP;function KSe(){if(rP)return Fw;rP=1,Fw=e,e.displayName="tremor",e.aliases=[];function e(t){(function(n){n.languages.tremor={comment:{pattern:/(^|[^\\])(?:\/\*[\s\S]*?\*\/|(?:--|\/\/|#).*)/,lookbehind:!0},"interpolated-string":null,extractor:{pattern:/\b[a-z_]\w*\|(?:[^\r\n\\|]|\\(?:\r\n|[\s\S]))*\|/i,greedy:!0,inside:{regex:{pattern:/(^re)\|[\s\S]+/,lookbehind:!0},function:/^\w+/,value:/\|[\s\S]+/}},identifier:{pattern:/`[^`]*`/,greedy:!0},function:/\b[a-z_]\w*(?=\s*(?:::\s*<|\())\b/,keyword:/\b(?:args|as|by|case|config|connect|connector|const|copy|create|default|define|deploy|drop|each|emit|end|erase|event|flow|fn|for|from|group|having|insert|into|intrinsic|let|links|match|merge|mod|move|of|operator|patch|pipeline|recur|script|select|set|sliding|state|stream|to|tumbling|update|use|when|where|window|with)\b/,boolean:/\b(?:false|null|true)\b/i,number:/\b(?:0b[01_]*|0x[0-9a-fA-F_]*|\d[\d_]*(?:\.\d[\d_]*)?(?:[Ee][+-]?[\d_]+)?)\b/,"pattern-punctuation":{pattern:/%(?=[({[])/,alias:"punctuation"},operator:/[-+*\/%~!^]=?|=[=>]?|&[&=]?|\|[|=]?|<<?=?|>>?>?=?|(?:absent|and|not|or|present|xor)\b/,punctuation:/::|[;\[\]()\{\},.:]/};var r=/#\{(?:[^"{}]|\{[^{}]*\}|"(?:[^"\\\r\n]|\\(?:\r\n|[\s\S]))*")*\}/.source;n.languages.tremor["interpolated-string"]={pattern:RegExp(/(^|[^\\])/.source+'(?:"""(?:'+/[^"\\#]|\\[\s\S]|"(?!"")|#(?!\{)/.source+"|"+r+')*"""|"(?:'+/[^"\\\r\n#]|\\(?:\r\n|[\s\S])|#(?!\{)/.source+"|"+r+')*")'),lookbehind:!0,greedy:!0,inside:{interpolation:{pattern:RegExp(r),inside:{punctuation:/^#\{|\}$/,expression:{pattern:/[\s\S]+/,inside:n.languages.tremor}}},string:/[\s\S]+/}},n.languages.troy=n.languages.tremor,n.languages.trickle=n.languages.tremor})(t)}return Fw}var Pw,aP;function XSe(){if(aP)return Pw;aP=1;var e=qH(),t=fA();Pw=n,n.displayName="tsx",n.aliases=[];function n(r){r.register(e),r.register(t),function(a){var o=a.util.clone(a.languages.typescript);a.languages.tsx=a.languages.extend("jsx",o),delete a.languages.tsx.parameter,delete a.languages.tsx["literal-property"];var s=a.languages.tsx.tag;s.pattern=RegExp(/(^|[^\w$]|(?=<\/))/.source+"(?:"+s.pattern.source+")",s.pattern.flags),s.lookbehind=!0}(r)}return Pw}var zw,oP;function ZSe(){if(oP)return zw;oP=1;var e=Wn();zw=t,t.displayName="tt2",t.aliases=[];function t(n){n.register(e),function(r){r.languages.tt2=r.languages.extend("clike",{comment:/#.*|\[%#[\s\S]*?%\]/,keyword:/\b(?:BLOCK|CALL|CASE|CATCH|CLEAR|DEBUG|DEFAULT|ELSE|ELSIF|END|FILTER|FINAL|FOREACH|GET|IF|IN|INCLUDE|INSERT|LAST|MACRO|META|NEXT|PERL|PROCESS|RAWPERL|RETURN|SET|STOP|SWITCH|TAGS|THROW|TRY|UNLESS|USE|WHILE|WRAPPER)\b/,punctuation:/[[\]{},()]/}),r.languages.insertBefore("tt2","number",{operator:/=[>=]?|!=?|<=?|>=?|&&|\|\|?|\b(?:and|not|or)\b/,variable:{pattern:/\b[a-z]\w*(?:\s*\.\s*(?:\d+|\$?[a-z]\w*))*\b/i}}),r.languages.insertBefore("tt2","keyword",{delimiter:{pattern:/^(?:\[%|%%)-?|-?%\]$/,alias:"punctuation"}}),r.languages.insertBefore("tt2","string",{"single-quoted-string":{pattern:/'[^\\']*(?:\\[\s\S][^\\']*)*'/,greedy:!0,alias:"string"},"double-quoted-string":{pattern:/"[^\\"]*(?:\\[\s\S][^\\"]*)*"/,greedy:!0,alias:"string",inside:{variable:{pattern:/\$(?:[a-z]\w*(?:\.(?:\d+|\$?[a-z]\w*))*)/i}}}}),delete r.languages.tt2.string,r.hooks.add("before-tokenize",function(a){var o=/\[%[\s\S]+?%\]/g;r.languages["markup-templating"].buildPlaceholders(a,"tt2",o)}),r.hooks.add("after-tokenize",function(a){r.languages["markup-templating"].tokenizePlaceholders(a,"tt2")})}(n)}return zw}var Bw,iP;function QSe(){if(iP)return Bw;iP=1;var e=Wn();Bw=t,t.displayName="twig",t.aliases=[];function t(n){n.register(e),n.languages.twig={comment:/^\{#[\s\S]*?#\}$/,"tag-name":{pattern:/(^\{%-?\s*)\w+/,lookbehind:!0,alias:"keyword"},delimiter:{pattern:/^\{[{%]-?|-?[%}]\}$/,alias:"punctuation"},string:{pattern:/("|')(?:\\.|(?!\1)[^\\\r\n])*\1/,inside:{punctuation:/^['"]|['"]$/}},keyword:/\b(?:even|if|odd)\b/,boolean:/\b(?:false|null|true)\b/,number:/\b0x[\dA-Fa-f]+|(?:\b\d+(?:\.\d*)?|\B\.\d+)(?:[Ee][-+]?\d+)?/,operator:[{pattern:/(\s)(?:and|b-and|b-or|b-xor|ends with|in|is|matches|not|or|same as|starts with)(?=\s)/,lookbehind:!0},/[=<>]=?|!=|\*\*?|\/\/?|\?:?|[-+~%|]/],punctuation:/[()\[\]{}:.,]/},n.hooks.add("before-tokenize",function(r){if(r.language==="twig"){var a=/\{(?:#[\s\S]*?#|%[\s\S]*?%|\{[\s\S]*?\})\}/g;n.languages["markup-templating"].buildPlaceholders(r,"twig",a)}}),n.hooks.add("after-tokenize",function(r){n.languages["markup-templating"].tokenizePlaceholders(r,"twig")})}return Bw}var Uw,sP;function JSe(){if(sP)return Uw;sP=1,Uw=e,e.displayName="typoscript",e.aliases=["tsconfig"];function e(t){(function(n){var r=/\b(?:ACT|ACTIFSUB|CARRAY|CASE|CLEARGIF|COA|COA_INT|CONSTANTS|CONTENT|CUR|EDITPANEL|EFFECT|EXT|FILE|FLUIDTEMPLATE|FORM|FRAME|FRAMESET|GIFBUILDER|GMENU|GMENU_FOLDOUT|GMENU_LAYERS|GP|HMENU|HRULER|HTML|IENV|IFSUB|IMAGE|IMGMENU|IMGMENUITEM|IMGTEXT|IMG_RESOURCE|INCLUDE_TYPOSCRIPT|JSMENU|JSMENUITEM|LLL|LOAD_REGISTER|NO|PAGE|RECORDS|RESTORE_REGISTER|TEMPLATE|TEXT|TMENU|TMENUITEM|TMENU_LAYERS|USER|USER_INT|_GIFBUILDER|global|globalString|globalVar)\b/;n.languages.typoscript={comment:[{pattern:/(^|[^\\])\/\*[\s\S]*?(?:\*\/|$)/,lookbehind:!0},{pattern:/(^|[^\\:= \t]|(?:^|[^= \t])[ \t]+)\/\/.*/,lookbehind:!0,greedy:!0},{pattern:/(^|[^"'])#.*/,lookbehind:!0,greedy:!0}],function:[{pattern:/<INCLUDE_TYPOSCRIPT:\s*source\s*=\s*(?:"[^"\r\n]*"|'[^'\r\n]*')\s*>/,inside:{string:{pattern:/"[^"\r\n]*"|'[^'\r\n]*'/,inside:{keyword:r}},keyword:{pattern:/INCLUDE_TYPOSCRIPT/}}},{pattern:/@import\s*(?:"[^"\r\n]*"|'[^'\r\n]*')/,inside:{string:/"[^"\r\n]*"|'[^'\r\n]*'/}}],string:{pattern:/^([^=]*=[< ]?)(?:(?!\]\n).)*/,lookbehind:!0,inside:{function:/\{\$.*\}/,keyword:r,number:/^\d+$/,punctuation:/[,|:]/}},keyword:r,number:{pattern:/\b\d+\s*[.{=]/,inside:{operator:/[.{=]/}},tag:{pattern:/\.?[-\w\\]+\.?/,inside:{punctuation:/\./}},punctuation:/[{}[\];(),.:|]/,operator:/[<>]=?|[!=]=?=?|--?|\+\+?|&&?|\|\|?|[?*/~^%]/},n.languages.tsconfig=n.languages.typoscript})(t)}return Uw}var jw,lP;function eEe(){if(lP)return jw;lP=1,jw=e,e.displayName="unrealscript",e.aliases=["uc","uscript"];function e(t){t.languages.unrealscript={comment:/\/\/.*|\/\*[\s\S]*?\*\//,string:{pattern:/(["'])(?:\\(?:\r\n|[\s\S])|(?!\1)[^\\\r\n])*\1/,greedy:!0},category:{pattern:/(\b(?:(?:autoexpand|hide|show)categories|var)\s*\()[^()]+(?=\))/,lookbehind:!0,greedy:!0,alias:"property"},metadata:{pattern:/(\w\s*)<\s*\w+\s*=[^<>|=\r\n]+(?:\|\s*\w+\s*=[^<>|=\r\n]+)*>/,lookbehind:!0,greedy:!0,inside:{property:/\b\w+(?=\s*=)/,operator:/=/,punctuation:/[<>|]/}},macro:{pattern:/`\w+/,alias:"property"},"class-name":{pattern:/(\b(?:class|enum|extends|interface|state(?:\(\))?|struct|within)\s+)\w+/,lookbehind:!0},keyword:/\b(?:abstract|actor|array|auto|autoexpandcategories|bool|break|byte|case|class|classgroup|client|coerce|collapsecategories|config|const|continue|default|defaultproperties|delegate|dependson|deprecated|do|dontcollapsecategories|editconst|editinlinenew|else|enum|event|exec|export|extends|final|float|for|forcescriptorder|foreach|function|goto|guid|hidecategories|hidedropdown|if|ignores|implements|inherits|input|int|interface|iterator|latent|local|material|name|native|nativereplication|noexport|nontransient|noteditinlinenew|notplaceable|operator|optional|out|pawn|perobjectconfig|perobjectlocalized|placeable|postoperator|preoperator|private|protected|reliable|replication|return|server|showcategories|simulated|singular|state|static|string|struct|structdefault|structdefaultproperties|switch|texture|transient|travel|unreliable|until|var|vector|while|within)\b/,function:/\b[a-z_]\w*(?=\s*\()/i,boolean:/\b(?:false|true)\b/,number:/\b0x[\da-f]+\b|(?:\b\d+(?:\.\d*)?|\B\.\d+)(?:e[+-]?\d+)?/i,operator:/>>|<<|--|\+\+|\*\*|[-+*/~!=<>$@]=?|&&?|\|\|?|\^\^?|[?:%]|\b(?:ClockwiseFrom|Cross|Dot)\b/,punctuation:/[()[\]{};,.]/},t.languages.uc=t.languages.uscript=t.languages.unrealscript}return jw}var Gw,uP;function tEe(){if(uP)return Gw;uP=1,Gw=e,e.displayName="uorazor",e.aliases=[];function e(t){t.languages.uorazor={"comment-hash":{pattern:/#.*/,alias:"comment",greedy:!0},"comment-slash":{pattern:/\/\/.*/,alias:"comment",greedy:!0},string:{pattern:/("|')(?:\\.|(?!\1)[^\\\r\n])*\1/,inside:{punctuation:/^['"]|['"]$/},greedy:!0},"source-layers":{pattern:/\b(?:arms|backpack|blue|bracelet|cancel|clear|cloak|criminal|earrings|enemy|facialhair|friend|friendly|gloves|gray|grey|ground|hair|head|innerlegs|innertorso|innocent|lefthand|middletorso|murderer|neck|nonfriendly|onehandedsecondary|outerlegs|outertorso|pants|red|righthand|ring|self|shirt|shoes|talisman|waist)\b/i,alias:"function"},"source-commands":{pattern:/\b(?:alliance|attack|cast|clearall|clearignore|clearjournal|clearlist|clearsysmsg|createlist|createtimer|dclick|dclicktype|dclickvar|dress|dressconfig|drop|droprelloc|emote|getlabel|guild|gumpclose|gumpresponse|hotkey|ignore|lasttarget|lift|lifttype|menu|menuresponse|msg|org|organize|organizer|overhead|pause|poplist|potion|promptresponse|pushlist|removelist|removetimer|rename|restock|say|scav|scavenger|script|setability|setlasttarget|setskill|settimer|setvar|sysmsg|target|targetloc|targetrelloc|targettype|undress|unignore|unsetvar|useobject|useonce|useskill|usetype|virtue|wait|waitforgump|waitformenu|waitforprompt|waitforstat|waitforsysmsg|waitfortarget|walk|wfsysmsg|wft|whisper|yell)\b/,alias:"function"},"tag-name":{pattern:/(^\{%-?\s*)\w+/,lookbehind:!0,alias:"keyword"},delimiter:{pattern:/^\{[{%]-?|-?[%}]\}$/,alias:"punctuation"},function:/\b(?:atlist|close|closest|count|counter|counttype|dead|dex|diffhits|diffmana|diffstam|diffweight|find|findbuff|finddebuff|findlayer|findtype|findtypelist|followers|gumpexists|hidden|hits|hp|hue|human|humanoid|ingump|inlist|insysmessage|insysmsg|int|invul|lhandempty|list|listexists|mana|maxhits|maxhp|maxmana|maxstam|maxweight|monster|mounted|name|next|noto|paralyzed|poisoned|position|prev|previous|queued|rand|random|rhandempty|skill|stam|str|targetexists|timer|timerexists|varexist|warmode|weight)\b/,keyword:/\b(?:and|as|break|continue|else|elseif|endfor|endif|endwhile|for|if|loop|not|or|replay|stop|while)\b/,boolean:/\b(?:false|null|true)\b/,number:/\b0x[\dA-Fa-f]+|(?:\b\d+(?:\.\d*)?|\B\.\d+)(?:[Ee][-+]?\d+)?/,operator:[{pattern:/(\s)(?:and|b-and|b-or|b-xor|ends with|in|is|matches|not|or|same as|starts with)(?=\s)/,lookbehind:!0},/[=<>]=?|!=|\*\*?|\/\/?|\?:?|[-+~%|]/],punctuation:/[()\[\]{}:.,]/}}return Gw}var Hw,cP;function nEe(){if(cP)return Hw;cP=1,Hw=e,e.displayName="uri",e.aliases=["url"];function e(t){t.languages.uri={scheme:{pattern:/^[a-z][a-z0-9+.-]*:/im,greedy:!0,inside:{"scheme-delimiter":/:$/}},fragment:{pattern:/#[\w\-.~!$&'()*+,;=%:@/?]*/,inside:{"fragment-delimiter":/^#/}},query:{pattern:/\?[\w\-.~!$&'()*+,;=%:@/?]*/,inside:{"query-delimiter":{pattern:/^\?/,greedy:!0},"pair-delimiter":/[&;]/,pair:{pattern:/^[^=][\s\S]*/,inside:{key:/^[^=]+/,value:{pattern:/(^=)[\s\S]+/,lookbehind:!0}}}}},authority:{pattern:RegExp(/^\/\//.source+/(?:[\w\-.~!$&'()*+,;=%:]*@)?/.source+("(?:"+/\[(?:[0-9a-fA-F:.]{2,48}|v[0-9a-fA-F]+\.[\w\-.~!$&'()*+,;=]+)\]/.source+"|"+/[\w\-.~!$&'()*+,;=%]*/.source+")")+/(?::\d*)?/.source,"m"),inside:{"authority-delimiter":/^\/\//,"user-info-segment":{pattern:/^[\w\-.~!$&'()*+,;=%:]*@/,inside:{"user-info-delimiter":/@$/,"user-info":/^[\w\-.~!$&'()*+,;=%:]+/}},"port-segment":{pattern:/:\d*$/,inside:{"port-delimiter":/^:/,port:/^\d+/}},host:{pattern:/[\s\S]+/,inside:{"ip-literal":{pattern:/^\[[\s\S]+\]$/,inside:{"ip-literal-delimiter":/^\[|\]$/,"ipv-future":/^v[\s\S]+/,"ipv6-address":/^[\s\S]+/}},"ipv4-address":/^(?:(?:[03-9]\d?|[12]\d{0,2})\.){3}(?:[03-9]\d?|[12]\d{0,2})$/}}}},path:{pattern:/^[\w\-.~!$&'()*+,;=%:@/]+/m,inside:{"path-separator":/\//}}},t.languages.url=t.languages.uri}return Hw}var $w,dP;function rEe(){if(dP)return $w;dP=1,$w=e,e.displayName="v",e.aliases=[];function e(t){(function(n){var r={pattern:/[\s\S]+/,inside:null};n.languages.v=n.languages.extend("clike",{string:{pattern:/r?(["'])(?:\\(?:\r\n|[\s\S])|(?!\1)[^\\\r\n])*\1/,alias:"quoted-string",greedy:!0,inside:{interpolation:{pattern:/((?:^|[^\\])(?:\\{2})*)\$(?:\{[^{}]*\}|\w+(?:\.\w+(?:\([^\(\)]*\))?|\[[^\[\]]+\])*)/,lookbehind:!0,inside:{"interpolation-variable":{pattern:/^\$\w[\s\S]*$/,alias:"variable"},"interpolation-punctuation":{pattern:/^\$\{|\}$/,alias:"punctuation"},"interpolation-expression":r}}}},"class-name":{pattern:/(\b(?:enum|interface|struct|type)\s+)(?:C\.)?\w+/,lookbehind:!0},keyword:/(?:\b(?:__global|as|asm|assert|atomic|break|chan|const|continue|defer|else|embed|enum|fn|for|go(?:to)?|if|import|in|interface|is|lock|match|module|mut|none|or|pub|return|rlock|select|shared|sizeof|static|struct|type(?:of)?|union|unsafe)|\$(?:else|for|if)|#(?:flag|include))\b/,number:/\b(?:0x[a-f\d]+(?:_[a-f\d]+)*|0b[01]+(?:_[01]+)*|0o[0-7]+(?:_[0-7]+)*|\d+(?:_\d+)*(?:\.\d+(?:_\d+)*)?)\b/i,operator:/~|\?|[*\/%^!=]=?|\+[=+]?|-[=-]?|\|[=|]?|&(?:=|&|\^=?)?|>(?:>=?|=)?|<(?:<=?|=|-)?|:=|\.\.\.?/,builtin:/\b(?:any(?:_float|_int)?|bool|byte(?:ptr)?|charptr|f(?:32|64)|i(?:8|16|64|128|nt)|rune|size_t|string|u(?:16|32|64|128)|voidptr)\b/}),r.inside=n.languages.v,n.languages.insertBefore("v","string",{char:{pattern:/`(?:\\`|\\?[^`]{1,2})`/,alias:"rune"}}),n.languages.insertBefore("v","operator",{attribute:{pattern:/(^[\t ]*)\[(?:deprecated|direct_array_access|flag|inline|live|ref_only|typedef|unsafe_fn|windows_stdcall)\]/m,lookbehind:!0,alias:"annotation",inside:{punctuation:/[\[\]]/,keyword:/\w+/}},generic:{pattern:/<\w+>(?=\s*[\)\{])/,inside:{punctuation:/[<>]/,"class-name":/\w+/}}}),n.languages.insertBefore("v","function",{"generic-function":{pattern:/\b\w+\s*<\w+>(?=\()/,inside:{function:/^\w+/,generic:{pattern:/<\w+>/,inside:n.languages.v.generic.inside}}}})})(t)}return $w}var qw,fP;function aEe(){if(fP)return qw;fP=1,qw=e,e.displayName="vala",e.aliases=[];function e(t){t.languages.vala=t.languages.extend("clike",{"class-name":[{pattern:/\b[A-Z]\w*(?:\.\w+)*\b(?=(?:\?\s+|\*?\s+\*?)\w)/,inside:{punctuation:/\./}},{pattern:/(\[)[A-Z]\w*(?:\.\w+)*\b/,lookbehind:!0,inside:{punctuation:/\./}},{pattern:/(\b(?:class|interface)\s+[A-Z]\w*(?:\.\w+)*\s*:\s*)[A-Z]\w*(?:\.\w+)*\b/,lookbehind:!0,inside:{punctuation:/\./}},{pattern:/((?:\b(?:class|enum|interface|new|struct)\s+)|(?:catch\s+\())[A-Z]\w*(?:\.\w+)*\b/,lookbehind:!0,inside:{punctuation:/\./}}],keyword:/\b(?:abstract|as|assert|async|base|bool|break|case|catch|char|class|const|construct|continue|default|delegate|delete|do|double|dynamic|else|ensures|enum|errordomain|extern|finally|float|for|foreach|get|if|in|inline|int|int16|int32|int64|int8|interface|internal|is|lock|long|namespace|new|null|out|override|owned|params|private|protected|public|ref|requires|return|set|short|signal|sizeof|size_t|ssize_t|static|string|struct|switch|this|throw|throws|try|typeof|uchar|uint|uint16|uint32|uint64|uint8|ulong|unichar|unowned|ushort|using|value|var|virtual|void|volatile|weak|while|yield)\b/i,function:/\b\w+(?=\s*\()/,number:/(?:\b0x[\da-f]+\b|(?:\b\d+(?:\.\d*)?|\B\.\d+)(?:e[+-]?\d+)?)(?:f|u?l?)?/i,operator:/\+\+|--|&&|\|\||<<=?|>>=?|=>|->|~|[+\-*\/%&^|=!<>]=?|\?\??|\.\.\./,punctuation:/[{}[\];(),.:]/,constant:/\b[A-Z0-9_]+\b/}),t.languages.insertBefore("vala","string",{"raw-string":{pattern:/"""[\s\S]*?"""/,greedy:!0,alias:"string"},"template-string":{pattern:/@"[\s\S]*?"/,greedy:!0,inside:{interpolation:{pattern:/\$(?:\([^)]*\)|[a-zA-Z]\w*)/,inside:{delimiter:{pattern:/^\$\(?|\)$/,alias:"punctuation"},rest:t.languages.vala}},string:/[\s\S]+/}}}),t.languages.insertBefore("vala","keyword",{regex:{pattern:/\/(?:\[(?:[^\]\\\r\n]|\\.)*\]|\\.|[^/\\\[\r\n])+\/[imsx]{0,4}(?=\s*(?:$|[\r\n,.;})\]]))/,greedy:!0,inside:{"regex-source":{pattern:/^(\/)[\s\S]+(?=\/[a-z]*$)/,lookbehind:!0,alias:"language-regex",inside:t.languages.regex},"regex-delimiter":/^\//,"regex-flags":/^[a-z]+$/}}})}return qw}var Vw,pP;function oEe(){if(pP)return Vw;pP=1,Vw=e,e.displayName="velocity",e.aliases=[];function e(t){(function(n){n.languages.velocity=n.languages.extend("markup",{});var r={variable:{pattern:/(^|[^\\](?:\\\\)*)\$!?(?:[a-z][\w-]*(?:\([^)]*\))?(?:\.[a-z][\w-]*(?:\([^)]*\))?|\[[^\]]+\])*|\{[^}]+\})/i,lookbehind:!0,inside:{}},string:{pattern:/"[^"]*"|'[^']*'/,greedy:!0},number:/\b\d+\b/,boolean:/\b(?:false|true)\b/,operator:/[=!<>]=?|[+*/%-]|&&|\|\||\.\.|\b(?:eq|g[et]|l[et]|n(?:e|ot))\b/,punctuation:/[(){}[\]:,.]/};r.variable.inside={string:r.string,function:{pattern:/([^\w-])[a-z][\w-]*(?=\()/,lookbehind:!0},number:r.number,boolean:r.boolean,punctuation:r.punctuation},n.languages.insertBefore("velocity","comment",{unparsed:{pattern:/(^|[^\\])#\[\[[\s\S]*?\]\]#/,lookbehind:!0,greedy:!0,inside:{punctuation:/^#\[\[|\]\]#$/}},"velocity-comment":[{pattern:/(^|[^\\])#\*[\s\S]*?\*#/,lookbehind:!0,greedy:!0,alias:"comment"},{pattern:/(^|[^\\])##.*/,lookbehind:!0,greedy:!0,alias:"comment"}],directive:{pattern:/(^|[^\\](?:\\\\)*)#@?(?:[a-z][\w-]*|\{[a-z][\w-]*\})(?:\s*\((?:[^()]|\([^()]*\))*\))?/i,lookbehind:!0,inside:{keyword:{pattern:/^#@?(?:[a-z][\w-]*|\{[a-z][\w-]*\})|\bin\b/,inside:{punctuation:/[{}]/}},rest:r}},variable:r.variable}),n.languages.velocity.tag.inside["attr-value"].inside.rest=n.languages.velocity})(t)}return Vw}var Ww,gP;function iEe(){if(gP)return Ww;gP=1,Ww=e,e.displayName="verilog",e.aliases=[];function e(t){t.languages.verilog={comment:{pattern:/\/\/.*|\/\*[\s\S]*?\*\//,greedy:!0},string:{pattern:/"(?:\\(?:\r\n|[\s\S])|[^"\\\r\n])*"/,greedy:!0},"kernel-function":{pattern:/\B\$\w+\b/,alias:"property"},constant:/\B`\w+\b/,function:/\b\w+(?=\()/,keyword:/\b(?:alias|and|assert|assign|assume|automatic|before|begin|bind|bins|binsof|bit|break|buf|bufif0|bufif1|byte|case|casex|casez|cell|chandle|class|clocking|cmos|config|const|constraint|context|continue|cover|covergroup|coverpoint|cross|deassign|default|defparam|design|disable|dist|do|edge|else|end|endcase|endclass|endclocking|endconfig|endfunction|endgenerate|endgroup|endinterface|endmodule|endpackage|endprimitive|endprogram|endproperty|endsequence|endspecify|endtable|endtask|enum|event|expect|export|extends|extern|final|first_match|for|force|foreach|forever|fork|forkjoin|function|generate|genvar|highz0|highz1|if|iff|ifnone|ignore_bins|illegal_bins|import|incdir|include|initial|inout|input|inside|instance|int|integer|interface|intersect|join|join_any|join_none|large|liblist|library|local|localparam|logic|longint|macromodule|matches|medium|modport|module|nand|negedge|new|nmos|nor|noshowcancelled|not|notif0|notif1|null|or|output|package|packed|parameter|pmos|posedge|primitive|priority|program|property|protected|pull0|pull1|pulldown|pullup|pulsestyle_ondetect|pulsestyle_onevent|pure|rand|randc|randcase|randsequence|rcmos|real|realtime|ref|reg|release|repeat|return|rnmos|rpmos|rtran|rtranif0|rtranif1|scalared|sequence|shortint|shortreal|showcancelled|signed|small|solve|specify|specparam|static|string|strong0|strong1|struct|super|supply0|supply1|table|tagged|task|this|throughout|time|timeprecision|timeunit|tran|tranif0|tranif1|tri|tri0|tri1|triand|trior|trireg|type|typedef|union|unique|unsigned|use|uwire|var|vectored|virtual|void|wait|wait_order|wand|weak0|weak1|while|wildcard|wire|with|within|wor|xnor|xor)\b/,important:/\b(?:always|always_comb|always_ff|always_latch)\b(?: *@)?/,number:/\B##?\d+|(?:\b\d+)?'[odbh] ?[\da-fzx_?]+|\b(?:\d*[._])?\d+(?:e[-+]?\d+)?/i,operator:/[-+{}^~%*\/?=!<>&|]+/,punctuation:/[[\];(),.:]/}}return Ww}var Yw,hP;function sEe(){if(hP)return Yw;hP=1,Yw=e,e.displayName="vhdl",e.aliases=[];function e(t){t.languages.vhdl={comment:/--.+/,"vhdl-vectors":{pattern:/\b[oxb]"[\da-f_]+"|"[01uxzwlh-]+"/i,alias:"number"},"quoted-function":{pattern:/"\S+?"(?=\()/,alias:"function"},string:/"(?:[^\\"\r\n]|\\(?:\r\n|[\s\S]))*"/,constant:/\b(?:library|use)\b/i,keyword:/\b(?:'active|'ascending|'base|'delayed|'driving|'driving_value|'event|'high|'image|'instance_name|'last_active|'last_event|'last_value|'left|'leftof|'length|'low|'path_name|'pos|'pred|'quiet|'range|'reverse_range|'right|'rightof|'simple_name|'stable|'succ|'transaction|'val|'value|access|after|alias|all|architecture|array|assert|attribute|begin|block|body|buffer|bus|case|component|configuration|constant|disconnect|downto|else|elsif|end|entity|exit|file|for|function|generate|generic|group|guarded|if|impure|in|inertial|inout|is|label|library|linkage|literal|loop|map|new|next|null|of|on|open|others|out|package|port|postponed|procedure|process|pure|range|record|register|reject|report|return|select|severity|shared|signal|subtype|then|to|transport|type|unaffected|units|until|use|variable|wait|when|while|with)\b/i,boolean:/\b(?:false|true)\b/i,function:/\w+(?=\()/,number:/'[01uxzwlh-]'|\b(?:\d+#[\da-f_.]+#|\d[\d_.]*)(?:e[-+]?\d+)?/i,operator:/[<>]=?|:=|[-+*/&=]|\b(?:abs|and|mod|nand|nor|not|or|rem|rol|ror|sla|sll|sra|srl|xnor|xor)\b/i,punctuation:/[{}[\];(),.:]/}}return Yw}var Kw,mP;function lEe(){if(mP)return Kw;mP=1,Kw=e,e.displayName="vim",e.aliases=[];function e(t){t.languages.vim={string:/"(?:[^"\\\r\n]|\\.)*"|'(?:[^'\r\n]|'')*'/,comment:/".*/,function:/\b\w+(?=\()/,keyword:/\b(?:N|Next|P|Print|X|XMLent|XMLns|ab|abbreviate|abc|abclear|abo|aboveleft|al|all|ar|arga|argadd|argd|argdelete|argdo|arge|argedit|argg|argglobal|argl|arglocal|args|argu|argument|as|ascii|b|bN|bNext|ba|bad|badd|ball|bd|bdelete|be|bel|belowright|bf|bfirst|bl|blast|bm|bmodified|bn|bnext|bo|botright|bp|bprevious|br|brea|break|breaka|breakadd|breakd|breakdel|breakl|breaklist|brewind|bro|browse|bufdo|buffer|buffers|bun|bunload|bw|bwipeout|c|cN|cNext|cNfcNfile|ca|cabbrev|cabc|cabclear|cad|caddb|caddbuffer|caddexpr|caddf|caddfile|cal|call|cat|catch|cb|cbuffer|cc|ccl|cclose|cd|ce|center|cex|cexpr|cf|cfile|cfir|cfirst|cg|cgetb|cgetbuffer|cgete|cgetexpr|cgetfile|change|changes|chd|chdir|che|checkpath|checkt|checktime|cl|cla|clast|clist|clo|close|cmapc|cmapclear|cn|cnew|cnewer|cnext|cnf|cnfile|cnorea|cnoreabbrev|co|col|colder|colo|colorscheme|comc|comclear|comp|compiler|con|conf|confirm|continue|cope|copen|copy|cp|cpf|cpfile|cprevious|cq|cquit|cr|crewind|cu|cuna|cunabbrev|cunmap|cw|cwindow|d|debugg|debuggreedy|delc|delcommand|delete|delf|delfunction|delm|delmarks|di|diffg|diffget|diffoff|diffpatch|diffpu|diffput|diffsplit|diffthis|diffu|diffupdate|dig|digraphs|display|dj|djump|dl|dlist|dr|drop|ds|dsearch|dsp|dsplit|e|earlier|echoe|echoerr|echom|echomsg|echon|edit|el|else|elsei|elseif|em|emenu|en|endf|endfo|endfor|endfun|endfunction|endif|endt|endtry|endw|endwhile|ene|enew|ex|exi|exit|exu|exusage|f|file|files|filetype|fin|fina|finally|find|fini|finish|fir|first|fix|fixdel|fo|fold|foldc|foldclose|foldd|folddoc|folddoclosed|folddoopen|foldo|foldopen|for|fu|fun|function|go|goto|gr|grep|grepa|grepadd|h|ha|hardcopy|help|helpf|helpfind|helpg|helpgrep|helpt|helptags|hid|hide|his|history|ia|iabbrev|iabc|iabclear|if|ij|ijump|il|ilist|imapc|imapclear|in|inorea|inoreabbrev|isearch|isp|isplit|iu|iuna|iunabbrev|iunmap|j|join|ju|jumps|k|kee|keepalt|keepj|keepjumps|keepmarks|l|lN|lNext|lNf|lNfile|la|lad|laddb|laddbuffer|laddexpr|laddf|laddfile|lan|language|last|later|lb|lbuffer|lc|lcd|lch|lchdir|lcl|lclose|left|lefta|leftabove|let|lex|lexpr|lf|lfile|lfir|lfirst|lg|lgetb|lgetbuffer|lgete|lgetexpr|lgetfile|lgr|lgrep|lgrepa|lgrepadd|lh|lhelpgrep|list|ll|lla|llast|lli|llist|lm|lmak|lmake|lmap|lmapc|lmapclear|ln|lne|lnew|lnewer|lnext|lnf|lnfile|lnoremap|lo|loadview|loc|lockmarks|lockv|lockvar|lol|lolder|lop|lopen|lp|lpf|lpfile|lprevious|lr|lrewind|ls|lt|ltag|lu|lunmap|lv|lvimgrep|lvimgrepa|lvimgrepadd|lw|lwindow|m|ma|mak|make|mark|marks|mat|match|menut|menutranslate|mk|mkexrc|mks|mksession|mksp|mkspell|mkv|mkvie|mkview|mkvimrc|mod|mode|move|mz|mzf|mzfile|mzscheme|n|nbkey|new|next|nmapc|nmapclear|noh|nohlsearch|norea|noreabbrev|nu|number|nun|nunmap|o|omapc|omapclear|on|only|open|opt|options|ou|ounmap|p|pc|pclose|pe|ped|pedit|perl|perld|perldo|po|pop|popu|popup|pp|ppop|pre|preserve|prev|previous|print|prof|profd|profdel|profile|promptf|promptfind|promptr|promptrepl|ps|psearch|ptN|ptNext|pta|ptag|ptf|ptfirst|ptj|ptjump|ptl|ptlast|ptn|ptnext|ptp|ptprevious|ptr|ptrewind|pts|ptselect|pu|put|pw|pwd|py|pyf|pyfile|python|q|qa|qall|quit|quita|quitall|r|read|rec|recover|red|redi|redir|redo|redr|redraw|redraws|redrawstatus|reg|registers|res|resize|ret|retab|retu|return|rew|rewind|ri|right|rightb|rightbelow|ru|rub|ruby|rubyd|rubydo|rubyf|rubyfile|runtime|rv|rviminfo|sN|sNext|sa|sal|sall|san|sandbox|sargument|sav|saveas|sb|sbN|sbNext|sba|sball|sbf|sbfirst|sbl|sblast|sbm|sbmodified|sbn|sbnext|sbp|sbprevious|sbr|sbrewind|sbuffer|scrip|scripte|scriptencoding|scriptnames|se|set|setf|setfiletype|setg|setglobal|setl|setlocal|sf|sfind|sfir|sfirst|sh|shell|sign|sil|silent|sim|simalt|sl|sla|slast|sleep|sm|smagic|smap|smapc|smapclear|sme|smenu|sn|snext|sni|sniff|sno|snomagic|snor|snoremap|snoreme|snoremenu|so|sor|sort|source|sp|spe|spelld|spelldump|spellgood|spelli|spellinfo|spellr|spellrepall|spellu|spellundo|spellw|spellwrong|split|spr|sprevious|sre|srewind|st|sta|stag|star|startg|startgreplace|startinsert|startr|startreplace|stj|stjump|stop|stopi|stopinsert|sts|stselect|sun|sunhide|sunm|sunmap|sus|suspend|sv|sview|syncbind|t|tN|tNext|ta|tab|tabN|tabNext|tabc|tabclose|tabd|tabdo|tabe|tabedit|tabf|tabfind|tabfir|tabfirst|tabl|tablast|tabm|tabmove|tabn|tabnew|tabnext|tabo|tabonly|tabp|tabprevious|tabr|tabrewind|tabs|tag|tags|tc|tcl|tcld|tcldo|tclf|tclfile|te|tearoff|tf|tfirst|th|throw|tj|tjump|tl|tlast|tm|tmenu|tn|tnext|to|topleft|tp|tprevious|tr|trewind|try|ts|tselect|tu|tunmenu|u|una|unabbreviate|undo|undoj|undojoin|undol|undolist|unh|unhide|unlet|unlo|unlockvar|unm|unmap|up|update|ve|verb|verbose|version|vert|vertical|vi|vie|view|vim|vimgrep|vimgrepa|vimgrepadd|visual|viu|viusage|vmapc|vmapclear|vne|vnew|vs|vsplit|vu|vunmap|w|wN|wNext|wa|wall|wh|while|win|winc|wincmd|windo|winp|winpos|winsize|wn|wnext|wp|wprevious|wq|wqa|wqall|write|ws|wsverb|wv|wviminfo|x|xa|xall|xit|xm|xmap|xmapc|xmapclear|xme|xmenu|xn|xnoremap|xnoreme|xnoremenu|xu|xunmap|y|yank)\b/,builtin:/\b(?:acd|ai|akm|aleph|allowrevins|altkeymap|ambiwidth|ambw|anti|antialias|arab|arabic|arabicshape|ari|arshape|autochdir|autocmd|autoindent|autoread|autowrite|autowriteall|aw|awa|background|backspace|backup|backupcopy|backupdir|backupext|backupskip|balloondelay|ballooneval|balloonexpr|bdir|bdlay|beval|bex|bexpr|bg|bh|bin|binary|biosk|bioskey|bk|bkc|bomb|breakat|brk|browsedir|bs|bsdir|bsk|bt|bufhidden|buflisted|buftype|casemap|ccv|cdpath|cedit|cfu|ch|charconvert|ci|cin|cindent|cink|cinkeys|cino|cinoptions|cinw|cinwords|clipboard|cmdheight|cmdwinheight|cmp|cms|columns|com|comments|commentstring|compatible|complete|completefunc|completeopt|consk|conskey|copyindent|cot|cpo|cpoptions|cpt|cscopepathcomp|cscopeprg|cscopequickfix|cscopetag|cscopetagorder|cscopeverbose|cspc|csprg|csqf|cst|csto|csverb|cuc|cul|cursorcolumn|cursorline|cwh|debug|deco|def|define|delcombine|dex|dg|dict|dictionary|diff|diffexpr|diffopt|digraph|dip|dir|directory|dy|ea|ead|eadirection|eb|ed|edcompatible|ef|efm|ei|ek|enc|encoding|endofline|eol|ep|equalalways|equalprg|errorbells|errorfile|errorformat|esckeys|et|eventignore|expandtab|exrc|fcl|fcs|fdc|fde|fdi|fdl|fdls|fdm|fdn|fdo|fdt|fen|fenc|fencs|fex|ff|ffs|fileencoding|fileencodings|fileformat|fileformats|fillchars|fk|fkmap|flp|fml|fmr|foldcolumn|foldenable|foldexpr|foldignore|foldlevel|foldlevelstart|foldmarker|foldmethod|foldminlines|foldnestmax|foldtext|formatexpr|formatlistpat|formatoptions|formatprg|fp|fs|fsync|ft|gcr|gd|gdefault|gfm|gfn|gfs|gfw|ghr|gp|grepformat|grepprg|gtl|gtt|guicursor|guifont|guifontset|guifontwide|guiheadroom|guioptions|guipty|guitablabel|guitabtooltip|helpfile|helpheight|helplang|hf|hh|hi|hidden|highlight|hk|hkmap|hkmapp|hkp|hl|hlg|hls|hlsearch|ic|icon|iconstring|ignorecase|im|imactivatekey|imak|imc|imcmdline|imd|imdisable|imi|iminsert|ims|imsearch|inc|include|includeexpr|incsearch|inde|indentexpr|indentkeys|indk|inex|inf|infercase|insertmode|invacd|invai|invakm|invallowrevins|invaltkeymap|invanti|invantialias|invar|invarab|invarabic|invarabicshape|invari|invarshape|invautochdir|invautoindent|invautoread|invautowrite|invautowriteall|invaw|invawa|invbackup|invballooneval|invbeval|invbin|invbinary|invbiosk|invbioskey|invbk|invbl|invbomb|invbuflisted|invcf|invci|invcin|invcindent|invcompatible|invconfirm|invconsk|invconskey|invcopyindent|invcp|invcscopetag|invcscopeverbose|invcst|invcsverb|invcuc|invcul|invcursorcolumn|invcursorline|invdeco|invdelcombine|invdg|invdiff|invdigraph|invdisable|invea|inveb|inved|invedcompatible|invek|invendofline|inveol|invequalalways|inverrorbells|invesckeys|invet|invex|invexpandtab|invexrc|invfen|invfk|invfkmap|invfoldenable|invgd|invgdefault|invguipty|invhid|invhidden|invhk|invhkmap|invhkmapp|invhkp|invhls|invhlsearch|invic|invicon|invignorecase|invim|invimc|invimcmdline|invimd|invincsearch|invinf|invinfercase|invinsertmode|invis|invjoinspaces|invjs|invlazyredraw|invlbr|invlinebreak|invlisp|invlist|invloadplugins|invlpl|invlz|invma|invmacatsui|invmagic|invmh|invml|invmod|invmodeline|invmodifiable|invmodified|invmore|invmousef|invmousefocus|invmousehide|invnu|invnumber|invodev|invopendevice|invpaste|invpi|invpreserveindent|invpreviewwindow|invprompt|invpvw|invreadonly|invremap|invrestorescreen|invrevins|invri|invrightleft|invrightleftcmd|invrl|invrlc|invro|invrs|invru|invruler|invsb|invsc|invscb|invscrollbind|invscs|invsecure|invsft|invshellslash|invshelltemp|invshiftround|invshortname|invshowcmd|invshowfulltag|invshowmatch|invshowmode|invsi|invsm|invsmartcase|invsmartindent|invsmarttab|invsmd|invsn|invsol|invspell|invsplitbelow|invsplitright|invspr|invsr|invssl|invsta|invstartofline|invstmp|invswapfile|invswf|invta|invtagbsearch|invtagrelative|invtagstack|invtbi|invtbidi|invtbs|invtermbidi|invterse|invtextauto|invtextmode|invtf|invtgst|invtildeop|invtimeout|invtitle|invto|invtop|invtr|invttimeout|invttybuiltin|invttyfast|invtx|invvb|invvisualbell|invwa|invwarn|invwb|invweirdinvert|invwfh|invwfw|invwildmenu|invwinfixheight|invwinfixwidth|invwiv|invwmnu|invwrap|invwrapscan|invwrite|invwriteany|invwritebackup|invws|isf|isfname|isi|isident|isk|iskeyword|isprint|joinspaces|js|key|keymap|keymodel|keywordprg|km|kmp|kp|langmap|langmenu|laststatus|lazyredraw|lbr|lcs|linebreak|lines|linespace|lisp|lispwords|listchars|loadplugins|lpl|lsp|lz|macatsui|magic|makeef|makeprg|matchpairs|matchtime|maxcombine|maxfuncdepth|maxmapdepth|maxmem|maxmempattern|maxmemtot|mco|mef|menuitems|mfd|mh|mis|mkspellmem|ml|mls|mm|mmd|mmp|mmt|modeline|modelines|modifiable|modified|more|mouse|mousef|mousefocus|mousehide|mousem|mousemodel|mouses|mouseshape|mouset|mousetime|mp|mps|msm|mzq|mzquantum|nf|noacd|noai|noakm|noallowrevins|noaltkeymap|noanti|noantialias|noar|noarab|noarabic|noarabicshape|noari|noarshape|noautochdir|noautoindent|noautoread|noautowrite|noautowriteall|noaw|noawa|nobackup|noballooneval|nobeval|nobin|nobinary|nobiosk|nobioskey|nobk|nobl|nobomb|nobuflisted|nocf|noci|nocin|nocindent|nocompatible|noconfirm|noconsk|noconskey|nocopyindent|nocp|nocscopetag|nocscopeverbose|nocst|nocsverb|nocuc|nocul|nocursorcolumn|nocursorline|nodeco|nodelcombine|nodg|nodiff|nodigraph|nodisable|noea|noeb|noed|noedcompatible|noek|noendofline|noeol|noequalalways|noerrorbells|noesckeys|noet|noex|noexpandtab|noexrc|nofen|nofk|nofkmap|nofoldenable|nogd|nogdefault|noguipty|nohid|nohidden|nohk|nohkmap|nohkmapp|nohkp|nohls|noic|noicon|noignorecase|noim|noimc|noimcmdline|noimd|noincsearch|noinf|noinfercase|noinsertmode|nois|nojoinspaces|nojs|nolazyredraw|nolbr|nolinebreak|nolisp|nolist|noloadplugins|nolpl|nolz|noma|nomacatsui|nomagic|nomh|noml|nomod|nomodeline|nomodifiable|nomodified|nomore|nomousef|nomousefocus|nomousehide|nonu|nonumber|noodev|noopendevice|nopaste|nopi|nopreserveindent|nopreviewwindow|noprompt|nopvw|noreadonly|noremap|norestorescreen|norevins|nori|norightleft|norightleftcmd|norl|norlc|noro|nors|noru|noruler|nosb|nosc|noscb|noscrollbind|noscs|nosecure|nosft|noshellslash|noshelltemp|noshiftround|noshortname|noshowcmd|noshowfulltag|noshowmatch|noshowmode|nosi|nosm|nosmartcase|nosmartindent|nosmarttab|nosmd|nosn|nosol|nospell|nosplitbelow|nosplitright|nospr|nosr|nossl|nosta|nostartofline|nostmp|noswapfile|noswf|nota|notagbsearch|notagrelative|notagstack|notbi|notbidi|notbs|notermbidi|noterse|notextauto|notextmode|notf|notgst|notildeop|notimeout|notitle|noto|notop|notr|nottimeout|nottybuiltin|nottyfast|notx|novb|novisualbell|nowa|nowarn|nowb|noweirdinvert|nowfh|nowfw|nowildmenu|nowinfixheight|nowinfixwidth|nowiv|nowmnu|nowrap|nowrapscan|nowrite|nowriteany|nowritebackup|nows|nrformats|numberwidth|nuw|odev|oft|ofu|omnifunc|opendevice|operatorfunc|opfunc|osfiletype|pa|para|paragraphs|paste|pastetoggle|patchexpr|patchmode|path|pdev|penc|pex|pexpr|pfn|ph|pheader|pi|pm|pmbcs|pmbfn|popt|preserveindent|previewheight|previewwindow|printdevice|printencoding|printexpr|printfont|printheader|printmbcharset|printmbfont|printoptions|prompt|pt|pumheight|pvh|pvw|qe|quoteescape|readonly|remap|report|restorescreen|revins|rightleft|rightleftcmd|rl|rlc|ro|rs|rtp|ruf|ruler|rulerformat|runtimepath|sbo|sc|scb|scr|scroll|scrollbind|scrolljump|scrolloff|scrollopt|scs|sect|sections|secure|sel|selection|selectmode|sessionoptions|sft|shcf|shellcmdflag|shellpipe|shellquote|shellredir|shellslash|shelltemp|shelltype|shellxquote|shiftround|shiftwidth|shm|shortmess|shortname|showbreak|showcmd|showfulltag|showmatch|showmode|showtabline|shq|si|sidescroll|sidescrolloff|siso|sj|slm|smartcase|smartindent|smarttab|smc|smd|softtabstop|sol|spc|spell|spellcapcheck|spellfile|spelllang|spellsuggest|spf|spl|splitbelow|splitright|sps|sr|srr|ss|ssl|ssop|stal|startofline|statusline|stl|stmp|su|sua|suffixes|suffixesadd|sw|swapfile|swapsync|swb|swf|switchbuf|sws|sxq|syn|synmaxcol|syntax|t_AB|t_AF|t_AL|t_CS|t_CV|t_Ce|t_Co|t_Cs|t_DL|t_EI|t_F1|t_F2|t_F3|t_F4|t_F5|t_F6|t_F7|t_F8|t_F9|t_IE|t_IS|t_K1|t_K3|t_K4|t_K5|t_K6|t_K7|t_K8|t_K9|t_KA|t_KB|t_KC|t_KD|t_KE|t_KF|t_KG|t_KH|t_KI|t_KJ|t_KK|t_KL|t_RI|t_RV|t_SI|t_Sb|t_Sf|t_WP|t_WS|t_ZH|t_ZR|t_al|t_bc|t_cd|t_ce|t_cl|t_cm|t_cs|t_da|t_db|t_dl|t_fs|t_k1|t_k2|t_k3|t_k4|t_k5|t_k6|t_k7|t_k8|t_k9|t_kB|t_kD|t_kI|t_kN|t_kP|t_kb|t_kd|t_ke|t_kh|t_kl|t_kr|t_ks|t_ku|t_le|t_mb|t_md|t_me|t_mr|t_ms|t_nd|t_op|t_se|t_so|t_sr|t_te|t_ti|t_ts|t_ue|t_us|t_ut|t_vb|t_ve|t_vi|t_vs|t_xs|tabline|tabpagemax|tabstop|tagbsearch|taglength|tagrelative|tagstack|tal|tb|tbi|tbidi|tbis|tbs|tenc|term|termbidi|termencoding|terse|textauto|textmode|textwidth|tgst|thesaurus|tildeop|timeout|timeoutlen|title|titlelen|titleold|titlestring|toolbar|toolbariconsize|top|tpm|tsl|tsr|ttimeout|ttimeoutlen|ttm|tty|ttybuiltin|ttyfast|ttym|ttymouse|ttyscroll|ttytype|tw|tx|uc|ul|undolevels|updatecount|updatetime|ut|vb|vbs|vdir|verbosefile|vfile|viewdir|viewoptions|viminfo|virtualedit|visualbell|vop|wak|warn|wb|wc|wcm|wd|weirdinvert|wfh|wfw|whichwrap|wi|wig|wildchar|wildcharm|wildignore|wildmenu|wildmode|wildoptions|wim|winaltkeys|window|winfixheight|winfixwidth|winheight|winminheight|winminwidth|winwidth|wiv|wiw|wm|wmh|wmnu|wmw|wop|wrap|wrapmargin|wrapscan|writeany|writebackup|writedelay|ww)\b/,number:/\b(?:0x[\da-f]+|\d+(?:\.\d+)?)\b/i,operator:/\|\||&&|[-+.]=?|[=!](?:[=~][#?]?)?|[<>]=?[#?]?|[*\/%?]|\b(?:is(?:not)?)\b/,punctuation:/[{}[\](),;:]/}}return Kw}var Xw,bP;function uEe(){if(bP)return Xw;bP=1,Xw=e,e.displayName="visualBasic",e.aliases=[];function e(t){t.languages["visual-basic"]={comment:{pattern:/(?:['‘’]|REM\b)(?:[^\r\n_]|_(?:\r\n?|\n)?)*/i,inside:{keyword:/^REM/i}},directive:{pattern:/#(?:Const|Else|ElseIf|End|ExternalChecksum|ExternalSource|If|Region)(?:\b_[ \t]*(?:\r\n?|\n)|.)+/i,alias:"property",greedy:!0},string:{pattern:/\$?["“”](?:["“”]{2}|[^"“”])*["“”]C?/i,greedy:!0},date:{pattern:/#[ \t]*(?:\d+([/-])\d+\1\d+(?:[ \t]+(?:\d+[ \t]*(?:AM|PM)|\d+:\d+(?::\d+)?(?:[ \t]*(?:AM|PM))?))?|\d+[ \t]*(?:AM|PM)|\d+:\d+(?::\d+)?(?:[ \t]*(?:AM|PM))?)[ \t]*#/i,alias:"number"},number:/(?:(?:\b\d+(?:\.\d+)?|\.\d+)(?:E[+-]?\d+)?|&[HO][\dA-F]+)(?:[FRD]|U?[ILS])?/i,boolean:/\b(?:False|Nothing|True)\b/i,keyword:/\b(?:AddHandler|AddressOf|Alias|And(?:Also)?|As|Boolean|ByRef|Byte|ByVal|Call|Case|Catch|C(?:Bool|Byte|Char|Date|Dbl|Dec|Int|Lng|Obj|SByte|Short|Sng|Str|Type|UInt|ULng|UShort)|Char|Class|Const|Continue|Currency|Date|Decimal|Declare|Default|Delegate|Dim|DirectCast|Do|Double|Each|Else(?:If)?|End(?:If)?|Enum|Erase|Error|Event|Exit|Finally|For|Friend|Function|Get(?:Type|XMLNamespace)?|Global|GoSub|GoTo|Handles|If|Implements|Imports|In|Inherits|Integer|Interface|Is|IsNot|Let|Lib|Like|Long|Loop|Me|Mod|Module|Must(?:Inherit|Override)|My(?:Base|Class)|Namespace|Narrowing|New|Next|Not(?:Inheritable|Overridable)?|Object|Of|On|Operator|Option(?:al)?|Or(?:Else)?|Out|Overloads|Overridable|Overrides|ParamArray|Partial|Private|Property|Protected|Public|RaiseEvent|ReadOnly|ReDim|RemoveHandler|Resume|Return|SByte|Select|Set|Shadows|Shared|short|Single|Static|Step|Stop|String|Structure|Sub|SyncLock|Then|Throw|To|Try|TryCast|Type|TypeOf|U(?:Integer|Long|Short)|Until|Using|Variant|Wend|When|While|Widening|With(?:Events)?|WriteOnly|Xor)\b/i,operator:/[+\-*/\\^<=>&#@$%!]|\b_(?=[ \t]*[\r\n])/,punctuation:/[{}().,:?]/},t.languages.vb=t.languages["visual-basic"],t.languages.vba=t.languages["visual-basic"]}return Xw}var Zw,yP;function cEe(){if(yP)return Zw;yP=1,Zw=e,e.displayName="warpscript",e.aliases=[];function e(t){t.languages.warpscript={comment:/#.*|\/\/.*|\/\*[\s\S]*?\*\//,string:{pattern:/"(?:[^"\\\r\n]|\\.)*"|'(?:[^'\\\r\n]|\\.)*'|<'(?:[^\\']|'(?!>)|\\.)*'>/,greedy:!0},variable:/\$\S+/,macro:{pattern:/@\S+/,alias:"property"},keyword:/\b(?:BREAK|CHECKMACRO|CONTINUE|CUDF|DEFINED|DEFINEDMACRO|EVAL|FAIL|FOR|FOREACH|FORSTEP|IFT|IFTE|MSGFAIL|NRETURN|RETHROW|RETURN|SWITCH|TRY|UDF|UNTIL|WHILE)\b/,number:/[+-]?\b(?:NaN|Infinity|\d+(?:\.\d*)?(?:[Ee][+-]?\d+)?|0x[\da-fA-F]+|0b[01]+)\b/,boolean:/\b(?:F|T|false|true)\b/,punctuation:/<%|%>|[{}[\]()]/,operator:/==|&&?|\|\|?|\*\*?|>>>?|<<|[<>!~]=?|[-/%^]|\+!?|\b(?:AND|NOT|OR)\b/}}return Zw}var Qw,vP;function dEe(){if(vP)return Qw;vP=1,Qw=e,e.displayName="wasm",e.aliases=[];function e(t){t.languages.wasm={comment:[/\(;[\s\S]*?;\)/,{pattern:/;;.*/,greedy:!0}],string:{pattern:/"(?:\\[\s\S]|[^"\\])*"/,greedy:!0},keyword:[{pattern:/\b(?:align|offset)=/,inside:{operator:/=/}},{pattern:/\b(?:(?:f32|f64|i32|i64)(?:\.(?:abs|add|and|ceil|clz|const|convert_[su]\/i(?:32|64)|copysign|ctz|demote\/f64|div(?:_[su])?|eqz?|extend_[su]\/i32|floor|ge(?:_[su])?|gt(?:_[su])?|le(?:_[su])?|load(?:(?:8|16|32)_[su])?|lt(?:_[su])?|max|min|mul|neg?|nearest|or|popcnt|promote\/f32|reinterpret\/[fi](?:32|64)|rem_[su]|rot[lr]|shl|shr_[su]|sqrt|store(?:8|16|32)?|sub|trunc(?:_[su]\/f(?:32|64))?|wrap\/i64|xor))?|memory\.(?:grow|size))\b/,inside:{punctuation:/\./}},/\b(?:anyfunc|block|br(?:_if|_table)?|call(?:_indirect)?|data|drop|elem|else|end|export|func|get_(?:global|local)|global|if|import|local|loop|memory|module|mut|nop|offset|param|result|return|select|set_(?:global|local)|start|table|tee_local|then|type|unreachable)\b/],variable:/\$[\w!#$%&'*+\-./:<=>?@\\^`|~]+/,number:/[+-]?\b(?:\d(?:_?\d)*(?:\.\d(?:_?\d)*)?(?:[eE][+-]?\d(?:_?\d)*)?|0x[\da-fA-F](?:_?[\da-fA-F])*(?:\.[\da-fA-F](?:_?[\da-fA-D])*)?(?:[pP][+-]?\d(?:_?\d)*)?)\b|\binf\b|\bnan(?::0x[\da-fA-F](?:_?[\da-fA-D])*)?\b/,punctuation:/[()]/}}return Qw}var Jw,SP;function fEe(){if(SP)return Jw;SP=1,Jw=e,e.displayName="webIdl",e.aliases=[];function e(t){(function(n){var r=/(?:\B-|\b_|\b)[A-Za-z][\w-]*(?![\w-])/.source,a="(?:"+/\b(?:unsigned\s+)?long\s+long(?![\w-])/.source+"|"+/\b(?:unrestricted|unsigned)\s+[a-z]+(?![\w-])/.source+"|"+/(?!(?:unrestricted|unsigned)\b)/.source+r+/(?:\s*<(?:[^<>]|<[^<>]*>)*>)?/.source+")"+/(?:\s*\?)?/.source,o={};n.languages["web-idl"]={comment:{pattern:/\/\/.*|\/\*[\s\S]*?\*\//,greedy:!0},string:{pattern:/"[^"]*"/,greedy:!0},namespace:{pattern:RegExp(/(\bnamespace\s+)/.source+r),lookbehind:!0},"class-name":[{pattern:/(^|[^\w-])(?:iterable|maplike|setlike)\s*<(?:[^<>]|<[^<>]*>)*>/,lookbehind:!0,inside:o},{pattern:RegExp(/(\b(?:attribute|const|deleter|getter|optional|setter)\s+)/.source+a),lookbehind:!0,inside:o},{pattern:RegExp("("+/\bcallback\s+/.source+r+/\s*=\s*/.source+")"+a),lookbehind:!0,inside:o},{pattern:RegExp(/(\btypedef\b\s*)/.source+a),lookbehind:!0,inside:o},{pattern:RegExp(/(\b(?:callback|dictionary|enum|interface(?:\s+mixin)?)\s+)(?!(?:interface|mixin)\b)/.source+r),lookbehind:!0},{pattern:RegExp(/(:\s*)/.source+r),lookbehind:!0},RegExp(r+/(?=\s+(?:implements|includes)\b)/.source),{pattern:RegExp(/(\b(?:implements|includes)\s+)/.source+r),lookbehind:!0},{pattern:RegExp(a+"(?="+/\s*(?:\.{3}\s*)?/.source+r+/\s*[(),;=]/.source+")"),inside:o}],builtin:/\b(?:ArrayBuffer|BigInt64Array|BigUint64Array|ByteString|DOMString|DataView|Float32Array|Float64Array|FrozenArray|Int16Array|Int32Array|Int8Array|ObservableArray|Promise|USVString|Uint16Array|Uint32Array|Uint8Array|Uint8ClampedArray)\b/,keyword:[/\b(?:async|attribute|callback|const|constructor|deleter|dictionary|enum|getter|implements|includes|inherit|interface|mixin|namespace|null|optional|or|partial|readonly|required|setter|static|stringifier|typedef|unrestricted)\b/,/\b(?:any|bigint|boolean|byte|double|float|iterable|long|maplike|object|octet|record|sequence|setlike|short|symbol|undefined|unsigned|void)\b/],boolean:/\b(?:false|true)\b/,number:{pattern:/(^|[^\w-])-?(?:0x[0-9a-f]+|(?:\d+(?:\.\d*)?|\.\d+)(?:e[+-]?\d+)?|NaN|Infinity)(?![\w-])/i,lookbehind:!0},operator:/\.{3}|[=:?<>-]/,punctuation:/[(){}[\].,;]/};for(var s in n.languages["web-idl"])s!=="class-name"&&(o[s]=n.languages["web-idl"][s]);n.languages.webidl=n.languages["web-idl"]})(t)}return Jw}var ex,EP;function pEe(){if(EP)return ex;EP=1,ex=e,e.displayName="wiki",e.aliases=[];function e(t){t.languages.wiki=t.languages.extend("markup",{"block-comment":{pattern:/(^|[^\\])\/\*[\s\S]*?\*\//,lookbehind:!0,alias:"comment"},heading:{pattern:/^(=+)[^=\r\n].*?\1/m,inside:{punctuation:/^=+|=+$/,important:/.+/}},emphasis:{pattern:/('{2,5}).+?\1/,inside:{"bold-italic":{pattern:/(''''').+?(?=\1)/,lookbehind:!0,alias:["bold","italic"]},bold:{pattern:/(''')[^'](?:.*?[^'])?(?=\1)/,lookbehind:!0},italic:{pattern:/('')[^'](?:.*?[^'])?(?=\1)/,lookbehind:!0},punctuation:/^''+|''+$/}},hr:{pattern:/^-{4,}/m,alias:"punctuation"},url:[/ISBN +(?:97[89][ -]?)?(?:\d[ -]?){9}[\dx]\b|(?:PMID|RFC) +\d+/i,/\[\[.+?\]\]|\[.+?\]/],variable:[/__[A-Z]+__/,/\{{3}.+?\}{3}/,/\{\{.+?\}\}/],symbol:[/^#redirect/im,/~{3,5}/],"table-tag":{pattern:/((?:^|[|!])[|!])[^|\r\n]+\|(?!\|)/m,lookbehind:!0,inside:{"table-bar":{pattern:/\|$/,alias:"punctuation"},rest:t.languages.markup.tag.inside}},punctuation:/^(?:\{\||\|\}|\|-|[*#:;!|])|\|\||!!/m}),t.languages.insertBefore("wiki","tag",{nowiki:{pattern:/<(nowiki|pre|source)\b[^>]*>[\s\S]*?<\/\1>/i,inside:{tag:{pattern:/<(?:nowiki|pre|source)\b[^>]*>|<\/(?:nowiki|pre|source)>/i,inside:t.languages.markup.tag.inside}}}})}return ex}var tx,wP;function gEe(){if(wP)return tx;wP=1,tx=e,e.displayName="wolfram",e.aliases=["mathematica","wl","nb"];function e(t){t.languages.wolfram={comment:/\(\*(?:\(\*(?:[^*]|\*(?!\)))*\*\)|(?!\(\*)[\s\S])*?\*\)/,string:{pattern:/"(?:\\.|[^"\\\r\n])*"/,greedy:!0},keyword:/\b(?:Abs|AbsArg|Accuracy|Block|Do|For|Function|If|Manipulate|Module|Nest|NestList|None|Return|Switch|Table|Which|While)\b/,context:{pattern:/\b\w+`+\w*/,alias:"class-name"},blank:{pattern:/\b\w+_\b/,alias:"regex"},"global-variable":{pattern:/\$\w+/,alias:"variable"},boolean:/\b(?:False|True)\b/,number:/(?:\b(?=\d)|\B(?=\.))(?:0[bo])?(?:(?:\d|0x[\da-f])[\da-f]*(?:\.\d*)?|\.\d+)(?:e[+-]?\d+)?j?\b/i,operator:/\/\.|;|=\.|\^=|\^:=|:=|<<|>>|<\||\|>|:>|\|->|->|<-|@@@|@@|@|\/@|=!=|===|==|=|\+|-|\^|\[\/-+%=\]=?|!=|\*\*?=?|\/\/?=?|<[<=>]?|>[=>]?|[&|^~]/,punctuation:/[{}[\];(),.:]/},t.languages.mathematica=t.languages.wolfram,t.languages.wl=t.languages.wolfram,t.languages.nb=t.languages.wolfram}return tx}var nx,xP;function hEe(){if(xP)return nx;xP=1,nx=e,e.displayName="wren",e.aliases=[];function e(t){t.languages.wren={comment:[{pattern:/\/\*(?:[^*/]|\*(?!\/)|\/(?!\*)|\/\*(?:[^*/]|\*(?!\/)|\/(?!\*)|\/\*(?:[^*/]|\*(?!\/)|\/(?!\*))*\*\/)*\*\/)*\*\//,greedy:!0},{pattern:/(^|[^\\:])\/\/.*/,lookbehind:!0,greedy:!0}],"triple-quoted-string":{pattern:/"""[\s\S]*?"""/,greedy:!0,alias:"string"},"string-literal":null,hashbang:{pattern:/^#!\/.+/,greedy:!0,alias:"comment"},attribute:{pattern:/#!?[ \t\u3000]*\w+/,alias:"keyword"},"class-name":[{pattern:/(\bclass\s+)\w+/,lookbehind:!0},/\b[A-Z][a-z\d_]*\b/],constant:/\b[A-Z][A-Z\d_]*\b/,null:{pattern:/\bnull\b/,alias:"keyword"},keyword:/\b(?:as|break|class|construct|continue|else|for|foreign|if|import|in|is|return|static|super|this|var|while)\b/,boolean:/\b(?:false|true)\b/,number:/\b(?:0x[\da-f]+|\d+(?:\.\d+)?(?:e[+-]?\d+)?)\b/i,function:/\b[a-z_]\w*(?=\s*[({])/i,operator:/<<|>>|[=!<>]=?|&&|\|\||[-+*/%~^&|?:]|\.{2,3}/,punctuation:/[\[\](){}.,;]/},t.languages.wren["string-literal"]={pattern:/(^|[^\\"])"(?:[^\\"%]|\\[\s\S]|%(?!\()|%\((?:[^()]|\((?:[^()]|\([^)]*\))*\))*\))*"/,lookbehind:!0,greedy:!0,inside:{interpolation:{pattern:/((?:^|[^\\])(?:\\{2})*)%\((?:[^()]|\((?:[^()]|\([^)]*\))*\))*\)/,lookbehind:!0,inside:{expression:{pattern:/^(%\()[\s\S]+(?=\)$)/,lookbehind:!0,inside:t.languages.wren},"interpolation-punctuation":{pattern:/^%\(|\)$/,alias:"punctuation"}}},string:/[\s\S]+/}}}return nx}var rx,kP;function mEe(){if(kP)return rx;kP=1,rx=e,e.displayName="xeora",e.aliases=["xeoracube"];function e(t){(function(n){n.languages.xeora=n.languages.extend("markup",{constant:{pattern:/\$(?:DomainContents|PageRenderDuration)\$/,inside:{punctuation:{pattern:/\$/}}},variable:{pattern:/\$@?(?:#+|[-+*~=^])?[\w.]+\$/,inside:{punctuation:{pattern:/[$.]/},operator:{pattern:/#+|[-+*~=^@]/}}},"function-inline":{pattern:/\$F:[-\w.]+\?[-\w.]+(?:,(?:(?:@[-#]*\w+\.[\w+.]\.*)*\|)*(?:(?:[\w+]|[-#*.~^]+[\w+]|=\S)(?:[^$=]|=+[^=])*=*|(?:@[-#]*\w+\.[\w+.]\.*)+(?:(?:[\w+]|[-#*~^][-#*.~^]*[\w+]|=\S)(?:[^$=]|=+[^=])*=*)?)?)?\$/,inside:{variable:{pattern:/(?:[,|])@?(?:#+|[-+*~=^])?[\w.]+/,inside:{punctuation:{pattern:/[,.|]/},operator:{pattern:/#+|[-+*~=^@]/}}},punctuation:{pattern:/\$\w:|[$:?.,|]/}},alias:"function"},"function-block":{pattern:/\$XF:\{[-\w.]+\?[-\w.]+(?:,(?:(?:@[-#]*\w+\.[\w+.]\.*)*\|)*(?:(?:[\w+]|[-#*.~^]+[\w+]|=\S)(?:[^$=]|=+[^=])*=*|(?:@[-#]*\w+\.[\w+.]\.*)+(?:(?:[\w+]|[-#*~^][-#*.~^]*[\w+]|=\S)(?:[^$=]|=+[^=])*=*)?)?)?\}:XF\$/,inside:{punctuation:{pattern:/[$:{}?.,|]/}},alias:"function"},"directive-inline":{pattern:/\$\w(?:#\d+\+?)?(?:\[[-\w.]+\])?:[-\/\w.]+\$/,inside:{punctuation:{pattern:/\$(?:\w:|C(?:\[|#\d))?|[:{[\]]/,inside:{tag:{pattern:/#\d/}}}},alias:"function"},"directive-block-open":{pattern:/\$\w+:\{|\$\w(?:#\d+\+?)?(?:\[[-\w.]+\])?:[-\w.]+:\{(?:![A-Z]+)?/,inside:{punctuation:{pattern:/\$(?:\w:|C(?:\[|#\d))?|[:{[\]]/,inside:{tag:{pattern:/#\d/}}},attribute:{pattern:/![A-Z]+$/,inside:{punctuation:{pattern:/!/}},alias:"keyword"}},alias:"function"},"directive-block-separator":{pattern:/\}:[-\w.]+:\{/,inside:{punctuation:{pattern:/[:{}]/}},alias:"function"},"directive-block-close":{pattern:/\}:[-\w.]+\$/,inside:{punctuation:{pattern:/[:{}$]/}},alias:"function"}}),n.languages.insertBefore("inside","punctuation",{variable:n.languages.xeora["function-inline"].inside.variable},n.languages.xeora["function-block"]),n.languages.xeoracube=n.languages.xeora})(t)}return rx}var ax,TP;function bEe(){if(TP)return ax;TP=1,ax=e,e.displayName="xmlDoc",e.aliases=[];function e(t){(function(n){function r(u,c){n.languages[u]&&n.languages.insertBefore(u,"comment",{"doc-comment":c})}var a=n.languages.markup.tag,o={pattern:/\/\/\/.*/,greedy:!0,alias:"comment",inside:{tag:a}},s={pattern:/'''.*/,greedy:!0,alias:"comment",inside:{tag:a}};r("csharp",o),r("fsharp",o),r("vbnet",s)})(t)}return ax}var ox,AP;function yEe(){if(AP)return ox;AP=1,ox=e,e.displayName="xojo",e.aliases=[];function e(t){t.languages.xojo={comment:{pattern:/(?:'|\/\/|Rem\b).+/i,greedy:!0},string:{pattern:/"(?:""|[^"])*"/,greedy:!0},number:[/(?:\b\d+(?:\.\d*)?|\B\.\d+)(?:E[+-]?\d+)?/i,/&[bchou][a-z\d]+/i],directive:{pattern:/#(?:Else|ElseIf|Endif|If|Pragma)\b/i,alias:"property"},keyword:/\b(?:AddHandler|App|Array|As(?:signs)?|Auto|Boolean|Break|By(?:Ref|Val)|Byte|Call|Case|Catch|CFStringRef|CGFloat|Class|Color|Const|Continue|CString|Currency|CurrentMethodName|Declare|Delegate|Dim|Do(?:uble|wnTo)?|Each|Else(?:If)?|End|Enumeration|Event|Exception|Exit|Extends|False|Finally|For|Function|Get|GetTypeInfo|Global|GOTO|If|Implements|In|Inherits|Int(?:8|16|32|64|eger|erface)?|Lib|Loop|Me|Module|Next|Nil|Object|Optional|OSType|ParamArray|Private|Property|Protected|PString|Ptr|Raise(?:Event)?|ReDim|RemoveHandler|Return|Select(?:or)?|Self|Set|Shared|Short|Single|Soft|Static|Step|String|Sub|Super|Text|Then|To|True|Try|Ubound|UInt(?:8|16|32|64|eger)?|Until|Using|Var(?:iant)?|Wend|While|WindowPtr|WString)\b/i,operator:/<[=>]?|>=?|[+\-*\/\\^=]|\b(?:AddressOf|And|Ctype|IsA?|Mod|New|Not|Or|WeakAddressOf|Xor)\b/i,punctuation:/[.,;:()]/}}return ox}var ix,RP;function vEe(){if(RP)return ix;RP=1,ix=e,e.displayName="xquery",e.aliases=[];function e(t){(function(n){n.languages.xquery=n.languages.extend("markup",{"xquery-comment":{pattern:/\(:[\s\S]*?:\)/,greedy:!0,alias:"comment"},string:{pattern:/(["'])(?:\1\1|(?!\1)[\s\S])*\1/,greedy:!0},extension:{pattern:/\(#.+?#\)/,alias:"symbol"},variable:/\$[-\w:]+/,axis:{pattern:/(^|[^-])(?:ancestor(?:-or-self)?|attribute|child|descendant(?:-or-self)?|following(?:-sibling)?|parent|preceding(?:-sibling)?|self)(?=::)/,lookbehind:!0,alias:"operator"},"keyword-operator":{pattern:/(^|[^:-])\b(?:and|castable as|div|eq|except|ge|gt|idiv|instance of|intersect|is|le|lt|mod|ne|or|union)\b(?=$|[^:-])/,lookbehind:!0,alias:"operator"},keyword:{pattern:/(^|[^:-])\b(?:as|ascending|at|base-uri|boundary-space|case|cast as|collation|construction|copy-namespaces|declare|default|descending|else|empty (?:greatest|least)|encoding|every|external|for|function|if|import|in|inherit|lax|let|map|module|namespace|no-inherit|no-preserve|option|order(?: by|ed|ing)?|preserve|return|satisfies|schema|some|stable|strict|strip|then|to|treat as|typeswitch|unordered|validate|variable|version|where|xquery)\b(?=$|[^:-])/,lookbehind:!0},function:/[\w-]+(?::[\w-]+)*(?=\s*\()/,"xquery-element":{pattern:/(element\s+)[\w-]+(?::[\w-]+)*/,lookbehind:!0,alias:"tag"},"xquery-attribute":{pattern:/(attribute\s+)[\w-]+(?::[\w-]+)*/,lookbehind:!0,alias:"attr-name"},builtin:{pattern:/(^|[^:-])\b(?:attribute|comment|document|element|processing-instruction|text|xs:(?:ENTITIES|ENTITY|ID|IDREFS?|NCName|NMTOKENS?|NOTATION|Name|QName|anyAtomicType|anyType|anyURI|base64Binary|boolean|byte|date|dateTime|dayTimeDuration|decimal|double|duration|float|gDay|gMonth|gMonthDay|gYear|gYearMonth|hexBinary|int|integer|language|long|negativeInteger|nonNegativeInteger|nonPositiveInteger|normalizedString|positiveInteger|short|string|time|token|unsigned(?:Byte|Int|Long|Short)|untyped(?:Atomic)?|yearMonthDuration))\b(?=$|[^:-])/,lookbehind:!0},number:/\b\d+(?:\.\d+)?(?:E[+-]?\d+)?/,operator:[/[+*=?|@]|\.\.?|:=|!=|<[=<]?|>[=>]?/,{pattern:/(\s)-(?=\s)/,lookbehind:!0}],punctuation:/[[\](){},;:/]/}),n.languages.xquery.tag.pattern=/<\/?(?!\d)[^\s>\/=$<%]+(?:\s+[^\s>\/=]+(?:=(?:("|')(?:\\[\s\S]|\{(?!\{)(?:\{(?:\{[^{}]*\}|[^{}])*\}|[^{}])+\}|(?!\1)[^\\])*\1|[^\s'">=]+))?)*\s*\/?>/,n.languages.xquery.tag.inside["attr-value"].pattern=/=(?:("|')(?:\\[\s\S]|\{(?!\{)(?:\{(?:\{[^{}]*\}|[^{}])*\}|[^{}])+\}|(?!\1)[^\\])*\1|[^\s'">=]+)/,n.languages.xquery.tag.inside["attr-value"].inside.punctuation=/^="|"$/,n.languages.xquery.tag.inside["attr-value"].inside.expression={pattern:/\{(?!\{)(?:\{(?:\{[^{}]*\}|[^{}])*\}|[^{}])+\}/,inside:n.languages.xquery,alias:"language-xquery"};var r=function(o){return typeof o=="string"?o:typeof o.content=="string"?o.content:o.content.map(r).join("")},a=function(o){for(var s=[],u=0;u<o.length;u++){var c=o[u],d=!1;if(typeof c!="string"&&(c.type==="tag"&&c.content[0]&&c.content[0].type==="tag"?c.content[0].content[0].content==="</"?s.length>0&&s[s.length-1].tagName===r(c.content[0].content[1])&&s.pop():c.content[c.content.length-1].content==="/>"||s.push({tagName:r(c.content[0].content[1]),openedBraces:0}):s.length>0&&c.type==="punctuation"&&c.content==="{"&&(!o[u+1]||o[u+1].type!=="punctuation"||o[u+1].content!=="{")&&(!o[u-1]||o[u-1].type!=="plain-text"||o[u-1].content!=="{")?s[s.length-1].openedBraces++:s.length>0&&s[s.length-1].openedBraces>0&&c.type==="punctuation"&&c.content==="}"?s[s.length-1].openedBraces--:c.type!=="comment"&&(d=!0)),(d||typeof c=="string")&&s.length>0&&s[s.length-1].openedBraces===0){var p=r(c);u<o.length-1&&(typeof o[u+1]=="string"||o[u+1].type==="plain-text")&&(p+=r(o[u+1]),o.splice(u+1,1)),u>0&&(typeof o[u-1]=="string"||o[u-1].type==="plain-text")&&(p=r(o[u-1])+p,o.splice(u-1,1),u--),/^\s+$/.test(p)?o[u]=p:o[u]=new n.Token("plain-text",p,null,p)}c.content&&typeof c.content!="string"&&a(c.content)}};n.hooks.add("after-tokenize",function(o){o.language==="xquery"&&a(o.tokens)})})(t)}return ix}var sx,_P;function SEe(){if(_P)return sx;_P=1,sx=e,e.displayName="yang",e.aliases=[];function e(t){t.languages.yang={comment:/\/\*[\s\S]*?\*\/|\/\/.*/,string:{pattern:/"(?:[^\\"]|\\.)*"|'[^']*'/,greedy:!0},keyword:{pattern:/(^|[{};\r\n][ \t]*)[a-z_][\w.-]*/i,lookbehind:!0},namespace:{pattern:/(\s)[a-z_][\w.-]*(?=:)/i,lookbehind:!0},boolean:/\b(?:false|true)\b/,operator:/\+/,punctuation:/[{};:]/}}return sx}var lx,CP;function EEe(){if(CP)return lx;CP=1,lx=e,e.displayName="zig",e.aliases=[];function e(t){(function(n){function r(p){return function(){return p}}var a=/\b(?:align|allowzero|and|anyframe|anytype|asm|async|await|break|cancel|catch|comptime|const|continue|defer|else|enum|errdefer|error|export|extern|fn|for|if|inline|linksection|nakedcc|noalias|nosuspend|null|or|orelse|packed|promise|pub|resume|return|stdcallcc|struct|suspend|switch|test|threadlocal|try|undefined|union|unreachable|usingnamespace|var|volatile|while)\b/,o="\\b(?!"+a.source+")(?!\\d)\\w+\\b",s=/align\s*\((?:[^()]|\([^()]*\))*\)/.source,u=/(?:\?|\bpromise->|(?:\[[^[\]]*\]|\*(?!\*)|\*\*)(?:\s*<ALIGN>|\s*const\b|\s*volatile\b|\s*allowzero\b)*)/.source.replace(/<ALIGN>/g,r(s)),c=/(?:\bpromise\b|(?:\berror\.)?<ID>(?:\.<ID>)*(?!\s+<ID>))/.source.replace(/<ID>/g,r(o)),d="(?!\\s)(?:!?\\s*(?:"+u+"\\s*)*"+c+")+";n.languages.zig={comment:[{pattern:/\/\/[/!].*/,alias:"doc-comment"},/\/{2}.*/],string:[{pattern:/(^|[^\\@])c?"(?:[^"\\\r\n]|\\.)*"/,lookbehind:!0,greedy:!0},{pattern:/([\r\n])([ \t]+c?\\{2}).*(?:(?:\r\n?|\n)\2.*)*/,lookbehind:!0,greedy:!0}],char:{pattern:/(^|[^\\])'(?:[^'\\\r\n]|[\uD800-\uDFFF]{2}|\\(?:.|x[a-fA-F\d]{2}|u\{[a-fA-F\d]{1,6}\}))'/,lookbehind:!0,greedy:!0},builtin:/\B@(?!\d)\w+(?=\s*\()/,label:{pattern:/(\b(?:break|continue)\s*:\s*)\w+\b|\b(?!\d)\w+\b(?=\s*:\s*(?:\{|while\b))/,lookbehind:!0},"class-name":[/\b(?!\d)\w+(?=\s*=\s*(?:(?:extern|packed)\s+)?(?:enum|struct|union)\s*[({])/,{pattern:RegExp(/(:\s*)<TYPE>(?=\s*(?:<ALIGN>\s*)?[=;,)])|<TYPE>(?=\s*(?:<ALIGN>\s*)?\{)/.source.replace(/<TYPE>/g,r(d)).replace(/<ALIGN>/g,r(s))),lookbehind:!0,inside:null},{pattern:RegExp(/(\)\s*)<TYPE>(?=\s*(?:<ALIGN>\s*)?;)/.source.replace(/<TYPE>/g,r(d)).replace(/<ALIGN>/g,r(s))),lookbehind:!0,inside:null}],"builtin-type":{pattern:/\b(?:anyerror|bool|c_u?(?:int|long|longlong|short)|c_longdouble|c_void|comptime_(?:float|int)|f(?:16|32|64|128)|[iu](?:8|16|32|64|128|size)|noreturn|type|void)\b/,alias:"keyword"},keyword:a,function:/\b(?!\d)\w+(?=\s*\()/,number:/\b(?:0b[01]+|0o[0-7]+|0x[a-fA-F\d]+(?:\.[a-fA-F\d]*)?(?:[pP][+-]?[a-fA-F\d]+)?|\d+(?:\.\d*)?(?:[eE][+-]?\d+)?)\b/,boolean:/\b(?:false|true)\b/,operator:/\.[*?]|\.{2,3}|[-=]>|\*\*|\+\+|\|\||(?:<<|>>|[-+*]%|[-+*/%^&|<>!=])=?|[?~]/,punctuation:/[.:,;(){}[\]]/},n.languages.zig["class-name"].forEach(function(p){p.inside===null&&(p.inside=n.languages.zig)})})(t)}return lx}var ux,NP;function wEe(){if(NP)return ux;NP=1;var e=zme();return ux=e,e.register(Ume()),e.register(jme()),e.register(Gme()),e.register(Hme()),e.register($me()),e.register(qme()),e.register(Vme()),e.register(Wme()),e.register(Yme()),e.register(Kme()),e.register(Xme()),e.register(Zme()),e.register(Qme()),e.register(Jme()),e.register(ebe()),e.register(tbe()),e.register(nbe()),e.register(rbe()),e.register(abe()),e.register(obe()),e.register(ibe()),e.register(sbe()),e.register(GH()),e.register(HH()),e.register(lbe()),e.register(ube()),e.register(cbe()),e.register(dbe()),e.register(fbe()),e.register(pbe()),e.register(gbe()),e.register(hbe()),e.register(mbe()),e.register(bbe()),e.register(Ai()),e.register(ybe()),e.register(vbe()),e.register(Sbe()),e.register(Ebe()),e.register(wbe()),e.register(xbe()),e.register(kbe()),e.register(Tbe()),e.register(Abe()),e.register(uA()),e.register(Rbe()),e.register(Mp()),e.register(_be()),e.register(Cbe()),e.register(Nbe()),e.register(Obe()),e.register(Dbe()),e.register(Ibe()),e.register(Lbe()),e.register(Mbe()),e.register(Fbe()),e.register(Pbe()),e.register(zbe()),e.register(Bbe()),e.register(Ube()),e.register(jbe()),e.register(Gbe()),e.register(Hbe()),e.register($be()),e.register(qbe()),e.register(Vbe()),e.register(Wbe()),e.register(Ybe()),e.register(Kbe()),e.register(Xbe()),e.register(Zbe()),e.register(Qbe()),e.register(Jbe()),e.register(eye()),e.register(tye()),e.register(nye()),e.register(rye()),e.register(aye()),e.register(oye()),e.register(iye()),e.register(sye()),e.register(lye()),e.register(uye()),e.register(cye()),e.register(dye()),e.register(fye()),e.register(pye()),e.register(gye()),e.register(hye()),e.register(mye()),e.register(bye()),e.register(yye()),e.register(vye()),e.register(Sye()),e.register(cA()),e.register(Eye()),e.register(wye()),e.register(xye()),e.register(kye()),e.register(Tye()),e.register(Aye()),e.register(Rye()),e.register(_ye()),e.register(Cye()),e.register(Nye()),e.register(Oye()),e.register(Dye()),e.register(Iye()),e.register(Lye()),e.register(Mye()),e.register(Fye()),e.register(Pye()),e.register(dA()),e.register(zye()),e.register(Pp()),e.register(Bye()),e.register(Uye()),e.register(jye()),e.register(Gye()),e.register(Hye()),e.register($ye()),e.register(qye()),e.register(pA()),e.register(Vye()),e.register(Wye()),e.register(Yye()),e.register(qH()),e.register(Kye()),e.register(Xye()),e.register(Zye()),e.register(Qye()),e.register(Jye()),e.register(eve()),e.register(tve()),e.register(nve()),e.register(rve()),e.register(ave()),e.register(ove()),e.register(ive()),e.register(sve()),e.register(lve()),e.register(uve()),e.register(cve()),e.register($H()),e.register(dve()),e.register(fve()),e.register(pve()),e.register(Wn()),e.register(gve()),e.register(hve()),e.register(mve()),e.register(bve()),e.register(yve()),e.register(vve()),e.register(Sve()),e.register(Eve()),e.register(wve()),e.register(xve()),e.register(kve()),e.register(Tve()),e.register(Ave()),e.register(Rve()),e.register(_ve()),e.register(Cve()),e.register(Nve()),e.register(Ove()),e.register(Dve()),e.register(Ive()),e.register(Lve()),e.register(Mve()),e.register(Fve()),e.register(Pve()),e.register(zve()),e.register(Bve()),e.register(Uve()),e.register(jve()),e.register(Gve()),e.register(Hve()),e.register($ve()),e.register(qve()),e.register(zp()),e.register(Vve()),e.register(Wve()),e.register(Yve()),e.register(Kve()),e.register(Xve()),e.register(Zve()),e.register(Qve()),e.register(Jve()),e.register(eSe()),e.register(tSe()),e.register(nSe()),e.register(rSe()),e.register(aSe()),e.register(oSe()),e.register(iSe()),e.register(sSe()),e.register(lSe()),e.register(uSe()),e.register(cSe()),e.register(dSe()),e.register(fSe()),e.register(pSe()),e.register(gSe()),e.register(hSe()),e.register(mSe()),e.register(bSe()),e.register(ySe()),e.register(vSe()),e.register(SSe()),e.register(ESe()),e.register(Fp()),e.register(wSe()),e.register(xSe()),e.register(kSe()),e.register(TSe()),e.register(gA()),e.register(ASe()),e.register(RSe()),e.register(_Se()),e.register(CSe()),e.register(NSe()),e.register(OSe()),e.register(DSe()),e.register(ISe()),e.register(LSe()),e.register(MSe()),e.register(FSe()),e.register(PSe()),e.register(lA()),e.register(zSe()),e.register(BSe()),e.register(USe()),e.register(jSe()),e.register(GSe()),e.register(HSe()),e.register(hA()),e.register($Se()),e.register(qSe()),e.register(VSe()),e.register(WSe()),e.register(YSe()),e.register(KSe()),e.register(XSe()),e.register(ZSe()),e.register(VH()),e.register(QSe()),e.register(fA()),e.register(JSe()),e.register(eEe()),e.register(tEe()),e.register(nEe()),e.register(rEe()),e.register(aEe()),e.register(WH()),e.register(oEe()),e.register(iEe()),e.register(sEe()),e.register(lEe()),e.register(uEe()),e.register(cEe()),e.register(dEe()),e.register(fEe()),e.register(pEe()),e.register(gEe()),e.register(hEe()),e.register(mEe()),e.register(bEe()),e.register(yEe()),e.register(vEe()),e.register(YH()),e.register(SEe()),e.register(EEe()),ux}var xEe=wEe();const kEe=un(xEe);var KH=cge(kEe,Bme);KH.supportedLanguages=dge;var cx={},dx={exports:{}},OP;function TEe(){return OP||(OP=1,function(e){function t(n){return n&&n.__esModule?n:{default:n}}e.exports=t,e.exports.__esModule=!0,e.exports.default=e.exports}(dx)),dx.exports}var fx={},DP;function AEe(){return DP||(DP=1,function(e){Object.defineProperty(e,"__esModule",{value:!0}),e.default=void 0,e.default={'code[class*="language-"]':{color:"black",background:"none",fontFamily:"Consolas, Monaco, 'Andale Mono', 'Ubuntu Mono', monospace",fontSize:"1em",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",wordWrap:"normal",lineHeight:"1.5",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none",maxHeight:"inherit",height:"inherit",padding:"0 1em",display:"block",overflow:"auto"},'pre[class*="language-"]':{color:"black",background:"none",fontFamily:"Consolas, Monaco, 'Andale Mono', 'Ubuntu Mono', monospace",fontSize:"1em",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",wordWrap:"normal",lineHeight:"1.5",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none",position:"relative",margin:".5em 0",overflow:"visible",padding:"1px",backgroundColor:"#fdfdfd",WebkitBoxSizing:"border-box",MozBoxSizing:"border-box",boxSizing:"border-box",marginBottom:"1em"},'pre[class*="language-"] > code':{position:"relative",zIndex:"1",borderLeft:"10px solid #358ccb",boxShadow:"-1px 0px 0px 0px #358ccb, 0px 0px 0px 1px #dfdfdf",backgroundColor:"#fdfdfd",backgroundImage:"linear-gradient(transparent 50%, rgba(69, 142, 209, 0.04) 50%)",backgroundSize:"3em 3em",backgroundOrigin:"content-box",backgroundAttachment:"local"},':not(pre) > code[class*="language-"]':{backgroundColor:"#fdfdfd",WebkitBoxSizing:"border-box",MozBoxSizing:"border-box",boxSizing:"border-box",marginBottom:"1em",position:"relative",padding:".2em",borderRadius:"0.3em",color:"#c92c2c",border:"1px solid rgba(0, 0, 0, 0.1)",display:"inline",whiteSpace:"normal"},'pre[class*="language-"]:before':{content:"''",display:"block",position:"absolute",bottom:"0.75em",left:"0.18em",width:"40%",height:"20%",maxHeight:"13em",boxShadow:"0px 13px 8px #979797",WebkitTransform:"rotate(-2deg)",MozTransform:"rotate(-2deg)",msTransform:"rotate(-2deg)",OTransform:"rotate(-2deg)",transform:"rotate(-2deg)"},'pre[class*="language-"]:after':{content:"''",display:"block",position:"absolute",bottom:"0.75em",left:"auto",width:"40%",height:"20%",maxHeight:"13em",boxShadow:"0px 13px 8px #979797",WebkitTransform:"rotate(2deg)",MozTransform:"rotate(2deg)",msTransform:"rotate(2deg)",OTransform:"rotate(2deg)",transform:"rotate(2deg)",right:"0.75em"},comment:{color:"#7D8B99"},"block-comment":{color:"#7D8B99"},prolog:{color:"#7D8B99"},doctype:{color:"#7D8B99"},cdata:{color:"#7D8B99"},punctuation:{color:"#5F6364"},property:{color:"#c92c2c"},tag:{color:"#c92c2c"},boolean:{color:"#c92c2c"},number:{color:"#c92c2c"},"function-name":{color:"#c92c2c"},constant:{color:"#c92c2c"},symbol:{color:"#c92c2c"},deleted:{color:"#c92c2c"},selector:{color:"#2f9c0a"},"attr-name":{color:"#2f9c0a"},string:{color:"#2f9c0a"},char:{color:"#2f9c0a"},function:{color:"#2f9c0a"},builtin:{color:"#2f9c0a"},inserted:{color:"#2f9c0a"},operator:{color:"#a67f59",background:"rgba(255, 255, 255, 0.5)"},entity:{color:"#a67f59",background:"rgba(255, 255, 255, 0.5)",cursor:"help"},url:{color:"#a67f59",background:"rgba(255, 255, 255, 0.5)"},variable:{color:"#a67f59",background:"rgba(255, 255, 255, 0.5)"},atrule:{color:"#1990b8"},"attr-value":{color:"#1990b8"},keyword:{color:"#1990b8"},"class-name":{color:"#1990b8"},regex:{color:"#e90"},important:{color:"#e90",fontWeight:"normal"},".language-css .token.string":{color:"#a67f59",background:"rgba(255, 255, 255, 0.5)"},".style .token.string":{color:"#a67f59",background:"rgba(255, 255, 255, 0.5)"},bold:{fontWeight:"bold"},italic:{fontStyle:"italic"},namespace:{Opacity:".7"},'pre[class*="language-"].line-numbers.line-numbers':{paddingLeft:"0"},'pre[class*="language-"].line-numbers.line-numbers code':{paddingLeft:"3.8em"},'pre[class*="language-"].line-numbers.line-numbers .line-numbers-rows':{left:"0"},'pre[class*="language-"][data-line]':{paddingTop:"0",paddingBottom:"0",paddingLeft:"0"},"pre[data-line] code":{position:"relative",paddingLeft:"4em"},"pre .line-highlight":{marginTop:"0"}}}(fx)),fx}var px={},IP;function REe(){return IP||(IP=1,function(e){Object.defineProperty(e,"__esModule",{value:!0}),e.default=void 0,e.default={'code[class*="language-"]':{color:"white",background:"none",textShadow:"0 -.1em .2em black",fontFamily:"Consolas, Monaco, 'Andale Mono', 'Ubuntu Mono', monospace",fontSize:"1em",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",wordWrap:"normal",lineHeight:"1.5",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none"},'pre[class*="language-"]':{color:"white",background:"hsl(30, 20%, 25%)",textShadow:"0 -.1em .2em black",fontFamily:"Consolas, Monaco, 'Andale Mono', 'Ubuntu Mono', monospace",fontSize:"1em",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",wordWrap:"normal",lineHeight:"1.5",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none",padding:"1em",margin:".5em 0",overflow:"auto",border:".3em solid hsl(30, 20%, 40%)",borderRadius:".5em",boxShadow:"1px 1px .5em black inset"},':not(pre) > code[class*="language-"]':{background:"hsl(30, 20%, 25%)",padding:".15em .2em .05em",borderRadius:".3em",border:".13em solid hsl(30, 20%, 40%)",boxShadow:"1px 1px .3em -.1em black inset",whiteSpace:"normal"},comment:{color:"hsl(30, 20%, 50%)"},prolog:{color:"hsl(30, 20%, 50%)"},doctype:{color:"hsl(30, 20%, 50%)"},cdata:{color:"hsl(30, 20%, 50%)"},punctuation:{Opacity:".7"},namespace:{Opacity:".7"},property:{color:"hsl(350, 40%, 70%)"},tag:{color:"hsl(350, 40%, 70%)"},boolean:{color:"hsl(350, 40%, 70%)"},number:{color:"hsl(350, 40%, 70%)"},constant:{color:"hsl(350, 40%, 70%)"},symbol:{color:"hsl(350, 40%, 70%)"},selector:{color:"hsl(75, 70%, 60%)"},"attr-name":{color:"hsl(75, 70%, 60%)"},string:{color:"hsl(75, 70%, 60%)"},char:{color:"hsl(75, 70%, 60%)"},builtin:{color:"hsl(75, 70%, 60%)"},inserted:{color:"hsl(75, 70%, 60%)"},operator:{color:"hsl(40, 90%, 60%)"},entity:{color:"hsl(40, 90%, 60%)",cursor:"help"},url:{color:"hsl(40, 90%, 60%)"},".language-css .token.string":{color:"hsl(40, 90%, 60%)"},".style .token.string":{color:"hsl(40, 90%, 60%)"},variable:{color:"hsl(40, 90%, 60%)"},atrule:{color:"hsl(350, 40%, 70%)"},"attr-value":{color:"hsl(350, 40%, 70%)"},keyword:{color:"hsl(350, 40%, 70%)"},regex:{color:"#e90"},important:{color:"#e90",fontWeight:"bold"},bold:{fontWeight:"bold"},italic:{fontStyle:"italic"},deleted:{color:"red"}}}(px)),px}var gx={},LP;function _Ee(){return LP||(LP=1,function(e){Object.defineProperty(e,"__esModule",{value:!0}),e.default=void 0,e.default={'code[class*="language-"]':{fontFamily:"Consolas, Monaco, 'Andale Mono', 'Ubuntu Mono', monospace",fontSize:"1em",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",wordWrap:"normal",lineHeight:"1.5",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none",background:"black",color:"white",boxShadow:"-.3em 0 0 .3em black, .3em 0 0 .3em black"},'pre[class*="language-"]':{fontFamily:"Consolas, Monaco, 'Andale Mono', 'Ubuntu Mono', monospace",fontSize:"1em",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",wordWrap:"normal",lineHeight:"1.5",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none",padding:".4em .8em",margin:".5em 0",overflow:"auto",background:`url('data:image/svg+xml;charset=utf-8,<svg%20version%3D"1.1"%20xmlns%3D"http%3A%2F%2Fwww.w3.org%2F2000%2Fsvg"%20width%3D"100"%20height%3D"100"%20fill%3D"rgba(0%2C0%2C0%2C.2)">%0D%0A<polygon%20points%3D"0%2C50%2050%2C0%200%2C0"%20%2F>%0D%0A<polygon%20points%3D"0%2C100%2050%2C100%20100%2C50%20100%2C0"%20%2F>%0D%0A<%2Fsvg>')`,backgroundSize:"1em 1em"},':not(pre) > code[class*="language-"]':{padding:".2em",borderRadius:".3em",boxShadow:"none",whiteSpace:"normal"},comment:{color:"#aaa"},prolog:{color:"#aaa"},doctype:{color:"#aaa"},cdata:{color:"#aaa"},punctuation:{color:"#999"},namespace:{Opacity:".7"},property:{color:"#0cf"},tag:{color:"#0cf"},boolean:{color:"#0cf"},number:{color:"#0cf"},constant:{color:"#0cf"},symbol:{color:"#0cf"},selector:{color:"yellow"},"attr-name":{color:"yellow"},string:{color:"yellow"},char:{color:"yellow"},builtin:{color:"yellow"},operator:{color:"yellowgreen"},entity:{color:"yellowgreen",cursor:"help"},url:{color:"yellowgreen"},".language-css .token.string":{color:"yellowgreen"},variable:{color:"yellowgreen"},inserted:{color:"yellowgreen"},atrule:{color:"deeppink"},"attr-value":{color:"deeppink"},keyword:{color:"deeppink"},regex:{color:"orange"},important:{color:"orange",fontWeight:"bold"},bold:{fontWeight:"bold"},italic:{fontStyle:"italic"},deleted:{color:"red"},"pre.diff-highlight.diff-highlight > code .token.deleted:not(.prefix)":{backgroundColor:"rgba(255, 0, 0, .3)",display:"inline"},"pre > code.diff-highlight.diff-highlight .token.deleted:not(.prefix)":{backgroundColor:"rgba(255, 0, 0, .3)",display:"inline"},"pre.diff-highlight.diff-highlight > code .token.inserted:not(.prefix)":{backgroundColor:"rgba(0, 255, 128, .3)",display:"inline"},"pre > code.diff-highlight.diff-highlight .token.inserted:not(.prefix)":{backgroundColor:"rgba(0, 255, 128, .3)",display:"inline"}}}(gx)),gx}var hx={},MP;function CEe(){return MP||(MP=1,function(e){Object.defineProperty(e,"__esModule",{value:!0}),e.default=void 0,e.default={'code[class*="language-"]':{color:"#f8f8f2",background:"none",textShadow:"0 1px rgba(0, 0, 0, 0.3)",fontFamily:"Consolas, Monaco, 'Andale Mono', 'Ubuntu Mono', monospace",fontSize:"1em",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",wordWrap:"normal",lineHeight:"1.5",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none"},'pre[class*="language-"]':{color:"#f8f8f2",background:"#272822",textShadow:"0 1px rgba(0, 0, 0, 0.3)",fontFamily:"Consolas, Monaco, 'Andale Mono', 'Ubuntu Mono', monospace",fontSize:"1em",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",wordWrap:"normal",lineHeight:"1.5",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none",padding:"1em",margin:".5em 0",overflow:"auto",borderRadius:"0.3em"},':not(pre) > code[class*="language-"]':{background:"#272822",padding:".1em",borderRadius:".3em",whiteSpace:"normal"},comment:{color:"#8292a2"},prolog:{color:"#8292a2"},doctype:{color:"#8292a2"},cdata:{color:"#8292a2"},punctuation:{color:"#f8f8f2"},namespace:{Opacity:".7"},property:{color:"#f92672"},tag:{color:"#f92672"},constant:{color:"#f92672"},symbol:{color:"#f92672"},deleted:{color:"#f92672"},boolean:{color:"#ae81ff"},number:{color:"#ae81ff"},selector:{color:"#a6e22e"},"attr-name":{color:"#a6e22e"},string:{color:"#a6e22e"},char:{color:"#a6e22e"},builtin:{color:"#a6e22e"},inserted:{color:"#a6e22e"},operator:{color:"#f8f8f2"},entity:{color:"#f8f8f2",cursor:"help"},url:{color:"#f8f8f2"},".language-css .token.string":{color:"#f8f8f2"},".style .token.string":{color:"#f8f8f2"},variable:{color:"#f8f8f2"},atrule:{color:"#e6db74"},"attr-value":{color:"#e6db74"},function:{color:"#e6db74"},"class-name":{color:"#e6db74"},keyword:{color:"#66d9ef"},regex:{color:"#fd971f"},important:{color:"#fd971f",fontWeight:"bold"},bold:{fontWeight:"bold"},italic:{fontStyle:"italic"}}}(hx)),hx}var mx={},FP;function NEe(){return FP||(FP=1,function(e){Object.defineProperty(e,"__esModule",{value:!0}),e.default=void 0,e.default={'code[class*="language-"]':{color:"#657b83",fontFamily:"Consolas, Monaco, 'Andale Mono', 'Ubuntu Mono', monospace",fontSize:"1em",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",wordWrap:"normal",lineHeight:"1.5",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none"},'pre[class*="language-"]':{color:"#657b83",fontFamily:"Consolas, Monaco, 'Andale Mono', 'Ubuntu Mono', monospace",fontSize:"1em",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",wordWrap:"normal",lineHeight:"1.5",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none",padding:"1em",margin:".5em 0",overflow:"auto",borderRadius:"0.3em",backgroundColor:"#fdf6e3"},'pre[class*="language-"]::-moz-selection':{background:"#073642"},'pre[class*="language-"] ::-moz-selection':{background:"#073642"},'code[class*="language-"]::-moz-selection':{background:"#073642"},'code[class*="language-"] ::-moz-selection':{background:"#073642"},'pre[class*="language-"]::selection':{background:"#073642"},'pre[class*="language-"] ::selection':{background:"#073642"},'code[class*="language-"]::selection':{background:"#073642"},'code[class*="language-"] ::selection':{background:"#073642"},':not(pre) > code[class*="language-"]':{backgroundColor:"#fdf6e3",padding:".1em",borderRadius:".3em"},comment:{color:"#93a1a1"},prolog:{color:"#93a1a1"},doctype:{color:"#93a1a1"},cdata:{color:"#93a1a1"},punctuation:{color:"#586e75"},namespace:{Opacity:".7"},property:{color:"#268bd2"},tag:{color:"#268bd2"},boolean:{color:"#268bd2"},number:{color:"#268bd2"},constant:{color:"#268bd2"},symbol:{color:"#268bd2"},deleted:{color:"#268bd2"},selector:{color:"#2aa198"},"attr-name":{color:"#2aa198"},string:{color:"#2aa198"},char:{color:"#2aa198"},builtin:{color:"#2aa198"},url:{color:"#2aa198"},inserted:{color:"#2aa198"},entity:{color:"#657b83",background:"#eee8d5",cursor:"help"},atrule:{color:"#859900"},"attr-value":{color:"#859900"},keyword:{color:"#859900"},function:{color:"#b58900"},"class-name":{color:"#b58900"},regex:{color:"#cb4b16"},important:{color:"#cb4b16",fontWeight:"bold"},variable:{color:"#cb4b16"},bold:{fontWeight:"bold"},italic:{fontStyle:"italic"}}}(mx)),mx}var bx={},PP;function OEe(){return PP||(PP=1,function(e){Object.defineProperty(e,"__esModule",{value:!0}),e.default=void 0,e.default={'code[class*="language-"]':{color:"#ccc",background:"none",fontFamily:"Consolas, Monaco, 'Andale Mono', 'Ubuntu Mono', monospace",fontSize:"1em",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",wordWrap:"normal",lineHeight:"1.5",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none"},'pre[class*="language-"]':{color:"#ccc",background:"#2d2d2d",fontFamily:"Consolas, Monaco, 'Andale Mono', 'Ubuntu Mono', monospace",fontSize:"1em",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",wordWrap:"normal",lineHeight:"1.5",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none",padding:"1em",margin:".5em 0",overflow:"auto"},':not(pre) > code[class*="language-"]':{background:"#2d2d2d",padding:".1em",borderRadius:".3em",whiteSpace:"normal"},comment:{color:"#999"},"block-comment":{color:"#999"},prolog:{color:"#999"},doctype:{color:"#999"},cdata:{color:"#999"},punctuation:{color:"#ccc"},tag:{color:"#e2777a"},"attr-name":{color:"#e2777a"},namespace:{color:"#e2777a"},deleted:{color:"#e2777a"},"function-name":{color:"#6196cc"},boolean:{color:"#f08d49"},number:{color:"#f08d49"},function:{color:"#f08d49"},property:{color:"#f8c555"},"class-name":{color:"#f8c555"},constant:{color:"#f8c555"},symbol:{color:"#f8c555"},selector:{color:"#cc99cd"},important:{color:"#cc99cd",fontWeight:"bold"},atrule:{color:"#cc99cd"},keyword:{color:"#cc99cd"},builtin:{color:"#cc99cd"},string:{color:"#7ec699"},char:{color:"#7ec699"},"attr-value":{color:"#7ec699"},regex:{color:"#7ec699"},variable:{color:"#7ec699"},operator:{color:"#67cdcc"},entity:{color:"#67cdcc",cursor:"help"},url:{color:"#67cdcc"},bold:{fontWeight:"bold"},italic:{fontStyle:"italic"},inserted:{color:"green"}}}(bx)),bx}var yx={},zP;function DEe(){return zP||(zP=1,function(e){Object.defineProperty(e,"__esModule",{value:!0}),e.default=void 0,e.default={'code[class*="language-"]':{color:"white",background:"none",fontFamily:"Consolas, Monaco, 'Andale Mono', 'Ubuntu Mono', monospace",fontSize:"1em",textAlign:"left",textShadow:"0 -.1em .2em black",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",wordWrap:"normal",lineHeight:"1.5",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none"},'pre[class*="language-"]':{color:"white",background:"hsl(0, 0%, 8%)",fontFamily:"Consolas, Monaco, 'Andale Mono', 'Ubuntu Mono', monospace",fontSize:"1em",textAlign:"left",textShadow:"0 -.1em .2em black",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",wordWrap:"normal",lineHeight:"1.5",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none",borderRadius:".5em",border:".3em solid hsl(0, 0%, 33%)",boxShadow:"1px 1px .5em black inset",margin:".5em 0",overflow:"auto",padding:"1em"},':not(pre) > code[class*="language-"]':{background:"hsl(0, 0%, 8%)",borderRadius:".3em",border:".13em solid hsl(0, 0%, 33%)",boxShadow:"1px 1px .3em -.1em black inset",padding:".15em .2em .05em",whiteSpace:"normal"},'pre[class*="language-"]::-moz-selection':{background:"hsla(0, 0%, 93%, 0.15)",textShadow:"none"},'pre[class*="language-"]::selection':{background:"hsla(0, 0%, 93%, 0.15)",textShadow:"none"},'pre[class*="language-"] ::-moz-selection':{textShadow:"none",background:"hsla(0, 0%, 93%, 0.15)"},'code[class*="language-"]::-moz-selection':{textShadow:"none",background:"hsla(0, 0%, 93%, 0.15)"},'code[class*="language-"] ::-moz-selection':{textShadow:"none",background:"hsla(0, 0%, 93%, 0.15)"},'pre[class*="language-"] ::selection':{textShadow:"none",background:"hsla(0, 0%, 93%, 0.15)"},'code[class*="language-"]::selection':{textShadow:"none",background:"hsla(0, 0%, 93%, 0.15)"},'code[class*="language-"] ::selection':{textShadow:"none",background:"hsla(0, 0%, 93%, 0.15)"},comment:{color:"hsl(0, 0%, 47%)"},prolog:{color:"hsl(0, 0%, 47%)"},doctype:{color:"hsl(0, 0%, 47%)"},cdata:{color:"hsl(0, 0%, 47%)"},punctuation:{Opacity:".7"},namespace:{Opacity:".7"},tag:{color:"hsl(14, 58%, 55%)"},boolean:{color:"hsl(14, 58%, 55%)"},number:{color:"hsl(14, 58%, 55%)"},deleted:{color:"hsl(14, 58%, 55%)"},keyword:{color:"hsl(53, 89%, 79%)"},property:{color:"hsl(53, 89%, 79%)"},selector:{color:"hsl(53, 89%, 79%)"},constant:{color:"hsl(53, 89%, 79%)"},symbol:{color:"hsl(53, 89%, 79%)"},builtin:{color:"hsl(53, 89%, 79%)"},"attr-name":{color:"hsl(76, 21%, 52%)"},"attr-value":{color:"hsl(76, 21%, 52%)"},string:{color:"hsl(76, 21%, 52%)"},char:{color:"hsl(76, 21%, 52%)"},operator:{color:"hsl(76, 21%, 52%)"},entity:{color:"hsl(76, 21%, 52%)",cursor:"help"},url:{color:"hsl(76, 21%, 52%)"},".language-css .token.string":{color:"hsl(76, 21%, 52%)"},".style .token.string":{color:"hsl(76, 21%, 52%)"},variable:{color:"hsl(76, 21%, 52%)"},inserted:{color:"hsl(76, 21%, 52%)"},atrule:{color:"hsl(218, 22%, 55%)"},regex:{color:"hsl(42, 75%, 65%)"},important:{color:"hsl(42, 75%, 65%)",fontWeight:"bold"},bold:{fontWeight:"bold"},italic:{fontStyle:"italic"},".language-markup .token.tag":{color:"hsl(33, 33%, 52%)"},".language-markup .token.attr-name":{color:"hsl(33, 33%, 52%)"},".language-markup .token.punctuation":{color:"hsl(33, 33%, 52%)"},"":{position:"relative",zIndex:"1"},".line-highlight.line-highlight":{background:"linear-gradient(to right, hsla(0, 0%, 33%, .1) 70%, hsla(0, 0%, 33%, 0))",borderBottom:"1px dashed hsl(0, 0%, 33%)",borderTop:"1px dashed hsl(0, 0%, 33%)",marginTop:"0.75em",zIndex:"0"},".line-highlight.line-highlight:before":{backgroundColor:"hsl(215, 15%, 59%)",color:"hsl(24, 20%, 95%)"},".line-highlight.line-highlight[data-end]:after":{backgroundColor:"hsl(215, 15%, 59%)",color:"hsl(24, 20%, 95%)"}}}(yx)),yx}var vx={},BP;function IEe(){return BP||(BP=1,function(e){Object.defineProperty(e,"__esModule",{value:!0}),e.default=void 0,e.default={'code[class*="language-"]':{color:"black",background:"none",textShadow:"0 1px white",fontFamily:"Consolas, Monaco, 'Andale Mono', 'Ubuntu Mono', monospace",fontSize:"1em",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",wordWrap:"normal",lineHeight:"1.5",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none"},'pre[class*="language-"]':{color:"black",background:"#f5f2f0",textShadow:"0 1px white",fontFamily:"Consolas, Monaco, 'Andale Mono', 'Ubuntu Mono', monospace",fontSize:"1em",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",wordWrap:"normal",lineHeight:"1.5",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none",padding:"1em",margin:".5em 0",overflow:"auto"},'pre[class*="language-"]::-moz-selection':{textShadow:"none",background:"#b3d4fc"},'pre[class*="language-"] ::-moz-selection':{textShadow:"none",background:"#b3d4fc"},'code[class*="language-"]::-moz-selection':{textShadow:"none",background:"#b3d4fc"},'code[class*="language-"] ::-moz-selection':{textShadow:"none",background:"#b3d4fc"},'pre[class*="language-"]::selection':{textShadow:"none",background:"#b3d4fc"},'pre[class*="language-"] ::selection':{textShadow:"none",background:"#b3d4fc"},'code[class*="language-"]::selection':{textShadow:"none",background:"#b3d4fc"},'code[class*="language-"] ::selection':{textShadow:"none",background:"#b3d4fc"},':not(pre) > code[class*="language-"]':{background:"#f5f2f0",padding:".1em",borderRadius:".3em",whiteSpace:"normal"},comment:{color:"slategray"},prolog:{color:"slategray"},doctype:{color:"slategray"},cdata:{color:"slategray"},punctuation:{color:"#999"},namespace:{Opacity:".7"},property:{color:"#905"},tag:{color:"#905"},boolean:{color:"#905"},number:{color:"#905"},constant:{color:"#905"},symbol:{color:"#905"},deleted:{color:"#905"},selector:{color:"#690"},"attr-name":{color:"#690"},string:{color:"#690"},char:{color:"#690"},builtin:{color:"#690"},inserted:{color:"#690"},operator:{color:"#9a6e3a",background:"hsla(0, 0%, 100%, .5)"},entity:{color:"#9a6e3a",background:"hsla(0, 0%, 100%, .5)",cursor:"help"},url:{color:"#9a6e3a",background:"hsla(0, 0%, 100%, .5)"},".language-css .token.string":{color:"#9a6e3a",background:"hsla(0, 0%, 100%, .5)"},".style .token.string":{color:"#9a6e3a",background:"hsla(0, 0%, 100%, .5)"},atrule:{color:"#07a"},"attr-value":{color:"#07a"},keyword:{color:"#07a"},function:{color:"#DD4A68"},"class-name":{color:"#DD4A68"},regex:{color:"#e90"},important:{color:"#e90",fontWeight:"bold"},variable:{color:"#e90"},bold:{fontWeight:"bold"},italic:{fontStyle:"italic"}}}(vx)),vx}var Sx={},UP;function LEe(){return UP||(UP=1,function(e){Object.defineProperty(e,"__esModule",{value:!0}),e.default=void 0,e.default={'code[class*="language-"]':{color:"#f8f8f2",background:"none",fontFamily:"Consolas, Monaco, 'Andale Mono', 'Ubuntu Mono', monospace",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",wordWrap:"normal",lineHeight:"1.5",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none"},'pre[class*="language-"]':{color:"#f8f8f2",background:"#2b2b2b",fontFamily:"Consolas, Monaco, 'Andale Mono', 'Ubuntu Mono', monospace",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",wordWrap:"normal",lineHeight:"1.5",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none",padding:"1em",margin:"0.5em 0",overflow:"auto",borderRadius:"0.3em"},':not(pre) > code[class*="language-"]':{background:"#2b2b2b",padding:"0.1em",borderRadius:"0.3em",whiteSpace:"normal"},comment:{color:"#d4d0ab"},prolog:{color:"#d4d0ab"},doctype:{color:"#d4d0ab"},cdata:{color:"#d4d0ab"},punctuation:{color:"#fefefe"},property:{color:"#ffa07a"},tag:{color:"#ffa07a"},constant:{color:"#ffa07a"},symbol:{color:"#ffa07a"},deleted:{color:"#ffa07a"},boolean:{color:"#00e0e0"},number:{color:"#00e0e0"},selector:{color:"#abe338"},"attr-name":{color:"#abe338"},string:{color:"#abe338"},char:{color:"#abe338"},builtin:{color:"#abe338"},inserted:{color:"#abe338"},operator:{color:"#00e0e0"},entity:{color:"#00e0e0",cursor:"help"},url:{color:"#00e0e0"},".language-css .token.string":{color:"#00e0e0"},".style .token.string":{color:"#00e0e0"},variable:{color:"#00e0e0"},atrule:{color:"#ffd700"},"attr-value":{color:"#ffd700"},function:{color:"#ffd700"},keyword:{color:"#00e0e0"},regex:{color:"#ffd700"},important:{color:"#ffd700",fontWeight:"bold"},bold:{fontWeight:"bold"},italic:{fontStyle:"italic"}}}(Sx)),Sx}var Ex={},jP;function MEe(){return jP||(jP=1,function(e){Object.defineProperty(e,"__esModule",{value:!0}),e.default=void 0,e.default={'code[class*="language-"]':{color:"#c5c8c6",textShadow:"0 1px rgba(0, 0, 0, 0.3)",fontFamily:"Inconsolata, Monaco, Consolas, 'Courier New', Courier, monospace",direction:"ltr",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",lineHeight:"1.5",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none"},'pre[class*="language-"]':{color:"#c5c8c6",textShadow:"0 1px rgba(0, 0, 0, 0.3)",fontFamily:"Inconsolata, Monaco, Consolas, 'Courier New', Courier, monospace",direction:"ltr",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",lineHeight:"1.5",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none",padding:"1em",margin:".5em 0",overflow:"auto",borderRadius:"0.3em",background:"#1d1f21"},':not(pre) > code[class*="language-"]':{background:"#1d1f21",padding:".1em",borderRadius:".3em"},comment:{color:"#7C7C7C"},prolog:{color:"#7C7C7C"},doctype:{color:"#7C7C7C"},cdata:{color:"#7C7C7C"},punctuation:{color:"#c5c8c6"},".namespace":{Opacity:".7"},property:{color:"#96CBFE"},keyword:{color:"#96CBFE"},tag:{color:"#96CBFE"},"class-name":{color:"#FFFFB6",textDecoration:"underline"},boolean:{color:"#99CC99"},constant:{color:"#99CC99"},symbol:{color:"#f92672"},deleted:{color:"#f92672"},number:{color:"#FF73FD"},selector:{color:"#A8FF60"},"attr-name":{color:"#A8FF60"},string:{color:"#A8FF60"},char:{color:"#A8FF60"},builtin:{color:"#A8FF60"},inserted:{color:"#A8FF60"},variable:{color:"#C6C5FE"},operator:{color:"#EDEDED"},entity:{color:"#FFFFB6",cursor:"help"},url:{color:"#96CBFE"},".language-css .token.string":{color:"#87C38A"},".style .token.string":{color:"#87C38A"},atrule:{color:"#F9EE98"},"attr-value":{color:"#F9EE98"},function:{color:"#DAD085"},regex:{color:"#E9C062"},important:{color:"#fd971f",fontWeight:"bold"},bold:{fontWeight:"bold"},italic:{fontStyle:"italic"}}}(Ex)),Ex}var wx={},GP;function FEe(){return GP||(GP=1,function(e){Object.defineProperty(e,"__esModule",{value:!0}),e.default=void 0,e.default={'code[class*="language-"]':{fontFamily:'Consolas, Menlo, Monaco, "Andale Mono WT", "Andale Mono", "Lucida Console", "Lucida Sans Typewriter", "DejaVu Sans Mono", "Bitstream Vera Sans Mono", "Liberation Mono", "Nimbus Mono L", "Courier New", Courier, monospace',fontSize:"14px",lineHeight:"1.375",direction:"ltr",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none",background:"#f5f7ff",color:"#5e6687"},'pre[class*="language-"]':{fontFamily:'Consolas, Menlo, Monaco, "Andale Mono WT", "Andale Mono", "Lucida Console", "Lucida Sans Typewriter", "DejaVu Sans Mono", "Bitstream Vera Sans Mono", "Liberation Mono", "Nimbus Mono L", "Courier New", Courier, monospace',fontSize:"14px",lineHeight:"1.375",direction:"ltr",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none",background:"#f5f7ff",color:"#5e6687",padding:"1em",margin:".5em 0",overflow:"auto"},'pre > code[class*="language-"]':{fontSize:"1em"},'pre[class*="language-"]::-moz-selection':{textShadow:"none",background:"#dfe2f1"},'pre[class*="language-"] ::-moz-selection':{textShadow:"none",background:"#dfe2f1"},'code[class*="language-"]::-moz-selection':{textShadow:"none",background:"#dfe2f1"},'code[class*="language-"] ::-moz-selection':{textShadow:"none",background:"#dfe2f1"},'pre[class*="language-"]::selection':{textShadow:"none",background:"#dfe2f1"},'pre[class*="language-"] ::selection':{textShadow:"none",background:"#dfe2f1"},'code[class*="language-"]::selection':{textShadow:"none",background:"#dfe2f1"},'code[class*="language-"] ::selection':{textShadow:"none",background:"#dfe2f1"},':not(pre) > code[class*="language-"]':{padding:".1em",borderRadius:".3em"},comment:{color:"#898ea4"},prolog:{color:"#898ea4"},doctype:{color:"#898ea4"},cdata:{color:"#898ea4"},punctuation:{color:"#5e6687"},namespace:{Opacity:".7"},operator:{color:"#c76b29"},boolean:{color:"#c76b29"},number:{color:"#c76b29"},property:{color:"#c08b30"},tag:{color:"#3d8fd1"},string:{color:"#22a2c9"},selector:{color:"#6679cc"},"attr-name":{color:"#c76b29"},entity:{color:"#22a2c9",cursor:"help"},url:{color:"#22a2c9"},".language-css .token.string":{color:"#22a2c9"},".style .token.string":{color:"#22a2c9"},"attr-value":{color:"#ac9739"},keyword:{color:"#ac9739"},control:{color:"#ac9739"},directive:{color:"#ac9739"},unit:{color:"#ac9739"},statement:{color:"#22a2c9"},regex:{color:"#22a2c9"},atrule:{color:"#22a2c9"},placeholder:{color:"#3d8fd1"},variable:{color:"#3d8fd1"},deleted:{textDecoration:"line-through"},inserted:{borderBottom:"1px dotted #202746",textDecoration:"none"},italic:{fontStyle:"italic"},important:{fontWeight:"bold",color:"#c94922"},bold:{fontWeight:"bold"},"pre > code.highlight":{Outline:"0.4em solid #c94922",OutlineOffset:".4em"},".line-numbers.line-numbers .line-numbers-rows":{borderRightColor:"#dfe2f1"},".line-numbers .line-numbers-rows > span:before":{color:"#979db4"},".line-highlight.line-highlight":{background:"linear-gradient(to right, rgba(107, 115, 148, 0.2) 70%, rgba(107, 115, 148, 0))"}}}(wx)),wx}var xx={},HP;function PEe(){return HP||(HP=1,function(e){Object.defineProperty(e,"__esModule",{value:!0}),e.default=void 0,e.default={'code[class*="language-"]':{color:"#fff",textShadow:"0 1px 1px #000",fontFamily:'Menlo, Monaco, "Courier New", monospace',direction:"ltr",textAlign:"left",wordSpacing:"normal",whiteSpace:"pre",wordWrap:"normal",lineHeight:"1.4",background:"none",border:"0",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none"},'pre[class*="language-"]':{color:"#fff",textShadow:"0 1px 1px #000",fontFamily:'Menlo, Monaco, "Courier New", monospace',direction:"ltr",textAlign:"left",wordSpacing:"normal",whiteSpace:"pre",wordWrap:"normal",lineHeight:"1.4",background:"#222",border:"0",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none",padding:"15px",margin:"1em 0",overflow:"auto",MozBorderRadius:"8px",WebkitBorderRadius:"8px",borderRadius:"8px"},'pre[class*="language-"] code':{float:"left",padding:"0 15px 0 0"},':not(pre) > code[class*="language-"]':{background:"#222",padding:"5px 10px",lineHeight:"1",MozBorderRadius:"3px",WebkitBorderRadius:"3px",borderRadius:"3px"},comment:{color:"#797979"},prolog:{color:"#797979"},doctype:{color:"#797979"},cdata:{color:"#797979"},selector:{color:"#fff"},operator:{color:"#fff"},punctuation:{color:"#fff"},namespace:{Opacity:".7"},tag:{color:"#ffd893"},boolean:{color:"#ffd893"},atrule:{color:"#B0C975"},"attr-value":{color:"#B0C975"},hex:{color:"#B0C975"},string:{color:"#B0C975"},property:{color:"#c27628"},entity:{color:"#c27628",cursor:"help"},url:{color:"#c27628"},"attr-name":{color:"#c27628"},keyword:{color:"#c27628"},regex:{color:"#9B71C6"},function:{color:"#e5a638"},constant:{color:"#e5a638"},variable:{color:"#fdfba8"},number:{color:"#8799B0"},important:{color:"#E45734"},deliminator:{color:"#E45734"},".line-highlight.line-highlight":{background:"rgba(255, 255, 255, .2)"},".line-highlight.line-highlight:before":{top:".3em",backgroundColor:"rgba(255, 255, 255, .3)",color:"#fff",MozBorderRadius:"8px",WebkitBorderRadius:"8px",borderRadius:"8px"},".line-highlight.line-highlight[data-end]:after":{top:".3em",backgroundColor:"rgba(255, 255, 255, .3)",color:"#fff",MozBorderRadius:"8px",WebkitBorderRadius:"8px",borderRadius:"8px"},".line-numbers .line-numbers-rows > span":{borderRight:"3px #d9d336 solid"}}}(xx)),xx}var kx={},$P;function zEe(){return $P||($P=1,function(e){Object.defineProperty(e,"__esModule",{value:!0}),e.default=void 0,e.default={'code[class*="language-"]':{color:"#111b27",background:"none",fontFamily:'Consolas, Monaco, "Andale Mono", "Ubuntu Mono", monospace',textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",wordWrap:"normal",lineHeight:"1.5",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none"},'pre[class*="language-"]':{color:"#111b27",background:"#e3eaf2",fontFamily:'Consolas, Monaco, "Andale Mono", "Ubuntu Mono", monospace',textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",wordWrap:"normal",lineHeight:"1.5",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none",padding:"1em",margin:"0.5em 0",overflow:"auto"},'pre[class*="language-"]::-moz-selection':{background:"#8da1b9"},'pre[class*="language-"] ::-moz-selection':{background:"#8da1b9"},'code[class*="language-"]::-moz-selection':{background:"#8da1b9"},'code[class*="language-"] ::-moz-selection':{background:"#8da1b9"},'pre[class*="language-"]::selection':{background:"#8da1b9"},'pre[class*="language-"] ::selection':{background:"#8da1b9"},'code[class*="language-"]::selection':{background:"#8da1b9"},'code[class*="language-"] ::selection':{background:"#8da1b9"},':not(pre) > code[class*="language-"]':{background:"#e3eaf2",padding:"0.1em 0.3em",borderRadius:"0.3em",whiteSpace:"normal"},comment:{color:"#3c526d"},prolog:{color:"#3c526d"},doctype:{color:"#3c526d"},cdata:{color:"#3c526d"},punctuation:{color:"#111b27"},"delimiter.important":{color:"#006d6d",fontWeight:"inherit"},"selector.parent":{color:"#006d6d"},tag:{color:"#006d6d"},"tag.punctuation":{color:"#006d6d"},"attr-name":{color:"#755f00"},boolean:{color:"#755f00"},"boolean.important":{color:"#755f00"},number:{color:"#755f00"},constant:{color:"#755f00"},"selector.attribute":{color:"#755f00"},"class-name":{color:"#005a8e"},key:{color:"#005a8e"},parameter:{color:"#005a8e"},property:{color:"#005a8e"},"property-access":{color:"#005a8e"},variable:{color:"#005a8e"},"attr-value":{color:"#116b00"},inserted:{color:"#116b00"},color:{color:"#116b00"},"selector.value":{color:"#116b00"},string:{color:"#116b00"},"string.url-link":{color:"#116b00"},builtin:{color:"#af00af"},"keyword-array":{color:"#af00af"},package:{color:"#af00af"},regex:{color:"#af00af"},function:{color:"#7c00aa"},"selector.class":{color:"#7c00aa"},"selector.id":{color:"#7c00aa"},"atrule.rule":{color:"#a04900"},combinator:{color:"#a04900"},keyword:{color:"#a04900"},operator:{color:"#a04900"},"pseudo-class":{color:"#a04900"},"pseudo-element":{color:"#a04900"},selector:{color:"#a04900"},unit:{color:"#a04900"},deleted:{color:"#c22f2e"},important:{color:"#c22f2e",fontWeight:"bold"},"keyword-this":{color:"#005a8e",fontWeight:"bold"},this:{color:"#005a8e",fontWeight:"bold"},bold:{fontWeight:"bold"},italic:{fontStyle:"italic"},entity:{cursor:"help"},".language-markdown .token.title":{color:"#005a8e",fontWeight:"bold"},".language-markdown .token.title .token.punctuation":{color:"#005a8e",fontWeight:"bold"},".language-markdown .token.blockquote.punctuation":{color:"#af00af"},".language-markdown .token.code":{color:"#006d6d"},".language-markdown .token.hr.punctuation":{color:"#005a8e"},".language-markdown .token.url > .token.content":{color:"#116b00"},".language-markdown .token.url-link":{color:"#755f00"},".language-markdown .token.list.punctuation":{color:"#af00af"},".language-markdown .token.table-header":{color:"#111b27"},".language-json .token.operator":{color:"#111b27"},".language-scss .token.variable":{color:"#006d6d"},"token.tab:not(:empty):before":{color:"#3c526d"},"token.cr:before":{color:"#3c526d"},"token.lf:before":{color:"#3c526d"},"token.space:before":{color:"#3c526d"},"div.code-toolbar > .toolbar.toolbar > .toolbar-item > a":{color:"#e3eaf2",background:"#005a8e"},"div.code-toolbar > .toolbar.toolbar > .toolbar-item > button":{color:"#e3eaf2",background:"#005a8e"},"div.code-toolbar > .toolbar.toolbar > .toolbar-item > a:hover":{color:"#e3eaf2",background:"#005a8eda",textDecoration:"none"},"div.code-toolbar > .toolbar.toolbar > .toolbar-item > a:focus":{color:"#e3eaf2",background:"#005a8eda",textDecoration:"none"},"div.code-toolbar > .toolbar.toolbar > .toolbar-item > button:hover":{color:"#e3eaf2",background:"#005a8eda",textDecoration:"none"},"div.code-toolbar > .toolbar.toolbar > .toolbar-item > button:focus":{color:"#e3eaf2",background:"#005a8eda",textDecoration:"none"},"div.code-toolbar > .toolbar.toolbar > .toolbar-item > span":{color:"#e3eaf2",background:"#3c526d"},"div.code-toolbar > .toolbar.toolbar > .toolbar-item > span:hover":{color:"#e3eaf2",background:"#3c526d"},"div.code-toolbar > .toolbar.toolbar > .toolbar-item > span:focus":{color:"#e3eaf2",background:"#3c526d"},".line-highlight.line-highlight":{background:"linear-gradient(to right, #8da1b92f 70%, #8da1b925)"},".line-highlight.line-highlight:before":{backgroundColor:"#3c526d",color:"#e3eaf2",boxShadow:"0 1px #8da1b9"},".line-highlight.line-highlight[data-end]:after":{backgroundColor:"#3c526d",color:"#e3eaf2",boxShadow:"0 1px #8da1b9"},"pre[id].linkable-line-numbers.linkable-line-numbers span.line-numbers-rows > span:hover:before":{backgroundColor:"#3c526d1f"},".line-numbers.line-numbers .line-numbers-rows":{borderRight:"1px solid #8da1b97a",background:"#d0dae77a"},".line-numbers .line-numbers-rows > span:before":{color:"#3c526dda"},".rainbow-braces .token.token.punctuation.brace-level-1":{color:"#755f00"},".rainbow-braces .token.token.punctuation.brace-level-5":{color:"#755f00"},".rainbow-braces .token.token.punctuation.brace-level-9":{color:"#755f00"},".rainbow-braces .token.token.punctuation.brace-level-2":{color:"#af00af"},".rainbow-braces .token.token.punctuation.brace-level-6":{color:"#af00af"},".rainbow-braces .token.token.punctuation.brace-level-10":{color:"#af00af"},".rainbow-braces .token.token.punctuation.brace-level-3":{color:"#005a8e"},".rainbow-braces .token.token.punctuation.brace-level-7":{color:"#005a8e"},".rainbow-braces .token.token.punctuation.brace-level-11":{color:"#005a8e"},".rainbow-braces .token.token.punctuation.brace-level-4":{color:"#7c00aa"},".rainbow-braces .token.token.punctuation.brace-level-8":{color:"#7c00aa"},".rainbow-braces .token.token.punctuation.brace-level-12":{color:"#7c00aa"},"pre.diff-highlight > code .token.token.deleted:not(.prefix)":{backgroundColor:"#c22f2e1f"},"pre > code.diff-highlight .token.token.deleted:not(.prefix)":{backgroundColor:"#c22f2e1f"},"pre.diff-highlight > code .token.token.inserted:not(.prefix)":{backgroundColor:"#116b001f"},"pre > code.diff-highlight .token.token.inserted:not(.prefix)":{backgroundColor:"#116b001f"},".command-line .command-line-prompt":{borderRight:"1px solid #8da1b97a"},".command-line .command-line-prompt > span:before":{color:"#3c526dda"}}}(kx)),kx}var Tx={},qP;function BEe(){return qP||(qP=1,function(e){Object.defineProperty(e,"__esModule",{value:!0}),e.default=void 0,e.default={'code[class*="language-"]':{color:"#e3eaf2",background:"none",fontFamily:'Consolas, Monaco, "Andale Mono", "Ubuntu Mono", monospace',textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",wordWrap:"normal",lineHeight:"1.5",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none"},'pre[class*="language-"]':{color:"#e3eaf2",background:"#111b27",fontFamily:'Consolas, Monaco, "Andale Mono", "Ubuntu Mono", monospace',textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",wordWrap:"normal",lineHeight:"1.5",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none",padding:"1em",margin:"0.5em 0",overflow:"auto"},'pre[class*="language-"]::-moz-selection':{background:"#3c526d"},'pre[class*="language-"] ::-moz-selection':{background:"#3c526d"},'code[class*="language-"]::-moz-selection':{background:"#3c526d"},'code[class*="language-"] ::-moz-selection':{background:"#3c526d"},'pre[class*="language-"]::selection':{background:"#3c526d"},'pre[class*="language-"] ::selection':{background:"#3c526d"},'code[class*="language-"]::selection':{background:"#3c526d"},'code[class*="language-"] ::selection':{background:"#3c526d"},':not(pre) > code[class*="language-"]':{background:"#111b27",padding:"0.1em 0.3em",borderRadius:"0.3em",whiteSpace:"normal"},comment:{color:"#8da1b9"},prolog:{color:"#8da1b9"},doctype:{color:"#8da1b9"},cdata:{color:"#8da1b9"},punctuation:{color:"#e3eaf2"},"delimiter.important":{color:"#66cccc",fontWeight:"inherit"},"selector.parent":{color:"#66cccc"},tag:{color:"#66cccc"},"tag.punctuation":{color:"#66cccc"},"attr-name":{color:"#e6d37a"},boolean:{color:"#e6d37a"},"boolean.important":{color:"#e6d37a"},number:{color:"#e6d37a"},constant:{color:"#e6d37a"},"selector.attribute":{color:"#e6d37a"},"class-name":{color:"#6cb8e6"},key:{color:"#6cb8e6"},parameter:{color:"#6cb8e6"},property:{color:"#6cb8e6"},"property-access":{color:"#6cb8e6"},variable:{color:"#6cb8e6"},"attr-value":{color:"#91d076"},inserted:{color:"#91d076"},color:{color:"#91d076"},"selector.value":{color:"#91d076"},string:{color:"#91d076"},"string.url-link":{color:"#91d076"},builtin:{color:"#f4adf4"},"keyword-array":{color:"#f4adf4"},package:{color:"#f4adf4"},regex:{color:"#f4adf4"},function:{color:"#c699e3"},"selector.class":{color:"#c699e3"},"selector.id":{color:"#c699e3"},"atrule.rule":{color:"#e9ae7e"},combinator:{color:"#e9ae7e"},keyword:{color:"#e9ae7e"},operator:{color:"#e9ae7e"},"pseudo-class":{color:"#e9ae7e"},"pseudo-element":{color:"#e9ae7e"},selector:{color:"#e9ae7e"},unit:{color:"#e9ae7e"},deleted:{color:"#cd6660"},important:{color:"#cd6660",fontWeight:"bold"},"keyword-this":{color:"#6cb8e6",fontWeight:"bold"},this:{color:"#6cb8e6",fontWeight:"bold"},bold:{fontWeight:"bold"},italic:{fontStyle:"italic"},entity:{cursor:"help"},".language-markdown .token.title":{color:"#6cb8e6",fontWeight:"bold"},".language-markdown .token.title .token.punctuation":{color:"#6cb8e6",fontWeight:"bold"},".language-markdown .token.blockquote.punctuation":{color:"#f4adf4"},".language-markdown .token.code":{color:"#66cccc"},".language-markdown .token.hr.punctuation":{color:"#6cb8e6"},".language-markdown .token.url .token.content":{color:"#91d076"},".language-markdown .token.url-link":{color:"#e6d37a"},".language-markdown .token.list.punctuation":{color:"#f4adf4"},".language-markdown .token.table-header":{color:"#e3eaf2"},".language-json .token.operator":{color:"#e3eaf2"},".language-scss .token.variable":{color:"#66cccc"},"token.tab:not(:empty):before":{color:"#8da1b9"},"token.cr:before":{color:"#8da1b9"},"token.lf:before":{color:"#8da1b9"},"token.space:before":{color:"#8da1b9"},"div.code-toolbar > .toolbar.toolbar > .toolbar-item > a":{color:"#111b27",background:"#6cb8e6"},"div.code-toolbar > .toolbar.toolbar > .toolbar-item > button":{color:"#111b27",background:"#6cb8e6"},"div.code-toolbar > .toolbar.toolbar > .toolbar-item > a:hover":{color:"#111b27",background:"#6cb8e6da",textDecoration:"none"},"div.code-toolbar > .toolbar.toolbar > .toolbar-item > a:focus":{color:"#111b27",background:"#6cb8e6da",textDecoration:"none"},"div.code-toolbar > .toolbar.toolbar > .toolbar-item > button:hover":{color:"#111b27",background:"#6cb8e6da",textDecoration:"none"},"div.code-toolbar > .toolbar.toolbar > .toolbar-item > button:focus":{color:"#111b27",background:"#6cb8e6da",textDecoration:"none"},"div.code-toolbar > .toolbar.toolbar > .toolbar-item > span":{color:"#111b27",background:"#8da1b9"},"div.code-toolbar > .toolbar.toolbar > .toolbar-item > span:hover":{color:"#111b27",background:"#8da1b9"},"div.code-toolbar > .toolbar.toolbar > .toolbar-item > span:focus":{color:"#111b27",background:"#8da1b9"},".line-highlight.line-highlight":{background:"linear-gradient(to right, #3c526d5f 70%, #3c526d55)"},".line-highlight.line-highlight:before":{backgroundColor:"#8da1b9",color:"#111b27",boxShadow:"0 1px #3c526d"},".line-highlight.line-highlight[data-end]:after":{backgroundColor:"#8da1b9",color:"#111b27",boxShadow:"0 1px #3c526d"},"pre[id].linkable-line-numbers.linkable-line-numbers span.line-numbers-rows > span:hover:before":{backgroundColor:"#8da1b918"},".line-numbers.line-numbers .line-numbers-rows":{borderRight:"1px solid #0b121b",background:"#0b121b7a"},".line-numbers .line-numbers-rows > span:before":{color:"#8da1b9da"},".rainbow-braces .token.token.punctuation.brace-level-1":{color:"#e6d37a"},".rainbow-braces .token.token.punctuation.brace-level-5":{color:"#e6d37a"},".rainbow-braces .token.token.punctuation.brace-level-9":{color:"#e6d37a"},".rainbow-braces .token.token.punctuation.brace-level-2":{color:"#f4adf4"},".rainbow-braces .token.token.punctuation.brace-level-6":{color:"#f4adf4"},".rainbow-braces .token.token.punctuation.brace-level-10":{color:"#f4adf4"},".rainbow-braces .token.token.punctuation.brace-level-3":{color:"#6cb8e6"},".rainbow-braces .token.token.punctuation.brace-level-7":{color:"#6cb8e6"},".rainbow-braces .token.token.punctuation.brace-level-11":{color:"#6cb8e6"},".rainbow-braces .token.token.punctuation.brace-level-4":{color:"#c699e3"},".rainbow-braces .token.token.punctuation.brace-level-8":{color:"#c699e3"},".rainbow-braces .token.token.punctuation.brace-level-12":{color:"#c699e3"},"pre.diff-highlight > code .token.token.deleted:not(.prefix)":{backgroundColor:"#cd66601f"},"pre > code.diff-highlight .token.token.deleted:not(.prefix)":{backgroundColor:"#cd66601f"},"pre.diff-highlight > code .token.token.inserted:not(.prefix)":{backgroundColor:"#91d0761f"},"pre > code.diff-highlight .token.token.inserted:not(.prefix)":{backgroundColor:"#91d0761f"},".command-line .command-line-prompt":{borderRight:"1px solid #0b121b"},".command-line .command-line-prompt > span:before":{color:"#8da1b9da"}}}(Tx)),Tx}var Ax={},VP;function UEe(){return VP||(VP=1,function(e){Object.defineProperty(e,"__esModule",{value:!0}),e.default=void 0,e.default={'code[class*="language-"]':{color:"black",background:"none",fontFamily:"Consolas, Monaco, 'Andale Mono', 'Ubuntu Mono', monospace",fontSize:"1em",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",wordWrap:"normal",lineHeight:"1.5",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none"},'pre[class*="language-"]':{color:"black",background:"none",fontFamily:"Consolas, Monaco, 'Andale Mono', 'Ubuntu Mono', monospace",fontSize:"1em",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",wordWrap:"normal",lineHeight:"1.5",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none",position:"relative",borderLeft:"10px solid #358ccb",boxShadow:"-1px 0 0 0 #358ccb, 0 0 0 1px #dfdfdf",backgroundColor:"#fdfdfd",backgroundImage:"linear-gradient(transparent 50%, rgba(69, 142, 209, 0.04) 50%)",backgroundSize:"3em 3em",backgroundOrigin:"content-box",backgroundAttachment:"local",margin:".5em 0",padding:"0 1em"},'pre[class*="language-"] > code':{display:"block"},':not(pre) > code[class*="language-"]':{position:"relative",padding:".2em",borderRadius:"0.3em",color:"#c92c2c",border:"1px solid rgba(0, 0, 0, 0.1)",display:"inline",whiteSpace:"normal",backgroundColor:"#fdfdfd",WebkitBoxSizing:"border-box",MozBoxSizing:"border-box",boxSizing:"border-box"},comment:{color:"#7D8B99"},"block-comment":{color:"#7D8B99"},prolog:{color:"#7D8B99"},doctype:{color:"#7D8B99"},cdata:{color:"#7D8B99"},punctuation:{color:"#5F6364"},property:{color:"#c92c2c"},tag:{color:"#c92c2c"},boolean:{color:"#c92c2c"},number:{color:"#c92c2c"},"function-name":{color:"#c92c2c"},constant:{color:"#c92c2c"},symbol:{color:"#c92c2c"},deleted:{color:"#c92c2c"},selector:{color:"#2f9c0a"},"attr-name":{color:"#2f9c0a"},string:{color:"#2f9c0a"},char:{color:"#2f9c0a"},function:{color:"#2f9c0a"},builtin:{color:"#2f9c0a"},inserted:{color:"#2f9c0a"},operator:{color:"#a67f59",background:"rgba(255, 255, 255, 0.5)"},entity:{color:"#a67f59",background:"rgba(255, 255, 255, 0.5)",cursor:"help"},url:{color:"#a67f59",background:"rgba(255, 255, 255, 0.5)"},variable:{color:"#a67f59",background:"rgba(255, 255, 255, 0.5)"},atrule:{color:"#1990b8"},"attr-value":{color:"#1990b8"},keyword:{color:"#1990b8"},"class-name":{color:"#1990b8"},regex:{color:"#e90"},important:{color:"#e90",fontWeight:"normal"},".language-css .token.string":{color:"#a67f59",background:"rgba(255, 255, 255, 0.5)"},".style .token.string":{color:"#a67f59",background:"rgba(255, 255, 255, 0.5)"},bold:{fontWeight:"bold"},italic:{fontStyle:"italic"},namespace:{Opacity:".7"}}}(Ax)),Ax}var Rx={},WP;function jEe(){return WP||(WP=1,function(e){Object.defineProperty(e,"__esModule",{value:!0}),e.default=void 0,e.default={'code[class*="language-"]':{color:"#a9b7c6",fontFamily:"Consolas, Monaco, 'Andale Mono', monospace",direction:"ltr",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",lineHeight:"1.5",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none"},'pre[class*="language-"]':{color:"#a9b7c6",fontFamily:"Consolas, Monaco, 'Andale Mono', monospace",direction:"ltr",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",lineHeight:"1.5",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none",padding:"1em",margin:".5em 0",overflow:"auto",background:"#2b2b2b"},'pre[class*="language-"]::-moz-selection':{color:"inherit",background:"rgba(33, 66, 131, .85)"},'pre[class*="language-"] ::-moz-selection':{color:"inherit",background:"rgba(33, 66, 131, .85)"},'code[class*="language-"]::-moz-selection':{color:"inherit",background:"rgba(33, 66, 131, .85)"},'code[class*="language-"] ::-moz-selection':{color:"inherit",background:"rgba(33, 66, 131, .85)"},'pre[class*="language-"]::selection':{color:"inherit",background:"rgba(33, 66, 131, .85)"},'pre[class*="language-"] ::selection':{color:"inherit",background:"rgba(33, 66, 131, .85)"},'code[class*="language-"]::selection':{color:"inherit",background:"rgba(33, 66, 131, .85)"},'code[class*="language-"] ::selection':{color:"inherit",background:"rgba(33, 66, 131, .85)"},':not(pre) > code[class*="language-"]':{background:"#2b2b2b",padding:".1em",borderRadius:".3em"},comment:{color:"#808080"},prolog:{color:"#808080"},cdata:{color:"#808080"},delimiter:{color:"#cc7832"},boolean:{color:"#cc7832"},keyword:{color:"#cc7832"},selector:{color:"#cc7832"},important:{color:"#cc7832"},atrule:{color:"#cc7832"},operator:{color:"#a9b7c6"},punctuation:{color:"#a9b7c6"},"attr-name":{color:"#a9b7c6"},tag:{color:"#e8bf6a"},"tag.punctuation":{color:"#e8bf6a"},doctype:{color:"#e8bf6a"},builtin:{color:"#e8bf6a"},entity:{color:"#6897bb"},number:{color:"#6897bb"},symbol:{color:"#6897bb"},property:{color:"#9876aa"},constant:{color:"#9876aa"},variable:{color:"#9876aa"},string:{color:"#6a8759"},char:{color:"#6a8759"},"attr-value":{color:"#a5c261"},"attr-value.punctuation":{color:"#a5c261"},"attr-value.punctuation:first-child":{color:"#a9b7c6"},url:{color:"#287bde",textDecoration:"underline"},function:{color:"#ffc66d"},regex:{background:"#364135"},bold:{fontWeight:"bold"},italic:{fontStyle:"italic"},inserted:{background:"#294436"},deleted:{background:"#484a4a"},"code.language-css .token.property":{color:"#a9b7c6"},"code.language-css .token.property + .token.punctuation":{color:"#a9b7c6"},"code.language-css .token.id":{color:"#ffc66d"},"code.language-css .token.selector > .token.class":{color:"#ffc66d"},"code.language-css .token.selector > .token.attribute":{color:"#ffc66d"},"code.language-css .token.selector > .token.pseudo-class":{color:"#ffc66d"},"code.language-css .token.selector > .token.pseudo-element":{color:"#ffc66d"}}}(Rx)),Rx}var _x={},YP;function GEe(){return YP||(YP=1,function(e){Object.defineProperty(e,"__esModule",{value:!0}),e.default=void 0,e.default={'code[class*="language-"]':{color:"#f8f8f2",background:"none",textShadow:"0 1px rgba(0, 0, 0, 0.3)",fontFamily:"Consolas, Monaco, 'Andale Mono', 'Ubuntu Mono', monospace",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",wordWrap:"normal",lineHeight:"1.5",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none"},'pre[class*="language-"]':{color:"#f8f8f2",background:"#282a36",textShadow:"0 1px rgba(0, 0, 0, 0.3)",fontFamily:"Consolas, Monaco, 'Andale Mono', 'Ubuntu Mono', monospace",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",wordWrap:"normal",lineHeight:"1.5",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none",padding:"1em",margin:".5em 0",overflow:"auto",borderRadius:"0.3em"},':not(pre) > code[class*="language-"]':{background:"#282a36",padding:".1em",borderRadius:".3em",whiteSpace:"normal"},comment:{color:"#6272a4"},prolog:{color:"#6272a4"},doctype:{color:"#6272a4"},cdata:{color:"#6272a4"},punctuation:{color:"#f8f8f2"},".namespace":{Opacity:".7"},property:{color:"#ff79c6"},tag:{color:"#ff79c6"},constant:{color:"#ff79c6"},symbol:{color:"#ff79c6"},deleted:{color:"#ff79c6"},boolean:{color:"#bd93f9"},number:{color:"#bd93f9"},selector:{color:"#50fa7b"},"attr-name":{color:"#50fa7b"},string:{color:"#50fa7b"},char:{color:"#50fa7b"},builtin:{color:"#50fa7b"},inserted:{color:"#50fa7b"},operator:{color:"#f8f8f2"},entity:{color:"#f8f8f2",cursor:"help"},url:{color:"#f8f8f2"},".language-css .token.string":{color:"#f8f8f2"},".style .token.string":{color:"#f8f8f2"},variable:{color:"#f8f8f2"},atrule:{color:"#f1fa8c"},"attr-value":{color:"#f1fa8c"},function:{color:"#f1fa8c"},"class-name":{color:"#f1fa8c"},keyword:{color:"#8be9fd"},regex:{color:"#ffb86c"},important:{color:"#ffb86c",fontWeight:"bold"},bold:{fontWeight:"bold"},italic:{fontStyle:"italic"}}}(_x)),_x}var Cx={},KP;function HEe(){return KP||(KP=1,function(e){Object.defineProperty(e,"__esModule",{value:!0}),e.default=void 0,e.default={'code[class*="language-"]':{fontFamily:'Consolas, Menlo, Monaco, "Andale Mono WT", "Andale Mono", "Lucida Console", "Lucida Sans Typewriter", "DejaVu Sans Mono", "Bitstream Vera Sans Mono", "Liberation Mono", "Nimbus Mono L", "Courier New", Courier, monospace',fontSize:"14px",lineHeight:"1.375",direction:"ltr",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none",background:"#2a2734",color:"#9a86fd"},'pre[class*="language-"]':{fontFamily:'Consolas, Menlo, Monaco, "Andale Mono WT", "Andale Mono", "Lucida Console", "Lucida Sans Typewriter", "DejaVu Sans Mono", "Bitstream Vera Sans Mono", "Liberation Mono", "Nimbus Mono L", "Courier New", Courier, monospace',fontSize:"14px",lineHeight:"1.375",direction:"ltr",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none",background:"#2a2734",color:"#9a86fd",padding:"1em",margin:".5em 0",overflow:"auto"},'pre > code[class*="language-"]':{fontSize:"1em"},'pre[class*="language-"]::-moz-selection':{textShadow:"none",background:"#6a51e6"},'pre[class*="language-"] ::-moz-selection':{textShadow:"none",background:"#6a51e6"},'code[class*="language-"]::-moz-selection':{textShadow:"none",background:"#6a51e6"},'code[class*="language-"] ::-moz-selection':{textShadow:"none",background:"#6a51e6"},'pre[class*="language-"]::selection':{textShadow:"none",background:"#6a51e6"},'pre[class*="language-"] ::selection':{textShadow:"none",background:"#6a51e6"},'code[class*="language-"]::selection':{textShadow:"none",background:"#6a51e6"},'code[class*="language-"] ::selection':{textShadow:"none",background:"#6a51e6"},':not(pre) > code[class*="language-"]':{padding:".1em",borderRadius:".3em"},comment:{color:"#6c6783"},prolog:{color:"#6c6783"},doctype:{color:"#6c6783"},cdata:{color:"#6c6783"},punctuation:{color:"#6c6783"},namespace:{Opacity:".7"},tag:{color:"#e09142"},operator:{color:"#e09142"},number:{color:"#e09142"},property:{color:"#9a86fd"},function:{color:"#9a86fd"},"tag-id":{color:"#eeebff"},selector:{color:"#eeebff"},"atrule-id":{color:"#eeebff"},"code.language-javascript":{color:"#c4b9fe"},"attr-name":{color:"#c4b9fe"},"code.language-css":{color:"#ffcc99"},"code.language-scss":{color:"#ffcc99"},boolean:{color:"#ffcc99"},string:{color:"#ffcc99"},entity:{color:"#ffcc99",cursor:"help"},url:{color:"#ffcc99"},".language-css .token.string":{color:"#ffcc99"},".language-scss .token.string":{color:"#ffcc99"},".style .token.string":{color:"#ffcc99"},"attr-value":{color:"#ffcc99"},keyword:{color:"#ffcc99"},control:{color:"#ffcc99"},directive:{color:"#ffcc99"},unit:{color:"#ffcc99"},statement:{color:"#ffcc99"},regex:{color:"#ffcc99"},atrule:{color:"#ffcc99"},placeholder:{color:"#ffcc99"},variable:{color:"#ffcc99"},deleted:{textDecoration:"line-through"},inserted:{borderBottom:"1px dotted #eeebff",textDecoration:"none"},italic:{fontStyle:"italic"},important:{fontWeight:"bold",color:"#c4b9fe"},bold:{fontWeight:"bold"},"pre > code.highlight":{Outline:".4em solid #8a75f5",OutlineOffset:".4em"},".line-numbers.line-numbers .line-numbers-rows":{borderRightColor:"#2c2937"},".line-numbers .line-numbers-rows > span:before":{color:"#3c3949"},".line-highlight.line-highlight":{background:"linear-gradient(to right, rgba(224, 145, 66, 0.2) 70%, rgba(224, 145, 66, 0))"}}}(Cx)),Cx}var Nx={},XP;function $Ee(){return XP||(XP=1,function(e){Object.defineProperty(e,"__esModule",{value:!0}),e.default=void 0,e.default={'code[class*="language-"]':{fontFamily:'Consolas, Menlo, Monaco, "Andale Mono WT", "Andale Mono", "Lucida Console", "Lucida Sans Typewriter", "DejaVu Sans Mono", "Bitstream Vera Sans Mono", "Liberation Mono", "Nimbus Mono L", "Courier New", Courier, monospace',fontSize:"14px",lineHeight:"1.375",direction:"ltr",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none",background:"#322d29",color:"#88786d"},'pre[class*="language-"]':{fontFamily:'Consolas, Menlo, Monaco, "Andale Mono WT", "Andale Mono", "Lucida Console", "Lucida Sans Typewriter", "DejaVu Sans Mono", "Bitstream Vera Sans Mono", "Liberation Mono", "Nimbus Mono L", "Courier New", Courier, monospace',fontSize:"14px",lineHeight:"1.375",direction:"ltr",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none",background:"#322d29",color:"#88786d",padding:"1em",margin:".5em 0",overflow:"auto"},'pre > code[class*="language-"]':{fontSize:"1em"},'pre[class*="language-"]::-moz-selection':{textShadow:"none",background:"#6f5849"},'pre[class*="language-"] ::-moz-selection':{textShadow:"none",background:"#6f5849"},'code[class*="language-"]::-moz-selection':{textShadow:"none",background:"#6f5849"},'code[class*="language-"] ::-moz-selection':{textShadow:"none",background:"#6f5849"},'pre[class*="language-"]::selection':{textShadow:"none",background:"#6f5849"},'pre[class*="language-"] ::selection':{textShadow:"none",background:"#6f5849"},'code[class*="language-"]::selection':{textShadow:"none",background:"#6f5849"},'code[class*="language-"] ::selection':{textShadow:"none",background:"#6f5849"},':not(pre) > code[class*="language-"]':{padding:".1em",borderRadius:".3em"},comment:{color:"#6a5f58"},prolog:{color:"#6a5f58"},doctype:{color:"#6a5f58"},cdata:{color:"#6a5f58"},punctuation:{color:"#6a5f58"},namespace:{Opacity:".7"},tag:{color:"#bfa05a"},operator:{color:"#bfa05a"},number:{color:"#bfa05a"},property:{color:"#88786d"},function:{color:"#88786d"},"tag-id":{color:"#fff3eb"},selector:{color:"#fff3eb"},"atrule-id":{color:"#fff3eb"},"code.language-javascript":{color:"#a48774"},"attr-name":{color:"#a48774"},"code.language-css":{color:"#fcc440"},"code.language-scss":{color:"#fcc440"},boolean:{color:"#fcc440"},string:{color:"#fcc440"},entity:{color:"#fcc440",cursor:"help"},url:{color:"#fcc440"},".language-css .token.string":{color:"#fcc440"},".language-scss .token.string":{color:"#fcc440"},".style .token.string":{color:"#fcc440"},"attr-value":{color:"#fcc440"},keyword:{color:"#fcc440"},control:{color:"#fcc440"},directive:{color:"#fcc440"},unit:{color:"#fcc440"},statement:{color:"#fcc440"},regex:{color:"#fcc440"},atrule:{color:"#fcc440"},placeholder:{color:"#fcc440"},variable:{color:"#fcc440"},deleted:{textDecoration:"line-through"},inserted:{borderBottom:"1px dotted #fff3eb",textDecoration:"none"},italic:{fontStyle:"italic"},important:{fontWeight:"bold",color:"#a48774"},bold:{fontWeight:"bold"},"pre > code.highlight":{Outline:".4em solid #816d5f",OutlineOffset:".4em"},".line-numbers.line-numbers .line-numbers-rows":{borderRightColor:"#35302b"},".line-numbers .line-numbers-rows > span:before":{color:"#46403d"},".line-highlight.line-highlight":{background:"linear-gradient(to right, rgba(191, 160, 90, 0.2) 70%, rgba(191, 160, 90, 0))"}}}(Nx)),Nx}var Ox={},ZP;function qEe(){return ZP||(ZP=1,function(e){Object.defineProperty(e,"__esModule",{value:!0}),e.default=void 0,e.default={'code[class*="language-"]':{fontFamily:'Consolas, Menlo, Monaco, "Andale Mono WT", "Andale Mono", "Lucida Console", "Lucida Sans Typewriter", "DejaVu Sans Mono", "Bitstream Vera Sans Mono", "Liberation Mono", "Nimbus Mono L", "Courier New", Courier, monospace',fontSize:"14px",lineHeight:"1.375",direction:"ltr",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none",background:"#2a2d2a",color:"#687d68"},'pre[class*="language-"]':{fontFamily:'Consolas, Menlo, Monaco, "Andale Mono WT", "Andale Mono", "Lucida Console", "Lucida Sans Typewriter", "DejaVu Sans Mono", "Bitstream Vera Sans Mono", "Liberation Mono", "Nimbus Mono L", "Courier New", Courier, monospace',fontSize:"14px",lineHeight:"1.375",direction:"ltr",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none",background:"#2a2d2a",color:"#687d68",padding:"1em",margin:".5em 0",overflow:"auto"},'pre > code[class*="language-"]':{fontSize:"1em"},'pre[class*="language-"]::-moz-selection':{textShadow:"none",background:"#435643"},'pre[class*="language-"] ::-moz-selection':{textShadow:"none",background:"#435643"},'code[class*="language-"]::-moz-selection':{textShadow:"none",background:"#435643"},'code[class*="language-"] ::-moz-selection':{textShadow:"none",background:"#435643"},'pre[class*="language-"]::selection':{textShadow:"none",background:"#435643"},'pre[class*="language-"] ::selection':{textShadow:"none",background:"#435643"},'code[class*="language-"]::selection':{textShadow:"none",background:"#435643"},'code[class*="language-"] ::selection':{textShadow:"none",background:"#435643"},':not(pre) > code[class*="language-"]':{padding:".1em",borderRadius:".3em"},comment:{color:"#535f53"},prolog:{color:"#535f53"},doctype:{color:"#535f53"},cdata:{color:"#535f53"},punctuation:{color:"#535f53"},namespace:{Opacity:".7"},tag:{color:"#a2b34d"},operator:{color:"#a2b34d"},number:{color:"#a2b34d"},property:{color:"#687d68"},function:{color:"#687d68"},"tag-id":{color:"#f0fff0"},selector:{color:"#f0fff0"},"atrule-id":{color:"#f0fff0"},"code.language-javascript":{color:"#b3d6b3"},"attr-name":{color:"#b3d6b3"},"code.language-css":{color:"#e5fb79"},"code.language-scss":{color:"#e5fb79"},boolean:{color:"#e5fb79"},string:{color:"#e5fb79"},entity:{color:"#e5fb79",cursor:"help"},url:{color:"#e5fb79"},".language-css .token.string":{color:"#e5fb79"},".language-scss .token.string":{color:"#e5fb79"},".style .token.string":{color:"#e5fb79"},"attr-value":{color:"#e5fb79"},keyword:{color:"#e5fb79"},control:{color:"#e5fb79"},directive:{color:"#e5fb79"},unit:{color:"#e5fb79"},statement:{color:"#e5fb79"},regex:{color:"#e5fb79"},atrule:{color:"#e5fb79"},placeholder:{color:"#e5fb79"},variable:{color:"#e5fb79"},deleted:{textDecoration:"line-through"},inserted:{borderBottom:"1px dotted #f0fff0",textDecoration:"none"},italic:{fontStyle:"italic"},important:{fontWeight:"bold",color:"#b3d6b3"},bold:{fontWeight:"bold"},"pre > code.highlight":{Outline:".4em solid #5c705c",OutlineOffset:".4em"},".line-numbers.line-numbers .line-numbers-rows":{borderRightColor:"#2c302c"},".line-numbers .line-numbers-rows > span:before":{color:"#3b423b"},".line-highlight.line-highlight":{background:"linear-gradient(to right, rgba(162, 179, 77, 0.2) 70%, rgba(162, 179, 77, 0))"}}}(Ox)),Ox}var Dx={},QP;function VEe(){return QP||(QP=1,function(e){Object.defineProperty(e,"__esModule",{value:!0}),e.default=void 0,e.default={'code[class*="language-"]':{fontFamily:'Consolas, Menlo, Monaco, "Andale Mono WT", "Andale Mono", "Lucida Console", "Lucida Sans Typewriter", "DejaVu Sans Mono", "Bitstream Vera Sans Mono", "Liberation Mono", "Nimbus Mono L", "Courier New", Courier, monospace',fontSize:"14px",lineHeight:"1.375",direction:"ltr",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none",background:"#faf8f5",color:"#728fcb"},'pre[class*="language-"]':{fontFamily:'Consolas, Menlo, Monaco, "Andale Mono WT", "Andale Mono", "Lucida Console", "Lucida Sans Typewriter", "DejaVu Sans Mono", "Bitstream Vera Sans Mono", "Liberation Mono", "Nimbus Mono L", "Courier New", Courier, monospace',fontSize:"14px",lineHeight:"1.375",direction:"ltr",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none",background:"#faf8f5",color:"#728fcb",padding:"1em",margin:".5em 0",overflow:"auto"},'pre > code[class*="language-"]':{fontSize:"1em"},'pre[class*="language-"]::-moz-selection':{textShadow:"none",background:"#faf8f5"},'pre[class*="language-"] ::-moz-selection':{textShadow:"none",background:"#faf8f5"},'code[class*="language-"]::-moz-selection':{textShadow:"none",background:"#faf8f5"},'code[class*="language-"] ::-moz-selection':{textShadow:"none",background:"#faf8f5"},'pre[class*="language-"]::selection':{textShadow:"none",background:"#faf8f5"},'pre[class*="language-"] ::selection':{textShadow:"none",background:"#faf8f5"},'code[class*="language-"]::selection':{textShadow:"none",background:"#faf8f5"},'code[class*="language-"] ::selection':{textShadow:"none",background:"#faf8f5"},':not(pre) > code[class*="language-"]':{padding:".1em",borderRadius:".3em"},comment:{color:"#b6ad9a"},prolog:{color:"#b6ad9a"},doctype:{color:"#b6ad9a"},cdata:{color:"#b6ad9a"},punctuation:{color:"#b6ad9a"},namespace:{Opacity:".7"},tag:{color:"#063289"},operator:{color:"#063289"},number:{color:"#063289"},property:{color:"#b29762"},function:{color:"#b29762"},"tag-id":{color:"#2d2006"},selector:{color:"#2d2006"},"atrule-id":{color:"#2d2006"},"code.language-javascript":{color:"#896724"},"attr-name":{color:"#896724"},"code.language-css":{color:"#728fcb"},"code.language-scss":{color:"#728fcb"},boolean:{color:"#728fcb"},string:{color:"#728fcb"},entity:{color:"#728fcb",cursor:"help"},url:{color:"#728fcb"},".language-css .token.string":{color:"#728fcb"},".language-scss .token.string":{color:"#728fcb"},".style .token.string":{color:"#728fcb"},"attr-value":{color:"#728fcb"},keyword:{color:"#728fcb"},control:{color:"#728fcb"},directive:{color:"#728fcb"},unit:{color:"#728fcb"},statement:{color:"#728fcb"},regex:{color:"#728fcb"},atrule:{color:"#728fcb"},placeholder:{color:"#93abdc"},variable:{color:"#93abdc"},deleted:{textDecoration:"line-through"},inserted:{borderBottom:"1px dotted #2d2006",textDecoration:"none"},italic:{fontStyle:"italic"},important:{fontWeight:"bold",color:"#896724"},bold:{fontWeight:"bold"},"pre > code.highlight":{Outline:".4em solid #896724",OutlineOffset:".4em"},".line-numbers.line-numbers .line-numbers-rows":{borderRightColor:"#ece8de"},".line-numbers .line-numbers-rows > span:before":{color:"#cdc4b1"},".line-highlight.line-highlight":{background:"linear-gradient(to right, rgba(45, 32, 6, 0.2) 70%, rgba(45, 32, 6, 0))"}}}(Dx)),Dx}var Ix={},JP;function WEe(){return JP||(JP=1,function(e){Object.defineProperty(e,"__esModule",{value:!0}),e.default=void 0,e.default={'code[class*="language-"]':{fontFamily:'Consolas, Menlo, Monaco, "Andale Mono WT", "Andale Mono", "Lucida Console", "Lucida Sans Typewriter", "DejaVu Sans Mono", "Bitstream Vera Sans Mono", "Liberation Mono", "Nimbus Mono L", "Courier New", Courier, monospace',fontSize:"14px",lineHeight:"1.375",direction:"ltr",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none",background:"#1d262f",color:"#57718e"},'pre[class*="language-"]':{fontFamily:'Consolas, Menlo, Monaco, "Andale Mono WT", "Andale Mono", "Lucida Console", "Lucida Sans Typewriter", "DejaVu Sans Mono", "Bitstream Vera Sans Mono", "Liberation Mono", "Nimbus Mono L", "Courier New", Courier, monospace',fontSize:"14px",lineHeight:"1.375",direction:"ltr",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none",background:"#1d262f",color:"#57718e",padding:"1em",margin:".5em 0",overflow:"auto"},'pre > code[class*="language-"]':{fontSize:"1em"},'pre[class*="language-"]::-moz-selection':{textShadow:"none",background:"#004a9e"},'pre[class*="language-"] ::-moz-selection':{textShadow:"none",background:"#004a9e"},'code[class*="language-"]::-moz-selection':{textShadow:"none",background:"#004a9e"},'code[class*="language-"] ::-moz-selection':{textShadow:"none",background:"#004a9e"},'pre[class*="language-"]::selection':{textShadow:"none",background:"#004a9e"},'pre[class*="language-"] ::selection':{textShadow:"none",background:"#004a9e"},'code[class*="language-"]::selection':{textShadow:"none",background:"#004a9e"},'code[class*="language-"] ::selection':{textShadow:"none",background:"#004a9e"},':not(pre) > code[class*="language-"]':{padding:".1em",borderRadius:".3em"},comment:{color:"#4a5f78"},prolog:{color:"#4a5f78"},doctype:{color:"#4a5f78"},cdata:{color:"#4a5f78"},punctuation:{color:"#4a5f78"},namespace:{Opacity:".7"},tag:{color:"#0aa370"},operator:{color:"#0aa370"},number:{color:"#0aa370"},property:{color:"#57718e"},function:{color:"#57718e"},"tag-id":{color:"#ebf4ff"},selector:{color:"#ebf4ff"},"atrule-id":{color:"#ebf4ff"},"code.language-javascript":{color:"#7eb6f6"},"attr-name":{color:"#7eb6f6"},"code.language-css":{color:"#47ebb4"},"code.language-scss":{color:"#47ebb4"},boolean:{color:"#47ebb4"},string:{color:"#47ebb4"},entity:{color:"#47ebb4",cursor:"help"},url:{color:"#47ebb4"},".language-css .token.string":{color:"#47ebb4"},".language-scss .token.string":{color:"#47ebb4"},".style .token.string":{color:"#47ebb4"},"attr-value":{color:"#47ebb4"},keyword:{color:"#47ebb4"},control:{color:"#47ebb4"},directive:{color:"#47ebb4"},unit:{color:"#47ebb4"},statement:{color:"#47ebb4"},regex:{color:"#47ebb4"},atrule:{color:"#47ebb4"},placeholder:{color:"#47ebb4"},variable:{color:"#47ebb4"},deleted:{textDecoration:"line-through"},inserted:{borderBottom:"1px dotted #ebf4ff",textDecoration:"none"},italic:{fontStyle:"italic"},important:{fontWeight:"bold",color:"#7eb6f6"},bold:{fontWeight:"bold"},"pre > code.highlight":{Outline:".4em solid #34659d",OutlineOffset:".4em"},".line-numbers.line-numbers .line-numbers-rows":{borderRightColor:"#1f2932"},".line-numbers .line-numbers-rows > span:before":{color:"#2c3847"},".line-highlight.line-highlight":{background:"linear-gradient(to right, rgba(10, 163, 112, 0.2) 70%, rgba(10, 163, 112, 0))"}}}(Ix)),Ix}var Lx={},ez;function YEe(){return ez||(ez=1,function(e){Object.defineProperty(e,"__esModule",{value:!0}),e.default=void 0,e.default={'code[class*="language-"]':{fontFamily:'Consolas, Menlo, Monaco, "Andale Mono WT", "Andale Mono", "Lucida Console", "Lucida Sans Typewriter", "DejaVu Sans Mono", "Bitstream Vera Sans Mono", "Liberation Mono", "Nimbus Mono L", "Courier New", Courier, monospace',fontSize:"14px",lineHeight:"1.375",direction:"ltr",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none",background:"#24242e",color:"#767693"},'pre[class*="language-"]':{fontFamily:'Consolas, Menlo, Monaco, "Andale Mono WT", "Andale Mono", "Lucida Console", "Lucida Sans Typewriter", "DejaVu Sans Mono", "Bitstream Vera Sans Mono", "Liberation Mono", "Nimbus Mono L", "Courier New", Courier, monospace',fontSize:"14px",lineHeight:"1.375",direction:"ltr",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none",background:"#24242e",color:"#767693",padding:"1em",margin:".5em 0",overflow:"auto"},'pre > code[class*="language-"]':{fontSize:"1em"},'pre[class*="language-"]::-moz-selection':{textShadow:"none",background:"#5151e6"},'pre[class*="language-"] ::-moz-selection':{textShadow:"none",background:"#5151e6"},'code[class*="language-"]::-moz-selection':{textShadow:"none",background:"#5151e6"},'code[class*="language-"] ::-moz-selection':{textShadow:"none",background:"#5151e6"},'pre[class*="language-"]::selection':{textShadow:"none",background:"#5151e6"},'pre[class*="language-"] ::selection':{textShadow:"none",background:"#5151e6"},'code[class*="language-"]::selection':{textShadow:"none",background:"#5151e6"},'code[class*="language-"] ::selection':{textShadow:"none",background:"#5151e6"},':not(pre) > code[class*="language-"]':{padding:".1em",borderRadius:".3em"},comment:{color:"#5b5b76"},prolog:{color:"#5b5b76"},doctype:{color:"#5b5b76"},cdata:{color:"#5b5b76"},punctuation:{color:"#5b5b76"},namespace:{Opacity:".7"},tag:{color:"#dd672c"},operator:{color:"#dd672c"},number:{color:"#dd672c"},property:{color:"#767693"},function:{color:"#767693"},"tag-id":{color:"#ebebff"},selector:{color:"#ebebff"},"atrule-id":{color:"#ebebff"},"code.language-javascript":{color:"#aaaaca"},"attr-name":{color:"#aaaaca"},"code.language-css":{color:"#fe8c52"},"code.language-scss":{color:"#fe8c52"},boolean:{color:"#fe8c52"},string:{color:"#fe8c52"},entity:{color:"#fe8c52",cursor:"help"},url:{color:"#fe8c52"},".language-css .token.string":{color:"#fe8c52"},".language-scss .token.string":{color:"#fe8c52"},".style .token.string":{color:"#fe8c52"},"attr-value":{color:"#fe8c52"},keyword:{color:"#fe8c52"},control:{color:"#fe8c52"},directive:{color:"#fe8c52"},unit:{color:"#fe8c52"},statement:{color:"#fe8c52"},regex:{color:"#fe8c52"},atrule:{color:"#fe8c52"},placeholder:{color:"#fe8c52"},variable:{color:"#fe8c52"},deleted:{textDecoration:"line-through"},inserted:{borderBottom:"1px dotted #ebebff",textDecoration:"none"},italic:{fontStyle:"italic"},important:{fontWeight:"bold",color:"#aaaaca"},bold:{fontWeight:"bold"},"pre > code.highlight":{Outline:".4em solid #7676f4",OutlineOffset:".4em"},".line-numbers.line-numbers .line-numbers-rows":{borderRightColor:"#262631"},".line-numbers .line-numbers-rows > span:before":{color:"#393949"},".line-highlight.line-highlight":{background:"linear-gradient(to right, rgba(221, 103, 44, 0.2) 70%, rgba(221, 103, 44, 0))"}}}(Lx)),Lx}var Mx={},tz;function KEe(){return tz||(tz=1,function(e){Object.defineProperty(e,"__esModule",{value:!0}),e.default=void 0,e.default={'code[class*="language-"]':{color:"#393A34",fontFamily:'"Consolas", "Bitstream Vera Sans Mono", "Courier New", Courier, monospace',direction:"ltr",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",fontSize:".9em",lineHeight:"1.2em",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none"},'pre[class*="language-"]':{color:"#393A34",fontFamily:'"Consolas", "Bitstream Vera Sans Mono", "Courier New", Courier, monospace',direction:"ltr",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",fontSize:".9em",lineHeight:"1.2em",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none",padding:"1em",margin:".5em 0",overflow:"auto",border:"1px solid #dddddd",backgroundColor:"white"},'pre > code[class*="language-"]':{fontSize:"1em"},'pre[class*="language-"]::-moz-selection':{background:"#b3d4fc"},'pre[class*="language-"] ::-moz-selection':{background:"#b3d4fc"},'code[class*="language-"]::-moz-selection':{background:"#b3d4fc"},'code[class*="language-"] ::-moz-selection':{background:"#b3d4fc"},'pre[class*="language-"]::selection':{background:"#b3d4fc"},'pre[class*="language-"] ::selection':{background:"#b3d4fc"},'code[class*="language-"]::selection':{background:"#b3d4fc"},'code[class*="language-"] ::selection':{background:"#b3d4fc"},':not(pre) > code[class*="language-"]':{padding:".2em",paddingTop:"1px",paddingBottom:"1px",background:"#f8f8f8",border:"1px solid #dddddd"},comment:{color:"#999988",fontStyle:"italic"},prolog:{color:"#999988",fontStyle:"italic"},doctype:{color:"#999988",fontStyle:"italic"},cdata:{color:"#999988",fontStyle:"italic"},namespace:{Opacity:".7"},string:{color:"#e3116c"},"attr-value":{color:"#e3116c"},punctuation:{color:"#393A34"},operator:{color:"#393A34"},entity:{color:"#36acaa"},url:{color:"#36acaa"},symbol:{color:"#36acaa"},number:{color:"#36acaa"},boolean:{color:"#36acaa"},variable:{color:"#36acaa"},constant:{color:"#36acaa"},property:{color:"#36acaa"},regex:{color:"#36acaa"},inserted:{color:"#36acaa"},atrule:{color:"#00a4db"},keyword:{color:"#00a4db"},"attr-name":{color:"#00a4db"},".language-autohotkey .token.selector":{color:"#00a4db"},function:{color:"#9a050f",fontWeight:"bold"},deleted:{color:"#9a050f"},".language-autohotkey .token.tag":{color:"#9a050f"},tag:{color:"#00009f"},selector:{color:"#00009f"},".language-autohotkey .token.keyword":{color:"#00009f"},important:{fontWeight:"bold"},bold:{fontWeight:"bold"},italic:{fontStyle:"italic"}}}(Mx)),Mx}var Fx={},nz;function XEe(){return nz||(nz=1,function(e){Object.defineProperty(e,"__esModule",{value:!0}),e.default=void 0,e.default={'code[class*="language-"]':{color:"#ebdbb2",fontFamily:'Consolas, Monaco, "Andale Mono", monospace',direction:"ltr",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",lineHeight:"1.5",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none"},'pre[class*="language-"]':{color:"#ebdbb2",fontFamily:'Consolas, Monaco, "Andale Mono", monospace',direction:"ltr",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",lineHeight:"1.5",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none",padding:"1em",margin:"0.5em 0",overflow:"auto",background:"#1d2021"},'pre[class*="language-"]::-moz-selection':{color:"#fbf1c7",background:"#7c6f64"},'pre[class*="language-"] ::-moz-selection':{color:"#fbf1c7",background:"#7c6f64"},'code[class*="language-"]::-moz-selection':{color:"#fbf1c7",background:"#7c6f64"},'code[class*="language-"] ::-moz-selection':{color:"#fbf1c7",background:"#7c6f64"},'pre[class*="language-"]::selection':{color:"#fbf1c7",background:"#7c6f64"},'pre[class*="language-"] ::selection':{color:"#fbf1c7",background:"#7c6f64"},'code[class*="language-"]::selection':{color:"#fbf1c7",background:"#7c6f64"},'code[class*="language-"] ::selection':{color:"#fbf1c7",background:"#7c6f64"},':not(pre) > code[class*="language-"]':{background:"#1d2021",padding:"0.1em",borderRadius:"0.3em"},comment:{color:"#a89984"},prolog:{color:"#a89984"},cdata:{color:"#a89984"},delimiter:{color:"#fb4934"},boolean:{color:"#fb4934"},keyword:{color:"#fb4934"},selector:{color:"#fb4934"},important:{color:"#fb4934"},atrule:{color:"#fb4934"},operator:{color:"#a89984"},punctuation:{color:"#a89984"},"attr-name":{color:"#a89984"},tag:{color:"#fabd2f"},"tag.punctuation":{color:"#fabd2f"},doctype:{color:"#fabd2f"},builtin:{color:"#fabd2f"},entity:{color:"#d3869b"},number:{color:"#d3869b"},symbol:{color:"#d3869b"},property:{color:"#fb4934"},constant:{color:"#fb4934"},variable:{color:"#fb4934"},string:{color:"#b8bb26"},char:{color:"#b8bb26"},"attr-value":{color:"#a89984"},"attr-value.punctuation":{color:"#a89984"},url:{color:"#b8bb26",textDecoration:"underline"},function:{color:"#fabd2f"},regex:{background:"#b8bb26"},bold:{fontWeight:"bold"},italic:{fontStyle:"italic"},inserted:{background:"#a89984"},deleted:{background:"#fb4934"}}}(Fx)),Fx}var Px={},rz;function ZEe(){return rz||(rz=1,function(e){Object.defineProperty(e,"__esModule",{value:!0}),e.default=void 0,e.default={'code[class*="language-"]':{color:"#3c3836",fontFamily:'Consolas, Monaco, "Andale Mono", monospace',direction:"ltr",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",lineHeight:"1.5",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none"},'pre[class*="language-"]':{color:"#3c3836",fontFamily:'Consolas, Monaco, "Andale Mono", monospace',direction:"ltr",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",lineHeight:"1.5",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none",padding:"1em",margin:"0.5em 0",overflow:"auto",background:"#f9f5d7"},'pre[class*="language-"]::-moz-selection':{color:"#282828",background:"#a89984"},'pre[class*="language-"] ::-moz-selection':{color:"#282828",background:"#a89984"},'code[class*="language-"]::-moz-selection':{color:"#282828",background:"#a89984"},'code[class*="language-"] ::-moz-selection':{color:"#282828",background:"#a89984"},'pre[class*="language-"]::selection':{color:"#282828",background:"#a89984"},'pre[class*="language-"] ::selection':{color:"#282828",background:"#a89984"},'code[class*="language-"]::selection':{color:"#282828",background:"#a89984"},'code[class*="language-"] ::selection':{color:"#282828",background:"#a89984"},':not(pre) > code[class*="language-"]':{background:"#f9f5d7",padding:"0.1em",borderRadius:"0.3em"},comment:{color:"#7c6f64"},prolog:{color:"#7c6f64"},cdata:{color:"#7c6f64"},delimiter:{color:"#9d0006"},boolean:{color:"#9d0006"},keyword:{color:"#9d0006"},selector:{color:"#9d0006"},important:{color:"#9d0006"},atrule:{color:"#9d0006"},operator:{color:"#7c6f64"},punctuation:{color:"#7c6f64"},"attr-name":{color:"#7c6f64"},tag:{color:"#b57614"},"tag.punctuation":{color:"#b57614"},doctype:{color:"#b57614"},builtin:{color:"#b57614"},entity:{color:"#8f3f71"},number:{color:"#8f3f71"},symbol:{color:"#8f3f71"},property:{color:"#9d0006"},constant:{color:"#9d0006"},variable:{color:"#9d0006"},string:{color:"#797403"},char:{color:"#797403"},"attr-value":{color:"#7c6f64"},"attr-value.punctuation":{color:"#7c6f64"},url:{color:"#797403",textDecoration:"underline"},function:{color:"#b57614"},regex:{background:"#797403"},bold:{fontWeight:"bold"},italic:{fontStyle:"italic"},inserted:{background:"#7c6f64"},deleted:{background:"#9d0006"}}}(Px)),Px}var zx={},az;function QEe(){return az||(az=1,function(e){Object.defineProperty(e,"__esModule",{value:!0}),e.default=void 0,e.default={"code[class*='language-']":{color:"#d6e7ff",background:"#030314",textShadow:"none",fontFamily:'Consolas, Monaco, "Andale Mono", "Ubuntu Mono", monospace',fontSize:"1em",lineHeight:"1.5",letterSpacing:".2px",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",wordWrap:"normal",textAlign:"left",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none"},"pre[class*='language-']":{color:"#d6e7ff",background:"#030314",textShadow:"none",fontFamily:'Consolas, Monaco, "Andale Mono", "Ubuntu Mono", monospace',fontSize:"1em",lineHeight:"1.5",letterSpacing:".2px",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",wordWrap:"normal",textAlign:"left",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none",border:"1px solid #2a4555",borderRadius:"5px",padding:"1.5em 1em",margin:"1em 0",overflow:"auto"},"pre[class*='language-']::-moz-selection":{color:"inherit",background:"#1d3b54",textShadow:"none"},"pre[class*='language-'] ::-moz-selection":{color:"inherit",background:"#1d3b54",textShadow:"none"},"code[class*='language-']::-moz-selection":{color:"inherit",background:"#1d3b54",textShadow:"none"},"code[class*='language-'] ::-moz-selection":{color:"inherit",background:"#1d3b54",textShadow:"none"},"pre[class*='language-']::selection":{color:"inherit",background:"#1d3b54",textShadow:"none"},"pre[class*='language-'] ::selection":{color:"inherit",background:"#1d3b54",textShadow:"none"},"code[class*='language-']::selection":{color:"inherit",background:"#1d3b54",textShadow:"none"},"code[class*='language-'] ::selection":{color:"inherit",background:"#1d3b54",textShadow:"none"},":not(pre) > code[class*='language-']":{color:"#f0f6f6",background:"#2a4555",padding:"0.2em 0.3em",borderRadius:"0.2em",boxDecorationBreak:"clone"},comment:{color:"#446e69"},prolog:{color:"#446e69"},doctype:{color:"#446e69"},cdata:{color:"#446e69"},punctuation:{color:"#d6b007"},property:{color:"#d6e7ff"},tag:{color:"#d6e7ff"},boolean:{color:"#d6e7ff"},number:{color:"#d6e7ff"},constant:{color:"#d6e7ff"},symbol:{color:"#d6e7ff"},deleted:{color:"#d6e7ff"},selector:{color:"#e60067"},"attr-name":{color:"#e60067"},builtin:{color:"#e60067"},inserted:{color:"#e60067"},string:{color:"#49c6ec"},char:{color:"#49c6ec"},operator:{color:"#ec8e01",background:"transparent"},entity:{color:"#ec8e01",background:"transparent"},url:{color:"#ec8e01",background:"transparent"},".language-css .token.string":{color:"#ec8e01",background:"transparent"},".style .token.string":{color:"#ec8e01",background:"transparent"},atrule:{color:"#0fe468"},"attr-value":{color:"#0fe468"},keyword:{color:"#0fe468"},function:{color:"#78f3e9"},"class-name":{color:"#78f3e9"},regex:{color:"#d6e7ff"},important:{color:"#d6e7ff"},variable:{color:"#d6e7ff"}}}(zx)),zx}var Bx={},oz;function JEe(){return oz||(oz=1,function(e){Object.defineProperty(e,"__esModule",{value:!0}),e.default=void 0,e.default={'code[class*="language-"]':{fontFamily:'"Fira Mono", Menlo, Monaco, "Lucida Console", "Courier New", Courier, monospace',fontSize:"16px",lineHeight:"1.375",direction:"ltr",textAlign:"left",wordSpacing:"normal",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none",whiteSpace:"pre-wrap",wordBreak:"break-all",wordWrap:"break-word",background:"#322931",color:"#b9b5b8"},'pre[class*="language-"]':{fontFamily:'"Fira Mono", Menlo, Monaco, "Lucida Console", "Courier New", Courier, monospace',fontSize:"16px",lineHeight:"1.375",direction:"ltr",textAlign:"left",wordSpacing:"normal",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none",whiteSpace:"pre-wrap",wordBreak:"break-all",wordWrap:"break-word",background:"#322931",color:"#b9b5b8",padding:"1em",margin:".5em 0",overflow:"auto"},'pre > code[class*="language-"]':{fontSize:"1em"},':not(pre) > code[class*="language-"]':{padding:".1em",borderRadius:".3em"},comment:{color:"#797379"},prolog:{color:"#797379"},doctype:{color:"#797379"},cdata:{color:"#797379"},punctuation:{color:"#b9b5b8"},".namespace":{Opacity:".7"},null:{color:"#fd8b19"},operator:{color:"#fd8b19"},boolean:{color:"#fd8b19"},number:{color:"#fd8b19"},property:{color:"#fdcc59"},tag:{color:"#1290bf"},string:{color:"#149b93"},selector:{color:"#c85e7c"},"attr-name":{color:"#fd8b19"},entity:{color:"#149b93",cursor:"help"},url:{color:"#149b93"},".language-css .token.string":{color:"#149b93"},".style .token.string":{color:"#149b93"},"attr-value":{color:"#8fc13e"},keyword:{color:"#8fc13e"},control:{color:"#8fc13e"},directive:{color:"#8fc13e"},unit:{color:"#8fc13e"},statement:{color:"#149b93"},regex:{color:"#149b93"},atrule:{color:"#149b93"},placeholder:{color:"#1290bf"},variable:{color:"#1290bf"},important:{color:"#dd464c",fontWeight:"bold"},"pre > code.highlight":{Outline:".4em solid red",OutlineOffset:".4em"}}}(Bx)),Bx}var Ux={},iz;function ewe(){return iz||(iz=1,function(e){Object.defineProperty(e,"__esModule",{value:!0}),e.default=void 0,e.default={'code[class*="language-"]':{color:"#f8f8f2",background:"none",textShadow:"0 1px rgba(0, 0, 0, 0.3)",fontFamily:"Monaco, Consolas, 'Andale Mono', 'Ubuntu Mono', monospace",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",wordWrap:"normal",lineHeight:"1.5",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none"},'pre[class*="language-"]':{color:"#f8f8f2",background:"#263E52",textShadow:"0 1px rgba(0, 0, 0, 0.3)",fontFamily:"Monaco, Consolas, 'Andale Mono', 'Ubuntu Mono', monospace",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",wordWrap:"normal",lineHeight:"1.5",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none",padding:"1em",margin:".5em 0",overflow:"auto",borderRadius:"0.3em"},':not(pre) > code[class*="language-"]':{background:"#263E52",padding:".1em",borderRadius:".3em",whiteSpace:"normal"},comment:{color:"#5c98cd"},prolog:{color:"#5c98cd"},doctype:{color:"#5c98cd"},cdata:{color:"#5c98cd"},punctuation:{color:"#f8f8f2"},".namespace":{Opacity:".7"},property:{color:"#F05E5D"},tag:{color:"#F05E5D"},constant:{color:"#F05E5D"},symbol:{color:"#F05E5D"},deleted:{color:"#F05E5D"},boolean:{color:"#BC94F9"},number:{color:"#BC94F9"},selector:{color:"#FCFCD6"},"attr-name":{color:"#FCFCD6"},string:{color:"#FCFCD6"},char:{color:"#FCFCD6"},builtin:{color:"#FCFCD6"},inserted:{color:"#FCFCD6"},operator:{color:"#f8f8f2"},entity:{color:"#f8f8f2",cursor:"help"},url:{color:"#f8f8f2"},".language-css .token.string":{color:"#f8f8f2"},".style .token.string":{color:"#f8f8f2"},variable:{color:"#f8f8f2"},atrule:{color:"#66D8EF"},"attr-value":{color:"#66D8EF"},function:{color:"#66D8EF"},"class-name":{color:"#66D8EF"},keyword:{color:"#6EB26E"},regex:{color:"#F05E5D"},important:{color:"#F05E5D",fontWeight:"bold"},bold:{fontWeight:"bold"},italic:{fontStyle:"italic"}}}(Ux)),Ux}var jx={},sz;function twe(){return sz||(sz=1,function(e){Object.defineProperty(e,"__esModule",{value:!0}),e.default=void 0,e.default={'code[class*="language-"]':{textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",wordWrap:"normal",color:"#eee",background:"#2f2f2f",fontFamily:"Roboto Mono, monospace",fontSize:"1em",lineHeight:"1.5em",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none"},'pre[class*="language-"]':{textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",wordWrap:"normal",color:"#eee",background:"#2f2f2f",fontFamily:"Roboto Mono, monospace",fontSize:"1em",lineHeight:"1.5em",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none",overflow:"auto",position:"relative",margin:"0.5em 0",padding:"1.25em 1em"},'code[class*="language-"]::-moz-selection':{background:"#363636"},'pre[class*="language-"]::-moz-selection':{background:"#363636"},'code[class*="language-"] ::-moz-selection':{background:"#363636"},'pre[class*="language-"] ::-moz-selection':{background:"#363636"},'code[class*="language-"]::selection':{background:"#363636"},'pre[class*="language-"]::selection':{background:"#363636"},'code[class*="language-"] ::selection':{background:"#363636"},'pre[class*="language-"] ::selection':{background:"#363636"},':not(pre) > code[class*="language-"]':{whiteSpace:"normal",borderRadius:"0.2em",padding:"0.1em"},".language-css > code":{color:"#fd9170"},".language-sass > code":{color:"#fd9170"},".language-scss > code":{color:"#fd9170"},'[class*="language-"] .namespace':{Opacity:"0.7"},atrule:{color:"#c792ea"},"attr-name":{color:"#ffcb6b"},"attr-value":{color:"#a5e844"},attribute:{color:"#a5e844"},boolean:{color:"#c792ea"},builtin:{color:"#ffcb6b"},cdata:{color:"#80cbc4"},char:{color:"#80cbc4"},class:{color:"#ffcb6b"},"class-name":{color:"#f2ff00"},comment:{color:"#616161"},constant:{color:"#c792ea"},deleted:{color:"#ff6666"},doctype:{color:"#616161"},entity:{color:"#ff6666"},function:{color:"#c792ea"},hexcode:{color:"#f2ff00"},id:{color:"#c792ea",fontWeight:"bold"},important:{color:"#c792ea",fontWeight:"bold"},inserted:{color:"#80cbc4"},keyword:{color:"#c792ea"},number:{color:"#fd9170"},operator:{color:"#89ddff"},prolog:{color:"#616161"},property:{color:"#80cbc4"},"pseudo-class":{color:"#a5e844"},"pseudo-element":{color:"#a5e844"},punctuation:{color:"#89ddff"},regex:{color:"#f2ff00"},selector:{color:"#ff6666"},string:{color:"#a5e844"},symbol:{color:"#c792ea"},tag:{color:"#ff6666"},unit:{color:"#fd9170"},url:{color:"#ff6666"},variable:{color:"#ff6666"}}}(jx)),jx}var Gx={},lz;function nwe(){return lz||(lz=1,function(e){Object.defineProperty(e,"__esModule",{value:!0}),e.default=void 0,e.default={'code[class*="language-"]':{textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",wordWrap:"normal",color:"#90a4ae",background:"#fafafa",fontFamily:"Roboto Mono, monospace",fontSize:"1em",lineHeight:"1.5em",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none"},'pre[class*="language-"]':{textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",wordWrap:"normal",color:"#90a4ae",background:"#fafafa",fontFamily:"Roboto Mono, monospace",fontSize:"1em",lineHeight:"1.5em",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none",overflow:"auto",position:"relative",margin:"0.5em 0",padding:"1.25em 1em"},'code[class*="language-"]::-moz-selection':{background:"#cceae7",color:"#263238"},'pre[class*="language-"]::-moz-selection':{background:"#cceae7",color:"#263238"},'code[class*="language-"] ::-moz-selection':{background:"#cceae7",color:"#263238"},'pre[class*="language-"] ::-moz-selection':{background:"#cceae7",color:"#263238"},'code[class*="language-"]::selection':{background:"#cceae7",color:"#263238"},'pre[class*="language-"]::selection':{background:"#cceae7",color:"#263238"},'code[class*="language-"] ::selection':{background:"#cceae7",color:"#263238"},'pre[class*="language-"] ::selection':{background:"#cceae7",color:"#263238"},':not(pre) > code[class*="language-"]':{whiteSpace:"normal",borderRadius:"0.2em",padding:"0.1em"},".language-css > code":{color:"#f76d47"},".language-sass > code":{color:"#f76d47"},".language-scss > code":{color:"#f76d47"},'[class*="language-"] .namespace':{Opacity:"0.7"},atrule:{color:"#7c4dff"},"attr-name":{color:"#39adb5"},"attr-value":{color:"#f6a434"},attribute:{color:"#f6a434"},boolean:{color:"#7c4dff"},builtin:{color:"#39adb5"},cdata:{color:"#39adb5"},char:{color:"#39adb5"},class:{color:"#39adb5"},"class-name":{color:"#6182b8"},comment:{color:"#aabfc9"},constant:{color:"#7c4dff"},deleted:{color:"#e53935"},doctype:{color:"#aabfc9"},entity:{color:"#e53935"},function:{color:"#7c4dff"},hexcode:{color:"#f76d47"},id:{color:"#7c4dff",fontWeight:"bold"},important:{color:"#7c4dff",fontWeight:"bold"},inserted:{color:"#39adb5"},keyword:{color:"#7c4dff"},number:{color:"#f76d47"},operator:{color:"#39adb5"},prolog:{color:"#aabfc9"},property:{color:"#39adb5"},"pseudo-class":{color:"#f6a434"},"pseudo-element":{color:"#f6a434"},punctuation:{color:"#39adb5"},regex:{color:"#6182b8"},selector:{color:"#e53935"},string:{color:"#f6a434"},symbol:{color:"#7c4dff"},tag:{color:"#e53935"},unit:{color:"#f76d47"},url:{color:"#e53935"},variable:{color:"#e53935"}}}(Gx)),Gx}var Hx={},uz;function rwe(){return uz||(uz=1,function(e){Object.defineProperty(e,"__esModule",{value:!0}),e.default=void 0,e.default={'code[class*="language-"]':{textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",wordWrap:"normal",color:"#c3cee3",background:"#263238",fontFamily:"Roboto Mono, monospace",fontSize:"1em",lineHeight:"1.5em",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none"},'pre[class*="language-"]':{textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",wordWrap:"normal",color:"#c3cee3",background:"#263238",fontFamily:"Roboto Mono, monospace",fontSize:"1em",lineHeight:"1.5em",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none",overflow:"auto",position:"relative",margin:"0.5em 0",padding:"1.25em 1em"},'code[class*="language-"]::-moz-selection':{background:"#363636"},'pre[class*="language-"]::-moz-selection':{background:"#363636"},'code[class*="language-"] ::-moz-selection':{background:"#363636"},'pre[class*="language-"] ::-moz-selection':{background:"#363636"},'code[class*="language-"]::selection':{background:"#363636"},'pre[class*="language-"]::selection':{background:"#363636"},'code[class*="language-"] ::selection':{background:"#363636"},'pre[class*="language-"] ::selection':{background:"#363636"},':not(pre) > code[class*="language-"]':{whiteSpace:"normal",borderRadius:"0.2em",padding:"0.1em"},".language-css > code":{color:"#fd9170"},".language-sass > code":{color:"#fd9170"},".language-scss > code":{color:"#fd9170"},'[class*="language-"] .namespace':{Opacity:"0.7"},atrule:{color:"#c792ea"},"attr-name":{color:"#ffcb6b"},"attr-value":{color:"#c3e88d"},attribute:{color:"#c3e88d"},boolean:{color:"#c792ea"},builtin:{color:"#ffcb6b"},cdata:{color:"#80cbc4"},char:{color:"#80cbc4"},class:{color:"#ffcb6b"},"class-name":{color:"#f2ff00"},color:{color:"#f2ff00"},comment:{color:"#546e7a"},constant:{color:"#c792ea"},deleted:{color:"#f07178"},doctype:{color:"#546e7a"},entity:{color:"#f07178"},function:{color:"#c792ea"},hexcode:{color:"#f2ff00"},id:{color:"#c792ea",fontWeight:"bold"},important:{color:"#c792ea",fontWeight:"bold"},inserted:{color:"#80cbc4"},keyword:{color:"#c792ea",fontStyle:"italic"},number:{color:"#fd9170"},operator:{color:"#89ddff"},prolog:{color:"#546e7a"},property:{color:"#80cbc4"},"pseudo-class":{color:"#c3e88d"},"pseudo-element":{color:"#c3e88d"},punctuation:{color:"#89ddff"},regex:{color:"#f2ff00"},selector:{color:"#f07178"},string:{color:"#c3e88d"},symbol:{color:"#c792ea"},tag:{color:"#f07178"},unit:{color:"#f07178"},url:{color:"#fd9170"},variable:{color:"#f07178"}}}(Hx)),Hx}var $x={},cz;function awe(){return cz||(cz=1,function(e){Object.defineProperty(e,"__esModule",{value:!0}),e.default=void 0,e.default={'code[class*="language-"]':{color:"#d6deeb",fontFamily:'Consolas, Monaco, "Andale Mono", "Ubuntu Mono", monospace',textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",wordWrap:"normal",lineHeight:"1.5",fontSize:"1em",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none"},'pre[class*="language-"]':{color:"white",fontFamily:'Consolas, Monaco, "Andale Mono", "Ubuntu Mono", monospace',textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",wordWrap:"normal",lineHeight:"1.5",fontSize:"1em",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none",padding:"1em",margin:"0.5em 0",overflow:"auto",background:"#011627"},'pre[class*="language-"]::-moz-selection':{textShadow:"none",background:"rgba(29, 59, 83, 0.99)"},'pre[class*="language-"] ::-moz-selection':{textShadow:"none",background:"rgba(29, 59, 83, 0.99)"},'code[class*="language-"]::-moz-selection':{textShadow:"none",background:"rgba(29, 59, 83, 0.99)"},'code[class*="language-"] ::-moz-selection':{textShadow:"none",background:"rgba(29, 59, 83, 0.99)"},'pre[class*="language-"]::selection':{textShadow:"none",background:"rgba(29, 59, 83, 0.99)"},'pre[class*="language-"] ::selection':{textShadow:"none",background:"rgba(29, 59, 83, 0.99)"},'code[class*="language-"]::selection':{textShadow:"none",background:"rgba(29, 59, 83, 0.99)"},'code[class*="language-"] ::selection':{textShadow:"none",background:"rgba(29, 59, 83, 0.99)"},':not(pre) > code[class*="language-"]':{color:"white",background:"#011627",padding:"0.1em",borderRadius:"0.3em",whiteSpace:"normal"},comment:{color:"rgb(99, 119, 119)",fontStyle:"italic"},prolog:{color:"rgb(99, 119, 119)",fontStyle:"italic"},cdata:{color:"rgb(99, 119, 119)",fontStyle:"italic"},punctuation:{color:"rgb(199, 146, 234)"},".namespace":{color:"rgb(178, 204, 214)"},deleted:{color:"rgba(239, 83, 80, 0.56)",fontStyle:"italic"},symbol:{color:"rgb(128, 203, 196)"},property:{color:"rgb(128, 203, 196)"},tag:{color:"rgb(127, 219, 202)"},operator:{color:"rgb(127, 219, 202)"},keyword:{color:"rgb(127, 219, 202)"},boolean:{color:"rgb(255, 88, 116)"},number:{color:"rgb(247, 140, 108)"},constant:{color:"rgb(130, 170, 255)"},function:{color:"rgb(130, 170, 255)"},builtin:{color:"rgb(130, 170, 255)"},char:{color:"rgb(130, 170, 255)"},selector:{color:"rgb(199, 146, 234)",fontStyle:"italic"},doctype:{color:"rgb(199, 146, 234)",fontStyle:"italic"},"attr-name":{color:"rgb(173, 219, 103)",fontStyle:"italic"},inserted:{color:"rgb(173, 219, 103)",fontStyle:"italic"},string:{color:"rgb(173, 219, 103)"},url:{color:"rgb(173, 219, 103)"},entity:{color:"rgb(173, 219, 103)"},".language-css .token.string":{color:"rgb(173, 219, 103)"},".style .token.string":{color:"rgb(173, 219, 103)"},"class-name":{color:"rgb(255, 203, 139)"},atrule:{color:"rgb(255, 203, 139)"},"attr-value":{color:"rgb(255, 203, 139)"},regex:{color:"rgb(214, 222, 235)"},important:{color:"rgb(214, 222, 235)",fontWeight:"bold"},variable:{color:"rgb(214, 222, 235)"},bold:{fontWeight:"bold"},italic:{fontStyle:"italic"}}}($x)),$x}var qx={},dz;function owe(){return dz||(dz=1,function(e){Object.defineProperty(e,"__esModule",{value:!0}),e.default=void 0,e.default={'code[class*="language-"]':{color:"#f8f8f2",background:"none",fontFamily:`"Fira Code", Consolas, Monaco, 'Andale Mono', 'Ubuntu Mono', monospace`,textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",wordWrap:"normal",lineHeight:"1.5",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none"},'pre[class*="language-"]':{color:"#f8f8f2",background:"#2E3440",fontFamily:`"Fira Code", Consolas, Monaco, 'Andale Mono', 'Ubuntu Mono', monospace`,textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",wordWrap:"normal",lineHeight:"1.5",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none",padding:"1em",margin:".5em 0",overflow:"auto",borderRadius:"0.3em"},':not(pre) > code[class*="language-"]':{background:"#2E3440",padding:".1em",borderRadius:".3em",whiteSpace:"normal"},comment:{color:"#636f88"},prolog:{color:"#636f88"},doctype:{color:"#636f88"},cdata:{color:"#636f88"},punctuation:{color:"#81A1C1"},".namespace":{Opacity:".7"},property:{color:"#81A1C1"},tag:{color:"#81A1C1"},constant:{color:"#81A1C1"},symbol:{color:"#81A1C1"},deleted:{color:"#81A1C1"},number:{color:"#B48EAD"},boolean:{color:"#81A1C1"},selector:{color:"#A3BE8C"},"attr-name":{color:"#A3BE8C"},string:{color:"#A3BE8C"},char:{color:"#A3BE8C"},builtin:{color:"#A3BE8C"},inserted:{color:"#A3BE8C"},operator:{color:"#81A1C1"},entity:{color:"#81A1C1",cursor:"help"},url:{color:"#81A1C1"},".language-css .token.string":{color:"#81A1C1"},".style .token.string":{color:"#81A1C1"},variable:{color:"#81A1C1"},atrule:{color:"#88C0D0"},"attr-value":{color:"#88C0D0"},function:{color:"#88C0D0"},"class-name":{color:"#88C0D0"},keyword:{color:"#81A1C1"},regex:{color:"#EBCB8B"},important:{color:"#EBCB8B",fontWeight:"bold"},bold:{fontWeight:"bold"},italic:{fontStyle:"italic"}}}(qx)),qx}var Vx={},fz;function iwe(){return fz||(fz=1,function(e){Object.defineProperty(e,"__esModule",{value:!0}),e.default=void 0,e.default={'code[class*="language-"]':{background:"hsl(220, 13%, 18%)",color:"hsl(220, 14%, 71%)",textShadow:"0 1px rgba(0, 0, 0, 0.3)",fontFamily:'"Fira Code", "Fira Mono", Menlo, Consolas, "DejaVu Sans Mono", monospace',direction:"ltr",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",lineHeight:"1.5",MozTabSize:"2",OTabSize:"2",tabSize:"2",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none"},'pre[class*="language-"]':{background:"hsl(220, 13%, 18%)",color:"hsl(220, 14%, 71%)",textShadow:"0 1px rgba(0, 0, 0, 0.3)",fontFamily:'"Fira Code", "Fira Mono", Menlo, Consolas, "DejaVu Sans Mono", monospace',direction:"ltr",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",lineHeight:"1.5",MozTabSize:"2",OTabSize:"2",tabSize:"2",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none",padding:"1em",margin:"0.5em 0",overflow:"auto",borderRadius:"0.3em"},'code[class*="language-"]::-moz-selection':{background:"hsl(220, 13%, 28%)",color:"inherit",textShadow:"none"},'code[class*="language-"] *::-moz-selection':{background:"hsl(220, 13%, 28%)",color:"inherit",textShadow:"none"},'pre[class*="language-"] *::-moz-selection':{background:"hsl(220, 13%, 28%)",color:"inherit",textShadow:"none"},'code[class*="language-"]::selection':{background:"hsl(220, 13%, 28%)",color:"inherit",textShadow:"none"},'code[class*="language-"] *::selection':{background:"hsl(220, 13%, 28%)",color:"inherit",textShadow:"none"},'pre[class*="language-"] *::selection':{background:"hsl(220, 13%, 28%)",color:"inherit",textShadow:"none"},':not(pre) > code[class*="language-"]':{padding:"0.2em 0.3em",borderRadius:"0.3em",whiteSpace:"normal"},comment:{color:"hsl(220, 10%, 40%)",fontStyle:"italic"},prolog:{color:"hsl(220, 10%, 40%)"},cdata:{color:"hsl(220, 10%, 40%)"},doctype:{color:"hsl(220, 14%, 71%)"},punctuation:{color:"hsl(220, 14%, 71%)"},entity:{color:"hsl(220, 14%, 71%)",cursor:"help"},"attr-name":{color:"hsl(29, 54%, 61%)"},"class-name":{color:"hsl(29, 54%, 61%)"},boolean:{color:"hsl(29, 54%, 61%)"},constant:{color:"hsl(29, 54%, 61%)"},number:{color:"hsl(29, 54%, 61%)"},atrule:{color:"hsl(29, 54%, 61%)"},keyword:{color:"hsl(286, 60%, 67%)"},property:{color:"hsl(355, 65%, 65%)"},tag:{color:"hsl(355, 65%, 65%)"},symbol:{color:"hsl(355, 65%, 65%)"},deleted:{color:"hsl(355, 65%, 65%)"},important:{color:"hsl(355, 65%, 65%)"},selector:{color:"hsl(95, 38%, 62%)"},string:{color:"hsl(95, 38%, 62%)"},char:{color:"hsl(95, 38%, 62%)"},builtin:{color:"hsl(95, 38%, 62%)"},inserted:{color:"hsl(95, 38%, 62%)"},regex:{color:"hsl(95, 38%, 62%)"},"attr-value":{color:"hsl(95, 38%, 62%)"},"attr-value > .token.punctuation":{color:"hsl(95, 38%, 62%)"},variable:{color:"hsl(207, 82%, 66%)"},operator:{color:"hsl(207, 82%, 66%)"},function:{color:"hsl(207, 82%, 66%)"},url:{color:"hsl(187, 47%, 55%)"},"attr-value > .token.punctuation.attr-equals":{color:"hsl(220, 14%, 71%)"},"special-attr > .token.attr-value > .token.value.css":{color:"hsl(220, 14%, 71%)"},".language-css .token.selector":{color:"hsl(355, 65%, 65%)"},".language-css .token.property":{color:"hsl(220, 14%, 71%)"},".language-css .token.function":{color:"hsl(187, 47%, 55%)"},".language-css .token.url > .token.function":{color:"hsl(187, 47%, 55%)"},".language-css .token.url > .token.string.url":{color:"hsl(95, 38%, 62%)"},".language-css .token.important":{color:"hsl(286, 60%, 67%)"},".language-css .token.atrule .token.rule":{color:"hsl(286, 60%, 67%)"},".language-javascript .token.operator":{color:"hsl(286, 60%, 67%)"},".language-javascript .token.template-string > .token.interpolation > .token.interpolation-punctuation.punctuation":{color:"hsl(5, 48%, 51%)"},".language-json .token.operator":{color:"hsl(220, 14%, 71%)"},".language-json .token.null.keyword":{color:"hsl(29, 54%, 61%)"},".language-markdown .token.url":{color:"hsl(220, 14%, 71%)"},".language-markdown .token.url > .token.operator":{color:"hsl(220, 14%, 71%)"},".language-markdown .token.url-reference.url > .token.string":{color:"hsl(220, 14%, 71%)"},".language-markdown .token.url > .token.content":{color:"hsl(207, 82%, 66%)"},".language-markdown .token.url > .token.url":{color:"hsl(187, 47%, 55%)"},".language-markdown .token.url-reference.url":{color:"hsl(187, 47%, 55%)"},".language-markdown .token.blockquote.punctuation":{color:"hsl(220, 10%, 40%)",fontStyle:"italic"},".language-markdown .token.hr.punctuation":{color:"hsl(220, 10%, 40%)",fontStyle:"italic"},".language-markdown .token.code-snippet":{color:"hsl(95, 38%, 62%)"},".language-markdown .token.bold .token.content":{color:"hsl(29, 54%, 61%)"},".language-markdown .token.italic .token.content":{color:"hsl(286, 60%, 67%)"},".language-markdown .token.strike .token.content":{color:"hsl(355, 65%, 65%)"},".language-markdown .token.strike .token.punctuation":{color:"hsl(355, 65%, 65%)"},".language-markdown .token.list.punctuation":{color:"hsl(355, 65%, 65%)"},".language-markdown .token.title.important > .token.punctuation":{color:"hsl(355, 65%, 65%)"},bold:{fontWeight:"bold"},italic:{fontStyle:"italic"},namespace:{Opacity:"0.8"},"token.tab:not(:empty):before":{color:"hsla(220, 14%, 71%, 0.15)",textShadow:"none"},"token.cr:before":{color:"hsla(220, 14%, 71%, 0.15)",textShadow:"none"},"token.lf:before":{color:"hsla(220, 14%, 71%, 0.15)",textShadow:"none"},"token.space:before":{color:"hsla(220, 14%, 71%, 0.15)",textShadow:"none"},"div.code-toolbar > .toolbar.toolbar > .toolbar-item":{marginRight:"0.4em"},"div.code-toolbar > .toolbar.toolbar > .toolbar-item > button":{background:"hsl(220, 13%, 26%)",color:"hsl(220, 9%, 55%)",padding:"0.1em 0.4em",borderRadius:"0.3em"},"div.code-toolbar > .toolbar.toolbar > .toolbar-item > a":{background:"hsl(220, 13%, 26%)",color:"hsl(220, 9%, 55%)",padding:"0.1em 0.4em",borderRadius:"0.3em"},"div.code-toolbar > .toolbar.toolbar > .toolbar-item > span":{background:"hsl(220, 13%, 26%)",color:"hsl(220, 9%, 55%)",padding:"0.1em 0.4em",borderRadius:"0.3em"},"div.code-toolbar > .toolbar.toolbar > .toolbar-item > button:hover":{background:"hsl(220, 13%, 28%)",color:"hsl(220, 14%, 71%)"},"div.code-toolbar > .toolbar.toolbar > .toolbar-item > button:focus":{background:"hsl(220, 13%, 28%)",color:"hsl(220, 14%, 71%)"},"div.code-toolbar > .toolbar.toolbar > .toolbar-item > a:hover":{background:"hsl(220, 13%, 28%)",color:"hsl(220, 14%, 71%)"},"div.code-toolbar > .toolbar.toolbar > .toolbar-item > a:focus":{background:"hsl(220, 13%, 28%)",color:"hsl(220, 14%, 71%)"},"div.code-toolbar > .toolbar.toolbar > .toolbar-item > span:hover":{background:"hsl(220, 13%, 28%)",color:"hsl(220, 14%, 71%)"},"div.code-toolbar > .toolbar.toolbar > .toolbar-item > span:focus":{background:"hsl(220, 13%, 28%)",color:"hsl(220, 14%, 71%)"},".line-highlight.line-highlight":{background:"hsla(220, 100%, 80%, 0.04)"},".line-highlight.line-highlight:before":{background:"hsl(220, 13%, 26%)",color:"hsl(220, 14%, 71%)",padding:"0.1em 0.6em",borderRadius:"0.3em",boxShadow:"0 2px 0 0 rgba(0, 0, 0, 0.2)"},".line-highlight.line-highlight[data-end]:after":{background:"hsl(220, 13%, 26%)",color:"hsl(220, 14%, 71%)",padding:"0.1em 0.6em",borderRadius:"0.3em",boxShadow:"0 2px 0 0 rgba(0, 0, 0, 0.2)"},"pre[id].linkable-line-numbers.linkable-line-numbers span.line-numbers-rows > span:hover:before":{backgroundColor:"hsla(220, 100%, 80%, 0.04)"},".line-numbers.line-numbers .line-numbers-rows":{borderRightColor:"hsla(220, 14%, 71%, 0.15)"},".command-line .command-line-prompt":{borderRightColor:"hsla(220, 14%, 71%, 0.15)"},".line-numbers .line-numbers-rows > span:before":{color:"hsl(220, 14%, 45%)"},".command-line .command-line-prompt > span:before":{color:"hsl(220, 14%, 45%)"},".rainbow-braces .token.token.punctuation.brace-level-1":{color:"hsl(355, 65%, 65%)"},".rainbow-braces .token.token.punctuation.brace-level-5":{color:"hsl(355, 65%, 65%)"},".rainbow-braces .token.token.punctuation.brace-level-9":{color:"hsl(355, 65%, 65%)"},".rainbow-braces .token.token.punctuation.brace-level-2":{color:"hsl(95, 38%, 62%)"},".rainbow-braces .token.token.punctuation.brace-level-6":{color:"hsl(95, 38%, 62%)"},".rainbow-braces .token.token.punctuation.brace-level-10":{color:"hsl(95, 38%, 62%)"},".rainbow-braces .token.token.punctuation.brace-level-3":{color:"hsl(207, 82%, 66%)"},".rainbow-braces .token.token.punctuation.brace-level-7":{color:"hsl(207, 82%, 66%)"},".rainbow-braces .token.token.punctuation.brace-level-11":{color:"hsl(207, 82%, 66%)"},".rainbow-braces .token.token.punctuation.brace-level-4":{color:"hsl(286, 60%, 67%)"},".rainbow-braces .token.token.punctuation.brace-level-8":{color:"hsl(286, 60%, 67%)"},".rainbow-braces .token.token.punctuation.brace-level-12":{color:"hsl(286, 60%, 67%)"},"pre.diff-highlight > code .token.token.deleted:not(.prefix)":{backgroundColor:"hsla(353, 100%, 66%, 0.15)"},"pre > code.diff-highlight .token.token.deleted:not(.prefix)":{backgroundColor:"hsla(353, 100%, 66%, 0.15)"},"pre.diff-highlight > code .token.token.deleted:not(.prefix)::-moz-selection":{backgroundColor:"hsla(353, 95%, 66%, 0.25)"},"pre.diff-highlight > code .token.token.deleted:not(.prefix) *::-moz-selection":{backgroundColor:"hsla(353, 95%, 66%, 0.25)"},"pre > code.diff-highlight .token.token.deleted:not(.prefix)::-moz-selection":{backgroundColor:"hsla(353, 95%, 66%, 0.25)"},"pre > code.diff-highlight .token.token.deleted:not(.prefix) *::-moz-selection":{backgroundColor:"hsla(353, 95%, 66%, 0.25)"},"pre.diff-highlight > code .token.token.deleted:not(.prefix)::selection":{backgroundColor:"hsla(353, 95%, 66%, 0.25)"},"pre.diff-highlight > code .token.token.deleted:not(.prefix) *::selection":{backgroundColor:"hsla(353, 95%, 66%, 0.25)"},"pre > code.diff-highlight .token.token.deleted:not(.prefix)::selection":{backgroundColor:"hsla(353, 95%, 66%, 0.25)"},"pre > code.diff-highlight .token.token.deleted:not(.prefix) *::selection":{backgroundColor:"hsla(353, 95%, 66%, 0.25)"},"pre.diff-highlight > code .token.token.inserted:not(.prefix)":{backgroundColor:"hsla(137, 100%, 55%, 0.15)"},"pre > code.diff-highlight .token.token.inserted:not(.prefix)":{backgroundColor:"hsla(137, 100%, 55%, 0.15)"},"pre.diff-highlight > code .token.token.inserted:not(.prefix)::-moz-selection":{backgroundColor:"hsla(135, 73%, 55%, 0.25)"},"pre.diff-highlight > code .token.token.inserted:not(.prefix) *::-moz-selection":{backgroundColor:"hsla(135, 73%, 55%, 0.25)"},"pre > code.diff-highlight .token.token.inserted:not(.prefix)::-moz-selection":{backgroundColor:"hsla(135, 73%, 55%, 0.25)"},"pre > code.diff-highlight .token.token.inserted:not(.prefix) *::-moz-selection":{backgroundColor:"hsla(135, 73%, 55%, 0.25)"},"pre.diff-highlight > code .token.token.inserted:not(.prefix)::selection":{backgroundColor:"hsla(135, 73%, 55%, 0.25)"},"pre.diff-highlight > code .token.token.inserted:not(.prefix) *::selection":{backgroundColor:"hsla(135, 73%, 55%, 0.25)"},"pre > code.diff-highlight .token.token.inserted:not(.prefix)::selection":{backgroundColor:"hsla(135, 73%, 55%, 0.25)"},"pre > code.diff-highlight .token.token.inserted:not(.prefix) *::selection":{backgroundColor:"hsla(135, 73%, 55%, 0.25)"},".prism-previewer.prism-previewer:before":{borderColor:"hsl(224, 13%, 17%)"},".prism-previewer-gradient.prism-previewer-gradient div":{borderColor:"hsl(224, 13%, 17%)",borderRadius:"0.3em"},".prism-previewer-color.prism-previewer-color:before":{borderRadius:"0.3em"},".prism-previewer-easing.prism-previewer-easing:before":{borderRadius:"0.3em"},".prism-previewer.prism-previewer:after":{borderTopColor:"hsl(224, 13%, 17%)"},".prism-previewer-flipped.prism-previewer-flipped.after":{borderBottomColor:"hsl(224, 13%, 17%)"},".prism-previewer-angle.prism-previewer-angle:before":{background:"hsl(219, 13%, 22%)"},".prism-previewer-time.prism-previewer-time:before":{background:"hsl(219, 13%, 22%)"},".prism-previewer-easing.prism-previewer-easing":{background:"hsl(219, 13%, 22%)"},".prism-previewer-angle.prism-previewer-angle circle":{stroke:"hsl(220, 14%, 71%)",strokeOpacity:"1"},".prism-previewer-time.prism-previewer-time circle":{stroke:"hsl(220, 14%, 71%)",strokeOpacity:"1"},".prism-previewer-easing.prism-previewer-easing circle":{stroke:"hsl(220, 14%, 71%)",fill:"transparent"},".prism-previewer-easing.prism-previewer-easing path":{stroke:"hsl(220, 14%, 71%)"},".prism-previewer-easing.prism-previewer-easing line":{stroke:"hsl(220, 14%, 71%)"}}}(Vx)),Vx}var Wx={},pz;function swe(){return pz||(pz=1,function(e){Object.defineProperty(e,"__esModule",{value:!0}),e.default=void 0,e.default={'code[class*="language-"]':{background:"hsl(230, 1%, 98%)",color:"hsl(230, 8%, 24%)",fontFamily:'"Fira Code", "Fira Mono", Menlo, Consolas, "DejaVu Sans Mono", monospace',direction:"ltr",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",lineHeight:"1.5",MozTabSize:"2",OTabSize:"2",tabSize:"2",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none"},'pre[class*="language-"]':{background:"hsl(230, 1%, 98%)",color:"hsl(230, 8%, 24%)",fontFamily:'"Fira Code", "Fira Mono", Menlo, Consolas, "DejaVu Sans Mono", monospace',direction:"ltr",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",lineHeight:"1.5",MozTabSize:"2",OTabSize:"2",tabSize:"2",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none",padding:"1em",margin:"0.5em 0",overflow:"auto",borderRadius:"0.3em"},'code[class*="language-"]::-moz-selection':{background:"hsl(230, 1%, 90%)",color:"inherit"},'code[class*="language-"] *::-moz-selection':{background:"hsl(230, 1%, 90%)",color:"inherit"},'pre[class*="language-"] *::-moz-selection':{background:"hsl(230, 1%, 90%)",color:"inherit"},'code[class*="language-"]::selection':{background:"hsl(230, 1%, 90%)",color:"inherit"},'code[class*="language-"] *::selection':{background:"hsl(230, 1%, 90%)",color:"inherit"},'pre[class*="language-"] *::selection':{background:"hsl(230, 1%, 90%)",color:"inherit"},':not(pre) > code[class*="language-"]':{padding:"0.2em 0.3em",borderRadius:"0.3em",whiteSpace:"normal"},comment:{color:"hsl(230, 4%, 64%)",fontStyle:"italic"},prolog:{color:"hsl(230, 4%, 64%)"},cdata:{color:"hsl(230, 4%, 64%)"},doctype:{color:"hsl(230, 8%, 24%)"},punctuation:{color:"hsl(230, 8%, 24%)"},entity:{color:"hsl(230, 8%, 24%)",cursor:"help"},"attr-name":{color:"hsl(35, 99%, 36%)"},"class-name":{color:"hsl(35, 99%, 36%)"},boolean:{color:"hsl(35, 99%, 36%)"},constant:{color:"hsl(35, 99%, 36%)"},number:{color:"hsl(35, 99%, 36%)"},atrule:{color:"hsl(35, 99%, 36%)"},keyword:{color:"hsl(301, 63%, 40%)"},property:{color:"hsl(5, 74%, 59%)"},tag:{color:"hsl(5, 74%, 59%)"},symbol:{color:"hsl(5, 74%, 59%)"},deleted:{color:"hsl(5, 74%, 59%)"},important:{color:"hsl(5, 74%, 59%)"},selector:{color:"hsl(119, 34%, 47%)"},string:{color:"hsl(119, 34%, 47%)"},char:{color:"hsl(119, 34%, 47%)"},builtin:{color:"hsl(119, 34%, 47%)"},inserted:{color:"hsl(119, 34%, 47%)"},regex:{color:"hsl(119, 34%, 47%)"},"attr-value":{color:"hsl(119, 34%, 47%)"},"attr-value > .token.punctuation":{color:"hsl(119, 34%, 47%)"},variable:{color:"hsl(221, 87%, 60%)"},operator:{color:"hsl(221, 87%, 60%)"},function:{color:"hsl(221, 87%, 60%)"},url:{color:"hsl(198, 99%, 37%)"},"attr-value > .token.punctuation.attr-equals":{color:"hsl(230, 8%, 24%)"},"special-attr > .token.attr-value > .token.value.css":{color:"hsl(230, 8%, 24%)"},".language-css .token.selector":{color:"hsl(5, 74%, 59%)"},".language-css .token.property":{color:"hsl(230, 8%, 24%)"},".language-css .token.function":{color:"hsl(198, 99%, 37%)"},".language-css .token.url > .token.function":{color:"hsl(198, 99%, 37%)"},".language-css .token.url > .token.string.url":{color:"hsl(119, 34%, 47%)"},".language-css .token.important":{color:"hsl(301, 63%, 40%)"},".language-css .token.atrule .token.rule":{color:"hsl(301, 63%, 40%)"},".language-javascript .token.operator":{color:"hsl(301, 63%, 40%)"},".language-javascript .token.template-string > .token.interpolation > .token.interpolation-punctuation.punctuation":{color:"hsl(344, 84%, 43%)"},".language-json .token.operator":{color:"hsl(230, 8%, 24%)"},".language-json .token.null.keyword":{color:"hsl(35, 99%, 36%)"},".language-markdown .token.url":{color:"hsl(230, 8%, 24%)"},".language-markdown .token.url > .token.operator":{color:"hsl(230, 8%, 24%)"},".language-markdown .token.url-reference.url > .token.string":{color:"hsl(230, 8%, 24%)"},".language-markdown .token.url > .token.content":{color:"hsl(221, 87%, 60%)"},".language-markdown .token.url > .token.url":{color:"hsl(198, 99%, 37%)"},".language-markdown .token.url-reference.url":{color:"hsl(198, 99%, 37%)"},".language-markdown .token.blockquote.punctuation":{color:"hsl(230, 4%, 64%)",fontStyle:"italic"},".language-markdown .token.hr.punctuation":{color:"hsl(230, 4%, 64%)",fontStyle:"italic"},".language-markdown .token.code-snippet":{color:"hsl(119, 34%, 47%)"},".language-markdown .token.bold .token.content":{color:"hsl(35, 99%, 36%)"},".language-markdown .token.italic .token.content":{color:"hsl(301, 63%, 40%)"},".language-markdown .token.strike .token.content":{color:"hsl(5, 74%, 59%)"},".language-markdown .token.strike .token.punctuation":{color:"hsl(5, 74%, 59%)"},".language-markdown .token.list.punctuation":{color:"hsl(5, 74%, 59%)"},".language-markdown .token.title.important > .token.punctuation":{color:"hsl(5, 74%, 59%)"},bold:{fontWeight:"bold"},italic:{fontStyle:"italic"},namespace:{Opacity:"0.8"},"token.tab:not(:empty):before":{color:"hsla(230, 8%, 24%, 0.2)"},"token.cr:before":{color:"hsla(230, 8%, 24%, 0.2)"},"token.lf:before":{color:"hsla(230, 8%, 24%, 0.2)"},"token.space:before":{color:"hsla(230, 8%, 24%, 0.2)"},"div.code-toolbar > .toolbar.toolbar > .toolbar-item":{marginRight:"0.4em"},"div.code-toolbar > .toolbar.toolbar > .toolbar-item > button":{background:"hsl(230, 1%, 90%)",color:"hsl(230, 6%, 44%)",padding:"0.1em 0.4em",borderRadius:"0.3em"},"div.code-toolbar > .toolbar.toolbar > .toolbar-item > a":{background:"hsl(230, 1%, 90%)",color:"hsl(230, 6%, 44%)",padding:"0.1em 0.4em",borderRadius:"0.3em"},"div.code-toolbar > .toolbar.toolbar > .toolbar-item > span":{background:"hsl(230, 1%, 90%)",color:"hsl(230, 6%, 44%)",padding:"0.1em 0.4em",borderRadius:"0.3em"},"div.code-toolbar > .toolbar.toolbar > .toolbar-item > button:hover":{background:"hsl(230, 1%, 78%)",color:"hsl(230, 8%, 24%)"},"div.code-toolbar > .toolbar.toolbar > .toolbar-item > button:focus":{background:"hsl(230, 1%, 78%)",color:"hsl(230, 8%, 24%)"},"div.code-toolbar > .toolbar.toolbar > .toolbar-item > a:hover":{background:"hsl(230, 1%, 78%)",color:"hsl(230, 8%, 24%)"},"div.code-toolbar > .toolbar.toolbar > .toolbar-item > a:focus":{background:"hsl(230, 1%, 78%)",color:"hsl(230, 8%, 24%)"},"div.code-toolbar > .toolbar.toolbar > .toolbar-item > span:hover":{background:"hsl(230, 1%, 78%)",color:"hsl(230, 8%, 24%)"},"div.code-toolbar > .toolbar.toolbar > .toolbar-item > span:focus":{background:"hsl(230, 1%, 78%)",color:"hsl(230, 8%, 24%)"},".line-highlight.line-highlight":{background:"hsla(230, 8%, 24%, 0.05)"},".line-highlight.line-highlight:before":{background:"hsl(230, 1%, 90%)",color:"hsl(230, 8%, 24%)",padding:"0.1em 0.6em",borderRadius:"0.3em",boxShadow:"0 2px 0 0 rgba(0, 0, 0, 0.2)"},".line-highlight.line-highlight[data-end]:after":{background:"hsl(230, 1%, 90%)",color:"hsl(230, 8%, 24%)",padding:"0.1em 0.6em",borderRadius:"0.3em",boxShadow:"0 2px 0 0 rgba(0, 0, 0, 0.2)"},"pre[id].linkable-line-numbers.linkable-line-numbers span.line-numbers-rows > span:hover:before":{backgroundColor:"hsla(230, 8%, 24%, 0.05)"},".line-numbers.line-numbers .line-numbers-rows":{borderRightColor:"hsla(230, 8%, 24%, 0.2)"},".command-line .command-line-prompt":{borderRightColor:"hsla(230, 8%, 24%, 0.2)"},".line-numbers .line-numbers-rows > span:before":{color:"hsl(230, 1%, 62%)"},".command-line .command-line-prompt > span:before":{color:"hsl(230, 1%, 62%)"},".rainbow-braces .token.token.punctuation.brace-level-1":{color:"hsl(5, 74%, 59%)"},".rainbow-braces .token.token.punctuation.brace-level-5":{color:"hsl(5, 74%, 59%)"},".rainbow-braces .token.token.punctuation.brace-level-9":{color:"hsl(5, 74%, 59%)"},".rainbow-braces .token.token.punctuation.brace-level-2":{color:"hsl(119, 34%, 47%)"},".rainbow-braces .token.token.punctuation.brace-level-6":{color:"hsl(119, 34%, 47%)"},".rainbow-braces .token.token.punctuation.brace-level-10":{color:"hsl(119, 34%, 47%)"},".rainbow-braces .token.token.punctuation.brace-level-3":{color:"hsl(221, 87%, 60%)"},".rainbow-braces .token.token.punctuation.brace-level-7":{color:"hsl(221, 87%, 60%)"},".rainbow-braces .token.token.punctuation.brace-level-11":{color:"hsl(221, 87%, 60%)"},".rainbow-braces .token.token.punctuation.brace-level-4":{color:"hsl(301, 63%, 40%)"},".rainbow-braces .token.token.punctuation.brace-level-8":{color:"hsl(301, 63%, 40%)"},".rainbow-braces .token.token.punctuation.brace-level-12":{color:"hsl(301, 63%, 40%)"},"pre.diff-highlight > code .token.token.deleted:not(.prefix)":{backgroundColor:"hsla(353, 100%, 66%, 0.15)"},"pre > code.diff-highlight .token.token.deleted:not(.prefix)":{backgroundColor:"hsla(353, 100%, 66%, 0.15)"},"pre.diff-highlight > code .token.token.deleted:not(.prefix)::-moz-selection":{backgroundColor:"hsla(353, 95%, 66%, 0.25)"},"pre.diff-highlight > code .token.token.deleted:not(.prefix) *::-moz-selection":{backgroundColor:"hsla(353, 95%, 66%, 0.25)"},"pre > code.diff-highlight .token.token.deleted:not(.prefix)::-moz-selection":{backgroundColor:"hsla(353, 95%, 66%, 0.25)"},"pre > code.diff-highlight .token.token.deleted:not(.prefix) *::-moz-selection":{backgroundColor:"hsla(353, 95%, 66%, 0.25)"},"pre.diff-highlight > code .token.token.deleted:not(.prefix)::selection":{backgroundColor:"hsla(353, 95%, 66%, 0.25)"},"pre.diff-highlight > code .token.token.deleted:not(.prefix) *::selection":{backgroundColor:"hsla(353, 95%, 66%, 0.25)"},"pre > code.diff-highlight .token.token.deleted:not(.prefix)::selection":{backgroundColor:"hsla(353, 95%, 66%, 0.25)"},"pre > code.diff-highlight .token.token.deleted:not(.prefix) *::selection":{backgroundColor:"hsla(353, 95%, 66%, 0.25)"},"pre.diff-highlight > code .token.token.inserted:not(.prefix)":{backgroundColor:"hsla(137, 100%, 55%, 0.15)"},"pre > code.diff-highlight .token.token.inserted:not(.prefix)":{backgroundColor:"hsla(137, 100%, 55%, 0.15)"},"pre.diff-highlight > code .token.token.inserted:not(.prefix)::-moz-selection":{backgroundColor:"hsla(135, 73%, 55%, 0.25)"},"pre.diff-highlight > code .token.token.inserted:not(.prefix) *::-moz-selection":{backgroundColor:"hsla(135, 73%, 55%, 0.25)"},"pre > code.diff-highlight .token.token.inserted:not(.prefix)::-moz-selection":{backgroundColor:"hsla(135, 73%, 55%, 0.25)"},"pre > code.diff-highlight .token.token.inserted:not(.prefix) *::-moz-selection":{backgroundColor:"hsla(135, 73%, 55%, 0.25)"},"pre.diff-highlight > code .token.token.inserted:not(.prefix)::selection":{backgroundColor:"hsla(135, 73%, 55%, 0.25)"},"pre.diff-highlight > code .token.token.inserted:not(.prefix) *::selection":{backgroundColor:"hsla(135, 73%, 55%, 0.25)"},"pre > code.diff-highlight .token.token.inserted:not(.prefix)::selection":{backgroundColor:"hsla(135, 73%, 55%, 0.25)"},"pre > code.diff-highlight .token.token.inserted:not(.prefix) *::selection":{backgroundColor:"hsla(135, 73%, 55%, 0.25)"},".prism-previewer.prism-previewer:before":{borderColor:"hsl(0, 0, 95%)"},".prism-previewer-gradient.prism-previewer-gradient div":{borderColor:"hsl(0, 0, 95%)",borderRadius:"0.3em"},".prism-previewer-color.prism-previewer-color:before":{borderRadius:"0.3em"},".prism-previewer-easing.prism-previewer-easing:before":{borderRadius:"0.3em"},".prism-previewer.prism-previewer:after":{borderTopColor:"hsl(0, 0, 95%)"},".prism-previewer-flipped.prism-previewer-flipped.after":{borderBottomColor:"hsl(0, 0, 95%)"},".prism-previewer-angle.prism-previewer-angle:before":{background:"hsl(0, 0%, 100%)"},".prism-previewer-time.prism-previewer-time:before":{background:"hsl(0, 0%, 100%)"},".prism-previewer-easing.prism-previewer-easing":{background:"hsl(0, 0%, 100%)"},".prism-previewer-angle.prism-previewer-angle circle":{stroke:"hsl(230, 8%, 24%)",strokeOpacity:"1"},".prism-previewer-time.prism-previewer-time circle":{stroke:"hsl(230, 8%, 24%)",strokeOpacity:"1"},".prism-previewer-easing.prism-previewer-easing circle":{stroke:"hsl(230, 8%, 24%)",fill:"transparent"},".prism-previewer-easing.prism-previewer-easing path":{stroke:"hsl(230, 8%, 24%)"},".prism-previewer-easing.prism-previewer-easing line":{stroke:"hsl(230, 8%, 24%)"}}}(Wx)),Wx}var Yx={},gz;function lwe(){return gz||(gz=1,function(e){Object.defineProperty(e,"__esModule",{value:!0}),e.default=void 0,e.default={'code[class*="language-"]':{MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none",whiteSpace:"pre-wrap",wordBreak:"break-all",wordWrap:"break-word",fontFamily:'Menlo, Monaco, "Courier New", monospace',fontSize:"15px",lineHeight:"1.5",color:"#dccf8f",textShadow:"0"},'pre[class*="language-"]':{MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none",whiteSpace:"pre-wrap",wordBreak:"break-all",wordWrap:"break-word",fontFamily:'Menlo, Monaco, "Courier New", monospace',fontSize:"15px",lineHeight:"1.5",color:"#DCCF8F",textShadow:"0",borderRadius:"5px",border:"1px solid #000",background:"#181914 url('data:image/jpeg;base64,/9j/4AAQSkZJRgABAgAAZABkAAD/7AARRHVja3kAAQAEAAAAMAAA/+4ADkFkb2JlAGTAAAAAAf/bAIQACQYGBgcGCQcHCQ0IBwgNDwsJCQsPEQ4ODw4OERENDg4ODg0RERQUFhQUERoaHBwaGiYmJiYmKysrKysrKysrKwEJCAgJCgkMCgoMDwwODA8TDg4ODhMVDg4PDg4VGhMRERERExoXGhYWFhoXHR0aGh0dJCQjJCQrKysrKysrKysr/8AAEQgAjACMAwEiAAIRAQMRAf/EAF4AAQEBAAAAAAAAAAAAAAAAAAABBwEBAQAAAAAAAAAAAAAAAAAAAAIQAAEDAwIHAQEAAAAAAAAAAADwAREhYaExkUFRcYGxwdHh8REBAAAAAAAAAAAAAAAAAAAAAP/aAAwDAQACEQMRAD8AyGFEjHaBS2fDDs2zkhKmBKktb7km+ZwwCnXPkLVmCTMItj6AXFxRS465/BTnkAJvkLkJe+7AKKoi2AtRS2zuAWsCb5GOlBN8gKfmuGHZ8MFqIth3ALmFoFwbwKWyAlTAp17uKqBvgBD8sM4fTjhvAhkzhaRkBMKBrfs7jGPIpzy7gFrAqnC0C0gB0EWwBDW2cBVQwm+QtPpa3wBO3sVvszCnLAhkzgL5/RLf13cLQd8/AGlu0Cb5HTx9KuAEieGJEdcehS3eRTp2ATdt3CpIm+QtZwAhROXFeb7swp/ahaM3kBE/jSIUBc/AWrgBN8uNFAl+b7sAXFxFn2YLUU5Ns7gFX8C4ib+hN8gFWXwK3bZglxEJm+gKdciLPsFV/TClsgJUwKJ5FVA7tvIFrfZhVfGJDcsCKaYgAqv6YRbE+RWOWBtu7+AL3yRalXLyKqAIIfk+zARbDgFyEsncYwJvlgFRW+GEWntIi2P0BooyFxcNr8Ep3+ANLbMO+QyhvbiqdgC0kVvgUUiLYgBS2QtPbiVI1/sgOmG9uO+Y8DW+7jS2zAOnj6O2BndwuIAUtkdRN8gFoK3wwXMQyZwHVbClsuNLd4E3yAUR6FVDBR+BafQGt93LVMxJTv8ABts4CVLhcfYWsCb5kC9/BHdU8CLYFY5bMAd+eX9MGthhpbA1vu4B7+RKkaW2Yq4AQtVBBFsAJU/AuIXBhN8gGWnstefhiZyWvLAEnbYS1uzSFP6Jvn4Baxx70JKkQojLib5AVTey1jjgkKJGO0AKWyOm7N7cSpgSpAdPH0Tfd/gp1z5C1ZgKqN9J2wFxcUUuAFLZAm+QC0Fb4YUVRFsAOvj4KW2dwtYE3yAWk/wS/PLMKfmuGHZ8MAXF/Ja32Yi5haAKWz4Ydm2cSpgU693Atb7km+Zwwh+WGcPpxw3gAkzCLY+iYUDW/Z3Adc/gpzyFrAqnALkJe+7DoItgAtRS2zuKqGE3yAx0oJvkdvYrfZmALURbDuL5/RLf13cAuDeBS2RpbtAm+QFVA3wR+3fUtFHoBDJnC0jIXH0HWsgMY8inPLuOkd9chp4z20ALQLSA8cI9jYAIa2zjzjBd8gRafS1vgiUho/kAKcsCGTOGWvoOpkAtB3z8Hm8x2Ff5ADp4+lXAlIvcmwH/2Q==') repeat left top",padding:"12px",overflow:"auto"},'pre > code[class*="language-"]':{fontSize:"1em"},':not(pre) > code[class*="language-"]':{borderRadius:"5px",border:"1px solid #000",color:"#DCCF8F",background:"#181914 url('data:image/jpeg;base64,/9j/4AAQSkZJRgABAgAAZABkAAD/7AARRHVja3kAAQAEAAAAMAAA/+4ADkFkb2JlAGTAAAAAAf/bAIQACQYGBgcGCQcHCQ0IBwgNDwsJCQsPEQ4ODw4OERENDg4ODg0RERQUFhQUERoaHBwaGiYmJiYmKysrKysrKysrKwEJCAgJCgkMCgoMDwwODA8TDg4ODhMVDg4PDg4VGhMRERERExoXGhYWFhoXHR0aGh0dJCQjJCQrKysrKysrKysr/8AAEQgAjACMAwEiAAIRAQMRAf/EAF4AAQEBAAAAAAAAAAAAAAAAAAABBwEBAQAAAAAAAAAAAAAAAAAAAAIQAAEDAwIHAQEAAAAAAAAAAADwAREhYaExkUFRcYGxwdHh8REBAAAAAAAAAAAAAAAAAAAAAP/aAAwDAQACEQMRAD8AyGFEjHaBS2fDDs2zkhKmBKktb7km+ZwwCnXPkLVmCTMItj6AXFxRS465/BTnkAJvkLkJe+7AKKoi2AtRS2zuAWsCb5GOlBN8gKfmuGHZ8MFqIth3ALmFoFwbwKWyAlTAp17uKqBvgBD8sM4fTjhvAhkzhaRkBMKBrfs7jGPIpzy7gFrAqnC0C0gB0EWwBDW2cBVQwm+QtPpa3wBO3sVvszCnLAhkzgL5/RLf13cLQd8/AGlu0Cb5HTx9KuAEieGJEdcehS3eRTp2ATdt3CpIm+QtZwAhROXFeb7swp/ahaM3kBE/jSIUBc/AWrgBN8uNFAl+b7sAXFxFn2YLUU5Ns7gFX8C4ib+hN8gFWXwK3bZglxEJm+gKdciLPsFV/TClsgJUwKJ5FVA7tvIFrfZhVfGJDcsCKaYgAqv6YRbE+RWOWBtu7+AL3yRalXLyKqAIIfk+zARbDgFyEsncYwJvlgFRW+GEWntIi2P0BooyFxcNr8Ep3+ANLbMO+QyhvbiqdgC0kVvgUUiLYgBS2QtPbiVI1/sgOmG9uO+Y8DW+7jS2zAOnj6O2BndwuIAUtkdRN8gFoK3wwXMQyZwHVbClsuNLd4E3yAUR6FVDBR+BafQGt93LVMxJTv8ABts4CVLhcfYWsCb5kC9/BHdU8CLYFY5bMAd+eX9MGthhpbA1vu4B7+RKkaW2Yq4AQtVBBFsAJU/AuIXBhN8gGWnstefhiZyWvLAEnbYS1uzSFP6Jvn4Baxx70JKkQojLib5AVTey1jjgkKJGO0AKWyOm7N7cSpgSpAdPH0Tfd/gp1z5C1ZgKqN9J2wFxcUUuAFLZAm+QC0Fb4YUVRFsAOvj4KW2dwtYE3yAWk/wS/PLMKfmuGHZ8MAXF/Ja32Yi5haAKWz4Ydm2cSpgU693Atb7km+Zwwh+WGcPpxw3gAkzCLY+iYUDW/Z3Adc/gpzyFrAqnALkJe+7DoItgAtRS2zuKqGE3yAx0oJvkdvYrfZmALURbDuL5/RLf13cAuDeBS2RpbtAm+QFVA3wR+3fUtFHoBDJnC0jIXH0HWsgMY8inPLuOkd9chp4z20ALQLSA8cI9jYAIa2zjzjBd8gRafS1vgiUho/kAKcsCGTOGWvoOpkAtB3z8Hm8x2Ff5ADp4+lXAlIvcmwH/2Q==') repeat left top",padding:"2px 6px"},namespace:{Opacity:".7"},comment:{color:"#586e75",fontStyle:"italic"},prolog:{color:"#586e75",fontStyle:"italic"},doctype:{color:"#586e75",fontStyle:"italic"},cdata:{color:"#586e75",fontStyle:"italic"},number:{color:"#b89859"},string:{color:"#468966"},char:{color:"#468966"},builtin:{color:"#468966"},inserted:{color:"#468966"},"attr-name":{color:"#b89859"},operator:{color:"#dccf8f"},entity:{color:"#dccf8f",cursor:"help"},url:{color:"#dccf8f"},".language-css .token.string":{color:"#dccf8f"},".style .token.string":{color:"#dccf8f"},selector:{color:"#859900"},regex:{color:"#859900"},atrule:{color:"#cb4b16"},keyword:{color:"#cb4b16"},"attr-value":{color:"#468966"},function:{color:"#b58900"},variable:{color:"#b58900"},placeholder:{color:"#b58900"},property:{color:"#b89859"},tag:{color:"#ffb03b"},boolean:{color:"#b89859"},constant:{color:"#b89859"},symbol:{color:"#b89859"},important:{color:"#dc322f"},statement:{color:"#dc322f"},deleted:{color:"#dc322f"},punctuation:{color:"#dccf8f"},bold:{fontWeight:"bold"},italic:{fontStyle:"italic"}}}(Yx)),Yx}var Kx={},hz;function uwe(){return hz||(hz=1,function(e){Object.defineProperty(e,"__esModule",{value:!0}),e.default=void 0,e.default={"code[class*='language-']":{color:"#9efeff",direction:"ltr",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none",fontFamily:"'Operator Mono', 'Fira Code', Consolas, Monaco, 'Andale Mono', 'Ubuntu Mono', monospace",fontWeight:"400",fontSize:"17px",lineHeight:"25px",letterSpacing:"0.5px",textShadow:"0 1px #222245"},"pre[class*='language-']":{color:"#9efeff",direction:"ltr",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none",fontFamily:"'Operator Mono', 'Fira Code', Consolas, Monaco, 'Andale Mono', 'Ubuntu Mono', monospace",fontWeight:"400",fontSize:"17px",lineHeight:"25px",letterSpacing:"0.5px",textShadow:"0 1px #222245",padding:"2em",margin:"0.5em 0",overflow:"auto",background:"#1e1e3f"},"pre[class*='language-']::-moz-selection":{color:"inherit",background:"#a599e9"},"pre[class*='language-'] ::-moz-selection":{color:"inherit",background:"#a599e9"},"code[class*='language-']::-moz-selection":{color:"inherit",background:"#a599e9"},"code[class*='language-'] ::-moz-selection":{color:"inherit",background:"#a599e9"},"pre[class*='language-']::selection":{color:"inherit",background:"#a599e9"},"pre[class*='language-'] ::selection":{color:"inherit",background:"#a599e9"},"code[class*='language-']::selection":{color:"inherit",background:"#a599e9"},"code[class*='language-'] ::selection":{color:"inherit",background:"#a599e9"},":not(pre) > code[class*='language-']":{background:"#1e1e3f",padding:"0.1em",borderRadius:"0.3em"},"":{fontWeight:"400"},comment:{color:"#b362ff"},prolog:{color:"#b362ff"},cdata:{color:"#b362ff"},delimiter:{color:"#ff9d00"},keyword:{color:"#ff9d00"},selector:{color:"#ff9d00"},important:{color:"#ff9d00"},atrule:{color:"#ff9d00"},operator:{color:"rgb(255, 180, 84)",background:"none"},"attr-name":{color:"rgb(255, 180, 84)"},punctuation:{color:"#ffffff"},boolean:{color:"rgb(255, 98, 140)"},tag:{color:"rgb(255, 157, 0)"},"tag.punctuation":{color:"rgb(255, 157, 0)"},doctype:{color:"rgb(255, 157, 0)"},builtin:{color:"rgb(255, 157, 0)"},entity:{color:"#6897bb",background:"none"},symbol:{color:"#6897bb"},number:{color:"#ff628c"},property:{color:"#ff628c"},constant:{color:"#ff628c"},variable:{color:"#ff628c"},string:{color:"#a5ff90"},char:{color:"#a5ff90"},"attr-value":{color:"#a5c261"},"attr-value.punctuation":{color:"#a5c261"},"attr-value.punctuation:first-child":{color:"#a9b7c6"},url:{color:"#287bde",textDecoration:"underline",background:"none"},function:{color:"rgb(250, 208, 0)"},regex:{background:"#364135"},bold:{fontWeight:"bold"},italic:{fontStyle:"italic"},inserted:{background:"#00ff00"},deleted:{background:"#ff000d"},"code.language-css .token.property":{color:"#a9b7c6"},"code.language-css .token.property + .token.punctuation":{color:"#a9b7c6"},"code.language-css .token.id":{color:"#ffc66d"},"code.language-css .token.selector > .token.class":{color:"#ffc66d"},"code.language-css .token.selector > .token.attribute":{color:"#ffc66d"},"code.language-css .token.selector > .token.pseudo-class":{color:"#ffc66d"},"code.language-css .token.selector > .token.pseudo-element":{color:"#ffc66d"},"class-name":{color:"#fb94ff"},".language-css .token.string":{background:"none"},".style .token.string":{background:"none"},".line-highlight.line-highlight":{marginTop:"36px",background:"linear-gradient(to right, rgba(179, 98, 255, 0.17), transparent)"},".line-highlight.line-highlight:before":{content:"''"},".line-highlight.line-highlight[data-end]:after":{content:"''"}}}(Kx)),Kx}var Xx={},mz;function cwe(){return mz||(mz=1,function(e){Object.defineProperty(e,"__esModule",{value:!0}),e.default=void 0,e.default={'code[class*="language-"]':{color:"#839496",textShadow:"0 1px rgba(0, 0, 0, 0.3)",fontFamily:"Inconsolata, Monaco, Consolas, 'Courier New', Courier, monospace",direction:"ltr",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",lineHeight:"1.5",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none"},'pre[class*="language-"]':{color:"#839496",textShadow:"0 1px rgba(0, 0, 0, 0.3)",fontFamily:"Inconsolata, Monaco, Consolas, 'Courier New', Courier, monospace",direction:"ltr",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",lineHeight:"1.5",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none",padding:"1em",margin:".5em 0",overflow:"auto",borderRadius:"0.3em",background:"#002b36"},':not(pre) > code[class*="language-"]':{background:"#002b36",padding:".1em",borderRadius:".3em"},comment:{color:"#586e75"},prolog:{color:"#586e75"},doctype:{color:"#586e75"},cdata:{color:"#586e75"},punctuation:{color:"#93a1a1"},".namespace":{Opacity:".7"},property:{color:"#268bd2"},keyword:{color:"#268bd2"},tag:{color:"#268bd2"},"class-name":{color:"#FFFFB6",textDecoration:"underline"},boolean:{color:"#b58900"},constant:{color:"#b58900"},symbol:{color:"#dc322f"},deleted:{color:"#dc322f"},number:{color:"#859900"},selector:{color:"#859900"},"attr-name":{color:"#859900"},string:{color:"#859900"},char:{color:"#859900"},builtin:{color:"#859900"},inserted:{color:"#859900"},variable:{color:"#268bd2"},operator:{color:"#EDEDED"},function:{color:"#268bd2"},regex:{color:"#E9C062"},important:{color:"#fd971f",fontWeight:"bold"},entity:{color:"#FFFFB6",cursor:"help"},url:{color:"#96CBFE"},".language-css .token.string":{color:"#87C38A"},".style .token.string":{color:"#87C38A"},bold:{fontWeight:"bold"},italic:{fontStyle:"italic"},atrule:{color:"#F9EE98"},"attr-value":{color:"#F9EE98"}}}(Xx)),Xx}var Zx={},bz;function dwe(){return bz||(bz=1,function(e){Object.defineProperty(e,"__esModule",{value:!0}),e.default=void 0,e.default={'code[class*="language-"]':{color:"#f92aad",textShadow:"0 0 2px #100c0f, 0 0 5px #dc078e33, 0 0 10px #fff3",background:"none",fontFamily:"Consolas, Monaco, 'Andale Mono', 'Ubuntu Mono', monospace",fontSize:"1em",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",wordWrap:"normal",lineHeight:"1.5",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none"},'pre[class*="language-"]':{color:"#f92aad",textShadow:"0 0 2px #100c0f, 0 0 5px #dc078e33, 0 0 10px #fff3",background:"none",fontFamily:"Consolas, Monaco, 'Andale Mono', 'Ubuntu Mono', monospace",fontSize:"1em",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",wordWrap:"normal",lineHeight:"1.5",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none",padding:"1em",margin:".5em 0",overflow:"auto",backgroundColor:"transparent !important",backgroundImage:"linear-gradient(to bottom, #2a2139 75%, #34294f)"},':not(pre) > code[class*="language-"]':{backgroundColor:"transparent !important",backgroundImage:"linear-gradient(to bottom, #2a2139 75%, #34294f)",padding:".1em",borderRadius:".3em",whiteSpace:"normal"},comment:{color:"#8e8e8e"},"block-comment":{color:"#8e8e8e"},prolog:{color:"#8e8e8e"},doctype:{color:"#8e8e8e"},cdata:{color:"#8e8e8e"},punctuation:{color:"#ccc"},tag:{color:"#e2777a"},"attr-name":{color:"#e2777a"},namespace:{color:"#e2777a"},number:{color:"#e2777a"},unit:{color:"#e2777a"},hexcode:{color:"#e2777a"},deleted:{color:"#e2777a"},property:{color:"#72f1b8",textShadow:"0 0 2px #100c0f, 0 0 10px #257c5575, 0 0 35px #21272475"},selector:{color:"#72f1b8",textShadow:"0 0 2px #100c0f, 0 0 10px #257c5575, 0 0 35px #21272475"},"function-name":{color:"#6196cc"},boolean:{color:"#fdfdfd",textShadow:"0 0 2px #001716, 0 0 3px #03edf975, 0 0 5px #03edf975, 0 0 8px #03edf975"},"selector.id":{color:"#fdfdfd",textShadow:"0 0 2px #001716, 0 0 3px #03edf975, 0 0 5px #03edf975, 0 0 8px #03edf975"},function:{color:"#fdfdfd",textShadow:"0 0 2px #001716, 0 0 3px #03edf975, 0 0 5px #03edf975, 0 0 8px #03edf975"},"class-name":{color:"#fff5f6",textShadow:"0 0 2px #000, 0 0 10px #fc1f2c75, 0 0 5px #fc1f2c75, 0 0 25px #fc1f2c75"},constant:{color:"#f92aad",textShadow:"0 0 2px #100c0f, 0 0 5px #dc078e33, 0 0 10px #fff3"},symbol:{color:"#f92aad",textShadow:"0 0 2px #100c0f, 0 0 5px #dc078e33, 0 0 10px #fff3"},important:{color:"#f4eee4",textShadow:"0 0 2px #393a33, 0 0 8px #f39f0575, 0 0 2px #f39f0575",fontWeight:"bold"},atrule:{color:"#f4eee4",textShadow:"0 0 2px #393a33, 0 0 8px #f39f0575, 0 0 2px #f39f0575"},keyword:{color:"#f4eee4",textShadow:"0 0 2px #393a33, 0 0 8px #f39f0575, 0 0 2px #f39f0575"},"selector.class":{color:"#f4eee4",textShadow:"0 0 2px #393a33, 0 0 8px #f39f0575, 0 0 2px #f39f0575"},builtin:{color:"#f4eee4",textShadow:"0 0 2px #393a33, 0 0 8px #f39f0575, 0 0 2px #f39f0575"},string:{color:"#f87c32"},char:{color:"#f87c32"},"attr-value":{color:"#f87c32"},regex:{color:"#f87c32"},variable:{color:"#f87c32"},operator:{color:"#67cdcc"},entity:{color:"#67cdcc",cursor:"help"},url:{color:"#67cdcc"},bold:{fontWeight:"bold"},italic:{fontStyle:"italic"},inserted:{color:"green"}}}(Zx)),Zx}var Qx={},yz;function fwe(){return yz||(yz=1,function(e){Object.defineProperty(e,"__esModule",{value:!0}),e.default=void 0,e.default={'code[class*="language-"]':{color:"#393A34",fontFamily:'"Consolas", "Bitstream Vera Sans Mono", "Courier New", Courier, monospace',direction:"ltr",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",fontSize:".9em",lineHeight:"1.2em",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none"},'pre[class*="language-"]':{color:"#393A34",fontFamily:'"Consolas", "Bitstream Vera Sans Mono", "Courier New", Courier, monospace',direction:"ltr",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",fontSize:".9em",lineHeight:"1.2em",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none",padding:"1em",margin:".5em 0",overflow:"auto",border:"1px solid #dddddd",backgroundColor:"white"},'pre > code[class*="language-"]':{fontSize:"1em"},'pre[class*="language-"]::-moz-selection':{background:"#C1DEF1"},'pre[class*="language-"] ::-moz-selection':{background:"#C1DEF1"},'code[class*="language-"]::-moz-selection':{background:"#C1DEF1"},'code[class*="language-"] ::-moz-selection':{background:"#C1DEF1"},'pre[class*="language-"]::selection':{background:"#C1DEF1"},'pre[class*="language-"] ::selection':{background:"#C1DEF1"},'code[class*="language-"]::selection':{background:"#C1DEF1"},'code[class*="language-"] ::selection':{background:"#C1DEF1"},':not(pre) > code[class*="language-"]':{padding:".2em",paddingTop:"1px",paddingBottom:"1px",background:"#f8f8f8",border:"1px solid #dddddd"},comment:{color:"#008000",fontStyle:"italic"},prolog:{color:"#008000",fontStyle:"italic"},doctype:{color:"#008000",fontStyle:"italic"},cdata:{color:"#008000",fontStyle:"italic"},namespace:{Opacity:".7"},string:{color:"#A31515"},punctuation:{color:"#393A34"},operator:{color:"#393A34"},url:{color:"#36acaa"},symbol:{color:"#36acaa"},number:{color:"#36acaa"},boolean:{color:"#36acaa"},variable:{color:"#36acaa"},constant:{color:"#36acaa"},inserted:{color:"#36acaa"},atrule:{color:"#0000ff"},keyword:{color:"#0000ff"},"attr-value":{color:"#0000ff"},".language-autohotkey .token.selector":{color:"#0000ff"},".language-json .token.boolean":{color:"#0000ff"},".language-json .token.number":{color:"#0000ff"},'code[class*="language-css"]':{color:"#0000ff"},function:{color:"#393A34"},deleted:{color:"#9a050f"},".language-autohotkey .token.tag":{color:"#9a050f"},selector:{color:"#800000"},".language-autohotkey .token.keyword":{color:"#00009f"},important:{color:"#e90",fontWeight:"bold"},bold:{fontWeight:"bold"},italic:{fontStyle:"italic"},"class-name":{color:"#2B91AF"},".language-json .token.property":{color:"#2B91AF"},tag:{color:"#800000"},"attr-name":{color:"#ff0000"},property:{color:"#ff0000"},regex:{color:"#ff0000"},entity:{color:"#ff0000"},"directive.tag.tag":{background:"#ffff00",color:"#393A34"},".line-numbers.line-numbers .line-numbers-rows":{borderRightColor:"#a5a5a5"},".line-numbers .line-numbers-rows > span:before":{color:"#2B91AF"},".line-highlight.line-highlight":{background:"linear-gradient(to right, rgba(193, 222, 241, 0.2) 70%, rgba(221, 222, 241, 0))"}}}(Qx)),Qx}var Jx={},vz;function pwe(){return vz||(vz=1,function(e){Object.defineProperty(e,"__esModule",{value:!0}),e.default=void 0,e.default={'pre[class*="language-"]':{color:"#d4d4d4",fontSize:"13px",textShadow:"none",fontFamily:'Menlo, Monaco, Consolas, "Andale Mono", "Ubuntu Mono", "Courier New", monospace',direction:"ltr",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",lineHeight:"1.5",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none",padding:"1em",margin:".5em 0",overflow:"auto",background:"#1e1e1e"},'code[class*="language-"]':{color:"#d4d4d4",fontSize:"13px",textShadow:"none",fontFamily:'Menlo, Monaco, Consolas, "Andale Mono", "Ubuntu Mono", "Courier New", monospace',direction:"ltr",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",lineHeight:"1.5",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none"},'pre[class*="language-"]::selection':{textShadow:"none",background:"#264F78"},'code[class*="language-"]::selection':{textShadow:"none",background:"#264F78"},'pre[class*="language-"] *::selection':{textShadow:"none",background:"#264F78"},'code[class*="language-"] *::selection':{textShadow:"none",background:"#264F78"},':not(pre) > code[class*="language-"]':{padding:".1em .3em",borderRadius:".3em",color:"#db4c69",background:"#1e1e1e"},".namespace":{Opacity:".7"},"doctype.doctype-tag":{color:"#569CD6"},"doctype.name":{color:"#9cdcfe"},comment:{color:"#6a9955"},prolog:{color:"#6a9955"},punctuation:{color:"#d4d4d4"},".language-html .language-css .token.punctuation":{color:"#d4d4d4"},".language-html .language-javascript .token.punctuation":{color:"#d4d4d4"},property:{color:"#9cdcfe"},tag:{color:"#569cd6"},boolean:{color:"#569cd6"},number:{color:"#b5cea8"},constant:{color:"#9cdcfe"},symbol:{color:"#b5cea8"},inserted:{color:"#b5cea8"},unit:{color:"#b5cea8"},selector:{color:"#d7ba7d"},"attr-name":{color:"#9cdcfe"},string:{color:"#ce9178"},char:{color:"#ce9178"},builtin:{color:"#ce9178"},deleted:{color:"#ce9178"},".language-css .token.string.url":{textDecoration:"underline"},operator:{color:"#d4d4d4"},entity:{color:"#569cd6"},"operator.arrow":{color:"#569CD6"},atrule:{color:"#ce9178"},"atrule.rule":{color:"#c586c0"},"atrule.url":{color:"#9cdcfe"},"atrule.url.function":{color:"#dcdcaa"},"atrule.url.punctuation":{color:"#d4d4d4"},keyword:{color:"#569CD6"},"keyword.module":{color:"#c586c0"},"keyword.control-flow":{color:"#c586c0"},function:{color:"#dcdcaa"},"function.maybe-class-name":{color:"#dcdcaa"},regex:{color:"#d16969"},important:{color:"#569cd6"},italic:{fontStyle:"italic"},"class-name":{color:"#4ec9b0"},"maybe-class-name":{color:"#4ec9b0"},console:{color:"#9cdcfe"},parameter:{color:"#9cdcfe"},interpolation:{color:"#9cdcfe"},"punctuation.interpolation-punctuation":{color:"#569cd6"},variable:{color:"#9cdcfe"},"imports.maybe-class-name":{color:"#9cdcfe"},"exports.maybe-class-name":{color:"#9cdcfe"},escape:{color:"#d7ba7d"},"tag.punctuation":{color:"#808080"},cdata:{color:"#808080"},"attr-value":{color:"#ce9178"},"attr-value.punctuation":{color:"#ce9178"},"attr-value.punctuation.attr-equals":{color:"#d4d4d4"},namespace:{color:"#4ec9b0"},'pre[class*="language-javascript"]':{color:"#9cdcfe"},'code[class*="language-javascript"]':{color:"#9cdcfe"},'pre[class*="language-jsx"]':{color:"#9cdcfe"},'code[class*="language-jsx"]':{color:"#9cdcfe"},'pre[class*="language-typescript"]':{color:"#9cdcfe"},'code[class*="language-typescript"]':{color:"#9cdcfe"},'pre[class*="language-tsx"]':{color:"#9cdcfe"},'code[class*="language-tsx"]':{color:"#9cdcfe"},'pre[class*="language-css"]':{color:"#ce9178"},'code[class*="language-css"]':{color:"#ce9178"},'pre[class*="language-html"]':{color:"#d4d4d4"},'code[class*="language-html"]':{color:"#d4d4d4"},".language-regex .token.anchor":{color:"#dcdcaa"},".language-html .token.punctuation":{color:"#808080"},'pre[class*="language-"] > code[class*="language-"]':{position:"relative",zIndex:"1"},".line-highlight.line-highlight":{background:"#f7ebc6",boxShadow:"inset 5px 0 0 #f7d87c",zIndex:"0"}}}(Jx)),Jx}var ek={},Sz;function gwe(){return Sz||(Sz=1,function(e){Object.defineProperty(e,"__esModule",{value:!0}),e.default=void 0,e.default={'code[class*="language-"]':{MozTabSize:"2",OTabSize:"2",tabSize:"2",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none",whiteSpace:"pre-wrap",wordWrap:"normal",fontFamily:'Menlo, Monaco, "Courier New", monospace',fontSize:"14px",color:"#76d9e6",textShadow:"none"},'pre[class*="language-"]':{MozTabSize:"2",OTabSize:"2",tabSize:"2",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none",whiteSpace:"pre-wrap",wordWrap:"normal",fontFamily:'Menlo, Monaco, "Courier New", monospace',fontSize:"14px",color:"#76d9e6",textShadow:"none",background:"#2a2a2a",padding:"15px",borderRadius:"4px",border:"1px solid #e1e1e8",overflow:"auto",position:"relative"},'pre > code[class*="language-"]':{fontSize:"1em"},':not(pre) > code[class*="language-"]':{background:"#2a2a2a",padding:"0.15em 0.2em 0.05em",borderRadius:".3em",border:"0.13em solid #7a6652",boxShadow:"1px 1px 0.3em -0.1em #000 inset"},'pre[class*="language-"] code':{whiteSpace:"pre",display:"block"},namespace:{Opacity:".7"},comment:{color:"#6f705e"},prolog:{color:"#6f705e"},doctype:{color:"#6f705e"},cdata:{color:"#6f705e"},operator:{color:"#a77afe"},boolean:{color:"#a77afe"},number:{color:"#a77afe"},"attr-name":{color:"#e6d06c"},string:{color:"#e6d06c"},entity:{color:"#e6d06c",cursor:"help"},url:{color:"#e6d06c"},".language-css .token.string":{color:"#e6d06c"},".style .token.string":{color:"#e6d06c"},selector:{color:"#a6e22d"},inserted:{color:"#a6e22d"},atrule:{color:"#ef3b7d"},"attr-value":{color:"#ef3b7d"},keyword:{color:"#ef3b7d"},important:{color:"#ef3b7d",fontWeight:"bold"},deleted:{color:"#ef3b7d"},regex:{color:"#76d9e6"},statement:{color:"#76d9e6",fontWeight:"bold"},placeholder:{color:"#fff"},variable:{color:"#fff"},bold:{fontWeight:"bold"},punctuation:{color:"#bebec5"},italic:{fontStyle:"italic"},"code.language-markup":{color:"#f9f9f9"},"code.language-markup .token.tag":{color:"#ef3b7d"},"code.language-markup .token.attr-name":{color:"#a6e22d"},"code.language-markup .token.attr-value":{color:"#e6d06c"},"code.language-markup .token.style":{color:"#76d9e6"},"code.language-markup .token.script":{color:"#76d9e6"},"code.language-markup .token.script .token.keyword":{color:"#76d9e6"},".line-highlight.line-highlight":{padding:"0",background:"rgba(255, 255, 255, 0.08)"},".line-highlight.line-highlight:before":{padding:"0.2em 0.5em",backgroundColor:"rgba(255, 255, 255, 0.4)",color:"black",height:"1em",lineHeight:"1em",boxShadow:"0 1px 1px rgba(255, 255, 255, 0.7)"},".line-highlight.line-highlight[data-end]:after":{padding:"0.2em 0.5em",backgroundColor:"rgba(255, 255, 255, 0.4)",color:"black",height:"1em",lineHeight:"1em",boxShadow:"0 1px 1px rgba(255, 255, 255, 0.7)"}}}(ek)),ek}var tk={},Ez;function hwe(){return Ez||(Ez=1,function(e){Object.defineProperty(e,"__esModule",{value:!0}),e.default=void 0,e.default={'code[class*="language-"]':{color:"#22da17",fontFamily:"monospace",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",wordWrap:"normal",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none",lineHeight:"25px",fontSize:"18px",margin:"5px 0"},'pre[class*="language-"]':{color:"white",fontFamily:"monospace",textAlign:"left",whiteSpace:"pre",wordSpacing:"normal",wordBreak:"normal",wordWrap:"normal",MozTabSize:"4",OTabSize:"4",tabSize:"4",WebkitHyphens:"none",MozHyphens:"none",msHyphens:"none",hyphens:"none",lineHeight:"25px",fontSize:"18px",margin:"0.5em 0",background:"#0a143c",padding:"1em",overflow:"auto"},'pre[class*="language-"] *':{fontFamily:"monospace"},':not(pre) > code[class*="language-"]':{color:"white",background:"#0a143c",padding:"0.1em",borderRadius:"0.3em",whiteSpace:"normal"},'pre[class*="language-"]::-moz-selection':{textShadow:"none",background:"rgba(29, 59, 83, 0.99)"},'pre[class*="language-"] ::-moz-selection':{textShadow:"none",background:"rgba(29, 59, 83, 0.99)"},'code[class*="language-"]::-moz-selection':{textShadow:"none",background:"rgba(29, 59, 83, 0.99)"},'code[class*="language-"] ::-moz-selection':{textShadow:"none",background:"rgba(29, 59, 83, 0.99)"},'pre[class*="language-"]::selection':{textShadow:"none",background:"rgba(29, 59, 83, 0.99)"},'pre[class*="language-"] ::selection':{textShadow:"none",background:"rgba(29, 59, 83, 0.99)"},'code[class*="language-"]::selection':{textShadow:"none",background:"rgba(29, 59, 83, 0.99)"},'code[class*="language-"] ::selection':{textShadow:"none",background:"rgba(29, 59, 83, 0.99)"},comment:{color:"rgb(99, 119, 119)",fontStyle:"italic"},prolog:{color:"rgb(99, 119, 119)",fontStyle:"italic"},cdata:{color:"rgb(99, 119, 119)",fontStyle:"italic"},punctuation:{color:"rgb(199, 146, 234)"},".namespace":{color:"rgb(178, 204, 214)"},deleted:{color:"rgba(239, 83, 80, 0.56)",fontStyle:"italic"},symbol:{color:"rgb(128, 203, 196)"},property:{color:"rgb(128, 203, 196)"},tag:{color:"rgb(127, 219, 202)"},operator:{color:"rgb(127, 219, 202)"},keyword:{color:"rgb(127, 219, 202)"},boolean:{color:"rgb(255, 88, 116)"},number:{color:"rgb(247, 140, 108)"},constant:{color:"rgb(34 183 199)"},function:{color:"rgb(34 183 199)"},builtin:{color:"rgb(34 183 199)"},char:{color:"rgb(34 183 199)"},selector:{color:"rgb(199, 146, 234)",fontStyle:"italic"},doctype:{color:"rgb(199, 146, 234)",fontStyle:"italic"},"attr-name":{color:"rgb(173, 219, 103)",fontStyle:"italic"},inserted:{color:"rgb(173, 219, 103)",fontStyle:"italic"},string:{color:"rgb(173, 219, 103)"},url:{color:"rgb(173, 219, 103)"},entity:{color:"rgb(173, 219, 103)"},".language-css .token.string":{color:"rgb(173, 219, 103)"},".style .token.string":{color:"rgb(173, 219, 103)"},"class-name":{color:"rgb(255, 203, 139)"},atrule:{color:"rgb(255, 203, 139)"},"attr-value":{color:"rgb(255, 203, 139)"},regex:{color:"rgb(214, 222, 235)"},important:{color:"rgb(214, 222, 235)",fontWeight:"bold"},variable:{color:"rgb(214, 222, 235)"},bold:{fontWeight:"bold"},italic:{fontStyle:"italic"}}}(tk)),tk}var wz;function mwe(){return wz||(wz=1,function(e){var t=TEe();Object.defineProperty(e,"__esModule",{value:!0}),Object.defineProperty(e,"a11yDark",{enumerable:!0,get:function(){return p.default}}),Object.defineProperty(e,"atomDark",{enumerable:!0,get:function(){return g.default}}),Object.defineProperty(e,"base16AteliersulphurpoolLight",{enumerable:!0,get:function(){return m.default}}),Object.defineProperty(e,"cb",{enumerable:!0,get:function(){return b.default}}),Object.defineProperty(e,"coldarkCold",{enumerable:!0,get:function(){return y.default}}),Object.defineProperty(e,"coldarkDark",{enumerable:!0,get:function(){return S.default}}),Object.defineProperty(e,"coy",{enumerable:!0,get:function(){return n.default}}),Object.defineProperty(e,"coyWithoutShadows",{enumerable:!0,get:function(){return x.default}}),Object.defineProperty(e,"darcula",{enumerable:!0,get:function(){return R.default}}),Object.defineProperty(e,"dark",{enumerable:!0,get:function(){return r.default}}),Object.defineProperty(e,"dracula",{enumerable:!0,get:function(){return k.default}}),Object.defineProperty(e,"duotoneDark",{enumerable:!0,get:function(){return A.default}}),Object.defineProperty(e,"duotoneEarth",{enumerable:!0,get:function(){return C.default}}),Object.defineProperty(e,"duotoneForest",{enumerable:!0,get:function(){return N.default}}),Object.defineProperty(e,"duotoneLight",{enumerable:!0,get:function(){return _.default}}),Object.defineProperty(e,"duotoneSea",{enumerable:!0,get:function(){return O.default}}),Object.defineProperty(e,"duotoneSpace",{enumerable:!0,get:function(){return F.default}}),Object.defineProperty(e,"funky",{enumerable:!0,get:function(){return a.default}}),Object.defineProperty(e,"ghcolors",{enumerable:!0,get:function(){return L.default}}),Object.defineProperty(e,"gruvboxDark",{enumerable:!0,get:function(){return I.default}}),Object.defineProperty(e,"gruvboxLight",{enumerable:!0,get:function(){return H.default}}),Object.defineProperty(e,"holiTheme",{enumerable:!0,get:function(){return $.default}}),Object.defineProperty(e,"hopscotch",{enumerable:!0,get:function(){return U.default}}),Object.defineProperty(e,"lucario",{enumerable:!0,get:function(){return Y.default}}),Object.defineProperty(e,"materialDark",{enumerable:!0,get:function(){return Z.default}}),Object.defineProperty(e,"materialLight",{enumerable:!0,get:function(){return j.default}}),Object.defineProperty(e,"materialOceanic",{enumerable:!0,get:function(){return G.default}}),Object.defineProperty(e,"nightOwl",{enumerable:!0,get:function(){return z.default}}),Object.defineProperty(e,"nord",{enumerable:!0,get:function(){return K.default}}),Object.defineProperty(e,"okaidia",{enumerable:!0,get:function(){return o.default}}),Object.defineProperty(e,"oneDark",{enumerable:!0,get:function(){return D.default}}),Object.defineProperty(e,"oneLight",{enumerable:!0,get:function(){return V.default}}),Object.defineProperty(e,"pojoaque",{enumerable:!0,get:function(){return B.default}}),Object.defineProperty(e,"prism",{enumerable:!0,get:function(){return d.default}}),Object.defineProperty(e,"shadesOfPurple",{enumerable:!0,get:function(){return M.default}}),Object.defineProperty(e,"solarizedDarkAtom",{enumerable:!0,get:function(){return W.default}}),Object.defineProperty(e,"solarizedlight",{enumerable:!0,get:function(){return s.default}}),Object.defineProperty(e,"synthwave84",{enumerable:!0,get:function(){return Q.default}}),Object.defineProperty(e,"tomorrow",{enumerable:!0,get:function(){return u.default}}),Object.defineProperty(e,"twilight",{enumerable:!0,get:function(){return c.default}}),Object.defineProperty(e,"vs",{enumerable:!0,get:function(){return oe.default}}),Object.defineProperty(e,"vscDarkPlus",{enumerable:!0,get:function(){return re.default}}),Object.defineProperty(e,"xonokai",{enumerable:!0,get:function(){return ie.default}}),Object.defineProperty(e,"zTouch",{enumerable:!0,get:function(){return Se.default}});var n=t(AEe()),r=t(REe()),a=t(_Ee()),o=t(CEe()),s=t(NEe()),u=t(OEe()),c=t(DEe()),d=t(IEe()),p=t(LEe()),g=t(MEe()),m=t(FEe()),b=t(PEe()),y=t(zEe()),S=t(BEe()),x=t(UEe()),R=t(jEe()),k=t(GEe()),A=t(HEe()),C=t($Ee()),N=t(qEe()),_=t(VEe()),O=t(WEe()),F=t(YEe()),L=t(KEe()),I=t(XEe()),H=t(ZEe()),$=t(QEe()),U=t(JEe()),Y=t(ewe()),Z=t(twe()),j=t(nwe()),G=t(rwe()),z=t(awe()),K=t(owe()),D=t(iwe()),V=t(swe()),B=t(lwe()),M=t(uwe()),W=t(cwe()),Q=t(dwe()),oe=t(fwe()),re=t(pwe()),ie=t(gwe()),Se=t(hwe())}(cx)),cx}var xz=mwe();const bwe=({message:e})=>{const{t}=Pt(),n=E.useCallback(async()=>{if(e.content)try{await navigator.clipboard.writeText(e.content)}catch(r){console.error(t("chat.copyError"),r)}},[e]);return w.jsxs("div",{className:`max-w-[80%] rounded-lg px-4 py-2 ${e.role==="user"?"bg-primary text-primary-foreground":e.isError?"bg-red-100 text-red-600 dark:bg-red-950 dark:text-red-400":"bg-muted"}`,children:[w.jsxs("pre",{className:"relative break-words whitespace-pre-wrap",children:[w.jsx(Tde,{className:"dark:prose-invert max-w-none text-base text-sm",remarkPlugins:[Ppe,Kpe],rehypePlugins:[zpe],skipHtml:!1,components:{code:vwe},children:e.content}),e.role==="assistant"&&e.content.length>0&&w.jsx(xt,{onClick:n,className:"absolute right-0 bottom-0 size-6 rounded-md opacity-20 transition-opacity hover:opacity-100",tooltip:t("retrievePanel.chatMessage.copyTooltip"),variant:"default",size:"icon",children:w.jsx(_W,{})})]}),e.content.length===0&&w.jsx($W,{className:"animate-spin duration-2000"})]})},ywe=e=>!(e.children||[]).filter(n=>n.type==="text").map(n=>n.value).join("").includes(`
`),vwe=({className:e,children:t,node:n,...r})=>{const{theme:a}=E5(),o=e==null?void 0:e.match(/language-(\w+)/),s=o?o[1]:void 0;return(n?ywe(n):!1)?w.jsx("code",{className:Me(e,"mx-1 rounded-xs bg-black/10 px-1 dark:bg-gray-100/20"),...r,children:t}):w.jsx(KH,{style:a==="dark"?xz.oneDark:xz.oneLight,PreTag:"div",language:s,...r,children:String(t).replace(/\n$/,"")})};function Swe(){const{t:e}=Pt(),[t,n]=E.useState(()=>Fe.getState().retrievalHistory||[]),[r,a]=E.useState(""),[o,s]=E.useState(!1),u=E.useRef(null),c=E.useCallback(()=>{var m;(m=u.current)==null||m.scrollIntoView({behavior:"smooth"})},[]),d=E.useCallback(async m=>{if(m.preventDefault(),!r.trim()||o)return;const b={content:r,role:"user"},y={content:"",role:"assistant"},S=[...t];n([...S,b,y]),a(""),s(!0);const x=(A,C)=>{y.content+=A,n(N=>{const _=[...N],O=_[_.length-1];return O.role==="assistant"&&(O.content=y.content,O.isError=C),_})},R=Fe.getState(),k={...R.querySettings,query:b.content,conversation_history:S.filter(A=>A.isError!==!0).map(A=>({role:A.role,content:A.content}))};try{if(R.querySettings.stream){let A="";await IV(k,x,C=>{A+=C}),A&&(y.content&&(A=y.content+`
`+A),x(A,!0))}else{const A=await DV(k);x(A.response)}}catch(A){x(`${e("retrievePanel.retrieval.error")}
${Sr(A)}`,!0)}finally{s(!1),Fe.getState().setRetrievalHistory([...S,b,y])}},[r,o,t,n,e]),p=LT(t,100);E.useEffect(()=>c(),[p,c]);const g=E.useCallback(()=>{n([]),Fe.getState().setRetrievalHistory([])},[n]);return w.jsxs("div",{className:"flex size-full gap-2 px-2 pb-12 overflow-hidden",children:[w.jsxs("div",{className:"flex grow flex-col gap-4",children:[w.jsx("div",{className:"relative grow",children:w.jsx("div",{className:"bg-primary-foreground/60 absolute inset-0 flex flex-col overflow-auto rounded-lg border p-2",children:w.jsxs("div",{className:"flex min-h-0 flex-1 flex-col gap-2",children:[t.length===0?w.jsx("div",{className:"text-muted-foreground flex h-full items-center justify-center text-lg",children:e("retrievePanel.retrieval.startPrompt")}):t.map((m,b)=>w.jsx("div",{className:`flex ${m.role==="user"?"justify-end":"justify-start"}`,children:w.jsx(bwe,{message:m})},b)),w.jsx("div",{ref:u,className:"pb-1"})]})})}),w.jsxs("form",{onSubmit:d,className:"flex shrink-0 items-center gap-2",children:[w.jsxs(xt,{type:"button",variant:"outline",onClick:g,disabled:o,size:"sm",children:[w.jsx(eU,{}),e("retrievePanel.retrieval.clear")]}),w.jsx($a,{className:"flex-1",value:r,onChange:m=>a(m.target.value),placeholder:e("retrievePanel.retrieval.placeholder"),disabled:o}),w.jsxs(xt,{type:"submit",variant:"default",disabled:o,size:"sm",children:[w.jsx(pY,{}),e("retrievePanel.retrieval.send")]})]})]}),w.jsx(gse,{})]})}const Ewe=()=>{const e=E.useContext(WB);if(!e)throw new Error("useTabVisibility must be used within a TabVisibilityProvider");return e};function wwe(){const{t:e}=Pt(),{isTabVisible:t}=Ewe(),n=t("api"),[r,a]=E.useState(!1);return E.useEffect(()=>{r||a(!0)},[r]),w.jsx("div",{className:`size-full ${n?"":"hidden"}`,children:r?w.jsx("iframe",{src:zB+"/docs",className:"size-full w-full h-full",style:{width:"100%",height:"100%",border:"none"}},"api-docs-iframe"):w.jsx("div",{className:"flex h-full w-full items-center justify-center bg-background",children:w.jsxs("div",{className:"text-center",children:[w.jsx("div",{className:"mb-2 h-8 w-8 animate-spin rounded-full border-4 border-primary border-t-transparent"}),w.jsx("p",{children:e("apiSite.loading")})]})})})}function xwe(){const e=Hn.use.message(),t=Fe.use.enableHealthCheck(),n=Fe.use.currentTab(),[r,a]=E.useState(!1);E.useEffect(()=>{Hn.getState().check();const s=setInterval(async()=>{await Hn.getState().check()},bV*1e3);return()=>clearInterval(s)},[t]);const o=E.useCallback(s=>Fe.getState().setCurrentTab(s),[]);return E.useEffect(()=>{if(e&&(e.includes(GB)||e.includes(HB))){a(!0);return}a(!1)},[e,a]),w.jsx(VB,{children:w.jsx(hW,{children:w.jsxs("main",{className:"flex h-screen w-screen overflow-hidden",children:[w.jsxs(eJ,{defaultValue:n,className:"!m-0 flex grow flex-col !p-0 overflow-hidden",onValueChange:o,children:[w.jsx(nJ,{}),w.jsxs("div",{className:"relative grow",children:[w.jsx(pu,{value:"documents",className:"absolute top-0 right-0 bottom-0 left-0 overflow-auto",children:w.jsx(Hie,{})}),w.jsx(pu,{value:"knowledge-graph",className:"absolute top-0 right-0 bottom-0 left-0 overflow-hidden",children:w.jsx(poe,{})}),w.jsx(pu,{value:"retrieval",className:"absolute top-0 right-0 bottom-0 left-0 overflow-hidden",children:w.jsx(Swe,{})}),w.jsx(pu,{value:"api",className:"absolute top-0 right-0 bottom-0 left-0 overflow-hidden",children:w.jsx(wwe,{})})]})]}),t&&w.jsx(uQ,{}),e!==null&&!r&&w.jsx(xY,{}),r&&w.jsx(LZ,{})]})})})}const kwe=()=>{const e=Vf(),{login:t,isAuthenticated:n}=na(),{t:r}=Pt(),[a,o]=E.useState(!1),[s,u]=E.useState(""),[c,d]=E.useState(""),[p,g]=E.useState(!0);if(E.useEffect(()=>{console.log("LoginPage mounted")},[]),E.useEffect(()=>{let b=!0;return(async()=>{try{if(n){e("/");return}const S=await N0();if(!b)return;if(!S.auth_configured&&S.access_token){t(S.access_token,!0),S.message&&Ft.info(S.message),e("/");return}}catch(S){console.error("Failed to check auth configuration:",S)}finally{b&&g(!1)}})(),()=>{b=!1}},[n,t,e]),p)return null;const m=async b=>{if(b.preventDefault(),!s||!c){Ft.error(r("login.errorEmptyFields"));return}try{o(!0);const y=await FV(s,c),S=y.auth_mode==="disabled";t(y.access_token,S),S?Ft.info(y.message||r("login.authDisabled","Authentication is disabled. Using guest access.")):Ft.success(r("login.successMessage")),e("/")}catch(y){console.error("Login failed...",y),Ft.error(r("login.errorInvalidCredentials")),na.getState().logout(),localStorage.removeItem("LIGHTRAG-API-TOKEN")}finally{o(!1)}};return w.jsxs("div",{className:"flex h-screen w-screen items-center justify-center bg-gradient-to-br from-emerald-50 to-teal-100 dark:from-gray-900 dark:to-gray-800",children:[w.jsx("div",{className:"absolute top-4 right-4 flex items-center gap-2",children:w.jsx(Mj,{className:"bg-white/30 dark:bg-gray-800/30 backdrop-blur-sm rounded-md"})}),w.jsxs(Ms,{className:"w-full max-w-[480px] shadow-lg mx-4",children:[w.jsx(Nu,{className:"flex items-center justify-center space-y-2 pb-8 pt-6",children:w.jsxs("div",{className:"flex flex-col items-center space-y-4",children:[w.jsxs("div",{className:"flex items-center gap-3",children:[w.jsx("img",{src:"logo.png",alt:"LightRAG Logo",className:"h-12 w-12"}),w.jsx(aU,{className:"size-10 text-emerald-400","aria-hidden":"true"})]}),w.jsxs("div",{className:"text-center space-y-2",children:[w.jsx("h1",{className:"text-3xl font-bold tracking-tight",children:"LightRAG"}),w.jsx("p",{className:"text-muted-foreground text-sm",children:r("login.description")})]})]})}),w.jsx(Du,{className:"px-8 pb-8",children:w.jsxs("form",{onSubmit:m,className:"space-y-6",children:[w.jsxs("div",{className:"flex items-center gap-4",children:[w.jsx("label",{htmlFor:"username",className:"text-sm font-medium w-16 shrink-0",children:r("login.username")}),w.jsx($a,{id:"username",placeholder:r("login.usernamePlaceholder"),value:s,onChange:b=>u(b.target.value),required:!0,className:"h-11 flex-1"})]}),w.jsxs("div",{className:"flex items-center gap-4",children:[w.jsx("label",{htmlFor:"password",className:"text-sm font-medium w-16 shrink-0",children:r("login.password")}),w.jsx($a,{id:"password",type:"password",placeholder:r("login.passwordPlaceholder"),value:c,onChange:b=>d(b.target.value),required:!0,className:"h-11 flex-1"})]}),w.jsx(xt,{type:"submit",className:"w-full h-11 text-base font-medium mt-2",disabled:a,children:r(a?"login.loggingIn":"login.loginButton")})]})})]})]})},Twe=({children:e})=>{const{isAuthenticated:t}=na(),[n,r]=E.useState(!0),a=Vf();return E.useEffect(()=>{xu.setNavigate(a)},[a]),E.useEffect(()=>{let o=!0;return(async()=>{try{if(t){o&&r(!1);return}const u=await N0();if(!o)return;!u.auth_configured&&u.access_token&&(na.getState().login(u.access_token,!0),u.message&&Ft.info(u.message))}catch(u){console.error("Failed to check auth status:",u)}finally{o&&r(!1)}})(),()=>{o=!1}},[t]),E.useEffect(()=>{!n&&!t&&(window.location.hash.slice(1)==="/login"||(console.log("Not authenticated, redirecting to login"),xu.navigateToLogin()))},[n,t]),n||!t&&window.location.hash.slice(1)==="/login"||!t?null:w.jsx(w.Fragment,{children:e})},Awe=()=>{const[e,t]=E.useState(!0),{isAuthenticated:n}=na(),r=Vf();return E.useEffect(()=>{xu.setNavigate(r)},[r]),E.useEffect(()=>{let a=!0;return(async()=>{try{const s=localStorage.getItem("LIGHTRAG-API-TOKEN");if(s&&n){a&&t(!1);return}const u=await N0();if(!a)return;!u.auth_configured&&u.access_token?(na.getState().login(u.access_token,!0),u.message&&Ft.info(u.message)):s||na.getState().logout()}catch(s){console.error("Auth initialization error:",s),a&&!n&&na.getState().logout()}finally{a&&t(!1)}})(),()=>{a=!1}},[n]),e?null:w.jsxs(dq,{children:[w.jsx(ak,{path:"/login",element:w.jsx(kwe,{})}),w.jsx(ak,{path:"/*",element:w.jsx(Twe,{children:w.jsx(xwe,{})})})]})},Rwe=()=>w.jsx(VB,{children:w.jsxs(Lq,{children:[w.jsx(Awe,{}),w.jsx(fW,{position:"bottom-center"})]})}),tt=e=>typeof e=="string",du=()=>{let e,t;const n=new Promise((r,a)=>{e=r,t=a});return n.resolve=e,n.reject=t,n},kz=e=>e==null?"":""+e,_we=(e,t,n)=>{e.forEach(r=>{t[r]&&(n[r]=t[r])})},Cwe=/###/g,Tz=e=>e&&e.indexOf("###")>-1?e.replace(Cwe,"."):e,Az=e=>!e||tt(e),Su=(e,t,n)=>{const r=tt(t)?t.split("."):t;let a=0;for(;a<r.length-1;){if(Az(e))return{};const o=Tz(r[a]);!e[o]&&n&&(e[o]=new n),Object.prototype.hasOwnProperty.call(e,o)?e=e[o]:e={},++a}return Az(e)?{}:{obj:e,k:Tz(r[a])}},Rz=(e,t,n)=>{const{obj:r,k:a}=Su(e,t,Object);if(r!==void 0||t.length===1){r[a]=n;return}let o=t[t.length-1],s=t.slice(0,t.length-1),u=Su(e,s,Object);for(;u.obj===void 0&&s.length;)o=`${s[s.length-1]}.${o}`,s=s.slice(0,s.length-1),u=Su(e,s,Object),u!=null&&u.obj&&typeof u.obj[`${u.k}.${o}`]<"u"&&(u.obj=void 0);u.obj[`${u.k}.${o}`]=n},Nwe=(e,t,n,r)=>{const{obj:a,k:o}=Su(e,t,Object);a[o]=a[o]||[],a[o].push(n)},Uf=(e,t)=>{const{obj:n,k:r}=Su(e,t);if(n&&Object.prototype.hasOwnProperty.call(n,r))return n[r]},Owe=(e,t,n)=>{const r=Uf(e,n);return r!==void 0?r:Uf(t,n)},XH=(e,t,n)=>{for(const r in t)r!=="__proto__"&&r!=="constructor"&&(r in e?tt(e[r])||e[r]instanceof String||tt(t[r])||t[r]instanceof String?n&&(e[r]=t[r]):XH(e[r],t[r],n):e[r]=t[r]);return e},ms=e=>e.replace(/[\-\[\]\/\{\}\(\)\*\+\?\.\\\^\$\|]/g,"\\$&");var Dwe={"&":"&amp;","<":"&lt;",">":"&gt;",'"':"&quot;","'":"&#39;","/":"&#x2F;"};const Iwe=e=>tt(e)?e.replace(/[&<>"'\/]/g,t=>Dwe[t]):e;class Lwe{constructor(t){this.capacity=t,this.regExpMap=new Map,this.regExpQueue=[]}getRegExp(t){const n=this.regExpMap.get(t);if(n!==void 0)return n;const r=new RegExp(t);return this.regExpQueue.length===this.capacity&&this.regExpMap.delete(this.regExpQueue.shift()),this.regExpMap.set(t,r),this.regExpQueue.push(t),r}}const Mwe=[" ",",","?","!",";"],Fwe=new Lwe(20),Pwe=(e,t,n)=>{t=t||"",n=n||"";const r=Mwe.filter(s=>t.indexOf(s)<0&&n.indexOf(s)<0);if(r.length===0)return!0;const a=Fwe.getRegExp(`(${r.map(s=>s==="?"?"\\?":s).join("|")})`);let o=!a.test(e);if(!o){const s=e.indexOf(n);s>0&&!a.test(e.substring(0,s))&&(o=!0)}return o},y0=function(e,t){let n=arguments.length>2&&arguments[2]!==void 0?arguments[2]:".";if(!e)return;if(e[t])return Object.prototype.hasOwnProperty.call(e,t)?e[t]:void 0;const r=t.split(n);let a=e;for(let o=0;o<r.length;){if(!a||typeof a!="object")return;let s,u="";for(let c=o;c<r.length;++c)if(c!==o&&(u+=n),u+=r[c],s=a[u],s!==void 0){if(["string","number","boolean"].indexOf(typeof s)>-1&&c<r.length-1)continue;o+=c-o+1;break}a=s}return a},jf=e=>e==null?void 0:e.replace("_","-"),zwe={type:"logger",log(e){this.output("log",e)},warn(e){this.output("warn",e)},error(e){this.output("error",e)},output(e,t){var n,r;(r=(n=console==null?void 0:console[e])==null?void 0:n.apply)==null||r.call(n,console,t)}};class Gf{constructor(t){let n=arguments.length>1&&arguments[1]!==void 0?arguments[1]:{};this.init(t,n)}init(t){let n=arguments.length>1&&arguments[1]!==void 0?arguments[1]:{};this.prefix=n.prefix||"i18next:",this.logger=t||zwe,this.options=n,this.debug=n.debug}log(){for(var t=arguments.length,n=new Array(t),r=0;r<t;r++)n[r]=arguments[r];return this.forward(n,"log","",!0)}warn(){for(var t=arguments.length,n=new Array(t),r=0;r<t;r++)n[r]=arguments[r];return this.forward(n,"warn","",!0)}error(){for(var t=arguments.length,n=new Array(t),r=0;r<t;r++)n[r]=arguments[r];return this.forward(n,"error","")}deprecate(){for(var t=arguments.length,n=new Array(t),r=0;r<t;r++)n[r]=arguments[r];return this.forward(n,"warn","WARNING DEPRECATED: ",!0)}forward(t,n,r,a){return a&&!this.debug?null:(tt(t[0])&&(t[0]=`${r}${this.prefix} ${t[0]}`),this.logger[n](t))}create(t){return new Gf(this.logger,{prefix:`${this.prefix}:${t}:`,...this.options})}clone(t){return t=t||this.options,t.prefix=t.prefix||this.prefix,new Gf(this.logger,t)}}var ra=new Gf;class Bp{constructor(){this.observers={}}on(t,n){return t.split(" ").forEach(r=>{this.observers[r]||(this.observers[r]=new Map);const a=this.observers[r].get(n)||0;this.observers[r].set(n,a+1)}),this}off(t,n){if(this.observers[t]){if(!n){delete this.observers[t];return}this.observers[t].delete(n)}}emit(t){for(var n=arguments.length,r=new Array(n>1?n-1:0),a=1;a<n;a++)r[a-1]=arguments[a];this.observers[t]&&Array.from(this.observers[t].entries()).forEach(s=>{let[u,c]=s;for(let d=0;d<c;d++)u(...r)}),this.observers["*"]&&Array.from(this.observers["*"].entries()).forEach(s=>{let[u,c]=s;for(let d=0;d<c;d++)u.apply(u,[t,...r])})}}class _z extends Bp{constructor(t){let n=arguments.length>1&&arguments[1]!==void 0?arguments[1]:{ns:["translation"],defaultNS:"translation"};super(),this.data=t||{},this.options=n,this.options.keySeparator===void 0&&(this.options.keySeparator="."),this.options.ignoreJSONStructure===void 0&&(this.options.ignoreJSONStructure=!0)}addNamespaces(t){this.options.ns.indexOf(t)<0&&this.options.ns.push(t)}removeNamespaces(t){const n=this.options.ns.indexOf(t);n>-1&&this.options.ns.splice(n,1)}getResource(t,n,r){var d,p;let a=arguments.length>3&&arguments[3]!==void 0?arguments[3]:{};const o=a.keySeparator!==void 0?a.keySeparator:this.options.keySeparator,s=a.ignoreJSONStructure!==void 0?a.ignoreJSONStructure:this.options.ignoreJSONStructure;let u;t.indexOf(".")>-1?u=t.split("."):(u=[t,n],r&&(Array.isArray(r)?u.push(...r):tt(r)&&o?u.push(...r.split(o)):u.push(r)));const c=Uf(this.data,u);return!c&&!n&&!r&&t.indexOf(".")>-1&&(t=u[0],n=u[1],r=u.slice(2).join(".")),c||!s||!tt(r)?c:y0((p=(d=this.data)==null?void 0:d[t])==null?void 0:p[n],r,o)}addResource(t,n,r,a){let o=arguments.length>4&&arguments[4]!==void 0?arguments[4]:{silent:!1};const s=o.keySeparator!==void 0?o.keySeparator:this.options.keySeparator;let u=[t,n];r&&(u=u.concat(s?r.split(s):r)),t.indexOf(".")>-1&&(u=t.split("."),a=n,n=u[1]),this.addNamespaces(n),Rz(this.data,u,a),o.silent||this.emit("added",t,n,r,a)}addResources(t,n,r){let a=arguments.length>3&&arguments[3]!==void 0?arguments[3]:{silent:!1};for(const o in r)(tt(r[o])||Array.isArray(r[o]))&&this.addResource(t,n,o,r[o],{silent:!0});a.silent||this.emit("added",t,n,r)}addResourceBundle(t,n,r,a,o){let s=arguments.length>5&&arguments[5]!==void 0?arguments[5]:{silent:!1,skipCopy:!1},u=[t,n];t.indexOf(".")>-1&&(u=t.split("."),a=r,r=n,n=u[1]),this.addNamespaces(n);let c=Uf(this.data,u)||{};s.skipCopy||(r=JSON.parse(JSON.stringify(r))),a?XH(c,r,o):c={...c,...r},Rz(this.data,u,c),s.silent||this.emit("added",t,n,r)}removeResourceBundle(t,n){this.hasResourceBundle(t,n)&&delete this.data[t][n],this.removeNamespaces(n),this.emit("removed",t,n)}hasResourceBundle(t,n){return this.getResource(t,n)!==void 0}getResourceBundle(t,n){return n||(n=this.options.defaultNS),this.getResource(t,n)}getDataByLanguage(t){return this.data[t]}hasLanguageSomeTranslations(t){const n=this.getDataByLanguage(t);return!!(n&&Object.keys(n)||[]).find(a=>n[a]&&Object.keys(n[a]).length>0)}toJSON(){return this.data}}var ZH={processors:{},addPostProcessor(e){this.processors[e.name]=e},handle(e,t,n,r,a){return e.forEach(o=>{var s;t=((s=this.processors[o])==null?void 0:s.process(t,n,r,a))??t}),t}};const Cz={},Nz=e=>!tt(e)&&typeof e!="boolean"&&typeof e!="number";class Hf extends Bp{constructor(t){let n=arguments.length>1&&arguments[1]!==void 0?arguments[1]:{};super(),_we(["resourceStore","languageUtils","pluralResolver","interpolator","backendConnector","i18nFormat","utils"],t,this),this.options=n,this.options.keySeparator===void 0&&(this.options.keySeparator="."),this.logger=ra.create("translator")}changeLanguage(t){t&&(this.language=t)}exists(t){let n=arguments.length>1&&arguments[1]!==void 0?arguments[1]:{interpolation:{}};if(t==null)return!1;const r=this.resolve(t,n);return(r==null?void 0:r.res)!==void 0}extractFromKey(t,n){let r=n.nsSeparator!==void 0?n.nsSeparator:this.options.nsSeparator;r===void 0&&(r=":");const a=n.keySeparator!==void 0?n.keySeparator:this.options.keySeparator;let o=n.ns||this.options.defaultNS||[];const s=r&&t.indexOf(r)>-1,u=!this.options.userDefinedKeySeparator&&!n.keySeparator&&!this.options.userDefinedNsSeparator&&!n.nsSeparator&&!Pwe(t,r,a);if(s&&!u){const c=t.match(this.interpolator.nestingRegexp);if(c&&c.length>0)return{key:t,namespaces:tt(o)?[o]:o};const d=t.split(r);(r!==a||r===a&&this.options.ns.indexOf(d[0])>-1)&&(o=d.shift()),t=d.join(a)}return{key:t,namespaces:tt(o)?[o]:o}}translate(t,n,r){if(typeof n!="object"&&this.options.overloadTranslationOptionHandler&&(n=this.options.overloadTranslationOptionHandler(arguments)),typeof n=="object"&&(n={...n}),n||(n={}),t==null)return"";Array.isArray(t)||(t=[String(t)]);const a=n.returnDetails!==void 0?n.returnDetails:this.options.returnDetails,o=n.keySeparator!==void 0?n.keySeparator:this.options.keySeparator,{key:s,namespaces:u}=this.extractFromKey(t[t.length-1],n),c=u[u.length-1],d=n.lng||this.language,p=n.appendNamespaceToCIMode||this.options.appendNamespaceToCIMode;if((d==null?void 0:d.toLowerCase())==="cimode"){if(p){const H=n.nsSeparator||this.options.nsSeparator;return a?{res:`${c}${H}${s}`,usedKey:s,exactUsedKey:s,usedLng:d,usedNS:c,usedParams:this.getUsedParamsDetails(n)}:`${c}${H}${s}`}return a?{res:s,usedKey:s,exactUsedKey:s,usedLng:d,usedNS:c,usedParams:this.getUsedParamsDetails(n)}:s}const g=this.resolve(t,n);let m=g==null?void 0:g.res;const b=(g==null?void 0:g.usedKey)||s,y=(g==null?void 0:g.exactUsedKey)||s,S=["[object Number]","[object Function]","[object RegExp]"],x=n.joinArrays!==void 0?n.joinArrays:this.options.joinArrays,R=!this.i18nFormat||this.i18nFormat.handleAsObject,k=n.count!==void 0&&!tt(n.count),A=Hf.hasDefaultValue(n),C=k?this.pluralResolver.getSuffix(d,n.count,n):"",N=n.ordinal&&k?this.pluralResolver.getSuffix(d,n.count,{ordinal:!1}):"",_=k&&!n.ordinal&&n.count===0,O=_&&n[`defaultValue${this.options.pluralSeparator}zero`]||n[`defaultValue${C}`]||n[`defaultValue${N}`]||n.defaultValue;let F=m;R&&!m&&A&&(F=O);const L=Nz(F),I=Object.prototype.toString.apply(F);if(R&&F&&L&&S.indexOf(I)<0&&!(tt(x)&&Array.isArray(F))){if(!n.returnObjects&&!this.options.returnObjects){this.options.returnedObjectHandler||this.logger.warn("accessing an object - but returnObjects options is not enabled!");const H=this.options.returnedObjectHandler?this.options.returnedObjectHandler(b,F,{...n,ns:u}):`key '${s} (${this.language})' returned an object instead of string.`;return a?(g.res=H,g.usedParams=this.getUsedParamsDetails(n),g):H}if(o){const H=Array.isArray(F),$=H?[]:{},U=H?y:b;for(const Y in F)if(Object.prototype.hasOwnProperty.call(F,Y)){const Z=`${U}${o}${Y}`;A&&!m?$[Y]=this.translate(Z,{...n,defaultValue:Nz(O)?O[Y]:void 0,joinArrays:!1,ns:u}):$[Y]=this.translate(Z,{...n,joinArrays:!1,ns:u}),$[Y]===Z&&($[Y]=F[Y])}m=$}}else if(R&&tt(x)&&Array.isArray(m))m=m.join(x),m&&(m=this.extendTranslation(m,t,n,r));else{let H=!1,$=!1;!this.isValidLookup(m)&&A&&(H=!0,m=O),this.isValidLookup(m)||($=!0,m=s);const Y=(n.missingKeyNoValueFallbackToKey||this.options.missingKeyNoValueFallbackToKey)&&$?void 0:m,Z=A&&O!==m&&this.options.updateMissing;if($||H||Z){if(this.logger.log(Z?"updateKey":"missingKey",d,c,s,Z?O:m),o){const K=this.resolve(s,{...n,keySeparator:!1});K&&K.res&&this.logger.warn("Seems the loaded translations were in flat JSON format instead of nested. Either set keySeparator: false on init or make sure your translations are published in nested format.")}let j=[];const G=this.languageUtils.getFallbackCodes(this.options.fallbackLng,n.lng||this.language);if(this.options.saveMissingTo==="fallback"&&G&&G[0])for(let K=0;K<G.length;K++)j.push(G[K]);else this.options.saveMissingTo==="all"?j=this.languageUtils.toResolveHierarchy(n.lng||this.language):j.push(n.lng||this.language);const z=(K,D,V)=>{var M;const B=A&&V!==m?V:Y;this.options.missingKeyHandler?this.options.missingKeyHandler(K,c,D,B,Z,n):(M=this.backendConnector)!=null&&M.saveMissing&&this.backendConnector.saveMissing(K,c,D,B,Z,n),this.emit("missingKey",K,c,D,m)};this.options.saveMissing&&(this.options.saveMissingPlurals&&k?j.forEach(K=>{const D=this.pluralResolver.getSuffixes(K,n);_&&n[`defaultValue${this.options.pluralSeparator}zero`]&&D.indexOf(`${this.options.pluralSeparator}zero`)<0&&D.push(`${this.options.pluralSeparator}zero`),D.forEach(V=>{z([K],s+V,n[`defaultValue${V}`]||O)})}):z(j,s,O))}m=this.extendTranslation(m,t,n,g,r),$&&m===s&&this.options.appendNamespaceToMissingKey&&(m=`${c}:${s}`),($||H)&&this.options.parseMissingKeyHandler&&(m=this.options.parseMissingKeyHandler(this.options.appendNamespaceToMissingKey?`${c}:${s}`:s,H?m:void 0))}return a?(g.res=m,g.usedParams=this.getUsedParamsDetails(n),g):m}extendTranslation(t,n,r,a,o){var d,p;var s=this;if((d=this.i18nFormat)!=null&&d.parse)t=this.i18nFormat.parse(t,{...this.options.interpolation.defaultVariables,...r},r.lng||this.language||a.usedLng,a.usedNS,a.usedKey,{resolved:a});else if(!r.skipInterpolation){r.interpolation&&this.interpolator.init({...r,interpolation:{...this.options.interpolation,...r.interpolation}});const g=tt(t)&&(((p=r==null?void 0:r.interpolation)==null?void 0:p.skipOnVariables)!==void 0?r.interpolation.skipOnVariables:this.options.interpolation.skipOnVariables);let m;if(g){const y=t.match(this.interpolator.nestingRegexp);m=y&&y.length}let b=r.replace&&!tt(r.replace)?r.replace:r;if(this.options.interpolation.defaultVariables&&(b={...this.options.interpolation.defaultVariables,...b}),t=this.interpolator.interpolate(t,b,r.lng||this.language||a.usedLng,r),g){const y=t.match(this.interpolator.nestingRegexp),S=y&&y.length;m<S&&(r.nest=!1)}!r.lng&&a&&a.res&&(r.lng=this.language||a.usedLng),r.nest!==!1&&(t=this.interpolator.nest(t,function(){for(var y=arguments.length,S=new Array(y),x=0;x<y;x++)S[x]=arguments[x];return(o==null?void 0:o[0])===S[0]&&!r.context?(s.logger.warn(`It seems you are nesting recursively key: ${S[0]} in key: ${n[0]}`),null):s.translate(...S,n)},r)),r.interpolation&&this.interpolator.reset()}const u=r.postProcess||this.options.postProcess,c=tt(u)?[u]:u;return t!=null&&(c!=null&&c.length)&&r.applyPostProcessor!==!1&&(t=ZH.handle(c,t,n,this.options&&this.options.postProcessPassResolved?{i18nResolved:{...a,usedParams:this.getUsedParamsDetails(r)},...r}:r,this)),t}resolve(t){let n=arguments.length>1&&arguments[1]!==void 0?arguments[1]:{},r,a,o,s,u;return tt(t)&&(t=[t]),t.forEach(c=>{if(this.isValidLookup(r))return;const d=this.extractFromKey(c,n),p=d.key;a=p;let g=d.namespaces;this.options.fallbackNS&&(g=g.concat(this.options.fallbackNS));const m=n.count!==void 0&&!tt(n.count),b=m&&!n.ordinal&&n.count===0,y=n.context!==void 0&&(tt(n.context)||typeof n.context=="number")&&n.context!=="",S=n.lngs?n.lngs:this.languageUtils.toResolveHierarchy(n.lng||this.language,n.fallbackLng);g.forEach(x=>{var R,k;this.isValidLookup(r)||(u=x,!Cz[`${S[0]}-${x}`]&&((R=this.utils)!=null&&R.hasLoadedNamespace)&&!((k=this.utils)!=null&&k.hasLoadedNamespace(u))&&(Cz[`${S[0]}-${x}`]=!0,this.logger.warn(`key "${a}" for languages "${S.join(", ")}" won't get resolved as namespace "${u}" was not yet loaded`,"This means something IS WRONG in your setup. You access the t function before i18next.init / i18next.loadNamespace / i18next.changeLanguage was done. Wait for the callback or Promise to resolve before accessing it!!!")),S.forEach(A=>{var _;if(this.isValidLookup(r))return;s=A;const C=[p];if((_=this.i18nFormat)!=null&&_.addLookupKeys)this.i18nFormat.addLookupKeys(C,p,A,x,n);else{let O;m&&(O=this.pluralResolver.getSuffix(A,n.count,n));const F=`${this.options.pluralSeparator}zero`,L=`${this.options.pluralSeparator}ordinal${this.options.pluralSeparator}`;if(m&&(C.push(p+O),n.ordinal&&O.indexOf(L)===0&&C.push(p+O.replace(L,this.options.pluralSeparator)),b&&C.push(p+F)),y){const I=`${p}${this.options.contextSeparator}${n.context}`;C.push(I),m&&(C.push(I+O),n.ordinal&&O.indexOf(L)===0&&C.push(I+O.replace(L,this.options.pluralSeparator)),b&&C.push(I+F))}}let N;for(;N=C.pop();)this.isValidLookup(r)||(o=N,r=this.getResource(A,x,N,n))}))})}),{res:r,usedKey:a,exactUsedKey:o,usedLng:s,usedNS:u}}isValidLookup(t){return t!==void 0&&!(!this.options.returnNull&&t===null)&&!(!this.options.returnEmptyString&&t==="")}getResource(t,n,r){var o;let a=arguments.length>3&&arguments[3]!==void 0?arguments[3]:{};return(o=this.i18nFormat)!=null&&o.getResource?this.i18nFormat.getResource(t,n,r,a):this.resourceStore.getResource(t,n,r,a)}getUsedParamsDetails(){let t=arguments.length>0&&arguments[0]!==void 0?arguments[0]:{};const n=["defaultValue","ordinal","context","replace","lng","lngs","fallbackLng","ns","keySeparator","nsSeparator","returnObjects","returnDetails","joinArrays","postProcess","interpolation"],r=t.replace&&!tt(t.replace);let a=r?t.replace:t;if(r&&typeof t.count<"u"&&(a.count=t.count),this.options.interpolation.defaultVariables&&(a={...this.options.interpolation.defaultVariables,...a}),!r){a={...a};for(const o of n)delete a[o]}return a}static hasDefaultValue(t){const n="defaultValue";for(const r in t)if(Object.prototype.hasOwnProperty.call(t,r)&&n===r.substring(0,n.length)&&t[r]!==void 0)return!0;return!1}}class Oz{constructor(t){this.options=t,this.supportedLngs=this.options.supportedLngs||!1,this.logger=ra.create("languageUtils")}getScriptPartFromCode(t){if(t=jf(t),!t||t.indexOf("-")<0)return null;const n=t.split("-");return n.length===2||(n.pop(),n[n.length-1].toLowerCase()==="x")?null:this.formatLanguageCode(n.join("-"))}getLanguagePartFromCode(t){if(t=jf(t),!t||t.indexOf("-")<0)return t;const n=t.split("-");return this.formatLanguageCode(n[0])}formatLanguageCode(t){if(tt(t)&&t.indexOf("-")>-1){let n;try{n=Intl.getCanonicalLocales(t)[0]}catch{}return n&&this.options.lowerCaseLng&&(n=n.toLowerCase()),n||(this.options.lowerCaseLng?t.toLowerCase():t)}return this.options.cleanCode||this.options.lowerCaseLng?t.toLowerCase():t}isSupportedCode(t){return(this.options.load==="languageOnly"||this.options.nonExplicitSupportedLngs)&&(t=this.getLanguagePartFromCode(t)),!this.supportedLngs||!this.supportedLngs.length||this.supportedLngs.indexOf(t)>-1}getBestMatchFromCodes(t){if(!t)return null;let n;return t.forEach(r=>{if(n)return;const a=this.formatLanguageCode(r);(!this.options.supportedLngs||this.isSupportedCode(a))&&(n=a)}),!n&&this.options.supportedLngs&&t.forEach(r=>{if(n)return;const a=this.getLanguagePartFromCode(r);if(this.isSupportedCode(a))return n=a;n=this.options.supportedLngs.find(o=>{if(o===a)return o;if(!(o.indexOf("-")<0&&a.indexOf("-")<0)&&(o.indexOf("-")>0&&a.indexOf("-")<0&&o.substring(0,o.indexOf("-"))===a||o.indexOf(a)===0&&a.length>1))return o})}),n||(n=this.getFallbackCodes(this.options.fallbackLng)[0]),n}getFallbackCodes(t,n){if(!t)return[];if(typeof t=="function"&&(t=t(n)),tt(t)&&(t=[t]),Array.isArray(t))return t;if(!n)return t.default||[];let r=t[n];return r||(r=t[this.getScriptPartFromCode(n)]),r||(r=t[this.formatLanguageCode(n)]),r||(r=t[this.getLanguagePartFromCode(n)]),r||(r=t.default),r||[]}toResolveHierarchy(t,n){const r=this.getFallbackCodes(n||this.options.fallbackLng||[],t),a=[],o=s=>{s&&(this.isSupportedCode(s)?a.push(s):this.logger.warn(`rejecting language code not found in supportedLngs: ${s}`))};return tt(t)&&(t.indexOf("-")>-1||t.indexOf("_")>-1)?(this.options.load!=="languageOnly"&&o(this.formatLanguageCode(t)),this.options.load!=="languageOnly"&&this.options.load!=="currentOnly"&&o(this.getScriptPartFromCode(t)),this.options.load!=="currentOnly"&&o(this.getLanguagePartFromCode(t))):tt(t)&&o(this.formatLanguageCode(t)),r.forEach(s=>{a.indexOf(s)<0&&o(this.formatLanguageCode(s))}),a}}const Dz={zero:0,one:1,two:2,few:3,many:4,other:5},Iz={select:e=>e===1?"one":"other",resolvedOptions:()=>({pluralCategories:["one","other"]})};class Bwe{constructor(t){let n=arguments.length>1&&arguments[1]!==void 0?arguments[1]:{};this.languageUtils=t,this.options=n,this.logger=ra.create("pluralResolver"),this.pluralRulesCache={}}addRule(t,n){this.rules[t]=n}clearCache(){this.pluralRulesCache={}}getRule(t){let n=arguments.length>1&&arguments[1]!==void 0?arguments[1]:{};const r=jf(t==="dev"?"en":t),a=n.ordinal?"ordinal":"cardinal",o=JSON.stringify({cleanedCode:r,type:a});if(o in this.pluralRulesCache)return this.pluralRulesCache[o];let s;try{s=new Intl.PluralRules(r,{type:a})}catch{if(!Intl)return this.logger.error("No Intl support, please use an Intl polyfill!"),Iz;if(!t.match(/-|_/))return Iz;const c=this.languageUtils.getLanguagePartFromCode(t);s=this.getRule(c,n)}return this.pluralRulesCache[o]=s,s}needsPlural(t){let n=arguments.length>1&&arguments[1]!==void 0?arguments[1]:{},r=this.getRule(t,n);return r||(r=this.getRule("dev",n)),(r==null?void 0:r.resolvedOptions().pluralCategories.length)>1}getPluralFormsOfKey(t,n){let r=arguments.length>2&&arguments[2]!==void 0?arguments[2]:{};return this.getSuffixes(t,r).map(a=>`${n}${a}`)}getSuffixes(t){let n=arguments.length>1&&arguments[1]!==void 0?arguments[1]:{},r=this.getRule(t,n);return r||(r=this.getRule("dev",n)),r?r.resolvedOptions().pluralCategories.sort((a,o)=>Dz[a]-Dz[o]).map(a=>`${this.options.prepend}${n.ordinal?`ordinal${this.options.prepend}`:""}${a}`):[]}getSuffix(t,n){let r=arguments.length>2&&arguments[2]!==void 0?arguments[2]:{};const a=this.getRule(t,r);return a?`${this.options.prepend}${r.ordinal?`ordinal${this.options.prepend}`:""}${a.select(n)}`:(this.logger.warn(`no plural rule found for: ${t}`),this.getSuffix("dev",n,r))}}const Lz=function(e,t,n){let r=arguments.length>3&&arguments[3]!==void 0?arguments[3]:".",a=arguments.length>4&&arguments[4]!==void 0?arguments[4]:!0,o=Owe(e,t,n);return!o&&a&&tt(n)&&(o=y0(e,n,r),o===void 0&&(o=y0(t,n,r))),o},nk=e=>e.replace(/\$/g,"$$$$");class Uwe{constructor(){var n;let t=arguments.length>0&&arguments[0]!==void 0?arguments[0]:{};this.logger=ra.create("interpolator"),this.options=t,this.format=((n=t==null?void 0:t.interpolation)==null?void 0:n.format)||(r=>r),this.init(t)}init(){let t=arguments.length>0&&arguments[0]!==void 0?arguments[0]:{};t.interpolation||(t.interpolation={escapeValue:!0});const{escape:n,escapeValue:r,useRawValueToEscape:a,prefix:o,prefixEscaped:s,suffix:u,suffixEscaped:c,formatSeparator:d,unescapeSuffix:p,unescapePrefix:g,nestingPrefix:m,nestingPrefixEscaped:b,nestingSuffix:y,nestingSuffixEscaped:S,nestingOptionsSeparator:x,maxReplaces:R,alwaysFormat:k}=t.interpolation;this.escape=n!==void 0?n:Iwe,this.escapeValue=r!==void 0?r:!0,this.useRawValueToEscape=a!==void 0?a:!1,this.prefix=o?ms(o):s||"{{",this.suffix=u?ms(u):c||"}}",this.formatSeparator=d||",",this.unescapePrefix=p?"":g||"-",this.unescapeSuffix=this.unescapePrefix?"":p||"",this.nestingPrefix=m?ms(m):b||ms("$t("),this.nestingSuffix=y?ms(y):S||ms(")"),this.nestingOptionsSeparator=x||",",this.maxReplaces=R||1e3,this.alwaysFormat=k!==void 0?k:!1,this.resetRegExp()}reset(){this.options&&this.init(this.options)}resetRegExp(){const t=(n,r)=>(n==null?void 0:n.source)===r?(n.lastIndex=0,n):new RegExp(r,"g");this.regexp=t(this.regexp,`${this.prefix}(.+?)${this.suffix}`),this.regexpUnescape=t(this.regexpUnescape,`${this.prefix}${this.unescapePrefix}(.+?)${this.unescapeSuffix}${this.suffix}`),this.nestingRegexp=t(this.nestingRegexp,`${this.nestingPrefix}(.+?)${this.nestingSuffix}`)}interpolate(t,n,r,a){var b;let o,s,u;const c=this.options&&this.options.interpolation&&this.options.interpolation.defaultVariables||{},d=y=>{if(y.indexOf(this.formatSeparator)<0){const k=Lz(n,c,y,this.options.keySeparator,this.options.ignoreJSONStructure);return this.alwaysFormat?this.format(k,void 0,r,{...a,...n,interpolationkey:y}):k}const S=y.split(this.formatSeparator),x=S.shift().trim(),R=S.join(this.formatSeparator).trim();return this.format(Lz(n,c,x,this.options.keySeparator,this.options.ignoreJSONStructure),R,r,{...a,...n,interpolationkey:x})};this.resetRegExp();const p=(a==null?void 0:a.missingInterpolationHandler)||this.options.missingInterpolationHandler,g=((b=a==null?void 0:a.interpolation)==null?void 0:b.skipOnVariables)!==void 0?a.interpolation.skipOnVariables:this.options.interpolation.skipOnVariables;return[{regex:this.regexpUnescape,safeValue:y=>nk(y)},{regex:this.regexp,safeValue:y=>this.escapeValue?nk(this.escape(y)):nk(y)}].forEach(y=>{for(u=0;o=y.regex.exec(t);){const S=o[1].trim();if(s=d(S),s===void 0)if(typeof p=="function"){const R=p(t,o,a);s=tt(R)?R:""}else if(a&&Object.prototype.hasOwnProperty.call(a,S))s="";else if(g){s=o[0];continue}else this.logger.warn(`missed to pass in variable ${S} for interpolating ${t}`),s="";else!tt(s)&&!this.useRawValueToEscape&&(s=kz(s));const x=y.safeValue(s);if(t=t.replace(o[0],x),g?(y.regex.lastIndex+=s.length,y.regex.lastIndex-=o[0].length):y.regex.lastIndex=0,u++,u>=this.maxReplaces)break}}),t}nest(t,n){let r=arguments.length>2&&arguments[2]!==void 0?arguments[2]:{},a,o,s;const u=(c,d)=>{const p=this.nestingOptionsSeparator;if(c.indexOf(p)<0)return c;const g=c.split(new RegExp(`${p}[ ]*{`));let m=`{${g[1]}`;c=g[0],m=this.interpolate(m,s);const b=m.match(/'/g),y=m.match(/"/g);(((b==null?void 0:b.length)??0)%2===0&&!y||y.length%2!==0)&&(m=m.replace(/'/g,'"'));try{s=JSON.parse(m),d&&(s={...d,...s})}catch(S){return this.logger.warn(`failed parsing options string in nesting for key ${c}`,S),`${c}${p}${m}`}return s.defaultValue&&s.defaultValue.indexOf(this.prefix)>-1&&delete s.defaultValue,c};for(;a=this.nestingRegexp.exec(t);){let c=[];s={...r},s=s.replace&&!tt(s.replace)?s.replace:s,s.applyPostProcessor=!1,delete s.defaultValue;let d=!1;if(a[0].indexOf(this.formatSeparator)!==-1&&!/{.*}/.test(a[1])){const p=a[1].split(this.formatSeparator).map(g=>g.trim());a[1]=p.shift(),c=p,d=!0}if(o=n(u.call(this,a[1].trim(),s),s),o&&a[0]===t&&!tt(o))return o;tt(o)||(o=kz(o)),o||(this.logger.warn(`missed to resolve ${a[1]} for nesting ${t}`),o=""),d&&(o=c.reduce((p,g)=>this.format(p,g,r.lng,{...r,interpolationkey:a[1].trim()}),o.trim())),t=t.replace(a[0],o),this.regexp.lastIndex=0}return t}}const jwe=e=>{let t=e.toLowerCase().trim();const n={};if(e.indexOf("(")>-1){const r=e.split("(");t=r[0].toLowerCase().trim();const a=r[1].substring(0,r[1].length-1);t==="currency"&&a.indexOf(":")<0?n.currency||(n.currency=a.trim()):t==="relativetime"&&a.indexOf(":")<0?n.range||(n.range=a.trim()):a.split(";").forEach(s=>{if(s){const[u,...c]=s.split(":"),d=c.join(":").trim().replace(/^'+|'+$/g,""),p=u.trim();n[p]||(n[p]=d),d==="false"&&(n[p]=!1),d==="true"&&(n[p]=!0),isNaN(d)||(n[p]=parseInt(d,10))}})}return{formatName:t,formatOptions:n}},bs=e=>{const t={};return(n,r,a)=>{let o=a;a&&a.interpolationkey&&a.formatParams&&a.formatParams[a.interpolationkey]&&a[a.interpolationkey]&&(o={...o,[a.interpolationkey]:void 0});const s=r+JSON.stringify(o);let u=t[s];return u||(u=e(jf(r),a),t[s]=u),u(n)}};class Gwe{constructor(){let t=arguments.length>0&&arguments[0]!==void 0?arguments[0]:{};this.logger=ra.create("formatter"),this.options=t,this.formats={number:bs((n,r)=>{const a=new Intl.NumberFormat(n,{...r});return o=>a.format(o)}),currency:bs((n,r)=>{const a=new Intl.NumberFormat(n,{...r,style:"currency"});return o=>a.format(o)}),datetime:bs((n,r)=>{const a=new Intl.DateTimeFormat(n,{...r});return o=>a.format(o)}),relativetime:bs((n,r)=>{const a=new Intl.RelativeTimeFormat(n,{...r});return o=>a.format(o,r.range||"day")}),list:bs((n,r)=>{const a=new Intl.ListFormat(n,{...r});return o=>a.format(o)})},this.init(t)}init(t){let n=arguments.length>1&&arguments[1]!==void 0?arguments[1]:{interpolation:{}};this.formatSeparator=n.interpolation.formatSeparator||","}add(t,n){this.formats[t.toLowerCase().trim()]=n}addCached(t,n){this.formats[t.toLowerCase().trim()]=bs(n)}format(t,n,r){let a=arguments.length>3&&arguments[3]!==void 0?arguments[3]:{};const o=n.split(this.formatSeparator);if(o.length>1&&o[0].indexOf("(")>1&&o[0].indexOf(")")<0&&o.find(u=>u.indexOf(")")>-1)){const u=o.findIndex(c=>c.indexOf(")")>-1);o[0]=[o[0],...o.splice(1,u)].join(this.formatSeparator)}return o.reduce((u,c)=>{var g;const{formatName:d,formatOptions:p}=jwe(c);if(this.formats[d]){let m=u;try{const b=((g=a==null?void 0:a.formatParams)==null?void 0:g[a.interpolationkey])||{},y=b.locale||b.lng||a.locale||a.lng||r;m=this.formats[d](u,y,{...p,...a,...b})}catch(b){this.logger.warn(b)}return m}else this.logger.warn(`there was no format function for ${d}`);return u},t)}}const Hwe=(e,t)=>{e.pending[t]!==void 0&&(delete e.pending[t],e.pendingCount--)};class $we extends Bp{constructor(t,n,r){var o,s;let a=arguments.length>3&&arguments[3]!==void 0?arguments[3]:{};super(),this.backend=t,this.store=n,this.services=r,this.languageUtils=r.languageUtils,this.options=a,this.logger=ra.create("backendConnector"),this.waitingReads=[],this.maxParallelReads=a.maxParallelReads||10,this.readingCalls=0,this.maxRetries=a.maxRetries>=0?a.maxRetries:5,this.retryTimeout=a.retryTimeout>=1?a.retryTimeout:350,this.state={},this.queue=[],(s=(o=this.backend)==null?void 0:o.init)==null||s.call(o,r,a.backend,a)}queueLoad(t,n,r,a){const o={},s={},u={},c={};return t.forEach(d=>{let p=!0;n.forEach(g=>{const m=`${d}|${g}`;!r.reload&&this.store.hasResourceBundle(d,g)?this.state[m]=2:this.state[m]<0||(this.state[m]===1?s[m]===void 0&&(s[m]=!0):(this.state[m]=1,p=!1,s[m]===void 0&&(s[m]=!0),o[m]===void 0&&(o[m]=!0),c[g]===void 0&&(c[g]=!0)))}),p||(u[d]=!0)}),(Object.keys(o).length||Object.keys(s).length)&&this.queue.push({pending:s,pendingCount:Object.keys(s).length,loaded:{},errors:[],callback:a}),{toLoad:Object.keys(o),pending:Object.keys(s),toLoadLanguages:Object.keys(u),toLoadNamespaces:Object.keys(c)}}loaded(t,n,r){const a=t.split("|"),o=a[0],s=a[1];n&&this.emit("failedLoading",o,s,n),!n&&r&&this.store.addResourceBundle(o,s,r,void 0,void 0,{skipCopy:!0}),this.state[t]=n?-1:2,n&&r&&(this.state[t]=0);const u={};this.queue.forEach(c=>{Nwe(c.loaded,[o],s),Hwe(c,t),n&&c.errors.push(n),c.pendingCount===0&&!c.done&&(Object.keys(c.loaded).forEach(d=>{u[d]||(u[d]={});const p=c.loaded[d];p.length&&p.forEach(g=>{u[d][g]===void 0&&(u[d][g]=!0)})}),c.done=!0,c.errors.length?c.callback(c.errors):c.callback())}),this.emit("loaded",u),this.queue=this.queue.filter(c=>!c.done)}read(t,n,r){let a=arguments.length>3&&arguments[3]!==void 0?arguments[3]:0,o=arguments.length>4&&arguments[4]!==void 0?arguments[4]:this.retryTimeout,s=arguments.length>5?arguments[5]:void 0;if(!t.length)return s(null,{});if(this.readingCalls>=this.maxParallelReads){this.waitingReads.push({lng:t,ns:n,fcName:r,tried:a,wait:o,callback:s});return}this.readingCalls++;const u=(d,p)=>{if(this.readingCalls--,this.waitingReads.length>0){const g=this.waitingReads.shift();this.read(g.lng,g.ns,g.fcName,g.tried,g.wait,g.callback)}if(d&&p&&a<this.maxRetries){setTimeout(()=>{this.read.call(this,t,n,r,a+1,o*2,s)},o);return}s(d,p)},c=this.backend[r].bind(this.backend);if(c.length===2){try{const d=c(t,n);d&&typeof d.then=="function"?d.then(p=>u(null,p)).catch(u):u(null,d)}catch(d){u(d)}return}return c(t,n,u)}prepareLoading(t,n){let r=arguments.length>2&&arguments[2]!==void 0?arguments[2]:{},a=arguments.length>3?arguments[3]:void 0;if(!this.backend)return this.logger.warn("No backend was added via i18next.use. Will not load resources."),a&&a();tt(t)&&(t=this.languageUtils.toResolveHierarchy(t)),tt(n)&&(n=[n]);const o=this.queueLoad(t,n,r,a);if(!o.toLoad.length)return o.pending.length||a(),null;o.toLoad.forEach(s=>{this.loadOne(s)})}load(t,n,r){this.prepareLoading(t,n,{},r)}reload(t,n,r){this.prepareLoading(t,n,{reload:!0},r)}loadOne(t){let n=arguments.length>1&&arguments[1]!==void 0?arguments[1]:"";const r=t.split("|"),a=r[0],o=r[1];this.read(a,o,"read",void 0,void 0,(s,u)=>{s&&this.logger.warn(`${n}loading namespace ${o} for language ${a} failed`,s),!s&&u&&this.logger.log(`${n}loaded namespace ${o} for language ${a}`,u),this.loaded(t,s,u)})}saveMissing(t,n,r,a,o){var c,d,p,g,m;let s=arguments.length>5&&arguments[5]!==void 0?arguments[5]:{},u=arguments.length>6&&arguments[6]!==void 0?arguments[6]:()=>{};if((d=(c=this.services)==null?void 0:c.utils)!=null&&d.hasLoadedNamespace&&!((g=(p=this.services)==null?void 0:p.utils)!=null&&g.hasLoadedNamespace(n))){this.logger.warn(`did not save key "${r}" as the namespace "${n}" was not yet loaded`,"This means something IS WRONG in your setup. You access the t function before i18next.init / i18next.loadNamespace / i18next.changeLanguage was done. Wait for the callback or Promise to resolve before accessing it!!!");return}if(!(r==null||r==="")){if((m=this.backend)!=null&&m.create){const b={...s,isUpdate:o},y=this.backend.create.bind(this.backend);if(y.length<6)try{let S;y.length===5?S=y(t,n,r,a,b):S=y(t,n,r,a),S&&typeof S.then=="function"?S.then(x=>u(null,x)).catch(u):u(null,S)}catch(S){u(S)}else y(t,n,r,a,u,b)}!t||!t[0]||this.store.addResource(t[0],n,r,a)}}}const Mz=()=>({debug:!1,initAsync:!0,ns:["translation"],defaultNS:["translation"],fallbackLng:["dev"],fallbackNS:!1,supportedLngs:!1,nonExplicitSupportedLngs:!1,load:"all",preload:!1,simplifyPluralSuffix:!0,keySeparator:".",nsSeparator:":",pluralSeparator:"_",contextSeparator:"_",partialBundledLanguages:!1,saveMissing:!1,updateMissing:!1,saveMissingTo:"fallback",saveMissingPlurals:!0,missingKeyHandler:!1,missingInterpolationHandler:!1,postProcess:!1,postProcessPassResolved:!1,returnNull:!1,returnEmptyString:!0,returnObjects:!1,joinArrays:!1,returnedObjectHandler:!1,parseMissingKeyHandler:!1,appendNamespaceToMissingKey:!1,appendNamespaceToCIMode:!1,overloadTranslationOptionHandler:e=>{let t={};if(typeof e[1]=="object"&&(t=e[1]),tt(e[1])&&(t.defaultValue=e[1]),tt(e[2])&&(t.tDescription=e[2]),typeof e[2]=="object"||typeof e[3]=="object"){const n=e[3]||e[2];Object.keys(n).forEach(r=>{t[r]=n[r]})}return t},interpolation:{escapeValue:!0,format:e=>e,prefix:"{{",suffix:"}}",formatSeparator:",",unescapePrefix:"-",nestingPrefix:"$t(",nestingSuffix:")",nestingOptionsSeparator:",",maxReplaces:1e3,skipOnVariables:!0}}),Fz=e=>{var t,n;return tt(e.ns)&&(e.ns=[e.ns]),tt(e.fallbackLng)&&(e.fallbackLng=[e.fallbackLng]),tt(e.fallbackNS)&&(e.fallbackNS=[e.fallbackNS]),((n=(t=e.supportedLngs)==null?void 0:t.indexOf)==null?void 0:n.call(t,"cimode"))<0&&(e.supportedLngs=e.supportedLngs.concat(["cimode"])),typeof e.initImmediate=="boolean"&&(e.initAsync=e.initImmediate),e},$d=()=>{},qwe=e=>{Object.getOwnPropertyNames(Object.getPrototypeOf(e)).forEach(n=>{typeof e[n]=="function"&&(e[n]=e[n].bind(e))})};class zu extends Bp{constructor(){let t=arguments.length>0&&arguments[0]!==void 0?arguments[0]:{},n=arguments.length>1?arguments[1]:void 0;if(super(),this.options=Fz(t),this.services={},this.logger=ra,this.modules={external:[]},qwe(this),n&&!this.isInitialized&&!t.isClone){if(!this.options.initAsync)return this.init(t,n),this;setTimeout(()=>{this.init(t,n)},0)}}init(){var t=this;let n=arguments.length>0&&arguments[0]!==void 0?arguments[0]:{},r=arguments.length>1?arguments[1]:void 0;this.isInitializing=!0,typeof n=="function"&&(r=n,n={}),n.defaultNS==null&&n.ns&&(tt(n.ns)?n.defaultNS=n.ns:n.ns.indexOf("translation")<0&&(n.defaultNS=n.ns[0]));const a=Mz();this.options={...a,...this.options,...Fz(n)},this.options.interpolation={...a.interpolation,...this.options.interpolation},n.keySeparator!==void 0&&(this.options.userDefinedKeySeparator=n.keySeparator),n.nsSeparator!==void 0&&(this.options.userDefinedNsSeparator=n.nsSeparator);const o=p=>p?typeof p=="function"?new p:p:null;if(!this.options.isClone){this.modules.logger?ra.init(o(this.modules.logger),this.options):ra.init(null,this.options);let p;this.modules.formatter?p=this.modules.formatter:p=Gwe;const g=new Oz(this.options);this.store=new _z(this.options.resources,this.options);const m=this.services;m.logger=ra,m.resourceStore=this.store,m.languageUtils=g,m.pluralResolver=new Bwe(g,{prepend:this.options.pluralSeparator,simplifyPluralSuffix:this.options.simplifyPluralSuffix}),p&&(!this.options.interpolation.format||this.options.interpolation.format===a.interpolation.format)&&(m.formatter=o(p),m.formatter.init(m,this.options),this.options.interpolation.format=m.formatter.format.bind(m.formatter)),m.interpolator=new Uwe(this.options),m.utils={hasLoadedNamespace:this.hasLoadedNamespace.bind(this)},m.backendConnector=new $we(o(this.modules.backend),m.resourceStore,m,this.options),m.backendConnector.on("*",function(b){for(var y=arguments.length,S=new Array(y>1?y-1:0),x=1;x<y;x++)S[x-1]=arguments[x];t.emit(b,...S)}),this.modules.languageDetector&&(m.languageDetector=o(this.modules.languageDetector),m.languageDetector.init&&m.languageDetector.init(m,this.options.detection,this.options)),this.modules.i18nFormat&&(m.i18nFormat=o(this.modules.i18nFormat),m.i18nFormat.init&&m.i18nFormat.init(this)),this.translator=new Hf(this.services,this.options),this.translator.on("*",function(b){for(var y=arguments.length,S=new Array(y>1?y-1:0),x=1;x<y;x++)S[x-1]=arguments[x];t.emit(b,...S)}),this.modules.external.forEach(b=>{b.init&&b.init(this)})}if(this.format=this.options.interpolation.format,r||(r=$d),this.options.fallbackLng&&!this.services.languageDetector&&!this.options.lng){const p=this.services.languageUtils.getFallbackCodes(this.options.fallbackLng);p.length>0&&p[0]!=="dev"&&(this.options.lng=p[0])}!this.services.languageDetector&&!this.options.lng&&this.logger.warn("init: no languageDetector is used and no lng is defined"),["getResource","hasResourceBundle","getResourceBundle","getDataByLanguage"].forEach(p=>{this[p]=function(){return t.store[p](...arguments)}}),["addResource","addResources","addResourceBundle","removeResourceBundle"].forEach(p=>{this[p]=function(){return t.store[p](...arguments),t}});const c=du(),d=()=>{const p=(g,m)=>{this.isInitializing=!1,this.isInitialized&&!this.initializedStoreOnce&&this.logger.warn("init: i18next is already initialized. You should call init just once!"),this.isInitialized=!0,this.options.isClone||this.logger.log("initialized",this.options),this.emit("initialized",this.options),c.resolve(m),r(g,m)};if(this.languages&&!this.isInitialized)return p(null,this.t.bind(this));this.changeLanguage(this.options.lng,p)};return this.options.resources||!this.options.initAsync?d():setTimeout(d,0),c}loadResources(t){var o,s;let r=arguments.length>1&&arguments[1]!==void 0?arguments[1]:$d;const a=tt(t)?t:this.language;if(typeof t=="function"&&(r=t),!this.options.resources||this.options.partialBundledLanguages){if((a==null?void 0:a.toLowerCase())==="cimode"&&(!this.options.preload||this.options.preload.length===0))return r();const u=[],c=d=>{if(!d||d==="cimode")return;this.services.languageUtils.toResolveHierarchy(d).forEach(g=>{g!=="cimode"&&u.indexOf(g)<0&&u.push(g)})};a?c(a):this.services.languageUtils.getFallbackCodes(this.options.fallbackLng).forEach(p=>c(p)),(s=(o=this.options.preload)==null?void 0:o.forEach)==null||s.call(o,d=>c(d)),this.services.backendConnector.load(u,this.options.ns,d=>{!d&&!this.resolvedLanguage&&this.language&&this.setResolvedLanguage(this.language),r(d)})}else r(null)}reloadResources(t,n,r){const a=du();return typeof t=="function"&&(r=t,t=void 0),typeof n=="function"&&(r=n,n=void 0),t||(t=this.languages),n||(n=this.options.ns),r||(r=$d),this.services.backendConnector.reload(t,n,o=>{a.resolve(),r(o)}),a}use(t){if(!t)throw new Error("You are passing an undefined module! Please check the object you are passing to i18next.use()");if(!t.type)throw new Error("You are passing a wrong module! Please check the object you are passing to i18next.use()");return t.type==="backend"&&(this.modules.backend=t),(t.type==="logger"||t.log&&t.warn&&t.error)&&(this.modules.logger=t),t.type==="languageDetector"&&(this.modules.languageDetector=t),t.type==="i18nFormat"&&(this.modules.i18nFormat=t),t.type==="postProcessor"&&ZH.addPostProcessor(t),t.type==="formatter"&&(this.modules.formatter=t),t.type==="3rdParty"&&this.modules.external.push(t),this}setResolvedLanguage(t){if(!(!t||!this.languages)&&!(["cimode","dev"].indexOf(t)>-1))for(let n=0;n<this.languages.length;n++){const r=this.languages[n];if(!(["cimode","dev"].indexOf(r)>-1)&&this.store.hasLanguageSomeTranslations(r)){this.resolvedLanguage=r;break}}}changeLanguage(t,n){var r=this;this.isLanguageChangingTo=t;const a=du();this.emit("languageChanging",t);const o=c=>{this.language=c,this.languages=this.services.languageUtils.toResolveHierarchy(c),this.resolvedLanguage=void 0,this.setResolvedLanguage(c)},s=(c,d)=>{d?(o(d),this.translator.changeLanguage(d),this.isLanguageChangingTo=void 0,this.emit("languageChanged",d),this.logger.log("languageChanged",d)):this.isLanguageChangingTo=void 0,a.resolve(function(){return r.t(...arguments)}),n&&n(c,function(){return r.t(...arguments)})},u=c=>{var p,g;!t&&!c&&this.services.languageDetector&&(c=[]);const d=tt(c)?c:this.services.languageUtils.getBestMatchFromCodes(c);d&&(this.language||o(d),this.translator.language||this.translator.changeLanguage(d),(g=(p=this.services.languageDetector)==null?void 0:p.cacheUserLanguage)==null||g.call(p,d)),this.loadResources(d,m=>{s(m,d)})};return!t&&this.services.languageDetector&&!this.services.languageDetector.async?u(this.services.languageDetector.detect()):!t&&this.services.languageDetector&&this.services.languageDetector.async?this.services.languageDetector.detect.length===0?this.services.languageDetector.detect().then(u):this.services.languageDetector.detect(u):u(t),a}getFixedT(t,n,r){var a=this;const o=function(s,u){let c;if(typeof u!="object"){for(var d=arguments.length,p=new Array(d>2?d-2:0),g=2;g<d;g++)p[g-2]=arguments[g];c=a.options.overloadTranslationOptionHandler([s,u].concat(p))}else c={...u};c.lng=c.lng||o.lng,c.lngs=c.lngs||o.lngs,c.ns=c.ns||o.ns,c.keyPrefix!==""&&(c.keyPrefix=c.keyPrefix||r||o.keyPrefix);const m=a.options.keySeparator||".";let b;return c.keyPrefix&&Array.isArray(s)?b=s.map(y=>`${c.keyPrefix}${m}${y}`):b=c.keyPrefix?`${c.keyPrefix}${m}${s}`:s,a.t(b,c)};return tt(t)?o.lng=t:o.lngs=t,o.ns=n,o.keyPrefix=r,o}t(){var a;for(var t=arguments.length,n=new Array(t),r=0;r<t;r++)n[r]=arguments[r];return(a=this.translator)==null?void 0:a.translate(...n)}exists(){var a;for(var t=arguments.length,n=new Array(t),r=0;r<t;r++)n[r]=arguments[r];return(a=this.translator)==null?void 0:a.exists(...n)}setDefaultNamespace(t){this.options.defaultNS=t}hasLoadedNamespace(t){let n=arguments.length>1&&arguments[1]!==void 0?arguments[1]:{};if(!this.isInitialized)return this.logger.warn("hasLoadedNamespace: i18next was not initialized",this.languages),!1;if(!this.languages||!this.languages.length)return this.logger.warn("hasLoadedNamespace: i18n.languages were undefined or empty",this.languages),!1;const r=n.lng||this.resolvedLanguage||this.languages[0],a=this.options?this.options.fallbackLng:!1,o=this.languages[this.languages.length-1];if(r.toLowerCase()==="cimode")return!0;const s=(u,c)=>{const d=this.services.backendConnector.state[`${u}|${c}`];return d===-1||d===0||d===2};if(n.precheck){const u=n.precheck(this,s);if(u!==void 0)return u}return!!(this.hasResourceBundle(r,t)||!this.services.backendConnector.backend||this.options.resources&&!this.options.partialBundledLanguages||s(r,t)&&(!a||s(o,t)))}loadNamespaces(t,n){const r=du();return this.options.ns?(tt(t)&&(t=[t]),t.forEach(a=>{this.options.ns.indexOf(a)<0&&this.options.ns.push(a)}),this.loadResources(a=>{r.resolve(),n&&n(a)}),r):(n&&n(),Promise.resolve())}loadLanguages(t,n){const r=du();tt(t)&&(t=[t]);const a=this.options.preload||[],o=t.filter(s=>a.indexOf(s)<0&&this.services.languageUtils.isSupportedCode(s));return o.length?(this.options.preload=a.concat(o),this.loadResources(s=>{r.resolve(),n&&n(s)}),r):(n&&n(),Promise.resolve())}dir(t){var a,o;if(t||(t=this.resolvedLanguage||(((a=this.languages)==null?void 0:a.length)>0?this.languages[0]:this.language)),!t)return"rtl";const n=["ar","shu","sqr","ssh","xaa","yhd","yud","aao","abh","abv","acm","acq","acw","acx","acy","adf","ads","aeb","aec","afb","ajp","apc","apd","arb","arq","ars","ary","arz","auz","avl","ayh","ayl","ayn","ayp","bbz","pga","he","iw","ps","pbt","pbu","pst","prp","prd","ug","ur","ydd","yds","yih","ji","yi","hbo","men","xmn","fa","jpr","peo","pes","prs","dv","sam","ckb"],r=((o=this.services)==null?void 0:o.languageUtils)||new Oz(Mz());return n.indexOf(r.getLanguagePartFromCode(t))>-1||t.toLowerCase().indexOf("-arab")>1?"rtl":"ltr"}static createInstance(){let t=arguments.length>0&&arguments[0]!==void 0?arguments[0]:{},n=arguments.length>1?arguments[1]:void 0;return new zu(t,n)}cloneInstance(){let t=arguments.length>0&&arguments[0]!==void 0?arguments[0]:{},n=arguments.length>1&&arguments[1]!==void 0?arguments[1]:$d;const r=t.forkResourceStore;r&&delete t.forkResourceStore;const a={...this.options,...t,isClone:!0},o=new zu(a);if((t.debug!==void 0||t.prefix!==void 0)&&(o.logger=o.logger.clone(t)),["store","services","language"].forEach(u=>{o[u]=this[u]}),o.services={...this.services},o.services.utils={hasLoadedNamespace:o.hasLoadedNamespace.bind(o)},r){const u=Object.keys(this.store.data).reduce((c,d)=>(c[d]={...this.store.data[d]},Object.keys(c[d]).reduce((p,g)=>(p[g]={...c[d][g]},p),{})),{});o.store=new _z(u,a),o.services.resourceStore=o.store}return o.translator=new Hf(o.services,a),o.translator.on("*",function(u){for(var c=arguments.length,d=new Array(c>1?c-1:0),p=1;p<c;p++)d[p-1]=arguments[p];o.emit(u,...d)}),o.init(a,n),o.translator.options=a,o.translator.backendConnector.services.utils={hasLoadedNamespace:o.hasLoadedNamespace.bind(o)},o}toJSON(){return{options:this.options,store:this.store,language:this.language,languages:this.languages,resolvedLanguage:this.resolvedLanguage}}}const vn=zu.createInstance();vn.createInstance=zu.createInstance;vn.createInstance;vn.dir;vn.init;vn.loadResources;vn.reloadResources;vn.use;vn.changeLanguage;vn.getFixedT;vn.t;vn.exists;vn.setDefaultNamespace;vn.hasLoadedNamespace;vn.loadNamespaces;vn.loadLanguages;const Vwe=Object.freeze(Object.defineProperty({__proto__:null,default:vn},Symbol.toStringTag,{value:"Module"})),Wwe={language:"Language",theme:"Theme",light:"Light",dark:"Dark",system:"System"},Ywe={documents:"Documents",knowledgeGraph:"Knowledge Graph",retrieval:"Retrieval",api:"API",projectRepository:"Project Repository",logout:"Logout",themeToggle:{switchToLight:"Switch to light theme",switchToDark:"Switch to dark theme"}},Kwe={description:"Please enter your account and password to log in to the system",username:"Username",usernamePlaceholder:"Please input a username",password:"Password",passwordPlaceholder:"Please input a password",loginButton:"Login",loggingIn:"Logging in...",successMessage:"Login succeeded",errorEmptyFields:"Please enter your username and password",errorInvalidCredentials:"Login failed, please check username and password",authDisabled:"Authentication is disabled. Using login free mode.",guestMode:"Login Free"},Xwe={clearDocuments:{button:"Clear",tooltip:"Clear documents",title:"Clear Documents",confirm:"Do you really want to clear all documents?",confirmButton:"YES",success:"Documents cleared successfully",failed:`Clear Documents Failed:
{{message}}`,error:`Clear Documents Failed:
{{error}}`},uploadDocuments:{button:"Upload",tooltip:"Upload documents",title:"Upload Documents",description:"Drag and drop your documents here or click to browse.",uploading:"Uploading {{name}}: {{percent}}%",success:`Upload Success:
{{name}} uploaded successfully`,failed:`Upload Failed:
{{name}}
{{message}}`,error:`Upload Failed:
{{name}}
{{error}}`,generalError:`Upload Failed
{{error}}`,fileTypes:"Supported types: TXT, MD, DOCX, PDF, PPTX, RTF, ODT, EPUB, HTML, HTM, TEX, JSON, XML, YAML, YML, CSV, LOG, CONF, INI, PROPERTIES, SQL, BAT, SH, C, CPP, PY, JAVA, JS, TS, SWIFT, GO, RB, PHP, CSS, SCSS, LESS"},documentManager:{title:"Document Management",scanButton:"Scan",scanTooltip:"Scan documents",uploadedTitle:"Uploaded Documents",uploadedDescription:"List of uploaded documents and their statuses.",emptyTitle:"No Documents",emptyDescription:"There are no uploaded documents yet.",columns:{id:"ID",summary:"Summary",status:"Status",length:"Length",chunks:"Chunks",created:"Created",updated:"Updated",metadata:"Metadata"},status:{completed:"Completed",processing:"Processing",pending:"Pending",failed:"Failed"},errors:{loadFailed:`Failed to load documents
{{error}}`,scanFailed:`Failed to scan documents
{{error}}`,scanProgressFailed:`Failed to get scan progress
{{error}}`}}},Zwe={sideBar:{settings:{settings:"Settings",healthCheck:"Health Check",showPropertyPanel:"Show Property Panel",showSearchBar:"Show Search Bar",showNodeLabel:"Show Node Label",nodeDraggable:"Node Draggable",showEdgeLabel:"Show Edge Label",hideUnselectedEdges:"Hide Unselected Edges",edgeEvents:"Edge Events",maxQueryDepth:"Max Query Depth",minDegree:"Minimum Degree",maxLayoutIterations:"Max Layout Iterations",depth:"Depth",degree:"Degree",apiKey:"API Key",enterYourAPIkey:"Enter your API key",save:"Save",refreshLayout:"Refresh Layout"},zoomControl:{zoomIn:"Zoom In",zoomOut:"Zoom Out",resetZoom:"Reset Zoom",rotateCamera:"Clockwise Rotate",rotateCameraCounterClockwise:"Counter-Clockwise Rotate"},layoutsControl:{startAnimation:"Continue layout animation",stopAnimation:"Stop layout animation",layoutGraph:"Layout Graph",layouts:{Circular:"Circular",Circlepack:"Circlepack",Random:"Random",Noverlaps:"Noverlaps","Force Directed":"Force Directed","Force Atlas":"Force Atlas"}},fullScreenControl:{fullScreen:"Full Screen",windowed:"Windowed"}},statusIndicator:{connected:"Connected",disconnected:"Disconnected"},statusCard:{unavailable:"Status information unavailable",storageInfo:"Storage Info",workingDirectory:"Working Directory",inputDirectory:"Input Directory",llmConfig:"LLM Configuration",llmBinding:"LLM Binding",llmBindingHost:"LLM Binding Host",llmModel:"LLM Model",maxTokens:"Max Tokens",embeddingConfig:"Embedding Configuration",embeddingBinding:"Embedding Binding",embeddingBindingHost:"Embedding Binding Host",embeddingModel:"Embedding Model",storageConfig:"Storage Configuration",kvStorage:"KV Storage",docStatusStorage:"Doc Status Storage",graphStorage:"Graph Storage",vectorStorage:"Vector Storage"},propertiesView:{node:{title:"Node",id:"ID",labels:"Labels",degree:"Degree",properties:"Properties",relationships:"Relationships",expandNode:"Expand Node",pruneNode:"Prune Node",deleteAllNodesError:"Refuse to delete all nodes in the graph",nodesRemoved:"{{count}} nodes removed, including orphan nodes",noNewNodes:"No expandable nodes found",propertyNames:{description:"Description",entity_id:"Name",entity_type:"Type",source_id:"SrcID",Neighbour:"Neigh"}},edge:{title:"Relationship",id:"ID",type:"Type",source:"Source",target:"Target",properties:"Properties"}},search:{placeholder:"Search nodes...",message:"And {count} others"},graphLabels:{selectTooltip:"Select query label",noLabels:"No labels found",label:"Label",placeholder:"Search labels...",andOthers:"And {count} others",refreshTooltip:"Reload graph data"}},Qwe={chatMessage:{copyTooltip:"Copy to clipboard",copyError:"Failed to copy text to clipboard"},retrieval:{startPrompt:"Start a retrieval by typing your query below",clear:"Clear",send:"Send",placeholder:"Type your query...",error:"Error: Failed to get response"},querySettings:{parametersTitle:"Parameters",parametersDescription:"Configure your query parameters",queryMode:"Query Mode",queryModeTooltip:`Select the retrieval strategy:
• Naive: Basic search without advanced techniques
• Local: Context-dependent information retrieval
• Global: Utilizes global knowledge base
• Hybrid: Combines local and global retrieval
• Mix: Integrates knowledge graph with vector retrieval`,queryModeOptions:{naive:"Naive",local:"Local",global:"Global",hybrid:"Hybrid",mix:"Mix"},responseFormat:"Response Format",responseFormatTooltip:`Defines the response format. Examples:
• Multiple Paragraphs
• Single Paragraph
• Bullet Points`,responseFormatOptions:{multipleParagraphs:"Multiple Paragraphs",singleParagraph:"Single Paragraph",bulletPoints:"Bullet Points"},topK:"Top K Results",topKTooltip:"Number of top items to retrieve. Represents entities in 'local' mode and relationships in 'global' mode",topKPlaceholder:"Number of results",maxTokensTextUnit:"Max Tokens for Text Unit",maxTokensTextUnitTooltip:"Maximum number of tokens allowed for each retrieved text chunk",maxTokensGlobalContext:"Max Tokens for Global Context",maxTokensGlobalContextTooltip:"Maximum number of tokens allocated for relationship descriptions in global retrieval",maxTokensLocalContext:"Max Tokens for Local Context",maxTokensLocalContextTooltip:"Maximum number of tokens allocated for entity descriptions in local retrieval",historyTurns:"History Turns",historyTurnsTooltip:"Number of complete conversation turns (user-assistant pairs) to consider in the response context",historyTurnsPlaceholder:"Number of history turns",hlKeywords:"High-Level Keywords",hlKeywordsTooltip:"List of high-level keywords to prioritize in retrieval. Separate with commas",hlkeywordsPlaceHolder:"Enter keywords",llKeywords:"Low-Level Keywords",llKeywordsTooltip:"List of low-level keywords to refine retrieval focus. Separate with commas",onlyNeedContext:"Only Need Context",onlyNeedContextTooltip:"If True, only returns the retrieved context without generating a response",onlyNeedPrompt:"Only Need Prompt",onlyNeedPromptTooltip:"If True, only returns the generated prompt without producing a response",streamResponse:"Stream Response",streamResponseTooltip:"If True, enables streaming output for real-time responses"}},Jwe={loading:"Loading API Documentation..."},exe={settings:Wwe,header:Ywe,login:Kwe,documentPanel:Xwe,graphPanel:Zwe,retrievePanel:Qwe,apiSite:Jwe},txe={language:"语言",theme:"主题",light:"浅色",dark:"深色",system:"系统"},nxe={documents:"文档",knowledgeGraph:"知识图谱",retrieval:"检索",api:"API",projectRepository:"项目仓库",logout:"退出登录",themeToggle:{switchToLight:"切换到浅色主题",switchToDark:"切换到深色主题"}},rxe={description:"请输入您的账号和密码登录系统",username:"用户名",usernamePlaceholder:"请输入用户名",password:"密码",passwordPlaceholder:"请输入密码",loginButton:"登录",loggingIn:"登录中...",successMessage:"登录成功",errorEmptyFields:"请输入您的用户名和密码",errorInvalidCredentials:"登录失败，请检查用户名和密码",authDisabled:"认证已禁用，使用无需登陆模式。",guestMode:"无需登陆"},axe={clearDocuments:{button:"清空",tooltip:"清空文档",title:"清空文档",confirm:"确定要清空所有文档吗？",confirmButton:"确定",success:"文档清空成功",failed:`清空文档失败：
{{message}}`,error:`清空文档失败：
{{error}}`},uploadDocuments:{button:"上传",tooltip:"上传文档",title:"上传文档",description:"拖拽文件到此处或点击浏览",uploading:"正在上传 {{name}}：{{percent}}%",success:`上传成功：
{{name}} 上传完成`,failed:`上传失败：
{{name}}
{{message}}`,error:`上传失败：
{{name}}
{{error}}`,generalError:`上传失败
{{error}}`,fileTypes:"支持的文件类型：TXT, MD, DOCX, PDF, PPTX, RTF, ODT, EPUB, HTML, HTM, TEX, JSON, XML, YAML, YML, CSV, LOG, CONF, INI, PROPERTIES, SQL, BAT, SH, C, CPP, PY, JAVA, JS, TS, SWIFT, GO, RB, PHP, CSS, SCSS, LESS"},documentManager:{title:"文档管理",scanButton:"扫描",scanTooltip:"扫描文档",uploadedTitle:"已上传文档",uploadedDescription:"已上传文档列表及其状态",emptyTitle:"无文档",emptyDescription:"还没有上传任何文档",columns:{id:"ID",summary:"摘要",status:"状态",length:"长度",chunks:"分块",created:"创建时间",updated:"更新时间",metadata:"元数据"},status:{completed:"已完成",processing:"处理中",pending:"等待中",failed:"失败"},errors:{loadFailed:`加载文档失败
{{error}}`,scanFailed:`扫描文档失败
{{error}}`,scanProgressFailed:`获取扫描进度失败
{{error}}`}}},oxe={sideBar:{settings:{settings:"设置",healthCheck:"健康检查",showPropertyPanel:"显示属性面板",showSearchBar:"显示搜索栏",showNodeLabel:"显示节点标签",nodeDraggable:"节点可拖动",showEdgeLabel:"显示边标签",hideUnselectedEdges:"隐藏未选中的边",edgeEvents:"边事件",maxQueryDepth:"最大查询深度",minDegree:"最小邻边数",maxLayoutIterations:"最大布局迭代次数",depth:"深度",degree:"邻边",apiKey:"API密钥",enterYourAPIkey:"输入您的API密钥",save:"保存",refreshLayout:"刷新布局"},zoomControl:{zoomIn:"放大",zoomOut:"缩小",resetZoom:"重置缩放",rotateCamera:"顺时针旋转图形",rotateCameraCounterClockwise:"逆时针旋转图形"},layoutsControl:{startAnimation:"继续布局动画",stopAnimation:"停止布局动画",layoutGraph:"图布局",layouts:{Circular:"环形",Circlepack:"圆形打包",Random:"随机",Noverlaps:"无重叠","Force Directed":"力导向","Force Atlas":"力地图"}},fullScreenControl:{fullScreen:"全屏",windowed:"窗口"}},statusIndicator:{connected:"已连接",disconnected:"未连接"},statusCard:{unavailable:"状态信息不可用",storageInfo:"存储信息",workingDirectory:"工作目录",inputDirectory:"输入目录",llmConfig:"LLM配置",llmBinding:"LLM绑定",llmBindingHost:"LLM绑定主机",llmModel:"LLM模型",maxTokens:"最大令牌数",embeddingConfig:"嵌入配置",embeddingBinding:"嵌入绑定",embeddingBindingHost:"嵌入绑定主机",embeddingModel:"嵌入模型",storageConfig:"存储配置",kvStorage:"KV存储",docStatusStorage:"文档状态存储",graphStorage:"图存储",vectorStorage:"向量存储"},propertiesView:{node:{title:"节点",id:"ID",labels:"标签",degree:"度数",properties:"属性",relationships:"关系",expandNode:"扩展节点",pruneNode:"修剪节点",deleteAllNodesError:"拒绝删除图中的所有节点",nodesRemoved:"已删除 {{count}} 个节点，包括孤立节点",noNewNodes:"没有发现可以扩展的节点",propertyNames:{description:"描述",entity_id:"名称",entity_type:"类型",source_id:"信源ID",Neighbour:"邻接"}},edge:{title:"关系",id:"ID",type:"类型",source:"源节点",target:"目标节点",properties:"属性"}},search:{placeholder:"搜索节点...",message:"还有 {count} 个"},graphLabels:{selectTooltip:"选择查询标签",noLabels:"未找到标签",label:"标签",placeholder:"搜索标签...",andOthers:"还有 {count} 个",refreshTooltip:"重新加载图形数据"}},ixe={chatMessage:{copyTooltip:"复制到剪贴板",copyError:"复制文本到剪贴板失败"},retrieval:{startPrompt:"输入查询开始检索",clear:"清空",send:"发送",placeholder:"输入查询...",error:"错误：获取响应失败"},querySettings:{parametersTitle:"参数",parametersDescription:"配置查询参数",queryMode:"查询模式",queryModeTooltip:`选择检索策略：
• Naive：基础搜索，无高级技术
• Local：上下文相关信息检索
• Global：利用全局知识库
• Hybrid：结合本地和全局检索
• Mix：整合知识图谱和向量检索`,queryModeOptions:{naive:"朴素",local:"本地",global:"全局",hybrid:"混合",mix:"混合"},responseFormat:"响应格式",responseFormatTooltip:`定义响应格式。例如：
• 多段落
• 单段落
• 要点`,responseFormatOptions:{multipleParagraphs:"多段落",singleParagraph:"单段落",bulletPoints:"要点"},topK:"Top K结果",topKTooltip:"检索的顶部项目数。在'local'模式下表示实体，在'global'模式下表示关系",topKPlaceholder:"结果数量",maxTokensTextUnit:"文本单元最大令牌数",maxTokensTextUnitTooltip:"每个检索文本块允许的最大令牌数",maxTokensGlobalContext:"全局上下文最大令牌数",maxTokensGlobalContextTooltip:"全局检索中关系描述的最大令牌数",maxTokensLocalContext:"本地上下文最大令牌数",maxTokensLocalContextTooltip:"本地检索中实体描述的最大令牌数",historyTurns:"历史轮次",historyTurnsTooltip:"响应上下文中考虑的完整对话轮次（用户-助手对）数量",historyTurnsPlaceholder:"历史轮次数",hlKeywords:"高级关键词",hlKeywordsTooltip:"检索中优先考虑的高级关键词列表。用逗号分隔",hlkeywordsPlaceHolder:"输入关键词",llKeywords:"低级关键词",llKeywordsTooltip:"用于细化检索重点的低级关键词列表。用逗号分隔",onlyNeedContext:"仅需上下文",onlyNeedContextTooltip:"如果为True，仅返回检索到的上下文而不生成响应",onlyNeedPrompt:"仅需提示",onlyNeedPromptTooltip:"如果为True，仅返回生成的提示而不产生响应",streamResponse:"流式响应",streamResponseTooltip:"如果为True，启用实时流式输出响应"}},sxe={loading:"正在加载 API 文档..."},lxe={settings:txe,header:nxe,login:rxe,documentPanel:axe,graphPanel:oxe,retrievePanel:ixe,apiSite:sxe},uxe=()=>{var e;try{const t=localStorage.getItem("settings-storage");if(t)return((e=JSON.parse(t).state)==null?void 0:e.language)||"en"}catch(t){console.error("Failed to get stored language:",t)}return"en"};vn.use(rQ).init({resources:{en:{translation:exe},zh:{translation:lxe}},lng:uxe(),fallbackLng:"en",interpolation:{escapeValue:!1}});E9.createRoot(document.getElementById("root")).render(w.jsx(E.StrictMode,{children:w.jsx(Rwe,{})}));
